#ifdef GRANDSTREAM_NETWORKS

#define AST_MODULE_LOG "Bridge"


#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 422177 $")

#include "asterisk/channel.h"
#include "asterisk/bridge.h"
#include "asterisk/stasis.h"
#include "asterisk/stasis_channels.h"
#include "asterisk/stasis_bridges.h"
#include "asterisk/manager.h"
#include "asterisk/stasis_message_router.h"
#include "asterisk/app.h"
#include "asterisk/astdb.h"
#include "asterisk/pjsip_cei.h"


#include "include/confbridge.h"

/* STRUCTS */
struct cei_user {
	enum cei_user_type user_type;
	int user_num;                    //Meetme user number
	char *entity;                    //Entity string
	char cid_name[33];
	char cid_number[33];
	char channel_name[101];
	char conference_id[41];          // conference joined
	char self_conf_id[41];           // seld conference id, if it's a remote user
	char user_hash[41];
	int admin;
	int user_role;
	int muted;
	int user_status;
	int can_invite;
	AST_LIST_ENTRY(cei_user) list;
};

struct cei_conference {
	char conference_num[100];         //Name of this conference, as configured in Meetme
	char conference_id[41];           //Conference hash
	char internal_publisher[41];      //Internal publisher's conference hash
	int locked;
	int local_size;
	int remote_size;                  //If remote, parsed local user-count
	int local_user_size;
	int remote_user_size;
	char *entity;
	int mcb_count;
	AST_LIST_ENTRY(cei_conference) list;
	int refcount;
	struct ao2_container *remote_container;
	struct ao2_container *user_container;
	int stop_propagating;
	int version;                      //If local, indicates the document version since a MCB entered the room. If remote, indicates last version received
	int version_reset;
	struct cei_conference* host_conf;
};
/* END STRUCTS */

/* VARIABLES */
static AST_LIST_HEAD_STATIC(confs, cei_conference);
#define TRUE 1
#define FALSE 0
#define CEI_EVENT_USER_LEN 250
#define CEI_EVENT_DATA_LEN 9500
#define MAX_CONFNUM 80
//static int room_count = 0;
static int conf_max_capability = 0;
static struct stasis_message_router *bridge_state_router;
static struct stasis_message_router *channel_state_router;
static struct stasis_message_router *cei_mcb_process_router;
/* END VARIABLES */

/* FUNCTION DECLARATION */
static struct cei_conference* build_conf_by_num(const char *conf_no);
static int user_hash_cmp(void *obj, void *arg, int flags);
static int conference_id_cmp(void *obj, void *arg, int flags);
static void hash_user(struct cei_user *user);
static void poll_cei_conference(struct cei_conference *conf,struct ast_str **extra_text);
static void poll_cei_partial_notify(struct cei_conference *conf, struct cei_conference *remote_conf, struct cei_user *user,
																		struct ast_str **data_string, enum cei_event_type event_type );
static void conf_free_all(void);
static int conf_free(struct cei_conference *conf);
static void clear_conf_users(struct cei_conference *conf);
static void process_remote_xml(struct cei_conference *conf, const char *publisher_private_num, const char *xml,
	                 struct ast_str **data_string);
static int check_xml_text_node(struct ast_xml_node *node);
static void process_remote_xml_users(struct cei_conference *conf,
									 struct cei_conference *remote_conference,
									 struct ast_xml_node *node_users);
static struct cei_user *process_remote_xml_user(struct cei_conference *host_conf,
	                     struct ast_xml_node *node_user);
static struct cei_conference* get_conf_by_id(const char *conf_id);
static void delete_remote_confs_by_publisher(struct cei_conference *host_conf,
											 const char *publisher);
static void delete_remote_conf_by_id(struct cei_conference *host_conf,
										 	 const char *remote_conf_id);
static struct cei_conference* get_conf_by_num(const char *conf_num);
static struct cei_user* get_user_by_cid_num(struct cei_conference *conf, const char *cid_num);
static void process_remote_string(struct cei_conference *conf,
												const char *publisher_num, const char *publisher_id,
												char *content, struct ast_str **data_string);
static int get_conf_max_capability(void);
static int check_role_as_conference(struct cei_user *user);
static void reset_conference(struct cei_conference *conf);

/* END FUNCTION DECLARATION */

static void cei_publish_manager_event(struct ast_json * json_object)
{
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

	// Sanity check
	if (json_object == NULL) {
		ast_debug(LOG_ERROR, "json_object IS NULL");
		return;
	}

	const char *conference_name = ast_json_string_get(ast_json_object_get(json_object, "conference"));

	if (conference_name == NULL) {
		ast_debug(LOG_ERROR, "conference_name IS NULL");
		return;
	}

	payload = ast_json_payload_create(json_object);
    if(!payload)
        return;
	msg = stasis_message_create(ast_cei_notify_type(), payload);
	if (msg == NULL) {
		return;
	}

	stasis_publish(ast_cei_topic(conference_name), msg);
}

static void cei_join_cb(void *data, struct stasis_subscription *sub,
	struct stasis_message *message)
{
	struct cei_conference *conf;
	struct cei_user *user = NULL;
	struct ast_str *full_text = ast_str_create(CEI_EVENT_DATA_LEN);
	struct ast_str *partial_text = ast_str_create(CEI_EVENT_DATA_LEN);

	const struct ast_bridge_blob *blob = stasis_message_data(message);
	unsigned int user_count= ast_json_integer_get(ast_json_object_get(blob->blob, "user_count"));
	unsigned int user_number = ast_json_integer_get(ast_json_object_get(blob->blob, "user_number"));
	const char *conference_name = ast_json_string_get(ast_json_object_get(blob->blob, "conference"));
	const char *cid_number = ast_json_string_get(ast_json_object_get(blob->blob, "caller_id_number"));
	const char *cid_name = ast_json_string_get(ast_json_object_get(blob->blob, "caller_id_name"));
	const char *channel_name = ast_json_string_get(ast_json_object_get(blob->blob, "channel_name"));
	const char *mcb_type = ast_json_string_get(ast_json_object_get(blob->blob, "mcb_type"));
	const char *mcb_remote_conf_id = ast_json_string_get(ast_json_object_get(blob->blob, "mcb_remote_conf_id"));

	unsigned int admin_status = 0;
	if (ast_json_is_true(ast_json_object_get(blob->blob, "admin"))) {
			admin_status = 1;
	}
	unsigned int mute_status = ast_json_integer_get(ast_json_object_get(blob->blob, "muted"));
	unsigned int can_invite = ast_json_integer_get(ast_json_object_get(blob->blob, "can_invite"));
	unsigned int invited_status = ast_json_integer_get(ast_json_object_get(blob->blob, "invited"));

	if (conference_name == NULL || cid_number == NULL || cid_name == NULL ||
		channel_name == NULL || mcb_type == NULL) {
		ast_debug(LOG_DEBUG, "[WL] EVENT INCOMPLETE");
		ast_free(full_text);
		ast_free(partial_text);
		return;
	}

	// get cei_conference
	conf = build_conf_by_num(conference_name);
	ast_debug(LOG_DEBUG, "[WL] Create new user struct with user_count: %d, user_num:%d, cid_name:%s, cid_num:%s, chan_name:%s, conf_name:%s, admin_status:%d, muted_status:%d, can_invite: %d",
												user_count,
												user_number,
												cid_name,
												cid_number,
												channel_name,
												conf->conference_id,
												admin_status,
												mute_status,
												can_invite
	);

	if (!conf) {
		ast_free(full_text);
		ast_free(partial_text);
		return;
	}

	if (!(user = ao2_alloc(sizeof(struct cei_user), NULL))) {
		ast_free(full_text);
		ast_free(partial_text);
		ast_debug(LOG_DEBUG, "[WL] CANNOT CREATE NEW USER");
		return;
	}

	if (user_count != (conf->local_size + 1)) {
		ast_debug(LOG_DEBUG, "[WL] CONF SIZE MISMATCH %d vs %d",
							 conf->local_size, user_count);
	}

	//add info to cei_user
	user->user_num = (int) user_number;
	ast_copy_string(user->cid_number, cid_number, sizeof(user->cid_number));
	ast_copy_string(user->channel_name, channel_name, sizeof(user->channel_name));
	ast_copy_string(user->conference_id, conf->conference_id, sizeof(user->conference_id));
	hash_user(user);
	user->admin = (int) admin_status;
	user->muted = (int) mute_status;
	user->can_invite = (int) can_invite;

	// User type, User role, cid name
	if (strcmp(mcb_type, CEI_MCB_TYPE_NONE) == 0) {
		user->user_type = CEI_TYPE_NORMAL_USER;
		ast_copy_string(user->cid_name, cid_name, sizeof(user->cid_name));

		if (user->admin) {
			user->user_role = CEI_ROLE_ADMIN;
		}
		else if (user->can_invite) {
			user->user_role = CEI_ROLE_USER_INVITE;
		}
		else {
			user->user_role = CEI_ROLE_USER_NORMAL;
		}

	} else if (strcmp(mcb_type, CEI_MCB_TYPE_EXTERNAL) == 0) {
		user->user_type = CEI_TYPE_EXTERNAL_MCB;
		user->user_role = CEI_ROLE_CONFERENCE;
		ast_copy_string(user->cid_name, CEI_USER_ROLE_CONFERENCE, sizeof(user->cid_name));
		if (mcb_remote_conf_id != NULL) {
			ast_copy_string(user->self_conf_id, mcb_remote_conf_id, sizeof(user->self_conf_id));
		}

	} else if (strcmp(mcb_type, CEI_MCB_TYPE_INTERNAL) == 0) {
		user->user_type = CEI_TYPE_INTERNAL_MCB;
		user->user_role = CEI_ROLE_CONFERENCE;
		ast_copy_string(user->cid_name, CEI_USER_ROLE_CONFERENCE, sizeof(user->cid_name));
		if (mcb_remote_conf_id != NULL) {
			ast_copy_string(user->self_conf_id, mcb_remote_conf_id, sizeof(user->self_conf_id));
		}

	} else {
		user->user_type = CEI_TYPE_UNKNOWN;
	}

	ao2_lock(conf->user_container);
	ao2_link(conf->user_container, user);
	ast_atomic_fetchadd_int(&conf->local_size, 1);
	ast_atomic_fetchadd_int(&conf->local_user_size, (!check_role_as_conference(user))? 1 : 0);
	ast_atomic_fetchadd_int(&conf->refcount, 1);
	ao2_unlock(conf->user_container);

	poll_cei_conference(conf, &full_text);
	poll_cei_partial_notify(conf, NULL, user, &partial_text, CEI_EVENT_JOIN);

	// Prepare the stasis data
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);

	json_object = ast_json_pack("{s: i, s: s, s: i, s: i, s: i, s: s, s: s, s: s, s: s}",
		"user_type", user->user_type,
		"conference", conference_name,
		"preverified_user", invited_status,
		"version_reset", conf->version_reset,
		"user_role", user->user_role,
		"user_name", check_role_as_conference(user) ? user->self_conf_id : channel_name,
		"full_text", ast_str_buffer(full_text),
		"partial_text", ast_str_buffer(partial_text),
		"event_type", CEI_NOTIFY_EVENT_JOIN
	);

	cei_publish_manager_event(json_object);

	ast_free(partial_text);
	ast_free(full_text);
}

static void cei_leave_cb(void *data, struct stasis_subscription *sub,
	struct stasis_message *message)
{
	enum cei_user_type user_type = CEI_TYPE_UNKNOWN;
	char self_conf_id[41] = {0};
	struct cei_conference *conf;
	struct cei_user *user = NULL;
	struct ast_str *extra_text = ast_str_create(CEI_EVENT_DATA_LEN);

	const struct ast_bridge_blob *blob = stasis_message_data(message);
	unsigned int user_count= ast_json_integer_get(ast_json_object_get(blob->blob, "user_count"));
	unsigned int user_number = ast_json_integer_get(ast_json_object_get(blob->blob, "user_number"));
	const char *conference_name = ast_json_string_get(ast_json_object_get(blob->blob, "conference"));
	const char *cid_number = ast_json_string_get(ast_json_object_get(blob->blob, "caller_id_number"));
	const char *channel_name = ast_json_string_get(ast_json_object_get(blob->blob, "channel_name"));

	ast_debug(LOG_DEBUG, "[WL] Delete a user : user_count: %d, user_num:%d, cid_num:%s, conf_name:%s",
												user_count,
												user_number,
												cid_number,
												conference_name
	);

	if (conference_name == NULL || cid_number == NULL || channel_name == NULL) {
		ast_debug(LOG_DEBUG, "[WL] EVENT INCOMPLETE");
		ast_free(extra_text);
		return;
	}

	// get cei_conference
	conf = build_conf_by_num(conference_name);

	if (!conf) {
		ast_free(extra_text);
		return;
	}

	if (user_count != (conf->local_size - 1)) {
		ast_debug(LOG_DEBUG, "[WL] CONF SIZE MISMATCH %d vs %d", conf->local_size, user_count);
	}

	ao2_lock(conf->user_container);

	user = get_user_by_cid_num(conf, cid_number);
	if (user) {
		user_type = user->user_type;
		if (check_role_as_conference(user)) {
			ast_copy_string(self_conf_id, user->self_conf_id, sizeof(self_conf_id));
		}
		ast_atomic_fetchadd_int(&conf->local_size, -1);
		ast_atomic_fetchadd_int(&conf->local_user_size, (!check_role_as_conference(user))? -1 : 0);

		poll_cei_partial_notify(conf, NULL, user, &extra_text, CEI_EVENT_PART);

		ao2_unlink(conf->user_container, user);
		ast_atomic_fetchadd_int(&conf->refcount, -1);

		ast_debug(LOG_DEBUG, "[WL] REMOVED USER WITH ADDRESS %p", user);
		ao2_ref(user, -1);
	}
	else {
		ast_debug(LOG_DEBUG, "[WL] ERROR, CANNOT FIND USER NUMBER %d IN CONFERENCE %s", user_number, conf->conference_num);
		ao2_unlock(conf->user_container);
		ast_free(extra_text);
		return;
	}

	ao2_unlock(conf->user_container);

	// Prepare the stasis data
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);

	json_object = ast_json_pack("{s: i, s: s, s: i, s: s, s: s, s: s}",
		"user_type", user_type,
		"conference", conference_name,
		"version_reset", conf->version_reset,
		"user_name", (!ast_strlen_zero(self_conf_id)) ? self_conf_id : channel_name,
		"partial_text", ast_str_buffer(extra_text),
		"event_type", CEI_NOTIFY_EVENT_LEAVE
	);

	cei_publish_manager_event(json_object);

	reset_conference(conf);

	ast_free(extra_text);
}


static void cei_lock_status_cb(void *data, struct stasis_subscription *sub,
	struct stasis_message *message)
{
	struct cei_conference *conf;
	struct ast_str *extra_text = ast_str_create(CEI_EVENT_DATA_LEN);;

	const struct ast_bridge_blob *blob = stasis_message_data(message);
	const char *lock_status= ast_json_string_get(ast_json_object_get(blob->blob, "status"));
	const char *conference_name = ast_json_string_get(ast_json_object_get(blob->blob, "conference"));

	if (lock_status == NULL || conference_name == NULL) {
		ast_debug(LOG_DEBUG, "[WL] EVENT INCOMPLETE");
		ast_free(extra_text);
		return;
	}

	// get cei_conference
	conf = build_conf_by_num(conference_name);
	ast_debug(LOG_DEBUG, "[WL] Lock conference: lock_status: %s, conf_name:%s;",
												lock_status,
												conference_name
	);
	if (!conf) {
		ast_free(extra_text);
		return;
	}

	conf->locked = !strcmp(lock_status, "locked") ? TRUE : FALSE;

	poll_cei_partial_notify(conf, NULL, NULL, &extra_text, CEI_EVENT_TOGGLE_LOCK);

	// Prepare the stasis data
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);

	json_object = ast_json_pack("{s: s, s: i, s: s, s: s}",
		"conference", conference_name,
		"version_reset", conf->version_reset,
		"partial_text", ast_str_buffer(extra_text),
		"event_type", CEI_NOTIFY_EVENT_LOCK
	);

	cei_publish_manager_event(json_object);

	ast_free(extra_text);
}

static void cei_toggle_mute_cb(void *data, struct stasis_subscription *sub,
	struct stasis_message *message)
{
	struct cei_conference *conf;
	struct cei_user *user = NULL;
	struct ast_str *extra_text = ast_str_create(CEI_EVENT_DATA_LEN);

	const struct ast_bridge_blob *blob = stasis_message_data(message);
	unsigned int user_number = ast_json_integer_get(ast_json_object_get(blob->blob, "user_number"));
	const char *conference_name = ast_json_string_get(ast_json_object_get(blob->blob, "conference"));
	unsigned int mute_status = ast_json_integer_get(ast_json_object_get(blob->blob, "muted"));
	unsigned int admin_status = 0;
	if (ast_json_is_true(ast_json_object_get(blob->blob, "admin"))) {
			admin_status = 1;
	}
	const char *cid_number = ast_json_string_get(ast_json_object_get(blob->blob, "caller_id_number"));

	ast_debug(LOG_DEBUG, "[WL] Mute user: user_num:%d, conf_name %s, must_status:%d, admin_status:%d",
												user_number,
												conference_name,
												mute_status,
												admin_status
	);

	if (conference_name == NULL) {
		ast_debug(LOG_DEBUG, "[WL] EVENT INCOMPLETE");
		ast_free(extra_text);
		return;
	}

	conf = build_conf_by_num(conference_name);

	ao2_lock(conf->user_container);

	user = get_user_by_cid_num(conf, cid_number);
	if (user) {
		user->muted = mute_status;
	}
	else {
		ast_debug(LOG_DEBUG, "[WL] ERROR, CANNOT FIND USER NUMBER %d IN CONFERENCE %s", user_number, conf->conference_num);
		ao2_unlock(conf->user_container);
		ast_free(extra_text);
		return;
	}
	ao2_unlock(conf->user_container);

	poll_cei_partial_notify(conf, NULL, user, &extra_text, CEI_EVENT_TOGGLE_MUTE);

	// Prepare the stasis data
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);

	json_object = ast_json_pack("{s: i, s: s, s: i, s: s, s: s}",
		"user_type", user->user_type,
		"conference", conference_name,
		"version_reset", conf->version_reset,
		"partial_text", ast_str_buffer(extra_text),
		"event_type", CEI_NOTIFY_EVENT_MUTE
	);

	cei_publish_manager_event(json_object);

	ast_free(extra_text);
}

// UNSUB for both ext MCB and int MCB
static void cei_mcb_process_unsub_cb(void *data, struct stasis_subscription *sub,
	struct stasis_message *message)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s \n", __FUNCTION__);
	struct cei_conference	*host_conf  = NULL;
	struct ast_str			*extra_text = ast_str_create(CEI_EVENT_DATA_LEN);
	struct ast_json_payload *payload    = NULL;
	struct ast_json			*json_object = NULL;

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_debug(LOG_DEBUG, "Got NULL payload from stasis message.");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_debug(LOG_DEBUG, "[WL] Imcomplete message.");
		return;
	}

	const char *conf_id = ast_json_string_get(ast_json_object_get(json_object, "conf_id"));
	const char *publisher = ast_json_string_get(ast_json_object_get(json_object, "publisher"));


	if (conf_id == NULL || publisher == NULL) {
		ast_debug(LOG_DEBUG, "[WL] Imcomplete data. \n");
		return;
	}

	ast_debug(LOG_DEBUG, "[WL] Got mcb leave data : %s,%s",
												conf_id,
												publisher);

	// Get host conference
	host_conf = get_conf_by_id(conf_id);
	if (host_conf == NULL) {
		ast_free(extra_text);
		return;
	}

	delete_remote_confs_by_publisher(host_conf, publisher);

	poll_cei_conference(host_conf, &extra_text);

	// Prepare the stasis data
	RAII_VAR(struct ast_json *, to_send_json_object, NULL, ast_json_unref);

	to_send_json_object = ast_json_pack("{s: s, s: i, s: s, s: s}",
		"conference", host_conf->conference_num,
		"version_reset", host_conf->version_reset,
		"full_text", ast_str_buffer(extra_text),
		"event_type", CEI_NOTIFY_EVENT_MCB_LEAVE
	);

	cei_publish_manager_event(to_send_json_object);

	ast_free(extra_text);
}

static void cei_ext_mcb_process_xml_cb(void *data, struct stasis_subscription *sub,
	struct stasis_message *message)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s \n", __FUNCTION__);
	struct cei_conference *host_conf = NULL;
	struct ast_str *extra_text = ast_str_create(CEI_EVENT_DATA_LEN);

	struct ast_json_payload *payload     = NULL;
	struct ast_json			*json_object = NULL;

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_debug(LOG_DEBUG, "Got NULL payload from stasis message.");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_debug(LOG_DEBUG, "[WL] Imcomplete message.");
		return;
	}
	const char *conf_no = ast_json_string_get(ast_json_object_get(json_object, "conf_num"));
	const char *xml = ast_json_string_get(ast_json_object_get(json_object, "xml"));
	const char *publisher_id = ast_json_string_get(ast_json_object_get(json_object, "publisher_id"));
	const char *publisher_private_num = ast_json_string_get(ast_json_object_get(json_object, "publisher_private_cid"));

	if (conf_no == NULL || xml == NULL || publisher_id == NULL || publisher_private_num == NULL) {
		ast_debug(LOG_DEBUG, "[WL] Imcomplete data. \n");
		return;
	}
	ast_debug(LOG_DEBUG, "[WL] Got mcb data : conf %s, publisher %s %s, content %s",
												conf_no,
												publisher_id,
												publisher_private_num,
												xml
												);

	// Get host conference
	host_conf = get_conf_by_num(conf_no);
	if (host_conf == NULL) {
		// No user in conf ever before MCB
		host_conf = build_conf_by_num(conf_no);
	}

	if (host_conf == NULL) {
		ast_free(extra_text);
		return;
	}

	process_remote_xml(host_conf, publisher_private_num, xml, &extra_text);

	// Prepare the stasis data
	RAII_VAR(struct ast_json *, to_send_json_object, NULL, ast_json_unref);


	to_send_json_object = ast_json_pack("{s: i, s: s, s: i, s: s, s: s, s: s}",
		"user_type", CEI_TYPE_EXTERNAL_MCB,
		"conference", conf_no,
		"version_reset", host_conf->version_reset,
		"user_name", publisher_id,
		"partial_text", ast_str_buffer(extra_text),
		"event_type", CEI_NOTIFY_EVENT_EXT_MCB
	);

	cei_publish_manager_event(to_send_json_object);

	ast_free(extra_text);
}

static void cei_int_mcb_process_string_cb(void *data, struct stasis_subscription *sub,
	struct stasis_message *message)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s \n", __FUNCTION__);
	struct cei_conference *host_conf = NULL;
	struct cei_conference *remote_conf = NULL;
	struct ast_str *extra_text = ast_str_create(CEI_EVENT_DATA_LEN);

	struct ast_json_payload *payload     = NULL;
	struct ast_json			*json_object = NULL;

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_debug(LOG_DEBUG, "[WL] Imcomplete message.");
		return;
	}
	const char *conf_no = ast_json_string_get(ast_json_object_get(json_object, "conf_num"));
	const char *content = ast_json_string_get(ast_json_object_get(json_object, "content"));
	const char *publisher_id = ast_json_string_get(ast_json_object_get(json_object, "publisher_id"));

	if (conf_no == NULL || content == NULL || publisher_id == NULL) {
		ast_debug(LOG_DEBUG, "[WL] Imcomplete data. \n");
		return;
	}
	ast_debug(LOG_DEBUG, "[WL] Got mcb data : conf %s, publisher %s,%s",
												conf_no,
												publisher_id,
												content
												);

	// Get host conference
	host_conf = get_conf_by_num(conf_no);
	if (host_conf == NULL) {
		host_conf = build_conf_by_num(conf_no);
	}

	if (host_conf == NULL) {
		ast_free(extra_text);
		return;
	}

	remote_conf = get_conf_by_id(publisher_id);
	if (remote_conf == NULL) {
		ast_free(extra_text);
		return;
	}

	//change to string
	char *parsing_content = ast_strdup(content);
	process_remote_string(host_conf, remote_conf->conference_num, publisher_id, parsing_content, &extra_text);

	// Prepare the stasis data
	RAII_VAR(struct ast_json *, to_send_json_object, NULL, ast_json_unref);

	to_send_json_object = ast_json_pack("{s: i, s: s, s: i, s: s, s: s, s: s}",
		"user_type", CEI_TYPE_INTERNAL_MCB,
		"conference", conf_no,
		"version_reset", host_conf->version_reset,
		"user_name", publisher_id,
		"partial_text", ast_str_buffer(extra_text),
		"event_type", CEI_NOTIFY_EVENT_INT_MCB
	);

	cei_publish_manager_event(to_send_json_object);

    // Clean up
	ast_free(extra_text);
    ast_free(parsing_content);
}


STASIS_MESSAGE_TYPE_DEFN(cei_join_type);
STASIS_MESSAGE_TYPE_DEFN(cei_leave_type);
STASIS_MESSAGE_TYPE_DEFN(cei_lock_status_type);
STASIS_MESSAGE_TYPE_DEFN(cei_toggle_mute_type);

void manager_cei_shutdown(void) {
	STASIS_MESSAGE_TYPE_CLEANUP(cei_join_type);
	STASIS_MESSAGE_TYPE_CLEANUP(cei_leave_type);
	STASIS_MESSAGE_TYPE_CLEANUP(cei_lock_status_type);
	STASIS_MESSAGE_TYPE_CLEANUP(cei_toggle_mute_type);

	if (bridge_state_router) {
		stasis_message_router_unsubscribe(bridge_state_router);
		bridge_state_router = NULL;
	}

	if (channel_state_router) {
		stasis_message_router_unsubscribe(channel_state_router);
		channel_state_router = NULL;
	}

	if (cei_mcb_process_router) {
		stasis_message_router_unsubscribe(cei_mcb_process_router);
		cei_mcb_process_router = NULL;
	}

	/*Clear up all struct*/
	conf_free_all();
}


int manager_cei_init(void)
{
	STASIS_MESSAGE_TYPE_INIT(cei_join_type);
	STASIS_MESSAGE_TYPE_INIT(cei_leave_type);
	STASIS_MESSAGE_TYPE_INIT(cei_lock_status_type);
	STASIS_MESSAGE_TYPE_INIT(cei_toggle_mute_type);

	bridge_state_router = stasis_message_router_create(
		ast_bridge_topic_all_cached());

	if (!bridge_state_router) {
		return -1;
	}

	if (stasis_message_router_add(bridge_state_router,
			cei_join_type(),
			cei_join_cb,
			NULL)) {
		manager_cei_shutdown();
		return -1;
	}
	if (stasis_message_router_add(bridge_state_router,
			cei_leave_type(),
			cei_leave_cb,
			NULL)) {
		manager_cei_shutdown();
		return -1;
	}
	if (stasis_message_router_add(bridge_state_router,
			cei_lock_status_type(),
			cei_lock_status_cb,
			NULL)) {
		manager_cei_shutdown();
		return -1;
	}
	if (stasis_message_router_add(bridge_state_router,
			cei_toggle_mute_type(),
			cei_toggle_mute_cb,
			NULL)) {
		manager_cei_shutdown();
		return -1;
	}


	channel_state_router = stasis_message_router_create(
		ast_channel_topic_all_cached());

	if (!channel_state_router) {
		manager_cei_shutdown();
		return -1;
	}

	if (stasis_message_router_add(channel_state_router,
		  cei_join_type(),
			cei_join_cb,
			NULL)) {
		manager_cei_shutdown();
		return -1;
	}
	if (stasis_message_router_add(channel_state_router,
			cei_leave_type(),
			cei_leave_cb,
			NULL)) {
		manager_cei_shutdown();
		return -1;
	}
	if (stasis_message_router_add(channel_state_router,
			cei_lock_status_type(),
			cei_lock_status_cb,
			NULL)) {
		manager_cei_shutdown();
		return -1;
	}
	if (stasis_message_router_add(channel_state_router,
			cei_toggle_mute_type(),
			cei_toggle_mute_cb,
			NULL)) {
		manager_cei_shutdown();
		return -1;
	}

	// Add CEI Manager-CEI NOTIFY generator message routing for MCB
	struct stasis_topic *cei_mcb_process_topic;

	cei_mcb_process_topic = ast_cei_mcb_process_topic_all();
	if (!cei_mcb_process_topic) {
		ast_log(LOG_WARNING, "[WL] Failed to get cei_mcb_process_topic\n");
		manager_cei_shutdown();
		return -1;
	}

	cei_mcb_process_router = stasis_message_router_create(cei_mcb_process_topic);
	if (cei_mcb_process_router == NULL) {
		ast_log(LOG_WARNING, "[WL] Failed to create cei_mcb_process_router\n");
		manager_cei_shutdown();
		return -1;
	}

	if (stasis_message_router_add(cei_mcb_process_router,
									ast_cei_mcb_process_unsub_type(),
									cei_mcb_process_unsub_cb, NULL))
	{
		ast_log(LOG_WARNING, "[WL] Failed to subscribe to ast_cei_mcb_process_unsub_type\n");
		manager_cei_shutdown();
		return -1;
	}

	if (stasis_message_router_add(cei_mcb_process_router,
									ast_cei_mcb_process_xml_type(),
									cei_ext_mcb_process_xml_cb, NULL))
	{
		ast_log(LOG_WARNING, "[WL] Failed to subscribe to ast_cei_mcb_process_xml_type\n");
		manager_cei_shutdown();
		return -1;
	}

	if (stasis_message_router_add(cei_mcb_process_router,
									ast_cei_mcb_process_string_type(),
									cei_int_mcb_process_string_cb, NULL))
	{
		ast_log(LOG_WARNING, "[WL] Failed to subscribe to ast_cei_mcb_process_string_type\n");
		manager_cei_shutdown();
		return -1;
	}

	// Get conf max user capability
	conf_max_capability = get_conf_max_capability();

	return 0;
}


static struct cei_conference* build_conf_by_num(const char *conf_no) {

	struct cei_conference *conf;
	char conf_id[41];
	char database_tree[MAX_CONFNUM + 5];

	if (conf_no == NULL) {
		return NULL;
	}

	memset(conf_id, 0, sizeof(conf_id));
	snprintf(database_tree, sizeof(database_tree), "MCB/%s", conf_no);

	if (ast_db_get(database_tree, "LOCAL_IDENTITY", conf_id, sizeof(conf_id)) == 0) {
		ast_debug(LOG_DEBUG, "[WL] ROOM %s HAS LOCAL_IDENTITY %s", conf_no, conf_id);
	}
	else {
		ast_debug(LOG_DEBUG, "[WL] ROOM %s HAS LOCAL_IDENTITY NONE", conf_no);

		char unix_time[16];
		snprintf(unix_time, sizeof(unix_time), "%u", (int) time(NULL));
		char random_num[3];
		snprintf(random_num, sizeof(random_num), "%d", (rand() % 90) + 10);
		char to_hash[strlen(conf_no) + strlen(unix_time) + strlen(random_num) + 1];
		strcpy(to_hash, conf_no);
		strcpy(to_hash + strlen(to_hash), unix_time);
		strcpy(to_hash + strlen(to_hash), random_num);

		char hash_result[41];
		ast_sha1_hash(hash_result, to_hash);

		ast_copy_string(conf_id, hash_result, sizeof(conf_id));
		ast_db_put(database_tree, "LOCAL_IDENTITY", hash_result);
	}
	ast_debug(LOG_DEBUG, "[WL] conference id IS %s", conf_id);

	AST_LIST_LOCK(&confs);
	AST_LIST_TRAVERSE(&confs, conf, list) {
		if (!strcmp(conf_id, conf->conference_id)){
			AST_LIST_UNLOCK(&confs);
			ast_debug(LOG_DEBUG, "[WL] Found the confernece %s", conf->conference_id);
			return conf;
		}
	}

	if (!(conf = ao2_alloc(sizeof(struct cei_conference), NULL)) ||
		!(conf->remote_container = ao2_container_alloc(1, NULL, conference_id_cmp)) ||
		!(conf->user_container = ao2_container_alloc(1, NULL, user_hash_cmp))) {
		if (conf) {
			conf_free(conf);
		}
		ast_debug(LOG_DEBUG, "[WL] CONFERENCE %s %s CANNOT BE CREATED\n", conf_no, conf_id);
		AST_LIST_UNLOCK(&confs);
	}
	else {
		ast_copy_string(conf->conference_num, conf_no, sizeof(conf->conference_num));
		ast_copy_string(conf->conference_id, conf_id, sizeof(conf->conference_id));
		ast_debug(LOG_DEBUG, "[WL] Create new conference struct %s", conf->conference_id);
		conf->locked = FALSE;
		conf->local_size = 0;
		conf->remote_size = 0;
		conf->local_user_size = 0;
		conf->remote_user_size = 0;
		conf->mcb_count = 0;
		conf->refcount = 0;
		conf->host_conf = NULL;
		conf->version = 0;
		conf->version_reset = 0;
		conf->stop_propagating = FALSE;

		AST_LIST_INSERT_HEAD(&confs, conf, list);
		ast_atomic_fetchadd_int(&conf->refcount, 1);

		AST_LIST_UNLOCK(&confs);

		return conf;
	}

	return NULL;
}

static int user_hash_cmp(void *obj, void *arg, int flags) {
	struct cei_user *user = obj;
	char *user_hash = arg;

	if (!strcmp(user->user_hash, user_hash)) {
		return (CMP_MATCH | CMP_STOP);
	}

	return 0;
}

static int conference_id_cmp(void *obj, void *arg, int flags) {
	struct cei_conference *conf = obj;
	char *conference_id = arg;

	if (!strcmp(conf->conference_id, conference_id)) {
		return (CMP_MATCH | CMP_STOP);
	}

	return 0;
}

static void hash_user(struct cei_user *user) {
	char to_hash[51];

	snprintf(to_hash, sizeof(to_hash), "%s%d", user->conference_id, user->user_num);
	ast_sha1_hash(user->user_hash, to_hash);
}

static void poll_cei_conference(struct cei_conference *conf, struct ast_str **data_string) {

	char user_string[CEI_EVENT_USER_LEN];
	struct ao2_iterator user_iter;
	struct cei_user *user;
	int user_status = CEI_STATUS_UNKNOWN;

	// Sanity check
	if (conf == NULL) {
		return;
	}

	ao2_lock(conf->user_container);
	ao2_lock(conf->remote_container);

	// Append conference information
	// notify_type|local_count|remote_count|local_user_count|remote_user_count|total_count_need_update|conf_locked|subject|free_text|when|maximum_user_count|conference_type\tconf_user_count\tuser_number\tcid_number\tcid_name\tchannel_name\tconference_id\tuser_role\tentity_status|

	ast_str_append(data_string, 0, "%d|%d|%d|%d|%d|%d|%d|%s|%s|%s|%d|",
									CEI_NOTIFY_FULL,
									conf->local_size,
									conf->remote_size,
									conf->local_user_size,
									conf->remote_user_size,
									conf->local_size + conf->remote_size + ao2_container_count(conf->remote_container),
									conf->locked,
									"NULL",
									"NULL",
									"NULL",
									conf_max_capability
									);

	AST_LIST_LOCK(&confs);

	// LOOP over user in conference
	user_iter = ao2_iterator_init(conf->user_container, 0);
	while ((user = ao2_iterator_next(&user_iter))) {

		// User status
		user_status = CEI_STATUS_CONNECTED;
		if (user->muted) {
			user_status = CEI_STATUS_MUTEVIAFOCUS;
		}

		// Append current conference users
		// conference_type\tconf_user_count\tuser_number\tcid_number\tcid_name\tchannel_name\t
		// conference_id\tself_conference_id\tuser_role\tentity_status|

		snprintf(user_string, CEI_EVENT_USER_LEN,
					   "%d\t%d\t%d\t%.32s\t%.32s\t%.100s\t%s\t%s\t%d\t%d|",
					   /* Modified by achan, Grandstream, bug 32194 */
					   CEI_CONFERENCE_LOCAL,
						 0,
					   user->user_num,
					   user->cid_number,
					   user->cid_name,
					   user->channel_name,
					   user->conference_id,
					   (!ast_strlen_zero(user->self_conf_id)) ? user->self_conf_id : "NULL",
					   user->user_role,
					   user_status);

		ast_str_append(data_string, 0, user_string);
		ao2_ref(user, -1);
	}

	ao2_iterator_destroy(&user_iter);

	if (ao2_container_count(conf->remote_container)) {
		struct ao2_iterator remote_conf_iter = ao2_iterator_init(conf->remote_container, 0);
		struct cei_conference *remote_conf;
		while ((remote_conf = ao2_iterator_next(&remote_conf_iter))) {
			// Append new remote conference
			// conference_type\tconf_user_count\tuser_number\tcid_number\tcid_name\tchannel_name\t
			// conference_id\tself_conference_id\tuser_role\tentity_status|
			snprintf(user_string, CEI_EVENT_USER_LEN,
						   "%d\t%d\t%d\t%.32s\t%.32s\t%.100s\t%s\t%s\t%d\t%d|",
						   /* Modified by achan, Grandstream, bug 32194 */
						   CEI_CONFERENCE_REMOTE,
							 remote_conf->local_user_size,
						   0,
						   remote_conf->conference_num,
						   remote_conf->conference_num,
						   "unknown",
						   remote_conf->conference_id,
						   "NULL",
						   CEI_ROLE_CONFERENCE,
						   CEI_STATUS_CONNECTED);

			ast_str_append(data_string, 0, user_string);

			if ((conf->local_size + conf->remote_size) <= MAX_NOTIFY_USERS_DISPLAY) {
				ao2_lock(remote_conf->user_container);
				user_iter = ao2_iterator_init(remote_conf->user_container, 0);
				while ((user = ao2_iterator_next(&user_iter))) {

					// User status
					user_status = CEI_STATUS_CONNECTED;
					if (user->muted) {
						user_status = CEI_STATUS_MUTEVIAFOCUS;
					}

					// Append current remote conference users
					// conference_type\tconf_user_count\tuser_number\tcid_number\tcid_name\tchannel_name\t
					// conference_id\tself_conference_id\tuser_role\tentity_status|
					snprintf(user_string, CEI_EVENT_USER_LEN,
								   "%d\t%d\t%d\t%.32s\t%.32s\t%.100s\t%s\t%s\t%d\t%d|",
								   /* Modified by achan, Grandstream, bug 32194 */
								   CEI_CONFERENCE_LOCAL,
									 0,
								   user->user_num,
								   user->cid_number,
								   user->cid_name,
								   user->channel_name,
								   user->conference_id,
									 (!ast_strlen_zero(user->self_conf_id)) ? user->self_conf_id : "NULL",
								   user->user_role,
								   user_status);

					ast_str_append(data_string, 0, user_string);
					ao2_ref(user, -1);
				}
				ao2_unlock(remote_conf->user_container);
				ao2_iterator_destroy(&user_iter);
			}
			ao2_ref(remote_conf, -1);
		}
		ao2_iterator_destroy(&remote_conf_iter);
	}

	ao2_unlock(conf->remote_container);
	ao2_unlock(conf->user_container);
	AST_LIST_UNLOCK(&confs);
	ast_log(LOG_DEBUG, "[WL] Exit poll_cei_conference with data: %s", ast_str_buffer(*data_string));
}

static void poll_cei_partial_notify(struct cei_conference *conf, struct cei_conference *remote_conf, struct cei_user *user,
	   																struct ast_str **data_string, enum cei_event_type event_type ) {
	char user_string[CEI_EVENT_USER_LEN];
	int user_status = CEI_STATUS_UNKNOWN;

	// Sanity check
	if (conf == NULL) {
		return;
	}
	switch (event_type)
	{
		case CEI_EVENT_JOIN:
		{
			user_status = (user_status == CEI_STATUS_UNKNOWN) ? CEI_STATUS_CONNECTED : user_status;
		}
		case CEI_EVENT_PART:
		{
			user_status = (user_status == CEI_STATUS_UNKNOWN) ? CEI_STATUS_DISCONNECTED : user_status;
		}
		case CEI_EVENT_TOGGLE_MUTE:
		{
			if (user == NULL) {
				return;
			}

			if (user_status == CEI_STATUS_UNKNOWN) {
				if(user->muted) {
					user_status = CEI_STATUS_MUTEVIAFOCUS;
				} else {
					user_status = CEI_STATUS_CONNECTED;
				}
			}

			// Append conference information
			// notify_type|local_count|remote_count|local_user_count|remote_user_count|total_count_need_update|conf_locked|subject|free_text|when|maximum_user_count|conference_type\tconf_user_count\tuser_number\tcid_number\tcid_name\tchannel_name\tconference_id\tuser_role\tentity_status|

			ast_str_append(data_string, 0, "%d|%d|%d|%d|%d|%d|%d|%s|%s|%s|%d|",
											CEI_NOTIFY_PARTIAl,
											conf->local_size,
											conf->remote_size,
											conf->local_user_size,
											conf->remote_user_size,
											1,
											conf->locked,
											"NULL",
											"NULL",
											"NULL",
											conf_max_capability
											);

			// Append current conference users
			// conference_type\tconf_user_count\tuser_number\tcid_number\tcid_name\tchannel_name\t
			// conference_id\tself_conference_id\tuser_role\tentity_status|
			snprintf(user_string, CEI_EVENT_USER_LEN,
							"%d\t%d\t%d\t%.32s\t%.32s\t%.100s\t%s\t%s\t%d\t%d|",
							/* Modified by achan, Grandstream, bug 32194 */
							CEI_CONFERENCE_LOCAL,
							0,
							user->user_num,
							user->cid_number,
							user->cid_name,
							user->channel_name,
							user->conference_id,
							(!ast_strlen_zero(user->self_conf_id)) ? user->self_conf_id : "NULL",
							user->user_role,
							user_status);

			ast_str_append(data_string, 0, user_string);
		}
		break;
		case CEI_EVENT_TOGGLE_LOCK:
		{
			// Append conference information
			// notify_type|local_count|remote_count|local_user_count|remote_user_count|total_count_need_update|conf_locked|subject|free_text|when|maximum_user_count|conference_type\tconf_user_count\tuser_number\tcid_number\tcid_name\tchannel_name\tconference_id\tuser_role\tentity_status|

			ast_str_append(data_string, 0, "%d|%d|%d|%d|%d|%d|%d|%s|%s|%s|%d|",
											CEI_NOTIFY_PARTIAl,
											conf->local_size,
											conf->remote_size,
											conf->local_user_size,
											conf->remote_user_size,
											0,
											conf->locked,
											"NULL",
											"NULL",
											"NULL",
											conf_max_capability
											);
		}
		break;
		case CEI_EVENT_MCB_USER_JOIN:
		{
			user_status = (user_status == CEI_STATUS_UNKNOWN) ? CEI_STATUS_CONNECTED : user_status;
		}
		case CEI_EVENT_MCB_USER_PART:
		{
			user_status = (user_status == CEI_STATUS_UNKNOWN) ? CEI_STATUS_DISCONNECTED : user_status;
		}
		case CEI_EVENT_MCB_USER_TOGGLE_MUTE:
		{
			if (remote_conf == NULL || user == NULL) {
				return;
			}

			if (user_status == CEI_STATUS_UNKNOWN) {
				if(user->muted) {
					user_status = CEI_STATUS_MUTEVIAFOCUS;
				} else {
					user_status = CEI_STATUS_CONNECTED;
				}
			}

			// Append conference information
			// notify_type|local_count|remote_count|local_user_count|remote_user_count|total_count_need_update|conf_locked|subject|free_text|when|maximum_user_count|conference_type\tconf_user_count\tuser_number\tcid_number\tcid_name\tchannel_name\tconference_id\tuser_role\tentity_status|

			ast_str_append(data_string, 0, "%d|%d|%d|%d|%d|%d|%d|%s|%s|%s|%d|",
											CEI_NOTIFY_PARTIAl,
											conf->local_size,
											conf->remote_size,
											conf->local_user_size,
											conf->remote_user_size,
											2,
											conf->locked,
											"NULL",
											"NULL",
											"NULL",
											conf_max_capability
											);
			// Append new remote conference
			// conference_type\tconf_user_count\tuser_number\tcid_number\tcid_name\tchannel_name\t
			// conference_id\tself_conference_id\tuser_role\tentity_status|
			snprintf(user_string, CEI_EVENT_USER_LEN,
							"%d\t%d\t%d\t%.32s\t%.32s\t%.100s\t%s\t%s\t%d\t%d|",
							/* Modified by achan, Grandstream, bug 32194 */
							CEI_CONFERENCE_REMOTE,
							remote_conf->local_user_size,
							0,
							remote_conf->conference_num,
							remote_conf->conference_num,
							"unknown",
							remote_conf->conference_id,
							"NULL",
							CEI_ROLE_CONFERENCE,
							CEI_STATUS_CONNECTED);

			ast_str_append(data_string, 0, user_string);

			// Append current remote conference users
			// conference_type\tconf_user_count\tuser_number\tcid_number\tcid_name\tchannel_name\t
			// conference_id\tself_conference_id\tuser_role\tentity_status|
			snprintf(user_string, CEI_EVENT_USER_LEN,
							"%d\t%d\t%d\t%.32s\t%.32s\t%.100s\t%s\t%s\t%d\t%d|",
							/* Modified by achan, Grandstream, bug 32194 */
							CEI_CONFERENCE_LOCAL,
							0,
							user->user_num,
							user->cid_number,
							user->cid_name,
							user->channel_name,
							user->conference_id,
							(!ast_strlen_zero(user->self_conf_id)) ? user->self_conf_id : "NULL",
							user->user_role,
							user_status);

			ast_str_append(data_string, 0, user_string);
		}
		break;
		default:
		{
			ast_log(LOG_DEBUG, "[WL] Invalid event type !");
		}
		break;
	}

	ast_log(LOG_DEBUG, "[WL] Exit poll_cei_partial_notify with data: %s", ast_str_buffer(*data_string));
}

static void conf_free_all(void) {
	struct cei_conference *conf = NULL;

	AST_LIST_LOCK(&confs);
	while ((conf = AST_LIST_REMOVE_HEAD(&confs, list))) {
		conf_free(conf);
	}
	AST_LIST_UNLOCK(&confs);
}

// Assume confs is locked
static int conf_free(struct cei_conference *conf) {

	struct ao2_iterator iter;
	if (conf->remote_container) {
		struct cei_conference *remote_conf = NULL;
		ao2_lock(conf->remote_container);
		iter = ao2_iterator_init(conf->remote_container, 0);
		while ((remote_conf = ao2_iterator_next(&iter))) {
			ao2_unlink(conf->remote_container, remote_conf);
			clear_conf_users(remote_conf);
			ao2_ref(remote_conf,-2);
		}
		ao2_iterator_destroy(&iter);
		ao2_unlock(conf->remote_container);
		ao2_ref(conf->remote_container,-1);
	}
	if (conf->user_container) {
		clear_conf_users(conf);
		ao2_ref(conf->user_container, -1);
	}
	ao2_cleanup(conf);
	return 0;
}

static void clear_conf_users(struct cei_conference *conf) {

	struct ao2_iterator user_iter;
	struct cei_user *user;
	struct cei_user *next_user;

	ao2_lock(conf->user_container);
	user_iter = ao2_iterator_init(conf->user_container, 0);
	user = ao2_iterator_next(&user_iter);
	while (user) {
		next_user = ao2_iterator_next(&user_iter);
		ao2_unlink(conf->user_container, user);
		ast_atomic_fetchadd_int(&conf->local_size, -1);
		ast_atomic_fetchadd_int(&conf->local_user_size, (!check_role_as_conference(user))? -1 : 0);

		if (conf->host_conf) {
			ast_atomic_fetchadd_int(&conf->host_conf->remote_size, -1);
			ast_atomic_fetchadd_int(&conf->host_conf->remote_user_size, (!check_role_as_conference(user))? -1 : 0);
		}
		ast_atomic_fetchadd_int(&conf->refcount, -1);
		ao2_ref(user, -2);
		user = next_user;
	}
	ao2_iterator_destroy(&user_iter);
	ao2_unlock(conf->user_container);

}

static struct cei_conference* get_remote_conf_by_id(struct cei_conference *host_conf, const char *remote_id)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);

	if (host_conf) {
		struct cei_conference *remote_conf = NULL;

		ao2_lock(host_conf->remote_container);
		struct ao2_iterator conf_iter = ao2_iterator_init(host_conf->remote_container, 0);
		while ((remote_conf = ao2_iterator_next(&conf_iter))) {
			if (!strcmp(remote_id, remote_conf->conference_id)) {
				ao2_ref(remote_conf, -1);
				break;
			}
			ao2_ref(remote_conf, -1);
		}
		ao2_iterator_destroy(&conf_iter);
		ao2_unlock(host_conf->remote_container);

		return remote_conf;
	}

	return NULL;
}

static struct cei_conference* build_remote_conf(struct cei_conference *host_conf, const char *remote_no, const char *remote_id, const char *publisher)
{
	struct cei_conference *conf = NULL;
	ast_debug(LOG_DEBUG, "[WL]---> LOOKING FOR CONF NUM %s ID %s", remote_no, remote_id);

	ao2_lock(host_conf->remote_container);

	conf = get_remote_conf_by_id(host_conf, remote_id);

	if (conf) {
		ast_debug(LOG_DEBUG, "[WL]---> CONFERENCE %s %s ALREADY EXISTS\n", remote_no, remote_id);
	}
	else {
		if (!(conf = ao2_alloc(sizeof(struct cei_conference), NULL)) ||
			!(conf->user_container = ao2_container_alloc(1, NULL, user_hash_cmp))) {
			if (conf) {
				ao2_cleanup(conf);
			}
			ast_debug(LOG_DEBUG, "[WL] CONFERENCE %s %s CANNOT BE CREATED\n", remote_no, remote_id);

		}
		else {
			ast_copy_string(conf->conference_num, remote_no, sizeof(conf->conference_num));
			ast_copy_string(conf->conference_id, remote_id, sizeof(conf->conference_id));
			ast_copy_string(conf->internal_publisher, publisher, sizeof(conf->internal_publisher));
			conf->locked = FALSE;
			conf->local_size = 0;
			conf->remote_size = 0;
			conf->local_user_size = 0;
			conf->remote_user_size = 0;
			conf->mcb_count = 0;
			conf->refcount = 0;
			conf->host_conf = host_conf;
			conf->version = 0;
			conf->version_reset = 0;
			ast_atomic_fetchadd_int(&conf->refcount, 1);

			ao2_link(host_conf->remote_container, conf);

			ast_debug(LOG_DEBUG, "[WL]---> REMOTE CONFERENCE %s %s IN ROOM %s CREATED\n",
														conf->conference_num,
														conf->conference_id,
														host_conf->conference_num);
		}
	}

	ao2_unlock(host_conf->remote_container);

	return conf;
}

static void process_remote_xml(struct cei_conference *conf, const char *publisher_private_num, const char *xml, struct ast_str **data_string)
{
	char publisher[41] = {0};
	char *xml_content = NULL;
	xml_content = ast_strdup(xml);
	if (!xml_content) {
		return;
	}
	else {
		struct ast_xml_doc *xml_doc = NULL;
		if (!(xml_doc = ast_xml_read_memory(xml_content, strlen(xml_content))))
			ast_log(LOG_DEBUG, "ZAKU DEBUG ---> FAILED TO READ XML DOC");
		else {
			ast_log(LOG_DEBUG, "ZAKU DEBUG ---> READ THE XML DOC CORRECTLY");
			struct ast_xml_node *conference_node = ast_xml_get_root(xml_doc);
			struct ast_xml_node *child_nodes;
			struct ast_xml_node *node_iterator;
			struct ast_xml_ns *ns;
			const char *entity;
			const char *notifyType;
			const char *state;
			const char *namespace;
			const char *node_name;


			if (!conference_node) {
				ast_log(LOG_DEBUG, "ZAKU DEBUG ---> NO ROOT NODE");
			}
			else {
				/*********
				Parse data from xml
				**********/
				char remote_num[MAX_CONFNUM];
				struct cei_conference *remote_conference = NULL;
				struct cei_user *user = NULL;
				struct cei_user *user_tmp = NULL;
				struct ast_xml_node *node_publisher = NULL;
				struct ast_xml_node *node_host_info = NULL;
				struct ast_xml_node *node_conference_state = NULL;
				struct ast_xml_node *node_users = NULL;
				struct ast_xml_node *node_sidebars = NULL;

				notifyType = ast_xml_get_attribute(conference_node, "state");

				if (strcmp(notifyType,"full") == 0) {
					entity = ast_xml_get_attribute(conference_node, "entity");
					char temp_entity[strlen(entity) + 1];
					strncpy(temp_entity, entity, strlen(entity));
					char *colon = strchr(temp_entity, ':');
					char *at = strchr(temp_entity, '@');
					*at = '\0';
					strncpy(remote_num, colon + 1, sizeof(remote_num));
					ast_xml_free_attr(entity);

					ns = ast_xml_find_namespace(xml_doc, conference_node, NULL);
					if (ns) {
						namespace = ast_xml_get_ns_href(ns);
						ast_log(LOG_DEBUG, "ZAKU DEBUG ---> NAMESPACE IS %s", namespace);
					}

					child_nodes = ast_xml_node_get_children(conference_node);
					if (child_nodes) {
						ast_log(LOG_DEBUG, "ZAKU DEBUG ---> HAS CHILDREN");

						for (node_iterator = child_nodes; node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
							const char *node_name = ast_xml_node_get_name(node_iterator);
							if (check_xml_text_node(node_iterator)) {
								ast_log(LOG_DEBUG, "ZAKU DEBUG ---> HAS CHILD NODE %s", node_name);
								if (!strcmp(node_name, "publisher") && !node_publisher)
									node_publisher = node_iterator;
								else if (!strcmp(node_name, "host-info") && !node_host_info)
									node_host_info = node_iterator;
								else if (!strcmp(node_name, "conference-state") && !node_conference_state)
									node_conference_state = node_iterator;
								else if (!strcmp(node_name, "users") && !node_users)
									node_users = node_iterator;
								else if (!strcmp(node_name, "sidebars-by-val") && !node_sidebars) {
									node_sidebars = node_iterator;
								}

								const char *content = ast_xml_get_text(node_iterator);
								if (content) {
									ast_log(LOG_DEBUG, "ZAKU DEBUG ---> HAS CHILD NODE CONTENT %s", content);
									ast_xml_free_text(content);
								}
								else
									ast_log(LOG_DEBUG, "ZAKU DEBUG ---> HAS CHILD NODE CONTENT NULL");

							}
							else
								ast_log(LOG_DEBUG, "ZAKU DEBUG ---> IGNORING BLANK TEXT NODE");
						}
					}

					if (node_publisher) {
						const char *content = ast_xml_get_text(node_publisher);

						if (content) {
							ast_copy_string(publisher, content, sizeof(publisher));
							ast_debug(LOG_DEBUG, "[TK] PUBLISHER IS %s", publisher);
						}

						ast_xml_free_text(content);
					}

					if (node_host_info) {
						const char *content = ast_xml_get_text(node_host_info);
						if (content) {
							char remote_id[strlen(content) + 1];
							strncpy(remote_id, content, sizeof(remote_id));
							ast_xml_free_text(content);
							char *strip_remote_id = ast_strip(remote_id);

							if (strlen(publisher) == 0) {
								ast_copy_string(publisher, strip_remote_id, sizeof(publisher));
								ast_debug(LOG_DEBUG, "[TK] PUBLISHER IS %s", publisher);
							}

							// clean up the remote conf published from publisher
							if(!conf){
								ast_debug(LOG_DEBUG, "[TK] no host conf !!!");
							}

							// Reinitialize published conferences
							delete_remote_confs_by_publisher(conf, strip_remote_id);

							ao2_lock(conf->remote_container);
							/***
							get or create the remote conf from host conf's remote container;
							the remote conf is basically this cei msg come FROM
							***/
							remote_conference = build_remote_conf(conf, publisher_private_num, strip_remote_id, publisher);
							ao2_unlock(conf->remote_container);
						}
					}

					int accept = TRUE;
					if (remote_conference) {
						const char *version = ast_xml_get_attribute(conference_node, "version");
						if (!version) {
							ast_xml_free_attr(version);
							ast_log(LOG_DEBUG, "ZAKU DEBUG ---> DOCUMENT HAS NO VERSION");
							accept = FALSE;
						}
						else {
							int xml_version = atoi(version);

							if (xml_version <= remote_conference->version) {
								ast_xml_free_attr(version);
								ast_log(LOG_DEBUG, "ZAKU DEBUG ---> DOCUMENT VERSION %d <= %d", xml_version, remote_conference->version);
								accept = FALSE;
							}
							else {
								ast_xml_free_attr(version);
								remote_conference->version = xml_version;
							}
						}

						ao2_lock(remote_conference->user_container);

						if (accept) {
							clear_conf_users(remote_conference);

							// Get remote conference-state items
							if (node_conference_state) {
								for (node_iterator = ast_xml_node_get_children(node_conference_state); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
									if (check_xml_text_node(node_iterator)) {
										const char *node_name = ast_xml_node_get_name(node_iterator);
										const char *content = ast_xml_get_text(node_iterator);
										ast_log(LOG_DEBUG, "ZAKU DEBUG ---> CONFERENCE STATE CHILD %s", node_name);
										if (content) {
											if (!strcmp(node_name, "user-count")) {
												char strip_content[strlen(content) + 1];
												strncpy(strip_content, content, strlen(content));
												ast_log(LOG_DEBUG, "ZAKU DEBUG ---> REMOTE LOCAL SIZE %d", remote_conference->local_user_size);
											}
											if (!strcmp(node_name, "locked")) {
												char strip_content[strlen(content) + 1];
												strncpy(strip_content, content, strlen(content));
												if (!strcmp(strip_content, "true"))
													remote_conference->locked = TRUE;
												else
													remote_conference->locked = FALSE;
												ast_log(LOG_DEBUG, "ZAKU DEBUG ---> REMOTE LOCKED %s", remote_conference->locked ? "TRUE" : "FALSE");
											}
										}
										ast_xml_free_text(content);
									}
								}
							}

							// Add users for remote publisher
							if (node_users) {
								process_remote_xml_users(conf, remote_conference,
														node_users);
							}

							// Add remote conference by sidebars-by-val
							if (node_sidebars) {
								// state
								accept = FALSE;
								state = ast_xml_get_attribute(node_sidebars, "state");
								if (!state || strcmp(state, "full") == 0) {
									accept = TRUE;
								}
								ast_xml_free_attr(state);

								if (accept) {
									for (node_iterator = ast_xml_node_get_children(node_sidebars); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
										const char *node_name = ast_xml_node_get_name(node_iterator);
										ast_log(LOG_DEBUG, "ZAKU DEBUG ---> sidebars-by-val CHILD %s", node_name);

										if (check_xml_text_node(node_iterator) == 0) {
											continue;
										}

										if (strcmp(node_name, "entry") == 0) {
											char *head_pointer;
											char *split_pointer;
											struct ast_xml_node *entry_child;

											// state
											state = ast_xml_get_attribute(node_iterator, "state");
											if (!state || strcmp(state, "full") != 0) {
												ast_xml_free_attr(state);
												continue;
											}
											ast_xml_free_attr(state);

											// entity conf info
											entity = ast_xml_get_attribute(node_iterator, "entity");
											char entry_entity[strlen(entity) + 1];
											strncpy(entry_entity, entity, sizeof(entry_entity));
											ast_xml_free_attr(entity);
											ast_log(LOG_DEBUG, "[TK] entry entity: %s", entry_entity);

											head_pointer = entry_entity;
											split_pointer = strchr(entry_entity, '-');
											if (split_pointer == NULL) {
												continue;
											}
											*split_pointer = '\0';
											// remote_num is the remote conf num  in sider bar
											strncpy(remote_num, head_pointer, sizeof(remote_num));

											head_pointer = strchr(split_pointer + 1, '@');
											if (head_pointer == NULL) {
												continue;
											}

											head_pointer = head_pointer + 1;
											split_pointer = strchr(head_pointer, '.');
											if (split_pointer == NULL) {
												continue;
											}
											*split_pointer = '\0';
											// remode_id is the remote conf id in sider bar
											char remote_id[strlen(head_pointer) + 1];
											strncpy(remote_id, head_pointer, sizeof(remote_id));
											char *strip_remote_id = ast_strip(remote_id);
											ast_log(LOG_DEBUG, "[TK] process remote container conf of this xml : %s", strip_remote_id);

											// Ignore invalid conference
											// Don't add myself as my remote conf
											if (strcmp(conf->conference_id, strip_remote_id) == 0) {
												ast_debug(LOG_DEBUG, "[TK] Ignore my own conference");
												ast_debug(LOG_DEBUG, "[TK] %s@%s", remote_num, strip_remote_id);
												continue;
											}
											// Don't add the remote conf which are my own remote conf again
											if (get_remote_conf_by_id(conf, strip_remote_id) != NULL) {
												ast_debug(LOG_DEBUG, "[TK] Ignore my own remote conference");
												ast_debug(LOG_DEBUG, "[TK] %s@%s", remote_num, strip_remote_id);
												continue;
											}

											// Get users
											node_users = NULL;
											for(entry_child = ast_xml_node_get_children(node_iterator); entry_child != NULL; entry_child = ast_xml_node_get_next(entry_child)) {
												node_name = ast_xml_node_get_name(entry_child);
												if (strcmp(node_name, "users") == 0) {
													node_users = entry_child;
													break;
												}
											}

											if (node_users == NULL) {
												ast_debug(LOG_DEBUG, "[TK] NO USERS");
												continue;
											}

											ast_log(LOG_DEBUG, "[TK] NEW REMOTE CONF %s@%s", remote_num, strip_remote_id);

											// Switch to new remote conference
											// Handle current remote conference
											ao2_unlock(remote_conference->user_container);

											// Handle new remote conference
											remote_conference = build_remote_conf(conf, remote_num, strip_remote_id, publisher);

											ao2_lock(remote_conference->user_container);
											clear_conf_users(remote_conference);

											process_remote_xml_users(conf, remote_conference,
																	node_users);
										}
									}
								}
							}
						}

						ao2_unlock(remote_conference->user_container);
					}
					poll_cei_conference(conf, data_string);
				}
				// PARTIAL NOTIFY
				else if (strcmp(notifyType,"partial") == 0) {
					entity = ast_xml_get_attribute(conference_node, "entity");
					char temp_entity[strlen(entity) + 1];
					strncpy(temp_entity, entity, strlen(entity));
					char *colon = strchr(temp_entity, ':');
					char *at = strchr(temp_entity, '@');
					*at = '\0';
					strncpy(remote_num, colon + 1, sizeof(remote_num));
					ast_xml_free_attr(entity);

					ns = ast_xml_find_namespace(xml_doc, conference_node, NULL);
					if (ns) {
						namespace = ast_xml_get_ns_href(ns);
						ast_log(LOG_DEBUG, "ZAKU DEBUG ---> NAMESPACE IS %s", namespace);
					}

					child_nodes = ast_xml_node_get_children(conference_node);
					if (child_nodes) {
						ast_log(LOG_DEBUG, "ZAKU DEBUG ---> HAS CHILDREN");

						for (node_iterator = child_nodes; node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
							const char *node_name = ast_xml_node_get_name(node_iterator);
							if (check_xml_text_node(node_iterator)) {
								ast_log(LOG_DEBUG, "ZAKU DEBUG ---> HAS CHILD NODE %s", node_name);
								if (!strcmp(node_name, "publisher") && !node_publisher)
									node_publisher = node_iterator;
								else if (!strcmp(node_name, "host-info") && !node_host_info)
									node_host_info = node_iterator;
								else if (!strcmp(node_name, "conference-state") && !node_conference_state)
									node_conference_state = node_iterator;
								else if (!strcmp(node_name, "users") && !node_users)
									node_users = node_iterator;
								else if (!strcmp(node_name, "sidebars-by-val") && !node_sidebars) {
									node_sidebars = node_iterator;
								}

								const char *content = ast_xml_get_text(node_iterator);
								if (content) {
									ast_log(LOG_DEBUG, "ZAKU DEBUG ---> HAS CHILD NODE CONTENT %s", content);
									ast_xml_free_text(content);
								}
								else
									ast_log(LOG_DEBUG, "ZAKU DEBUG ---> HAS CHILD NODE CONTENT NULL");

							}
							else
								ast_log(LOG_DEBUG, "ZAKU DEBUG ---> IGNORING BLANK TEXT NODE");
						}
					}

					if (node_publisher) {
						const char *content = ast_xml_get_text(node_publisher);

						if (content) {
							ast_copy_string(publisher, content, sizeof(publisher));
							ast_debug(LOG_DEBUG, "[TK] PUBLISHER IS %s", publisher);
						}

						ast_xml_free_text(content);
					}

					if (node_host_info) {
						const char *content = ast_xml_get_text(node_host_info);
						if (content) {
							char remote_id[strlen(content) + 1];
							strncpy(remote_id, content, sizeof(remote_id));
							ast_xml_free_text(content);
							char *strip_remote_id = ast_strip(remote_id);

							if (strlen(publisher) == 0) {
								ast_copy_string(publisher, strip_remote_id, sizeof(publisher));
								ast_debug(LOG_DEBUG, "[TK] PUBLISHER IS %s", publisher);
							}

							// clean up the remote conf published from publisher
							if(!conf){
								ast_debug(LOG_DEBUG, "[TK] no host conf !!!");
							}

							/***
							get the remote conf from host conf's remote container;
							the remote conf is basically this cei msg come FROM
							***/
							remote_conference = get_remote_conf_by_id(conf, publisher);
						}
					}

					int accept = TRUE;
					if (remote_conference) {
						const char *version = ast_xml_get_attribute(conference_node, "version");
						if (!version) {
							ast_xml_free_attr(version);
							ast_log(LOG_DEBUG, "ZAKU DEBUG ---> DOCUMENT HAS NO VERSION");
							accept = FALSE;
						}
						else {
							int xml_version = atoi(version);

							if (xml_version <= remote_conference->version) {
								ast_xml_free_attr(version);
								ast_log(LOG_DEBUG, "ZAKU DEBUG ---> DOCUMENT VERSION %d <= %d", xml_version, remote_conference->version);
								accept = FALSE;
							}
							else {
								ast_xml_free_attr(version);
								remote_conference->version = xml_version;
							}
						}


						if (accept) {

							// Get remote conference-state items
							if (node_conference_state) {
								for (node_iterator = ast_xml_node_get_children(node_conference_state); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
									if (check_xml_text_node(node_iterator)) {
										const char *node_name = ast_xml_node_get_name(node_iterator);
										const char *content = ast_xml_get_text(node_iterator);
										ast_log(LOG_DEBUG, "ZAKU DEBUG ---> CONFERENCE STATE CHILD %s", node_name);
										if (content) {
											if (!strcmp(node_name, "user-count")) {
												char strip_content[strlen(content) + 1];
												strncpy(strip_content, content, strlen(content));
												ast_log(LOG_DEBUG, "ZAKU DEBUG ---> REMOTE LOCAL SIZE %d", remote_conference->local_user_size);
											}
											if (!strcmp(node_name, "locked")) {
												char strip_content[strlen(content) + 1];
												strncpy(strip_content, content, strlen(content));
												if (!strcmp(strip_content, "true"))
													remote_conference->locked = TRUE;
												else
													remote_conference->locked = FALSE;
												ast_log(LOG_DEBUG, "ZAKU DEBUG ---> REMOTE LOCKED %s", remote_conference->locked ? "TRUE" : "FALSE");
											}
										}
										ast_xml_free_text(content);
									}
								}
							}

							// Add users for remote publisher
							if (node_users) {
								for (node_iterator = ast_xml_node_get_children(node_users); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
									if (check_xml_text_node(node_iterator) == 0) {
										continue;
									}
									node_name = ast_xml_node_get_name(node_iterator);
									if (!strcmp(node_name, "user")) {

										user = process_remote_xml_user(conf, node_iterator);
										if (user) {
											switch (user->user_status) {
												case CEI_STATUS_CONNECTED:
												{
													ao2_lock(remote_conference->user_container);
													//CEI_EVENT_MCB_USER_TOGGLE_MUTE unmute
													user_tmp = get_user_by_cid_num(remote_conference, user->cid_number);
													if (user_tmp) {
														user_tmp->muted = 0;
														poll_cei_partial_notify(conf, remote_conference, user_tmp, data_string, CEI_EVENT_MCB_USER_TOGGLE_MUTE);
														ao2_ref(user, -1);
													}
													else {
														hash_user(user);
														ast_log(LOG_DEBUG, "ZAKU DEBUG ---> add user : %s to remote conf user container", user->channel_name);
														ao2_link(remote_conference->user_container, user);
														ast_atomic_fetchadd_int(&remote_conference->local_size, 1);
														ast_atomic_fetchadd_int(&remote_conference->local_user_size, (!check_role_as_conference(user))? 1 : 0);
														ast_atomic_fetchadd_int(&conf->remote_size, 1);
														ast_atomic_fetchadd_int(&conf->remote_user_size, (!check_role_as_conference(user))? 1 : 0);
														ast_atomic_fetchadd_int(&remote_conference->refcount, 1);
														poll_cei_partial_notify(conf, remote_conference, user, data_string, CEI_EVENT_MCB_USER_JOIN);
													}
													ao2_unlock(remote_conference->user_container);
												}
												break;
												case CEI_STATUS_DISCONNECTED:
												{
													ao2_lock(remote_conference->user_container);

													user_tmp = get_user_by_cid_num(remote_conference, user->cid_number);

													if (user_tmp) {

														ast_atomic_fetchadd_int(&remote_conference->local_size, -1);
														ast_atomic_fetchadd_int(&remote_conference->local_user_size, (!check_role_as_conference(user))? -1 : 0);
														ast_atomic_fetchadd_int(&conf->remote_size, -1);
														ast_atomic_fetchadd_int(&conf->remote_user_size, (!check_role_as_conference(user))? -1 : 0);

														poll_cei_partial_notify(conf, remote_conference, user, data_string, CEI_EVENT_MCB_USER_PART);

														ao2_unlink(remote_conference->user_container, user_tmp);
														ast_atomic_fetchadd_int(&remote_conference->refcount, -1);

														ast_debug(LOG_DEBUG, "[WL] REMOVED USER WITH ADDRESS %p", user_tmp);
														ao2_ref(user_tmp, -1);
													}
													else {
														ast_debug(LOG_DEBUG, "[WL] ERROR, CANNOT FIND USER  %s IN CONFERENCE %s", user->cid_number, remote_conference->conference_num);
													}
													ao2_ref(user, -1);
													ao2_unlock(remote_conference->user_container);
												}
												break;
												case CEI_STATUS_MUTEVIAFOCUS:
												{
													ao2_lock(remote_conference->user_container);

													user_tmp = get_user_by_cid_num(remote_conference, user->cid_number);

													if (user_tmp) {
														user_tmp->muted = 1;
														poll_cei_partial_notify(conf, remote_conference, user_tmp, data_string, CEI_EVENT_MCB_USER_TOGGLE_MUTE);
													}
													ao2_ref(user, -1);
													ao2_unlock(remote_conference->user_container);
												}
												break;
												default:
												{

												}
												break;
											}
										}
									}
								}
							}

							// Add remote conference by sidebars-by-val
							if (node_sidebars) {
								// state
								accept = FALSE;
								state = ast_xml_get_attribute(node_sidebars, "state");
								if (!state || strcmp(state, "partial") == 0) {
									accept = TRUE;
								}
								ast_xml_free_attr(state);

								if (accept) {
									for (node_iterator = ast_xml_node_get_children(node_sidebars); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
										const char *node_name = ast_xml_node_get_name(node_iterator);
										ast_log(LOG_DEBUG, "ZAKU DEBUG ---> sidebars-by-val CHILD %s", node_name);

										if (check_xml_text_node(node_iterator) == 0) {
											continue;
										}

										if (strcmp(node_name, "entry") == 0) {
											char *head_pointer;
											char *split_pointer;
											struct ast_xml_node *node_remote_users;
											struct ast_xml_node *node_remote_child_user;

											// state
											state = ast_xml_get_attribute(node_iterator, "state");

											// entity conf info
											entity = ast_xml_get_attribute(node_iterator, "entity");
											char entry_entity[strlen(entity) + 1];
											strncpy(entry_entity, entity, sizeof(entry_entity));
											ast_xml_free_attr(entity);
											ast_log(LOG_DEBUG, "[TK] entry entity: %s", entry_entity);

											head_pointer = entry_entity;
											split_pointer = strchr(entry_entity, '-');
											if (split_pointer == NULL) {
												continue;
											}
											*split_pointer = '\0';
											// remote_num is the remote conf num  in sider bar
											strncpy(remote_num, head_pointer, sizeof(remote_num));

											head_pointer = strchr(split_pointer + 1, '@');
											if (head_pointer == NULL) {
												continue;
											}

											head_pointer = head_pointer + 1;
											split_pointer = strchr(head_pointer, '.');
											if (split_pointer == NULL) {
												continue;
											}
											*split_pointer = '\0';
											// remode_id is the remote conf id in sider bar
											char remote_id[strlen(head_pointer) + 1];
											strncpy(remote_id, head_pointer, sizeof(remote_id));
											char *strip_remote_id = ast_strip(remote_id);
											ast_log(LOG_DEBUG, "[TK] process remote container conf of this xml : %s", strip_remote_id);

											// Ignore invalid conference
											// Don't add myself as my remote conf
											if (strcmp(conf->conference_id, strip_remote_id) == 0) {
												ast_debug(LOG_DEBUG, "[TK] Ignore my own conference");
												ast_debug(LOG_DEBUG, "[TK] %s@%s", remote_num, strip_remote_id);
												continue;
											}

											remote_conference = get_remote_conf_by_id(conf, strip_remote_id);

											if (strcmp(state, "partial") == 0) {
												if (remote_conference != NULL) {
													// Get users
													for(node_remote_users = ast_xml_node_get_children(node_iterator); node_remote_users != NULL; node_remote_users = ast_xml_node_get_next(node_remote_users)) {
														node_name = ast_xml_node_get_name(node_remote_users);
														if (strcmp(node_name, "users") == 0) {
															break;
														}
													}

													if (node_remote_users == NULL) {
														ast_debug(LOG_DEBUG, "[TK] NO USERS");
														continue;
													}

													ast_log(LOG_DEBUG, "[TK] NEW REMOTE CONF %s@%s", remote_num, strip_remote_id);

													// Switch to new remote conference
													// Handle current remote conference

													for (node_remote_child_user = ast_xml_node_get_children(node_remote_users); node_remote_child_user; node_remote_child_user = ast_xml_node_get_next(node_remote_child_user)) {
														if (check_xml_text_node(node_remote_child_user) == 0) {
															continue;
														}
														node_name = ast_xml_node_get_name(node_remote_child_user);
														if (!strcmp(node_name, "user")) {
															user = process_remote_xml_user(conf, node_remote_child_user);
															if (user) {
																switch (user->user_status) {
																	case CEI_STATUS_CONNECTED:
																	{
																		ao2_lock(remote_conference->user_container);
																		user_tmp = get_user_by_cid_num(remote_conference, user->cid_number);
																		//CEI_EVENT_MCB_USER_TOGGLE_MUTE unmute
																		if (user_tmp) {
																			user_tmp->muted = 0;
																			poll_cei_partial_notify(conf, remote_conference, user_tmp, data_string, CEI_EVENT_MCB_USER_TOGGLE_MUTE);
																			ao2_ref(user, -1);
																		}
																		//CEI_EVENT_MCB_USER_JOIN
																		else {
																			hash_user(user);
																			ast_log(LOG_DEBUG, "ZAKU DEBUG ---> add user : %s to remote conf user container", user->channel_name);
																			ao2_link(remote_conference->user_container, user);
																			ast_atomic_fetchadd_int(&remote_conference->local_size, 1);
																			ast_atomic_fetchadd_int(&remote_conference->local_user_size, (!check_role_as_conference(user))? 1 : 0);
																			ast_atomic_fetchadd_int(&conf->remote_size, 1);
																			ast_atomic_fetchadd_int(&conf->remote_user_size, (!check_role_as_conference(user))? 1 : 0);
																			ast_atomic_fetchadd_int(&remote_conference->refcount, 1);
																			poll_cei_partial_notify(conf, remote_conference, user, data_string, CEI_EVENT_MCB_USER_JOIN);
																		}
																		ao2_unlock(remote_conference->user_container);
																	}
																	break;
																	case CEI_STATUS_DISCONNECTED:
																	{
																		ao2_lock(remote_conference->user_container);

																		user_tmp = get_user_by_cid_num(remote_conference, user->cid_number);

																		if (user_tmp) {

																			ast_atomic_fetchadd_int(&remote_conference->local_size, -1);
																			ast_atomic_fetchadd_int(&remote_conference->local_user_size, (!check_role_as_conference(user))? -1 : 0);
																			ast_atomic_fetchadd_int(&conf->remote_size, -1);
																			ast_atomic_fetchadd_int(&conf->remote_user_size, (!check_role_as_conference(user))? -1 : 0);

																			poll_cei_partial_notify(conf, remote_conference, user, data_string, CEI_EVENT_MCB_USER_PART);

																			ao2_unlink(remote_conference->user_container, user_tmp);
																			ast_atomic_fetchadd_int(&remote_conference->refcount, -1);

																			ao2_ref(user_tmp, -1);
																		}
																		else {
																			ast_debug(LOG_DEBUG, "[WL] ERROR, CANNOT FIND USER  %s IN CONFERENCE %s", user->cid_number, remote_conference->conference_num);
																		}
																		ao2_ref(user, -1);
																		ao2_unlock(remote_conference->user_container);
																	}
																	break;
																	case CEI_STATUS_MUTEVIAFOCUS:
																	{
																		ao2_lock(remote_conference->user_container);

																		user_tmp = get_user_by_cid_num(remote_conference, user->cid_number);

																		if (user_tmp) {
																			user_tmp->muted = 1;
																			poll_cei_partial_notify(conf, remote_conference, user_tmp, data_string, CEI_EVENT_MCB_USER_TOGGLE_MUTE);
																		}
																		ao2_ref(user, -1);
																		ao2_unlock(remote_conference->user_container);
																	}
																	break;
																	default:
																	{
																		ast_debug(LOG_DEBUG, "[WL] INVALID STATUS");
																	}
																	break;
																}
															}
														}
													}
												}

											}
											else if(strcmp(state, "deleted") == 0) {
												if (remote_conference !=  NULL) {
													delete_remote_conf_by_id(conf, publisher);
													poll_cei_conference(conf, data_string);
												}
											}
											else {
												ast_debug(LOG_DEBUG, "[WL] INVALID STATE");
											}
											ast_xml_free_attr(state);
										}
									}
								}
							}
						}

					}
				}
				ast_xml_free_attr(notifyType);
			}
		}
		ast_xml_close(xml_doc);
	}
	ast_free(xml_content);
	ast_debug(LOG_DEBUG, "ZAKU DEBUG ---> END TEST XML\n");
}

static int check_xml_text_node(struct ast_xml_node *node)
{
	return strcmp(ast_xml_node_get_name(node), "text") ? 1 : 0;
}

static void process_remote_xml_users(struct cei_conference *conf,
									 struct cei_conference *remote_conference,
									 struct ast_xml_node *node_users)
{
	const char *node_name;
	struct cei_user *user;
	struct ast_xml_node *node_iterator = NULL;

	// Sanity check
	if (conf == NULL || remote_conference == NULL || node_users == NULL) {
		return;
	}

	for (node_iterator = ast_xml_node_get_children(node_users); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
		if (check_xml_text_node(node_iterator) == 0) {
			continue;
		}

		node_name = ast_xml_node_get_name(node_iterator);
		ast_log(LOG_DEBUG, "ZAKU DEBUG ---> USERS CHILD %s", node_name);

		if (!strcmp(node_name, "user")) {
			user = process_remote_xml_user(conf, node_iterator);
			if (user == NULL) {
				continue;
			}
			ast_log(LOG_DEBUG, "ZAKU DEBUG ---> add user : %s to remote conf user container", user->channel_name);
			ao2_link(remote_conference->user_container, user);
			ast_atomic_fetchadd_int(&remote_conference->local_size, 1);
			ast_atomic_fetchadd_int(&remote_conference->local_user_size, (!check_role_as_conference(user))? 1 : 0);
			ast_atomic_fetchadd_int(&conf->remote_size, 1);
			ast_atomic_fetchadd_int(&conf->remote_user_size, (!check_role_as_conference(user))? 1 : 0);
			ast_atomic_fetchadd_int(&remote_conference->refcount, 1);
		}
	}
}

static struct cei_user *process_remote_xml_user(struct cei_conference *host_conf,
	                                                struct ast_xml_node *node_iterator)
{
	char *head_pointer;
	char *split_pointer;
	const char *content;
	const char *self_conf_id;
	const char *entity;
	const char *state;
	struct cei_user *user;
	struct ast_xml_node *node_display_text = NULL;
	struct ast_xml_node *node_aors = NULL;
	struct ast_xml_node *node_roles = NULL;
	struct ast_xml_node *node_endpoint = NULL;
	struct ast_xml_node *user_inner_iterator = NULL;
	struct ast_xml_node *node_role_iterator = NULL;

	// Sanity check
	if (node_iterator == NULL || host_conf == NULL) {
		return NULL;
	}

	if (!(user = ao2_alloc(sizeof(struct cei_user), NULL))) {
		ast_debug(LOG_DEBUG, "ZAKU DEBUG ---> CANNOT CREATE NEW USER");
		return NULL;
	}

	state = ast_xml_get_attribute(node_iterator, "state");

	if (!state || strcmp(state, "full")) {
		ast_xml_free_attr(state);
		ast_log(LOG_DEBUG, "ZAKU DEBUG ---> USER STATE NOT full");
		ao2_ref(user, -1);
		return NULL;
	}
	ast_xml_free_attr(state);

	entity = ast_xml_get_attribute(node_iterator, "entity");
	if (entity == NULL) {
		ao2_ref(user, -1);
		return NULL;
	}

	char user_entity[strlen(entity) + 1];
	strncpy(user_entity, entity, sizeof(user_entity));
	ast_xml_free_attr(entity);
	head_pointer = user_entity;
	split_pointer = strchr(user_entity, '-');
	if (!split_pointer) {
		ast_log(LOG_DEBUG, "ZAKU DEBUG ---> USER ENTITY SPLIT 1");
		ao2_ref(user, -1);
		return NULL;
	}
	*split_pointer = '\0';
	strncpy(user->cid_number, head_pointer, sizeof(user->cid_number));

	head_pointer = split_pointer + 1;
	split_pointer = strchr(head_pointer, '@');
	if (!split_pointer) {
		ast_log(LOG_DEBUG, "ZAKU DEBUG ---> USER ENTITY SPLIT 2");
		ao2_ref(user, -1);
		return NULL;
	}
	*split_pointer = '\0';
	user->user_num = atoi(head_pointer);

	head_pointer = split_pointer + 1;
	split_pointer = strchr(head_pointer, '.');
	if (!split_pointer) {
		ast_log(LOG_DEBUG, "ZAKU DEBUG ---> USER ENTITY SPLIT 3");
		ao2_ref(user, -1);
		return NULL;
	}
	*split_pointer = '\0';
	strncpy(user->conference_id, head_pointer, sizeof(user->conference_id));

	ast_log(LOG_DEBUG, "ZAKU DEBUG ---> GOT USER CID NUM, CONFERENCE ID, USER NUM");

	// Remove local users in host conf
	if (strcmp(user->conference_id, host_conf->conference_id) == 0) {
		ast_log(LOG_DEBUG, "[TK] Ignore local conference users");
		ast_log(LOG_DEBUG, "[TK] %s-%d@%s",
		user->cid_number, user->user_num, user->conference_id);
		ao2_ref(user, -1);
		return NULL;
	}

	node_display_text = NULL;
	node_aors = NULL;
	node_roles = NULL;
	node_endpoint = NULL;

	for (user_inner_iterator = ast_xml_node_get_children(node_iterator); user_inner_iterator; user_inner_iterator = ast_xml_node_get_next(user_inner_iterator)) {
		if (check_xml_text_node(user_inner_iterator)) {
			const char *user_inner_node_name = ast_xml_node_get_name(user_inner_iterator);
			ast_log(LOG_DEBUG, "ZAKU DEBUG ---> USER CHILD %s", user_inner_node_name);
			if (!strcmp(user_inner_node_name, "display-text") && !node_display_text) {
				node_display_text = user_inner_iterator;
			}
			else if (!strcmp(user_inner_node_name, "associated-aors") && !node_aors) {
				node_aors = user_inner_iterator;
			}
			else if (!strcmp(user_inner_node_name, "roles") && !node_roles) {
				node_roles = user_inner_iterator;
			}
			else if (!strcmp(user_inner_node_name, "endpoint") && !node_endpoint) {
				node_endpoint = user_inner_iterator;
			}
		}
	}

	if (!node_display_text || !node_roles || !node_endpoint ) {
		ast_log(LOG_DEBUG, "ZAKU DEBUG ---> USER MISSING DISPLAY/ROLES/ENDPOINT");
		ao2_ref(user, -1);
		return NULL;
	}

	//<display-text>
	content = ast_xml_get_text(node_display_text);
	if (content) {
		strncpy(user->cid_name, content, sizeof(user->cid_name));
		ast_xml_free_text(content);
	}
	else {
		ast_log(LOG_DEBUG, "ZAKU DEBUG ---> USER EMPTY DISPLAY");
		ao2_ref(user, -1);
		return NULL;
	}

	//<associated-aors> the self conf id if conference user
	if (node_aors) {

		self_conf_id = ast_xml_get_text(node_aors);
		ast_log(LOG_DEBUG, "[WL] Got self conf id : %s",self_conf_id);
		if (self_conf_id) {
			if (strcmp(host_conf->conference_id, self_conf_id) == 0) {
				memset(user->cid_number, 0, sizeof(user->cid_number));
				strncpy(user->cid_number, host_conf->conference_num, sizeof(user->cid_number));
			}
		}
	}

	//<roles>
	user->admin = 0;
	user->can_invite = 0;
	user->user_role = CEI_ROLE_UNKNOWN;
	for (node_role_iterator = ast_xml_node_get_children(node_roles); node_role_iterator; node_role_iterator = ast_xml_node_get_next(node_role_iterator)) {
		if (check_xml_text_node(node_role_iterator) == 0){
			continue;
		}

		if (!strcmp(ast_xml_node_get_name(node_role_iterator), "entry")) {
			content = ast_xml_get_text(node_role_iterator);
			if (content) {
				if (!strcmp(content, "admin")) {
					user->admin = 1;
					user->user_role = CEI_ROLE_ADMIN;
					break;
				}
				else if (!strcmp(content, "conference")) {
					user->user_role = CEI_ROLE_CONFERENCE;
					break;
				}
				else if (!strcmp(content, "invite")) {
					user->can_invite = 1;
					break;
				}
				else {
					user->user_role = CEI_ROLE_USER_NORMAL;
				}

				ast_xml_free_text(content);
			}
		}
	}

	if (user->can_invite && (user->user_role == CEI_ROLE_USER_NORMAL)) {
		user->user_role = CEI_ROLE_USER_INVITE;
	}

	//<endpoint>
	entity = ast_xml_get_attribute(node_endpoint, "entity");
	strncpy(user->channel_name, entity, sizeof(user->channel_name));
	ast_xml_free_attr(entity);

	state = ast_xml_get_attribute(node_endpoint, "state");
	if (!state || strcmp(state, "full")) {
		ast_xml_free_attr(state);
		ast_log(LOG_DEBUG, "ZAKU DEBUG ---> ENDPOINT STATE NOT full");
		ao2_ref(user, -1);
		return NULL;
	}
	ast_xml_free_attr(state);

	// Get user status
	user->muted = 0;
	user->user_status = CEI_STATUS_UNKNOWN;

	for (user_inner_iterator = ast_xml_node_get_children(node_endpoint); user_inner_iterator; user_inner_iterator = ast_xml_node_get_next(user_inner_iterator)) {
		const char *endpoint_node_name = ast_xml_node_get_name(user_inner_iterator);
		ast_log(LOG_DEBUG, "ZAKU DEBUG ---> ENDPOINT CHILD %s", endpoint_node_name);

		// [TK] TODO: replace muted with user_status
		const char *content = ast_xml_get_text(user_inner_iterator);
		if (content) {
			if (strstr(content, "disconnected")) {
				user->user_status = CEI_STATUS_DISCONNECTED;
			}
			else if (strstr(content, "connected")) {
				user->user_status = CEI_STATUS_CONNECTED;
			}
			else if (strstr(content, "on-hold")) {
				user->user_status = CEI_STATUS_ONHOLD;
			}
			else if (strstr(content, "muted-via-focus")) {
				user->muted = 1;
				user->user_status = CEI_STATUS_MUTEVIAFOCUS;
			}
			else if (strstr(content, "pending")) {
				user->user_status = CEI_STATUS_PENDING;
			}
			else if (strstr(content, "alerting")) {
				user->user_status = CEI_STATUS_ALERTING;
			}
			else if (strstr(content, "dialing-in")) {
				user->user_status = CEI_STATUS_DIALINGIN;
			}
			else if (strstr(content, "dialing-out")) {
				user->user_status = CEI_STATUS_DIALINGOUT;
			}
			else if (strstr(content, "disconnecting")) {
				user->user_status = CEI_STATUS_DISCONNECTING;
			}

			ast_xml_free_text(content);
		}
	}
	if (user->user_status == CEI_STATUS_UNKNOWN) {
		ast_log(LOG_DEBUG, "ZAKU DEBUG ---> USER EMPTY STATUS");
		ao2_ref(user, -1);
		return NULL;
	}
	return user;

}


static void delete_remote_confs_by_publisher(struct cei_conference *host_conf,
											 const char *publisher)
{
	ast_debug(LOG_DEBUG, "[WL] Deleting remote conf by publisher: %s", publisher);

	struct cei_conference *conf = NULL;
	struct ao2_iterator iter;

	if (!host_conf || !publisher) {
		return;
	}

	ao2_lock(host_conf->remote_container);
	iter = ao2_iterator_init(host_conf->remote_container, 0);
	while ((conf = ao2_iterator_next(&iter))) {
		if (strcmp(conf->internal_publisher, publisher) == 0) {
			ao2_unlink(host_conf->remote_container, conf);
			clear_conf_users(conf);
			ao2_ref(conf, -1);
		}
		ao2_ref(conf, -1);
	}
	ao2_iterator_destroy(&iter);
	ao2_unlock(host_conf->remote_container);
}

static void delete_remote_conf_by_id(struct cei_conference *host_conf,
											 const char *remote_conf_id)
{
	ast_debug(LOG_DEBUG, "[WL] Deleting remote conf by id: %s", remote_conf_id);

	struct cei_conference *conf = NULL;
	struct ao2_iterator iter;

	if (!host_conf || !remote_conf_id) {
		return;
	}

	ao2_lock(host_conf->remote_container);
	iter = ao2_iterator_init(host_conf->remote_container, 0);
	while ((conf = ao2_iterator_next(&iter))) {
		if (strcmp(conf->conference_id, remote_conf_id) == 0) {
			ao2_unlink(host_conf->remote_container, conf);
			clear_conf_users(conf);
			ao2_ref(conf, -1);
		}
		ao2_ref(conf, -1);
	}
	ao2_iterator_destroy(&iter);
	ao2_unlock(host_conf->remote_container);
}


static struct cei_conference* get_conf_by_id(const char *conf_id)
{
	if (conf_id) {
		struct cei_conference *conf = NULL;

		AST_LIST_LOCK(&confs);
		AST_LIST_TRAVERSE(&confs, conf, list) {
			if (!strcmp(conf_id, conf->conference_id)) {
				break;
			}
		}
		AST_LIST_UNLOCK(&confs);
		return conf;
	}
	return NULL;
}

static struct cei_conference* get_conf_by_num(const char *conf_num)
{
	if (conf_num) {
		struct cei_conference *conf = NULL;

		AST_LIST_LOCK(&confs);
		AST_LIST_TRAVERSE(&confs, conf, list) {
			if (!strcmp(conf_num, conf->conference_num)) {
				break;
			}
		}

		AST_LIST_UNLOCK(&confs);
		return conf;
	}
	return NULL;
}

static struct cei_user* get_user_by_cid_num(struct cei_conference *conf, const char *cid_num)
{
	if (conf == NULL || cid_num == NULL) {
		return NULL;
	}

	struct cei_user *user = NULL;

	ao2_lock(conf->user_container);
	struct ao2_iterator user_iter = ao2_iterator_init(conf->user_container, 0);
	while ((user = ao2_iterator_next(&user_iter))) {
		if (strcmp(cid_num,user->cid_number) == 0){
			ao2_ref(user, -1);
			break;
		}
		ao2_ref(user, -1);
	}
    ao2_iterator_destroy(&user_iter);
	ao2_unlock(conf->user_container);
	if (user) {
		return user;
	}
	return NULL;
}

// update struct for conf
static void process_remote_string(struct cei_conference *conf,
		const char *publisher_num, const char *publisher_id,
		char *content, struct ast_str **data_string)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);

	enum cei_notify_type notifyType = CEI_NOTIFY_UNKNOWN;
	int total_count = 0;
	int index = 0;
	char *saveptr = NULL;
	char *notifyPointer = NULL;
	char *local_count = NULL;
	char *remote_count = NULL;
	char *local_user_count = NULL;
	char *remote_user_count = NULL;
	char *total_update_count = NULL;
	char *conf_locked = NULL;
	char *subject = NULL;
	char *free_text = NULL;
	char *conf_start = NULL;
	char *conf_max_user = NULL;
	char *conference_type = NULL;
	char *conf_user_count = NULL;
	char *user_number = NULL;
	char *cid_number = NULL;
	char *cid_name = NULL;
	char *channel_name = NULL;
	char *conference_id = NULL;
	char *self_conference_id = NULL;
	char *user_role = NULL;
	char *entity_status = NULL;
	struct cei_conference *remote_conference = NULL;

	int flag_ignored_conf = FALSE;


	// Sanatity check
	if (content == NULL || conf == NULL || publisher_num == NULL || publisher_id == NULL) {
		ast_log(LOG_WARNING, "[TK] Invalid CEI data");
		return;
	}

	// Notify type
	notifyPointer = strtok_r(content, "|", &saveptr);
	if (notifyPointer == NULL) {
		ast_log(LOG_DEBUG, "[TK] Invalid notify type\n");
		return;
	}
	notifyType = atoi(notifyPointer);
	ast_log(LOG_DEBUG, "[TK] notifyType: %d\n", notifyType);

	local_count = strtok_r(NULL, "|", &saveptr);
	remote_count = strtok_r(NULL, "|", &saveptr);
	local_user_count = strtok_r(NULL, "|", &saveptr);
	remote_user_count = strtok_r(NULL, "|", &saveptr);
	total_update_count = strtok_r(NULL, "|", &saveptr);
	conf_locked = strtok_r(NULL, "|", &saveptr);
	subject = strtok_r(NULL, "|", &saveptr);
	free_text = strtok_r(NULL, "|", &saveptr);
	conf_start = strtok_r(NULL, "|", &saveptr);
	conf_max_user = strtok_r(NULL, "|", &saveptr);

	// Sanity check
	if (local_count == NULL || remote_count == NULL ||
		local_user_count == NULL || remote_user_count == NULL ||
		total_update_count == NULL || conf_locked == NULL ||
		subject == NULL || free_text == NULL ||
		conf_start == NULL || conf_max_user == NULL) {
		ast_log(LOG_WARNING, "[TK] Invalid CEI conference header data");
		return;
	}

	switch(notifyType) {
		case CEI_NOTIFY_FULL:
		{
			// Calculate user count
			total_count = atoi(local_user_count) + atoi(remote_user_count);

			// Reinitialize published conferences
			delete_remote_confs_by_publisher(conf, publisher_id);

			ao2_lock(conf->remote_container);
			/***
			get or create the remote conf from host conf's remote container;
			the remote conf is basically this cei msg come FROM
			***/
			remote_conference = build_remote_conf(conf, publisher_num, publisher_id, publisher_id);
			ao2_unlock(conf->remote_container);

			if (remote_conference) {
				ao2_lock(remote_conference->user_container);
				clear_conf_users(remote_conference);
				remote_conference->remote_user_size = total_count;
				remote_conference->locked = atoi(conf_locked) ? TRUE : FALSE;

				// Append users for remote publisher
				for (index = 0; index < atoi(total_update_count); index++) {
					conference_type = strtok_r(NULL, "\t", &saveptr);
					conf_user_count = strtok_r(NULL, "\t", &saveptr);
					user_number = strtok_r(NULL, "\t", &saveptr);
					cid_number = strtok_r(NULL, "\t", &saveptr);
					cid_name = strtok_r(NULL, "\t", &saveptr);
					channel_name = strtok_r(NULL, "\t", &saveptr);
					conference_id = strtok_r(NULL, "\t", &saveptr);
					self_conference_id = strtok_r(NULL, "\t", &saveptr);
					user_role = strtok_r(NULL, "\t", &saveptr);
					entity_status = strtok_r(NULL, "|", &saveptr);

					if (conference_type == NULL || user_number == NULL || conf_user_count == NULL ||
						cid_number == NULL || cid_name == NULL || channel_name == NULL ||
						conference_id == NULL || user_role == NULL ||
						entity_status == NULL || self_conference_id == NULL) {
						ast_log(LOG_WARNING, "[TK] Invalid CEI conference data");
						continue;
					}

					switch (atoi(conference_type)) {
						case CEI_CONFERENCE_LOCAL:
						{
							if (flag_ignored_conf == TRUE) {
								ast_debug(LOG_DEBUG, "[WL] User of ignored remote conf, then ignor this user");
								continue;
							}
							struct cei_user *user;
							if (!(user = ao2_alloc(sizeof(struct cei_user), NULL))) {
								ast_debug(LOG_DEBUG, "ZAKU DEBUG ---> CANNOT CREATE NEW USER");
								continue;
							}

							//add info to cei_user
							user->user_num = atoi(user_number);
							ast_copy_string(user->cid_name, cid_name, sizeof(user->cid_name));
							ast_copy_string(user->cid_number, cid_number, sizeof(user->cid_number));
							ast_copy_string(user->channel_name, channel_name, sizeof(user->channel_name));
							ast_copy_string(user->conference_id, remote_conference->conference_id, sizeof(user->conference_id));
							hash_user(user);
							user->muted = (atoi(entity_status) == CEI_STATUS_MUTEVIAFOCUS) ? 1 : 0;
							user->user_role = atoi(user_role);
							user->admin = 0;
							user->can_invite = 0;
							if (atoi(user_role) == CEI_ROLE_ADMIN) {
								user->admin = 1;
							} else if (atoi(user_role) == CEI_ROLE_USER_INVITE) {
								user->can_invite = 1;
							}

							if (entity_status == CEI_STATUS_UNKNOWN) {
								ast_log(LOG_DEBUG, "ZAKU DEBUG ---> USER EMPTY STATUS");
								ao2_ref(user, -1);
								continue;
							}
							ao2_link(remote_conference->user_container, user);
							ast_atomic_fetchadd_int(&remote_conference->local_size, 1);
							ast_atomic_fetchadd_int(&remote_conference->local_user_size, (!check_role_as_conference(user))? 1 : 0);
							ast_atomic_fetchadd_int(&conf->remote_size, 1);
							ast_atomic_fetchadd_int(&conf->remote_user_size, (!check_role_as_conference(user))? 1 : 0);
							ast_atomic_fetchadd_int(&remote_conference->refcount, 1);
						}
						break;
						case CEI_CONFERENCE_REMOTE:
						{
							// Ignore invalid conference
							// Don't add myself as my remote conf
							if (strcmp(conf->conference_id, conference_id) == 0) {
								ast_debug(LOG_DEBUG, "[TK] Ignore my own conference");
								ast_debug(LOG_DEBUG, "[TK] %s@%s", cid_number, conference_id);
								flag_ignored_conf = TRUE;
								continue;
							}
							// Don't add the remote conf which are my own remote conf again
							if (get_remote_conf_by_id(conf, conference_id) != NULL) {
								ast_debug(LOG_DEBUG, "[TK] Ignore my own remote conference");
								ast_debug(LOG_DEBUG, "[TK] %s@%s", cid_number, conference_id);
								flag_ignored_conf = TRUE;
								continue;
							}

							ao2_unlock(remote_conference->user_container);
							/***
							get or create the remote conf from host conf's remote container;
							the remote conf is basically this cei msg come FROM
							***/
							remote_conference = build_remote_conf(conf, cid_number, conference_id, publisher_id);

							ao2_lock(remote_conference->user_container);
							clear_conf_users(remote_conference);

							flag_ignored_conf = FALSE;
						}
						break;
						default:
						{
							ast_debug(LOG_DEBUG, "[WL] INVALID TYPE");
						}
					}
				}
				ao2_unlock(remote_conference->user_container);
			}
			poll_cei_conference(conf, data_string);
		}
		break;
		case CEI_NOTIFY_PARTIAl:
		{
			// Calculate user count
			total_count = atoi(local_user_count) + atoi(remote_user_count);

			remote_conference = get_remote_conf_by_id(conf, publisher_id);

			if (remote_conference) {

				// Append users for remote publisher
				for (index = 0; index < atoi(total_update_count); index++) {
					conference_type = strtok_r(NULL, "\t", &saveptr);
					conf_user_count = strtok_r(NULL, "\t", &saveptr);
					user_number = strtok_r(NULL, "\t", &saveptr);
					cid_number = strtok_r(NULL, "\t", &saveptr);
					cid_name = strtok_r(NULL, "\t", &saveptr);
					channel_name = strtok_r(NULL, "\t", &saveptr);
					conference_id = strtok_r(NULL, "\t", &saveptr);
					self_conference_id = strtok_r(NULL, "\t", &saveptr);
					user_role = strtok_r(NULL, "\t", &saveptr);
					entity_status = strtok_r(NULL, "|", &saveptr);

					if (conference_type == NULL || user_number == NULL || conf_user_count == NULL ||
						cid_number == NULL || cid_name == NULL || channel_name == NULL ||
						conference_id == NULL || user_role == NULL ||
						entity_status == NULL || self_conference_id == NULL) {
						ast_log(LOG_WARNING, "[TK] Invalid CEI conference data");
						continue;
					}

					switch (atoi(conference_type)) {
						case CEI_CONFERENCE_LOCAL:
						{
							if (flag_ignored_conf == TRUE) {
								ast_debug(LOG_DEBUG, "[WL] User of ignored remote conf, then ignor this user");
								continue;
							}
							struct cei_user *user;

							switch (atoi(entity_status)) {
								// CEI_EVENT_MCB_USER_JOIN
								case CEI_STATUS_CONNECTED:
								{
									ao2_lock(remote_conference->user_container);
									user = get_user_by_cid_num(remote_conference, cid_number);
									//CEI_EVENT_MCB_USER_TOGGLE_MUTE unmute
									if (user) {
										user->muted = 0;
										poll_cei_partial_notify(conf, remote_conference, user, data_string, CEI_EVENT_MCB_USER_TOGGLE_MUTE);

									}
									//CEI_EVENT_MCB_USER_JOIN
									else {
										if (!(user = ao2_alloc(sizeof(struct cei_user), NULL))) {
											ast_debug(LOG_DEBUG, "ZAKU DEBUG ---> CANNOT CREATE NEW USER");
											ao2_unlock(remote_conference->user_container);
											continue;
										}

										//add info to cei_user
										user->user_num = atoi(user_number);
										ast_copy_string(user->cid_name, cid_name, sizeof(user->cid_name));
										ast_copy_string(user->cid_number, cid_number, sizeof(user->cid_number));
										ast_copy_string(user->channel_name, channel_name, sizeof(user->channel_name));
										ast_copy_string(user->conference_id, remote_conference->conference_id, sizeof(user->conference_id));
										hash_user(user);
										user->muted = (atoi(entity_status) == CEI_STATUS_MUTEVIAFOCUS) ? 1 : 0;
										user->user_role = atoi(user_role);
										user->admin = 0;
										user->can_invite = 0;
										if (atoi(user_role) == CEI_ROLE_ADMIN) {
											user->admin = 1;
										} else if (atoi(user_role) == CEI_ROLE_USER_INVITE) {
											user->can_invite = 1;
										}

										ao2_link(remote_conference->user_container, user);
										ast_atomic_fetchadd_int(&remote_conference->local_size, 1);
										ast_atomic_fetchadd_int(&remote_conference->local_user_size, (!check_role_as_conference(user))? 1 : 0);
										ast_atomic_fetchadd_int(&conf->remote_size, 1);
										ast_atomic_fetchadd_int(&conf->remote_user_size, (!check_role_as_conference(user))? 1 : 0);
										ast_atomic_fetchadd_int(&remote_conference->refcount, 1);

										poll_cei_partial_notify(conf, remote_conference, user, data_string, CEI_EVENT_MCB_USER_JOIN);
									}
									ao2_unlock(remote_conference->user_container);
								}
								break;
								//CEI_EVENT_MCB_USER_PART
								case CEI_STATUS_DISCONNECTED:
								{
									ao2_lock(remote_conference->user_container);

									user = get_user_by_cid_num(remote_conference, cid_number);

									if (user) {

										ast_atomic_fetchadd_int(&remote_conference->local_size, -1);
										ast_atomic_fetchadd_int(&remote_conference->local_user_size, (!check_role_as_conference(user))? -1 : 0);
										ast_atomic_fetchadd_int(&conf->remote_size, -1);
										ast_atomic_fetchadd_int(&conf->remote_user_size, (!check_role_as_conference(user))? -1 : 0);

										poll_cei_partial_notify(conf, remote_conference, user, data_string, CEI_EVENT_MCB_USER_PART);

										ao2_unlink(remote_conference->user_container, user);
										ast_atomic_fetchadd_int(&remote_conference->refcount, -1);

										ast_debug(LOG_DEBUG, "[WL] REMOVED USER WITH ADDRESS %p", user);
										ao2_ref(user, -1);
									}
									else {
										ast_debug(LOG_DEBUG, "[WL] ERROR, CANNOT FIND USER  %s IN CONFERENCE %s", cid_number, remote_conference->conference_num);
									}

									ao2_unlock(remote_conference->user_container);
								}
								break;
								case CEI_STATUS_MUTEVIAFOCUS:
								{
									ao2_lock(remote_conference->user_container);

									user = get_user_by_cid_num(remote_conference, cid_number);

									if (user) {
										user->muted = 1;
										poll_cei_partial_notify(conf, remote_conference, user, data_string, CEI_EVENT_MCB_USER_TOGGLE_MUTE);
									}

									ao2_unlock(remote_conference->user_container);
								}
								break;
								default:
								{
									ast_debug(LOG_DEBUG, "[WL] INVALID STATUS");
								}
								break;
							}
						}
						break;
						case CEI_CONFERENCE_REMOTE:
						{
							// Ignore invalid conference
							// Don't add myself as my remote conf
							if (strcmp(conf->conference_id, conference_id) == 0) {
								ast_debug(LOG_DEBUG, "[TK] Ignore my own conference");
								ast_debug(LOG_DEBUG, "[TK] %s@%s", cid_number, conference_id);
								flag_ignored_conf = TRUE;
								continue;
							}

							/***
							get or create the remote conf from host conf's remote container;
							the remote conf is basically this cei msg come FROM
							***/

							remote_conference = get_remote_conf_by_id(conf, conference_id);
							if (atoi(entity_status) == CEI_STATUS_CONNECTED) {
								if (remote_conference != NULL) {
									ast_debug(LOG_DEBUG, "[TK] CEI_EVENT_MCB_USER_* event");
									continue;
								}

							} else if (atoi(entity_status) == CEI_STATUS_DISCONNECTED) {
								// CEI_EVENT_MCB_CONF_PART
								if (remote_conference != NULL) {
									delete_remote_conf_by_id(conf, publisher_id);
									poll_cei_conference(conf, data_string);
								}
							}
						}
						break;
						default:
						{
							ast_debug(LOG_DEBUG, "[WL] INVALID TYPE");
						}
					}
				}
			}
		}
		break;
		default:
		{
			ast_debug(LOG_DEBUG, "[WL] INVALID TYPE");
		}
	}
	ast_log(LOG_DEBUG, "[TK] Exit...%s\n", __FUNCTION__);

	return;
}

static int get_conf_max_capability(void)
{
	// Get device model
	FILE* fp = fopen( "/proc/gxvboard/dev_info/dev_alias", "r" );
	char* devModel = NULL;

	if ( fp != NULL ){
		size_t len = 0;

		if(getline( &devModel, &len, fp ) < 0){
			ast_log(LOG_ERROR,"get device error");
		}

		fclose( fp );
	}

	// Get max user count for conference according to system
	int maxUserCount = 0;
	if ( strcmp(devModel, "UCM6102") == 0 || strcmp(devModel, "UCM6104") == 0 ){
		maxUserCount = 25;
	}else if ( strcmp(devModel, "UCM6108") == 0 || strcmp(devModel, "UCM6116") == 0 ){
		maxUserCount = 32;
	}else if ( strcmp(devModel, "UCM6510") == 0 ){
		maxUserCount = 64;
	}
	return maxUserCount;
}

static int check_role_as_conference(struct cei_user *user)
{
		if (user == NULL) {
			return 0;
		}

		if (user->user_role == CEI_ROLE_CONFERENCE) {
			return 1;
		} else {
			return 0;
		}
}

static void reset_conference(struct cei_conference *conf)
{
	if (conf == NULL) {
		return;
	}

	if (ast_atomic_dec_and_test(&conf->refcount)) {
		//reset conf
		conf->locked = FALSE;
		conf->version = 0;
		conf->version_reset = 0;
		conf->stop_propagating = FALSE;
		ast_atomic_fetchadd_int(&conf->refcount, 1);
	}
	else
		ast_atomic_fetchadd_int(&conf->refcount, 1);
}
#endif
