/*
 * Vinetic-SVIP native transcoding support
 *
 * Copyright (C) 2015, Grandstream Networks, Inc.
 *
 * jwli@grandstream.com 
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Translate between various formats natively through
 *  Vinetic-SVIP transcoding
 *
 * \ingroup codecs
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision$");

#include <fcntl.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/socket.h>

#include <sys/poll.h>

#include "asterisk/lock.h"
#include "asterisk/translate.h"
#include "asterisk/config.h"
#include "asterisk/module.h"
#include "asterisk/cli.h"
#include "asterisk/channel.h"
#include "asterisk/utils.h"
#include "asterisk/linkedlists.h"
#include "asterisk/ulaw.h"
#include "asterisk/channel.h"
#include "asterisk/timing.h"
#include "asterisk/format_compatibility.h"

#include "svip_codec/gxe_svip_codec.h"

#ifndef SVIP_RES_MANAGER
#error "you must build svip-resouce-manager into asterisk"
#endif

#define MAX_TRANSLATOR_NUM MAX_SVIP_CODEC_CHANNEL

#define BUFFER_SIZE          8000

#define G723_SENDRATE_53     53
#define G723_SENDRATE_63     63
#define G723_FRAME_LEN_53    20
#define G723_FRAME_LEN_63    48 //24*2

#define ILBC_SENDRATE_133 133
#define ILBC_SENDRATE_152 152

/* #define SOFTSLIN_AULAW */
/* #define USESOFTLIN_ONLY */

/* #define SVIP_DEBUG */

#define USE_SINGLE_THREAD
#define USE_AST_TS_SEQNUM
#define MAX_RINGBUF_NUM 32 /*if frame length is 20 ms,ringbuffer length is 640ms*/
#define RINGBUF_LEN 180 /*ringbuffer length in ms, 180 ms frames length in total*/

#define TRANSLATOR_ID(n, m) (61000 + (n) * MAX_VINETIC_CHAN_PER_DEV + (m))

#ifdef USESOFTLIN_ONLY
	#ifndef SOFTSLIN_AULAW
	#error "you must define SOFTSLIN_AULAW for USESOFTLIN_ONLY"
	#endif
#endif 

typedef struct {
	void *buffer;
	int buf_lens;
	int frame_offset;
	int frame_lens;
	unsigned int timestamp;
#if defined(USE_AST_TS_SEQNUM)
	unsigned short seq_num;
#endif
} svip_frame_t;

/* Circular buffer object for receive svip rtp packets */
typedef struct {
	int count;   /* maximum number of frame              */
	int start;  /* index of oldest frame                */
	int end;    /* index at which to write new frame    */
	svip_frame_t *frames[MAX_RINGBUF_NUM];
} svip_rbuf_t;

typedef struct {
	uint8_t dev_ind;
	uint8_t encchannel;
	uint8_t decchannel;
} coder_id_t;

struct svip_trans_par {
	int p_type;
	int p_lens;
	int enc_lens;
	int enc_type;
	int required_samples;
	int frameout_samples;
	int ts_inc;
};

struct codec_svip_pvt {
	coder_id_t cod_id;
	struct svip_trans_formats fmts;
	uint32_t softslin:1;
	uint32_t fake:2;
	uint16_t required_samples;
	uint16_t frameout_samples;
	uint16_t samples_in_buffer;
	uint8_t ulaw_buffer[1024*2];
	svip_rbuf_t *frame_out_srb;
	svip_rtp_sess_t rtp_sess;
	struct svip_trans_par trsc_pars;
#if !defined(USE_SINGLE_THREAD)
	pthread_t svip_rw_thread;
	struct ast_timer *timer;
#endif
};

#if defined(USE_SINGLE_THREAD)
	pthread_t g_svip_rw_thread = AST_PTHREADT_NULL;
	struct ast_timer *g_svip_timer = NULL;
#endif

ast_mutex_t pmutex;
static int close_svip_thread;

typedef struct {
	int total;
	int coder_lbr;
	int coder_g711;
} chan_usage_t;

static struct vinetic_usage {
	int dev_total;
	int dev_nch[MAX_VINETIC_DEV_NUM];
	int svip_dev_state;
	chan_usage_t channels;
} devices = {
	.dev_total = 0,
	.svip_dev_state = 0,
	.channels = {
		.total = 0,
	},
};

typedef struct {
	unsigned int svip_new_total;
	unsigned int svip_new_succ;
	unsigned int svip_new_fail;
	unsigned int svip_dstr_total;
} svip_callstat_t;

static svip_callstat_t svip_callstat;

struct svip_translator_usage {
	struct codec_svip_pvt *svipp;
	AST_LIST_ENTRY(svip_translator_usage) entry;
};

struct svip_translator_show {
	int trans_id;
	unsigned int srcfmt;
	unsigned int dstfmt;
	int softslin;
	int dev_id;
	int encchan;
	int decchan;
	AST_LIST_ENTRY(svip_translator_show) entry;
};

static AST_RWLIST_HEAD_STATIC(g_svip_translator_list, svip_translator_usage);
static AST_LIST_HEAD_STATIC(g_svip_translator_show, svip_translator_show);

#if defined(USE_SINGLE_THREAD)
static int g_svip_translator_socket[2];
#endif 

typedef enum {
	COD_PAIR_UNCONFIG,
	COD_PAIR_CONFIG_S1,
	COD_PAIR_CONFIG_S2,
} cod_pair_stat_t;

struct svip_codec_pair {
	int dev_ind;
	int encchannel;
	int decchannel;
	cod_pair_stat_t cod_pair_stat;   /*'>0' configed, '0'  unconfiged*/
	struct svip_trans_formats fmts;
	svip_rtp_sess_t rtp_sess;
	struct svip_trans_par trsc_pars;
	AST_LIST_ENTRY(svip_codec_pair) entry;
};

static AST_LIST_HEAD_STATIC(svip_rem_codec_pairs, svip_codec_pair);
static AST_LIST_HEAD_STATIC(svip_inused_codec_pairs, svip_codec_pair);

int module_load_fail = 0;

int svip_fd;

#define SET_SVIP_DEV_STATE_START(nDev, state) ((state) |= ( 1 << (nDev)))
#define SET_SVIP_DEV_STATE_STOP(nDev, state) ((state) &= ~( 1 << (nDev)))

#define SVIP_DEV_ISSTARTED(nDev, state) (((state) & (1 << (nDev))) ? 1 : 0)

const char *dev_filename = "/dev/svip";

static char *svip_db_levels[9] = {
	"L_EMERG",
	"L_ALERT",
	"L_CRIT",
	"L_ERR",
	"L_WARNING",
	"L_NOTICE",
	"L_INFO",
	"L_DBG",
	"L_THREAD_DEBUG",
};

uint8_t g723_sendrate = G723_SENDRATE_63;  /*default g723 send rate: 6.3 kbit/s*/
uint8_t ilbc_sendrate = ILBC_SENDRATE_133; /*default iLBC send rate: 133 kbit/s*/

struct svip_transcoder_info {
	unsigned int tcnum;
	char name[80];
	unsigned int numchannels;
	unsigned int dstfmts;
	unsigned int srcfmts;
} svip_tc_info []= {
#ifdef USESOFTLIN_ONLY
	{
		.tcnum = 5,
		.name = "svip-codec",
		.numchannels = 32,
		.dstfmts = (AST_FORMAT_G723| 
			AST_FORMAT_G729|
			AST_FORMAT_ILBC|
			AST_FORMAT_G726|
			AST_FORMAT_G726_AAL2),
		.srcfmts = AST_FORMAT_ULAW,
	},
	{
		.tcnum = 5,
		.name = "svip-codec",
		.numchannels = 32,
		.dstfmts =  AST_FORMAT_ULAW,
		.srcfmts = (AST_FORMAT_G723| 
			AST_FORMAT_G729|
			AST_FORMAT_ILBC|
			AST_FORMAT_G726|
			AST_FORMAT_G726_AAL2),
	},
#else
	{
		.tcnum = 5,
		.name = "svip-codec",
		.numchannels = 32,
		.dstfmts = (AST_FORMAT_G723| 
			AST_FORMAT_G729|
			AST_FORMAT_ILBC|
			AST_FORMAT_G726|
			AST_FORMAT_G726_AAL2),
		.srcfmts = AST_FORMAT_ULAW,
	},
	{
		.tcnum = 5,
		.name = "svip-codec",
		.numchannels = 32,
		.dstfmts = (AST_FORMAT_G723| 
			AST_FORMAT_G729|
			AST_FORMAT_ILBC|
			AST_FORMAT_G726|
			AST_FORMAT_G726_AAL2),
		.srcfmts = AST_FORMAT_ALAW,
	},
	{
		.tcnum = 6,
		.name = "svip-codec",
		.numchannels = 32,
		.dstfmts = (AST_FORMAT_ULAW|
			AST_FORMAT_ALAW|
			AST_FORMAT_G729|
			AST_FORMAT_ILBC|
			AST_FORMAT_G726|
			AST_FORMAT_G726_AAL2),
		.srcfmts =  AST_FORMAT_G723,
	},
	{
		.tcnum = 6,
		.name = "svip-codec",
		.numchannels = 32,
		.dstfmts = (AST_FORMAT_ULAW|
			AST_FORMAT_ALAW|
			AST_FORMAT_G723|
			AST_FORMAT_ILBC|
			AST_FORMAT_G726|
			AST_FORMAT_G726_AAL2),
		.srcfmts =  AST_FORMAT_G729,
	},
	{
		.tcnum = 6,
		.name = "svip-codec",
		.numchannels = 32,
		.dstfmts = (AST_FORMAT_ULAW|
			AST_FORMAT_ALAW|
			AST_FORMAT_G723|
			AST_FORMAT_G729|
			AST_FORMAT_G726|
			AST_FORMAT_G726_AAL2),          
		.srcfmts =  AST_FORMAT_ILBC,
	},
	{
		.tcnum = 6,
		.name = "svip-codec",
		.numchannels = 32,
		.dstfmts = (AST_FORMAT_ULAW|
			AST_FORMAT_ALAW|
			AST_FORMAT_G723|
			AST_FORMAT_G729|
			AST_FORMAT_ILBC|
			AST_FORMAT_G726_AAL2),          
		.srcfmts =  AST_FORMAT_G726,
	},
	{
		.tcnum = 6,
		.name = "svip-codec",
		.numchannels = 32,
		.dstfmts = (AST_FORMAT_ULAW|
			AST_FORMAT_ALAW|
			AST_FORMAT_G723|
			AST_FORMAT_G729|
			AST_FORMAT_ILBC|
			AST_FORMAT_G726),          
		.srcfmts = AST_FORMAT_G726_AAL2,
	},
#endif
};

static const struct ast_codec *get_svip_codec(uint32_t fmt)
{
	const struct ast_codec *codec;

	static const struct ast_codec svip_g723 = {
		.name = "g723",
		.type = AST_MEDIA_TYPE_AUDIO,
		.sample_rate = 8000,
	};
	static const struct ast_codec svip_g726 = {
		.name = "g726",
		.type = AST_MEDIA_TYPE_AUDIO,
		.sample_rate = 8000,
	};
	static const struct ast_codec svip_g726aal2 = {
		.name = "g726aal2",
		.type = AST_MEDIA_TYPE_AUDIO,
		.sample_rate = 8000,
	};
	static const struct ast_codec svip_g729 = {
		.name = "g729",
		.type = AST_MEDIA_TYPE_AUDIO,
		.sample_rate = 8000,
	};
	static const struct ast_codec svip_ilbc = {
		.name = "ilbc",
		.type = AST_MEDIA_TYPE_AUDIO,
		.sample_rate = 8000,
	};
	static const struct ast_codec svip_ulaw = {
		.name = "ulaw",
		.type = AST_MEDIA_TYPE_AUDIO,
		.sample_rate = 8000,
	};
	static const struct ast_codec svip_alaw = {
		.name = "alaw",
		.type = AST_MEDIA_TYPE_AUDIO,
		.sample_rate = 8000,
	};
	static const struct ast_codec svip_slinear = {
		.name = "slin",
		.type = AST_MEDIA_TYPE_AUDIO,
		.sample_rate = 8000,
	};

	switch (fmt) {
	case AST_FORMAT_G723:
		codec = &svip_g723;
		break;
	case AST_FORMAT_G726:
		codec = &svip_g726;
		break;
	case AST_FORMAT_G726_AAL2:
		codec = &svip_g726aal2;
		break;
	case AST_FORMAT_G729:
		codec = &svip_g729;
		break;
	case AST_FORMAT_ILBC:
		codec = &svip_ilbc;
		break;
	case AST_FORMAT_ULAW:
		codec = &svip_ulaw;
		break;
	case AST_FORMAT_ALAW:
		codec = &svip_alaw;
		break;
	case AST_FORMAT_SLIN:
		codec = &svip_slinear;
		break;
	default:
		codec = NULL;
		break;
	}

	return codec;
}

static const char *get_svip_codec_name(uint32_t fmt) 
{
	const struct ast_codec *codec;

	codec = get_svip_codec(fmt);
	if (codec)
		return codec->name;
	else
		return "Unknown";
}

static char *handle_cli_transcoder_show(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_svip_debug_set(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_svip_debug_get(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_svip_jitterbuf_set(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_svip_jitterbuf_get(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);

static struct ast_cli_entry cli[] = {
	AST_CLI_DEFINE(handle_cli_transcoder_show, "Display Svip transcoder utilization."),
	AST_CLI_DEFINE(handle_cli_svip_debug_set, "Svip debug utilization."),
	AST_CLI_DEFINE(handle_cli_svip_debug_get, "Svip debug utilization."),
	AST_CLI_DEFINE(handle_cli_svip_jitterbuf_set, "Svip jitter buffer set type and length."),
	AST_CLI_DEFINE(handle_cli_svip_jitterbuf_get, "Svip jitter buffer get type and length."),
};

struct translator {
	struct ast_translator t;
	uint32_t src_svip_fmt;
	uint32_t dst_svip_fmt;
	AST_LIST_ENTRY(translator) entry;
};

#ifndef container_of
#define container_of(ptr, type, member) \
	((type *)((char *)(ptr) - offsetof(type, member)))
#endif

static AST_LIST_HEAD_STATIC(translators, translator);

/* Workaround for Bug 45427, Asterisk crash on FD_ISSET
*   If GCC flag _FORTIFY_SOURCE be enabled, it will cause an "abort"  as soon as
*  the socket  file description fd >1024(FD_SETSIZE) , for more information, check
*  the definition FD_ISSET in GLIBC.
*/
static long int svip_fdelt_warn( long int fd)
{
	ast_log(LOG_ERROR, "svip_fdelt_warn: FD %ld exceeds"
		"the maximum size of ast_fdset %d!\n", fd, ast_FD_SETSIZE);
	return 0;
}
#define __SVIP_FD_ISSET(d, s) \
	((__FDS_BITS (s)[(__d / __NFDBITS)] & __FD_MASK (d)) != 0)

#define SVIP_FD_ISSET(d, s) \
({ long int __d = (d);			           \
	(0 <= __d && __d < ast_FD_SETSIZE		\
		   ? __SVIP_FD_ISSET(d, s)	       \
		   : svip_fdelt_warn(__d));})

#define NEW_DESTROY_LOCK
#ifdef NEW_DESTROY_LOCK
ast_mutex_t tr_lock;
/* #define DETECT_SVIP_DEADLOCKS */
static int __svip_mutex_lock(const char *filename, int lineno, const char *func,
	const char* mutex_name, ast_mutex_t *t);

#define svip_mutex_lock(a)  __svip_mutex_lock(__FILE__, __LINE__, __PRETTY_FUNCTION__, #a, a)
#define svip_mutex_unlock(a)  ast_mutex_unlock(a)

static int __svip_mutex_lock(const char *filename, int lineno, const char *func,
	const char* mutex_name, ast_mutex_t *t)
{
	int res;

#if defined(DETECT_SVIP_DEADLOCKS)

	time_t seconds = time(NULL);
	time_t wait_time, reported_wait = 0;
	while ((res = ast_mutex_trylock(t))) {
		wait_time = time(NULL) - seconds;
		if (wait_time > reported_wait && (wait_time % 5) == 0) {
			ast_log(LOG_ERROR, "%s line %d (%s): Deadlock? waited %d sec for mutex '%s'?\n",
				filename, lineno, func, (int) wait_time, mutex_name);
#ifdef HAVE_BKTR
			ast_log(LOG_ERROR, "ast_backtrace:\n");
			ast_log_backtrace();
#else
#warning "HAVE_BKTR! not defined!" 
#endif
			reported_wait = wait_time;
		}
		if (wait_time > 60) {
			ast_log(LOG_ERROR, "%s line %d (%s): take mutex '%s' more than %d s, Out of lock.\n",
				filename, lineno, func, mutex_name, (int)wait_time);
			break;
		}
		usleep(abs(random())%200);
	};

#else /* !DETECT_SVIP_DEADLOCKS*/
	res = ast_mutex_lock(t);
#endif

	return res;
}

#endif /* #ifdef NEW_DESTROY_LOCK */

/* Only used by a decoder */
static int ulawtolin(struct ast_trans_pvt *pvt, int samples)
{
	struct codec_svip_pvt *svipp = pvt->pvt;
	int i = samples;
	uint8_t *src = &svipp->ulaw_buffer[0];
	int16_t *dst = pvt->outbuf.i16 + pvt->datalen;

	/* convert and copy in outbuf */
	while (i--) {
		*dst++ = AST_MULAW(*src++);
	}

	return 0;
}

/* Only used by an encoder. */
static int lintoulaw(struct ast_trans_pvt *pvt, struct ast_frame *f)
{
	struct codec_svip_pvt *svipp = pvt->pvt;
	int i = f->samples;
	uint8_t *dst = &svipp->ulaw_buffer[svipp->samples_in_buffer];
	int16_t *src = f->data.ptr;

	if (svipp->samples_in_buffer + i > sizeof(svipp->ulaw_buffer)) {
		ast_log(LOG_ERROR, "Out of buffer space!\n");
		return -i;
	}

	while (i--) {
		*dst++ = AST_LIN2MU(*src++);
	}

	svipp->samples_in_buffer += f->samples;

	return 0;
}

static char *handle_cli_transcoder_show(struct ast_cli_entry *e, int cmd,
	struct ast_cli_args *a)
{
	int total;
	struct svip_translator_usage *trans_usage;
	struct svip_codec_pair *codec_pair;
	struct svip_translator_show *trans_show = NULL;
	int trans_num = 0;
	int cpair_num = 0;

	switch (cmd) {
	case CLI_INIT:
		e->command = "svip transcoder show";
		e->usage =
			"Usage: svip transcoder show\n"
			"       Displays channel utilization of SVIP transcoder(s).\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3)
		return CLI_SHOWUSAGE;

	total = devices.channels.total;

	ast_cli(a->fd, "Translator New Total:%u | New Ok:%u | New Failed:%u | Destroy Total:%u \n",
		svip_callstat.svip_new_total, svip_callstat.svip_new_succ,
		svip_callstat.svip_new_fail, svip_callstat.svip_dstr_total);

	if (total == 0)
		ast_cli(a->fd, "No SVIP transcoders found.\n");
	else {
		ast_cli(a->fd, "%d LBR coder(s) %d G711 coder(s) of %d channels are in use.\n", 
			devices.channels.coder_lbr, devices.channels.coder_g711, total);
		ast_cli(a->fd, "[NO.][translator ID]  (transcode path) [device]  [encoder channel] [decoder channel]\n");

		AST_RWLIST_RDLOCK(&g_svip_translator_list);
		AST_RWLIST_TRAVERSE_SAFE_BEGIN(&g_svip_translator_list, trans_usage, entry) {
			/*copy translator info to g_svip_translator_show list*/
			if ((trans_show = ast_calloc(1, sizeof(*trans_show)))) {
				trans_show->trans_id = TRANSLATOR_ID(trans_usage->svipp->cod_id.dev_ind,
					trans_usage->svipp->cod_id.decchannel);
				trans_show->srcfmt = trans_usage->svipp->fmts.srcfmt;
				trans_show->dstfmt = trans_usage->svipp->fmts.dstfmt;
				trans_show->softslin = trans_usage->svipp->softslin;
				trans_show->dev_id = trans_usage->svipp->cod_id.dev_ind;
				trans_show->encchan = trans_usage->svipp->cod_id.encchannel;
				trans_show->decchan = trans_usage->svipp->cod_id.decchannel;
				AST_LIST_INSERT_TAIL(&g_svip_translator_show, trans_show, entry);
			}
		}
		AST_RWLIST_TRAVERSE_SAFE_END;
		AST_RWLIST_UNLOCK(&g_svip_translator_list);

		AST_LIST_TRAVERSE_SAFE_BEGIN(&g_svip_translator_show, trans_show, entry) {
			ast_cli(a->fd, "[%d][%d] ( \e[34m\e[1m%s ------->  %s %s\e[0m ) dev:%d encc:%d decc:%d\n",
				trans_num++,
				trans_show->trans_id,
				get_svip_codec_name(trans_show->srcfmt),
				get_svip_codec_name(trans_show->dstfmt),
				trans_show->softslin ?"(softslin)":" ",
				trans_show->dev_id,
				trans_show->encchan,
				trans_show->decchan);
			AST_LIST_REMOVE_CURRENT(entry);
			ast_free(trans_show);
		}
		AST_LIST_TRAVERSE_SAFE_END;

		ast_cli(a->fd, "Idle Transcoding Path:\n");

		/*print coder pair rem list*/
		AST_LIST_LOCK(&svip_rem_codec_pairs);
		AST_LIST_TRAVERSE_SAFE_BEGIN(&svip_rem_codec_pairs, codec_pair, entry) {
			ast_cli(a->fd, "[%d][%d] ( \e[32m\e[1m%s ------->  %s \e[0m ) dev:%d encc:%d decc:%d\n",
				cpair_num++,
				TRANSLATOR_ID(codec_pair->dev_ind, codec_pair->decchannel),
				get_svip_codec_name(codec_pair->fmts.srcfmt),
				get_svip_codec_name(codec_pair->fmts.dstfmt),
				codec_pair->dev_ind,
				codec_pair->encchannel,
				codec_pair->decchannel);
		  }
		AST_LIST_TRAVERSE_SAFE_END;
		AST_LIST_UNLOCK(&svip_rem_codec_pairs);
	}

	return CLI_SUCCESS;
}

static char *handle_cli_svip_jitterbuf_set(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int res;
	svip_jb_setting_t jb_set;

	switch (cmd) {
	case CLI_INIT:
		e->command = "svip set jitterbuff {FIXED|ADAPTIVE} {LOW|MEDIUM|HIGH}";
		e->usage =
			"Usage: svip set jitterbuff {FIXED|ADAPTIVE} {LOW|MEDIUM|HIGH}\n"
			"       set SVIP jitter buffer type and length.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 5)
		return CLI_SHOWUSAGE;

	if (!strcasecmp(S_OR(a->argv[3], ""), "FIXED")) {
		jb_set.type = JB_TYPE_FIXED;
	} else if (!strcasecmp(S_OR(a->argv[3], ""), "ADAPTIVE")) {
		jb_set.type = JB_TYPE_ADAPTIVE;
	} else {
		return CLI_SHOWUSAGE;
	}
	if (!strcasecmp(S_OR(a->argv[4], ""), "LOW")) {
		jb_set.length = JB_LEN_LOW;
	} else if (!strcasecmp(S_OR(a->argv[4], ""), "MEDIUM")) {
		jb_set.length = JB_LEN_MEDIUM;
	} else if (!strcasecmp(S_OR(a->argv[4], ""), "HIGH")) {
		jb_set.length = JB_LEN_HIGH;
	}else {
		return CLI_SHOWUSAGE;
	}

	res = vinetic_set_jb_typelen (jb_set);
	if (res)
		ast_cli(a->fd, "Set svip jitter buffer type and length error\n");

	ast_cli(a->fd, "Set svip jitter buffer type and length sucessful\n");
	
	return CLI_SUCCESS;
}

static char *handle_cli_svip_jitterbuf_get(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	svip_jb_setting_t jb_set;

	switch (cmd) {
	case CLI_INIT:
		e->command = "svip get jbsetting";
		e->usage =
			"Usage: svip get jbsetting\n"
			"       get SVIP jitter buffer setting.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3)
		return CLI_SHOWUSAGE;
	
	vinetic_get_jb_typelen(&jb_set);

	ast_cli(a->fd, "svip jitter buffer setting: type: %s length: %s\n", 
		(jb_set.type==JB_TYPE_FIXED) ? "FIXED" : ((jb_set.type==JB_TYPE_ADAPTIVE) ? "ADAPTIVE" : "UNKNOW"),
		(jb_set.length==JB_LEN_LOW) ? "LOW" : 
		((jb_set.length==JB_LEN_MEDIUM) ? "MEDIUM" :
		((jb_set.length==JB_LEN_HIGH) ? "HIGH" : "UNKNOW")));

	return CLI_SUCCESS;
}

static char *handle_cli_svip_debug_set(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int res;
	int u;

	switch (cmd) {
	case CLI_INIT:
		e->command = "svip set debug level {L_EMERG|L_ALERT|L_CRIT|L_ERR|L_WARNING|L_NOTICE|L_INFO|L_DBG|L_THREAD_DEBUG}";
		e->usage =
			"Usage: svip set debug level {L_EMERG|L_ALERT|L_CRIT|L_ERR|L_WARNING|L_NOTICE|L_INFO|L_DBG|L_THREAD_DEBUG}\n"
			"       set SVIP debug level.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 5)
		return CLI_SHOWUSAGE;

	
	for (u = 0; u < ARRAY_LEN(svip_db_levels); u++) {
		if (svip_db_levels[u] && !strcasecmp(S_OR(a->argv[4], ""), svip_db_levels[u]))
		{
			if (u==8) {
				ast_cli(a->fd, "Set thread lock debug level\n");
			} else {
				res = vinetic_set_loglevel(u);
				ast_cli(a->fd, "Set svip debug level %s(%d)\n", svip_db_levels[u], res);
			}
			break;
		}
	}

	return CLI_SUCCESS;
}

static char *handle_cli_svip_debug_get(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int res;

	switch (cmd) {
	case CLI_INIT:
		e->command = "svip get debug level";
		e->usage =
			"Usage: svip get debug level\n"
			"       get SVIP debug level.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 4)
		return CLI_SHOWUSAGE;
	
	res = vinetic_get_loglevel();
	ast_cli(a->fd, "the current svip debug level is %s(%d)\n", svip_db_levels[res], res);

	return CLI_SUCCESS;
}

static int parse_config(int reload) 
{
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	struct ast_config *cfg = ast_config_load("codecs.conf", config_flags);
	struct ast_variable *var;
	int res;

	if (cfg == CONFIG_STATUS_FILEMISSING ||
		cfg == CONFIG_STATUS_FILEUNCHANGED ||
		cfg == CONFIG_STATUS_FILEINVALID)
		return 0;

	for (var = ast_variable_browse(cfg, "g723_1"); var; var = var->next) {
		if (!strcasecmp(var->name, "sendrate")) {
			res = abs(atoi(var->value));
			if (res == G723_SENDRATE_53 || res == G723_SENDRATE_63) {
				ast_verb(3, "CODEC G723.1: Setting Send rate: %d\n",res);
				g723_sendrate = res;
			} else 
				ast_log(LOG_ERROR,"Error G723.1 Sendrate must be set 53 or 63\n");
		} 
	}
#ifdef CONFIG_SENDRATE_ILBC
	for (var = ast_variable_browse(cfg, "ilbc"); var; var = var->next) {
		if (!strcasecmp(var->name, "sendrate")) {
			res = abs(atoi(var->value));
			if (res == ILBC_SENDRATE_133 || res == ILBC_SENDRATE_152) {
				ast_verb(3, "CODEC iLBC: Setting Send rate: %d\n",res);
				ilbc_sendrate = res;
			} else 
				ast_log(LOG_ERROR,"Error iLBC Sendrate must be set 133 or 152\n");
		}
	}
#endif
	ast_config_destroy(cfg);
	return 0;
}

#if defined(SVIP_DEBUG)
static void svip_dump_frame_info (struct ast_trans_pvt *pvt, char *cust_inf)
{
	struct ast_frame pvt_f = pvt->f; 

	ast_debug(LOG_DEBUG, "(%s)====PVT frame info====\n\
		datalen:%d \n\
		samples:%d \n\
		mallocd:%d \n\
		mallocd_hdr_len:%d \n\
		offset:%d \n\
		frame flags:%d \n\
		ts:%ld \n\
		Length in milliseconds:%ld \n\
		seqno:%d \n",
		cust_inf, pvt_f.datalen, pvt_f.samples,
		pvt_f.mallocd, pvt_f.mallocd_hdr_len,
		pvt_f.offset, pvt_f.flags, pvt_f.ts,
		pvt_f.len, pvt_f.seqno
	);
}
#endif

static int svip_rb_allocInit( svip_rbuf_t *srb, int num, int size) 
{
	int i;

	srb->start = 0;
	srb->end   = 0;

	srb->count  = num + 1; /*include a empty buffer*/  
	if (srb->count > MAX_RINGBUF_NUM)
		srb->count = MAX_RINGBUF_NUM;

	for (i=0; i<srb->count; i++) {
		srb->frames[i] = (svip_frame_t *)ast_calloc(1, sizeof(svip_frame_t));
		if (!srb->frames[i])
			break;
		srb->frames[i]->buffer = (void *)ast_calloc(1, size);
		if (!srb->frames[i]) {
			free(srb->frames[i]);
			break;
		}
		srb->frames[i]->buf_lens = size;
	}

	if (i < srb->count) {
		--i;
		for (; i>=0; i--) {
			ast_free(srb->frames[i]->buffer);
			ast_free(srb->frames[i]);
		}
		srb->count  = 0;
		return -ENOMEM;
	}

	return 0;
}

static void svip_rb_free(svip_rbuf_t *srb)
{
	int i;

	for (i=0; i<srb->count; i++) {
		ast_free(srb->frames[i]->buffer);
		ast_free(srb->frames[i]);
	}
}

static int svip_rb_isempty(svip_rbuf_t *srb)
{
	return srb->end == srb->start;
}

static svip_frame_t* svip_rb_end_frame_get(svip_rbuf_t *srb)
{
	return  srb->frames[srb->end];
}

static void svip_rb_end_frame_put(svip_rbuf_t *srb)
{
	srb->end = (srb->end + 1) % srb->count;
	if (srb->end == srb->start)
		srb->start = (srb->start + 1) % srb->count;
}

#if defined(USE_AST_TS_SEQNUM)
static int svip_rb_read(svip_rbuf_t *srb, void *buff, int *count, unsigned int *ts, unsigned short *seq) 
#else
static int svip_rb_read(svip_rbuf_t *srb, void *buff, int *count, unsigned int *ts) 
#endif
{
	void *framep;
	int frame_lens = 0;

	count[0] = 0;

#if defined(USE_AST_TS_SEQNUM)
	*seq = srb->frames[srb->start]->seq_num;
#endif
	*ts = srb->frames[srb->start]->timestamp;
	while (!svip_rb_isempty(srb)) {
		framep = srb->frames[srb->start]->buffer + srb->frames[srb->start]->frame_offset;
		memcpy(buff, framep, srb->frames[srb->start]->frame_lens);

		frame_lens += srb->frames[srb->start]->frame_lens;
		buff += srb->frames[srb->start]->frame_lens;
		srb->start = (srb->start + 1) % srb->count;
		count[0] ++;
	}
	return frame_lens;
}

static void svip_write_frame(struct codec_svip_pvt *svipp, const uint8_t *buffer, const ssize_t count)
{
	rtp_packet_t snd_packet = svipp->rtp_sess.rtp_pkt_snd;
	unsigned char rtp_pt = svipp->rtp_sess.rtp_pt;
	unsigned int *seq_num = &svipp->rtp_sess.seq_num;
	unsigned int *ts_current = &svipp->rtp_sess.timestamp;
	int *rtp_ssrc = &svipp->rtp_sess.rtp_ssrc;
#if !defined(USE_AST_TS_SEQNUM)
	unsigned int timestamp_increse = svipp->rtp_sess.timestamp_increse; 
#endif
	int *m_bit = &svipp->rtp_sess.m_bit;
	int sock = svipp->rtp_sess.sock;
	struct sockaddr_in *svip_saddr = &svipp->rtp_sess.svip_saddr;
	struct rtp_packet packet;
	int res;
	int rval;
	struct pollfd p = {0};

	void *payload = snd_packet->data + 12;
	int payload_buf_len = snd_packet->data_len -12;

	packet.data = snd_packet->data;

#if defined(SVIP_DEBUG)
	ast_log(LOG_DEBUG, "svip_write_frame codec enc[%d] dec[%d] device [%d] wirte count %d payload_buf_len %d \n", 
		svipp->cod_id.encchannel,
		svipp->cod_id.decchannel,
		svipp->cod_id.dev_ind,
		count,
		payload_buf_len);
#endif
	if (!count) return;

	res = count;
	while (res) {
		if (res > payload_buf_len) {
			memcpy(payload, buffer, payload_buf_len);
			packet.data_len = payload_buf_len + 12;
			res -= payload_buf_len;
			buffer += payload_buf_len;
		} else {
			memcpy(payload, buffer, res);
			packet.data_len = res + 12;
			res = 0;
		}

		rtp_packet_set_version(&packet, 2);
		rtp_packet_set_padding(&packet, 0);
		rtp_packet_set_extension(&packet, 0);
		rtp_packet_set_csrc_count(&packet, 0);

		if (*m_bit == 1 && svip_fd) {
			if (IFX_SUCCESS != vinetic_cod_ctrl(svip_fd , svipp->cod_id.dev_ind,
				svipp->cod_id.decchannel, COD_DEC_START)) {
				ast_log(LOG_ERROR, "Start Channel %d of device %d deccode error \n",
					svipp->cod_id.decchannel, svipp->cod_id.dev_ind);
				return;
			}
			/*start encoder*/
			if (IFX_SUCCESS != vinetic_cod_ctrl(svip_fd , svipp->cod_id.dev_ind,
				svipp->cod_id.encchannel, COD_ENC_START)) {
				ast_log(LOG_ERROR, "Start Channel %d of device %d encode error \n",
					svipp->cod_id.encchannel, svipp->cod_id.dev_ind);
				return;
			}
		}
		rtp_packet_set_marker(&packet, *m_bit);
		*m_bit = 0;

		rtp_packet_set_payload_type(&packet, rtp_pt);
		rtp_packet_set_seq(&packet, *seq_num);
		rtp_packet_set_timestamp(&packet, *ts_current);
		rtp_packet_set_ssrc(&packet, *rtp_ssrc);

		rval =  rtp_packet_send(sock, &packet, svip_saddr);

		if (option_verbose > 10) {
			if (-1 == rval) {
				ast_log(LOG_ERROR, "Failed to write to transcoder: %s\n", strerror(errno));
			}
		}
#if !defined(USE_AST_TS_SEQNUM)
		++*seq_num;
		*ts_current = *ts_current + timestamp_increse;
#endif
		memset(packet.data, 0, packet.data_len);

		p.fd = sock;
		p.events = POLLOUT;
		rval = poll(&p, 1, 50);
	}
}

#if !defined(USE_SINGLE_THREAD)
/*read svip frame to frame ring buffer*/
static void svip_rw_frame_handle_thread (void *arg)
{
	struct codec_svip_pvt *svipp = (struct codec_svip_pvt *)arg;
	struct rtp_packet_t packet = {NULL,0};

	int sock = svipp->rtp_sess.sock;

	void *payload;
	int payload_len;
	struct timeval tv;
	ast_fdset fds;

	struct sockaddr fromaddr;
	socklen_t fromlen = sizeof(fromaddr);

	int res = 0; 
	ast_timer_set_rate(svipp->timer, 1000/10); /*10ms*/
	int timingfd = ast_timer_fd(svipp->timer);
	int timeout = -1;
	svip_frame_t *sframe;

	for (;;) {
		/**for write (frame in) --we dont write frame in thread*/
		/**for read (frame out)*/
		FD_ZERO(&fds);
		FD_SET(sock, &fds);
		tv.tv_sec = 0;
		tv.tv_usec = 0;
		res= ast_select(sock+1, &fds, NULL, NULL, &tv);

		if (res < 0) {
			ast_log(LOG_ERROR, "select error");
			break;
		}
		if (SVIP_FD_ISSET(sock, &fds)) {
			sframe = svip_rb_end_frame_get(svipp->frame_out_srb);
			if (!sframe->buffer) {
				usleep(10000);
				continue;
			}
			packet.data = sframe->buffer;
			packet.data_len = sframe->buf_lens;
			res = recvfrom(sock, packet.data, packet.data_len, 0, &fromaddr, &fromlen);
			packet.data_len = res;
			if (res < 0) {
				ast_log(LOG_ERROR, "rtp_packet_read: recvfrom: %s", strerror(errno));
				break;
			}
			payload = rtp_packet_get_payload(&packet);
			payload_len = rtp_packet_get_payload_len(&packet);
			sframe->frame_offset = payload - packet.data;
			sframe->frame_lens = payload_len;
			svip_rb_end_frame_put(svipp->frame_out_srb);
		}
		ast_waitfor_n_fd(&timingfd, 1, &timeout, NULL);
		ast_timer_ack(svipp->timer, 1);
	}
}
#else
static int notify_svip_rw_thread(void)
{
	char data = 0;
	int res;
	res = write(g_svip_translator_socket[1], &data, sizeof(data));
	return res;
}

static void svip_rw_frame_handle_single_thread (void *arg)
{
	struct rtp_packet packet = {NULL,0};
	struct svip_translator_usage *trans_usage;

	int sock;
	void *payload;
	int payload_len;
	ast_fdset fds;

	struct sockaddr fromaddr;
	socklen_t fromlen = sizeof(fromaddr);

	int res = 0;
	svip_frame_t *sframe;
	int max_fd;

	for (;;) {
		FD_ZERO(&fds);
		FD_SET(g_svip_translator_socket[0], &fds);
		max_fd = g_svip_translator_socket[0];
		
		AST_RWLIST_RDLOCK(&g_svip_translator_list);
		AST_RWLIST_TRAVERSE_SAFE_BEGIN(&g_svip_translator_list, trans_usage, entry) {
			/**for write (frame in) --we dont write frame in thread*/
			/**for read (frame out)*/
			sock = trans_usage->svipp->rtp_sess.sock;
			FD_SET(sock, &fds);
			if (sock > max_fd)
				max_fd = sock;
		  }
		AST_RWLIST_TRAVERSE_SAFE_END;
		AST_RWLIST_UNLOCK(&g_svip_translator_list);

		res= ast_select(max_fd+1, &fds, NULL, NULL, NULL);

		if (res < 0) {
			ast_log(LOG_ERROR, "select error %s, select maxfd %d\n",
				strerror(errno), max_fd);
			sched_yield();
			usleep(2000);
			continue;
		}

		if (SVIP_FD_ISSET(g_svip_translator_socket[0], &fds)) {
			char buf[128];
			int notify_res;
			notify_res = read(g_svip_translator_socket[0], buf, sizeof(buf));
			if (notify_res < 0)
				ast_log(LOG_ERROR, "notify read %s\n", strerror(errno));
		}

		AST_RWLIST_RDLOCK(&g_svip_translator_list);
		AST_RWLIST_TRAVERSE_SAFE_BEGIN(&g_svip_translator_list, trans_usage, entry) {
			/**for write (frame in) --we dont write frame in thread*/
			/**for read (frame out)*/
			sock = trans_usage->svipp->rtp_sess.sock;
			if (SVIP_FD_ISSET(sock, &fds)) {
				sframe = svip_rb_end_frame_get(trans_usage->svipp->frame_out_srb);
				if (!sframe->buffer) {
					continue;
				}
				packet.data = sframe->buffer;
				packet.data_len = sframe->buf_lens;
				res = recvfrom(sock, packet.data, packet.data_len,
					MSG_DONTWAIT, &fromaddr, &fromlen);
				packet.data_len = res;
				if (res < 0 ) {
					if (errno == EAGAIN || errno == EWOULDBLOCK ) {
						ast_log(LOG_WARNING, "rtp_packet_read: recvfrom: %s", strerror(errno));
						continue;
					}
					ast_log(LOG_WARNING, "rtp_packet_read: recvfrom: %s", strerror(errno));
					break;
				}
				payload = rtp_packet_get_payload(&packet);
				payload_len = rtp_packet_get_payload_len(&packet);
				sframe->timestamp = rtp_packet_get_timestamp(&packet);
#if defined(USE_AST_TS_SEQNUM)
				sframe->seq_num = rtp_packet_get_seqnum(&packet);
#endif
				sframe->frame_offset = payload - packet.data;
				sframe->frame_lens = payload_len;
				svip_rb_end_frame_put(trans_usage->svipp->frame_out_srb);
			}
		  }
		AST_RWLIST_TRAVERSE_SAFE_END;
		AST_RWLIST_UNLOCK(&g_svip_translator_list);
		sched_yield();
		usleep(2000);

		if (close_svip_thread) {
			break;
		}
	} 
}
#endif

static int svip_get_ts_ratio(uint32_t ast_fmt)
{
	return (ast_fmt == AST_FORMAT_SLIN16) ? 16 : 8;
}

static int svip_framein(struct ast_trans_pvt *pvt, struct ast_frame *f)
{
	struct codec_svip_pvt *svipp = pvt->pvt;
	void *data_ptr = f->data.ptr;

#if defined(SVIP_DEBUG)
	svip_dump_frame_info(pvt, "svip_framein");
#endif

	if (!f->subclass.format) {
		/* We're just faking a return for calculation purposes. */
		svipp->fake = 2;
		pvt->samples = f->samples;
		return 0;
	}

	if (svipp->softslin && (svipp->fmts.srcfmt&(AST_FORMAT_ULAW|AST_FORMAT_ALAW|AST_FORMAT_SLIN))) {
		if (lintoulaw(pvt, f)) {
			 return -1;
		}
	} else {
		svipp->samples_in_buffer = f->samples;
	}

#if defined(SVIP_DEBUG)
	ast_log(LOG_DEBUG, "enc[%d]-dec[%d] of device [%d]===svip_framein svipp->trsc_pars.required_samples %d len %d \
f->samples %d svipp->samples_in_buffer %d\n",
		svipp->cod_id.encchannel, svipp->cod_id.decchannel, 
		svipp->cod_id.dev_ind, svipp->trsc_pars.required_samples, f->datalen, f->samples, svipp->samples_in_buffer);
#endif

#if defined(USE_AST_TS_SEQNUM)
	if (f->seqno == 0)
		svipp->rtp_sess.seq_num ++;
	else
		svipp->rtp_sess.seq_num = f->seqno;

	if (f->ts == 0)
		svipp->rtp_sess.timestamp += svipp->rtp_sess.timestamp_increse;
	else
		svipp->rtp_sess.timestamp = f->ts * svip_get_ts_ratio(svipp->fmts.srcfmt);

	if (svipp->samples_in_buffer > 0) {
		if (svipp->softslin && (svipp->fmts.srcfmt & (AST_FORMAT_ULAW|AST_FORMAT_ALAW|AST_FORMAT_SLIN))) {
			svip_write_frame(svipp, svipp->ulaw_buffer, svipp->samples_in_buffer);
		} else {
			svip_write_frame(svipp, data_ptr, f->datalen);
		}
		svipp->samples_in_buffer = 0;
	}
#else
	while (svipp->samples_in_buffer >= svipp->trsc_pars.required_samples) {
		if (svipp->softslin && (svipp->fmts.srcfmt&(AST_FORMAT_ULAW|AST_FORMAT_ALAW|AST_FORMAT_SLIN))) {
			svip_write_frame(svipp, svipp->ulaw_buffer, svipp->trsc_pars.required_samples);
			svipp->samples_in_buffer -= svipp->trsc_pars.required_samples;
			if (svipp->samples_in_buffer > 0) {
				/* Shift any remaining bytes down. */
				memmove(svipp->ulaw_buffer, &svipp->ulaw_buffer[svipp->trsc_pars.required_samples],
				svipp->samples_in_buffer);
			}
		} else {
			svip_write_frame(svipp, data_ptr, f->datalen * svipp->trsc_pars.required_samples / f->samples);
			data_ptr += f->datalen * svipp->trsc_pars.required_samples / f->samples;
			svipp->samples_in_buffer -= svipp->trsc_pars.required_samples;
		}
	}
#endif
	pvt->samples += f->samples;
	pvt->datalen = 0;

	return -1;
}

static struct ast_frame *svip_frameout(struct ast_trans_pvt *pvt)
{
	int res = 0;
	struct codec_svip_pvt *svipp = pvt->pvt;
	int samples = 0;
	int count;
	unsigned int ts;
#if defined(USE_AST_TS_SEQNUM)
	unsigned short seq_num;
#endif

#if defined(SVIP_DEBUG)
	svip_dump_frame_info(pvt, "svip_decoder_frameout");
#endif

	if (2 == svipp->fake) {
		struct ast_frame frm = {
			.frametype = AST_FRAME_VOICE,
			.samples = svipp->required_samples,
			.src = pvt->t->name,
		};
		svipp->fake = 1;
		pvt->samples = 0;
		return ast_frisolate(&frm);
	} else if (1 == svipp->fake) {
		pvt->samples = 0;
		svipp->fake = 0;
		return NULL;
	}

	/* Let's check to see if there is a new frame for us.... */
	pvt->f.datalen = 0;
	pvt->f.samples = 0;

//  if(pvt->samples >= svipp->trsc_pars.frameout_samples) {
	if (svipp->softslin && (svipp->fmts.dstfmt&(AST_FORMAT_ULAW|AST_FORMAT_ALAW|AST_FORMAT_SLIN))) {
		res = svip_rb_read(svipp->frame_out_srb, svipp->ulaw_buffer,
			&count, &ts
#if defined(USE_AST_TS_SEQNUM)
		, &seq_num
#endif
		);
		if (res > 0) {
			ulawtolin(pvt, res);
			res = res * 2;
			samples += (svipp->trsc_pars.frameout_samples * count);
		}
	} else {
		res = svip_rb_read(svipp->frame_out_srb, (uint8_t *)pvt->outbuf.c,
			&count, &ts
#if defined(USE_AST_TS_SEQNUM)
			, &seq_num
#endif
			);

		if (res > 0)
			samples += (svipp->trsc_pars.frameout_samples * count);
	}
//   }

#if defined(SVIP_DEBUG)
	ast_log(LOG_DEBUG, "enc[%d]-dec[%d] of device [%d] svip_frameout read len %d"
#if defined(USE_AST_TS_SEQNUM)
		" seq num %u ts %u"
#endif
		"\n", svipp->cod_id.encchannel, svipp->cod_id.decchannel, svipp->cod_id.dev_ind, res
#if defined(USE_AST_TS_SEQNUM)
		, seq_num, ts
#endif
	);
#endif

	if (res < 0) {
		if (EWOULDBLOCK == errno) {
			/* Nothing waiting... */
			return NULL;
		} else {
			ast_log(LOG_ERROR, "Failed to read from transcoder: %s\n", strerror(errno));
			return NULL;
		}
	} if (0 == res) {
		return NULL;
	}

#if defined(USE_AST_TS_SEQNUM)
	pvt->f.seqno = seq_num;
#endif
	pvt->f.ts = ts/svip_get_ts_ratio(svipp->fmts.dstfmt);

	return ast_trans_frameout(pvt, res, samples);
}

static void __svip_destroy(struct ast_trans_pvt *pvt)
{
	struct codec_svip_pvt *svipp = pvt->pvt;
	struct svip_translator_usage *trans_usage;
	struct svip_codec_pair *codec_pair;
	int found_codec_pair =0;
	int n_lbr_rm = 0;
	int n_g711_rm = 0;
	int ret = 0;
	int stop_retry = 3;

	ast_debug(1, "[%p]Vinetic SVIP coder enc[%d]-dec[%d] of device [%d] ---destroy translate\n",
		pvt,
		svipp->cod_id.encchannel,
		svipp->cod_id.decchannel,
		svipp->cod_id.dev_ind);

	AST_RWLIST_WRLOCK(&g_svip_translator_list);
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&g_svip_translator_list, trans_usage, entry) {
		if (trans_usage->svipp != svipp)
			continue;
#if !defined(USE_SINGLE_THREAD)
		if ( svipp->svip_rw_thread &&
			( svipp->svip_rw_thread  != AST_PTHREADT_STOP) &&
			( svipp->svip_rw_thread  != AST_PTHREADT_NULL)) {
			pthread_cancel( svipp->svip_rw_thread );
			pthread_kill( svipp->svip_rw_thread , SIGURG);
			pthread_join( svipp->svip_rw_thread , NULL);
			svipp->svip_rw_thread  = AST_PTHREADT_NULL;
		}
		ast_timer_close(svipp->timer);
		svipp->timer = NULL;
#endif
		AST_RWLIST_REMOVE_CURRENT(entry);
		ast_free(trans_usage);
		break;
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
	AST_RWLIST_UNLOCK(&g_svip_translator_list);

	do {
		ret = vinetic_svip_codec_stop(svip_fd, svipp->cod_id.dev_ind, svipp->cod_id.encchannel,
			svipp->cod_id.decchannel);

		if (ret < 0) {
			if (stop_retry-- < 0)
				break;
			ast_log(LOG_ERROR, "Failed to stop codec: dev %d encch %d decch %d socket port %d\n",
				svipp->cod_id.dev_ind, svipp->cod_id.encchannel, svipp->cod_id.decchannel,
				svipp->rtp_sess.snd_srcport);
			continue;
		} else
			break;
	} while(1);

	if (svipp->cod_id.decchannel == svipp->cod_id.encchannel) {
		vinetic_unmap_cod_ru(svip_fd, svipp->cod_id.dev_ind,
			svipp->cod_id.decchannel,svipp->cod_id.encchannel);
		svip_user_rtp_release(svipp->rtp_sess.sock, &svipp->rtp_sess.rtp_pkt_snd,
			&svipp->rtp_sess.rtp_pkt_rcv);
		ast_put_svip_codec(svipp->cod_id.dev_ind, svipp->cod_id.encchannel);
		n_lbr_rm ++;
	} else {
		AST_LIST_LOCK(&svip_inused_codec_pairs);
		AST_LIST_TRAVERSE_SAFE_BEGIN(&svip_inused_codec_pairs, codec_pair, entry) {
		if ((codec_pair->dev_ind == svipp->cod_id.dev_ind) &&
			(codec_pair->decchannel == svipp->cod_id.encchannel) && 
			(codec_pair->encchannel == svipp->cod_id.decchannel)) {
				found_codec_pair = 1;
			}
		  }
		AST_LIST_TRAVERSE_SAFE_END;
		AST_LIST_UNLOCK(&svip_inused_codec_pairs);

		AST_LIST_LOCK(&svip_inused_codec_pairs);
		AST_RWLIST_TRAVERSE_SAFE_BEGIN(&svip_inused_codec_pairs, codec_pair, entry) {
			if ((codec_pair->dev_ind == svipp->cod_id.dev_ind) &&
				(codec_pair->decchannel == svipp->cod_id.decchannel) && 
				(codec_pair->encchannel == svipp->cod_id.encchannel)) {
				if (found_codec_pair) {
					/*found another coder_pairs of svip coders  is inused*/
					/*add it to the list--svip_rem_codec_pairs*/
					AST_RWLIST_REMOVE_CURRENT(entry);
					AST_LIST_LOCK(&svip_rem_codec_pairs);
					AST_LIST_INSERT_TAIL(&svip_rem_codec_pairs, codec_pair, entry);
					AST_LIST_UNLOCK(&svip_rem_codec_pairs);
				} else {
					 /*another coder_pairs pairs not inuse release all*/
					vinetic_unmap_cod_ru(svip_fd, svipp->cod_id.dev_ind,
						svipp->cod_id.decchannel,svipp->cod_id.encchannel);
					svip_user_rtp_release(svipp->rtp_sess.sock, &svipp->rtp_sess.rtp_pkt_snd,
						&svipp->rtp_sess.rtp_pkt_rcv);
					AST_LIST_REMOVE_CURRENT(entry);
					ast_free(codec_pair);

					AST_LIST_LOCK(&svip_rem_codec_pairs);
					AST_LIST_TRAVERSE_SAFE_BEGIN(&svip_rem_codec_pairs, codec_pair, entry) {
						if ((codec_pair->dev_ind == svipp->cod_id.dev_ind) &&
							(codec_pair->decchannel == svipp->cod_id.encchannel) && 
							(codec_pair->encchannel == svipp->cod_id.decchannel)) {
							svip_user_rtp_release(codec_pair->rtp_sess.sock, &codec_pair->rtp_sess.rtp_pkt_snd,
								&codec_pair->rtp_sess.rtp_pkt_rcv);

							ast_put_svip_codec(svipp->cod_id.dev_ind, svipp->cod_id.decchannel);
							ast_put_svip_codec(svipp->cod_id.dev_ind, svipp->cod_id.encchannel);
							n_lbr_rm++;
							n_g711_rm++;
							AST_LIST_REMOVE_CURRENT(entry);
							ast_free(codec_pair);
						}
					  }
					AST_LIST_TRAVERSE_SAFE_END;
					AST_LIST_UNLOCK(&svip_rem_codec_pairs);
				};
			}
		}
		AST_LIST_TRAVERSE_SAFE_END;
		AST_LIST_UNLOCK(&svip_inused_codec_pairs);
	}

	svip_rb_free(svipp->frame_out_srb);
	ast_free(svipp->frame_out_srb);

	ast_atomic_fetchadd_int(&devices.channels.coder_lbr, -n_lbr_rm);
	ast_atomic_fetchadd_int(&devices.channels.coder_g711, -n_g711_rm);
}

static void svip_destroy(struct ast_trans_pvt *pvt)
{
	sigset_t set, oset;

	/* Block signals; make sure svip resource release at process exit. */
	(void)sigfillset(&set);
	(void)sigprocmask(SIG_BLOCK, &set, &oset);
#ifdef NEW_DESTROY_LOCK
	if (!svip_mutex_lock(&tr_lock)) {
#endif
		__svip_destroy(pvt);
#ifdef NEW_DESTROY_LOCK
		svip_mutex_unlock(&tr_lock);
	} else
		ast_log(LOG_ERROR, "SVIP Critical error occured: cannot get lock.\n");
#endif
	(void)sigprocmask(SIG_SETMASK, &oset, (sigset_t *)NULL);
	svip_callstat.svip_dstr_total++;
}

static unsigned char get_sample_accord_format(uint32_t ast_fmt)
{
	switch(ast_fmt) {
	/*! G.723.1 compression */
	case AST_FORMAT_G723:
		return 240;
	/*! G.729A audio */
	case AST_FORMAT_G729:
		return 160;
	/*! Raw mu-law data (G.711) */
	case AST_FORMAT_ULAW:
		return 160;
	/*! Raw A-law data (G.711) */
	case AST_FORMAT_ALAW:
		return 160;
	/*! G.722 */
	case AST_FORMAT_G722:
		return 160;
	/*! iLBC Free Compression */
	case AST_FORMAT_ILBC:
		if (ilbc_sendrate == ILBC_SENDRATE_133)
			return 240;
	else
		return 160;
	/*! ADPCM (G.726, 32kbps, RFC3551 codeword packing) */
	/*! ADPCM (G.726, 32kbps, AAL2 codeword packing) */
	case AST_FORMAT_G726:
	case AST_FORMAT_G726_AAL2:
		return 160;
	/** Linear codec, 16 bits 8 kHz. */
	case AST_FORMAT_SLIN:
		return 160;
	default:
		break;
	}

	return 160;
}

static unsigned char get_payload_type_accord_format(uint32_t ast_fmt)
{
	switch(ast_fmt) {
	/*! G.723.1 compression */
	case AST_FORMAT_G723:
		return 4;
	/*! G.729A audio */
	case AST_FORMAT_G729:
		return 18;
	/*! Raw mu-law data (G.711) */
	case AST_FORMAT_ULAW:
		return 0;
	/*! Raw A-law data (G.711) */
	case AST_FORMAT_ALAW:
		return 8;
	/*! G.722 */
	case AST_FORMAT_G722:
		return 9;
	/*! iLBC Free Compression */
	case AST_FORMAT_ILBC:
		return 103;
	/*! ADPCM (G.726, 32kbps, RFC3551 codeword packing) */
	/*! ADPCM (G.726, 32kbps, AAL2 codeword packing) */
	case AST_FORMAT_G726:
	case AST_FORMAT_G726_AAL2:
		return 100;
	/** Linear codec, 16 bits 8 kHz. */
	case AST_FORMAT_SLIN:
		return 113;
	default:
		break;
	}

	return 0;
}

static int  ast_to_tapi_format (uint32_t ast_fmt)
{
	switch(ast_fmt) {
	/*! G.723.1 compression */
	case AST_FORMAT_G723:
		if (g723_sendrate == G723_SENDRATE_53)
			return IFX_TAPI_ENC_TYPE_G723_53;
	else
		return IFX_TAPI_ENC_TYPE_G723_63;
	/*! G.729A audio */
	case AST_FORMAT_G729:
		return IFX_TAPI_ENC_TYPE_G729;
	/*! Raw mu-law data (G.711) */
	case AST_FORMAT_ULAW:
		return IFX_TAPI_ENC_TYPE_MLAW;
	/*! Raw A-law data (G.711) */
	case AST_FORMAT_ALAW:
		return IFX_TAPI_ENC_TYPE_ALAW;
	/*! ADPCM (G.726, 32kbps, AAL2 codeword packing) */
	/*! ADPCM (G.726, 32kbps, RFC3551 codeword packing) */
	case AST_FORMAT_G726_AAL2:
	case AST_FORMAT_G726:
		return IFX_TAPI_ENC_TYPE_G726_32;
	/*! G.722 */
	case AST_FORMAT_G722:
		return IFX_TAPI_ENC_TYPE_G722_64;
	/*! G.722.1 (also known as Siren7, 32kbps assumed) */
	case AST_FORMAT_SIREN7:
		return IFX_TAPI_ENC_TYPE_G7221_32;
	/*! Raw 16-bit Signed Linear (8000 Hz) PCM */
	case AST_FORMAT_SLIN:
		return IFX_TAPI_ENC_TYPE_LIN16_8;
	/*! Raw 16-bit Signed Linear (16000 Hz) PCM */
	case AST_FORMAT_SLIN16:
		return IFX_TAPI_ENC_TYPE_LIN16_16; 
	/*! iLBC Free Compression */
	case AST_FORMAT_ILBC:
		if(ilbc_sendrate == ILBC_SENDRATE_133)
			return IFX_TAPI_COD_TYPE_ILBC_133;
	else
		return IFX_TAPI_COD_TYPE_ILBC_152; 
	default:
		break;
	}

	return -1;
}

static inline int get_len_accord_sample_format(uint32_t format, int samples)
{
	int len = 0;

	switch(format) {
	case AST_FORMAT_G723:
		len = (samples / 240) * 20;
		break;
	case AST_FORMAT_G726:
	case AST_FORMAT_G726_AAL2:
		len = samples / 2;
		break;
	case AST_FORMAT_G729:
		len = samples / 8;
		break;
	case AST_FORMAT_ILBC:
		len = (samples / 240) * 50;
		break;
	case AST_FORMAT_SLIN:
	case AST_FORMAT_SLIN16:
		len = samples * 2;
		break;
	case AST_FORMAT_ULAW:
	case AST_FORMAT_ALAW:
		len = samples;
		break;
	default:
		len = samples;
	}

	return len;
}

static inline void figure_out_transc_params_for_svip (uint32_t srcfmt, uint32_t dstfmt,
	struct svip_trans_par *trsc_par) 
{

	/**
	*    figure out payload type, required_samples and timestamp increasement for tanslator framein
	*    figure out  frame length, svip encode type and frameout_samples for tanslator frameout
	*/
	trsc_par->p_type = get_payload_type_accord_format(srcfmt);
	trsc_par->required_samples = get_sample_accord_format(srcfmt);

	if (dstfmt & (AST_FORMAT_ILBC | AST_FORMAT_G723)) {
		trsc_par->enc_lens = 30;
	} else {
		trsc_par->enc_lens = 20;
	}
	trsc_par->p_lens = get_len_accord_sample_format(srcfmt, trsc_par->required_samples);
	if (srcfmt & AST_FORMAT_G723) {
		trsc_par->p_lens = G723_FRAME_LEN_63;  /*max payload len*/
	}
	trsc_par->ts_inc = trsc_par->required_samples;
	trsc_par->frameout_samples = trsc_par->enc_lens * 8; 
	trsc_par->enc_type = ast_to_tapi_format(dstfmt);
}

#if defined(SVIP_DEBUG)
static void dump_svip_coder_debug_info (struct codec_svip_pvt *svipp)
{
	ast_debug(1, "\n \
		Config Codec device: %d \n \
		Config encode channel: %d \n \
		Config decode channel: %d \n \
		enctype: %d \n \
		Length in milliseconds (svip encode): %d \n \
		rtp payload buffer length in bytes (svip framein) %d \n \
		required_samples %d \n \
		frameout_samples %d \n \
		timestamp increase %d \n \
		snd_dstport: %d \n \
		rcv_dstport:%d \n \
		snd_srcport:%d \n",
		svipp->cod_id.dev_ind, svipp->cod_id.encchannel,
		svipp->cod_id.decchannel, ast_to_tapi_format(svipp->fmts.dstfmt),
		svipp->trsc_pars.enc_lens, svipp->trsc_pars.p_lens,
		svipp->trsc_pars.required_samples, svipp->trsc_pars.frameout_samples,
		svipp->trsc_pars.ts_inc, svipp->rtp_sess.snd_dstport,
		svipp->rtp_sess.rcv_dstport, svipp->rtp_sess.snd_srcport);
}
#endif

static inline void copy_codec_pair_to_svipp (struct codec_svip_pvt *svipp, 
	const struct svip_codec_pair *cod_pair)
{
	svipp->cod_id.dev_ind = cod_pair->dev_ind;
	svipp->cod_id.decchannel = cod_pair->decchannel;
	svipp->cod_id.encchannel = cod_pair->encchannel;
	svipp->fmts.srcfmt = cod_pair->fmts.srcfmt;
	svipp->fmts.dstfmt = cod_pair->fmts.dstfmt;

	svipp->rtp_sess.snd_dstport = cod_pair->rtp_sess.snd_dstport;
	svipp->rtp_sess.rcv_dstport = cod_pair->rtp_sess.rcv_dstport;
	svipp->rtp_sess.snd_srcport = cod_pair->rtp_sess.snd_srcport;
	memcpy(&svipp->rtp_sess.svip_saddr, &cod_pair->rtp_sess.svip_saddr, sizeof(struct sockaddr_in));
	memcpy(&svipp->rtp_sess.rcv_saddr, &cod_pair->rtp_sess.rcv_saddr, sizeof(struct sockaddr_in));
	svipp->rtp_sess.sock =  cod_pair->rtp_sess.sock;
	svipp->rtp_sess.rtp_pkt_snd = cod_pair->rtp_sess.rtp_pkt_snd;
	svipp->rtp_sess.rtp_pkt_rcv = cod_pair->rtp_sess.rtp_pkt_rcv;
	svipp->rtp_sess.rtp_ssrc = cod_pair->rtp_sess.rtp_ssrc;
	svipp->rtp_sess.rtp_pt = cod_pair->rtp_sess.rtp_pt;
	svipp->rtp_sess.seq_num = cod_pair->rtp_sess.seq_num;
	svipp->rtp_sess.timestamp = cod_pair->rtp_sess.timestamp;
	svipp->rtp_sess.timestamp_increse = cod_pair->rtp_sess.timestamp_increse;
	svipp->rtp_sess.m_bit = cod_pair->rtp_sess.m_bit;
	svipp->rtp_sess.rtpinf = cod_pair->rtp_sess.rtpinf;

	svipp->trsc_pars.p_type = cod_pair->trsc_pars.p_type;
	svipp->trsc_pars.p_lens = cod_pair->trsc_pars.p_lens;
	svipp->trsc_pars.enc_lens = cod_pair->trsc_pars.enc_lens;
	svipp->trsc_pars.enc_type = cod_pair->trsc_pars.enc_type;
	svipp->trsc_pars.required_samples = cod_pair->trsc_pars.required_samples;
	svipp->trsc_pars.frameout_samples = cod_pair->trsc_pars.frameout_samples;
	svipp->trsc_pars.ts_inc =  cod_pair->trsc_pars.ts_inc;
}

static inline void copy_codec_svipp_to_pair (struct svip_codec_pair *cod_pair,
    const struct codec_svip_pvt *svipp)
{
	cod_pair->dev_ind = svipp->cod_id.dev_ind;
	cod_pair->decchannel = svipp->cod_id.decchannel;
	cod_pair->encchannel = svipp->cod_id.encchannel;
	cod_pair->fmts.srcfmt = svipp->fmts.srcfmt;
	cod_pair->fmts.dstfmt = svipp->fmts.dstfmt;

	cod_pair->rtp_sess.snd_dstport = svipp->rtp_sess.snd_dstport;
	cod_pair->rtp_sess.rcv_dstport = svipp->rtp_sess.rcv_dstport;
	cod_pair->rtp_sess.snd_srcport = svipp->rtp_sess.snd_srcport;
	memcpy(&cod_pair->rtp_sess.svip_saddr, &svipp->rtp_sess.svip_saddr, sizeof(struct sockaddr_in));
	memcpy(&cod_pair->rtp_sess.rcv_saddr, &svipp->rtp_sess.rcv_saddr, sizeof(struct sockaddr_in));
	cod_pair->rtp_sess.sock = svipp->rtp_sess.sock;
	cod_pair->rtp_sess.rtp_pkt_snd = svipp->rtp_sess.rtp_pkt_snd;
	cod_pair->rtp_sess.rtp_pkt_rcv = svipp->rtp_sess.rtp_pkt_rcv;
	cod_pair->rtp_sess.rtp_ssrc = svipp->rtp_sess.rtp_ssrc;
	cod_pair->rtp_sess.rtp_pt = svipp->rtp_sess.rtp_pt;
	cod_pair->rtp_sess.seq_num = svipp->rtp_sess.seq_num;
	cod_pair->rtp_sess.timestamp = svipp->rtp_sess.timestamp;
	cod_pair->rtp_sess.timestamp_increse = svipp->rtp_sess.timestamp_increse;
	cod_pair->rtp_sess.m_bit = svipp->rtp_sess.m_bit;
	cod_pair->rtp_sess.rtpinf = svipp->rtp_sess.rtpinf;

	cod_pair->trsc_pars.p_type = svipp->trsc_pars.p_type;
	cod_pair->trsc_pars.p_lens = svipp->trsc_pars.p_lens;
	cod_pair->trsc_pars.enc_lens = svipp->trsc_pars.enc_lens;
	cod_pair->trsc_pars.enc_type = svipp->trsc_pars.enc_type;
	cod_pair->trsc_pars.required_samples = svipp->trsc_pars.required_samples;
	cod_pair->trsc_pars.frameout_samples = svipp->trsc_pars.frameout_samples;
	cod_pair->trsc_pars.ts_inc = svipp->trsc_pars.ts_inc;
}

static int svip_translate(struct ast_trans_pvt *pvt, uint32_t dst_svip_fmt, uint32_t src_svip_fmt)
{
	int ii;
	int has_get_coders = 0;
	int use_coder_pairs = 0;
	int use_old_coder_pair = 0;
	int lbr_ch;
	int g711_ch;
	int n_ringbuf;
	int codec_udpdst;
	int codec_udprcv;
	int codec_udpsrc;
	int error_flag;

	struct codec_svip_pvt *svipp = pvt->pvt;
	struct svip_translator_usage *trans_entry;
	struct svip_codec_pair *codec_pair;
	struct svip_codec_pair *codec_pair_entry0 = NULL;
	struct svip_codec_pair *codec_pair_entry1 = NULL;

	if (!svip_fd) {
		return -1;
	}

	svipp->fmts.srcfmt = src_svip_fmt;
	svipp->fmts.dstfmt = dst_svip_fmt;

	ast_assert(pvt->f.subclass.format == NULL);
	pvt->f.subclass.format = 
		ao2_bump(ast_format_compatibility_bitfield2format(svipp->fmts.dstfmt));

#if defined(SOFTSLIN_AULAW)
	if (AST_FORMAT_SLIN == svipp->fmts.srcfmt) {
		ast_debug(1, "Using soft_slin support on source\n");
		svipp->softslin = 1;
		svipp->fmts.srcfmt = AST_FORMAT_ULAW;
	} else if (AST_FORMAT_SLIN == svipp->fmts.dstfmt) {
		ast_debug(1, "Using soft_slin support on destination\n");
		svipp->softslin = 1;
		svipp->fmts.dstfmt = AST_FORMAT_ULAW;
	}
#else
	svipp->softslin = 0;
#endif

	if ((svipp->fmts.srcfmt&(AST_FORMAT_ULAW|AST_FORMAT_ALAW|AST_FORMAT_SLIN))||
		(svipp->fmts.dstfmt&(AST_FORMAT_ULAW|AST_FORMAT_ALAW|AST_FORMAT_SLIN))) {
#if 0   /*print coder pair rem list*/
		ast_debug(1,"coder pair rem list:\n");
		AST_LIST_LOCK(&svip_rem_codec_pairs);
		AST_LIST_TRAVERSE_SAFE_BEGIN(&svip_rem_codec_pairs, codec_pair, entry) {
			ast_debug(1,"%s(%x)---->%s(%x) enc[%d] dec[%d] addr %p\n",
				get_svip_codec_name(codec_pair->fmts.srcfmt),
				codec_pair->fmts.srcfmt,
				get_svip_codec_name(codec_pair->fmts.dstfmt),
				codec_pair->fmts.dstfmt,
				codec_pair->encchannel, codec_pair->decchannel,
				codec_pair);
		  }
		AST_LIST_TRAVERSE_SAFE_END;
		AST_LIST_UNLOCK(&svip_rem_codec_pairs);
#endif
		AST_LIST_LOCK(&svip_inused_codec_pairs);
		AST_LIST_LOCK(&svip_rem_codec_pairs);
		AST_LIST_TRAVERSE_SAFE_BEGIN(&svip_rem_codec_pairs, codec_pair, entry) {
			if ((codec_pair->fmts.srcfmt == svipp->fmts.srcfmt)&&
				(codec_pair->fmts.dstfmt == svipp->fmts.dstfmt)) {
				copy_codec_pair_to_svipp(svipp, codec_pair);
				use_old_coder_pair = 1;
				has_get_coders = 1;
				use_coder_pairs = 1;
				AST_LIST_REMOVE_CURRENT(entry);
				AST_LIST_INSERT_TAIL(&svip_inused_codec_pairs, codec_pair, entry);
				break;
			}
		  }
		AST_LIST_TRAVERSE_SAFE_END;
		AST_LIST_UNLOCK(&svip_rem_codec_pairs);
		AST_LIST_UNLOCK(&svip_inused_codec_pairs);

		if (0 == has_get_coders) {
			ii = ast_choose_optimal_dev();
			lbr_ch = ast_get_svip_codec(ii, COD_LBR);
			if (lbr_ch < 0)
				ast_debug(1, "can not get coder LBR channel of device [%d]\n",ii);
			else {
				use_coder_pairs = 1;
				g711_ch = ast_get_svip_codec(ii, COD_G711_ONLY);
				if (g711_ch < 0) {
					ast_debug(1, "can not get coder G711 channel of device [%d]\n",ii);
					use_coder_pairs = 0;
					g711_ch = lbr_ch;
				}
				if(svipp->fmts.srcfmt&(AST_FORMAT_ULAW|AST_FORMAT_ALAW|AST_FORMAT_SLIN)){
					svipp->cod_id.encchannel = lbr_ch;
					svipp->cod_id.decchannel = g711_ch;
				} else {
					svipp->cod_id.encchannel = g711_ch;
					svipp->cod_id.decchannel = lbr_ch;
				}
				has_get_coders = 1;
				svipp->cod_id.dev_ind = ii;

				/*save remnant codec pair*/
				if (use_coder_pairs) {
					if (!(codec_pair_entry0 = ast_calloc(1, sizeof(*codec_pair_entry0)))) {
						ast_put_svip_codec(ii, lbr_ch);
						ast_put_svip_codec(ii, g711_ch);
						return -1;
					}
					codec_pair_entry0->dev_ind = svipp->cod_id.dev_ind;
					codec_pair_entry0->decchannel = svipp->cod_id.encchannel;
					codec_pair_entry0->encchannel = svipp->cod_id.decchannel;
					codec_pair_entry0->fmts.srcfmt = svipp->fmts.dstfmt;
					codec_pair_entry0->fmts.dstfmt = svipp->fmts.srcfmt;
					codec_pair_entry0->cod_pair_stat = COD_PAIR_UNCONFIG; //unconfiged
				}
			}
		}
	} else {
		ii = ast_choose_optimal_dev();
		lbr_ch = ast_get_svip_codec(ii, COD_LBR);
		if (lbr_ch < 0) {
			ast_debug(1, "can not get coder LBR channel of device [%d]\n",ii);
		} else {
			svipp->cod_id.dev_ind = ii;
			svipp->cod_id.encchannel = lbr_ch;
			svipp->cod_id.decchannel = lbr_ch;
			has_get_coders = 1;
			use_coder_pairs = 0;
		}
	}

	if (0 == has_get_coders) {
		ast_log(LOG_WARNING, "Cannot get a idle COD-RM of SVIP devices");
		return -1;
	}

	ast_debug(1, "[%p]Opening SVIP transcoder from"
		" %s to %s: encoder chan[%d], decoder chan[%d] of device[%d]\n",
		pvt,
		get_svip_codec_name(svipp->fmts.srcfmt),
		get_svip_codec_name(svipp->fmts.dstfmt),
		svipp->cod_id.encchannel,
		svipp->cod_id.decchannel,
		svipp->cod_id.dev_ind);

	if (use_coder_pairs && use_old_coder_pair) {/*had been configed */
#if defined(SVIP_DEBUG)
		dump_svip_coder_debug_info (svipp);
#endif
		ast_debug(1, "use remnant coder pair\n");
	} else {
		/* initialize socket address for receive and send rtp packets */
		figure_out_transc_params_for_svip(svipp->fmts.srcfmt, svipp->fmts.dstfmt, &svipp->trsc_pars);
		codec_udprcv = vinetic_svip_get_cod_rcvport(svipp->cod_id.dev_ind, svipp->cod_id.decchannel);
		if (IFX_SUCCESS != svip_user_rtp_init(svipp->cod_id.dev_ind,
			codec_udprcv,
 #if defined(USE_AST_TS_SEQNUM)
			RTP_MTU-12,
 #else
			svipp->trsc_pars.p_lens, 
 #endif
			svipp->trsc_pars.p_type, 
			svipp->trsc_pars.ts_inc,
			&svipp->rtp_sess)) {
			ast_log(LOG_ERROR, "Svip Codec init error\n");
			error_flag = 0;
			goto cfg_error_handle;
		}

		if (use_coder_pairs && (COD_PAIR_UNCONFIG == codec_pair_entry0->cod_pair_stat)) {
			/*config another translator which use the coder pair*/
			figure_out_transc_params_for_svip(codec_pair_entry0->fmts.srcfmt, codec_pair_entry0->fmts.dstfmt,
			    &codec_pair_entry0->trsc_pars);
			codec_pair_entry0->trsc_pars.required_samples = codec_pair_entry0->trsc_pars.required_samples;
			codec_pair_entry0->trsc_pars.frameout_samples = codec_pair_entry0->trsc_pars.frameout_samples;
			codec_udprcv = vinetic_svip_get_cod_rcvport(codec_pair_entry0->dev_ind, codec_pair_entry0->decchannel);
			if (IFX_SUCCESS != svip_user_rtp_init(codec_pair_entry0->dev_ind,
				codec_udprcv,
 #if defined(USE_AST_TS_SEQNUM)
				RTP_MTU-12,
 #else
				codec_pair_entry0->trsc_pars.p_lens,
 #endif 
				codec_pair_entry0->trsc_pars.p_type, 
				codec_pair_entry0->trsc_pars.ts_inc,
				&codec_pair_entry0->rtp_sess)) {
				ast_log(LOG_ERROR, "Svip Codec pair init error\n");
				error_flag = 1;
				goto cfg_error_handle;
			}
			codec_pair_entry0->cod_pair_stat = COD_PAIR_CONFIG_S1;
		}

#if defined(SVIP_DEBUG)
		dump_svip_coder_debug_info (svipp);
#endif
		if (use_coder_pairs) {
			codec_udpdst = svipp->rtp_sess.rcv_dstport;
			codec_udprcv = codec_pair_entry0->rtp_sess.snd_dstport;
			codec_udpsrc = codec_udprcv;
		} else {
			codec_udpdst = svipp->rtp_sess.rcv_dstport;
			codec_udprcv = svipp->rtp_sess.snd_dstport;
			codec_udpsrc = codec_udprcv;
		}

		/*Vinetic Svip Codec Configuration*/
		if (IFX_SUCCESS != vinetic_codec_cfg(svip_fd,
			svipp->cod_id.dev_ind,
			svipp->cod_id.encchannel,
			svipp->trsc_pars.enc_type,
			svipp->trsc_pars.enc_lens,
			codec_udpdst,
			codec_udprcv,
			codec_udpsrc)) {
			ast_log(LOG_ERROR, "Vinetic-Svip codec configure error \n");
			error_flag = 1;
			goto cfg_error_handle;
		}

		if (use_coder_pairs) {
			if (IFX_SUCCESS != vinetic_codec_cfg(svip_fd,
				codec_pair_entry0->dev_ind,
				codec_pair_entry0->encchannel,
				codec_pair_entry0->trsc_pars.enc_type,
				codec_pair_entry0->trsc_pars.enc_lens,
				codec_pair_entry0->rtp_sess.rcv_dstport,
				svipp->rtp_sess.snd_dstport,
				svipp->rtp_sess.snd_dstport)) {
				ast_log(LOG_ERROR, "Vinetic-Svip codec pair configure error \n");
				error_flag = 1;
				goto cfg_error_handle;
			}
			codec_pair_entry0->cod_pair_stat = COD_PAIR_CONFIG_S2;
		}

		if (IFX_SUCCESS != vinetic_map_cod_ru(svip_fd,
			svipp->cod_id.dev_ind,
			svipp->cod_id.decchannel,
			svipp->cod_id.encchannel)) {
			ast_log(LOG_ERROR, "Vinetic-Svip codec map error \n");
			error_flag = 1;
			goto cfg_error_handle;
		};
	}

	n_ringbuf = RINGBUF_LEN/ svipp->trsc_pars.enc_lens;

	if (!(svipp->frame_out_srb = ast_calloc(1, sizeof(svip_rbuf_t)))) {
		if(use_old_coder_pair)
			return -1;
		error_flag = 2;
		goto cfg_error_handle;
	}

	if (svip_rb_allocInit(svipp->frame_out_srb, n_ringbuf, RTP_MTU) < 0) {
		ast_log(LOG_ERROR, "Alloc ring buffers for svip frame receive error\n");
		ast_free(svipp->frame_out_srb);
		if(use_old_coder_pair)
			return -1;
		error_flag = 2;
		goto cfg_error_handle;
	}

	if (!(trans_entry = ast_calloc(1, sizeof(*trans_entry)))) {
		svip_rb_free(svipp->frame_out_srb);
		ast_free(svipp->frame_out_srb);
		if(use_old_coder_pair)
			return -1;
		error_flag = 2;
		goto cfg_error_handle;
	}

#if !defined(USE_SINGLE_THREAD)
	if (!(svipp->timer = ast_timer_open())) {
		ast_free(trans_entry);
		svip_rb_free(svipp->frame_out_srb);
		ast_free(svipp->frame_out_srb);
		if (use_old_coder_pair)
			return -1;
		error_flag = 2;
		goto cfg_error_handle;
	}

	svipp->svip_rw_thread = AST_PTHREADT_NULL;
	ii = ast_pthread_create(&svipp->svip_rw_thread, NULL, (void *)svip_rw_frame_handle_thread, svipp);
	if (ii < 0) {
		ast_debug(1, "svip read thread create failed\n");
		ast_timer_close(svipp->timer);
		ast_free(trans_entry);
		svip_rb_free(svipp->frame_out_srb);
		ast_free(svipp->frame_out_srb);
		if (use_old_coder_pair)
			return -1;
		error_flag = 2;
		goto cfg_error_handle;
	}
#endif

	if (use_coder_pairs && !use_old_coder_pair) {
		AST_LIST_LOCK(&svip_rem_codec_pairs);
		AST_LIST_INSERT_TAIL(&svip_rem_codec_pairs, codec_pair_entry0, entry);
		AST_LIST_UNLOCK(&svip_rem_codec_pairs);
		if (!(codec_pair_entry1 = ast_calloc(1, sizeof(*codec_pair_entry1)))) {
#if !defined(USE_SINGLE_THREAD)
			pthread_cancel( svipp->svip_rw_thread );
			pthread_kill( svipp->svip_rw_thread , SIGURG);
			pthread_join( svipp->svip_rw_thread , NULL);
			svipp->svip_rw_thread  = AST_PTHREADT_NULL;
			ast_timer_close(svipp->timer);
			svipp->timer = NULL;
#endif
			ast_free(trans_entry);
			svip_rb_free(svipp->frame_out_srb);
			ast_free(svipp->frame_out_srb);
			error_flag = 2;
			goto cfg_error_handle;
		}
		copy_codec_svipp_to_pair(codec_pair_entry1, svipp);
		AST_LIST_LOCK(&svip_inused_codec_pairs);
		AST_LIST_INSERT_TAIL(&svip_inused_codec_pairs, codec_pair_entry1, entry);
		AST_LIST_UNLOCK(&svip_inused_codec_pairs);
	}

	trans_entry->svipp = svipp;
	AST_RWLIST_WRLOCK(&g_svip_translator_list);
	AST_RWLIST_INSERT_TAIL(&g_svip_translator_list, trans_entry, entry);
	AST_RWLIST_UNLOCK(&g_svip_translator_list);

#if defined(USE_SINGLE_THREAD)
	/* notify svip_rw_frame_handle_single_thread */
	notify_svip_rw_thread();
#endif

	if (use_coder_pairs) {
		if (0 == use_old_coder_pair) {
			ast_atomic_fetchadd_int(&devices.channels.coder_lbr, +1);
			ast_atomic_fetchadd_int(&devices.channels.coder_g711, +1);
		}
	} else {
		ast_atomic_fetchadd_int(&devices.channels.coder_lbr, +1);
	}

	return 0;

cfg_error_handle:
	if ((2 == error_flag) && (0 == use_old_coder_pair)) {
		vinetic_unmap_cod_ru(svip_fd, 
		svipp->cod_id.dev_ind,
		svipp->cod_id.decchannel,
		svipp->cod_id.encchannel);
	}
	if (error_flag >= 1) {
		svip_user_rtp_release(svipp->rtp_sess.sock, &svipp->rtp_sess.rtp_pkt_snd,
			&svipp->rtp_sess.rtp_pkt_rcv);
		if (use_coder_pairs) {
			if (COD_PAIR_UNCONFIG != codec_pair_entry0->cod_pair_stat) {
				svip_user_rtp_release(codec_pair_entry0->rtp_sess.sock, 
					&codec_pair_entry0->rtp_sess.rtp_pkt_snd,
					&codec_pair_entry0->rtp_sess.rtp_pkt_rcv);
			}
		}
	}
	if (error_flag >= 0) {
		if (use_coder_pairs) { 
			ast_put_svip_codec(svipp->cod_id.dev_ind, svipp->cod_id.decchannel);
			ast_put_svip_codec(svipp->cod_id.dev_ind, svipp->cod_id.encchannel);
			ast_free(codec_pair_entry0);
		} else {
			ast_put_svip_codec(svipp->cod_id.dev_ind, svipp->cod_id.encchannel);
		}
	}
	return -1;
}

static int svip_new(struct ast_trans_pvt *pvt)
{
	int ret;
	sigset_t set, oset;
	struct translator *zt = container_of(pvt->t, struct translator, t);

	ast_debug(1, "[%p]Vinetic SVIP coder ---new translate\n", pvt);
	svip_callstat.svip_new_total++;
	/* Block signals; make sure svip resource alloc at process exit. */
	(void)sigfillset(&set);
	(void)sigprocmask(SIG_BLOCK, &set, &oset);
#ifdef NEW_DESTROY_LOCK
	if (svip_mutex_lock(&tr_lock)) {
		svip_callstat.svip_new_fail++;
		ast_log(LOG_ERROR, "SVIP Critical error occured: cannot get lock.\n");
		return -1;
	}
#endif
	ret = svip_translate(pvt, zt->dst_svip_fmt, zt->src_svip_fmt);
#ifdef NEW_DESTROY_LOCK
	svip_mutex_unlock(&tr_lock);
#endif
	(void)sigprocmask(SIG_SETMASK, &oset, (sigset_t *)NULL);
	if (ret)
		svip_callstat.svip_new_fail++;
	else
		svip_callstat.svip_new_succ++;

	return ret;
}

static struct ast_frame *fakesrc_sample(void)
{
	/* Don't bother really trying to test hardware ones. */
	static struct ast_frame f = {
		.frametype = AST_FRAME_VOICE,
		.samples = 160,
		.src = __PRETTY_FUNCTION__
	};

	return &f;
}

static int register_translator(uint32_t dst_svip_fmt, uint32_t src_svip_fmt)
{
	const struct ast_codec *dst_codec;
	const struct ast_codec *src_codec;
	struct translator *zt;
	int res;

	dst_codec = get_svip_codec(dst_svip_fmt);
	src_codec = get_svip_codec(src_svip_fmt);
	if (!dst_codec || !src_codec) {
		return -1;
	}

	if (!(zt = ast_calloc(1, sizeof(*zt)))) {
		return -1;
	}

	zt->src_svip_fmt = src_svip_fmt;
	zt->dst_svip_fmt = dst_svip_fmt;
	snprintf((char *) (zt->t.name), sizeof(zt->t.name), "svip_%s_to_%s",
		 src_codec->name, dst_codec->name);

	memcpy(&zt->t.src_codec, src_codec, sizeof(*src_codec));
	memcpy(&zt->t.dst_codec, dst_codec, sizeof(*dst_codec));

	zt->t.buf_size = BUFFER_SIZE;
	zt->t.framein = svip_framein;
	zt->t.frameout = svip_frameout;
	zt->t.destroy = svip_destroy;
	zt->t.buffer_samples = 0;
	zt->t.newpvt = svip_new;
	zt->t.sample = fakesrc_sample;
	zt->t.native_plc = 0;
	zt->t.force_default = 1; /*Force to use the default cost, as the SVIP transcoding is an async procedure*/
	zt->t.default_cost = 10;  /*Used when can't calculate the real cost*/

	zt->t.desc_size = sizeof(struct codec_svip_pvt);
	if ((res = ast_register_translator(&zt->t))) {
		ast_free(zt);
		return -1;
	}

	AST_LIST_INSERT_HEAD(&translators, zt, entry);

	return res;
}

static void unregister_translators(void)
{
	struct translator *cur;

	AST_LIST_LOCK(&translators);
	while ((cur = AST_LIST_REMOVE_HEAD(&translators, entry))) {
		ast_unregister_translator(&cur->t);
		ast_free(cur);
	}
	AST_LIST_UNLOCK(&translators);
}

/* Must be called with the translators list locked. */
static int is_already_registered(uint32_t dstfmt, uint32_t srcfmt)
{
	int res = 0;
	const struct translator *zt;

	AST_LIST_TRAVERSE(&translators, zt, entry) {
		if ((zt->src_svip_fmt == srcfmt) && (zt->dst_svip_fmt == dstfmt)) {
			res = 1;
			break;
		}
	}

	return res;
}

static void build_translators(uint32_t dstfmts, uint32_t srcfmts)
{
	uint32_t srcfmt;
	uint32_t dstfmt;

	AST_LIST_LOCK(&translators);

	for (srcfmt = 1; srcfmt != 0; srcfmt <<= 1) {
		for (dstfmt = 1; dstfmt != 0; dstfmt <<= 1) {
			if (!(dstfmts & dstfmt) || !(srcfmts & srcfmt)) {
				continue;
			}
			if (is_already_registered(dstfmt, srcfmt)) {
				continue;
			}
			register_translator(dstfmt, srcfmt);
		}
	}

	AST_LIST_UNLOCK(&translators);
}

static int find_transcoders(void)
{
	struct svip_transcoder_info info = { 0, };

	unsigned int tcnum;

	for (tcnum=0; tcnum<sizeof(svip_tc_info)/sizeof(svip_tc_info[0]); tcnum++) {
		info = svip_tc_info[tcnum];
		if (option_verbose > 1)
			ast_verbose(VERBOSE_PREFIX_2 "Found transcoder '%s'.\n", info.name);

#if defined(SOFTSLIN_AULAW)
		/* Complex codecs need to support signed linear.  If the
		 * hardware transcoder does not natively support signed linear
		 * format, we will emulate it in software directly in this
		 * module. Also, do not allow direct ulaw/alaw to complex
		 * codec translation */
		if (info.dstfmts & (AST_FORMAT_ULAW | AST_FORMAT_ALAW)) {
			info.dstfmts |= AST_FORMAT_SLIN;
			info.dstfmts &= ~(AST_FORMAT_ULAW | AST_FORMAT_ALAW);
		}
		if (info.srcfmts & (AST_FORMAT_ULAW | AST_FORMAT_ALAW)) {
			info.srcfmts |= AST_FORMAT_SLIN;
			info.srcfmts &= ~(AST_FORMAT_ULAW | AST_FORMAT_ALAW);
		}
#endif
		build_translators(info.dstfmts, info.srcfmts);
	}

	ast_atomic_fetchadd_int(&devices.channels.total, info.numchannels); 

	if (!info.tcnum && (option_verbose > 1))
		ast_verbose(VERBOSE_PREFIX_2 "No hardware transcoders found.\n");

	return 0;

}

static int svip_codec_rm_int(void)
{
	int ii;
	int nchannels = 0;

	svip_fd = open( dev_filename, O_RDWR, 0x666);
	if (svip_fd < 0)
	{
		ast_log(LOG_ERROR, "Failed to open %s: %s\n", dev_filename, strerror(errno));
		return -1;
	}
	/*initialize Vinetic svip device*/
	for (ii=0; ii<MAX_VINETIC_DEV_NUM; ii++) {
		if (!SVIP_DEV_ISSTARTED(ii, devices.svip_dev_state)) {
			if (IFX_SUCCESS != vinetic_svip_devstart(svip_fd, ii)) {
				break;
			}
			SET_SVIP_DEV_STATE_START(ii, devices.svip_dev_state);
			devices.dev_total++;
			devices.dev_nch[ii] = vinetic_check_codec_num(svip_fd, ii);
			ast_svip_register(ii, devices.dev_nch[ii]);
			nchannels += devices.dev_nch[ii];
		}
	}

	if (0 == ii) {
		ast_log(LOG_ERROR, "start SVIP device %d error\n", ii);
		goto error0;
	} else {
		ast_log(LOG_VERBOSE, "start %d devices, %d coders \n", ii, nchannels);
		for (ii=0; ii<sizeof(svip_tc_info)/sizeof(svip_tc_info[0]); ii++) {
			svip_tc_info[ii].numchannels = nchannels;
		}
	}
#ifdef NEW_DESTROY_LOCK
	ast_mutex_init(&tr_lock);
#endif
#if defined(USE_SINGLE_THREAD)
	if (NULL == g_svip_timer) {
		if (!(g_svip_timer = ast_timer_open())) {
			goto error1;
		}
	}

	if (AST_PTHREADT_NULL == g_svip_rw_thread) {
		ii = ast_pthread_create(&g_svip_rw_thread, NULL, (void *)svip_rw_frame_handle_single_thread, NULL);
		if (ii < 0) {
			ast_debug(1, "svip read thread create failed\n");
			ast_timer_close(g_svip_timer);
			g_svip_timer = NULL;
			goto error1;
		}
	}
#endif
	return 0;

error1:
#ifdef NEW_DESTROY_LOCK
	ast_mutex_destroy(&tr_lock);
#endif
	for (ii=0; ii<MAX_VINETIC_DEV_NUM; ii++) {
		if (SVIP_DEV_ISSTARTED(ii, devices.svip_dev_state)) {
/* FIXME:  SVIP BUG, If we enable "SVIP Command ACK" by "sysinit --mirror port",
   and stop SVIP devices here.  the kernel tSVIPx_Ack threads cannot be exited after
   removing svip driver modules.
*/
#if 0
			if (IFX_SUCCESS != vinetic_svip_devstop(svip_fd, ii))
				ast_log(LOG_ERROR, "release device %d error\n", ii);
#endif
			SET_SVIP_DEV_STATE_STOP(ii, devices.svip_dev_state);
			ast_svip_unregister(ii);
		}
	}

error0:
	close(svip_fd);
	return -1;
}

static void svip_codec_rm_release(void)
{
	int ii;

	while (!AST_RWLIST_EMPTY(&g_svip_translator_list)) {
		sleep(1);
	}

#if defined(USE_SINGLE_THREAD)
	close_svip_thread = 1;
	notify_svip_rw_thread();
	if (AST_PTHREADT_NULL != g_svip_rw_thread) {
		pthread_join( g_svip_rw_thread , NULL);
		g_svip_rw_thread  = AST_PTHREADT_NULL;
	}
	if (g_svip_timer) {
		ast_timer_close(g_svip_timer);
		g_svip_timer = NULL;
	}
#endif

	for (ii=0; ii<MAX_VINETIC_DEV_NUM; ii++) {
		if (SVIP_DEV_ISSTARTED(ii, devices.svip_dev_state)) {
/* FIXME:  SVIP BUG, If we enable "SVIP Command ACK" by "sysinit --mirror port",
   and stop SVIP devices here.  the kernel tSVIPx_Ack threads cannot be exited after
   removing svip driver modules.
*/
#if 0
			if (IFX_SUCCESS != vinetic_svip_devstop(svip_fd, ii))
				ast_log(LOG_ERROR, "release device %d error\n", ii);
#endif
			SET_SVIP_DEV_STATE_STOP(ii, devices.svip_dev_state);
			ast_svip_unregister(ii);
		}
	}

#ifdef NEW_DESTROY_LOCK
	ast_mutex_destroy(&tr_lock);
#endif
	close(svip_fd);
}

static int reload(void)
{
	if (parse_config(1))
		return AST_MODULE_LOAD_DECLINE;

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	if (0 == module_load_fail) {
		svip_codec_rm_release();
	}
	ast_cli_unregister_multiple(cli, ARRAY_LEN(cli));
	unregister_translators();

#if defined(USE_SINGLE_THREAD)
	close(g_svip_translator_socket[0]);
	close(g_svip_translator_socket[1]);
#endif
	return 0;
}

static int load_module(void)
{
	int rval;
	int flags;

	ast_verb(2, "Loaded codec_svip.so: build data %s %s\n", __DATE__, __TIME__);

#if defined(USE_SINGLE_THREAD)
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, g_svip_translator_socket) < 0) {
		ast_log(LOG_ERROR, "create socket pair error\n");
		return AST_MODULE_LOAD_DECLINE;
	}
	/*set no block flags, for g_svip_translator_socket[1]*/
	flags = fcntl(g_svip_translator_socket[1], F_GETFL);
	fcntl(g_svip_translator_socket[1], F_SETFL, flags | O_NONBLOCK);
#endif

	if (parse_config(0))
		return AST_MODULE_LOAD_DECLINE;

	rval = svip_codec_rm_int();

	if (rval < 0) {
		/*we dont return FAILURE to prevent Asterisk exit when this module load error*/
		module_load_fail = 1;
		//return AST_MODULE_LOAD_FAILURE;
	}
	find_transcoders();
	ast_cli_register_multiple(cli, ARRAY_LEN(cli));
	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "Vinetic-SVIP Transcoder Codec Translator",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
		);
