/*
* Vinetic-SVIP Codec control.
*
* Copyright (C) 2012 - 2014, Grandstream Networks, Inc.
*
* jwli@grandstream.com
*
* This program is free software, distributed under the terms of
* the GNU General Public License Version 2. See the LICENSE file
* at the top of the source tree.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <arpa/inet.h>
#include <errno.h>

#include <time.h>
#include "gxe_svip_codec.h"

#define RETURN_VAL_IF_FAILL(x, rt) if(!(x))return (rt)
#define RETURN_IF_FAILL(x) if(!(x))return

llevel_t dlog_level = L_ERR;

static svip_jb_setting_t g_jb_setting = {
	.type = JB_TYPE_FIXED,
	.length = JB_LEN_LOW
};

//#define UCM_SVIP_DBG
#define SVIP_SP6_1 //SVIP_SP5_0

/*
* Vinetic-Svip set jitter buffer type and length
*
*  Param:
*        jb_setting: jitter buffer setting.
*
*/
IFX_return_t vinetic_set_jb_typelen (svip_jb_setting_t jb_setting)
{
	if ((jb_setting.type < 0) ||
		(jb_setting.type >= JB_TYPE_END))
		return IFX_ERROR;

	if ((jb_setting.length < 0) ||
		(jb_setting.type >= /*JB_LEN_END*/ JB_TYPE_END))
		return IFX_ERROR;

	g_jb_setting.type = jb_setting.type;
	g_jb_setting.length = jb_setting.length;

	return IFX_SUCCESS;
}

/*
* Vinetic-Svip get jitter buffer type and length
*
*  Param:
*        jb_setting: return jitter buffer setting.
*/
IFX_return_t vinetic_get_jb_typelen (svip_jb_setting_t *jb_setting)
{
	jb_setting->type = g_jb_setting.type;
	jb_setting->length = g_jb_setting.length;

	return IFX_SUCCESS;
}

/*
* Vinetic-Svip Set log level
*
*  Param:
*      level: log level to set.
*
*  return return current log level.
*/
llevel_t vinetic_set_loglevel (int level)
{
	if(level > L_DBG)
		dlog_level = L_DBG;
	else if(level < L_EMERG)
		dlog_level = L_EMERG;
	else
		dlog_level = level;
	return dlog_level;
}

/*
* Vinetic-Svip get log level
*
*  return current log level
*/
llevel_t vinetic_get_loglevel (void)
{
	return dlog_level;
}

/*
*  Checks whether a specific codec type is supported
*
*  Param:
*      fd:  file descriptor of svip device 
*      dev:  device index.
*      type:  encode type (IFX_TAPI_COD_TYPE_t)
*
*  return IFX_TRUE if Support, otherwise IFX_FALSE.
*/
IFX_return_t vinetic_check_codec_type(int fd,
	IFX_uint16_t dev, IFX_TAPI_COD_TYPE_t type)
{
	IFX_TAPI_CAP_t caplist;

	memset(&caplist, 0, sizeof(IFX_TAPI_CAP_t));
	caplist.dev = dev;
	caplist.captype = IFX_TAPI_CAP_TYPE_CODEC;
	caplist.cap = type;

	if (ioctl(fd, IFX_TAPI_CAP_CHECK, (IFX_int32_t)&caplist) > 0) {
		return IFX_TRUE;
	}

	return IFX_FALSE;
}

/*
*  Check how many data channels are supported  
*
*  Param:
*      fd:  file descriptor of svip device 
*      DevInd:  device index.
*
*  return the num of svip data channels
*/
IFX_int32_t vinetic_check_codec_num(int fd,
	IFX_uint16_t dev)
{
	IFX_TAPI_CAP_t caplist;

	memset(&caplist, 0, sizeof(IFX_TAPI_CAP_t));
	/* Check how many data channels are supported */
	caplist.dev = dev;
	caplist.captype = IFX_TAPI_CAP_TYPE_CODECS;
	if (ioctl(fd, IFX_TAPI_CAP_CHECK, (IFX_int32_t)&caplist) > 0) {
		return caplist.cap;
	}

	return 0;
}

/*
* Get master side network device MAC and IP address
*
*  Param:
*      devname:  network device name 
*      devmac:  pionter to save device MAC address.
*      addr:  pionter to save IP address (struct in_addr *)
*
*  return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
static int get_master_dev_mac_ip(char *devname,
	unsigned char *devmac, struct in_addr *addr)
{
	int sock;
	int rval = -1;
	struct sockaddr_in sin;
	struct sockaddr sa;
	struct ifreq ifr;

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock == -1) {
		perror("socket");
		return -1;
	}

	strncpy(ifr.ifr_name, devname, IFNAMSIZ);
	ifr.ifr_name[IFNAMSIZ-1] = 0;

	if (devmac) {
		rval = ioctl(sock, SIOCGIFHWADDR, &ifr);
		if (rval < 0) {
			ERR("ioctl error\n");
			goto out0;
		}
		memcpy(&sa, &ifr.ifr_addr, sizeof(sa));
		memcpy((void*)devmac, &sa.sa_data, IFX_TAPI_MAC_LEN);
	}

	if (addr) {
		rval = ioctl(sock, SIOCGIFADDR, &ifr);
		if(rval < 0) {
			ERR("ioctl error\n");
			goto out0;
		}
		memcpy(&sin, (struct sockaddr_in *)&ifr.ifr_addr, sizeof(sin));
		memcpy(addr, &sin.sin_addr, sizeof(struct in_addr));
	}

out0:
	close(sock);
	return rval;
}

/*
* Set encode and decode type for codec RU
*
*  Param:
*      fd:  file descriptor of svip device 
*      dev:  device index.
*      ch:  data(Codec) channel index.
*      enctype:  encode type (IFX_TAPI_COD_TYPE_t)
*      framelen: Codec Frame timespan(enum IFX_TAPI_COD_LENGTH_t )
*
*  return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
static IFX_return_t vinetic_codec_type_cfg (int fd, IFX_uint16_t dev,
	IFX_uint16_t ch, IFX_TAPI_COD_TYPE_t enctype, IFX_uint16_t framelen)
{
	IFX_TAPI_PKT_VOLUME_t cod_vol;
	IFX_TAPI_ENC_VAD_CFG_t enc_vad;
	IFX_TAPI_COD_DEC_HP_t cod_dec;
	IFX_TAPI_ENC_CFG_t enc_cfg;
	IFX_TAPI_DEC_CFG_t dec_cfg;

	memset(&cod_vol, 0, sizeof(IFX_TAPI_PKT_VOLUME_t));
	memset(&enc_cfg, 0,sizeof(IFX_TAPI_ENC_CFG_t));
	memset(&dec_cfg, 0,sizeof(IFX_TAPI_DEC_CFG_t));
	memset(&enc_vad, 0,sizeof(IFX_TAPI_ENC_VAD_CFG_t));
	memset(&cod_dec, 0,sizeof(IFX_TAPI_COD_DEC_HP_t));
#if defined(UCM_SVIP_DBG)
	INFO("Set the encoder type\n");
#endif
	/* Set the encoder type */
	enc_cfg.dev = dev;
	enc_cfg.ch = ch;
	enc_cfg.nEncType = enctype;

	switch(framelen)
	{
	case 5:
		enc_cfg.nFrameLen = IFX_TAPI_COD_LENGTH_5;
		break;
	case 10:
		enc_cfg.nFrameLen = IFX_TAPI_COD_LENGTH_10;
		break;
	case 11:
		enc_cfg.nFrameLen = IFX_TAPI_COD_LENGTH_11;
		break;
	case 20:
		enc_cfg.nFrameLen = IFX_TAPI_COD_LENGTH_20;
		break;
	case 30:
		enc_cfg.nFrameLen = IFX_TAPI_COD_LENGTH_30;
		break;
	case 40:
		enc_cfg.nFrameLen = IFX_TAPI_COD_LENGTH_40;
		break;
	case 50:
		enc_cfg.nFrameLen = IFX_TAPI_COD_LENGTH_50;
		break;
	case 60:
		enc_cfg.nFrameLen = IFX_TAPI_COD_LENGTH_60;
		break;
	default:
		enc_cfg.nFrameLen = IFX_TAPI_COD_LENGTH_ZERO;
		break;
	}

	if (enctype == IFX_TAPI_COD_TYPE_G726_16 ||
		IFX_TAPI_COD_TYPE_G726_24 ||
		IFX_TAPI_COD_TYPE_G726_32 ||
		IFX_TAPI_COD_TYPE_G726_40 ) {
		enc_cfg.AAL2BitPack = IFX_TAPI_COD_AAL2_BITPACK;
	} else {
		enc_cfg.AAL2BitPack = IFX_TAPI_COD_RTP_BITPACK;
	}

	if (ioctl(fd, IFX_TAPI_ENC_CFG_SET, (IFX_int32_t) &enc_cfg)
		!= IFX_SUCCESS)
	return IFX_ERROR;

#if defined(UCM_SVIP_DBG)
	INFO("Set the decoder type\n");
#endif
	/* Config decoder */
	dec_cfg.dev = dev;
	dec_cfg.ch = ch;
	dec_cfg.AAL2BitPack = enc_cfg.AAL2BitPack;
	dec_cfg.Plc = IFX_TAPI_DEC_PLC_CODEC;
	if (ioctl(fd, IFX_TAPI_DEC_CFG_SET, (IFX_int32_t) &dec_cfg)
		!= IFX_SUCCESS)
		return IFX_ERROR;

#if defined(UCM_SVIP_DBG)
	INFO("SET VAD \n");
#endif
	/* Enable/Disable VAD */
	enc_vad.dev = dev;
	enc_vad.ch = ch;
	enc_vad.vadMode = IFX_TAPI_ENC_VAD_NOVAD;//IFX_TAPI_ENC_VAD_CNG_ONLY//IFX_TAPI_ENC_VAD_ON//;
	if (ioctl(fd, IFX_TAPI_ENC_VAD_CFG_SET, (IFX_int32_t) &enc_vad)
		!= IFX_SUCCESS)
		return IFX_ERROR;

#if defined(UCM_SVIP_DBG)
	INFO("Configure encoder and decoder gains \n");
#endif
	/* Configure encoder and decoder gains : 0 dB */
	cod_vol.dev = dev;
	cod_vol.ch = ch;
	cod_vol.nEnc = 0;
	cod_vol.nDec = 0;
	if (ioctl(fd, IFX_TAPI_COD_VOLUME_SET, (IFX_int32_t) &cod_vol)
		!= IFX_SUCCESS)
	return IFX_ERROR;

#if defined(UCM_SVIP_DBG)
	INFO("Activate high-pass filter \n");
#endif
/* Activate high-pass filter */
	cod_dec.dev = dev;
	cod_dec.ch = ch;
	cod_dec.hpMode = IFX_ENABLE; //IFX_DISABLE;//
	if (ioctl(fd, IFX_TAPI_COD_DEC_HP_SET, (IFX_int32_t) &cod_dec)
		!= IFX_SUCCESS)
		return IFX_ERROR;

	return IFX_SUCCESS;
}

/*
* Set rtp payload type for codec RU 
*
*  Param:
*      fd:  file descriptor of svip device 
*      dev:  device index.
*      ch:  data(Codec) channel index.
*
*  return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
static IFX_return_t cfg_rtp_payload_type(int fd, IFX_uint16_t dev,
	IFX_uint16_t ch)
{
	IFX_TAPI_PKT_RTP_PT_CFG_t rtp_pt_conf;
	/* rtp payload type */

	memset(&rtp_pt_conf, 0, sizeof(IFX_TAPI_PKT_RTP_PT_CFG_t));
	/* Configure RTP payload type tables */
	/* Payload type table - Upstream */
	rtp_pt_conf.nPTup[IFX_TAPI_COD_TYPE_MLAW] = 0;
	rtp_pt_conf.nPTup[IFX_TAPI_COD_TYPE_ALAW] = 8;
	rtp_pt_conf.nPTup[IFX_TAPI_COD_TYPE_G723_63] = 4;
	rtp_pt_conf.nPTup[IFX_TAPI_COD_TYPE_G723_53] = 4;
	rtp_pt_conf.nPTup[IFX_TAPI_COD_TYPE_G729] = 18;
	rtp_pt_conf.nPTup[IFX_TAPI_COD_TYPE_ILBC_152] = 103;
	rtp_pt_conf.nPTup[IFX_TAPI_COD_TYPE_ILBC_133] = 103;
	rtp_pt_conf.nPTup[IFX_TAPI_COD_TYPE_G726_32] = 100;
	rtp_pt_conf.nPTup[IFX_TAPI_ENC_TYPE_LIN16_8] = 113;
	/* Payload type table - Downstream*/
	rtp_pt_conf.nPTdown[IFX_TAPI_COD_TYPE_MLAW] = 0;
	rtp_pt_conf.nPTdown[IFX_TAPI_COD_TYPE_ALAW] = 8;
	rtp_pt_conf.nPTdown[IFX_TAPI_COD_TYPE_G723_63] = 4;
	rtp_pt_conf.nPTdown[IFX_TAPI_COD_TYPE_G723_53] = 4;
	rtp_pt_conf.nPTdown[IFX_TAPI_COD_TYPE_G729] = 18;
	rtp_pt_conf.nPTdown[IFX_TAPI_COD_TYPE_ILBC_152] = 103;
	rtp_pt_conf.nPTdown[IFX_TAPI_COD_TYPE_ILBC_133] = 103;
	rtp_pt_conf.nPTdown[IFX_TAPI_COD_TYPE_G726_32] = 100;
	rtp_pt_conf.nPTdown[IFX_TAPI_ENC_TYPE_LIN16_8] = 113;

	rtp_pt_conf.dev = dev;
	rtp_pt_conf.ch = ch;
	/* Program the channel with the specified payload types */
#if defined(UCM_SVIP_DBG)
	DBG("Program the channel [%d] of device [%d] with the specified payload types\n",
		ch, dev);
#endif
	if (ioctl(fd, IFX_TAPI_PKT_RTP_PT_CFG_SET, (IFX_int32_t)&rtp_pt_conf)
		!= IFX_SUCCESS) {
		DBG("cfg rtp payload type Error\n");
		return IFX_ERROR;
	}

	return IFX_SUCCESS;
}

/*
* Set rtp session parameters for codec RU 
*
*  Param:
*      fd:  file descriptor of svip device 
*      dev:  device index.
*      ch:  data(Codec) channel index.
*
*  return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
static IFX_return_t cfg_rtp_session(int fd,
	IFX_uint16_t dev, IFX_uint16_t ch)
{
	IFX_TAPI_PKT_RTP_CFG_t rtpconf;

	memset(&rtpconf, 0, sizeof(IFX_TAPI_PKT_RTP_CFG_t));

	rtpconf.dev = dev;
	rtpconf.ch = ch;

	rtpconf.nSeqNr = random() % 0xfffful;
	rtpconf.nSsrc = random() % 0xfffffffful;
	rtpconf.nTimestamp = random() % 0xfffffffful;

	rtpconf.nPlayEvents = IFX_TAPI_PKT_EV_OOBPLAY_MUTE;

	rtpconf.nEventPT = 101;
	rtpconf.nEventPlayPT = 101;

	rtpconf.bRedundancyDecoder = IFX_FALSE;
	rtpconf.nRedundancy = 0;
	if (ioctl(fd, IFX_TAPI_PKT_RTP_CFG_SET, (IFX_int32_t) &rtpconf)
		!= IFX_SUCCESS)
		return IFX_ERROR;

	return IFX_SUCCESS;
}

/*
* Set ethernet packet parameters for codec RU 
*
*  Param:
*      fd:  file descriptor of svip device 
*      dev:  device index.
*      ch:  data(Codec) channel index.
*
*  return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
static IFX_return_t cfg_eth_packet_par(int fd,
	IFX_uint16_t dev, IFX_uint16_t ch,
	IFX_uint16_t udpdst, IFX_uint16_t udpsrc, IFX_uint16_t udprcv)
{
	IFX_TAPI_PKT_NET_CFG_t session;
	unsigned char rem_mac[IFX_TAPI_MAC_LEN] = {0};
	unsigned char local_mac[IFX_TAPI_MAC_LEN] = {0};

	struct in_addr rem_addr;
	struct in_addr local_addr;

	unsigned long rem_ip,local_ip;

	memset(&session, 0, sizeof (IFX_TAPI_PKT_NET_CFG_t));
	memset(&rem_addr, 0, sizeof (struct in_addr));
	memset(&local_addr, 0, sizeof (struct in_addr));

	session.dev = dev;
	session.ch = ch;

	/*get local(SVIP side) IP and MAC address */
	if (SVIP_libDevIP_Get((const IFX_uint32_t)dev,
		(IFX_int32_t *)&local_addr.s_addr) != SVIP_libStatusOk)
	{
		ERR("Get SVIP dev [%d] IP addr error\n", dev);
	}
	local_ip = local_addr.s_addr;

	if (SVIP_libDevMAC_Get((const IFX_uint32_t)dev, local_mac) !=
		SVIP_libStatusOk)
	{
		ERR("Get SVIP dev [%d] MAC addr error\n", dev);
	}
#ifdef SVIP_SP5_0
	session.MacSrc[0] =  local_mac[1];
	session.MacSrc[1] =  local_mac[0];
	session.MacSrc[2] =  local_mac[3];
	session.MacSrc[3] =  local_mac[2];
	session.MacSrc[4] =  local_mac[5];
	session.MacSrc[5] =  local_mac[4];
#else
	session.MacSrc[0] =  local_mac[0];
	session.MacSrc[1] =  local_mac[1];
	session.MacSrc[2] =  local_mac[2];
	session.MacSrc[3] =  local_mac[3];
	session.MacSrc[4] =  local_mac[4];
	session.MacSrc[5] =  local_mac[5];
#endif
	/*get remote(CPU side) IP and MAC address*/
	get_master_dev_mac_ip(VETH_NAME, rem_mac, &rem_addr);
	rem_ip = rem_addr.s_addr;
#ifdef SVIP_SP5_0
	session.MacDst[0] =  rem_mac[1];
	session.MacDst[1] =  rem_mac[0];
	session.MacDst[2] =  rem_mac[3];
	session.MacDst[3] =  rem_mac[2];
	session.MacDst[4] =  rem_mac[5];
	session.MacDst[5] =  rem_mac[4];
#else
	session.MacDst[0] =  rem_mac[0];
	session.MacDst[1] =  rem_mac[1];
	session.MacDst[2] =  rem_mac[2];
	session.MacDst[3] =  rem_mac[3];
	session.MacDst[4] =  rem_mac[4];
	session.MacDst[5] =  rem_mac[5];
#endif

	/* Configure the Ethernet packet parameter
	for the RTP data packets */
	session.nPktType = Voice;

	/* IPv4 is used. Configure the protocol
	information for transmission */
	session.nIpMode = IFX_TAPI_PKT_NET_IPV4;
#if SVIP_SP5_0
	session.nIp.v4.IpDst = htonl(rem_ip);
	session.nIp.v4.IpSrc = htonl(local_ip);
#else
	session.nIp.v4.IpDst = rem_ip;
	session.nIp.v4.IpSrc = local_ip;
#endif
	/* Configure the DSCP/TTL/Flag parameter
	inside the IPv4 protocol header */
	session.nIp.v4.nDSCP = 0x38;
	session.nIp.v4.nFlags = 0;
	session.nIp.v4.nTtl = 5;
	/* UDP parameter to receive and transmit. local port
	is the same for transmission and reception.*/
	session.bUdpCrcGen = IFX_ENABLE;
	session.bUdpCrcCheck = IFX_ENABLE;
#if SVIP_SP5_0
	session.nUdpDst = udpdst;
	session.nUdpSrc = udpsrc;
	session.nUdpRcv = udprcv;
#else
	session.nUdpDst = htons(udpdst);
	session.nUdpSrc = htons(udpsrc);
	session.nUdpRcv = htons(udprcv);
#endif

#if defined(UCM_SVIP_DBG)
	DBG("\n####### Configure Ethernet session ########\n"
	"\tDestination MAC %02X:%02X:%02X:%02X:%02X:%02X\n"
	"\tSource      MAC %02X:%02X:%02X:%02X:%02X:%02X\n"
	"\tPorts: Rec-%d, Des-%d, Sou-%d\n"
	"\tchan %d, dev %d\n"
	"\tType: %s, Using: %s\n"
	"\tIP Dst %s\n",
#if SVIP_SP5_0
	session.MacDst[1], session.MacDst[0], session.MacDst[3],
	session.MacDst[2], session.MacDst[5], session.MacDst[4],
	session.MacSrc[1], session.MacSrc[0], session.MacSrc[3],
	session.MacSrc[2], session.MacSrc[5], session.MacSrc[4],
#else
	session.MacDst[0], session.MacDst[1], session.MacDst[2],
	session.MacDst[3], session.MacDst[4], session.MacDst[5],
	session.MacSrc[0], session.MacSrc[1], session.MacSrc[2],
	session.MacSrc[3], session.MacSrc[4], session.MacSrc[5],
#endif
	session.nUdpRcv, session.nUdpDst, session.nUdpSrc,
	session.ch, session.dev,
	session.nPktType == Voice ? "Voice" :
	session.nPktType == Fax ? "Fax" :
	session.nPktType == FaxTrace ? "FaxTrace" : "unknown",
	session.nIpMode == IFX_TAPI_PKT_NET_IPV4 ? "IPV4" : "IPV6",
	inet_ntoa(rem_addr));
	DBG("\n"
	"\tIP Src %s\n"
	"\tnVLAN_Tagging %d\n"
	"###############\n",
	inet_ntoa(local_addr),
	session.nVLAN_Tagging);
#endif
	if (ioctl(fd, IFX_TAPI_PKT_NET_CFG_SET, (IFX_int32_t) &session)
		!= IFX_SUCCESS)
	return IFX_ERROR;

	return IFX_SUCCESS;
}

/*
* Retrieves RTCP statistics 
*
*  Param:
*      fd:  file descriptor of svip device 
*      dev:  device index.
*      ch:  data(Codec) channel index.
*/
IFX_return_t vinetic_get_jb_statistic(int fd, IFX_uint16_t dev,
	IFX_uint16_t ch, IFX_TAPI_JB_STATISTICS_t *jbstat,
	IFX_TAPI_PKT_RTCP_STATISTICS_t *rtcpstat)
{
	memset(rtcpstat, 0, sizeof(*rtcpstat));
	memset(jbstat, 0, sizeof(*jbstat));

	jbstat->dev = rtcpstat->dev = dev;
	jbstat->ch = rtcpstat->ch = ch;
	if (ioctl(fd, IFX_TAPI_PKT_RTCP_STATISTICS_GET, (IFX_int32_t) rtcpstat)
		!= IFX_SUCCESS)
		return IFX_ERROR;
	if (ioctl(fd, IFX_TAPI_JB_STATISTICS_GET, (IFX_int32_t) jbstat)
		!= IFX_SUCCESS)
		return IFX_ERROR;

	return IFX_SUCCESS;
}

/*
* Set jitter buffer
*
*  Param:
*      fd:  file descriptor of svip device 
*      dev:  device index.
*      ch:  data(Codec) channel index.
*      type:  jitter buffer type, '0'-- fixed JB. '1'-- Adaptive JB 
*      length: jitter buffer length, '0'--50ms; '1'--100ms; '2'--200ms
*      voice:  1--Jitter buffer optimized for voice.
*                 0--Jitter buffer optimized for data but without doing packet repetition. 
*/
static int vinetic_setjb( int fd, IFX_uint16_t dev, IFX_uint16_t ch,
	IFX_uint8_t type, IFX_uint8_t length, IFX_uint8_t voice)
{
	IFX_TAPI_JB_CFG_t jbcfg;

	memset(&jbcfg, 0, sizeof(IFX_TAPI_JB_CFG_t));

	jbcfg.dev = dev;
	jbcfg.ch = ch;

	if ( voice )
	{
		/* Adaptive or fixed JB */
		if ( type == 0 )
			jbcfg.nJbType = IFX_TAPI_JB_TYPE_FIXED;
		else
		{
			jbcfg.nJbType = IFX_TAPI_JB_TYPE_ADAPTIVE;
			jbcfg.nLocalAdpt = IFX_TAPI_JB_LOCAL_ADAPT_ON;
		}
		/* Optimization for voice */
		jbcfg.nPckAdpt = IFX_TAPI_JB_PKT_ADAPT_VOICE;
		switch ( length )
		{
		case 0:
			jbcfg.nScaling = 0x8;
			/* Initial JB size 50 ms = 0x190 * 125 us */
			jbcfg.nInitialSize = 0x190;
			/* Minimum JB size 10 ms = 0x50 * 125 us */
			jbcfg.nMinSize = 0x50;
			/* Maximum JB size 100 ms = 0x320 * 125 us */
			jbcfg.nMaxSize = 0x320;
			break;
		case 1:
			jbcfg.nScaling = 0x10;
			/* Initial JB size 100 ms = 0x320 * 125 us */
			jbcfg.nInitialSize = 0x320;
			/* Minimum JB size 20 ms = 0xa0 * 125 us */
			jbcfg.nMinSize = 0xa0;
			/* Maximum JB size 200 ms = 0x640 * 125 us */
			jbcfg.nMaxSize = 0x640;
			break;
		case 3:
		default:
			jbcfg.nScaling = 0x16;
			/* Initial JB size 200 ms = 0x2D0 * 125 us */
			jbcfg.nInitialSize = 0x640;
			/* Minimum JB size 40 ms = 0x140 * 125 us */
			jbcfg.nMinSize = 0x140;
			/* Maximum JB size 600 ms = 0x12C0 * 125 us */
			jbcfg.nMaxSize = 0x12C0;
			break;
		}
	}
	else
	{
		jbcfg.nJbType = IFX_TAPI_JB_TYPE_FIXED;
		jbcfg.nPckAdpt = IFX_TAPI_JB_PKT_ADAPT_DATA;
		jbcfg.nInitialSize = 0x190;
		jbcfg.nMinSize = 0x50;
		jbcfg.nMaxSize = 0x640;
	}

	if (ioctl(fd, IFX_TAPI_JB_CFG_SET, (IFX_int32_t)&jbcfg) != IFX_SUCCESS)
		return IFX_ERROR;

	return IFX_SUCCESS;
}


/**
*  Map a  COD RU to a PCM/COD RU.
*  
*   param: fd - file descriptor of svip device 
*          dev - device index.
*          ch  - data channel index.
*          dstch - channel index for mapping
*          type  - channel type for mapping
*          
*  return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
static IFX_return_t map_ru_to_cod_ru(int fd, IFX_uint16_t dev,
	IFX_uint16_t ch, IFX_uint16_t dstch, IFX_TAPI_MAP_TYPE_t type)
{
	IFX_TAPI_MAP_DATA_t datamap;
	memset(&datamap, 0, sizeof(IFX_TAPI_MAP_DATA_t));

	datamap.dev = dev;
	datamap.ch = ch;
	datamap.nDstCh = dstch;
	datamap.nChType = type;

	if (ioctl(fd, IFX_TAPI_MAP_DATA_ADD, (int)&datamap) != IFX_SUCCESS)
		return IFX_ERROR;

	return IFX_SUCCESS;
}

/**
*  UNMap COD RUs
*  
*   param: fd - file descriptor of svip device 
*          dev - device index.
*          ch  - data channel index.
*          dstch - channel index for mapping
*          Type  - channel type for mapping
*          
*  return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
static IFX_return_t unmap_ru_to_cod_ru(int fd, IFX_uint16_t dev,
	IFX_uint16_t ch, IFX_uint16_t dstch, IFX_TAPI_MAP_TYPE_t type)
{
	IFX_TAPI_MAP_DATA_t param;
	memset(&param, 0, sizeof (param));

	param.nPlayStart = IFX_TAPI_MAP_DATA_UNCHANGED;
	param.nRecStart = IFX_TAPI_MAP_DATA_UNCHANGED;

	param.nChType = type;
	param.nDstCh = dstch;

	param.ch = ch;
	param.dev = dev;

	if (ioctl (fd, IFX_TAPI_MAP_DATA_REMOVE, (unsigned long) &param)
		!= IFX_SUCCESS)
		return IFX_ERROR;

	return IFX_SUCCESS;
}

/**
*  Map a  COD RU to a COD RU.
*  
*   param: fd - file descriptor of svip device 
*          dev - device index.
*          ch  - data channel index.
*          dstch - channel index for mapping
*
*  return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
IFX_return_t vinetic_map_cod_ru(int fd, IFX_uint16_t dev,
	IFX_uint16_t ch, IFX_uint16_t dstch)
{
	return map_ru_to_cod_ru(fd, dev, ch, dstch, IFX_TAPI_MAP_TYPE_CODER);
}

/**
*  UNMap a  COD RU to a COD RU.
*  
*   param: fd - file descriptor of svip device 
*          dev - device index.
*          ch  - data channel index.
*          dstch - channel index for mapping
*
*  return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
IFX_return_t vinetic_unmap_cod_ru(int fd, IFX_uint16_t dev,
	IFX_uint16_t ch, IFX_uint16_t dstch)
{
	return unmap_ru_to_cod_ru(fd, dev, ch, dstch, IFX_TAPI_MAP_TYPE_CODER);
}


/**
* Configure encode/decode type, rtp payload type, rtp session 
* parameters, eth packet parameters and map encode-decode RU.
*  
* Param:
*      fd:  svip device descriptor
*      ch :  Codec Channel
*      dev - device index
*      enctype:  encode type
*      framelen: Codec Frame timespan(enum IFX_TAPI_COD_LENGTH_t )
*      udpdst:  UDP destination port number for send packets of codec channel
*      udprcv:  UDP destination port number for receive packets of codec channel
*      udpsrc:  UDP source port number for send packets of codec channel
*
* return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*
*/
IFX_return_t vinetic_codec_cfg(int fd, IFX_uint16_t dev,
	IFX_uint16_t ch, IFX_TAPI_COD_TYPE_t enctype, IFX_uint16_t framelen,
	IFX_uint16_t udpdst, IFX_uint16_t udprcv, IFX_uint16_t udpsrc)
{
	int retry_n = 2;
	/*Vinetic Svip Codec Configuration*/
#if defined(UCM_SVIP_DBG)
	DBG("vinetic_codec_cfg:fd:%d devid %d chid %d enctype %d nFrameLen %d \n",
		fd, dev, ch, enctype, framelen);
#endif
retry:
	if(0 == retry_n)
		return IFX_ERROR;

	retry_n--;

	/*Set encode and decode type*/
	if (vinetic_codec_type_cfg(fd, dev,
		ch , enctype, framelen)!= IFX_SUCCESS) {
		ERR("Vinetic COD Cfg error \n");
		vinetic_svip_codec_stop(fd, dev, ch, ch);
		goto retry;
	}

	/* Config Codec RTP Payload type table */
	if (cfg_rtp_payload_type(fd, dev, ch) != IFX_SUCCESS) {
		ERR("config svip codec rtp payload type error\n");
		vinetic_svip_codec_stop(fd, dev, ch, ch);
		goto retry;
	}

	/* Config Codec RTP Session parameters */
	if (cfg_rtp_session(fd, dev, ch) != IFX_SUCCESS) {
		ERR("config svip codec channel %d of device %d rtp session error\n",
			ch, dev);
		vinetic_svip_codec_stop(fd, dev, ch, ch);
		goto retry;
	}

	/* Config Codec eth packet parameters */
	if (cfg_eth_packet_par(fd, dev, ch,
		udpdst, udpsrc, udprcv) != IFX_SUCCESS) {
		ERR("config channel %d of device %d ethernet packet Parameters error\n",
			ch, dev);
		vinetic_svip_codec_stop(fd, dev, ch, ch);
		goto retry;
	}

	/* Set jitter buffer */
	if (vinetic_setjb(fd, dev, ch, g_jb_setting.type,
		g_jb_setting.length, 1)!= IFX_SUCCESS) {
		ERR("Failed to set jitter buffer on channel %d \n", ch);
		vinetic_svip_codec_stop(fd, dev, ch, ch);
		goto retry;
	}

	return IFX_SUCCESS;
}

/*
*  Vinetic-Svip Codec ctrol
*  Param:
*      fd:   svip device descriptor
*      dev   device index
*      ch:   codec channel
*      cctrl:  ctrol command
*
* return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
IFX_return_t vinetic_cod_ctrl(int fd, IFX_uint16_t dev,
	IFX_uint16_t ch, cod_ctrl cctrl)
{
	IFX_TAPI_ENC_MODE_t enc_mode;
	IFX_TAPI_DEC_t dec_mode;
	IFX_TAPI_ENC_HOLD_t enc_hold;
	IFX_return_t ret;

	memset(&enc_mode, 0, sizeof (enc_mode));
	memset(&dec_mode, 0, sizeof (dec_mode));
	memset(&enc_hold, 0, sizeof (enc_hold));

	/* Start encoding on COD channel [ChInd] of device [DevInd]:
	this also starts RTP packet flow */
	/* Parameter is not needed */
	enc_mode.dev = dec_mode.dev = enc_hold.dev = dev;
	enc_mode.ch = dec_mode.ch = enc_hold.ch = ch;

	ret = IFX_SUCCESS;
	switch (cctrl) {
	case COD_ENC_START:
#if defined(UCM_SVIP_DBG)
		INFO("ENCODE channel %d of device %d start...\n", ch, dev);
#endif
		if (ioctl(fd, IFX_TAPI_ENC_START, (IFX_int32_t) &enc_mode)
			!= IFX_SUCCESS)
			ret = IFX_ERROR;
		break;
	case COD_ENC_HOLD:
		/* Now the remote (VoIP) party has to be put on hold */
		/* Enable encoder hold */
#if defined(UCM_SVIP_DBG)
		INFO("ENCODE channel %d of device %d hold...\n", ch, dev);
#endif
		enc_hold.hold = IFX_ENABLE;
		if (ioctl(fd, IFX_TAPI_ENC_HOLD, (IFX_int32_t) &enc_hold)
			!= IFX_SUCCESS)
			ret = IFX_ERROR;
		break;
	case COD_ENC_STOP:
		/* Stop encoding: this also stops RTP packet flow */
		/* Parameter is not needed */
#if defined(UCM_SVIP_DBG)
		INFO("ENCODE channel %d of device %d stop...\n", ch, dev);
#endif
		if (ioctl(fd, IFX_TAPI_ENC_STOP, (IFX_int32_t) &enc_mode)
			!= IFX_SUCCESS)
			ret = IFX_ERROR;
		break;
	case COD_DEC_START:
#if defined(UCM_SVIP_DBG)
	INFO("DECODE channel %d of device %d start...\n", ch, dev);
#endif
		if (ioctl(fd, IFX_TAPI_DEC_START, (IFX_int32_t) &dec_mode)
			!= IFX_SUCCESS)
			ret = IFX_ERROR;
		break;
	case COD_DEC_STOP:
	/* Stop encoding: this also stops RTP packet flow */
	/* Parameter is not needed */
#if defined(UCM_SVIP_DBG)
		INFO("DECODE channel %d of device %d stop...\n", ch, dev);
#endif
		if (ioctl(fd, IFX_TAPI_DEC_STOP, (IFX_int32_t) &dec_mode)
			!= IFX_SUCCESS)
			ret = IFX_ERROR;
		break;
	default:
		break;
	}

	return ret;
}

/*
*  Start Vinetic-Svip device 
*  Param:
*      fd:   svip device descriptor
*      dev   device index
*
* return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
IFX_return_t vinetic_svip_devstart(int fd, IFX_uint16_t dev)
{
	IFX_TAPI_DEV_START_CFG_t devstart;

	INFO("Start Vinetic initialization device %d \n", dev);
	memset(&devstart, 0, sizeof(IFX_TAPI_DEV_START_CFG_t));

	/* Init for VoIP applications */
	/* devstart.nMode = IFX_TAPI_INIT_MODE_VOICE_CODER;*/
	/* devstart.nMode = IFX_TAPI_INIT_MODE_NONE;*/
	devstart.nMode = IFX_TAPI_INIT_MODE_DEFAULT;

	devstart.dev = dev;
	if (ioctl(fd, IFX_TAPI_DEV_START, (IFX_int32_t)&devstart) < 0)
	{
		ERR("err: start dev [%d] failed \n", dev);
		return IFX_ERROR;
	}

	return IFX_SUCCESS;
}

/*
*  Stop Vinetic-Svip device 
*  Param:
*      fd:   svip device descriptor
*      dev   device index
*
* return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
IFX_return_t vinetic_svip_devstop(int fd, IFX_uint16_t dev)
{
	IFX_TAPI_DEV_START_CFG_t devstop;

	memset(&devstop, 0, sizeof(IFX_TAPI_DEV_START_CFG_t));
	/* Stop the TAPI device instance*/
	devstop.dev = dev;
	if (ioctl(fd, IFX_TAPI_DEV_STOP, (int)&devstop) != 0)
		return IFX_ERROR;

	return IFX_SUCCESS;
}

/*
*  Stop encode and decode ,umap codec RUs.
*
*  Param:
*      fd:   svip device descriptor
*      dev   device index
*      EncCh:  decoder channel index
*      DecCh:  encoder channel index 
*
* return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
IFX_return_t vinetic_svip_codec_stop(int fd, IFX_uint16_t dev,
	IFX_uint16_t encch, IFX_uint16_t decch)
{
	int err = 0;
	/*stop dec*/
	if (IFX_SUCCESS != vinetic_cod_ctrl(fd, dev, decch, COD_DEC_STOP)) {
		ERR("Stop channel %d of device %d decode error \n",
			decch, dev);
		err++;
	}

	/*stop enc*/
	if (IFX_SUCCESS != vinetic_cod_ctrl(fd, dev, encch, COD_ENC_STOP)) {
		ERR("Stop channel %d of device %d encode error \n",
			encch, dev);
		err++;
	}

	if (err)
		return IFX_ERROR;

	return IFX_SUCCESS;
}

/*
*  Flushes the transmit and receive packet buffer
*
*  Param:
*      fd:  svip device descriptor
*      nChInd:  Codec channel index
*
* return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
IFX_return_t vinetic_svip_pkt_flush(int fd, IFX_uint16_t dev,
	IFX_uint16_t ch)
{
	IFX_TAPI_PKT_FLUSH_t pktflush;

	pktflush.dev = dev;
	pktflush.ch = ch;

	if (ioctl(fd, IFX_TAPI_PKT_FLUSH, (IFX_int32_t) &pktflush)
		!= IFX_SUCCESS)
		return IFX_ERROR;

	return IFX_SUCCESS;
}

/*
*  Vinetic svip coder udp port get for coder receive packet
*
*  Param:
*      DevInd:  index of device
*      nChInd:  index of Codec channel
*
* return port num;
*/
IFX_int16_t vinetic_svip_get_cod_rcvport(IFX_uint16_t dev,
	IFX_uint16_t ch)
{
	return CODEC_UDPPORT0 + dev * MAX_VINETIC_CHAN_PER_DEV + ch;
}

static rtp_packet_t svip_codec_rtp_packet_alloc(
	unsigned int payload_len,
	unsigned int pad_len,
	unsigned int csrc_count)
{
	unsigned int len;
	rtp_packet_t packet;

	RETURN_VAL_IF_FAILL(csrc_count <= 15, NULL);

	len = RTP_HEADER_LEN + csrc_count * 4
		+ payload_len + pad_len;

	RETURN_VAL_IF_FAILL(len <= RTP_MTU, NULL);

	packet = malloc(sizeof *packet);

	if (!packet)
		return NULL;

	packet->data_len = len;
	packet->data = malloc(len);

	if (!packet->data) {
		free(packet);
		return NULL;
	}

	return(packet);
}

static void svip_codec_rtp_packet_free(rtp_packet_t *packet)
{
	if((*packet)->data) {
		free((*packet)->data);
		(*packet)->data = NULL;
	}
	if((*packet)) {
		free(*packet);
		*packet = NULL;
	}
}

/**
* initialize socket address for receive and send rtp packets
*
* Param:
*      dev: index of device
*      send_dstport: must get it by call vinetic_svip_get_cod_rcvport();
*      rtp_payload_len: length of rtp payload
*      rtp_sess: rtp session parameters  for caller  
*      rtp_pt : rtp payload type
*      ts_inc : timestamp increse
*
*  return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
IFX_return_t svip_user_rtp_init(IFX_uint16_t dev,
	IFX_uint16_t send_dport, IFX_uint32_t rtp_payload_len,
	IFX_uint8_t rtp_pt, IFX_uint32_t ts_inc, svip_rtp_sess_t * rtp_sess)
{
	int sock;
	struct sockaddr_in *svip_saddr;
	struct sockaddr_in *rcv_saddr;
	struct in_addr master_in_addr;
	struct in_addr svip_in_addr;
	unsigned int rcv_dport;
	time_t t;
	int rval;

	if(!rtp_sess)
		goto error0;

	svip_saddr = &rtp_sess->svip_saddr;
	rcv_saddr = &rtp_sess->rcv_saddr;

#if defined(UCM_SVIP_DBG)
	DBG("===================Svip_User_Rtp_Init===============\n");
#endif
	rval = get_master_dev_mac_ip(VETH_NAME, NULL, &master_in_addr);
	if (rval < 0)
		goto error0;

	if (SVIP_libDevIP_Get((const IFX_uint32_t)dev,
		(IFX_int32_t *)&svip_in_addr.s_addr) != SVIP_libStatusOk)
	{
		ERR("Get SVIP dev [%d] IP addr error\n", dev);
		goto error0;
	}

	srandom((unsigned) time(&t));

	/*initialize sock addr of svip codec */
	svip_saddr->sin_family = AF_INET;
	svip_saddr->sin_port = htons(send_dport);
	svip_saddr->sin_addr = svip_in_addr;
	bzero(&(svip_saddr->sin_zero), 8);

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
		ERR("socket: %s\n",strerror(errno));
		goto error0;
	}

	/*local socket address for receive*/
retry:
	rcv_dport = 10000 + abs(random())/50000;/*10000~60000*/;
	rcv_saddr->sin_family = AF_INET;
	rcv_saddr->sin_port = htons(rcv_dport);
	rcv_saddr->sin_addr = master_in_addr;
	bzero(&(rcv_saddr->sin_zero), 8);

	if (bind(sock, (struct sockaddr *)rcv_saddr,
		sizeof(struct sockaddr)) == -1)
	{
		if (errno == EADDRINUSE) {
			goto retry;
		}
		perror("Bind");
		goto error1;
	}

	rtp_sess->snd_dstport = send_dport;
	rtp_sess->rcv_dstport = rcv_dport;
	rtp_sess->snd_srcport = rcv_dport;

	rtp_sess->seq_num = 0;
	rtp_sess->timestamp =0 ;
	rtp_sess->rtp_ssrc = random()&0xffffffff;
	rtp_sess->timestamp_increse = ts_inc;
	rtp_sess->rtp_pt = rtp_pt;
	rtp_sess->m_bit = 1;

	rtp_sess->sock = sock;
	rtp_sess->rtp_pkt_snd =
		svip_codec_rtp_packet_alloc(rtp_payload_len, 0, 0);

	if (!rtp_sess->rtp_pkt_snd)
		goto error1;

	rtp_sess->rtp_pkt_rcv =
		svip_codec_rtp_packet_alloc(RTP_MTU-12, 0, 0);

	if (!rtp_sess->rtp_pkt_rcv)
		goto error2;

	return IFX_SUCCESS;

error2:
	svip_codec_rtp_packet_free(&rtp_sess->rtp_pkt_snd);

error1:
	close(sock);
error0:
	return IFX_ERROR;
}

/**
* free struct Rtp_Packet allocated by Svip_User_Rtp_Init and close socket 
*
* Param:
*      sock: scoket descriptor
*      pkt_snd: struct Rtp_Packet alloc by Svip_User_Rtp_Init for send packets
*      pkt_rcv: struct Rtp_Packet alloc by Svip_User_Rtp_Init for receive packets
*   return IFX_SUCCESS if successful, otherwise IFX_ERROR.
*/
int svip_user_rtp_release (int sock, rtp_packet_t *pkt_snd,
	rtp_packet_t *pkt_rcv)
{
	svip_codec_rtp_packet_free(pkt_snd);
	svip_codec_rtp_packet_free(pkt_rcv);
	close(sock);

	return 0;
}

unsigned char rtp_packet_get_padding(rtp_packet_t packet)
{
	RETURN_VAL_IF_FAILL(packet != NULL, 0);

	return ((rtp_header_t)packet->data)->padding;
}

unsigned char rtp_packet_get_csrc_count(rtp_packet_t packet)
{
	RETURN_VAL_IF_FAILL(packet != NULL, 0);

	return ((rtp_header_t)packet->data)->csrc_count;
}

void *rtp_packet_get_payload(rtp_packet_t packet)
{
	RETURN_VAL_IF_FAILL(packet != NULL, NULL);

	return ((char *)packet->data)
		+ RTP_HEADER_LEN
		+ rtp_packet_get_csrc_count(packet) * sizeof(unsigned int);
}

unsigned int rtp_packet_get_payload_len(rtp_packet_t packet)
{
	unsigned int len;

	RETURN_VAL_IF_FAILL(packet != NULL, 0);

	len = packet->data_len - RTP_HEADER_LEN
		- rtp_packet_get_csrc_count(packet) * sizeof(unsigned int);

	if (rtp_packet_get_padding(packet))
		len -= ((unsigned char *) packet->data)[packet->data_len - 1];

	return len;
}

unsigned int rtp_packet_get_timestamp(rtp_packet_t packet)
{
	RETURN_VAL_IF_FAILL(packet != NULL, 0);

	return ntohl(((rtp_header_t) packet->data)->timestamp);
}

unsigned short rtp_packet_get_seqnum(rtp_packet_t packet)
{
	RETURN_VAL_IF_FAILL(packet != NULL, 0);

	return ntohs(((rtp_header_t) packet->data)->seq);
}

void rtp_packet_set_version(rtp_packet_t packet, unsigned char version)
{
	RETURN_IF_FAILL(packet != NULL);
	RETURN_IF_FAILL(version < 0x04);

	((rtp_header_t)packet->data)->version = version;
}

void rtp_packet_set_padding(rtp_packet_t packet, unsigned char padding)
{
	RETURN_IF_FAILL(packet != NULL);
	RETURN_IF_FAILL(padding < 0x02);

	((rtp_header_t)packet->data)->padding = padding;
}
void rtp_packet_set_extension(rtp_packet_t packet, unsigned char extension)
{
	RETURN_IF_FAILL(packet != NULL);
	RETURN_IF_FAILL(extension < 0x02);

	((rtp_header_t)packet->data)->extension = extension;
}

void rtp_packet_set_csrc_count(rtp_packet_t packet, unsigned char csrc_count)
{
	RETURN_IF_FAILL(packet != NULL);
	RETURN_IF_FAILL(csrc_count < 0x04);

	((rtp_header_t)packet->data)->csrc_count = csrc_count;
}

void rtp_packet_set_marker(rtp_packet_t packet, unsigned char marker)
{
	RETURN_IF_FAILL(packet != NULL);
	RETURN_IF_FAILL(marker < 0x02);

	((rtp_header_t)packet->data)->marker = marker;
}

void rtp_packet_set_payload_type(rtp_packet_t packet,
	unsigned char payload_type)
{
	RETURN_IF_FAILL(packet != NULL);
	RETURN_IF_FAILL(payload_type < 0x80);

	((rtp_header_t)packet->data)->payload_type = payload_type;
}

void rtp_packet_set_seq(rtp_packet_t packet, unsigned int seq)
{
	RETURN_IF_FAILL(packet != NULL);

	((rtp_header_t)packet->data)->seq = htons(seq);
}

void rtp_packet_set_timestamp(rtp_packet_t packet,
	unsigned int timestamp)
{
	RETURN_IF_FAILL(packet != NULL);

	((rtp_header_t)packet->data)->timestamp = htonl(timestamp);
}

void rtp_packet_set_ssrc(rtp_packet_t packet, unsigned int ssrc)
{
	RETURN_IF_FAILL(packet != NULL);

	((rtp_header_t)packet->data)->ssrc = htonl(ssrc);
}

int rtp_packet_send(int sock, rtp_packet_t packet, struct sockaddr_in *toaddr)
{
	RETURN_VAL_IF_FAILL(toaddr != NULL, 0);
	RETURN_VAL_IF_FAILL(packet != NULL, 0);
	RETURN_VAL_IF_FAILL(packet->data != NULL, 0);

	return (sendto(sock, (void *) packet->data, packet->data_len,
		0, (struct sockaddr *)toaddr, sizeof(struct sockaddr)));
}
