/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2006, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Asterisk Logger
 *
 * Logging routines
 *
 * \author Mark Spencer <markster@digium.com>
 */

/*! \li \ref logger.c uses the configuration file \ref logger.conf
 * \addtogroup configuration_file Configuration Files
 */

/*!
 * \page logger.conf logger.conf
 * \verbinclude logger.conf.sample
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#ifdef GRANDSTREAM_NETWORKS
#define AST_MODULE_LOG "logger"     
#endif

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision$")

/* When we include logger.h again it will trample on some stuff in syslog.h, but
 * nothing we care about in here. */
#include <syslog.h>
#include <signal.h>
#include <time.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef GRANDSTREAM_NETWORKS
#include <sqlite3.h>
#endif

#include "asterisk/_private.h"
#include "asterisk/paths.h"	/* use ast_config_AST_LOG_DIR */
#include "asterisk/logger.h"
#include "asterisk/lock.h"
#include "asterisk/channel.h"
#include "asterisk/config.h"
#include "asterisk/term.h"
#include "asterisk/cli.h"
#include "asterisk/utils.h"
#include "asterisk/manager.h"
#include "asterisk/astobj2.h"
#include "asterisk/threadstorage.h"
#include "asterisk/strings.h"
#include "asterisk/pbx.h"
#include "asterisk/app.h"
#include "asterisk/syslog.h"
#include "asterisk/buildinfo.h"
#include "asterisk/ast_version.h"
#include "asterisk/backtrace.h"

/*** DOCUMENTATION
 ***/
#ifdef GRANDSTREAM_NETWORKS
#define LOGGER_VERBOSE_REGISTER_OFF 65536  /* turn off logger verbose register 'ast_register_verbose' */
static char dateformat[256] = "%Y%m%d %H%M%S.%3q";	
#else
static char dateformat[256] = "%b %e %T";		/* Original Asterisk Format */
#endif

static char queue_log_name[256] = QUEUELOG;
static char exec_after_rotate[256] = "";

static int filesize_reload_needed;
static unsigned int global_logmask = 0xFFFF;
static int queuelog_init;
static int logger_initialized;
static volatile int next_unique_callid; /* Used to assign unique call_ids to calls */
static int display_callids;
static void unique_callid_cleanup(void *data);

struct ast_callid {
	int call_identifier; /* Numerical value of the call displayed in the logs */
};

AST_THREADSTORAGE_CUSTOM(unique_callid, NULL, unique_callid_cleanup);

static enum rotatestrategy {
	NONE = 0,                /* Do not rotate log files at all, instead rely on external mechanisms */
	SEQUENTIAL = 1 << 0,     /* Original method - create a new file, in order */
	ROTATE = 1 << 1,         /* Rotate all files, such that the oldest file has the highest suffix */
	TIMESTAMP = 1 << 2,      /* Append the epoch timestamp onto the end of the archived file */
} rotatestrategy = SEQUENTIAL;

static struct {
	unsigned int queue_log:1;
	unsigned int queue_log_to_file:1;
	unsigned int queue_adaptive_realtime:1;
	unsigned int queue_log_realtime_use_gmt:1;
} logfiles = { 1 };

static char hostname[MAXHOSTNAMELEN];
AST_THREADSTORAGE_RAW(in_safe_log);

enum logtypes {
	LOGTYPE_SYSLOG,
	LOGTYPE_FILE,
	LOGTYPE_CONSOLE,
};

struct logchannel {
	/*! What to log to this channel */
	unsigned int logmask;
	/*! If this channel is disabled or not */
	int disabled;
	/*! syslog facility */
	int facility;
	/*! Verbosity level. (-1 if use option_verbose for the level.) */
	int verbosity;
	/*! Type of log channel */
	enum logtypes type;
	/*! logfile logging file pointer */
	FILE *fileptr;
	/*! Filename */
	char filename[PATH_MAX];
	/*! field for linking to list */
	AST_LIST_ENTRY(logchannel) list;
	/*! Line number from configuration file */
	int lineno;
	/*! Whether this log channel was created dynamically */
	int dynamic;
	/*! Components (levels) from last config load */
	char components[0];
};

static AST_RWLIST_HEAD_STATIC(logchannels, logchannel);

#ifndef GRANDSTREAM_NETWORKS
enum logmsgtypes {
	LOGMSG_NORMAL = 0,
	LOGMSG_VERBOSE,
};
#endif

struct logmsg {
#ifdef GRANDSTREAM_NETWORKS
    struct timeval datetime; 
    unsigned int level     :5;
    unsigned int reserve_27:27;
#else
	enum logmsgtypes type;
	int level;
#endif
    int line;
	int lwp;
#ifndef GRANDSTREAM_NETWORKS	
	struct ast_callid *callid; 
#endif
    AST_DECLARE_STRING_FIELDS(
#ifndef GRANDSTREAM_NETWORKS		
        AST_STRING_FIELD(date);  
#endif
        AST_STRING_FIELD(file);
        AST_STRING_FIELD(function);
        AST_STRING_FIELD(message);
#ifndef GRANDSTREAM_NETWORKS		
        AST_STRING_FIELD(level_name);  
#endif		
    );
    AST_LIST_ENTRY(logmsg) list;
};

static void logmsg_free(struct logmsg *msg)
{
#ifndef GRANDSTREAM_NETWORKS		
	if (msg->callid) {
		ast_callid_unref(msg->callid);
	}
#endif
	ast_free(msg);
}

static AST_LIST_HEAD_STATIC(logmsgs, logmsg);

#ifdef GRANDSTREAM_NETWORKS
static unsigned long logmsgs_size = 0;
static unsigned long logmsgs_drop_total = 0;
static unsigned long logmsgs_drop_level[32] = {0};
#endif

static pthread_t logthread = AST_PTHREADT_NULL;
static ast_cond_t logcond;
static int close_logger_thread = 0;

static FILE *qlog;

/*! \brief Logging channels used in the Asterisk logging system
 *
 * The first 16 levels are reserved for system usage, and the remaining
 * levels are reserved for usage by dynamic levels registered via
 * ast_logger_register_level.
 */

/* Modifications to this array are protected by the rwlock in the
 * logchannels list.
 */
#ifdef GRANDSTREAM_NETWORKS 
static char *levels[NUMLOGLEVELS] = {
	"DEBUG",
	"---EVENT---",		/* no longer used */
	"NOTIC",
	"WARN",
	"ERROR",
	"VERB",
	"DTMF",
	"CODEC_NEG",
};
#else
static char *levels[NUMLOGLEVELS] = {
	"DEBUG",
	"---EVENT---",		/* no longer used */
	"NOTICE",
	"WARNING",
	"ERROR",
	"VERBOSE",
	"DTMF",
};
#endif

#ifdef GRANDSTREAM_NETWORKS
/* When log record number larger than drop_limit[level], drop that log! */
static const unsigned long drop_limit[32] = {
    3000,
    5000,
    6000,
    7000,
    9999,
    4000,
    8000,
    6000,

    6000,
    6000,
    6000,
    6000,
    6000,
    6000,
    6000,
    6000,

    5000,
    5000,
    5000,
    5000,
    5000,
    5000,
    5000,
    5000,

    5000,
    5000,
    5000,
    5000,
    5000,
    5000,
    5000,
    5000
};

#define DYNAMIC_LEVEL_START 16 /* dynamic level start at 16, before is static level, see levels */

static char *log_formats[16] = {
	"date",
	"level",		
	"PID",
	"filename",
	"line",
	"funcname",
	"MAC",
	"VERSION",
	NULL, NULL, NULL, NULL, 
	NULL, NULL, NULL, NULL
};

typedef enum 
{
    EN_DATE = 0,
    EN_LEVEL,
    EN_PID,
    EN_FILENAME,
    EN_LINE,
    EN_FUNCNAME,
    EN_MAC,
    EN_VERSION
    
} EnLogFormat;

static unsigned short g_log_format = 0x001F; /* default log format */

/** struct for files with the module mapping. */
typedef struct __LOG_FILE_MODULE__
{
    char file_name[64];
    char module_name[64];
    int  file_load_index;  /* order by file load */
    int  module_index;
    AST_RWLIST_ENTRY(__LOG_FILE_MODULE__) entry;

} LogFileModule;

AST_RWLIST_HEAD(LogFileModuleList, __LOG_FILE_MODULE__);

/* list of files name and module name map */
static struct LogFileModuleList g_log_file_module_list;


/** list struct for files on one module. */
typedef struct __LOG_FILE_NAME__
{
    char file_name[64];
    int  file_index;   /* File index whitin module */
    int  module_index; /* Module index, this is a redundancy field. Is used to verify 'module_index' in '__LOG_MODULE_FILE__' */
    AST_RWLIST_ENTRY(__LOG_FILE_NAME__) entry;

} LogFilename;

AST_RWLIST_HEAD(LogFilenameList, __LOG_FILE_NAME__);

/** struct for modules with the file mapping. */
typedef struct __LOG_MODULE_FILE__
{
    char module_name[64];
    int  module_index;
    int  file_count; 
    struct LogFilenameList files_list;
    
    AST_RWLIST_ENTRY(__LOG_MODULE_FILE__) entry;
    
} LogModuleFile;

AST_RWLIST_HEAD(LogModuleFileList, __LOG_MODULE_FILE__);

static struct LogModuleFileList g_log_module_file_list; /* order by module */

static int g_log_file_module_list_count = 0;

/* All modules level. 32 bit; bit 0 is __LOG_DEBUG, bit 2 is __LOG_NOTICE... */
int ast_log_modules_level[256] = {0};

/* modules count */
int ast_log_modules_count = 0;

/* 32 bit, dynamic level control switch  */
unsigned long ast_log_dynamic_levels_switch = 0xFFFFFFFF; 

static unsigned long g_log_dynamic_register_count = 0;
static unsigned long g_log_dynamic_unregister_count = 0;

static char g_log_unvalid_file[32]     = {0};
static int  g_log_unvalid_line         = -10;
static char g_log_unvalid_callfile[32] = {0};
static int  g_log_unvalid_callline     = -10;
static int  g_log_unvalid_value        = -10;

/** log to file; not implements **/
#define LOG_FILE_MAX_LENGTH (3000 * 1024) /* Max logfile length is 3000K */
#define LOG_FILE_MAX_COUNT  10   /* max file count */
#define LOG_FILE_MAX_SUBFIX 65535  /* log_MAC_VERSION.0 log_MAC_VERSION.1 ... log_MAC_VERSION.65535 */

static FILE *g_log_fileptr = NULL; /* log to file pointer */

#endif

/*! \brief Colors used in the console for logging */
static const int colors[NUMLOGLEVELS] = {
	COLOR_BRGREEN,
	COLOR_BRBLUE,		/* no longer used */
	COLOR_YELLOW,
	COLOR_BRRED,
	COLOR_RED,
	COLOR_GREEN,
	COLOR_BRGREEN,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
	COLOR_BRBLUE,
};

AST_THREADSTORAGE(verbose_buf);
AST_THREADSTORAGE(verbose_build_buf);
#define VERBOSE_BUF_INIT_SIZE   256

AST_THREADSTORAGE(log_buf);
#define LOG_BUF_INIT_SIZE       256

static void logger_queue_init(void);

#ifdef GRANDSTREAM_NETWORKS
static int logger_switch_open_sqlite3(sqlite3** db);
static int logger_sqlite3_add_record(sqlite3* db, const char *table, const char **fields, const char **values, int size);
static int logger_sqlite3_delete_record(sqlite3* db, const char *table, const char *field, const char *value);
static int logger_sqlite3_update_record(sqlite3* db, const char *table, const char **fieldset, const char **newset, int size, const char *where, const char *value);
static int logger_sqlite3_record_isexist(sqlite3* db, const char *table, const char *field, const char *value);

static char *handle_logger_save_format(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_save_switch(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_set_module_level(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_set_dlevel(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_set_format(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_set_default(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_test_module_level(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_test_dlevel(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_test_printf(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_show_file2module(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_show_module2file(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_show_information(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_show_level(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_show_switch(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_logger_show_format(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
#endif

static void logger_queue_init(void);

static void make_components(struct logchannel *chan)
{
	char *w;
	unsigned int logmask = 0;
	char *stringp = ast_strdupa(chan->components);
	unsigned int x;
	unsigned int verb_level;

	/* Default to using option_verbose as the verbosity level of the logging channel.  */
	verb_level = -1;

	while ((w = strsep(&stringp, ","))) {
		w = ast_strip(w);
		if (ast_strlen_zero(w)) {
			continue;
		}
		if (!strcmp(w, "*")) {
			logmask = 0xFFFFFFFF;
		} else if (!strncasecmp(w, "verbose(", 8)) {
			if (levels[__LOG_VERBOSE] && sscanf(w + 8, "%30u)", &verb_level) == 1) {
				logmask |= (1 << __LOG_VERBOSE);
			}
		} else {
			for (x = 0; x < ARRAY_LEN(levels); ++x) {
				if (levels[x] && !strcasecmp(w, levels[x])) {
					logmask |= (1 << x);
					break;
				}
			}
		}
	}
	if (chan->type == LOGTYPE_CONSOLE) {
		/*
		 * Force to use the root console verbose level so if the
		 * user specified any verbose level then it does not interfere
		 * with calculating the ast_verb_sys_level value.
		 */
		chan->verbosity = -1;
	} else {
		chan->verbosity = verb_level;
	}
	chan->logmask = logmask;
}

static struct logchannel *make_logchannel(const char *channel, const char *components, int lineno, int dynamic)
{
	struct logchannel *chan;
	char *facility;
	struct ast_tm tm;
	struct timeval now = ast_tvnow();
	char datestring[256];

	if (ast_strlen_zero(channel) || !(chan = ast_calloc(1, sizeof(*chan) + strlen(components) + 1)))
		return NULL;

	strcpy(chan->components, components);
	chan->lineno = lineno;
	chan->dynamic = dynamic;

	if (!strcasecmp(channel, "console")) {
		chan->type = LOGTYPE_CONSOLE;
	} else if (!strncasecmp(channel, "syslog", 6)) {
		/*
		* syntax is:
		*  syslog.facility => level,level,level
		*/
		facility = strchr(channel, '.');
		if (!facility++ || !facility) {
			facility = "local0";
		}

		chan->facility = ast_syslog_facility(facility);

		if (chan->facility < 0) {
			fprintf(stderr, "Logger Warning: bad syslog facility in logger.conf\n");
			ast_free(chan);
			return NULL;
		}

		chan->type = LOGTYPE_SYSLOG;
		ast_copy_string(chan->filename, channel, sizeof(chan->filename));
		openlog("asterisk", LOG_PID, chan->facility);
	} else {
		const char *log_dir_prefix = "";
		const char *log_dir_separator = "";

		if (channel[0] != '/') {
			log_dir_prefix = ast_config_AST_LOG_DIR;
			log_dir_separator = "/";
		}

		if (!ast_strlen_zero(hostname)) {
			snprintf(chan->filename, sizeof(chan->filename), "%s%s%s.%s",
				log_dir_prefix, log_dir_separator, channel, hostname);
		} else {
			snprintf(chan->filename, sizeof(chan->filename), "%s%s%s",
				log_dir_prefix, log_dir_separator, channel);
		}

		if (!(chan->fileptr = fopen(chan->filename, "a"))) {
			/* Can't do real logging here since we're called with a lock
			 * so log to any attached consoles */
			ast_console_puts_mutable("ERROR: Unable to open log file '", __LOG_ERROR);
			ast_console_puts_mutable(chan->filename, __LOG_ERROR);
			ast_console_puts_mutable("': ", __LOG_ERROR);
			ast_console_puts_mutable(strerror(errno), __LOG_ERROR);
			ast_console_puts_mutable("'\n", __LOG_ERROR);
			ast_free(chan);
			return NULL;
		} else {
			/* Create our date/time */
			ast_localtime(&now, &tm, NULL);
			ast_strftime(datestring, sizeof(datestring), dateformat, &tm);

			fprintf(chan->fileptr, "[%s] Asterisk %s built by %s @ %s on a %s running %s on %s\n",
				datestring, ast_get_version(), ast_build_user, ast_build_hostname,
				ast_build_machine, ast_build_os, ast_build_date);
			fflush(chan->fileptr);
		}
		chan->type = LOGTYPE_FILE;
	}
	make_components(chan);

	return chan;
}

/* \brief Read config, setup channels.
 * \param locked The logchannels list is locked and this is a reload
 * \param altconf Alternate configuration file to read.
 *
 * \retval 0 Success
 * \retval -1 No config found or Failed
 */
static int init_logger_chain(int locked, const char *altconf)
{
	struct logchannel *chan;
	struct ast_config *cfg;
	struct ast_variable *var;
	const char *s;
	struct ast_flags config_flags = { 0 };

	if (!(cfg = ast_config_load2(S_OR(altconf, "logger.conf"), "logger", config_flags)) || cfg == CONFIG_STATUS_FILEINVALID) {
		cfg = NULL;
	}

	if (!locked) {
		AST_RWLIST_WRLOCK(&logchannels);
	}

	/* Set defaults */
	hostname[0] = '\0';
	display_callids = 1;
	memset(&logfiles, 0, sizeof(logfiles));
	logfiles.queue_log = 1;
#ifdef GRANDSTREAM_NETWORKS
	ast_copy_string(dateformat, "%Y%m%d %H:%M:%S.%3q", sizeof(dateformat));
#else
        ast_copy_string(dateformat, "%b %e %T", sizeof(dateformat));
#endif
	ast_copy_string(queue_log_name, QUEUELOG, sizeof(queue_log_name));
	exec_after_rotate[0] = '\0';
	rotatestrategy = SEQUENTIAL;

	/* delete our list of log channels */
	while ((chan = AST_RWLIST_REMOVE_HEAD(&logchannels, list))) {
		ast_free(chan);
	}
	global_logmask = 0;
	if (!locked) {
		AST_RWLIST_UNLOCK(&logchannels);
	}

	errno = 0;
	/* close syslog */
	closelog();

	/* If no config file, we're fine, set default options. */
	if (!cfg) {
		if (!(chan = ast_calloc(1, sizeof(*chan)))) {
			fprintf(stderr, "Failed to initialize default logging\n");
			return -1;
		}
		chan->type = LOGTYPE_CONSOLE;
		chan->logmask = __LOG_WARNING | __LOG_NOTICE | __LOG_ERROR;
		if (!locked) {
			AST_RWLIST_WRLOCK(&logchannels);
		}
		AST_RWLIST_INSERT_HEAD(&logchannels, chan, list);
		global_logmask |= chan->logmask;
		if (!locked) {
			AST_RWLIST_UNLOCK(&logchannels);
		}

		return -1;
	}

	if ((s = ast_variable_retrieve(cfg, "general", "appendhostname"))) {
		if (ast_true(s)) {
			if (gethostname(hostname, sizeof(hostname) - 1)) {
				ast_copy_string(hostname, "unknown", sizeof(hostname));
				fprintf(stderr, "What box has no hostname???\n");
			}
		}
	}
	if ((s = ast_variable_retrieve(cfg, "general", "display_callids"))) {
		display_callids = ast_true(s);
	}
	if ((s = ast_variable_retrieve(cfg, "general", "dateformat"))) {
		ast_copy_string(dateformat, s, sizeof(dateformat));
	}
	if ((s = ast_variable_retrieve(cfg, "general", "queue_log"))) {
		logfiles.queue_log = ast_true(s);
	}
	if ((s = ast_variable_retrieve(cfg, "general", "queue_log_to_file"))) {
		logfiles.queue_log_to_file = ast_true(s);
	}
	if ((s = ast_variable_retrieve(cfg, "general", "queue_log_name"))) {
		ast_copy_string(queue_log_name, s, sizeof(queue_log_name));
	}
	if ((s = ast_variable_retrieve(cfg, "general", "queue_log_realtime_use_gmt"))) {
		logfiles.queue_log_realtime_use_gmt = ast_true(s);
	}
	if ((s = ast_variable_retrieve(cfg, "general", "exec_after_rotate"))) {
		ast_copy_string(exec_after_rotate, s, sizeof(exec_after_rotate));
	}
	if ((s = ast_variable_retrieve(cfg, "general", "rotatestrategy"))) {
		if (strcasecmp(s, "timestamp") == 0) {
			rotatestrategy = TIMESTAMP;
		} else if (strcasecmp(s, "rotate") == 0) {
			rotatestrategy = ROTATE;
		} else if (strcasecmp(s, "sequential") == 0) {
			rotatestrategy = SEQUENTIAL;
		} else if (strcasecmp(s, "none") == 0) {
			rotatestrategy = NONE;
		} else {
			fprintf(stderr, "Unknown rotatestrategy: %s\n", s);
		}
	} else {
		if ((s = ast_variable_retrieve(cfg, "general", "rotatetimestamp"))) {
			rotatestrategy = ast_true(s) ? TIMESTAMP : SEQUENTIAL;
			fprintf(stderr, "rotatetimestamp option has been deprecated.  Please use rotatestrategy instead.\n");
		}
	}

	if (!locked) {
		AST_RWLIST_WRLOCK(&logchannels);
	}
	var = ast_variable_browse(cfg, "logfiles");
	for (; var; var = var->next) {
		if (!(chan = make_logchannel(var->name, var->value, var->lineno, 0))) {
			/* Print error message directly to the consoles since the lock is held
			 * and we don't want to unlock with the list partially built */
			ast_console_puts_mutable("ERROR: Unable to create log channel '", __LOG_ERROR);
			ast_console_puts_mutable(var->name, __LOG_ERROR);
			ast_console_puts_mutable("'\n", __LOG_ERROR);
			continue;
		}
		AST_RWLIST_INSERT_HEAD(&logchannels, chan, list);
		global_logmask |= chan->logmask;
	}

	if (qlog) {
		fclose(qlog);
		qlog = NULL;
	}

	if (!locked) {
		AST_RWLIST_UNLOCK(&logchannels);
	}

	ast_config_destroy(cfg);

	return 0;
}

void ast_child_verbose(int level, const char *fmt, ...)
{
	char *msg = NULL, *emsg = NULL, *sptr, *eptr;
	va_list ap, aq;
	int size;

	va_start(ap, fmt);
	va_copy(aq, ap);
	if ((size = vsnprintf(msg, 0, fmt, ap)) < 0) {
		va_end(ap);
		va_end(aq);
		return;
	}
	va_end(ap);

	if (!(msg = ast_malloc(size + 1))) {
		va_end(aq);
		return;
	}

	vsnprintf(msg, size + 1, fmt, aq);
	va_end(aq);

	if (!(emsg = ast_malloc(size * 2 + 1))) {
		ast_free(msg);
		return;
	}

	for (sptr = msg, eptr = emsg; ; sptr++) {
		if (*sptr == '"') {
			*eptr++ = '\\';
		}
		*eptr++ = *sptr;
		if (*sptr == '\0') {
			break;
		}
	}
	ast_free(msg);

	fprintf(stdout, "verbose \"%s\" %d\n", emsg, level);
	fflush(stdout);
	ast_free(emsg);
}

void ast_queue_log(const char *queuename, const char *callid, const char *agent, const char *event, const char *fmt, ...)
{
	va_list ap;
	struct timeval tv;
	struct ast_tm tm;
	char qlog_msg[8192];
	int qlog_len;
	char time_str[30];

	if (!logger_initialized) {
		/* You are too early.  We are not open yet! */
		return;
	}
	if (!queuelog_init) {
		/* We must initialize now since someone is trying to log something. */
		logger_queue_start();
	}

	if (ast_check_realtime("queue_log")) {
		tv = ast_tvnow();
		ast_localtime(&tv, &tm, logfiles.queue_log_realtime_use_gmt ? "GMT" : NULL);
		ast_strftime(time_str, sizeof(time_str), "%F %T.%6q", &tm);
		va_start(ap, fmt);
		vsnprintf(qlog_msg, sizeof(qlog_msg), fmt, ap);
		va_end(ap);
		if (logfiles.queue_adaptive_realtime) {
			AST_DECLARE_APP_ARGS(args,
				AST_APP_ARG(data)[5];
			);
			AST_NONSTANDARD_APP_ARGS(args, qlog_msg, '|');
			/* Ensure fields are large enough to receive data */
			ast_realtime_require_field("queue_log",
				"data1", RQ_CHAR, strlen(S_OR(args.data[0], "")),
				"data2", RQ_CHAR, strlen(S_OR(args.data[1], "")),
				"data3", RQ_CHAR, strlen(S_OR(args.data[2], "")),
				"data4", RQ_CHAR, strlen(S_OR(args.data[3], "")),
				"data5", RQ_CHAR, strlen(S_OR(args.data[4], "")),
				SENTINEL);

			/* Store the log */
			ast_store_realtime("queue_log", "time", time_str,
				"callid", callid,
				"queuename", queuename,
				"agent", agent,
				"event", event,
				"data1", S_OR(args.data[0], ""),
				"data2", S_OR(args.data[1], ""),
				"data3", S_OR(args.data[2], ""),
				"data4", S_OR(args.data[3], ""),
				"data5", S_OR(args.data[4], ""),
				SENTINEL);
		} else {
			ast_store_realtime("queue_log", "time", time_str,
				"callid", callid,
				"queuename", queuename,
				"agent", agent,
				"event", event,
				"data", qlog_msg,
				SENTINEL);
		}

		if (!logfiles.queue_log_to_file) {
			return;
		}
	}

	if (qlog) {
		va_start(ap, fmt);
		qlog_len = snprintf(qlog_msg, sizeof(qlog_msg), "%ld|%s|%s|%s|%s|", (long)time(NULL), callid, queuename, agent, event);
		vsnprintf(qlog_msg + qlog_len, sizeof(qlog_msg) - qlog_len, fmt, ap);
		va_end(ap);
		AST_RWLIST_RDLOCK(&logchannels);
		if (qlog) {
			fprintf(qlog, "%s\n", qlog_msg);
			fflush(qlog);
		}
		AST_RWLIST_UNLOCK(&logchannels);
	}
}

static int rotate_file(const char *filename)
{
	char old[PATH_MAX];
	char new[PATH_MAX];
	int x, y, which, found, res = 0, fd;
	char *suffixes[4] = { "", ".gz", ".bz2", ".Z" };

	switch (rotatestrategy) {
	case NONE:
		/* No rotation */
		break;
	case SEQUENTIAL:
		for (x = 0; ; x++) {
			snprintf(new, sizeof(new), "%s.%d", filename, x);
			fd = open(new, O_RDONLY);
			if (fd > -1)
				close(fd);
			else
				break;
		}
		if (rename(filename, new)) {
			fprintf(stderr, "Unable to rename file '%s' to '%s'\n", filename, new);
			res = -1;
		} else {
			filename = new;
		}
		break;
	case TIMESTAMP:
		snprintf(new, sizeof(new), "%s.%ld", filename, (long)time(NULL));
		if (rename(filename, new)) {
			fprintf(stderr, "Unable to rename file '%s' to '%s'\n", filename, new);
			res = -1;
		} else {
			filename = new;
		}
		break;
	case ROTATE:
		/* Find the next empty slot, including a possible suffix */
		for (x = 0; ; x++) {
			found = 0;
			for (which = 0; which < ARRAY_LEN(suffixes); which++) {
				snprintf(new, sizeof(new), "%s.%d%s", filename, x, suffixes[which]);
				fd = open(new, O_RDONLY);
				if (fd > -1) {
					close(fd);
					found = 1;
					break;
				}
			}
			if (!found) {
				break;
			}
		}

		/* Found an empty slot */
		for (y = x; y > 0; y--) {
			for (which = 0; which < ARRAY_LEN(suffixes); which++) {
				snprintf(old, sizeof(old), "%s.%d%s", filename, y - 1, suffixes[which]);
				fd = open(old, O_RDONLY);
				if (fd > -1) {
					/* Found the right suffix */
					close(fd);
					snprintf(new, sizeof(new), "%s.%d%s", filename, y, suffixes[which]);
					if (rename(old, new)) {
						fprintf(stderr, "Unable to rename file '%s' to '%s'\n", old, new);
						res = -1;
					}
					break;
				}
			}
		}

		/* Finally, rename the current file */
		snprintf(new, sizeof(new), "%s.0", filename);
		if (rename(filename, new)) {
			fprintf(stderr, "Unable to rename file '%s' to '%s'\n", filename, new);
			res = -1;
		} else {
			filename = new;
		}
	}

	if (!ast_strlen_zero(exec_after_rotate)) {
		struct ast_channel *c = ast_dummy_channel_alloc();
		char buf[512];

		pbx_builtin_setvar_helper(c, "filename", filename);
		pbx_substitute_variables_helper(c, exec_after_rotate, buf, sizeof(buf));
		if (c) {
			c = ast_channel_unref(c);
		}
		if (ast_safe_system(buf) == -1) {
			ast_log(LOG_WARNING, "error executing '%s'\n", buf);
		}
	}
	return res;
}

/*!
 * \internal
 * \brief Start the realtime queue logging if configured.
 *
 * \retval TRUE if not to open queue log file.
 */
static int logger_queue_rt_start(void)
{
	if (ast_check_realtime("queue_log")) {
		if (!ast_realtime_require_field("queue_log",
			"time", RQ_DATETIME, 26,
			"data1", RQ_CHAR, 20,
			"data2", RQ_CHAR, 20,
			"data3", RQ_CHAR, 20,
			"data4", RQ_CHAR, 20,
			"data5", RQ_CHAR, 20,
			SENTINEL)) {
			logfiles.queue_adaptive_realtime = 1;
		} else {
			logfiles.queue_adaptive_realtime = 0;
		}

		if (!logfiles.queue_log_to_file) {
			/* Don't open the log file. */
			return 1;
		}
	}
	return 0;
}

/*!
 * \internal
 * \brief Rotate the queue log file and restart.
 *
 * \param queue_rotate Log queue rotation mode.
 *
 * \note Assumes logchannels is write locked on entry.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int logger_queue_restart(int queue_rotate)
{
	int res = 0;
	char qfname[PATH_MAX];

	if (logger_queue_rt_start()) {
		return res;
	}

	snprintf(qfname, sizeof(qfname), "%s/%s", ast_config_AST_LOG_DIR, queue_log_name);
	if (qlog) {
		/* Just in case it was still open. */
		fclose(qlog);
		qlog = NULL;
	}
	if (queue_rotate) {
		rotate_file(qfname);
	}

	/* Open the log file. */
	qlog = fopen(qfname, "a");
	if (!qlog) {
		ast_log(LOG_ERROR, "Unable to create queue log: %s\n", strerror(errno));
		res = -1;
	}
	return res;
}

static int reload_logger(int rotate, const char *altconf)
{
	int queue_rotate = rotate;
	struct logchannel *f;
	int res = 0;

	AST_RWLIST_WRLOCK(&logchannels);

	if (qlog) {
		if (rotate < 0) {
			/* Check filesize - this one typically doesn't need an auto-rotate */
			if (ftello(qlog) > 0x40000000) { /* Arbitrarily, 1 GB */
				fclose(qlog);
				qlog = NULL;
			} else {
				queue_rotate = 0;
			}
		} else {
			fclose(qlog);
			qlog = NULL;
		}
	} else {
		queue_rotate = 0;
	}

	ast_mkdir(ast_config_AST_LOG_DIR, 0777);

	AST_RWLIST_TRAVERSE(&logchannels, f, list) {
		if (f->disabled) {
			f->disabled = 0;	/* Re-enable logging at reload */
			/*** DOCUMENTATION
				<managerEventInstance>
					<synopsis>Raised when a logging channel is re-enabled after a reload operation.</synopsis>
					<syntax>
						<parameter name="Channel">
							<para>The name of the logging channel.</para>
						</parameter>
					</syntax>
				</managerEventInstance>
			***/
			manager_event(EVENT_FLAG_SYSTEM, "LogChannel", "Channel: %s\r\nEnabled: Yes\r\n", f->filename);
		}
		if (f->fileptr && (f->fileptr != stdout) && (f->fileptr != stderr)) {
			int rotate_this = 0;
			if (rotatestrategy != NONE && ftello(f->fileptr) > 0x40000000) { /* Arbitrarily, 1 GB */
				/* Be more proactive about rotating massive log files */
				rotate_this = 1;
			}
			fclose(f->fileptr);	/* Close file */
			f->fileptr = NULL;
			if (rotate || rotate_this) {
				rotate_file(f->filename);
			}
		}
	}

	filesize_reload_needed = 0;

	init_logger_chain(1 /* locked */, altconf);

	ast_unload_realtime("queue_log");
	if (logfiles.queue_log) {
		res = logger_queue_restart(queue_rotate);
		AST_RWLIST_UNLOCK(&logchannels);
#ifndef GRANDSTREAM_NETWORKS
		ast_verb_update(); 
#endif
		ast_queue_log("NONE", "NONE", "NONE", "CONFIGRELOAD", "%s", "");
		ast_verb(1, "Asterisk Queue Logger restarted\n");
	} else {
		AST_RWLIST_UNLOCK(&logchannels);
#ifndef GRANDSTREAM_NETWORKS
		ast_verb_update();
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
    logger_init_switch_from_db();
    logger_init_format_from_db();
#endif  
	
	return res;
}

/*! \brief Reload the logger module without rotating log files (also used from loader.c during
	a full Asterisk reload) */
int logger_reload(void)
{
	if (reload_logger(0, NULL)) {
		return RESULT_FAILURE;
	}
	return RESULT_SUCCESS;
}

static char *handle_logger_reload(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "logger reload";
		e->usage =
			"Usage: logger reload [<alt-conf>]\n"
			"       Reloads the logger subsystem state.  Use after restarting syslogd(8) if you are using syslog logging.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}
	if (reload_logger(0, a->argc == 3 ? a->argv[2] : NULL)) {
		ast_cli(a->fd, "Failed to reload the logger\n");
		return CLI_FAILURE;
	}
	return CLI_SUCCESS;
}

static char *handle_logger_rotate(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "logger rotate";
		e->usage =
			"Usage: logger rotate\n"
			"       Rotates and Reopens the log files.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}
	if (reload_logger(1, NULL)) {
		ast_cli(a->fd, "Failed to reload the logger and rotate log files\n");
		return CLI_FAILURE;
	}
	return CLI_SUCCESS;
}

int ast_logger_rotate()
{
	return reload_logger(1, NULL);
}

#ifndef GRANDSTREAM_NETWORKS
static char *handle_logger_set_level(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int x;
	int state;
	int level = -1;

	switch (cmd) {
	case CLI_INIT:
		e->command = "logger set level {DEBUG|NOTICE|WARNING|ERROR|VERBOSE|DTMF} {on|off}";
		e->usage =
			"Usage: logger set level {DEBUG|NOTICE|WARNING|ERROR|VERBOSE|DTMF} {on|off}\n"
			"       Set a specific log level to enabled/disabled for this console.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc < 5)
		return CLI_SHOWUSAGE;

	AST_RWLIST_WRLOCK(&logchannels);

	for (x = 0; x < ARRAY_LEN(levels); x++) {
		if (levels[x] && !strcasecmp(a->argv[3], levels[x])) {
			level = x;
			break;
		}
	}

	AST_RWLIST_UNLOCK(&logchannels);

	state = ast_true(a->argv[4]) ? 1 : 0;

	if (level != -1) {
		ast_console_toggle_loglevel(a->fd, level, state);
		ast_cli(a->fd, "Logger status for '%s' has been set to '%s'.\n", levels[level], state ? "on" : "off");
	} else
		return CLI_SHOWUSAGE;

	return CLI_SUCCESS;
}
#endif

/*! \brief CLI command to show logging system configuration */
static char *handle_logger_show_channels(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
#define FORMATL	"%-35.35s %-8.8s %-9.9s "
	struct logchannel *chan;
	switch (cmd) {
	case CLI_INIT:
		e->command = "logger show channels";
		e->usage =
			"Usage: logger show channels\n"
			"       List configured logger channels.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}
	ast_cli(a->fd, FORMATL, "Channel", "Type", "Status");
	ast_cli(a->fd, "Configuration\n");
	ast_cli(a->fd, FORMATL, "-------", "----", "------");
	ast_cli(a->fd, "-------------\n");
	AST_RWLIST_RDLOCK(&logchannels);
	AST_RWLIST_TRAVERSE(&logchannels, chan, list) {
		unsigned int level;

		ast_cli(a->fd, FORMATL, chan->filename, chan->type == LOGTYPE_CONSOLE ? "Console" : (chan->type == LOGTYPE_SYSLOG ? "Syslog" : "File"),
			chan->disabled ? "Disabled" : "Enabled");
		ast_cli(a->fd, " - ");
		for (level = 0; level < ARRAY_LEN(levels); level++) {
			if ((chan->logmask & (1 << level)) && levels[level]) {
				ast_cli(a->fd, "%s ", levels[level]);
			}
		}
		ast_cli(a->fd, "\n");
	}
	AST_RWLIST_UNLOCK(&logchannels);
	ast_cli(a->fd, "\n");

	return CLI_SUCCESS;
}

static char *handle_logger_add_channel(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct logchannel *chan;

	switch (cmd) {
	case CLI_INIT:
		e->command = "logger add channel";
		e->usage =
			"Usage: logger add channel <name> <levels>\n"
			"       Adds a temporary logger channel. This logger channel\n"
			"       will exist until removed or until Asterisk is restarted.\n"
			"       <levels> is a comma-separated list of desired logger\n"
			"       levels such as: verbose,warning,error\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc < 5) {
		return CLI_SHOWUSAGE;
	}

	AST_RWLIST_WRLOCK(&logchannels);
	AST_RWLIST_TRAVERSE(&logchannels, chan, list) {
		if (!strcmp(chan->filename, a->argv[3])) {
			break;
		}
	}

	if (chan) {
		AST_RWLIST_UNLOCK(&logchannels);
		ast_cli(a->fd, "Logger channel '%s' already exists\n", a->argv[3]);
		return CLI_SUCCESS;
	}

	chan = make_logchannel(a->argv[3], a->argv[4], 0, 1);
	if (chan) {
		AST_RWLIST_INSERT_HEAD(&logchannels, chan, list);
		global_logmask |= chan->logmask;
		AST_RWLIST_UNLOCK(&logchannels);
		return CLI_SUCCESS;
	}

	AST_RWLIST_UNLOCK(&logchannels);
	ast_cli(a->fd, "ERROR: Unable to create log channel '%s'\n", a->argv[3]);

	return CLI_FAILURE;
}

static char *handle_logger_remove_channel(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct logchannel *chan;
	int gen_count = 0;
	char *gen_ret = NULL;

	switch (cmd) {
	case CLI_INIT:
		e->command = "logger remove channel";
		e->usage =
			"Usage: logger remove channel <name>\n"
			"       Removes a temporary logger channel.\n";
		return NULL;
	case CLI_GENERATE:
		if (a->argc > 4 || (a->argc == 4 && a->pos > 3)) {
			return NULL;
		}
		AST_RWLIST_RDLOCK(&logchannels);
		AST_RWLIST_TRAVERSE(&logchannels, chan, list) {
			if (chan->dynamic && (ast_strlen_zero(a->argv[3])
				|| !strncmp(a->argv[3], chan->filename, strlen(a->argv[3])))) {
				if (gen_count == a->n) {
					gen_ret = ast_strdup(chan->filename);
					break;
				}
				gen_count++;
			}
		}
		AST_RWLIST_UNLOCK(&logchannels);
		return gen_ret;
	}

	if (a->argc < 4) {
		return CLI_SHOWUSAGE;
	}

	AST_RWLIST_WRLOCK(&logchannels);
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&logchannels, chan, list) {
		if (chan->dynamic && !strcmp(chan->filename, a->argv[3])) {
			AST_RWLIST_REMOVE_CURRENT(list);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
	AST_RWLIST_UNLOCK(&logchannels);

	if (!chan) {
		ast_cli(a->fd, "Unable to find dynamic logger channel '%s'\n", a->argv[3]);
		return CLI_SUCCESS;
	}

	ast_cli(a->fd, "Removed dynamic logger channel '%s'\n", chan->filename);
	if (chan->fileptr) {
		fclose(chan->fileptr);
		chan->fileptr = NULL;
	}
	ast_free(chan);
	chan = NULL;

	return CLI_SUCCESS;
}

struct verb {
	void (*verboser)(const char *string);
	AST_LIST_ENTRY(verb) list;
};

static AST_RWLIST_HEAD_STATIC(verbosers, verb);

static struct ast_cli_entry cli_logger[] = {
	AST_CLI_DEFINE(handle_logger_show_channels, "List configured log channels"),
	AST_CLI_DEFINE(handle_logger_reload, "Reopens the log files"),
	AST_CLI_DEFINE(handle_logger_rotate, "Rotates and reopens the log files"),
#ifndef GRANDSTREAM_NETWORKS
	AST_CLI_DEFINE(handle_logger_set_level, "Enables/Disables a specific logging level for this console"), 
#endif	
	AST_CLI_DEFINE(handle_logger_add_channel, "Adds a new logging channel"),
	AST_CLI_DEFINE(handle_logger_remove_channel, "Removes a logging channel"),

#ifdef GRANDSTREAM_NETWORKS
	AST_CLI_DEFINE(handle_logger_save_format, "Save logger format, see 'logger show format'"),
    AST_CLI_DEFINE(handle_logger_save_switch, "Save logger switch, contain dynamic level and static level"),
    AST_CLI_DEFINE(handle_logger_set_module_level, "Enables/Disables a specific logging level with module for log output"),
    AST_CLI_DEFINE(handle_logger_set_dlevel, "Set dynamic level log switch"),
    AST_CLI_DEFINE(handle_logger_set_format, "Set log format"),
    AST_CLI_DEFINE(handle_logger_set_default, "Set default Logger switch"),
    AST_CLI_DEFINE(handle_logger_show_file2module, "Show file with the module mapping"),
    AST_CLI_DEFINE(handle_logger_show_module2file, "Show module with the file mapping"),
    AST_CLI_DEFINE(handle_logger_show_information, "Show logger information"),
    AST_CLI_DEFINE(handle_logger_show_level, "Show logger level information"),
    AST_CLI_DEFINE(handle_logger_show_switch, "show logger switch"),
    AST_CLI_DEFINE(handle_logger_show_format, "show logger format for current set"),
    AST_CLI_DEFINE(handle_logger_test_module_level, "Simulation a log and output "),
    AST_CLI_DEFINE(handle_logger_test_dlevel, "Simulation a dynamic level log and output "),
    AST_CLI_DEFINE(handle_logger_test_printf, "Simulation a printf output string "),
#endif
};

static void _handle_SIGXFSZ(int sig)
{
	/* Indicate need to reload */
	filesize_reload_needed = 1;
}

static struct sigaction handle_SIGXFSZ = {
	.sa_handler = _handle_SIGXFSZ,
	.sa_flags = SA_RESTART,
};

#ifdef GRANDSTREAM_NETWORKS
static inline void ast_log_vsyslog(int level, char *buf, int max_buf_len)
{
    int syslog_level = ast_syslog_priority_from_loglevel(level);

    if (syslog_level < 0) {
        /* we are locked here, so cannot ast_log() */
        fprintf(stderr, "ast_log_vsyslog called with bogus level: %d\n", level);

        ast_log_exception(__FILE__, __LINE__, __FILE__, __LINE__, level);
        return;
    }

    term_strip(buf, buf, max_buf_len);
    syslog(syslog_level, "%s", buf);
}
#else
static void ast_log_vsyslog(struct logmsg *msg)
{
	char buf[BUFSIZ];
	int syslog_level = ast_syslog_priority_from_loglevel(msg->level);
	char call_identifier_str[13];

	if (msg->callid) {
		snprintf(call_identifier_str, sizeof(call_identifier_str), "[C-%08x]", (unsigned)msg->callid->call_identifier);
	} else {
		call_identifier_str[0] = '\0';
	}

	if (syslog_level < 0) {
		/* we are locked here, so cannot ast_log() */
		fprintf(stderr, "ast_log_vsyslog called with bogus level: %d\n", msg->level);
		return;
	}

	snprintf(buf, sizeof(buf), "%s[%d]%s: %s:%d in %s: %s",
		 levels[msg->level], msg->lwp, call_identifier_str, msg->file, msg->line, msg->function, msg->message);

	term_strip(buf, buf, strlen(buf) + 1);
	syslog(syslog_level, "%s", buf);
}

#endif

#ifndef GRANDSTREAM_NETWORKS
static char *logger_strip_verbose_magic(const char *message, int level)
{
	const char *begin, *end;
	char *stripped_message, *dst;
	char magic = -(level + 1);

	if (!(stripped_message = ast_malloc(strlen(message) + 1))) {
		return NULL;
	}

	begin = message;
	dst = stripped_message;
	do {
		end = strchr(begin, magic);
		if (end) {
			size_t len = end - begin;
			memcpy(dst, begin, len);
			begin = end + 1;
			dst += len;
		} else {
			strcpy(dst, begin); /* safe */
			break;
		}
	} while (1);

	return stripped_message;
}
#endif

/*! \brief Print a normal log message to the channels */
#ifdef GRANDSTREAM_NETWORKS
/*! \brief Print a normal log message to the channels */
static void logger_print_normal(struct logmsg *logmsg, char isLogThread)
{
    /* struct logchannel *chan = NULL; */
    char buf[BUFSIZ];
    int  buf_len;
    struct verb *v = NULL;
    struct ast_tm tm;
    char color_buf[64];
    char temp_buf[16];

    if (option_verbose != LOGGER_VERBOSE_REGISTER_OFF && logmsg->level == __LOG_VERBOSE) {
        char *tmpmsg = ast_strdupa(logmsg->message + 1);
        /* Iterate through the list of verbosers and pass them the log message string */
        AST_RWLIST_RDLOCK(&verbosers);
        AST_RWLIST_TRAVERSE(&verbosers, v, list)
            v->verboser(logmsg->message);
        AST_RWLIST_UNLOCK(&verbosers);
        ast_string_field_set(logmsg, message, tmpmsg);
    }

    buf[0] = '\n'; /* log output, start with '\n' */
    buf_len = 1;

    /** format log **/
    if (g_log_format & (1 << EN_DATE))
    {
        buf[buf_len++] = '[';
        ast_localtime(&logmsg->datetime, &tm, NULL);
        buf_len += ast_strftime(buf + buf_len, sizeof(buf) - buf_len, dateformat, &tm);
        buf[buf_len++] = ']';
    }

    if (g_log_format & (1 << EN_LEVEL))
    {
        snprintf(temp_buf, sizeof(temp_buf), "%-5s", levels[logmsg->level]);
        temp_buf[sizeof(temp_buf) - 1] = '\0';
        
        buf_len += snprintf(buf + buf_len, sizeof(buf) - buf_len, 
            " %s", term_color(color_buf, temp_buf, colors[logmsg->level], 0, sizeof(color_buf)) );
    }

    if (!isLogThread)
    {
        buf_len += snprintf(buf + buf_len, sizeof(buf) - buf_len, 
            "[*%05d]", logmsg->lwp);
    }
    else if (g_log_format & (1 << EN_PID))
    {
        buf_len += snprintf(buf + buf_len, sizeof(buf) - buf_len, 
            "[%05d]", logmsg->lwp);
    }

    if (g_log_format & (1 << EN_FILENAME))
    {
        buf_len += snprintf(buf + buf_len, sizeof(buf) - buf_len, 
            " %s", term_color(color_buf, logmsg->file, COLOR_BRWHITE, 0, sizeof(color_buf)) );
    }
    
    if (g_log_format & (1 << EN_LINE))
    {
        /* Turn the numerical line number into a string */
        snprintf(temp_buf, sizeof(temp_buf), "%d", logmsg->line);
        temp_buf[sizeof(temp_buf) - 1] = '\0';
        
        term_color(color_buf, temp_buf, COLOR_BRWHITE, 0, sizeof(color_buf) );
        
        buf_len += snprintf(buf + buf_len, sizeof(buf) - buf_len, 
            (g_log_format & (1 << EN_FILENAME) ? ":%s" : " %s"),  color_buf);
    }

    if (g_log_format & (1 << EN_FUNCNAME))
    {
        buf_len += snprintf(buf + buf_len, sizeof(buf) - buf_len, 
            " %s", term_color(color_buf, logmsg->function, COLOR_BRWHITE, 0, sizeof(color_buf)) );
    }

    if (g_log_format & (1 << EN_MAC))
    {
        ast_eid_to_str(color_buf, sizeof(color_buf), &ast_eid_default);

        buf_len += snprintf(buf + buf_len, sizeof(buf) - buf_len, 
            " [%s]", color_buf);
    }

    if (g_log_format & (1 << EN_VERSION))
    {
        buf_len += snprintf(buf + buf_len, sizeof(buf) - buf_len, 
            " [%s]", ast_get_version());
    }

    buf_len += snprintf(buf + buf_len, sizeof(buf) - buf_len, 
        ": %s", logmsg->message);

    if ((unsigned int)buf_len >= BUFSIZ)
    {
        ast_log_exception(__FILE__, __LINE__, logmsg->file, logmsg->line, buf_len);
        printf("\r\n[%s %d] Log buf(%u >= %u) overflow!!!\n", 
            __FILE__, __LINE__, buf_len, BUFSIZ);
    }
    
    buf[BUFSIZ - 1] = '\0';
    
    /** Output console channels **/
    if (option_verbose == LOGGER_VERBOSE_REGISTER_OFF || logmsg->level != __LOG_VERBOSE)
    {
        ast_console_puts_mutable(buf, logmsg->level);
    }
    
    /** Output file channels **/
    ast_log_file(logmsg->level, buf, buf_len);

    /** Output syslog channels **/
    ast_log_vsyslog(logmsg->level, buf, buf_len);

	/* If we need to reload because of the file size, then do so */
	if (filesize_reload_needed) {
		reload_logger(-1, NULL);
		ast_verb(1, "Rotated Logs Per SIGXFSZ (Exceeded file size limit)\n");
	}

	return;
}

#else   /* the original asterisk function, Grandstream networks */
static void logger_print_normal(struct logmsg *logmsg)
{
	struct logchannel *chan = NULL;
	char buf[BUFSIZ];
	struct verb *v = NULL;
	char *tmpmsg;
	int level = 0;

	if (logmsg->level == __LOG_VERBOSE) {

		/* Iterate through the list of verbosers and pass them the log message string */
		AST_RWLIST_RDLOCK(&verbosers);
		AST_RWLIST_TRAVERSE(&verbosers, v, list)
			v->verboser(logmsg->message);
		AST_RWLIST_UNLOCK(&verbosers);

		level = VERBOSE_MAGIC2LEVEL(logmsg->message);

		tmpmsg = logger_strip_verbose_magic(logmsg->message, level);
		if (tmpmsg) {
			ast_string_field_set(logmsg, message, tmpmsg);
			ast_free(tmpmsg);
		}
	}

	AST_RWLIST_RDLOCK(&logchannels);

	if (!AST_RWLIST_EMPTY(&logchannels)) {
		AST_RWLIST_TRAVERSE(&logchannels, chan, list) {
			char call_identifier_str[13];

			if (logmsg->callid) {
				snprintf(call_identifier_str, sizeof(call_identifier_str), "[C-%08x]", (unsigned)logmsg->callid->call_identifier);
			} else {
				call_identifier_str[0] = '\0';
			}


			/* If the channel is disabled, then move on to the next one */
			if (chan->disabled) {
				continue;
			}
			if (logmsg->level == __LOG_VERBOSE
				&& (((chan->verbosity < 0) ? option_verbose : chan->verbosity)) < level) {
				continue;
			}

			/* Check syslog channels */
			if (chan->type == LOGTYPE_SYSLOG && (chan->logmask & (1 << logmsg->level))) {
				ast_log_vsyslog(logmsg);
			/* Console channels */
			} else if (chan->type == LOGTYPE_CONSOLE && (chan->logmask & (1 << logmsg->level))) {
				char linestr[128];

				/* If the level is verbose, then skip it */
				if (logmsg->level == __LOG_VERBOSE)
					continue;

				/* Turn the numerical line number into a string */
				snprintf(linestr, sizeof(linestr), "%d", logmsg->line);
				/* Build string to print out */
				snprintf(buf, sizeof(buf), "[%s] " COLORIZE_FMT "[%d]%s: " COLORIZE_FMT ":" COLORIZE_FMT " " COLORIZE_FMT ": %s",
					 logmsg->date,
					 COLORIZE(colors[logmsg->level], 0, logmsg->level_name),
					 logmsg->lwp,
					 call_identifier_str,
					 COLORIZE(COLOR_BRWHITE, 0, logmsg->file),
					 COLORIZE(COLOR_BRWHITE, 0, linestr),
					 COLORIZE(COLOR_BRWHITE, 0, logmsg->function),
					 logmsg->message);
				/* Print out */
				ast_console_puts_mutable(buf, logmsg->level);
			/* File channels */
			} else if (chan->type == LOGTYPE_FILE && (chan->logmask & (1 << logmsg->level))) {
				int res = 0;

				/* If no file pointer exists, skip it */
				if (!chan->fileptr) {
					continue;
				}

				/* Print out to the file */
				res = fprintf(chan->fileptr, "[%s] %s[%d]%s %s: %s",
					      logmsg->date, logmsg->level_name, logmsg->lwp, call_identifier_str,
					      logmsg->file, term_strip(buf, logmsg->message, BUFSIZ));
				if (res <= 0 && !ast_strlen_zero(logmsg->message)) {
					fprintf(stderr, "**** Asterisk Logging Error: ***********\n");
					if (errno == ENOMEM || errno == ENOSPC)
						fprintf(stderr, "Asterisk logging error: Out of disk space, can't log to log file %s\n", chan->filename);
					else
						fprintf(stderr, "Logger Warning: Unable to write to log file '%s': %s (disabled)\n", chan->filename, strerror(errno));
					/*** DOCUMENTATION
						<managerEventInstance>
							<synopsis>Raised when a logging channel is disabled.</synopsis>
							<syntax>
								<parameter name="Channel">
									<para>The name of the logging channel.</para>
								</parameter>
							</syntax>
						</managerEventInstance>
					***/
					manager_event(EVENT_FLAG_SYSTEM, "LogChannel", "Channel: %s\r\nEnabled: No\r\nReason: %d - %s\r\n", chan->filename, errno, strerror(errno));
					chan->disabled = 1;
				} else if (res > 0) {
					fflush(chan->fileptr);
				}
			}
		}
	} else if (logmsg->level != __LOG_VERBOSE) {
		fputs(logmsg->message, stdout);
	}

	AST_RWLIST_UNLOCK(&logchannels);

	/* If we need to reload because of the file size, then do so */
	if (filesize_reload_needed) {
		reload_logger(-1, NULL);
		ast_verb(1, "Rotated Logs Per SIGXFSZ (Exceeded file size limit)\n");
	}

	return;
}
#endif

/*! \brief Actual logging thread */
static void *logger_thread(void *data)
{
	struct logmsg *next = NULL, *msg = NULL;

	for (;;) {
		/* We lock the message list, and see if any message exists... if not we wait on the condition to be signalled */
		AST_LIST_LOCK(&logmsgs);
		if (AST_LIST_EMPTY(&logmsgs)) {
			if (close_logger_thread) {
				AST_LIST_UNLOCK(&logmsgs);
				break;
			} else {
				ast_cond_wait(&logcond, &logmsgs.lock);
			}
		}
		next = AST_LIST_FIRST(&logmsgs);
		AST_LIST_HEAD_INIT_NOLOCK(&logmsgs);
#ifdef GRANDSTREAM_NETWORKS
        logmsgs_size = 0;   
#endif
		AST_LIST_UNLOCK(&logmsgs);

		/* Otherwise go through and process each message in the order added */
		while ((msg = next)) {
			/* Get the next entry now so that we can free our current structure later */
			next = AST_LIST_NEXT(msg, list);

			/* Depending on the type, send it to the proper function */
#ifdef GRANDSTREAM_NETWORKS
			logger_print_normal(msg, 1); 
#else
			logger_print_normal(msg);
#endif

			/* Free the data since we are done */
			logmsg_free(msg);
		}
	}

	return NULL;
}

/*!
 * \internal
 * \brief Initialize the logger queue.
 *
 * \note Assumes logchannels is write locked on entry.
 *
 * \return Nothing
 */
static void logger_queue_init(void)
{
	ast_unload_realtime("queue_log");
	if (logfiles.queue_log) {
		char qfname[PATH_MAX];

		if (logger_queue_rt_start()) {
			return;
		}

		/* Open the log file. */
		snprintf(qfname, sizeof(qfname), "%s/%s", ast_config_AST_LOG_DIR,
			queue_log_name);
		if (qlog) {
			/* Just in case it was already open. */
			fclose(qlog);
		}
		qlog = fopen(qfname, "a");
		if (!qlog) {
			ast_log(LOG_ERROR, "Unable to create queue log: %s\n", strerror(errno));
		}
	}
}

/*!
 * \brief Start the ast_queue_log() logger.
 *
 * \note Called when the system is fully booted after startup
 * so preloaded realtime modules can get up.
 *
 * \return Nothing
 */
void logger_queue_start(void)
{
	/* Must not be called before the logger is initialized. */
	ast_assert(logger_initialized);

	AST_RWLIST_WRLOCK(&logchannels);
	if (!queuelog_init) {
		logger_queue_init();
		queuelog_init = 1;
		AST_RWLIST_UNLOCK(&logchannels);
		ast_queue_log("NONE", "NONE", "NONE", "QUEUESTART", "%s", "");
	} else {
		AST_RWLIST_UNLOCK(&logchannels);
	}
}

int init_logger(void)
{
	int res;
	/* auto rotate if sig SIGXFSZ comes a-knockin */
	sigaction(SIGXFSZ, &handle_SIGXFSZ, NULL);

	/* Re-initialize the logmsgs mutex.  The recursive mutex can be accessed prior
 	 * to Asterisk being forked into the background, which can cause the thread
 	 * ID tracked by the underlying pthread mutex to be different than the ID of
 	 * the thread that unlocks the mutex.  Since init_logger is called after the
 	 * fork, it is safe to initialize the mutex here for future accesses.
 	 */
	ast_mutex_destroy(&logmsgs.lock);
	ast_mutex_init(&logmsgs.lock);
	ast_cond_init(&logcond, NULL);

	/* start logger thread */
	if (ast_pthread_create(&logthread, NULL, logger_thread, NULL) < 0) {
		ast_cond_destroy(&logcond);
		return -1;
	}

	/* register the logger cli commands */
	ast_cli_register_multiple(cli_logger, ARRAY_LEN(cli_logger));

	ast_mkdir(ast_config_AST_LOG_DIR, 0777);

	/* create log channels */
	res = init_logger_chain(0 /* locked */, NULL);
#ifdef GRANDSTREAM_NETWORKS
    logger_init_file_output();
#else
	ast_verb_update();
#endif
	logger_initialized = 1;
	if (res) {
		ast_log(LOG_ERROR, "Errors detected in logger.conf.  Default console logging is being used.\n");
	}

	return 0;
}

void close_logger(void)
{
	struct logchannel *f = NULL;
	struct verb *cur = NULL;

	ast_cli_unregister_multiple(cli_logger, ARRAY_LEN(cli_logger));

	logger_initialized = 0;

	/* Stop logger thread */
	AST_LIST_LOCK(&logmsgs);
	close_logger_thread = 1;
	ast_cond_signal(&logcond);
	AST_LIST_UNLOCK(&logmsgs);

	if (logthread != AST_PTHREADT_NULL)
		pthread_join(logthread, NULL);

	AST_RWLIST_WRLOCK(&verbosers);
	while ((cur = AST_LIST_REMOVE_HEAD(&verbosers, list))) {
		ast_free(cur);
	}
	AST_RWLIST_UNLOCK(&verbosers);

	AST_RWLIST_WRLOCK(&logchannels);

	if (qlog) {
		fclose(qlog);
		qlog = NULL;
	}

	while ((f = AST_LIST_REMOVE_HEAD(&logchannels, list))) {
		if (f->fileptr && (f->fileptr != stdout) && (f->fileptr != stderr)) {
			fclose(f->fileptr);
			f->fileptr = NULL;
		}
		ast_free(f);
	}

	closelog(); /* syslog */

	AST_RWLIST_UNLOCK(&logchannels);
#ifdef GRANDSTREAM_NETWORKS
    logger_close_file_output();   
#endif
}

void ast_callid_strnprint(char *buffer, size_t buffer_size, struct ast_callid *callid)
{
	snprintf(buffer, buffer_size, "[C-%08x]", (unsigned)callid->call_identifier);
}

struct ast_callid *ast_create_callid(void)
{
	struct ast_callid *call;

	call = ao2_alloc_options(sizeof(*call), NULL, AO2_ALLOC_OPT_LOCK_NOLOCK);
	if (!call) {
		ast_log(LOG_ERROR, "Could not allocate callid struct.\n");
		return NULL;
	}

	call->call_identifier = ast_atomic_fetchadd_int(&next_unique_callid, +1);
#ifdef TEST_FRAMEWORK
	ast_debug(3, "CALL_ID [C-%08x] created by thread.\n", (unsigned)call->call_identifier);
#endif
	return call;
}

struct ast_callid *ast_read_threadstorage_callid(void)
{
	struct ast_callid **callid;

	callid = ast_threadstorage_get(&unique_callid, sizeof(*callid));
	if (callid && *callid) {
		ast_callid_ref(*callid);
		return *callid;
	}

	return NULL;

}

int ast_callid_threadassoc_change(struct ast_callid *callid)
{
	struct ast_callid **id = ast_threadstorage_get(&unique_callid, sizeof(*id));

	if (!id) {
		ast_log(LOG_ERROR, "Failed to allocate thread storage.\n");
		return -1;
	}

	if (*id && (*id != callid)) {
#ifdef TEST_FRAMEWORK
		ast_debug(3, "CALL_ID [C-%08x] being removed from thread.\n", (unsigned)(*id)->call_identifier);
#endif
		*id = ast_callid_unref(*id);
		*id = NULL;
	}

	if (!(*id) && callid) {
		/* callid will be unreffed at thread destruction */
		ast_callid_ref(callid);
		*id = callid;
#ifdef TEST_FRAMEWORK
		ast_debug(3, "CALL_ID [C-%08x] bound to thread.\n", (unsigned)callid->call_identifier);
#endif
	}

	return 0;
}

int ast_callid_threadassoc_add(struct ast_callid *callid)
{
	struct ast_callid **pointing;

	pointing = ast_threadstorage_get(&unique_callid, sizeof(*pointing));
	if (!(pointing)) {
		ast_log(LOG_ERROR, "Failed to allocate thread storage.\n");
		return -1;
	}

	if (!(*pointing)) {
		/* callid will be unreffed at thread destruction */
		ast_callid_ref(callid);
		*pointing = callid;
#ifdef TEST_FRAMEWORK
		ast_debug(3, "CALL_ID [C-%08x] bound to thread.\n", (unsigned)callid->call_identifier);
#endif
	} else {
		ast_log(LOG_WARNING, "Attempted to ast_callid_threadassoc_add on thread already associated with a callid.\n");
		return 1;
	}

	return 0;
}

int ast_callid_threadassoc_remove(void)
{
	struct ast_callid **pointing;

	pointing = ast_threadstorage_get(&unique_callid, sizeof(*pointing));
	if (!(pointing)) {
		ast_log(LOG_ERROR, "Failed to allocate thread storage.\n");
		return -1;
	}

	if (!(*pointing)) {
		ast_log(LOG_ERROR, "Tried to clean callid thread storage with no callid in thread storage.\n");
		return -1;
	} else {
#ifdef TEST_FRAMEWORK
		ast_debug(3, "CALL_ID [C-%08x] being removed from thread.\n", (unsigned)(*pointing)->call_identifier);
#endif
		*pointing = ast_callid_unref(*pointing);
		return 0;
	}
}

int ast_callid_threadstorage_auto(struct ast_callid **callid)
{
	struct ast_callid *tmp;

	/* Start by trying to see if a callid is available from thread storage */
	tmp = ast_read_threadstorage_callid();
	if (tmp) {
		*callid = tmp;
		return 0;
	}

	/* If that failed, try to create a new one and bind it. */
	tmp = ast_create_callid();
	if (tmp) {
		ast_callid_threadassoc_add(tmp);
		*callid = tmp;
		return 1;
	}

	/* If neither worked, then something must have gone wrong. */
	return -1;
}

void ast_callid_threadstorage_auto_clean(struct ast_callid *callid, int callid_created)
{
	if (callid) {
		/* If the callid was created rather than simply grabbed from the thread storage, we need to unbind here. */
		if (callid_created == 1) {
			ast_callid_threadassoc_remove();
		}
		callid = ast_callid_unref(callid);
	}
}

/*!
 * \internal
 * \brief thread storage cleanup function for unique_callid
 */
static void unique_callid_cleanup(void *data)
{
	struct ast_callid **callid = data;

	if (*callid) {
		ast_callid_unref(*callid);
	}

	ast_free(data);
}

/*!
 * \brief send log messages to syslog and/or the console
 */
#ifndef GRANDSTREAM_NETWORKS
static void __attribute__((format(printf, 6, 0))) ast_log_full(int level, const char *file, int line, const char *function, struct ast_callid *callid, const char *fmt, va_list ap)
{
	struct logmsg *logmsg = NULL;
	struct ast_str *buf = NULL;
	struct ast_tm tm;
	struct timeval now = ast_tvnow();
	int res = 0;
	char datestring[256];

	if (!(buf = ast_str_thread_get(&log_buf, LOG_BUF_INIT_SIZE)))
		return;

	if (level != __LOG_VERBOSE && AST_RWLIST_EMPTY(&logchannels)) {
		/*
		 * we don't have the logger chain configured yet,
		 * so just log to stdout
		 */
		int result;
		result = ast_str_set_va(&buf, BUFSIZ, fmt, ap); /* XXX BUFSIZ ? */
		if (result != AST_DYNSTR_BUILD_FAILED) {
			term_filter_escapes(ast_str_buffer(buf));
			fputs(ast_str_buffer(buf), stdout);
		}
		return;
	}

	/* Ignore anything that never gets logged anywhere */
	if (level != __LOG_VERBOSE && !(global_logmask & (1 << level)))
		return;

	/* Build string */
	res = ast_str_set_va(&buf, BUFSIZ, fmt, ap);

	/* If the build failed, then abort and free this structure */
	if (res == AST_DYNSTR_BUILD_FAILED)
		return;

	/* Create a new logging message */
	if (!(logmsg = ast_calloc_with_stringfields(1, struct logmsg, res + 128)))
		return;

	/* Copy string over */
	ast_string_field_set(logmsg, message, ast_str_buffer(buf));

	/* Set type */
	if (level == __LOG_VERBOSE) {
		logmsg->type = LOGMSG_VERBOSE;
	} else {
		logmsg->type = LOGMSG_NORMAL;
	}

	if (display_callids && callid) {
		logmsg->callid = ast_callid_ref(callid);
		/* callid will be unreffed at logmsg destruction */
	}

	/* Create our date/time */
	ast_localtime(&now, &tm, NULL);
	ast_strftime(datestring, sizeof(datestring), dateformat, &tm);
	ast_string_field_set(logmsg, date, datestring);

	/* Copy over data */
	logmsg->level = level;
	logmsg->line = line;
	ast_string_field_set(logmsg, level_name, levels[level]);
	ast_string_field_set(logmsg, file, file);
	ast_string_field_set(logmsg, function, function);
	logmsg->lwp = ast_get_tid();

	/* If the logger thread is active, append it to the tail end of the list - otherwise skip that step */
	if (logthread != AST_PTHREADT_NULL) {
		AST_LIST_LOCK(&logmsgs);
		if (close_logger_thread) {
			/* Logger is either closing or closed.  We cannot log this message. */
			logmsg_free(logmsg);
		} else {
			AST_LIST_INSERT_TAIL(&logmsgs, logmsg, list);
			ast_cond_signal(&logcond);
		}
		AST_LIST_UNLOCK(&logmsgs);
	} else {
		logger_print_normal(logmsg);
		logmsg_free(logmsg);
	}
}
#endif

#ifdef GRANDSTREAM_NETWORKS
void __ast_log_send(int level, const char *file, int line, const char *function, const char *fmt, ...)
{
	struct logmsg *logmsg = NULL;
	struct ast_str *buf = NULL;
	int res = 0;
	va_list ap;

    /* Drop log */
    if (logmsgs_size > drop_limit[level])
    {
        logmsgs_drop_level[level]++;
        logmsgs_drop_total++;

        if (!(logmsgs_drop_total & 0x1FF)) /* logmsgs_drop_size % 512 */
        {
            printf("\r\n\r\n\r\n(%lu > %lu) Drop '%s' log, total drop %lu already!!\r\n\r\n\r\n ",
                logmsgs_size, drop_limit[level], levels[level], logmsgs_drop_total);
        }

        return;
    }

	if (!(buf = ast_str_thread_get(&log_buf, LOG_BUF_INIT_SIZE)))
	{
        ast_log_exception(__FILE__, __LINE__, file, line, (int)(ptrdiff_t)buf);
        printf("\r\n[%s %d] Get log thread memory(%d) failed !!\n", __FILE__, __LINE__, LOG_BUF_INIT_SIZE);
        return;
    }
    

	/* Build string */
	va_start(ap, fmt);
	res = ast_str_set_va(&buf, BUFSIZ, fmt, ap);
	va_end(ap);

	/* If the build failed, then abort and free this structure */
	if (res == AST_DYNSTR_BUILD_FAILED)
	{
        ast_log_exception(__FILE__, __LINE__, file, line, res);
        printf("\r\n[%s %d] Logger va_start(%d, %s) failed !!\n", __FILE__, __LINE__, BUFSIZ, fmt);
        return;
    }
    
	/* Create a new logging message */
	if (!(logmsg = ast_calloc_with_stringfields(1, struct logmsg, res + 128)))
    {
        ast_log_exception(__FILE__, __LINE__, file, line, (int)(ptrdiff_t)logmsg);
        printf("\r\n[%s %d] Create a new logging message(%zu, %d) failed !!\n", 
            __FILE__, __LINE__, 1 * sizeof(struct logmsg), res + 128);
        return;
    }

	/* Copy string over */
	ast_string_field_set(logmsg, message, ast_str_buffer(buf));

	/* Create our date/time */
    logmsg->datetime = ast_tvnow();

	/* Copy over data */
	logmsg->level = level;
	logmsg->line = line;
	ast_string_field_set(logmsg, file, file);
	ast_string_field_set(logmsg, function, function);
	logmsg->lwp = ast_get_tid();

    /* If the logger thread is active, append it to the tail end of the list - otherwise skip that step */
    if (logthread != AST_PTHREADT_NULL) {
        AST_LIST_LOCK(&logmsgs);
        AST_LIST_INSERT_TAIL(&logmsgs, logmsg, list);
        logmsgs_size++;
        ast_cond_signal(&logcond);
        AST_LIST_UNLOCK(&logmsgs);
    } else {
        logger_print_normal(logmsg, 0);
        ast_free(logmsg);
    }

	return;
}
#else    /* original asterisk ast_log function */
void ast_log(int level, const char *file, int line, const char *function, const char *fmt, ...)
{
	struct ast_callid *callid;
	va_list ap;

	callid = ast_read_threadstorage_callid();

	va_start(ap, fmt);
	if (level == __LOG_VERBOSE) {
		__ast_verbose_ap(file, line, function, 0, callid, fmt, ap);
	} else {
		ast_log_full(level, file, line, function, callid, fmt, ap);
	}
	va_end(ap);

	if (callid) {
		ast_callid_unref(callid);
	}
}
#endif

#ifndef GRANDSTREAM_NETWORKS
void ast_log_safe(int level, const char *file, int line, const char *function, const char *fmt, ...)
{
	va_list ap;
	void *recursed = ast_threadstorage_get_ptr(&in_safe_log);
	struct ast_callid *callid;

	if (recursed) {
		return;
	}

	if (ast_threadstorage_set_ptr(&in_safe_log, (void*)1)) {
		/* We've failed to set the flag that protects against
		 * recursion, so bail. */
		return;
	}

	callid = ast_read_threadstorage_callid();

	va_start(ap, fmt);
	ast_log_full(level, file, line, function, callid, fmt, ap);
	va_end(ap);

	if (callid) {
		ast_callid_unref(callid);
	}

	/* Clear flag so the next allocation failure can be logged. */
	ast_threadstorage_set_ptr(&in_safe_log, NULL);
}

void ast_log_callid(int level, const char *file, int line, const char *function, struct ast_callid *callid, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	ast_log_full(level, file, line, function, callid, fmt, ap);
	va_end(ap);
}
#endif

void ast_log_backtrace(void)
{
#ifdef HAVE_BKTR
	struct ast_bt *bt;
	int i = 0;
	char **strings;

	if (!(bt = ast_bt_create())) {
		ast_log(LOG_WARNING, "Unable to allocate space for backtrace structure\n");
		return;
	}

	if ((strings = ast_bt_get_symbols(bt->addresses, bt->num_frames))) {
		ast_verbose("Got %d backtrace record%c\n", bt->num_frames, bt->num_frames != 1 ? 's' : ' ');
		for (i = 3; i < bt->num_frames - 2; i++) {
			ast_verbose("#%d: [%p] %s\n", i - 3, bt->addresses[i], strings[i]);
		}

		ast_std_free(strings);
	} else {
		ast_verbose("Could not allocate memory for backtrace\n");
	}
	ast_bt_destroy(bt);
#else
	ast_log(LOG_WARNING, "Must run configure with '--with-execinfo' for stack backtraces.\n");
#endif /* defined(HAVE_BKTR) */
}

#ifndef GRANDSTREAM_NETWORKS
void __ast_verbose_ap(const char *file, int line, const char *func, int level, struct ast_callid *callid, const char *fmt, va_list ap)
{
	const char *p;
	struct ast_str *prefixed, *buf;
	int res = 0;
	signed char magic = level > 9 ? -10 : -level - 1; /* 0 => -1, 1 => -2, etc.  Can't pass NUL, as it is EOS-delimiter */

	/* For compatibility with modules still calling ast_verbose() directly instead of using ast_verb() */
	if (level < 0) {
		if (!strncmp(fmt, VERBOSE_PREFIX_4, strlen(VERBOSE_PREFIX_4))) {
			magic = -5;
		} else if (!strncmp(fmt, VERBOSE_PREFIX_3, strlen(VERBOSE_PREFIX_3))) {
			magic = -4;
		} else if (!strncmp(fmt, VERBOSE_PREFIX_2, strlen(VERBOSE_PREFIX_2))) {
			magic = -3;
		} else if (!strncmp(fmt, VERBOSE_PREFIX_1, strlen(VERBOSE_PREFIX_1))) {
			magic = -2;
		} else {
			magic = -1;
		}
	}

	if (!(prefixed = ast_str_thread_get(&verbose_buf, VERBOSE_BUF_INIT_SIZE)) ||
	    !(buf = ast_str_thread_get(&verbose_build_buf, VERBOSE_BUF_INIT_SIZE))) {
		return;
	}

	res = ast_str_set_va(&buf, 0, fmt, ap);
	/* If the build failed then we can drop this allocated message */
	if (res == AST_DYNSTR_BUILD_FAILED) {
		return;
	}

	ast_str_reset(prefixed);

	/* for every newline found in the buffer add verbose prefix data */
	fmt = ast_str_buffer(buf);
	do {
		if (!(p = strchr(fmt, '\n'))) {
			p = strchr(fmt, '\0') - 1;
		}
		++p;

		ast_str_append(&prefixed, 0, "%c", (char)magic);
		ast_str_append_substr(&prefixed, 0, fmt, p - fmt);
		fmt = p;
	} while (p && *p);

	ast_log_callid(__LOG_VERBOSE, file, line, func, callid, "%s", ast_str_buffer(prefixed));
}
#endif

#ifdef GRANDSTREAM_NETWORKS
void __ast_verbose(int module_index, const char *file, int line, const char *func, const char *fmt, ...)
{
	struct ast_str *buf = NULL;
	int res = 0;
	va_list ap;

	if (!(buf = ast_str_thread_get(&verbose_buf, VERBOSE_BUF_INIT_SIZE)))
		return;

	if (ast_opt_timestamp) {
		struct timeval now;
		struct ast_tm tm;
		char date[40];
		char *datefmt;

		now = ast_tvnow();
		ast_localtime(&now, &tm, NULL);
		ast_strftime(date, sizeof(date), dateformat, &tm);
		datefmt = alloca(strlen(date) + 3 + strlen(fmt) + 1);
		sprintf(datefmt, "%c[%s] %s", 127, date, fmt);
		fmt = datefmt;
	} 

	/* Build string */
    va_start(ap, fmt);
    res = ast_str_set_va(&buf, 0, fmt, ap);
    va_end(ap);
    
	/* If the build failed then we can drop this allocated message */
	if (res == AST_DYNSTR_BUILD_FAILED)
		return;

	__ast_log_send(__LOG_VERBOSE, file, line, func, "%s", ast_str_buffer(buf));
}
#else  /* original asterisk __ast_verbose */
void __ast_verbose(const char *file, int line, const char *func, int level, const char *fmt, ...)
{
	struct ast_callid *callid;
	va_list ap;

	callid = ast_read_threadstorage_callid();

	va_start(ap, fmt);
	__ast_verbose_ap(file, line, func, level, callid, fmt, ap);
	va_end(ap);

	if (callid) {
		ast_callid_unref(callid);
	}
}
#endif 

#ifndef GRANDSTREAM_NETWORKS
void __ast_verbose_callid(const char *file, int line, const char *func, int level, struct ast_callid *callid, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	__ast_verbose_ap(file, line, func, level, callid, fmt, ap);
	va_end(ap);
}

/* No new code should use this directly, but we have the ABI for backwards compat */
#undef ast_verbose
void __attribute__((format(printf, 1,2))) ast_verbose(const char *fmt, ...);
void ast_verbose(const char *fmt, ...)
{
	struct ast_callid *callid;
	va_list ap;

	callid = ast_read_threadstorage_callid();

	va_start(ap, fmt);
	__ast_verbose_ap("", 0, "", 0, callid, fmt, ap);
	va_end(ap);

	if (callid) {
		ast_callid_unref(callid);
	}
}

/*! Console verbosity level node. */
struct verb_console {
	/*! List node link */
	AST_LIST_ENTRY(verb_console) node;
	/*! Console verbosity level. */
	int *level;
};

/*! Registered console verbosity levels */
static AST_RWLIST_HEAD_STATIC(verb_consoles, verb_console);

/*! ast_verb_update() reentrancy protection lock. */
AST_MUTEX_DEFINE_STATIC(verb_update_lock);

void ast_verb_update(void)
{
	struct logchannel *log;
	struct verb_console *console;
	int verb_level;

	ast_mutex_lock(&verb_update_lock);

	AST_RWLIST_RDLOCK(&verb_consoles);

	/* Default to the root console verbosity. */
	verb_level = option_verbose;

	/* Determine max remote console level. */
	AST_LIST_TRAVERSE(&verb_consoles, console, node) {
		if (verb_level < *console->level) {
			verb_level = *console->level;
		}
	}
	AST_RWLIST_UNLOCK(&verb_consoles);

	/* Determine max logger channel level. */
	AST_RWLIST_RDLOCK(&logchannels);
	AST_RWLIST_TRAVERSE(&logchannels, log, list) {
		if (verb_level < log->verbosity) {
			verb_level = log->verbosity;
		}
	}
	AST_RWLIST_UNLOCK(&logchannels);

	ast_verb_sys_level = verb_level;

	ast_mutex_unlock(&verb_update_lock);
}

/*!
 * \internal
 * \brief Unregister a console verbose level.
 *
 * \param console Which console to unregister.
 *
 * \return Nothing
 */
static void verb_console_unregister(struct verb_console *console)
{
	AST_RWLIST_WRLOCK(&verb_consoles);
	console = AST_RWLIST_REMOVE(&verb_consoles, console, node);
	AST_RWLIST_UNLOCK(&verb_consoles);
	if (console) {
		ast_verb_update();
	}
}

static void verb_console_free(void *v_console)
{
	struct verb_console *console = v_console;

	verb_console_unregister(console);
	ast_free(console);
}

/*! Thread specific console verbosity level node. */
AST_THREADSTORAGE_CUSTOM(my_verb_console, NULL, verb_console_free);

void ast_verb_console_register(int *level)
{
	struct verb_console *console;

	console = ast_threadstorage_get(&my_verb_console, sizeof(*console));
	if (!console || !level) {
		return;
	}
	console->level = level;

	AST_RWLIST_WRLOCK(&verb_consoles);
	AST_RWLIST_INSERT_HEAD(&verb_consoles, console, node);
	AST_RWLIST_UNLOCK(&verb_consoles);
	ast_verb_update();
}

void ast_verb_console_unregister(void)
{
	struct verb_console *console;

	console = ast_threadstorage_get(&my_verb_console, sizeof(*console));
	if (!console) {
		return;
	}
	verb_console_unregister(console);
}

int ast_verb_console_get(void)
{
	struct verb_console *console;
	int verb_level;

	console = ast_threadstorage_get(&my_verb_console, sizeof(*console));
	AST_RWLIST_RDLOCK(&verb_consoles);
	if (!console) {
		verb_level = 0;
	} else if (console->level) {
		verb_level = *console->level;
	} else {
		verb_level = option_verbose;
	}
	AST_RWLIST_UNLOCK(&verb_consoles);
	return verb_level;
}

void ast_verb_console_set(int verb_level)
{
	struct verb_console *console;

	console = ast_threadstorage_get(&my_verb_console, sizeof(*console));
	if (!console) {
		return;
	}

	AST_RWLIST_WRLOCK(&verb_consoles);
	if (console->level) {
		*console->level = verb_level;
	} else {
		option_verbose = verb_level;
	}
	AST_RWLIST_UNLOCK(&verb_consoles);
	ast_verb_update();
}

#endif

int ast_register_verbose(void (*v)(const char *string))
{
	struct verb *verb;

	if (!(verb = ast_malloc(sizeof(*verb))))
		return -1;

	verb->verboser = v;

	AST_RWLIST_WRLOCK(&verbosers);
	AST_RWLIST_INSERT_HEAD(&verbosers, verb, list);
	AST_RWLIST_UNLOCK(&verbosers);

	return 0;
}

int ast_unregister_verbose(void (*v)(const char *string))
{
	struct verb *cur;

	AST_RWLIST_WRLOCK(&verbosers);
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&verbosers, cur, list) {
		if (cur->verboser == v) {
			AST_RWLIST_REMOVE_CURRENT(list);
			ast_free(cur);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
	AST_RWLIST_UNLOCK(&verbosers);

	return cur ? 0 : -1;
}

static void update_logchannels(void)
{
	struct logchannel *cur;

	AST_RWLIST_WRLOCK(&logchannels);

	global_logmask = 0;

	AST_RWLIST_TRAVERSE(&logchannels, cur, list) {
		make_components(cur);
		global_logmask |= cur->logmask;
	}

	AST_RWLIST_UNLOCK(&logchannels);
}

int ast_logger_register_level(const char *name)
{
	unsigned int level;
	unsigned int available = 0;

	AST_RWLIST_WRLOCK(&logchannels);

	for (level = 0; level < ARRAY_LEN(levels); level++) {
#ifdef GRANDSTREAM_NETWORKS		
		if ((level >= DYNAMIC_LEVEL_START) && !available && !levels[level]) {
#else
		if ((level >= 16) && !available && !levels[level]) {
#endif
			available = level;
			continue;
		}

		if (levels[level] && !strcasecmp(levels[level], name)) {
			ast_log(LOG_WARNING,
				"Unable to register dynamic logger level '%s': a standard logger level uses that name.\n",
				name);
			AST_RWLIST_UNLOCK(&logchannels);

			return -1;
		}
	}

	if (!available) {
		ast_log(LOG_WARNING,
			"Unable to register dynamic logger level '%s'; maximum number of levels registered.\n",
			name);
		AST_RWLIST_UNLOCK(&logchannels);

		return -1;
	}

	levels[available] = ast_strdup(name);
#ifdef GRANDSTREAM_NETWORKS
    g_log_dynamic_register_count++;   
#endif
	AST_RWLIST_UNLOCK(&logchannels);

	ast_debug(1, "Registered dynamic logger level '%s' with index %u.\n", name, available);

	update_logchannels();

	return available;
}

void ast_logger_unregister_level(const char *name)
{
	unsigned int found = 0;
	unsigned int x;

	AST_RWLIST_WRLOCK(&logchannels);
#ifdef GRANDSTREAM_NETWORKS
	for (x = DYNAMIC_LEVEL_START; x < ARRAY_LEN(levels); x++) {
#else
	for (x = 16; x < ARRAY_LEN(levels); x++) {
#endif
		if (!levels[x]) {
			continue;
		}

		if (strcasecmp(levels[x], name)) {
			continue;
		}

		found = 1;
		break;
	}

	if (found) {
		/* take this level out of the global_logmask, to ensure that no new log messages
		 * will be queued for it
		 */

		global_logmask &= ~(1 << x);

		ast_free(levels[x]);
		levels[x] = NULL;
#ifdef GRANDSTREAM_NETWORKS
        g_log_dynamic_unregister_count++;  
#endif
		AST_RWLIST_UNLOCK(&logchannels);

		ast_debug(1, "Unregistered dynamic logger level '%s' with index %u.\n", name, x);

		update_logchannels();
	} else {
		AST_RWLIST_UNLOCK(&logchannels);
	}
}

const char *ast_logger_get_dateformat(void)
{
	return dateformat;
}


#ifdef GRANDSTREAM_NETWORKS
static char *handle_logger_save_format(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int   result;
    
    switch (cmd) {
    case CLI_INIT:
        e->command = "logger save format";
        e->usage =
        "Usage: logger save format \n"
        "       Save logger format, see 'logger show format' .\n";
        return NULL;
        
    case CLI_GENERATE:

        return NULL;
    }

    if (a->argc != 3)
    {
        return CLI_SHOWUSAGE;
    }

    result = logger_save_format_sqlite3();

    if (result)
    {
        ast_cli(a->fd, "\nLogger format save failed!\n");
        return CLI_FAILURE;
    }

    ast_cli(a->fd, "\nLogger format save success.\n");
    
    return CLI_SUCCESS;
}

static char *handle_logger_save_switch(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int   result;
    
    switch (cmd) {
    case CLI_INIT:
        e->command = "logger save switch";
        e->usage =
        "Usage: logger save switch \n"
        "       Save logger switch, contain dynamic level and static level.\n";
        return NULL;
        
    case CLI_GENERATE:

        return NULL;
    }

    if (a->argc != 3)
    {
        return CLI_SHOWUSAGE;
    }

    result = logger_save_switch_sqlite3();

    if (result)
    {
        ast_cli(a->fd, "\nLogger switch save failed!\n");
        return CLI_FAILURE;
    }

    ast_cli(a->fd, "\nLogger switch save success.\n");
    
    return CLI_SUCCESS;
}

static char *handle_logger_set_module_level(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int flag;
    int level        = -1;
    int module_index = -1;

    int u;

    switch (cmd) {
    case CLI_INIT:
        e->command = "logger set module";
        e->usage = 
        "Usage: logger set module {module} level {ALL|CODEC_NEG|DEBUG|VERB|NOTIC|WARN|ERROR} {on|off}\n"
        "       Set a specific log level with module to enabled/disabled for log output.\n";
        return NULL;
        
    case CLI_GENERATE:

        if (3 == a->pos)
        {
            return logger_search_modulename(S_OR(a->argv[3], ""), a->n);
        }
        else if (4 == a->pos && 0 == a->n)
        {
            return ast_strdup("level");
        }
        else if (5 == a->pos)
        {
            return logger_search_levelname(S_OR(a->argv[5], ""), a->n);
        }
        else if (6 == a->pos)
        {
            return logger_search_flagname(S_OR(a->argv[6], ""), a->n);
        }

        return NULL;
    }

    if (a->argc != 7)
    {
        return CLI_SHOWUSAGE;
    }

    if (strcasecmp(a->argv[4], "level"))
    {
        return CLI_SHOWUSAGE;
    }

    if (!strcasecmp(a->argv[3], "ALL"))
    {
        module_index = -10;
    }
    else
    {
        module_index = logger_get_moduleindex_by_name(a->argv[3]);

        if (module_index < 0)
        {
            ast_cli(a->fd, " module name [%s] not exists!\n", a->argv[3]);

            return CLI_SHOWUSAGE;
        }
    }

    if (!strcasecmp(a->argv[5], "ALL"))
    {
        level = -10;
    }
    else if (!strcasecmp(a->argv[5], levels[__LOG_DEBUG]))
    {
        level = __LOG_DEBUG;
    }
    else if (!strcasecmp(a->argv[5], levels[__LOG_VERBOSE]))
    {
        level = __LOG_VERBOSE;
    }
    else if (!strcasecmp(a->argv[5], levels[__LOG_NOTICE]))
    {
        level = __LOG_NOTICE;
    }
    else if (!strcasecmp(a->argv[5], levels[__LOG_WARNING]))
    {
        level = __LOG_WARNING;
    }
    else if (!strcasecmp(a->argv[5], levels[__LOG_ERROR]))
    {
        level = __LOG_ERROR;
    }	
#ifdef GRANDSTREAM_NETWORKS
	else if (!strcasecmp(a->argv[5], levels[__LOG_CODEC_NEG]))
	{
		level = __LOG_CODEC_NEG;
	}
#endif
    else
    {
        ast_cli(a->fd, " Unknow level[%s] !\n", a->argv[5]);

        return CLI_SHOWUSAGE;
    }

    if (!strcasecmp(a->argv[6], "ON"))
    {
        flag = 1;
    }
    else if (!strcasecmp(a->argv[6], "OFF"))
    {
        flag = 0;
    }
    else
    {
        ast_cli(a->fd, " Unknow switch [%s] !\n", a->argv[6]);

        return CLI_SHOWUSAGE;
    }

    AST_RWLIST_RDLOCK(&g_log_module_file_list);

    if (module_index == -10) /* -10 is 'ALL', turn on/off all module  */
    {

        for (u = 0; u < ARRAY_LEN(ast_log_modules_level); u++)
        {
            if (level == -10)
            {    
                if (flag)
                {
                    
                    ast_log_modules_level[u] = 
                        ((1 << __LOG_ERROR) | (1 << __LOG_WARNING) | (1 << __LOG_NOTICE) | (1 << __LOG_VERBOSE) | (1 << __LOG_DEBUG));
                }
                else
                {
                    ast_log_modules_level[u] = 0;
                }

            }
            else
            {
                if (flag)
                {
                    ast_log_modules_level[u] |= (1 << level);
                }
                else
                {
                    ast_log_modules_level[u] &= ~(1 << level);
                }
            }

        }
    }
    else
    {
        if (level == -10)
        {
            if (flag)
            {
                ast_log_modules_level[module_index] = 
                    ((1 << __LOG_ERROR) | (1 << __LOG_WARNING) | (1 << __LOG_NOTICE) | (1 << __LOG_VERBOSE) | (1 << __LOG_DEBUG));
            }
            else
            {
                ast_log_modules_level[module_index] = 0;
            }
        }
        else
        {
            if (flag)
            {
                ast_log_modules_level[module_index] |= (1 << level);
            }
            else
            {
                ast_log_modules_level[module_index] &= ~(1 << level);
            }
        }
    }

    AST_RWLIST_UNLOCK(&g_log_module_file_list);


    /* BEGIN: add by wangjiafeng 20130108 [optimization] */
    if (__LOG_DEBUG == level || -10 == level)
    {
        char *exe_cmd;
        int   ret;
        int   out_fd = -64;

        if (!strcasecmp("chan_iax2", a->argv[3]) || !strcasecmp("ALL", a->argv[3]))
        {
            if (flag)
            {
                exe_cmd = "iax2 set debug on";
            }
            else
            {
                exe_cmd = "iax2 set debug off";
            }
            
            ret = ast_cli_command_full(CLI_NO_PERMS, CLI_NO_PERMS, out_fd, exe_cmd);
            if (ret)
            {
                ast_log(LOG_ERROR, "Command '%s' execute failed[%d] !!", exe_cmd, ret);
            }

        }

        if (!strcasecmp("chan_sip", a->argv[3]) || !strcasecmp("ALL", a->argv[3]))
        {
            if (flag)
            {
                exe_cmd = "sip set debug on";
            }
            else
            {
                exe_cmd = "sip set debug off";
            }
            
            ret = ast_cli_command_full(CLI_NO_PERMS, CLI_NO_PERMS, out_fd, exe_cmd);
            if (ret)
            {
                ast_log(LOG_ERROR, "Command '%s' execute failed[%d] !!", exe_cmd, ret);
            }

        }

        if (!strcasecmp("res_fax", a->argv[3]) || !strcasecmp("ALL", a->argv[3]))
        {
            if (flag)
            {
                exe_cmd = "fax set debug on";
            }
            else
            {
                exe_cmd = "fax set debug off";
            }
            
            ret = ast_cli_command_full(CLI_NO_PERMS, CLI_NO_PERMS, out_fd, exe_cmd);
            if (ret)
            {
                ast_log(LOG_ERROR, "Command '%s' execute failed[%d] !!", exe_cmd, ret);
            }
        }

        if (!strcasecmp("RTP", a->argv[3]) || !strcasecmp("ALL", a->argv[3]))
        {
            if (flag)
            {
                exe_cmd = "rtp set debug on";
            }
            else
            {
                exe_cmd = "rtp set debug off";
            }
            
            ret = ast_cli_command_full(CLI_NO_PERMS, CLI_NO_PERMS, out_fd, exe_cmd);
            if (ret)
            {
                ast_log(LOG_ERROR, "Command '%s' execute failed[%d] !!", exe_cmd, ret);
            }
        }

        if (!strcasecmp("NAT", a->argv[3]) || !strcasecmp("ALL", a->argv[3]))
        {
            if (flag)
            {
                exe_cmd = "stun set debug on";
            }
            else
            {
                exe_cmd = "stun set debug off";
            }
            
            ret = ast_cli_command_full(CLI_NO_PERMS, CLI_NO_PERMS, out_fd, exe_cmd);
            if (ret)
            {
                ast_log(LOG_ERROR, "Command '%s' execute failed[%d] !!", exe_cmd, ret);
            }
        }

    } /*? if (__LOG_DEBUG == level || -10 == level) ?*/
    /* END: add by wangjiafeng 20130108 [optimization] */

    return CLI_SUCCESS;
}

static char *handle_logger_set_dlevel(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int u;
    int flag;
    int dlevel = 0;

    switch (cmd) 
    {
    case CLI_INIT:
        e->command = "logger set dlevel";
        e->usage = 
            "Usage: logger set dlevel <DTMF|CC|FAX|SECURITY> <on|off> \n"
            "       Set dynamic level log switch.\n";
        return NULL;
        
    case CLI_GENERATE:

        if (3 == a->pos)
        {
            return logger_search_dlevelname(S_OR(a->argv[3], ""), a->n);
        }
        else if (4 == a->pos)
        {
            return logger_search_flagname(S_OR(a->argv[4], ""), a->n);
        }

        return NULL;        
    }

    if (a->argc != 5)
    {
        return CLI_SHOWUSAGE;
    }

    
    flag = ast_true(a->argv[4]) ? 1 : 0;
    
    if (!strcasecmp(S_OR(a->argv[3], ""), levels[__LOG_DTMF]))
    {
        if (flag)
        {
            ast_log_dynamic_levels_switch |= (1 << __LOG_DTMF);
        }
        else
        {
            ast_log_dynamic_levels_switch &= ~(1 << __LOG_DTMF);
        }

        ast_cli(a->fd, "Logger status for '%s' has been set to '%s'.\n", 
            levels[__LOG_DTMF], flag ? "on" : "off");
        return CLI_SUCCESS;
    }

    AST_RWLIST_RDLOCK(&logchannels);

    for (u = DYNAMIC_LEVEL_START; u < ARRAY_LEN(levels); u++) {
    
        if (levels[u] && !strcasecmp(S_OR(a->argv[3], ""), levels[u]))
        {
            dlevel = u;
            break;
        }
        
    }
    
    AST_RWLIST_UNLOCK(&logchannels);


    if (0 == dlevel)
    {
        ast_cli(a->fd, " Unknow dynamic level [%s] !\n", a->argv[3]);
        return CLI_SHOWUSAGE;
    }


    if (flag)
    {
        ast_log_dynamic_levels_switch |= (1 << dlevel);
    }
    else
    {
        ast_log_dynamic_levels_switch &= ~(1 << dlevel);
    }
    
    ast_cli(a->fd, "Logger status for '%s' has been set to '%s'.\n", 
        levels[dlevel], flag ? "on" : "off");
    

    return CLI_SUCCESS;
}

static char *handle_logger_set_format(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int index;

    switch (cmd) 
    {
    case CLI_INIT:
        e->command = "logger set format {date|level|PID|filename|line|funcname|MAC|VERSION} {on|off}";
        e->usage = 
            "Usage: logger set format {date|level|PID|filename|line|funcname|MAC|VERSION} {on|off}\n"
            "       Set logger output format.\n";
        return NULL;
        
    case CLI_GENERATE:
        return NULL;
        
    }

    if (a->argc != 5)
    {
        return CLI_SHOWUSAGE;
    }

    index = logger_get_formatindex_by_name(S_OR(a->argv[3], ""), 0);

    if (index >= 0 && index < ARRAY_LEN(log_formats))
    {
        if (ast_true(a->argv[4]))
        {
            g_log_format |= 1 << index;
        }
        else
        {
            g_log_format &= ~(1 << index);
        }
    }
    else
    {
        ast_cli(a->fd, " Unknow format type [%s] !\n", a->argv[3]);

        return CLI_SHOWUSAGE;
    }

    ast_cli(a->fd, "Logger format for '%s' has been set to '%s'.\n", 
        S_OR(a->argv[3], ""), ast_true(a->argv[4]) ? "on" : "off");
    
    return CLI_SUCCESS;
}


static char *handle_logger_set_default(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{

    switch (cmd) 
    {
    case CLI_INIT:
        e->command = "logger set default ";
        e->usage = 
            "Usage: logger set default \n"
            "       Set default logger switch.\n";
        return NULL;
        
    case CLI_GENERATE:
        return NULL;
        
    }

    if (a->argc != 3)
    {
        return CLI_SHOWUSAGE;
    }

    logger_set_default_level();

    return CLI_SUCCESS;
}

static char *handle_logger_test_module_level(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    LogModuleFile *u;
    int level        = -1;
    int module_index = -1;

    switch (cmd) 
    {
    case CLI_INIT:
        e->command = "logger test module";
        e->usage = 
            "Usage: logger test module {module} level {ALL|DEBUG|NOTIC|WARN|ERROR|VERB|DTMF} \n"
            "       Simulation a specific log to this console.\n";
        return NULL;
        
    case CLI_GENERATE:

        if (3 == a->pos)
        {
            return logger_search_modulename(S_OR(a->argv[3], ""), a->n);
        }
        else if (4 == a->pos && 0 == a->n)
        {
            return ast_strdup("level");
        }
        else if (5 == a->pos)
        {
            return logger_search_levelname(S_OR(a->argv[5], ""), a->n);
        }

        return NULL;
    }

    if (a->argc != 6)
    {
        return CLI_SHOWUSAGE;
    }

    if (strcasecmp(a->argv[4], "level"))
    {
        return CLI_SHOWUSAGE;
    }

    if (!strcasecmp(a->argv[3], "ALL"))
    {
        module_index = -10;
    }
    else
    {
        module_index = logger_get_moduleindex_by_name(a->argv[3]);

        if (module_index < 0)
        {
            ast_cli(a->fd, " module name [%s] not exists!\n", a->argv[3]);

            return CLI_SHOWUSAGE;
        }
    }

    if (!strcasecmp(a->argv[5], "ALL"))
    {
        level = -10;
    }
    else if (!strcasecmp(a->argv[5], levels[__LOG_DEBUG]))
    {
        level = __LOG_DEBUG;
    }
    else if (!strcasecmp(a->argv[5], levels[__LOG_VERBOSE]))
    {
        level = __LOG_VERBOSE;
    }
    else if (!strcasecmp(a->argv[5], levels[__LOG_NOTICE]))
    {
        level = __LOG_NOTICE;
    }
    else if (!strcasecmp(a->argv[5], levels[__LOG_WARNING]))
    {
        level = __LOG_WARNING;
    }
    else if (!strcasecmp(a->argv[5], levels[__LOG_ERROR]))
    {
        level = __LOG_ERROR;
    }
    else
    {
        ast_cli(a->fd, " Unknow level[%s] !\n", a->argv[5]);

        return CLI_SHOWUSAGE;
    }

    AST_RWLIST_RDLOCK(&g_log_module_file_list);
    
    AST_LIST_TRAVERSE(&g_log_module_file_list, u, entry)
    {
        if (module_index != -10 && module_index != u->module_index)
        {
            continue;
        }

        if (level == -10)
        {
            ast_log_module_index(__LOG_ERROR, u->module_index,
                "Simulation a log, module[%s (%d)] level '%s' ^_^", 
                u->module_name, u->module_index, levels[__LOG_ERROR]);
            ast_log_module_index(__LOG_WARNING, u->module_index,
                "Simulation a log, module[%s (%d)] level '%s' ^_^", 
                u->module_name, u->module_index, levels[__LOG_WARNING]);
            ast_log_module_index(__LOG_NOTICE, u->module_index,
                "Simulation a log, module[%s (%d)] level '%s' ^_^",
                u->module_name, u->module_index, levels[__LOG_NOTICE]);
            ast_log_module_index(__LOG_VERBOSE, u->module_index,
                "Simulation a log, module[%s (%d)] level '%s' ^_^",
                u->module_name, u->module_index, levels[__LOG_VERBOSE]);
            ast_log_module_index(__LOG_DEBUG, u->module_index,
                "Simulation a log, module[%s (%d)] level '%s' ^_^",
                u->module_name, u->module_index, levels[__LOG_DEBUG]);
        }
        else
        {
            ast_log_module_index(level, u->module_index,
                "Simulation a log, module[%s (%d)] level '%s' ^_^",
                u->module_name, u->module_index, levels[level]);
        }
    }

    AST_RWLIST_UNLOCK(&g_log_module_file_list);
    

    if (module_index == -10 && level == -10)
    {
    
        ast_verb(0, "Simulation a ast_verb(0) ^_^ ");
        ast_verb(3, "Simulation a ast_verb(3) ^_^ ");
        ast_verb(9, "Simulation a ast_verb(9) ^_^ ");

        ast_verbose("Simulation a ast_verbose ^_^ ");

        ast_debug(0, "Simulation a ast_debug(0) ^_^ ");
        ast_debug(3, "Simulation a ast_debug(3) ^_^ ");
        ast_debug(9, "Simulation a ast_debug(9) ^_^ ");

        ast_queue_log("Simulation_astqueuelog", "CALLID", "AGENT", "EVENT", "FMT");
    }

    return CLI_SUCCESS;
}

static char *handle_logger_test_dlevel(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int u;

    switch (cmd) 
    {
    case CLI_INIT:
        e->command = "logger test dlevel";
        e->usage = 
            "Usage: logger test dlevel <DTMF|CC|FAX|SECURITY> \n"
            "       Simulation a specific log for dynamic level to this console.\n";
        return NULL;
        
    case CLI_GENERATE:
    
        if (3 == a->pos)
        {
            return logger_search_dlevelname(S_OR(a->argv[3], ""), a->n);
        }
 
        return NULL;
    }

    if (a->argc != 4)
    {
        return CLI_SHOWUSAGE;
    }

    for (u = 0; u < ARRAY_LEN(levels); u++) {

        if (u < DYNAMIC_LEVEL_START && u != __LOG_DTMF)
        {
            continue;
        }

        if (u == __LOG_DTMF)
        {
            ast_log(LOG_DTMF, "Simulation a dynamic log, level is [%s:%02d] ^_^", 
                levels[u], u);
            continue;
        }
    
        if (levels[u] && (!strcasecmp(a->argv[3], "ALL") || !strcasecmp(S_OR(a->argv[3], ""), levels[u])))
        {
            ast_log_dynamic_level(u,
                "Simulation a dynamic log, level is [%s:%02d] ^_^", 
                levels[u], u);
        }
        
    }

    return CLI_SUCCESS;
}

static char *handle_logger_test_printf(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{

    switch (cmd) 
    {
    case CLI_INIT:
        e->command = "logger test printf";
        e->usage = 
            "Usage: logger test printf <string> \n"
            "       Simulation a specific log for dynamic level to this console.\n";
        return NULL;
        
    case CLI_GENERATE:
    
        return NULL;
    }

    if (a->argc < 4)
    {
        return CLI_SHOWUSAGE;
    }

    printf("%s\n", S_OR(a->argv[3], ""));

    return CLI_SUCCESS;
}

static char *handle_logger_show_file2module(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    LogFileModule *u;
    int            v;
    const char    *partial = NULL;
    
    switch (cmd) 
    {
    case CLI_INIT:
        e->command = "logger show file2module [like]";
        e->usage = 
            "Usage: logger show file2module [like partial_filename] \n"
            "       Show files with the module mapping.\n";
        return NULL;
        
    case CLI_GENERATE:


        if (4 == a->pos)
        {
            return logger_search_filename(S_OR(a->argv[4], ""), a->n);
        }
        else
        {
            return NULL;
        }
        
    }

    if (a->argc < 3)
    {
        return CLI_SHOWUSAGE;
    }

    if (a->argc >= 4) //partial_filename
    {
        if (strcasecmp(S_OR(a->argv[3], ""), "like"))
        {
            return CLI_SHOWUSAGE;
        }
        
        partial = S_OR(a->argv[4], "");
    }

    ast_cli(a->fd, "%-3s %-32s %-32s\n", "No", "file name", "module name");
    ast_cli(a->fd, "======================================================================\n");

    AST_RWLIST_RDLOCK(&g_log_file_module_list);

    v = -1;
    AST_LIST_TRAVERSE(&g_log_file_module_list, u, entry)
    {
        v++;
        
        if (NULL != partial && NULL == strcasestr(u->file_name, partial))
        {
            continue;
        }
        
        ast_cli(a->fd, "%03d %-32s %-32s\n", v, u->file_name, u->module_name);
        
    }

    AST_RWLIST_UNLOCK(&g_log_file_module_list);

    ast_cli(a->fd, "======================================================================\n");

    return CLI_SUCCESS;
}

static char *handle_logger_show_module2file(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    LogModuleFile *u;
    LogFilename   *v;
    const char    *partial = NULL;
    int       number;
    int        first;
    char module_buf[128];
    char   file_buf[128];

    switch (cmd) 
    {
    case CLI_INIT:
        e->command = "logger show module2file [like]"; /* TODO: Can input module name, and output module's filename only. [20120711] */
        e->usage = 
            "Usage: logger show module2file [like partial_modulename]\n"
            "       Show modules with the file mapping.\n";
        return NULL;
        
    case CLI_GENERATE:

        if (4 == a->pos)
        {
            return logger_search_modulename(S_OR(a->argv[4], ""), a->n);
        }
        else
        {
            return NULL;
        }
    }

    if (a->argc < 3)
    {
        return CLI_SHOWUSAGE;
    }

    if (a->argc >= 4) //partial_filename
    {
        if (strcasecmp(S_OR(a->argv[3], ""), "like"))
        {
            return CLI_SHOWUSAGE;
        }
    
        partial = S_OR(a->argv[4], "");
    }

    ast_cli(a->fd, "%-3s %-32s %-32s\n", "No", "module name", "file name");
    ast_cli(a->fd, "======================================================================\n");

    number = -1;
    
    AST_RWLIST_RDLOCK(&g_log_module_file_list);
    
    AST_LIST_TRAVERSE(&g_log_module_file_list, u, entry)
    {
        number++;
        
        if (partial != NULL && NULL == strcasestr(u->module_name, partial))
        {
            continue;
        }
    
        first  = 1;
        
        snprintf(module_buf, sizeof(module_buf), "%s (%d)", u->module_name, u->module_index);
        
        AST_LIST_TRAVERSE(&u->files_list, v, entry)
        {
            snprintf(file_buf, sizeof(file_buf), "%s (%d)", v->file_name, v->file_index);

            if (first)
            {
                first = 0;
                ast_cli(a->fd, "%03d %-32s %-32s\n", number, module_buf, file_buf);
            }
            else
            {
                ast_cli(a->fd, "%-3s %-32s %-32s\n", "", module_buf, file_buf);
            }
        }

        

    }

    AST_RWLIST_UNLOCK(&g_log_module_file_list);
    
    ast_cli(a->fd, "======================================================================\n");

    return CLI_SUCCESS;
}

static char *handle_logger_show_information(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int u;

    switch (cmd) 
    {
    case CLI_INIT:
        e->command = "logger show information ";
        e->usage = 
            "Usage: logger show information \n"
            "       show logger information.\n";
        return NULL;
        
    case CLI_GENERATE:
        return NULL;
        
    }

    if (a->argc != 3)
    {
        return CLI_SHOWUSAGE;
    }

    ast_cli(a->fd, "\n");
    ast_cli(a->fd, " %-32s %-32s\n", "Global variable", "value");
    ast_cli(a->fd, "======================================================================\n");
    ast_cli(a->fd, " %-32s: %d \n",  "ast_log_modules_count", ast_log_modules_count);
    ast_cli(a->fd, " %-32s: %d \n",  "g_log_file_module_list_count", g_log_file_module_list_count);
    ast_cli(a->fd, " %-32s: %zu \n", "ast_log_modules_level size(MAX)", ARRAY_LEN(ast_log_modules_level));
    ast_cli(a->fd, "----------------------------------------------------------------------\n");

    AST_RWLIST_RDLOCK(&logchannels);

    for (u = 0; u < ARRAY_LEN(levels); u++) {
    
        if (ast_log_level_is_dynamic(u, 0, __FILE__, __LINE__))
        {
            ast_cli(a->fd, " %-32s: %u \n", levels[u],  u);
        }
        
    }
    AST_RWLIST_UNLOCK(&logchannels);
    
    ast_cli(a->fd, " %-32s: %lu \n", "g_log_dynamic_register_count",   g_log_dynamic_register_count);
    ast_cli(a->fd, " %-32s: %lu \n", "g_log_dynamic_unregister_count", g_log_dynamic_unregister_count);
    ast_cli(a->fd, " %-32s: %lu \n", "logmsgs_queue_size",             logmsgs_size);
    ast_cli(a->fd, " %-32s: %lu \n", "logmsgs_drop_total",             logmsgs_drop_total);
    ast_cli(a->fd, "----------------------------------------------------------------------\n");
    ast_cli(a->fd, " %-32s: %s \n", "g_log_unvalid_file",     g_log_unvalid_file);
    ast_cli(a->fd, " %-32s: %d \n", "g_log_unvalid_line",     g_log_unvalid_line);
    ast_cli(a->fd, " %-32s: %s \n", "g_log_unvalid_callfile", g_log_unvalid_callfile);
    ast_cli(a->fd, " %-32s: %d \n", "g_log_unvalid_callline", g_log_unvalid_callline);
    ast_cli(a->fd, " %-32s: %d \n", "g_log_unvalid_value",    g_log_unvalid_value);
    ast_cli(a->fd, "======================================================================\n");

    return CLI_SUCCESS;
}

static char *handle_logger_show_level(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int u;
    int level_size;

    switch (cmd) 
    {
    case CLI_INIT:
        e->command = "logger show level ";
        e->usage = 
            "Usage: logger show level \n"
            "       show logger level information.\n";
        return NULL;
        
    case CLI_GENERATE:
        return NULL;
        
    }

    if (a->argc != 3)
    {
        return CLI_SHOWUSAGE;
    }

    ast_cli(a->fd, "\n");
    ast_cli(a->fd, "=================================\n");
    ast_cli(a->fd, " %3s %-12s %5s %5s \n", "idx", "name", "drop", "limit");
    ast_cli(a->fd, "---------------------------------\n");

    AST_RWLIST_RDLOCK(&logchannels);

    level_size = ARRAY_LEN(logmsgs_drop_level) < ARRAY_LEN(levels) ? ARRAY_LEN(logmsgs_drop_level) : ARRAY_LEN(levels);
    level_size = ARRAY_LEN(drop_limit) < level_size ? ARRAY_LEN(drop_limit) : level_size;

    for (u = 0; u < level_size; u++)
    {
        ast_cli(a->fd, " %02u  %-12s %5lu %5lu \n",
            u, levels[u] ? levels[u] : "(NULL)", logmsgs_drop_level[u], drop_limit[u]);
    }
    AST_RWLIST_UNLOCK(&logchannels);
    
    ast_cli(a->fd, "=================================\n");

    return CLI_SUCCESS;
}


static char *handle_logger_show_switch(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    LogModuleFile *u;
    int            v;
    int            w;

    int            level;

    switch (cmd) 
    {
    case CLI_INIT:
        e->command = "logger show switch ";
        e->usage = 
            "Usage: logger show switch \n"
            "       show logger switch for module.\n";
        return NULL;
        
    case CLI_GENERATE:
        return NULL;
        
    }

    if (a->argc != 3)
    {
        return CLI_SHOWUSAGE;
    }

    ast_cli(a->fd, "\r\n");
    ast_cli(a->fd, "%-3s %-25s %-5s %-5s %-5s %-5s %-5s \n", "No", "module name", 
        levels[__LOG_ERROR], levels[__LOG_WARNING], levels[__LOG_NOTICE], levels[__LOG_VERBOSE], levels[__LOG_DEBUG]);
    ast_cli(a->fd, "======================================================================\n");

    v = 0;
    
    AST_RWLIST_RDLOCK(&g_log_module_file_list);
    
    AST_LIST_TRAVERSE(&g_log_module_file_list, u, entry)
    {
    
        level = ast_log_modules_level[u->module_index];
        
        ast_cli(a->fd, "%03d %-25s %-5d %-5d %-5d %-5d %-5d\n", 
            v++, u->module_name, 
            (level >> __LOG_ERROR  & 1), (level >> __LOG_WARNING & 1),
            (level >> __LOG_NOTICE & 1), (level >> __LOG_VERBOSE & 1),
            (level >> __LOG_DEBUG  & 1)
            );
        
    }

    AST_RWLIST_UNLOCK(&g_log_module_file_list);

    ast_cli(a->fd, "----------------------------------------------------------------------\n");

    //ast_cli(a->fd, "%3d %-25s %-5s \n", __LOG_DTMF, levels[__LOG_DTMF], 
    //    (g_log_dynamic_level & (1 << __LOG_DTMF)) ? "ON" : "OFF");

    AST_RWLIST_RDLOCK(&logchannels);

    for (w = 0; w < ARRAY_LEN(levels); w++) {
    
        if (ast_log_level_is_dynamic(w, 0, __FILE__, __LINE__))
        {
            ast_cli(a->fd, "%3d %-25s %-5s \n", w, levels[w], 
                (ast_log_dynamic_levels_switch & (1 << w)) ? "ON" : "OFF");
        }
        
    }
    AST_RWLIST_UNLOCK(&logchannels);

    ast_cli(a->fd, "======================================================================\n");


    return CLI_SUCCESS;
}

static char *handle_logger_show_format(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int u;
    
    switch (cmd) 
    {
    case CLI_INIT:
        e->command = "logger show format";
        e->usage = 
            "Usage: logger show format \n"
            "       show logger format for current set.\n";
        return NULL;
        
    case CLI_GENERATE:
        return NULL;
        
    }

    if (a->argc != 3)
    {
        return CLI_SHOWUSAGE;
    }

    ast_cli(a->fd, "Current logger format set: 0x%04x\n", g_log_format);

    ast_cli(a->fd, "======================================================================\n");

    for (u = 0; u < ARRAY_LEN(log_formats); u++)
    {
        if (NULL == log_formats[u])
        {
            continue;
        }
        
        ast_cli(a->fd, " %-12s: %-2d \n", log_formats[u], (g_log_format >> u) & 0x1);
    }

    ast_cli(a->fd, "======================================================================\n");

    return CLI_SUCCESS;
}

int logger_get_moduleindex_by_name(const char *modulename)
{
    LogModuleFile *u;
    int   module_index = -1;

    AST_RWLIST_RDLOCK(&g_log_module_file_list);
    
    AST_LIST_TRAVERSE(&g_log_module_file_list, u, entry)
    {
        if (!strcasecmp(u->module_name, modulename))
        {
            module_index = u->module_index;
            break;
        }
    }

    AST_RWLIST_UNLOCK(&g_log_module_file_list);

    return module_index;
}

int logger_get_dlevelindex_by_name(const char *dlevelname)
{
    int  u;
    int dlevel_index = -1;

    if (NULL == dlevelname)
    {
        return dlevel_index;
    }

    AST_RWLIST_RDLOCK(&logchannels);

    for (u = 0; u < ARRAY_LEN(levels); u++)
    {
        if (NULL == levels[u] || strcmp(dlevelname, levels[u]))
        {
            continue;
        }

        if ((u >= DYNAMIC_LEVEL_START) || !strcmp(dlevelname, "DTMF"))
        {
            dlevel_index = u;
            break;
        }
    }

    AST_RWLIST_UNLOCK(&logchannels);

    return dlevel_index;
}

int logger_get_formatindex_by_name(const char *formatname, char sensitive)
{
    int u;
    int index = -1;

    if (NULL == formatname)
    {
        return index;
    }

    for (u = 0; u < ARRAY_LEN(log_formats); u++)
    {
        if (NULL == log_formats[u])
        {
            continue;
        }

        if (sensitive)
        {
            if (!strcmp(formatname, log_formats[u]))
            {
                index = u;
                break;
            }
        }
        else
        {
            if (!strcasecmp(formatname, log_formats[u]))
            {
                index = u;
                break;
            }
        }
        
    }

    return index;
}

int logger_get_levelindex_by_name(const char *levelname)
{
    int  u;
    int level_index = -1;

    if (NULL == levelname)
    {
        return level_index;
    }

    AST_RWLIST_RDLOCK(&logchannels);

    for (u = 0; u < ARRAY_LEN(levels); u++)
    {
        if (levels[u] && !strcmp(levelname, levels[u]))
        {
            level_index = u;
            break;
        }
    }

    AST_RWLIST_UNLOCK(&logchannels);

    return level_index;
}

char* logger_search_modulename(const char *partial, int skip_count)
{
    LogModuleFile *u;
    int            partial_len = strlen(partial);
    int            count = 0;
    char          *modulename = NULL;
    

    if (!strncasecmp("ALL", partial, partial_len)
            && ++count > skip_count)
    {
        return ast_strdup("ALL");
    }

    AST_RWLIST_RDLOCK(&g_log_module_file_list);
    
    AST_LIST_TRAVERSE(&g_log_module_file_list, u, entry)
    {
        if (!strncasecmp(u->module_name, partial, partial_len)
            && ++count > skip_count)
        {
            modulename = ast_strdup(u->module_name);
            break;
        }
    }

    if (NULL == modulename && 0 == count && partial_len > 1)
    {
        AST_LIST_TRAVERSE(&g_log_module_file_list, u, entry)
        {
            if (strcasestr(u->module_name, partial) != NULL
                && ++count > skip_count)
            {
                modulename = ast_strdup(u->module_name);
                break;
            }
        }
    }

    AST_RWLIST_UNLOCK(&g_log_module_file_list);

    return modulename;
}


char* logger_search_filename(const char *partial, int skip_count)
{
    LogFileModule *u;
    int            partial_len = strlen(partial);
    int            count = 0;
    char          *filename = NULL;
    

    if (!strncasecmp("ALL", partial, partial_len)
            && ++count > skip_count)
    {
        return ast_strdup("ALL");
    }

    AST_RWLIST_RDLOCK(&g_log_file_module_list);
    
    AST_LIST_TRAVERSE(&g_log_file_module_list, u, entry)
    {
        if (!strncasecmp(u->file_name, partial, partial_len)
            && ++count > skip_count)
        {
            filename = ast_strdup(u->file_name);
            break;
        }
    }

    if (NULL == filename && 0 == count && partial_len > 1)
    {
        AST_LIST_TRAVERSE(&g_log_file_module_list, u, entry)
        {
            if (strcasestr(u->file_name, partial) != NULL
                && ++count > skip_count)
            {
                filename = ast_strdup(u->file_name);
                break;
            }
        }
    }

    AST_RWLIST_UNLOCK(&g_log_file_module_list);

    return filename;
}


char* logger_search_levelname(const char *partial, int skip_count)
{
    int          partial_len = strlen(partial);
    int          count = 0;
    int          u;
    char        *levelname = NULL;
    static char *levels_name[6] = {"ALL"};

        
    levels_name[1] = levels[__LOG_DEBUG]; 
    levels_name[2] = levels[__LOG_VERBOSE]; 
    levels_name[3] = levels[__LOG_NOTICE]; 
    levels_name[4] = levels[__LOG_WARNING]; 
    levels_name[5] = levels[__LOG_ERROR];

    for (u = 0; u < ARRAY_LEN(levels_name); u++)
    {
        if (!strncasecmp(levels_name[u], partial, partial_len)
            && ++count > skip_count)
        {
            levelname = ast_strdup(levels_name[u]);
            break;
        }
    }

    return levelname;
}

char* logger_search_dlevelname(const char *partial, int skip_count)
{
    int          partial_len = strlen(partial);
    int          count = 0;
    int          u;
    char        *dlevelname = NULL;

    if (!strncasecmp(levels[__LOG_DTMF], partial, partial_len)
            && ++count > skip_count)
    {
        return ast_strdup(levels[__LOG_DTMF]);
    }

    AST_RWLIST_RDLOCK(&logchannels);

    for (u = DYNAMIC_LEVEL_START; u < ARRAY_LEN(levels); u++) {
    
        if (levels[u] && !strncasecmp(levels[u], partial, partial_len)
            && ++count > skip_count)
        {
            dlevelname = ast_strdup(levels[u]);
            break;
        }
        
    }
    AST_RWLIST_UNLOCK(&logchannels);


    return dlevelname;
}

char* logger_search_flagname(const char *partial, int skip_count)
{
    int          partial_len = strlen(partial);
    int          count = 0;
    int          u;
    char        *flagname = NULL;
    static char *flags_name[] = {"off", "on"};

    
    for (u = 0; u < ARRAY_LEN(flags_name); u++)
    {
        if (!strncasecmp(flags_name[u], partial, partial_len)
            && ++count > skip_count)
        {
            flagname = ast_strdup(flags_name[u]);
            break;
        }
    }

    return flagname;
}


void logger_set_default_level(void)
{
    int u;

    for (u = 0; u < ARRAY_LEN(ast_log_modules_level); u++)
    {
        /* log level default open 'ERROR' only */
        ast_log_modules_level[u] = (1 << __LOG_ERROR);
    }

    ast_log_dynamic_levels_switch = 0xFFFFFFF0;
}

void logger_init_all_switch(int out_fd)
{
    /** init option **/
    option_debug   = LOGGER_VERBOSE_REGISTER_OFF;
    option_verbose = LOGGER_VERBOSE_REGISTER_OFF;

    /* ignore debug_modules and verbose_modules */

    //ast_log(LOG_NOTICE, "%s fd:%d", __FUNCTION__, out_fd);
    
    /** init log level **/
    logger_set_default_level();
    logger_init_switch_from_db();
    logger_init_format_from_db();
    logger_save_switch_sqlite3();
    logger_save_format_sqlite3();

    /** init dynamic level log switch **/
}

inline void ast_log_exception(
    const char *filename, 
    int line, 
    const char *call_filename, 
    int call_line,
    int value)
{
    /* if (NULL == filename))  ???? */

    strncpy(g_log_unvalid_file, filename, sizeof(g_log_unvalid_file));
    g_log_unvalid_file[sizeof(g_log_unvalid_file) - 1] = '\0';
    g_log_unvalid_line     = line;

    strncpy(g_log_unvalid_callfile, call_filename, sizeof(g_log_unvalid_callfile));
    g_log_unvalid_callfile[sizeof(g_log_unvalid_callfile) - 1] = '\0';
    g_log_unvalid_callline = call_line;
    
    g_log_unvalid_value    = value;
}

int logger_add_file_module_list(const char *filename, const char *modulename)
{
    LogFileModule *fm;
    LogFileModule *u;
    LogFileModule *prev_fm;
    static int file_load_index = 0;

    LogModuleFile *mf;
    LogModuleFile *v;
    LogModuleFile *prev_mf;

    LogFilename   *fn;
    LogFilename   *w;
    LogFilename   *prev_fn;


    if (NULL == filename || NULL == modulename)
    {
        ast_log(LOG_ERROR, "Parameter(0x%08zx, 0x%08zx) is NULL!!\r\n", (ptrdiff_t)filename, (ptrdiff_t)modulename);
        return -100;
    }  


    /** insert into module_file list */
    AST_RWLIST_WRLOCK(&g_log_module_file_list);

    mf = NULL;
    
    AST_LIST_TRAVERSE(&g_log_module_file_list, v, entry)
    {
            
        if (!strcmp(v->module_name, modulename)) {
            mf = v;
            break;
        }

    }

    /* new file */
    fn = ast_calloc(1, sizeof(LogFilename));
    if (!fn) {
        ast_log(LOG_ERROR, "Malloc(%zu) filename structor failed!!", sizeof(LogFilename));
        AST_RWLIST_UNLOCK(&g_log_module_file_list);
        return -110;
    }

    strncpy(fn->file_name, filename, sizeof(fn->file_name));
    fn->file_name[sizeof(fn->file_name) - 1] = '\0';

    if (NULL == mf)
    {

        mf = ast_calloc(1, sizeof(LogModuleFile));
        if (!mf) {
            ast_log(LOG_ERROR, "Malloc(%zu) module structor failed!!", sizeof(LogModuleFile));
            AST_RWLIST_UNLOCK(&g_log_module_file_list);
            ast_free(fn);
            fn = NULL;
            return -120;
        }

        strncpy(mf->module_name, modulename, sizeof(mf->module_name));
        mf->module_name[sizeof(mf->module_name) - 1] = '\0';  
        
        mf->file_count = 0;
        mf->module_index = ast_log_modules_count++;

        if (ast_log_modules_count > (sizeof(ast_log_modules_level) / sizeof(int)) )
        {
            ast_log(LOG_ERROR, "Array g_log_modules_level[%zu] overflow[%d]!!\r\n",
                (sizeof(ast_log_modules_level) / sizeof(int)), 
                ast_log_modules_count);

            AST_RWLIST_UNLOCK(&g_log_module_file_list);
            ast_free(fn);
            ast_free(mf);
            fn = NULL;
            mf = NULL;
            return -130;
        }

        /* default switch value */
        ast_log_modules_level[mf->module_index] = 0xFFFFFFFF;

        /* iterate module_file list, insert order by module_name */
        prev_mf = NULL;
        AST_LIST_TRAVERSE(&g_log_module_file_list, v, entry)
        {
            if (strcasecmp(v->module_name, mf->module_name) > 0)
            {
                break;
            }
            
            prev_mf = v;
        }

        if (NULL == prev_mf)
        {
            AST_RWLIST_INSERT_HEAD(&g_log_module_file_list, mf, entry);
        }
        else
        {
            AST_RWLIST_INSERT_AFTER(&g_log_module_file_list, prev_mf, mf, entry);
        }
    
    }

    prev_fn = NULL;
    /* iterate filename list, insert order by file_name */
    AST_LIST_TRAVERSE(&mf->files_list, w, entry)
    {
        if (strcasecmp(w->file_name, fn->file_name) > 0)
        {
            break;
        }
        
        prev_fn = w;
    }

    fn->module_index = mf->module_index;

    if (NULL == prev_fn)
    {
        AST_RWLIST_INSERT_HEAD(&mf->files_list, fn, entry);
    }
    else
    {
        AST_RWLIST_INSERT_AFTER(&mf->files_list, prev_fn, fn, entry);
    }

    fn->file_index = ++mf->file_count;

    AST_RWLIST_UNLOCK(&g_log_module_file_list);


    /* insert file_module list */
    AST_RWLIST_WRLOCK(&g_log_file_module_list);

    fm = ast_calloc(1, sizeof(LogFileModule));
    if (!fm) {
        ast_log(LOG_ERROR, "Malloc(%zu) failed!!\r\n", sizeof(LogFileModule));
        AST_RWLIST_UNLOCK(&g_log_file_module_list);
        return -140;
    }

    strncpy(fm->file_name, filename, sizeof(fm->file_name));
    fm->file_name[sizeof(fm->file_name) - 1] = '\0';

    strncpy(fm->module_name, modulename, sizeof(fm->module_name));
    fm->module_name[sizeof(fm->module_name) - 1] = '\0';  

    fm->module_index = mf->module_index;
    fm->file_load_index = file_load_index++;

    prev_fm = NULL; /* the node before 'u' */

    /* order by file_name */
    AST_LIST_TRAVERSE(&g_log_file_module_list, u, entry)
    {
        
        if (strcasecmp(u->file_name, fm->file_name) > 0)
        {
            break;
        }

        prev_fm = u;
    }

    if (NULL == prev_fm)
    {
        AST_RWLIST_INSERT_HEAD(&g_log_file_module_list, fm, entry);
    }
    else
    {
        AST_RWLIST_INSERT_AFTER(&g_log_file_module_list, prev_fm, fm, entry);
    }
    
    g_log_file_module_list_count++;

    AST_RWLIST_UNLOCK(&g_log_file_module_list);
    

    return mf->module_index;
} 

void logger_del_file_module_list(const char *file_name)
{
    LogFileModule *u;
    LogFileModule *fm = NULL;

    LogModuleFile *v;
    LogModuleFile *mf = NULL;

    LogFilename   *w;
    LogFilename   *fn = NULL;

    /* remove element in file_module list */
    AST_RWLIST_WRLOCK(&g_log_file_module_list);

    AST_LIST_TRAVERSE(&g_log_file_module_list, u, entry)
    {
        if (!strcmp(u->file_name, file_name))
        {
            fm = u;
            break;
        }
    }
    
    if (fm)
    {
        AST_LIST_REMOVE(&g_log_file_module_list, fm, entry);
        g_log_file_module_list_count--;
        ast_free(fm);
        fm = NULL;
    }
    
    AST_RWLIST_UNLOCK(&g_log_file_module_list);
    
    /* remove element in module_file list */
    AST_RWLIST_WRLOCK(&g_log_module_file_list);

    AST_LIST_TRAVERSE(&g_log_module_file_list, v, entry)
    {
        AST_LIST_TRAVERSE(&v->files_list, w, entry)
        {
            if (!strcmp(w->file_name, file_name))
            {
                fn = w;
                mf = v;
                break;
            }
        }
    }

    if (fn)
    {
        
        AST_LIST_REMOVE(&mf->files_list, fn, entry);
        mf->file_count--;

        /* Order file index */
        AST_LIST_TRAVERSE(&mf->files_list, w, entry)
        {
            if (w->file_index > fn->file_index)
            {
                w->file_index--;
            }
        }
        
        ast_free(fn);
        fn = NULL;
    }
    
    AST_RWLIST_UNLOCK(&g_log_module_file_list);
}


int logger_save_switch_sqlite3(void)
{
    LogModuleFile *u;
    sqlite3       *db = NULL;
    int            v;
    int            result;
    int            ret = 0;
    const char    *fieldset[6] = {
        levels[__LOG_ERROR], 
        levels[__LOG_WARNING], 
        levels[__LOG_NOTICE], 
        levels[__LOG_VERBOSE], 
        levels[__LOG_DEBUG], 
        LOGGER_FIELD_MODULENAME};
    const char    *valueset[6];
    
    result = logger_switch_open_sqlite3(&db);

    if (result != SQLITE_OK)
    {
        ast_log(LOG_ERROR, "Cannot open sqlite3 '%s/etc/%s', result %d, save logger switch failed!!", getenv("AST_BASE"), LOGGER_SWITCH_DB_FILE, result);
        return -1;
    }

    // jcli modify 2013-12-02
    if(SQLITE_OK != (ret = sqlite3_exec(db, "BEGIN IMMEDIATE;", 0, 0, 0)))
    {
        ast_log(LOG_WARNING, "Can't begin a IMMEDIATE transaction!!");
        return ret;
    }
    // jcli end

    /** static level table */
    AST_RWLIST_RDLOCK(&g_log_module_file_list);

    AST_LIST_TRAVERSE(&g_log_module_file_list, u, entry)
    {
        if (NULL == u)
        {
            ast_log(LOG_ERROR, "Have NULL element in 'g_log_module_file_list'!!");
            continue;
        }

        if ('\0' == u->module_name[0])
        {
            ast_log(LOG_ERROR, "[%d]Have empty module_name in 'g_log_module_file_list'!! file_count:%d.",
                u->module_index, u->file_count);
            continue;
        }

        if (u->module_index < 0 || u->module_index >= ARRAY_LEN(ast_log_modules_level))
        {
            ast_log(LOG_ERROR, "Module(%s)'s index[%d] is unvalid!! file_count:%d.",
                u->module_name, u->module_index, u->file_count);
            continue;
        }
        
        valueset[0] = (ast_log_modules_level[u->module_index] & (1 << __LOG_ERROR))   ? "1" : "0";
        valueset[1] = (ast_log_modules_level[u->module_index] & (1 << __LOG_WARNING)) ? "1" : "0";
        valueset[2] = (ast_log_modules_level[u->module_index] & (1 << __LOG_NOTICE))  ? "1" : "0";
        valueset[3] = (ast_log_modules_level[u->module_index] & (1 << __LOG_VERBOSE)) ? "1" : "0";
        valueset[4] = (ast_log_modules_level[u->module_index] & (1 << __LOG_DEBUG))   ? "1" : "0";
        valueset[5] = u->module_name;

        /* Query static level table */
        result = logger_sqlite3_record_isexist(db, LOGGER_TABLE_NAME_STATIC, LOGGER_FIELD_MODULENAME, u->module_name);

        if (1 == result)
        {
            logger_sqlite3_update_record(db, LOGGER_TABLE_NAME_STATIC, fieldset, valueset, 5, LOGGER_FIELD_MODULENAME, u->module_name);
        }
        else if (0 == result)
        {
            ast_log(LOG_WARNING, "%s(%s) not exists in table '%s', so add a new record! ", LOGGER_FIELD_MODULENAME, u->module_name, LOGGER_TABLE_NAME_STATIC);
            logger_sqlite3_add_record(db, LOGGER_TABLE_NAME_STATIC, fieldset, valueset, 6);
        }
        else
        {
            ast_log(LOG_WARNING, "Select %s(%s) from %s failed %d, try add a new record! ", LOGGER_FIELD_MODULENAME, u->module_name, LOGGER_TABLE_NAME_STATIC, result);
            logger_sqlite3_add_record(db, LOGGER_TABLE_NAME_STATIC, fieldset, valueset, 6);
        }

    }

    AST_RWLIST_UNLOCK(&g_log_module_file_list);

    /** synamic level table */
    AST_RWLIST_RDLOCK(&logchannels);

    for (v = 0; v < ARRAY_LEN(levels); v++)
    {
        if (NULL == levels[v] || (v < DYNAMIC_LEVEL_START && v != __LOG_DTMF))
        {
            continue;
        }

        /* Query dynamic level table */
        result = logger_sqlite3_record_isexist(db, LOGGER_TABLE_NAME_DYNAMIC, LOGGER_FIELD_DLEVEL, levels[v]);

        fieldset[0] = LOGGER_FIELD_SWITCH;
        fieldset[1] = LOGGER_FIELD_DLEVEL;
        valueset[0] = (ast_log_dynamic_levels_switch & (1 << v)) ? "1" : "0";
        valueset[1] = levels[v];
        
        if (1 == result)
        {
            logger_sqlite3_update_record(db, LOGGER_TABLE_NAME_DYNAMIC, fieldset, valueset, 1, LOGGER_FIELD_DLEVEL, levels[v]);
        }
        else if (0 == result)
        {
            ast_log(LOG_WARNING, "%s(%s) not exists in table '%s', so add a new record! ", LOGGER_FIELD_DLEVEL, levels[v], LOGGER_TABLE_NAME_DYNAMIC);
            logger_sqlite3_add_record(db, LOGGER_TABLE_NAME_DYNAMIC, fieldset, valueset, 2);
        }
        else
        {
            ast_log(LOG_WARNING, "Select %s(%s) from %s failed %d, try add a new record! ", LOGGER_FIELD_DLEVEL, levels[v], LOGGER_TABLE_NAME_DYNAMIC, result);
            logger_sqlite3_add_record(db, LOGGER_TABLE_NAME_DYNAMIC, fieldset, valueset, 2);
        }

    }

    AST_RWLIST_UNLOCK(&logchannels);

    // jcli modify 2013-12-02
    if(SQLITE_OK != (ret = sqlite3_exec(db, "commit;", 0, 0, 0)))
    {    
        ast_log(LOG_WARNING, "Logger save switch failed! ERR_REASON: %s. ERRCODE: %d.\n", sqlite3_errmsg(db), ret);
        sqlite3_exec(db, "rollback;", 0, 0, 0);
    }
    // jcli end    
    sqlite3_close(db);
    return ret;
}

int logger_save_format_sqlite3(void)
{
    sqlite3    *db = NULL;
    int         u;
    int         result;
    int         ret = 0;
    const char *fieldset[2];
    const char *valueset[2];
    
    result = logger_switch_open_sqlite3(&db);

    if (result != SQLITE_OK)
    {
        ast_log(LOG_ERROR, "Cannot open sqlite3 '%s/etc/%s', result %d, save logger switch failed!!", getenv("AST_BASE"), LOGGER_SWITCH_DB_FILE, result);
        return -1;
    }

    // jcli modify 2013-12-02
    if(SQLITE_OK != (ret = sqlite3_exec(db, "BEGIN IMMEDIATE;", 0, 0, 0)))
    {
        ast_log(LOG_WARNING, "Can't begin a IMMEDIATE transaction!!");
        return ret;
    }
    // jcli end
    
    for (u = 0; u < ARRAY_LEN(log_formats); u++)
    {
        if (NULL == log_formats[u])
        {
            continue;
        }
        
        fieldset[0] = LOGGER_FIELD_SWITCH;
        fieldset[1] = LOGGER_FIELD_FORMAT;
        valueset[0] = (g_log_format & (1 << u)) ? "1" : "0";
        valueset[1] = log_formats[u];

        /* Query log format table */
        result = logger_sqlite3_record_isexist(db, LOGGER_TABLE_NAME_FORMAT, LOGGER_FIELD_FORMAT, log_formats[u]);

        if (1 == result)
        {
            logger_sqlite3_update_record(db, LOGGER_TABLE_NAME_FORMAT, fieldset, valueset, 1, LOGGER_FIELD_FORMAT, log_formats[u]);
        }
        else if (0 == result)
        {
            ast_log(LOG_WARNING, "%s(%s) not exists in table '%s', so add a new record! ", LOGGER_FIELD_FORMAT, log_formats[u], LOGGER_TABLE_NAME_FORMAT);
            logger_sqlite3_add_record(db, LOGGER_TABLE_NAME_FORMAT, fieldset, valueset, 2);
        }
        else
        {
            ast_log(LOG_WARNING, "Select %s(%s) from %s failed %d, try add a new record! ", LOGGER_FIELD_FORMAT, log_formats[u], LOGGER_TABLE_NAME_FORMAT, result);
            logger_sqlite3_add_record(db, LOGGER_TABLE_NAME_FORMAT, fieldset, valueset, 2);
        }
    }

    // jcli modify 2013-12-02
    if(SQLITE_OK != (ret = sqlite3_exec(db, "COMMIT;", 0, 0, 0)))
    {
        ast_log(LOG_WARNING, "Logger save format failed! ERR_REASON: %s. ERRCODE: %d.\n", sqlite3_errmsg(db), ret);
        sqlite3_exec(db, "ROLLBACK;", NULL, 0, NULL);
    }    
    // jcli end
    sqlite3_close(db);    
    return ret;
}

int logger_init_format_from_db(void)
{
    sqlite3     *db = NULL;
    char    *errmsg = NULL;
    char **dbResult = NULL;
    char sql[128];
    int nRow = 0, nColumn = 0;
    int u;
    int index;
    int result;

    result = logger_switch_open_sqlite3(&db);

    if (result != SQLITE_OK)
    {
        ast_log(LOG_ERROR, "Cannot open sqlite3, result %d, init logger format failed!!", result);
        return -1;
    }

    snprintf(sql, sizeof(sql), "select * from `%s`", LOGGER_TABLE_NAME_FORMAT);

    result = sqlite3_get_table(db, sql, &dbResult, &nRow, &nColumn, &errmsg);
    if (SQLITE_OK != result)
    {
        ast_log(LOG_ERROR, "sqlite3_get_table(\"%s\") result %d failed; errmsg:%s !!", sql, result, errmsg);
        sqlite3_close(db);
        return -1;
    }

    if (nRow <= 0 || nColumn <= 0)
    {
        ast_log(LOG_WARNING, "[%s] no record[%d, %d]!", sql, nRow, nColumn);
    }

    /* Table 'astlog_format' must have
      'id' 'format' 'switch' three fields. */
    if (nColumn > 0 && nColumn < 3) 
    {
        ast_log(LOG_ERROR, "Table %s's field(%d) less than 3!!", LOGGER_TABLE_NAME_FORMAT, nColumn);
        sqlite3_free_table(dbResult);
        sqlite3_close(db);
        return -1;
    }
 

    for (u = 1; u <= nRow; u++ )
    {
        
        index = logger_get_formatindex_by_name(S_OR(dbResult[u * nColumn + 1], ""), 1);

        if (index >= 0 && index < ARRAY_LEN(log_formats))
        {
            if (strcmp(dbResult[u * nColumn + 2], "0"))
            {
                g_log_format |= 1 << index;
            }
            else
            {
                g_log_format &= ~(1 << index);
            }
        }
        else
        { /* this log format is not exits, so delete this record in LOGGER_TABLE_NAME_FORMAT table */
            ast_log(LOG_WARNING, "Unknow %s(%s) index is %d, so delete record from table '%s'", 
                LOGGER_FIELD_FORMAT, dbResult[u * nColumn + 1], index, LOGGER_TABLE_NAME_FORMAT);
            result = logger_sqlite3_delete_record(
                        db,
                        LOGGER_TABLE_NAME_FORMAT, 
                        LOGGER_FIELD_FORMAT, 
                        dbResult[u * nColumn + 1]);

            continue;
        }
        
    } // ? for (u = 1; u < nRow; u++ ) ?  end static

    sqlite3_free_table(dbResult);
    
    sqlite3_close(db);
    
    return 0;
}

int logger_init_switch_from_db(void)
{
    sqlite3     *db = NULL;
    char    *errmsg = NULL;
    char **dbResult = NULL;
    char sql[128];
    int nRow = 0, nColumn = 0;
    int u, v;
    int index, module_index;
    int result;

    result = logger_switch_open_sqlite3(&db);

    if (result != SQLITE_OK)
    {
        ast_log(LOG_ERROR, "Cannot open sqlite3, result %d, init logger switch failed!!", result);
        return -1;
    }

    /** static level table */
    snprintf(sql, sizeof(sql), "select * from `%s`", LOGGER_TABLE_NAME_STATIC);

    result = sqlite3_get_table(db, sql, &dbResult, &nRow, &nColumn, &errmsg);
    if (SQLITE_OK != result)
    {
        ast_log(LOG_ERROR, "sqlite3_get_table(\"%s\") result %d failed; errmsg:%s !!", sql, result, errmsg);
        sqlite3_close(db);
        return -1;
    }

    if (nRow <= 0 || nColumn <= 0)
    {
        ast_log(LOG_WARNING, "[%s] no record[%d, %d]!", sql, nRow, nColumn);
    }

    /* Table 'astlog_switch_static' must have
      'id' 'module_name' 'ERROR' 'WARN' 'NOTIC' 'VERB' 'DEBUG' seven fields. */
    if (nColumn > 0 && nColumn < 7) 
    {
        ast_log(LOG_ERROR, "Table %s's field(%d) less than 7!!", LOGGER_TABLE_NAME_STATIC, nColumn);
        sqlite3_free_table(dbResult);
        sqlite3_close(db);
        return -1;
    }
 

    for (u = 1; u <= nRow; u++ )
    {
    
        module_index = logger_get_moduleindex_by_name(dbResult[u * nColumn + 1]); /* 'module_name' is second field */

        if (module_index < 0 || module_index >= ARRAY_LEN(ast_log_modules_level))
        { /* this module_name is not exits, so delete this record in LOGGER_TABLE_NAME_STATIC table */
            ast_log(LOG_WARNING, "Unknow %s(%s) index is %d, so delete record in table '%s'", 
                LOGGER_FIELD_MODULENAME, dbResult[u * nColumn + 1], module_index, LOGGER_TABLE_NAME_STATIC);
            result = logger_sqlite3_delete_record(
                        db,
                        LOGGER_TABLE_NAME_STATIC, 
                        LOGGER_FIELD_MODULENAME, 
                        dbResult[u * nColumn + 1]);

            continue;
        }

        for (v = 2; v < nColumn; v++) /* first level field startwith 2 */
        {
        
            index = logger_get_levelindex_by_name(dbResult[v]); /* dbResult[0 * nColumn + v] */

            if (index < 0 || index >= ARRAY_LEN(levels))
            { /* this level name is not exits, so must recreate LOGGER_TABLE_NAME_STATIC table */
                ast_log(LOG_ERROR, "Logger level[%s] is not exists(%d), need recreate table '%s' !!", 
                    dbResult[v], index, LOGGER_TABLE_NAME_STATIC);
                continue;
            }

            if (strcmp("0" , dbResult[u * nColumn + v])) 
            { /* ON */
                ast_log_modules_level[module_index] |= (1 << index);
            }
            else
            { /* OFF */
                ast_log_modules_level[module_index] &= ~(1 << index);
            }

        }
        
    } // ? for (u = 0; u < nRow; u++ ) ?  end static

    sqlite3_free_table(dbResult);
    

    /** dynamic level table */
    snprintf(sql, sizeof(sql), "select * from `%s`", LOGGER_TABLE_NAME_DYNAMIC);

    result = sqlite3_get_table(db, sql, &dbResult, &nRow, &nColumn, &errmsg);
    if (SQLITE_OK != result)
    {
        ast_log(LOG_ERROR, "sqlite3_get_table(\"%s\") result %d failed; errmsg:%s !!", sql, result, errmsg);
        sqlite3_close(db);
        return -1;
    }

    if (nRow <= 0 || nColumn <= 0)
    {
        ast_log(LOG_WARNING, "[%s] no record(%d, %d)!", sql, nRow, nColumn);
    }
    
    if (nColumn > 0 && nColumn < 3) /* Table 'astlog_switch_dynamic' must have 'id' 'dlevel' 'switch' fields. */
    {
        ast_log(LOG_ERROR, "Table %s's field(%d) less than 3 !!", LOGGER_TABLE_NAME_DYNAMIC, nColumn);
        sqlite3_free_table(dbResult);
        sqlite3_close(db);
        return -1;
    }
 
    for (u = 1; u <= nRow; u++ )
    {

        index = logger_get_dlevelindex_by_name(dbResult[u * nColumn + 1]); /* 'dlevel' is second field */

        if (index < 0 || index >= ARRAY_LEN(levels))
        { /* this dlevel is not exits, so delete this record in LOGGER_TABLE_NAME_DYNAMIC table */
            ast_log(LOG_WARNING, "Unknow %s(%s) index is %d, so delete record in table '%s'", 
                LOGGER_FIELD_DLEVEL, dbResult[u * nColumn + 1], index, LOGGER_TABLE_NAME_DYNAMIC);

            result = logger_sqlite3_delete_record(
                        db,
                        LOGGER_TABLE_NAME_DYNAMIC, 
                        LOGGER_FIELD_DLEVEL, 
                        dbResult[u * nColumn + 1]);

            continue;
        }

        if (strcmp("0" , dbResult[u * nColumn + 2])) /* field 'switch' is third */
        { /* ON */
            ast_log_dynamic_levels_switch |= (1 << index);
        }
        else
        { /* OFF */
            ast_log_dynamic_levels_switch &= ~(1 << index);
        }
        
    }

    sqlite3_free_table(dbResult);
    
    sqlite3_close(db);
    
    return 0;
}

static int logger_switch_open_sqlite3(sqlite3** db)
{
    char db_file[256];
    //char *errMsg;
    int result;

    if (NULL == db)
    {
        ast_log(LOG_ERROR, "logger_switch_open_sqlite3(0x%08zx) paramenter is NULL!!", (ptrdiff_t)db);
        return -1;
    }
    
    snprintf(db_file, sizeof(db_file), "%s/etc/%s", getenv("AST_BASE"), LOGGER_SWITCH_DB_FILE);

    result = sqlite3_open(db_file, db);

    if (result != SQLITE_OK)
    {
        ast_log(LOG_ERROR, "Cannot open '%s' result:%d", db_file, result);
        return result;
    }

    /* jcli add 2013-12-04 */
    sqlite3_busy_timeout(*db, 5000);
    /* jcli end */

    return SQLITE_OK;
}

/* insert into 'table' (f1, f2, f3...) values (v1, v2, v3...) */
static int logger_sqlite3_add_record(sqlite3* db, const char *table, const char **fields, const char **values, int size)
{
    sqlite3_stmt *pStmt = NULL;
    char          sql[128];
    int           result;
    int           sql_len = 0;
    int           v;

    if (NULL == db || NULL == table || NULL == fields || NULL == values || size < 1)
    {
        ast_log(LOG_ERROR, "logger_sqlite3_add_record(0x%08zx, 0x%08zx, 0x%08zx, 0x%08zx, %d) paramenter is NULL!!",
            (ptrdiff_t)db, (ptrdiff_t)table, (ptrdiff_t)fields, (ptrdiff_t)values, size);
        return -1;
    }
    
    sql_len += snprintf(sql, sizeof(sql), "insert into `%s` (`%s`", table, fields[0]);

    for (v = 1; v < size; v++)
    {
        sql_len += snprintf(sql + sql_len, sizeof(sql) - sql_len, ", `%s`", fields[v]);
    }

    sql_len += snprintf(sql + sql_len, sizeof(sql) - sql_len, ") values (?");

    for (v = 1; v < size; v++)
    {
        sql_len += snprintf(sql + sql_len, sizeof(sql) - sql_len, ", ?");
    }

    sql_len += snprintf(sql + sql_len, sizeof(sql) - sql_len, ");");


    ast_log(LOG_NOTICE, "Execute <%s> (value=%s ...)", sql, values[0]);


    result = sqlite3_prepare(db, sql, -1, &pStmt, NULL);
    if (result != SQLITE_OK)  
    {
        ast_log(LOG_ERROR, "sqlite3_prepare(0x%08zx, '%s', -1, stmt, NULL) return %d failed!!", (ptrdiff_t)db, sql, result);
        return result;
    }

    for (v = 0; v < size; v++)
    {
        result = sqlite3_bind_text(pStmt, v + 1, values[v], strlen(values[v]), NULL);
        if (result != SQLITE_OK)  
        {
            ast_log(LOG_ERROR, "sqlite3_bind_text(0x%08zx, %d, '%s', %zd, NULL) return %d failed!!", (ptrdiff_t)pStmt, v + 1, values[v], strlen(values[v]), result);
            sqlite3_finalize(pStmt); 
            return result;
        }
    }

    result = sqlite3_step(pStmt);
    if (result != SQLITE_DONE)  
    {
        ast_log(LOG_ERROR, "sqlite3_step(0x%08zx) return %d failed!!sql:%s.", (ptrdiff_t)pStmt, result, sql);
        sqlite3_finalize(pStmt); 
        return result;
    }

    sqlite3_finalize(pStmt);
   
    return SQLITE_OK;
}


/* delete from 'table' where 'field' = 'value' */
static int logger_sqlite3_delete_record(sqlite3* db, const char *table, const char *field, const char *value)
{
    sqlite3_stmt *pStmt = NULL;
    char          sql[128];
    int           result;

    if (NULL == db || NULL == table || NULL == field || NULL == value)
    {
        ast_log(LOG_ERROR, "logger_sqlite3_delete_record(0x%08zx, 0x%08zx, 0x%08zx, 0x%08zx) paramenter is NULL!!",
            (ptrdiff_t)db, (ptrdiff_t)table, (ptrdiff_t)field, (ptrdiff_t)value);
        return -1;
    }
    
    snprintf(sql, sizeof(sql), "delete from `%s` where `%s` = ?", table, field);

    ast_log(LOG_NOTICE, "Execute <%s> (value=%s) ...", sql, value);
   
    result = sqlite3_prepare(db, sql, -1, &pStmt, NULL);
    if (result != SQLITE_OK)  
    {
        ast_log(LOG_ERROR, "sqlite3_prepare(0x%08zx, '%s', -1, stmt, NULL) return %d failed!!", (ptrdiff_t)db, sql, result);
        return result;
    }
    
    result = sqlite3_bind_text(pStmt, 1, value, strlen(value), NULL);
    if (result != SQLITE_OK)  
    {
        ast_log(LOG_ERROR, "sqlite3_bind_text(0x%08zx, 1, '%s', %zd, NULL) return %d failed!!", (ptrdiff_t)pStmt, value, strlen(value), result);
        sqlite3_finalize(pStmt); 
        return result;
    }

    result = sqlite3_step(pStmt);
    if (result != SQLITE_DONE)  
    {
        ast_log(LOG_ERROR, "sqlite3_step(0x%08zx) return %d failed!!", (ptrdiff_t)pStmt, result);
        sqlite3_finalize(pStmt); 
        return result;
    }

    sqlite3_finalize(pStmt);
    
    return SQLITE_OK;
}

/* update 'table' set 'field1' = 'new'  where 'field2' = 'value' */
static int logger_sqlite3_update_record(sqlite3* db, const char *table, const char **fieldset, const char **newset, int size, const char *where, const char *value)
{
    sqlite3_stmt *pStmt = NULL;
    char          sql[256];
    int           result;
    int           sql_len = 0;
    int           v;

    if (NULL == db || NULL == table || NULL == fieldset || NULL == newset || size < 1 || NULL == where || NULL == value)
    {
        ast_log(LOG_ERROR, "logger_sqlite3_update_record(0x%08zx, 0x%08zx, 0x%08zx, 0x%08zx, %d, 0x%08zx, 0x%08zx) paramenter is NULL!!",
            (ptrdiff_t)db, (ptrdiff_t)table, (ptrdiff_t)fieldset, (ptrdiff_t)newset, size, (ptrdiff_t)where, (ptrdiff_t)value);
        return -1;
    }
    
    sql_len += snprintf(sql + sql_len, sizeof(sql) - sql_len, "update `%s` set `%s` = ?", table, fieldset[0]);

    for (v = 1; v < size; v++)
    {
        sql_len += snprintf(sql + sql_len, sizeof(sql) - sql_len, ", `%s` = ? ", fieldset[v]);
    }

    sql_len += snprintf(sql + sql_len, sizeof(sql) - sql_len, " where `%s` = ?", where);

    ast_log(LOG_NOTICE, "Execute <%s> (value=%s) ...", sql, value);

    result = sqlite3_prepare(db, sql, -1, &pStmt, NULL);
    if (result != SQLITE_OK)  
    {
        ast_log(LOG_ERROR, "sqlite3_prepare(0x%08zx, '%s', -1, stmt, NULL) return %d failed!!", (ptrdiff_t)db, sql, result);
        return result;
    }

    for (v = 0; v < size; v++)
    {
        result = sqlite3_bind_text(pStmt, v + 1, newset[v], strlen(newset[v]), NULL);
        if (result != SQLITE_OK)  
        {
            ast_log(LOG_ERROR, "sqlite3_bind_text(0x%08zx, %d, '%s', %zd, NULL) return %d failed!!",
                (ptrdiff_t)pStmt, v + 1, newset[v], strlen(newset[v]), result);
            sqlite3_finalize(pStmt); 
            return result;
        }
    }

    result = sqlite3_bind_text(pStmt, size + 1, value, strlen(value), NULL);
    if (result != SQLITE_OK)  
    {
        ast_log(LOG_ERROR, "sqlite3_bind_text(0x%08zx, %d, '%s', %zd, NULL) return %d failed!!",
            (ptrdiff_t)pStmt, size + 1, value, strlen(value), result);
        sqlite3_finalize(pStmt); 
        return result;
    }

    result = sqlite3_step(pStmt);
    if (result != SQLITE_DONE)  
    {
        ast_log(LOG_ERROR, "sqlite3_step(0x%08zx) return %d failed!! sql:%s.", (ptrdiff_t)pStmt, result, sql);
        sqlite3_finalize(pStmt); 
        return result;
    }

    sqlite3_finalize(pStmt);
   
    return SQLITE_OK;
}

/* select * from 'table' where 'field' = 'value' */
static int logger_sqlite3_record_isexist(sqlite3* db, const char *table, const char *field, const char *value)
{
    sqlite3_stmt *pStmt = NULL;
    char          sql[128];
    int           result;

    if (NULL == db || NULL == table || NULL == field || NULL == value)
    {
        ast_log(LOG_ERROR, "logger_sqlite3_record_isexist(0x%08zx, 0x%08zx, 0x%08zx, 0x%08zx) paramenter is NULL!!",
            (ptrdiff_t)db, (ptrdiff_t)table, (ptrdiff_t)field, (ptrdiff_t)value);
        return -1;
    }

    snprintf(sql, sizeof(sql), "select * from `%s` where `%s` = ?", table, field);

    ast_log(LOG_NOTICE, "Execute <%s> (value=%s) ...", sql, value);

    result = sqlite3_prepare(db, sql, -1, &pStmt, NULL);
    if (result != SQLITE_OK)  
    {
        ast_log(LOG_ERROR, "sqlite3_prepare(0x%08zx, <%s>, -1, stmt, NULL) return %d failed!!", (ptrdiff_t)db, sql, result);
        return -1;
    }
    
    result = sqlite3_bind_text(pStmt, 1, value, strlen(value), NULL);
    if (result != SQLITE_OK)  
    {
        ast_log(LOG_ERROR, "sqlite3_bind_text(0x%08zx, 1, '%s', %zd, NULL) return %d failed!!", (ptrdiff_t)pStmt, value, strlen(value), result);
        sqlite3_finalize(pStmt); 
        return -1;
    }

    result = sqlite3_step(pStmt);

    sqlite3_finalize(pStmt);
    
    if (SQLITE_DONE == result)
    {
        ast_log(LOG_DEBUG, "Execute <%s> (value=%s) success, but not record.", sql, value);
        return 0;
    }
    else if (SQLITE_ROW == result)
    {
        ast_log(LOG_DEBUG, "Execute <%s> (value=%s) success, and have record.", sql, value);
        return 1;
    }
    else 
    {
        ast_log(LOG_ERROR, "sqlite3_step(0x%08zx) return %d failed!! sql:%s.", (ptrdiff_t)pStmt, result, sql);
        sqlite3_finalize(pStmt); 
        return -1;
    }

}

/*
 * \brief Level is valid dynamic level?
 */
inline int ast_log_level_is_dynamic(int level, int out_log, const char *file, int line)
{

    if (__LOG_DTMF == level)
    {
        return 1;
    }

    if (level < DYNAMIC_LEVEL_START)
    {
        if (out_log)
        {
            ast_log_exception(__FILE__, __LINE__, file, line, level);
            printf("\r\n[%s %d] Invalid 'dynamic level' [%d < %d] !!\n", file, line, level, DYNAMIC_LEVEL_START);
        }
        return 0;
    }

    if (level >= ARRAY_LEN(levels))
    {
        if (out_log)
        {
            ast_log_exception(__FILE__, __LINE__, file, line, level);
            printf("\r\n[%s %d] Invalid 'dynamic level' [%d >= %zu] !!\n", file, line, level, ARRAY_LEN(levels));
        }
        return 0;
    }

    if (NULL == levels[level])
    {
        if (out_log)
        {
            ast_log_exception(__FILE__, __LINE__, file, line, level);
            printf("\r\n[%s %d] Invalid 'dynamic level' [%d] !!\n", file, line, level);
        }
        return 0;
    }

    return 1;
}

void logger_init_file_output(void)
{
    g_log_fileptr = g_log_fileptr;
    /* TODO: not implements */
}

void logger_close_file_output(void)
{
    /* TODO: not implements */
}

void ast_log_file(int level, char *buf, int max_buf_len)
{
    /* TODO: not implements */
}

#endif
