/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Yang Yanfei <yfyang@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_session</depend>
	<support_level>core</support_level>
 ***/

#ifdef GRANDSTREAM_NETWORKS
#define AST_MODULE_LOG "pjsip"
#endif

#include "asterisk.h"

#ifdef GRANDSTREAM_NETWORKS
ASTERISK_FILE_VERSION(__FILE__, "$Revision: 13.1 $");
#endif

#include <pjsip.h>
#include <pjsip_simple.h>
#include <pjsip_ua.h>
#include <pjlib.h>

#include "asterisk/module.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_pubsub.h"
#include "asterisk/res_pjsip_presence_xml.h"
#include "asterisk/res_pjsip_body_generator_types.h"
#include "asterisk/res_pjsip_outbound_publish.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/callerid.h"
#include "asterisk/channel.h"
#include "asterisk/strings.h"


#define  CC_BUFSIZE   256
#define  MAX_STRING_GROWTHS 5
#define  DEFAULT_CC_EXPIRES 3600
#define  DEFAULT_CC_BODY  "application/call-completion"
#define  CC_DATASTORE     "CC datastore"

#define  CC_TYPE         "application"
#define  CC_PIDF_SUBTYPE "pidf+xml"
#define  CC_SUBTYPE      "call-completion"
#define  CC_STATE_STR    "cc-state: "
#define  CC_STATE_LEN    10
#define  CC_URI_STR      "cc-URI: "
#define  CC_URI_LEN      8

static struct ast_sched_context *cc_sched = NULL;
static const pj_str_t callinfo_name = { "Call-Info", 9 };

struct ao2_container *pjsip_monitor_instances;


/*! \brief Structure which contains the information to construct "pidf+xml" body */
struct pjsip_monitor_data {
	char peername[128];
	char is_open;    
	pj_pool_t *pool;  
};

/*! \brief Structure which contains the old publish data */
struct pjsip_cc_publish_data {
	int    valid;
	struct ast_sip_outbound_publish_client *client;
	char   notify_user[AST_MAX_EXTENSION];
};

/*! \brief Structure which contains the sip ccss monitor information */
struct pjsip_monitor_instance {
	int    core_id;
	char   device_name[128];
	struct ast_sip_subscription *sub;
	struct cc_publisher_state  *state;
	char   subscribe_user[AST_MAX_EXTENSION];
	char   notify_user[AST_MAX_EXTENSION];
	char   notify_uri[PJSIP_MAX_URL_SIZE];
	struct pjsip_monitor_data  monitor_data;
	struct pjsip_cc_publish_data  publish_data;
	int    iExpires;
	int    iSchedID;
};


/*! \brief Structure which contains the sip ccss agent information */
struct pjsip_cc_agent_pvt {
	int  offer_timer_id;
	char original_callid[CC_BUFSIZE];   /* A copy of the original call's Call-ID. */
	char original_exten[CC_BUFSIZE];    /* A copy of the exten called originally.*/
	struct ast_sip_subscription *sub;
	char  subscribe_user[AST_UUID_STR_LEN];
	char  notify_user[AST_UUID_STR_LEN];
	pjsip_uri  *notify_uri;             /* When we send a ready NOTIFY, we include a notify_uri in the body */
	pjsip_uri  *subscribe_uri;          /* When we advertise call completion to a caller, provide a subscribe_uri for the caller to use */
	char is_available;
	char is_cc_ready;
	pj_pool_t *pool;  /* Allocation pool */
};


/*! \brief Structure which contains ccss agent data */
struct pjsip_cc_agent_data {
	struct ast_cc_agent *agent;
	struct ast_channel  *chan;
};

/*! \brief Structure which contains the information to construct "call-completion" body */
struct pjsip_cc_notify_data {
	char        is_cc_ready;
	pjsip_uri  *notify_uri;              
};

/*! \brief Structure which contains ccss publisher state information */
struct cc_publisher_state {
	struct ast_sip_outbound_publish_client *client;
	struct pjsip_monitor_instance  *monitor_instance;
};

static int pjsip_cc_stop_publishing(struct ast_sip_outbound_publish_client *client);
static int pjsip_cc_start_publishing(struct pjsip_monitor_instance *monitor_instance,struct ast_sip_outbound_publish_client *client);
static int pjsip_cc_create_subscription(void *obj);

/*! \brief The hash_fn used by the container "pjsip_monitor_instances" */
static int pjsip_monitor_instance_hash_fn(const void *obj, const int flags)
{
	const struct pjsip_monitor_instance *monitor_instance = obj;
	return monitor_instance->core_id;
}

/*! \brief The cmp_fn used by the container "pjsip_monitor_instances" */
static int pjsip_monitor_instance_cmp_fn(void *obj, void *arg, int flags)
{
	struct pjsip_monitor_instance *monitor_instance1 = obj;
	struct pjsip_monitor_instance *monitor_instance2 = arg;

	return monitor_instance1->core_id == monitor_instance2->core_id ? CMP_MATCH | CMP_STOP : 0;
}


static int pjsip_cc_monitor_instance_release(void *obj)
{
	struct pjsip_monitor_instance *monitor_instance = obj;

	if (monitor_instance && monitor_instance->sub) {
		ast_sip_terminated_subscribe(monitor_instance->sub, PJ_FALSE);
		//ao2_cleanup(monitor_instance->sub);
	}

	return 0;
}

/*! \brief The destructor_fn of the struct pjsip_monitor_instance */
static void pjsip_monitor_instance_destructor(void *data)
{
	struct pjsip_monitor_instance *monitor_instance = data;
	struct ast_sip_outbound_publish_client *client;

	if (monitor_instance->state) {
		pjsip_cc_stop_publishing(monitor_instance->state->client); 
		ast_sip_destroy_cc_outbound_publish(monitor_instance->notify_user);
	} else {
		client = ast_sip_cc_publish_client_get(monitor_instance->notify_user);
		if (client) {
			ast_sip_destroy_cc_outbound_publish(monitor_instance->notify_user);
		}
	}

	if ((monitor_instance->iSchedID >= 0) && (0 != monitor_instance->publish_data.valid)) {
		AST_SCHED_DEL(cc_sched, monitor_instance->iSchedID);
		pjsip_cc_stop_publishing(monitor_instance->publish_data.client); 
		ast_sip_destroy_cc_outbound_publish(monitor_instance->publish_data.notify_user);

		monitor_instance->publish_data.valid = 0;
	}

	ast_sip_push_task_synchronous(NULL, pjsip_cc_monitor_instance_release, monitor_instance); 

	return;
}


/*! \brief Alloc and init one object of the struct pjsip_monitor_instance */
static struct pjsip_monitor_instance *pjsip_monitor_instance_init(int core_id, char *subscribe_user, 
																  const char * const peername, const char * const device_name)
{
	struct pjsip_monitor_data     *monitor_data;
	struct pjsip_monitor_instance *monitor_instance = ao2_alloc(sizeof(*monitor_instance), pjsip_monitor_instance_destructor);

	if (!monitor_instance) {
		return NULL;
	}

	monitor_data = &monitor_instance->monitor_data;

	monitor_data->pool = NULL;
	monitor_data->is_open = 0;

	strncpy(monitor_data->peername, peername, sizeof(monitor_data->peername));
	strncpy(monitor_instance->device_name, device_name, sizeof(monitor_instance->device_name));

	monitor_instance->publish_data.valid = 0;
	monitor_instance->publish_data.client = NULL;
	monitor_instance->state = NULL;
	monitor_instance->sub = NULL;
	monitor_instance->core_id = core_id;
	monitor_instance->iExpires = DEFAULT_CC_EXPIRES;
	monitor_instance->iSchedID = -1;

	ast_log(LOG_NOTICE, "pjsip_monitor_instance_init, subscribe_user[%s]", subscribe_user);
	strncpy(monitor_instance->subscribe_user, subscribe_user, sizeof(monitor_instance->subscribe_user));

	ao2_link(pjsip_monitor_instances, monitor_instance);

	return monitor_instance;
}


static int pjsip_find_cc_monitor_instance_by_subscribe_user(void *obj, void *arg, int flags)
{
	struct pjsip_monitor_instance *monitor_instance = obj;
	const  char  *user = arg;

	return !strcmp(monitor_instance->subscribe_user, user) ? CMP_MATCH | CMP_STOP : 0;
}


static int pjsip_find_cc_monitor_instance_by_subscription(void *obj, void *arg, int flags)
{
	struct pjsip_monitor_instance *monitor_instance = obj;

	return monitor_instance->sub == arg ? CMP_MATCH | CMP_STOP : 0;
}

static int pjsip_find_by_subscribe_user_helper(void *obj, void *arg, int flags)
{
	struct ast_cc_agent *agent = obj;
	struct pjsip_cc_agent_pvt *agent_pvt = agent->private_data;
	const char *user = arg;

	return !strcmp(agent_pvt->subscribe_user, user) ? CMP_MATCH | CMP_STOP : 0;
}

static struct ast_cc_agent *pjsip_find_cc_agent_by_subscribe_user(const char * const user)
{
	struct ast_cc_agent *agent = ast_cc_agent_callback(0, pjsip_find_by_subscribe_user_helper, (char *)user, "PJSIP");
	return agent;
}


static int pjsip_find_by_notify_user_helper(void *obj, void *arg, int flags)
{
	struct ast_cc_agent *agent = obj;
	struct pjsip_cc_agent_pvt *agent_pvt = agent->private_data;
	const char *user = arg;

	return !strcmp(agent_pvt->notify_user, user) ? CMP_MATCH | CMP_STOP : 0;
}

static struct ast_cc_agent *pjsip_find_cc_agent_by_notify_user(const char * const user)
{
	struct ast_cc_agent *agent = ast_cc_agent_callback(0, pjsip_find_by_notify_user_helper, (char *)user, "PJSIP");
	return agent;
}

static int pjsip_find_by_callid_helper(void *obj, void *arg, int flags)
{
	struct ast_cc_agent *agent = obj;
	struct pjsip_cc_agent_pvt *agent_pvt = agent->private_data;
	struct ast_sip_session *session = arg;

	if ((NULL == session) || (NULL == session->channel)) {
		return 0;
	}

    return !strcmp(agent_pvt->original_callid, ast_channel_uniqueid(session->channel)) ? CMP_MATCH | CMP_STOP : 0; 
}

static struct ast_cc_agent *pjsip_find_cc_agent_by_original_callid(struct ast_sip_session *session)
{
	struct ast_cc_agent *agent = ast_cc_agent_callback(0, pjsip_find_by_callid_helper, session, "PJSIP");
	return agent;
}

static int pjsip_find_cc_exten_by_notify_user(char *exten, void *data)
{
	struct ast_sip_cc_original_exten_data *exten_data = data;
	struct ast_cc_agent  *agent;
	struct pjsip_cc_agent_pvt *agent_pvt;

	if (!exten || !data) {
		return -1;
	}

	ast_log(LOG_NOTICE, "pjsip_find_cc_exten_by_notify_user, notify_user[%s]", exten);
	agent = pjsip_find_cc_agent_by_notify_user(exten);
	if (!agent) {
		return -1;
	}

	agent_pvt = agent->private_data;

	exten_data->core_id = agent->core_id;
	strncpy(exten_data->original_exten, agent_pvt->original_exten, sizeof(exten_data->original_exten)-1);
	strncpy(exten_data->device_name, agent->device_name, sizeof(exten_data->device_name)-1);

	ast_log(LOG_NOTICE, "Get core_id[%d], original_exten[%s], device_name[%s]", agent->core_id, agent_pvt->original_exten, agent->device_name);
	ao2_ref(agent, -1);

	return 0;
}

struct ast_sip_exten_supplement cc_original_exten = {
	.exten_type = AST_SIP_CCSS_ORIGINAL_EXTEN,
	.find_exten = pjsip_find_cc_exten_by_notify_user,
};


static int pjsip_find_cc_notify_user_by_monitor_instance(char *exten, void *data)
{
	struct ast_sip_cc_notify_user_data  *user_data = data;
	struct pjsip_monitor_instance *monitor_instance;

    if (!user_data || !user_data->monitor) {
		return -1;
    }

	monitor_instance = user_data->monitor->private_data;
	strncpy(user_data->notify_user, monitor_instance->notify_user, sizeof(user_data->notify_user)-1);

	ast_log(LOG_NOTICE, "Get notify_user[%s]", monitor_instance->notify_user);
    return 0; 
}

struct ast_sip_exten_supplement cc_notify_user = {
	.exten_type = AST_SIP_CCSS_NOTIFY_USER,
	.find_exten = pjsip_find_cc_notify_user_by_monitor_instance,
};

static struct ast_datastore_info  cc_ds_info = { };

static int pjsip_cc_add_datastore(struct pjsip_cc_agent_pvt  *agent_pvt)
{
	RAII_VAR(struct ast_datastore *, cc_datastore, NULL, ao2_cleanup);

	cc_datastore = ast_sip_subscription_alloc_datastore(&cc_ds_info, CC_DATASTORE);
	if (!cc_datastore) {
		return -1;
	}
	cc_datastore->data = agent_pvt;

	ast_sip_subscription_add_datastore(agent_pvt->sub, cc_datastore);
	return 0;
}

static int pjsip_cc_new_subscribe(struct ast_sip_endpoint *endpoint, const char *resource)
{
	struct ast_cc_agent *agent;

	agent = pjsip_find_cc_agent_by_subscribe_user(resource);
    if (!agent) {
		ast_log(LOG_WARNING, "can't find cc agent by subscribe uuid[%s]\n", resource);
		return 404;
	}

	ao2_ref(agent, -1);
	return 200;
}

static int pjsip_cc_subscription_established(struct ast_sip_subscription *sip_sub)
{
	const char *resource = ast_sip_subscription_get_resource_name(sip_sub);
	struct pjsip_cc_agent_pvt  *agent_pvt;
	struct ast_cc_agent        *agent;

	agent = pjsip_find_cc_agent_by_subscribe_user(resource);
	if (!agent) {
		ast_log(LOG_WARNING, "can't find cc agent by subscribe uuid[%s]\n", resource);
		return -1;
	}

	agent_pvt = agent->private_data;
	agent_pvt->sub = ao2_bump(sip_sub);

	if (pjsip_cc_add_datastore(agent_pvt)) {
		ast_log(LOG_WARNING, "cc_add_datastore failed, resource[%s]\n", resource);
		ao2_ref(agent, -1);
		return -1;
	}

	ast_cc_agent_accept_request(agent->core_id, "SIP caller %s has requested CC via SUBSCRIBE", agent->device_name); 

	ao2_ref(agent, -1);
	return 0;
}


static int pjsip_cc_publish_clear_timeout(const void *data)
{
	struct pjsip_monitor_instance *monitor_instance = (struct pjsip_monitor_instance  *)data;

	if (NULL == monitor_instance || (ao2_ref(monitor_instance, 0) <= 0) || (0 == monitor_instance->publish_data.valid)) {
		return 0;
	}

	monitor_instance->iSchedID = -1;

	ast_log(LOG_NOTICE, "pjsip_cc_publish_clear_timeout, stop publish[%s]\n", monitor_instance->publish_data.notify_user);

	pjsip_cc_stop_publishing(monitor_instance->publish_data.client); 
	ast_sip_destroy_cc_outbound_publish(monitor_instance->publish_data.notify_user);

	monitor_instance->publish_data.valid = 0;

	return 0;
}

static void pjsip_cc_subscriber_state_change(struct ast_sip_subscription *sub, pjsip_msg_body *body, enum pjsip_evsub_state state)
{
	struct pjsip_monitor_instance *monitor_instance = ao2_callback(pjsip_monitor_instances, 0, 
																   pjsip_find_cc_monitor_instance_by_subscription, sub);
	char  *status;
	char  *cc_uri;
	char  *eol;
	char  *user;
	char  *user_end;

	if (!monitor_instance) {
		ast_log(LOG_WARNING, "can't find monitor_instance by subscription, resource[%s]\n", ast_sip_subscription_get_resource_name(sub));
		return;
	}

	ast_log(LOG_NOTICE, "pjsip_cc_subscriber_state_change start process\n");

	/* check content type*/
    if (0 != pj_stricmp2(&body->content_type.type, CC_TYPE) || 0 != pj_stricmp2(&body->content_type.subtype, CC_SUBTYPE)) {
		ast_log(LOG_WARNING, "the content type is [%s/%s]\n", body->content_type.type.ptr, body->content_type.subtype.ptr);
		ao2_ref(monitor_instance, -1); 
		return;
	}

	/* get body data */
	status = strstr((const char*)body->data, CC_STATE_STR);
	if (!status) {
		ast_log(LOG_WARNING, "no cc-state in the body[%s]\n", (char *)body->data);
		ao2_ref(monitor_instance, -1); 
		return;
	}

	/* if the cc-state is "queued", it means the callee is not available */
	status = status + CC_STATE_LEN;
	if (!strncmp(status, "queued", 6)) {
		ast_cc_monitor_request_acked(monitor_instance->core_id, "endpoint %s accepted request", monitor_instance->device_name);
		ao2_ref(monitor_instance, -1); 
		return;
	}

	/* if the cc-state is "ready", get the cc-URI info */
	cc_uri = strstr((const char*)body->data, CC_URI_STR);
	if (!cc_uri) {
		ast_log(LOG_WARNING, "no cc-URI in the body[%s]\n", (char *)body->data);
		ao2_ref(monitor_instance, -1); 
		return;
	}

	cc_uri = cc_uri + CC_URI_LEN;
    eol = strstr(cc_uri, "\r\n");
    if (eol) {
		*eol = '\0';
    }
	
	/* get notify user */
	strncpy(monitor_instance->notify_uri, cc_uri, sizeof(monitor_instance->notify_uri)-1);
	user = strchr(cc_uri, ':');
    if (!user) {
		ao2_ref(monitor_instance, -1); 
		return;
    }

	user = user+1;
	user_end = strchr(cc_uri, '@');
	if (user_end) {
		*user_end = '\0';
	}

	/* start process outbound publish */
	if (!monitor_instance->state) {
		strncpy(monitor_instance->notify_user, user, sizeof(monitor_instance->notify_user)-1);

		ast_log(LOG_NOTICE, "Create cc outbound publish, notify_user[%s]\n", monitor_instance->notify_user);
		if (ast_sip_create_cc_outbound_publish(monitor_instance->notify_user, CC_SUBTYPE, monitor_instance->notify_uri)) {
			ast_log(LOG_WARNING, "Unable create sip outbound_publish\n");

			ao2_ref(monitor_instance, -1); 
			return;
		}

		ast_cc_monitor_callee_available(monitor_instance->core_id, "PJSIP monitored callee has become available");
	} else {
		/* check if there is an old publish data need to release */
		if (monitor_instance->publish_data.valid && monitor_instance->publish_data.client) {
			monitor_instance->publish_data.valid = 0;
			pjsip_cc_stop_publishing(monitor_instance->publish_data.client); 
			ast_sip_destroy_cc_outbound_publish(monitor_instance->publish_data.notify_user);
		}

		/* record the current publish data */
		monitor_instance->publish_data.client = monitor_instance->state->client; 
		strncpy(monitor_instance->publish_data.notify_user, monitor_instance->notify_user, AST_MAX_EXTENSION-1);

		ast_log(LOG_NOTICE, "the monitor instance has state, prepare to stop old publish[%s]\n", monitor_instance->notify_user);
		monitor_instance->iSchedID = ast_sched_replace(monitor_instance->iSchedID, cc_sched, 1000, pjsip_cc_publish_clear_timeout, monitor_instance);
		if (monitor_instance->iSchedID < 0) {
			pjsip_cc_stop_publishing(monitor_instance->state->client); 
			ast_sip_destroy_cc_outbound_publish(monitor_instance->notify_user);
		} else {
			monitor_instance->publish_data.valid = 1;
		}

		strncpy(monitor_instance->notify_user, user, sizeof(monitor_instance->notify_user)-1);
		ast_log(LOG_NOTICE, "Start new publish[%s]\n", monitor_instance->notify_user);
		if (ast_sip_create_cc_outbound_publish(monitor_instance->notify_user, CC_SUBTYPE, monitor_instance->notify_uri)) {
			ast_log(LOG_WARNING, "Unable create sip outbound_publish\n");
			ao2_ref(monitor_instance, -1); 
			return;
		}

		ast_cc_monitor_callee_available(monitor_instance->core_id, "PJSIP monitored callee has become available");
	}

    ao2_ref(monitor_instance, -1); 
	return;
}

static void pjsip_cc_subscription_destroy(struct ast_sip_subscription *sub)
{
	const char *resource = ast_sip_subscription_get_resource_name(sub);
	struct ast_cc_agent  *agent;

	agent = pjsip_find_cc_agent_by_subscribe_user(resource);
	if (!agent) {
		ast_log(LOG_WARNING, "can't find cc agent by subscribe uuid[%s]\n", resource);
		return;
	}

	ast_cc_failed(agent->core_id, "core[%d], pjsip_cc_subscription_destroy", agent->core_id);
	ao2_ref(agent, -1);

	return;
}


/*! \brief Send a subscription or resubscription for dialog */
static int pjsip_cc_subscribe_check_timeout(const void *data)
{
	struct pjsip_monitor_instance *monitor_instance = (struct pjsip_monitor_instance  *)data;

	if (NULL == monitor_instance) {
		return 0;
	}

	monitor_instance->iSchedID = -1;
	ao2_ref(monitor_instance, -1); 

	ast_log(LOG_NOTICE, "pjsip_cc_subscribe_check_timeout, core[%d] try to send subscribe\n", monitor_instance->core_id);

	if (ast_sip_push_task_synchronous(NULL, pjsip_cc_create_subscription, monitor_instance)) {
		return -1;
	}

	return 0;
}


static void pjsip_cc_subscriber_destroy(struct ast_sip_subscription *sub)
{
	struct pjsip_monitor_instance *monitor_instance = ao2_callback(pjsip_monitor_instances, 0, 
																   pjsip_find_cc_monitor_instance_by_subscription, sub);
	if (!monitor_instance) {
		ast_log(LOG_WARNING, "can't find monitor_instance by subscription, resource[%s]\n", ast_sip_subscription_get_resource_name(sub));
		return;
	}

	monitor_instance->sub = NULL;
	monitor_instance->iExpires = ast_sip_get_sub_expires(sub);

	if (-1 != monitor_instance->iExpires) {
		/* wait 1 s and try again*/
		monitor_instance->iSchedID = ast_sched_replace(monitor_instance->iSchedID, cc_sched, 1000, pjsip_cc_subscribe_check_timeout, monitor_instance);
		if (monitor_instance->iSchedID >= 0) {
			ast_log(LOG_NOTICE, "pjsip_cc_subscriber_destroy, core[%d], iExpires[%d], iSchedID[%d]", 
					monitor_instance->core_id, monitor_instance->iExpires, monitor_instance->iSchedID);
			return;
		}
	}

	ast_cc_failed(monitor_instance->core_id, "core[%d], pjsip_cc_subscriber_destroy", monitor_instance->core_id);
	ao2_ref(monitor_instance, -1); 
	return;
}

static void *pjsip_cc_get_notify_data(struct ast_sip_subscription *sub)
{
	struct pjsip_cc_agent_pvt   *agent_pvt;
	struct pjsip_cc_notify_data *cc_notify;
	struct ast_datastore        *cc_datastore;

	cc_datastore = ast_sip_subscription_get_datastore(sub, CC_DATASTORE);
	if (!cc_datastore) {
		return NULL;
	}

	agent_pvt = (struct pjsip_cc_agent_pvt *)cc_datastore->data;

	cc_notify = ao2_alloc(sizeof(*cc_notify), NULL);
	if (!cc_notify) {
		ao2_cleanup(cc_notify);
		return NULL;
	}

	cc_notify->is_cc_ready = agent_pvt->is_cc_ready;
	if (cc_notify->is_cc_ready) {
		cc_notify->notify_uri = pjsip_uri_get_uri(agent_pvt->notify_uri);
	}
	ao2_cleanup(cc_datastore);
	return cc_notify;
}

static struct ast_sip_notifier cc_notifier = {
	.default_accept = DEFAULT_CC_BODY,
	.new_subscribe = pjsip_cc_new_subscribe,
	.subscription_established = pjsip_cc_subscription_established,
	.subscription_destroy = pjsip_cc_subscription_destroy,
	.get_notify_data = pjsip_cc_get_notify_data,
};


static struct ast_sip_subscriber cc_subscriber = {
	.state_change = pjsip_cc_subscriber_state_change,
	.subscriber_destroy = pjsip_cc_subscriber_destroy,
};

static struct ast_sip_subscription_handler cc_handler = {
	.event_name = "call-completion",
	.body_type = AST_SIP_CALL_COMPLETION,
	.accept = { DEFAULT_CC_BODY, },
	.subscriber = &cc_subscriber,
	.notifier = &cc_notifier,
};

static int pjsip_cc_transmit_notify(void *userdata)
{
	struct pjsip_cc_notify_data  cc_data;
	struct pjsip_cc_agent_pvt *agent_pvt = userdata;

	struct ast_sip_body_data data = {
		.body_type = AST_SIP_CALL_COMPLETION,
		.body_data = &cc_data,
	};

	cc_data.is_cc_ready = agent_pvt->is_cc_ready;
    if (agent_pvt->is_cc_ready) {
		cc_data.notify_uri = (pjsip_uri *)pjsip_uri_clone(agent_pvt->pool, agent_pvt->notify_uri); 
    }

	ast_sip_subscription_notify(agent_pvt->sub, &data, 0);
	ao2_ref(agent_pvt, -1);

	return 0;
}

static int pjsip_cc_agent_release_pool(void *obj)
{
	struct pjsip_cc_agent_pvt *agent_pvt = obj;

	if (agent_pvt) {
		if (agent_pvt->sub) {
			ao2_cleanup(agent_pvt->sub);
		}

		pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), agent_pvt->pool); 
		agent_pvt->pool = NULL;
	}

	return 0;
}

static void pjsip_cc_agent_destroy(void *obj)
{
	struct pjsip_cc_agent_pvt *agent_pvt = obj;

	if (!agent_pvt) {
		return;
	}

	if (agent_pvt->pool) {
		ast_sip_push_task_synchronous(NULL, pjsip_cc_agent_release_pool, agent_pvt);
	}
}

static int pjsip_cc_agent_create(void *data)
{
	struct pjsip_cc_agent_data  *agent_data = data;
	struct pjsip_cc_agent_pvt   *agent_pvt = NULL;
	struct ast_sip_channel_pvt  *call_pvt  = NULL;
	struct ast_sip_session      *session;

	call_pvt = ast_channel_tech_pvt(agent_data->chan);
	if (!call_pvt || (NULL == call_pvt->session)) {
		return -1;
	}

	agent_pvt = ao2_alloc(sizeof(*agent_pvt), pjsip_cc_agent_destroy);
	if (!agent_pvt) {
		return -1;
	}

	session = call_pvt->session;
	agent_pvt->pool = pjsip_endpt_create_pool(ast_sip_get_pjsip_endpoint(), "cc_agent", 1024, 1024);
	if (!agent_pvt->pool) {
		ao2_cleanup(agent_pvt);
		return -1;
	}

    ast_copy_string(agent_pvt->original_callid, ast_channel_uniqueid(agent_data->chan), sizeof(agent_pvt->original_callid)); 
	ast_copy_string(agent_pvt->original_exten, call_pvt->session->exten, sizeof(agent_pvt->original_exten));

	if (session->cc_subs_user.slen && session->cc_subs_uri) {
		ast_copy_string(agent_pvt->subscribe_user, session->cc_subs_user.ptr, sizeof(agent_pvt->subscribe_user));
		agent_pvt->subscribe_uri = pjsip_uri_clone(agent_pvt->pool, session->cc_subs_uri); 
	}

	ast_log(LOG_NOTICE, "pjsip_cc_agent_create, original_callid[%s], original_exten[%s]", 
			agent_pvt->original_callid, agent_pvt->original_exten);

	agent_pvt->offer_timer_id = -1;
	agent_data->agent->private_data = agent_pvt;
	call_pvt->session->cc_offer = 1;

	return 0;
}

/* when the call is wait_for_answer, create cc agent */
static int pjsip_cc_agent_init(struct ast_cc_agent *agent, struct ast_channel *chan)
{
	struct pjsip_cc_agent_data  agent_data;

	ast_log(LOG_NOTICE, "pjsip_cc_agent_init, core_id[%d], chan[%s] ", agent->core_id, ast_channel_name(chan));

	agent_data.agent = agent;
	agent_data.chan  = chan;
	if (ast_sip_push_task_synchronous(NULL, pjsip_cc_agent_create, &agent_data)) {
		return -1;
	}

	return 0;
}

/* if the caller not request the cc service and the offer timer expired, then will destroy the cc core */
static int pjsip_cc_offer_timer_expire(const void *data)
{
	struct ast_cc_agent *agent = (struct ast_cc_agent *) data;
	struct pjsip_cc_agent_pvt *agent_pvt = agent->private_data;

	agent_pvt->offer_timer_id = -1;

	return ast_cc_failed(agent->core_id, "PJSIP agent %s's offer timer expired", agent->device_name);
}

/* when hangup the call, the core state is cc_offered, then start the offer timer */
static int pjsip_cc_agent_start_offer_timer(struct ast_cc_agent *agent)
{
	struct pjsip_cc_agent_pvt *agent_pvt = agent->private_data;
	int when;

	when = ast_get_cc_offer_timer(agent->cc_params) * 1000;
	agent_pvt->offer_timer_id = ast_sched_add(cc_sched, when, pjsip_cc_offer_timer_expire, agent);
	return 0;
}

/* if the caller request the cc service succeed, then stop the offer timer*/
static int pjsip_cc_agent_stop_offer_timer(struct ast_cc_agent *agent)
{
	struct pjsip_cc_agent_pvt *agent_pvt = agent->private_data;

	AST_SCHED_DEL(cc_sched, agent_pvt->offer_timer_id);
	return 0;
}

/* send response to the cc request. if succeed, send 200 OK, then send Notify; if failed, send 500 */
static void pjsip_cc_agent_respond(struct ast_cc_agent *agent, enum ast_cc_agent_response_reason reason)
{
	struct pjsip_cc_agent_pvt *agent_pvt = agent->private_data;

	if (reason == AST_CC_AGENT_RESPONSE_SUCCESS || agent_pvt->notify_uri) {
		ast_log(LOG_NOTICE, "pjsip_cc_agent_respond, success");
	} else {
		ast_log(LOG_NOTICE, "pjsip_cc_agent_respond, failed, reason[%d]", reason);
	}

	agent_pvt->is_available = 1;
}

/*get cc agent's status*/
static int pjsip_cc_agent_status_request(struct ast_cc_agent *agent)
{
	struct pjsip_cc_agent_pvt *agent_pvt = agent->private_data;
	enum ast_device_state state = agent_pvt->is_available ? AST_DEVICE_NOT_INUSE : AST_DEVICE_INUSE;

	return ast_cc_agent_status_response(agent->core_id, state);
}

/* when callee is available, the caller is busy, just wait */
static int pjsip_cc_agent_start_monitoring(struct ast_cc_agent *agent)
{
	/* To start monitoring just means to wait for an incoming PUBLISH
	 * to tell us that the caller has become available again. No special
	 * action is needed
	 */
	return 0;
}


/* the callee become available, notify the caller */
static int pjsip_cc_agent_recall(struct ast_cc_agent *agent)
{
	struct pjsip_cc_agent_pvt *agent_pvt = agent->private_data;
	pjsip_sip_uri    *uri;
	char uuid_str[AST_UUID_STR_LEN] = {0};
	char *user = NULL;
	char  cc_uri[PJSIP_MAX_URL_SIZE] = {0};

	ast_log(LOG_NOTICE, "pjsip_cc_agent_recall, is_available[%d]", agent_pvt->is_available);

	/* If we have received a PUBLISH beforehand stating that the caller in question
	 * is not available, we can save ourself a bit of effort here and just report
	 * the caller as busy*/
	if (!agent_pvt->is_available) {
		return ast_cc_agent_caller_busy(agent->core_id, "Caller %s is busy, reporting to the core",
				agent->device_name);
	}

	/* Otherwise, we transmit a NOTIFY to the caller and await either a PUBLISH or an INVITE */
	agent_pvt->is_cc_ready = 1;

	/* generate notify uri */
	user = ast_uuid_generate_str(uuid_str, sizeof(uuid_str));
	strncpy(agent_pvt->notify_user, uuid_str, sizeof(agent_pvt->notify_user));

	if (!agent_pvt->notify_uri) {
		agent_pvt->notify_uri = pjsip_uri_clone(agent_pvt->pool, agent_pvt->subscribe_uri);
		if (!agent_pvt->notify_uri) {
			ast_log(LOG_WARNING, "pjsip_cc_agent_recall, create notify uri failed");
			return 0;
		}
	}

	uri = pjsip_uri_get_uri(agent_pvt->notify_uri);
    pj_strdup2(agent_pvt->pool, &uri->user, user); 

	pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, agent_pvt->notify_uri, cc_uri, sizeof(cc_uri)); 
	ast_log(LOG_NOTICE, "pjsip_cc_agent_recall, notify_uri[%s]\n", cc_uri);

	ao2_ref(agent_pvt, +1);
	ast_sip_push_task(ast_sip_subscription_get_serializer(agent_pvt->sub), pjsip_cc_transmit_notify, agent_pvt);

	return 0;
}

/* destroy the cc agent */
static void pjsip_cc_agent_destructor(struct ast_cc_agent *agent)
{
	struct pjsip_cc_agent_pvt *agent_pvt = agent->private_data;

	if (!agent_pvt) {
		/* The agent constructor probably failed. */
		return;
	}

	pjsip_cc_agent_stop_offer_timer(agent);

	ao2_cleanup(agent_pvt);
}


static struct ast_cc_agent_callbacks cc_agent_callbacks = {
	.type = "PJSIP",
	.init = pjsip_cc_agent_init,
	.start_offer_timer = pjsip_cc_agent_start_offer_timer,
	.stop_offer_timer = pjsip_cc_agent_stop_offer_timer,
	.respond = pjsip_cc_agent_respond,
	.status_request = pjsip_cc_agent_status_request,
	.start_monitoring = pjsip_cc_agent_start_monitoring,
	.callee_available = pjsip_cc_agent_recall,
	.destructor = pjsip_cc_agent_destructor,
};

static int pjsip_cc_create_subscription(void *obj)
{
	struct pjsip_monitor_instance *monitor_instance = obj;
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);

	if (!monitor_instance) {
		return -1;
	}

	ast_log(LOG_NOTICE, "pjsip_cc_create_subscription, endpoint[%s], resource[%s]", monitor_instance->monitor_data.peername, monitor_instance->subscribe_user);
	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", monitor_instance->monitor_data.peername);

	monitor_instance->sub = ast_sip_create_subscription(&cc_handler, endpoint, monitor_instance->subscribe_user, monitor_instance->iExpires);
	if (!monitor_instance->sub) {
		ast_log(LOG_WARNING, "create subscription failed!");
		return -1;
	}

	return 0;
}

/* to subscribe the callee's status */
static int pjsip_cc_monitor_request_cc(struct ast_cc_monitor *monitor, int *available_timer_id)
{
	struct pjsip_monitor_instance *monitor_instance = monitor->private_data;
	enum   ast_cc_service_type service = monitor->service_offered;
	int    when;

	if (!monitor_instance) {
		return -1;
	}

	ast_log(LOG_NOTICE, "pjsip_cc_monitor_request_cc");

	if (ast_sip_push_task_synchronous(NULL, pjsip_cc_create_subscription, monitor_instance)) {
		return -1;
	}

    if (service == AST_CC_CCBS) {
		when = ast_get_ccbs_available_timer(monitor->interface->config_params);
    } else {
		when = ast_get_ccnr_available_timer(monitor->interface->config_params); 
	}

	ao2_t_ref(monitor, +1, "Adding a ref to the monitor for the scheduler");
	*available_timer_id = ast_sched_add(cc_sched, when * 1000, ast_cc_available_timer_expire, monitor);

    return 0; 
}


static void *pjsip_cc_pidf_allocate_body(void *data)
{
	struct pjsip_monitor_data *monitor_data = data;

	return ast_sip_presence_xml_create_node(monitor_data->pool, NULL, "presence");
}

/*
<?xml version="1.0" encoding="UTF-8"?>
<presence xmlns="urn:ietf:params:xml:ns:pidf" entity="trunk_3">
<tuple id="29135-554e966c4b3dd44512685ed56">
<status><basic>closed</basic></status>
</tuple>
</presence>
*/
static int pjsip_cc_pidf_generate_body_content(void *body, void *data)
{
	pj_xml_node *presence_info = body, *tuple, *status, *basic;
	struct pjsip_monitor_data *monitor_data = data;
	char tuple_id[32];
	char state[8];

	if (!body || !data) {
		return -1;
	}

    if (monitor_data->is_open) {
		snprintf(state, sizeof(state), "%s", "open"); 
    } else {
		snprintf(state, sizeof(state), "%s", "closed"); 
	}

	ast_generate_random_string(tuple_id, sizeof(tuple_id));	

	ast_sip_presence_xml_create_attr(monitor_data->pool, presence_info, "xmlns", "urn:ietf:params:xml:ns:pidf");
	ast_sip_presence_xml_create_attr(monitor_data->pool, presence_info, "entity", monitor_data->peername);

	tuple = ast_sip_presence_xml_create_node(monitor_data->pool, presence_info, "tuple");
	ast_sip_presence_xml_create_attr(monitor_data->pool, tuple, "id", tuple_id);

	status = ast_sip_presence_xml_create_node(monitor_data->pool, tuple, "status");
	basic  = ast_sip_presence_xml_create_node(monitor_data->pool, status, "basic");
	pj_strdup2(monitor_data->pool, &basic->content, state);

	return 0;
}


static void pjsip_cc_pidf_to_string(void *body, struct ast_str **str)
{
	pjpidf_pres *cc_pub = body;
	int growths = 0;
	int size;

	do {
		size = pjpidf_print(cc_pub, ast_str_buffer(*str), ast_str_size(*str));
		if (size == AST_PJSIP_XML_PROLOG_LEN) {
			ast_str_make_space(str, ast_str_size(*str) * 2);
			++growths;
		}
	} while (size == AST_PJSIP_XML_PROLOG_LEN && growths < MAX_STRING_GROWTHS);

	if (size == AST_PJSIP_XML_PROLOG_LEN) {
		ast_log(LOG_WARNING, "PIDF body text too large\n");
		return;
	}

	*(ast_str_buffer(*str) + size) = '\0';
	ast_str_update(*str);
}


static int pjsip_cc_transmit_publish(void *data)
{
	struct pjsip_monitor_instance *monitor_instance = data;
	struct ast_sip_body body;

	struct ast_str *body_text;
	void  *body_buf;

    if (!monitor_instance) {
		return -1;
	}

    monitor_instance->monitor_data.pool = pjsip_endpt_create_pool(ast_sip_get_pjsip_endpoint(), "public_state", 1024, 1024); 
	if (!monitor_instance->monitor_data.pool) {
		return -1;
	}

	body.type = CC_TYPE;
	body.subtype = CC_PIDF_SUBTYPE;

	body_buf = pjsip_cc_pidf_allocate_body(&monitor_instance->monitor_data);
	if (!body_buf) {
		ast_log(LOG_WARNING, "Unable to allocate body\n");
		return -1;
	}

	if (pjsip_cc_pidf_generate_body_content(body_buf, &monitor_instance->monitor_data)) {
		ast_log(LOG_WARNING, "Unable to allocate body\n");
		return -1;
	}

	body_text = ast_str_create(CC_BUFSIZE);
	if (!body_text) {
		return -1;
	}

	pjsip_cc_pidf_to_string(body_buf, &body_text);
	body.body_text = ast_str_buffer(body_text);

	ast_log(LOG_NOTICE, "pjsip_cc_transmit_publish\n");
	ast_sip_publish_client_send(monitor_instance->state->client, &body);

	ast_free(body_text);

	pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), monitor_instance->monitor_data.pool);
	monitor_instance->monitor_data.pool = NULL;

	return 0;
}

/* when callee available, if the caller is busy, send PUBLISH to suspend the monitor */
static int pjsip_cc_monitor_suspend(struct ast_cc_monitor *monitor)
{
	struct pjsip_monitor_instance *monitor_instance = monitor->private_data;
	struct ast_sip_outbound_publish_client *client;

	if (NULL == monitor_instance) {
		return -1;
	}

	if (!monitor_instance->state) {
		ast_log(LOG_NOTICE, "pjsip_cc_monitor_suspend, no state\n");

		client = ast_sip_cc_publish_client_get(monitor_instance->notify_user);
		if (!client) {
			ast_log(LOG_ERROR, "ast_sip_publish_client_get failed, notify_user[%s]\n",monitor_instance->notify_user);
			return -1;
		}

		if (pjsip_cc_start_publishing(monitor_instance, client)) {
			ast_log(LOG_ERROR, "pjsip_cc_start_publishing failed, notify_user[%s]\n",monitor_instance->notify_user);
			ao2_ref(client, -1);
			return -1;
		}

		ao2_ref(client, -1);
	} 

	monitor_instance->monitor_data.is_open = 0;

	ast_sip_push_task(NULL,pjsip_cc_transmit_publish, (void *)monitor_instance);

	return 0;
}

/* caller become available, send PUBLISH to unsuspend the monitor */
static int pjsip_cc_monitor_unsuspend(struct ast_cc_monitor *monitor)
{
	struct pjsip_monitor_instance *monitor_instance = monitor->private_data;

	if (NULL == monitor_instance) {
		return -1;
	}

	if (!monitor_instance->state) {
		ast_log(LOG_WARNING, "pjsip_cc_monitor_unsuspend, no state\n");
		return -1;
	} 

	monitor_instance->monitor_data.is_open = 1;
	ast_sip_push_task(NULL, pjsip_cc_transmit_publish, (void *)monitor_instance);

	return 0;
}

static int pjsip_cc_monitor_cancel_available_timer(struct ast_cc_monitor *monitor, int *sched_id)
{
	if (*sched_id != -1) {
		AST_SCHED_DEL(cc_sched, *sched_id);
		ao2_t_ref(monitor, -1, "Removing scheduler's reference to the monitor");
	}
	return 0;
}

static void pjsip_cc_monitor_destructor(void *private_data)
{
	struct pjsip_monitor_instance *monitor_instance = private_data;

	ao2_unlink(pjsip_monitor_instances, monitor_instance);
	ast_module_unref(ast_module_info->self);
}


static struct ast_cc_monitor_callbacks cc_monitor_callbacks = {
	.type = "PJSIP",
	.request_cc = pjsip_cc_monitor_request_cc,
	.suspend = pjsip_cc_monitor_suspend,
	.unsuspend = pjsip_cc_monitor_unsuspend,
	.cancel_available_timer = pjsip_cc_monitor_cancel_available_timer,
	.destructor = pjsip_cc_monitor_destructor,
};



/*! \brief Destroy callback for ccss publisher state information from datastore */
static void pjsip_cc_publisher_state_destroy(void *obj)
{
	struct cc_publisher_state *publisher_state = obj;
	struct pjsip_monitor_instance *monitor_instance = publisher_state->monitor_instance;

    if (!monitor_instance || !monitor_instance->state) {
		return;
    }

	ao2_cleanup(publisher_state->client);
	ast_free(monitor_instance->state);
	monitor_instance->state = NULL;
    return; 
}


/*! \brief Datastore for attaching ccss publisher state information */
static const struct ast_datastore_info  cc_publisher_state_datastore = {
	.type = "ccss-publisher",
	.destroy = pjsip_cc_publisher_state_destroy,
};


static int pjsip_cc_publication_new(struct ast_sip_endpoint *endpoint, const char *resource, const char *event_configuration)
{
	struct ast_cc_agent *agent;

	ast_log(LOG_NOTICE, "pjsip_cc_publication_new, resource[%s]\n", resource);
	agent = pjsip_find_cc_agent_by_notify_user(resource);
    if (!agent) {
		ast_log(LOG_WARNING, "can't find cc agent by notify uuid[%s]\n", resource);
		return 404;
	}

	ao2_ref(agent, -1);
	return 200;
}


static int pjsip_cc_publication_state_change(struct ast_sip_publication *pub, pjsip_msg_body *body, enum ast_sip_publish_state state)
{
	struct ast_xml_doc *pidf_doc;
	struct ast_cc_agent *agent;
	struct pjsip_cc_agent_pvt *agent_pvt;
	struct ast_xml_node *presence_node;
	struct ast_xml_node *presence_children;
	struct ast_xml_node *tuple_node;
	struct ast_xml_node *tuple_children;
	struct ast_xml_node *status_node;
	struct ast_xml_node *status_children;
	struct ast_xml_node *basic_node;
	const char *basic_status = NULL;
	int res = 0;

    if (!pub || !body) {
		return -1;
    }

    if (0 != pj_stricmp2(&body->content_type.type, CC_TYPE) || 0 != pj_stricmp2(&body->content_type.subtype, CC_PIDF_SUBTYPE)) {
		ast_log(LOG_WARNING, "the content type is [%s/%s]\n", body->content_type.type.ptr, body->content_type.subtype.ptr);
		return -1;
	}

    agent = pjsip_find_cc_agent_by_notify_user(ast_sip_publication_get_resource(pub)); 
    if (!agent) {
		ast_log(LOG_WARNING, "can't find cc agent by notify uuid[%s]\n", ast_sip_publication_get_resource(pub));
		return -1;
	}

	agent_pvt = agent->private_data;
    
	if (!(pidf_doc = ast_xml_read_memory((char *)body->data, body->len))) {
		ast_log(LOG_WARNING, "Unable to open XML PIDF document. Is it malformed?\n");
		ao2_ref(agent, -1);
		return -1;
	}

	presence_node = ast_xml_get_root(pidf_doc);
	if (!(presence_children = ast_xml_node_get_children(presence_node))) {
		ast_log(LOG_WARNING, "No tuples within presence element.\n");
		res = -1;
		goto cc_publish_cleanup;
	}

	if (!(tuple_node = ast_xml_find_element(presence_children, "tuple", NULL, NULL))) {
		ast_log(LOG_NOTICE, "Couldn't find tuple node?\n");
		res = -1;
		goto cc_publish_cleanup;
	}

	tuple_children = ast_xml_node_get_children(tuple_node);
 	status_node = ast_xml_find_element(tuple_children, "status", NULL, NULL);

	if (!(status_children = ast_xml_node_get_children(status_node))) {
		ast_log(LOG_WARNING, "No basic elements within status element.\n");
		res = -1;
		goto cc_publish_cleanup;
	}

	if (!(basic_node = ast_xml_find_element(status_children, "basic", NULL, NULL))) {
		ast_log(LOG_WARNING, "Couldn't find basic node?\n");
		res = -1;
		goto cc_publish_cleanup;
	}

	basic_status = ast_xml_get_text(basic_node);

	if (ast_strlen_zero(basic_status)) {
		ast_log(LOG_NOTICE, "NOthing in basic node?\n");
		res = -1;
		goto cc_publish_cleanup;
	}

	if (!strcmp(basic_status, "open")) {
		agent_pvt->is_available = 1;
		ast_cc_agent_caller_available(agent->core_id, "Received PUBLISH stating PJSIP caller %s is available",agent->device_name);
	} else if (!strcmp(basic_status, "closed")) {
		agent_pvt->is_available = 0;
		ast_cc_agent_caller_busy(agent->core_id, "Received PUBLISH stating PJSIP caller %s is busy",agent->device_name);
	} else {
		ast_log(LOG_NOTICE, "Invalid content in basic element: %s\n", basic_status);
	}

cc_publish_cleanup:
	if (basic_status) {
		ast_xml_free_text(basic_status);
	}
	if (pidf_doc) {
		ast_xml_close(pidf_doc);
	}
	ao2_ref(agent, -1);

	return res;
}

struct ast_sip_publish_handler cc_publish_handler = {
	.event_name = "call-completion",
	.new_publication = pjsip_cc_publication_new,
	.publication_state_change = pjsip_cc_publication_state_change,
};


static int pjsip_cc_start_publishing(struct pjsip_monitor_instance *monitor_instance,
	struct ast_sip_outbound_publish_client *client)
{
	RAII_VAR(struct ast_datastore *, datastore, NULL, ao2_cleanup);
	struct cc_publisher_state *publisher_state;

    if (!monitor_instance || !client) {
		return -1;
    }

	datastore = ast_sip_publish_client_alloc_datastore(&cc_publisher_state_datastore, "ccss-publisher");
	if (!datastore) {
		return -1;
	}

	publisher_state = ast_calloc(1, sizeof(struct cc_publisher_state));
	if (!publisher_state) {
		return -1;
	}
	datastore->data = publisher_state;

	publisher_state->client = ao2_bump(client);
	publisher_state->monitor_instance = monitor_instance;
	monitor_instance->state = publisher_state;

	if (ast_sip_publish_client_add_datastore(client, datastore)) {
		return -1;
	}

	ast_log(LOG_NOTICE, "pjsip_cc_start_publishing\n");
	return 0;
}

static int pjsip_cc_stop_publishing(struct ast_sip_outbound_publish_client *client)
{
	RAII_VAR(struct ast_datastore *, datastore, ast_sip_publish_client_get_datastore(client, "ccss-publisher"),
		ao2_cleanup);

	if (!datastore) {
		return 0;
	}

	ast_log(LOG_NOTICE, "pjsip_cc_stop_publishing\n");
	ast_sip_publish_client_remove_datastore(client, "ccss-publisher");

	return 0;
}

static int pjsip_cc_check_resp_code(int code)
{
	switch (code) {
		case 180:
		case 181:
		case 183:
		case 200:
		case 300:
		case 301:
		case 302:
		case 303:
		case 305:
		case 486:
		case 600:
		case 603:
			return 0;
		default:
			return 1;
    }
}

/* Call-Info:<sip:456@z.b.com>;purpose=call-completion;m=NR */
static int pjsip_cc_get_call_info(pjsip_rx_data *rdata, char *subscribe_user,enum ast_cc_service_type *service)
{
	pjsip_generic_string_hdr *hdr;
	pj_str_t callinfo_value;

	static const pj_str_t purpose_str = { "purpose", 7 };
	static const pj_str_t m_str = { "m", 1 };
	static const pj_str_t from_name = { "From", 4 };
    static pjsip_fromto_hdr *pFrom;
	pjsip_sip_uri  *sip_uri;
	pjsip_param  *purpose;
	pjsip_param  *m_param;
	int    size;

    if (!rdata || !subscribe_user || !service) {
		return -1;
    }

    if (!(hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &callinfo_name, NULL))) {
		return -1;
	}

	pj_strdup_with_null(rdata->tp_info.pool, &callinfo_value, &hdr->hvalue);

	if (0 == callinfo_value.slen) {
		return -1;
	}

	pFrom = pjsip_parse_hdr(rdata->tp_info.pool, &from_name, callinfo_value.ptr, pj_strlen(&callinfo_value), &size);
    if (!pFrom) {
        return -1;
    }

	purpose = pjsip_param_find(&pFrom->other_param, &purpose_str);
	m_param = pjsip_param_find(&pFrom->other_param, &m_str);
	if (NULL == purpose || 0 != pj_stricmp2(&purpose->value, "call-completion")) {
        return -1;
	}

	if (!m_param) {
        return -1;
	}

	if (!pj_stricmp2(&m_param->value, "BS")) {
		*service = AST_CC_CCBS;
	} else if (!pj_stricmp2(&m_param->value, "NR")) {
		*service = AST_CC_CCNR;
	} else {
		*service = AST_CC_NONE;
		return -1;
	}

	sip_uri = pjsip_uri_get_uri(pFrom->uri);
	ast_copy_pj_str(subscribe_user, &sip_uri->user, AST_MAX_EXTENSION);

	ast_log(LOG_NOTICE, "pjsip_cc_get_call_info, subscribe_user[%s]\n", subscribe_user);

	return 0;
}


static void pjsip_cc_add_call_info_to_response(struct ast_sip_session *session, pjsip_tx_data *tdata, int code)
{
	char uuid_str[AST_UUID_STR_LEN] = {0};
	char *user = NULL;
	struct ast_cc_agent *agent = pjsip_find_cc_agent_by_original_callid(session);
	struct pjsip_cc_agent_pvt *agent_pvt;
	pjsip_fromto_hdr *hdr;
	pjsip_sip_uri    *uri;
	pjsip_param *param_p;
	pjsip_param *param_m;
	pjsip_tpmgr_fla2_param prm;
	pjsip_uri *base;
	pj_str_t  *host;	
    char buf[128]={0};
	char cc_uri[PJSIP_MAX_URL_SIZE] = {0};
	pjsip_generic_string_hdr  *hdr_old;

	if (!agent) {
		ast_log(LOG_WARNING, "Can't find SIP CC agent, remove the existed call-info\n");

		if ((hdr_old = pjsip_msg_find_hdr_by_name(tdata->msg, &callinfo_name, NULL))) {
			pj_list_erase(hdr_old);
		}

		return;
	}

	agent_pvt = agent->private_data; 
    if (agent_pvt->subscribe_uri) {
		pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, agent_pvt->subscribe_uri, cc_uri, sizeof(cc_uri));
		ast_log(LOG_NOTICE, "cc_add_call_info_to_response, already has subscribe_uri[%s]\n", cc_uri);
		ao2_ref(agent, -1);
		return;
	} 

	pjsip_tpmgr_fla2_param_default(&prm);
	prm.tp_type = tdata->tp_info.transport->key.type;
	pj_strset2(&prm.dst_host, tdata->tp_info.dst_name);
	prm.local_if = PJ_TRUE;

	/* get the local address*/
	if (pjsip_tpmgr_find_local_addr2(pjsip_endpt_get_tpmgr(ast_sip_get_pjsip_endpoint()), tdata->pool, &prm) != PJ_SUCCESS) {
		ao2_ref(agent, -1);
		ast_log(LOG_NOTICE, "Can't get local addr\n");
		return;
	} 

	if ((hdr_old = pjsip_msg_find_hdr_by_name(tdata->msg, &callinfo_name, NULL))) {
		pj_list_erase(hdr_old);
	}

	hdr = pjsip_from_hdr_create(tdata->pool);
	hdr->type = PJSIP_H_OTHER;
	pj_strdup(tdata->pool, &hdr->name, &callinfo_name);
	hdr->sname = hdr->name;

	/* get call-info uri */
	base = (pjsip_uri *)pjsip_sip_uri_create(tdata->pool, 0);
	uri = pjsip_uri_get_uri(base);
	user = ast_uuid_generate_str(uuid_str, sizeof(uuid_str));
	pj_strdup2(tdata->pool, &uri->user, user);

	host = &prm.ret_addr;
	pj_strdup_with_null( tdata->pool, &uri->host, host);

	strncpy(agent_pvt->subscribe_user, uuid_str, sizeof(agent_pvt->subscribe_user));
	agent_pvt->subscribe_uri = pjsip_uri_clone(agent_pvt->pool, base); 

	pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, agent_pvt->subscribe_uri, cc_uri, sizeof(cc_uri)); 
	ast_log(LOG_NOTICE, "Get subscribe_uri[%s]\n", cc_uri);

	if (AST_CC_NONE == session->cc_service) {
		if (code < 200) {
			session->cc_service = AST_CC_CCNR;
		} else {
			session->cc_service = AST_CC_CCBS;
		}
	}

	/* set hdr content */
	hdr->uri = base;

	param_p = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
	param_p->name = pj_str("purpose");
	param_p->value = pj_str("call-completion");
	pj_list_insert_before(&hdr->other_param, param_p);

	param_m = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
	param_m->name = pj_str("m");
	param_m->value = (session->cc_service == AST_CC_CCNR) ? pj_str("NR") : pj_str("BS");
	pj_list_insert_before(&hdr->other_param, param_m);

	if (pjsip_hdr_print_on((pjsip_hdr*)hdr, buf, sizeof(buf)-1) < 0) {
		ast_log(LOG_NOTICE, "cc_add_call_info_to_response, print failed\n");
		ao2_ref(agent, -1);
		return;
	}

	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr);
	ast_log(LOG_NOTICE, "cc_add_call_info_to_response, hdr[%s]\n", buf);
	ao2_ref(agent, -1);
}


static void pjsip_cc_session_incoming_response(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	pjsip_status_line status = rdata->msg_info.msg->line.status;
	enum ast_cc_monitor_policies monitor_policy;
	enum ast_cc_service_type  service = AST_CC_CCBS;
	char interface_name[AST_CHANNEL_NAME];
	char subscribe_user[AST_MAX_EXTENSION] = {0};
	int core_id;

    ast_log(LOG_NOTICE, "pjsip_cc_session_incoming_response, status.code[%d]\n", status.code);
	if ((NULL == session->endpoint) || (NULL == session->endpoint->cc_params) || (NULL == session->channel)) {
		return;
	}

	if (pjsip_cc_check_resp_code(status.code)) {
		ast_log(LOG_NOTICE, "status[%d], ignore it\n", status.code);
		return;
	}

	if (status.code < 200) {
		service = AST_CC_CCNR; 
	}

    monitor_policy = ast_get_cc_monitor_policy(session->endpoint->cc_params); 
	if (monitor_policy == AST_CC_MONITOR_NEVER) {
		ast_log(LOG_NOTICE, "The monitor_policy of endpoint[%s] is never\n", ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if ((core_id = ast_cc_get_current_core_id(session->channel)) == -1) {
		ast_log(LOG_NOTICE, "endpoint[%s], the cc core id is invalid\n", ast_sorcery_object_get_id(session->endpoint));
		return;
	}

    ast_log(LOG_NOTICE, "channel[%s], endpoint[%s], core id[%d], monitor policy[%d]\n", ast_channel_name(session->channel),
            ast_sorcery_object_get_id(session->endpoint), core_id, monitor_policy);

	ast_channel_get_device_name(session->channel, interface_name, sizeof(interface_name));  // PJSIP/5000
	if (pjsip_cc_get_call_info(rdata, subscribe_user, &service)) {
		if (status.code == 200) {
			ast_cc_failed(core_id, "core[%d], 200 response without call-info", core_id);
			return;
		}
		goto generic; 
	}

	if (monitor_policy == AST_CC_MONITOR_ALWAYS || monitor_policy == AST_CC_MONITOR_NATIVE) {
		struct pjsip_monitor_instance *monitor_instance = NULL;

		if (AST_CC_NONE != session->cc_service) {
			monitor_instance = ao2_callback(pjsip_monitor_instances, 0, pjsip_find_cc_monitor_instance_by_subscribe_user, subscribe_user);
            if (monitor_instance) {
				ast_log(LOG_NOTICE, "find monitor_instance!\n");
				ao2_ref(monitor_instance, -1);
				return;
            }
        }

		if (!(monitor_instance = pjsip_monitor_instance_init(core_id, subscribe_user,
														     ast_sorcery_object_get_id(session->endpoint), 
															 interface_name))) {
			goto generic;
		}

		/* We bump the refcount of chan_sip because once we queue this frame, the CC core
		 * will have a reference to callbacks in this module. We decrement the module
		 * refcount once the monitor destructor is called
		 */
		session->cc_service = service;
		ast_module_ref(ast_module_info->self);
        ast_log(LOG_NOTICE, "native monitor, interface[%s]\n", interface_name);
		ast_queue_cc_frame(session->channel, "PJSIP", interface_name, service, monitor_instance);
		ao2_ref(monitor_instance, -1);
		return;
	}

generic:
	if (monitor_policy == AST_CC_MONITOR_GENERIC || monitor_policy == AST_CC_MONITOR_ALWAYS) {
        ast_log(LOG_NOTICE, "generic monitor, interface[%s]\n", interface_name);
		session->cc_service = service;
		ast_queue_cc_frame(session->channel, AST_CC_GENERIC_MONITOR_TYPE, interface_name, service, NULL);
	}
}


/*!
 * \internal
 * \brief Adds a diversion header to an outgoing 3XX response
 *
 * \param session The session on which the INVITE response is to be sent
 * \param tdata The outbound INVITE response
 */
static void pjsip_cc_session_outgoing_response(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	struct pjsip_status_line status = tdata->msg->line.status;

	if (pjsip_cc_check_resp_code(status.code)) {
		ast_log(LOG_NOTICE, "status code[%d], ignore it\n", status.code);
		return;
	}

    ast_log(LOG_NOTICE, "pjsip_cc_session_outgoing_response, cc_offer[%d]\n", session->cc_offer);
	if (session->cc_offer) {
		pjsip_cc_add_call_info_to_response(session, tdata, status.code);
	}
	return;
}

static struct ast_sip_session_supplement cc_session_supplement = {
	.method = "INVITE",
	/* this supplement needs to be called after caller id and after the channel has been created */
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_CHANNEL + 100,
	.incoming_response = pjsip_cc_session_incoming_response,
	.outgoing_response = pjsip_cc_session_outgoing_response,
};


static void *pjsip_cc_allocate_body(void *data)
{
	struct ast_str **cc_str;

	cc_str = ast_malloc(sizeof(*cc_str));
	if (!cc_str) {
		return NULL;
	}
	*cc_str = ast_str_create(CC_BUFSIZE);
	if (!*cc_str) {
		ast_free(cc_str);
		return NULL;
	}
	return cc_str;
}

static int pjsip_cc_generate_body_content(void *body, void *data)
{
	struct ast_str **cc = body;
	struct pjsip_cc_notify_data *cc_data = data;
	char   cc_uri[PJSIP_MAX_URL_SIZE];

	ast_str_append(cc, 0, "cc-state: %s\r\n", cc_data->is_cc_ready ? "ready" : "queued");

	if (cc_data->is_cc_ready) {
		char *uri_ptr;
		int   uri_len;

		uri_len = pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, cc_data->notify_uri, cc_uri, sizeof(cc_uri)); 
        if (cc_uri[0] == '<' && cc_uri[uri_len-1] == '>') {
			cc_uri[uri_len-1] = '\0';
			uri_ptr = &cc_uri[1];
        } else {
			uri_ptr = cc_uri;
		}
		ast_str_append(cc, 0, "cc-URI: %s\r\n", uri_ptr);
	}

	return 0;
}

static void pjsip_cc_to_string(void *body, struct ast_str **str)
{
	struct ast_str **cc = body;

	ast_str_set(str, 0, "%s", ast_str_buffer(*cc));
}

static void pjsip_cc_destroy_body(void *body)
{
	struct ast_str **cc = body;

	ast_free(*cc);
	ast_free(cc);
}

static struct ast_sip_pubsub_body_generator cc_body_generator = {
	.type = "application",
	.subtype = "call-completion",
	.body_type = AST_SIP_CALL_COMPLETION,
	.allocate_body = pjsip_cc_allocate_body,
	.generate_body_content = pjsip_cc_generate_body_content,
	.to_string = pjsip_cc_to_string,
	.destroy_body = pjsip_cc_destroy_body,
};



static int unload_module(void)
{
	ast_sip_session_unregister_supplement(&cc_session_supplement);
	ast_cc_monitor_unregister(&cc_monitor_callbacks);
	ast_cc_agent_unregister(&cc_agent_callbacks);
	ast_sip_pubsub_unregister_body_generator(&cc_body_generator);
	ast_sip_unregister_publish_handler(&cc_publish_handler);
	ast_sip_unregister_subscription_handler(&cc_handler);
	ast_sip_exten_unregister_supplement(&cc_original_exten);
	ast_sip_exten_unregister_supplement(&cc_notify_user);

	ao2_t_cleanup(pjsip_monitor_instances, "unref the pjsip_monitor_instances table");

	if (cc_sched) {
		ast_sched_context_destroy(cc_sched);
		cc_sched = NULL;
	}
	return 0;
}


static int load_module(void)
{
	CHECK_PJSIP_SESSION_MODULE_LOADED();
	CHECK_PJSIP_PUBSUB_MODULE_LOADED();

	cc_sched = ast_sched_context_create();
	if (!cc_sched) {
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (ast_sched_start_thread(cc_sched)) {
		ast_sched_context_destroy(cc_sched);
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (!(pjsip_monitor_instances = ao2_container_alloc(37, pjsip_monitor_instance_hash_fn, pjsip_monitor_instance_cmp_fn))) {
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (ast_cc_monitor_register(&cc_monitor_callbacks)) {
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

    if (ast_cc_agent_register(&cc_agent_callbacks)) {
        unload_module();
        return AST_MODULE_LOAD_DECLINE;
    }

	if (ast_sip_pubsub_register_body_generator(&cc_body_generator)) {
        unload_module();
        return AST_MODULE_LOAD_DECLINE;
    }

	if (ast_sip_register_publish_handler(&cc_publish_handler)) {
        unload_module();
        return AST_MODULE_LOAD_DECLINE;
    }

	if (ast_sip_register_subscription_handler(&cc_handler)) {
        unload_module();
        return AST_MODULE_LOAD_DECLINE;
	}

	ast_sip_session_register_supplement(&cc_session_supplement); 
	ast_sip_exten_register_supplement(&cc_original_exten);
	ast_sip_exten_register_supplement(&cc_notify_user);

	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP CCSS Support",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.load_pri = AST_MODPRI_APP_DEPEND,
	       );
