/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2014, Digium, Inc.
 *
 * Thammanoon Kawinfruangfukul <thammanoon@grandstream.com>
 * Weiling Liao <weiling.liao@grandstream.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_pubsub</depend>
	<support_level>core</support_level>
 ***/

#ifdef GRANDSTREAM_NETWORKS

#define AST_MODULE_LOG "pjsip"

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 13.0 $");

#include <pjsip.h>
#include <pjsip_simple.h>
#include <pjlib.h>
#include <zlib.h>

#include "asterisk/app.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_pubsub.h"
#include "asterisk/res_pjsip_body_generator_types.h"
#include "asterisk/module.h"
#include "asterisk/strings.h"
#include "asterisk/pjsip_cei.h"
#include "asterisk/stasis.h"
#include "asterisk/stasis_message_router.h"
#include "asterisk/astdb.h"

static struct stasis_message_router *cei_mcb_sub_router = NULL;

static int unload_module(void);
static void cei_subscription_shutdown(struct ast_sip_subscription *sub);
static int cei_new_subscribe(struct ast_sip_endpoint *endpoint,
		const char *resource);
static int cei_subscription_established(struct ast_sip_subscription *sub);
static void *cei_get_notify_data(struct ast_sip_subscription *sub);
static int cei_subscription_activation(struct ast_sip_subscription *sub, int active);
static void cei_subscriber_state_change(struct ast_sip_subscription *sub,
										pjsip_msg_body *body,
										enum pjsip_evsub_state state);
static struct cei_subscription *cei_subscribe_stasis(
		struct ast_sip_endpoint *endpoint, struct ast_sip_subscription *sip_sub, const char *name);
static void cei_stasis_cb(void *userdata, struct stasis_subscription *sub,
		struct stasis_message *msg);

/*!
 * \brief Wrapper for stasis subscription
 *
 * An CEI subscription has a container of these. This
 * represents a stasis subscription for CEI state.
 */
struct cei_stasis_subscription {
	/*! The CEI stasis subscription */
	struct stasis_subscription *stasis_sub;
	/*! The conference corresponding with the CEI subscription. Used as a hash key */
	char conference[1];
};

/*!
 * \brief A subscription for CEI NOTIFY
 *
 * This subscription is the basis for CEI for an endpoint. Each
 * endpoint that uses CEI will have a corresponding cei_subscription.
 *
 * This structure acts as the owner for the underlying SIP subscription.
 * When the cei_subscription is destroyed, the SIP subscription dies, too.
 * The cei_subscription's lifetime is governed by its underlying stasis
 * subscriptions. When all stasis subscriptions are destroyed, the
 * cei_subscription is destroyed as well.
 */
struct cei_subscription {
	/*!	Subscription active status */
	int active;
	/*!	The stasis sub. */
 	struct ao2_container *stasis_subs;
	/*! The SIP subscription. */
	struct ast_sip_subscription *sip_sub;
	/*! The notify data. */
	struct pjsip_cei_notify_data *notify_raw_data;
	/*! Retrived URI from To header of SIP SUBSCRIBE */
	char ceiURI[PJSIP_MAX_URL_SIZE];
	/*! the local conference num user joined*/
	char *local_conference;
	/*! The notify version */
	int ceiNotifyVer;
	/*! The type of cei user*/
	enum cei_user_type user_type;
	/*! callerid */
	char *user_name;
	/*! peer channnel; mostly used for identify normal user*/
	char *channelName;
	/*! self conf id if external mcb user; NULL if normal user; mostly used for identifymcb user*/
	char *self_conf_id;
	/*! Identifier for the subscription.
	 * The identifier is the same as the corresponding endpoint's stasis ID.
	 * Used as a hash key
	 */
	char id[1];
};


static struct ast_sip_notifier cei_notifier = {
	.default_accept = CEI_TYPE"/"CEI_SUBTYPE,
	.new_subscribe = cei_new_subscribe,
	.subscription_established = cei_subscription_established,
	.get_notify_data = cei_get_notify_data,
	.subscription_activation = cei_subscription_activation,
};

static struct ast_sip_subscriber cei_subscriber = {
	.state_change = cei_subscriber_state_change,
};

static struct ast_sip_subscription_handler cei_handler = {
	.event_name = "conference",
	.body_type = AST_SIP_CEI_NOTIFY_DATA,
	.accept = { CEI_TYPE"/"CEI_SUBTYPE, },
	.subscription_shutdown = cei_subscription_shutdown,
	.notifier = &cei_notifier,
	.subscriber = &cei_subscriber,
};

struct ao2_container *cei_subs;


static int select_sip_info(struct ast_sip_endpoint *endpoint,
						   const char **user, const char **domain)
{
	// Sanity check
	if (endpoint == NULL) {
		return -1;
	}

	// Contact header
	if (!ast_strlen_zero(endpoint->cei_contactuser) &&
		!ast_strlen_zero(endpoint->cei_contactdomain)) {
		*user = endpoint->cei_contactuser;
		*domain = endpoint->cei_contactdomain;
	}
	// From header
	else {
		*user = endpoint->cei_fromuser;
		*domain = endpoint->cei_fromdomain;
	}

	return 0;
}

static int unsubscribe_stasis(void *obj, void *arg, int flags)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct cei_stasis_subscription *cei_stasis = obj;
	if (cei_stasis->stasis_sub) {
		ast_debug(3, "Removing stasis subscription to conference %s\n", cei_stasis->conference);
		cei_stasis->stasis_sub = stasis_unsubscribe(cei_stasis->stasis_sub);
	}
	return 0;
}

static struct cei_subscription *find_cei_subscription_by_datastore(struct ast_sip_subscription *sub)
{
	struct cei_subscription *cei_sub;
	RAII_VAR(struct ast_datastore *, cei_datastore,
			ast_sip_subscription_get_datastore(sub, CEI_DATASTORE), ao2_cleanup);

	if (!cei_datastore) {
		return NULL;
	}

	cei_sub = cei_datastore->data;

	if (!cei_sub) {
		ast_log(LOG_DEBUG, "[TK] NO CEI object \n");
		return NULL;
	}
	return cei_sub;
}

static void cei_subscription_shutdown(struct ast_sip_subscription *sub)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct cei_subscription *cei_sub;

	cei_sub = find_cei_subscription_by_datastore(sub);

	if (!cei_sub) {
		ast_log(LOG_DEBUG, "[TK] NO CEI object \n");
		return;
	}
	ao2_callback(cei_sub->stasis_subs, OBJ_UNLINK | OBJ_NODATA | OBJ_MULTIPLE, unsubscribe_stasis, NULL);

}

/*! \brief Handle new subscription for a conference user */
static int cei_new_subscribe_conf_user(struct ast_sip_endpoint *endpoint,
									   const char *endpoint_id,
							 		   const char *resource)
{
	struct sip_conference* conf = NULL;

	// Sanity check
	if (endpoint == NULL || endpoint_id == NULL || resource == NULL) {
		return 404;
	}

	conf = ast_sip_cei_find_peer_conference(endpoint, resource);
	if (conf == NULL) {
		ast_log(LOG_WARNING, "Invalid subscription from %s to %s\n",
							 endpoint_id, resource);
		return 404;
	}

	if (conf->state != SIP_CONFERENCE_STATE_WAITSUB) {
		ast_log(LOG_WARNING, "%s already subscribed to %s\n",
							 endpoint_id, resource);
		return 404;
	}

	return 200;
}

/*! \brief Handle new subscription for an external MCB peer */
static int cei_new_subscribe_mcb_peer(struct ast_sip_endpoint *endpoint,
									  const char *endpoint_id,
							 		  const char *resource)
{
	struct cei_external_mcb_peer* mcb_peer = NULL;
	const char *remote_conf = NULL;
	const char *remote_conf_host = NULL;

	// Sanity check
	if (endpoint == NULL || endpoint_id == NULL || resource == NULL) {
		return 404;
	}

	select_sip_info(endpoint, &remote_conf, &remote_conf_host);

	mcb_peer = ast_cei_find_external_mcb_peer(endpoint, resource, remote_conf,
									  		  endpoint_id, remote_conf_host);

	// Create new MCB peer if needed.
	// Local conference ID, remote conference id, and channel name will be
	// filled in by sip_external_mcb_sub in chan_pjsip.c
	if (mcb_peer == NULL) {
		char *user = ast_strdupa(remote_conf);
		char *domain = ast_strdupa(remote_conf_host);

		mcb_peer = ast_cei_add_external_mcb_peer(endpoint, resource, "",
												 user, "", domain, "");
	}

    AST_LIST_LOCK(&endpoint->external_mcb_peers);

	if (mcb_peer == NULL) {
		ast_log(LOG_WARNING, "Invalid subscription from %s:%s via %s to Conference %s\n",
							 S_OR(remote_conf, ""), S_OR(remote_conf_host, ""),
							 endpoint_id, resource);
        AST_LIST_UNLOCK(&endpoint->external_mcb_peers);

		return 404;
	}

    // Handle duplicate SIP subscription
	if (mcb_peer->state != SIP_CONFERENCE_STATE_WAITSUB) {
		ast_log(LOG_WARNING, "%s:%s via %s already subscribed to Conference %s\n",
							 S_OR(remote_conf, ""), S_OR(remote_conf_host, ""),
							 endpoint_id, resource);

        // [TK] On case II of Bug 49550, we will need to destory current incoming SIP subscription.
        // We could not find sip_sub since it does not exist in subscriptions. We may need to add reference
        // to cei_sub in mcb_peer after establishing the subscription. However, this process will be done
        // by Asterisk after 30 minutes.

        // Update subscription status
        mcb_peer->state = SIP_CONFERENCE_STATE_WAITSUB;
	}

    AST_LIST_UNLOCK(&endpoint->external_mcb_peers);

    return 200;
}

static int cei_new_subscribe(struct ast_sip_endpoint *endpoint,
							 const char *resource)
{
	const char *id = NULL;
	enum cei_user_type user_type = CEI_TYPE_NORMAL_USER;

	// Sanity check
	if (endpoint == NULL || ast_strlen_zero(resource)) {
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return 404;
	}

	id = ast_sorcery_object_get_id(endpoint);
	if (id == NULL) {
		ast_log(LOG_WARNING, "id IS NULL\n");
		return 404;
	}

	// This subscription is for an external MCB
	if (!ast_strlen_zero(endpoint->transport_type) &&
		strstr(endpoint->transport_type, "trunk") != NULL) {
		user_type = CEI_TYPE_EXTERNAL_MCB;
	}

	// Verify subscription
	switch(user_type) {
		case CEI_TYPE_NORMAL_USER: {
			return cei_new_subscribe_conf_user(endpoint, id, resource);
		}
		break;

		case CEI_TYPE_EXTERNAL_MCB: {
			return cei_new_subscribe_mcb_peer(endpoint, id, resource);
		}
		break;

		default:
			ast_log(LOG_WARNING, "We do not support type %d for endpoint id: %s, resource: %s\n",
								 user_type, id, resource);
	}

	return 404;
}

/*! \brief Handle established subscription for a conference user */
static int cei_subscription_established_conf_user(struct ast_sip_subscription *sip_sub,
												  struct ast_sip_endpoint *endpoint,
											      const char *endpoint_id,
									 		      const char *resource,
												  char *cei_uri)
{
	struct sip_conference* conf = NULL;
	struct cei_subscription *cei_sub = NULL;

	// Sanity check
	if (endpoint == NULL || endpoint_id == NULL ||
		resource == NULL || cei_uri == NULL || sip_sub == NULL) {
		return -1;
	}

	conf = ast_sip_cei_find_peer_conference(endpoint, resource);
	if (conf == NULL) {
		ast_log(LOG_WARNING, "No subscription from %s to %s\n",
							 endpoint_id, resource);
		return -1;
	}

	ao2_lock(cei_subs);

	cei_sub = cei_subscribe_stasis(endpoint, sip_sub, resource);
	if (!cei_sub) {
		ast_log(LOG_WARNING, "[WL] test NO CEI\n");
		ao2_unlock(cei_subs);
		return -1;
	}

	// Set user_type
	cei_sub->user_type = CEI_TYPE_NORMAL_USER;
	cei_sub->channelName = ast_strdup(conf->channelName);
	cei_sub->local_conference = ast_strdup(conf->conference);
	cei_sub->self_conf_id = NULL;

	// Set peer conference's subscription if a user join a conference
	// with correct password.
	cei_sub->active = conf->ceiJoined;

	// Set CEI URI
	strncpy(cei_sub->ceiURI, cei_uri, sizeof(cei_sub->ceiURI));

	if (!ao2_link(cei_subs, cei_sub)) {
		ao2_ref(cei_sub, -2);
		cei_sub = NULL;
		ao2_unlock(cei_subs);
		return -1;
	}

	ao2_unlock(cei_subs);

	// Update subscription state
	conf->state = SIP_CONFERENCE_STATE_SUBSCRIBED;

	// Update channel
	ast_sip_set_sub_chan(sip_sub, conf->channelName);

	return 0;
}

/*! \brief Handle established subscription for an external MCB peer */
static int cei_subscription_established_mcb_peer(struct ast_sip_subscription *sip_sub,
												 struct ast_sip_endpoint *endpoint,
											     const char *endpoint_id,
									 		     const char *resource,
												 char *cei_uri)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);

	struct cei_external_mcb_peer* mcb_peer = NULL;
	struct cei_subscription *cei_sub = NULL;
	const char *remote_conf = NULL;
    const char *remote_conf_host = NULL;

	// Sanity check
	if (endpoint == NULL || endpoint_id == NULL ||
		resource == NULL || cei_uri == NULL) {
		return -1;
	}

	select_sip_info(endpoint, &remote_conf, &remote_conf_host);

	mcb_peer = ast_cei_find_external_mcb_peer(endpoint, resource, remote_conf,
											  endpoint_id, remote_conf_host);

	if (mcb_peer == NULL) {
		ast_log(LOG_WARNING, "No subscription from %s:%s via %s to Conference %s\n",
							 S_OR(remote_conf, ""), S_OR(remote_conf_host, ""),
							 endpoint_id, resource);
		return -1;
	}


	ao2_lock(cei_subs);

	cei_sub = cei_subscribe_stasis(endpoint, sip_sub, mcb_peer->local_conf_private_num);
	if (!cei_sub) {
		ast_log(LOG_WARNING, "[WL] test NO CEI\n");
		ao2_unlock(cei_subs);
		return -1;
	}

	// Set user_type
	cei_sub->user_type = CEI_TYPE_EXTERNAL_MCB;


	// Set CEI URI
	strncpy(cei_sub->ceiURI, cei_uri, sizeof(cei_sub->ceiURI));

	if (!ast_strlen_zero(mcb_peer->remote_conf_id) && cei_sub->self_conf_id == NULL) {
		cei_sub->self_conf_id = ast_strdup(mcb_peer->remote_conf_id);
	}
	if (!ast_strlen_zero(mcb_peer->channelName) && cei_sub->channelName == NULL) {
		cei_sub->channelName = ast_strdup(mcb_peer->channelName);
	}

	if (!ao2_link(cei_subs, cei_sub)) {
		ao2_ref(cei_sub, -2);
		cei_sub = NULL;
		ao2_unlock(cei_subs);
		return -1;
	}

	ao2_unlock(cei_subs);

	// Update subscription state
	mcb_peer->state = SIP_CONFERENCE_STATE_SUBSCRIBED;

	// MCB's subscription is always active.
	cei_sub->active = 1;

	return 0;
}

static int cei_subscription_established(struct ast_sip_subscription *sip_sub)
{
	struct ast_sip_endpoint *endpoint = ast_sip_subscription_get_endpoint(sip_sub);
	const char *resource = ast_sip_subscription_get_resource_name(sip_sub);
	const char *id = (endpoint != NULL) ? ast_sorcery_object_get_id(endpoint) : NULL;
	char cei_uri[PJSIP_MAX_URL_SIZE] = {0};
	char *uri_port = NULL;
	char *iter_uri_port = NULL;
	enum cei_user_type user_type = CEI_TYPE_NORMAL_USER;
	int ret = -1;

	// Sanity check
	if (endpoint == NULL) {
		return -1;
	}

	if (id == NULL || ast_strlen_zero(resource)) {
		ao2_cleanup(endpoint);
		return -1;
	}

	// Get CEI URI
	ast_sip_subscription_get_local_uri(sip_sub, cei_uri, sizeof(cei_uri));
	uri_port = iter_uri_port = strstr(cei_uri, ":");

	while (iter_uri_port != NULL) {
		iter_uri_port = iter_uri_port + 1;
		iter_uri_port = strstr(iter_uri_port, ":");
		if (iter_uri_port != NULL) {
			uri_port = iter_uri_port;
		}
	}

	if (uri_port != NULL) {
		*uri_port = '\0';
	}

	if (ast_strlen_zero(cei_uri)) {
		ao2_cleanup(endpoint);
		return -1;
	}

	// This subscription is for an external MCB
	if (!ast_strlen_zero(endpoint->transport_type) &&
		strstr(endpoint->transport_type, "trunk") != NULL) {
		user_type = CEI_TYPE_EXTERNAL_MCB;
	}

	// Handle established subscription
	switch(user_type) {
		case CEI_TYPE_NORMAL_USER: {
			ret = cei_subscription_established_conf_user(sip_sub, endpoint, id,
														 resource, cei_uri);
		}
		break;

		case CEI_TYPE_EXTERNAL_MCB: {
			ret = cei_subscription_established_mcb_peer(sip_sub, endpoint, id,
														resource, cei_uri);
		}
		break;

		default: {
			ast_log(LOG_WARNING, "We do not support type %d for endpoint id: %s, resource: %s\n",
								 user_type, id, resource);
			ret = -1;
		}
		break;
	}

	// Clean up
	ao2_cleanup(endpoint);

	return ret;
}

static void cei_notify_data_dtor(void *obj)
{
	struct pjsip_cei_notify_data *notify_data = obj;

	if (notify_data->channelName != NULL) {
		ast_free(notify_data->channelName);
		notify_data->channelName = NULL;
	}

	if (notify_data->conference != NULL) {
		ast_free(notify_data->conference);
		notify_data->conference = NULL;
	}

	if (notify_data->content != NULL) {
		ast_free(notify_data->content);
		notify_data->content = NULL;
	}
}


// Used to generate the notify data for terminate notify
static void *cei_get_notify_data(struct ast_sip_subscription *sub)
{
	ast_log(LOG_DEBUG, "[WL] Enter...%s\n", __FUNCTION__);
	struct cei_subscription *cei_sub;
	struct pjsip_cei_notify_data *notify_data = NULL;

	cei_sub = find_cei_subscription_by_datastore(sub);
	if (cei_sub == NULL) {
		ast_log(LOG_DEBUG, "[TK] NO CEI object \n");
		return NULL;
	}

	notify_data = ao2_alloc(sizeof(*notify_data), cei_notify_data_dtor);
	if (notify_data == NULL) {
		ast_log(LOG_WARNING, "[TK] Failed to create notify_data");
		return NULL;
	}

	strncpy(notify_data->ceiURI, cei_sub->ceiURI, sizeof(notify_data->ceiURI));
	notify_data->notifyVer = ++cei_sub->ceiNotifyVer;
	notify_data->isTerminated = 1;
	notify_data->content = NULL;
	notify_data->channelName = NULL;
	notify_data->conference = NULL;

	ast_log(LOG_DEBUG, "[WL] Exit...%s\n", __FUNCTION__);

	return notify_data;
}

static void *new_cei_get_notify_data(struct cei_subscription *cei_sub)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);

	if(!cei_sub) {
		ast_log(LOG_DEBUG, "NO CEI object \n");
		return NULL;
	}
	ast_log(LOG_DEBUG, "[TK] User name : %s\n", cei_sub->user_name);

	struct pjsip_cei_notify_data *raw_data = cei_sub->notify_raw_data;
	const char *channel = NULL;
	const char *conference = NULL;
	struct pjsip_cei_notify_data *notify_data = NULL;

	if (raw_data == NULL) {
		ast_debug(LOG_DEBUG, "[WL] Invaild CEI data");
		return NULL;
	}

	if (raw_data->conference == NULL || raw_data->content == NULL) {
		ast_debug(LOG_DEBUG, "[WL] Invaild CEI data: conf %s, content %s",
                             S_OR(raw_data->conference, "NULL"),
                             S_OR(raw_data->content, "NULL"));
		return NULL;
	}

	ast_debug(LOG_DEBUG, "[WL] Got raw notify data for conf: %s with content: %s", raw_data->conference, raw_data->content);


	channel = cei_sub->channelName;
	conference = cei_sub->local_conference;
	if ( channel == NULL || conference == NULL ) {
		return NULL;
	}


	notify_data = ao2_alloc(sizeof(*notify_data), cei_notify_data_dtor);
	if (notify_data == NULL) {
		ast_log(LOG_WARNING, "[TK] Failed to create notify_data");
		return NULL;
	}

	// Prepare XML body
	if (raw_data->notifyVer) {
		cei_sub->ceiNotifyVer = 1;
	} else {
		cei_sub->ceiNotifyVer++;
	}

	ast_log(LOG_WARNING, "[TK] ready to push the data: channel %s; conference %s; notify version %d; content %s;",
									channel,
									conference,
									cei_sub->ceiNotifyVer,
									raw_data->content);

	notify_data->channelName = ast_strdup(channel);
	notify_data->conference = ast_strdup(conference);
	strncpy(notify_data->ceiURI, cei_sub->ceiURI, sizeof(notify_data->ceiURI));
	notify_data->notifyVer = cei_sub->ceiNotifyVer;
	notify_data->content = ast_strdup(raw_data->content);
	notify_data->isTerminated = 0;

	ast_log(LOG_DEBUG, "[TK] Exit...%s\n", __FUNCTION__);

	return notify_data;
}

static int cei_subscription_activation(struct ast_sip_subscription *sub, int active)
{
	struct cei_subscription *cei_sub;

	cei_sub = find_cei_subscription_by_datastore(sub);
	if (cei_sub == NULL) {
		ast_log(LOG_WARNING, "NO CEI object\n");
		return -1;
	}
	ao2_lock(cei_sub);
	cei_sub->active = active;
	ast_log(LOG_DEBUG, "Subscription of channel %s is %s\n",
						cei_sub->channelName,
						cei_sub->active ? "active" : "inactive");

	ao2_unlock(cei_sub);

	return 0;
}

static int stasis_sub_hash(const void *obj, const int flags)
{
	const struct cei_stasis_subscription *object;
	const char *key;

	switch (flags & OBJ_SEARCH_MASK) {
	case OBJ_SEARCH_KEY:
		key = obj;
		break;
	case OBJ_SEARCH_OBJECT:
		object = obj;
		key = object->conference;
		break;
	default:
		ast_assert(0);
		return 0;
	}
	return ast_str_hash(key);
}

static int stasis_sub_cmp(void *obj, void *arg, int flags)
{
	const struct cei_stasis_subscription *sub_left = obj;
	const struct cei_stasis_subscription *sub_right = arg;
	const char *right_key = arg;
	int cmp;

	switch (flags & OBJ_SEARCH_MASK) {
	case OBJ_SEARCH_OBJECT:
		right_key = sub_right->conference;
		/* Fall through */
	case OBJ_SEARCH_KEY:
		cmp = strcmp(sub_left->conference, right_key);
		break;
	case OBJ_SEARCH_PARTIAL_KEY:
		cmp = strncmp(sub_left->conference, right_key, strlen(right_key));
		break;
	default:
		cmp = 0;
		break;
	}
	if (cmp) {
		return 0;
	}
	return 0;
}


static struct cei_stasis_subscription *cei_stasis_subscription_alloc(const char *conference, struct cei_subscription *cei_sub)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct cei_stasis_subscription *cei_stasis_sub;
	struct stasis_topic *topic;

	if (!cei_sub) {
		return NULL;
	}
	cei_stasis_sub = ao2_alloc(sizeof(*cei_stasis_sub) + strlen(conference), NULL);
	if (!cei_stasis_sub) {
		ao2_cleanup(cei_sub);
		return NULL;
	}

	topic = ast_cei_topic(conference);

	if (topic == NULL) {
		ast_log(LOG_DEBUG, "[WL] Can NOT find stasis topic \n");
		return NULL;
	}

	/* Safe strcpy */
	strcpy(cei_stasis_sub->conference, conference);
	cei_sub->local_conference = ast_strdup(conference);
	ao2_ref(cei_sub, +1);
	cei_stasis_sub->stasis_sub = stasis_subscribe(topic, cei_stasis_cb, cei_sub);
	return cei_stasis_sub;
}


static void cei_subscription_destructor(void *obj)
{
	struct cei_subscription *cei_sub = obj;
	struct ao2_iterator iter;

	if (!cei_sub) {
		return;
	}

	if (cei_sub->stasis_subs) {
		struct cei_stasis_subscription *stasis_sub = NULL;
		ao2_lock(cei_sub->stasis_subs);

		iter = ao2_iterator_init(cei_sub->stasis_subs, 0);
		while ((stasis_sub = ao2_iterator_next(&iter))) {
			ao2_unlink(cei_sub->stasis_subs, stasis_sub);
			ao2_ref(stasis_sub, -2);
		}
		ao2_iterator_destroy(&iter);
		ao2_unlock(cei_sub->stasis_subs);

		ao2_ref(cei_sub->stasis_subs, -1);
	}

	if (cei_sub->sip_sub != NULL) {
		ao2_cleanup(cei_sub->sip_sub);
		cei_sub->sip_sub = NULL;
	}
	if (cei_sub->notify_raw_data != NULL) {
		ao2_cleanup(cei_sub->notify_raw_data);
		cei_sub->notify_raw_data = NULL;
	}

	if (cei_sub->user_name != NULL) {
		ast_free(cei_sub->user_name);
		cei_sub->user_name = NULL;
	}

	if (cei_sub->channelName != NULL) {
		ast_free(cei_sub->channelName);
		cei_sub->channelName = NULL;
	}

	if (cei_sub->local_conference != NULL) {
		ast_free(cei_sub->local_conference);
		cei_sub->local_conference = NULL;
	}

}

static struct cei_subscription *cei_subscription_alloc(struct ast_sip_endpoint *endpoint,
	struct ast_sip_subscription *sip_sub)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct cei_subscription *cei_sub;
	const char *endpoint_id = ast_sorcery_object_get_id(endpoint);
	const char *user = NULL;
    const char *domain = NULL;

	cei_sub = ao2_alloc(sizeof(*cei_sub) + strlen(endpoint_id),
											cei_subscription_destructor);

	if (!cei_sub) {
		return NULL;
	}

	select_sip_info(endpoint, &user, &domain);

	/* Safe strcpy */
	strcpy(cei_sub->id, endpoint_id);

	cei_sub->sip_sub = ao2_bump(sip_sub);
	cei_sub->user_name = ast_strdup(user);
	cei_sub->channelName = NULL;
	cei_sub->local_conference = NULL;
	cei_sub->stasis_subs = ao2_container_alloc(STASIS_BUCKETS, stasis_sub_hash, stasis_sub_cmp);
	if (!cei_sub->stasis_subs) {
		ao2_cleanup(cei_sub);
		return NULL;
	}

	return cei_sub;
}

static struct ast_datastore_info cei_ds_info = { };

static int add_cei_datastore(struct cei_subscription *cei_sub)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);

	RAII_VAR(struct ast_datastore *, cei_datastore, NULL, ao2_cleanup);

	cei_datastore = ast_sip_subscription_alloc_datastore(&cei_ds_info, CEI_DATASTORE);
	if (!cei_datastore) {
		return -1;
	}
	cei_datastore->data = cei_sub;

	ast_sip_subscription_add_datastore(cei_sub->sip_sub, cei_datastore);
	return 0;
}



static struct cei_subscription *cei_create_subscription(
	struct ast_sip_endpoint *endpoint, struct ast_sip_subscription *sip_sub)
{
	struct cei_subscription *cei_sub = cei_subscription_alloc(endpoint, sip_sub);

	if (!cei_sub) {
		return NULL;
	}

	if (add_cei_datastore(cei_sub)) {
		ao2_cleanup(cei_sub);
		return NULL;
	}

	memset(cei_sub->ceiURI, 0, sizeof(cei_sub->ceiURI));
	cei_sub->ceiNotifyVer = 0;

	return cei_sub;
}

static struct cei_subscription *cei_subscribe_stasis(
	struct ast_sip_endpoint *endpoint, struct ast_sip_subscription *sip_sub, const char *name)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);

	if (name == NULL || endpoint == NULL || sip_sub == NULL) {
		return NULL;
	}
	struct cei_subscription *cei_sub;
	char *conference = ast_strdup(name);

	if (!(cei_sub = cei_create_subscription(endpoint, sip_sub))) {
		ast_free(conference);
		return NULL;
	}

	RAII_VAR(struct cei_stasis_subscription *, cei_stasis_sub,
		cei_stasis_subscription_alloc(conference, cei_sub), ao2_cleanup);
	if (cei_stasis_sub) {
		ao2_link(cei_sub->stasis_subs, cei_stasis_sub);
	}else{
		ast_free(conference);
		return NULL;
	}
	ast_free(conference);
	return cei_sub;
}

static void send_cei_notify(struct cei_subscription *cei_sub)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct pjsip_cei_notify_data *notify_data = NULL;
	struct ast_sip_body_data data = {
		.body_type = AST_SIP_CEI_NOTIFY_DATA,
	};

  	if(!cei_sub) {
		ast_log(LOG_ERROR, "NO CEI object\n");
		return;
	}
	notify_data = new_cei_get_notify_data(cei_sub);
	if (!notify_data) {
		return;
	}

	data.body_data = notify_data;
	ast_sip_subscription_notify(cei_sub->sip_sub, &data, 0);

	// Clean up
	ao2_cleanup(notify_data);
}

static int serialized_notify(void *userdata)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct cei_subscription *cei_sub = userdata;

	// Sanity check
	if (cei_sub == NULL) {
		return -1;
	}

	send_cei_notify(cei_sub);

	// Clean up
	ao2_cleanup(cei_sub->notify_raw_data);
	cei_sub->notify_raw_data = NULL;

	return 0;
}

static int serialized_cleanup(void *userdata)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct cei_subscription *cei_sub = userdata;

	ao2_lock(cei_subs);
	ao2_unlink(cei_subs, cei_sub);

	/* This is getting rid of the reference that was added
	 * just before this serialized task was pushed.
	 */
	ao2_cleanup(cei_sub);
	/* This is getting rid of the reference held by the
	 * stasis subscription
	 */
	ao2_cleanup(cei_sub);
	ao2_unlock(cei_subs);
	return 0;
}

static void cei_stasis_cb(void *userdata, struct stasis_subscription *sub,
		struct stasis_message *msg)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s \n", __FUNCTION__);
	struct cei_subscription *cei_sub = userdata;
	int flagFullNotify = 0;

	if (!cei_sub) {
		ast_log(LOG_ERROR, "NO CEI object\n");
		return;
	}
	ao2_lock(cei_sub);
	
	if (stasis_subscription_final_message(sub, msg)) {
		ast_log(LOG_ERROR, "[WL] Got final stasis message of user : %s\n", cei_sub->user_name);
		ao2_unlock(cei_sub);
		ast_sip_push_task(NULL, serialized_cleanup, cei_sub);
		return;
	}


	if (ast_cei_notify_type() == stasis_message_type(msg)) {
		ast_log(LOG_ERROR, "[WL] Got stasis message of user %s\n", cei_sub->user_name);
		ao2_ref(cei_sub, +1);
		if (cei_sub->sip_sub == NULL) {
			goto cleanup_ceisub;
		}

		if (ast_sip_subscription_is_terminated(cei_sub->sip_sub)) {
			goto cleanup_ceisub;
		}

		ast_sip_change_tree_ref(cei_sub->sip_sub, +1);

    struct ast_json_payload *payload = stasis_message_data(msg);
		struct ast_json *json_object = payload->json;

		if (json_object == NULL) {
			ast_debug(LOG_DEBUG, "[WL] Imcomplete message.");
			goto cleanup_tree;
		}

		enum cei_user_type user_type = ast_json_integer_get(ast_json_object_get(json_object, "user_type"));
		const char *conference_name = ast_json_string_get(ast_json_object_get(json_object, "conference"));
		unsigned int version_reset = ast_json_integer_get(ast_json_object_get(json_object, "version_reset"));
		const char *partial_text = ast_json_string_get(ast_json_object_get(json_object, "partial_text"));
		const char *full_text = ast_json_string_get(ast_json_object_get(json_object, "full_text"));
		const char *event_type = ast_json_string_get(ast_json_object_get(json_object, "event_type"));
		const char *user_name = ast_json_string_get(ast_json_object_get(json_object, "user_name"));
		unsigned int user_role = ast_json_integer_get(ast_json_object_get(json_object, "user_role"));
		unsigned int preverified_user = ast_json_integer_get(ast_json_object_get(json_object, "preverified_user"));


		if ( conference_name == NULL || event_type == NULL ) {
			ast_debug(LOG_DEBUG, "[WL] Imcomplete data. \n");
			goto cleanup_tree;
		}

		if (cei_sub->active == 0 && !preverified_user) {
			ast_log(LOG_WARNING, "Subscription of channel %s is inactive\n", cei_sub->channelName);
			goto cleanup_tree;
		}

		ast_debug(LOG_DEBUG, "[WL] Got notify data : %d,%d,%s,%s,%s,%s,%s",
		                          user_type,
															cei_sub->user_type,
															conference_name,
															event_type,
															user_name,
															cei_sub->channelName,
															cei_sub->self_conf_id);

  	// Send pull state notify to the new joined user
		if (strcmp(event_type, CEI_NOTIFY_EVENT_JOIN) == 0 && user_type == cei_sub->user_type) {
			ast_debug(LOG_DEBUG, "[WL] It's a join event\n");

			switch (cei_sub->user_type) {
				case CEI_TYPE_NORMAL_USER: {
					if (user_name == NULL || cei_sub->channelName == NULL) {
						goto cleanup_tree;
					}
					if (ast_sip_cei_same_channel(user_name, cei_sub->channelName) == 0) {

						if (preverified_user == 1) {
							cei_sub->active = 1;
						}
						ast_debug(LOG_DEBUG, "[WL] This is a join event for me, send me the full state notify.");
						flagFullNotify = 1;

						// Also set up the user role
						struct ast_sip_endpoint *endpoint = ast_sip_subscription_get_endpoint(cei_sub->sip_sub);
						if (endpoint == NULL) {
							goto cleanup_tree;
						}
						struct sip_conference* conf = ast_sip_cei_find_peer_conference(endpoint, conference_name);
						if (conf == NULL) {
							ao2_cleanup(endpoint);
							goto cleanup_tree;
						}
						conf->ceiRole = user_role;
						ao2_cleanup(endpoint);
					}
				}
				break;
				case CEI_TYPE_EXTERNAL_MCB: {
					if (user_name == NULL || cei_sub->self_conf_id == NULL) {
						goto cleanup_tree;
					}
					if (strcmp(user_name, cei_sub->self_conf_id) == 0) {
						ast_debug(LOG_DEBUG, "[WL] This is a join event for me, send me the full state notify.");
						flagFullNotify = 1;
					}
				}
				break;
				default : {
					ast_debug(LOG_DEBUG, "[WL] Invalid subsciption type");
				}
			}
		}
		// Stop sending notify for the user who leave the conf
		else if (strcmp(event_type, CEI_NOTIFY_EVENT_LEAVE) == 0 && user_type == cei_sub->user_type) {
			ast_debug(LOG_DEBUG, "[WL] It's a leave event\n");

			switch (cei_sub->user_type) {
				case CEI_TYPE_NORMAL_USER: {
					if (user_name == NULL || cei_sub->channelName == NULL) {
						goto cleanup_tree;
					}
					if (ast_sip_cei_same_channel(user_name, cei_sub->channelName) == 0) {
						ast_debug(LOG_DEBUG, "[WL] This is leave event for me, return.");
						goto cleanup_tree;
					}
				}
				break;
				case CEI_TYPE_EXTERNAL_MCB: {
					if (user_name == NULL || cei_sub->self_conf_id == NULL) {
						goto cleanup_tree;
					}
					if (strcmp(user_name, cei_sub->self_conf_id) == 0) {
						ast_debug(LOG_DEBUG, "[WL] This is leave event for me, return.");
						goto cleanup_tree;
					}
				}
				break;
				default: {
					ast_debug(LOG_DEBUG, "[WL] Invalid subscription type");
				}
			}
		}
		// Stop sending notify for the mcb conf who publish this notify
		else if (strcmp(event_type, CEI_NOTIFY_EVENT_EXT_MCB) == 0 && user_type == cei_sub->user_type) {
			ast_debug(LOG_DEBUG, "[WL] It's a mcb event\n");
			if (user_name == NULL || cei_sub->self_conf_id == NULL|| strcmp(user_name, cei_sub->self_conf_id) == 0){
				ast_debug(LOG_DEBUG, "[WL] This is mcb event I published, return.");
				goto cleanup_tree;
			}
		}
		//Send full state notify when mcb leave
		else if (strcmp(event_type, CEI_NOTIFY_EVENT_MCB_LEAVE) == 0) {
			ast_debug(LOG_DEBUG, "[WL] This is a mcb leave event, send me the full state notify.");
			flagFullNotify = 1;
		}


		struct pjsip_cei_notify_data *notify_data;

		notify_data = ao2_alloc(sizeof(*notify_data), cei_notify_data_dtor);
		if (!notify_data) {
			goto cleanup_tree;
		}

		notify_data->conference = ast_strdup(conference_name);
		notify_data->notifyVer = version_reset;
		notify_data->content = NULL;
		notify_data->channelName = NULL;

		if (flagFullNotify) {
			if (full_text == NULL) {
				goto cleanup_data;
			}
			notify_data->content = ast_strdup(full_text);
		} else {
			if (partial_text == NULL) {
				goto cleanup_data;
			}
			notify_data->content = ast_strdup(partial_text);
		}

		cei_sub->notify_raw_data = notify_data;
		struct ast_taskprocessor *serializer = ast_sip_subscription_get_serializer(cei_sub->sip_sub);
		ast_sip_push_task_synchronous(serializer, serialized_notify, cei_sub);
		goto cleanup_tree;

cleanup_data :
		ao2_cleanup(notify_data);
cleanup_tree:
		ast_sip_change_tree_ref(cei_sub->sip_sub, -1);
cleanup_ceisub:
		ao2_ref(cei_sub, -1);
	}
	ao2_unlock(cei_sub);
}

static int int_mcb_cei_subscriber_state_change(void *obj)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);

	struct ast_json *json_object = obj;

	// Sanity check
	if (json_object == NULL) {
		return -1;
	}

	const char *content = ast_json_string_get(ast_json_object_get(json_object, "content"));
	const char *mcb_peer_id = ast_json_string_get(ast_json_object_get(json_object, "mcb_peer_id"));

	if (content == NULL || mcb_peer_id == NULL) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return -1;
	}

	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, to_send_json_object, NULL, ast_json_unref);
	struct cei_internal_mcb_peer *mcb_peer = NULL;

	mcb_peer = ast_cei_find_int_mcb_peer_by_id(mcb_peer_id);
	if (mcb_peer == NULL) {
		return -1;
	}
	const char *local_conf = mcb_peer->local_conf;
	const char *publisher_id = mcb_peer->remote_conf_id;
	if (local_conf == NULL || publisher_id == NULL) {
		return -1;
	}

	to_send_json_object = ast_json_pack("{s: s, s: s, s: s}",
		"conf_num", local_conf,
		"content", content,
		"publisher_id", publisher_id
	);

	payload = ast_json_payload_create(to_send_json_object);
	if (NULL == payload) {
		return -1;
	}

	msg = stasis_message_create(ast_cei_mcb_process_string_type(), payload);
	if (msg == NULL) {
		return -1;
	}

	stasis_publish(ast_cei_mcb_process_topic_all(), msg);
	return 0;
}


static void int_mcb_cei_stasis_cb(void *userdata, struct stasis_subscription *sub,
		struct stasis_message *msg)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s \n", __FUNCTION__);

	char *mcb_peer_id = userdata;
	if (mcb_peer_id == NULL) {
		return;
	}

	if (stasis_subscription_final_message(sub, msg)) {
		ast_log(LOG_ERROR, "[WL] Got final stasis message \n");
		return;
	}

	if (ast_cei_notify_type() == stasis_message_type(msg)) {
		ast_log(LOG_ERROR, "[WL] Got internal stasis message\n");

		int flagFullNotify = 0;
		RAII_VAR(struct ast_json *, to_send_json_object, NULL, ast_json_unref);
		struct cei_internal_mcb_peer *mcb_peer = NULL;

		mcb_peer = ast_cei_find_int_mcb_peer_by_id(mcb_peer_id);
		if (mcb_peer == NULL) {
			return;
		}

		const char *local_conf_id = mcb_peer->local_conf_id;
		if (local_conf_id == NULL) {
			return;
		}

    struct ast_json_payload *payload = stasis_message_data(msg);
		struct ast_json *json_object = payload->json;

		if (json_object == NULL) {
			ast_debug(LOG_DEBUG, "[WL] Imcomplete message.");
			return;
		}

		enum cei_user_type user_type = ast_json_integer_get(ast_json_object_get(json_object, "user_type"));
		const char *event_type = ast_json_string_get(ast_json_object_get(json_object, "event_type"));
		const char *user_name = ast_json_string_get(ast_json_object_get(json_object, "user_name"));
		const char *partial_text = ast_json_string_get(ast_json_object_get(json_object, "partial_text"));
		const char *full_text = ast_json_string_get(ast_json_object_get(json_object, "full_text"));

		if ( event_type == NULL ) {
			ast_debug(LOG_DEBUG, "[WL] Imcomplete data. \n");
			return;
		}

		// Send pull state notify to the new joined user
		if (strcmp(event_type, CEI_NOTIFY_EVENT_JOIN) == 0 && user_type == CEI_TYPE_INTERNAL_MCB) {
			if (user_name == NULL) {
				return;
			}
			ast_debug(LOG_DEBUG, "[WL] It's a join event\n");
			if (strcmp(user_name, local_conf_id) == 0){
				ast_debug(LOG_DEBUG, "[WL] It's a join event for me, send me the full notify\n");
				flagFullNotify = 1;
			}
		}
		// Stop sending notify for the mcb conf who publish this notify
		else if (strcmp(event_type, CEI_NOTIFY_EVENT_INT_MCB) == 0 &&  user_type == CEI_TYPE_INTERNAL_MCB){
			ast_debug(LOG_DEBUG, "[WL] It's a mcb event\n");
			if( user_name == NULL) {
				return;
			}
			if (strcmp(user_name, local_conf_id) == 0){
				ast_debug(LOG_DEBUG, "[WL] This is mcb event I published, return.");
				return;
			}
		}
		//Send full state notify when mcb leave
		else if (strcmp(event_type, CEI_NOTIFY_EVENT_MCB_LEAVE) == 0) {
			ast_debug(LOG_DEBUG, "[WL] This is a mcb leave event, send me the full state notify.");
			flagFullNotify = 1;
		}


		if (flagFullNotify) {
			if (full_text == NULL) {
				return;
			}
			to_send_json_object = ast_json_pack("{s: s, s: s }",
				"content", full_text,
				"mcb_peer_id", mcb_peer_id
			);
		} else {
			if (partial_text == NULL) {
				return;
			}
			to_send_json_object = ast_json_pack("{s: s, s: s}",
				"content", partial_text,
				"mcb_peer_id", mcb_peer_id
			);
		}
		if (!to_send_json_object) {
			ast_debug(LOG_DEBUG, "[WL] return here.");
			return;
		}

		ast_sip_push_task_synchronous(NULL, int_mcb_cei_subscriber_state_change, to_send_json_object);

	}
}

static struct stasis_subscription *int_mcb_cei_subscribe_stasis(const char *mcb_peer_id)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct stasis_topic *topic;
	struct stasis_subscription *stasis_sub;

	if (mcb_peer_id == NULL) {
		return NULL;
	}

	struct cei_internal_mcb_peer *mcb_peer = NULL;
	mcb_peer = ast_cei_find_int_mcb_peer_by_id(mcb_peer_id);
	if (mcb_peer == NULL) {
		return NULL;
	}
	const char *remote_conf = mcb_peer->remote_conf;
	if( remote_conf == NULL) {
		return NULL;
	}

	topic = ast_cei_topic(remote_conf);

	if (topic == NULL) {
		ast_log(LOG_DEBUG, "[WL] Can NOT find stasis topic \n");
		return NULL;
	}

	stasis_sub = stasis_subscribe(topic, int_mcb_cei_stasis_cb, mcb_peer->ref_id);

	return stasis_sub;
}

static void cei_subscriber_state_change(struct ast_sip_subscription *sub,
										pjsip_msg_body *body,
										enum pjsip_evsub_state state)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);

	char dest_to[256];
	size_t content_size;
	char *content_type = NULL;
	char *content_subtype = NULL;
	char *content = NULL;
	char *decoded_content = NULL;
	char *uncompressed_content = NULL;
	struct cei_external_mcb_peer* mcb_peer = NULL;

	RAII_VAR(struct ast_sip_endpoint *, endpoint, ast_sip_subscription_get_endpoint(sub), ao2_cleanup);
	if (!endpoint) {
		return;
	}

	const char *endpoint_id = (endpoint != NULL) ? ast_sorcery_object_get_id(endpoint) : NULL;
	if (endpoint_id == NULL) {
		return;
	}

	/* check content type*/
	content_size = pj_strlen(&body->content_type.type) + 1;
	content_type = alloca(content_size);
	ast_copy_pj_str(content_type, &body->content_type.type, content_size);

	content_size = pj_strlen(&body->content_type.subtype) + 1;
	content_subtype = alloca(content_size);
	ast_copy_pj_str(content_subtype, &body->content_type.subtype, content_size);

  if (strcmp(content_type, CEI_TYPE) || strcmp(content_subtype, CEI_SUBTYPE)) {
		ast_log(LOG_WARNING, "[WL] Invalid type! \n");
		return;
	}

	// Get conference info
	// CEI notify Contact : Dest_conf_num@from_host
	if (!ast_strlen_zero(endpoint->cei_contactuser)) {
		strncpy(dest_to, endpoint->cei_contactuser, sizeof(dest_to));
	} else {
		strncpy(dest_to, endpoint->cei_localuser, sizeof(dest_to));
	}

	content_size = pj_strlen((pj_str_t *)&body->data) + 1;
	content = alloca(content_size);
	ast_copy_pj_str(content, (pj_str_t *)&body->data, content_size);

	if (dest_to == NULL || content == NULL) {
		return;
	}

	mcb_peer = ast_cei_find_external_mcb_peer(endpoint, dest_to, endpoint->cei_fromuser,
											  endpoint_id, endpoint->cei_fromdomain);
	if (!mcb_peer) {
		return;
	}

	decoded_content = (char*) ast_malloc(PJSIP_NOTIFY_MAX_LEN);
	uncompressed_content = (char*) ast_malloc(PJSIP_NOTIFY_MAX_LEN);

	if (decoded_content == NULL || uncompressed_content == NULL) {
		return;
	}

	int isCEIContentEncoded = endpoint->isCEIContentEncoded;
	// Decode and umcompress When Content-Encoding is included in header
	if (isCEIContentEncoded == 1) {
		ast_log(LOG_WARNING, "[WL] Received encoded cei notify, start enconding and uncompression! \n");

		//Base64 Decoding
		ast_base64decode((unsigned char *)decoded_content, content, PJSIP_NOTIFY_MAX_LEN);

		//Zlib Uncompression
		z_stream infstream;
		infstream.zalloc = Z_NULL;
		infstream.zfree = Z_NULL;
		infstream.opaque = Z_NULL;
		infstream.avail_in = (uInt)(PJSIP_NOTIFY_MAX_LEN);
		infstream.next_in = (Bytef *)decoded_content;
		infstream.avail_out = (uInt)(PJSIP_NOTIFY_MAX_LEN);
		infstream.next_out = (Bytef *)uncompressed_content;

		inflateInit(&infstream);
		inflate(&infstream, Z_NO_FLUSH);
		inflateEnd(&infstream);

		ast_log(LOG_DEBUG, "[WL] Got uncompressed data : %s \n", uncompressed_content);

		if (uncompressed_content == NULL) {
	        ast_free(decoded_content);
			return;
		}
	}

	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);

	ast_log(LOG_DEBUG, "[WL] Prepare to send data to: %s, publisher: %s, content: %s\n",
	                    mcb_peer->local_conf_private_num,
											mcb_peer->conf_private_num,
											(isCEIContentEncoded == 1)?uncompressed_content:content);

	json_object = ast_json_pack("{s: s, s: s, s: s, s: s}",
		"conf_num", mcb_peer->local_conf_private_num,
		"xml", (isCEIContentEncoded == 1)?uncompressed_content:content,
		"publisher_id", mcb_peer->remote_conf_id,
		"publisher_private_cid", mcb_peer->conf_private_num
	);

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
		return;
	}

	msg = stasis_message_create(ast_cei_mcb_process_xml_type(), payload);
	if (msg == NULL) {
        ast_free(decoded_content);
        ast_free(uncompressed_content);
		return;
	}

    stasis_publish(ast_cei_mcb_process_topic_all(), msg);

    // Clean up
    ast_free(decoded_content);
    ast_free(uncompressed_content);
}

static void cei_ext_mcb_sub_update_endpoint(struct ast_sip_endpoint *endpoint,
											struct cei_external_mcb_peer *mcb_peer)
{
	const char *eid = NULL;
	char trunk_username[40];
	struct ast_str *database_tree = NULL;

	memset(trunk_username, 0, sizeof(trunk_username));

	// Sanity check
	if (endpoint == NULL || mcb_peer == NULL) {
		return;
	}

	eid = ast_sorcery_object_get_id(endpoint);

	// Get trunk username
	if (eid != NULL && strstr(eid, "trunk_") != NULL) {
		database_tree = ast_str_create(20);
		ast_str_append(&database_tree, 0, "TRUNK/SIP_%s", eid+6);
		ast_ucm_db_get(ast_str_buffer(database_tree), "USER_NAME", trunk_username, sizeof(trunk_username));
		ast_free(database_tree);
	}

	// Use trunk user as FROM user when a user makes a call via SIP trunk registration
	if (strlen(trunk_username)) {
		ast_string_field_set(endpoint, cei_ext_fromuser, trunk_username);
	}
	else {
		ast_string_field_set(endpoint, cei_ext_fromuser, mcb_peer->local_conf);
	}
	ast_string_field_set(endpoint, cei_contactuser, mcb_peer->local_conf);

	// Update X-Grandstream-Transport type
	ast_string_field_set(endpoint, transport_type, "trunk");
}

static int cei_ext_mcb_sub(void *obj)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct ast_json *json_object = obj;
	const char *peer_name = NULL;
	const char *mcb_peer_id = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
	struct cei_external_mcb_peer *mcb_peer = NULL;

	// Sanity check
	if (json_object == NULL) {
		return -1;
	}

	peer_name = ast_json_string_get(ast_json_object_get(json_object, "peer_name"));
	mcb_peer_id = ast_json_string_get(ast_json_object_get(json_object, "mcb_peer_id"));
	if (peer_name == NULL || mcb_peer_id == NULL) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return -1;
	}

	ast_log(LOG_DEBUG, "[TK] peer_name: %s, mcb_peer_id: %s\n", peer_name, mcb_peer_id);

	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", peer_name);
	if (endpoint == NULL) {
		ast_log(LOG_WARNING, "Failed to get endpoint of peer name %s", peer_name);
		return -1;
	}

	mcb_peer = ast_cei_find_external_mcb_peer_by_ref_id(endpoint, mcb_peer_id);
	if (mcb_peer == NULL) {
		ast_log(LOG_WARNING, "Failed to get MCB peer with id %s", mcb_peer_id);
		return -1;
	}

	cei_ext_mcb_sub_update_endpoint(endpoint, mcb_peer);

	mcb_peer->sip_sub = ast_sip_create_subscription(&cei_handler, endpoint, mcb_peer->conf, -1);
	if (mcb_peer->sip_sub == NULL) {
		ast_log(LOG_WARNING, "Failed to create CEI external MCB SIP subscription");
		return -1;
	}

	//Update cei_sub
	const char *endpoint_id = ast_sorcery_object_get_id(endpoint);
	struct cei_subscription *cei_sub;

	ao2_lock(cei_subs);

	struct ao2_iterator cei_sub_iter = ao2_iterator_init(cei_subs, 0);
	while ((cei_sub = ao2_iterator_next(&cei_sub_iter))) {
		if (strcmp(endpoint_id, cei_sub->id) == 0 && strcmp(mcb_peer->conf, cei_sub->user_name) == 0
				&& strcmp(mcb_peer->local_conf_private_num, cei_sub->local_conference) == 0){
			ao2_ref(cei_sub, -1);
			break;
		}
		ao2_ref(cei_sub, -1);
	}

  ao2_iterator_destroy(&cei_sub_iter);
	ao2_unlock(cei_subs);

	if (cei_sub == NULL) {
		ast_log(LOG_DEBUG, "[TK] NO CEI object \n");
		return 0;
	}

	if (!ast_strlen_zero(mcb_peer->remote_conf_id) && cei_sub->self_conf_id == NULL) {
		cei_sub->self_conf_id = ast_strdup(mcb_peer->remote_conf_id);
	}
	if (!ast_strlen_zero(mcb_peer->channelName) && cei_sub->channelName == NULL) {
		cei_sub->channelName = ast_strdup(mcb_peer->channelName);
	}

	ast_log(LOG_DEBUG, "[TK] Exit...%s\n", __FUNCTION__);
	return 0;
}

static int cei_ext_mcb_unsub(void *obj)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct ast_json *json_object = obj;
	const char *peer_name = NULL;
	const char *mcb_peer_id = NULL;
	const char *remote_conf_id = NULL;
	const char *local_conf_id = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
	struct cei_external_mcb_peer *mcb_peer = NULL;

	// Sanity check
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "Failed to call ast_cei_remove_external_mcb_peer");

		return -1;
	}

	peer_name = ast_json_string_get(ast_json_object_get(json_object, "peer_name"));
	mcb_peer_id = ast_json_string_get(ast_json_object_get(json_object, "mcb_peer_id"));
	remote_conf_id = ast_json_string_get(ast_json_object_get(json_object, "remote_conf_id"));
	local_conf_id = ast_json_string_get(ast_json_object_get(json_object, "local_conf_id"));

	if (peer_name == NULL || mcb_peer_id == NULL || remote_conf_id == NULL || local_conf_id == NULL) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		ast_log(LOG_WARNING, "Failed to call ast_cei_remove_external_mcb_peer");

		return -1;
	}

	ast_log(LOG_DEBUG, "[TK] peer_name: %s, mcb_peer_id: %s, remote_conf_id: %s, local_conf_id : %s\n",
												peer_name,
												mcb_peer_id,
												remote_conf_id,
												local_conf_id);

	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", peer_name);
	if (endpoint == NULL) {
		ast_log(LOG_WARNING, "Failed to get endpoint of peer name %s", peer_name);
		ast_log(LOG_WARNING, "Failed to call ast_cei_remove_external_mcb_peer");

		return -1;
	}

	mcb_peer = ast_cei_find_external_mcb_peer_by_ref_id(endpoint, mcb_peer_id);

	if (mcb_peer == NULL) {
		ast_log(LOG_WARNING, "Failed to get MCB peer with id %s", mcb_peer_id);
		ast_cei_remove_external_mcb_peer(endpoint, local_conf_id, remote_conf_id);

		return -1;
	}

	if (mcb_peer->sip_sub == NULL) {
		ast_log(LOG_WARNING, "No SIP SUBSCRIBE of MCB peer id %s", mcb_peer_id);
		ast_cei_remove_external_mcb_peer(endpoint, local_conf_id, remote_conf_id);

		return -1;
	}

	// Send leave event to CEI manager for MCB
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, to_send_json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

	to_send_json_object = ast_json_pack("{s: s, s: s}",
		"conf_id", local_conf_id,
		"publisher", remote_conf_id
	);

	payload = ast_json_payload_create(to_send_json_object);
	if (NULL == payload) {
		return -1;
	}

	msg = stasis_message_create(ast_cei_mcb_process_unsub_type(), payload);
	if (msg == NULL) {
		ast_cei_remove_external_mcb_peer(endpoint, local_conf_id, remote_conf_id);

		return -1;
	}

	stasis_publish(ast_cei_mcb_process_topic_all(), msg);

	// Clean up
	ast_cei_remove_external_mcb_peer(endpoint, local_conf_id, remote_conf_id);

	ast_log(LOG_DEBUG, "[TK] Exit...%s\n", __FUNCTION__);
	return 0;
}

static void cei_ext_mcb_sub_cb(void *data, struct stasis_subscription *sub,
							   struct stasis_message *message)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct ast_json_payload *payload	 = NULL;
	struct ast_json			*json_object = NULL;

	// Sanity check
	if (message == NULL) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	if (ast_sip_push_task_synchronous(NULL, cei_ext_mcb_sub, json_object)) {
		ast_log(LOG_WARNING, "Failed to create CEI external MCB subscription");
		return;
	}

	ast_log(LOG_DEBUG, "[TK] Exit...%s\n", __FUNCTION__);
}

static void cei_ext_mcb_unsub_cb(void *data, struct stasis_subscription *sub,
							   struct stasis_message *message)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct ast_json_payload *payload		= NULL;
	struct ast_json			*json_object	= NULL;

	// Sanity check
	if (message == NULL) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		ast_log(LOG_WARNING, "Failed to call ast_cei_remove_external_mcb_peer");
		return;
	}

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		ast_log(LOG_WARNING, "Failed to call ast_cei_remove_external_mcb_peer");
		return;
	}

	if (ast_sip_push_task_synchronous(NULL, cei_ext_mcb_unsub, json_object)) {
		ast_log(LOG_WARNING, "Failed to create CEI external MCB unsubscription");
		ast_log(LOG_WARNING, "Failed to call ast_cei_remove_external_mcb_peer");
		return;
	}

	ast_log(LOG_DEBUG, "[TK] Exit...%s\n", __FUNCTION__);
}

static int cei_int_mcb_sub(void *obj)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct ast_json *json_object = obj;
	struct cei_internal_mcb_peer *mcb_peer = NULL;

	// Sanity check
	if (json_object == NULL) {
		return -1;
	}

	const char *mcb_peer_id = ast_json_string_get(ast_json_object_get(json_object, "mcb_peer_id"));
	if (mcb_peer_id == NULL) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return -1;
	}

	mcb_peer = ast_cei_find_int_mcb_peer_by_id(mcb_peer_id);
	if (mcb_peer == NULL) {
		return -1;
	}

	RAII_VAR(struct stasis_subscription *, stasis_sub,
		int_mcb_cei_subscribe_stasis(mcb_peer_id), ao2_cleanup);
	if (stasis_sub) {
		mcb_peer->stasis_sub = ao2_bump(stasis_sub);
	} else {
		return -1;
	}

	ast_log(LOG_DEBUG, "[TK] Exit...%s\n", __FUNCTION__);
	return 0;
}

static int cei_int_mcb_unsub(void *obj)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct ast_json *json_object = obj;
	struct cei_internal_mcb_peer *mcb_peer = NULL;

	// Sanity check
	if (json_object == NULL) {
		return -1;
	}

	const char *mcb_peer_id = ast_json_string_get(ast_json_object_get(json_object, "mcb_peer_id"));
	if (mcb_peer_id == NULL) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return -1;
	}

	mcb_peer = ast_cei_find_int_mcb_peer_by_id(mcb_peer_id);
	if (mcb_peer == NULL) {
		return -1;
	}

	const char *local_conf_id = mcb_peer->local_conf_id;
	const char *remote_conf_id = mcb_peer->remote_conf_id;

	if (local_conf_id == NULL || remote_conf_id == NULL) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return -1;
	}

	ast_log(LOG_DEBUG, "[TK] local_conf_id: %s, remote_conf_id: %s\n",
												local_conf_id,
												remote_conf_id);


	// Send leave event to CEI manager for MCB
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, to_send_json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

	to_send_json_object = ast_json_pack("{s: s, s: s}",
		"conf_id", local_conf_id,
		"publisher", remote_conf_id
	);

	payload = ast_json_payload_create(to_send_json_object);
	if (NULL == payload) {
		return -1;
	}

	msg = stasis_message_create(ast_cei_mcb_process_unsub_type(), payload);
	if (msg == NULL) {
		return -1;
	}

	stasis_publish(ast_cei_mcb_process_topic_all(), msg);

	//Unsubscribe stasis subscription
	if (mcb_peer->stasis_sub) {
		mcb_peer->stasis_sub = stasis_unsubscribe(mcb_peer->stasis_sub);
	}

	// Remove given subscription in endpoint
	ast_cei_remove_int_mcb_peer(local_conf_id, remote_conf_id);

	ast_log(LOG_DEBUG, "[TK] Exit...%s\n", __FUNCTION__);
	return 0;
}

static void cei_int_mcb_sub_cb(void *data, struct stasis_subscription *sub,
							   struct stasis_message *message)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct ast_json_payload *payload	 = NULL;
	struct ast_json			*json_object = NULL;

	// Sanity check
	if (message == NULL) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	if (ast_sip_push_task_synchronous(NULL, cei_int_mcb_sub, json_object)) {
		ast_log(LOG_WARNING, "Failed to create CEI internal MCB subscription");
		return;
	}

	ast_log(LOG_DEBUG, "[TK] Exit...%s\n", __FUNCTION__);
}

static void cei_int_mcb_unsub_cb(void *data, struct stasis_subscription *sub,
							   struct stasis_message *message)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct ast_json_payload *payload	 = NULL;
	struct ast_json			*json_object = NULL;

	// Sanity check
	if (message == NULL) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	if (ast_sip_push_task_synchronous(NULL, cei_int_mcb_unsub, json_object)) {
		ast_log(LOG_WARNING, "Failed to create CEI internal MCB unsubscription");
		return;
	}

	ast_log(LOG_DEBUG, "[TK] Exit...%s\n", __FUNCTION__);
}

/*! \brief Comparison function used for cei subscriptions container */
static int cei_subscription_cmp(void *obj, void *arg, int flags)
{
	struct cei_subscription *cei_sub = obj;
	char *cei_sub_id = arg;

	if (!strcmp(cei_sub->id, cei_sub_id)) {
		return (CMP_MATCH | CMP_STOP);
	}

	return 0;

}

static int load_module(void)
{
	struct stasis_topic *cei_mcb_sub_topic;

	CHECK_PJSIP_PUBSUB_MODULE_LOADED();

	if (ast_sip_register_subscription_handler(&cei_handler)) {
		ast_log(LOG_WARNING, "Unable to register subscription handler %s\n",
							 cei_handler.event_name);
		return AST_MODULE_LOAD_DECLINE;
	}

	/* Create a container to hold the cei_sub */
	cei_subs = ao2_container_alloc(CEI_SUBSCRIPTION_BUCKETS,
		NULL, cei_subscription_cmp);
	if (!cei_subs) {
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}
	// Handle CEI stasis topics and types
	cei_mcb_sub_topic = ast_cei_mcb_sub_topic_all();
	if (!cei_mcb_sub_topic) {
		ast_log(LOG_WARNING, "Failed to get ast_cei_mcb_sub_topic_all\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	cei_mcb_sub_router = stasis_message_router_create(cei_mcb_sub_topic);
	if (cei_mcb_sub_router == NULL) {
		ast_log(LOG_WARNING, "Failed to create cei_mcb_sub_router\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (stasis_message_router_add(cei_mcb_sub_router,
								  ast_cei_ext_mcb_sub_type(),
								  cei_ext_mcb_sub_cb, NULL))
	{
		ast_log(LOG_WARNING, "Failed to subscribe to ast_cei_ext_mcb_sub_type\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (stasis_message_router_add(cei_mcb_sub_router,
								  ast_cei_ext_mcb_unsub_type(),
								  cei_ext_mcb_unsub_cb, NULL))
	{
		ast_log(LOG_WARNING, "Failed to subscribe to ast_cei_ext_mcb_unsub_type\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (stasis_message_router_add(cei_mcb_sub_router,
									ast_cei_int_mcb_sub_type(),
									cei_int_mcb_sub_cb, NULL))
	{
		ast_log(LOG_WARNING, "Failed to subscribe to ast_cei_int_mcb_sub_type\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (stasis_message_router_add(cei_mcb_sub_router,
								  ast_cei_int_mcb_unsub_type(),
								  cei_int_mcb_unsub_cb, NULL))
	{
		ast_log(LOG_WARNING, "Failed to subscribe to ast_cei_int_mcb_unsub_type\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_unregister_subscription_handler(&cei_handler);

	if (cei_mcb_sub_router) {
		stasis_message_router_unsubscribe(cei_mcb_sub_router);
		cei_mcb_sub_router = NULL;
	}

	/* Get rid of the cei subscription container. Since we only allow dynamic ones none will be active. */
	ao2_cleanup(cei_subs);
	cei_subs = NULL;

	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP CEI resource",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.load_pri = AST_MODPRI_CHANNEL_DEPEND,
);

#endif
