/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Mark Michelson <mmichelson@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#ifdef GRANDSTREAM_NETWORKS
#define AST_MODULE_LOG "pjsip"
#endif

#include "asterisk.h"

#ifdef GRANDSTREAM_NETWORKS
ASTERISK_FILE_VERSION(__FILE__, "$Revision: 13.0 $");
#endif

#include <pjsip.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_cli.h"
#include "asterisk/module.h"
#include "asterisk/acl.h"
#include "asterisk/manager.h"
#include "res_pjsip/include/res_pjsip_private.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/linkedlists.h"

struct srv_entry {
	unsigned short priority;
	unsigned short weight;
	unsigned short port;
	unsigned int weight_sum;
	AST_LIST_ENTRY(srv_entry) list;
	char host[1];
};

struct srv_context {
	unsigned int have_weights:1;
	struct srv_entry *prev;
	unsigned int num_records;
	AST_LIST_HEAD_NOLOCK(srv_entries, srv_entry) entries;
};

static struct ast_sched_context *dns_sched = NULL;

#define IP_IDENTIFY_DNS_REFRESH_INTERVAL   300
#endif

/*** DOCUMENTATION
	<configInfo name="res_pjsip_endpoint_identifier_ip" language="en_US">
		<synopsis>Module that identifies endpoints via source IP address</synopsis>
		<configFile name="pjsip.conf">
			<configObject name="identify">
				<synopsis>Identifies endpoints via source IP address</synopsis>
				<configOption name="endpoint">
					<synopsis>Name of Endpoint</synopsis>
				</configOption>
				<configOption name="match">
					<synopsis>IP addresses or networks to match against</synopsis>
					<description><para>
						The value is a comma-delimited list of IP addresses. IP addresses may
						have a subnet mask appended. The subnet mask may be written in either
						CIDR or dot-decimal notation. Separate the IP address and subnet
						mask with a slash ('/')
					</para></description>
				</configOption>
				<configOption name="type">
					<synopsis>Must be of type 'identify'.</synopsis>
				</configOption>
			</configObject>
		</configFile>
	</configInfo>
 ***/

/*! \brief Structure for an IP identification matching object */
struct ip_identify_match {
	/*! \brief Sorcery object details */
	SORCERY_OBJECT(details);
	/*! \brief Stringfields */
	AST_DECLARE_STRING_FIELDS(
		/*! The name of the endpoint */
		AST_STRING_FIELD(endpoint_name);
#ifdef GRANDSTREAM_NETWORKS
		AST_STRING_FIELD(host);   /*! the host of identify */
		AST_STRING_FIELD(service);/*! the service of dnssrv */
#endif
	);
	/*! \brief Networks or addresses that should match this */
	struct ast_ha *matches;
#ifdef GRANDSTREAM_NETWORKS
	int sched_id;    /*! the sched to refresh dns resolve*/
#endif
};

/*! \brief Destructor function for a matching object */
static void ip_identify_destroy(void *obj)
{
	struct ip_identify_match *identify = obj;

	ast_string_field_free_memory(identify);
	ast_free_ha(identify->matches);

#ifdef GRANDSTREAM_NETWORKS
	AST_SCHED_DEL(dns_sched, identify->sched_id);
#endif
}

/*! \brief Allocator function for a matching object */
static void *ip_identify_alloc(const char *name)
{
	struct ip_identify_match *identify = ast_sorcery_generic_alloc(sizeof(*identify), ip_identify_destroy);

	if (!identify || ast_string_field_init(identify, 256)) {
		ao2_cleanup(identify);
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	identify->sched_id = -1;
#endif
	return identify;
}

/*! \brief Comparator function for a matching object */
static int ip_identify_match_check(void *obj, void *arg, int flags)
{
	struct ip_identify_match *identify = obj;
	struct ast_sockaddr *addr = arg;
	int sense;

	sense = ast_apply_ha(identify->matches, addr);
	if (sense != AST_SENSE_ALLOW) {
		ast_debug(3, "Source address %s matches identify '%s'\n",
				ast_sockaddr_stringify(addr),
				ast_sorcery_object_get_id(identify));
		return CMP_MATCH | CMP_STOP;
	} else {
		ast_debug(3, "Source address %s does not match identify '%s'\n",
				ast_sockaddr_stringify(addr),
				ast_sorcery_object_get_id(identify));
		return 0;
	}
}

#ifdef GRANDSTREAM_NETWORKS
static struct ast_sip_endpoint *ip_identify_match_contact(const char* contact_str)
{
	struct ast_sip_endpoint *pstEndpoint = NULL;
	struct ao2_iterator      j;
	struct ast_sip_contact  *contact;
	int    found_endpoint = 0;

	char *contact_id     = NULL;
	char *uri_start      = NULL;
	char endpoint_id[80] = {0};
	char contact_uri[80] = {0};
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);

	if (!(contacts = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "contact", AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL)) ||
		!ao2_container_count(contacts)) {
		ast_log(LOG_NOTICE, "No contact sections to match against\n");
		return NULL;
	}

	j = ao2_iterator_init(contacts, 0);
	while ((contact = ao2_t_iterator_next(&j, "iterate contacts table"))) {
		memset(contact_uri, 0, sizeof(contact_uri));
		memset(endpoint_id, 0, sizeof(endpoint_id));

		contact_id = (char *) ast_sorcery_object_get_id(contact);

		uri_start = strstr(contact_id, "sip:");
		if (!uri_start) {
			uri_start = strstr(contact_id, "sips:");
		}
		if (uri_start) {
			sscanf(uri_start, "%79[^@]@%79[^;]", endpoint_id, contact_uri);
		}

		if (!strcmp(contact_str, contact_uri)) {
			sscanf(contact_id, "%79[^;]", endpoint_id);
			found_endpoint = 1;
			ao2_ref(contact, -1);
			break;
		}

		ao2_ref(contact, -1);
	}
	ao2_iterator_destroy(&j);

	if (found_endpoint) {
		pstEndpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", endpoint_id);
		return pstEndpoint;
	}
	return NULL;
}

static int get_endpoint_details(pjsip_rx_data *rdata, char *endpoint, size_t endpoint_size, char *domain, size_t domain_size)
{
    pjsip_uri *from = rdata->msg_info.from->uri;
    pjsip_sip_uri *sip_from;
    if (!PJSIP_URI_SCHEME_IS_SIP(from) && !PJSIP_URI_SCHEME_IS_SIPS(from)) {
            return -1;
        }
    sip_from = (pjsip_sip_uri *) pjsip_uri_get_uri(from);
    ast_copy_pj_str(endpoint, &sip_from->user, endpoint_size);
    ast_copy_pj_str(domain, &sip_from->host, domain_size);
    return 0;
}
#endif

static struct ast_sip_endpoint *ip_identify(pjsip_rx_data *rdata)
{
	struct ast_sockaddr addr = { { 0, } };
	RAII_VAR(struct ao2_container *, candidates, NULL, ao2_cleanup);
	RAII_VAR(struct ip_identify_match *, match, NULL, ao2_cleanup);
	struct ast_sip_endpoint *endpoint;

	/* If no possibilities exist return early to save some time */
	if (!(candidates = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "identify", AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL)) ||
		!ao2_container_count(candidates)) {
		ast_debug(3, "No identify sections to match against\n");
#ifdef GRANDSTREAM_NETWORKS
		if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG && 
			rdata->msg_info.msg->line.req.method.id == PJSIP_INVITE_METHOD) {
			ast_sockaddr_parse(&addr, rdata->pkt_info.src_name, PARSE_PORT_FORBID);
			ast_sockaddr_set_port(&addr, rdata->pkt_info.src_port);

			endpoint = ip_identify_match_contact(ast_sockaddr_stringify(&addr));
			if (endpoint) {
				ast_debug(3, "Retrieved endpoint %s from contact\n", ast_sorcery_object_get_id(endpoint));
				return endpoint;
			}
		}
#endif
		return NULL;
	}

	ast_sockaddr_parse(&addr, rdata->pkt_info.src_name, PARSE_PORT_FORBID);
	ast_sockaddr_set_port(&addr, rdata->pkt_info.src_port);

	if (!(match = ao2_callback(candidates, 0, ip_identify_match_check, &addr))) {
		ast_debug(3, "'%s' did not match any identify section rules\n",
				ast_sockaddr_stringify(&addr));
#ifdef GRANDSTREAM_NETWORKS
		if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG && 
		    rdata->msg_info.msg->line.req.method.id == PJSIP_INVITE_METHOD) {
			endpoint = ip_identify_match_contact(ast_sockaddr_stringify(&addr));
			if (endpoint) {
				char endpoint_name[64], domain_name[64];
				get_endpoint_details(rdata, endpoint_name, sizeof(endpoint_name), domain_name, sizeof(domain_name));
				if (strcasecmp(endpoint_name, "anonymous") == 0) {
					 ast_string_field_set(endpoint, anonymous, "anonymous");
				}
				ast_debug(3, "Retrieved endpoint %s from contact\n", ast_sorcery_object_get_id(endpoint));
				return endpoint;
			}
		}
#endif
		return NULL;
	}

	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", match->endpoint_name);
	if (endpoint) {
		ast_debug(3, "Retrieved endpoint %s\n", ast_sorcery_object_get_id(endpoint));
	} else {
		ast_log(LOG_WARNING, "Identify section '%s' points to endpoint '%s' but endpoint could not be looked up\n",
				ast_sorcery_object_get_id(match), match->endpoint_name);
	}

	return endpoint;
}

#ifdef GRANDSTREAM_NETWORKS
static char *ip_identify_get_srcaddr_next(pjsip_rx_data *rdata, char* endpoint_id) 
{
	struct ast_ha *match;
	struct ast_sockaddr addr = { { 0, } };
	char   identify_id[80] = {0};
	RAII_VAR(struct ip_identify_match *, identify, NULL, ao2_cleanup);

	if (!rdata) {
		return NULL;
	}

	snprintf(identify_id, sizeof(identify_id), "id_%s", endpoint_id);

	identify = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "identify", identify_id);
	if (!identify) {
		ast_log(LOG_NOTICE, "ip_identify_get_address_next, identify[%s] is NULL", identify_id);
		return NULL;
	}

	ast_sockaddr_parse(&addr, rdata->pkt_info.src_name, PARSE_PORT_FORBID);
	ast_sockaddr_set_port(&addr, rdata->pkt_info.src_port);
	for (match = identify->matches; match; match = match->next) {
		if (!ast_sockaddr_cmp_addr(&match->addr, &addr)) {
			break;
		}
	}

	if (match && match->next) {
		return ast_sockaddr_stringify(&match->next->addr);
	}

	ast_log(LOG_NOTICE, "ip_identify_get_address_next, next is NULL");
	return NULL;
}
#endif

static struct ast_sip_endpoint_identifier ip_identifier = {
	.identify_endpoint = ip_identify,
#ifdef GRANDSTREAM_NETWORKS
	.identify_address_next = ip_identify_get_srcaddr_next,
#endif
};

#ifdef GRANDSTREAM_NETWORKS

/* returns   1: network address was successfully converted
             0: is returned if src does not contain a character string representing a valid network address in the specified address family.
            -1: af does not contain a valid address family
*/
static int is_a_valid_IP_address(const char *str)
{
	struct in_addr addr;

	return inet_pton(AF_INET, str, &addr);
}

static int add_ip_into_identify(const char *hostname, void *obj, int need_reset)
{
	struct ip_identify_match *identify = obj;
	struct ast_sockaddr *addrs;
	int num_addrs = 0, error = 0, i;

	if (ast_strlen_zero(hostname)) {
		return 0;
	}

	num_addrs = ast_sockaddr_resolve(&addrs, hostname, PARSE_PORT_FORBID, AST_AF_UNSPEC);
	if (!num_addrs) {
		ast_log(LOG_ERROR, "Address '%s' provided on ip endpoint identifier '%s' did not resolve to any address\n",
			hostname, ast_sorcery_object_get_id(obj));
		return -1;
	}

	if (need_reset) {
		ast_free_ha(identify->matches);
		identify->matches = NULL;
	}

	for (i = 0; i < num_addrs; ++i) {
		/* We deny what we actually want to match because there is an implicit permit all rule for ACLs */
		identify->matches = ast_append_ha("d", ast_sockaddr_stringify_addr(&addrs[i]), identify->matches, &error);

		if (!identify->matches || error) {
			ast_log(LOG_ERROR, "Failed to add address '%s' to ip endpoint identifier '%s'\n",
				ast_sockaddr_stringify_addr(&addrs[i]), ast_sorcery_object_get_id(obj));
			error = -1;
			break;
		}
	}

	ast_free(addrs);

	if (error) {
		return -1;
	}

	return 0;
}


/*! \brief do dns resolve every 5 mins */
static int ip_identify_refresh_timeout(const void *data)
{
	struct ip_identify_match *identify = (struct ip_identify_match *)data;
	struct ip_identify_match *match;
	struct srv_context context = { .entries = AST_LIST_HEAD_NOLOCK_INIT_VALUE };
	struct srv_entry *current;
	int    need_reset = 1;
	int    found      = 0;
	struct ao2_iterator j;
	RAII_VAR(struct ao2_container *, candidates, NULL, ao2_cleanup);

	if (NULL == data) {
		return 0;
	}

	if (!(candidates = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "identify", AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL)) ||
		!ao2_container_count(candidates)) {
		ast_debug(3, "No identify sections to refresh\n");
		return 0;
	}

	j = ao2_iterator_init(candidates, 0);
	while ((match = ao2_t_iterator_next(&j, "iterate identify table"))) {
		if (match == identify) {
			found = 1;
			ao2_ref(match, -1);
			break;
		}
		ao2_ref(match, -1);
	}
	ao2_iterator_destroy(&j);

	if (!found) {
		ast_log(LOG_WARNING, "No identify matched to refresh\n");
		return 0;
	}

	if (ast_strlen_zero(identify->host) && ast_strlen_zero(identify->service)) {
		return 0;
	}

	ao2_ref(identify, +1);
	ast_log(LOG_NOTICE, "ip_identify_refresh_timeout, service[%s]", identify->service);
	if (!ast_strlen_zero(identify->service) && ast_get_srv_all(NULL, identify->service, &context) > 0) {
		while ((current = AST_LIST_REMOVE_HEAD(&context.entries, list))) {
			ast_log(LOG_NOTICE, "SRV lookup for '%s' mapped to host %s, port %d\n", identify->service, current->host, current->port);
			if (!add_ip_into_identify(current->host, identify, need_reset)) {
				need_reset = 0;
			}
			ast_free(current);
		}
	} else {
		ast_log(LOG_NOTICE, "ip_identify_refresh_timeout, hostname[%s]", identify->host);
		add_ip_into_identify(identify->host, identify, need_reset);
	}

	identify->sched_id = ast_sched_add(dns_sched, IP_IDENTIFY_DNS_REFRESH_INTERVAL*1000, ip_identify_refresh_timeout, identify);
	ao2_ref(identify, -1);
	return 0;
}

/*! \brief Custom handler for match field */
static int ip_identify_match_handler(const struct aco_option *opt, struct ast_variable *var, void *obj)
{
	struct ip_identify_match *identify = obj;
	char *input_string = ast_strdupa(var->value);
	char *current_string;
	char *mask = NULL;
	char tmp_current[MAXHOSTNAMELEN] = {0};
	char *tmp_port = NULL;
	char *tmp_transport = NULL;
	int error = 0;

	if (ast_strlen_zero(var->value)) {
		return 0;
	}

	while ((current_string = strsep(&input_string, ","))) {
		char transport[MAXHOSTNAMELEN] = {0};
		char *tmp_ts = NULL;

		snprintf(tmp_current, MAXHOSTNAMELEN, "%s", current_string);

		tmp_port = strrchr(tmp_current, ':');
		tmp_transport = strstr(tmp_current, ";transport=");

		if (tmp_port == NULL) {
			tmp_ts = tmp_transport ? tmp_transport + strlen(";transport=") : NULL;
			if (tmp_ts && !strcmp(tmp_ts, "TLS")) {
				strncpy(transport, "_sips._tcp", MAXHOSTNAMELEN); 
			} else {
				snprintf(transport, MAXHOSTNAMELEN, "_SIP._%s", tmp_ts ? tmp_ts : "UDP"); 
			}

			if (tmp_transport != NULL) {
				*tmp_transport = '\0';
			}
		}
		else {
			*tmp_port = '\0';
		}

		mask = strrchr(tmp_current, '/');
		if (mask) {
			identify->matches = ast_append_ha("d", tmp_current, identify->matches, &error);

			if (!identify->matches || error) {
				ast_log(LOG_ERROR, "Failed to add address '%s' to ip endpoint identifier '%s'\n",
					tmp_current, ast_sorcery_object_get_id(obj));
				return -1;
			}

			continue;
		}

		ast_log(LOG_VERBOSE, "Transport[%s], host[%s]", transport, tmp_current);

		/*
		* If it's actually an IP address and not a name, there's no
		* need for a managed lookup.
		*/
		if (is_a_valid_IP_address(tmp_current) < 1) {
			struct srv_context context = { .entries = AST_LIST_HEAD_NOLOCK_INIT_VALUE };
			struct srv_entry *current;
			int    srv_flag = 0;

			if (transport[0] != '\0') {
				srv_flag = 1;
				snprintf(transport + strlen(transport), MAXHOSTNAMELEN - strlen(transport), ".%s", tmp_current); 
				ast_string_field_set(identify, service, transport);
			}

			ast_string_field_set(identify, host, tmp_current);
			identify->sched_id = ast_sched_add(dns_sched, IP_IDENTIFY_DNS_REFRESH_INTERVAL*1000, ip_identify_refresh_timeout, identify);

			if (srv_flag && (ast_get_srv_all(NULL, transport, &context) > 0)) {
				while ((current = AST_LIST_REMOVE_HEAD(&context.entries, list))) {
					ast_log(LOG_VERBOSE, "SRV lookup for '%s' mapped to host %s, port %d\n", transport, current->host, current->port);
					add_ip_into_identify(current->host, obj, 0);
					ast_free(current);
				}
				continue;
			}
		}

		add_ip_into_identify(tmp_current, obj, 0);
	}

	return 0;
}
#else

/*! \brief Custom handler for match field */
static int ip_identify_match_handler(const struct aco_option *opt, struct ast_variable *var, void *obj)
{
	struct ip_identify_match *identify = obj;
	char *input_string = ast_strdupa(var->value);
	char *current_string;

	if (ast_strlen_zero(var->value)) {
		return 0;
	}

	while ((current_string = strsep(&input_string, ","))) {
		struct ast_sockaddr *addrs;
		int num_addrs = 0, error = 0, i;
		char *mask = strrchr(current_string, '/');

		if (mask) {
			identify->matches = ast_append_ha("d", current_string, identify->matches, &error);

			if (!identify->matches || error) {
				ast_log(LOG_ERROR, "Failed to add address '%s' to ip endpoint identifier '%s'\n",
					current_string, ast_sorcery_object_get_id(obj));
				return -1;
			}

			continue;
		}

		num_addrs = ast_sockaddr_resolve(&addrs, current_string, PARSE_PORT_FORBID, AST_AF_UNSPEC);

		if (!num_addrs) {
			ast_log(LOG_ERROR, "Address '%s' provided on ip endpoint identifier '%s' did not resolve to any address\n",
				var->value, ast_sorcery_object_get_id(obj));
			return -1;
		}

		for (i = 0; i < num_addrs; ++i) {
			/* We deny what we actually want to match because there is an implicit permit all rule for ACLs */
			identify->matches = ast_append_ha("d", ast_sockaddr_stringify_addr(&addrs[i]), identify->matches, &error);

			if (!identify->matches || error) {
				ast_log(LOG_ERROR, "Failed to add address '%s' to ip endpoint identifier '%s'\n",
					ast_sockaddr_stringify_addr(&addrs[i]), ast_sorcery_object_get_id(obj));
				error = -1;
				break;
			}
		}

		ast_free(addrs);

		if (error) {
			return -1;
		}
	}

	return 0;
}
#endif

static int match_to_str(const void *obj, const intptr_t *args, char **buf)
{
	RAII_VAR(struct ast_str *, str, ast_str_create(MAX_OBJECT_FIELD), ast_free);
	const struct ip_identify_match *identify = obj;

	ast_ha_join(identify->matches, &str);
	*buf = ast_strdup(ast_str_buffer(str));
	return 0;
}

static int match_to_var_list(const void *obj, struct ast_variable **fields)
{
	char str[MAX_OBJECT_FIELD];
	const struct ip_identify_match *identify = obj;
	struct ast_variable *head = NULL;
	struct ast_ha *ha = identify->matches;

	for (; ha; ha = ha->next) {
		const char *addr = ast_strdupa(ast_sockaddr_stringify_addr(&ha->addr));
		snprintf(str, MAX_OBJECT_FIELD, "%s%s/%s", ha->sense == AST_SENSE_ALLOW ? "!" : "",
			addr, ast_sockaddr_stringify_addr(&ha->netmask));

		ast_variable_list_append(&head, ast_variable_new("match", str, ""));

	}

	if (head) {
		*fields = head;
	}

	return 0;
}

static int sip_identify_to_ami(const struct ip_identify_match *identify,
			       struct ast_str **buf)
{
	return ast_sip_sorcery_object_to_ami(identify, buf);
}

static int find_identify_by_endpoint(void *obj, void *arg, int flags)
{
	struct ip_identify_match *identify = obj;
	const char *endpoint_name = arg;

	return strcmp(identify->endpoint_name, endpoint_name) ? 0 : CMP_MATCH;
}

static int format_ami_endpoint_identify(const struct ast_sip_endpoint *endpoint,
					struct ast_sip_ami *ami)
{
	RAII_VAR(struct ao2_container *, identifies, NULL, ao2_cleanup);
	RAII_VAR(struct ip_identify_match *, identify, NULL, ao2_cleanup);
	RAII_VAR(struct ast_str *, buf, NULL, ast_free);

	identifies = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "identify",
		AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL);
	if (!identifies) {
		return -1;
	}

	identify = ao2_callback(identifies, 0, find_identify_by_endpoint,
		(void *) ast_sorcery_object_get_id(endpoint));
	if (!identify) {
		return 1;
	}

	if (!(buf = ast_sip_create_ami_event("IdentifyDetail", ami))) {
		return -1;
	}

	if (sip_identify_to_ami(identify, &buf)) {
		return -1;
	}

	ast_str_append(&buf, 0, "EndpointName: %s\r\n",
		ast_sorcery_object_get_id(endpoint));

	astman_append(ami->s, "%s\r\n", ast_str_buffer(buf));
	ami->count++;

	return 0;
}

struct ast_sip_endpoint_formatter endpoint_identify_formatter = {
	.format_ami = format_ami_endpoint_identify
};

static int cli_populate_container(void *obj, void *arg, int flags)
{
	ao2_link(arg, obj);

	return 0;
}

static int cli_iterator(void *container, ao2_callback_fn callback, void *args)
{
	const struct ast_sip_endpoint *endpoint = container;
	struct ao2_container *identifies;

	struct ast_variable fields = {
		.name = "endpoint",
		.value = ast_sorcery_object_get_id(endpoint),
		.next = NULL,
	};

	identifies = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "identify",
		AST_RETRIEVE_FLAG_MULTIPLE, &fields);
	if (!identifies) {
		return -1;
	}

	ao2_callback(identifies, OBJ_NODATA, callback, args);
	ao2_cleanup(identifies);

	return 0;
}

static int cli_endpoint_gather_identifies(void *obj, void *arg, int flags)
{
	struct ast_sip_endpoint *endpoint = obj;
	struct ao2_container *container = arg;

	cli_iterator(endpoint, cli_populate_container, container);

	return 0;
}

static struct ao2_container *cli_get_container(void)
{
	RAII_VAR(struct ao2_container *, parent_container, NULL, ao2_cleanup);
	RAII_VAR(struct ao2_container *, s_parent_container, NULL, ao2_cleanup);
	struct ao2_container *child_container;

	parent_container =  ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "endpoint",
		AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL);
	if (!parent_container) {
		return NULL;
	}

	s_parent_container = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_NOLOCK, 0,
		ast_sorcery_object_id_sort, ast_sorcery_object_id_compare);
	if (!s_parent_container) {
		return NULL;
	}

	if (ao2_container_dup(s_parent_container, parent_container, 0)) {
		return NULL;
	}

	child_container = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_NOLOCK, 0,
		ast_sorcery_object_id_sort, ast_sorcery_object_id_compare);
	if (!child_container) {
		return NULL;
	}

	ao2_callback(s_parent_container, OBJ_NODATA, cli_endpoint_gather_identifies, child_container);

	return child_container;
}

static void *cli_retrieve_by_id(const char *id)
{
	return ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "identify", id);
}

static int cli_print_header(void *obj, void *arg, int flags)
{
	struct ast_sip_cli_context *context = arg;
	int indent = CLI_INDENT_TO_SPACES(context->indent_level);
	int filler = CLI_MAX_WIDTH - indent - 22;

	ast_assert(context->output_buffer != NULL);

	ast_str_append(&context->output_buffer, 0,
		"%*s:  <Identify/Endpoint%*.*s>\n",
		indent, "Identify", filler, filler, CLI_HEADER_FILLER);

	if (context->recurse) {
		context->indent_level++;
		indent = CLI_INDENT_TO_SPACES(context->indent_level);
		filler = CLI_LAST_TABSTOP - indent - 24;

		ast_str_append(&context->output_buffer, 0,
			"%*s:  <ip/cidr%*.*s>\n",
			indent, "Match", filler, filler, CLI_HEADER_FILLER);

		context->indent_level--;
	}

	return 0;
}

static int cli_print_body(void *obj, void *arg, int flags)
{
	RAII_VAR(struct ast_str *, str, ast_str_create(MAX_OBJECT_FIELD), ast_free);
	struct ip_identify_match *ident = obj;
	struct ast_sip_cli_context *context = arg;
	struct ast_ha *match;
	int indent;

	ast_assert(context->output_buffer != NULL);

	ast_str_append(&context->output_buffer, 0, "%*s:  %s/%s\n",
		CLI_INDENT_TO_SPACES(context->indent_level), "Identify",
		ast_sorcery_object_get_id(ident), ident->endpoint_name);

	if (context->recurse) {
		context->indent_level++;
		indent = CLI_INDENT_TO_SPACES(context->indent_level);

		for (match = ident->matches; match; match = match->next) {
			const char *addr = ast_sockaddr_stringify_addr(&match->addr);

			ast_str_append(&context->output_buffer, 0, "%*s: %s%s/%d\n",
				indent,
				"Match",
				match->sense == AST_SENSE_ALLOW ? "!" : "",
				addr, ast_sockaddr_cidr_bits(&match->netmask));
		}

		context->indent_level--;

		if (context->indent_level == 0) {
			ast_str_append(&context->output_buffer, 0, "\n");
		}
	}

	if (context->show_details
		|| (context->show_details_only_level_0 && context->indent_level == 0)) {
		ast_str_append(&context->output_buffer, 0, "\n");
		ast_sip_cli_print_sorcery_objectset(ident, context, 0);
	}

	return 0;
}

/*
 * A function pointer to callback needs to be within the
 * module in order to avoid problems with an undefined
 * symbol when the module is loaded.
 */
static char *my_cli_traverse_objects(struct ast_cli_entry *e, int cmd,
	struct ast_cli_args *a)
{
	return ast_sip_cli_traverse_objects(e, cmd, a);
}

static struct ast_cli_entry cli_identify[] = {
AST_CLI_DEFINE(my_cli_traverse_objects, "List PJSIP Identifies",
	.command = "pjsip list identifies",
	.usage = "Usage: pjsip list identifies\n"
	"       List the configured PJSIP Identifies\n"),
AST_CLI_DEFINE(my_cli_traverse_objects, "Show PJSIP Identifies",
	.command = "pjsip show identifies",
	.usage = "Usage: pjsip show identifies\n"
	"       Show the configured PJSIP Identifies\n"),
AST_CLI_DEFINE(my_cli_traverse_objects, "Show PJSIP Identify",
	.command = "pjsip show identify",
	.usage = "Usage: pjsip show identify <id>\n"
	"       Show the configured PJSIP Identify\n"),
};

static struct ast_sip_cli_formatter_entry *cli_formatter;

static int load_module(void)
{
	CHECK_PJSIP_MODULE_LOADED();

	ast_sorcery_apply_config(ast_sip_get_sorcery(), "res_pjsip_endpoint_identifier_ip");
	ast_sorcery_apply_default(ast_sip_get_sorcery(), "identify", "config", "pjsip.conf,criteria=type=identify");

	if (ast_sorcery_object_register(ast_sip_get_sorcery(), "identify", ip_identify_alloc, NULL, NULL)) {
		return AST_MODULE_LOAD_DECLINE;
	}

#ifdef GRANDSTREAM_NETWORKS
	dns_sched = ast_sched_context_create();
	if (!dns_sched) {
		return AST_MODULE_LOAD_DECLINE;
	}

	if (ast_sched_start_thread(dns_sched)) {
		ast_log(LOG_ERROR, "Could not start scheduler thread for identify dns refresh\n");
		ast_sched_context_destroy(dns_sched);
		return AST_MODULE_LOAD_FAILURE;
	}
#endif

	ast_sorcery_object_field_register(ast_sip_get_sorcery(), "identify", "type", "", OPT_NOOP_T, 0, 0);
	ast_sorcery_object_field_register(ast_sip_get_sorcery(), "identify", "endpoint", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ip_identify_match, endpoint_name));
	ast_sorcery_object_field_register_custom(ast_sip_get_sorcery(), "identify", "match", "", ip_identify_match_handler, match_to_str, match_to_var_list, 0, 0);
	ast_sorcery_reload_object(ast_sip_get_sorcery(), "identify");

	ast_sip_register_endpoint_identifier_with_name(&ip_identifier, "ip");
	ast_sip_register_endpoint_formatter(&endpoint_identify_formatter);

	cli_formatter = ao2_alloc(sizeof(struct ast_sip_cli_formatter_entry), NULL);
	if (!cli_formatter) {
		ast_log(LOG_ERROR, "Unable to allocate memory for cli formatter\n");
		return AST_MODULE_LOAD_DECLINE;
	}
	cli_formatter->name = "identify";
	cli_formatter->print_header = cli_print_header;
	cli_formatter->print_body = cli_print_body;
	cli_formatter->get_container = cli_get_container;
	cli_formatter->iterate = cli_iterator;
	cli_formatter->get_id = ast_sorcery_object_get_id;
	cli_formatter->retrieve_by_id = cli_retrieve_by_id;

	ast_sip_register_cli_formatter(cli_formatter);
	ast_cli_register_multiple(cli_identify, ARRAY_LEN(cli_identify));

	return AST_MODULE_LOAD_SUCCESS;
}

static int reload_module(void)
{
	ast_sorcery_reload_object(ast_sip_get_sorcery(), "identify");

	return 0;
}

static int unload_module(void)
{
	ast_cli_unregister_multiple(cli_identify, ARRAY_LEN(cli_identify));
	ast_sip_unregister_cli_formatter(cli_formatter);
	ast_sip_unregister_endpoint_formatter(&endpoint_identify_formatter);
	ast_sip_unregister_endpoint_identifier(&ip_identifier);

#ifdef GRANDSTREAM_NETWORKS
	if (dns_sched) {
		ast_sched_context_destroy(dns_sched);
		dns_sched = NULL;
	}
#endif
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP IP endpoint identifier",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.reload = reload_module,
		.unload = unload_module,
		.load_pri = AST_MODPRI_APP_DEPEND,
	       );
