/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Mark Michelson <mmichelson@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#ifdef GRANDSTREAM_NETWORKS
#define AST_MODULE_LOG "pjsip"
#endif

#include "asterisk.h"

#ifdef GRANDSTREAM_NETWORKS
ASTERISK_FILE_VERSION(__FILE__, "$Revision: 13.0 $");
#endif

#include <pjsip.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/module.h"

static int get_endpoint_details(pjsip_rx_data *rdata, char *endpoint, size_t endpoint_size, char *domain, size_t domain_size)
{
	pjsip_uri *from = rdata->msg_info.from->uri;
	pjsip_sip_uri *sip_from;
	if (!PJSIP_URI_SCHEME_IS_SIP(from) && !PJSIP_URI_SCHEME_IS_SIPS(from)) {
		return -1;
	}
	sip_from = (pjsip_sip_uri *) pjsip_uri_get_uri(from);
	ast_copy_pj_str(endpoint, &sip_from->user, endpoint_size);
	ast_copy_pj_str(domain, &sip_from->host, domain_size);
	return 0;
}

static int find_transport_in_use(void *obj, void *arg, int flags)
{
	struct ast_sip_transport *transport = obj;
	pjsip_rx_data *rdata = arg;

	if ((transport->state->transport == rdata->tp_info.transport) ||
		(transport->state->factory && !pj_strcmp(&transport->state->factory->addr_name.host, &rdata->tp_info.transport->local_name.host) &&
			transport->state->factory->addr_name.port == rdata->tp_info.transport->local_name.port)) {
		return CMP_MATCH | CMP_STOP;
	}

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
/* identify user by username+ip, after username matched, we check the user'contact */
static int username_check_contact(pjsip_rx_data *rdata, struct ast_sip_endpoint *endpoint)
{
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);
	struct ao2_iterator i;
	struct ast_sip_contact *contact;
	char aor_id[80]      = {0};
	char contact_uri[80] = {0};
	char src_addr[80]    = {0};
	int  contact_matched = 0;

	if (!rdata || !endpoint) {
		return -1;
	}

	aor = ast_sip_location_retrieve_aor(endpoint->aors); 
	if (!aor) {
		return -1;
	}

	contacts = ast_sip_location_retrieve_aor_contacts(aor);
	if (!contacts || (aor->max_contacts > ao2_container_count(contacts))) {
		ast_log(LOG_NOTICE, "The endpoint[%s] not reach its max contacts, pass!", ast_sorcery_object_get_id(endpoint));
		return 0;
	}

	snprintf(src_addr, sizeof(src_addr)-1, "%s:%d", rdata->pkt_info.src_name, rdata->pkt_info.src_port);
	i = ao2_iterator_init(contacts, 0);
	while ((contact = ao2_iterator_next(&i))) {
		sscanf(contact->uri, "%79[^@]@%79[^;]", aor_id, contact_uri);
		ao2_ref(contact, -1); 

		if (!strcmp(contact_uri, src_addr)) {
			contact_matched = 1;
			break;
		}
	}
	ao2_iterator_destroy(&i);

	if (contact_matched) {
		ast_log(LOG_NOTICE, "contact_matched!");
		return 0;
	}

	return -1;
}
#endif

static struct ast_sip_endpoint *username_identify(pjsip_rx_data *rdata)
{
	char endpoint_name[64], domain_name[64], id[AST_UUID_STR_LEN];
	struct ast_sip_endpoint *endpoint;
	RAII_VAR(struct ast_sip_domain_alias *, alias, NULL, ao2_cleanup);
	RAII_VAR(struct ao2_container *, transports, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_transport *, transport, NULL, ao2_cleanup);

	if (get_endpoint_details(rdata, endpoint_name, sizeof(endpoint_name), domain_name, sizeof(domain_name))) {
		return NULL;
	}

	/* Attempt to find the endpoint given the name and domain provided */
	snprintf(id, sizeof(id), "%s@%s", endpoint_name, domain_name);
	if ((endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", id))) {
		goto done;
	}

	/* See if an alias exists for the domain provided */
	if ((alias = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "domain_alias", domain_name))) {
		snprintf(id, sizeof(id), "%s@%s", endpoint_name, alias->domain);
		if ((endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", id))) {
			goto done;
		}
	}

	/* See if the transport this came in on has a provided domain */
	if ((transports = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "transport", AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL)) &&
		(transport = ao2_callback(transports, 0, find_transport_in_use, rdata)) &&
		!ast_strlen_zero(transport->domain)) {
		snprintf(id, sizeof(id), "%s@%s", endpoint_name, transport->domain);
		if ((endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", id))) {
			goto done;
		}
	}

	/* Fall back to no domain */
	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", endpoint_name);

done:
	if (endpoint) {
		if (!(endpoint->ident_method & AST_SIP_ENDPOINT_IDENTIFY_BY_USERNAME)) {
			ao2_ref(endpoint, -1);
			return NULL;
		}
#ifdef GRANDSTREAM_NETWORKS
		if (!((rdata->msg_info.msg->type == PJSIP_REQUEST_MSG) && (rdata->msg_info.msg->line.req.method.id == PJSIP_REGISTER_METHOD)) 
            && username_check_contact(rdata, endpoint)) {
			ast_debug(3, "Checkout contact failed, endpoint[%s]\n", ast_sorcery_object_get_id(endpoint));
			ao2_ref(endpoint, -1);
			return NULL;
		}
#endif
		ast_debug(3, "Retrieved endpoint %s\n", ast_sorcery_object_get_id(endpoint));
	} else {
		ast_debug(3, "Could not identify endpoint by username '%s'\n", endpoint_name);
	}
	return endpoint;
}

static struct ast_sip_endpoint_identifier username_identifier = {
	.identify_endpoint = username_identify,
};

static int load_module(void)
{
	CHECK_PJSIP_MODULE_LOADED();

	ast_sip_register_endpoint_identifier_with_name(&username_identifier, "username");
	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_unregister_endpoint_identifier(&username_identifier);
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP username endpoint identifier",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.load_pri = AST_MODPRI_APP_DEPEND,
	       );
