/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * JiaQun Chen <jqchen@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#define AST_MODULE_LOG "pjsip"

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 13.0 $");

#include <pjsip.h>
#include <pjsip_simple.h>
#include <pjsip_ua.h>
#include <pjlib.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <ctype.h>

#include "asterisk/sched.h"
#include "asterisk/lock.h"
#include "asterisk/file.h"
#include "asterisk/astobj2.h"

#include "asterisk/module.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_pubsub.h"
#include "asterisk/res_pjsip_presence_xml.h"
#include "asterisk/res_pjsip_body_generator_types.h"
#include "asterisk/res_pjsip_outbound_publish.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/callerid.h"
#include "asterisk/channel.h"
#include "asterisk/strings.h"
#include "asterisk/cli.h"

#include "asterisk/pbx.h"
#include "asterisk/manager.h"
#include "asterisk/config.h"
#include "asterisk/monitor.h"
#include "asterisk/utils.h"
#include "asterisk/astdb.h"
#include "asterisk/devicestate.h"
#include "asterisk/global_datastores.h"
#include "asterisk/taskprocessor.h"
#include "asterisk/data.h"
#include "asterisk/stasis_channels.h"
#include "asterisk/stasis_message_router.h"
#include "asterisk/core_local.h"
#include "asterisk/mixmonitor.h"
#include "asterisk/linkedlists.h"
#include "asterisk/datastore.h"
#include "asterisk/uuid.h"
#include "asterisk/test.h"



/*!
   \brief copy from pbx.c for eventlists' remote subscribe needs.
    ast_exten: An extension - one item per
	priority.
*/
#define  SIP_LOCAL_CONTEXT     "from-internal"
#define  SIP_STR_LOCAL_SUBSR   "ucm6100"
#define  SIP_EXT_CONFERENCE    "ext-conference"
#define  SIP_EXT_QUEUES        "ext-queues"
#define  SIP_EXT_GROUP         "ext-group"
#define  SIP_PARKEDCALLS         "parkedcalls"

#define  SIP_DEFAULT_CONTEXT    "ext-local"
#define  RES_PARKING_CONFIG_FILE "res_parking.conf"
#define  RESOURCE_LIST_CONFIG_FILE    "pjsip_resource_list.conf"
#define  EVENTLIST_CONFIG_FILE    "eventlist.conf"
#define  AST_SIP_EXTEN_STATE_DATA "ast_sip_exten_state_data"
#define  DEFAULT_DIALOG_BODY "application/dialog-info+xml"
#define  EVENT_TYPE         "application"
#define  EVENT_PIDF_SUBTYPE "pidf+xml"
#define  EVENT_SUBTYPE      "dialog-info+xml"
#define  EVENT_STATE_STR    "<state: "
#define  SIP_MAX_RES_ONE_TIME     3
#define  SIP_MAX_STATE_ERR_NUM    30
#define  SIP_REMOTE_SUBS_EXPIRES  600
/*! --- Hash tables of various objects --------*/
#ifdef LOW_MEMORY
static const int HASH_EVENTLIST_SIZE = 17;
#else
static const int HASH_EVENTLIST_SIZE = 563;	/*!< Size of eventlist hash table, prime number preferred! */
#endif

/*! The eventlist list container */
struct ao2_container *con_eventlists = NULL;

/*! This is the thread for manange the eventlists' subscribe state. */
pthread_t eventlist_state_thread = AST_PTHREADT_NULL;


int g_loading_flag = TRUE;
int g_doing_state  = TRUE;
struct timeval  eventlist_loading_start;

/*! \brief Scheduler used for automatically expiring publications */
static struct ast_sched_context *sched;

enum eventlisttype
{
	EN_EVENTLIST_DIALOG,
	EN_EVENTLIST_PRESENCE,
	EN_EVENTLIST_BUTT
};

typedef enum {
	EN_EVENTLISTS_MARKED,
	EN_EVENTLISTS_ALL
} EN_EVENTLIST_UNLINK_FLAG;

typedef enum {
	EN_EVENTLIST_LOCAL_USER,
	EN_EVENTLIST_REMOTE_USER,
	EN_EVENTLIST_CONF_USER,
	EN_EVENTLIST_QUEUE_USER,
	EN_EVENTLIST_RGROUP_USER,
	EN_EVENTLIST_PARK_USER,
	EN_EVENTLIST_UNKNOWN_USER,
	EN_EVENTLIST_USER_BUTT
} EN_EVENTLIST_USER_TYPE;

/* Added by yfyang, [Bug 17631], use rotate state to void endless loop */
typedef enum {
	EN_EVENTLIST_ROTATE_INIT,
	EN_EVENTLIST_ROTATE_START,
	EN_EVENTLIST_ROTATE_INPROGRESS,
	EN_EVENTLIST_ROTATE_BUTT    
} EN_EVENTLIST_ROTATE_STATE;

struct st_state_exten {
	char cExten[80];                       /*!< extension number */
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(cEndpoint);       /*!< remote peer name */
		AST_STRING_FIELD(cOriginExten);    /*!< original exten string */ 
		AST_STRING_FIELD(cDeviceName);     /*!< device string, for eventlist enhancement feature */ 
		AST_STRING_FIELD(cHostName);       /*!< host name of remote peer */ 
	);
 
	enum ast_extension_states enState;     /*!< current extension state */
	EN_EVENTLIST_USER_TYPE    enType;      /*!< user type */
	struct pjsip_sub_pvt     *pPvt;        /*!< remote subscribe dialog */
	int  iStateID;                         /*!< SUBSCRIBE: ID for devicestate subscriptions */
	int  iSchedID;                         /*!< Sched id to retry to send subscribe*/
	int  iExpires;                         /*!< subscribe expires, Bug 16242 */
	int  iStateErrCount;                   /*!< the count of stated error */
	int  bMark;                            /*! Mark used during reload processing */
	int  bReload;                          /*! Mark the exten need to update during reloading */
	int  bTypeUpdate;                      /*! Mark whether the exten type has update or not during reloading */
	AST_LIST_ENTRY(st_state_exten) entry;  /*!< next exten record */
};

static void pjsip_event_subscriber_destroy(struct ast_sip_subscription *sub);
static void pjsip_event_subscriber_state_change(struct ast_sip_subscription *sub, pjsip_msg_body *body, enum pjsip_evsub_state state);
static void pjsip_event_subscriber_record(const char *eventlist, const char *subscriber, int add);

static struct ast_sip_subscriber event_subscriber = {
	.state_change = pjsip_event_subscriber_state_change,
	.subscriber_destroy = pjsip_event_subscriber_destroy,
};


static struct ast_sip_subscriber_info_handler  info_handler = {
	.event_name = "dialog",
	.subscriber_record = pjsip_event_subscriber_record,
};

static struct ast_sip_subscription_handler event_handler = {
	.event_name = "dialog",
	.body_type = AST_SIP_EXTEN_STATE_DATA,
	.accept = { DEFAULT_DIALOG_BODY, },
	.subscriber = &event_subscriber,
};
/*! \brief Structure which contains the sip ccss agent information */
struct pjsip_sub_pvt {
	struct ast_sip_subscription    *pSub;  /*!< remote subscribe dialog */
	struct sip_eventlist  *pstEventlist;   /*!< If this is a SUBSCRIBE dialog, relate to a eventlist struct, Added by yfyang, for blf eventlist feature */
	struct st_state_exten *pstRemoteExten; /*!< remote exten for subscribe, Added by yfyang, for blf eventlist */
};

/* to define a eventlist struct */
struct sip_eventlist {
	char               cName[80];          /*!< the unique name of this object */
	enum eventlisttype enType;             /*!< list type, default dialog */
	int                bMark;              /*!< moved out of ASTOBJ into struct proper; That which bears the_mark should be deleted! */
	char               cExtenlistStr[1024]; /*!< the extens string in the eventlist configure */
	AST_LIST_HEAD(, st_state_exten) ast_exten_list;           /*!< relate to the state exten list*/
	AST_VECTOR(, const char *) subscribers; /*!< record the subscribers to the list */
};

/*function declare*/

static void  eventlist_destroy_list(void *data);
static struct sip_eventlist *eventlist_build(const char *cListName, struct ast_variable *pstVar, struct ast_variable *pstVarregs);

static int eventlist_match_and_cleanup(void *eventlistObj, void *arg, int flags);
static void eventlist_unlink_marked_from_tables(void);
static void eventlist_unlink_all_from_tables(void);
static void *eventlist_unref(struct sip_eventlist *pstEventlist, char *tag);
static void eventlist_clear_exten(struct sip_eventlist *pstEventlist);
static void eventlist_create_exten_list(struct sip_eventlist *pstEventlist, const char* listStr);
static int  eventlist_extenstate(char *context, char *exten, struct ast_state_cb_info *info, void *data);

static void *eventlist_do_state(void *data);
static int eventlist_hash_cb(const void *obj, const int flags);
static int eventlist_cmp_cb(void *obj, void *arg, int flags);
static struct st_state_exten *eventlist_create_state_exten(const char *pcExten, char *pExtenName);
static char *eventlist_extentype_2_str(EN_EVENTLIST_USER_TYPE  enType);
static int pjsip_do_remote_blf_subscribe(struct sip_eventlist  *pstEventlist, struct st_state_exten *pstStateExten);
static void eventlist_set_exten_type(struct st_state_exten *pstStateExten, char *device, EN_EVENTLIST_USER_TYPE enType, int exten_changed);

struct statechange {
	AST_LIST_ENTRY(statechange) entry;
	int state;
	char dev[0];
};

#define ______EVENTLIST_BODY_BUILD_______

/*! \brief get exten type string */
static char *eventlist_extentype_2_str(EN_EVENTLIST_USER_TYPE  enType)
{
	switch (enType) {
	case  EN_EVENTLIST_LOCAL_USER:
		return "local";

	case  EN_EVENTLIST_REMOTE_USER:
		return "remote";

	case  EN_EVENTLIST_CONF_USER:
		return "confRoom";

	case  EN_EVENTLIST_QUEUE_USER:
		return "queue";

	case  EN_EVENTLIST_RGROUP_USER:
		return "ringgroup";

	case  EN_EVENTLIST_PARK_USER:
		return "parkLot";

	case  EN_EVENTLIST_UNKNOWN_USER:
		return "unknown";

	default:
		return "unknown";
	}
}


static char *eventlist_get_context_by_usertype(EN_EVENTLIST_USER_TYPE user_type)
{
	switch (user_type) {
	case EN_EVENTLIST_CONF_USER:
		return "ext-conference";
	case EN_EVENTLIST_QUEUE_USER:
		return "ext-queues";
	case EN_EVENTLIST_RGROUP_USER:
		return "ext-group";
	case EN_EVENTLIST_PARK_USER:
		return "parkedcalls";
	default:
		return "ext-local";
	}
}


static EN_EVENTLIST_USER_TYPE eventlist_get_usertype_by_context(char *context)
{
	if (!context) {
		return EN_EVENTLIST_UNKNOWN_USER;
	}

	if (!strcmp(context, "ext-conference")) {
		return EN_EVENTLIST_CONF_USER;
	}

	if (!strcmp(context, "ext-queues")) {
		return EN_EVENTLIST_QUEUE_USER;
	}

	if (!strcmp(context, "ext-group")) {
		return EN_EVENTLIST_RGROUP_USER;
	}

	if (!strcmp(context, "parkedcalls")) {
		return EN_EVENTLIST_PARK_USER;
	}

	if (!strcmp(context, "ext-local")) {
		return EN_EVENTLIST_LOCAL_USER;
	}

    return EN_EVENTLIST_UNKNOWN_USER;
}


static void *eventlist_unref(struct sip_eventlist *pstEventlist, char *tag)
{
	ao2_t_ref(pstEventlist, -1, tag);
	return NULL;
}

static void eventlist_terminated_subscribe(struct pjsip_sub_pvt *sub_pvt, pj_bool_t notify)
{
	if (NULL == sub_pvt) {
		return;
	}

	ast_sip_terminated_subscribe(sub_pvt->pSub, notify);

	sub_pvt->pSub = NULL;
	sub_pvt->pstEventlist = NULL;
	sub_pvt->pstRemoteExten = NULL;
	ao2_ref(sub_pvt, -1);
	return;
}


/*! Destroy remote subscribcription */
static int eventlist_clear_remote_subscribe(void *obj)
{
	struct st_state_exten  *pstExten  = obj;

	if (NULL == pstExten) {
		return 0;
	}

	eventlist_terminated_subscribe(pstExten->pPvt, PJ_FALSE);
	pstExten->pPvt = NULL;
	pstExten->iStateID = -1;

	return 0;
}

static void eventlist_destroy_state_exten(void *obj)
{
	struct st_state_exten  *pstStateExten = obj;

	pstStateExten->iStateID = -1;

	if (NULL != pstStateExten->pPvt) {
		ast_sip_push_task_synchronous(NULL, eventlist_clear_remote_subscribe, pstStateExten);
	}

	if (-1 != pstStateExten->iSchedID) {
		if (0 != AST_SCHED_DEL(sched, pstStateExten->iSchedID)) {
			ast_log(LOG_ERROR, "exten subscribe sched[%d] del failed!\n", pstStateExten->iSchedID);
		}
		pstStateExten->iSchedID = -1;
	}

	ast_string_field_free_memory(pstStateExten);
	return;
}

/*! \brief create a new st_state_exten */
static struct st_state_exten *eventlist_create_state_exten(const char *pcExten, char *pExtenName)
{
	struct st_state_exten  *pstStateExten = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);

	if (NULL == pcExten) {
		return NULL;
	}

	pstStateExten = ao2_alloc(sizeof(*pstStateExten), eventlist_destroy_state_exten);
	if (NULL == pstStateExten) {
		return NULL;
	}

	if (ast_string_field_init(pstStateExten, 240)) {
		ao2_ref(pstStateExten, -1);
		return NULL;
	}

	ast_copy_string(pstStateExten->cExten, pcExten, sizeof(pstStateExten->cExten));
	ast_string_field_build(pstStateExten, cOriginExten, "%s", pstStateExten->cExten);
	if (pExtenName) {
		ast_string_field_build(pstStateExten, cDeviceName, "%s", pExtenName); 
	}

	pstStateExten->iStateID    = -1;
	pstStateExten->iSchedID    = -1;
	pstStateExten->pPvt        = NULL;
	pstStateExten->enState     = AST_EXTENSION_UNAVAILABLE;
	pstStateExten->enType      = EN_EVENTLIST_UNKNOWN_USER;
	pstStateExten->iExpires    = SIP_REMOTE_SUBS_EXPIRES;   /* Bug 16242*/
	pstStateExten->iStateErrCount = 0;
	pstStateExten->bMark          = 0;
	pstStateExten->bReload        = 0;
	pstStateExten->bTypeUpdate    = 0;

	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", pcExten);

	if (NULL != endpoint) {
		/* sip user */
		pstStateExten->enState = ast_extension_state(NULL, SIP_DEFAULT_CONTEXT, pcExten);
		pstStateExten->enType  = EN_EVENTLIST_LOCAL_USER;
        pstStateExten->iStateID = 1;
	} 

	ast_debug(LOG_DEBUG, "Created a state exten [%s], state[%s], type[%s].\n", 
			pcExten, ast_extension_state2str(pstStateExten->enState), eventlist_extentype_2_str(pstStateExten->enType));

	return pstStateExten;
}

/*! \brief create st_state_exten list by parse the extenlist string*/
static void eventlist_create_exten_list(struct sip_eventlist *pstEventlist, const char* listStr)
{
	char *pclist = ast_strdupa(listStr);
	char *pcExtens     = NULL;
	int   bSame        = FALSE;
	char  cExten[80]   = {0};
	char  cExtenStart[400]= {0};
	char  cExtenName[80] = {0};
	char  cNameTmp[80]   = {0};
	char *pNameStart     = NULL;
	char *pNameEnd       = NULL;
	char *pSpecial       = NULL;
	struct st_state_exten  *pstStateExten = NULL;
	unsigned int   ulExtenCount  = 0;

	if (NULL == pstEventlist) {
		return;
	}

    /* parse the extenlist string to a exten list, format: 1000-1004,1009 */
	AST_LIST_LOCK(&pstEventlist->ast_exten_list);
	while ((pcExtens = strsep(&pclist, ","))) {
		pcExtens = ast_skip_blanks(pcExtens);

		if ((NULL == pcExtens) || (0 == strlen(pcExtens))) {
			continue;
		}

		memset(cExtenName, 0, sizeof(cExtenName));
		memset(cNameTmp, 0, sizeof(cNameTmp));

		pNameStart = strchr(pcExtens, '(');
		pNameEnd = strrchr(pcExtens, ')');
		if ((NULL != pNameStart) && (NULL != pNameEnd) && (pNameEnd > pNameStart)) {
			strncpy(cExtenName, pNameStart+1, ((pNameEnd-pNameStart-1) >= sizeof(cExtenName)) ? (sizeof(cExtenName)-1):(pNameEnd-pNameStart-1));
			*pNameStart = '\0';

			pNameStart = cExtenName;
			while ((pSpecial = strstr(pNameStart, "%2c"))) {
				strncpy(cNameTmp, cExtenName, pSpecial-pNameStart);
				strcat(cNameTmp, ",");
              
				pSpecial = pSpecial+3;
				if ('\0' != *pSpecial) {
					strcat(cNameTmp, pSpecial);
				}
				strncpy(cExtenName, cNameTmp, sizeof(cExtenName));
				memset(cNameTmp, 0, sizeof(cNameTmp));
			}
		}

		strncpy(cExtenStart, pcExtens, sizeof(cExtenStart));

		ast_log(LOG_VERBOSE, "parse exten, exten[%s].\n", cExtenStart);
                    
		do {
			sprintf(cExten, "%s", cExtenStart);  
			AST_LIST_TRAVERSE(&(pstEventlist->ast_exten_list), pstStateExten, entry) {
				if (0 == strcmp(cExten, pstStateExten->cOriginExten)) {
					if (strlen(cExtenName)) {
						ast_string_field_build(pstStateExten, cDeviceName, "%s", cExtenName); 
					}
					pstStateExten->bMark = 0;
					pstStateExten->iStateErrCount = 0;
					pstStateExten->bTypeUpdate    = 0;
					bSame = TRUE;
					break;
				}
			}

			if (!bSame) {
				pstStateExten = eventlist_create_state_exten(cExten, cExtenName);
				if (pstStateExten) {
					AST_LIST_INSERT_TAIL(&pstEventlist->ast_exten_list, pstStateExten, entry);
					ulExtenCount++;
					ast_sip_add_subscription_to_tree(pstEventlist->cName, cExten);
				}
			}

			bSame = FALSE;
		} while (0); 
	}

	AST_LIST_UNLOCK(&pstEventlist->ast_exten_list);

	return;
}

/*************************************************
 * Function Name : eventlist_hash_cb 
 * Description: The only member of the eventlist used here is the name field
 * Input:  data 
 * Output: NA 
 * Changes: 
*************************************************/
static int eventlist_hash_cb(const void *obj, const int flags)
{
	const struct sip_eventlist  *pstEventlist = obj;

	if (NULL == obj) {
		return -1;
	}

	return ast_str_case_hash(pstEventlist->cName);
}


/*************************************************
 * Function Name : eventlist_hash_cb 
 * Description: The only member of the eventlist used here is the name field
 * Input:  data 
 * Output: NA 
 * Changes: 
*************************************************/
static int eventlist_cmp_cb(void *obj, void *arg, int flags)
{
	struct sip_eventlist *pstEventlist  = obj; 
	struct sip_eventlist *pstEventlist2 = arg;

	if (NULL == obj || NULL == arg) {
		ast_log(LOG_ERROR, "eventlist_cmp_cb, input null\n");
		return -1;
	}

	return !strcasecmp(pstEventlist->cName, pstEventlist2->cName) ? CMP_MATCH | CMP_STOP : 0;
}

/*! this func is used with ao2_callback to unlink/delete all marked or linked eventlists, depending on arg */
static int eventlist_match_and_cleanup(void *eventlistObj, void *arg, int flags)
{
	struct sip_eventlist  *pstEventlist = eventlistObj;
	struct st_state_exten *pstExten;
	EN_EVENTLIST_UNLINK_FLAG  flag = *(EN_EVENTLIST_UNLINK_FLAG *)arg;

	if ((EN_EVENTLISTS_ALL == flag) || (pstEventlist->bMark)) {
		ast_log(LOG_DEBUG, "eventlist_match_and_cleanup, clean list[%s]\n", pstEventlist->cName);
		return CMP_MATCH;
	}
  
	AST_LIST_TRAVERSE_SAFE_BEGIN(&pstEventlist->ast_exten_list, pstExten, entry) {
		if (!pstExten->bMark) {
			continue;
		}
		AST_LIST_REMOVE_CURRENT(entry);
		ast_sip_remove_subscription_from_tree(pstEventlist->cName, (char *)pstExten->cOriginExten);
		ast_log(LOG_NOTICE, "eventlist_match_and_cleanup, clean exten[%s]\n", pstExten->cExten);
		ao2_ref(pstExten, -1);
	}
	AST_LIST_TRAVERSE_SAFE_END

	return 0;
}


/*! Unlink all marked eventlists from ao2 containers */
static void eventlist_unlink_marked_from_tables(void)
{
	EN_EVENTLIST_UNLINK_FLAG  flag = EN_EVENTLISTS_MARKED;

	ao2_t_callback(con_eventlists, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE,
			eventlist_match_and_cleanup, &flag, "initiating callback to remove marked eventlists");
}

/*! \brief set mark flag to eventlist, prepare for destroy */
static int eventlist_markall(void *device, void *arg, int flags)
{
	struct sip_eventlist  *pstEventlist = device;
	struct st_state_exten *pstExten;

	pstEventlist->bMark = 1;

	ast_log(LOG_NOTICE, "eventlist_markall, eventlist[%s].\n", pstEventlist->cName);
	AST_LIST_TRAVERSE(&pstEventlist->ast_exten_list, pstExten, entry) {
		pstExten->bMark   = 1;
		pstExten->bReload = 1;
		pstExten->bTypeUpdate = 0;
		ast_log(LOG_NOTICE, "eventlist_markall, exten[%s].\n", pstExten->cExten);
	}

	return 0;
}

/*! Unlink all eventlists from ao2 containers */
static void eventlist_unlink_all_from_tables(void)
{
	EN_EVENTLIST_UNLINK_FLAG  flag = EN_EVENTLISTS_ALL;

	ao2_t_callback(con_eventlists, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE,
			eventlist_match_and_cleanup, &flag, "initiating callback to remove marked eventlists");
}
/*! \brief Get state string express */
static char* eventlist_get_dialogstate_by_extenstate(int state)
{
	switch (state) {
		case (AST_EXTENSION_RINGING | AST_EXTENSION_INUSE):
		case AST_EXTENSION_RINGING:
			return "early";

		case AST_EXTENSION_INUSE:  /* no break*/
		case AST_EXTENSION_BUSY:   /* no break*/
		case AST_EXTENSION_ONHOLD:
			return "confirmed";

		case AST_EXTENSION_UNAVAILABLE:
		case AST_EXTENSION_REMOVED:    /* Added by yfyang, [Bug 16135] BLF:pbx sent wrong status when deleting extension on web*/
			return "offline";  

		default:
			return "terminated";  
	}

	return "terminated";  
}


/*************************************************
 * Function Name : eventlist_extenstate 
 * Description: Callback for the devicestate notification (SUBSCRIBE) support subsystem
 * Input:   
 * Output: NA 
 * Changes: 
*************************************************/
static int eventlist_extenstate(char *context, char *exten, struct ast_state_cb_info *info, void *data)
{
	struct sip_eventlist  *pstEventlist  = NULL;
	struct st_state_exten *pstStateExten = NULL;
	struct ao2_iterator    stIterator;  
	int    state = info->exten_state;
	EN_EVENTLIST_USER_TYPE  enType;

	if (NULL == context || NULL == exten) {
		return -1;
	}
	
	if (TRUE == g_loading_flag) {
		ast_log(LOG_NOTICE, "eventlist_extenstate, during reloading, do nothing!\n");
		return 0;
	}

	/* only interested in extension state updates involving device states */
	if (info->reason != AST_HINT_UPDATE_DEVICE) {
		return 0;
	}

	enType = eventlist_get_usertype_by_context(context);
	ast_log(LOG_NOTICE, "eventlist_extenstate, context[%s], exten[%s], state[%d]=[%s]\n", context,exten, state, ast_extension_state2str(state));

	stIterator = ao2_iterator_init(con_eventlists, 0);
	while ((pstEventlist = ao2_t_iterator_next(&stIterator, "iterate thru eventlists table"))) {
		/* find the exten in the list by exten name */
		AST_LIST_TRAVERSE(&(pstEventlist->ast_exten_list), pstStateExten, entry) {
			if ((pstStateExten->enType == enType) && (0 == strcmp(exten, pstStateExten->cExten))) {
				/* update the state of the exten */
				if ((AST_EXTENSION_REMOVED == state) || 
				    (0 != strcmp(eventlist_get_dialogstate_by_extenstate(pstStateExten->enState), eventlist_get_dialogstate_by_extenstate(state)))) {
					pstStateExten->enState = state;
					if (AST_EXTENSION_REMOVED == state) {
						pstStateExten->enType = EN_EVENTLIST_UNKNOWN_USER;
					}

					if ((AST_EXTENSION_UNAVAILABLE == pstStateExten->enState) && (pstStateExten->enType == EN_EVENTLIST_CONF_USER)) {
						pstStateExten->enState = AST_EXTENSION_NOT_INUSE;
					}

					manager_event(EVENT_FLAG_SYSTEM, "EventlistMemberStatus", "List: %s\r\nExten: %s\r\nLocation: %s\r\nStatus: %s\r\n", 
					        pstEventlist->cName, pstStateExten->cExten, eventlist_extentype_2_str(pstStateExten->enType),
					        eventlist_get_dialogstate_by_extenstate(pstStateExten->enState)); 

					ast_log(LOG_NOTICE, "eventlist_extenstate, update exten[%s], state[%d].\n", exten, pstStateExten->enState);
				}

				break;
			}
		}
		eventlist_unref(pstEventlist, "Done with iterator"); 
	}
	ao2_iterator_destroy(&stIterator);

	return 0;
}

/*! Destroy all eventlist-related subscriptions */
static void eventlist_clear_exten(struct sip_eventlist *pstEventlist)
{
	struct st_state_exten  *pstStateExten = NULL;

	if (NULL == pstEventlist) {
		return;
	}

	AST_LIST_LOCK(&pstEventlist->ast_exten_list);
	while ((pstStateExten = AST_LIST_REMOVE_HEAD(&pstEventlist->ast_exten_list, entry))) {
		ao2_ref(pstStateExten, -1);
    }

	AST_LIST_UNLOCK(&pstEventlist->ast_exten_list);
}

/*! \brief Destroy eventlist object from memory */
static void eventlist_destroy_list(void *data)
{
	struct sip_eventlist *pstEventlist = data;
	int  i;

	if (NULL == pstEventlist) {
		return;
	}

	ast_debug(LOG_VERBOSE, "Destroying SIP eventlist %s\n", pstEventlist->cName);

	/* Remove any eventlist subscriptions for this eventlist before we destroy anything */
	eventlist_clear_exten(pstEventlist);
	AST_LIST_HEAD_DESTROY(&pstEventlist->ast_exten_list);

	/* clear subscriber list */
	for (i = 0; i < AST_VECTOR_SIZE(&pstEventlist->subscribers); ++i) {
		ast_free((char *) AST_VECTOR_GET(&pstEventlist->subscribers, i));
	}
	AST_VECTOR_FREE(&pstEventlist->subscribers);

	ast_sip_remove_subscription_tree_for_list(pstEventlist->cName);

	memset(pstEventlist->cName, 0, sizeof(pstEventlist->cName));
	memset(pstEventlist->cExtenlistStr, 0, sizeof(pstEventlist->cExtenlistStr));
}

#define ______EVENTLIST_REMOTE_STATE_SUBSCRIBE______
/*! exten --> extension's dialplan (outbound-allroutes) --> trunk --> peername*/
static int eventlist_get_endpointname(struct st_state_exten  *pstStateExten, struct ast_exten *pstExten)
{
	char   cExten[80]  = {0};
	char   family[80]  = {0};
	char   data[256]   = {0};
	char   endpt_id[80]= {0};
	char   cExten0[80] = {0};
	char   *arr[3];
	char   *buf  = data;
	unsigned int  i = 0;
	unsigned int  strip = 0;
	int    exten_changed = 0;
//	char   *uri = NULL;
	RAII_VAR(struct ast_sip_endpoint *, pstEndpoint, NULL, ao2_cleanup);
//	RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);
//	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);

	if (NULL == pstStateExten || NULL == pstExten) {
		return -1;
	}

	/* get AAUSER DIAL data from ucmdb */
	snprintf(family, sizeof(family), "TRUNK/%s", pstExten->exten);

    /* get DIAL */
	ast_ucm_db_get(family, "EVENTLIST", data, sizeof(data));

	if (strlen(data) == 0) {
		return -1;   /*! not found*/
	}
	
	while((i < 3) && ((arr[i] = strtok(buf,"_")) != NULL)) {
		i++;
		buf = NULL; 
	}

	snprintf(endpt_id, sizeof(endpt_id), "trunk_%s", arr[0]);
	if (!(pstEndpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", endpt_id))) {
		return -1;
	}

	strip = atoi(arr[1]);
	if (strip < strlen(pstStateExten->cOriginExten)) {
		snprintf(cExten, sizeof(cExten), "%s", pstStateExten->cOriginExten + strip);
	} else {
		snprintf(cExten, sizeof(cExten), "%s", pstStateExten->cOriginExten);
	}

	if (strcmp(arr[2], "a")) {
		snprintf(cExten0, sizeof(cExten0), "%s%s", arr[2], cExten);
		snprintf(cExten, sizeof(cExten), "%s", cExten0);
	}

	exten_changed = strcmp(pstStateExten->cExten, cExten);
	memset(pstStateExten->cExten, 0, sizeof(pstStateExten->cExten)); 
	strncpy(pstStateExten->cExten, cExten, strlen(cExten));
#if 0  /*for Bug 49909 */
	if ((aor = ast_sip_location_retrieve_aor(endpt_id)) && (contact = ast_sip_location_retrieve_first_aor_contact(aor))) {
		uri = (char*)contact->uri;
		if (!exten_changed && pstStateExten->bReload && (pstStateExten->enType == EN_EVENTLIST_REMOTE_USER)) {
			exten_changed = strcmp(pstStateExten->cHostName, uri);
		}

		ast_string_field_set(pstStateExten, cHostName, uri); 
    }
#endif
	eventlist_set_exten_type(pstStateExten, endpt_id, EN_EVENTLIST_REMOTE_USER, exten_changed);
	ast_string_field_build(pstStateExten, cEndpoint, "%s", endpt_id);
	ast_log(LOG_VERBOSE, "o-exten[%s], exten[%s], peer[%s]\n",pstStateExten->cOriginExten, pstStateExten->cExten, pstStateExten->cEndpoint);
	return 0;
}

#define ______EVENTLIST_ENHANCED_FEATURE_______

static int eventlist_get_device(char *pExten, char *pContext ,struct ast_exten *pstExten, struct pbx_find_info *pstFindInfo)
{
	int    iPriority = 0;
	char   cContext[80] = {0};

	if (NULL == pExten || NULL == pContext || NULL == pstExten || NULL == pstFindInfo) {
		return -1;
	}

	if (!strcmp(pstExten->exten, pExten)) {
		return 0;
	}

    /*! continue searching in context */
	strncpy(cContext, pstFindInfo->foundcontext, sizeof(cContext));

	pstFindInfo->incstack[pstFindInfo->stacklen++] = pContext;	/* Setup the stack */
	for (iPriority = 1; ;iPriority++) {
		if ((pstExten = pbx_find_extension(NULL, NULL, pstFindInfo, cContext, pExten, iPriority, NULL, NULL, E_MATCH))) {
			if (!strcmp(pstExten->exten, pExten)) {
				return 0;
			}
		}
	}

    /*! not found*/
	return -1;
}
static int eventlist_park_pos_check(char* exten)
{
	struct ast_flags config_flags = {0};
	struct ast_config    *pstCfg  = NULL;
	struct ast_variable  *pstVarGen = NULL;
	int gs_park_pos_start = 0;
	int gs_park_pos_end = 0;
	int extennum = 0;
	
    /* load features.conf file*/
	if ( exten == NULL || exten[0] == '0') {
		return -1;
	}
	extennum = atoi(exten);
	pstCfg = ast_config_load(RES_PARKING_CONFIG_FILE, config_flags);

	if (pstCfg == CONFIG_STATUS_FILEUNCHANGED || NULL == pstCfg) {
		return -1;
	} 

	if (pstCfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Contents of res_parking.conf are invalid and cannot be parsed\n");
		return -1;
	}
	pstVarGen = ast_variable_browse(pstCfg, "default");

	for (; (NULL != pstVarGen); pstVarGen = pstVarGen->next) {
		if (0 == strcasecmp(pstVarGen->name, "parkpos")) {
			if (NULL == pstVarGen->value) {
				break;
			}

			sscanf(pstVarGen->value, "%30d-%30d", &gs_park_pos_start, &gs_park_pos_end);			
			break;
		}    
	}

	ast_config_destroy(pstCfg);
	if (extennum >= gs_park_pos_start && extennum <= gs_park_pos_end) {
		return 0;
	}

	return -1;
}

/*! to check the user type of exten */
static int eventlist_check_device(char *pExten, struct st_state_exten  *pstStateExten)
{
	struct pbx_find_info   stFindInfo = { .stacklen = 0 }; /* the rest is reset in pbx_find_extension */
	struct ast_exten       *pstExten = NULL;

	if ((NULL == pExten) || ((NULL != pstStateExten) && (0 != strcmp(pExten, pstStateExten->cExten)))) {
		return -1;
	}

	stFindInfo.stacklen = 0;
	pstExten = pbx_find_extension(NULL, NULL, &stFindInfo, SIP_LOCAL_CONTEXT, pExten, 1, NULL, NULL, E_MATCH);
	if (NULL != pstExten) {
		if ((NULL != strstr(stFindInfo.foundcontext, SIP_EXT_CONFERENCE)) &&
			(0 == eventlist_get_device(pExten, SIP_EXT_CONFERENCE, pstExten, &stFindInfo))) {
			if (pstStateExten) {
				eventlist_set_exten_type(pstStateExten, NULL, EN_EVENTLIST_CONF_USER, 0);
				ast_string_field_build(pstStateExten, cDeviceName, "confbridge:%s", pExten);
			}

			ast_log(LOG_NOTICE, "get conf user exten[%s]\n", pExten);
			return EN_EVENTLIST_CONF_USER;
		}

		if ((NULL != strstr(stFindInfo.foundcontext, SIP_EXT_QUEUES)) &&
			(0 == eventlist_get_device(pExten, SIP_EXT_QUEUES, pstExten, &stFindInfo))) {
			if (pstStateExten) {
				eventlist_set_exten_type(pstStateExten, NULL, EN_EVENTLIST_QUEUE_USER, 0);
				ast_string_field_build(pstStateExten, cDeviceName, "queue:%s", pExten);
			}

			ast_log(LOG_NOTICE, "get queue user exten[%s]\n", pExten);
			return EN_EVENTLIST_QUEUE_USER;
		}

		if ((NULL != strstr(stFindInfo.foundcontext, SIP_EXT_GROUP)) &&
			(0 == eventlist_get_device(pExten, SIP_EXT_GROUP, pstExten, &stFindInfo))) {
			if (pstStateExten) {
				eventlist_set_exten_type(pstStateExten, NULL, EN_EVENTLIST_RGROUP_USER, 0);
				ast_string_field_build(pstStateExten, cDeviceName, "ringgroup:%s", pExten);
			}

			ast_log(LOG_NOTICE, "get ringgroup user exten[%s]\n", pExten);
			return EN_EVENTLIST_RGROUP_USER;
		}

		if (0 == eventlist_park_pos_check(pExten)) {
			if (pstStateExten) {
				eventlist_set_exten_type(pstStateExten, NULL, EN_EVENTLIST_PARK_USER, 0);
				ast_string_field_build(pstStateExten, cDeviceName, "park:%s@parkedcalls", pExten);
			}

			ast_log(LOG_NOTICE, "get park pos[%s]\n", pExten);
			return EN_EVENTLIST_PARK_USER;
		}
	}

    return -1;
}
#define ______EVENTLIST_PROCESS_______

/*************************************************
 * Function Name : eventlist_build 
 * Description: build a struct sip_eventlist according to the variable
 * Input: cListName, the name of the eventlist 
 *        pstVar
 *        pstVarregs 
 * Output:  NA 
 * Changes: 
*************************************************/
static struct sip_eventlist *eventlist_build(const char *cListName, struct ast_variable *pstVar, struct ast_variable *pstVarregs)
{
	struct sip_eventlist  *pstEventlist  = NULL;
	struct sip_eventlist   stEventlist;
	struct st_state_exten *pstStateExten = NULL;
	int    bFound      = FALSE;
	int    bExtenOK    = FALSE;

	ast_copy_string(stEventlist.cName, cListName, sizeof(stEventlist.cName));
	pstEventlist = ao2_t_find(con_eventlists, &stEventlist, OBJ_POINTER | OBJ_UNLINK, "find and unlink eventlist from eventlists table");

	if (NULL == pstEventlist) {
        /* alloc a new eventlist */
		if (!(pstEventlist = ao2_t_alloc_options(sizeof(struct sip_eventlist), eventlist_destroy_list, AO2_ALLOC_OPT_LOCK_RWLOCK, "allocate a eventlist"))) {
			return NULL;
		}

        /* init the new eventlist struct */
		strncpy(pstEventlist->cName, cListName, sizeof(pstEventlist->cName));
		pstEventlist->enType = EN_EVENTLIST_DIALOG; /* default event type */
		pstEventlist->bMark  = FALSE;
		memset(pstEventlist->cExtenlistStr, 0, sizeof(pstEventlist->cExtenlistStr));

        /* Initialize list of eventlist */
		AST_LIST_HEAD_INIT(&pstEventlist->ast_exten_list);
		AST_VECTOR_INIT(&pstEventlist->subscribers, 10);
    } else {
		pstEventlist->bMark = FALSE;   /* clear the mark flag */
		bFound = TRUE;
	}

	for (; pstVar || ((pstVar = pstVarregs) && !(pstVarregs = NULL)); pstVar = pstVar->next) {
		if (!strcasecmp(pstVar->name, "event")) {
			pstEventlist->enType = (!strcmp(pstVar->value, "dialog"))? EN_EVENTLIST_DIALOG : EN_EVENTLIST_BUTT;
		} else if (!strcasecmp(pstVar->name, "exten") && !ast_strlen_zero(pstVar->value)) {
			bExtenOK = TRUE;
			strncpy(pstEventlist->cExtenlistStr, pstVar->value, sizeof(pstEventlist->cExtenlistStr));
			eventlist_create_exten_list(pstEventlist, pstVar->value);
		}
	}

    /*clear the invalid list */
	if (!bExtenOK) {
		if (bFound) {
			pstEventlist->bMark = 1;   /* set the mark flag */
		}

		eventlist_unref(pstEventlist, "failed to get exten value, drop eventlist");
		return NULL;
	}

    /*subsribe device state event */
	AST_LIST_LOCK(&pstEventlist->ast_exten_list);
	AST_LIST_TRAVERSE(&(pstEventlist->ast_exten_list), pstStateExten, entry) {
		if ((pstStateExten->enType == EN_EVENTLIST_LOCAL_USER) && (0 == pstStateExten->bReload)) {
			manager_event(EVENT_FLAG_SYSTEM, "EventlistMemberStatus", "List: %s\r\nExten: %s\r\nLocation: local\r\nStatus: %s\r\n", 
						pstEventlist->cName, pstStateExten->cExten, eventlist_get_dialogstate_by_extenstate(pstStateExten->enState)); 
			ast_log(LOG_DEBUG, "eventlist_build, Exten[%s], stateid[%d].\n", pstStateExten->cExten, pstStateExten->iStateID);
		}
	}
	AST_LIST_UNLOCK(&pstEventlist->ast_exten_list);

	return pstEventlist;
}

/*! find the eventlist that contain the exten */
static struct pjsip_sub_pvt *sub_pvt_find_by_sub(struct ast_sip_subscription *sub)
{
	struct sip_eventlist  *pstList  = NULL;
	struct st_state_exten *pstExten = NULL;
	struct pjsip_sub_pvt  *pstPvt   = NULL;
	struct ao2_iterator    stIterator;  
	int    bFound = FALSE;  

	if (sub == NULL) {
		return NULL;
	}

	pstPvt = (struct pjsip_sub_pvt *)ast_sip_get_userdata(sub);
	if (NULL != pstPvt) {
		if (pstPvt->pSub == sub) {
			return pstPvt;
		}
	}

	stIterator = ao2_iterator_init(con_eventlists, 0);
	while ((pstList = ao2_t_iterator_next(&stIterator, "iterate thru eventlists table"))) {
		AST_LIST_TRAVERSE(&(pstList->ast_exten_list), pstExten, entry) {
			pstPvt = pstExten->pPvt;
			if ((NULL != pstPvt) && (pstPvt->pSub == sub)) {
				bFound = TRUE;
				break;
			}
		}

		eventlist_unref(pstList, "Done with iterator");

		if (bFound) {
			break;
		}
	}

	ao2_iterator_destroy(&stIterator);

	if (bFound) {
		return pstPvt;
	}
	
	return NULL;
}


static int pjsip_event_create_subscription(void *obj)
{
	struct pjsip_sub_pvt    *pstPvt = obj;
	struct st_state_exten   *pstExten;
	RAII_VAR(struct ast_sip_endpoint *, pstEndpoint, NULL, ao2_cleanup);

	if (!pstPvt || !pstPvt->pstRemoteExten) {
		return -1;
	}

	pstExten = pstPvt->pstRemoteExten;
	ast_log(LOG_NOTICE, "pjsip_event_create_subscription, endpoint[%s], resource[%s], expires[%d]", 
            pstExten->cEndpoint, pstExten->cExten, pstExten->iExpires);
	pstEndpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", pstPvt->pstRemoteExten->cEndpoint);

	if (NULL == pstEndpoint) {
		return -1;
	}

	pstPvt->pSub = ast_sip_create_subscription(&event_handler, pstEndpoint, pstExten->cExten, pstExten->iExpires);
	if (!pstPvt->pSub) {
		ast_log(LOG_WARNING, "create subscription failed!");
		return -1;
	}

	return 0;
}

/*! find the eventlist that contain the exten */
static struct sip_eventlist *eventlist_find_by_name(const char *name)
{
	struct sip_eventlist  *pstList  = NULL;
	struct ao2_iterator    stIterator;  
	int    bFound = FALSE;  

	if (name == NULL) {
		return 0;
	}
         
	stIterator = ao2_iterator_init(con_eventlists, 0);
	while ((pstList = ao2_t_iterator_next(&stIterator, "iterate thru eventlists table"))) {
		if (!strcmp(pstList->cName, name)) {
			bFound = TRUE;
			eventlist_unref(pstList, "Done with iterator");
			break;
		}

		eventlist_unref(pstList, "Done with iterator");
	}
	ao2_iterator_destroy(&stIterator);

	if (bFound) {
		return pstList;
	}
	
	return NULL;
}

/*! find the eventlist that contain the exten */
static struct sip_eventlist *eventlist_find_by_exten_st(struct st_state_exten *pstExten)
{
	struct sip_eventlist  *pstList  = NULL;
	struct st_state_exten *pstExten0 = NULL;
	struct ao2_iterator    stIterator;  
	int    bFound = FALSE;  

	if (pstExten == NULL) {
		return 0;
	}
         
	stIterator = ao2_iterator_init(con_eventlists, 0);
	while ((pstList = ao2_t_iterator_next(&stIterator, "iterate thru eventlists table"))) {
		AST_LIST_TRAVERSE(&(pstList->ast_exten_list), pstExten0, entry) {
			if (pstExten0 == pstExten) {
				bFound = TRUE;
				break;
			}
		}

		eventlist_unref(pstList, "Done with iterator");

		if (bFound) {
			break;
		}
	}

	ao2_iterator_destroy(&stIterator);

	if (bFound) {
		return pstList;
	}
	
	return NULL;
}

/*! this func is used for compare two eventlist's name  */
static int eventlist_comparefunc(const void *a, const void *b)
{
	struct sip_eventlist **ap = (struct sip_eventlist **)a;
	struct sip_eventlist **bp = (struct sip_eventlist **)b;

	if (*ap == NULL || *bp == NULL) {
		return -1;
	}

	return strcmp((*ap)->cName, (*bp)->cName);
}

/*! \brief Send a subscription or resubscription for dialog */
static int pjsip_subscribe_check_timeout(const void *data)
{
	struct st_state_exten  *pstExten = (struct st_state_exten  *)data;
	struct sip_eventlist   *pstEventlist = NULL;

	if (NULL == pstExten) {
		return 0;
	}

	if (TRUE == g_loading_flag) {
		ast_log(LOG_NOTICE, "pjsip_subscribe_check_timeout, the module is reloading, do nothing!\n");
		return 0;
	}

	pstEventlist = eventlist_find_by_exten_st(pstExten);
	if (NULL == pstEventlist) {
		return 0;
	}

	pstExten->iSchedID = -1;
	if (NULL != pstExten->pPvt) {
		ast_log(LOG_NOTICE, "pjsip_subscribe_check_timeout, exten[%s] with a pvt, just add schedule\n", pstExten->cExten);
		if ((pstExten->iSchedID = ast_sched_add(sched, SIP_REMOTE_SUBS_EXPIRES * 1000/2, pjsip_subscribe_check_timeout, pstExten)) < 0) {
			ast_log(LOG_ERROR, "pjsip_subscribe_check_timeout, add schedule failed\n");
		}
		return 0;
	}

	ast_log(LOG_DEBUG, "pjsip_subscribe_check_timeout, try to send subscribe to [%s]\n", pstExten->cExten);

	pjsip_do_remote_blf_subscribe(pstEventlist, pstExten);

	return 0;
}

/*! \brief send blf subscribe */
int pjsip_do_remote_blf_subscribe(struct sip_eventlist  *pstEventlist, struct st_state_exten *pstStateExten)
{
	struct pjsip_sub_pvt    *pstPvt = NULL;

	if (NULL == pstStateExten) {
		return -1;
	}

	/* If we already have a subscription, wait for pjsip stack to do refresh subscribe */
	if (NULL != pstStateExten->pPvt) {
		ast_log(LOG_NOTICE, "pjsip_do_remote_blf_subscribe, exten[%s] already have a subscribe!", pstStateExten->cExten);
		return 0;
	}

	pstPvt = ao2_alloc(sizeof(*pstPvt), NULL);
	if (!pstPvt) {
		return -1;
	}

	pstPvt->pstRemoteExten = pstStateExten;
	pstPvt->pstEventlist   = pstEventlist;
	pstStateExten->pPvt    = pstPvt;

	if (ast_sip_push_task_synchronous(NULL, pjsip_event_create_subscription, pstPvt)) {
		ao2_ref(pstPvt, -1);
		return -1;
	}

	if (pstStateExten->iSchedID < 0) {
		pstStateExten->iSchedID = ast_sched_add(sched, SIP_REMOTE_SUBS_EXPIRES * 1000 / 2, pjsip_subscribe_check_timeout, pstStateExten); 
		if (pstStateExten->iSchedID < 0) {
			ast_log(LOG_ERROR, "pjsip_do_remote_blf_subscribe, add schedule failed\n");
		}
	}

	return 0;
}

/*! \brief Get state enum expression */
static int eventlist_get_extenstate_by_string(char* pState)
{
	if (!pState) {
		return -1;
	}

	if (!strcmp(pState, "confirmed")) {
		return AST_EXTENSION_INUSE;
	}

	if (!strcmp(pState, "early")) {
		return AST_EXTENSION_RINGING;
	}

	if (!strcmp(pState, "terminated")) {
		return AST_EXTENSION_NOT_INUSE;
	}

	if (!strcmp(pState, "offline")) {
		return AST_EXTENSION_UNAVAILABLE;
	}

    /* all other string act as not_inuse */
	return AST_EXTENSION_NOT_INUSE;
}

static void pjsip_event_subscriber_record(const char *eventlist, const char *subscriber, int add)
{
	struct sip_eventlist *pstList = eventlist_find_by_name(eventlist);
	int i;

	if (!pstList || !subscriber) {
		return;
	}

	ast_log(LOG_NOTICE, "pjsip_event_subscriber_record, eventlist[%s], subscriber[%s], add[%d]\n", eventlist, subscriber, add);
	for (i = 0; i < AST_VECTOR_SIZE(&pstList->subscribers); ++i) {
		if (!strcmp(subscriber, AST_VECTOR_GET(&pstList->subscribers, i))) {
			if (!add) {
				ast_free((void *)AST_VECTOR_GET(&pstList->subscribers, i));
				AST_VECTOR_REMOVE_UNORDERED(&pstList->subscribers, i);
			}
			return; 
		}
	}

	if (add) {
		AST_VECTOR_APPEND(&pstList->subscribers, ast_strdup(subscriber)); 
	}

	return;
}

static void pjsip_event_subscriber_destroy(struct ast_sip_subscription *sub)
{
	struct st_state_exten *pstExten = NULL;
	struct pjsip_sub_pvt  *pvt = sub_pvt_find_by_sub(sub);

	if (!pvt) {
		return;
	}

	pstExten = pvt->pstRemoteExten;
	if (pstExten) {
		pstExten->pPvt = NULL;
		pvt->pSub = NULL;
		ao2_ref(pvt, -1);

		pstExten->iExpires = ast_sip_get_sub_expires(sub);
		if (-1 != pstExten->iExpires) {
			/* wait 1 s and try again*/
			pstExten->iSchedID = ast_sched_replace(pstExten->iSchedID, sched, 1000, pjsip_subscribe_check_timeout, pstExten);
			if (pstExten->iSchedID < 0) {
				ast_log(LOG_ERROR, "pjsip_event_subscriber_destroy, Exten[%p][%s] replace sched failed", pstExten, pstExten->cExten);
			}
		}
		ast_log(LOG_NOTICE, "pjsip_event_subscriber_destroy, Exten[%p][%s], iExpires[%d], iSchedID[%d]", 
                pstExten, pstExten->cExten, pstExten->iExpires, pstExten->iSchedID);
	}
	return;
}

static void pjsip_event_subscriber_state_change(struct ast_sip_subscription *sub, pjsip_msg_body *body, enum pjsip_evsub_state state)
{
	char *pEntityStart = NULL;
	char *pEntityEnd   = NULL;
	char *pStateStart  = NULL;
	char *pStateEnd    = NULL;
	char *pState       = NULL;
	char  eState[20]   = {0};
	char  cExten[80]   = {0};
	int   iState = 0;
	int   iExtenLen = sizeof(cExten);
	struct st_state_exten *pstExten = NULL;
	struct pjsip_sub_pvt  *pvt;

	if (TRUE == g_loading_flag) {
		ast_log(LOG_NOTICE, "state_change, during reloading, do nothing!\n");
		return;
	}

	pvt = sub_pvt_find_by_sub(sub);

	if ((!pvt) || (!body)) {
		ast_log(LOG_WARNING, "can't find pvt by subscription, resource[%s]\n", ast_sip_subscription_get_resource_name(sub));
		return;
	}

	ast_log(LOG_NOTICE, "the content type is [%.*s/%.*s]\n", 
				(int) pj_strlen(&body->content_type.type), pj_strbuf(&body->content_type.type),
				(int) pj_strlen(&body->content_type.subtype), pj_strbuf(&body->content_type.subtype));

	/* check content type*/
	if (0 != pj_stricmp2(&body->content_type.type, EVENT_TYPE) || 0 != pj_stricmp2(&body->content_type.subtype, EVENT_SUBTYPE)) {
		ast_log(LOG_WARNING, "the content type is [%s/%s]\n", body->content_type.type.ptr, body->content_type.subtype.ptr);
		return;
	}

	ast_log(LOG_NOTICE, "pjsip_event_subscriber_state_change start process\n");
    /* get entity attribute */
	pEntityStart = strstr((const char*)body->data, "entity=");
	
	if (!pEntityStart) {
		return;
	}

	pEntityStart += (strlen("entity=") + 5);   //entity="sip:user@host", offset to the user part
	pEntityEnd = strchr(pEntityStart, '@');
	if (!pEntityEnd) {
		return;
	}

	strncpy(cExten, pEntityStart, (((pEntityEnd - pEntityStart) >= iExtenLen) ? (iExtenLen-1) : (pEntityEnd - pEntityStart)));

	/* get body data */
	pStateStart = strstr((const char*)body->data, "<state");
	pStateEnd   = strstr((const char*)body->data, "</state>");

	if (!pStateStart || !pStateEnd) {
		ast_log(LOG_WARNING, "no event-state in the body[%s]\n", (char *)body->data);
		return;
	}

	pState = strchr(pStateStart, '>');

	if (!pState || (pState > pStateEnd)) {
		return;
	}

	pState++;
	strncpy(eState, pState, (((pStateEnd - pState) >= sizeof(eState)) ? (sizeof(eState)-1) : (pStateEnd - pState)));
	
	pstExten = pvt->pstRemoteExten;

	if (NULL == pstExten || NULL == pvt->pstEventlist) {
		return;
	}

	if (strcmp(cExten, pstExten->cExten)) {
		ast_log(LOG_ERROR, "handle_dialog_notify, different exten: cExten[%s], e_exten[%s] \n", cExten, pstExten->cExten);
		return;
	}

	if (ast_strlen_zero(pstExten->cDeviceName)) {
		char *pIdentityStart  = NULL;
		char *pIdentityEnd    = NULL;
		char *pDisplay        = NULL;
		char *pDisplayEnd     = NULL;
		char  cDeviceName[80] = {0};
		int   len = 0;

		pIdentityStart = strstr((const char*)body->data, "<identity");
		pIdentityEnd   = strstr((const char*)body->data, "</identity>");
		if (pIdentityStart && pIdentityEnd) {
			pDisplay = strstr((const char*)pIdentityStart, "display=\"");
			if (pDisplay) {
				pDisplayEnd = strstr((const char*)pDisplay, "\">");
				if (pDisplayEnd) {
					pDisplay = pDisplay+strlen("display=\"");
					len = pDisplayEnd-pDisplay;

					strncpy(cDeviceName, pDisplay, ((len >= sizeof(cDeviceName)) ? (sizeof(cDeviceName)-1):len));
					ast_string_field_build(pstExten, cDeviceName, "%s", cDeviceName); 
					ast_log(LOG_NOTICE, "Get exten[%s]--display[%s] from NOTIFY body", pstExten->cExten, cDeviceName);
				}
			}
		}
	}
	ast_sip_set_userdata(sub, (void *)pvt); 

	iState = eventlist_get_extenstate_by_string(eState);
	ast_sip_set_remote_exten_state(sub, pstExten->cOriginExten, iState, pstExten->cDeviceName);

	/* check the state, if changed, update state value */
	if (pstExten->enState != iState) {
		ast_log(LOG_VERBOSE, "handle_dialog_notify, exten[%s], old state[%d] --> new state[%d]\n", cExten, pstExten->enState,iState);
		pstExten->enState = iState;

        /* manager event report */
		manager_event(EVENT_FLAG_SYSTEM, "EventlistMemberStatus", "List: %s\r\nExten: %s\r\nLocation: remote\r\nTrunk: %s\r\nStatus: %s\r\n",
				pvt->pstEventlist->cName, pstExten->cOriginExten, pstExten->cEndpoint, eventlist_get_dialogstate_by_extenstate(iState));

		ast_sip_send_remote_notify(pvt->pstEventlist->cName, (char *)pstExten->cOriginExten, iState);
	}

	return;
}

static void eventlist_set_exten_type(struct st_state_exten *pstStateExten, char *device, EN_EVENTLIST_USER_TYPE enType, int exten_changed)
{
	if (NULL == pstStateExten)  {
		return;
	}

	if (0 == pstStateExten->bReload) {
		ast_log(LOG_NOTICE, "eventlist_set_exten_type, cExten[%s], type[%d]\n", pstStateExten->cExten, enType);
		pstStateExten->enType = enType;
		return;
	}

	ast_log(LOG_NOTICE, "eventlist_set_exten_type, cExten[%s], device[%s]-->[%s], type[%d]-->[%d], exten_changed[%d]\n", pstStateExten->cExten,
		   pstStateExten->cEndpoint, device, pstStateExten->enType, enType, exten_changed);

	if ((pstStateExten->enType == enType) && (enType != EN_EVENTLIST_REMOTE_USER)) {
		return;
	}

	if (pstStateExten->enType != EN_EVENTLIST_REMOTE_USER) {
		pstStateExten->iStateID = -1;
		pstStateExten->enState = AST_EXTENSION_UNAVAILABLE;
		pstStateExten->enType = enType;
		return;
	}

	if ((enType != EN_EVENTLIST_REMOTE_USER) || exten_changed || ((NULL != device) && strcmp(device, pstStateExten->cEndpoint))) {
		ast_log(LOG_NOTICE, "eventlist_set_exten_type, clear remote subscribe, endpoint[%s]\n", pstStateExten->cEndpoint);
		ast_sip_push_task_synchronous(NULL, eventlist_clear_remote_subscribe, pstStateExten);
		pstStateExten->enType = enType;
		pstStateExten->enState = AST_EXTENSION_UNAVAILABLE;

		if (-1 != pstStateExten->iSchedID) {
			if (0 != AST_SCHED_DEL(sched, pstStateExten->iSchedID)) {
				ast_log(LOG_ERROR, "remote subscribe sched[%d] del failed!\n", pstStateExten->iSchedID);
			}
			pstStateExten->iSchedID = -1;
		}

		if (enType != EN_EVENTLIST_REMOTE_USER)  {
			ast_copy_string(pstStateExten->cExten, pstStateExten->cOriginExten, sizeof(pstStateExten->cExten)); 
			ast_string_field_set(pstStateExten, cEndpoint, NULL); 
		}
		return;
	}

	return;
}

/*************************************************
 * Function Name : eventlist_do_state 
 * Description: Go through the eventlist queue and pro state subscribe changes of the eventlist
 * Input:  data 
 * Output: NA 
 * Changes: 
*************************************************/
static void *eventlist_do_state(void *data)
{
	struct sip_eventlist   *pstEventlist = NULL;
	struct st_state_exten  *pstStateExten = NULL;
	struct ast_exten       *pstExten = NULL;
	struct pbx_find_info    stFindInfo = { .stacklen = 0 }; /* the rest is reset in pbx_find_extension */
	char   hint[80]  = {0};
	struct ao2_iterator     stIterator;
	int    do_something = TRUE;   /* to indicate if the state exten is update here */

    for (;;) 
    {
		if (TRUE == g_loading_flag) {
			usleep(100000); /* sleep 100 ms */
			continue;
		}

		if (FALSE == g_doing_state) {
			usleep(100000); /* sleep 100 ms */
			continue;
		}

		do_something = FALSE;
		stIterator = ao2_iterator_init(con_eventlists, 0);
		while ((pstEventlist = ao2_t_iterator_next(&stIterator, "iterate thru eventlists table"))) {
			ao2_rdlock(pstEventlist);

			AST_LIST_LOCK(&(pstEventlist->ast_exten_list));
			AST_LIST_TRAVERSE(&(pstEventlist->ast_exten_list), pstStateExten, entry) {
				if (-1 == pstStateExten->iStateID || pstStateExten->bReload) {
					if ((pstStateExten->enType == EN_EVENTLIST_UNKNOWN_USER || pstStateExten->bReload) && 
					    (pstStateExten->iStateErrCount < SIP_MAX_STATE_ERR_NUM)) {
						do_something = TRUE;

						if (0 != ast_get_hint(hint, sizeof(hint), NULL, 0, NULL, SIP_DEFAULT_CONTEXT, pstStateExten->cOriginExten)) {				   	
							eventlist_set_exten_type(pstStateExten, NULL, EN_EVENTLIST_LOCAL_USER, 0);
							pstStateExten->bReload = 0;
							pstStateExten->bTypeUpdate = 1;
							pstStateExten->iStateID = 1;
						} else {					   	
							pstStateExten->iStateErrCount++;
							stFindInfo.stacklen = 0;						
							pstExten = pbx_find_extension(NULL, NULL, &stFindInfo, SIP_LOCAL_CONTEXT, pstStateExten->cOriginExten, 1, NULL, NULL, E_MATCH);
							if (NULL != pstExten) {							
                               /* BEGIN: Modified by yfyang, Grandstream, [Bug 21916], for eventlist enhancement feature */
								if (-1 != eventlist_check_device((char *)pstStateExten->cOriginExten, pstStateExten)) {
									pstStateExten->bReload = 0;
									pstStateExten->bTypeUpdate = 1;
									pstStateExten->iStateID = 1;
								} else if (((NULL != strstr(stFindInfo.foundcontext, "outrt-")) || (NULL != strstr(stFindInfo.foundcontext, "outbound-allroutes"))) && 
									(0 == eventlist_get_endpointname(pstStateExten, pstExten))) {
									pstStateExten->bTypeUpdate = 1;
                                   continue;
								} else {
									continue;
								}
                               /* END: Modified by yfyang, Grandstream, for eventlist enhancement feature */
							} else{
								continue;
							}
						}
					}
                   /* [Bug 21527] eventlist:the status of remote members can not update when delete outbound route/trunk */
					else if ((pstStateExten->enType == EN_EVENTLIST_UNKNOWN_USER || 0 == pstStateExten->bTypeUpdate) && (pstStateExten->iStateErrCount == SIP_MAX_STATE_ERR_NUM)) {
						eventlist_set_exten_type(pstStateExten, NULL, EN_EVENTLIST_UNKNOWN_USER, 0);
						do_something = TRUE;

						pstStateExten->iStateErrCount++;
						pstStateExten->bReload = 0;
						manager_event(EVENT_FLAG_SYSTEM, "EventlistMemberStatus", "List: %s\r\nExten: %s\r\nLocation: %s\r\nStatus: %s\r\n", 
								pstEventlist->cName, pstStateExten->cOriginExten, eventlist_extentype_2_str(pstStateExten->enType),eventlist_get_dialogstate_by_extenstate(pstStateExten->enState)); 
					}

					if (pstStateExten->bTypeUpdate && (pstStateExten->enType != EN_EVENTLIST_REMOTE_USER)) {
						pstStateExten->enState = ast_extension_state(NULL, eventlist_get_context_by_usertype(pstStateExten->enType), pstStateExten->cExten);
						pstStateExten->bTypeUpdate = 0;
						do_something = TRUE;
						if ((AST_EXTENSION_UNAVAILABLE == pstStateExten->enState) && (pstStateExten->enType == EN_EVENTLIST_CONF_USER)) {
							pstStateExten->enState = AST_EXTENSION_NOT_INUSE;
						}

                        /* manage event report*/
						manager_event(EVENT_FLAG_SYSTEM, "EventlistMemberStatus", "List: %s\r\nExten: %s\r\nLocation: %s\r\nStatus: %s\r\n", 
									pstEventlist->cName, pstStateExten->cOriginExten, eventlist_extentype_2_str(pstStateExten->enType),eventlist_get_dialogstate_by_extenstate(pstStateExten->enState)); 
					}
				}
			}
            /* send one subscriber once */
			AST_LIST_TRAVERSE(&(pstEventlist->ast_exten_list), pstStateExten, entry) {
				if ((-1 == pstStateExten->iStateID || (pstStateExten->bReload && pstStateExten->bTypeUpdate)) && (pstStateExten->enType == EN_EVENTLIST_REMOTE_USER)) {
	                 /* manage event report remote initial state*/
					manager_event(EVENT_FLAG_SYSTEM, "EventlistMemberStatus", "List: %s\r\nExten: %s\r\nLocation: remote\r\nTrunk: %s\r\nStatus: %s\r\n", 
								pstEventlist->cName, pstStateExten->cOriginExten, 
								pstStateExten->cEndpoint, eventlist_get_dialogstate_by_extenstate(pstStateExten->enState)); 

					do_something = TRUE;
					pstStateExten->bReload = 0;
					if ((-1 == pstStateExten->iStateID) || (NULL == pstStateExten->pPvt)) {
						/* send remote subscribe msg */
						ast_log(LOG_NOTICE, "eventlist_do_state, start to send subscribe to exten[%s].\n", pstStateExten->cExten);
						pstStateExten->iStateID = pjsip_do_remote_blf_subscribe(pstEventlist, pstStateExten); 
                        usleep(100000); /* sleep 100 ms */
					}
				}
			}
			AST_LIST_UNLOCK(&(pstEventlist->ast_exten_list));
			ao2_unlock(pstEventlist);
			eventlist_unref(pstEventlist, "Done with iterator");
		}

		ao2_iterator_destroy(&stIterator);

		if (FALSE == do_something) {
			ast_log(LOG_NOTICE, "eventlist_do_state, Set g_doing_state = FALSE.\n");
			g_doing_state = FALSE;
		}
		usleep(100000); /* sleep 100 ms */
	}

    /* Never reached */
	return NULL;
}

#define ______EVENTLIST_CLI________

#define LISTS_FORMAT  "%-25.25s %-10.10s %-45.45s \n"

/* Show one list in detail (main function)*/
static char *_eventlist_show_list(int type, int fd, struct mansession *s, const struct message *m, int argc, const char *argv[])
{
	struct sip_eventlist   *pstEventlist  = NULL;
	struct st_state_exten  *pstStateExten = NULL;
	char   strExten[256]    = {0};
	char   strSubs[256]     = {0};
	const char  *subscriber = NULL;
    int    i;

	pstEventlist = eventlist_find_by_name(argv[3]);
	if (NULL == pstEventlist) {
		ast_cli(fd, "Eventlist %s not found.\n", argv[3]);
		ast_cli(fd, "\n");
		return CLI_SUCCESS;
	}

	ao2_rdlock(pstEventlist);

	for (i = 0; i < AST_VECTOR_SIZE(&pstEventlist->subscribers); ++i) {
		subscriber = AST_VECTOR_GET(&pstEventlist->subscribers, i);
		if ((strlen(strSubs) + strlen(subscriber)) < (sizeof(strSubs) -1)) {
			strcat(strSubs, subscriber);
			strcat(strSubs, ",");
		}
	}
	ast_cli(fd, "\n");
	ast_cli(fd, "  * Name       : %s\n", pstEventlist->cName);
	ast_cli(fd, "  Type         : %s\n", (pstEventlist->enType == EN_EVENTLIST_DIALOG) ? "dialog" : "unknown");
	ast_cli(fd, "  Exten list   : %s\n", pstEventlist->cExtenlistStr);
	ast_cli(fd, "  Subscriber   : %s\n", strSubs);
	ast_cli(fd, "  bMark        : %d\n", pstEventlist->bMark);

	AST_LIST_TRAVERSE(&(pstEventlist->ast_exten_list), pstStateExten, entry) {
		memset(strExten, 0, sizeof(strExten));

		snprintf(strExten, sizeof(strExten), "%s, name=%s, o_exten=%s, %s, %s, state id=%d, endpoint=%s, %d/%d/%d/%d/%d",
			pstStateExten->cExten,
			(ast_strlen_zero(pstStateExten->cDeviceName)) ? "null":pstStateExten->cDeviceName,
			pstStateExten->cOriginExten,
			eventlist_get_dialogstate_by_extenstate(pstStateExten->enState),
			eventlist_extentype_2_str(pstStateExten->enType),
			pstStateExten->iStateID,
			(ast_strlen_zero(pstStateExten->cEndpoint)) ? "null": pstStateExten->cEndpoint,
			pstStateExten->iSchedID,pstStateExten->iStateErrCount,pstStateExten->bMark,pstStateExten->bReload,pstStateExten->bTypeUpdate
			);

		ast_cli(fd, "  Exten state  : %s\n", strExten);
	}

	ao2_unlock(pstEventlist);
	ast_cli(fd, "\n");

	return CLI_SUCCESS;
}                               

/* CLI Show Eventlists command (main function)*/
static char *_eventlist_show_lists(int fd, int *total, struct mansession *s, const struct message *m, int argc, const char *argv[])
{
	int  iObjcount = ao2_container_count(con_eventlists);
	struct sip_eventlist **ppstEventlistarray;
	struct sip_eventlist  *pstEventlist;
	struct ao2_iterator    stIterator;
	char   cListname[80] = {0};
	char   cExtenlist[80]= {0};
	int    iEventlistNum = 0;
	int    k;

	if (s) {	
		ast_log(LOG_WARNING, "_eventlist_show_lists, don't support manager show now\n");
		return CLI_SUCCESS;
	}

	switch (argc) {
	case 5:
		if (!strcasecmp(argv[3], "like")) {
			return CLI_SHOWUSAGE;
		}
		break;

	case 3:
		break;

	default:
		return CLI_SHOWUSAGE;
	}

	ast_cli(fd, "\n");
	if (0 == iObjcount) {
		ast_cli(fd, "------------------------------------------\n");
		ast_cli(fd, "total 0 sip eventlists \n");

		return CLI_SUCCESS;
	}

	ast_cli(fd, LISTS_FORMAT, "Name", "Type", "Extensions");
	ast_cli(fd, LISTS_FORMAT, "----", "----", "----------");

	ppstEventlistarray = ast_calloc(sizeof(struct sip_eventlist *), iObjcount); 
	stIterator = ao2_iterator_init(con_eventlists, 0);

	while ((pstEventlist = ao2_t_iterator_next(&stIterator, "iterate thru eventlists table"))) {
		ppstEventlistarray[iEventlistNum++] = pstEventlist;
	}
	ao2_iterator_destroy(&stIterator);

	qsort(ppstEventlistarray, iEventlistNum, sizeof(struct sip_eventlist *), eventlist_comparefunc);

	for (k = 0; k < iEventlistNum; k++) {
		pstEventlist = ppstEventlistarray[k];
		
		ao2_rdlock(pstEventlist);

		ast_copy_string(cListname, pstEventlist->cName, sizeof(cListname));
		ast_copy_string(cExtenlist, pstEventlist->cExtenlistStr, sizeof(cExtenlist));

		ast_cli(fd, LISTS_FORMAT, 
				cListname,
				(pstEventlist->enType == EN_EVENTLIST_DIALOG) ? "dialog" : "unknown",
				cExtenlist);

		ao2_unlock(pstEventlist);
		pstEventlist = ppstEventlistarray[k] = eventlist_unref(pstEventlist, "toss iterator eventlist ptr");
	}

	ast_cli(fd, "------------------------------------------\n");
	ast_cli(fd, "total %d sip eventlists \n",iEventlistNum);


	if (total) {
		*total = iEventlistNum;
	}
	
	ast_free(ppstEventlistarray);
	
	return CLI_SUCCESS;
}

/* Show one list in detail */
static char *eventlist_show_list(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "eventlist show list";
		e->usage   = "Usage: eventlist show list <name> \n"
		     	"       Shows all details on one eventlist and the current status.\n";
		return NULL;

	case CLI_GENERATE:
		return NULL;
	}

	return _eventlist_show_list(0, a->fd, NULL, NULL, a->argc, (const char **) a->argv);
}


/*  CLI Show Eventlists command */
static char *eventlist_show_lists(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "eventlist show lists";
		e->usage   = "Usage: eventlist show lists [like <pattern>]\n"
			"       Lists all known Eventlists.\n"
			"       Optional regular expression pattern is used to filter the peer list.\n";
		return NULL;

	case CLI_GENERATE:
		return NULL;
	default:
		break;
	}
	return _eventlist_show_lists(a->fd, NULL, NULL, NULL, a->argc, (const char **) a->argv);
}

struct ast_cli_entry cli_eventlist[] = {
	AST_CLI_DEFINE(eventlist_show_lists, "List all configured eventlists"),
	AST_CLI_DEFINE(eventlist_show_list, "Show details on specific eventlist"),
};


static int action_eventlists(struct mansession *s, const struct message *m)
{
    const char *actionid = astman_get_header(m, "ActionID");
    const char *listname = astman_get_header(m, "List");
    char   idText[80]    = "";
    struct sip_eventlist   *pstEventlist;
    struct ao2_iterator     stIterator;
    struct st_state_exten  *pstStateExten = NULL;
    int    iObjcount = ao2_container_count(con_eventlists);
    int    exten_num = 0;
    int    subs_num  = 0;
    int    i;

    if (!ast_strlen_zero(actionid)) {
        snprintf(idText, sizeof(idText), "ActionID: %s\r\n", actionid);
    }

    astman_send_listack(s, m, "Eventlist subscriber and exten list will follow", "start");

    stIterator = ao2_iterator_init(con_eventlists, 0);
    while ((pstEventlist = ao2_t_iterator_next(&stIterator, "iterate thru eventlists table"))) {
        /* If we ask for one particular, and this isn't it, skip it */
        if (!ast_strlen_zero(listname) && strcmp(pstEventlist->cName, listname)) {
            continue;
        }

        ao2_lock(pstEventlist);

        /* send EventlistSubscriber */
        subs_num  = 0;
        for (i = 0; i < AST_VECTOR_SIZE(&pstEventlist->subscribers); ++i) {
            astman_append(s, "Event: EventlistSubscriber\r\n"
                             "%s"
                             "List: %s\r\n"
                             "Subscriber: %s\r\n"
                             "Status: Subscribe\r\n\r\n", 
                             idText, pstEventlist->cName, AST_VECTOR_GET(&pstEventlist->subscribers, i));

            subs_num++;
            if ((subs_num%5) == 0) {
                usleep(100000); /* sleep 100 ms */
            }
        }

        /* send EventlistMemberStatus */
        exten_num = 0;
        AST_LIST_TRAVERSE(&(pstEventlist->ast_exten_list), pstStateExten, entry) {
            astman_append(s, "Event: EventlistMemberStatus\r\n"
                                 "%s"
                                 "List: %s\r\n"
                                 "Exten: %s\r\n"
                                 "Location: %s\r\n"
                                 "Trunk: %s\r\n"
                                 "Status: %s\r\n\r\n", 
                                 idText,pstEventlist->cName, pstStateExten->cOriginExten, 
                                 eventlist_extentype_2_str(pstStateExten->enType),
                                 ast_strlen_zero(pstStateExten->cEndpoint) ? "null" : pstStateExten->cEndpoint,
                                 eventlist_get_dialogstate_by_extenstate(pstStateExten->enState)); 
            exten_num++;
            if ((exten_num%5) == 0) {
                usleep(100000); /* sleep 100 ms */
            }
        }

        ao2_unlock(pstEventlist);
        eventlist_unref(pstEventlist, "Done with iterator");

        usleep(100000); /* sleep 100 ms */
    }
    ao2_iterator_destroy(&stIterator);

    /* Send final confirmation */
    astman_append(s,
	"Event: EventlistsComplete\r\n"
	"EventList: Complete\r\n"
	"ListItems: %d\r\n"
	"%s"
	"\r\n", iObjcount, idText);

    return 0;
}

#define ______EVENTLIST_CONFIGURE________

/* load eventlist configure data, refresh eventlist cb */
static int eventlist_load_proc(int bReload)
{
	struct ast_flags config_flags = {0};
	struct ast_config    *pstCfg = NULL;
	struct ast_variable  *pstVarGen = NULL;
	struct sip_eventlist *pstEventlist = NULL;
	char  *pCategory       = NULL;
	int    iEventlistCount = 0;

    /* load eventlist.conf file*/
	pstCfg = ast_config_load(EVENTLIST_CONFIG_FILE, config_flags);

	if (pstCfg == CONFIG_STATUS_FILEUNCHANGED || NULL == pstCfg) {
		return -1;
	} else if (pstCfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Contents of pjsip_resource_list.conf are invalid and cannot be parsed\n");
		return -1;
    }

	pstVarGen = ast_variable_browse(pstCfg, "general");
	pCategory = ast_category_browse(pstCfg, NULL);

	while (pCategory) {
		if (strcasecmp(pCategory, "general")) {
			pstEventlist = eventlist_build(pCategory, pstVarGen, ast_variable_browse(pstCfg, pCategory));
			if (pstEventlist) {
				ao2_t_link(con_eventlists, pstEventlist, "link eventlist into eventlist table");
				eventlist_unref(pstEventlist, "eventlist_unref: from eventlist_load_proc");
				iEventlistCount++;
			}
		}
		pCategory = ast_category_browse(pstCfg, pCategory);
	}

	ast_config_destroy(pstCfg);

    /* Prune eventlists who still are supposed to be deleted */
	eventlist_unlink_marked_from_tables();
	ast_debug(LOG_VERBOSE, "--------------- Done destroy pruned eventlists\n");

	sleep(1); /* sleep 1 s */
	g_loading_flag = FALSE;    /* clear reloading flag */

	/* start a thread used for eventlists' state*/
	if (eventlist_state_thread != AST_PTHREADT_NULL) {
        /* Wake up the thread */
		pthread_kill(eventlist_state_thread, SIGURG);
	} else {
        /* Start a new thread */
		if (ast_pthread_create_background(&eventlist_state_thread, NULL, eventlist_do_state, NULL) < 0) {
			ast_log(LOG_ERROR, "Unable to start eventlist state thread.\n");
			return -1;
		}
	}

	return 0;
}

#define ______EVENTLIST_SUBMODULE INFO________

/* load eventlist submodule - initialization */
static int eventlist_load(void)
{
	int err = 0;
	ast_verbose("SIP Event list loading...\n");

	con_eventlists = ao2_t_container_alloc(HASH_EVENTLIST_SIZE, eventlist_hash_cb, eventlist_cmp_cb, "allocate eventlists");
	if (!con_eventlists) {
		ast_log(LOG_ERROR, "Unable to create primary eventlist container(s)\n");
		return AST_MODULE_LOAD_FAILURE;
	}

    //g_loading_flag = TRUE;    /* Added by yfyang, [Bug 17854], set loading flag */
	eventlist_loading_start = ast_tvnow();  /* Added by yfyang, Grandstream, [Bug 19182] */

	ast_cli_register_multiple(cli_eventlist, ARRAY_LEN(cli_eventlist));
	ast_manager_register_xml("Eventlists", EVENT_FLAG_SYSTEM, action_eventlists);

    /* reload eventlist configure info*/
	err = eventlist_load_proc(FALSE);
	if (err != 0) {
		return AST_MODULE_LOAD_FAILURE;
	}

	return 0;
}


/*! unload eventlist submodule */
static void eventlist_unload(void)
{
	ast_log(LOG_NOTICE, "unloading SIP Event list...");

    /* unlink all eventlist*/
	eventlist_unlink_all_from_tables();

	ast_manager_unregister("Eventlists");
	ast_cli_unregister_multiple(cli_eventlist, ARRAY_LEN(cli_eventlist));

    /* cancel eventlist state thread */
	if (eventlist_state_thread && (eventlist_state_thread != AST_PTHREADT_STOP) && (eventlist_state_thread != AST_PTHREADT_NULL)) {
		pthread_cancel(eventlist_state_thread);
		pthread_kill(eventlist_state_thread, SIGURG);
		pthread_join(eventlist_state_thread, NULL);
	}

	eventlist_state_thread = AST_PTHREADT_STOP;

	ao2_t_ref(con_eventlists, -1, "unref the eventlists table");
}
static int unload_module(void)
{
	eventlist_unload();
	if (sched) {
		ast_sched_context_destroy(sched);
		sched = NULL;
	}

	ast_sip_unregister_subscriber_info_handler(&info_handler);
	ast_extension_state_del(0, eventlist_extenstate);
	return 0;
}


static int load_module(void)
{	
	int err = 0;
	CHECK_PJSIP_SESSION_MODULE_LOADED();
	CHECK_PJSIP_PUBSUB_MODULE_LOADED();

	sched = ast_sched_context_create();
	if (!sched) {
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (ast_sched_start_thread(sched)) {
		ast_log(LOG_ERROR, "Could not start scheduler thread for publication expiration\n");
		unload_module();
		return AST_MODULE_LOAD_FAILURE;
	}

	ast_sip_register_subscriber_info_handler(&info_handler);

	err = eventlist_load();

	if (err) {
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	ast_extension_state_add(NULL, NULL, eventlist_extenstate, NULL);
	return AST_MODULE_LOAD_SUCCESS;
}

static int reload(void)
{
	ast_verbose("SIP Event list reloading...\n");

	g_loading_flag = TRUE;    /* set reloading flag */
	g_doing_state  = TRUE;
	ao2_t_callback(con_eventlists, OBJ_NODATA, eventlist_markall, NULL, "callback to mark all eventlists");

	eventlist_loading_start = ast_tvnow();  /* Added by yfyang, Grandstream, [Bug 19182] */

	eventlist_load_proc(TRUE);

	return 0;

}


AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP EVENTLIST",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
		.load_pri = AST_MODPRI_APP_DEPEND,
);



