/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2016, GrandStream, Inc.
 *
 * pxie <pxie@grandstream.cn> and others.
 *
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*!
 * \file
 * \brief Custom SQLite3 Chat records(Message).
 *
 * \author Adapted by Alejandro Rios <alejandro.rios@avatar.com.co> and
 *  Russell Bryant <russell@digium.com> from
 *  cdr_mysql_custom by Edward Eastman <ed@dm3.co.uk>,
 *	and cdr_sqlite by Holger Schurig <hs4233@mail.mn-solutions.de>
 */

/*** MODULEINFO
	<depend>sqlite3</depend>
	<support_level>core</support_level>
 ***/

#define AST_MODULE_LOG "messages"

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision$")

#include <sqlite3.h>
#include "asterisk/message.h"
#include "asterisk/paths.h"
#include "asterisk/channel.h"
#include "asterisk/module.h"
#include "asterisk/config.h"
#include "asterisk/pbx.h"
#include "asterisk/utils.h"
#include "asterisk/cli.h"
#include "asterisk/app.h"
#include "asterisk/astdb.h"

/*** DOCUMENTATION
	
 ***/

static sqlite3 *messdb = NULL;

static char *message_chat_record_db = "		\
CREATE TABLE IF NOT EXISTS chat_record (	\
    ActId       INTEGER PRIMARY KEY,		\
    src         VARCHAR(80),				\
    dst         VARCHAR(80),				\
    context     VARCHAR(80),				\
    callid      VARCHAR(80),				\
    channel     VARCHAR(80),				\
    app         VARCHAR(80),				\
    start_time  CHAR(80),					\
    end_time    CHAR(80),					\
    chat_data   TEXT						\
);";

struct char_record_value {
	char *value;
};

struct char_record_head {
	char *name;
	struct char_record_value record_value;
} record_head[] = { {// 0
		.name = "ActID",
		.record_value = {NULL}
	}, { // 1
		.name = "src",
		.record_value = {NULL}
 	}, { // 2
		.name = "dst",
		.record_value = {NULL}
	}, { // 3
		.name = "context",
		.record_value = {NULL}
	}, { // 4
		.name = "callid",
		.record_value = {NULL}
	}, { // 5
		.name = "channel",
		.record_value = {NULL}
	}, { // 6
		.name = "app",
		.record_value = {NULL}
	}, { // 7
		.name = "start_time",
		.record_value = {NULL}
	}, { // 8
		.name = "end_time",
		.record_value = {NULL}
	}, { // 9
		.name = "chat_data",
		.record_value = {NULL}
	}
};

static void *sip_msg_chat_record_pool(int size)
{
	void *pool = calloc(1, sizeof(void) * 3 * size);
	if (!pool) {
		return NULL;
	}

	return pool;
}

static int sip_msg_chat_record_size(void)
{
	return sizeof(record_head) / sizeof(struct char_record_head);
}

static char *sip_msg_chat_record_add_char(const char *str)
{
	if (!str) {
		return NULL;
	}

	int len = strlen(str);
	char *dst_str = sip_msg_chat_record_pool(sizeof(char) * len);
	snprintf(dst_str, len * 2 - 1, "\'%s\'", str);

	return dst_str;
}

static int sip_msg_chat_record_append(const struct ast_msg *msg, const char *to)
{
	if (ast_msg_get_from(msg)) {
		record_head[1].record_value.value = sip_msg_chat_record_add_char(ast_msg_get_from(msg));
	}

	if (to) {
		record_head[2].record_value.value = sip_msg_chat_record_add_char(to);
	}

	if (ast_msg_get_context(msg)) {
		record_head[3].record_value.value = sip_msg_chat_record_add_char(ast_msg_get_context(msg));
	}

	if (ast_msg_get_endpoint(msg)) {
		record_head[4].record_value.value = sip_msg_chat_record_add_char(ast_msg_get_endpoint(msg));
	}

	if (ast_msg_get_channel(msg)) {
		record_head[5].record_value.value = sip_msg_chat_record_add_char(ast_msg_get_channel(msg));
	}

	if (ast_msg_get_app(msg)) {
		record_head[6].record_value.value = sip_msg_chat_record_add_char(ast_msg_get_app(msg));
	}

	if (ast_msg_get_start_time(msg)) {
		record_head[7].record_value.value = sip_msg_chat_record_add_char(ast_msg_get_start_time(msg));

		RAII_VAR(char *, end_time, NULL, ast_free);
		ast_msg_get_system_time(&end_time);
		if (end_time) {
			record_head[8].record_value.value = sip_msg_chat_record_add_char(end_time);
		}
	}

	if (ast_msg_get_body(msg)) {
		record_head[9].record_value.value = sip_msg_chat_record_add_char(ast_msg_get_body(msg));
	}

	return 0;
}

static int sip_msg_chat_record_destructor(void)
{
	int i = 0;
	for (; i < sip_msg_chat_record_size(); i++) {
		if (record_head[i].record_value.value) {
			free(record_head[i].record_value.value);
			record_head[i].record_value.value = NULL;
		}
	}

	return 0;
}

static int sip_msg_chat_record_append_str(char *dst, const char *src)
{
	if (!dst || !src) {
		return -1;
	}

	strcat(dst, src);
	strcat(dst, ", ");

	return 0;
}

static int sip_msg_chat_record_sql_constructor(char **key, char **value)
{
	int i;
	int key_len = 0, value_len = 0;
	for (i = 0; i < sip_msg_chat_record_size(); i++) {
		if (record_head[i].record_value.value) {
			key_len += (strlen(record_head[i].name) + 1);
			value_len += (strlen(record_head[i].record_value.value) + 1);
		}
	}

	*key = sip_msg_chat_record_pool(key_len + sizeof(record_head));
	*value = sip_msg_chat_record_pool(value_len + sizeof(record_head));

	if (*key == NULL || *value == NULL) {
		ast_log(LOG_ERROR, "Calloc memory failed!\n");
		return -1;
	}

	for (i = 0; i < sip_msg_chat_record_size(); i++) {
		if (record_head[i].record_value.value) {
			sip_msg_chat_record_append_str(*key, record_head[i].name);
			sip_msg_chat_record_append_str(*value, record_head[i].record_value.value);
		}
	}
	ast_log(LOG_NOTICE, "*key[%s], *value[%s]\n", *key, *value);

	return 0;
}

static int sip_msg_chat_record_exesql(const char *sql_cmd)
{
	if (!sql_cmd) {
		return -1;
	}

	char *errmsg;
	if (sqlite3_exec(messdb, sql_cmd, NULL, NULL, &errmsg)) {
		ast_log(LOG_WARNING, "Insert database(chat_record) is failed, because of '%s'!\n", errmsg);
		errmsg ? sqlite3_free(errmsg) : NULL;
		return -1;
	}

	return 0;
}

static int sip_msg_chat_record(const struct ast_msg *msg, const char *to, const char *from)
{
	sip_msg_chat_record_append(msg, to);

    RAII_VAR(char *, sql_key, NULL, ast_free);
    RAII_VAR(char *, sql_value, NULL, ast_free);
	sip_msg_chat_record_sql_constructor(&sql_key, &sql_value);

	int sql_cmd_len = strlen(sql_key) + strlen(sql_value);
	RAII_VAR(char *, sql_cmd, sip_msg_chat_record_pool(sql_cmd_len), ast_free);
	snprintf(sql_cmd, sql_cmd_len * 3 - 1, "INSERT INTO chat_record (%s ActID) VALUES (%s NULL)", sql_key, sql_value);
	ast_log(LOG_VERBOSE, "sql_key[%s], sql_value[%s]\n", sql_key, sql_value);
	ast_log(LOG_VERBOSE, "sql_cmd[%s]\n", sql_cmd);

	sip_msg_chat_record_exesql(sql_cmd);
	sip_msg_chat_record_destructor();

	return 0;
}

static const struct ast_msg_tech msg_chat_record_tech = {
	.name = "pjsip message chat record",
	.msg_chat_record = sip_msg_chat_record,
};

static int unload_module(void)
{
	int res = ast_msg_tech_unregister(&msg_chat_record_tech);

	if (messdb) {
		sqlite3_close(messdb);
		messdb = NULL;
	}

	return res;
}

static int load_module(void)
{
	char filename[PATH_MAX] = {0};
	snprintf(filename, sizeof(filename), "%s/chat_record.db", ast_config_AST_LOG_DIR);
	int res = sqlite3_open(filename, &messdb);
	if (res != SQLITE_OK) {
		ast_log(LOG_ERROR, "Could not open database %s.\n", filename);
		return AST_MODULE_LOAD_DECLINE;
	}

	char *errmsg;
	if (sqlite3_exec(messdb, message_chat_record_db, NULL, NULL, &errmsg)) {
		ast_log(LOG_ERROR, "Create database(%s) is failed, because of '%s'\n", filename, errmsg);
		errmsg ? sqlite3_free(errmsg) : NULL;
		sqlite3_close(messdb);
		messdb = NULL;
		return AST_MODULE_LOAD_DECLINE;
	}

	if (ast_msg_tech_register(&msg_chat_record_tech)) {
		sqlite3_close(messdb);
		messdb = NULL;
		return AST_MODULE_LOAD_DECLINE;
	}

	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "SQLite3 Message Chat Record Module",
	.support_level = AST_MODULE_SUPPORT_EXTENDED,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_DEFAULT,
);
