/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Kevin Harwell <kharwell@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_session</depend>
	<support_level>core</support_level>
 ***/

/*** DOCUMENTATION
	<info name="PJSIPMessageFromInfo" language="en_US" tech="PJSIP">
		<para>The <literal>from</literal> parameter can be a configured endpoint
		or in the form of "display-name" &lt;URI&gt;.</para>
	</info>
	<info name="PJSIPMessageToInfo" language="en_US" tech="PJSIP">
		<para>Specifying a prefix of <literal>pjsip:</literal> will send the
		message as a SIP MESSAGE request.</para>
	</info>
 ***/
#ifdef GRANDSTREAM_NETWORKS
#define AST_MODULE_LOG "pjsip"
#endif

#include "asterisk.h"

#ifdef GRANDSTREAM_NETWORKS
ASTERISK_FILE_VERSION(__FILE__, "$Revision: 13.0 $");
#endif

#include "pjsua-lib/pjsua.h"

#include "asterisk/message.h"
#include "asterisk/module.h"
#include "asterisk/pbx.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/taskprocessor.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/app.h"
#include "asterisk/astdb.h"
#include "asterisk/pjsip_cei.h"
#include "asterisk/http.h"
#endif

const pjsip_method pjsip_message_method = {PJSIP_OTHER_METHOD, {"MESSAGE", 7} };

#define MAX_HDR_SIZE 512

#ifdef GRANDSTREAM_NETWORKS
#define MAX_BODY_SIZE 2048
#define CCM_MAX_NUM_VARS	5
#define CCM_VAR_LEN			5
#else
#define MAX_BODY_SIZE 1024
#endif

#define MAX_USER_SIZE 128

static struct ast_taskprocessor *message_serializer;

/*!
 * \internal
 * \brief Checks to make sure the request has the correct content type.
 *
 * \details This module supports the following media types: "text/plain".
 * Return unsupported otherwise.
 *
 * \param rdata The SIP request
 */
static enum pjsip_status_code check_content_type(const pjsip_rx_data *rdata)
{
	int res;
	if (rdata->msg_info.msg->body && rdata->msg_info.msg->body->len) {
		res = ast_sip_is_content_type(
			&rdata->msg_info.msg->body->content_type, "text", "plain");
	} else {
		res = rdata->msg_info.ctype &&
			!pj_strcmp2(&rdata->msg_info.ctype->media.type, "text") &&
			!pj_strcmp2(&rdata->msg_info.ctype->media.subtype, "plain");
	}

	return res ? PJSIP_SC_OK : PJSIP_SC_UNSUPPORTED_MEDIA_TYPE;
}

/*!
 * \internal
 * \brief Puts pointer past 'sip[s]:' string that should be at the
 * front of the given 'fromto' parameter
 *
 * \param fromto 'From' or 'To' field containing 'sip:'
 */
static const char *skip_sip(const char *fromto)
{
	const char *p;

	/* need to be one past 'sip:' or 'sips:' */
	if (!(p = strstr(fromto, "sip"))) {
		return fromto;
	}

	p += 3;
	if (*p == 's') {
		++p;
	}

	return ++p;
}

#ifdef GRANDSTREAM_NETWORKS
static struct ast_sip_endpoint* get_outbound_trunk_endpoint_by_exten(char *exten, char **uri)
{
	RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);
	struct ast_sip_endpoint* endpoint;
	struct ast_exten       *pstExten = NULL;
	struct pbx_find_info    stFindInfo = { .stacklen = 0 }; /* the rest is reset in pbx_find_extension */
	char   family[80]  = {0};
	char   data[256]   = {0};
	char   endpt_id[80]= {0};
	char   cExten[80]  = {0};
	char   cExten0[80] = {0};
	char   remUri[128] = {0};
	char   *buf  = data;
	char   *puri = NULL;
	char   *arr[3];
	unsigned int  i = 0;
	unsigned int  strip = 0;

	pstExten = pbx_find_extension(NULL, NULL, &stFindInfo, "from-internal", exten, 1, NULL, NULL, E_MATCH);
	if (!pstExten) {
		return NULL;
	}

	/* get outbound route data from ucmdb */
	snprintf(family, sizeof(family), "TRUNK/%s", ast_get_extension_name(pstExten));
	ast_ucm_db_get(family, "EVENTLIST", data, sizeof(data));
	
	if (!strlen(data)) {
		return NULL;
	}

	while((i < 3) && ((arr[i]=strtok(buf,"_")) != NULL)) {
		i++;
		buf = NULL; 
	}

	snprintf(endpt_id, sizeof(endpt_id), "trunk_%s", arr[0]);
	if (!(endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", endpt_id))) {
		return NULL;
	}

	strip = atoi(arr[1]);
	if (strip < strlen(exten)) {
		snprintf(cExten, sizeof(cExten), "%s", exten + strip);
	} else {
		snprintf(cExten, sizeof(cExten), "%s", exten);
	}

	if (strcmp(arr[2], "a")) {
		snprintf(cExten0, sizeof(cExten0), "%s%s", arr[2], cExten);
		snprintf(cExten, sizeof(cExten), "%s", cExten0);
	}

	if ((aor = ast_sip_location_retrieve_aor(endpt_id)) &&
		(contact = ast_sip_location_retrieve_first_aor_contact(aor))) {
		puri = (char*)contact->uri;
	}

	if (puri) {
		char *pColon = NULL;
		char *pAt    = NULL;

		pColon = strchr(puri, ':');
		if (!pColon) {
			return endpoint;
		}

		if ((pAt = strchr(puri, '@'))) {
			strncpy(remUri, puri, (int)(pColon-puri+1));
			strncat(remUri, cExten, strlen(cExten));
			strncat(remUri, pAt, strlen(pAt));
		} else {
			strncpy(remUri, puri, (int)(pColon-puri+1));
			strncat(remUri, cExten, strlen(cExten));
			snprintf(remUri+strlen(remUri), sizeof(remUri)-strlen(remUri), "@%s", pColon+1);
		}
	}

	*uri = ast_strdup(remUri);
	return endpoint;
}
#endif

/*!
 * \internal
 * \brief Retrieves an endpoint if specified in the given 'to'
 *
 * Expects the given 'to' to be in one of the following formats:
 *      sip[s]:endpoint[/aor]
 *      sip[s]:endpoint[/uri]
 *      sip[s]:uri <-- will use default outbound endpoint
 *
 * If an optional aor is given it will try to find an associated uri
 * to return.  If an optional uri is given then that will be returned,
 * otherwise uri will be NULL.
 *
 * \param to 'From' or 'To' field with possible endpoint
 * \param uri Optional uri to return
 */
static struct ast_sip_endpoint* get_outbound_endpoint(
	const char *to, char **uri)
{
	char *name, *aor_uri;
	struct ast_sip_endpoint* endpoint;
	RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);

	name = ast_strdupa(skip_sip(to));

	/* attempt to extract the endpoint name */
	if ((aor_uri = strchr(name, '/'))) {
		/* format was 'endpoint/' */
		*aor_uri++ = '\0';
	} else if ((aor_uri = strchr(name, '@'))) {
		/* format was 'endpoint@' - don't use the rest */
		*aor_uri = '\0';
	}

	/* at this point, if name is not empty then it
	   might be an endpoint, so try to retrieve it */
	if (ast_strlen_zero(name) || !(endpoint = ast_sorcery_retrieve_by_id(
		      ast_sip_get_sorcery(), "endpoint", name))) {
#ifdef GRANDSTREAM_NETWORKS
		if ((endpoint = get_outbound_trunk_endpoint_by_exten(name, uri))) {
			return endpoint;
		}
#endif
		/* an endpoint was not found, so assume sending directly
		   to a uri and use the default outbound endpoint */
		*uri = ast_strdup(to);
		return ast_sip_default_outbound_endpoint();
	}

	*uri = aor_uri;
	if (*uri) {
		char *end = strchr(*uri, '>');
		if (end) {
			*end++ = '\0';
		}

		/* if what's in 'uri' is a retrievable aor use the uri on it
		   instead, otherwise assume what's there is already a uri*/
		if ((aor = ast_sip_location_retrieve_aor(*uri)) &&
			(contact = ast_sip_location_retrieve_first_aor_contact(aor))) {
			*uri = (char*)contact->uri;
		}
		/* need to copy because underlying uri goes away */
		*uri = ast_strdup(*uri);
	}

	return endpoint;
}

/*!
 * \internal
 * \brief Overwrite fields in the outbound 'To' header
 *
 * Updates display name in an outgoing To header.
 *
 * \param tdata the outbound message data structure
 * \param to info to copy into the header
 */
static void update_to(pjsip_tx_data *tdata, char *to)
{
	pjsip_name_addr *name_addr = (pjsip_name_addr *)
		PJSIP_MSG_TO_HDR(tdata->msg)->uri;
	pjsip_uri *parsed;

	if ((parsed = pjsip_parse_uri(tdata->pool, to, strlen(to),
				      PJSIP_PARSE_URI_AS_NAMEADDR))) {
		pjsip_name_addr *parsed_name_addr = (pjsip_name_addr *)parsed;
		if (pj_strlen(&parsed_name_addr->display)) {
			pj_strdup(tdata->pool, &name_addr->display,
				  &parsed_name_addr->display);
		}
	}
}

/*!
 * \internal
 * \brief Overwrite fields in the outbound 'From' header
 *
 * The outbound 'From' header is created/added in ast_sip_create_request with
 * default data.  If available that data may be info specified in the 'from_user'
 * and 'from_domain' options found on the endpoint.  That information will be
 * overwritten with data in the given 'from' parameter.
 *
 * \param tdata the outbound message data structure
 * \param from info to copy into the header
 */
#ifdef GRANDSTREAM_NETWORKS
static void update_from(pjsip_tx_data *tdata, char *from, struct ast_sip_endpoint *endpoint)
#else
static void update_from(pjsip_tx_data *tdata, char *from)
#endif
{
	pjsip_name_addr *name_addr = (pjsip_name_addr *)
		PJSIP_MSG_FROM_HDR(tdata->msg)->uri;
	pjsip_sip_uri *uri = pjsip_uri_get_uri(name_addr);
	pjsip_uri *parsed;

#ifdef GRANDSTREAM_NETWORKS
	const char *from_user = NULL;
	const char *endpoint_id = ast_sorcery_object_get_id(endpoint);
	RAII_VAR(struct ast_sip_auth *, auth, ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "auth", endpoint_id), ao2_cleanup);

	if (!strncmp(endpoint_id, "trunk_", 6) && auth) {
		from_user = auth->auth_user;
	}
#endif

	if (ast_strlen_zero(from)) {
		return;
	}

	if ((parsed = pjsip_parse_uri(tdata->pool, from, strlen(from),
				      PJSIP_PARSE_URI_AS_NAMEADDR))) {
		pjsip_name_addr *parsed_name_addr = (pjsip_name_addr *)parsed;
		pjsip_sip_uri *parsed_uri = pjsip_uri_get_uri(parsed_name_addr->uri);
		if (pj_strlen(&parsed_name_addr->display)) {
			pj_strdup(tdata->pool, &name_addr->display,
				  &parsed_name_addr->display);
		}

		pj_strdup(tdata->pool, &uri->user, &parsed_uri->user);
#ifdef GRANDSTREAM_NETWORKS
		if (from_user) {
			pj_strdup2(tdata->pool, &uri->user, from_user);
		}
#endif
		pj_strdup(tdata->pool, &uri->host, &parsed_uri->host);
		uri->port = parsed_uri->port;
	} else {
		/* assume it is 'user[@domain]' format */
		char *domain = strchr(from, '@');
		if (domain) {
			*domain++ = '\0';
			pj_strdup2(tdata->pool, &uri->host, domain);
		}
		pj_strdup2(tdata->pool, &uri->user, from);
#ifdef GRANDSTREAM_NETWORKS
		if (from_user) {
			pj_strdup2(tdata->pool, &uri->user, from_user);
		}
#endif
	}
}

/*!
 * \internal
 * \brief Checks if the given msg var name should be blocked.
 *
 * \details Some headers are not allowed to be overriden by the user.
 *  Determine if the given var header name from the user is blocked for
 *  an outgoing MESSAGE.
 *
 * \param name name of header to see if it is blocked.
 *
 * \retval TRUE if the given header is blocked.
 */
static int is_msg_var_blocked(const char *name)
{
	int i;

	/*
	 * Don't block Content-Type or Max-Forwards headers because the
	 * user can override them.
	 */
	static const char *hdr[] = {
		"To",
		"From",
		"Via",
		"Route",
		"Contact",
		"Call-ID",
		"CSeq",
		"Allow",
		"Content-Length",
		"Request-URI",
	};

	for (i = 0; i < ARRAY_LEN(hdr); ++i) {
		if (!strcasecmp(name, hdr[i])) {
			/* Block addition of this header. */
			return 1;
		}
	}
	return 0;
}

/*!
 * \internal
 * \brief Copies any other msg vars over to the request headers.
 *
 * \param msg The msg structure to copy headers from
 * \param tdata The SIP transmission data
 */
static enum pjsip_status_code vars_to_headers(const struct ast_msg *msg, pjsip_tx_data *tdata)
{
	const char *name;
	const char *value;
	int max_forwards;
	struct ast_msg_var_iterator *iter;

	for (iter = ast_msg_var_iterator_init(msg);
		ast_msg_var_iterator_next(msg, iter, &name, &value);
		ast_msg_var_unref_current(iter)) {
		if (!strcasecmp(name, "Max-Forwards")) {
			/* Decrement Max-Forwards for SIP loop prevention. */
			if (sscanf(value, "%30d", &max_forwards) != 1 || --max_forwards == 0) {
				ast_msg_var_iterator_destroy(iter);
				ast_log(LOG_NOTICE, "MESSAGE(Max-Forwards) reached zero.  MESSAGE not sent.\n");
				return -1;
			}
			sprintf((char *) value, "%d", max_forwards);
			ast_sip_add_header(tdata, name, value);
		} else if (!is_msg_var_blocked(name)) {
			ast_sip_add_header(tdata, name, value);
		}
	}
	ast_msg_var_iterator_destroy(iter);

	return PJSIP_SC_OK;
}

/*!
 * \internal
 * \brief Copies any other request header data over to ast_msg structure.
 *
 * \param rdata The SIP request
 * \param msg The msg structure to copy headers into
 */
static int headers_to_vars(const pjsip_rx_data *rdata, struct ast_msg *msg)
{
	char *c;
	char name[MAX_HDR_SIZE];
	char buf[MAX_HDR_SIZE];
	int res = 0;
	pjsip_hdr *h = rdata->msg_info.msg->hdr.next;
	pjsip_hdr *end= &rdata->msg_info.msg->hdr;

	while (h != end) {
		if ((res = pjsip_hdr_print_on(h, buf, sizeof(buf)-1)) > 0) {
			buf[res] = '\0';
			if ((c = strchr(buf, ':'))) {
				ast_copy_string(buf, ast_skip_blanks(c + 1), sizeof(buf));
			}

			ast_copy_pj_str(name, &h->name, sizeof(name));
			if ((res = ast_msg_set_var(msg, name, buf)) != 0) {
				break;
			}
		}
		h = h->next;
	}
	return 0;
}

/*!
 * \internal
 * \brief Prints the message body into the given char buffer.
 *
 * \details Copies body content from the received data into the given
 * character buffer removing any extra carriage return/line feeds.
 *
 * \param rdata The SIP request
 * \param buf Buffer to fill
 * \param len The length of the buffer
 */
static int print_body(pjsip_rx_data *rdata, char *buf, int len)
{
	int res;

	if (!rdata->msg_info.msg->body || !rdata->msg_info.msg->body->len) {
		return 0;
	}

#ifdef GRANDSTREAM_NETWORKS
	pjsip_msg_body *body_tmp; 
	res = 0;
	body_tmp = pjsip_msg_body_clone(rdata->tp_info.pool, rdata->msg_info.msg->body);
	if (NULL != body_tmp) {
		if (body_tmp->len > len) {
			body_tmp->len = len; 
		}

		if ((res = rdata->msg_info.msg->body->print_body(body_tmp, buf, len)) < 0) {
			return res;
		}
	}
#else
	if ((res = rdata->msg_info.msg->body->print_body(
		     rdata->msg_info.msg->body, buf, len)) < 0) {
		return res;
	}
#endif

	/* remove any trailing carriage return/line feeds */
	while (res > 0 && ((buf[--res] == '\r') || (buf[res] == '\n')));

	buf[++res] = '\0';

	return res;
}

/*!
 * \internal
 * \brief Converts a 'sip:' uri to a 'pjsip:' so it can be found by
 * the message tech.
 *
 * \param buf uri to insert 'pjsip' into
 * \param size length of the uri in buf
 * \param capacity total size of buf
 */
static char *sip_to_pjsip(char *buf, int size, int capacity)
{
	int count;
	const char *scheme;
	char *res = buf;

	/* remove any wrapping brackets */
	if (*buf == '<') {
		++buf;
		--size;
	}

	scheme = strncmp(buf, "sip", 3) ? "pjsip:" : "pj";
	count = strlen(scheme);
	if (count + size >= capacity) {
		ast_log(LOG_WARNING, "Unable to handle MESSAGE- incoming uri "
			"too large for given buffer\n");
		return NULL;
	}

	memmove(res + count, buf, size);
	memcpy(res, scheme, count);

	buf += size - 1;
	if (*buf == '>') {
		*buf = '\0';
	}

	return res;
}

/*!
 * \internal
 * \brief Converts a pjsip_rx_data structure to an ast_msg structure.
 *
 * \details Attempts to fill in as much information as possible into the given
 * msg structure copied from the given request data.
 *
 * \param rdata The SIP request
 * \param msg The asterisk message structure to fill in.
 */
static enum pjsip_status_code rx_data_to_ast_msg(pjsip_rx_data *rdata, struct ast_msg *msg)
{
	RAII_VAR(struct ast_sip_endpoint *, endpt, NULL, ao2_cleanup);
	pjsip_uri *ruri = rdata->msg_info.msg->line.req.uri;
	pjsip_sip_uri *sip_ruri;
	pjsip_name_addr *name_addr;
	char buf[MAX_BODY_SIZE];
	const char *field;
	const char *context;
	char exten[AST_MAX_EXTENSION];
	int res = 0;
	int size;

	if (!PJSIP_URI_SCHEME_IS_SIP(ruri) && !PJSIP_URI_SCHEME_IS_SIPS(ruri)) {
		return PJSIP_SC_UNSUPPORTED_URI_SCHEME;
	}

	sip_ruri = pjsip_uri_get_uri(ruri);
	ast_copy_pj_str(exten, &sip_ruri->user, AST_MAX_EXTENSION);

	endpt = ast_pjsip_rdata_get_endpoint(rdata);
	ast_assert(endpt != NULL);

	context = S_OR(endpt->message_context, endpt->context);
	res |= ast_msg_set_context(msg, "%s", context);
	res |= ast_msg_set_exten(msg, "%s", exten);

	/* to header */
	name_addr = (pjsip_name_addr *)rdata->msg_info.to->uri;
	size = pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, name_addr, buf, sizeof(buf) - 1);
	if (size <= 0) {
		return PJSIP_SC_INTERNAL_SERVER_ERROR;
	}
	buf[size] = '\0';
	res |= ast_msg_set_to(msg, "%s", sip_to_pjsip(buf, ++size, sizeof(buf) - 1));

	/* from header */
	name_addr = (pjsip_name_addr *)rdata->msg_info.from->uri;
	size = pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, name_addr, buf, sizeof(buf) - 1);
	if (size <= 0) {
		return PJSIP_SC_INTERNAL_SERVER_ERROR;
	}
	buf[size] = '\0';
	res |= ast_msg_set_from(msg, "%s", buf);

	field = pj_sockaddr_print(&rdata->pkt_info.src_addr, buf, sizeof(buf) - 1, 1);
	res |= ast_msg_set_var(msg, "PJSIP_RECVADDR", field);

	if (print_body(rdata, buf, sizeof(buf) - 1) > 0) {
		res |= ast_msg_set_body(msg, "%s", buf);
	}

	/* endpoint name */
	res |= ast_msg_set_tech(msg, "%s", "PJSIP");
	res |= ast_msg_set_endpoint(msg, "%s", ast_sorcery_object_get_id(endpt));
	if (endpt->id.self.name.valid) {
		res |= ast_msg_set_var(msg, "PJSIP_ENDPOINT", endpt->id.self.name.str);
	}

	res |= headers_to_vars(rdata, msg);

	return !res ? PJSIP_SC_OK : PJSIP_SC_INTERNAL_SERVER_ERROR;
}

struct msg_data {
	struct ast_msg *msg;
	char *to;
	char *from;
};

static void msg_data_destroy(void *obj)
{
	struct msg_data *mdata = obj;

	ast_free(mdata->from);
	ast_free(mdata->to);

	ast_msg_destroy(mdata->msg);
}

static struct msg_data* msg_data_create(const struct ast_msg *msg, const char *to, const char *from)
{
	char *tag;
	struct msg_data *mdata = ao2_alloc(sizeof(*mdata), msg_data_destroy);

	if (!mdata) {
		return NULL;
	}

	/* typecast to suppress const warning */
	mdata->msg = ast_msg_ref((struct ast_msg*)msg);

	/* starts with 'pjsip:' the 'pj' needs to be removed and maybe even
	   the entire string. */
	if (!(to = strchr(to, ':'))) {
		ao2_ref(mdata, -1);
		return NULL;
	}

	/* Make sure we start with sip: */
	mdata->to = ast_begins_with(to, "sip:") ? ast_strdup(++to) : ast_strdup(to - 3);

#ifdef GRANDSTREAM_NETWORKS
	if (strstr(from, "webrtc_anonymous")) {
		char tmp_buf[128] = {0};
		snprintf(tmp_buf, (strstr(from, "webrtc_anonymous") - from) + 1, "%s", from);
		char tmp_from[256] = {0};
		snprintf(tmp_from, sizeof(tmp_from) - 1, "%s%s%s", tmp_buf,
			ast_msg_get_webrtc_message_callee(msg), strrchr(from, '@'));
		mdata->from = ast_strdup(tmp_from);
	} else {
		mdata->from = ast_strdup(from);
	}
#else
	mdata->from = ast_strdup(from);
#endif

	/* sometimes from can still contain the tag at this point, so remove it */
	if ((tag = strchr(mdata->from, ';'))) {
		*tag = '\0';
	}
	return mdata;
}

#ifdef GRANDSTREAM_NETWORKS
static int ast_sip_add_extend_contact_header(pjsip_tx_data *tdata, const char *header_value)
{
	if (header_value == NULL) {
		return -1;
	}

	char *tmp_str = strrchr(header_value, '@');
	char tmp_buf[40] = {0};
	snprintf(tmp_buf, tmp_str - header_value - strlen("<sip:") + 1, "%s", header_value + strlen("<sip:"));

	RAII_VAR(struct ast_sip_endpoint *, endpoint, ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", tmp_buf), ao2_cleanup);
	RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);
	if (endpoint == NULL) {
		return -1;
	}

	aor = ast_sip_location_retrieve_aor(ast_sorcery_object_get_id(endpoint));
	if (!aor) {
		ast_log(LOG_ERROR, "no AOR to the endpoint[%s]\n", ast_sorcery_object_get_id(endpoint));
		return -1;
	}

	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
	contacts = ast_sip_location_retrieve_aor_contacts(aor);
	struct ao2_iterator     stIterator;
	struct ast_sip_contact *contact;
	RAII_VAR(char *, edp_contacts, NULL, ast_free);

	if (!contacts || (ao2_container_count(contacts) == 0)) {
		ast_log(LOG_ERROR, "no contact to the endpoint[%s]\n", ast_sorcery_object_get_id(endpoint));
		return -1;
	}

	stIterator = ao2_iterator_init(contacts, 0);
	if (!contacts || (ao2_container_count(contacts) == 0)) {
		ast_log(LOG_ERROR, "no contact to the endpoint[%s]\n", ast_sorcery_object_get_id(endpoint));
		return -1;
	}

	while ((contact = ao2_t_iterator_next(&stIterator, "iterate thru contacts table"))) {
		edp_contacts = strdup(contact->uri);
		ao2_ref(contact, -1);
	}
	ao2_iterator_destroy(&stIterator);
	if (edp_contacts) {
		ast_sip_add_header(tdata, "X-GS-Contact", edp_contacts);
	}

	return 0;
}
#endif

static int msg_send(void *data)
{
	RAII_VAR(struct msg_data *, mdata, data, ao2_cleanup);

	const struct ast_sip_body body = {
		.type = "text",
		.subtype = "plain",
		.body_text = ast_msg_get_body(mdata->msg)
	};

	pjsip_tx_data *tdata;
	RAII_VAR(char *, uri, NULL, ast_free);

#ifdef GRANDSTREAM_NETWORKS
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
	if (ast_http_rwlock()) {
		ast_rwlock_rdlock(ast_http_rwlock());
		RAII_VAR(char *, request_uri, NULL, ast_free);
		RAII_VAR(char *, transport_name, NULL, ast_free);
		if (!ast_http_webrtc_uri_find(strstr(mdata->to, "sip:"), &request_uri, &transport_name)) {
			endpoint = get_outbound_endpoint("sip:webrtc_anonymous", &uri);
			if (!uri) {
				char tmp_request_uri[256] = {0};
				snprintf(tmp_request_uri, sizeof(tmp_request_uri) - 1, "%s@%s;transport=%s", "sip:webrtc_anonymous", request_uri, transport_name);
				uri = ast_strdup(tmp_request_uri);
				ast_log(LOG_NOTICE, "uri[%s]\n", uri);
			}
		} else {
			endpoint = get_outbound_endpoint(mdata->to, &uri);
		}
		ast_rwlock_unlock(ast_http_rwlock());
	}
#else
	RAII_VAR(struct ast_sip_endpoint *, endpoint, get_outbound_endpoint(
			 mdata->to, &uri), ao2_cleanup);
#endif

	if (!endpoint) {
		ast_log(LOG_ERROR, "PJSIP MESSAGE - Could not find endpoint '%s' and "
			"no default outbound endpoint configured\n", mdata->to);
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (ast_strlen_zero(uri)) {
		RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
		RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);
		struct ao2_iterator     stIterator;
		struct ast_sip_contact *contact;

		aor = ast_sip_location_retrieve_aor(ast_sorcery_object_get_id(endpoint));
		if (!aor) {
			ast_log(LOG_ERROR, "no AOR to the endpoint[%s]\n", ast_sorcery_object_get_id(endpoint));
			return -1;
		}

		contacts = ast_sip_location_retrieve_aor_contacts(aor); 
		if (!contacts || (ao2_container_count(contacts) == 0)) {
			ast_log(LOG_ERROR, "no contact to the endpoint[%s]\n", ast_sorcery_object_get_id(endpoint));
			return -1;
		}

		stIterator = ao2_iterator_init(contacts, 0);
		while ((contact = ao2_t_iterator_next(&stIterator, "iterate thru contacts table"))) {
			if (ast_sip_create_request("MESSAGE", NULL, endpoint, contact->uri, NULL, &tdata)) {
				ast_log(LOG_ERROR, "PJSIP MESSAGE - Could not create request\n");
				ao2_ref(contact, -1);
				break;
			}

			update_to(tdata, mdata->to);
			update_from(tdata, mdata->from, endpoint);

			if (ast_sip_add_body(tdata, &body)) {
				pjsip_tx_data_dec_ref(tdata);
				ast_log(LOG_ERROR, "PJSIP MESSAGE - Could not add body to request\n");
				ao2_ref(contact, -1);
				break;
			}

			vars_to_headers(mdata->msg, tdata);

			ast_debug(1, "Sending message to '%s' (via endpoint %s) from '%s'\n",
				contact->uri, ast_sorcery_object_get_id(endpoint), mdata->from);

			if (ast_sip_send_request(tdata, NULL, endpoint, NULL, NULL)) {
				ast_log(LOG_ERROR, "PJSIP MESSAGE - Could not send request\n");
				ao2_ref(contact, -1);
				break;
			}

			ao2_ref(contact, -1);
		}

		ao2_iterator_destroy(&stIterator);

		return PJ_SUCCESS;
	}
#endif

	if (ast_sip_create_request("MESSAGE", NULL, endpoint, uri, NULL, &tdata)) {
		ast_log(LOG_ERROR, "PJSIP MESSAGE - Could not create request\n");
		return -1;
	}

	update_to(tdata, mdata->to);

#ifdef GRANDSTREAM_NETWORKS
	update_from(tdata, mdata->from, endpoint);
#else
	update_from(tdata, mdata->from);
#endif

	if (ast_sip_add_body(tdata, &body)) {
		pjsip_tx_data_dec_ref(tdata);
		ast_log(LOG_ERROR, "PJSIP MESSAGE - Could not add body to request\n");
		return -1;
	}

	vars_to_headers(mdata->msg, tdata);

#ifdef GRANDSTREAM_NETWORKS
	ast_sip_add_extend_contact_header(tdata, mdata->from);
#endif

	ast_debug(1, "Sending message to '%s' (via endpoint %s) from '%s'\n",
		mdata->to, ast_sorcery_object_get_id(endpoint), mdata->from);

	if (ast_sip_send_request(tdata, NULL, endpoint, NULL, NULL)) {
		ast_log(LOG_ERROR, "PJSIP MESSAGE - Could not send request\n");
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "Sending Message to '%s' (via endpoint %s) from '%s', transport is '%s'\n",
		mdata->to, ast_sorcery_object_get_id(endpoint), mdata->from,
		tdata->tp_info.transport ? tdata->tp_info.transport->type_name : "NULL");
#endif

	return PJ_SUCCESS;
}

static int sip_msg_send(const struct ast_msg *msg, const char *to, const char *from)
{
	struct msg_data *mdata;

	if (ast_strlen_zero(to)) {
		ast_log(LOG_ERROR, "SIP MESSAGE - a 'To' URI  must be specified\n");
		return -1;
	}

	if (!(mdata = msg_data_create(msg, to, from)) ||
	    ast_sip_push_task(message_serializer, msg_send, mdata)) {
		ao2_ref(mdata, -1);
		return -1;
	}
	return 0;
}

static const struct ast_msg_tech msg_tech = {
	.name = "pjsip",
	.msg_send = sip_msg_send,
};

static pj_status_t send_response(pjsip_rx_data *rdata, enum pjsip_status_code code,
				 pjsip_dialog *dlg, pjsip_transaction *tsx)
{
	pjsip_tx_data *tdata;
	pj_status_t status;

	status = ast_sip_create_response(rdata, code, NULL, &tdata);
	if (status != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Unable to create response (%d)\n", status);
		return status;
	}

	if (dlg && tsx) {
		status = pjsip_dlg_send_response(dlg, tsx, tdata);
	} else {
		struct ast_sip_endpoint *endpoint;

		endpoint = ast_pjsip_rdata_get_endpoint(rdata);
		status = ast_sip_send_stateful_response(rdata, tdata, endpoint);
		ao2_cleanup(endpoint);
	}

	if (status != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Unable to send response (%d)\n", status);
	}

	return status;
}

static pj_bool_t module_on_rx_request(pjsip_rx_data *rdata)
{
	enum pjsip_status_code code;
	struct ast_msg *msg;

	/* if not a MESSAGE, don't handle */
	if (pjsip_method_cmp(&rdata->msg_info.msg->line.req.method, &pjsip_message_method)) {
		return PJ_FALSE;
	}

	code = check_content_type(rdata);
	if (code != PJSIP_SC_OK) {
		send_response(rdata, code, NULL, NULL);
		return PJ_TRUE;
	}

	msg = ast_msg_alloc();
	if (!msg) {
		send_response(rdata, PJSIP_SC_INTERNAL_SERVER_ERROR, NULL, NULL);
		return PJ_TRUE;
	}

#ifdef GRANDSTREAM_NETWORKS
	char request_uri[256] = {0};
	char request_callee_id[20] = {0};
	char request_ip_port [128] = {0};

	ast_msg_set_webrtc_request_method(msg, rdata->msg_info.msg->line.req.method.name.ptr);
	snprintf(request_uri, 255 * sizeof(char), "sip:webrtc_anonymous@%s:%d", rdata->pkt_info.src_name, rdata->pkt_info.src_port);
	ast_msg_set_webrtc_message_uri(msg, request_uri);

	if (ast_http_rwlock()) {
		ast_rwlock_wrlock(ast_http_rwlock());
		snprintf(request_ip_port, sizeof(request_ip_port) - 1, "%s:%d", rdata->pkt_info.src_name, rdata->pkt_info.src_port);
		int request_number[5] = {0};
		sscanf(request_ip_port, "%03d.%03d.%03d.%03d:%05d", &request_number[0], &request_number[1], &request_number[2], &request_number[3], &request_number[4]);
		snprintf(request_callee_id, sizeof(request_callee_id) - 1, "%d%d%d%d%d", request_number[0], request_number[1], request_number[2], request_number[3], request_number[4]);
		ast_msg_set_webrtc_message_callee(msg, request_callee_id);
		ast_http_webrtc_alter(request_ip_port, request_callee_id, rdata->tp_info.transport->type_name, -1);
		ast_rwlock_unlock(ast_http_rwlock());
	}
#endif

	code = rx_data_to_ast_msg(rdata, msg);
	if (code != PJSIP_SC_OK) {
		send_response(rdata, code, NULL, NULL);
		ast_msg_destroy(msg);
		return PJ_TRUE;
	}

	if (!ast_msg_has_destination(msg)) {
		ast_debug(1, "MESSAGE request received, but no handler wanted it\n");
		send_response(rdata, PJSIP_SC_NOT_FOUND, NULL, NULL);
		ast_msg_destroy(msg);
		return PJ_TRUE;
	}

	/* Send it to the messaging core.
	 *
	 * If we are unable to send a response, the most likely reason is that we
	 * are handling a retransmission of an incoming MESSAGE and were unable to
	 * create a transaction due to a duplicate key. If we are unable to send
	 * a response, we should not queue the message to the dialplan
	 */
	if (!send_response(rdata, PJSIP_SC_ACCEPTED, NULL, NULL)) {
		ast_msg_queue(msg);
	}

	return PJ_TRUE;
}

#ifdef GRANDSTREAM_NETWORKS
/*!
 * \brief Prepare and publish an \ref ast_ccm_command_kick_type
 * message to the \ref ast_ccm_command_topic_all.
 */
static enum pjsip_status_code ccm_publish_command_kick(char *conference,
													   char *user_channel,
													   char *entity)
{
    struct stasis_topic *ccm_command_topic = NULL;
    RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
    RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

    // Sanity check
    if (conference == NULL || user_channel == NULL || entity == NULL) {
        return PJSIP_SC_FORBIDDEN;
    }

    ccm_command_topic = ast_ccm_command_topic_all();
	if (ccm_command_topic == NULL) {
        ast_log(LOG_WARNING, "Failed to get ast_ccm_command_topic_all");
        return PJSIP_SC_INTERNAL_SERVER_ERROR;
	}

    json_object = ast_json_pack("{s: s, s: s, s: s}",
                                "conference", conference,
								"user_channel", user_channel,
								"entity", entity);

    if (json_object == NULL) {
        ast_log(LOG_WARNING, "Failed to create json_object");
        return PJSIP_SC_INTERNAL_SERVER_ERROR;
    }

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
        return PJSIP_SC_INTERNAL_SERVER_ERROR;
	}

	msg = stasis_message_create(ast_ccm_command_kick_type(), payload);

    if (msg == NULL) {
        ast_log(LOG_WARNING, "Failed to create stasis message, "
							 "conference: %s, user_channel: %s, entity: %s",
							 conference, user_channel, entity);
        return PJSIP_SC_INTERNAL_SERVER_ERROR;
    }

    stasis_publish(ccm_command_topic, msg);

    return PJSIP_SC_OK;
}

/*!
 * \brief Prepare and publish an \ref ast_ccm_command_lock_type or \ref ast_ccm_command_unlock_type
 * message to the \ref ast_ccm_command_topic_all.
 */
static enum pjsip_status_code ccm_publish_command_lock_unlock(char *conference,
															  int lock)
{
    struct stasis_topic *ccm_command_topic = NULL;
    RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
    RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

    // Sanity check
    if (conference == NULL) {
        return PJSIP_SC_FORBIDDEN;
    }

    ccm_command_topic = ast_ccm_command_topic_all();
	if (ccm_command_topic == NULL) {
        ast_log(LOG_WARNING, "Failed to get ast_ccm_command_topic_all");
        return PJSIP_SC_INTERNAL_SERVER_ERROR;
	}

    json_object = ast_json_pack("{s: s}",
                                "conference", conference);

    if (json_object == NULL) {
        ast_log(LOG_WARNING, "Failed to create json_object");
        return PJSIP_SC_INTERNAL_SERVER_ERROR;
    }

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
        return PJSIP_SC_INTERNAL_SERVER_ERROR;
	}

	if (lock) {
		msg = stasis_message_create(ast_ccm_command_lock_type(), payload);
	}
	else {
		msg = stasis_message_create(ast_ccm_command_unlock_type(), payload);
	}

    if (msg == NULL) {
        ast_log(LOG_WARNING, "Failed to create stasis message, "
							 "conference: %s, lock: %d", conference, lock);
        return PJSIP_SC_INTERNAL_SERVER_ERROR;
    }

    stasis_publish(ccm_command_topic, msg);

    return PJSIP_SC_OK;
}

/*!
 * \brief Prepare and publish an \ref ast_ccm_command_mute_unmute_type
 * message to the \ref ast_ccm_command_topic_all.
 */
static enum pjsip_status_code ccm_publish_command_mute_unmute(char *conference,
															  char *entity,
															  int muted)
{
    struct stasis_topic *ccm_command_topic = NULL;
    RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
    RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

    // Sanity check
    if (conference == NULL || entity == NULL) {
        return PJSIP_SC_FORBIDDEN;
    }

    ccm_command_topic = ast_ccm_command_topic_all();
	if (ccm_command_topic == NULL) {
        ast_log(LOG_WARNING, "Failed to get ast_ccm_command_topic_all");
        return PJSIP_SC_INTERNAL_SERVER_ERROR;
	}

    json_object = ast_json_pack("{s: s, s: s, s: i}",
                                "conference", conference,
								"entity", entity,
								"muted", muted);

    if (json_object == NULL) {
        ast_log(LOG_WARNING, "Failed to create json_object");
        return PJSIP_SC_INTERNAL_SERVER_ERROR;
    }

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
        return PJSIP_SC_INTERNAL_SERVER_ERROR;
	}

	msg = stasis_message_create(ast_ccm_command_mute_unmute_type(), payload);

    if (msg == NULL) {
        ast_log(LOG_WARNING, "Failed to create stasis message, "
							 "conference: %s, entity: %s, muted: %d",
							 conference, entity, muted);
        return PJSIP_SC_INTERNAL_SERVER_ERROR;
    }

    stasis_publish(ccm_command_topic, msg);

    return PJSIP_SC_OK;
}

/*! \brief Retrive user from TO Header.  */
/* \note Returned string MUST be freed. */
static char* get_to_user(pjsip_rx_data *rdata)
{
	pjsip_to_hdr *to_hdr = NULL;
	pjsip_sip_uri *to_uri_sip = NULL;
	size_t size;
	char *touser = NULL;

	// Sanity check
	if (rdata == NULL) {
		return NULL;
	}

	// Get TO header
	to_hdr = PJSIP_MSG_TO_HDR(rdata->msg_info.msg);
	if (to_hdr == NULL || to_hdr->uri == NULL) {
		return NULL;
	}

	// Get TO information
	to_uri_sip = pjsip_uri_get_uri(to_hdr->uri);
	if (to_uri_sip == NULL) {
		return NULL;
	}

	size = pj_strlen(&to_uri_sip->user) + 1;
	touser = ast_malloc(size);
	if (touser == NULL) {
		return NULL;
	}

	ast_copy_pj_str(touser, &to_uri_sip->user, size);

	return touser;
}

/*! \brief Retrive user from Request-URI Header.  */
/* \note Returned string MUST be freed. */
static char* get_request_uri_user(struct pjsip_rx_data *rdata)
{
	const pjsip_msg *rx_msg = NULL;
	pjsip_sip_uri *uri_sip = NULL;
	size_t size;
	char *user = NULL;

	// Sanity check
	if (rdata == NULL) {
		return NULL;
	}

	rx_msg = rdata->msg_info.msg;
	if (rx_msg == NULL) {
		return NULL;
	}

	uri_sip = pjsip_uri_get_uri(rx_msg->line.req.uri);
	if (uri_sip == NULL) {
		return NULL;
	}

	size = pj_strlen(&uri_sip->user) + 1;
	user = ast_malloc(size);
	if (user == NULL) {
		return NULL;
	}

	ast_copy_pj_str(user, &uri_sip->user, size);

	return user;
}

/*! \brief Retrive FROM URI.  */
/* \note Returned string MUST be freed. */
static char* get_from_uri(struct pjsip_rx_data *rdata)
{
	pjsip_from_hdr *from_hdr;
	char uri[512] = {0};
	int size;

	// Sanity check
	if (rdata == NULL) {
		return NULL;
	}

	from_hdr = PJSIP_MSG_FROM_HDR(rdata->msg_info.msg);
	if (from_hdr == NULL || from_hdr->uri == NULL) {
		return NULL;
	}

	size = pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, from_hdr->uri, uri, sizeof(uri) -1);
	if (size <= 0) {
		return NULL;
	}
	uri[size] = '\0';

	return ast_strdup(uri);
}

/*! \brief Forward CCM message and informatin to Dialplan.  */
static enum pjsip_status_code ccm_msg_queue(struct ast_sip_endpoint *endpoint,
										   	struct sip_conference* conf,
											struct pjsip_rx_data *rdata,
											char *ccm_msg,
											char *role, char *command,
											char **vars, int num_vars)
{
	int res;
	int var_index = 0;
	char *context = NULL;
	char *dest_to = NULL;
	char *from_uri = NULL;
	char varname[CCM_VAR_LEN];
	struct ast_msg *msg = NULL;
	enum pjsip_status_code status_code = PJSIP_SC_OK;

	// Sanity check
	if (endpoint == NULL || conf == NULL || rdata == NULL ||
		ccm_msg == NULL || role == NULL || command == NULL ||
		vars == NULL) {
		return PJSIP_SC_FORBIDDEN;
	}

	context = ast_strdup(endpoint->context);
	ast_string_field_set(endpoint, context, "ccm-processing");

	// switch (get_destination(p, NULL, NULL)) {
	// 	case SIP_GET_DEST_EXTEN_ANONYMOUS:
	// 	case SIP_GET_DEST_REFUSED:
	// 		transmit_response(p, "403 Forbidden", req);
	// 		return;
	// 	case SIP_GET_DEST_INVALID_URI:
	// 		transmit_response(p, "416 Unsupported URI Scheme", req);
	// 		return;
	// 	case SIP_GET_DEST_EXTEN_NOT_FOUND:
	// 	case SIP_GET_DEST_EXTEN_MATCHMORE:
	// 		transmit_response(p, "404 Not Found", req);
	// 		return;
	// 	case SIP_GET_DEST_EXTEN_FOUND:
	// 		break;
	// }

	msg = ast_msg_alloc();
	if (msg == NULL) {
		ast_log(LOG_DEBUG, "Failed to allocate ast_msg");
		ast_string_field_set(endpoint, context, context);
		ast_free(context);
		return PJSIP_SC_INTERNAL_SERVER_ERROR;
	}

	dest_to = get_request_uri_user(rdata);
	from_uri = get_from_uri(rdata);
	if (dest_to == NULL || from_uri == NULL) {
		ast_log(LOG_DEBUG, "Failed to get Request-URI user (%s) or FROM URI (%s)",
						   S_OR(dest_to, "NULL"), S_OR(from_uri, "NULL"));
		ast_free(dest_to);
		ast_free(from_uri);
		ast_string_field_set(endpoint, context, context);
		ast_free(context);
		return PJSIP_SC_INTERNAL_SERVER_ERROR;
	}

	res = ast_msg_set_to(msg, "%s", dest_to);
	res = ast_msg_set_from(msg, "%s", from_uri);

	res |= ast_msg_set_body(msg, "%s", ccm_msg);
	res |= ast_msg_set_context(msg, "%s", endpoint->context);
	res |= ast_msg_set_exten(msg, "%s", dest_to);

	res |= ast_msg_set_var(msg, "CCM_ACTION", command);
	for (var_index = 0; var_index < num_vars; var_index++) {
		snprintf(varname, CCM_VAR_LEN, "VAR%d", var_index);
		res |= ast_msg_set_var(msg, varname, vars[var_index]);
	}
	res |= ast_msg_set_var(msg, "ROLE", role);
	res |= ast_msg_set_var(msg, "ORIG_CHANNEL", conf->channelName);

	if (res) {
		ast_msg_destroy(msg);
		status_code = PJSIP_SC_INTERNAL_SERVER_ERROR;
	}
	else {
		ast_msg_queue(msg);
	}

	// Switch to original context
	ast_string_field_set(endpoint, context, context);

	// Clean up
	ast_free(dest_to);
	ast_free(from_uri);
	ast_free(context);

	return status_code;
}

static enum pjsip_status_code ccm_command_handler(struct ast_sip_endpoint *endpoint,
										   		  struct sip_conference* conf,
												  struct pjsip_rx_data *rdata,
												  char *ccm_msg,
												  char *role, char *command,
												  char **vars, int num_vars)
{
	enum pjsip_status_code status_code = PJSIP_SC_OK;
	char *conference = NULL;

	// Sanity check
	if (endpoint == NULL || conf == NULL || rdata == NULL ||
		ccm_msg == NULL || role == NULL || command == NULL ||
		vars == NULL) {
		return PJSIP_SC_FORBIDDEN;
	}

	conference = conf->conference;

	// Handlers
	if (strcmp(command, "KICK") == 0) {
		if (num_vars == 1) {
			status_code = ccm_publish_command_kick(conference,
												   conf->channelName, vars[0]);
		}
		else {
			status_code = PJSIP_SC_FORBIDDEN;
		}
	}
	else if (strcmp(command, "MUTE") == 0) {
		if (num_vars == 1) {
			status_code = ccm_publish_command_mute_unmute(conference, vars[0], 1);
		}
		else {
			status_code = PJSIP_SC_FORBIDDEN;
		}
	}
	else if (strcmp(command, "UNMUTE") == 0) {
		if (num_vars == 1) {
			status_code = ccm_publish_command_mute_unmute(conference, vars[0], 0);
		}
		else {
			status_code = PJSIP_SC_FORBIDDEN;
		}
	}
	else if (strcmp(command, "LOCK") == 0) {
		status_code = ccm_publish_command_lock_unlock(conference, 1);
	}
	else if (strcmp(command, "UNLOCK") == 0) {
		status_code = ccm_publish_command_lock_unlock(conference, 0);
	}
	else if (strcmp(command, "INVITE") == 0) {
		if (num_vars == 2) {
			status_code = ccm_msg_queue(endpoint, conf, rdata, ccm_msg,
										role, command, vars, num_vars);
		}
		else {
			status_code = PJSIP_SC_FORBIDDEN;
		}
	}
	// COMMAND|TARGET CONFERENCE ROOM NUMBER|SHA1 HASH of target conference room user password (optional)
	// MCB|5300|40bd001563085fc35165329ea1ff5c5ecbdbbeef
	else if (strcmp(command, "MCB") == 0) {
		if (num_vars == 1 || num_vars == 2) {
			status_code = ccm_msg_queue(endpoint, conf, rdata, ccm_msg,
										role, command, vars, num_vars);
		}
		else {
			status_code = PJSIP_SC_FORBIDDEN;
		}
	}
	else if (strcmp(command, "DROP") == 0) {
		status_code = ccm_publish_command_kick(conf->conference,
											   conf->channelName, "ALL_MCB");
	}
	else {
		ast_log(LOG_DEBUG, "CCM command %s is not supported", command);
		status_code = PJSIP_SC_FORBIDDEN;
	}

	return status_code;
}

static enum pjsip_status_code process_incoming_ccm_message(struct ast_sip_session *session,
														   struct pjsip_rx_data *rdata,
														   char *ccm_msg)
{
	struct ast_sip_endpoint *endpoint = NULL;
	struct sip_conference* conf = NULL;
	char *conference = NULL;
	char *role = "0";
	char *buffer = NULL;
	char *saveptr = NULL;
	char *command = NULL;
	char *vars[CCM_MAX_NUM_VARS];
	int num_vars = 0;
	int index = 0;
	enum pjsip_status_code status_code = PJSIP_SC_OK;

	// Sanity check
	if (session == NULL || session->endpoint == NULL ||
		rdata == NULL || ast_strlen_zero(ccm_msg)) {
		return PJSIP_SC_FORBIDDEN;
	}

	//===========================================
	// Get peer conference
	//===========================================
	endpoint = session->endpoint;
	conference = get_to_user(rdata);
	if (conference == NULL) {
		return PJSIP_SC_FORBIDDEN;
	}

	conf = ast_sip_cei_find_peer_conference(endpoint, conference);
	if (conf == NULL) {
		ast_log(LOG_DEBUG, "Received request from a user with no CEI");
		ast_free(conference);
		return PJSIP_SC_FORBIDDEN;
	}

	ast_log(LOG_DEBUG, "[TK] conference: %s, ceiRole: %d, data: %s\n",
                       conference, conf->ceiRole, ccm_msg);

	//===========================================
	// Verify CEI role
	//===========================================
	switch (conf->ceiRole) {
		case CEI_ROLE_USER_NORMAL:
		{
			ast_log(LOG_DEBUG, "Received request from a user with no powers, "
							   "ignore contents but return 200 OK");
			ast_free(conference);
			return PJSIP_SC_OK;
		}
		break;

		case CEI_ROLE_USER_INVITE:
		{
			role = "2";
		}
		break;

		case CEI_ROLE_ADMIN:
		{
			role = "3";
		}
		break;

		default:
		{
			ast_log(LOG_DEBUG, "Received request from a user with no CEI role (%d)",
							   conf->ceiRole);
			ast_free(conference);
			return PJSIP_SC_FORBIDDEN;
		}
		break;
	}

	//===========================================
	// Parse CCM message
	//===========================================
	// Get COMMAND
	buffer = strtok_r(ccm_msg, "|", &saveptr);
	if (buffer == NULL) {
		ast_log(LOG_DEBUG, "No command in CCM message");
		ast_free(conference);
		return PJSIP_SC_FORBIDDEN;
	}
	command = ast_strdup(buffer);

	// Get variables
	buffer = strtok_r(NULL, "|", &saveptr);
	while (buffer != NULL && num_vars < CCM_MAX_NUM_VARS) {
		vars[num_vars] = ast_strdup(buffer);
		num_vars++;
		buffer = strtok_r(NULL, "|", &saveptr);
	}

	// Handle CCM message
	status_code = ccm_command_handler(endpoint, conf, rdata, ccm_msg,
									  role, command, vars, num_vars);

	// Clean up
	ast_free(conference);
	ast_free(command);

	for (index = 0; index < num_vars; index++) {
		ast_free(vars[index]);
	}

	return status_code;
}
#endif

static int incoming_in_dialog_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	char buf[MAX_BODY_SIZE];
	enum pjsip_status_code code;
	struct ast_frame f;
	pjsip_dialog *dlg = session->inv_session->dlg;
	pjsip_transaction *tsx = pjsip_rdata_get_tsx(rdata);

	if (!session->channel) {
		send_response(rdata, PJSIP_SC_NOT_FOUND, dlg, tsx);
		return 0;
	}

	if ((code = check_content_type(rdata)) != PJSIP_SC_OK) {
		send_response(rdata, code, dlg, tsx);
		return 0;
	}

	if (print_body(rdata, buf, sizeof(buf)-1) < 1) {
		/* invalid body size */
		send_response(rdata, PJSIP_SC_REQUEST_ENTITY_TOO_LARGE, dlg, tsx);
		return 0;
	}

	ast_debug(3, "Received in dialog SIP message\n");

	memset(&f, 0, sizeof(f));
	f.frametype = AST_FRAME_TEXT;
	f.subclass.integer = 0;
	f.offset = 0;
	f.data.ptr = buf;
	f.datalen = strlen(buf) + 1;
	ast_queue_frame(session->channel, &f);

#ifdef GRANDSTREAM_NETWORKS
	// In-dialog MESSAGE support for Conference Control Message
	if (strncmp(buf, "CCM|", 4) == 0 && strlen(buf) > 4) {
		code = process_incoming_ccm_message(session, rdata, buf + 4);
		send_response(rdata, code, dlg, tsx);
	}
	else {
		send_response(rdata, PJSIP_SC_ACCEPTED, dlg, tsx);
	}
#else
	send_response(rdata, PJSIP_SC_ACCEPTED, dlg, tsx);
#endif
	return 0;
}

static struct ast_sip_session_supplement messaging_supplement = {
	.method = "MESSAGE",
	.incoming_request = incoming_in_dialog_request
};

static pjsip_module messaging_module = {
	.name = {"Messaging Module", 16},
	.id = -1,
	.priority = PJSIP_MOD_PRIORITY_APPLICATION,
	.on_rx_request = module_on_rx_request,
};

static int load_module(void)
{
	CHECK_PJSIP_SESSION_MODULE_LOADED();

	if (ast_sip_register_service(&messaging_module) != PJ_SUCCESS) {
		return AST_MODULE_LOAD_DECLINE;
	}

	if (pjsip_endpt_add_capability(ast_sip_get_pjsip_endpoint(),
				       NULL, PJSIP_H_ALLOW, NULL, 1,
				       &pjsip_message_method.name) != PJ_SUCCESS) {

		ast_sip_unregister_service(&messaging_module);
		return AST_MODULE_LOAD_DECLINE;
	}

	if (ast_msg_tech_register(&msg_tech)) {
		ast_sip_unregister_service(&messaging_module);
		return AST_MODULE_LOAD_DECLINE;
	}

	message_serializer = ast_sip_create_serializer();
	if (!message_serializer) {
		ast_sip_unregister_service(&messaging_module);
		ast_msg_tech_unregister(&msg_tech);
		return AST_MODULE_LOAD_DECLINE;
	}

	ast_sip_session_register_supplement(&messaging_supplement);
	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_session_unregister_supplement(&messaging_supplement);
	ast_msg_tech_unregister(&msg_tech);
	ast_sip_unregister_service(&messaging_module);
	ast_taskprocessor_unreference(message_serializer);
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP Messaging Support",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.load_pri = AST_MODPRI_APP_DEPEND,
	       );
