/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Chuandong Ruan <chdruan@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#ifdef GRANDSTREAM_NETWORKS
#define AST_MODULE_LOG "pjsip"

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 13.0 $");

#include <sys/time.h>
#include <sys/signal.h>
#include <netinet/in.h>
#include <ctype.h>

#include "asterisk/lock.h"
#include "asterisk/file.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/app.h"
#include "asterisk/module.h"
#include "asterisk/manager.h"
#include "asterisk/config.h"
#include "asterisk/monitor.h"
#include "asterisk/utils.h"
#include "asterisk/astdb.h"
#include "asterisk/devicestate.h"
#include "asterisk/astobj2.h"
#include "asterisk/strings.h"
#include "asterisk/global_datastores.h"
#include "asterisk/taskprocessor.h"
#include "asterisk/data.h"
#include "asterisk/stasis_channels.h"
#include "asterisk/stasis_message_router.h"
#include "asterisk/core_local.h"
#include "asterisk/mixmonitor.h"


#define RINGGROUP_CONFIG_FILE  "ringgroups.conf"
#define SIP_EXT_GROUP         "ext-group"

typedef enum {
	EN_EVENTLISTS_MARKED,
	EN_EVENTLISTS_ALL
} EN_EVENTLIST_UNLINK_FLAG;

struct ao2_container *con_ringgroups = NULL;
/* to define a ringgroup member */
struct st_state_member
{
	char   cMember[80];
	char   cInterface[80];
	enum   ast_extension_states enState;     /*!< current extension state */
	AST_LIST_ENTRY(st_state_member) entry;  /*!< next member record */
};

#ifdef LOW_MEMORY
static const int HASH_EVENTLIST_SIZE = 17;
#else
static const int HASH_EVENTLIST_SIZE = 563;	/*!< Size of eventlist hash table, prime number preferred! */
#endif

/* to define a simplify struct of ringgroup */
struct ringgroup_state
{
	char   cExten[80];
	char   cName[80];
	char   cMemberlistStr[1024]; 
	int    iDevState;          /*! to record the device state of ringgroup */
	int    iMemberNum;
	int    bMark;              /*!< moved out of ASTOBJ into struct proper; That which bears the_mark should be deleted! */
	AST_LIST_HEAD(, st_state_member) ast_member_list;           /*!< members in the ring group */
};

static struct stasis_subscription *ringgroup_state_sub;
static int ringgroup_check_ringing(char* rgrp_exten, char* member_intf);

#define ______EVENTLIST_RINGGROUP_STATE_______

/*************************************************
 * Function Name : ringgroup_hash_cb 
 * Description: The only member of the ringgroup_state used here is the exten field
 * Changes: 
*************************************************/
static int ringgroup_hash_cb(const void *obj, const int flags)
{
	ast_log(LOG_ERROR, "*****ringgroup_hash_cb.\n");
	const struct ringgroup_state  *pstRinggroup = obj;

	if (NULL == obj) {
	return -1;
	}

	return ast_str_case_hash(pstRinggroup->cExten);
}

/*************************************************
 * Function Name : ringgroup_cmp_cb 
 * Description: The only member of the ringgroup_state used here is the exten field
 * Changes: 
*************************************************/
static int ringgroup_cmp_cb(void *obj, void *arg, int flags)
{
	struct ringgroup_state  *pstRinggroup  = obj; 
	struct ringgroup_state  *pstRinggroup2 = arg;

	if (NULL == obj || NULL == arg) {
		ast_log(LOG_ERROR, "ringgroup_cmp_cb, input null\n");
		return -1;
	}

	return !strcasecmp(pstRinggroup->cExten, pstRinggroup2->cExten) ? CMP_MATCH | CMP_STOP : 0;
}

/*! \brief set mark flag to ringgroup, prepare for destroy */
static int ringgroup_markall(void *device, void *arg, int flags)
{
	struct ringgroup_state  *pstRinggroup = device;

	pstRinggroup->bMark = 1;
	return 0;
}

/*! this func is used with ao2_callback to unlink/delete all marked or linked ringgroups, depending on arg */
static int ringgroup_match_and_cleanup(void *ringgroupObj, void *arg, int flags)
{
	struct ringgroup_state   *pstRinggroup = ringgroupObj;
	EN_EVENTLIST_UNLINK_FLAG  flag = *(EN_EVENTLIST_UNLINK_FLAG *)arg;

	if ((EN_EVENTLISTS_ALL == flag) || (pstRinggroup->bMark)) {
		ast_log(LOG_DEBUG, "ringgroup_match_and_cleanup, clean list[%s]\n", pstRinggroup->cExten);
		return CMP_MATCH;
    }

    return 0;
}

/*! Unlink all marked ringgroups from ao2 containers */
static void ringgroup_unlink_marked_from_tables(void)
{
	EN_EVENTLIST_UNLINK_FLAG  flag = EN_EVENTLISTS_MARKED;

	ao2_t_callback(con_ringgroups, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE,
				ringgroup_match_and_cleanup, &flag, "initiating callback to remove marked ringgroups");
}

/*! Unlink all ringgroup from ao2 containers */
static void ringgroup_unlink_all_from_tables(void)
{
	EN_EVENTLIST_UNLINK_FLAG  flag = EN_EVENTLISTS_ALL;

	ao2_t_callback(con_ringgroups, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE,
				ringgroup_match_and_cleanup, &flag, "initiating callback to remove all ringgroups");
}

/*! this func is used for compare two ringgroup's name  */
#if 0
static int ringgroup_comparefunc(const void *a, const void *b)
{
	struct ringgroup_state **ap = (struct ringgroup_state **)a;
	struct ringgroup_state **bp = (struct ringgroup_state **)b;

	if (*ap == NULL || *bp == NULL) {
	return -1;
	}

	return strcmp((*ap)->cName, (*bp)->cName);
}

static int ringgroup_find_by_exten(void *obj, void *arg, int flags)
{
	struct ringgroup_state *search = obj;
	struct ringgroup_state *match  = arg;

	if (obj == NULL || arg == NULL) {
		return 0;
	}
         
	if (strcmp(search->cExten, match->cExten)) {
		return 0;
	}

	return CMP_MATCH | CMP_STOP;
}

static struct ringgroup_state *ringgroup_find(const char *cExten)
{
	struct ringgroup_state *pstRinggroup    = NULL;
	struct ringgroup_state  stRinggroup_tmp;

	if (NULL == cExten) {
		ast_log(LOG_ERROR, "ringgroup_find, cExten is null\n");
		return NULL;
	}

	ast_copy_string(stRinggroup_tmp.cExten, cExten, sizeof(stRinggroup_tmp.cExten));
	pstRinggroup = ao2_t_callback(con_ringgroups, OBJ_POINTER, ringgroup_find_by_exten, &stRinggroup_tmp, "ao2_find in ringgroups table");

	return pstRinggroup;
}
#endif

/*! \brief Destroy ringgroup object from memory */
static void ringgroup_destroy(void *data)
{
	struct ringgroup_state  *pstRinggroup = data;
	struct st_state_member  *pstStateMem  = NULL;

	if (NULL == pstRinggroup) {
		return;
	}

	ast_debug(LOG_VERBOSE, "Destroying SIP ringgroup %s\n", pstRinggroup->cExten);

   // ast_devstate_changed(AST_DEVICE_REMOVED, AST_DEVSTATE_CACHABLE, "ringgroup:%s", pstRinggroup->cExten);

	AST_LIST_LOCK(&pstRinggroup->ast_member_list);
	while ((pstStateMem = AST_LIST_REMOVE_HEAD(&pstRinggroup->ast_member_list, entry))) {
	ast_free(pstStateMem);
	}

	AST_LIST_UNLOCK(&pstRinggroup->ast_member_list);
}

static void *ringgroup_unref(struct ringgroup_state *pstRinggroup, char *tag)
{
    ao2_t_ref(pstRinggroup, -1, tag);
	return NULL;
}

#define SIP_DEFAULT_CONTEXT    "ext-local"

/*! \brief create st_state_member list by parse the member string*/
static void ringgroup_create_member_list(struct ringgroup_state *pstRinggroup)
{
	struct st_state_member *pstStateMem = NULL;
	char   *pclist      = ast_strdupa(pstRinggroup->cMemberlistStr);
	int    iMemberCount = 0;
	char   *pcMember    = NULL;
	char   family[80]   = {0};
	char   data[256]    = {0};

	if (NULL == pstRinggroup) {
		return;
	}

    /* parse the member string to a member list, format: 1000-1004-1009 */
	AST_LIST_LOCK(&pstRinggroup->ast_member_list);
	while ((pcMember = strsep(&pclist, ","))) {
		pcMember = ast_skip_blanks(pcMember);
		if (NULL == pcMember) {
			continue;
		}

		pstStateMem = ast_calloc(1, sizeof(struct st_state_member));
		if (NULL == pstStateMem) {
			AST_LIST_UNLOCK(&pstRinggroup->ast_member_list);
			pstRinggroup->iMemberNum = iMemberCount;
			return;
		}
		
		/* get AAUSER DIAL data from ucmdb */
		snprintf(family, sizeof(family), "AAUSER/%s", pcMember);

        /* get DIAL */
		if (ast_ucm_db_get(family, "DIAL", data, sizeof(data))) {
			ast_free(pstStateMem);
			continue;
        }

		ast_copy_string(pstStateMem->cMember, pcMember, sizeof(pstStateMem->cMember));
		ast_copy_string(pstStateMem->cInterface, data, sizeof(pstStateMem->cInterface));

		pstStateMem->enState = ast_extension_state(NULL, SIP_DEFAULT_CONTEXT, pcMember);

		ast_log(LOG_VERBOSE, "Created a state member [%s], interface[%s], state[%s].\n", 
				pcMember, pstStateMem->cInterface, ast_extension_state2str(pstStateMem->enState));

		AST_LIST_INSERT_TAIL(&pstRinggroup->ast_member_list, pstStateMem, entry);
		iMemberCount++;
	}

	AST_LIST_UNLOCK(&pstRinggroup->ast_member_list);
	pstRinggroup->iMemberNum = iMemberCount;

	return;
}

static int ringgroup_getstate(struct ringgroup_state  *pstRinggroup)
{
	int dev_state = AST_DEVICE_INUSE;
	struct st_state_member *pstStateMem;

	if (NULL == pstRinggroup) {
		return dev_state;
	}

	AST_LIST_TRAVERSE(&(pstRinggroup->ast_member_list), pstStateMem, entry) {
		if ((AST_EXTENSION_REMOVED != pstStateMem->enState && AST_EXTENSION_DEACTIVATED != pstStateMem->enState) &&
		    (AST_EXTENSION_RINGING & pstStateMem->enState) && ringgroup_check_ringing(pstRinggroup->cExten, pstStateMem->cInterface)) {
			dev_state = AST_DEVICE_RINGING;
			break;
		} else if (AST_EXTENSION_NOT_INUSE == pstStateMem->enState) {
			dev_state = AST_DEVICE_NOT_INUSE;
		} else if (AST_DEVICE_NOT_INUSE != dev_state) {
			dev_state = AST_DEVICE_INUSE;
		}
	}

	ast_log(LOG_VERBOSE,"ringgroup_getstate, group[%s]--state[%s]",pstRinggroup->cExten, ast_devstate2str(dev_state));
	return dev_state;
}

static struct ringgroup_state *ringgroup_build(const char *cExten, struct ast_variable *pstVar, struct ast_variable *pstVarregs)
{
	struct ringgroup_state  *pstRinggroup = NULL;
	struct st_state_member  *pstStateMem  = NULL;
	struct ringgroup_state   stRinggroup;
	int    bFound = 0;

	ast_copy_string(stRinggroup.cExten, cExten, sizeof(stRinggroup.cExten));
	pstRinggroup = ao2_t_find(con_ringgroups, &stRinggroup, OBJ_POINTER, "find ringgroup from ringgroups table");

	if (NULL == pstRinggroup) {
        /* alloc a new ringgroup */
		if (!(pstRinggroup = ao2_t_alloc(sizeof(struct ringgroup_state), ringgroup_destroy, "allocate a ringgroup struct"))) {
			return NULL;
		}
		strncpy(pstRinggroup->cExten, cExten, sizeof(pstRinggroup->cExten));
	} else {
		bFound = 1;
		AST_LIST_LOCK(&pstRinggroup->ast_member_list);
		while ((pstStateMem = AST_LIST_REMOVE_HEAD(&pstRinggroup->ast_member_list, entry))) {
			ast_free(pstStateMem);
		}

		AST_LIST_UNLOCK(&pstRinggroup->ast_member_list);
	}

	pstRinggroup->bMark = 0; 
	pstRinggroup->iMemberNum = 0; 
	memset(pstRinggroup->cMemberlistStr, 0, sizeof(pstRinggroup->cMemberlistStr));

	AST_LIST_HEAD_INIT(&pstRinggroup->ast_member_list);

	for (; pstVar || ((pstVar = pstVarregs) && !(pstVarregs = NULL)); pstVar = pstVar->next) {
		if (!strcasecmp(pstVar->name, "name")) {
            strncpy(pstRinggroup->cName, pstVar->value, sizeof(pstRinggroup->cName));
		} else if (!strcasecmp(pstVar->name, "member") && !ast_strlen_zero(pstVar->value)) {
            strncpy(pstRinggroup->cMemberlistStr, pstVar->value, sizeof(pstRinggroup->cMemberlistStr));
			ringgroup_create_member_list(pstRinggroup);
		}
	}

	ast_log(LOG_VERBOSE, "ringgroup_build, build [%s] succeed!\n", pstRinggroup->cExten);

	pstRinggroup->iDevState = ringgroup_getstate(pstRinggroup);
	ast_devstate_changed(pstRinggroup->iDevState, AST_DEVSTATE_CACHABLE, "Ringgroup:%s", pstRinggroup->cExten);

	if (!bFound) {
		ao2_t_link(con_ringgroups, pstRinggroup, "link ringgroup into ringgroup table");
	}

	return pstRinggroup;
}

static void ringgroup_rebuild_all(void)
{
	struct ringgroup_state  *pstRinggroup = NULL;
	struct ast_flags         config_flags = {0};
	struct ast_config       *pstCfg       = NULL;
	struct ast_variable     *pstVarGen    = NULL;
	char  *pCategory   = NULL;
	int    iObjcount   = 0;

	/* load ringgroups.conf file*/
	pstCfg = ast_config_load(RINGGROUP_CONFIG_FILE, config_flags);
	if (pstCfg == CONFIG_STATUS_FILEUNCHANGED || NULL == pstCfg) {
		return;
	}

	if (pstCfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Contents of ringgroups.conf are invalid and cannot be parsed\n");
		return;
	}

	ao2_t_callback(con_ringgroups, OBJ_NODATA, ringgroup_markall, NULL, "callback to mark all ringgroups");

	pstVarGen = ast_variable_browse(pstCfg, "general");
	pCategory = ast_category_browse(pstCfg, NULL);
	while (pCategory) {
	if (strcasecmp(pCategory, "general")) {
			pstRinggroup = ringgroup_build(pCategory, pstVarGen, ast_variable_browse(pstCfg, pCategory));
			if (pstRinggroup) {
				ringgroup_unref(pstRinggroup, "ringgroup_unref: from ringgroup_rebuild_all");
				iObjcount++;
			}
		}
		pCategory = ast_category_browse(pstCfg, pCategory);
	}

	ast_config_destroy(pstCfg);

	ast_log(LOG_VERBOSE, "ringgroup_rebuild_all, total count[%d]!\n", iObjcount);
	ringgroup_unlink_marked_from_tables();
	ast_log(LOG_VERBOSE, "ringgroup_rebuild_all, finished!\n");
}

static int ringgroup_check_ringing(char* rgrp_exten, char* member_intf)
{
	struct ast_channel_iterator *iter;
	struct ast_channel *c = NULL;
	char   chan_name[120] = {0};
	char  *chan_dev       = NULL;
	int    ret = 0;

	if (!(iter = ast_channel_iterator_all_new())) {
		return 0;
	}

	while ((c = ast_channel_iterator_next(iter))) {
		ast_channel_lock(c);

        strncpy(chan_name, ast_channel_name(c), sizeof(chan_name));
		chan_dev = strrchr(chan_name, '-');
		if (chan_dev) {
			*chan_dev = '\0';
		}

		if ((0 == strcmp(chan_name, member_intf)) && (0 == strcmp(ast_channel_exten(c), rgrp_exten))) {
			ret = 1;
			ast_channel_unlock(c);
			ast_channel_unref(c);
			break;
		}

		ast_channel_unlock(c);
		ast_channel_unref(c);
	}

	ast_channel_iterator_destroy(iter);

	return ret;
}


/*! \brief set a member's status based on device state of that member's interface*/
static void ringgroup_state_cb(void *unused, struct stasis_subscription *sub, struct stasis_message *msg)
{
	struct ao2_iterator qiter;
	struct ast_device_state_message *dev_state;
	struct st_state_member *pstStateMem;
	struct ringgroup_state *pstRinggroup;
	enum   ast_extension_states iState;
	int state_changed = 0;
	int device_state = AST_DEVICE_NOT_INUSE;

	if (ast_device_state_message_type() != stasis_message_type(msg)) {
		return;
	}

	dev_state = stasis_message_data(msg);
	if (dev_state->eid) {
		/* ignore non-aggregate states */
		return;
	}

	iState = ast_devstate_to_extenstate((enum ast_device_state)dev_state->state);
	qiter = ao2_iterator_init(con_ringgroups, 0);
	while ((pstRinggroup = ao2_t_iterator_next(&qiter, "Iterate over ringgroups"))) {
		device_state = AST_DEVICE_INUSE; /* [Bug 21164] */
		state_changed = 0;

		ao2_lock(pstRinggroup);

		AST_LIST_LOCK(&pstRinggroup->ast_member_list);
		AST_LIST_TRAVERSE(&(pstRinggroup->ast_member_list), pstStateMem, entry) {
			if (!strcasecmp(pstStateMem->cInterface, dev_state->device)) {
				if (pstStateMem->enState != iState) {
					pstStateMem->enState = iState;
					state_changed = 1;
				}
				break;
			}
		}

		if (state_changed) {
			AST_LIST_TRAVERSE(&(pstRinggroup->ast_member_list), pstStateMem, entry) {
				if ((AST_EXTENSION_REMOVED != pstStateMem->enState && AST_EXTENSION_DEACTIVATED != pstStateMem->enState) && 
					(AST_EXTENSION_RINGING & pstStateMem->enState) && ringgroup_check_ringing(pstRinggroup->cExten, pstStateMem->cInterface)) {
					device_state = AST_DEVICE_RINGING;
					break;
				} else if (AST_EXTENSION_NOT_INUSE == pstStateMem->enState) {
					device_state = AST_DEVICE_NOT_INUSE;
				} else if (AST_DEVICE_NOT_INUSE != device_state) {
					device_state = AST_DEVICE_INUSE;
				}
			}

			if (device_state != pstRinggroup->iDevState) {
				ast_debug(1, "ringgroup '%s' changed from '%d' [%s] to state '%d' [%s], caused by member[%s] in state[%s]\n", pstRinggroup->cExten, 
							  pstRinggroup->iDevState, ast_devstate2str(pstRinggroup->iDevState),
							  device_state, ast_devstate2str(device_state), dev_state->device, ast_devstate2str(dev_state->state));
				pstRinggroup->iDevState = device_state;
				ast_devstate_changed(device_state, AST_DEVSTATE_CACHABLE, "Ringgroup:%s", pstRinggroup->cExten);
			}
		}

		AST_LIST_UNLOCK(&pstRinggroup->ast_member_list);
		ao2_unlock(pstRinggroup);
		ringgroup_unref(pstRinggroup, "Done with iterator");
	}

	ao2_iterator_destroy(&qiter);


	return;
}

/* load eventlist submodule - initialization */
static int eventlist_ringgroup_load(void)
{
	ast_verbose("Event list Ringgroup loading...\n");

	con_ringgroups = ao2_t_container_alloc(HASH_EVENTLIST_SIZE, ringgroup_hash_cb, ringgroup_cmp_cb, "allocate ringgroups");
	if (!con_ringgroups) {
		ast_log(LOG_ERROR, "Unable to create primary con_ringgroups container(s)\n");
		return AST_MODULE_LOAD_FAILURE;
	}

    /* rebuild ringgroups*/
	ringgroup_rebuild_all();

	ringgroup_state_sub = stasis_subscribe(ast_device_state_topic_all(), ringgroup_state_cb, NULL);
	if (!ringgroup_state_sub) {
		return -1;
	}

	return 0;
}
/*! unload eventlist submodule */
static void eventlist_ringgroup_unload(void)
{
	ast_log(LOG_NOTICE, "unloading SIP Event list...");

    /* unlink all ringgroups*/
	ringgroup_unlink_all_from_tables();
	ringgroup_state_sub = stasis_unsubscribe_and_join(ringgroup_state_sub);

}

static int unload_module(void)
{
	eventlist_ringgroup_unload();
	return 0;
}


static int load_module(void)
{
	int err = 0;
	err = eventlist_ringgroup_load();

	if (err) {
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}
	return AST_MODULE_LOAD_SUCCESS;
}

static int reload(void)
{
	/* rebuild ringgroups*/
	ringgroup_rebuild_all();
	return 0;
}


AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP EVENT Ringgroup",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
		.load_pri = AST_MODPRI_APP_DEPEND,
);
#endif
