/*
* Asterisk -- An open source telephony toolkit.
*
* Copyright (C) 2013, Digium, Inc.
*
* Mark Michelson <mmichelson@digium.com>
*
* See http://www.asterisk.org for more information about
* the Asterisk project. Please do not directly contact
* any of the maintainers of this project for assistance;
* the project provides a web site, mailing lists and IRC
* channels for your use.
*
* This program is free software, distributed under the terms of
* the GNU General Public License Version 2. See the LICENSE file
* at the top of the source tree.
*/

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#ifdef GRANDSTREAM_NETWORKS
#define AST_MODULE_LOG "pjsip"
#endif

#include "asterisk.h"

#ifdef GRANDSTREAM_NETWORKS
ASTERISK_FILE_VERSION(__FILE__, "$Revision: 13.0 $");
#endif

#include <pjsip.h>
#include <pjsip_ua.h>
#include <pjlib.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/datastore.h"
#include "asterisk/module.h"
#include "asterisk/logger.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/astobj2.h"
#include "asterisk/lock.h"
#include "asterisk/uuid.h"
#include "asterisk/pbx.h"
#include "asterisk/taskprocessor.h"
#include "asterisk/causes.h"
#include "asterisk/sdp_srtp.h"
#include "asterisk/dsp.h"
#include "asterisk/acl.h"
#include "asterisk/features_config.h"
#include "asterisk/pickup.h"

#ifdef GRANDSTREAM_NETWORKS
/*add by wxu@grandstream.cn*/
#include "asterisk/security_events.h"
/*end*/
#endif

#define SDP_HANDLER_BUCKETS 11

#define MOD_DATA_ON_RESPONSE "on_response"
#define MOD_DATA_NAT_HOOK "nat_hook"

/* Some forward declarations */
static void handle_incoming_request(struct ast_sip_session *session, pjsip_rx_data *rdata, pjsip_event_id_e type);
static void handle_incoming_response(struct ast_sip_session *session, pjsip_rx_data *rdata, pjsip_event_id_e type,
		enum ast_sip_session_response_priority response_priority);
static int handle_incoming(struct ast_sip_session *session, pjsip_rx_data *rdata, pjsip_event_id_e type,
		enum ast_sip_session_response_priority response_priority);
static void handle_outgoing_request(struct ast_sip_session *session, pjsip_tx_data *tdata);
static void handle_outgoing_response(struct ast_sip_session *session, pjsip_tx_data *tdata);
static void handle_outgoing(struct ast_sip_session *session, pjsip_tx_data *tdata);

#ifdef GRANDSTREAM_NETWORKS
/*add by wxu@grandstream.cn*/
static int hangup_sip2cause(int cause);
static void ast_sip_report_invite_failed(struct ast_sip_session *session,struct pjsip_rx_data *rdata,struct pjsip_tx_data *tdata,
	const char *sipmethod, int resp, const char* resp_text);
/*end*/
static pj_bool_t session_search_audio_port_with_image_audio(struct ast_sip_session *session, int image_audio, int is_t38_enable, pjmedia_sdp_session *receive_sdp, const char *media_type_name);
#endif

/*! \brief NAT hook for modifying outgoing messages with SDP */
static struct ast_sip_nat_hook *nat_hook;

/*!
 * \brief Registered SDP stream handlers
 *
 * This container is keyed on stream types. Each
 * object in the container is a linked list of
 * handlers for the stream type.
 */
static struct ao2_container *sdp_handlers;

/*!
 * These are the objects in the sdp_handlers container
 */
struct sdp_handler_list {
	/* The list of handlers to visit */
	AST_LIST_HEAD_NOLOCK(, ast_sip_session_sdp_handler) list;
	/* The handlers in this list handle streams of this type */
#ifdef GRANDSTREAM_NETWORKS
	/* Modify by bfdu for BFCP, 2015.07.23 */
	char stream_type[16];
#else
	char stream_type[1];
#endif
};

static struct pjmedia_sdp_session *create_local_sdp(pjsip_inv_session *inv, struct ast_sip_session *session, const pjmedia_sdp_session *offer);

static int sdp_handler_list_hash(const void *obj, int flags)
{
	const struct sdp_handler_list *handler_list = obj;
	const char *stream_type = flags & OBJ_KEY ? obj : handler_list->stream_type;

	return ast_str_hash(stream_type);
}

static int sdp_handler_list_cmp(void *obj, void *arg, int flags)
{
	struct sdp_handler_list *handler_list1 = obj;
	struct sdp_handler_list *handler_list2 = arg;
	const char *stream_type2 = flags & OBJ_KEY ? arg : handler_list2->stream_type;

	return strcmp(handler_list1->stream_type, stream_type2) ? 0 : CMP_MATCH | CMP_STOP;
}

static int session_media_hash(const void *obj, int flags)
{
	const struct ast_sip_session_media *session_media = obj;
	const char *stream_type = flags & OBJ_KEY ? obj : session_media->stream_type;

	return ast_str_hash(stream_type);
}

static int session_media_cmp(void *obj, void *arg, int flags)
{
	struct ast_sip_session_media *session_media1 = obj;
	struct ast_sip_session_media *session_media2 = arg;
	const char *stream_type2 = flags & OBJ_KEY ? arg : session_media2->stream_type;

	return strcmp(session_media1->stream_type, stream_type2) ? 0 : CMP_MATCH | CMP_STOP;
}

int ast_sip_session_register_sdp_handler(struct ast_sip_session_sdp_handler *handler, const char *stream_type)
{
	RAII_VAR(struct sdp_handler_list *, handler_list,
			ao2_find(sdp_handlers, stream_type, OBJ_KEY), ao2_cleanup);
	SCOPED_AO2LOCK(lock, sdp_handlers);

	if (handler_list) {
		struct ast_sip_session_sdp_handler *iter;
		/* Check if this handler is already registered for this stream type */
		AST_LIST_TRAVERSE(&handler_list->list, iter, next) {
			if (!strcmp(iter->id, handler->id)) {
				ast_log(LOG_WARNING, "Handler '%s' already registered for stream type '%s'.\n", handler->id, stream_type);
				return -1;
			}
		}
		AST_LIST_INSERT_TAIL(&handler_list->list, handler, next);
		ast_debug(1, "Registered SDP stream handler '%s' for stream type '%s'\n", handler->id, stream_type);
		ast_module_ref(ast_module_info->self);
		return 0;
	}

	/* No stream of this type has been registered yet, so we need to create a new list */
	handler_list = ao2_alloc(sizeof(*handler_list) + strlen(stream_type), NULL);
	if (!handler_list) {
		return -1;
	}
	/* Safe use of strcpy */
	strcpy(handler_list->stream_type, stream_type);
	AST_LIST_HEAD_INIT_NOLOCK(&handler_list->list);
	AST_LIST_INSERT_TAIL(&handler_list->list, handler, next);
	if (!ao2_link(sdp_handlers, handler_list)) {
		return -1;
	}
	ast_debug(1, "Registered SDP stream handler '%s' for stream type '%s'\n", handler->id, stream_type);
	ast_module_ref(ast_module_info->self);
	return 0;
}

static int remove_handler(void *obj, void *arg, void *data, int flags)
{
	struct sdp_handler_list *handler_list = obj;
	struct ast_sip_session_sdp_handler *handler = data;
	struct ast_sip_session_sdp_handler *iter;
	const char *stream_type = arg;

	AST_LIST_TRAVERSE_SAFE_BEGIN(&handler_list->list, iter, next) {
		if (!strcmp(iter->id, handler->id)) {
			AST_LIST_REMOVE_CURRENT(next);
			ast_debug(1, "Unregistered SDP stream handler '%s' for stream type '%s'\n", handler->id, stream_type);
			ast_module_unref(ast_module_info->self);
		}
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "remove registered SDP stream handler '%s' for stream type '%s'\n", handler->id, stream_type);
#endif
	AST_LIST_TRAVERSE_SAFE_END;

	if (AST_LIST_EMPTY(&handler_list->list)) {
		ast_debug(3, "No more handlers exist for stream type '%s'\n", stream_type);
		return CMP_MATCH;
	} else {
		return CMP_STOP;
	}
}

void ast_sip_session_unregister_sdp_handler(struct ast_sip_session_sdp_handler *handler, const char *stream_type)
{
	ao2_callback_data(sdp_handlers, OBJ_KEY | OBJ_UNLINK | OBJ_NODATA, remove_handler, (void *)stream_type, handler);
}

/*!
 * \brief Set an SDP stream handler for a corresponding session media.
 *
 * \note Always use this function to set the SDP handler for a session media.
 *
 * This function will properly free resources on the SDP handler currently being
 * used by the session media, then set the session media to use the new SDP
 * handler.
 */
static void session_media_set_handler(struct ast_sip_session_media *session_media,
		struct ast_sip_session_sdp_handler *handler)
{
	ast_assert(session_media->handler != handler);

	if (session_media->handler) {
		session_media->handler->stream_destroy(session_media);
	}
	session_media->handler = handler;
}

static int handle_incoming_sdp(struct ast_sip_session *session, const pjmedia_sdp_session *sdp)
{
	int i;
	int handled = 0;
#ifdef GRANDSTREAM_NETWORKS
    ast_log(LOG_VERBOSE, "starting handler incoming sdp ...\n");
#endif
	for (i = 0; i < sdp->media_count; ++i) {
		/* See if there are registered handlers for this media stream type */
		char media[20];
		struct ast_sip_session_sdp_handler *handler;
		RAII_VAR(struct sdp_handler_list *, handler_list, NULL, ao2_cleanup);
		RAII_VAR(struct ast_sip_session_media *, session_media, NULL, ao2_cleanup);
		int res;

		/* We need a null-terminated version of the media string */
		ast_copy_pj_str(media, &sdp->media[i]->desc.media, sizeof(media));

		session_media = ao2_find(session->media, media, OBJ_KEY);
		if (!session_media) {
			/* if the session_media doesn't exist, there weren't
			 * any handlers at the time of its creation */
			continue;
		}

		if (session_media->handler) {
			handler = session_media->handler;
			ast_debug(1, "Negotiating incoming SDP media stream '%s' using %s SDP handler\n",
				session_media->stream_type,
				session_media->handler->id);
			res = handler->negotiate_incoming_sdp_stream(session, session_media, sdp,
				sdp->media[i]);
			if (res < 0) {
				/* Catastrophic failure. Abort! */
				return -1;
			} else if (res > 0) {
				ast_debug(1, "Media stream '%s' handled by %s\n",
					session_media->stream_type,
					session_media->handler->id);
				/* Handled by this handler. Move to the next stream */
				handled = 1;
				continue;
			}			
#ifdef GRANDSTREAM_NETWORKS
			else if (0 == res && 0 == strncmp(session_media->stream_type, "audio", 5)) {
				ast_log(LOG_NOTICE, "no audio codecs available!");
				return -1;
			}
#endif
		}

		handler_list = ao2_find(sdp_handlers, media, OBJ_KEY);
		if (!handler_list) {
			ast_debug(1, "No registered SDP handlers for media type '%s'\n", media);
			continue;
		}
		AST_LIST_TRAVERSE(&handler_list->list, handler, next) {
			if (handler == session_media->handler) {
				continue;
			}
			ast_debug(1, "Negotiating incoming SDP media stream '%s' using %s SDP handler\n",
				session_media->stream_type,
				handler->id);
			res = handler->negotiate_incoming_sdp_stream(session, session_media, sdp,
				sdp->media[i]);
			if (res < 0) {
				/* Catastrophic failure. Abort! */
				return -1;
			}
			if (res > 0) {
				ast_debug(1, "Media stream '%s' handled by %s\n",
					session_media->stream_type,
					handler->id);
				/* Handled by this handler. Move to the next stream */
				session_media_set_handler(session_media, handler);
				handled = 1;
				break;
			}			
#ifdef GRANDSTREAM_NETWORKS
			else if (0 == res && 0 == strncmp(session_media->stream_type, "audio", 5)) {				
				ast_log(LOG_NOTICE, "no audio codecs available!");
				return -1;
			}
#endif
		}
	}
	if (!handled) {
		return -1;
	}
	return 0;
}

struct handle_negotiated_sdp_cb {
	struct ast_sip_session *session;
	const pjmedia_sdp_session *local;
	const pjmedia_sdp_session *remote;
};

static int handle_negotiated_sdp_session_media(void *obj, void *arg, int flags)
{
	struct ast_sip_session_media *session_media = obj;
	struct handle_negotiated_sdp_cb *callback_data = arg;
	struct ast_sip_session *session = callback_data->session;
	const pjmedia_sdp_session *local = callback_data->local;
	const pjmedia_sdp_session *remote = callback_data->remote;
	int i;
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "starting handle negotiated sdp session media ...\n");
#endif
	for (i = 0; i < local->media_count; ++i) {
		/* See if there are registered handlers for this media stream type */
		char media[20];
		struct ast_sip_session_sdp_handler *handler;
		RAII_VAR(struct sdp_handler_list *, handler_list, NULL, ao2_cleanup);
		int res;

		if (!remote->media[i]) {
			continue;
		}

		/* We need a null-terminated version of the media string */
		ast_copy_pj_str(media, &local->media[i]->desc.media, sizeof(media));

		/* stream type doesn't match the one we're looking to fill */
		if (strcasecmp(session_media->stream_type, media)) {
			continue;
		}

		handler = session_media->handler;
		if (handler) {
			ast_debug(1, "Applying negotiated SDP media stream '%s' using %s SDP handler\n",
				session_media->stream_type,
				handler->id);
			res = handler->apply_negotiated_sdp_stream(session, session_media, local,
				local->media[i], remote, remote->media[i]);
			if (res >= 0) {
				ast_debug(1, "Applied negotiated SDP media stream '%s' using %s SDP handler\n",
					session_media->stream_type,
					handler->id);
#ifdef GRANDSTREAM_NETWORKS
				/* Modify by bfdu for BFCP, 2015.07.15 */
				ast_log(LOG_VERBOSE, "Need to apply next same media type stream with media handler\n");
#else
				return CMP_MATCH;
#endif
			}
#ifdef GRANDSTREAM_NETWORKS
			/* Modify by bfdu for BFCP, 2015.07.15 */
			if (res < 0) {
				/* Catastrophic failure. Abort! */
				return 0;
			}
#else
			return 0;
#endif
		}

		handler_list = ao2_find(sdp_handlers, media, OBJ_KEY);
		if (!handler_list) {
			ast_debug(1, "No registered SDP handlers for media type '%s'\n", media);
			continue;
		}
		AST_LIST_TRAVERSE(&handler_list->list, handler, next) {
			if (handler == session_media->handler) {
				continue;
			}
			ast_debug(1, "Applying negotiated SDP media stream '%s' using %s SDP handler\n",
				session_media->stream_type,
				handler->id);
			res = handler->apply_negotiated_sdp_stream(session, session_media, local,
				local->media[i], remote, remote->media[i]);
			if (res < 0) {
				/* Catastrophic failure. Abort! */
				return 0;
			}
			if (res > 0) {
				ast_debug(1, "Applied negotiated SDP media stream '%s' using %s SDP handler\n",
					session_media->stream_type,
					handler->id);
				/* Handled by this handler. Move to the next stream */
				session_media_set_handler(session_media, handler);
#ifdef GRANDSTREAM_NETWORKS
				/* Modify by bfdu for BFCP, 2015.07.15 */
				ast_log(LOG_VERBOSE, "Need to apply next same media type stream with sdp handlers\n");
#else
				return CMP_MATCH;
#endif
			}
		}
	}
	return CMP_MATCH;
}

static int handle_negotiated_sdp(struct ast_sip_session *session, const pjmedia_sdp_session *local, const pjmedia_sdp_session *remote)
{
	RAII_VAR(struct ao2_iterator *, successful, NULL, ao2_iterator_cleanup);
	struct handle_negotiated_sdp_cb callback_data = {
		.session = session,
		.local = local,
		.remote = remote,
	};
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "starting handle negotiated sdp ...\n");
	/* Delete by bfdu for canreinvite, remove caps when apply, 2015.07.28 */
	//ast_format_cap_remove_by_type(session->req_caps, AST_MEDIA_TYPE_UNKNOWN);
#endif
	successful = ao2_callback(session->media, OBJ_MULTIPLE, handle_negotiated_sdp_session_media, &callback_data);
#ifdef GRANDSTREAM_NETWORKS
	ast_format_cap_remove_by_type(session->final_caps, AST_MEDIA_TYPE_UNKNOWN);
	ast_format_cap_append_from_cap(session->final_caps, session->req_caps, AST_MEDIA_TYPE_UNKNOWN);
#endif

	if (successful && ao2_iterator_count(successful) == ao2_container_count(session->media)) {
		/* Nothing experienced a catastrophic failure */
		ast_queue_frame(session->channel, &ast_null_frame);
		return 0;
	}
	return -1;
}

AST_RWLIST_HEAD_STATIC(session_supplements, ast_sip_session_supplement);

int ast_sip_session_register_supplement(struct ast_sip_session_supplement *supplement)
{
	struct ast_sip_session_supplement *iter;
	int inserted = 0;
	SCOPED_LOCK(lock, &session_supplements, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);

	if (!supplement->response_priority) {
		supplement->response_priority = AST_SIP_SESSION_BEFORE_MEDIA;
	}

	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&session_supplements, iter, next) {
		if (iter->priority > supplement->priority) {
			AST_RWLIST_INSERT_BEFORE_CURRENT(supplement, next);
			inserted = 1;
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;

	if (!inserted) {
		AST_RWLIST_INSERT_TAIL(&session_supplements, supplement, next);
	}
	ast_module_ref(ast_module_info->self);
	return 0;
}

void ast_sip_session_unregister_supplement(struct ast_sip_session_supplement *supplement)
{
	struct ast_sip_session_supplement *iter;
	SCOPED_LOCK(lock, &session_supplements, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&session_supplements, iter, next) {
		if (supplement == iter) {
			AST_RWLIST_REMOVE_CURRENT(next);
			ast_module_unref(ast_module_info->self);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
}

#ifdef GRANDSTREAM_NETWORKS

AST_RWLIST_HEAD_STATIC(exten_supplements, ast_sip_exten_supplement);

int ast_sip_exten_register_supplement(struct ast_sip_exten_supplement *supplement)
{
	if (!supplement->find_exten) {
		return -1;
	}

	SCOPED_LOCK(lock, &exten_supplements, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_INSERT_TAIL(&exten_supplements, supplement, next); 
	ast_module_ref(ast_module_info->self);

	return 0;
}

void ast_sip_exten_unregister_supplement(struct ast_sip_exten_supplement *supplement)
{
	struct ast_sip_exten_supplement *iter;
	SCOPED_LOCK(lock, &exten_supplements, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&exten_supplements, iter, next) {
		if (supplement == iter) {
			AST_RWLIST_REMOVE_CURRENT(next);
			ast_module_unref(ast_module_info->self);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
}

static int find_sub_handler_for_supplement_exten(enum ast_sip_supplement_type type, char *exten, void *data)
{
	struct ast_sip_exten_supplement *iter;

	SCOPED_LOCK(lock, &exten_supplements, AST_RWLIST_RDLOCK, AST_RWLIST_UNLOCK);

	AST_RWLIST_TRAVERSE(&exten_supplements, iter, next) {
		if (iter->exten_type != type) {
			continue;
		}

		if (0 == iter->find_exten(exten, data)) {
			return 0;
		}
	}
	return -1;
}

/* Begin add by jklou for BFCP/FEC/FECC */
enum ast_sip_application_type ast_sip_session_get_application_type(pj_str_t transport)
{
	if (NULL == transport.ptr) {
		ast_log(LOG_DEBUG, "get application type has no transport");
		return AST_SIP_APP_INVALID;
	}
	RAII_VAR(char *, transport_str, ast_strndup(transport.ptr, transport.slen), ast_free);

	if (strstr(transport_str, "UDP/BFCP")) {
		return AST_SIP_APP_UDP_BFCP;
	} else if (strstr(transport_str, "TCP/BFCP")) {
		return AST_SIP_APP_TCP_BFCP;
	} else if (strstr(transport_str, "RTP")) {
		return AST_SIP_APP_RTP;
	}

	return AST_SIP_APP_INVALID;
}

#endif

static struct ast_sip_session_supplement *supplement_dup(const struct ast_sip_session_supplement *src)
{
	struct ast_sip_session_supplement *dst = ast_calloc(1, sizeof(*dst));
	if (!dst) {
		return NULL;
	}
	/* Will need to revisit if shallow copy becomes an issue */
	*dst = *src;
	return dst;
}

#define DATASTORE_BUCKETS 53
#define MEDIA_BUCKETS 7

static void session_datastore_destroy(void *obj)
{
	struct ast_datastore *datastore = obj;

	/* Using the destroy function (if present) destroy the data */
	if (datastore->info->destroy != NULL && datastore->data != NULL) {
		datastore->info->destroy(datastore->data);
		datastore->data = NULL;
	}

	ast_free((void *) datastore->uid);
	datastore->uid = NULL;
}

struct ast_datastore *ast_sip_session_alloc_datastore(const struct ast_datastore_info *info, const char *uid)
{
	RAII_VAR(struct ast_datastore *, datastore, NULL, ao2_cleanup);
	char uuid_buf[AST_UUID_STR_LEN];
	const char *uid_ptr = uid;

	if (!info) {
		return NULL;
	}

	datastore = ao2_alloc(sizeof(*datastore), session_datastore_destroy);
	if (!datastore) {
		return NULL;
	}

	datastore->info = info;
	if (ast_strlen_zero(uid)) {
		/* They didn't provide an ID so we'll provide one ourself */
		uid_ptr = ast_uuid_generate_str(uuid_buf, sizeof(uuid_buf));
	}

	datastore->uid = ast_strdup(uid_ptr);
	if (!datastore->uid) {
		return NULL;
	}

	ao2_ref(datastore, +1);
	return datastore;
}

int ast_sip_session_add_datastore(struct ast_sip_session *session, struct ast_datastore *datastore)
{
	ast_assert(datastore != NULL);
	ast_assert(datastore->info != NULL);
	ast_assert(ast_strlen_zero(datastore->uid) == 0);

	if (!ao2_link(session->datastores, datastore)) {
		return -1;
	}
	return 0;
}

struct ast_datastore *ast_sip_session_get_datastore(struct ast_sip_session *session, const char *name)
{
	return ao2_find(session->datastores, name, OBJ_KEY);
}

void ast_sip_session_remove_datastore(struct ast_sip_session *session, const char *name)
{
	ao2_callback(session->datastores, OBJ_KEY | OBJ_UNLINK | OBJ_NODATA, NULL, (void *) name);
}

enum delayed_method {
	DELAYED_METHOD_INVITE,
	DELAYED_METHOD_UPDATE,
	DELAYED_METHOD_BYE,
};

/*!
 * \internal
 * \brief Convert delayed method enum value to to a string.
 * \since 13.3.0
 *
 * \param method Delayed method enum value to convert to a string.
 *
 * \return String value of delayed method.
 */
static const char *delayed_method2str(enum delayed_method method)
{
	const char *str = "<unknown>";

	switch (method) {
	case DELAYED_METHOD_INVITE:
		str = "INVITE";
		break;
	case DELAYED_METHOD_UPDATE:
		str = "UPDATE";
		break;
	case DELAYED_METHOD_BYE:
		str = "BYE";
		break;
	}

	return str;
}

/*!
 * \brief Structure used for sending delayed requests
 *
 * Requests are typically delayed because the current transaction
 * state of an INVITE. Once the pending INVITE transaction terminates,
 * the delayed request will be sent
 */
struct ast_sip_session_delayed_request {
	/*! Method of the request */
	enum delayed_method method;
	/*! Callback to call when the delayed request is created. */
	ast_sip_session_request_creation_cb on_request_creation;
	/*! Callback to call when the delayed request SDP is created */
	ast_sip_session_sdp_creation_cb on_sdp_creation;
	/*! Callback to call when the delayed request receives a response */
	ast_sip_session_response_cb on_response;
	/*! Whether to generate new SDP */
	int generate_new_sdp;
	AST_LIST_ENTRY(ast_sip_session_delayed_request) next;
};

static struct ast_sip_session_delayed_request *delayed_request_alloc(
	enum delayed_method method,
	ast_sip_session_request_creation_cb on_request_creation,
	ast_sip_session_sdp_creation_cb on_sdp_creation,
	ast_sip_session_response_cb on_response,
	int generate_new_sdp)
{
	struct ast_sip_session_delayed_request *delay = ast_calloc(1, sizeof(*delay));

	if (!delay) {
		return NULL;
	}
	delay->method = method;
	delay->on_request_creation = on_request_creation;
	delay->on_sdp_creation = on_sdp_creation;
	delay->on_response = on_response;
	delay->generate_new_sdp = generate_new_sdp;
	return delay;
}

static int send_delayed_request(struct ast_sip_session *session, struct ast_sip_session_delayed_request *delay)
{
	ast_debug(3, "Endpoint '%s(%s)' sending delayed %s request.\n",
		ast_sorcery_object_get_id(session->endpoint),
		session->channel ? ast_channel_name(session->channel) : "",
		delayed_method2str(delay->method));

	switch (delay->method) {
	case DELAYED_METHOD_INVITE:
		ast_sip_session_refresh(session, delay->on_request_creation,
			delay->on_sdp_creation, delay->on_response,
			AST_SIP_SESSION_REFRESH_METHOD_INVITE, delay->generate_new_sdp);
		return 0;
	case DELAYED_METHOD_UPDATE:
		ast_sip_session_refresh(session, delay->on_request_creation,
			delay->on_sdp_creation, delay->on_response,
			AST_SIP_SESSION_REFRESH_METHOD_UPDATE, delay->generate_new_sdp);
		return 0;
	case DELAYED_METHOD_BYE:
		ast_sip_session_terminate(session, 0);
		return 0;
	}
	ast_log(LOG_WARNING, "Don't know how to send delayed %s(%d) request.\n",
		delayed_method2str(delay->method), delay->method);
	return -1;
}

/*!
 * \internal
 * \brief The current INVITE transaction is in the PROCEEDING state.
 * \since 13.3.0
 *
 * \param vsession Session object.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int invite_proceeding(void *vsession)
{
	struct ast_sip_session *session = vsession;
	struct ast_sip_session_delayed_request *delay;
	int found = 0;
	int res = 0;

	AST_LIST_TRAVERSE_SAFE_BEGIN(&session->delayed_requests, delay, next) {
		switch (delay->method) {
		case DELAYED_METHOD_INVITE:
			break;
		case DELAYED_METHOD_UPDATE:
			AST_LIST_REMOVE_CURRENT(next);
			res = send_delayed_request(session, delay);
			ast_free(delay);
			found = 1;
			break;
		case DELAYED_METHOD_BYE:
			/* A BYE is pending so don't bother anymore. */
			found = 1;
			break;
		}
		if (found) {
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	ao2_ref(session, -1);
	return res;
}

/*!
 * \internal
 * \brief The current INVITE transaction is in the TERMINATED state.
 * \since 13.3.0
 *
 * \param vsession Session object.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int invite_terminated(void *vsession)
{
	struct ast_sip_session *session = vsession;
	struct ast_sip_session_delayed_request *delay;
	int found = 0;
	int res = 0;
	int timer_running;

	/* re-INVITE collision timer running? */
	timer_running = pj_timer_entry_running(&session->rescheduled_reinvite);

	AST_LIST_TRAVERSE_SAFE_BEGIN(&session->delayed_requests, delay, next) {
		switch (delay->method) {
		case DELAYED_METHOD_INVITE:
			if (!timer_running) {
				found = 1;
			}
			break;
		case DELAYED_METHOD_UPDATE:
		case DELAYED_METHOD_BYE:
			found = 1;
			break;
		}
		if (found) {
			AST_LIST_REMOVE_CURRENT(next);
			res = send_delayed_request(session, delay);
			ast_free(delay);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	ao2_ref(session, -1);
	return res;
}

/*!
 * \internal
 * \brief INVITE collision timeout.
 * \since 13.3.0
 *
 * \param vsession Session object.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int invite_collision_timeout(void *vsession)
{
	struct ast_sip_session *session = vsession;
	int res;

	if (session->inv_session->invite_tsx) {
		/*
		 * INVITE transaction still active.  Let it send
		 * the collision re-INVITE when it terminates.
		 */
		ao2_ref(session, -1);
		res = 0;
	} else {
		res = invite_terminated(session);
	}

	return res;
}

/*!
 * \internal
 * \brief The current UPDATE transaction is in the COMPLETED state.
 * \since 13.3.0
 *
 * \param vsession Session object.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int update_completed(void *vsession)
{
	struct ast_sip_session *session = vsession;
	int res;

	if (session->inv_session->invite_tsx) {
		res = invite_proceeding(session);
	} else {
		res = invite_terminated(session);
	}

	return res;
}

static void check_delayed_requests(struct ast_sip_session *session,
	int (*cb)(void *vsession))
{
	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, cb, session)) {
		ao2_ref(session, -1);
	}
}

static int delay_request(struct ast_sip_session *session,
	ast_sip_session_request_creation_cb on_request,
	ast_sip_session_sdp_creation_cb on_sdp_creation,
	ast_sip_session_response_cb on_response,
	int generate_new_sdp,
	enum delayed_method method)
{
	struct ast_sip_session_delayed_request *delay = delayed_request_alloc(method,
			on_request, on_sdp_creation, on_response, generate_new_sdp);

	if (!delay) {
		return -1;
	}

	if (method == DELAYED_METHOD_BYE) {
		/* Send BYE as early as possible */
		AST_LIST_INSERT_HEAD(&session->delayed_requests, delay, next);
	} else {
		AST_LIST_INSERT_TAIL(&session->delayed_requests, delay, next);
	}
	return 0;
}

static pjmedia_sdp_session *generate_session_refresh_sdp(struct ast_sip_session *session)
{
	pjsip_inv_session *inv_session = session->inv_session;
	const pjmedia_sdp_session *previous_sdp;

	if (pjmedia_sdp_neg_was_answer_remote(inv_session->neg)) {
		pjmedia_sdp_neg_get_active_remote(inv_session->neg, &previous_sdp);
	} else {
		pjmedia_sdp_neg_get_active_local(inv_session->neg, &previous_sdp);
	}
	return create_local_sdp(inv_session, session, previous_sdp);
}

int ast_sip_session_refresh(struct ast_sip_session *session,
		ast_sip_session_request_creation_cb on_request_creation,
		ast_sip_session_sdp_creation_cb on_sdp_creation,
		ast_sip_session_response_cb on_response,
		enum ast_sip_session_refresh_method method, int generate_new_sdp)
{
	pjsip_inv_session *inv_session = session->inv_session;
	pjmedia_sdp_session *new_sdp = NULL;
	pjsip_tx_data *tdata;

	if (inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		/* Don't try to do anything with a hung-up call */
		ast_debug(3, "Not sending reinvite to %s because of disconnected state...\n",
				ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	/* If the dialog has not yet been established we have to defer until it has */
	if (inv_session->dlg->state != PJSIP_DIALOG_STATE_ESTABLISHED) {
		ast_debug(3, "Delay sending request to %s because dialog has not been established...\n",
			ast_sorcery_object_get_id(session->endpoint));
		return delay_request(session, on_request_creation, on_sdp_creation, on_response,
			generate_new_sdp,
			method == AST_SIP_SESSION_REFRESH_METHOD_INVITE
				? DELAYED_METHOD_INVITE : DELAYED_METHOD_UPDATE);
	}

	if (method == AST_SIP_SESSION_REFRESH_METHOD_INVITE) {
		if (inv_session->invite_tsx) {
			/* We can't send a reinvite yet, so delay it */
			ast_debug(3, "Delay sending reinvite to %s because of outstanding transaction...\n",
					ast_sorcery_object_get_id(session->endpoint));
			return delay_request(session, on_request_creation, on_sdp_creation,
				on_response, generate_new_sdp, DELAYED_METHOD_INVITE);
		} else if (inv_session->state != PJSIP_INV_STATE_CONFIRMED) {
			/* Initial INVITE transaction failed to progress us to a confirmed state
			 * which means re-invites are not possible
			 */
			ast_debug(3, "Not sending reinvite to %s because not in confirmed state...\n",
					ast_sorcery_object_get_id(session->endpoint));
			return 0;
		}
	}

	if (generate_new_sdp) {
		/* SDP can only be generated if current negotiation has already completed */
		if (pjmedia_sdp_neg_get_state(inv_session->neg) != PJMEDIA_SDP_NEG_STATE_DONE) {
			ast_debug(3, "Delay session refresh with new SDP to %s because SDP negotiation is not yet done...\n",
				ast_sorcery_object_get_id(session->endpoint));
			return delay_request(session, on_request_creation, on_sdp_creation,
				on_response, generate_new_sdp,
				method == AST_SIP_SESSION_REFRESH_METHOD_INVITE
					? DELAYED_METHOD_INVITE : DELAYED_METHOD_UPDATE);
		}

		new_sdp = generate_session_refresh_sdp(session);
		if (!new_sdp) {
			ast_log(LOG_ERROR, "Failed to generate session refresh SDP. Not sending session refresh\n");
			return -1;
		}
		if (on_sdp_creation) {
			if (on_sdp_creation(session, new_sdp)) {
				return -1;
			}
		}
	}

	if (method == AST_SIP_SESSION_REFRESH_METHOD_INVITE) {
		if (pjsip_inv_reinvite(inv_session, NULL, new_sdp, &tdata)) {
			ast_log(LOG_WARNING, "Failed to create reinvite properly.\n");
			return -1;
		}
	} else if (pjsip_inv_update(inv_session, NULL, new_sdp, &tdata)) {
		ast_log(LOG_WARNING, "Failed to create UPDATE properly.\n");
		return -1;
	}
	if (on_request_creation) {
		if (on_request_creation(session, tdata)) {
			return -1;
		}
	}
	ast_debug(3, "Sending session refresh SDP via %s to %s\n",
		method == AST_SIP_SESSION_REFRESH_METHOD_INVITE ? "re-INVITE" : "UPDATE",
		ast_sorcery_object_get_id(session->endpoint));
	ast_sip_session_send_request_with_cb(session, tdata, on_response);
	return 0;
}

void ast_sip_session_send_response(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
#ifdef GRANDSTREAM_NETWORKS
    ast_log(LOG_VERBOSE, "sip session starting send response ...\n");
#endif
	handle_outgoing_response(session, tdata);
	pjsip_inv_send_msg(session->inv_session, tdata);
	return;
}

static pj_bool_t session_on_rx_request(pjsip_rx_data *rdata);

static pjsip_module session_module = {
	.name = {"Session Module", 14},
	.priority = PJSIP_MOD_PRIORITY_APPLICATION,
	.on_rx_request = session_on_rx_request,
};

/*! \brief Determine whether the SDP provided requires deferral of negotiating or not
 *
 * \retval 1 re-invite should be deferred and resumed later
 * \retval 0 re-invite should not be deferred
 */
static int sdp_requires_deferral(struct ast_sip_session *session, const pjmedia_sdp_session *sdp)
{
	int i;
#ifdef GRANDSTREAM_NETWORKS
		ast_debug(1, "sdp_requires_deferral enter...\n");
#endif
	for (i = 0; i < sdp->media_count; ++i) {
		/* See if there are registered handlers for this media stream type */
		char media[20];
		struct ast_sip_session_sdp_handler *handler;
		RAII_VAR(struct sdp_handler_list *, handler_list, NULL, ao2_cleanup);
		RAII_VAR(struct ast_sip_session_media *, session_media, NULL, ao2_cleanup);
		enum ast_sip_session_sdp_stream_defer res;

		/* We need a null-terminated version of the media string */
		ast_copy_pj_str(media, &sdp->media[i]->desc.media, sizeof(media));

		session_media = ao2_find(session->media, media, OBJ_KEY);
		if (!session_media) {
			/* if the session_media doesn't exist, there weren't
			 * any handlers at the time of its creation */
			continue;
		}

		if (session_media->handler) {
#ifdef GRANDSTREAM_NETWORKS
			ast_debug(1, "sdp_requires_deferral,handler\n");
#endif
			handler = session_media->handler;
			if (handler->defer_incoming_sdp_stream) {
				res = handler->defer_incoming_sdp_stream(session, session_media, sdp,
					sdp->media[i]);
#ifdef GRANDSTREAM_NETWORKS
				ast_debug(1, "defer_incoming_sdp_stream process result[%d]!\n", res);
#endif
				switch (res) {
				case AST_SIP_SESSION_SDP_DEFER_NOT_HANDLED:
					break;
				case AST_SIP_SESSION_SDP_DEFER_ERROR:
					return 0;
				case AST_SIP_SESSION_SDP_DEFER_NOT_NEEDED:
					break;
				case AST_SIP_SESSION_SDP_DEFER_NEEDED:
					return 1;
				}
			}
			/* Handled by this handler. Move to the next stream */
			continue;
		}

		handler_list = ao2_find(sdp_handlers, media, OBJ_KEY);
		if (!handler_list) {
			ast_debug(1, "No registered SDP handlers for media type '%s'\n", media);
			continue;
		}
		AST_LIST_TRAVERSE(&handler_list->list, handler, next) {
			if (handler == session_media->handler) {
				continue;
			}
			if (!handler->defer_incoming_sdp_stream) {
				continue;
			}
#ifdef GRANDSTREAM_NETWORKS
			ast_debug(1, "sdp_requires_deferral, traverse next media\n");
#endif
			res = handler->defer_incoming_sdp_stream(session, session_media, sdp,
				sdp->media[i]);
			switch (res) {
			case AST_SIP_SESSION_SDP_DEFER_NOT_HANDLED:
				continue;
			case AST_SIP_SESSION_SDP_DEFER_ERROR:
				session_media_set_handler(session_media, handler);
				return 0;
			case AST_SIP_SESSION_SDP_DEFER_NOT_NEEDED:
				/* Handled by this handler. */
				session_media_set_handler(session_media, handler);
				break;
			case AST_SIP_SESSION_SDP_DEFER_NEEDED:
				/* Handled by this handler. */
				session_media_set_handler(session_media, handler);
				return 1;
			}
			/* Move to the next stream */
			break;
		}
	}

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
static pj_bool_t session_check_sdp_media_type(const struct ast_sip_session *session, const pjmedia_sdp_session *sdp, const char *media_type_name)
{
	int i = 0;
	if (!sdp) {
		return PJ_FALSE;
	}

	for (i = 0; i < sdp->media_count; ++i) {
		RAII_VAR(struct ast_sip_session_media *, session_media, NULL, ao2_cleanup);
		session_media = ao2_find(session->media, media_type_name, OBJ_KEY);
		if (!session_media) {
			continue;
		}
		if (!pj_strcmp2(&sdp->media[i]->desc.media, media_type_name)) {
			return PJ_TRUE;
		}
	}

	return PJ_FALSE;
}

static pj_bool_t session_search_audio_port_with_image_audio(struct ast_sip_session *session, int image_audio, int is_t38_enable, pjmedia_sdp_session *receive_sdp, const char *media_type_name)
{
	if (!image_audio || !is_t38_enable) {
		return PJ_FALSE;
	}

	int i = 0;
	char media[20] = {0};
	for (; i < receive_sdp->media_count; ++i) {
		if (pj_strcmp2(&receive_sdp->media[i]->desc.media, media_type_name)) {
			continue;
		}

		RAII_VAR(struct ast_sip_session_media *, session_media, NULL, ao2_cleanup);
		session_media = ao2_find(session->media, media, OBJ_KEY);
		if (!session_media) {
			break;;
		}
	}

	ast_log(LOG_VERBOSE, "receive_sdp->media[i]->desc.port[%d]", receive_sdp->media[i]->desc.port);
	receive_sdp->media[i]->desc.port = 0;

	return PJ_TRUE;
}
#endif

static pj_bool_t session_reinvite_on_rx_request(pjsip_rx_data *rdata)
{
	pjsip_dialog *dlg;
	RAII_VAR(struct ast_sip_session *, session, NULL, ao2_cleanup);
	pjsip_rdata_sdp_info *sdp_info;
#ifdef GRANDSTREAM_NETWORKS
	ast_debug(1, "session_reinvite_on_rx_request enter...\n");
#endif
	if (rdata->msg_info.msg->line.req.method.id != PJSIP_INVITE_METHOD ||
		!(dlg = pjsip_ua_find_dialog(&rdata->msg_info.cid->id, &rdata->msg_info.to->tag, &rdata->msg_info.from->tag, PJ_FALSE)) ||
		!(session = ast_sip_dialog_get_session(dlg)) ||
		!session->channel) {
		return PJ_FALSE;
	}
#ifdef GRANDSTREAM_NETWORKS  /*By vanevan, Grandstream Networks*/
	if (session->inv_session != NULL && session->inv_session->state != PJSIP_INV_STATE_CONNECTING && 
			session->inv_session->state != PJSIP_INV_STATE_CONFIRMED) {
		pjsip_tx_data *tdata = NULL;

		ast_log(LOG_WARNING, "Receive a reinvite while the state is not connecting or confirmed, return 491!");
		if (pjsip_dlg_create_response(dlg, rdata, 491, NULL, &tdata) == PJ_SUCCESS) {
			pjsip_endpt_send_response2(ast_sip_get_pjsip_endpoint(), rdata, tdata, NULL, NULL);
		}

		return PJ_TRUE;
	}
#endif

	if (session->deferred_reinvite) {
		pj_str_t key, deferred_key;
		pjsip_tx_data *tdata;

		/* We use memory from the new request on purpose so the deferred reinvite pool does not grow uncontrollably */
		pjsip_tsx_create_key(rdata->tp_info.pool, &key, PJSIP_ROLE_UAS, &rdata->msg_info.cseq->method, rdata);
		pjsip_tsx_create_key(rdata->tp_info.pool, &deferred_key, PJSIP_ROLE_UAS, &session->deferred_reinvite->msg_info.cseq->method,
			session->deferred_reinvite);

		/* If this is a retransmission ignore it */
		if (!pj_strcmp(&key, &deferred_key)) {
			return PJ_TRUE;
		}

		/* Otherwise this is a new re-invite, so reject it */
		if (pjsip_dlg_create_response(dlg, rdata, 491, NULL, &tdata) == PJ_SUCCESS) {
			pjsip_endpt_send_response2(ast_sip_get_pjsip_endpoint(), rdata, tdata, NULL, NULL);
		}

		return PJ_TRUE;
	}

	if (!(sdp_info = pjsip_rdata_get_sdp_info(rdata)) ||
		(sdp_info->sdp_err != PJ_SUCCESS)) {
		return PJ_FALSE;
	}

	if (!sdp_info->sdp) {
		ast_queue_unhold(session->channel);
		return PJ_FALSE;
	}
	
#ifdef GRANDSTREAM_NETWORKS
	int session_media_image = 0, session_media_audio = 0;
	session_media_image = session_check_sdp_media_type(session, sdp_info->sdp, "image");
	session_media_audio = session_check_sdp_media_type(session, sdp_info->sdp, "audio");
	if (session_media_image && session_media_audio) {
		session->is_image_audio_two_media = 1;
	} else {
		session->is_image_audio_two_media = 0;
	}

	unsigned short intelligent_route_flag = 0;
	const char *callee_num = NULL;
	const char *from_trunk = NULL;

	if (session->is_intelligent_route) {
		intelligent_route_flag = 1;
		ast_channel_set_is_sdp_image_reinvite(session->channel, 1); //Added by pxie@grandstream.cn for bug[46127], 2015-07-20.
	}

	callee_num = pbx_builtin_getvar_helper(session->channel, "EXTTOCALL");
	from_trunk = pbx_builtin_getvar_helper(session->channel, "FROM_TRUNK");
	if (callee_num && session->endpoint->fax_intelligent_route_destination && !strcasecmp(callee_num, session->endpoint->fax_intelligent_route_destination)) {
		intelligent_route_flag = 1;
		ast_log(LOG_VERBOSE, "The same purpose(destination[%s]) and called extension(callee[%s])", session->endpoint->fax_intelligent_route_destination, callee_num);
	}
	if (!intelligent_route_flag && from_trunk && !strcasecmp(from_trunk, "yes") && session->endpoint->fax_intelligent_route_destination && !strcasecmp(session->endpoint->fax_intelligent_route, "yes") && session_media_image) {
		if (!strcasecmp(session->endpoint->faxcontext, "faxcontext")) {
			ast_verb(2, "Redirecting '%s' to fax extension due to T.38 Re-Invite 'image'", ast_channel_name(session->channel));
			pbx_builtin_setvar_helper(session->channel, "FAXEXTEN", ast_channel_exten(session->channel));
			/*! Begin: Fax intelligent route*/
			if (!strcasecmp(session->endpoint->fax_intelligent_route, "yes") && strcasecmp(session->endpoint->fax_intelligent_route_destination, "no")) {
				pbx_builtin_setvar_helper(session->channel, "_FAX_INTELLIGENT_ROUTE", "yes");
				pbx_builtin_setvar_helper(session->channel, "_FAX_INTELLIGENT_ROUTE_DETINATION", session->endpoint->fax_intelligent_route_destination);
				session->is_intelligent_route = 1;
				ast_channel_set_is_sdp_image_reinvite(session->channel, 1); //Added by pxie@grandstream.cn for bug[46127], 2015-07-20.
			}
			/*! End: Fax intelligent route*/
			if (ast_async_goto(session->channel, session->endpoint->faxcontext, "fax", 1)) {
				ast_log(LOG_NOTICE, "Failed to async goto '%s' into fax of '%s'\n", ast_channel_name(session->channel), session->endpoint->faxcontext);
			}
			return PJ_TRUE;
		}
	}

	/*! Begin: Added by pxie@grandstream.cn for bug[46209, 46202], 2015-07-17.*/
	if (session && session->channel && !ast_channel_get_vfax_flag(session->channel)) {
		struct ast_bridge_channel *peer_bridge_channel = ast_channel_get_bridge_channel(session->channel);
		struct ast_channel *peer_chan = ast_bridge_channel_get_channel_from_bridge_channel(peer_bridge_channel);
		if (!peer_chan && session_media_image && session->is_intelligent_route) {
			ao2_ref(peer_bridge_channel, -1);
			return PJ_TRUE;
		}

		if (session_media_image && !peer_chan) {
			ast_log(LOG_VERBOSE, "Peer channel don't bridge success, wait bridge finished!\n");
			ao2_ref(peer_bridge_channel, -1);
			return PJ_TRUE;
		}
		ao2_ref(peer_bridge_channel, -1);
	}
	/*! End: Added by pxie@grandstream.cn for bug[46209, 46202], 2015-07-17.*/
#endif

	if (!sdp_requires_deferral(session, sdp_info->sdp)) {
		return PJ_FALSE;
	}

	pjsip_rx_data_clone(rdata, 0, &session->deferred_reinvite);

	return PJ_TRUE;
}

void ast_sip_session_resume_reinvite(struct ast_sip_session *session)
{
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "starting sip session resume reinvite ...\n");
#endif
	if (!session->deferred_reinvite) {
		return;
	}

	pjsip_endpt_process_rx_data(ast_sip_get_pjsip_endpoint(), session->deferred_reinvite, NULL, NULL);
	pjsip_rx_data_free_cloned(session->deferred_reinvite);
	session->deferred_reinvite = NULL;
}

static pjsip_module session_reinvite_module = {
	.name = { "Session Re-Invite Module", 24 },
	.priority = PJSIP_MOD_PRIORITY_UA_PROXY_LAYER - 1,
	.on_rx_request = session_reinvite_on_rx_request,
};

void ast_sip_session_send_request_with_cb(struct ast_sip_session *session, pjsip_tx_data *tdata,
		ast_sip_session_response_cb on_response)
{
	pjsip_inv_session *inv_session = session->inv_session;
#ifdef GRANDSTREAM_NETWORKS
	const char *cid_num = pbx_builtin_getvar_helper(session->channel, "CID_NUM");
	const char *conf_room = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");
#endif

	if (inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		/* Don't try to do anything with a hung-up call */
		return;
	}

	ast_sip_mod_data_set(tdata->pool, tdata->mod_data, session_module.id,
			     MOD_DATA_ON_RESPONSE, on_response);

#ifdef GRANDSTREAM_NETWORKS
	if ((tdata->msg->type == PJSIP_REQUEST_MSG && 
		 tdata->msg->line.req.method.id == PJSIP_INVITE_METHOD) &&
		(!ast_strlen_zero(session->endpoint->fromuser) ||
		!ast_strlen_zero(session->endpoint->fromdomain))) {
#else
	if (!ast_strlen_zero(session->endpoint->fromuser) ||
		!ast_strlen_zero(session->endpoint->fromdomain)) {
#endif
		pjsip_fromto_hdr *from = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_FROM, tdata->msg->hdr.next);

#ifdef GRANDSTREAM_NETWORKS
		if (PJSIP_URI_SCHEME_IS_SIP(from->uri) || PJSIP_URI_SCHEME_IS_SIPS(from->uri)) {
#endif
		pjsip_sip_uri *uri = pjsip_uri_get_uri(from->uri);

#ifdef GRANDSTREAM_NETWORKS
		// Assign From header information
		if (!ast_strlen_zero(conf_room) && !ast_strlen_zero(cid_num)) {
			pj_strdup2(tdata->pool, &uri->user, cid_num);
		}
		else if (!ast_strlen_zero(session->endpoint->fromuser)) {
			pj_strdup2(tdata->pool, &uri->user, session->endpoint->fromuser);
		}
#else
		if (!ast_strlen_zero(session->endpoint->fromuser)) {
			pj_strdup2(tdata->pool, &uri->user, session->endpoint->fromuser);
		}
#endif
		if (!ast_strlen_zero(session->endpoint->fromdomain)) {
			pj_strdup2(tdata->pool, &uri->host, session->endpoint->fromdomain);
		}
#ifdef GRANDSTREAM_NETWORKS
		}
#endif
	}


	handle_outgoing_request(session, tdata);
	pjsip_inv_send_msg(session->inv_session, tdata);
	return;
}

void ast_sip_session_send_request(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	ast_sip_session_send_request_with_cb(session, tdata, NULL);
}

int ast_sip_session_create_invite(struct ast_sip_session *session, pjsip_tx_data **tdata)
{
	pjmedia_sdp_session *offer;
#ifdef GRANDSTREAM_NETWORKS
	int  cc_core_id;
    ast_log(LOG_VERBOSE, "sip session create local sdp ...\n");
#endif

	if (!(offer = create_local_sdp(session->inv_session, session, NULL))) {
		pjsip_inv_terminate(session->inv_session, 500, PJ_FALSE);
		return -1;
	}
	
#ifdef GRANDSTREAM_NETWORKS
	{
		char buf[1024] = {0};

		pjmedia_sdp_print(offer, buf, sizeof(buf) - 1);
		ast_log(LOG_CODEC_NEG, "[callee] UCM ===>>> B (the callee). send invite sdp! \n%s", buf);
	}
#endif

	pjsip_inv_set_local_sdp(session->inv_session, offer);
	pjmedia_sdp_neg_set_prefer_remote_codec_order(session->inv_session->neg, PJ_FALSE);
#ifdef PJMEDIA_SDP_NEG_ANSWER_MULTIPLE_CODECS
	pjmedia_sdp_neg_set_answer_multiple_codecs(session->inv_session->neg, PJ_TRUE);
#endif

#ifdef GRANDSTREAM_NETWORKS   /* Added by yfyang, for ccss feature, get notify user to set as target user */
	if (ast_cc_is_recall(session->channel, &cc_core_id, "PJSIP")) {
		char   device_name[AST_CHANNEL_NAME] = {0};
		struct ast_cc_monitor *recall_monitor;

		ast_channel_get_device_name(session->channel, device_name, sizeof(device_name));
		if ((recall_monitor = ast_cc_get_monitor_by_recall_core_id(cc_core_id, device_name))) {
			struct ast_sip_cc_notify_user_data  notify_user_data = {
				.monitor = recall_monitor,
			};

			if (0 == find_sub_handler_for_supplement_exten(AST_SIP_CCSS_NOTIFY_USER, NULL, &notify_user_data)) {
				if (session->inv_session && session->inv_session->dlg) {
					pjsip_dialog   *dlg = session->inv_session->dlg;
					pjsip_sip_uri  *sip_uri;

                    ast_log(LOG_NOTICE, "ast_sip_session_create_invite, get notify user[%s].\n", notify_user_data.notify_user);
					if (PJSIP_URI_SCHEME_IS_SIP(dlg->target) || PJSIP_URI_SCHEME_IS_SIPS(dlg->target)) {
						sip_uri = pjsip_uri_get_uri(dlg->target);
						pj_strdup2(dlg->pool, &sip_uri->user, notify_user_data.notify_user);
					}

					if (PJSIP_URI_SCHEME_IS_SIP(dlg->remote.info->uri) || PJSIP_URI_SCHEME_IS_SIPS(dlg->remote.info->uri)) {
						sip_uri = pjsip_uri_get_uri(dlg->remote.info->uri);
						pj_strdup2(dlg->pool, &sip_uri->user, notify_user_data.notify_user);
					}
				}
			}

			ao2_t_ref(recall_monitor, -1, "Got the URI we need so unreffing monitor");
		}
	}
#endif

	if (pjsip_inv_invite(session->inv_session, tdata) != PJ_SUCCESS) {
		return -1;
	}
	return 0;
}

static int datastore_hash(const void *obj, int flags)
{
	const struct ast_datastore *datastore = obj;
	const char *uid = flags & OBJ_KEY ? obj : datastore->uid;

	ast_assert(uid != NULL);

	return ast_str_hash(uid);
}

static int datastore_cmp(void *obj, void *arg, int flags)
{
	const struct ast_datastore *datastore1 = obj;
	const struct ast_datastore *datastore2 = arg;
	const char *uid2 = flags & OBJ_KEY ? arg : datastore2->uid;

	ast_assert(datastore1->uid != NULL);
	ast_assert(uid2 != NULL);

	return strcmp(datastore1->uid, uid2) ? 0 : CMP_MATCH | CMP_STOP;
}

static void session_media_dtor(void *obj)
{
	struct ast_sip_session_media *session_media = obj;
	struct sdp_handler_list *handler_list;
	/* It is possible for SDP handlers to allocate memory on a session_media but
	 * not end up getting set as the handler for this session_media. This traversal
	 * ensures that all memory allocated by SDP handlers on the session_media is
	 * cleared (as well as file descriptors, etc.).
	 */
	handler_list = ao2_find(sdp_handlers, session_media->stream_type, OBJ_KEY);
	if (handler_list) {
		struct ast_sip_session_sdp_handler *handler;

		AST_LIST_TRAVERSE(&handler_list->list, handler, next) {
			handler->stream_destroy(session_media);
		}
	}
	ao2_cleanup(handler_list);
	if (session_media->srtp) {
		ast_sdp_srtp_destroy(session_media->srtp);
	}
}

static void session_destructor(void *obj)
{
	struct ast_sip_session *session = obj;
	struct ast_sip_session_supplement *supplement;
	struct ast_sip_session_delayed_request *delay;

	ast_debug(3, "Destroying SIP session with endpoint %s(%p)\n",
			ast_sorcery_object_get_id(session->endpoint), session);

	while ((supplement = AST_LIST_REMOVE_HEAD(&session->supplements, next))) {
		if (supplement->session_destroy) {
			supplement->session_destroy(session);
		}
		ast_free(supplement);
	}

	ast_taskprocessor_unreference(session->serializer);
	ao2_cleanup(session->datastores);
	ao2_cleanup(session->media);

	AST_LIST_HEAD_DESTROY(&session->supplements);
	while ((delay = AST_LIST_REMOVE_HEAD(&session->delayed_requests, next))) {
		ast_free(delay);
	}
	ast_party_id_free(&session->id);
	ao2_cleanup(session->endpoint);
	ao2_cleanup(session->aor);
	ao2_cleanup(session->contact);
	ao2_cleanup(session->req_caps);
	ao2_cleanup(session->direct_media_cap);
#ifdef GRANDSTREAM_NETWORKS
	if (session->contact_user) {
		ast_free(session->contact_user);
		session->contact_user = NULL;
	}

    /* Add by bfdu for BFCP, 2014.07.14 */
	ao2_cleanup(session->slides_video_req_caps);
	ao2_cleanup(session->fec_req_caps);
	ao2_cleanup(session->fec_multiplexed_req_caps);
	ao2_cleanup(session->fecc_req_caps);
	ao2_cleanup(session->final_caps);
	if (session->slides_video_media.rtp) {
		ast_rtp_instance_stop(session->slides_video_media.rtp);
		ast_rtp_instance_destroy(session->slides_video_media.rtp);
		session->slides_video_media.rtp = NULL;
	}
	if (session->slides_video_media.srtp) {
		ast_sdp_srtp_destroy(session->slides_video_media.srtp);
		session->slides_video_media.srtp = NULL;
	}
	if (session->fec_media.rtp) {
		ast_rtp_instance_stop(session->fec_media.rtp);
		ast_rtp_instance_destroy(session->fec_media.rtp);
		session->fec_media.rtp = NULL;
	}
	if (session->fec_media.srtp) {
		ast_sdp_srtp_destroy(session->fec_media.srtp);
		session->fec_media.srtp = NULL;
	}
	if (session->fecc_media.rtp) {
		ast_rtp_instance_stop(session->fecc_media.rtp);
		ast_rtp_instance_destroy(session->fecc_media.rtp);
		session->fecc_media.rtp = NULL;
	}
	if (session->fecc_media.srtp) {
		ast_sdp_srtp_destroy(session->fecc_media.srtp);
		session->fecc_media.srtp = NULL;
	}
#endif

	if (session->dsp) {
		ast_dsp_free(session->dsp);
	}

	if (session->inv_session) {
#ifdef GRANDSTREAM_NETWORKS
		session->inv_session->mod_data[session_module.id] = NULL;
#endif
		pjsip_dlg_dec_session(session->inv_session->dlg, &session_module);
	}
}

static int add_supplements(struct ast_sip_session *session)
{
	struct ast_sip_session_supplement *iter;
	SCOPED_LOCK(lock, &session_supplements, AST_RWLIST_RDLOCK, AST_RWLIST_UNLOCK);

	AST_RWLIST_TRAVERSE(&session_supplements, iter, next) {
		struct ast_sip_session_supplement *copy = supplement_dup(iter);
		if (!copy) {
			return -1;
		}
		AST_LIST_INSERT_TAIL(&session->supplements, copy, next);
	}
	return 0;
}

static int add_session_media(void *obj, void *arg, int flags)
{
	struct sdp_handler_list *handler_list = obj;
	struct ast_sip_session * session = arg;
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "Add media for sip session ...\n");
#endif
	RAII_VAR(struct ast_sip_session_media *, session_media, NULL, ao2_cleanup);
	session_media = ao2_alloc(sizeof(*session_media) + strlen(handler_list->stream_type), session_media_dtor);
	if (!session_media) {
		return CMP_STOP;
	}
	session_media->encryption = session->endpoint->media.rtp.encryption;
#ifdef GRANDSTREAM_NETWORKS
	session_media->keepalive_sched_id = -1;
	session_media->timeout_sched_id = -1;
#endif
	/* Safe use of strcpy */
	strcpy(session_media->stream_type, handler_list->stream_type);
	ao2_link(session->media, session_media);
	return 0;
}

/*! \brief Destructor for SIP channel */
static void sip_channel_destroy(void *obj)
{
	struct ast_sip_channel_pvt *channel = obj;

	ao2_cleanup(channel->pvt);
	ao2_cleanup(channel->session);	
#ifdef GRANDSTREAM_NETWORKS
	ao2_cleanup(channel->req_caps_of_the_other_leg);
#endif
}

#ifdef GRANDSTREAM_NETWORKS
struct ast_sip_channel_pvt *ast_sip_channel_pvt_alloc(void *pvt, struct ast_sip_session *session, struct ast_format_cap * req_caps_requestor)
#else
struct ast_sip_channel_pvt *ast_sip_channel_pvt_alloc(void *pvt, struct ast_sip_session *session)
#endif
{
	struct ast_sip_channel_pvt *channel = ao2_alloc(sizeof(*channel), sip_channel_destroy);

	if (!channel) {
		return NULL;
	}

	ao2_ref(pvt, +1);
	channel->pvt = pvt;
	ao2_ref(session, +1);
	channel->session = session;	
#ifdef GRANDSTREAM_NETWORKS
	if (req_caps_requestor) {
		ao2_ref(req_caps_requestor, +1);
		channel->req_caps_of_the_other_leg = req_caps_requestor;
	}
#endif

	return channel;
}

#ifdef GRANDSTREAM_NETWORKS
static void ast_reinvite_attr_init(struct ast_sip_session *session)
{
	session->reinvite_attr.reinvite_attr_audio = REINVITE_SENDRECV;
	session->reinvite_attr.reinvite_attr_m_video = REINVITE_SENDRECV;
	session->reinvite_attr.reinvite_attr_s_video = REINVITE_SENDRECV;
	session->reinvite_attr.reinvite_attr_fec = REINVITE_SENDRECV;
	session->reinvite_attr.reinvite_attr_fecc = REINVITE_SENDRECV;
	session->reinvite_attr.reinvite_attr_bfcp = REINVITE_SENDRECV;
}
#endif

struct ast_sip_session *ast_sip_session_alloc(struct ast_sip_endpoint *endpoint,
	struct ast_sip_contact *contact, pjsip_inv_session *inv_session)
{
	RAII_VAR(struct ast_sip_session *, session, ao2_alloc(sizeof(*session), session_destructor), ao2_cleanup);
	struct ast_sip_session_supplement *iter;
	int dsp_features = 0;
	if (!session) {
		return NULL;
	}
	AST_LIST_HEAD_INIT(&session->supplements);
	session->datastores = ao2_container_alloc(DATASTORE_BUCKETS, datastore_hash, datastore_cmp);
	if (!session->datastores) {
		return NULL;
	}

	session->endpoint = ao2_bump(endpoint);

	session->media = ao2_container_alloc(MEDIA_BUCKETS, session_media_hash, session_media_cmp);
	if (!session->media) {
		return NULL;
	}
	/* fill session->media with available types */
	ao2_callback(sdp_handlers, OBJ_NODATA, add_session_media, session);

	session->serializer = ast_sip_create_serializer();
	if (!session->serializer) {
		return NULL;
	}
	ast_sip_dialog_set_serializer(inv_session->dlg, session->serializer);
	ast_sip_dialog_set_endpoint(inv_session->dlg, endpoint);
	pjsip_dlg_inc_session(inv_session->dlg, &session_module);
	inv_session->mod_data[session_module.id] = ao2_bump(session);
	session->contact = ao2_bump(contact);
	session->inv_session = inv_session;
	session->req_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);

#ifdef GRANDSTREAM_NETWORKS
	session->core_id = -1;
	session->cc_offer = 0;
	session->cc_service = AST_CC_NONE;
	session->cc_subs_user.slen = 0;
	session->cc_subs_user.ptr = NULL;
	session->cc_subs_uri = NULL;
	session->contact_user = NULL;
	session->is_outgoing = 0;
	/* Begin add by bfdu for video reinivite/BFCP, 2015.05.27 */
	session->is_video_media_change = 0;
	memset(&session->bfcp_para, 0, sizeof(struct ast_bfcp_para));
	memset(&session->main_video_para, 0, sizeof(struct ast_main_video_para));
	memset(&session->slides_video_para, 0, sizeof(struct ast_slides_video_para));
	session->slides_video_req_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	memset(&session->slides_video_media, 0, sizeof(struct ast_sip_extra_media));
	session->slides_video_media.encryption = session->endpoint->media.rtp.encryption;
	memset(&session->fec_para, 0, sizeof(struct ast_fec_para));
	memset(&session->fecc_para, 0, sizeof(struct ast_fecc_para));
	memset(&session->fec_media, 0, sizeof(struct ast_sip_extra_media));
	session->fec_media.encryption = AST_SIP_MEDIA_ENCRYPT_NONE; /* FEC stream default encrypt is none */
	memset(&session->fecc_media, 0, sizeof(struct ast_sip_extra_media));
	session->fecc_media.encryption = session->endpoint->media.rtp.encryption;
	session->fec_req_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	session->fec_multiplexed_req_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	session->fecc_req_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	/* End add by bfdu for video reinivite/BFCP, 2015.05.27 */
	memset(&session->reinvite_attr, 0, sizeof(struct ast_reinvite_attrs));
	ast_reinvite_attr_init(session);
	session->is_sendonly_to_peer = 0;
	session->final_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	session->has_early_media = 0;
#endif

	if ((endpoint->dtmf == AST_SIP_DTMF_INBAND) || (endpoint->dtmf == AST_SIP_DTMF_AUTO)) {
		dsp_features |= DSP_FEATURE_DIGIT_DETECT;
	}

	if (endpoint->faxdetect) {
		dsp_features |= DSP_FEATURE_FAX_DETECT;
#ifdef GRANDSTREAM_NETWORKS
		/*! Added by pxie@grandstream.cn for bug[42104], 2015-04-14*/
		session->wFaxDetect = 1;
		/*! Added by pxie@grandstream.cn for bug[42104], 2015-04-14*/
#endif
	}
#ifdef GRANDSTREAM_NETWORKS
	/*! Added by pxie@grandstream.cn for bug[42104], 2015-04-14*/
	else {
		session->wFaxDetect = 0;
	}
	/*! Added by pxie@grandstream.cn for bug[42104], 2015-04-14*/
#endif
	if (dsp_features) {
		if (!(session->dsp = ast_dsp_new())) {
/** The ref count of session will be reduced when exit the functin,
**  so there is no need to do it manually
**  Remove by vanevan, Grandstream Networks
**/
#ifndef GRANDSTREAM_NETWORKS
			ao2_ref(session, -1);
#endif
			return NULL;
		}

		ast_dsp_set_features(session->dsp, dsp_features);
	}

	if (add_supplements(session)) {
/** The ref count of session will be reduced when exit the functin,
**  so there is no need to do it manually
**  Remove by vanevan, Grandstream Networks
**/
#ifndef GRANDSTREAM_NETWORKS
		ao2_ref(session, -1);
#endif
		return NULL;
	}
	AST_LIST_TRAVERSE(&session->supplements, iter, next) {
		if (iter->session_begin) {
			iter->session_begin(session);
		}
	}
	session->direct_media_cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	AST_LIST_HEAD_INIT_NOLOCK(&session->delayed_requests);
	ast_party_id_init(&session->id);
	ao2_ref(session, +1);
	return session;
}

/*! \brief struct controlling the suspension of the session's serializer. */
struct ast_sip_session_suspender {
	ast_cond_t cond_suspended;
	ast_cond_t cond_complete;
	int suspended;
	int complete;
};

static void sip_session_suspender_dtor(void *vdoomed)
{
	struct ast_sip_session_suspender *doomed = vdoomed;

	ast_cond_destroy(&doomed->cond_suspended);
	ast_cond_destroy(&doomed->cond_complete);
}

/*!
 * \internal
 * \brief Block the session serializer thread task.
 *
 * \param data Pushed serializer task data for suspension.
 *
 * \retval 0
 */
static int sip_session_suspend_task(void *data)
{
	struct ast_sip_session_suspender *suspender = data;

	ao2_lock(suspender);

	/* Signal that the serializer task is now suspended. */
	suspender->suspended = 1;
	ast_cond_signal(&suspender->cond_suspended);

	/* Wait for the the serializer suspension to be completed. */
	while (!suspender->complete) {
		ast_cond_wait(&suspender->cond_complete, ao2_object_get_lockaddr(suspender));
	}

	ao2_unlock(suspender);
	ao2_ref(suspender, -1);

	return 0;
}

void ast_sip_session_suspend(struct ast_sip_session *session)
{
	struct ast_sip_session_suspender *suspender;
	int res;

	ast_assert(session->suspended == NULL);

	if (ast_taskprocessor_is_task(session->serializer)) {
		/* I am the session's serializer thread so I cannot suspend. */
		return;
	}

	suspender = ao2_alloc(sizeof(*suspender), sip_session_suspender_dtor);
	if (!suspender) {
		/* We will just have to hope that the system does not deadlock */
		return;
	}
	ast_cond_init(&suspender->cond_suspended, NULL);
	ast_cond_init(&suspender->cond_complete, NULL);

	ao2_ref(suspender, +1);
	res = ast_sip_push_task(session->serializer, sip_session_suspend_task, suspender);
	if (res) {
		/* We will just have to hope that the system does not deadlock */
		ao2_ref(suspender, -2);
		return;
	}

	session->suspended = suspender;

	/* Wait for the serializer to get suspended. */
	ao2_lock(suspender);
	while (!suspender->suspended) {
		ast_cond_wait(&suspender->cond_suspended, ao2_object_get_lockaddr(suspender));
	}
	ao2_unlock(suspender);
}

void ast_sip_session_unsuspend(struct ast_sip_session *session)
{
	struct ast_sip_session_suspender *suspender = session->suspended;

	if (!suspender) {
		/* Nothing to do */
		return;
	}
	session->suspended = NULL;

	/* Signal that the serializer task suspension is now complete. */
	ao2_lock(suspender);
	suspender->complete = 1;
	ast_cond_signal(&suspender->cond_complete);
	ao2_unlock(suspender);

	ao2_ref(suspender, -1);
}

static int session_outbound_auth(pjsip_dialog *dlg, pjsip_tx_data *tdata, void *user_data)
{
	pjsip_inv_session *inv = pjsip_dlg_get_inv_session(dlg);
	struct ast_sip_session *session = inv->mod_data[session_module.id];

	if (inv->state < PJSIP_INV_STATE_CONFIRMED && tdata->msg->line.req.method.id == PJSIP_INVITE_METHOD) {
		pjsip_inv_uac_restart(inv, PJ_FALSE);
	}
	ast_sip_session_send_request(session, tdata);
	return 0;
}

struct ast_sip_session *ast_sip_session_create_outgoing(struct ast_sip_endpoint *endpoint,
	struct ast_sip_contact *contact, const char *location, const char *request_user,
	struct ast_format_cap *req_caps)
{
	const char *uri = NULL;
	RAII_VAR(struct ast_sip_aor *, found_aor, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_contact *, found_contact, NULL, ao2_cleanup);
	pjsip_timer_setting timer;
	pjsip_dialog *dlg;
	struct pjsip_inv_session *inv_session;
	RAII_VAR(struct ast_sip_session *, session, NULL, ao2_cleanup);	
#ifdef GRANDSTREAM_NETWORKS
	struct ast_str *format_buf = ast_str_alloca(64);
	static pj_str_t STR_100REL = {"100rel", 6};
	static pj_str_t STR_TIMER = {"timer", 5};
#endif

	/* If no location has been provided use the AOR list from the endpoint itself */
	if (location || !contact) {
		location = S_OR(location, endpoint->aors);

		ast_sip_location_retrieve_contact_and_aor_from_list(location, &found_aor, &found_contact);
		if (!found_contact || ast_strlen_zero(found_contact->uri)) {
			uri = location;
		} else {
			uri = found_contact->uri;
		}
	} else {
		uri = contact->uri;
	}

	/* If we still have no URI to dial fail to create the session */
	if (ast_strlen_zero(uri)) {
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (!(dlg = ast_sip_create_dialog_uac(endpoint, uri, request_user, PJSIP_INVITE_METHOD))) {
		return NULL;
	}
#else
	if (!(dlg = ast_sip_create_dialog_uac(endpoint, uri, request_user))) {
		return NULL;
	}
#endif

	if (ast_sip_dialog_setup_outbound_authentication(dlg, endpoint, session_outbound_auth, NULL)) {
		pjsip_dlg_terminate(dlg);
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
/* add by chdruan, Bug 41991 - 100rel:ucm should not support 100rel when setting 100rel "no"*/
	if (!(endpoint->extensions.flags & PJSIP_INV_SUPPORT_100REL
			|| endpoint->extensions.flags & PJSIP_INV_REQUIRE_100REL)) {
		pjsip_endpt_remove_capability(ast_sip_get_pjsip_endpoint(), PJSIP_H_SUPPORTED, NULL, &STR_100REL);
	}

	if (!(endpoint->extensions.flags & PJSIP_INV_SUPPORT_TIMER
			|| endpoint->extensions.flags & PJSIP_INV_REQUIRE_TIMER
			|| endpoint->extensions.flags & PJSIP_INV_ALWAYS_USE_TIMER)) {
		pjsip_endpt_remove_capability(ast_sip_get_pjsip_endpoint(), PJSIP_H_SUPPORTED, NULL, &STR_TIMER);
	}

#endif

	if (pjsip_inv_create_uac(dlg, NULL, endpoint->extensions.flags, &inv_session) != PJ_SUCCESS) {
		pjsip_dlg_terminate(dlg);
		return NULL;
	}
#if defined(HAVE_PJSIP_REPLACE_MEDIA_STREAM) || defined(PJMEDIA_SDP_NEG_ALLOW_MEDIA_CHANGE)
	inv_session->sdp_neg_flags = PJMEDIA_SDP_NEG_ALLOW_MEDIA_CHANGE;
#endif

	pjsip_timer_setting_default(&timer);
	timer.min_se = endpoint->extensions.timer.min_se;
	timer.sess_expires = endpoint->extensions.timer.sess_expires;
	pjsip_timer_init_session(inv_session, &timer);

	if (!(session = ast_sip_session_alloc(endpoint, found_contact ? found_contact : contact, inv_session))) {
		pjsip_inv_terminate(inv_session, 500, PJ_FALSE);
		return NULL;
	}
	
#ifdef GRANDSTREAM_NETWORKS
	session->is_outgoing = 1;	
#endif
	session->aor = ao2_bump(found_aor);
	ast_party_id_copy(&session->id, &endpoint->id.self);

	if (ast_format_cap_count(req_caps)) {
		/* get joint caps between req_caps and endpoint caps */
		struct ast_format_cap *joint_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
		ast_format_cap_get_compatible(req_caps, session->endpoint->media.codecs, joint_caps);
		
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_CODEC_NEG, "[callee] req_caps inherited from the caller: %s\n",			 
			ast_format_cap_get_names(req_caps, &format_buf));		
		ast_log(LOG_CODEC_NEG, "[callee] endpoint codecs: %s\n",			 
			ast_format_cap_get_names(session->endpoint->media.codecs, &format_buf));
#endif

		/* if joint caps */
		if (ast_format_cap_count(joint_caps)) {            
#ifdef GRANDSTREAM_NETWORKS
			ast_format_cap_append_from_cap(session->req_caps, joint_caps, AST_MEDIA_TYPE_UNKNOWN);
			if (!ast_format_cap_has_type(joint_caps, AST_MEDIA_TYPE_AUDIO)) {				
				ast_format_cap_append_from_cap(session->req_caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_AUDIO);
			}
			
			ast_log(LOG_CODEC_NEG, "[callee] joint codecs(endpoint & req_caps): %s\n",			 
				ast_format_cap_get_names(joint_caps, &format_buf));			
			ast_log(LOG_CODEC_NEG, "[callee] final req_caps of the callee: %s\n",			 
				ast_format_cap_get_names(session->req_caps, &format_buf));
#else
			/* copy endpoint caps into session->req_caps */
			ast_format_cap_append_from_cap(session->req_caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_UNKNOWN);
#endif
			/* replace instances of joint caps equivalents in session->req_caps */
			ast_format_cap_replace_from_cap(session->req_caps, joint_caps, AST_MEDIA_TYPE_UNKNOWN);
		}
		ao2_cleanup(joint_caps);
	}

	if ((pjsip_dlg_add_usage(dlg, &session_module, NULL) != PJ_SUCCESS)) {
		pjsip_inv_terminate(inv_session, 500, PJ_FALSE);
		/* Since we are not notifying ourselves that the INVITE session is being terminated
		 * we need to manually drop its reference to session
		 */
		ao2_ref(session, -1);
		return NULL;
	}

	ao2_ref(session, +1);
	return session;
}

void ast_sip_session_terminate(struct ast_sip_session *session, int response)
{
	pj_status_t status;
	pjsip_tx_data *packet = NULL;

#ifdef GRANDSTREAM_NETWORKS
	if (session == NULL || session->inv_session == NULL || session->inv_session->dlg == NULL) {
		ast_log(LOG_ERROR, "SIP session is invalid or PJSIP invite session/dialog is NULL!");
		return;
	}
	/**Added by vanevan, Grandstream Networks
	**
	** There is another thread which is running PJSIP handling, it will get pjsip dialog lock first, and then call session_inv_on_tsx_state_changed to get ast_sip_session object lock,
	** if we try to get ast_sip_session object lock, and then in pjsip_inv_end_session function will try to get pjsip dialog lock, which may cause deadlock.
	**
	** to avoid deadlock, we must make sure the same lock sequence, so try to get pjsip dialog lock first and then get ast_sip_session object lock.
	**/
	pjsip_dlg_inc_lock(session->inv_session->dlg);
	ao2_lock(session);
#endif

	if (session->defer_terminate) {
		session->terminate_while_deferred = 1;
#ifdef GRANDSTREAM_NETWORKS
		ao2_unlock(session);
		pjsip_dlg_dec_lock(session->inv_session->dlg);
#endif
		return;
	}

	if (!response) {
		response = 603;
	}

	if ((session->inv_session->state == PJSIP_INV_STATE_CONFIRMED) && session->inv_session->invite_tsx) {
		ast_debug(3, "Delay sending BYE to %s because of outstanding transaction...\n",
				ast_sorcery_object_get_id(session->endpoint));
		/* If this is delayed the only thing that will happen is a BYE request so we don't
		 * actually need to store the response code for when it happens.
		 */
		delay_request(session, NULL, NULL, NULL, 0, DELAYED_METHOD_BYE);
	} else if (session->inv_session->state == PJSIP_INV_STATE_NULL) {
		pjsip_inv_terminate(session->inv_session, response, PJ_TRUE);
	} else if (((status = pjsip_inv_end_session(session->inv_session, response, NULL, &packet)) == PJ_SUCCESS)
		&& packet) {
		struct ast_sip_session_delayed_request *delay;

		/* Flush any delayed requests so they cannot overlap this transaction. */
		while ((delay = AST_LIST_REMOVE_HEAD(&session->delayed_requests, next))) {
			ast_free(delay);
		}

		if (packet->msg->type == PJSIP_RESPONSE_MSG) {
			ast_sip_session_send_response(session, packet);
		} else {
			ast_sip_session_send_request(session, packet);
		}
	}
#ifdef GRANDSTREAM_NETWORKS
	ao2_unlock(session);
	pjsip_dlg_dec_lock(session->inv_session->dlg);
#endif
}

static int session_termination_task(void *data)
{
	struct ast_sip_session *session = data;

	if (session->defer_terminate) {
		session->defer_terminate = 0;
		if (session->inv_session) {
			ast_sip_session_terminate(session, 0);
		}
	}

	ao2_ref(session, -1);
	return 0;
}

static void session_termination_cb(pj_timer_heap_t *timer_heap, struct pj_timer_entry *entry)
{
	struct ast_sip_session *session = entry->user_data;

	if (ast_sip_push_task(session->serializer, session_termination_task, session)) {
		ao2_cleanup(session);
	}
}

int ast_sip_session_defer_termination(struct ast_sip_session *session)
{
	pj_time_val delay = { .sec = 60, };
	int res;

	/* The session should not have an active deferred termination request. */
	ast_assert(!session->defer_terminate);

	session->defer_terminate = 1;

	session->scheduled_termination.id = 0;
	ao2_ref(session, +1);
	session->scheduled_termination.user_data = session;
	session->scheduled_termination.cb = session_termination_cb;

	res = (pjsip_endpt_schedule_timer(ast_sip_get_pjsip_endpoint(),
		&session->scheduled_termination, &delay) != PJ_SUCCESS) ? -1 : 0;
	if (res) {
		session->defer_terminate = 0;
		ao2_ref(session, -1);
	}
	return res;
}

void ast_sip_session_defer_termination_cancel(struct ast_sip_session *session)
{
	if (!session->defer_terminate) {
		/* Already canceled or timer fired. */
		return;
	}
	session->defer_terminate = 0;

	if (session->terminate_while_deferred) {
		/* Complete the termination started by the upper layer. */
		ast_sip_session_terminate(session, 0);
	}

	/* Stop the termination timer if it is still running. */
	if (pj_timer_heap_cancel(pjsip_endpt_get_timer_heap(ast_sip_get_pjsip_endpoint()),
		&session->scheduled_termination)) {
		ao2_ref(session, -1);
	}
}

struct ast_sip_session *ast_sip_dialog_get_session(pjsip_dialog *dlg)
{
	pjsip_inv_session *inv_session = pjsip_dlg_get_inv_session(dlg);
	struct ast_sip_session *session;

	if (!inv_session ||
		!(session = inv_session->mod_data[session_module.id])) {
		return NULL;
	}

	ao2_ref(session, +1);

	return session;
}

enum sip_get_destination_result {
	/*! The extension was successfully found */
	SIP_GET_DEST_EXTEN_FOUND,
	/*! The extension specified in the RURI was not found */
	SIP_GET_DEST_EXTEN_NOT_FOUND,
	/*! The extension specified in the RURI was a partial match */
	SIP_GET_DEST_EXTEN_PARTIAL,
	/*! The RURI is of an unsupported scheme */
	SIP_GET_DEST_UNSUPPORTED_URI,
};

#ifdef GRANDSTREAM_NETWORKS
static int get_endpoint_details(pjsip_rx_data *rdata, char *endpoint, size_t endpoint_size, char *domain, size_t domain_size)
{
    pjsip_uri *from = rdata->msg_info.from->uri;
    pjsip_sip_uri *sip_from;
    if (!PJSIP_URI_SCHEME_IS_SIP(from) && !PJSIP_URI_SCHEME_IS_SIPS(from)) {
            return -1;
        }
    sip_from = (pjsip_sip_uri *) pjsip_uri_get_uri(from);
    ast_copy_pj_str(endpoint, &sip_from->user, endpoint_size);
    ast_copy_pj_str(domain, &sip_from->host, domain_size);
    return 0;
}
#endif
/*!
 * \brief Determine where in the dialplan a call should go
 *
 * This uses the username in the request URI to try to match
 * an extension in the endpoint's configured context in order
 * to route the call.
 *
 * \param session The inbound SIP session
 * \param rdata The SIP INVITE
 */
static enum sip_get_destination_result get_destination(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	pjsip_uri *ruri = rdata->msg_info.msg->line.req.uri;
	pjsip_sip_uri *sip_ruri;
	struct ast_features_pickup_config *pickup_cfg;
	const char *pickupexten;
    
#ifdef GRANDSTREAM_NETWORKS
	char endpoint_name[64]={0}, domain_name[64]={0};
	pjsip_tel_uri  *tel_ruri;
#endif

#ifdef GRANDSTREAM_NETWORKS
	if (!PJSIP_URI_SCHEME_IS_SIP(ruri) && !PJSIP_URI_SCHEME_IS_SIPS(ruri) && !PJSIP_URI_SCHEME_IS_TEL(ruri)) {
#else
	if (!PJSIP_URI_SCHEME_IS_SIP(ruri) && !PJSIP_URI_SCHEME_IS_SIPS(ruri)) {
#endif
		return SIP_GET_DEST_UNSUPPORTED_URI;
	}
#ifdef GRANDSTREAM_NETWORKS
	pjsip_uri *to = rdata->msg_info.to->uri;
	ast_log(LOG_DEBUG,"didmode = %s\n", session->endpoint->didmode);
	if(strcasecmp(session->endpoint->didmode, "to-header") == 0) {
		sip_ruri = pjsip_uri_get_uri(to);

		if (PJSIP_URI_SCHEME_IS_SIP(ruri) || PJSIP_URI_SCHEME_IS_SIPS(ruri)) {
			pjsip_sip_uri *sip_requri = pjsip_uri_get_uri(ruri);
			if (sip_requri->user.slen) {
				session->contact_user = ast_strndup(sip_requri->user.ptr, sip_requri->user.slen); 
			}
		} else if (PJSIP_URI_SCHEME_IS_TEL(ruri)) {
			pjsip_tel_uri *tel_requri = pjsip_uri_get_uri(ruri);
			if (tel_requri->number.slen) {
				session->contact_user = ast_strndup(tel_requri->number.ptr, tel_requri->number.slen); 
			}
		}
		ast_log(LOG_NOTICE, "contact_user = %s\n", session->contact_user);
	} else {
		sip_ruri = pjsip_uri_get_uri(ruri);
	}

    if (PJSIP_URI_SCHEME_IS_SIP(to) || PJSIP_URI_SCHEME_IS_SIPS(to)) {
		ast_copy_pj_str(session->exten, &sip_ruri->user, sizeof(session->exten));
    } else if (PJSIP_URI_SCHEME_IS_TEL(to)) {
		tel_ruri = (pjsip_tel_uri *)sip_ruri;
		ast_copy_pj_str(session->exten, &tel_ruri->number, sizeof(session->exten));
    }
#else
	sip_ruri = pjsip_uri_get_uri(ruri);
	ast_copy_pj_str(session->exten, &sip_ruri->user, sizeof(session->exten));
#endif

	pickup_cfg = ast_get_chan_features_pickup_config(session->channel);
	if (!pickup_cfg) {
		ast_log(LOG_ERROR, "Unable to retrieve pickup configuration options. Unable to detect call pickup extension\n");
		pickupexten = "";
	} else {
		pickupexten = ast_strdupa(pickup_cfg->pickupexten);
		ao2_ref(pickup_cfg, -1);
	}
#ifdef GRANDSTREAM_NETWORKS      //add by zhrliu for caller number match 20150121
	if (get_endpoint_details(rdata, endpoint_name, sizeof(endpoint_name), domain_name, sizeof(domain_name))) {
		if (!strcmp(session->exten, pickupexten) ||
				 ast_exists_extension(NULL, session->endpoint->context, session->exten, 1, NULL)) {
			return SIP_GET_DEST_EXTEN_FOUND;
		}
	} else {
		struct ast_sip_cc_original_exten_data exten_data;

        /* Added by yfyang, Grandstream networks, for ccss feature */
		if (0 == find_sub_handler_for_supplement_exten(AST_SIP_CCSS_ORIGINAL_EXTEN, session->exten, &exten_data)) {
			strncpy(session->exten, exten_data.original_exten, sizeof(session->exten)-1);
			session->core_id = exten_data.core_id;

			ast_cc_agent_recalling(exten_data.core_id, "SIP caller %s is attempting recall", exten_data.device_name);
			return SIP_GET_DEST_EXTEN_FOUND;
		}

		if (!strcmp(session->exten, pickupexten) ||
			ast_exists_extension(NULL, session->endpoint->context, session->exten, 1, endpoint_name)) {
			return SIP_GET_DEST_EXTEN_FOUND;
		}

    }
#else

	if (!strcmp(session->exten, pickupexten) ||
		ast_exists_extension(NULL, session->endpoint->context, session->exten, 1, NULL)) {
		return SIP_GET_DEST_EXTEN_FOUND;
	}
#endif

	/* XXX In reality, we'll likely have further options so that partial matches
	 * can be indicated here, but for getting something up and running, we're going
	 * to return a "not exists" error here.
	 */
	return SIP_GET_DEST_EXTEN_NOT_FOUND;
}

static pjsip_inv_session *pre_session_setup(pjsip_rx_data *rdata, const struct ast_sip_endpoint *endpoint)
{
	pjsip_tx_data *tdata;
	pjsip_dialog *dlg;
	pjsip_inv_session *inv_session;
	unsigned int options = endpoint->extensions.flags;
	pj_status_t dlg_status;

	if (pjsip_inv_verify_request(rdata, &options, NULL, NULL, ast_sip_get_pjsip_endpoint(), &tdata) != PJ_SUCCESS) {
		if (tdata) {
			pjsip_endpt_send_response2(ast_sip_get_pjsip_endpoint(), rdata, tdata, NULL, NULL);
		} else {
			pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 500, NULL, NULL, NULL);
		}
		return NULL;
	}
	dlg = ast_sip_create_dialog_uas(endpoint, rdata, &dlg_status);
	if (!dlg) {
		if (dlg_status != PJ_EEXISTS) {
			pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 500, NULL, NULL, NULL);
		}
		return NULL;
	}
	if (pjsip_inv_create_uas(dlg, rdata, NULL, options, &inv_session) != PJ_SUCCESS) {
		pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 500, NULL, NULL, NULL);
		pjsip_dlg_terminate(dlg);
		return NULL;
	}

#if defined(HAVE_PJSIP_REPLACE_MEDIA_STREAM) || defined(PJMEDIA_SDP_NEG_ALLOW_MEDIA_CHANGE)
	inv_session->sdp_neg_flags = PJMEDIA_SDP_NEG_ALLOW_MEDIA_CHANGE;
#endif
	if (pjsip_dlg_add_usage(dlg, &session_module, NULL) != PJ_SUCCESS) {
		if (pjsip_inv_initial_answer(inv_session, rdata, 500, NULL, NULL, &tdata) != PJ_SUCCESS) {
			pjsip_inv_terminate(inv_session, 500, PJ_FALSE);
		}
		pjsip_inv_send_msg(inv_session, tdata);
		return NULL;
	}
	return inv_session;
}

struct new_invite {
	/*! \brief Session created for the new INVITE */
	struct ast_sip_session *session;

	/*! \brief INVITE request itself */
	pjsip_rx_data *rdata;
};

static void new_invite_destroy(void *obj)
{
	struct new_invite *invite = obj;

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "Destroy invite %p, session:%p", invite, invite->session);
#endif
	ao2_cleanup(invite->session);

	if (invite->rdata) {
		pjsip_rx_data_free_cloned(invite->rdata);
	}
}

static struct new_invite *new_invite_alloc(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	struct new_invite *invite = ao2_alloc(sizeof(*invite), new_invite_destroy);

	if (!invite) {
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "Create new invite %p for SIP session %p", invite, session);
#endif
	ao2_ref(session, +1);
	invite->session = session;

#ifdef GRANDSTREAM_NETWORKS
	char buf[1024] = {0};
	const pjsip_rdata_sdp_info *sdp_info = pjsip_rdata_get_sdp_info(rdata);
	if (sdp_info && sdp_info->sdp) {
		pjmedia_sdp_print(sdp_info->sdp, buf, sizeof(buf) - 1);
		if ((!strcasecmp(rdata->tp_info.transport->type_name, "WS")
			|| !strcasecmp(rdata->tp_info.transport->type_name, "WSS"))
			&& strcasestr(buf, "candidate")
			&& ((session->endpoint->media.rtp.encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS)
			|| (session->endpoint->media.rtp.encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP)
			|| (session->endpoint->media.rtp.encryption == AST_SIP_MEDIA_AUTO_NO_ENCRYPT))) {
			session->is_incoming_webrtc_call = 1;
		}
	}
#endif

	if (pjsip_rx_data_clone(rdata, 0, &invite->rdata) != PJ_SUCCESS) {
		ao2_ref(invite, -1);
		return NULL;
	}

	return invite;
}

#ifdef GRANDSTREAM_NETWORKS
void ast_sip_session_set_use_final_sdp_val(struct ast_sip_session *session)
{
	if (!session || !session->inv_session) {
		ast_log(LOG_WARNING, "Set final sdp val, param is invalid\n");
		return;
	}

	char *use_final_rsp_sdp = ast_sip_get_use_final_sdp();
	if (strcasecmp(use_final_rsp_sdp, "yes") == 0) {
		session->inv_session->use_final_rsp_sdp = TRUE;
	} else {
		session->inv_session->use_final_rsp_sdp = FALSE;
	}

	ast_free(use_final_rsp_sdp);
	use_final_rsp_sdp = NULL;

	return;
}
#endif
static int new_invite(void *data)
{
	RAII_VAR(struct new_invite *, invite, data, ao2_cleanup);
	pjsip_tx_data *tdata = NULL;
	pjsip_timer_setting timer;
	pjsip_rdata_sdp_info *sdp_info;
	pjmedia_sdp_session *local = NULL;
#ifdef GRANDSTREAM_NETWORKS
	pj_status_t status;
	int  terminated_flag = 0;
	int  from_trunk = 0;
	const char *ept_id;
	ast_log(LOG_VERBOSE, "new invite starting ...\n");

	ept_id = ast_sorcery_object_get_id(invite->session->endpoint);
	if (!ept_id || strstr(ept_id, "trunk_")) {
		from_trunk = 1;
	}

	/* Added by yfyang, grandstream networks, for Bug 47622*/
	if (!from_trunk) {
		ast_devstate_changed(AST_DEVICE_INUSE, AST_DEVSTATE_CACHABLE, "PJSIP/%s", ept_id);
	}
	/* Added by vanevan, Grandstream Networks, for thread synchronization when handling new invite request and meeting a transport/transaction error at the same time */
	pjsip_dialog *dlg = invite->session->inv_session->dlg;
	if (dlg == NULL) {
		ast_log(LOG_WARNING, "PJSIP dialog is NULL!");
		goto invite_end;
	}
	pjsip_dlg_inc_lock(dlg);
	if (invite->session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_WARNING, "Invite state is disconnected before it is handled!");
		goto invite_end;
	}
#endif

	/* From this point on, any calls to pjsip_inv_terminate have the last argument as PJ_TRUE
	 * so that we will be notified so we can destroy the session properly
	 */

	switch (get_destination(invite->session, invite->rdata)) {
	case SIP_GET_DEST_EXTEN_FOUND:
		/* Things worked. Keep going */
		break;
	case SIP_GET_DEST_UNSUPPORTED_URI:
		if (pjsip_inv_initial_answer(invite->session->inv_session, invite->rdata, 416, NULL, NULL, &tdata) == PJ_SUCCESS) {
			ast_sip_session_send_response(invite->session, tdata);
		} else  {
			pjsip_inv_terminate(invite->session->inv_session, 416, PJ_TRUE);
		}
#ifdef GRANDSTREAM_NETWORKS
		terminated_flag = 1;
		goto invite_end;
#else
		return 0;
#endif
	case SIP_GET_DEST_EXTEN_NOT_FOUND:
	case SIP_GET_DEST_EXTEN_PARTIAL:
	default:
		ast_log(LOG_NOTICE, "Call from '%s' (%s:%s:%d) to extension '%s' rejected because extension not found in context '%s'.\n",
			ast_sorcery_object_get_id(invite->session->endpoint), invite->rdata->tp_info.transport->type_name, invite->rdata->pkt_info.src_name,
			invite->rdata->pkt_info.src_port, invite->session->exten, invite->session->endpoint->context);

		if (pjsip_inv_initial_answer(invite->session->inv_session, invite->rdata, 404, NULL, NULL, &tdata) == PJ_SUCCESS) {
			ast_sip_session_send_response(invite->session, tdata);
		} else  {
			pjsip_inv_terminate(invite->session->inv_session, 404, PJ_TRUE);
		}
#ifdef GRANDSTREAM_NETWORKS
		terminated_flag = 1;
		goto invite_end;
#else
		return 0;
#endif
	};

	if ((sdp_info = pjsip_rdata_get_sdp_info(invite->rdata)) && (sdp_info->sdp_err == PJ_SUCCESS) && sdp_info->sdp) {
#ifdef GRANDSTREAM_NETWORKS
		{
			char buf[1024] = {0};

			pjmedia_sdp_print(sdp_info->sdp, buf, sizeof(buf) - 1);
			ast_log(LOG_CODEC_NEG, "[caller] A ===>>> UCM. received invite sdp!\n%s", buf);
		}
#endif
		if (handle_incoming_sdp(invite->session, sdp_info->sdp)) {
			if (pjsip_inv_initial_answer(invite->session->inv_session, invite->rdata, 488, NULL, NULL, &tdata) == PJ_SUCCESS) {
				ast_sip_session_send_response(invite->session, tdata);
			} else  {
				pjsip_inv_terminate(invite->session->inv_session, 488, PJ_TRUE);
			}
#ifdef GRANDSTREAM_NETWORKS
			terminated_flag = 1;
			goto invite_end;
#else
			return 0;
#endif
		}
		/* We are creating a local SDP which is an answer to their offer */
		local = create_local_sdp(invite->session->inv_session, invite->session, sdp_info->sdp);
	} else {
		/* We are creating a local SDP which is an offer */
		local = create_local_sdp(invite->session->inv_session, invite->session, NULL);
	}

	/* If we were unable to create a local SDP terminate the session early, it won't go anywhere */
	if (!local) {
		if (pjsip_inv_initial_answer(invite->session->inv_session, invite->rdata, 500, NULL, NULL, &tdata) == PJ_SUCCESS) {
			ast_sip_session_send_response(invite->session, tdata);
		} else  {
			pjsip_inv_terminate(invite->session->inv_session, 500, PJ_TRUE);
		}
#ifdef GRANDSTREAM_NETWORKS
		terminated_flag = 1;
		goto invite_end;
#else
		return 0;
#endif
	} else {
		pjsip_inv_set_local_sdp(invite->session->inv_session, local);
		pjmedia_sdp_neg_set_prefer_remote_codec_order(invite->session->inv_session->neg, PJ_FALSE);
#ifdef PJMEDIA_SDP_NEG_ANSWER_MULTIPLE_CODECS
		pjmedia_sdp_neg_set_answer_multiple_codecs(invite->session->inv_session->neg, PJ_TRUE);
#endif
#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for 183, 2016.02.01 */
		ast_sip_session_set_use_final_sdp_val(invite->session);
#endif
	}

	pjsip_timer_setting_default(&timer);
	timer.min_se = invite->session->endpoint->extensions.timer.min_se;
	timer.sess_expires = invite->session->endpoint->extensions.timer.sess_expires;
	pjsip_timer_init_session(invite->session->inv_session, &timer);

	/* At this point, we've verified what we can, so let's go ahead and send a 100 Trying out */
#ifdef GRANDSTREAM_NETWORKS
	if ((status = pjsip_inv_initial_answer(invite->session->inv_session, invite->rdata, 100, NULL, NULL, &tdata)) != PJ_SUCCESS) {
		if (PJSIP_ERRNO_TO_SIP_STATUS(status) != 422) {
			pjsip_inv_terminate(invite->session->inv_session, 500, PJ_TRUE);
			terminated_flag = 1;
			goto invite_end;
        }
	}
#else
	if (pjsip_inv_initial_answer(invite->session->inv_session, invite->rdata, 100, NULL, NULL, &tdata) != PJ_SUCCESS) {
		pjsip_inv_terminate(invite->session->inv_session, 500, PJ_TRUE);
		return 0;
	}
#endif
	ast_sip_session_send_response(invite->session, tdata);

	handle_incoming_request(invite->session, invite->rdata, PJSIP_EVENT_RX_MSG);

#ifdef GRANDSTREAM_NETWORKS
invite_end:  
	/* Added by yfyang, grandstream networks, for Bug 47622*/
	if (terminated_flag && !from_trunk) {
		ast_devstate_changed(AST_DEVICE_UNKNOWN, AST_DEVSTATE_CACHABLE, "PJSIP/%s", ept_id);
	}
	pjsip_dlg_dec_lock(dlg);
#endif
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for yealink flag, 2015.12.01 */
static void set_yealink_flag(pjsip_rx_data *rdata, struct ast_sip_session *session)
{
	pjsip_msg *msg = NULL;
	pjsip_user_agent_hdr *user_agent_hdr = NULL;
	const pj_str_t name = { "User-Agent", 10 };
	const pj_str_t str_yealink = { "Yealink", 7 };
	char *user_agent_str = NULL;
	int user_agent_size = 0;

	ast_log(LOG_DEBUG, "Set yealink flag begin\n");

	if (!rdata || !session) {
		ast_log(LOG_ERROR, "Set yealink flag, para is invalid\n");
		return;
	}

	msg = rdata->msg_info.msg;
	if (msg) {
		user_agent_hdr = (pjsip_user_agent_hdr*) pjsip_msg_find_hdr_by_name(msg, &name, NULL);
	}

	if (!user_agent_hdr) {
		ast_log(LOG_DEBUG, "Set yealink flag, can't find user-agent head\n");
		return;
	}

	user_agent_size = pj_strlen(&user_agent_hdr->hvalue) + 1;
	user_agent_str = ast_malloc(user_agent_size);
	ast_copy_pj_str(user_agent_str, &user_agent_hdr->hvalue, user_agent_size);

	ast_log(LOG_DEBUG, "User-Agent value:%s", user_agent_str);

	if (pj_strstr(&user_agent_hdr->hvalue, &str_yealink) != NULL) {
		ast_log(LOG_DEBUG, "Set yealink flag true");
		session->is_yealink_msg = TRUE;
	}

	ast_free(user_agent_str);
	user_agent_str = NULL;

	return;
}

/* To force change profile-level-id value when receive yealink message, 2015.12.02 */
static pjmedia_sdp_attr* pjmedia_sdp_attr_change_fmtp_param(pj_pool_t *pool,
	pjmedia_sdp_attr *attr_fmtp, const pj_str_t *payload)
{
	pj_str_t attr_name = pj_str("fmtp");
	pj_str_t find_str = pj_str("profile-level-id");
	char* fmtp_profile = NULL;
	char* change_param_str = "profile-level-id=428014";
	char* fmtp_buf = NULL;
	pj_str_t fmtp_value;
	int len_fmtp_value = 0;

	if (!pool || !attr_fmtp || !payload) {
		ast_log(LOG_DEBUG, "Param is null, don't change");
		return NULL;
	}

	if (pj_strcmp(&attr_fmtp->name, &attr_name)) {
		ast_log(LOG_DEBUG, "It's not a fmtp, don't change");
		return NULL;
	}

	len_fmtp_value = attr_fmtp->value.slen;

	fmtp_buf = pj_pool_alloc(pool, len_fmtp_value);
	pj_strset(&fmtp_value, fmtp_buf, 0);
	pj_strcpy(&fmtp_value, &attr_fmtp->value);
	fmtp_profile = pj_strstr(&fmtp_value, &find_str);
	if (!fmtp_profile) {
		ast_log(LOG_DEBUG, "Can't find profile-level-id string, don't need change\n");
		return pjmedia_sdp_attr_create(pool, "fmtp", &fmtp_value);
	}

	/* don't copy '\0' to fmtp_profile, 23 is the len of "profile-level-id=428014" */
	strncpy(fmtp_profile, change_param_str, 23);

	return pjmedia_sdp_attr_create(pool, "fmtp", &fmtp_value);
}

void ast_sip_session_modify_outgoing_yealink_sdp(const struct ast_sip_session *session,
	pj_pool_t *pool, pjmedia_sdp_media *media)
{
	int i = 0;

	if (!session->is_yealink_msg) {
		ast_log(LOG_DEBUG, "Isn't a yealink message, don't need modify");
		return;
	}

	ast_log(LOG_DEBUG, "Modify outgoing yealink sdp begin");

	for (i = 0; i < media->desc.fmt_count; ++i) {
		pjmedia_sdp_attr *attr_fmtp = NULL;
		pjmedia_sdp_attr *new_attr_fmtp = NULL;

		attr_fmtp = pjmedia_sdp_media_find_attr2(media, "fmtp", &media->desc.fmt[i]);

		if (attr_fmtp) {
			ast_log(LOG_DEBUG, "Modify outgoing yealink sdp, find fmtp attr");
			pjmedia_sdp_media_remove_attr(media, attr_fmtp);

			new_attr_fmtp = pjmedia_sdp_attr_change_fmtp_param(pool, attr_fmtp, &media->desc.fmt[i]);
			if (new_attr_fmtp) {
				pjmedia_sdp_media_add_attr(media, new_attr_fmtp);
			}
		}
	}

	return;
}

#endif

static void handle_new_invite_request(pjsip_rx_data *rdata)
{
	RAII_VAR(struct ast_sip_endpoint *, endpoint,
			ast_pjsip_rdata_get_endpoint(rdata), ao2_cleanup);
	pjsip_tx_data *tdata = NULL;
	pjsip_inv_session *inv_session = NULL;
	struct ast_sip_session *session;
	struct new_invite *invite;

	ast_assert(endpoint != NULL);
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "handle new sip invite request ...\n");
#endif

	inv_session = pre_session_setup(rdata, endpoint);
	if (!inv_session) {
		/* pre_session_setup() returns a response on failure */
		return;
	}

	session = ast_sip_session_alloc(endpoint, NULL, inv_session);
	if (!session) {
		if (pjsip_inv_initial_answer(inv_session, rdata, 500, NULL, NULL, &tdata) == PJ_SUCCESS) {
			pjsip_inv_terminate(inv_session, 500, PJ_FALSE);
		} else {
			pjsip_inv_send_msg(inv_session, tdata);
		}
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Begin add by bfdu for yealink flag, 2015.12.01 */
	set_yealink_flag(rdata, session);
	/* End add by bfdu for yealink flag, 2015.12.01 */
	session->session_uas = 1;
#endif

	invite = new_invite_alloc(session, rdata);
	if (!invite || ast_sip_push_task(session->serializer, new_invite, invite)) {
		if (pjsip_inv_initial_answer(inv_session, rdata, 500, NULL, NULL, &tdata) == PJ_SUCCESS) {
			pjsip_inv_terminate(inv_session, 500, PJ_FALSE);
		} else {
			pjsip_inv_send_msg(inv_session, tdata);
		}
		ao2_cleanup(invite);
	}
	ao2_ref(session, -1);
}

static pj_bool_t does_method_match(const pj_str_t *message_method, const char *supplement_method)
{
	pj_str_t method;

	if (ast_strlen_zero(supplement_method)) {
		return PJ_TRUE;
	}

	pj_cstr(&method, supplement_method);

	return pj_stristr(&method, message_method) ? PJ_TRUE : PJ_FALSE;
}

static pj_bool_t has_supplement(const struct ast_sip_session *session, const pjsip_rx_data *rdata)
{
	struct ast_sip_session_supplement *supplement;
	struct pjsip_method *method = &rdata->msg_info.msg->line.req.method;

	if (!session) {
		return PJ_FALSE;
	}

	AST_LIST_TRAVERSE(&session->supplements, supplement, next) {
		if (does_method_match(&method->name, supplement->method)) {
			return PJ_TRUE;
		}
	}
	return PJ_FALSE;
}
/*!
 * \brief Called when a new SIP request comes into PJSIP
 *
 * This function is called under two circumstances
 * 1) An out-of-dialog request is received by PJSIP
 * 2) An in-dialog request that the inv_session layer does not
 *    handle is received (such as an in-dialog INFO)
 *
 * In all cases, there is very little we actually do in this function
 * 1) For requests we don't handle, we return PJ_FALSE
 * 2) For new INVITEs, throw the work into the SIP threadpool to be done
 *    there to free up the thread(s) handling incoming requests
 * 3) For in-dialog requests we handle, we defer handling them until the
 *    on_inv_state_change() callback instead (where we will end up putting
 *    them into the threadpool).
 */
static pj_bool_t session_on_rx_request(pjsip_rx_data *rdata)
{
	pj_status_t handled = PJ_FALSE;
	pjsip_dialog *dlg = pjsip_rdata_get_dlg(rdata);
	pjsip_inv_session *inv_session;
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "starting sip session on receive request ...\n");
#endif

	switch (rdata->msg_info.msg->line.req.method.id) {
	case PJSIP_INVITE_METHOD:
		if (dlg) {
			ast_log(LOG_WARNING, "on_rx_request called for INVITE in mid-dialog?\n");
			break;
		}
		handled = PJ_TRUE;
		handle_new_invite_request(rdata);
		break;
	default:
		/* Handle other in-dialog methods if their supplements have been registered */
		handled = dlg && (inv_session = pjsip_dlg_get_inv_session(dlg)) &&
			has_supplement(inv_session->mod_data[session_module.id], rdata);
		break;
	}

	return handled;
}

static void resend_reinvite(pj_timer_heap_t *timer, pj_timer_entry *entry)
{
	struct ast_sip_session *session = entry->user_data;

	ast_debug(3, "Endpoint '%s(%s)' re-INVITE collision timer expired.\n",
		ast_sorcery_object_get_id(session->endpoint),
		session->channel ? ast_channel_name(session->channel) : "");
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for crash when BYE, 2015.08.26 */
	ao2_ref(session, +1);
#endif
	if (AST_LIST_EMPTY(&session->delayed_requests)) {
		/* No delayed request pending, so just return */
		ao2_ref(session, -1);
		return;
	}
	if (ast_sip_push_task(session->serializer, invite_collision_timeout, session)) {
		/*
		 * Uh oh.  We now have nothing in the foreseeable future
		 * to trigger sending the delayed requests.
		 */
		ao2_ref(session, -1);
	}
}

static void reschedule_reinvite(struct ast_sip_session *session, ast_sip_session_response_cb on_response)
{
	pjsip_inv_session *inv = session->inv_session;
	pj_time_val tv;

	ast_debug(3, "Endpoint '%s(%s)' re-INVITE collision.\n",
		ast_sorcery_object_get_id(session->endpoint),
		session->channel ? ast_channel_name(session->channel) : "");
	if (delay_request(session, NULL, NULL, on_response, 1, DELAYED_METHOD_INVITE)) {
		return;
	}
	if (pj_timer_entry_running(&session->rescheduled_reinvite)) {
		/* Timer already running.  Something weird is going on. */
		ast_debug(1, "Endpoint '%s(%s)' re-INVITE collision while timer running!!!\n",
			ast_sorcery_object_get_id(session->endpoint),
			session->channel ? ast_channel_name(session->channel) : "");
		return;
	}

	tv.sec = 0;
	if (inv->role == PJSIP_ROLE_UAC) {
		tv.msec = 2100 + ast_random() % 2000;
	} else {
		tv.msec = ast_random() % 2000;
	}
	pj_timer_entry_init(&session->rescheduled_reinvite, 0, session, resend_reinvite);

	pjsip_endpt_schedule_timer(ast_sip_get_pjsip_endpoint(), &session->rescheduled_reinvite, &tv);
}

static void __print_debug_details(const char *function, pjsip_inv_session *inv, pjsip_transaction *tsx, pjsip_event *e)
{
	struct ast_sip_session *session;

	if (!DEBUG_ATLEAST(5)) {
		/* Debug not spamy enough */
		return;
	}

	ast_log(LOG_DEBUG, "Function %s called on event %s\n",
		function, pjsip_event_str(e->type));
	if (!inv) {
		ast_log(LOG_DEBUG, "Transaction %p does not belong to an inv_session?\n", tsx);
		ast_log(LOG_DEBUG, "The transaction state is %s\n",
			pjsip_tsx_state_str(tsx->state));
		return;
	}
	session = inv->mod_data[session_module.id];
	if (!session) {
		ast_log(LOG_DEBUG, "inv_session %p has no ast session\n", inv);
	} else {
		ast_log(LOG_DEBUG, "The state change pertains to the endpoint '%s(%s)'\n",
			ast_sorcery_object_get_id(session->endpoint),
			session->channel ? ast_channel_name(session->channel) : "");
	}
	if (inv->invite_tsx) {
		ast_log(LOG_DEBUG, "The inv session still has an invite_tsx (%p)\n",
			inv->invite_tsx);
	} else {
		ast_log(LOG_DEBUG, "The inv session does NOT have an invite_tsx\n");
	}
	if (tsx) {
		ast_log(LOG_DEBUG, "The %s %.*s transaction involved in this state change is %p\n",
			pjsip_role_name(tsx->role),
			(int) pj_strlen(&tsx->method.name), pj_strbuf(&tsx->method.name),
			tsx);
		ast_log(LOG_DEBUG, "The current transaction state is %s\n",
			pjsip_tsx_state_str(tsx->state));
		ast_log(LOG_DEBUG, "The transaction state change event is %s\n",
			pjsip_event_str(e->body.tsx_state.type));
	} else {
		ast_log(LOG_DEBUG, "There is no transaction involved in this state change\n");
	}
	ast_log(LOG_DEBUG, "The current inv state is %s\n", pjsip_inv_state_name(inv->state));
}

#define print_debug_details(inv, tsx, e) __print_debug_details(__PRETTY_FUNCTION__, (inv), (tsx), (e))

static void handle_incoming_request(struct ast_sip_session *session, pjsip_rx_data *rdata, pjsip_event_id_e type)
{
	struct ast_sip_session_supplement *supplement;
	struct pjsip_request_line req = rdata->msg_info.msg->line.req;

	ast_debug(3, "Method is %.*s\n", (int) pj_strlen(&req.method.name), pj_strbuf(&req.method.name));
	AST_LIST_TRAVERSE(&session->supplements, supplement, next) {
		if (supplement->incoming_request && does_method_match(&req.method.name, supplement->method)) {
			if (supplement->incoming_request(session, rdata)) {
				break;
			}
		}
	}
}

static void handle_incoming_response(struct ast_sip_session *session, pjsip_rx_data *rdata, pjsip_event_id_e type,
		enum ast_sip_session_response_priority response_priority)
{
	struct ast_sip_session_supplement *supplement;
	struct pjsip_status_line status = rdata->msg_info.msg->line.status;

	ast_debug(3, "Response is %d %.*s\n", status.code, (int) pj_strlen(&status.reason),
			pj_strbuf(&status.reason));
#ifdef GRANDSTREAM_NETWORKS
	session->is_outgoing = 0;
	/*add by wxu@grandstream.cn for add sip_invite_failed*/
	if(does_method_match(&rdata->msg_info.cseq->method.name, "INVITE")){
		if(status.code == 486){
			if(response_priority == AST_SIP_SESSION_BEFORE_MEDIA)
				ast_sip_report_invite_failed(session,rdata,NULL,"INVITE",status.code,ast_cause2str(hangup_sip2cause(status.code)));
		}else{
			ast_sip_report_invite_failed(session,rdata,NULL,"INVITE",status.code,ast_cause2str(hangup_sip2cause(status.code)));
		}
	}
	/*end*/
#endif

	AST_LIST_TRAVERSE(&session->supplements, supplement, next) {
		if (!(supplement->response_priority & response_priority)) {
			continue;
		}
		if (supplement->incoming_response && does_method_match(&rdata->msg_info.cseq->method.name, supplement->method)) {
			supplement->incoming_response(session, rdata);
		}
	}
}

static int handle_incoming(struct ast_sip_session *session, pjsip_rx_data *rdata, pjsip_event_id_e type,
		enum ast_sip_session_response_priority response_priority)
{
	ast_debug(3, "Received %s\n", rdata->msg_info.msg->type == PJSIP_REQUEST_MSG ?
			"request" : "response");

#ifdef GRANDSTREAM_NETWORKS
	if ((session == NULL) || (ao2_ref(session, 0) <= 0) || (session->channel == NULL)) {
		ast_log(LOG_NOTICE, "Invalid SIP session or channel, do nothing!");
		return -1;
	}

	ao2_ref(session, +1);

	/* Begin add by bfdu for yealink flag, 2015.12.01 */
	set_yealink_flag(rdata, session);
	/* End add by bfdu for yealink flag, 2015.12.01 */
#endif
	if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG) {
		handle_incoming_request(session, rdata, type);
	} else {
		handle_incoming_response(session, rdata, type, response_priority);
	}
#ifdef GRANDSTREAM_NETWORKS
	ao2_ref(session, -1);
#endif
	return 0;
}

static void handle_outgoing_request(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	struct ast_sip_session_supplement *supplement;
	struct pjsip_request_line req = tdata->msg->line.req;

	ast_debug(3, "Method is %.*s\n", (int) pj_strlen(&req.method.name), pj_strbuf(&req.method.name));
	AST_LIST_TRAVERSE(&session->supplements, supplement, next) {
		if (supplement->outgoing_request && does_method_match(&req.method.name, supplement->method)) {
			supplement->outgoing_request(session, tdata);
		}
	}
}

static void handle_outgoing_response(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	struct ast_sip_session_supplement *supplement;
	struct pjsip_status_line status = tdata->msg->line.status;
	pjsip_cseq_hdr *cseq = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_CSEQ, NULL);
	ast_debug(3, "Method is %.*s, Response is %d %.*s\n", (int) pj_strlen(&cseq->method.name),
		pj_strbuf(&cseq->method.name), status.code, (int) pj_strlen(&status.reason),
		pj_strbuf(&status.reason));

	AST_LIST_TRAVERSE(&session->supplements, supplement, next) {
		if (supplement->outgoing_response && does_method_match(&cseq->method.name, supplement->method)) {
			supplement->outgoing_response(session, tdata);
		}
	}
}

static void handle_outgoing(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	ast_debug(3, "Sending %s\n", tdata->msg->type == PJSIP_REQUEST_MSG ?
			"request" : "response");
#ifdef GRANDSTREAM_NETWORKS
	if ((session == NULL) || (ao2_ref(session, 0) <= 0)) {
		ast_log(LOG_ERROR, "Invalid SIP session!");
		return;
	}
	ao2_lock(session);
#endif
	if (tdata->msg->type == PJSIP_REQUEST_MSG) {
#ifdef GRANDSTREAM_NETWORKS
		/*add by wxu@grandstream.cn for add sip_invite_failed*/
		struct pjsip_request_line req = tdata->msg->line.req;
		/*! begin: Added by pxie@grandstream.cn for bug[42228], 2015-04-13*/
		if(session && session->channel && does_method_match(&req.method.name, "BYE")) {
		/*! end:   Added by pxie@grandstream.cn for bug[42228], 2015-04-13*/
		/*add by zhrliu for bug[44336] 2015-6-10*/
		if(AST_CAUSE_NORMAL_CLEARING != ast_channel_hangupcause(session->channel))
			ast_sip_report_invite_failed(session,NULL,tdata,"BYE",0,ast_cause2str(ast_channel_hangupcause(session->channel)));
		}
		/*end*/
#endif
		handle_outgoing_request(session, tdata);
	} else {
		handle_outgoing_response(session, tdata);
	}
#ifdef GRANDSTREAM_NETWORKS
	ao2_unlock(session);
#endif
}

#ifdef GRANDSTREAM_NETWORKS
static void session_end(struct ast_sip_session *session)
{
	struct ast_sip_session_supplement *iter;

	ao2_lock(session);
	/* Stop the scheduled termination */
	if (pj_timer_heap_cancel(pjsip_endpt_get_timer_heap(ast_sip_get_pjsip_endpoint()), &session->scheduled_termination)) {
		ao2_ref(session, -1);
	}

	/* Session is dead. Notify the supplements */
	AST_LIST_TRAVERSE(&session->supplements, iter, next) {
		if (iter->session_end) {
			iter->session_end(session);
		}
	}
	ao2_unlock(session);
}

 /*!
 * \internal
 * \brief Complete ending session activities.
 * \since 13.5.0
 *
 * \param vsession Which session to complete stopping.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int session_end_completion(void *vsession)
{
	struct ast_sip_session *session = vsession;

	ao2_lock(session);
	session->inv_session->mod_data[session_module.id] = NULL;
	ast_sip_dialog_set_serializer(session->inv_session->dlg, NULL);
	ast_sip_dialog_set_endpoint(session->inv_session->dlg, NULL);
	ao2_unlock(session);
	ao2_cleanup(session);
	return 0;
}

#else
static int session_end(struct ast_sip_session *session)
{
	struct ast_sip_session_supplement *iter;

	/* Stop the scheduled termination */
	if (pj_timer_heap_cancel(pjsip_endpt_get_timer_heap(ast_sip_get_pjsip_endpoint()), &session->scheduled_termination)) {
		ao2_ref(session, -1);
	}

	/* Session is dead. Let's get rid of the reference to the session */
	AST_LIST_TRAVERSE(&session->supplements, iter, next) {
		if (iter->session_end) {
			iter->session_end(session);
		}
	}

	session->inv_session->mod_data[session_module.id] = NULL;
	ast_sip_dialog_set_serializer(session->inv_session->dlg, NULL);
	ast_sip_dialog_set_endpoint(session->inv_session->dlg, NULL);
	ao2_cleanup(session);
	return 0;
}
#endif

static void session_inv_on_state_changed(pjsip_inv_session *inv, pjsip_event *e)
{
	struct ast_sip_session *session = inv->mod_data[session_module.id];
	pjsip_event_id_e type;

#ifdef GRANDSTREAM_NETWORKS
	if ((session == NULL) || (ao2_ref(session, 0) <= 0)) {
		ast_log(LOG_ERROR, "Invalid SIP session when state changed!");
		return;
	}
#endif

	if (e) {
		print_debug_details(inv, NULL, e);
		type = e->type;
	} else {
		type = PJSIP_EVENT_UNKNOWN;
	}

	if (!session) {
		return;
	}

	switch(type) {
	case PJSIP_EVENT_TX_MSG:
		handle_outgoing(session, e->body.tx_msg.tdata);
		break;
	case PJSIP_EVENT_RX_MSG:
		handle_incoming(session, e->body.rx_msg.rdata, type,
				AST_SIP_SESSION_BEFORE_MEDIA);
		break;
	case PJSIP_EVENT_TSX_STATE:
		ast_debug(3, "Source of transaction state change is %s\n", pjsip_event_str(e->body.tsx_state.type));
		/* Transaction state changes are prompted by some other underlying event. */
		switch(e->body.tsx_state.type) {
		case PJSIP_EVENT_TX_MSG:
			handle_outgoing(session, e->body.tsx_state.src.tdata);
			break;
		case PJSIP_EVENT_RX_MSG:
			handle_incoming(session, e->body.tsx_state.src.rdata, type,
					AST_SIP_SESSION_BEFORE_MEDIA);
			break;
		case PJSIP_EVENT_TRANSPORT_ERROR:
		case PJSIP_EVENT_TIMER:
		case PJSIP_EVENT_USER:
		case PJSIP_EVENT_UNKNOWN:
		case PJSIP_EVENT_TSX_STATE:
			/* Inception? */
			break;
		}
		break;
	case PJSIP_EVENT_TRANSPORT_ERROR:
	case PJSIP_EVENT_TIMER:
	case PJSIP_EVENT_UNKNOWN:
	case PJSIP_EVENT_USER:
	default:
		break;
	}

	if (inv->state == PJSIP_INV_STATE_DISCONNECTED) {
		session_end(session);
	}
}

static void session_inv_on_new_session(pjsip_inv_session *inv, pjsip_event *e)
{
	/* XXX STUB */
}

static void session_inv_on_tsx_state_changed(pjsip_inv_session *inv, pjsip_transaction *tsx, pjsip_event *e)
{
	ast_sip_session_response_cb cb;
	struct ast_sip_session *session = inv->mod_data[session_module.id];

#ifdef GRANDSTREAM_NETWORKS
	if ((session == NULL) || (ao2_ref(session, 0) <= 0)) {
		return;
	}
#endif

	print_debug_details(inv, tsx, e);

	if (!session) {
		/* Transaction likely timed out after the call was hung up. Just
		 * ignore such transaction changes
		 */
		return;
	}

	switch (e->body.tsx_state.type) {
	case PJSIP_EVENT_TX_MSG:
		handle_outgoing(session, e->body.tsx_state.src.tdata);
		/* When we create an outgoing request, we do not have access to the transaction that
		 * is created. Instead, We have to place transaction-specific data in the tdata. Here,
		 * we transfer the data into the transaction. This way, when we receive a response, we
		 * can dig this data out again
		 */
		tsx->mod_data[session_module.id] = e->body.tsx_state.src.tdata->mod_data[session_module.id];
		break;
	case PJSIP_EVENT_RX_MSG:
		cb = ast_sip_mod_data_get(tsx->mod_data, session_module.id, MOD_DATA_ON_RESPONSE);
		handle_incoming(session, e->body.tsx_state.src.rdata, e->type,
				AST_SIP_SESSION_AFTER_MEDIA);
		if (tsx->method.id == PJSIP_INVITE_METHOD) {
			if (tsx->role == PJSIP_ROLE_UAC) {
				if (tsx->state == PJSIP_TSX_STATE_COMPLETED) {
					/* This means we got a non 2XX final response to our outgoing INVITE */
					if (tsx->status_code == PJSIP_SC_REQUEST_PENDING) {
						reschedule_reinvite(session, cb);
						return;
					} else if (inv->state == PJSIP_INV_STATE_CONFIRMED &&
						   tsx->status_code != 488) {
						/* Other reinvite failures (except 488) result in destroying the session. */
						pjsip_tx_data *tdata;
						if (pjsip_inv_end_session(inv, 500, NULL, &tdata) == PJ_SUCCESS) {
							ast_sip_session_send_request(session, tdata);
						}
					}
				} else if (tsx->state == PJSIP_TSX_STATE_TERMINATED) {
					if (inv->cancelling && tsx->status_code == PJSIP_SC_OK) {
						/* This is a race condition detailed in RFC 5407 section 3.1.2.
						 * We sent a CANCEL at the same time that the UAS sent us a 200 OK for
						 * the original INVITE. As a result, we have now received a 200 OK for
						 * a cancelled call. Our role is to immediately send a BYE to end the
						 * dialog.
						 */
						pjsip_tx_data *tdata;

						if (pjsip_inv_end_session(inv, 500, NULL, &tdata) == PJ_SUCCESS) {
							ast_sip_session_send_request(session, tdata);
						}
					}
				}
			}
		}
		if (cb) {
			cb(session, e->body.tsx_state.src.rdata);
		}
#ifdef GRANDSTREAM_NETWORKS
	/*! Begin: Added by pxie@grandstream.cn	for bug[43844], 2015-06-19.*/
	if (session->wImageReInvite && (tsx->status_code == 488) && (tsx->state == PJSIP_TSX_STATE_COMPLETED) && (session->wSwitchImageInvite488ToAudioInvite)) {
		ast_log(LOG_VERBOSE, "Send Re-Invite with audio sdp, when T.38 negotiate failed!");
		ast_sip_session_refresh(session, NULL, NULL, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1);
	}
	/*! End: Added by pxie@grandstream.cn	for bug[43844], 2015-06-19.*/
#endif
		break;
	case PJSIP_EVENT_TRANSPORT_ERROR:
#ifdef GRANDSTREAM_NETWORKS
		/*
		 * Clear the module data now to block session_inv_on_state_changed()
		 * from calling session_end() if it hasn't already done so.
		 */
		inv->mod_data[session_module.id] = NULL;

		if (inv->state != PJSIP_INV_STATE_DISCONNECTED) {
			session_end(session);
		}

		/*
		 * Pass the session ref held by session->inv_session to
		 * session_end_completion().
		 */
		session_end_completion(session);
		return;
#endif
	case PJSIP_EVENT_TIMER:
#ifdef GRANDSTREAM_NETWORKS
		/*
		 * The timer event is run by the pjsip monitor thread and not
		 * by the session serializer.
		 */
		if (inv->state == PJSIP_INV_STATE_DISCONNECTED) {
			/*
			 * We are locking because ast_sip_dialog_get_session() needs
			 * the dialog locked to get the session by other threads.
			 */
			pjsip_dlg_inc_lock(inv->dlg);
			session = inv->mod_data[session_module.id];
			inv->mod_data[session_module.id] = NULL;
			pjsip_dlg_dec_lock(inv->dlg);

			/*
			 * Pass the session ref held by session->inv_session to
			 * session_end_completion().
			 */
			if (ast_sip_push_task(session->serializer, session_end_completion, session)) {
				/* Do it anyway even though this is not the right thread. */
				session_end_completion(session);
			}
			return;
		}
		break;
#endif
	case PJSIP_EVENT_USER:
	case PJSIP_EVENT_UNKNOWN:
	case PJSIP_EVENT_TSX_STATE:
		/* Inception? */
		break;
	}

	if (AST_LIST_EMPTY(&session->delayed_requests)) {
		/* No delayed request pending, so just return */
		return;
	}

	if (tsx->method.id == PJSIP_INVITE_METHOD) {
		if (tsx->state == PJSIP_TSX_STATE_PROCEEDING) {
			ast_debug(3, "Endpoint '%s(%s)' INVITE delay check. tsx-state:%s\n",
				ast_sorcery_object_get_id(session->endpoint),
				session->channel ? ast_channel_name(session->channel) : "",
				pjsip_tsx_state_str(tsx->state));
			check_delayed_requests(session, invite_proceeding);
		} else if (tsx->state == PJSIP_TSX_STATE_TERMINATED) {
			/*
			 * Terminated INVITE transactions always should result in
			 * queuing delayed requests, no matter what event caused
			 * the transaction to terminate.
			 */
			ast_debug(3, "Endpoint '%s(%s)' INVITE delay check. tsx-state:%s\n",
				ast_sorcery_object_get_id(session->endpoint),
				session->channel ? ast_channel_name(session->channel) : "",
				pjsip_tsx_state_str(tsx->state));
			check_delayed_requests(session, invite_terminated);
		}
	} else if (tsx->role == PJSIP_ROLE_UAC
		&& tsx->state == PJSIP_TSX_STATE_COMPLETED
		&& !pj_strcmp2(&tsx->method.name, "UPDATE")) {
		ast_debug(3, "Endpoint '%s(%s)' UPDATE delay check. tsx-state:%s\n",
			ast_sorcery_object_get_id(session->endpoint),
			session->channel ? ast_channel_name(session->channel) : "",
			pjsip_tsx_state_str(tsx->state));
		check_delayed_requests(session, update_completed);
	}
}

static int add_sdp_streams(void *obj, void *arg, void *data, int flags)
{
	struct ast_sip_session_media *session_media = obj;
	pjmedia_sdp_session *answer = arg;
	struct ast_sip_session *session = data;
	struct ast_sip_session_sdp_handler *handler = session_media->handler;
	RAII_VAR(struct sdp_handler_list *, handler_list, NULL, ao2_cleanup);
	int res;
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "add_sdp_streams(), handler[%p]\n", handler);
#endif

	if (handler) {
		/* if an already assigned handler does not handle the session_media or reports a catastrophic error, fail */
		res = handler->create_outgoing_sdp_stream(session, session_media, answer);
		if (res < 0) {
			return 0;
		}
		return CMP_MATCH;
	}

	handler_list = ao2_find(sdp_handlers, session_media->stream_type, OBJ_KEY);
	if (!handler_list) {
		return CMP_MATCH;
	}

	/* no handler for this stream type and we have a list to search */
	AST_LIST_TRAVERSE(&handler_list->list, handler, next) {
		if (handler == session_media->handler) {
			continue;
		}
		res = handler->create_outgoing_sdp_stream(session, session_media, answer);
		if (res < 0) {
			/* catastrophic error */
			return 0;
		}
		if (res > 0) {
			/* Handled by this handler. Move to the next stream */
			session_media_set_handler(session_media, handler);
			return CMP_MATCH;
		}
	}

	/* streams that weren't handled won't be included in generated outbound SDP */
	return CMP_MATCH;
}

static struct pjmedia_sdp_session *create_local_sdp(pjsip_inv_session *inv, struct ast_sip_session *session, const pjmedia_sdp_session *offer)
{
	RAII_VAR(struct ao2_iterator *, successful, NULL, ao2_iterator_cleanup);
	static const pj_str_t STR_IN = { "IN", 2 };
	static const pj_str_t STR_IP4 = { "IP4", 3 };
	static const pj_str_t STR_IP6 = { "IP6", 3 };
	pjmedia_sdp_session *local;

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "create_local_sdp(),enter ...\n");
#endif
	if (!(local = PJ_POOL_ZALLOC_T(inv->pool_prov, pjmedia_sdp_session))) {
		return NULL;
	}

	if (!offer) {
		local->origin.version = local->origin.id = (pj_uint32_t)(ast_random());
	} else {
		local->origin.version = offer->origin.version + 1;
		local->origin.id = offer->origin.id;
	}
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu, CT add when has video caps, 2015.10.22 */
	RAII_VAR(struct ast_format_cap *, video_caps, NULL, ao2_cleanup);
	int has_video = FALSE;
	if (!(video_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate video capabilities\n");
		has_video = FALSE;
	} else {
		ast_format_cap_append_from_cap(video_caps, session->req_caps, AST_MEDIA_TYPE_VIDEO);
		if (!ast_format_cap_count(video_caps)) {
			ast_log(LOG_DEBUG, "Req_caps has no video, don't add CT\n");
			has_video = FALSE;
		} else {
			has_video = TRUE;
		}
	}
	/* add jklou maxcallbitrate*/
	unsigned int maxbit = ast_sip_get_maxcallbitrate();
	if (maxbit && has_video){
		const pj_str_t STR_BANDW_MODIFIER = { "CT", 2 };
		pjmedia_sdp_bandw *b;

		b = PJ_POOL_ALLOC_T(inv->pool_prov, pjmedia_sdp_bandw);
		b->modifier = STR_BANDW_MODIFIER;
		b->value = maxbit;
		local->bandw[local->bandw_count++] = b;
	}
#endif

	pj_strdup2(inv->pool_prov, &local->origin.user, session->endpoint->media.sdpowner);
	pj_strdup2(inv->pool_prov, &local->name, session->endpoint->media.sdpsession);

	/* Now let the handlers add streams of various types, pjmedia will automatically reorder the media streams for us */
	successful = ao2_callback_data(session->media, OBJ_MULTIPLE, add_sdp_streams, local, session);
#ifdef GRANDSTREAM_NETWORKS
	if (!successful || ((ao2_iterator_count(successful) != ao2_container_count(session->media)) && !ast_channel_get_originate_call_value(session->channel)))
#else
	if (!successful || ao2_iterator_count(successful) != ao2_container_count(session->media))
#endif
	{
		/* Something experienced a catastrophic failure */
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_VERBOSE, "Something experienced a catastrophic failure\n");
#endif
		return NULL;
	}
#ifdef GRANDSTREAM_NETWORKS
	unsigned i;
	for (i = 0; i < local->media_count; ++i) {
		pjmedia_sdp_media *im;
		im = local->media[i];
		char media_type[16];
		memset(media_type, 0, sizeof(media_type));
		ast_copy_pj_str(media_type, &im->desc.media, sizeof(media_type));

		if (pjmedia_sdp_media_find_attr2(im, "sendrecv", NULL)) {
			pjmedia_sdp_attr *attr;
			pjmedia_sdp_media_remove_all_attr(im, "sendrecv");
			enum ast_reinvite_attr switch_case = REINVITE_SENDRECV;
			if (strncmp(media_type, "audio", 5) == 0) {
				switch_case = session->reinvite_attr.reinvite_attr_audio;
			} else if (strncmp(media_type, "video", 5) == 0) {
				pjmedia_sdp_attr *tmp_attr;
				char content[8];
				memset(content, 0, sizeof(content));
				if ((tmp_attr = pjmedia_sdp_media_find_attr2(im, "content", NULL))) {
					ast_copy_pj_str(content, pj_strltrim(&tmp_attr->value), sizeof(content));
					if (0 != strlen(content)) {
						if (!strncmp(content, "slides", 6)) {
							switch_case = session->reinvite_attr.reinvite_attr_s_video;
						} else {
							switch_case = session->reinvite_attr.reinvite_attr_m_video;
						}
					}
				} else {
					switch_case = session->reinvite_attr.reinvite_attr_m_video;
				}
			} else if (strncmp(media_type, "application", 11) == 0) {
				enum ast_sip_application_type app_type = ast_sip_session_get_application_type(im->desc.transport);
				switch (app_type) {
					case AST_SIP_APP_UDP_BFCP:
						switch_case = session->reinvite_attr.reinvite_attr_bfcp;
						break;
					case AST_SIP_APP_TCP_BFCP:
						switch_case = session->reinvite_attr.reinvite_attr_bfcp;
						break;
					case AST_SIP_APP_RTP:
						if (ast_sip_session_judge_if_fec_stream(session, im)) {
							switch_case = session->reinvite_attr.reinvite_attr_fec;
							break;
						} else if (ast_sip_session_judge_if_fecc_stream(session, im)) {
						    switch_case = session->reinvite_attr.reinvite_attr_fecc;
							break;
						} else {
							ast_log(LOG_DEBUG, "Don't have rtpmap, just deactivate media");
							break;
						}
						break;
					default:
						ast_log(LOG_ERROR, "Can't process other application media, ignore this");
						break;
				}
				switch_case = session->reinvite_attr.reinvite_attr_audio;
			}
			ast_log(LOG_DEBUG, "switch_case = %d\n", switch_case);
			switch (switch_case) {
				case REINVITE_SENDONLY:
					attr = pjmedia_sdp_attr_create(inv->pool_prov, "sendonly", NULL);
					break;
				case REINVITE_INACTIVE:
					attr = pjmedia_sdp_attr_create(inv->pool_prov, "inactive", NULL);
					break;
				default:
					attr = pjmedia_sdp_attr_create(inv->pool_prov, "sendrecv", NULL);
					break;
			}
			pjmedia_sdp_media_add_attr(im, attr);
		}
	}
	ast_reinvite_attr_init(session);
#endif
	/* Use the connection details of the first media stream if possible for SDP level */
	if (local->media_count) {
		int stream;

		/* Since we are using the first media stream as the SDP level we can get rid of it
		 * from the stream itself
		 */
		local->conn = local->media[0]->conn;
		local->media[0]->conn = NULL;
		pj_strassign(&local->origin.net_type, &local->conn->net_type);
		pj_strassign(&local->origin.addr_type, &local->conn->addr_type);
		pj_strassign(&local->origin.addr, &local->conn->addr);

		/* Go through each media stream seeing if the connection details actually differ,
		 * if not just use SDP level and reduce the SDP size
		 */
		for (stream = 1; stream < local->media_count; stream++) {
#ifdef GRANDSTREAM_NETWORKS
			/* Add by bfdu for BFCP, 2014.07.20 */
			if (local->conn && local->media[stream]->conn &&
#else
			if (
#endif
				!pj_strcmp(&local->conn->net_type, &local->media[stream]->conn->net_type) &&
				!pj_strcmp(&local->conn->addr_type, &local->media[stream]->conn->addr_type) &&
				!pj_strcmp(&local->conn->addr, &local->media[stream]->conn->addr)) {
				local->media[stream]->conn = NULL;
			}
		}
	} else {
		local->origin.net_type = STR_IN;
		local->origin.addr_type = session->endpoint->media.rtp.ipv6 ? STR_IP6 : STR_IP4;

		if (!ast_strlen_zero(session->endpoint->media.address)) {
			pj_strdup2(inv->pool_prov, &local->origin.addr, session->endpoint->media.address);
		} else {
			pj_sockaddr localaddr;
			char our_ip[PJ_INET6_ADDRSTRLEN];

			pj_gethostip(session->endpoint->media.rtp.ipv6 ? pj_AF_INET6() : pj_AF_INET(), &localaddr);
			pj_sockaddr_print(&localaddr, our_ip, sizeof(our_ip), 0);
			pj_strdup2(inv->pool_prov, &local->origin.addr, our_ip);
		}
	}

	return local;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for BFCP, 2015.07.21 */
static void init_bfcp_para(struct ast_sip_session *session)
{
	session->is_video_media_change = 0;		/* init video reinvite flag */
	memset(&session->bfcp_para, 0, sizeof(struct ast_bfcp_para));
	memset(&session->main_video_para, 0, sizeof(struct ast_main_video_para));
	memset(&session->slides_video_para, 0, sizeof(struct ast_slides_video_para));
	if (session->slides_video_req_caps) {
		ast_format_cap_remove_by_type(session->slides_video_req_caps, AST_MEDIA_TYPE_UNKNOWN);
	}
	return;
}

static void init_fec_para(struct ast_sip_session *session)
{
	memset(&session->fec_para, 0, sizeof(struct ast_fec_para));
	if (session->fec_req_caps) {
		ast_format_cap_remove_by_type(session->fec_req_caps, AST_MEDIA_TYPE_UNKNOWN);
	}
	return;
}

static void init_fecc_para(struct ast_sip_session *session)
{
	memset(&session->fecc_para, 0, sizeof(struct ast_fecc_para));
	if (session->fecc_req_caps) {
		ast_format_cap_remove_by_type(session->fecc_req_caps, AST_MEDIA_TYPE_UNKNOWN);
	}
	return;
}

static void init_fec_multiplexed_para(struct ast_sip_session *session)
{
	memset(&session->fec_multiplexed_para, 0, sizeof(struct ast_fec_multiplexed_para));
	if (session->fec_multiplexed_req_caps) {
		ast_format_cap_remove_by_type(session->fec_multiplexed_req_caps, AST_MEDIA_TYPE_UNKNOWN);
	}
	return;
}

static void init_yealink_para(struct ast_sip_session *session)
{
	session->is_yealink_msg = FALSE;
	return;
}
#endif

static void session_inv_on_rx_offer(pjsip_inv_session *inv, const pjmedia_sdp_session *offer)
{
	struct ast_sip_session *session = inv->mod_data[session_module.id];
	pjmedia_sdp_session *answer;

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "session_inv_on_rx_offer(),enter ...\n");
	session->is_outgoing = 0;
	/* Begin add by bfdu for video reinvite, 2015.05.25 */
	/*if (session->req_caps) {
		ast_format_cap_remove_by_type(session->req_caps, AST_MEDIA_TYPE_UNKNOWN);
	}*/
	init_bfcp_para(session);
	init_fec_para(session);
	init_fecc_para(session);
	ast_reinvite_attr_init(session);
	init_fec_multiplexed_para(session);
	init_yealink_para(session);
	/* End add by bfdu for video reinvite, 2015.05.25 */
#endif
	if (handle_incoming_sdp(session, offer)) {
		return;
	}

	if ((answer = create_local_sdp(inv, session, offer))) {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_VERBOSE, "'%s' two media(audo and image), because '%s'!\n",
			session->is_image_audio_two_media ? "Create" : "Not Create",
			session->two_media_is_support_image ? "support" : "not support");
			session_search_audio_port_with_image_audio(session, session->is_image_audio_two_media,
			session->two_media_is_support_image, answer, "audio");
#endif
		pjsip_inv_set_sdp_answer(inv, answer);
	}
}

#if 0
static void session_inv_on_create_offer(pjsip_inv_session *inv, pjmedia_sdp_session **p_offer)
{
	/* XXX STUB */
}
#endif

static void session_inv_on_media_update(pjsip_inv_session *inv, pj_status_t status)
{
	struct ast_sip_session *session = inv->mod_data[session_module.id];
	const pjmedia_sdp_session *local, *remote;

	if (!session->channel) {
		/* If we don't have a channel. We really don't care about media updates.
		 * Just ignore
		 */
		return;
	}

	if ((status != PJ_SUCCESS) || (pjmedia_sdp_neg_get_active_local(inv->neg, &local) != PJ_SUCCESS) ||
		(pjmedia_sdp_neg_get_active_remote(inv->neg, &remote) != PJ_SUCCESS)) {
		ast_channel_hangupcause_set(session->channel, AST_CAUSE_BEARERCAPABILITY_NOTAVAIL);
		ast_set_hangupsource(session->channel, ast_channel_name(session->channel), 0);
		ast_queue_hangup(session->channel);
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	if(handle_negotiated_sdp(session, local, remote))
	{
		ast_channel_hangupcause_set(session->channel, AST_CAUSE_BEARERCAPABILITY_NOTAVAIL);
		ast_set_hangupsource(session->channel, ast_channel_name(session->channel), 0);
		ast_queue_hangup(session->channel);
		return;
	}
#else
	handle_negotiated_sdp(session, local, remote);
#endif
}

static pjsip_redirect_op session_inv_on_redirected(pjsip_inv_session *inv, const pjsip_uri *target, const pjsip_event *e)
{
	struct ast_sip_session *session = inv->mod_data[session_module.id];
	const pjsip_sip_uri *uri;

#ifdef GRANDSTREAM_NETWORKS
	if ((session == NULL) || (ao2_ref(session, 0) <= 0) || (session->channel == NULL)) {
		ast_log(LOG_ERROR, "Invalid SIP session or channel on redirected!");
		return PJSIP_REDIRECT_STOP;
	}
#endif
	if (session->endpoint->redirect_method == AST_SIP_REDIRECT_URI_PJSIP) {
		return PJSIP_REDIRECT_ACCEPT;
	}

	if (!PJSIP_URI_SCHEME_IS_SIP(target) && !PJSIP_URI_SCHEME_IS_SIPS(target)) {
		return PJSIP_REDIRECT_STOP;
	}
	handle_incoming(session, e->body.rx_msg.rdata, PJSIP_EVENT_RX_MSG,
			AST_SIP_SESSION_BEFORE_REDIRECTING);

	uri = pjsip_uri_get_uri(target);

	if (session->endpoint->redirect_method == AST_SIP_REDIRECT_USER) {
		char exten[AST_MAX_EXTENSION];

		ast_copy_pj_str(exten, &uri->user, sizeof(exten));
		ast_channel_call_forward_set(session->channel, exten);
	} else if (session->endpoint->redirect_method == AST_SIP_REDIRECT_URI_CORE) {
		char target_uri[PJSIP_MAX_URL_SIZE];
		/* PJSIP/ + endpoint length + / + max URL size */
		char forward[8 + strlen(ast_sorcery_object_get_id(session->endpoint)) + PJSIP_MAX_URL_SIZE];

		pjsip_uri_print(PJSIP_URI_IN_REQ_URI, uri, target_uri, sizeof(target_uri));
		sprintf(forward, "PJSIP/%s/%s", ast_sorcery_object_get_id(session->endpoint), target_uri);
		ast_channel_call_forward_set(session->channel, forward);
	}

	return PJSIP_REDIRECT_STOP;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for video reinivite */
static void handle_request_video_reinvite(struct ast_sip_session *session, struct ast_sip_video_reinvite_data *reinvite_data)
{
	ast_log(LOG_DEBUG, "handle_request_video_reinvite start\n");
	if (session && session->channel && ast_channel_tech(session->channel) && reinvite_data) {
		if (0 == strcmp(ast_channel_tech(session->channel)->type, "PJSIP")) {
			memcpy(&reinvite_data->reinvite_session, &session, sizeof(reinvite_data->reinvite_session));
			ast_log(LOG_DEBUG, "handle video re-invite req\n");
		}
		else {
			ast_log(LOG_WARNING, "channel type is not SIP, can not handle video reinvite req\n");
		}
	} else {
		ast_log(LOG_WARNING, "param illegal, can not handle video re-invite\n");
	}
	return;
}

static unsigned char judge_if_support_video_media(struct ast_sip_session *session)
{
	ast_log(LOG_DEBUG, "judge_if_support_video_media begin\n");
	unsigned char res = FALSE;

	RAII_VAR(struct ast_sip_session_media *, session_media, NULL, ao2_cleanup);

	session_media = ao2_find(session->media, "video", OBJ_KEY);
	if (!session_media) {
		/* if the session_media doesn't exist, there weren't
		 * any handlers at the time of its creation */
		return FALSE;
	}

	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate %s capabilities\n", session_media->stream_type);
		res = FALSE;
	}

	ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_VIDEO);

	if (!ast_format_cap_count(caps)) {
		ast_log(LOG_WARNING, "endpoint has no video cap\n");
		res = FALSE;
	} else {
		res = TRUE;
	}

	return res;
}

static unsigned char judge_if_both_type_is_sip(struct ast_channel* chan)
{
	if (NULL == chan){
		ast_log(LOG_WARNING, "Judge if both type is sip, chan is NULL\n");
		return FALSE;
	}
	RAII_VAR(struct ast_channel *, bridged, ast_channel_bridge_peer(chan), ast_channel_cleanup);
	if (NULL == bridged) {
		ast_log(LOG_DEBUG, "Judge if both type is sip, bridge is not exsit\n");
		return FALSE;
	}
	if (ast_channel_tech(bridged) && (0 == strcmp(ast_channel_tech(bridged)->type, "PJSIP")) &&
		ast_channel_tech(chan) && (0 == strcmp(ast_channel_tech(chan)->type, "PJSIP"))) {
		return TRUE;
	} else {
		ast_log(LOG_DEBUG, "Judge if both type is sip, type is not PJSIP\n");
		return FALSE;
	}
}

static void get_codecs(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream, struct ast_rtp_codecs *codecs)
{
	pjmedia_sdp_attr *attr;
	pjmedia_sdp_rtpmap *rtpmap;
	pjmedia_sdp_fmtp fmtp;
	struct ast_format *format;
	int i, num = 0;
	char name[256];
	char media[20];
	char fmt_param[256];

	ast_rtp_codecs_payloads_initialize(codecs);

	/* Iterate through provided formats */
	for (i = 0; i < stream->desc.fmt_count; ++i) {
		/* The payload is kept as a string for things like t38 but for video it is always numerical */
		ast_rtp_codecs_payloads_set_m_type(codecs, NULL, pj_strtoul(&stream->desc.fmt[i]));
		/* Look for the optional rtpmap attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "rtpmap", &stream->desc.fmt[i]))) {
			continue;
		}

		/* Interpret the attribute as an rtpmap */
		if ((pjmedia_sdp_attr_to_rtpmap(session->inv_session->pool_prov, attr, &rtpmap)) != PJ_SUCCESS) {
			continue;
		}

		ast_copy_pj_str(name, &rtpmap->enc_name, sizeof(name));
		ast_copy_pj_str(media, (pj_str_t*)&stream->desc.media, sizeof(media));
		ast_rtp_codecs_payloads_set_rtpmap_type_rate(codecs, NULL, pj_strtoul(&stream->desc.fmt[i]),
							     media, name, 0, rtpmap->clock_rate);
		/* Look for an optional associated fmtp attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "fmtp", &rtpmap->pt))) {
			continue;
		}

		if ((pjmedia_sdp_attr_get_fmtp(attr, &fmtp)) == PJ_SUCCESS) {
			ast_copy_pj_str(fmt_param, &fmtp.fmt, sizeof(fmt_param));
			if (sscanf(fmt_param, "%30d", &num) != 1) {
				continue;
			}

			if ((format = ast_rtp_codecs_get_payload_format(codecs, num))) {
				struct ast_format *format_parsed;

				ast_copy_pj_str(fmt_param, &fmtp.fmt_param, sizeof(fmt_param));

				format_parsed = ast_format_parse_sdp_fmtp(format, fmt_param);
				if (format_parsed) {
					ast_rtp_codecs_payload_replace_format(codecs, num, format_parsed);
					ao2_ref(format_parsed, -1);
				}

				ao2_ref(format, -1);
			}
		}
	}

	/* Get the packetization, if it exists */
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "ptime", NULL))) {
		unsigned long framing = pj_strtoul(pj_strltrim(&attr->value));
		if (framing && session->endpoint->media.rtp.use_ptime) {
			ast_rtp_codecs_set_framing(codecs, framing);
		}
	}
}

struct pjmedia_sdp_neg
{
    pjmedia_sdp_neg_state state;	    /**< Negotiator state.	     */
    pj_bool_t			prefer_remote_codec_order;
    pj_bool_t			answer_with_multiple_codecs;
    pj_bool_t			has_remote_answer;
    pj_bool_t			answer_was_remote;

    pjmedia_sdp_session	*initial_sdp,	    /**< Initial local SDP	     */
			*active_local_sdp,  /**< Currently active local SDP. */
			*active_remote_sdp, /**< Currently active remote's.  */
			*neg_local_sdp,	    /**< Temporary local SDP.	     */
			*neg_remote_sdp;    /**< Temporary remote SDP.	     */
};

static int compare_preanswer_offer_sdp(const pjmedia_sdp_media *preanswer, const pjmedia_sdp_media *offer, 
				struct ast_sip_session *session, const char* media_type,
				const pjmedia_sdp_session *media_session)
{
	unsigned i = 0;
	RAII_VAR(struct ast_sip_session_media *, session_media, NULL, ao2_cleanup);
	session_media = ao2_find(session->media, "audio", OBJ_KEY);

	/* If offer has zero port, just clone the offer */
	if ((offer->desc.port == 0 && preanswer->desc.port != 0)
		|| (offer->desc.port != 0 && preanswer->desc.port == 0)) {
		ast_log(LOG_DEBUG, "compare_offer, offer port %d, preanswer port %d\n", offer->desc.port, preanswer->desc.port);
		return 1;
	}
	/* need to compare video framerate, if change, reinvite should send to peer */
	pjmedia_sdp_attr *pre_attr = NULL;
	pjmedia_sdp_attr *offer_attr = NULL;
	unsigned int pre_framerate = 0;
	unsigned int offer_framerate = 0;
	if ((pre_attr = pjmedia_sdp_media_find_attr2(preanswer, "framerate", NULL))) {
		pre_framerate = pj_strtoul(pj_strltrim(&pre_attr->value));
	}
	if ((offer_attr = pjmedia_sdp_media_find_attr2(offer, "framerate", NULL))) {
		offer_framerate = pj_strtoul(pj_strltrim(&offer_attr->value));
	}
	if (pre_framerate != offer_framerate) {
		ast_log(LOG_DEBUG, "framerate is not the same, pre:%d, offer:%d\n", pre_framerate, offer_framerate);
		return 1;
	}

	unsigned j;
	pjmedia_sdp_media *im;

	for (im = NULL, j = 0; j < media_session->media_count; ++j) {
		enum ast_reinvite_attr switch_case = REINVITE_SENDRECV;
		im = media_session->media[j];

		if (pjmedia_sdp_media_find_attr2(im, "sendonly", NULL)) {
			switch_case = REINVITE_SENDONLY;
			ast_log(LOG_CODEC_NEG, "the offer with a=sendonly will send reinvite to peer\n");
		} else if (pjmedia_sdp_media_find_attr2(im, "inactive", NULL)) {
			switch_case = REINVITE_INACTIVE;
			ast_log(LOG_CODEC_NEG, "the offer with a=inactive will send reinvite to peer\n");
		} else if (pjmedia_sdp_media_find_attr2(im, "sendrecv", NULL) && session_media->held) {
			switch_case = REINVITE_SENDRECV;
			ast_log(LOG_CODEC_NEG, "the offer that session is held will send reinvite to peer\n");
			if (session->is_sendonly_to_peer == 1) {
				session->is_sendonly_to_peer = 0;
				return 1;
			}
		}
		char m_type[16];
		memset(m_type, 0, sizeof(m_type));
		ast_copy_pj_str(m_type, &im->desc.media, sizeof(m_type));
		if (strncasecmp(m_type, "audio", 5) == 0) {
			session->reinvite_attr.reinvite_attr_audio = switch_case;
		} else if (strncasecmp(m_type, "video", 5) == 0) {
			pjmedia_sdp_attr *tmp_attr;
			char content[8];
			memset(content, 0, sizeof(content));
			if ((tmp_attr = pjmedia_sdp_media_find_attr2(im, "content", NULL))) {
				ast_copy_pj_str(content, pj_strltrim(&tmp_attr->value), sizeof(content));
				if (0 != strlen(content)) {
					if (!strncmp(content, "slides", 6)) {
						session->reinvite_attr.reinvite_attr_s_video = switch_case;
					} else {
						session->reinvite_attr.reinvite_attr_m_video = switch_case;
					}
				}
			} else {
				session->reinvite_attr.reinvite_attr_m_video = switch_case;
			}
		} else if (strncasecmp(m_type, "application",11) == 0) {
			enum ast_sip_application_type app_type = ast_sip_session_get_application_type(im->desc.transport);
			switch (app_type) {
				case AST_SIP_APP_UDP_BFCP:
					session->reinvite_attr.reinvite_attr_bfcp = switch_case;
					break;
				case AST_SIP_APP_TCP_BFCP:
					session->reinvite_attr.reinvite_attr_bfcp = switch_case;
					break;
				case AST_SIP_APP_RTP:
					if (ast_sip_session_judge_if_fec_stream(session, im)) {
						session->reinvite_attr.reinvite_attr_fec = switch_case;
						break;
					} else if (ast_sip_session_judge_if_fec_stream(session, im)) {
					    session->reinvite_attr.reinvite_attr_fecc = switch_case;
						break;
					} else {
						ast_log(LOG_DEBUG, "Don't have rtpmap, just deactivate media\n");
						break;
					}
					break;
				default:
					ast_log(LOG_ERROR, "Can't process other application media, ignore this\n");
					break;
			}
		}
	}

	/* If hold request has bfcp info, need send to peer
	 * GVC hold request slides video attr is inactive, may also not contain main video,
	 * so we just think audio hold, all media will hold
	 * Todo: need to distinguish different media type as not a hold request, 2015.12.29
	 */
	if ((session->reinvite_attr.reinvite_attr_audio == REINVITE_SENDONLY) &&
		(session->bfcp_para.has_bfcp || session->bfcp_para.has_tcp_bfcp)) {
		ast_log(LOG_DEBUG, "This hold request has bfcp info, need send to peer\n");
		return 1;
	}

	for (i = 0; i < preanswer->desc.fmt_count; ++i) {
		if (pj_isdigit(*preanswer->desc.fmt[i].ptr)) {
			/* This is normal/standard payload type, where it's identified
			 * by payload number.
			 */
			unsigned preanswer_pt;
			preanswer_pt = pj_strtoul(&preanswer->desc.fmt[i]);

			/* if fmt's value\order\num is not same, send reinvite to peer */
			if (preanswer_pt < 96) {
				if (i < offer->desc.fmt_count) {
					unsigned offer_pt;
					offer_pt = pj_strtoul(&offer->desc.fmt[i]);
					if (preanswer_pt == offer_pt) {
						ast_log(LOG_CODEC_NEG, "payload match\n");
					} else {
						ast_log(LOG_CODEC_NEG, "payload value don't match, postion %d, preanswer: %d, offer: %d\n", i, preanswer_pt, offer_pt);
						return 1;
					}
				} else {
					ast_log(LOG_CODEC_NEG, "preanswer(%d) and offer(%d) is media fmt num is not the same\n", preanswer->desc.fmt_count, offer->desc.fmt_count);
					return 1;
				}
			} else {
				/* This is dynamic payload type.
				 * For dynamic payload type, we must look the rtpmap and
				 * compare the encoding name.
				 */
				const pjmedia_sdp_attr *a;
				pjmedia_sdp_rtpmap or_;

				/* Get the rtpmap for the payload type in the master. */
				a = pjmedia_sdp_media_find_attr2(preanswer, "rtpmap",
								 &preanswer->desc.fmt[i]);
				if (!a) {
					pj_assert(!"Bug! Offer should have been validated");
					return 0;
				}
				pjmedia_sdp_attr_get_rtpmap(a, &or_);

				if (i < offer->desc.fmt_count) {
					a = pjmedia_sdp_media_find_attr2(offer, "rtpmap",
									 &offer->desc.fmt[i]);
					if (a) {
						pjmedia_sdp_rtpmap lr;
						pjmedia_sdp_attr_get_rtpmap(a, &lr);
						if (!(!pj_stricmp(&or_.enc_name, &lr.enc_name) &&
							or_.clock_rate == lr.clock_rate &&
							(pj_stricmp(&or_.param, &lr.param)==0 ||
							 (lr.param.slen==0 && or_.param.slen==1 &&
									 *or_.param.ptr=='1') ||
							 (or_.param.slen==0 && lr.param.slen==1 &&
									  *lr.param.ptr=='1'))))
						{
							/* Not Match! */
							ast_log(LOG_CODEC_NEG, "rtpmap param not match\n");
							return 1;
						}

						struct ast_rtp_codecs pre_codecs = AST_RTP_CODECS_NULL_INIT;
						struct ast_rtp_codecs offer_codecs = AST_RTP_CODECS_NULL_INIT;;

						get_codecs(session, preanswer, &pre_codecs);
						get_codecs(session, offer, &offer_codecs);

						unsigned preanswer_pt;
						unsigned offer_pt;
						preanswer_pt = pj_strtoul(&preanswer->desc.fmt[i]);
						offer_pt = pj_strtoul(&offer->desc.fmt[i]);

						struct ast_format *pre_format = NULL;
						struct ast_format *offer_format = NULL;

						pre_format = ast_rtp_codecs_get_payload_format(&pre_codecs, preanswer_pt);
						offer_format = ast_rtp_codecs_get_payload_format(&offer_codecs, offer_pt);
						if (pre_format && offer_format) {
							if (AST_FORMAT_CMP_NOT_EQUAL == ast_format_cmp(pre_format, offer_format)) {
								ast_log(LOG_CODEC_NEG, "preanswer and offer format are not equal\n");
								ao2_ref(pre_format, -1);
								ao2_ref(offer_format, -1);
								ast_rtp_codecs_payloads_destroy(&pre_codecs);
								ast_rtp_codecs_payloads_destroy(&offer_codecs);
								return 1;
							} else {
								ao2_ref(pre_format, -1);
								ao2_ref(offer_format, -1);
							}
						}

						ast_rtp_codecs_payloads_destroy(&pre_codecs);
						ast_rtp_codecs_payloads_destroy(&offer_codecs);
					}
				} else {
					ast_log(LOG_CODEC_NEG, "dynamic payload preanswer(%d) and offer(%d) is media fmt num is not the same\n", preanswer->desc.fmt_count, offer->desc.fmt_count);
					return 1;
				}
			}
		} else {
			ast_log(LOG_CODEC_NEG, "payload is not digit, support later\n");
			return 0;
		}
	}
	return 0;
}

static int judge_if_need_send_video_reinvite_to_peer(struct ast_sip_session *session,
				  const pjmedia_sdp_session *offer,
				  const pjmedia_sdp_session *initial)
{
    unsigned i;

	if (initial->media_count != offer->media_count) {
		/* Asterisk1.8 don't set media port zero, media count may not the same */
		ast_log(LOG_DEBUG, "initial offer media_count is not same\n");
		return 1;
	}

    /* For each media line, create our answer based on our initial
     * capability.
     */
    for (i = 0; i < offer->media_count; ++i) {
		const pjmedia_sdp_media *om;	/* offer */
		const pjmedia_sdp_media *im;	/* initial media */

		om = offer->media[i];

		if (i < initial->media_count) {
			im = initial->media[i];
			/* if audio or video trun on/off SRTP, send to peer */
			if (pj_strcmp(&im->desc.transport, &om->desc.transport) != 0) {
				ast_log(LOG_DEBUG, "transport is not the same, need to send to peer\n");
				return 1;
			}
			if (pj_strcmp(&om->desc.media, &im->desc.media)) {
				ast_log(LOG_DEBUG, "media type is not the same, need to send to peer\n");
				return 1;
			}
			char media_type[16];
			memset(media_type, 0, sizeof(media_type));
			ast_copy_pj_str(media_type, &om->desc.media, sizeof(media_type));
		    if (pj_strcmp(&om->desc.media, &im->desc.media) == 0 &&
				(strncmp(media_type, "video", 5) == 0 ||
				strncmp(media_type, "audio", 5) == 0)) {
				/* See if it has matched. */
				int res;
				res = compare_preanswer_offer_sdp(im, om, session, media_type, offer);
				if (res == 1) {
					ast_log(LOG_DEBUG, "initial offer sdp is not same\n");
					return 1;
			    }
		    }
		} else {
			ast_log(LOG_DEBUG, "initial offer sdp is not same\n");
			return 1;
		}
    }

    return 0;
}

static int judge_if_sdp_has_image(const pjmedia_sdp_session *offer)
{
    unsigned i;

    for (i = 0; i < offer->media_count; ++i) {
		const pjmedia_sdp_media *om;	/* offer */
		om = offer->media[i];
		char media_type[16];
		memset(media_type, 0, sizeof(media_type));
		ast_copy_pj_str(media_type, &om->desc.media, sizeof(media_type));
	    if (strncmp(media_type, "image", 5) == 0) {
			ast_log(LOG_DEBUG, "offer sdp has image, don't send to peer\n");
			return 1;
	    }
    }

    return 0;
}

int ast_sip_session_judge_if_slides_video(const pjmedia_sdp_media *offer_media)
{
	pjmedia_sdp_attr *attr;
	char content[8];
	memset(content, 0, sizeof(content));
	if ((attr = pjmedia_sdp_media_find_attr2(offer_media, "content", NULL))) {
		ast_copy_pj_str(content, pj_strltrim(&attr->value), sizeof(content));
		if (0 != strlen(content)) {
			if (!strncmp(content, "slides", 6)) {
				ast_log(LOG_DEBUG, "content is slides\n");
				return TRUE;
			}
		}
	}

	return FALSE;
}

int ast_sip_session_judge_if_fec_stream(const struct ast_sip_session *session, const struct pjmedia_sdp_media *stream)
{
	pjmedia_sdp_attr *attr;
	int i = 0;
	const pj_str_t str_gs_fec = { "GS-FEC", 6 };

	if (!session || !stream) {
		ast_log(LOG_WARNING, "Judge fec stream, param is null\n");
		return FALSE;
	}

	for (i = 0; i < stream->desc.fmt_count; ++i) {
		/* Look for the optional rtpmap attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "rtpmap", &stream->desc.fmt[i]))) {
			continue;
		}

		if (pj_strstr(&attr->value, &str_gs_fec) != NULL) {
			return TRUE;
		}
	}
	return FALSE;
}

int ast_sip_session_judge_if_fecc_stream(const struct ast_sip_session *session, const struct pjmedia_sdp_media *stream)
{
	pjmedia_sdp_attr *attr;
	int i = 0;
	const pj_str_t str_fecc = { "H224", 4 };

	if (!session || !stream) {
		ast_log(LOG_WARNING, "Judge fecc stream, param is null\n");
		return FALSE;
	}

	for (i = 0; i < stream->desc.fmt_count; ++i) {
		/* Look for the optional rtpmap attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "rtpmap", &stream->desc.fmt[i]))) {
			continue;
		}

		if (pj_strstr(&attr->value, &str_fecc) != NULL) {
			return TRUE;
		}
	}
	return FALSE;
}


static int change_ssrc_for_hold(struct ast_sip_session *session, const pjmedia_sdp_session *offer)
{
	if (!session || !offer) {
		ast_log(LOG_WARNING, "Change ssrc for hold, param is invalid\n");
		return 0;
	}
	RAII_VAR(struct ast_sip_session_media *, audio_media, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_session_media *, video_media, NULL, ao2_cleanup);
	int i = 0, is_changed = 0;
	ao2_ref(session, +1);

	audio_media = ao2_find(session->media, "audio", OBJ_KEY);
	video_media = ao2_find(session->media, "video", OBJ_KEY);

	for (i = 0; i < offer->media_count; ++i) {
		const pjmedia_sdp_media *om;
		om = offer->media[i];
		if (NULL == pjmedia_sdp_media_find_attr2(om, "sendonly", NULL)) {
			continue;
		}

		char media_type[16];
		memset(media_type, 0, sizeof(media_type));
		ast_copy_pj_str(media_type, &om->desc.media, sizeof(media_type));

		if (strncmp(media_type, "audio", 5) == 0) {
			if (audio_media && audio_media->rtp) {
				ast_log(LOG_DEBUG, "audio change source for hold\n");
				ast_rtp_instance_change_source(audio_media->rtp);
				is_changed = 1;
			}
		}

		if (strncmp(media_type, "video", 5) == 0) {
			if (!ast_sip_session_judge_if_slides_video(om)) {
				if (video_media && video_media->rtp) {
					ast_log(LOG_DEBUG, "video change source for hold\n");
					ast_rtp_instance_change_source(video_media->rtp);
					is_changed = 1;
				}
			} else {
				if (session->slides_video_media.rtp) {
					ast_log(LOG_DEBUG, "slides video change source for hold\n");
					ast_rtp_instance_change_source(session->slides_video_media.rtp);
					is_changed = 1;
				}
			}
		}

		if (ast_sip_session_judge_if_fecc_stream(session, om)) {
			if (session->fecc_media.rtp) {
				ast_log(LOG_DEBUG, "fecc change source for hold\n");
				ast_rtp_instance_change_source(session->fecc_media.rtp);
				is_changed = 1;
			}
		}

		if (ast_sip_session_judge_if_fec_stream(session, om)) {
			if (session->fec_media.rtp) {
				ast_log(LOG_DEBUG, "fec change source for hold\n");
				ast_rtp_instance_change_source(session->fec_media.rtp);
				if (video_media && video_media->rtp) {
					ast_rtp_instance_set_ssrc(session->fec_media.rtp, video_media->rtp);
					is_changed = 1;
				}
			}
		}
	}
	ao2_ref(session, -1);

	return is_changed;
}

static int change_ssrc_for_unhold(struct ast_sip_session *session, const pjmedia_sdp_session *offer)
{
	if (!session || !offer) {
		ast_log(LOG_WARNING, "Change ssrc for unhold, param is invalid\n");
		return 0;
	}
	RAII_VAR(struct ast_sip_session_media *, audio_media, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_session_media *, video_media, NULL, ao2_cleanup);
	int i = 0, is_changed = 0;
	ao2_ref(session, +1);

	audio_media = ao2_find(session->media, "audio", OBJ_KEY);
	video_media = ao2_find(session->media, "video", OBJ_KEY);

	for (i = 0; i < offer->media_count; ++i) {
		const pjmedia_sdp_media *om;
		om = offer->media[i];

		/* if is not a unhold operate, just return */
		if (NULL == pjmedia_sdp_media_find_attr2(om, "sendrecv", NULL)) {
			continue;
		}

		char media_type[16];
		memset(media_type, 0, sizeof(media_type));
		ast_copy_pj_str(media_type, &om->desc.media, sizeof(media_type));

		if (strncmp(media_type, "audio", 5) == 0) {
			if (audio_media && audio_media->rtp && audio_media->held) {
				ast_log(LOG_DEBUG, "audio change source for unhold\n");
				ast_rtp_instance_change_source(audio_media->rtp);
				is_changed = 1;
			}
		}

		if (strncmp(media_type, "video", 5) == 0) {
			if (!ast_sip_session_judge_if_slides_video(om)) {
				if (video_media && video_media->rtp && video_media->held) {
					ast_log(LOG_DEBUG, "video change source for unhold\n");
					ast_rtp_instance_change_source(video_media->rtp);
					is_changed = 1;
				}
			} else {
				if (session->slides_video_media.rtp && session->slides_video_media.held) {
					ast_log(LOG_DEBUG, "slides video change source for unhold\n");
					ast_rtp_instance_change_source(session->slides_video_media.rtp);
					is_changed = 1;
				}
			}
		}

		if (ast_sip_session_judge_if_fecc_stream(session, om)) {
			if (session->fecc_media.rtp && session->fecc_media.held) {
				ast_log(LOG_DEBUG, "fecc change source for unhold\n");
				ast_rtp_instance_change_source(session->fecc_media.rtp);
				is_changed = 1;
			}
		}

		if (ast_sip_session_judge_if_fec_stream(session, om)) {
			if (session->fec_media.rtp && session->fec_media.held) {
				ast_log(LOG_DEBUG, "fec change source for unhold\n");
				ast_rtp_instance_change_source(session->fec_media.rtp);
				if (video_media && video_media->rtp) {
					ast_rtp_instance_set_ssrc(session->fec_media.rtp, video_media->rtp);
					is_changed = 1;
				}
			}
		}
	}
	ao2_ref(session, -1);

	return is_changed;
}

static void change_ssrc_for_media_change(struct ast_sip_session *session)
{
	if (!session) {
		ast_log(LOG_WARNING, "Change ssrc for media change, param is invalid\n");
		return;
	}

	RAII_VAR(struct ast_sip_session_media *, audio_media, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_session_media *, video_media, NULL, ao2_cleanup);

	ao2_ref(session, +1);

	audio_media = ao2_find(session->media, "audio", OBJ_KEY);
	if (audio_media && audio_media->rtp) {
		ast_log(LOG_DEBUG, "Audio change source for media change\n");
		ast_rtp_instance_change_source(audio_media->rtp);
	}

	video_media = ao2_find(session->media, "video", OBJ_KEY);
	if (video_media && video_media->rtp) {
		ast_log(LOG_DEBUG, "video change source for media change\n");
		ast_rtp_instance_change_source(video_media->rtp);
	}

	if (session->slides_video_media.rtp) {
		ast_log(LOG_DEBUG, "slides video change source for media change\n");
		ast_rtp_instance_change_source(session->slides_video_media.rtp);
	}

	if (session->fecc_media.rtp) {
		ast_log(LOG_DEBUG, "fecc change source for media change\n");
		ast_rtp_instance_change_source(session->fecc_media.rtp);
	}

	if (session->fec_media.rtp) {
		ast_log(LOG_DEBUG, "fec change source for media change\n");
		ast_rtp_instance_change_source(session->fec_media.rtp);
		if (video_media && video_media->rtp) {
			ast_rtp_instance_set_ssrc(session->fec_media.rtp, video_media->rtp);
		}
	}

	ao2_ref(session, -1);

	return;
}

static pj_status_t session_inv_on_rx_reinvite(pjsip_inv_session *inv, const pjmedia_sdp_session *offer, pjsip_rx_data *rdata)
{
	if (NULL == inv || NULL == offer || NULL == rdata) {
		ast_log(LOG_DEBUG, "session_inv_on_rx_reinvite, param is NULL\n");
		return PJ_TRUE;
	}
	ast_log(LOG_DEBUG, "session_inv_on_rx_reinvite start ...\n");

	if(judge_if_sdp_has_image(offer)) {
		return PJ_TRUE;
	}

	if (pjmedia_sdp_neg_get_state(inv->neg) != PJMEDIA_SDP_NEG_STATE_WAIT_NEGO) {
       	ast_log(LOG_DEBUG, "session_inv_on_rx_reinvite, neg state is not wait neg");
		return PJ_TRUE;
    }

	struct ast_sip_session *session = inv->mod_data[session_module.id];
	if (NULL == session) {
		ast_log(LOG_DEBUG, "session_inv_on_rx_reinvite, video media change is false");
		return PJ_TRUE;
	}
	pjsip_tx_data *tdata = NULL;
	if (pjsip_inv_initial_answer(session->inv_session, rdata, 100, NULL, NULL, &tdata) != PJ_SUCCESS) {
		pjsip_inv_terminate(session->inv_session, 500, PJ_TRUE);
		return PJ_TRUE;
	}
	ast_sip_session_send_response(session, tdata);

	if (!session->inv_session || !session->inv_session->neg) {
		ast_log(LOG_WARNING, "session_inv_on_rx_reinvite, inv_session or neg is null\n");
		return PJ_TRUE;
	}

	pjmedia_sdp_session *caller_sdp_local, *caller_sdp_neg_local;
	int send_video_reinvite = 0;

	caller_sdp_local = session->inv_session->neg->active_local_sdp;
	caller_sdp_neg_local = session->inv_session->neg->neg_local_sdp;

	if (!caller_sdp_local || !caller_sdp_neg_local) {
		ast_log(LOG_WARNING, "session_inv_on_rx_reinvite, active_local_sdp or neg_local_sdp is null\n");
		return PJ_TRUE;
	}

	char buf[1024];
	{
		memset(buf, 0, sizeof(buf));
		pjmedia_sdp_print(caller_sdp_neg_local, buf, sizeof(buf) - 1); //Must use sizeof(buf) - 1, otherwise there is a potential overflow
		ast_log(LOG_CODEC_NEG, "caller_sdp_neg_local sdp is:\n%s", buf);

		memset(buf, 0, sizeof(buf));
		pjmedia_sdp_print(caller_sdp_local, buf, sizeof(buf) - 1); //Must use sizeof(buf) - 1, otherwise there is a potential overflow
		ast_log(LOG_CODEC_NEG, "initial sdp is:\n%s", buf);

		memset(buf, 0, sizeof(buf));
		pjmedia_sdp_print(offer, buf, sizeof(buf) - 1); //Must use sizeof(buf) - 1, otherwise there is a potential overflow
		ast_log(LOG_CODEC_NEG, "offer sdp is:\n%s", buf);
	}

	set_yealink_flag(rdata, session);

	/* ssrc only need to change for one time, hold maybe also a media change reinvite */
	int is_ssrc_changed = 0;
	is_ssrc_changed = change_ssrc_for_hold(session, offer);
	if (!is_ssrc_changed) {
		change_ssrc_for_unhold(session, offer);
	}

	send_video_reinvite = judge_if_need_send_video_reinvite_to_peer(session, offer, caller_sdp_local);

	if (send_video_reinvite) {
		if ((FALSE == judge_if_support_video_media(session)) ||
			(FALSE == judge_if_both_type_is_sip(session->channel))) {
			ast_log(LOG_DEBUG, "endpoint can't support video or isn't PJSIP\n");
			return PJ_TRUE;
		} else {
			session->is_video_media_change = 1;
			ast_log(LOG_DEBUG, "session_inv_on_rx_reinvite set is_video_media_change true\n");
			if (session->reinvite_attr.reinvite_attr_audio== REINVITE_SENDONLY) {
				session->is_sendonly_to_peer = 1;
			}
			if (!is_ssrc_changed) {
				change_ssrc_for_media_change(session);
			}
			ast_log(LOG_DEBUG, "session_inv_on_rx_reinvite set is_video_media_change true\n");
			struct ast_sip_video_reinvite_data reinvite_data;
			handle_request_video_reinvite(session, &reinvite_data);
			ast_queue_control_data(session->channel, AST_CONTROL_VIDEO_REINVITE_REQ, &reinvite_data, sizeof(reinvite_data));
			return PJ_SUCCESS;
		}
	} else {
		/* todo: when audio reinvite, want use other fmt, need to send to peer */
		ast_reinvite_attr_init(session);
		ast_log(LOG_DEBUG, "don't need to send reinvite to peer\n");
		return PJ_TRUE;
	}
}
#endif


static pjsip_inv_callback inv_callback = {
	.on_state_changed = session_inv_on_state_changed,
	.on_new_session = session_inv_on_new_session,
	.on_tsx_state_changed = session_inv_on_tsx_state_changed,
	.on_rx_offer = session_inv_on_rx_offer,
	.on_media_update = session_inv_on_media_update,
	.on_redirected = session_inv_on_redirected,
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for video reinivite */
	.on_rx_reinvite = session_inv_on_rx_reinvite,
#endif
};

/*! \brief Hook for modifying outgoing messages with SDP to contain the proper address information */
static void session_outgoing_nat_hook(pjsip_tx_data *tdata, struct ast_sip_transport *transport)
{
	struct ast_sip_nat_hook *hook = ast_sip_mod_data_get(
		tdata->mod_data, session_module.id, MOD_DATA_NAT_HOOK);
	struct pjmedia_sdp_session *sdp;
	int stream;

	/* SDP produced by us directly will never be multipart */
	if (hook || !tdata->msg->body || pj_stricmp2(&tdata->msg->body->content_type.type, "application") ||
		pj_stricmp2(&tdata->msg->body->content_type.subtype, "sdp") || ast_strlen_zero(transport->external_media_address)) {
		return;
	}

	sdp = tdata->msg->body->data;

	if (sdp->conn) {
		char host[NI_MAXHOST];
		struct ast_sockaddr addr = { { 0, } };

		ast_copy_pj_str(host, &sdp->conn->addr, sizeof(host));
		ast_sockaddr_parse(&addr, host, PARSE_PORT_FORBID);

		if (ast_apply_ha(transport->localnet, &addr) != AST_SENSE_ALLOW) {
			pj_strdup2(tdata->pool, &sdp->conn->addr, transport->external_media_address);
		}
	}

	for (stream = 0; stream < sdp->media_count; ++stream) {
		/* See if there are registered handlers for this media stream type */
		char media[20];
		struct ast_sip_session_sdp_handler *handler;
		RAII_VAR(struct sdp_handler_list *, handler_list, NULL, ao2_cleanup);

		/* We need a null-terminated version of the media string */
		ast_copy_pj_str(media, &sdp->media[stream]->desc.media, sizeof(media));

		handler_list = ao2_find(sdp_handlers, media, OBJ_KEY);
		if (!handler_list) {
			ast_debug(1, "No registered SDP handlers for media type '%s'\n", media);
			continue;
		}
		AST_LIST_TRAVERSE(&handler_list->list, handler, next) {
			if (handler->change_outgoing_sdp_stream_media_address) {
				handler->change_outgoing_sdp_stream_media_address(tdata, sdp->media[stream], transport);
			}
		}
	}

	/* We purposely do this so that the hook will not be invoked multiple times, ie: if a retransmit occurs */
	ast_sip_mod_data_set(tdata->pool, tdata->mod_data, session_module.id, MOD_DATA_NAT_HOOK, nat_hook);
}

#ifdef GRANDSTREAM_NETWORKS
/*add by wxu@grandstream.cn for add sip_invite_failed*/
/*! \brief Convert SIP hangup causes to Asterisk hangup causes */
static int hangup_sip2cause(int cause)
{
	/* Possible values taken from causes.h */

	switch(cause) {
	case 401:       /* Unauthorized */
		return AST_CAUSE_CALL_REJECTED;
	case 403:       /* Not found */
		return AST_CAUSE_CALL_REJECTED;
	case 404:       /* Not found */
		return AST_CAUSE_UNALLOCATED;
	case 405:       /* Method not allowed */
		return AST_CAUSE_INTERWORKING;
	case 407:       /* Proxy authentication required */
		return AST_CAUSE_CALL_REJECTED;
	case 408:       /* No reaction */
		return AST_CAUSE_NO_USER_RESPONSE;
	case 409:       /* Conflict */
		return AST_CAUSE_NORMAL_TEMPORARY_FAILURE;
	case 410:       /* Gone */
		return AST_CAUSE_NUMBER_CHANGED;
	case 411:       /* Length required */
		return AST_CAUSE_INTERWORKING;
	case 413:       /* Request entity too large */
		return AST_CAUSE_INTERWORKING;
	case 414:       /* Request URI too large */
		return AST_CAUSE_INTERWORKING;
	case 415:       /* Unsupported media type */
		return AST_CAUSE_INTERWORKING;
	case 420:       /* Bad extension */
		return AST_CAUSE_NO_ROUTE_DESTINATION;
	case 480:       /* No answer */
		return AST_CAUSE_NO_ANSWER;
	case 481:       /* No answer */
		return AST_CAUSE_INTERWORKING;
	case 482:       /* Loop detected */
		return AST_CAUSE_INTERWORKING;
	case 483:       /* Too many hops */
		return AST_CAUSE_NO_ANSWER;
	case 484:       /* Address incomplete */
		return AST_CAUSE_INVALID_NUMBER_FORMAT;
	case 485:       /* Ambiguous */
		return AST_CAUSE_UNALLOCATED;
	case 486:       /* Busy everywhere */
		return AST_CAUSE_BUSY;
	case 487:       /* Request terminated */
		return AST_CAUSE_INTERWORKING;
	case 488:       /* No codecs approved */
		return AST_CAUSE_BEARERCAPABILITY_NOTAVAIL;
	case 491:       /* Request pending */
		return AST_CAUSE_INTERWORKING;
	case 493:       /* Undecipherable */
		return AST_CAUSE_INTERWORKING;
	case 500:       /* Server internal failure */
		return AST_CAUSE_FAILURE;
	case 501:       /* Call rejected */
		return AST_CAUSE_FACILITY_REJECTED;
	case 502:
		return AST_CAUSE_DESTINATION_OUT_OF_ORDER;
	case 503:       /* Service unavailable */
		return AST_CAUSE_CONGESTION;
	case 504:       /* Gateway timeout */
		return AST_CAUSE_RECOVERY_ON_TIMER_EXPIRE;
	case 505:       /* SIP version not supported */
		return AST_CAUSE_INTERWORKING;
	case 600:       /* Busy everywhere */
		return AST_CAUSE_USER_BUSY;
	case 603:       /* Decline */
		return AST_CAUSE_CALL_REJECTED;
	case 604:       /* Does not exist anywhere */
		return AST_CAUSE_UNALLOCATED;
	case 606:       /* Not acceptable */
		return AST_CAUSE_BEARERCAPABILITY_NOTAVAIL;
	default:
		if (cause < 500 && cause >= 400) {
			/* 4xx class error that is unknown - someting wrong with our request */
			return AST_CAUSE_INTERWORKING;
		} else if (cause < 600 && cause >= 500) {
			/* 5xx class error - problem in the remote end */
			return AST_CAUSE_CONGESTION;
		} else if (cause < 700 && cause >= 600) {
			/* 6xx - global errors in the 4xx class */
			return AST_CAUSE_INTERWORKING;
		}
		return AST_CAUSE_NORMAL;
	}
	/* Never reached */
	return 0;
}

static void ast_sip_report_invite_failed(struct ast_sip_session *session,struct pjsip_rx_data *rdata,struct pjsip_tx_data *tdata,
	const char *sipmethod, int resp, const char* resp_text)
{	
	if( (!strcmp(sipmethod,"INVITE") && resp >= 400 && resp != 401) || (!strcmp(sipmethod,"BYE") && 0 == resp) ){
		char reason_text[64] = {0};
		char peer_name[64] = {0};
		char from_user[256] = {0};
		char to_user[256] = {0};
		
		struct ast_security_event_sip_invite_failed sip_invite_failed = {
			.common.event_type = AST_SECURITY_EVENT_SIP_INVITE_FAILED,
			.common.version    = AST_SECURITY_EVENT_SIP_INVITE_FAILED_VERSION,
			.common.service    = "PJSIP",
			.is_outgoing = session->is_outgoing,
			.resp_code = resp,
			.reason_text = reason_text,
			.peer_name = peer_name,
			.from_user = from_user,
			.to_user = to_user,
		};

		pjsip_sip_uri *uri_to = NULL, *uri_from = NULL;

		if(!strcmp(sipmethod,"INVITE") && rdata != NULL){
			uri_to = pjsip_uri_get_uri(rdata->msg_info.to->uri);
			uri_from = pjsip_uri_get_uri(rdata->msg_info.from->uri);
		}else if( !strcmp(sipmethod,"BYE") && tdata != NULL ){
			pjsip_from_hdr *from_hd = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_FROM, NULL);
			pjsip_from_hdr *to_hd = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_TO, NULL);

			uri_to = pjsip_uri_get_uri(from_hd->uri);
			uri_from = pjsip_uri_get_uri(to_hd->uri);
		} else {
			ast_log(LOG_ERROR,"ast_sip_report_invite_failed error...");
		}

		if(uri_to != NULL && uri_from != NULL){
			ast_copy_pj_str(to_user, &uri_to->user, sizeof(to_user));
			ast_copy_pj_str(from_user, &uri_from->user, sizeof(from_user));
		}

		ast_log(LOG_DEBUG,"to:%s,from:%s",to_user,from_user);

		ast_copy_string(reason_text, resp_text, sizeof(reason_text));
		ast_copy_string(peer_name, ast_sorcery_object_get_id(session->endpoint), sizeof(peer_name));
		ast_log(LOG_VERBOSE, "endpoint: %s, peer_name: %s, from_user: %s, to_user: %s, resp code: %d %s", ast_sorcery_object_get_id(session->endpoint),peer_name, from_user, to_user, resp, reason_text);

		ast_security_event_report(AST_SEC_EVT(&sip_invite_failed));
	}
}
/*end*/
#endif

static int load_module(void)
{
	pjsip_endpoint *endpt;

	CHECK_PJSIP_MODULE_LOADED();

	if (!ast_sip_get_sorcery() || !ast_sip_get_pjsip_endpoint()) {
		return AST_MODULE_LOAD_DECLINE;
	}
	if (!(nat_hook = ast_sorcery_alloc(ast_sip_get_sorcery(), "nat_hook", NULL))) {
		return AST_MODULE_LOAD_DECLINE;
	}
	nat_hook->outgoing_external_message = session_outgoing_nat_hook;
	ast_sorcery_create(ast_sip_get_sorcery(), nat_hook);
	sdp_handlers = ao2_container_alloc(SDP_HANDLER_BUCKETS,
			sdp_handler_list_hash, sdp_handler_list_cmp);
	if (!sdp_handlers) {
		return AST_MODULE_LOAD_DECLINE;
	}
	endpt = ast_sip_get_pjsip_endpoint();
	pjsip_inv_usage_init(endpt, &inv_callback);
	pjsip_100rel_init_module(endpt);
	pjsip_timer_init_module(endpt);
	if (ast_sip_register_service(&session_module)) {
		return AST_MODULE_LOAD_DECLINE;
	}
	ast_sip_register_service(&session_reinvite_module);

	ast_module_shutdown_ref(ast_module_info->self);

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_unregister_service(&session_reinvite_module);
	ast_sip_unregister_service(&session_module);
	ast_sorcery_delete(ast_sip_get_sorcery(), nat_hook);
	ao2_cleanup(nat_hook);
	ao2_cleanup(sdp_handlers);
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
static int reload_module(void)
{
	pjsip_endpoint *endpt;
	static pj_str_t STR_100REL = {"100rel", 6};
	static pj_str_t STR_TIMER = {"timer", 5};

	endpt = ast_sip_get_pjsip_endpoint();

	if (!pjsip_endpt_has_capability(endpt, PJSIP_H_SUPPORTED, NULL, &STR_TIMER)) {
		pjsip_endpt_add_capability(endpt, NULL, PJSIP_H_SUPPORTED, NULL, 1, &STR_TIMER);
	}

	if (!pjsip_endpt_has_capability(endpt, PJSIP_H_SUPPORTED, NULL, &STR_100REL)) {
		pjsip_endpt_add_capability(endpt, NULL, PJSIP_H_SUPPORTED, NULL, 1, &STR_100REL);
	}

	return 1;
}
#endif

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "PJSIP Session resource",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
#ifdef GRANDSTREAM_NETWORKS
		.reload = reload_module,
#endif
		.load_pri = AST_MODPRI_APP_DEPEND,
	       );
