#ifdef GRANDSTREAM_NETWORKS
/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#define AST_MODULE_LOG "res_ucm_warning"

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 381281 $")

#include "asterisk/module.h"
#include "asterisk/logger.h"
#include "asterisk/event.h"
#include "asterisk/stasis.h"
#include "asterisk/threadstorage.h"
#include "asterisk/strings.h"
#include "asterisk/security_events.h"
#include <sys/socket.h>
#include <sys/un.h>                                                                                                                          
#include <sys/types.h>
#include <signal.h>
#include <sys/signal.h>

#define LEN_ENDFLAG  2 // "\r\n" or "\n\n"

#define UNIX_PATH_MAX 108
#define WARNING_SOCK_PATH "/tmp/res_ucm_warning.sock"

static pthread_t warning_thread = -1;
static struct stasis_subscription *warning_event_sub;
static int disconnect = 0;
static int connection_fd = -1, socket_fd = -1 ;

enum ie_required {
	NOT_REQUIRED,
	REQUIRED
};

AST_THREADSTORAGE(security_event_buf);
static const size_t SECURITY_EVENT_BUF_INIT_LEN = 256;

static void* monitor_warning_process(void* data)
{
	struct sockaddr_un address;
	socklen_t address_length;
 
	socket_fd = socket(AF_UNIX, SOCK_STREAM, 0);
	if(socket_fd < 0)
	{
		ast_log(LOG_WARNING, "%s:socket() failed!\n", WARNING_SOCK_PATH);
		return NULL;
	} 

	unlink(WARNING_SOCK_PATH);

	memset(&address, 0, sizeof(struct sockaddr_un));

	address.sun_family = AF_UNIX;
	snprintf(address.sun_path, UNIX_PATH_MAX, WARNING_SOCK_PATH);

	if(bind(socket_fd, (struct sockaddr*) &address, sizeof(struct sockaddr_un)) != 0)
	{
		ast_log(LOG_WARNING, "%s:bind() failed!\n", WARNING_SOCK_PATH);
		return NULL;
	}

	if(listen(socket_fd, 0) != 0)
	{
		ast_log(LOG_WARNING, "%s:listen() failed!\n", WARNING_SOCK_PATH);
		return NULL;
	}

	while((connection_fd = accept(socket_fd, (struct sockaddr *) &address,
		&address_length)) > -1)
	{
		struct pollfd pfd;
		char buffer[512];
		int nbytes;
		int cnt;
		int res;
		int err_num;
		char str[64];

		ast_debug(3, "%s:accept a connection!\n", WARNING_SOCK_PATH);

		pfd.fd = connection_fd;
		pfd.events = POLLIN | POLLHUP | POLLRDNORM;
		pfd.revents = 0;

		/* init disconnect flag */		
		disconnect = 0;

		while(1){
			res = poll(&pfd, 1, 60000);// use a 60s timeout
			if(res > 0){
			// if result > 0, this means that there is either data available on the
			// socket, or the socket has been closed
				cnt = recv(connection_fd, buffer, sizeof(buffer), MSG_PEEK | MSG_DONTWAIT);
				if( 0 == cnt ){
					if(res > 0){
					// if recv returns zero, that means the connection has been closed:
						ast_debug(3, "connection disconnect 1!\n");
						disconnect = 1; 
						break;
					}
				}else if( -1 == cnt ){
					err_num = errno;
					ast_debug(3, "recv failed:%s!\n", strerror_r(err_num, str, sizeof(str)));
					switch(err_num){
						case ECONNRESET:
						case EBADF:
						case EPIPE: 
						case ENOTSOCK:
							disconnect = 1;							
							break;
						default:
							break;
					}
					if( disconnect ){
						ast_debug(3, "connection disconnect 2!\n");
						break;
					}
				}else if( cnt > 0 ){
					/* discard everything received from warning process.*/
					while((nbytes = recv(connection_fd, buffer, sizeof(buffer)-1, MSG_DONTWAIT)) > 0){
						buffer[nbytes] = 0;
					}
					if( 0 == nbytes ){
						ast_debug(3, "All incoming data discarded!\n");
					}
				}
			}
			else if(res == -1){
				/* This case shouldn't happen, we sleep 5 seconds here in case and retry it. */
				ast_log(LOG_ERROR, "poll failed!\n");
				sleep(5);
			}
		}
		close(connection_fd);		
		connection_fd = -1;
	}

	close(socket_fd);
	unlink(WARNING_SOCK_PATH);
	return NULL;
}

static void append_json_single(struct ast_str **str, struct ast_json *json,
		const enum ast_event_ie_type ie_type, enum ie_required required)
{
	const char *ie_type_key = ast_event_get_ie_type_name(ie_type);

	struct ast_json *json_string;

	json_string = ast_json_object_get(json, ie_type_key);

	if (!required && !json_string) {
		/* Optional IE isn't present. Ignore. */
		return;
	}

	/* At this point, it _better_ be there! */
	ast_assert(json_string != NULL);

	ast_str_append(str, 0, ",%s=\"%s\"",
			ie_type_key,
			ast_json_string_get(json_string));
}

static void append_json(struct ast_str **str, struct ast_json *json,
		const struct ast_security_event_ie_type *ies, enum ie_required required)
{
	unsigned int i;

	for (i = 0; ies[i].ie_type != AST_EVENT_IE_END; i++) {
		append_json_single(str, json, ies[i].ie_type, required);
	}
}

static void warning_event_cb(void *data, struct stasis_subscription *sub,
	struct stasis_message *message)
{
	
	struct ast_str *str;
	enum ast_security_event_type event_type;
	struct ast_json *event_type_json;
	int res, err_num, len;
	char err_str[64];
	int discard = 0;

	struct ast_json_payload *payload = stasis_message_data(message);

	if (stasis_message_type(message) != ast_security_event_type()) {
		return;
	}

	if (!payload) {
		return;
	}

	event_type_json = ast_json_object_get(payload->json, "SecurityEvent");
	event_type = ast_json_integer_get(event_type_json);

	ast_assert(event_type >= 0 && event_type < AST_SECURITY_EVENT_NUM_TYPES);

	if (!(str = ast_str_thread_get(&security_event_buf,
			SECURITY_EVENT_BUF_INIT_LEN))) {
		return;
	}

	ast_str_set(&str, 0, "%s=\"%s\"",
			ast_event_get_ie_type_name(AST_EVENT_IE_SECURITY_EVENT),
			ast_security_event_get_name(event_type));

	append_json(&str, payload->json,
			ast_security_event_get_required_ies(event_type), REQUIRED);
	append_json(&str, payload->json,
			ast_security_event_get_optional_ies(event_type), NOT_REQUIRED);
	
	/* use \r\n to separate multiple events information. */
	ast_str_append(&str, 0, "\r\n");

	ast_debug(3, "Warning Event:%s", ast_str_buffer(str));
#if 0
	if(ast_security_event_get_severity(event_type) != AST_SECURITY_EVENT_SEVERITY_ERROR ){
		/* only process error events, ignore all informational events. */
		ast_debug(3, "Informational Event, discard it!\n");
		return;
	}
#endif
	
	while(!disconnect){
		if(discard){
			len = strlen("\n\n");
			/* use \n\n as a flag to discard incomplete information. */
			res = send(connection_fd, "\n\n", len, MSG_NOSIGNAL | MSG_DONTWAIT);
		}else{
			len = ast_str_strlen(str);
			res = send(connection_fd, ast_str_buffer(str), len, MSG_NOSIGNAL | MSG_DONTWAIT);
		}
		if( -1 == res ){
			err_num = errno;			
			ast_debug(3, "send failed:%s!", strerror_r(err_num, err_str, sizeof(err_str)));	

			switch(err_num){
				case ECONNRESET:
				case EBADF:
				case EPIPE: 
				case ENOTSOCK:
					disconnect = 1;
					break;
				//case EWOULDBLOCK:
				case EAGAIN:					
					usleep(10000);
					break;
				case EINTR:
					break;
				default:
					break;
			}			
		}
		if( res > 0 ){
			if( res < len ){
				/* Incomplete information, discard. */
				discard = 1;
			}else if( res == len ){
				if(discard){
					discard = 0;
				}else{
					/* message has sended successfully. */
					break;
				}
			}			
		}
	}
}

static int load_module(void)
{
	if (!(warning_event_sub = stasis_subscribe(ast_security_topic(),warning_event_cb,NULL))) {
		ast_log(LOG_ERROR, "subscribe security events failed!\n");
		return AST_MODULE_LOAD_DECLINE;
	}

	if (ast_pthread_create(&warning_thread, NULL, &monitor_warning_process, NULL) < 0) {		
		warning_event_sub = stasis_unsubscribe(warning_event_sub);
		
		ast_log(LOG_ERROR, "res_ucm_warning module load failed!\n");
		return AST_MODULE_LOAD_DECLINE;
	}

	ast_verb(3, "Warning Event Notifications Enabled!\n");

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	if (warning_event_sub) {
		warning_event_sub = stasis_unsubscribe(warning_event_sub);
	}

	if (warning_thread && (warning_thread != AST_PTHREADT_STOP) && (warning_thread != AST_PTHREADT_NULL)) {
		if(connection_fd >= 0){
			close(connection_fd);
			connection_fd = -1;
		}
		if(socket_fd >= 0){
			close(socket_fd);
			socket_fd = -1;
		}		
		unlink(WARNING_SOCK_PATH);
		
		pthread_cancel(warning_thread);
		pthread_kill(warning_thread, SIGURG);
		pthread_join(warning_thread, NULL);
	}
	warning_thread = AST_PTHREADT_STOP;

	ast_verb(3, "Warning Event Notifications Disabled!\n");

	return 0;
}

AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "Warning Event Notifications");
#endif
