/*
 * A tool for PSTN parameters auto detection 
 *
 * Copyright (C) 2012 - 2014, Grandstream Networks, Inc.
 *     
 *  20130328: pkasbekar@grandstream.com (fxotune.c and freqcadest.c)                 
 *  20130329: merged and modifed by jwli@grandstream.com
 *                    add multithread for ACIM auto detection.
 *  20130409: modifed by jwli@grandstream.com
 *                    add thread cancel for ACIM/CPT auto detection.
 *                    add asterisk CLI for get detect results.
 *                    ...           
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Auto detect Channel Driver
 *
 * \ingroup channel_drivers
 *
 */

/*** MODULEINFO
    <support_level>extended</support_level>
 ***/

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 1.14 $")

#if defined(__NetBSD__) || defined(__FreeBSD__)
    #include <pthread.h>
    #include <signal.h>
#else
    #include <sys/signal.h>
#endif
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/sem.h>
#include <math.h>
#include <ctype.h>

#include <fcntl.h>
#include <dahdi/user.h>
#include <dahdi/wctdm_user.h>
#include <dahdi/tonezone.h>

#include "asterisk/paths.h"
#include "asterisk/cli.h"
#include "asterisk/module.h"
#include "asterisk/channel.h"
#include "asterisk/app.h"
#include "asterisk/lock.h"
/* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
#include "asterisk/manager.h"
/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
static int load_chan_dahdi(void);
static int unload_chan_dahdi(void);
static int svip_lec_control(int enable);
static inline int dahdi_get_event(int fd);

typedef enum
{
    ADT_NON=0,
    ADT_CPT,
    ADT_ACIM
}g_adt_item_t;

g_adt_item_t g_adt_item = ADT_NON;

static ast_mutex_t adt_lock;

#define BUSYTONE_DETECT_ONLY

/*========================PSTN CPT Auto Detect=================================*/
typedef enum
{
    OFF_HOOK=0,
    ON_HOOK
}port_state_e;

/* Value for DAHDI_HOOK, set to ON hook */
#define DAHDI_ONHOOK    0

/* Value for DAHDI_HOOK, set to OFF hook */
#define DAHDI_OFFHOOK   1

#define	LV_MIN_ONHOOK_VOLTAGE		15 //was 18 but could not drive 18volts paging systems
#define	LV_MIN_OFFHOOK_VOLTAGE		3

#define AUD_TEST_FFT_SAMPLES    2048 
#define	STATIC_TABLE_ALLOCATION
#if defined(STATIC_TABLE_ALLOCATION)
static float cosary[AUD_TEST_FFT_SAMPLES];
static float sinary[AUD_TEST_FFT_SAMPLES];
#else
float *cosary = NULL;
float *sinary = NULL;
#endif

#define NUM_SAMPLES_10MS                 80
#define BUFFER_LENGTH     2*NUM_SAMPLES_10MS

/* AUD_TEST_FFT_CAPTURE_SAMPLES = smallest value >= NUM_FFT_SAMPLES that is a multiple of 80 */
#define AUD_TEST_FFT_CAPTURE_SAMPLES    \
        (((AUD_TEST_FFT_SAMPLES / NUM_SAMPLES_10MS) + 1) * NUM_SAMPLES_10MS)
#define AUD_TEST_CAPTURE_SAMPLES                AUD_TEST_FFT_CAPTURE_SAMPLES

#define PI                                              (4.0 * atan(1.0))
#define TWOPI                                   (2.0 * PI)
#define SQR(x)                                  ((x)*(x))
#define MAX_DELTA_FREQ_DB       10

#define TRUE   1
#define FALSE  0

#define CPT_MIN_FREQ                            (300)   /*min cpt frequency */
#define CPT_MAX_FREQ                            (650)   /*max cpt frequency */
#define CPT_MIN_FREQ_STEPS                      25              /*min space between frequencies */
#define CPT_MIN_BIN_STEPS                       (int)(((float)CPT_MIN_FREQ_STEPS*AUD_TEST_FFT_SAMPLES/8000+1)/2)             /*min space between frequencies */
#define CPT_CADENCE_NUM                         4
#define CPT_FREQ_NUM                            2
#define CPT_MAX_SILENCE_THRESHOLD       400
#define VAD_MIN_COUNT                     5

#define CPT_MAX_CADENCE_DIFF_MS         (50) //ms
#define CPT_MAX_CADENCE_DIFF            ((int)CPT_MAX_CADENCE_DIFF_MS/*ms*/ * (8000/1000))
/* Frequency rounding resolution*/
#define CPT_FREQ_RES_ROUND                              (5)

#define TONE_DETECT_NUM 3
typedef enum
{
    SILENCE_TONE = -1,
    DIAL_TONE = 0,
    RINGBACK_TONE,
    BUSY_TONE,
}tone_type_e;

static const char * const adt_cpt_tonetype_str[] = {
    "DIAL TONE",
    "RINGBACK TONE",
    "BUSY TONE",
};

typedef enum
{
    CPT_VAD_VOICE,
    CPT_VAD_SILENCE
} cpt_vad_e;

typedef enum
{
    CPT_CAD_ON,
    CPT_CAD_OFF
} cpt_cad_e;

typedef struct
{
    int on;
    int off;
} cadence_t;

typedef struct
{
    char active;
    float f_sample_hz;
    int num_samples;
    float f_hz[CPT_FREQ_NUM];
    cadence_t cadence[CPT_CADENCE_NUM];
    int last_cad;
    int cad_count;
    int freq_lock;
    int cad_lock;
    int vad_count;
    int cpt_timeout;
} cpt_param_t;

typedef struct
{
    int schannum;
    int dchannum;
    char sextension[AST_MAX_EXTENSION];
    char dextension[AST_MAX_EXTENSION];
}adect_cpt_usrcfg_t;

typedef struct
{
    int get_result;
    tone_type_e ttype;
    int f_hz[CPT_FREQ_NUM];
    cadence_t cadence[CPT_CADENCE_NUM];
    int cur_disc;  //current disconnect
    int pr; //polarity reversal
} cpt_detect_result_t;

typedef enum
{
    OFFHOOK = 0,
    DIAL,
    WAITFORRING,
    RUNWAITFORRING,
    RINGBACK,
    SENDDTMF,
    PICKUPCALL,
    HANGUP,
    BUSY,
    MONITOR_EXIT,
}adt_monitor_state_t;

typedef enum
{
    USER_DONONE=0,
    USER_PICKUPED,
    USER_HANGUPED,
}useact_t;

typedef enum
{
    ADT_CPT_NON=0,
    ADT_CPT_START,
    ADT_CPT_DIAL_TONE,
    ADT_CPT_RB_TONE,
    ADT_CPT_WAIT_PICKUP,
    ADT_CPT_WAIT_HANGUP,
    ADT_CPT_BUSY_TONE,
    ADT_CPT_EXIT,
    ADT_CPT_CANCEL,
}adt_cpt_state_t;

typedef struct
{
    int fd_caller;
    int fd_called;
    char *adt_bigbuffer; 
}cpt_adt_thread_t;

static const char * const adt_cpt_state_str[] = {
    "idle",
    "started",
    "dial tone",
    "ring back tone",
    "waiting for pickup",
    "waiting for hangup",
    "busy tone",
    "done",
    "canceled",
};

typedef enum
{
    ADT_CPT_ERR_NONE=0,
    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    ADT_CPT_ERR_ACIM_DETECT,
    ADT_CPT_ERR_DETECT_BUSY,
    ADT_CPT_ERR_UNLOAD_DAHDI_FAILED,
    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    ADT_CPT_ERR_INVALID_EXT,
    ADT_CPT_ERR_INVALID_CHAN,
    ADT_CPT_ERR_FXO_ABSENT,
    ADT_CPT_ERR_FXO_DISCONNECT,
    ADT_CPT_ERR_MTASK_INIT_FAILED,
    ADT_CPT_ERR_FXO_CONFIG_FAILED,
    ADT_CPT_ERR_FXO_OFFHOOK_FAILED,
    ADT_CPT_ERR_FXO_DIAL_FAILED,
    ADT_CPT_ERR_NO_DIALTONE,
    ADT_CPT_ERR_FXO_PICKUP_FAILED,
    ADT_CPT_ERR_USER_PICKUP_TIMEOUT,
    ADT_CPT_ERR_USER_NO_RBTONE,
    ADT_CPT_ERR_FXO_DTMFSEND_FAILED,
    ADT_CPT_ERR_FXO_NO_RINGING,
    ADT_CPT_ERR_FXO_HANGUP_FAILED,
    ADT_CPT_ERR_USER_HANGUP_TIMEOUT,
    ADT_CPT_ERR_USER_NO_BZTONE,
    ADT_CPT_ERR_THREAD_CREAT_FAILED,
    ADT_CPT_ERR_END,
}adt_cpt_error_t;

static const char * const adt_cpt_error_str[] = {
    "no error return",
    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    "ACIM detect is running",
    "CPT detect is running",
    "unload dahdi module failed",
    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    "invaid extension number",
    "invaid fxo channel number",
    "fxo absent/busy",
    "fxo disconnect",
    "monitor task init failed",
    "fxo configure failed",
    "fxo offhook failed",
    "fxo dial failed",
    "no dial tone detected",
    "pickup the call failed",
    "user pickup the call timeout",
    "no ringback tone detected",
    "fxo dtmf send failed",
    "fxo no ringing, call setup failed",
    "hang up the call failed",
    "user hangup the call timeout",
    "no busy tone detected",
    "thread creat failed",
    "end of error string",
};

typedef struct
{
    adt_cpt_state_t stat_id;
    adt_cpt_error_t err_code;
    int pickup_timeout;  //for semi-automatic detection
    int hangup_timeout;  //for semi-automatic detection
}adt_cpt_rtstate_t;

cpt_param_t cpt_param = {0};
cpt_detect_result_t cpt_detect_result[TONE_DETECT_NUM];
short *bigbuffer = NULL;

static int tables_initialized = FALSE;
static unsigned int table_init_size = 0;

/*The amount of FXO port*/
static int g_fxo_amount = 0;
static int fxo_dis_in_bz_dect = 0;   //fxo disconnect when busy tone detecting.

pthread_t g_cpt_auto_detect_thread = AST_PTHREADT_NULL;
static int g_cpt_auto_detect_thread_pri = 0;

static pthread_t adt_cpt_monitor_thread = AST_PTHREADT_NULL;
static int adt_cpt_monitor_thread_pri = 0;

static void* monitor_task(void* thread_id);
static void acquire_lock(void);
static void release_lock(void);
static int monitor_start(void);

static ast_mutex_t app_lock;
static long monitor_counter;
adt_monitor_state_t monitorstate;
double elapsedtime = -1;
int cd_timeout = 3000;   //3000ms default, if CD(current disconnect) duration >  $cd_timeout,then the CD will be ignored.
/* BEGIN: jizhang 20140418 [Bug 24082] modify */
int intervaltime = 1000;//1000 ms
/* END: jizhang 20140418 [Bug 24082] modify */

int fd_array[2];
int port[2];
char call_prog_detect_running = -1;
char ringing = FALSE;
char busycheck = FALSE;
char battery_lost = FALSE;
int detectmode = 0;  // '0' for full-automatic; '1' semi-automatic
int user_action = USER_DONONE; //user action for semi-automatic detection

adect_cpt_usrcfg_t g_adect_cpt_usrcfg;
//adt_cpt_state_t g_adt_cpt_state;
adt_cpt_rtstate_t g_adt_rt_state;   //detect state return to user.

static int get_max_val( short const * const pBuffer, unsigned int const nSize, int * const pdc);
static cpt_vad_e  cpt_vad( short * pcm_in, int num_samples );
static char abs_greater(int a, int b, int thresh);
static char cpt_is_cadence_valid(void);
static char cpt_update_cadence( cpt_cad_e cad );
static void cpt_report_detected_tone( tone_type_e tone_type, char state );
static void swap( float * const a, float * const b);
static void cfft( float * const xreal, float * const yimag, int const numdat, int const flag );

static void FFT_n( float * const xreal, float * const yimag, int const transform_size );   
static void estimate_frequency( short* psbuffer,int transform_size,int* mag_max1,int* mag_max2 );
char fxo_port_is_connected(int fd, int port, port_state_e port_state);
char dial_extension(int fd,int port,char* dialstr);
char configure_port(int fd,int port);
char go_offhook(int fd,int port);
char pickup_call(int fd,int port);
char send_dtmf(int fd,int port,char* dialstr);
char hangup(int fd,int port);
char tone_detected(int fd,int port,tone_type_e tone_type,int num_samples);
static int fxotune_read(int fd, void *buffer, int len);
static int cpt_autodetect_is_finished(void);
static void* dahdi_pstn_cpt_autodetect_thread( void *data );
static int dahdi_pstn_cpt_autodetect_start( int cli_fd, adect_cpt_usrcfg_t *usrcfg, int concise );
static void dahdi_pstn_cpt_autodetect_thread_release(int cli_fd, int concise);
static char *dahdi_pstn_cpt_autodetect_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *dahdi_pstn_cpt_autodetect_getstate_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *dahdi_pstn_cpt_autodetect_useract_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *dahdi_pstn_cpt_autodetect_stop_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *dahdi_pstn_cpt_autodetect_var_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);

/*============== ===PSTN ACIM Auto Detect=================================*/
#define TEST_DURATION 4000 /*pkasbekar - just a pacifier to keep both calibration and dump happy*/
#define TEST_BUFFER_LENGTH (2 * TEST_DURATION)
#define SKIP_SAMPLES 800
#define SINE_SAMPLES 8000

#define MAX_RESULTS 	(5)

#define MAX_FXO_CHANNEL 32   

#define SVIP_LEC_CHANNEL_USED 16
#define FXO_AMOUNT_FILE "/proc/gxvboard/dev_info/num_fxo"

struct result_catalog
{
    int     idx;
    float   echo;
    float   freqres;
    struct wctdm_echo_coefs settings;
};

typedef struct
{
    struct result_catalog results[MAX_RESULTS];
    int numactive;
} topresults_t;

#define OUT_OF_BOUNDS(x) ((x) < 0 || (x) > 255)

struct silence_info
{
    char *dialstr;
    /** fd of device we are working with */
    int device; 
    /** seconds we should wait after dialing the dialstring before we know for sure we'll have silence */
    int initial_delay;
    /** seconds after which a reset should occur */
    int reset_after;
    /** time of last reset */
    struct timeval last_reset; 
};

typedef enum
{
    ADT_ACIM_NON=0,
    ADT_ACIM_START,
    ADT_ACIM_EXIT,
    ADT_ACIM_CANCEL,
}adt_acim_state_t;

static const char * const adt_acim_state_str[] = {
    "idle",
    "started",
    "done",
    "canceled",
};

typedef enum
{
    ADT_ACIM_ERR_NONE=0,
    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    ADT_ACIM_ERR_CPT_DETECT_BUSY,
    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    ADT_ACIM_ERR_DETECT_BUSY,
    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    ADT_ACIM_ERR_UNLOAD_DAHDI_FAILED,
    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    ADT_ACIM_ERR_THREAD_CREAT_FAILED,
    ADT_ACIM_FXO_ABSENT,
    ADT_ACIM_ERR_READ_FXO_STATS_FAILED,
    ADT_ACIM_ERR_FXO_NOT_CONNECT,
    ADT_ACIM_ERR_SET_LINEAR_MODE_FAILED,
    ADT_ACIM_ERR_GET_BUFF_INFO_FAILED,
    ADT_ACIM_ERR_SET_BUFF_INFO_FAILED,
    ADT_ACIM_ERR_BRING_OFFHOOK_FAILED,
    ADT_ACIM_ERR_FXO_CANNOT_CLEAR,
    ADT_ACIM_ERR_FXO_UNABLE_SET_ECHO_COEFF,
    ADT_ACIM_ERR_FXO_UNABLE_FLUSH_BUFF,
    ADT_ACIM_ERR_FXO_UNABLE_WRITE_LINE,
    ADT_ACIM_ERR_END,
}adt_acim_error_t;

static const char * const adt_acim_error_str[] = {
    "no error return",
    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    "CPT detect is running",
    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    "ACIM detect is running",
    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    "unload dahdi module failed",
    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    "ACIM thread create failed",
    "fxo: cannot be opened, absent or busy",
    "fxo: read status failed",
    "fxo: not connect",
    "fxo: set signed linear mode failed",
    "fxo: cannot get buffer information",
    "fxo: set buffer information failed",
    "fxo: bring offhook failed",
    "fxo: unable to get a clear outside line",
    "fxo: unable to set echo coefficients",
    "fxo: unable to flush port buffer",
    "fxo: could not write all data to line",
    "end of error string",
};

typedef struct
{
    int dfd;
    int cfg_index; 
}acim_adt_thread_t;

typedef enum
{
    ACIM_DOSET_AUTO,
    ACIM_DOSET_MANUAL,
}acim_doset_mode_t;

typedef struct
{
    int get_result;
    struct wctdm_echo_coefs coefs;
    float erl;
}acim_detect_result_t;

typedef struct
{
    int cfg_index;
    int tostart;
    int release;
    int channum;
    pthread_t acim_thread;
    int acim_thread_pri;
    adt_acim_state_t acim_state;
    adt_acim_error_t acim_errcode;
    topresults_t topresults;
    acim_detect_result_t best_result;
    acim_doset_mode_t acim_doset_mode;
}adect_acim_usrcfg_t;

//pkasbekar - smaller set of impedance and echo coefficients
struct wctdm_echo_coefs local_echo_trys [] =
{               
    { 0, 0, 0, 0, 0, 0, 0, 0, 0},
    { 1, 0, 0, 0, 0, 0, 0, 0, 0},
    { 2, 0, 0, 0, 0, 0, 0, 0, 0},
    { 3, 0, 0, 0, 0, 0, 0, 0, 0},
    { 4, 0, 0, 0, 0, 0, 0, 0, 0},
    { 5, 0, 0, 0, 0, 0, 0, 0, 0},
    { 6, 0, 0, 0, 0, 0, 0, 0, 0},
    { 7, 0, 0, 0, 0, 0, 0, 0, 0},

    { 8, 0, 0, 0, 0, 0, 0, 0, 0},
    { 9, 0, 0, 0, 0, 0, 0, 0, 0},
    { 10, 0, 0, 0, 0, 0, 0, 0, 0},
    { 11, 0, 0, 0, 0, 0, 0, 0, 0},
    { 12, 0, 0, 0, 0, 0, 0, 0, 0},
    { 13, 0, 0, 0, 0, 0, 0, 0, 0},
    { 14, 0, 0, 0, 0, 0, 0, 0, 0},
    { 15, 0, 0, 0, 0, 0, 0, 0, 0},
};

struct wctdm_echo_coefs echo_trys [] =
{
    /* 600 ohm echo settings */
    { 0, 0, 0, 0, 0, 0, 0, 0, 0},
    { 10, 0, 6, 1, 254, 2, 255, 0, 0},
    { 3, 255, 255, 0, 1, 0, 0, 0, 0},
    { 3, 1, 253, 253, 2, 255, 0, 0, 0},
    { 9, 254, 251, 255, 2, 0, 1, 0, 0},
    { 5, 3, 251, 250, 2, 254, 0, 0, 255},
    { 8, 253, 2, 244, 255, 10, 244, 3, 253},
    { 10, 249, 244, 8, 12, 245, 252, 0, 1},

    /* 900 ohm echo settings */
    { 1, 0, 0, 0, 0, 0, 0, 0, 0},
    { 10, 252, 255, 1, 255, 0, 0, 0, 0},
    { 7, 255, 251, 251, 2, 255, 255, 1, 255},
    { 3, 1, 251, 250, 1, 254, 255, 0, 255},
    { 5, 252, 250, 0, 0, 255, 1, 0, 0},
    { 5, 3, 251, 250, 1, 253, 0, 0, 255},
    { 8, 253, 2, 244, 255, 10, 244, 3, 253},
    { 10, 249, 244, 8, 12, 245, 252, 0, 1},

    /* 270 ohm + (750 ohm || 150 nF) (CTR21) */
    { 2, 0, 0, 0, 0, 0, 0, 0, 0},
    { 7, 0, 0, 255, 254, 0, 0, 0, 0},
    { 9, 0, 253, 254, 2, 255, 0, 0, 0},
    { 5, 1, 249, 254, 4, 253, 1, 0, 0},
    { 5, 252, 250, 1, 1, 254, 0, 255, 0},
    { 5, 3, 251, 250, 2, 253, 255, 255, 255},
    { 8, 253, 2, 244, 255, 10, 244, 3, 253},
    { 10, 249, 244, 8, 12, 245, 252, 0, 1},

    /* 220 ohm + (820 ohm || 120 nF) (Australia/NewZealand) and 220 ohm + (820 ohm
     * || 115nF) (Slovakia/SAfrica/Germany/Austria/Bulgaria)
     */
    { 3, 0, 0, 0, 0, 0, 0, 0, 0},
    { 7, 0, 255, 254, 255, 0, 255, 0, 0},
    { 9, 0, 253, 253, 1, 255, 0, 0, 0},
    { 5, 1, 249, 254, 3, 253, 1, 0, 0},
    { 5, 252, 250, 1, 1, 254, 0, 255, 0},
    { 5, 3, 251, 251, 2, 253, 255, 255, 255},
    { 8, 253, 2, 244, 255, 10, 244, 3, 253},
    { 10, 249, 244, 8, 12, 245, 252, 0, 1},

    /* 370 ohm + (620ohm || 310nF) (New Zealand #2/India) CO Termination */
    { 4, 0, 0, 0, 0, 0, 0, 0, 0},
    { 9, 255, 1, 4, 0, 0, 1, 255, 0},
    { 9, 0, 253, 0, 3, 254, 0, 0, 255},
    { 9, 2, 250, 253, 5, 253, 1, 0 ,255},
    { 5, 252, 250, 1, 2, 255, 0 ,255, 0},
    { 5, 3, 251, 250, 3, 254, 255, 255, 255},
    { 8, 253, 2, 244, 255, 10, 244, 3, 253},
    { 10, 249, 244, 8, 12, 245, 252, 0, 1},

    /* 320 ohm + (1050ohm || 230 nF) (England) CO Termination */
    { 5, 0, 0, 0, 0, 0, 0, 0, 0},
    { 9, 0 ,255, 1, 255, 255, 0, 255, 0},
    { 5, 255, 252, 0, 2, 254, 0, 255, 255},
    { 9, 2, 250, 253, 4, 252, 0, 255, 255},
    { 5, 252, 250, 1, 1, 254, 0 ,255, 255},
    { 5, 3, 251, 250, 2, 253, 255, 255, 254},
    { 3, 1, 1, 242, 2, 9, 245, 3, 253},
    { 10, 249, 244, 8, 12, 245, 252, 0, 1},

    /* 370 ohm + (820 ohm || 110 nF) CO Termination */
    { 6, 0, 0, 0, 0, 0, 0, 0, 0},
    { 6, 1, 254, 253, 0, 255, 0, 0, 0},
    { 9, 0, 251, 252, 2, 255, 0, 0, 0},
    { 5, 1, 248, 252, 4, 253, 1, 0, 0},
    { 5, 252, 250, 0, 0, 254, 0 , 255, 0},
    { 5, 3, 251, 250, 2, 253, 255, 255, 254},
    { 3, 1, 1, 242, 2, 9, 245, 3, 253},
    { 10, 249, 244, 8, 12, 245, 252, 0, 1},

    /* 275 ohm + (780 ohm || 115 nF) CO Termination */
    { 7, 0, 0, 0, 0, 0, 0, 0, 0},
    { 7, 255, 255, 255, 255, 0, 0, 0, 0},
    { 9, 0, 253, 254, 2, 255, 0, 0, 0},
    { 5, 1, 249, 254, 4, 253, 1, 0, 0},
    { 5, 252, 250, 1, 1, 254, 0, 255, 0},
    { 5, 3, 251, 250, 2, 253, 255, 255, 255},
    { 8, 253, 2, 244, 255, 10, 244, 3, 253},
    { 10, 249, 244, 8, 12, 245, 252, 0, 1},

    /* Make sure we include the rest of the impedances */
    { 8, 0, 0, 0, 0, 0, 0, 0, 0},
    { 9, 0, 0, 0, 0, 0, 0, 0, 0},
    { 10, 0, 0, 0, 0, 0, 0, 0, 0},
    { 11, 0, 0, 0, 0, 0, 0, 0, 0},
    { 12, 0, 0, 0, 0, 0, 0, 0, 0},
    { 13, 0, 0, 0, 0, 0, 0, 0, 0},
    { 14, 0, 0, 0, 0, 0, 0, 0, 0},
    { 15, 0, 0, 0, 0, 0, 0, 0, 0},
};
static float sintable[SINE_SAMPLES];
static const float amplitude = 16384.0;

static adect_acim_usrcfg_t adect_acim_usrcfg[MAX_FXO_CHANNEL];
static int acim_adt_counter = 0;

static int printbest = 1;
int acim_adt_inited = 0;
static int use_table = 1;

static char *dahdipath = "/dev/dahdi";

static int init_acim_autodetect(void);
static int ensure_silence(struct silence_info *info);
static short inline gentone(int hz, int idx);
static short inline genwaveform(int idx);
static float power_of(void *prebuf, int bufsize, int short_format);
void init_sinetable(void);
float sin_tbl(int arg, int num_per_period);
float cos_tbl(int arg, int num_per_period);
static float db_loss(float measured, float reference);
static void one_point_dft(const short *inbuf, int len, int frequency, float *real, float *imaginary);
static float calc_magnitude(short *inbuf, int insamps);
static void init_topresults(int usrcfg_index);
static void set_topresults(int usrcfg_index, int tbloffset, struct wctdm_echo_coefs *setting, float echo, float freqres);
static void print_topresults(int usrcfg_index, char * header);
static int acim_tune2(int whichdahdi, int usrcfg_index, int freq, char *dialstr, int delayuntilsilence,
                      int silencegoodfor, struct wctdm_echo_coefs *coefs_out,float *erl);
static int acim_tune(int whichdahdi, int usrcfg_index, char *dialstr, int delayuntilsilence, 
                     int silencegoodfor, struct wctdm_echo_coefs *coefs_out);
static int do_set(int dfd, int cfg_index);
static int do_calibrate(int dev_fd, int usrcfg_index, int calibtype,
                        char* dialstr, int delayuntilsilence, int silencegoodfor);
static int acim_autodetect_is_finished(void);
static void* dahdi_pstn_acim_autodetect_thread( void *data );
static void dahdi_pstn_acim_autodetect_thread_release( int cli_fd, int concise );
static int dahdi_pstn_acim_autodetect_start( int cli_fd, int concise );
static char *dahdi_pstn_acim_autodetect_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *dahdi_pstn_acim_autodetect_stop_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *dahdi_pstn_acim_autodetect_getstate_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *dahdi_pstn_acim_autodetect_doset_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);

static struct ast_cli_entry cli[] = {
    AST_CLI_DEFINE(dahdi_pstn_acim_autodetect_cli, "PSTN ACIM auto detect start"),
    AST_CLI_DEFINE(dahdi_pstn_acim_autodetect_stop_cli, "PSTN ACIM auto detect exit"),
    AST_CLI_DEFINE(dahdi_pstn_acim_autodetect_getstate_cli, "Get PSTN ACIM auto detect state"),
    AST_CLI_DEFINE(dahdi_pstn_acim_autodetect_doset_cli, "PSTN ACIM auto detect doset mode setting"),

    AST_CLI_DEFINE(dahdi_pstn_cpt_autodetect_cli, "PSTN CPT auto detect start"),
    AST_CLI_DEFINE(dahdi_pstn_cpt_autodetect_stop_cli, "PSTN CPT auto detect exit"),
    AST_CLI_DEFINE(dahdi_pstn_cpt_autodetect_getstate_cli, "Get PSTN CPT auto detect state"),
    AST_CLI_DEFINE(dahdi_pstn_cpt_autodetect_useract_cli, "Get PSTN CPT auto detect state"),
    AST_CLI_DEFINE(dahdi_pstn_cpt_autodetect_var_cli, "Set/Get PSTN CPT auto detect variables"),
};

static int load_chan_dahdi(void)
{
    if ( ast_load_resource("chan_dahdi.so") )
    {
        ast_log(LOG_ERROR, "Unable to load module:chan_dahdi.so\n");
        return -1;
    }
    ast_log(LOG_DEBUG, "Loaded: chan_dahdi.so\n");

    return 0;
}

static int unload_chan_dahdi(void)
{
    int force = AST_FORCE_SOFT;

    if ( ast_unload_resource("chan_dahdi.so", force) )
    {
        ast_log(LOG_ERROR, "Unable to unload resource chan_dahdi.so\n");
        return -1;
    }
    ast_log(LOG_DEBUG, "Unloaded chan_dahdi.so\n");
    return 0;
}

static int svip_lec_control(int enable)
{
    int i;
    int ret;
    struct ast_str *cmdline = ast_str_create(50);

    if ( !cmdline )
        return -1;

    ast_str_set(&cmdline, 0, "grep echocanceller /cfg/etc/dahdi/system.conf");

    ast_log(LOG_DEBUG,"system call %s\n", ast_str_buffer(cmdline));

    ret = ast_safe_system(ast_str_buffer(cmdline));
    if ( ret<0 )
    {
        ast_log(LOG_ERROR, "system call %s error!! %s\n", ast_str_buffer(cmdline),strerror(errno));
        goto safe_exit;
    }

    ast_log(LOG_DEBUG,"System Call exit status %d\n",ret);
    if ( ret )//svip lec not be used!!
    {
        goto safe_exit;
    }

    if ( !enable )
    {
        for ( i=0; i<SVIP_LEC_CHANNEL_USED; i++ )
        {
            ast_str_set(&cmdline, 0, "/usr/bin/svip_ctl stop_lec %d", i);
            ast_log(LOG_DEBUG,"system call %s\n", ast_str_buffer(cmdline));
            ret = ast_safe_system(ast_str_buffer(cmdline));
            if ( ret<0 )
            {
                ast_log(LOG_ERROR, "system call %s error!!\n", ast_str_buffer(cmdline));
                goto safe_exit;
            }
        }
    }
    else
    {
        for ( i=0; i<SVIP_LEC_CHANNEL_USED; i++ )
        {
            ast_str_set(&cmdline, 0, "/usr/bin/svip_ctl start_lec %d", i);
            ast_log(LOG_DEBUG,"system call %s\n", ast_str_buffer(cmdline));
            ret = ast_safe_system(ast_str_buffer(cmdline));
            if ( ret<0 )
            {
                ast_log(LOG_ERROR, "system call %s error!!\n", ast_str_buffer(cmdline));
                goto safe_exit;
            }
        }
    }
    safe_exit:
    ast_free(cmdline);
    return ret;
}

static inline int dahdi_get_event(int fd)
{
    int j;
    if ( ioctl(fd, DAHDI_GETEVENT, &j) == -1 )
        return -1;
    return j;
}

/*========================PSTN CPT Auto Detect=================================*/

/****************************************************************************/
/*           Cadence Estimation Functions Start here                      */
/****************************************************************************/

/* Description: Returns the maximum value from the pbuffer  */
static int get_max_val( short const * const pBuffer, unsigned int const nSize, int * const pdc)
{
    long dc = 0;
    long maxVal = -32768;
    unsigned int i;

    for ( i=0; i<nSize; i++ )
    {
        dc += pBuffer[i];
        if ( pBuffer[i] > maxVal )
            maxVal = pBuffer[i];
    }
    dc /= (float)nSize;
    maxVal -= dc;    //adjust to offset AC value

    if ( pdc!=NULL )
        *pdc = (int)dc;
    return maxVal;
}

static cpt_vad_e  cpt_vad( short * pcm_in, int num_samples )
{
    cpt_vad_e ret_val;

    if ( CPT_MAX_SILENCE_THRESHOLD < get_max_val(pcm_in, num_samples, NULL) )
    {
        cpt_param.vad_count++;
    }
    else
    {
        if ( cpt_param.vad_count!=0 && cpt_param.vad_count<VAD_MIN_COUNT )
        {
            ast_log(AST_LOG_DEBUG, "Spike skipped %d\n",cpt_param.vad_count*10);
        }
        cpt_param.vad_count = 0;
    }

    if ( cpt_param.vad_count>=VAD_MIN_COUNT )
    {
        ret_val = CPT_VAD_VOICE;
    }
    else
    {
        ret_val = CPT_VAD_SILENCE;
    }

    return ret_val;
}


static char abs_greater(int a, int b, int thresh)
{
    int diff;
    diff = a-b;

    if ( 0>diff )
        diff = -diff;

    if ( thresh<diff )
        return TRUE;
    else
        return FALSE;
}

static char cpt_is_cadence_valid(void)
{
    unsigned int i;
    unsigned long avg_on_even=0, avg_off_even=0;
    unsigned long avg_on_odd=0, avg_off_odd=0;

    for ( i=0; i<CPT_CADENCE_NUM; i++ )
    {
        /* If any of the cadences are null then return and continue filling untill
                   there are non NULL cadences*/
        if ( 0 == cpt_param.cadence[i].on || 0 == cpt_param.cadence[i].off )
            return FALSE;

        if ( i & 0x01 )
        {
            /*Gets the average of the odd cadences and uses it as benchmark to veryfy variability */
            avg_on_odd += (cpt_param.cadence[i].on)/(CPT_CADENCE_NUM>>1);
            avg_off_odd+= (cpt_param.cadence[i].off)/(CPT_CADENCE_NUM>>1);
        }
        else
        {
            /*Gets the average of the even cadences and uses it as benchmark to veryfy variability */
            avg_on_even += (cpt_param.cadence[i].on)/(CPT_CADENCE_NUM>>1);
            avg_off_even+= (cpt_param.cadence[i].off)/(CPT_CADENCE_NUM>>1);
        }
    }

    /*this is a multi-cadence tone */
    if ( (abs_greater(avg_on_odd, avg_on_even, CPT_MAX_CADENCE_DIFF)) ||
         (abs_greater(avg_off_odd, avg_off_even, CPT_MAX_CADENCE_DIFF)) )
    {
        for ( i=0; i<CPT_CADENCE_NUM; i++ )
        {
            if ( i & 0x01 )
            {
                /* this checks to see if all 3 cadences are within the tolerance of CPT_MAX_CADENCE_DIFF */
                if ( abs_greater(avg_on_odd, cpt_param.cadence[i].on, CPT_MAX_CADENCE_DIFF) )
                    return FALSE;

                if ( abs_greater(avg_off_odd, cpt_param.cadence[i].off, CPT_MAX_CADENCE_DIFF) )
                    return FALSE;
            }
            else
            {
                /* this checks to see if all 3 cadences are within the tolerance of CPT_MAX_CADENCE_DIFF */
                if ( abs_greater(avg_on_even, cpt_param.cadence[i].on, CPT_MAX_CADENCE_DIFF) )
                    return FALSE;

                if ( abs_greater(avg_off_even, cpt_param.cadence[i].off, CPT_MAX_CADENCE_DIFF) )
                    return FALSE;
            }
        }

        /*If the ON and OFF cadences are within CPT_MAX_CADENCE_DIFF ms then assume equal cadences so average them */
        if ( !abs_greater(avg_off_even, avg_on_even, CPT_MAX_CADENCE_DIFF) )
        {
            cpt_param.cadence[0].on = (avg_on_even+avg_off_even)/2;
            cpt_param.cadence[0].off = (avg_on_even+avg_off_even)/2;
        }

        if ( !abs_greater(avg_off_odd, avg_on_odd, CPT_MAX_CADENCE_DIFF) )
        {
            cpt_param.cadence[1].on = (avg_on_odd+avg_off_odd)/2;
            cpt_param.cadence[1].off = (avg_on_odd+avg_off_odd)/2;
        }

    }
    else
    {
        for ( i=0; i<CPT_CADENCE_NUM; i++ )
        {
            /* this checks to see if all 3 cadences are within the tolerance of CPT_MAX_CADENCE_DIFF */
            if ( abs_greater(avg_on_odd, cpt_param.cadence[i].on, CPT_MAX_CADENCE_DIFF) )
                return FALSE;

            if ( abs_greater(avg_off_odd, cpt_param.cadence[i].off, CPT_MAX_CADENCE_DIFF) )
                return FALSE;
        }

        /*If the ON and OFF cadences are within CPT_MAX_CADENCE_DIFF ms then assume equal cadences so average them */
        if ( !abs_greater(avg_off_odd, avg_on_odd, CPT_MAX_CADENCE_DIFF) )
        {
            cpt_param.cadence[0].on = (avg_on_odd+avg_off_odd)/2;
            cpt_param.cadence[0].off = (avg_on_odd+avg_off_odd)/2;
        }

        /*If the tone is a single cadence then zero out the second cadence. 
                this is mainly for printing purposes so it is not printed in teh syslog */
        cpt_param.cadence[1].on = 0;
        cpt_param.cadence[1].off = 0;

    }

    return TRUE;
}

static char cpt_update_cadence( cpt_cad_e cad )
{
    if ( cpt_param.last_cad!=cad && cpt_param.last_cad==CPT_CAD_OFF )
    {
        /*Subtract the offiset due to the ON cadence VAD. */
        cpt_param.cadence[cpt_param.cad_count].off -= VAD_MIN_COUNT*NUM_SAMPLES_10MS;

        ast_log(LOG_NOTICE, "Cadence %d estimation %d ms ON and %d ms OFF\n",
                cpt_param.cad_count,
                (int)((float)cpt_param.cadence[cpt_param.cad_count].on/cpt_param.f_sample_hz*(1000/*ms*/)),
                (int)((float)cpt_param.cadence[cpt_param.cad_count].off/cpt_param.f_sample_hz*(1000/*ms*/)));

        /*Confirm if there are three similar cadences and */
        /*Analyze the pariodicity of every cadence with respect to its adjasent cadences*/
        if ( cpt_is_cadence_valid() )
            return TRUE;

        cpt_param.cad_count++;
        if ( CPT_CADENCE_NUM<=cpt_param.cad_count )
        {
            cpt_param.cad_count = 0;
        }

        /* Reset the values and start estimating a new cadence */
        cpt_param.cadence[cpt_param.cad_count].on = VAD_MIN_COUNT*NUM_SAMPLES_10MS; //the on cadence count should be offset for the 
        cpt_param.cadence[cpt_param.cad_count].off = 0;
    }

    cpt_param.last_cad = cad;

    if ( CPT_CAD_ON==cad )
    {
        /* This is to catch the long continuos tones about */
        if ( ((unsigned int)(0xFFFF) - cpt_param.cadence[cpt_param.cad_count].on) < NUM_SAMPLES_10MS )
        {
            /*Reset all cadences after a continuos tone is detected */
            memset(cpt_param.cadence, 0x0000, sizeof(cpt_param.cadence));
            return TRUE;
        }
        cpt_param.cadence[cpt_param.cad_count].on  += NUM_SAMPLES_10MS;
    }
    else
    {
        cpt_param.cadence[cpt_param.cad_count].off += NUM_SAMPLES_10MS;
    }

    return FALSE;
}

static void cpt_report_detected_tone( tone_type_e tone_type, char state )
{
#if CPT_CADENCE_ROUND
    int cad_round = 0;
#else
    float cad_round = 0;
#endif
    unsigned int freq_round = 0;
    /*Remove the locks for next test entry */
    cpt_param.freq_lock = FALSE;
    cpt_param.cad_lock = FALSE;

    if ( state==FALSE )
    {
        /*If this was a Timeout case then no need to log the results */
        ast_log(LOG_NOTICE, "Time out. No activity on the line for %d seconds. Please check line ",cpt_param.cpt_timeout/100);
        return;
    }
    if (1)//( busycheck == FALSE )
    {
        switch ( tone_type )
        {
            case DIAL_TONE:
                ast_log(LOG_DEBUG, "Dial tone detected report:\n");
                break;
            case RINGBACK_TONE:
                ast_log(LOG_DEBUG, "Ring Back tone detected report:\n");
                break;
            case BUSY_TONE:
                ast_log(LOG_DEBUG, "Busy Back tone detected report:\n");
              //  cpt_detect_result[BUSY_TONE].cur_disc = FALSE;
                break;
            case SILENCE_TONE:
                return;
        }
    }
    else
    {
        ast_log(LOG_DEBUG, "Current disconnect detected\n");
        return;
    }

    /*Round freq1 to the nearest divisible by CPT_FREQ_RES_ROUND*/
    freq_round = (((unsigned int)cpt_param.f_hz[0]+CPT_FREQ_RES_ROUND/2)/CPT_FREQ_RES_ROUND)*CPT_FREQ_RES_ROUND;
    cpt_detect_result[tone_type].f_hz[0] = freq_round;

    /*Round freq1 to the nearest divisible by CPT_FREQ_RES_ROUND*/
    freq_round = (((unsigned int)cpt_param.f_hz[1]+CPT_FREQ_RES_ROUND/2)/CPT_FREQ_RES_ROUND)*CPT_FREQ_RES_ROUND;
    cpt_detect_result[tone_type].f_hz[1] = freq_round;

    memset(cpt_param.f_hz, 0x0000, sizeof(cpt_param.f_hz));

    /*ON Round cadence to the nearest multiple of CPT_MAX_CADENCE_DIFF_MS*/
    cad_round = (float)cpt_param.cadence[0].on/cpt_param.f_sample_hz*(1000/*ms*/);
#if CPT_CADENCE_ROUND
    cad_round = (((unsigned int)cad_round+CPT_MAX_CADENCE_DIFF_MS/2)/CPT_MAX_CADENCE_DIFF_MS)*CPT_MAX_CADENCE_DIFF_MS;
#endif

    cpt_detect_result[tone_type].cadence[0].on = cad_round;

    /*ON Round cadence to the nearest multiple of CPT_MAX_CADENCE_DIFF_MS*/
    cad_round = (float)cpt_param.cadence[1].on/cpt_param.f_sample_hz*(1000/*ms*/);
#if CPT_CADENCE_ROUND
    cad_round = (((unsigned int)cad_round+CPT_MAX_CADENCE_DIFF_MS/2)/CPT_MAX_CADENCE_DIFF_MS)*CPT_MAX_CADENCE_DIFF_MS;
#endif
    cpt_detect_result[tone_type].cadence[1].on = cad_round;

    /*OFF Round cadence to the nearest multiple of CPT_MAX_CADENCE_DIFF_MS*/
    cad_round = (float)cpt_param.cadence[0].off/cpt_param.f_sample_hz*(1000/*ms*/);
#if CPT_CADENCE_ROUND
    cad_round = (((unsigned int)cad_round+CPT_MAX_CADENCE_DIFF_MS/2)/CPT_MAX_CADENCE_DIFF_MS)*CPT_MAX_CADENCE_DIFF_MS;
#endif
    cpt_detect_result[tone_type].cadence[0].off = cad_round;

    /*OFF Round cadence to the nearest multiple of CPT_MAX_CADENCE_DIFF_MS*/
    cad_round = (float)cpt_param.cadence[1].off/cpt_param.f_sample_hz*(1000/*ms*/);
#if CPT_CADENCE_ROUND
    cad_round = (((unsigned int)cad_round+CPT_MAX_CADENCE_DIFF_MS/2)/CPT_MAX_CADENCE_DIFF_MS)*CPT_MAX_CADENCE_DIFF_MS;
#endif
    cpt_detect_result[tone_type].cadence[1].off = cad_round;

    memset(cpt_param.cadence, 0x0000, sizeof(cpt_param.cadence));

    cpt_param.cad_count = 0;
    cpt_param.last_cad = CPT_CAD_ON;

    if ( cpt_detect_result[tone_type].cadence[0].on )
    {
        ast_log(LOG_NOTICE, "Tone frequence locked [%d]HZ and [%d]HZ"
                " wtih cadence[0] %dms ON and %d ms OFF;" 
                "cadence[1] %dms ON and %d ms OFF;",
                cpt_detect_result[tone_type].f_hz[0], cpt_detect_result[tone_type].f_hz[1],
                cpt_detect_result[tone_type].cadence[0].on, cpt_detect_result[tone_type].cadence[0].off,
                cpt_detect_result[tone_type].cadence[1].on, cpt_detect_result[tone_type].cadence[1].off
               );
    }
    else
    {
        ast_log(LOG_NOTICE, "Tone frequence locked [%d]HZ and [%d]HZ wtih continous cadence\n",
                cpt_detect_result[tone_type].f_hz[0], cpt_detect_result[tone_type].f_hz[1]);
    } 
    cpt_detect_result[tone_type].get_result = TRUE;
}
/****************************************************************************/
/*           Cadence Estimation Functions end here                                                                          */
/****************************************************************************/


/****************************************************************************/
/*           Frequency Estimation Functions start  here                                                                     */
/****************************************************************************/
static void swap( float * const a, float * const b)
{
    float temp;
    temp = *a;
    *a = *b;
    *b = temp;
}
static void cfft( float * const xreal, float * const yimag, int const numdat, int const flag )
{
    int maxpower, arg, cntr, pnt0, pnt1, i, j, a, b, k;
    float sign, prodreal, prodimag, harm;

#if !defined(STATIC_TABLE_ALLOCATION)
    cosary = (float *)ast_calloc( 1, numdat*sizeof(float) );
    sinary = (float *)ast_calloc( 1, numdat*sizeof(float) );
    if ( cosary == NULL || sinary == NULL )
    {
        ast_log(AST_LOG_ERROR, "Allocation error in cfft.\n");
        return;
    }
#endif

    j = 0;
    if ( flag != 0 )
    {
        sign = 1.0;
        for ( i = 0; i < numdat; i++ )
        {
            xreal[i] = xreal[i] / numdat;
            yimag[i] = yimag[i] / numdat;
        }
    }
    else
    {
        sign = -1.0;
    }
    for ( i = 0; i <= numdat - 2; i++ )
    {
        if ( i < j )
        {
            swap(&xreal[i],&xreal[j]);
            swap(&yimag[i],&yimag[j]);
        }
        k = numdat / 2;
        while ( k <= j )
        {
            j = j - k;
            k = k / 2;
        }
        j = j + k;
    }
    maxpower = 0;
    i = numdat;
    while ( i != 1 )
    {
        maxpower = maxpower + 1;
        i = i / 2;
    }
    harm = TWOPI / numdat;

    if ( (!tables_initialized) || (table_init_size != numdat) )
    {
        for ( i = 0; i < numdat; i++ )
        {
            sinary[i] = sign * sin(harm * i);
            cosary[i] = cos(harm * i);
        }
        tables_initialized = TRUE;
        table_init_size = numdat;
    }
    a = 2;
    b = 1;
    for ( cntr = 1; cntr <= maxpower; ++cntr )
    {
        pnt0 = numdat / a;
        pnt1 = 0;
        for ( k = 0; k <= b - 1; k++ )
        {
            i = k;
            while ( i < numdat )
            {
                arg = i + b;
                if ( k == 0 )
                {
                    prodreal = xreal[arg];
                    prodimag = yimag[arg];
                }
                else
                {
                    prodreal = xreal[arg] * cosary[pnt1] - yimag[arg] * sinary[pnt1];
                    prodimag = xreal[arg] * sinary[pnt1] + yimag[arg] * cosary[pnt1];
                }
                xreal[arg] = xreal[i] - prodreal;
                yimag[arg] = yimag[i] - prodimag;
                xreal[i] = xreal[i] + prodreal;
                yimag[i] = yimag[i] + prodimag;
                i += a;
            }
            pnt1 += pnt0;
        }
        a = a * 2;
        b = b * 2;
    }
#if !defined(STATIC_TABLE_ALLOCATION)
    ast_free(cosary);
    cosary = NULL;
    ast_free(sinary);
    sinary = NULL;
#endif
}

static void FFT_n( float * const xreal, float * const yimag, int const transform_size )       
{
    cfft(xreal,yimag, transform_size, 0);
}

static void estimate_frequency( short* psbuffer,int transform_size,int* mag_max1,int* mag_max2 )
{
    int i,j;
    float xreal[AUD_TEST_FFT_SAMPLES] = {0.0};
    float yimg[AUD_TEST_FFT_SAMPLES] = {0.0};
    float Pin_f = 0.0;
    float f_res_hz = 0.0;
    int bin_max1 = 0;
    int tf_max1 = 0;
    int bin_max2 = 0;
    int tf_max2 = 0;

    for ( i = 0; i < transform_size; ++i )
    {
        xreal[i] = (float)(psbuffer[i]);
        yimg[i] = 0.0;
    }

    FFT_n(xreal,yimg,transform_size);

    /* Find max TF value and frequency */
    f_res_hz = 8000 / (float)transform_size;

    /* compute transfer function at each frequency bin up to fs/2 */
    for ( j = 0; j < transform_size; ++j )
    {
        /* compute power in current frequency bin */
        Pin_f = SQR(xreal[j]/transform_size) + SQR(yimg[j]/transform_size);

        /* overwrite real arrays with power spectrum */
        xreal[j] = 10.0 * log10(Pin_f);
        //FIXME must do this in NON-dB format to save on mips
        //		X_in_real[i] = Pin_f;

        if ( (int)xreal[j] > tf_max1 )
        {
            tf_max1 =(int) xreal[j];
            bin_max1 = j;
        }

    }

    *mag_max1 = (int)(f_res_hz * (float)bin_max1);
    tf_max2 = 0;

    for ( j = 0; j < transform_size/2; ++j )
    {
        if ( ((int)xreal[j] > tf_max2) && \
             /* Make sure that the power of the second frequency component is 
                        the most MAX_DELTA_FREQ_DB smaller then the first one */
             ((int)xreal[j] > (tf_max1 - MAX_DELTA_FREQ_DB)) && \
             /* Make sure that the power of the second frequency component is 
                        at lesat CPT_MIN_BIN_STEPS away */
             ( ((bin_max1+CPT_MIN_BIN_STEPS) < j )|| ((bin_max1-CPT_MIN_BIN_STEPS) > j) ) )
        {
            tf_max2 =(int) xreal[j];
            bin_max2 = j;
        }

    }

    if ( bin_max2 )
        *mag_max2 = (int)(f_res_hz * (float)bin_max2);
    else
        *mag_max2 = *mag_max1;
}
/****************************************************************************/
/*           Frequency Estimation Functions end here                      */
/****************************************************************************/

char fxo_port_is_connected(int fd, int port, port_state_e port_state)
{
    struct wctdm_stats fxo_stat;
    /* pkasbekar - Check line voltage status to determine if a particular FXO port is connected */
    if ( ioctl(fd, WCTDM_GET_STATS, &fxo_stat) )
    {
        ast_log(AST_LOG_ERROR, "Unable to read status of fxo port\n");
        return FALSE;
    }

    if ( port_state == ON_HOOK )
    {
        if ( fxo_stat.tipvolt >= (-LV_MIN_ONHOOK_VOLTAGE*1000) && fxo_stat.tipvolt <= (LV_MIN_ONHOOK_VOLTAGE*1000) )
        {
            ast_log(AST_LOG_ERROR, "FXO port %d not connected \n",port);
            ast_log(AST_LOG_DEBUG, "FXO %d tip voltage = %d, FXO %d ring voltage= %d "
                    "FXO %d battery voltage = %d\n" , port, fxo_stat.tipvolt,
                    port, fxo_stat.ringvolt, port, fxo_stat.batvolt);
            return FALSE;
        }
    }
    else
    {
        if ( fxo_stat.tipvolt >= (-LV_MIN_OFFHOOK_VOLTAGE*1000) && fxo_stat.tipvolt <= (LV_MIN_OFFHOOK_VOLTAGE*1000) )
        {
            ast_log(AST_LOG_ERROR, "FXO port %d not connected \n", port);
            ast_log(AST_LOG_DEBUG,"FXO %d tip voltage = %d, FXO %d ring voltage = %d "
                    "FXO %d battery voltage = %d\n", port, fxo_stat.tipvolt, port, 
                    fxo_stat.ringvolt, port, fxo_stat.batvolt);
            return FALSE;
        }
    }
    return TRUE;     
}

char dial_extension(int fd,int port,char* dialstr)
{
    int diallength = 0;
    int i;
    struct dahdi_dialoperation dop;

    diallength = strlen(dialstr);

    ast_log(LOG_NOTICE, "dial_extension %s \n", dialstr);

    for ( i = 0; i < diallength; ++i )
    {
        char c[2];
        sprintf(c,"%c",dialstr[i]);
        c[1]='\0';
        memset(&dop, 0, sizeof(dop));
        dop.op = DAHDI_DIAL_OP_APPEND;
        dop.dialstr[0] = 'T';
        dahdi_copy_string(dop.dialstr + 1,c, sizeof(dop.dialstr));

        if ( ioctl(fd, DAHDI_DIAL, &dop) )
        {
            ast_log(LOG_NOTICE, "Unable to dial!\n");
            return FALSE;
        }
    }
    return TRUE;
}

char configure_port(int fd,int port)
{
    struct dahdi_bufferinfo bi;
    int x;

    x = 1;
    if ( ioctl(fd, DAHDI_SETLINEAR, &x) )
    {
        ast_log(AST_LOG_ERROR, "Unable to set channel to signed linear mode.\n");
        return FALSE;
    }

    memset(&bi, 0, sizeof(bi));
    if ( ioctl(fd, DAHDI_GET_BUFINFO, &bi) )
    {
        ast_log(AST_LOG_ERROR, "Unable to get buffer information!\n");
        return FALSE;
    }

    bi.numbufs = 2;
    bi.bufsize = 8000;
    bi.txbufpolicy = DAHDI_POLICY_IMMEDIATE;
    bi.rxbufpolicy = DAHDI_POLICY_IMMEDIATE;
    if ( ioctl(fd, DAHDI_SET_BUFINFO, &bi) )
    {
        ast_log(AST_LOG_ERROR, "Unable to set buffer information!\n");
        return FALSE;
    }

    return TRUE;  
}

char go_offhook(int fd,int port)
{
    int x;

    ast_log(AST_LOG_NOTICE, "go_offhook\n");

    x = DAHDI_OFFHOOK;

    if ( ioctl(fd, DAHDI_HOOK, &x) )
    {
        ast_log(AST_LOG_ERROR, "Cannot bring port %d off hook", port);
        return FALSE;
    }

    monitorstate=DIAL;

    return TRUE;
}

char pickup_call(int fd,int port)
{
    ast_log(LOG_NOTICE, "pickup_call\n");
    sleep(2);
    if ( FALSE==go_offhook(fd,port) )
    {
        return FALSE;
    }
    return TRUE;    
}

char send_dtmf(int fd,int port,char* dialstr)
{
    ast_log(LOG_NOTICE, "send_dtmf\n");
    if ( TRUE == dial_extension(fd,port,dialstr) )
    {
        return TRUE;
    }
    return FALSE;
}

char hangup(int fd,int port)
{
    int x;
    x = DAHDI_ONHOOK;
    ast_log(LOG_NOTICE, "hangup\n");

    if ( ioctl(fd, DAHDI_HOOK, &x) )
    {
        ast_log(AST_LOG_ERROR, "Cannot bring port %d on hook", port);
        return FALSE;
    }

    return TRUE;
}

static int fxotune_read(int fd, void *buffer, int len)
{
    int res;
    res = read(fd, buffer, len);
    return res;
}

//#define TONEDECT_DUMP_PCM_FILE  "/data/dump.bin"

char tone_detected(int fd,int port,tone_type_e tone_type,int num_samples)
{
    int retval = FALSE;
    short buffer[80];
    short fftbuffer[AUD_TEST_FFT_SAMPLES];
    short tempbuffer[8000];

    int f1,f2;
    int cpt_fail_count = 0;
    char start_capture;
    int in_index = 0;
    int x,res,y,z,k;
    int index,result;

#if defined(TONEDECT_DUMP_PCM_FILE)
    FILE* pdump;
#endif

    memset(&cpt_param, 0x0000,  sizeof(cpt_param_t));
    memset(fftbuffer, 0, sizeof(fftbuffer));
    memset(buffer, 0, sizeof(buffer));

    cpt_param.f_sample_hz = 8000.0;

    /*25 sec of CPT inactivity allowed to ensure slow Current disconnect detection*/
    cpt_param.cpt_timeout = 2500;

    cpt_param.freq_lock = FALSE;
    cpt_param.cad_lock = FALSE;
    cpt_param.active = TRUE;
    start_capture =  FALSE;

#if defined(TONEDECT_DUMP_PCM_FILE)
    pdump=fopen(TONEDECT_DUMP_PCM_FILE, "wb");
    if ( pdump == NULL )
    {
        ast_log(AST_LOG_ERROR, "could not open dump file\n");
        return FALSE;
    }
#endif
    y = 0; z = 0;k=0;
    ast_log(LOG_VERBOSE, "x = %d, y = %d, z = %d\n", x, y, z);

    /* Flush buffers */
    x = DAHDI_FLUSH_READ | DAHDI_FLUSH_WRITE | DAHDI_FLUSH_EVENT;
    if ( ioctl(fd, DAHDI_FLUSH, &x) )
    {
        ast_log(AST_LOG_ERROR, "Unable to flush I/O: %s\n", strerror(errno));
        return FALSE;
    }

    if ( bigbuffer != NULL )
    {
        ast_free(bigbuffer);
        bigbuffer = NULL;
    }
    bigbuffer = (short *)ast_calloc(1, 320000*2);
    if ( NULL == bigbuffer )
    {
        ast_log(AST_LOG_ERROR, "can not alloc memery for tone detect %s\n", strerror(errno));
        return FALSE;
    }

    retry:      
    if ( TRUE==fxo_port_is_connected(fd,port,OFF_HOOK) )
    {
        fxo_dis_in_bz_dect = 0;
        res = fxotune_read(fd,tempbuffer,sizeof(tempbuffer));
        if ( res!= sizeof(tempbuffer) )
        {
            int dummy;
            ioctl(fd, DAHDI_GETEVENT, &dummy);
            goto retry;
        }
        if ( k < num_samples )
        {
            if ( k+res/2 <= 320000 )
            {
                memcpy(&bigbuffer[k],tempbuffer,res); 
                k = k+res/2;
                goto retry;
            }
            else
            {
                ast_log(LOG_WARNING, "fxotune_read out of buffer\n");
            }
        }
    }
    else
    {
        fxo_dis_in_bz_dect = 1;
        ast_log(LOG_ERROR, "fxo port %d not connect\n", port);
        retval = FALSE;
        goto safe_out;
    }
#if defined(TONEDECT_DUMP_PCM_FILE)
    y=fwrite(bigbuffer,sizeof(char),sizeof(bigbuffer),pdump);
    z=z+y;
#else
    z=z+k*2;
#endif

#if defined(TONEDECT_DUMP_PCM_FILE)
    fclose(pdump);
#endif

    index=0;
    memset(buffer, 0, sizeof(buffer));

    while ( (index + NUM_SAMPLES_10MS*2 <= z) && (cpt_param.active == TRUE) )
    {
        result = NUM_SAMPLES_10MS*2;
        memcpy(buffer, &bigbuffer[index/2], result); 

        if ( CPT_VAD_VOICE == cpt_vad(buffer,NUM_SAMPLES_10MS) )
        {
            cpt_fail_count = 0;
            if ( cpt_update_cadence(CPT_CAD_ON) )
            {
                /*cadence has been verified by multiple detection,
                             this runs if 3 cadences have been verified and upon
                             the end of the thirs OFF period */
                cpt_param.cad_lock = TRUE;
                ast_log(LOG_NOTICE, "cadence locked\n");
            }
            else
            {
                start_capture = TRUE;
            }
        }
        else
        {
            /*if the frequency is already known then start taking into account the OFF cadence*/
            if ( cpt_param.freq_lock == TRUE )
            {
                cpt_update_cadence(CPT_CAD_OFF);
            }

            /*If after 5 sec there is no tone on the line then drop the detection and continue */
            cpt_fail_count++;
            if ( cpt_fail_count>cpt_param.cpt_timeout )
            {
                cpt_param.active = FALSE;
                cpt_report_detected_tone(tone_type, FALSE);
            }
        }

        if ( TRUE == start_capture )
        {
            memcpy(&fftbuffer[in_index],buffer,NUM_SAMPLES_10MS*sizeof(short));

            in_index+=NUM_SAMPLES_10MS;
        }
        if ( in_index >= AUD_TEST_CAPTURE_SAMPLES && cpt_param.freq_lock == FALSE )
        {
            estimate_frequency(fftbuffer,AUD_TEST_FFT_SAMPLES,&f1,&f2);

            ast_log(LOG_NOTICE, "fft complete\n");
            memset(&fftbuffer, 0, sizeof(fftbuffer));
            in_index = 0;
            cpt_fail_count = 0;
            start_capture = FALSE;

            if ( (f1 > CPT_MIN_FREQ) && (f1 < CPT_MAX_FREQ) && (f2 > CPT_MIN_FREQ) && (f2 < CPT_MAX_FREQ) )
            {
                cpt_fail_count = 0;
                cpt_param.f_hz[0] = (float)f1;
                cpt_param.f_hz[1] = (float)f2;
                cpt_param.freq_lock = TRUE;

                ast_log(LOG_NOTICE, "frequency locked  f1 %d  f2 %d\n",f1,f2);
            }
        }
        else if ( in_index >= AUD_TEST_CAPTURE_SAMPLES && cpt_param.freq_lock == TRUE )
        {
            in_index = 0;
            memset(&fftbuffer, 0, sizeof(fftbuffer));
        }
        /* After the cadence has been detected then finish the test and report the results.*/
        if ( cpt_param.cad_lock == TRUE )
        {
            /* Stop the audio capturing and continue with the nex CPT*/
            cpt_param.active = FALSE;

            /* All of the sanitty checking is done here for freq and cadence*/
            cpt_report_detected_tone(tone_type, TRUE);
        }
        index+=result;
    }

    retval = TRUE;

    safe_out:
    ast_free(bigbuffer);
    bigbuffer = NULL;
    return retval;

}

static void acquire_lock(void)
{
    ast_mutex_lock( &app_lock );
}

static void release_lock(void)
{
    ast_mutex_unlock( &app_lock );
}

static void release_lock_cleanup(void *data)
{
    ast_mutex_unlock( &app_lock );
}

static void monitor_task_clean_up(void* data)
{
    int lock_owner = *(int *)data;

    if ( lock_owner )
        release_lock();

    monitorstate = MONITOR_EXIT;
}

static void* monitor_task(void* thread_id)
{
    int lock_owner = 0;

    pthread_cleanup_push(monitor_task_clean_up, &lock_owner);

    ringing = FALSE;
    busycheck = FALSE;

    while ( call_prog_detect_running )
    {
        ast_fdset rfds;
        int fdValueMax = fd_array[0];
        int j;
        int rc,res;
        struct timeval tv,t1,t2;

        tv.tv_sec = 1;
        tv.tv_usec = 0;

        FD_ZERO(&rfds);
        /* Open device file descriptor */
        for ( j = 0; j < 2; j++ )
        {
            if ( fd_array[j] != -1 )
            {
                FD_SET( fd_array[j], &rfds );
                if ( fd_array[j] > fdValueMax )
                    fdValueMax = fd_array[j];
            }
        }

        rc = ast_select( fdValueMax + 1, &rfds, NULL, NULL, &tv );
        if ( rc < 0 )
        {
            ast_log(LOG_ERROR, "select returned -1\n");
            break;
        }
        else if ( rc == 0 )
        {
            continue;
        }
        for ( j = 0; j < 2; j++ )
        {
            if ( fd_array[j] != -1 )
            {
                if ( FD_ISSET( fd_array[j], &rfds ) )
                {
                    res = dahdi_get_event(fd_array[j]);
                    switch ( res )
                    {
                        case DAHDI_EVENT_DIALCOMPLETE: monitor_counter = 0;
                            if ( call_prog_detect_running && (monitorstate == WAITFORRING) )
                            {
                                monitorstate = RUNWAITFORRING;
                                ast_log(LOG_DEBUG, "Dial Completed!!....\n");
                                acquire_lock();
                                lock_owner = 1;
                            }
                            break;
                        case DAHDI_EVENT_RINGBEGIN:
                        case DAHDI_EVENT_RINGOFFHOOK: 
                            if ( ringing == FALSE )
                            {
                                ringing = TRUE;
                                monitorstate=RINGBACK;
                                if ( lock_owner )
                                {
                                    release_lock();
                                    lock_owner = 0;
                                }
                            }
                            break;
                        case DAHDI_EVENT_BATTERY_LOST:
                            if(j==1)break; //we deal battery lost event only for caller fxo.
                            battery_lost = TRUE;
                            if ( monitorstate == BUSY && busycheck == FALSE )
                            {
                                gettimeofday(&t1,NULL);
                                busycheck = TRUE;
                            }
                            else if ( monitorstate == BUSY && busycheck == TRUE )
                            {
                            }
                            else
                            {
                                if ( FALSE==fxo_port_is_connected(fd_array[j],port[j],OFF_HOOK) )
                                {
                                    if ( lock_owner )
                                    {
                                        release_lock();
                                        lock_owner = 0;
                                    }
                                    //call_prog_detect_running = 0;
                                }
                            }
                            break;
                        case DAHDI_EVENT_BATTERY_PRESENT:
                            if(j==1)break;
                            battery_lost = FALSE;
                            ast_log(LOG_DEBUG, "Got DAHDI_EVENT_BATTERY_PRESENT monitorstate: %d \n", monitorstate);
                            if ( monitorstate == BUSY )
                            {
                                gettimeofday(&t2,NULL);
                                elapsedtime = (t2.tv_sec - t1.tv_sec) *1000.0 + (t2.tv_usec - t1.tv_usec)/1000;
                                cpt_detect_result[BUSY_TONE].cur_disc = TRUE;
                                ast_log(LOG_DEBUG, "Current disconnect for %.2f ms\n", elapsedtime);
                               // call_prog_detect_running = 0;
                                if ( lock_owner )
                                {
                                    release_lock();
                                    lock_owner = 0;
                                }
                            }
                            break;
                        case DAHDI_EVENT_POLARITY:
   /** FIXEME(BUG16128): I dont think we need to deal event of polarity reversal. It will lead busytone detect failed.
        the 'call_prog_detect_running' will be set to zero when the event of polarity reversal be detected between 
        we hangup called extension and busytone detection not started in main() . then autodetection will be exit.
    */
  #if 0
                            if ( fxo_port_is_connected(fd_array[j],port[j],OFF_HOOK) )
                            {
                                if ( monitorstate == BUSY )
                                {
                                    gettimeofday(&t2,NULL);
                                    elapsedtime = (t2.tv_sec - t1.tv_sec) *1000.0 + (t2.tv_usec - t1.tv_usec)/1000;
                                    if ( elapsedtime < 3000.00 )
                                        ast_log(LOG_ERROR, "No busy tone, current disconnect for %.2f ms\n",elapsedtime);
                                    call_prog_detect_running = 0;
                                    if ( lock_owner )
                                    {
                                        release_lock();
                                        lock_owner = 0;
                                    }
                                }
                            }
                            else
                            {
                                if ( monitorstate == BUSY && busycheck == FALSE )
                                {
                                    gettimeofday(&t1,NULL);
                                    busycheck=TRUE;  
                                }
                                else if ( monitorstate == BUSY && busycheck == TRUE )
                                {
                                }
                                else
                                {
                                    if ( lock_owner )
                                    {
                                        release_lock();
                                        lock_owner = 0;
                                    }
                                    call_prog_detect_running = 0;
                                }
                            }
#else
                            if(j==1)break; //we ignore polarity reversal event for called fxo.
                            if ( monitorstate == BUSY )
                            {
                                ast_log(LOG_DEBUG, "No busy tone, polarity reversal detected\n");
                                cpt_detect_result[BUSY_TONE].pr = TRUE;
                             //   call_prog_detect_running = 0;
                                if ( lock_owner )
                                {
                                    release_lock();
                                    lock_owner = 0;
                                }
                            }
#endif
                        case DAHDI_EVENT_NONE:  
                        default: 
                            switch ( monitorstate )
                            {
                                case RUNWAITFORRING:
                                    if(detectmode)   //for semi-automatic detection, we donot waitting for ringing and the thread must releas lock.
                                    {
                                       if ( monitor_counter > 250000 )
                                       {
                                          if ( ringing == FALSE )
                                          {
                                             ringing = TRUE;
                                             monitorstate=RINGBACK;
                                          }
                                          if ( lock_owner )
                                          {
                                              ast_log(LOG_DEBUG, "Semi-automatic waiting for ringing end!!....\n");
                                              release_lock();
                                              lock_owner = 0;
                                          }
                                       }
                                    }
                                    if ( monitor_counter > 1000000 )//about 20s
                                    {
                                        ast_log(LOG_ERROR,"Did not get ringing\n");
                                        ringing = FALSE;
                                        if ( lock_owner )
                                        {
                                            release_lock();
                                            lock_owner = 0;
                                        }
                                        monitor_counter = 0;
                                    }
                                    else
                                    {
                                        //fprintf(stdout,"monitor_counter = %d\n",monitor_counter);
                                    }
                                    break;

                                default:         
                                    break;
                            }
                            break;
                    }
                }
            }
        }
        monitor_counter++;
    }
    monitorstate = MONITOR_EXIT;
    if ( lock_owner )
    {
        release_lock();
        lock_owner = 0;
    }
    pthread_cleanup_pop(0);
    return NULL;
}

static int monitor_start(void)
{
    pthread_attr_t attr;
    struct sched_param param;
    int policy = SCHED_OTHER;
    int priority_max;
    int priority_min;
    int res = -1;

    if ( adt_cpt_monitor_thread!= AST_PTHREADT_NULL )
    {
        res = pthread_kill(adt_cpt_monitor_thread, SIGURG);

        ast_log(AST_LOG_WARNING,
                "adt_cpt_monitor_thread is running, kill(%lu, NULL) res:%d, errno:%d(%s)",
                adt_cpt_monitor_thread, res, errno, strerror(errno));

        adt_cpt_monitor_thread = AST_PTHREADT_NULL;
    }

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    pthread_attr_setstacksize(&attr, 512 * 1024);

    param.sched_priority = adt_cpt_monitor_thread_pri;
    res = pthread_attr_setschedparam(&attr, &param);
    if ( res )
    {
        ast_log(AST_LOG_ERROR, "adt_cpt_monitor_thread set sched_priority=%d failed, res:%d, errno:%d(%s)!\n",
                param.sched_priority, res, errno, strerror(errno));
    }

    res = pthread_attr_getschedparam(&attr, &param);
    if ( res )
    {
        ast_log(AST_LOG_ERROR,  "adt_cpt_monitor_thread get schedparam failed, res:%d, errno:%d(%s)!\n",
                res, errno, strerror(errno));
    }
    else
    {
        //  ast_log(AST_LOG_DEBUG, "adt_cpt_monitor_thread get priority:%d.\n", param.sched_priority);
    }

    res = pthread_attr_getschedpolicy(&attr, &policy);
    if ( res )
    {
        ast_log(AST_LOG_ERROR, "adt_cpt_monitor_thread get schedpolicy failed, res:%d, errno:%d(%s)!\n",
                res, errno, strerror(errno));
    }
    else
    {
        //  ast_log(AST_LOG_DEBUG, "adt_cpt_monitor_thread get policy:%d.\n", policy);
    }

    priority_max = sched_get_priority_max(policy);
    priority_min = sched_get_priority_min(policy);
    ast_log(LOG_VERBOSE, "priority_max = %d, priority_min = %d\n", priority_max, priority_min);

    res = ast_pthread_create_background(&adt_cpt_monitor_thread, &attr,
                                        monitor_task, NULL);

    pthread_attr_destroy(&attr);

    if ( res != 0 )
    {
        ast_log(AST_LOG_ERROR, "adt_cpt_monitor_thread thread failed! res:%d, errno:%d(%s) !!\n",
                res, errno, strerror(errno));
        adt_cpt_monitor_thread = AST_PTHREADT_NULL;
        return -1;
    }

    ast_log(AST_LOG_DEBUG, "adt_cpt_monitor_thread start successful. adt_cpt_monitor_thread %lu ...\n",
            adt_cpt_monitor_thread);

    return 0;

}

static void cpt_autodetect_cleanup(void *data)
{
    cpt_adt_thread_t  *cpt_adt_thread = (cpt_adt_thread_t *)data;

    if ( bigbuffer != NULL )
    {
        ast_free(bigbuffer);
        bigbuffer = NULL;
    }
#if !defined(STATIC_TABLE_ALLOCATION)
    if ( sinary != NULL )
    {
        ast_free(sinary);
        sinary = NULL;
    }
    if ( cosary != NULL )
    {
        ast_free(cosary);
        cosary = NULL;
    }
#endif
    call_prog_detect_running = 0;

    if ( adt_cpt_monitor_thread!= AST_PTHREADT_NULL )
    {
        pthread_cancel(adt_cpt_monitor_thread);

        while ( 1 )
        {
            usleep(10000);
            if ( monitorstate == MONITOR_EXIT )
                break;
        }
        ast_log(AST_LOG_DEBUG, "cancel monitor task thread done!!\n");

        adt_cpt_monitor_thread = AST_PTHREADT_NULL;
    }

    if ( cpt_adt_thread->fd_caller >0 )
    {
        close(cpt_adt_thread->fd_caller);
    }
    if ( cpt_adt_thread->fd_called >0 )
    {
        close(cpt_adt_thread->fd_called);
    }

    load_chan_dahdi();
    g_adt_rt_state.stat_id = ADT_CPT_CANCEL;

}

static void* dahdi_pstn_cpt_autodetect_thread( void *data )
{
    int i;
    cpt_adt_thread_t  cpt_adt_thread;
    char dahdidev[80]="";
    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    char freq[80] = "";
    char cadence[80] = "";
    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */

    char *sextension = g_adect_cpt_usrcfg.sextension;
    char *dextension = g_adect_cpt_usrcfg.dextension;
    int schannum = g_adect_cpt_usrcfg.schannum;
    int dchannum = g_adect_cpt_usrcfg.dchannum;
    int dial_timeout = 20;   //20s
    int wait_bat_present = 10; //10s

    g_adt_rt_state.stat_id = ADT_CPT_START;
    g_adt_rt_state.err_code = ADT_CPT_ERR_NONE;

    pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
    for ( i=0;i <TONE_DETECT_NUM; i++ )
    {
        cpt_detect_result[i].get_result = FALSE;
        cpt_detect_result[i].cur_disc = FALSE;
        cpt_detect_result[i].pr = FALSE;
    }

    if ( (NULL == sextension)||(NULL == dextension) )
    {
        g_adt_rt_state.err_code = ADT_CPT_ERR_INVALID_EXT;
        goto safe_exit1;
    }

    if ( (strlen(sextension)>AST_MAX_EXTENSION)||
         (strlen(dextension)>AST_MAX_EXTENSION) )
    {
        g_adt_rt_state.err_code = ADT_CPT_ERR_INVALID_EXT;
        goto safe_exit1;
    }

    if ( schannum > DAHDI_MAX_CHANNELS || 
         dchannum > DAHDI_MAX_CHANNELS )
    {
        g_adt_rt_state.err_code = ADT_CPT_ERR_INVALID_CHAN;
        goto safe_exit1;
    }

    snprintf(dahdidev,sizeof(dahdidev),"%s/%d",dahdipath, schannum);

    cpt_adt_thread.fd_caller = -1;
    cpt_adt_thread.fd_called = -1;

    cpt_adt_thread.fd_caller = open(dahdidev,O_RDWR);
    if ( cpt_adt_thread.fd_caller < 0 )
    {
        g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_ABSENT;
        ast_log(LOG_ERROR,"%s absent: %s\n",dahdidev,strerror(errno));
        goto safe_exit1;
    }
    if( dchannum < 0 )
    {
        detectmode = 1;  
        user_action = USER_DONONE; 
        g_adt_rt_state.pickup_timeout = 45;
        g_adt_rt_state.hangup_timeout = 45;
        ast_log(LOG_DEBUG,"Start semi-automatic detection, External Number %s, \n", dextension);
        cpt_adt_thread.fd_called = -1;
    } 
    else
    {
        detectmode = 0;
        user_action = USER_DONONE; 
        g_adt_rt_state.pickup_timeout = 0;
        g_adt_rt_state.hangup_timeout = 0;
        ast_log(LOG_DEBUG,"Start automatic detection, Internal Number %s\n", dextension);
        snprintf(dahdidev,sizeof(dahdidev),"%s/%d",dahdipath,dchannum); 
        cpt_adt_thread.fd_called = open(dahdidev,O_RDWR);
        if ( cpt_adt_thread.fd_called < 0 )
        {
            g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_ABSENT;
            ast_log(LOG_ERROR,"%s absent: %s\n",dahdidev,strerror(errno));
            close(cpt_adt_thread.fd_caller);
            cpt_adt_thread.fd_caller = -1;
            goto safe_exit1; 
        }
    }

    pthread_cleanup_push(cpt_autodetect_cleanup, &cpt_adt_thread);
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
    
    if ( !(fxo_port_is_connected(cpt_adt_thread.fd_caller,schannum,ON_HOOK)) )
    {
        g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_DISCONNECT;
        goto safe_exit0;
    }

    if ((!detectmode) && cpt_adt_thread.fd_called >0 )
    {
        if( !fxo_port_is_connected(cpt_adt_thread.fd_called,dchannum,ON_HOOK) ){
            g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_DISCONNECT;
            goto safe_exit0;
        }
    }

    port[0] = schannum;
    port[1] = dchannum;
    fd_array[0] = cpt_adt_thread.fd_caller;
    fd_array[1] = cpt_adt_thread.fd_called;

    call_prog_detect_running = 1;
    if ( monitor_start() < 0 )
    {
        g_adt_rt_state.err_code = ADT_CPT_ERR_MTASK_INIT_FAILED;
        ast_log(LOG_ERROR, "Failed to init monitor task\n");
        goto safe_exit0;
    }

    /** Configure both ports for full-automatic detection and 
              Configure 'caller' port for semi-automatic detection */
    if ( (FALSE==configure_port(cpt_adt_thread.fd_caller,schannum)))
    {
        g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_CONFIG_FAILED;
        ast_log(AST_LOG_ERROR, "Failed to configure ports %d\n", schannum);
        goto safe_exit0;
    }
    if((!detectmode) && cpt_adt_thread.fd_called >0)
    {
        if(FALSE == configure_port(cpt_adt_thread.fd_called,dchannum))
        {
            g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_CONFIG_FAILED;
            ast_log(AST_LOG_ERROR, "Failed to configure ports %d\n", dchannum);
            goto safe_exit0;
        }
    }
    
    monitorstate = OFFHOOK;
    /*caller goes offhook*/
    if ( !call_prog_detect_running||(FALSE==go_offhook(cpt_adt_thread.fd_caller,schannum)) )
    {
        g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_OFFHOOK_FAILED;
        goto safe_exit0;
    }
    /* start detection after 1s */
    pthread_testcancel();
    sleep(1);

#if !defined(BUSYTONE_DETECT_ONLY)    
    g_adt_rt_state.stat_id = ADT_CPT_DIAL_TONE;
    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    manager_event(EVENT_FLAG_SYSTEM, "cpt_autodetect_state", "FXO: %d\r\nNumber: %s\r\nPort: %d\r\nDest: %s\r\nstate: %s\r\n", 
                  g_adect_cpt_usrcfg.schannum, g_adect_cpt_usrcfg.sextension, 
                  g_adect_cpt_usrcfg.dchannum, g_adect_cpt_usrcfg.dextension,
                  adt_cpt_state_str[g_adt_rt_state.stat_id]); 
    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    /* Detection logic for dial tone on caller port */
    if ( call_prog_detect_running && (TRUE==tone_detected(cpt_adt_thread.fd_caller, schannum, DIAL_TONE, 80000)) )
    {
        if ( call_prog_detect_running && (FALSE==dial_extension(cpt_adt_thread.fd_caller, schannum, dextension)) )
        {
            g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_DIAL_FAILED;
            goto safe_exit0;
        }
    }
    else
    {
        if ( fxo_dis_in_bz_dect )
            g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_DISCONNECT;
        else
            g_adt_rt_state.err_code = ADT_CPT_ERR_NO_DIALTONE;

        ast_log(AST_LOG_ERROR, "No dial tone detected on caller port\n");
        goto safe_exit0;
    }
#else 
    sleep(1);
    if ( call_prog_detect_running && (FALSE==dial_extension(cpt_adt_thread.fd_caller, schannum, dextension)) )
    {
        g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_DIAL_FAILED;
        goto safe_exit0;
    }
#endif
    pthread_testcancel();
    monitorstate=WAITFORRING;
    while(1)
    {
        if((RUNWAITFORRING == monitorstate)||(TRUE == ringing))
        {
            ast_log(AST_LOG_DEBUG, "Waiting for ringing\n");
            break;
        }
        sleep(1);
        dial_timeout --;
        if(dial_timeout < 0){
            ast_log(AST_LOG_ERROR, "Waiting dial completed timeout!!\n");
            break;
        }
    }; 
    pthread_cleanup_push(release_lock_cleanup, NULL);
    acquire_lock();
    pthread_testcancel();
    sleep(1);
    pthread_testcancel();
    release_lock();
    pthread_cleanup_pop(0);

    /* Detect ringing tone on called port */
    g_adt_rt_state.stat_id = ADT_CPT_RB_TONE;
    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    manager_event(EVENT_FLAG_SYSTEM, "cpt_autodetect_state", "FXO: %d\r\nNumber: %s\r\nPort: %d\r\nDest: %s\r\nstate: %s\r\n", 
                  g_adect_cpt_usrcfg.schannum, g_adect_cpt_usrcfg.sextension, 
                  g_adect_cpt_usrcfg.dchannum, g_adect_cpt_usrcfg.dextension,
                  adt_cpt_state_str[g_adt_rt_state.stat_id]);  
    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
                  
    if ( ringing == TRUE )
    {
#if !defined(BUSYTONE_DETECT_ONLY)
        ast_log(AST_LOG_DEBUG, "Detect ringback tone ...%s\n", detectmode ? "Dont pick up the call!!\n":"!");
        monitorstate=RINGBACK;
        /* Detect ringback tone on caller port */
        if ( call_prog_detect_running && TRUE == tone_detected(cpt_adt_thread.fd_caller,
                                                               schannum,RINGBACK_TONE,320000) )
        {
            ast_log(AST_LOG_NOTICE, "Detect ringback tone finished\n");
            pthread_testcancel();
            if(!detectmode)//full-automatic
            {
                sleep(10);
                monitorstate=PICKUPCALL;
                if ( !call_prog_detect_running ||FALSE==pickup_call(cpt_adt_thread.fd_called,
                                                                    dchannum) )
                {
                    g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_PICKUP_FAILED;
                    goto safe_exit0;
                }
            }
            else //semi-automatic
            {
                user_action = USER_DONONE;
                sleep(2);
                monitorstate=PICKUPCALL;
                g_adt_rt_state.stat_id = ADT_CPT_WAIT_PICKUP;
                /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
                manager_event(EVENT_FLAG_SYSTEM, "cpt_autodetect_state", "FXO: %d\r\nNumber: %s\r\nPort: %d\r\nDest: %s\r\nstate: %s\r\n", 
                              g_adect_cpt_usrcfg.schannum, g_adect_cpt_usrcfg.sextension, 
                              g_adect_cpt_usrcfg.dchannum, g_adect_cpt_usrcfg.dextension,
                              adt_cpt_state_str[g_adt_rt_state.stat_id]); 
                /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
                ast_log(AST_LOG_NOTICE, "Pickup the External call within a few seconds!!\n");
                pthread_testcancel();
            #if 0
                sleep(8);
            #else
                while ( g_adt_rt_state.pickup_timeout > 0)
                /* waiting for web ui tell us that the call has been pick up and go to next stage.*/
                { 
                    if(user_action == USER_PICKUPED)
                    {
                        user_action = USER_DONONE;
                        break;
                    }
                    
                    g_adt_rt_state.pickup_timeout--;
                    sleep(1);
                }
                if( g_adt_rt_state.pickup_timeout <= 0)
                {
                    ast_log(AST_LOG_NOTICE, "Pickup the External call timeout!!\n");
                    user_action = USER_DONONE;
                    g_adt_rt_state.err_code = ADT_CPT_ERR_USER_PICKUP_TIMEOUT;
                    goto safe_exit0;
                }
            #endif
            }
        }
        else
        {
            ast_log(AST_LOG_ERROR, "No ringback tone detected on caller port\n");

            if ( fxo_dis_in_bz_dect )
                g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_DISCONNECT;
            else
                g_adt_rt_state.err_code = ADT_CPT_ERR_USER_NO_RBTONE;

            goto safe_exit0;
        }
#else
        if(!detectmode)//full-automatic
        {
            ast_log(AST_LOG_DEBUG, "Detect ringing on port %d.\n", dchannum);
            pthread_testcancel();
            sleep(5);
            monitorstate=PICKUPCALL;
            if ( !call_prog_detect_running ||FALSE==pickup_call(cpt_adt_thread.fd_called,
                                                                dchannum) )
            {
                g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_PICKUP_FAILED;
                goto safe_exit0;
            }
        }
        else
        {
              pthread_testcancel();
              user_action = USER_DONONE;
              sleep(2);
              monitorstate=PICKUPCALL;
              g_adt_rt_state.stat_id = ADT_CPT_WAIT_PICKUP;
              /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
              manager_event(EVENT_FLAG_SYSTEM, "cpt_autodetect_state", "FXO: %d\r\nNumber: %s\r\nPort: %d\r\nDest: %s\r\nstate: %s\r\n", 
                            g_adect_cpt_usrcfg.schannum, g_adect_cpt_usrcfg.sextension, 
                            g_adect_cpt_usrcfg.dchannum, g_adect_cpt_usrcfg.dextension,
                            adt_cpt_state_str[g_adt_rt_state.stat_id]);
              /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
              ast_log(AST_LOG_NOTICE, "Pickup the External call within a few seconds!!\n");
              pthread_testcancel();
#if 0
              sleep(8);
#else
              while (g_adt_rt_state.pickup_timeout > 0)
              /* waiting for web ui tell us that the call has been pick up and go to next stage.*/
              { 
                  if(user_action == USER_PICKUPED)
                  {
                      user_action = USER_DONONE;
                      break;
                  }
                  
                  g_adt_rt_state.pickup_timeout--;
                  sleep(1);
              }
              if(g_adt_rt_state.pickup_timeout <= 0)
              {
                  ast_log(AST_LOG_NOTICE, "Pickup the External call timeout!!\n");
                  user_action = USER_DONONE;
                  g_adt_rt_state.err_code = ADT_CPT_ERR_USER_PICKUP_TIMEOUT;
                  goto safe_exit0;
              }
#endif
        }
#endif
        pthread_testcancel();
        sleep(1);
        monitorstate=SENDDTMF;
        if ( !call_prog_detect_running || FALSE==send_dtmf(cpt_adt_thread.fd_caller,
                                                           schannum,sextension) )
        {
            g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_DTMFSEND_FAILED;
            goto safe_exit0;
        }
        pthread_testcancel();
        sleep(2);
    }
    else
    {
        ast_log(AST_LOG_ERROR, "No ringing detected on port %d\n", dchannum);
        call_prog_detect_running = 0;
        if ( !call_prog_detect_running || (FALSE==hangup(cpt_adt_thread.fd_caller,
                                                         schannum)) )
        {
            g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_NO_RINGING;
            goto safe_exit0;
        }
    }
    if(!detectmode)//full-automatic
    {
        monitorstate=HANGUP;
        if ( !call_prog_detect_running || FALSE==hangup(cpt_adt_thread.fd_called,
                                                        dchannum) )
        {
            g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_HANGUP_FAILED;
            goto safe_exit0;
        }
    }
    else
    {
        monitorstate=BUSY;
        g_adt_rt_state.stat_id = ADT_CPT_WAIT_HANGUP;
        /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        manager_event(EVENT_FLAG_SYSTEM, "cpt_autodetect_state", "FXO: %d\r\nNumber: %s\r\nPort: %d\r\nDest: %s\r\nstate: %s\r\n", 
                      g_adect_cpt_usrcfg.schannum, g_adect_cpt_usrcfg.sextension, 
                      g_adect_cpt_usrcfg.dchannum, g_adect_cpt_usrcfg.dextension,
                      adt_cpt_state_str[g_adt_rt_state.stat_id]);         
        /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        ast_log(AST_LOG_NOTICE, "Hangup the External call for Busy Tone detect!!\n");
#if 0
        sleep(4);
#else
        while (g_adt_rt_state.hangup_timeout > 0)
        /* waiting for web ui tell us that the call has been pick up and go to next stage.*/
        { 
            if(user_action == USER_HANGUPED)
            {
                user_action = USER_DONONE;
                break;
            }
          
            g_adt_rt_state.hangup_timeout--;
            sleep(1);
        }
        if(g_adt_rt_state.hangup_timeout <= 0)
        {
            ast_log(AST_LOG_NOTICE, "Hangup the External call timeout!!\n");
            user_action = USER_DONONE;
            g_adt_rt_state.err_code = ADT_CPT_ERR_USER_HANGUP_TIMEOUT;
            goto safe_exit0;
        }
#endif
    }
    g_adt_rt_state.stat_id = ADT_CPT_BUSY_TONE;
    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    manager_event(EVENT_FLAG_SYSTEM, "cpt_autodetect_state", "FXO: %d\r\nNumber: %s\r\nPort: %d\r\nDest: %s\r\nstate: %s\r\n", 
                  g_adect_cpt_usrcfg.schannum, g_adect_cpt_usrcfg.sextension, 
                  g_adect_cpt_usrcfg.dchannum, g_adect_cpt_usrcfg.dextension,
                  adt_cpt_state_str[g_adt_rt_state.stat_id]); 
   /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    monitorstate=BUSY;
    pthread_testcancel();
    sleep(4);
    //[Bug 18304], more time for detect current disconnect.
    while(busycheck && battery_lost) {
        sleep(1);
        if(--wait_bat_present < 0)
            break;
    }
    sleep(4);
    ast_log(AST_LOG_DEBUG, "Detect busy tone ...\n");
    if ( !call_prog_detect_running || FALSE==tone_detected(cpt_adt_thread.fd_caller,
                                                           schannum,BUSY_TONE,320000) )
    {
         if ( fxo_dis_in_bz_dect )
             g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_DISCONNECT;
         else
             g_adt_rt_state.err_code = ADT_CPT_ERR_USER_NO_BZTONE;

        goto safe_exit0;
    }
    monitorstate = HANGUP;
    if ( !call_prog_detect_running || FALSE==hangup(cpt_adt_thread.fd_caller,
                                                    schannum) )
    {
        g_adt_rt_state.err_code = ADT_CPT_ERR_FXO_HANGUP_FAILED;
        goto safe_exit0;
    }

    safe_exit0:
    call_prog_detect_running = 0;

    if ( adt_cpt_monitor_thread!= AST_PTHREADT_NULL )
    {
        sleep(1);
        if ( monitorstate != MONITOR_EXIT )
        {
            ast_log(LOG_ERROR,"cpt monitor thread exit failed\n");
            pthread_kill(adt_cpt_monitor_thread, SIGURG);
        }
        adt_cpt_monitor_thread = AST_PTHREADT_NULL;
    }
    if(cpt_adt_thread.fd_caller>0)
    {
        close(cpt_adt_thread.fd_caller);
        cpt_adt_thread.fd_caller = -1;
    }
    if(cpt_adt_thread.fd_called>0) 
    {
        close(cpt_adt_thread.fd_called);
        cpt_adt_thread.fd_called = -1;
    }
    load_chan_dahdi();
    g_adt_rt_state.stat_id = ADT_CPT_EXIT;
    
	/* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    if ( cpt_detect_result[BUSY_TONE].get_result == TRUE )
    {
        if ( cpt_detect_result[BUSY_TONE].cadence[0].on )
        {
            snprintf(freq, sizeof(freq), "%d", cpt_detect_result[BUSY_TONE].f_hz[0]);
            if ( cpt_detect_result[BUSY_TONE].f_hz[1] )
            {
                snprintf(freq + strlen(freq), sizeof(freq) - strlen(freq), 
                         "+%d", cpt_detect_result[BUSY_TONE].f_hz[1]);
            }
           
            snprintf(cadence + strlen(cadence), sizeof(cadence) - strlen(cadence), 
                                 "c=%d/%d", cpt_detect_result[BUSY_TONE].cadence[0].on, cpt_detect_result[BUSY_TONE].cadence[0].off);
            if ( cpt_detect_result[BUSY_TONE].cadence[1].on && cpt_detect_result[BUSY_TONE].cadence[1].off )
            {
                snprintf(cadence + strlen(cadence), sizeof(cadence) - strlen(cadence), 
                         "-%d/%d", cpt_detect_result[BUSY_TONE].cadence[1].on, cpt_detect_result[BUSY_TONE].cadence[1].off);
            }
        }
    }   
    manager_event(EVENT_FLAG_SYSTEM, "cpt_autodetect_state", "FXO: %d\r\nNumber: %s\r\nPort: %d\r\nDest: %s\r\nstate: %s\r\nErrCod: %d\r\nCurrentDisconnect: %s\r\nlapsedtime: %.2f\r\nPolarityReversal: %s\r\n%s: %s\r\nfrequencies: %s\r\ncadence: %s\r\n", 
                  g_adect_cpt_usrcfg.schannum, g_adect_cpt_usrcfg.sextension, 
                  g_adect_cpt_usrcfg.dchannum, g_adect_cpt_usrcfg.dextension,
                  adt_cpt_state_str[g_adt_rt_state.stat_id],g_adt_rt_state.err_code,
                  cpt_detect_result[BUSY_TONE].cur_disc?"yes":"no",
                  elapsedtime,
                  cpt_detect_result[BUSY_TONE].pr?"yes":"no",
                  adt_cpt_tonetype_str[BUSY_TONE],
                  cpt_detect_result[BUSY_TONE].cadence[0].on?"yes":"no",
                  cpt_detect_result[BUSY_TONE].cadence[0].on?freq:"",
                  cpt_detect_result[BUSY_TONE].cadence[0].on?cadence:""
                  ); 
    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
                  
    g_adt_item = ADT_NON;
    ast_log(AST_LOG_NOTICE, "Auto Detect CPT finished\n");
    pthread_cleanup_pop(0);
    return NULL;

    safe_exit1:
    load_chan_dahdi();
    g_adt_rt_state.stat_id = ADT_CPT_EXIT;
   /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */   
    if ( cpt_detect_result[BUSY_TONE].get_result == TRUE )
    {
        if ( cpt_detect_result[BUSY_TONE].cadence[0].on )
        {
            snprintf(freq, sizeof(freq), "%d", cpt_detect_result[BUSY_TONE].f_hz[0]);
            if ( cpt_detect_result[BUSY_TONE].f_hz[1] )
            {
                snprintf(freq + strlen(freq), sizeof(freq) - strlen(freq), 
                         "+%d", cpt_detect_result[BUSY_TONE].f_hz[1]);
            }
           
            snprintf(cadence + strlen(cadence), sizeof(cadence) - strlen(cadence), 
                                 "c=%d/%d", cpt_detect_result[BUSY_TONE].cadence[0].on, cpt_detect_result[BUSY_TONE].cadence[0].off);
            if ( cpt_detect_result[BUSY_TONE].cadence[1].on && cpt_detect_result[BUSY_TONE].cadence[1].off )
            {
                snprintf(cadence + strlen(cadence), sizeof(cadence) - strlen(cadence), 
                         "-%d/%d", cpt_detect_result[BUSY_TONE].cadence[1].on, cpt_detect_result[BUSY_TONE].cadence[1].off);
            }
        }
    }
    manager_event(EVENT_FLAG_SYSTEM, "cpt_autodetect_state", "FXO: %d\r\nNumber: %s\r\nPort: %d\r\nDest: %s\r\nstate: %s\r\nErrCod: %d\r\nCurrentDisconnect: %s\r\nlapsedtime: %.2f\r\nPolarityReversal: %s\r\n%s: %s\r\nfrequencies: %s\r\ncadence: %s\r\n", 
                  g_adect_cpt_usrcfg.schannum, g_adect_cpt_usrcfg.sextension, 
                  g_adect_cpt_usrcfg.dchannum, g_adect_cpt_usrcfg.dextension,
                  adt_cpt_state_str[g_adt_rt_state.stat_id],g_adt_rt_state.err_code,
                  cpt_detect_result[BUSY_TONE].cur_disc?"yes":"no",
                  elapsedtime,
                  cpt_detect_result[BUSY_TONE].pr?"yes":"no",
                  adt_cpt_tonetype_str[BUSY_TONE],
                  cpt_detect_result[BUSY_TONE].cadence[0].on?"yes":"no",
                  cpt_detect_result[BUSY_TONE].cadence[0].on?freq:"",
                  cpt_detect_result[BUSY_TONE].cadence[0].on?cadence:""
                  ); 
    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    g_adt_item = ADT_NON;
    return NULL; 
}

static int cpt_autodetect_is_finished(void)
{
    if ( g_cpt_auto_detect_thread == AST_PTHREADT_NULL )
        return TRUE;

    if ( (g_adt_rt_state.stat_id == ADT_CPT_NON)||
         (g_adt_rt_state.stat_id == ADT_CPT_EXIT)||
         (g_adt_rt_state.stat_id ==ADT_CPT_CANCEL) )
    {
        return TRUE;
    }

    return FALSE;
}

static void dahdi_pstn_cpt_autodetect_thread_release(int cli_fd, int concise)
{
    int cancel_timeout = 0;

    if ( g_cpt_auto_detect_thread != AST_PTHREADT_NULL )
    {
        if ( (g_adt_rt_state.stat_id == ADT_CPT_NON) || 
             (g_adt_rt_state.stat_id == ADT_CPT_EXIT) )
        {
            if ( concise == 1 )
            {
                ast_cli(cli_fd, "FXO=%d!type=CPT!state=exited\n", g_adect_cpt_usrcfg.schannum);
            }
            else
            {
                ast_cli(cli_fd, "dahdi pstn cpt auto detect thread has been exit!!\n");
            }
        }
        else
        {

            pthread_cancel(g_cpt_auto_detect_thread);

            while ( 1 )
            {
                usleep(10000);
                cancel_timeout ++;
                if ( (g_adt_rt_state.stat_id == ADT_CPT_NON) || 
                     (g_adt_rt_state.stat_id == ADT_CPT_EXIT) ||
                     (g_adt_rt_state.stat_id == ADT_CPT_CANCEL) )
                {
                    if ( concise == 1 )
                    {
                        ast_cli(cli_fd, "FXO=%d!type=CPT!state=canceled\n", g_adect_cpt_usrcfg.schannum);
                    }
                    else
                    {
                        ast_cli(cli_fd, "dahdi pstn cpt auto detect thread be canceled!!\n");
                    }
                    break;
                }
                if ( cancel_timeout > 1000 )
                {
                    if ( concise == 1 )
                    {
                        ast_cli(cli_fd, "FXO=%d!type=CPT!state=cancel timeout\n", g_adect_cpt_usrcfg.schannum);
                    }
                    else
                    {
                        ast_cli(cli_fd, "dahdi pstn cpt auto detect thread cancel timeout!!\n");
                    }
                    break;
                }
            }
        }

        g_cpt_auto_detect_thread = AST_PTHREADT_NULL;
    }

    if ( concise != 1 )
    {
        ast_cli(cli_fd, "dahdi pstn cpt auto detect thread exit done!! g_adt_item %d\n", g_adt_item);
    }

    if ( g_adt_item == ADT_CPT )
    {
        g_adt_item = ADT_NON;
    }
}

static int dahdi_pstn_cpt_autodetect_start( int cli_fd, adect_cpt_usrcfg_t *usrcfg, int concise )
{
    pthread_attr_t attr;
    struct sched_param param;
    int policy = SCHED_OTHER;
    int priority_max;
    int priority_min;
    int res = -1;

    if ( ADT_ACIM==g_adt_item )
    {
        if ( !acim_autodetect_is_finished() )
        {
            ast_log(LOG_ERROR, "PSTN Auto detect ACIM is running!\n");

            /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
            if(concise == 1 )
            {
                ast_cli(cli_fd, "FXO=%d!state=failed!ErrCod=%d\n", usrcfg->schannum, ADT_CPT_ERR_ACIM_DETECT);
            }
            /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
            
            return -1;
        }
    }

    if ( !cpt_autodetect_is_finished() )
    {
        ast_log(LOG_ERROR, "PSTN Auto detect CPT is running!\n");

        /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        if(concise == 1 )
        {
            ast_cli(cli_fd, "FXO=%d!state=failed!ErrCod=%d\n", usrcfg->schannum, ADT_CPT_ERR_DETECT_BUSY);
        }
        /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        
        return -1;
    }
    else if ( g_cpt_auto_detect_thread != AST_PTHREADT_NULL )
    {
        g_cpt_auto_detect_thread = AST_PTHREADT_NULL;
    }

    g_adt_item = ADT_CPT;

    res = unload_chan_dahdi();
    if ( res<0 )
    {
        g_adt_item = ADT_NON;

        /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        if(concise == 1 )
        {
            ast_cli(cli_fd, "FXO=%d!state=failed!ErrCod=%d\n", usrcfg->schannum, ADT_CPT_ERR_UNLOAD_DAHDI_FAILED);
        }
        /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        
        return -1;
    }

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    pthread_attr_setstacksize(&attr, 1024 * 1024);

    param.sched_priority = g_cpt_auto_detect_thread_pri;
    res = pthread_attr_setschedparam(&attr, &param);
    if ( res )
    {
        if ( concise != 1 )
        {
            ast_cli(cli_fd, "[WARNING] Pthread set sched_priority=%d failed, res:%d, errno:%d(%s)!\n",
                    param.sched_priority, res, errno, strerror(errno));
        }
    }

    res = pthread_attr_getschedparam(&attr, &param);
    if ( res )
    {
        if ( concise != 1 )
        {
            ast_cli(cli_fd, "[WARNING] Pthread get schedparam failed, res:%d, errno:%d(%s)!\n",
                    res, errno, strerror(errno));
        }
    }
    else
    {
        //ast_cli(cli_fd, "[NOTICE] Dahdi pstn cpt autodetect thread get priority:%d.\n", param.sched_priority);
    }

    res = pthread_attr_getschedpolicy(&attr, &policy);
    if ( res )
    {
        if ( concise != 1 )
        {
            ast_cli(cli_fd, "[WARNING] Pthread get schedpolicy failed, res:%d, errno:%d(%s)!\n",
                    res, errno, strerror(errno));
        }
    }
    else
    {
        // ast_cli(cli_fd, "[NOTICE] pstn cpt autodetect thread get policy:%d.\n", policy);
    }

    priority_max = sched_get_priority_max(policy);
    priority_min = sched_get_priority_min(policy);

    ast_cli(cli_fd, "[NOTICE] priority max:%d, min:%d.\n", priority_max, priority_min);

    g_adect_cpt_usrcfg.dchannum = usrcfg->dchannum;
    g_adect_cpt_usrcfg.schannum = usrcfg->schannum;
    ast_copy_string(g_adect_cpt_usrcfg.sextension, usrcfg->sextension, AST_MAX_EXTENSION);
    ast_copy_string(g_adect_cpt_usrcfg.dextension, usrcfg->dextension, AST_MAX_EXTENSION);

    res = ast_pthread_create_background(&g_cpt_auto_detect_thread, &attr,
                                        dahdi_pstn_cpt_autodetect_thread, NULL);

    pthread_attr_destroy(&attr);

    if ( res != 0 )
    {
        if ( concise == 1 )
        {
            /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
            ast_cli(cli_fd, "FXO=%d!type=CPT!state=failed!ErrCod=%d\n", usrcfg->schannum, ADT_CPT_ERR_THREAD_CREAT_FAILED);
            /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
        }
        else
        {
            ast_cli(cli_fd, "[ERROR] Create pstn auto detect thread failed! res:%d, errno:%d(%s) !!\n",
                    res, errno, strerror(errno));
        }
        g_cpt_auto_detect_thread = AST_PTHREADT_NULL;
        load_chan_dahdi();
        g_adt_item = ADT_NON;
        g_adt_rt_state.stat_id = ADT_CPT_EXIT;
        g_adt_rt_state.err_code = ADT_CPT_ERR_THREAD_CREAT_FAILED;
        return -1;
    }

    if ( concise == 1 )
    {
        ast_cli(cli_fd, "FXO=%d!type=CPT!state=started\n", usrcfg->schannum);
    }
    else
    {
        ast_cli(cli_fd, "[NOTICE] pstn auto detect start successful. pstn auto detect thread:%lu ...\n",
                g_cpt_auto_detect_thread);
    }

    return 0;
}

static char *dahdi_pstn_cpt_autodetect_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int ret;
    int concise = 0;
    adect_cpt_usrcfg_t usrcfg;

    switch ( cmd )
    {
        case CLI_INIT:
            e->command = "pstn autodetect cpt_start";
            e->usage = "Usage: pstn autodetect cpt_start \n \
       [calling fxo port] [called fxo port] \n \
       [Calling extension] [Called extension] [concise]\n";
            return NULL;
        case CLI_GENERATE:
            return NULL;
    }

    if ( a->argc < 7 )
    {
        return CLI_SHOWUSAGE;
    }
    
    if ( a->argv[7] != NULL && strcmp(a->argv[7], "concise") == 0 )
    {
        concise = 1;
    }

    usrcfg.schannum = atoi(a->argv[3]);
    usrcfg.dchannum = atoi(a->argv[4]);
    
    if ( concise != 1 )
    {
        ast_cli(a->fd, "PSTN CPT Auto Detect Start.... fxo%d fxo%d \n",usrcfg.schannum , usrcfg.dchannum ); 
    }
    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    else
    {
        ast_cli(a->fd, "PSTN AutoDetect Start\n"); 
        ast_cli(a->fd, "type=CPT\n");
    }
    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
/* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
#if 0
/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    if ( (strlen(a->argv[5])>AST_MAX_EXTENSION)||(strlen(a->argv[6])>AST_MAX_EXTENSION) )
    {
        ast_cli(a->fd, "[ERR]PSTN extension length too long!!\n");
        return CLI_SUCCESS;
    }
/* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
#endif
/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    ast_copy_string(usrcfg.sextension, a->argv[5], strlen(a->argv[5])+1);
    ast_copy_string(usrcfg.dextension, a->argv[6], strlen(a->argv[6])+1);

    ast_mutex_lock(&adt_lock);
    ret = dahdi_pstn_cpt_autodetect_start(a->fd, &usrcfg, concise);
    ast_mutex_unlock(&adt_lock);

    if ( ret<0 )
    {
/* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
#if 0
/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        if ( concise == 1 )
        {
            ast_cli(a->fd, "FXO=%d!type=CPT!state=failed\n", usrcfg.schannum); 
        }
        else
/* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
#endif
/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        {
            ast_cli(a->fd, "PSTN CPT Auto Detect failed....\n"); 
        }
    }

    return CLI_SUCCESS;

}
static char *dahdi_pstn_cpt_autodetect_getstate_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int i;
    int concise=0;
    switch ( cmd )
    {
        case CLI_INIT:
            e->command = "pstn autodetect cpt_state";
            e->usage = "Usage: pstn autodetect cpt_state [concise]\n";
            return NULL;
        case CLI_GENERATE:
            return NULL;
    }

    if ( a->argc < 3 )
    {
        return CLI_SHOWUSAGE;
    }
    
    if ( a->argv[3] != NULL && strcmp( a->argv[3], "concise") == 0 )
    {
        concise = 1;
    }
    
    if ( concise == 1 )
    {
        /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        ast_cli(a->fd, "PSTN AutoDetect State\n"); 
        /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        ast_cli(a->fd, "FXO=%d!Number=%s!Port=%d!Dest=%s!type=CPT!state=%s\nstateID=%d!ErrCod=%d!T1OUT=%d!T2OUT=%d\n",
            g_adect_cpt_usrcfg.schannum, g_adect_cpt_usrcfg.sextension, 
            g_adect_cpt_usrcfg.dchannum, g_adect_cpt_usrcfg.dextension,
            adt_cpt_state_str[g_adt_rt_state.stat_id], g_adt_rt_state.stat_id,
            g_adt_rt_state.err_code, g_adt_rt_state.pickup_timeout, g_adt_rt_state.hangup_timeout);
    }
    else
    {
        ast_cli(a->fd, "PSTN CPT Auto Detect state %s error:%s....\n", 
            adt_cpt_state_str[g_adt_rt_state.stat_id], adt_cpt_error_str[g_adt_rt_state.err_code]);
    }
    
    for ( i=0; i<TONE_DETECT_NUM; i++ )
    {
        if ( i == BUSY_TONE )
        {
            if ( cpt_detect_result[BUSY_TONE].cur_disc )
            {
                if ( concise == 1 )
                {
                     if ( elapsedtime < cd_timeout )
                    {
                        /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                        ast_cli(a->fd,"CurrentDisconnect=yes!lapsedtime=%.2f!", elapsedtime );
                        /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                    }
                }
                else
                {
                    ast_cli(a->fd,"[%s] NON; current disconnect detected: %.2f ms\n",
                            adt_cpt_tonetype_str[i], elapsedtime );
                }
                //continue;
            }
            else
            {
                if ( concise == 1 )
                {
                    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                    ast_cli(a->fd,"CurrentDisconnect=no!");
                    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                }
            }
            if ( cpt_detect_result[BUSY_TONE].pr )
            {
                if ( concise == 1 )
                {
                    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                    ast_cli(a->fd,"PolarityReversal=yes!\n");
                    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                }
                else
                {
                    ast_cli(a->fd,"[%s] NON; Polarity Reversal Detected!\n",
                            adt_cpt_tonetype_str[i] );
                }
                //continue;
            }
            else
            {
                if ( concise == 1 )
                {
				    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                    ast_cli(a->fd,"PolarityReversal=no!\n");
                    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                }
            }
        }
        if ( cpt_detect_result[i].get_result == TRUE )
        {

            if ( cpt_detect_result[i].cadence[0].on )
            {
                if ( concise == 1 )
                {
                    char pstn_buf[1024] = {0};
                    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                    snprintf(pstn_buf, sizeof(pstn_buf), "%s=yes!frequencies=%d", adt_cpt_tonetype_str[i],
                             cpt_detect_result[i].f_hz[0]);
				    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                    if ( cpt_detect_result[i].f_hz[1] )
                    {
					    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                        snprintf(pstn_buf + strlen(pstn_buf), sizeof(pstn_buf) - strlen(pstn_buf), 
                                 "+%d", cpt_detect_result[i].f_hz[1]);
					     /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                    }
                    
					/* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                    snprintf(pstn_buf + strlen(pstn_buf), sizeof(pstn_buf) - strlen(pstn_buf), 
                             "!c=%d/%d", cpt_detect_result[i].cadence[0].on, cpt_detect_result[i].cadence[0].off);
                    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                    if ( cpt_detect_result[i].cadence[1].on && cpt_detect_result[i].cadence[1].off )
                    {
                        snprintf(pstn_buf + strlen(pstn_buf), sizeof(pstn_buf) - strlen(pstn_buf), 
                                 "-%d/%d", cpt_detect_result[i].cadence[1].on, cpt_detect_result[i].cadence[1].off);
                    }
                    ast_cli(a->fd,"%s\n", pstn_buf);
                }
                else
                {
                    ast_cli(a->fd, "[%s] Freq1:[%d]HZ Freq2:[%d]HZ "
                            "wtih cadence[0] %dms ON and %d ms OFF;" 
                            "cadence[1] %dms ON and %d ms OFF;\n",
                            adt_cpt_tonetype_str[i],
                            cpt_detect_result[i].f_hz[0], cpt_detect_result[i].f_hz[1],
                            cpt_detect_result[i].cadence[0].on, cpt_detect_result[i].cadence[0].off,
                            cpt_detect_result[i].cadence[1].on, cpt_detect_result[i].cadence[1].off
                           );
                }
            }
            else if (i != BUSY_TONE)
            {
                if ( concise == 1 )
                {
				    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                    ast_cli(a->fd,"%s=yes!frequencies=%d+%d\n", adt_cpt_tonetype_str[i],
                            cpt_detect_result[i].f_hz[0], cpt_detect_result[i].f_hz[1]);
				    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                }
                else
                {
                    ast_cli(a->fd,"[%s] Freq1:[%d]HZ Freq2:[%d]HZ wtih continous cadence\n",
                            adt_cpt_tonetype_str[i],
                            cpt_detect_result[i].f_hz[0], cpt_detect_result[i].f_hz[1]);
                }
            } 
        }
		/* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        else
        {
            if ( concise == 1 )
            {
                ast_cli(a->fd,"%s=no\n", adt_cpt_tonetype_str[i]);
            }
        }
		/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    }
    return CLI_SUCCESS;
}

static char *dahdi_pstn_cpt_autodetect_useract_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    switch ( cmd )
    {
        case CLI_INIT:
            e->command = "pstn autodetect cpt_useract {PICKUP|HANGUP}";
            e->usage = "Usage: pstn autodetect cpt_useract PICKUP|HANGUP \n";
            return NULL;
        case CLI_GENERATE:
            return NULL;
    }

    if ( a->argc < 4 )
    {
        return CLI_SHOWUSAGE;
    }
    
    if(detectmode)
    {
        if ( !strcasecmp(S_OR(a->argv[3], ""), "PICKUP") )
        {
            ast_cli(a->fd, "PSTN CPT Auto Detect user action: PICKUP\n");
            user_action = USER_PICKUPED;
        } 
        else if ( !strcasecmp(S_OR(a->argv[3], ""), "HANGUP") )
        {
            ast_cli(a->fd, "PSTN CPT Auto Detect user action: HANGUP\n");
            user_action = USER_HANGUPED;
        }
        else
        {
            user_action = USER_DONONE;
        }
    }
    
    return CLI_SUCCESS;
}

/* BEGIN: jizhang 20140418 [Bug 24082] modify */
static char *dahdi_pstn_cpt_autodetect_var_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int act = 0;

    switch ( cmd )
    {
        case CLI_INIT:
            e->command = "pstn autodetect cpt_var {set|get} {cd_timeout|intervaltime}";
            e->usage = "Usage: pstn autodetect cpt_var {set|get} [name] [value]\n";
            return NULL;
        case CLI_GENERATE:
            return NULL;
    }

    if ( a->argc < 5 )
    {
        return CLI_SHOWUSAGE;
    }

    if ( !strcasecmp(S_OR(a->argv[3], ""), "set") )
    {
        if ( a->argc != 6)
        {
            return CLI_SHOWUSAGE;
        }
        act = 1;
    }
    else if( !strcasecmp(S_OR(a->argv[3], ""), "get") )
    {
        act = 0;
    }
    else
    {
        return CLI_SHOWUSAGE;
    }

    if ( !strcasecmp(S_OR(a->argv[4], ""), "cd_timeout") )
    {
        if(act)
        {
            cd_timeout = atoi(a->argv[5]);
            if(cd_timeout > 10000)
                cd_timeout = 10000;
            if(cd_timeout < 100)
                cd_timeout = 100;      //100ms~10s
            ast_cli(a->fd, "PSTN CPT Auto Detect var SET: cd_timeout = %dms\n", cd_timeout);
        }
        else
        {
            ast_cli(a->fd, "PSTN CPT Auto Detect var GET: cd_timeout = %dms\n", cd_timeout);
        }
    }
    else if(!strcasecmp(S_OR(a->argv[4], ""), "intervaltime"))
    {
        if(act)
        {
            intervaltime = atoi(a->argv[5]);
            ast_cli(a->fd, "PSTN CPT Auto Detect var SET: intervaltime = %dms\n", intervaltime);
        }
        else
        {
            ast_cli(a->fd, "PSTN CPT Auto Detect var GET: intervaltime = %dms\n", intervaltime);
        }
    }
    else
    {
        ast_cli(a->fd, "PSTN CPT Auto Detect var: unknown variable\n");
    }

    return CLI_SUCCESS;
}
/* END: jizhang 20140418 [Bug 24082] modify */

static char *dahdi_pstn_cpt_autodetect_stop_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int concise = 0;
    switch ( cmd )
    {
        case CLI_INIT:
            e->command = "pstn autodetect cpt_stop";
            e->usage = "Usage: pstn autodetect cpt_stop [concise]\n";
            return NULL;
        case CLI_GENERATE:
            return NULL;
    }

    if ( a->argc < 3 )
    {
        return CLI_SHOWUSAGE;
    }
    
    if ( a->argv[3] != NULL && strcmp(a->argv[3], "concise") == 0 )
    {
        concise = 1;
    }
    
    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    ast_cli(a->fd, "PSTN AutoDetect Stop\n"); 
	/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */    
    ast_cli(a->fd, "FXO=all!type=CPT!state=Stoping\n");
    ast_mutex_lock(&adt_lock);
    dahdi_pstn_cpt_autodetect_thread_release(a->fd, concise);
    ast_mutex_unlock(&adt_lock);

    ast_cli(a->fd, "FXO=all!type=CPT!state=Stop done\n");
    return CLI_SUCCESS;
}

/*========================PSTN ACIM Auto Detect=================================*/
static int init_acim_autodetect(void)
{
    int i;
    if ( acim_adt_inited )
        return 0;

    if ( use_table )
    {
        init_sinetable();
    }

    for ( i=0; i<MAX_FXO_CHANNEL;i++ )
    {
        adect_acim_usrcfg[i].tostart = 0;
        adect_acim_usrcfg[i].channum = -1;
        adect_acim_usrcfg[i].acim_thread = AST_PTHREADT_NULL;
        adect_acim_usrcfg[i].acim_state = ADT_ACIM_NON;
        adect_acim_usrcfg[i].acim_thread_pri = 0;
    }

    for ( i=0; i<MAX_FXO_CHANNEL;i++ )
    {
        adect_acim_usrcfg[i].acim_doset_mode = ACIM_DOSET_AUTO;   //auto mode default
    }

    /*Init FXO amount*/
    FILE *file = fopen( FXO_AMOUNT_FILE, "r" );
    if ( file != NULL )
    {
        char buff[8] = {0};
	char *ret = NULL;
        ret = fgets (buff, sizeof(buff), file);
        fclose(file);
        ast_log(LOG_DEBUG, "Auto Detecting - init FXO amount as (%s), ret is %s. \n", buff, ret ? ret : "");
        g_fxo_amount = atoi(buff);
    }

    acim_adt_inited =1;
    return 0;
}

/**
 * Makes sure that the line is clear.
 * Right now, we do this by relying on the user to specify how long after dialing the
 * dialstring we can rely on the line being silent (before the telco complains about
 * the user not hitting the next digit).
 * 
 * A more robust way to do this would be to actually measure the sound levels on the line,
 * but that's a lot more complicated, and this should work.
 * 
 * @return 0 if succesful (no errors), 1 if unsuccesful
 */
static int ensure_silence(struct silence_info *info)
{
    struct timeval tv;
    long int elapsedms;
    int x = DAHDI_ONHOOK;
    struct dahdi_dialoperation dop;

    gettimeofday(&tv, NULL);

    if ( info->last_reset.tv_sec == 0 )
    {
        /* this is the first request, we will force it to run */
        elapsedms = -1;
    }
    else
    {
        /* this is not the first request, we will compute elapsed time */
        elapsedms = ((tv.tv_sec - info->last_reset.tv_sec) * 1000L + (tv.tv_usec - info->last_reset.tv_usec) / 1000L);
    }

    ast_log(LOG_DEBUG, "Reset line request received - elapsed ms = %li / reset after = %ld\n", elapsedms, info->reset_after * 1000L);

    if ( elapsedms > 0 && elapsedms < info->reset_after * 1000L )
        return 0;

    ast_log(LOG_DEBUG, "Resetting line\n");

    /* do a line reset */
    /* prepare line for silence */
    /* Do line hookstate reset */

    if ( ioctl(info->device, DAHDI_HOOK, &x) )
    {
        ast_log(LOG_ERROR, "Unable to hang up fd %d\n", info->device);
        return -1;
    }

    sleep(2);
    x = DAHDI_OFFHOOK;
    if ( ioctl(info->device, DAHDI_HOOK, &x) )
    {
        ast_log(LOG_ERROR, "Cannot bring fd %d off hook\n", info->device);
        return -1;
    }
    sleep(2); /* Added to ensure that dial can actually takes place */

    memset(&dop, 0, sizeof(dop));
    dop.op = DAHDI_DIAL_OP_REPLACE;
    dop.dialstr[0] = 'T';
    dahdi_copy_string(dop.dialstr + 1, info->dialstr, sizeof(dop.dialstr));

    if ( ioctl(info->device, DAHDI_DIAL, &dop) )
    {
        ast_log(LOG_ERROR, "Unable to dial!\n");
        return -1;
    }
    sleep(1); 
    sleep(info->initial_delay);  

    gettimeofday(&info->last_reset, NULL);

    return 0;
}

/**
 * Generates a tone of specified frequency.
 * 
 * @param hz the frequency of the tone to be generated
 * @param idx the current sample
 * 		to begenerated.  For a normal waveform you need to increment
 * 		this every time you execute the function.
 *
 * @return 16bit slinear sample for the specified index
 */
static short inline gentone(int hz, int idx)
{
    return amplitude * sin((idx * 2.0 * M_PI * hz)/8000);
}

/* Using DTMF tones for now since they provide good mid band testing 
 * while not being harmonics of each other */
static int freqs[] = {697, 770, 941, 1209, 1336, 1633};
static int freqcount = 6;

/**
 * Generates a waveform of several frequencies.
 * 
 * @param idx the current sample
 * 		to begenerated.  For a normal waveform you need to increment
 * 		this every time you execute the function.
 *
 * @return 16bit slinear sample for the specified index
 */
static short inline genwaveform(int idx)
{
    int i = 0;
    float response = (float)0;
    for ( i = 0; i < freqcount; i++ )
    {
        response += sin((idx * 2.0 * M_PI * freqs[i])/8000);
    }

    return amplitude * response / freqcount;
}


/**
 *  Calculates the RMS of the waveform buffer of samples in 16bit slinear format.
 *  prebuf the buffer of either shorts or floats
 *  bufsize the number of elements in the prebuf buffer (not the number of bytes!)
 *  short_format 1 if prebuf points to an array of shorts, 0 if it points to an array of floats
 *  
 *  Formula for RMS (http://en.wikipedia.org/wiki/Root_mean_square): 
 *  
 *  Xrms = sqrt(1/N Sum(x1^2, x2^2, ..., xn^2))
 *  
 *  Note:  this isn't really a power calculation - but it gives a good measure of the level of the response
 *  
 *  @param prebuf the buffer containing the values to compute
 *  @param bufsize the size of the buffer
 *  @param short_format 1 if prebuf contains short values, 0 if it contains float values
 */
static float power_of(void *prebuf, int bufsize, int short_format)
{
    float sum_of_squares = 0;
    int numsamples = 0;
    float finalanswer = 0;
    short *sbuf = (short*)prebuf;
    float *fbuf = (float*)prebuf;
    int i = 0;

    if ( short_format )
    {
        /* idiot proof checks */
        if ( bufsize <= 0 )
            return -1;

        numsamples = bufsize; /* Got rid of divide by 2 - the bufsize parameter should give the number of samples (that's what it does for the float computation, and it should do it here as well) */

        for ( i = 0; i < numsamples; i++ )
        {
            sum_of_squares += ((float)sbuf[i] * (float)sbuf[i]);
        }
    }
    else
    {
        /* Version for float inputs */
        for ( i = 0; i < bufsize; i++ )
        {
            sum_of_squares += (fbuf[i] * fbuf[i]);
        }
    }

    finalanswer = sum_of_squares/(float)bufsize; /* need to divide by the number of elements in the sample for RMS calc */

    if ( finalanswer < 0 )
    {
        ast_log(LOG_ERROR, "Error: Final answer negative number %f\n", finalanswer);
        return -3;
    }

    return sqrtf(finalanswer);
}

/* 
 * In an effort to eliminate as much as possible the effect of outside noise, we use principles
 * from the Fourier Transform to attempt to calculate the return loss of our signal for each setting.
 *
 * To begin, we send our output signal out on the line.  We then receive back the reflected
 * response.  In the Fourier Transform, each evenly distributed frequency within the window
 * is correlated (multiplied against, then the resulting samples are added together) with
 * the real (cos) and imaginary (sin) portions of that frequency base to detect that frequency.
 * 
 * Instead of doing a complete Fourier Transform, we solve the transform for only our signal
 * by multiplying the received signal by the real and imaginary portions of our reference
 * signal.  This then gives us the real and imaginary values that we can use to calculate
 * the return loss of the sinusoids that we sent out on the line.  This is done by finding
 * the magnitude (think polar form) of the vector resulting from the real and imaginary
 * portions calculated above.
 *
 * This essentially filters out any other noise which maybe present on the line which is outside
 * the frequencies used in our test multi-tone.
 */
void init_sinetable(void)
{
    int i;

    ast_log(LOG_DEBUG, "Using sine tables with %d samples\n", SINE_SAMPLES);

    for ( i = 0; i < SINE_SAMPLES; i++ )
    {
        sintable[i] = sin(((float)i * 2.0 * M_PI )/(float)(SINE_SAMPLES));
    }
}

/* Sine and cosine table lookup to use periodicity of the calculations being done */
float sin_tbl(int arg, int num_per_period)
{
    arg = arg % num_per_period;

    arg = (arg * SINE_SAMPLES)/num_per_period;

    return sintable[arg];
}

float cos_tbl(int arg, int num_per_period)
{
    arg = arg  % num_per_period;

    arg = (arg * SINE_SAMPLES)/num_per_period;

    arg = (arg + SINE_SAMPLES/4) % SINE_SAMPLES;  /* Pi/2 adjustment */

    return sintable[arg];
}


static float db_loss(float measured, float reference)
{
    return 20 * (logf(measured/reference)/logf(10));
}

static void one_point_dft(const short *inbuf, int len, int frequency, float *real, float *imaginary)
{
    float myreal = 0, myimag = 0;
    int i;

    for ( i = 0; i < len; i++ )
    {
        if ( use_table )
        {
            myreal += (float) inbuf[i] * cos_tbl(i*frequency, 8000);
            myimag += (float) inbuf[i] * sin_tbl(i*frequency, 8000);
        }
        else
        {
            myreal += (float) inbuf[i] * cos((i * 2.0 * M_PI * frequency)/8000);
            myimag += (float) inbuf[i] * sin((i * 2.0 * M_PI * frequency)/8000);
        }
    }

    myimag *= -1;

    *real = myreal / (float) len;
    *imaginary = myimag / (float) len;
}


static float calc_magnitude(short *inbuf, int insamps)
{
    float real, imaginary, magnitude;
    float totalmagnitude = 0;
    int i;

    for ( i = 0; i < freqcount; i++ )
    {
        one_point_dft(inbuf, insamps, freqs[i], &real, &imaginary);
        magnitude = sqrtf((real * real) + (imaginary * imaginary));
        totalmagnitude += magnitude;
    }

    return totalmagnitude;
}


/**
 *  Initialize the data store for storing off best calculated results
 */
static void init_topresults(int usrcfg_index)
{
    adect_acim_usrcfg[usrcfg_index].topresults.numactive = 0;
}


/**
 *  If this is a best result candidate, store in the top results data store
 * 		This is dependent on being the lowest echo value
 *
 *  @param tbleoffset - The offset into the echo_trys table used
 *  @param setting - Pointer to the settings used to achieve the fgiven value
 *  @param echo - The calculated echo return value (in dB)
 *  @param echo - The calculated magnitude of the response
 */
static void set_topresults(int usrcfg_index, int tbloffset, struct wctdm_echo_coefs *setting, float echo, float freqres)
{
    int place;
    int idx;
    topresults_t *ptopresults;

    ptopresults = &adect_acim_usrcfg[usrcfg_index].topresults;

    for ( place = 0; place < MAX_RESULTS && place < ptopresults->numactive; place++ )
    {
        if ( echo < ptopresults->results[place].echo )
        {
            break;
        }
    }

    if ( place < MAX_RESULTS )
    {
        /*  move results to the bottom */
        for ( idx = ptopresults->numactive-2; idx >= place; idx-- )
        {
            ptopresults->results[idx+1] = ptopresults->results[idx];
        }
        ptopresults->results[place].idx = tbloffset;
        ptopresults->results[place].settings = *setting;
        ptopresults->results[place].echo = echo;
        ptopresults->results[place].freqres = freqres;
        if ( MAX_RESULTS > ptopresults->numactive )
        {
            ptopresults->numactive++;
        }
    }
}

/**
 *  Prints the top results stored to stdout
 *
 *  @param header - Text that goes in the header of the response
 */
static void print_topresults(int usrcfg_index, char * header)
{
    int item;
    topresults_t *ptopresults;

    ptopresults = &adect_acim_usrcfg[usrcfg_index].topresults;

    ast_log(LOG_VERBOSE, "Top %d results for %s\n", ptopresults->numactive, header);
    for ( item = 0; item < ptopresults->numactive; item++ )
    {
        ast_log(LOG_VERBOSE, "Res #%d: index=%d, %3d,%3d,%3d,%3d,%3d,%3d,%3d,%3d,%3d: magnitude = %0.0f, echo = %0.4f dB\n",
                item+1, ptopresults->results[item].idx, ptopresults->results[item].settings.acim,
                ptopresults->results[item].settings.coef1, ptopresults->results[item].settings.coef2,
                ptopresults->results[item].settings.coef3, ptopresults->results[item].settings.coef4,
                ptopresults->results[item].settings.coef5, ptopresults->results[item].settings.coef6,
                ptopresults->results[item].settings.coef7, ptopresults->results[item].settings.coef8,
                ptopresults->results[item].freqres, ptopresults->results[item].echo);   
    }
}

/**
 * Perform calibration type 2 on the specified device
 * 
 * Determine optimum echo coefficients for the specified device
 * 
 * New tuning strategy.  If we have a number that we can dial that will result in silence from the
 * switch, the tune will be *much* faster (we don't have to keep hanging up and dialing a digit, etc...)
 * The downside is that the user needs to actually find a 'no tone' phone number at their CO's switch - but for
 * really fixing echo problems, this is what it takes.
 *
 * Also, for the purposes of optimizing settings, if we pick a single frequency and test with that,
 * we can try a whole bunch of impedence/echo coefficients.  This should give better results than trying
 * a bunch of frequencies, and we can always do a a frequency sweep to pick between the best 3 or 4
 * impedence/coefficients configurations.
 *   
 * Note:  It may be possible to take this even further and do some pertubation analysis on the echo coefficients
 * 		 themselves (maybe use the 72 entry sweep to find some settings that are close to working well, then
 * 		 deviate the coefficients a bit to see if we can improve things).  A better way to do this would be to
 * 		 use the optimization strategy from silabs.  For reference, here is an application note that describes
 * 		 the echo coefficients (and acim values):
 * 		 
 * 		 http://www.silabs.com/Support%20Documents/TechnicalDocs/an84.pdf
 *
 * 		 See Table 13 in this document for a breakdown of acim values by region.
 *
 * 		 http://www.silabs.com/Support%20Documents/TechnicalDocs/si3050-18-19.pdf
 * 		 
 */
static int acim_tune2(int whichdahdi, int usrcfg_index, int freq, char *dialstr, int delayuntilsilence,
                      int silencegoodfor, struct wctdm_echo_coefs *coefs_out,float *erl)
{
    int i = 0;
    int res = 0, x = 0;
    int lowesttry = -1;
    float lowesttryresult = 999999999999.0;
    float lowestecho = 999999999999.0;
    struct dahdi_bufferinfo bi;
    short inbuf[TEST_DURATION * 2];
    short outbuf[TEST_DURATION];

    struct silence_info sinfo;
#ifdef ACCURATE_CALIBRATE
    int echo_trys_size = 72;
#else
    int echo_trys_size = 16;
#endif
    int trys = 0;
    float waveform_power;
    float freq_result;
    float echo;
    struct wctdm_stats fxo_stat;
    int channel = adect_acim_usrcfg[usrcfg_index].channum;

    init_topresults(usrcfg_index);

    /* Set echo settings */
#ifdef ACCURATE_CALIBRATE //pkasbekar - Use the complete set of 72 ac impedances and echo co-efficients
    if ( ioctl(whichdahdi, WCTDM_SET_ECHOTUNE, &echo_trys[0]) )
    {
        adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_FXO_UNABLE_SET_ECHO_COEFF;
        ast_log(LOG_ERROR, "Unable to set impedance on fd %d, port %d\n", whichdahdi, channel);
        return -1;
    }
#else   //pkasbekar - use the reduced local set in this file.
    if ( ioctl(whichdahdi, WCTDM_SET_ECHOTUNE, &local_echo_trys[0]) )
    {
        adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_FXO_UNABLE_SET_ECHO_COEFF;
        ast_log(LOG_ERROR, "Unable to set impedance on fd %d port %d\n", whichdahdi, channel);
        return -1;
    }
#endif
    // pkasbekar - Check line voltage status to determine if a particular FXO port is connected
    if ( ioctl(whichdahdi, WCTDM_GET_STATS, &fxo_stat) )
    {
        ast_log(LOG_ERROR, "Unable to read status of fxo port %d\n", channel);
        adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_READ_FXO_STATS_FAILED;
        return -1;
    }
    ast_log(LOG_DEBUG, "FXO port %d FXO tip voltage = %d, FXO ring voltage = %d FXO battery voltage = %d\n",
            channel, fxo_stat.tipvolt,fxo_stat.ringvolt,fxo_stat.batvolt);

    if ( fxo_stat.tipvolt >= (-LV_MIN_ONHOOK_VOLTAGE*1000) && fxo_stat.tipvolt <= (LV_MIN_ONHOOK_VOLTAGE*1000) )
    {
        ast_log(LOG_ERROR, "FXO port %d not connected\n", channel);
        adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_FXO_NOT_CONNECT;
        return -1;
    }

    x = 1;
    if ( ioctl(whichdahdi, DAHDI_SETLINEAR, &x) )
    {
        ast_log(LOG_ERROR, "Unable to set channel %d to signed linear mode.\n", channel);
        adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_SET_LINEAR_MODE_FAILED;
        return -1;
    }

    memset(&bi, 0, sizeof(bi));
    if ( ioctl(whichdahdi, DAHDI_GET_BUFINFO, &bi) )
    {
        ast_log(LOG_ERROR, "Unable to get buffer information of port %d! \n", channel);
        adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_GET_BUFF_INFO_FAILED;
        return -1;
    }

    bi.numbufs = 2;
    bi.bufsize = TEST_BUFFER_LENGTH;
    bi.txbufpolicy = DAHDI_POLICY_IMMEDIATE;
    bi.rxbufpolicy = DAHDI_POLICY_IMMEDIATE;
    if ( ioctl(whichdahdi, DAHDI_SET_BUFINFO, &bi) )
    {
        ast_log(LOG_ERROR, "Unable to set buffer information of port %d!\n", channel);
        adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_SET_BUFF_INFO_FAILED;
        return -1;
    }
    x = DAHDI_OFFHOOK;
    if ( ioctl(whichdahdi, DAHDI_HOOK, &x) )
    {
        ast_log(LOG_ERROR, "Cannot bring fd %d off hook, port %d\n", whichdahdi, channel);
        adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_BRING_OFFHOOK_FAILED;
        return -1;
    }

    /* BEGIN: jizhang 20140418 [Bug 24082] add */
    ast_log(LOG_ERROR, "bring fd %d off hook, port %d\n", whichdahdi, channel);
    if(intervaltime > 0)
    {
        ast_log(LOG_ERROR, "sleep %d ms\n",intervaltime);
        usleep(intervaltime*1000);

        if ( ioctl(whichdahdi, WCTDM_GET_STATS, &fxo_stat) )
        {
            ast_log(LOG_ERROR, "Unable to read status of fxo port %d\n", channel);
            adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_READ_FXO_STATS_FAILED;
            return -1;
        }
    }

    ast_log(LOG_ERROR, "trys %d FXO port %d FXO tip voltage = %d, FXO ring voltage = %d FXO battery voltage = %d\n",
        trys, channel, fxo_stat.tipvolt,fxo_stat.ringvolt,fxo_stat.batvolt);
    /* END: jizhang 20140418 [Bug 24082] add */

    /* Set up silence settings */
    memset(&sinfo, 0, sizeof(sinfo));
    sinfo.device = whichdahdi;
    sinfo.dialstr = dialstr;
    sinfo.initial_delay = delayuntilsilence;
    sinfo.reset_after = silencegoodfor;

    /* Fill the output buffers */
    for ( i = 0; i < TEST_DURATION; i++ )
        outbuf[i] = freq > 0 ? gentone(freq, i) : genwaveform(i); /* if freq is negative, use a multi-frequency waveform */

    /* compute power of input (so we can later compute echo levels relative to input) */
    waveform_power = calc_magnitude(outbuf, TEST_DURATION);

    /* sweep through the various coefficient settings and see how our responses look */
    for ( trys = 0; trys < echo_trys_size; trys++ )
    {

        //pkasbekar - Check each time if the FXO port is connected or not.
        if ( ioctl(whichdahdi, WCTDM_GET_STATS, &fxo_stat) )
        {
            ast_log(LOG_ERROR, "Unable to read status of fxo port %d\n", channel);
            adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_READ_FXO_STATS_FAILED;
            return -1;
        }

        if ( fxo_stat.tipvolt >= (-LV_MIN_OFFHOOK_VOLTAGE*1000) && fxo_stat.tipvolt <= (LV_MIN_OFFHOOK_VOLTAGE*1000) )
        {
            ast_log(LOG_ERROR, "FXO port %d not connected \n", channel);
            adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_FXO_NOT_CONNECT;
            return -1;
        }

        /* ensure silence on the line */
        if ( ensure_silence(&sinfo) )
        {
            ast_log(LOG_ERROR, "Unable to get a clear outside line on port %d\n", channel);
            adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_FXO_CANNOT_CLEAR;
            return -1;
        }

#ifdef ACCURATE_CALIBRATE	
        if ( ioctl(whichdahdi, WCTDM_SET_ECHOTUNE, &echo_trys[trys]) )
        {
            ast_log(LOG_ERROR, "Unable to set echo coefficients on fd %d, port %d\n", whichdahdi, channel);
            adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_FXO_UNABLE_SET_ECHO_COEFF;
            return -1;
        }
#else
        if ( ioctl(whichdahdi, WCTDM_SET_ECHOTUNE, &local_echo_trys[trys]) )
        {
            ast_log(LOG_ERROR, "Unable to set echo coefficients on fd %d,port %d\n", whichdahdi, channel);
            adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_FXO_UNABLE_SET_ECHO_COEFF;
            return -1;
        }
#endif
        /* Flush buffers */
        x = DAHDI_FLUSH_READ | DAHDI_FLUSH_WRITE | DAHDI_FLUSH_EVENT;
        if ( ioctl(whichdahdi, DAHDI_FLUSH, &x) )
        {
            ast_log(LOG_ERROR, "Unable to flush port %d I/O: %s\n", channel, strerror(errno));
            adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_FXO_UNABLE_FLUSH_BUFF;
            return -1;
        }

        /* send data out on line */
        res = write(whichdahdi, outbuf, TEST_BUFFER_LENGTH);
        if ( res != TEST_BUFFER_LENGTH )
        {
            ast_log(LOG_ERROR, "Could not write all data to line %d\n", channel);
            adect_acim_usrcfg[usrcfg_index].acim_errcode = ADT_ACIM_ERR_FXO_UNABLE_WRITE_LINE;
            return -1;
        }
        retry:
        /* read return response */
        res = fxotune_read(whichdahdi, inbuf, TEST_BUFFER_LENGTH * 2);
        if ( res != TEST_BUFFER_LENGTH * 2 )
        {
            int dummy;
            ioctl(whichdahdi, DAHDI_GETEVENT, &dummy);
            goto retry;
        }

        freq_result = calc_magnitude(inbuf, TEST_DURATION * 2);
        echo = db_loss(freq_result, waveform_power);

#if 0
        ast_log(LOG_DEBUG, "%3d,%d,%d,%d,%d,%d,%d,%d,%d: magnitude = %0.0f, echo = %0.4f dB\n", 
                echo_trys[trys].acim, echo_trys[trys].coef1, echo_trys[trys].coef2,
                echo_trys[trys].coef3, echo_trys[trys].coef4, echo_trys[trys].coef5,
                echo_trys[trys].coef6, echo_trys[trys].coef7, echo_trys[trys].coef8,
                freq_result, echo);
#endif

        if ( freq_result < lowesttryresult )
        {
            lowesttry = trys;
            lowesttryresult = freq_result;
            lowestecho = echo;
        }

        char result[256];
        snprintf(result, sizeof(result), "%3d,%3d,%3d,%3d,%3d,%3d,%3d,%3d,%3d,%f,%f", 
                 echo_trys[trys].acim, echo_trys[trys].coef1, echo_trys[trys].coef2, 
                 echo_trys[trys].coef3, echo_trys[trys].coef4, echo_trys[trys].coef5, 
                 echo_trys[trys].coef6, echo_trys[trys].coef7, echo_trys[trys].coef8, 
                 freq_result, echo);

        ast_log(LOG_DEBUG, "%3d,%3d,%3d,%3d,%3d,%3d,%3d,%3d,%3d: magnitude = %0.0f, echo = %0.4f dB\n",
                echo_trys[trys].acim, echo_trys[trys].coef1, echo_trys[trys].coef2,
                echo_trys[trys].coef3, echo_trys[trys].coef4, echo_trys[trys].coef5,
                echo_trys[trys].coef6, echo_trys[trys].coef7, echo_trys[trys].coef8,
                freq_result, echo);

#ifdef ACCURATE_CALIBRATE
        if ( printbest )
        {
            set_topresults(usrcfg_index, trys, &echo_trys[trys], echo, freq_result);
        }
#else
        if ( printbest )
        {
            set_topresults(usrcfg_index, trys, &local_echo_trys[trys], echo, freq_result);
        }
#endif                
    }

    ast_log(LOG_DEBUG, "Config with lowest response = %d, magnitude = %0.0f, echo = %0.4f dB\n", lowesttry, lowesttryresult, lowestecho);
#ifdef ACCURATE_CALIBRATE
    memcpy(coefs_out, &echo_trys[lowesttry], sizeof(struct wctdm_echo_coefs));
#else
    memcpy(coefs_out, &local_echo_trys[lowesttry], sizeof(struct wctdm_echo_coefs));
#endif
    // pkasbekar - Send the ERL to do_calibration
    *erl = lowestecho;
    if ( printbest )
    {
        print_topresults(usrcfg_index, "Acim2_tune Test");
    }

    return 0;
}

/**
 *  Perform calibration type 1 on the specified device.  Only tunes the line impedance.  Look for best response range 
 */
static int acim_tune(int whichdahdi, int usrcfg_index, char *dialstr, int delayuntilsilence, 
                     int silencegoodfor, struct wctdm_echo_coefs *coefs_out)
{
    int i = 0, freq = 0, acim = 0;
    int res = 0, x = 0;
    struct dahdi_bufferinfo bi;
    struct wctdm_echo_coefs coefs;
    short inbuf[TEST_DURATION]; /* changed from TEST_BUFFER_LENGTH - this buffer is for short values, so it should be allocated using the length of the test */
    short outbuf[TEST_DURATION];
    int lowest = 0;
    float acim_results[16];
    struct silence_info sinfo;
    struct wctdm_stats fxo_stat;

    /* Set up silence settings */
    memset(&sinfo, 0, sizeof(sinfo));
    sinfo.device = whichdahdi;
    sinfo.dialstr = dialstr;
    sinfo.initial_delay = delayuntilsilence;
    sinfo.reset_after = silencegoodfor;

    /* Set echo settings */
    memset(&coefs, 0, sizeof(coefs));
    if ( ioctl(whichdahdi, WCTDM_SET_ECHOTUNE, &coefs) )
    {
        ast_log(LOG_ERROR, "Skipping non-TDM / non-FXO\n");
        return -1;
    }
    //pkasbekar - check if FXO port is connected.
    if ( ioctl(whichdahdi, WCTDM_GET_STATS, &fxo_stat) )
    {
        ast_log(LOG_ERROR, "Unable to read status of fxo port\n");
        return -1;
    }

    ast_log(LOG_DEBUG,"FXO tip voltage = %d, FXO ring voltage = %d FXO battery voltage = %d\n",fxo_stat.tipvolt,fxo_stat.ringvolt,fxo_stat.batvolt);

    if ( fxo_stat.tipvolt >= (-LV_MIN_ONHOOK_VOLTAGE*1000) && fxo_stat.tipvolt <= (LV_MIN_ONHOOK_VOLTAGE*1000) )
    {
        ast_log(LOG_ERROR, "FXO port not connected \n");
        return -1;
    }

    x = 1;
    if ( ioctl(whichdahdi, DAHDI_SETLINEAR, &x) )
    {
        ast_log(LOG_ERROR, "Unable to set channel to signed linear mode.\n");
        return -1;
    }

    memset(&bi, 0, sizeof(bi));
    if ( ioctl(whichdahdi, DAHDI_GET_BUFINFO, &bi) )
    {
        ast_log(LOG_ERROR, "Unable to get buffer information!\n");
        return -1;
    }
    bi.numbufs = 2;
    bi.bufsize = TEST_BUFFER_LENGTH;
    bi.txbufpolicy = DAHDI_POLICY_IMMEDIATE;
    bi.rxbufpolicy = DAHDI_POLICY_IMMEDIATE;
    if ( ioctl(whichdahdi, DAHDI_SET_BUFINFO, &bi) )
    {
        ast_log(LOG_ERROR, "Unable to set buffer information!\n");
        return -1;
    }

    for ( acim = 0; acim < 16; acim++ )
    {
        float freq_results[15];

        coefs.acim = acim;
        if ( ioctl(whichdahdi, WCTDM_SET_ECHOTUNE, &coefs) )
        {
            ast_log(LOG_ERROR, "Unable to set impedance on fd %d\n", whichdahdi);
            return -1;
        }

        for ( freq = 200; freq <=3000; freq+=200 )
        {
            /* Fill the output buffers */
            for ( i = 0; i < TEST_DURATION; i++ )
                outbuf[i] = gentone(freq, i);

            /* Make sure line is ready for next test iteration */
            if ( ensure_silence(&sinfo) )
            {
                ast_log(LOG_ERROR,"Unable to get a clear line\n");
                return -1;
            }

            /* Flush buffers */
            x = DAHDI_FLUSH_READ | DAHDI_FLUSH_WRITE | DAHDI_FLUSH_EVENT;
            if ( ioctl(whichdahdi, DAHDI_FLUSH, &x) )
            {
                ast_log(LOG_ERROR, "Unable to flush I/O: %s\n", strerror(errno));
                return -1;
            }

            /* send data out on line */
            res = write(whichdahdi, outbuf, TEST_BUFFER_LENGTH);
            if ( res != TEST_BUFFER_LENGTH )
            {
                ast_log(LOG_ERROR, "Could not write all data to line\n");
                return -1;
            }
            retry:
            /* read return response */
            res = fxotune_read(whichdahdi, inbuf, TEST_BUFFER_LENGTH);
            if ( res != TEST_BUFFER_LENGTH )
            {
                int dummy;
                ioctl(whichdahdi, DAHDI_GETEVENT, &dummy);
                goto retry;
            }

            /* calculate power of response */
            freq_results[(freq/200)-1] = power_of(inbuf+SKIP_SAMPLES, TEST_DURATION-SKIP_SAMPLES, 1); /* changed from inbuf+SKIP_BYTES, TEST_BUFFER_LENGTH-SKIP_BYTES, 1 */
            ast_log(LOG_DEBUG, "%d,%d,%f\n", acim, freq, freq_results[(freq/200)-1]);
        }

        acim_results[acim] = power_of(freq_results, 15, 0);
    }

    for ( i = 0; i < 16; i++ )
    {
        ast_log(LOG_DEBUG, "acim_results[%d] = %f\n", i, acim_results[i]);
    }

    /* Find out what the "best" impedance is for the line */
    lowest = 0;
    for ( i = 0; i < 16; i++ )
    {
        if ( acim_results[i] < acim_results[lowest] )
        {
            lowest = i;
        }
    }

    coefs_out->acim = lowest;
    coefs_out->coef1 = 0;
    coefs_out->coef2 = 0;
    coefs_out->coef3 = 0;
    coefs_out->coef4 = 0;
    coefs_out->coef5 = 0;
    coefs_out->coef6 = 0;
    coefs_out->coef7 = 0;
    coefs_out->coef8 = 0;

    return 0;
}

/**
 * Reads echo register settings and pushes them into
 * the appropriate devices
 * 
 * @return 0 if successful, !0 otherwise
 */ 
static int do_set(int dfd, int cfg_index)
{
    struct wctdm_echo_coefs mycoefs;

    int myacim,mycoef1,mycoef2,mycoef3,mycoef4,mycoef5,mycoef6,mycoef7,mycoef8;

    myacim = adect_acim_usrcfg[cfg_index].best_result.coefs.acim;
    mycoef1 = adect_acim_usrcfg[cfg_index].best_result.coefs.coef1;
    mycoef2 = adect_acim_usrcfg[cfg_index].best_result.coefs.coef2;
    mycoef3 = adect_acim_usrcfg[cfg_index].best_result.coefs.coef3;
    mycoef4 = adect_acim_usrcfg[cfg_index].best_result.coefs.coef4;
    mycoef5 = adect_acim_usrcfg[cfg_index].best_result.coefs.coef5;
    mycoef6 = adect_acim_usrcfg[cfg_index].best_result.coefs.coef6;
    mycoef7 = adect_acim_usrcfg[cfg_index].best_result.coefs.coef7;
    mycoef8 = adect_acim_usrcfg[cfg_index].best_result.coefs.coef8;


    /* Check to be sure conversion is done correctly */
    if ( OUT_OF_BOUNDS(myacim) || OUT_OF_BOUNDS(mycoef1)||
         OUT_OF_BOUNDS(mycoef2)|| OUT_OF_BOUNDS(mycoef3)||
         OUT_OF_BOUNDS(mycoef4)|| OUT_OF_BOUNDS(mycoef5)||
         OUT_OF_BOUNDS(mycoef6)|| OUT_OF_BOUNDS(mycoef7)|| OUT_OF_BOUNDS(mycoef8) )
    {

        ast_log(LOG_ERROR, "Acim Coefs Bounds check error \n");
        return -1;
    }

    mycoefs.acim = myacim;
    mycoefs.coef1 = mycoef1;
    mycoefs.coef2 = mycoef2;
    mycoefs.coef3 = mycoef3;
    mycoefs.coef4 = mycoef4;
    mycoefs.coef5 = mycoef5;
    mycoefs.coef6 = mycoef6;
    mycoefs.coef7 = mycoef7;
    mycoefs.coef8 = mycoef8;

    if ( ioctl(dfd, WCTDM_SET_ECHOTUNE, &mycoefs) )
    {
        ast_log(LOG_ERROR, "WCTDM_SET_ECHOTUNE %s\n", strerror(errno));
        return -1;
    }

    ast_log(LOG_DEBUG, "fxotune: successfully set echo coeffecients on FXO modules\n");
    return 0;   
}

/**
* Performs calibration on specified devices
* 
* @param devfd the file descriptor of the device to check
* @param calibtype the type of calibration to perform.  1=old style (loops through individual frequencies
* 			doesn't optimize echo coefficients.  2=new style (uses multi-tone and optimizes echo coefficients
* 			and acim setting)
* @param usrcfg_index the index of configbuffer that the calibration results should be written to
* @param dialstr the string that should be dialed to clear the dialtone from the line
* @param delayuntilsilence the number of seconds to wait after dialing dialstr before starting the test
* @param silencegoodfor the number of seconds that the test can run before having to reset the line again
* 			(this is basically the amount of time it takes before the 'if you'd like to make a call...' message
* 			kicks in after you dial dialstr
* 
* @return 0 if successful, -1 for serious error such as device not available , > 0 indicates the number of channels
*/  
static int do_calibrate(int dev_fd, int usrcfg_index, int calibtype,
                        char* dialstr, int delayuntilsilence, int silencegoodfor)
{
    int problems = 0;
    int res = 0;
    int devno = 0;

    struct wctdm_echo_coefs *coefs;

    float *erl; //pkasbekar - ERL

    coefs = &adect_acim_usrcfg[usrcfg_index].best_result.coefs;
    erl = &adect_acim_usrcfg[usrcfg_index].best_result.erl; 

    if ( 1 == calibtype )//not used!!!
        res = acim_tune(dev_fd, usrcfg_index, dialstr, delayuntilsilence, silencegoodfor, coefs);
    else
        //pkasbekar - additional parameter ERL
        res = acim_tune2(dev_fd, usrcfg_index, -1, dialstr, delayuntilsilence, silencegoodfor, coefs, erl);

    if ( res )
    {
        ast_log(LOG_ERROR, "Failure!\n");
        problems++;
    }
    else
    {
        ast_log(LOG_DEBUG, "Done!\n");
    }

    if ( res == 0 )
    {

        static char output[255] = "";

        snprintf(output, sizeof(output), "%d=%d,%d,%d,%d,%d,%d,%d,%d,%d Echo Return Loss %.4f dB\n", 
                 devno, coefs->acim, coefs->coef1, coefs->coef2, coefs->coef3, coefs->coef4, 
                 coefs->coef5, coefs->coef6, coefs->coef7, coefs->coef8, (-1*(*erl)));//pkasbekar - print ERL in the file

        adect_acim_usrcfg[usrcfg_index].best_result.get_result = TRUE; 
        ast_log(LOG_DEBUG, "Found best echo coefficients: %s\n", output);

    }

    if ( problems )
        ast_log(LOG_ERROR, "Unable to tune %d devices, even though those devices are present\n", problems);

    return problems;
}   

static void acim_autodetect_cleanup(void *data)
{
    acim_adt_thread_t *acim_adt_thread = (acim_adt_thread_t *)data;

    if ( acim_adt_thread->dfd > 0 )
        close(acim_adt_thread->dfd);

    ast_atomic_fetchadd_int(&acim_adt_counter, -1);
    if ( acim_adt_counter<=0 )
    {
        acim_adt_counter = 0;
        load_chan_dahdi();
        svip_lec_control(1);
        g_adt_item = ADT_NON;
    }

    adect_acim_usrcfg[acim_adt_thread->cfg_index].acim_state = ADT_ACIM_CANCEL;
}
static void* dahdi_pstn_acim_autodetect_thread( void *data )
{
    int cfg_index = *(int *)data;
    int calibtype = 2; 
    int delaytosilence = 0; 
    int silencegoodfor = 1; 
    char* dialstr = "5";
    int res = 0;

    char dahdidev[80]="";
    int dfd = -1;
    int chan;
    acim_adt_thread_t acim_adt_thread;

    acim_adt_thread.dfd = -1;
    acim_adt_thread.cfg_index = cfg_index;

    adect_acim_usrcfg[cfg_index].acim_state = ADT_ACIM_START;
    adect_acim_usrcfg[cfg_index].acim_errcode = ADT_ACIM_ERR_NONE;

    pthread_cleanup_push(acim_autodetect_cleanup, &acim_adt_thread);

    chan = adect_acim_usrcfg[cfg_index].channum;

    snprintf(dahdidev,sizeof(dahdidev),"%s/%d",dahdipath, chan);

    dfd = open(dahdidev,O_RDWR);
    if ( dfd < 0 )
    {
        adect_acim_usrcfg[cfg_index].acim_errcode = ADT_ACIM_FXO_ABSENT;
        ast_log(LOG_ERROR,"%s absent: %s\n",dahdidev,strerror(errno));
        goto safe_exit;
    }

    acim_adt_thread.dfd = dfd;

    adect_acim_usrcfg[cfg_index].best_result.get_result = FALSE;

    res = do_calibrate(dfd, cfg_index, calibtype, dialstr, delaytosilence, silencegoodfor);
    if ( (!res)&&(adect_acim_usrcfg[cfg_index].acim_doset_mode == ACIM_DOSET_AUTO) )
        do_set(dfd, cfg_index);

    close(dfd);
    acim_adt_thread.dfd = -1;

    safe_exit:   
    ast_atomic_fetchadd_int(&acim_adt_counter, -1);
    if ( acim_adt_counter<=0 )
    {
        acim_adt_counter = 0;
        load_chan_dahdi();
        svip_lec_control(1);
        g_adt_item = ADT_NON;
    }

    adect_acim_usrcfg[cfg_index].acim_state = ADT_ACIM_EXIT;
	/* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    manager_event(EVENT_FLAG_SYSTEM, "acim_autodetect_state", "FXO: %d\r\nstate: %s\r\nErrCod: %d\r\nACIM: %d\r\ncoefs: %d,%d,%d,%d,%d,%d,%d,%d\r\n", 
                  adect_acim_usrcfg[cfg_index].channum, 
                  adt_acim_state_str[adect_acim_usrcfg[cfg_index].acim_state], 
                  adect_acim_usrcfg[cfg_index].acim_errcode, 
                  adect_acim_usrcfg[cfg_index].best_result.coefs.acim,
                  adect_acim_usrcfg[cfg_index].best_result.coefs.coef1,
                  adect_acim_usrcfg[cfg_index].best_result.coefs.coef2,
                  adect_acim_usrcfg[cfg_index].best_result.coefs.coef3,
                  adect_acim_usrcfg[cfg_index].best_result.coefs.coef4,
                  adect_acim_usrcfg[cfg_index].best_result.coefs.coef5,
                  adect_acim_usrcfg[cfg_index].best_result.coefs.coef6,
                  adect_acim_usrcfg[cfg_index].best_result.coefs.coef7,
                  adect_acim_usrcfg[cfg_index].best_result.coefs.coef8
                  ); 
	/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    pthread_cleanup_pop(0);

    return NULL;

}

static int acim_autodetect_is_finished(void)
{
    int i;
    int finished = TRUE;

    for ( i=0;i<g_fxo_amount;i++ )
    {
        if ( adect_acim_usrcfg[i].acim_thread == AST_PTHREADT_NULL )
            continue;
        if ( (adect_acim_usrcfg[i].acim_state == ADT_ACIM_NON)||
             (adect_acim_usrcfg[i].acim_state == ADT_ACIM_EXIT)||
             (adect_acim_usrcfg[i].acim_state == ADT_ACIM_CANCEL) )
        {
            continue;
        }
        finished = FALSE;
    }

    return finished;
}

static void dahdi_pstn_acim_autodetect_thread_release( int cli_fd, int concise )
{
    int i;
    int cancel_timeout = 0;

    for ( i=0; i<g_fxo_amount; i++ )
    {

        if ( !adect_acim_usrcfg[i].release )
            continue;

        adect_acim_usrcfg[i].release = 0;

        if ( adect_acim_usrcfg[i].acim_thread != AST_PTHREADT_NULL )
        {
            if ( (adect_acim_usrcfg[i].acim_state == ADT_ACIM_NON) &&
                 (adect_acim_usrcfg[i].acim_state == ADT_ACIM_EXIT) )
            {
                if ( concise == 1 )
                {
                    ast_cli(cli_fd, "FXO=%d!state=exit\n", adect_acim_usrcfg[i].channum);
                }
                else
                {
                    ast_cli(cli_fd, "port %d dahdi pstn acim auto detect thread has been exit!!\n", 
                            adect_acim_usrcfg[i].channum);   
                }
            }
            else
            {
                if ( concise != 1 )
                {
                    ast_cli(cli_fd, "port %d dahdi pstn acim auto detect thread cancel index %d !!\n",
                            adect_acim_usrcfg[i].channum, i);
                }
                pthread_cancel(adect_acim_usrcfg[i].acim_thread);
                while ( 1 )
                {
                    usleep(10000);
                    cancel_timeout++;
                    if ( (adect_acim_usrcfg[i].acim_state == ADT_ACIM_NON) || 
                         (adect_acim_usrcfg[i].acim_state == ADT_ACIM_EXIT) ||
                         (adect_acim_usrcfg[i].acim_state == ADT_ACIM_CANCEL) )
                    {
                        if ( concise == 1 )
                        {
                            ast_cli(cli_fd, "FXO=%d!state=canceled\n", adect_acim_usrcfg[i].channum);
                        }
                        else
                        {
                            ast_cli(cli_fd, "port %d index %d dahdi pstn acim auto detect thread be canceled!!\n",
                                    adect_acim_usrcfg[i].channum, i);
                        }
                        break;
                    }
                    if ( cancel_timeout>1000 )
                    {
                        if ( concise == 1 )
                        {
                            ast_cli(cli_fd, "FXO=%d!state=cancel time out\n", adect_acim_usrcfg[i].channum);
                        }
                        else
                        {
                            ast_cli(cli_fd, "ERR: port %d dahdi pstn acim auto detect thread cancel time out!! adect_acim_usrcfg[%d].acim_state %d \n",
                                    adect_acim_usrcfg[i].channum, i, adect_acim_usrcfg[i].acim_state);
                        }
                        break;
                    }
                }
            }

            adect_acim_usrcfg[i].acim_thread = AST_PTHREADT_NULL;
            if ( concise == 1 )
            {
                //ast_cli(cli_fd, "FXO=all!state=exited\n");
            }
            else
            {
                ast_cli(cli_fd, "dahdi pstn acim auto detect thread exit done!!\n");
            }
        }
    }

    if ( acim_adt_counter>0 )
        return;

    if ( g_adt_item == ADT_ACIM )
    {
        g_adt_item = ADT_NON;
    }
}

static int dahdi_pstn_acim_autodetect_start( int cli_fd, int concise)
{
    pthread_attr_t attr;
    struct sched_param param;
    int policy = SCHED_OTHER;
    int priority_max;
    int priority_min;
    int res = -1;
    int i;
    int nstarted = 0;

    if ( ADT_CPT == g_adt_item )
    {
        if ( !cpt_autodetect_is_finished() )
        {
            ast_log(LOG_ERROR, "PSTN Auto detect CPT is running!\n");
            /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
            if(concise == 1 )
            {
                ast_cli(cli_fd, "FXO=all!state=failed!ErrCod=%d\n", ADT_ACIM_ERR_CPT_DETECT_BUSY);
            }
            /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
            return -1;
        }
    }
    if ( !acim_autodetect_is_finished() )
    {
        ast_log(LOG_ERROR, "PSTN Auto detect ACIM is running!\n");
        /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        if(concise == 1 )
        {
            ast_cli(cli_fd, "FXO=all!state=failed!ErrCod=%d\n", ADT_ACIM_ERR_DETECT_BUSY);
        }
        /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        return -2;
    }

    g_adt_item = ADT_ACIM;

    if ( acim_adt_counter == 0 )
    {
        res = unload_chan_dahdi();
        if ( res<0 )
        {
            g_adt_item = ADT_NON;
            /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
            if(concise == 1 )
            {
                ast_cli(cli_fd, "FXO=all!state=failed!ErrCod=%d\n", ADT_ACIM_ERR_UNLOAD_DAHDI_FAILED);
            }
            /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
            return -3;
        }
        svip_lec_control(0);
    }

    for ( i=0; i<g_fxo_amount; i++ )
    {
        if ( !adect_acim_usrcfg[i].tostart )
            continue;
        ast_atomic_fetchadd_int(&acim_adt_counter, +1);
    }

    for ( i=0; i<g_fxo_amount; i++ )
    {

        if ( !adect_acim_usrcfg[i].tostart )
            continue;

        adect_acim_usrcfg[i].tostart = 0;

        if ( concise != 1 )
        {
            ast_cli(cli_fd, "[NOTICE] pstn acim autodetect thread start fxo %d index %d.\n",
                    adect_acim_usrcfg[i].channum, i);
        }

        if ( adect_acim_usrcfg[i].acim_thread != AST_PTHREADT_NULL )
        {
            if ( (adect_acim_usrcfg[i].acim_state==ADT_ACIM_EXIT)||
                 (adect_acim_usrcfg[i].acim_state==ADT_ACIM_NON) )
            {

                adect_acim_usrcfg[i].acim_thread = AST_PTHREADT_NULL;

            }
            else
            {
                adect_acim_usrcfg[i].acim_errcode = ADT_ACIM_ERR_DETECT_BUSY;
                ast_log(AST_LOG_ERROR,"Port %d PSTN ACIM Auto Detect thread has been started!!\n", i);
                /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
                if ( concise == 1 )
                {
                    ast_cli(cli_fd, "FXO=%d!state=failed!ErrCod=%d\n", i+1, ADT_ACIM_ERR_DETECT_BUSY);
                }
                /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
                continue;
            }
        }
        adect_acim_usrcfg[i].cfg_index = i;

        pthread_attr_init(&attr);
        pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
        pthread_attr_setstacksize(&attr, 1024 * 1024);
        param.sched_priority = adect_acim_usrcfg[i].acim_thread_pri;
        res = pthread_attr_setschedparam(&attr, &param);

        if ( res )
        {
            if ( concise != 1 )
            {
                ast_cli(cli_fd, "[WARNING] Pthread set sched_priority=%d failed, res:%d, errno:%d(%s)!\n",
                        param.sched_priority, res, errno, strerror(errno));
            }
        }

        res = pthread_attr_getschedparam(&attr, &param);
        if ( res )
        {
            if ( concise != 1 )
            {
                ast_cli(cli_fd, "[WARNING] Pthread get schedparam failed, res:%d, errno:%d(%s)!\n",
                        res, errno, strerror(errno));
            }
        }
        else
        {
            // ast_cli(cli_fd, "[NOTICE] Dahdi pstn acim autodetect thread get priority:%d.\n", param.sched_priority);
        }

        res = pthread_attr_getschedpolicy(&attr, &policy);
        if ( res )
        {
            if ( concise != 1 )
            {
                ast_cli(cli_fd, "[WARNING] Pthread get schedpolicy failed, res:%d, errno:%d(%s)!\n",
                        res, errno, strerror(errno));
            }
        }
        else
        {
            // ast_cli(cli_fd, "[NOTICE] pstn acim autodetect thread get policy:%d.\n", policy);
        }

        priority_max = sched_get_priority_max(policy);
        priority_min = sched_get_priority_min(policy);

        ast_cli(cli_fd, "[NOTICE] priority max:%d, min:%d.\n", priority_max, priority_min);

        res = ast_pthread_create_background(&adect_acim_usrcfg[i].acim_thread, &attr,
                                            dahdi_pstn_acim_autodetect_thread, &adect_acim_usrcfg[i].cfg_index);

        pthread_attr_destroy(&attr);

        if ( res != 0 )
        {
            if ( concise != 1 )
            {
                ast_cli(cli_fd, "[ERROR] Create pstn acim auto detect thread failed! res:%d, errno:%d(%s) port %d!!\n",
                        res, errno, strerror(errno), i);
            }
		    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
            else
            {
                ast_cli(cli_fd, "FXO=%d!state=failed!ErrCod=%d\n", i, ADT_ACIM_ERR_THREAD_CREAT_FAILED);
            }
            /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
            adect_acim_usrcfg[i].acim_thread = AST_PTHREADT_NULL;
            adect_acim_usrcfg[i].acim_errcode = ADT_ACIM_ERR_THREAD_CREAT_FAILED;
            ast_atomic_fetchadd_int(&acim_adt_counter, -1);
        }
        else
        {
            nstarted ++;
            /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
            if(concise == 1 )
            {
                ast_cli(cli_fd, "FXO=%d!state=started\n", i+1);
            }
			/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        }
    }

    if ( concise != 1 )
    {
        ast_cli(cli_fd, "%d ports of ACIM-detection are started, %d ports are detecting\n", nstarted, acim_adt_counter);
    }

    if ( 0 == nstarted )
    {
        load_chan_dahdi();
        svip_lec_control(1);
        g_adt_item = ADT_NON;
    }
    return 0;
}

static char *dahdi_pstn_acim_autodetect_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int ret;
    int concise = 0;
    int port;
    char *portlist;
    char *portstr;

    switch ( cmd )
    {
        case CLI_INIT:
            e->command = "pstn autodetect acim_start";
            e->usage = "Usage: pstn autodetect acim_start [port list] [concise]\n"
                       "   [port list]: separated by ',' eg: 1,2,3 \n";
            return NULL;
        case CLI_GENERATE:
            return NULL;
    }

    if ( a->argc < 4 )
    {
        return CLI_SHOWUSAGE;
    }

    if ( ast_strlen_zero(a->argv[3]) )
    {
        return CLI_SHOWUSAGE;
    }
    
    if ( a->argv[4] != NULL && strcmp(a->argv[4], "concise") == 0 )
    {
        concise = 1;
    }
    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    if(concise == 1)
    {
        ast_cli(a->fd, "PSTN AutoDetect Start\n"); 
    }
    /* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    portlist = ast_strdupa(a->argv[3]);
    while ( (portstr = strsep(&portlist, ",")) )
    {
        port = atoi(portstr);
        if ( port > g_fxo_amount || port < 1 )
        {
            if ( concise == 1 )
            {
			    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                ast_cli(a->fd, "FXO=%d!state=failed!ErrCod=%d\n", port, ADT_CPT_ERR_INVALID_CHAN);
				/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
            }
            else
            {
                ast_cli(a->fd, "PSTN ACIM Auto Detect invalid port %d\n", port); 
            }
            continue;
        }
/* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
#if 0
/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        if ( concise == 1 )
        {
            ast_cli(a->fd, "FXO=%d!state=start\n", port); 
        }
        else
        {
            ast_cli(a->fd, "PSTN ACIM Auto Detect Start.... fxo%d \n", port); 
        }
/* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
#endif
/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
        adect_acim_usrcfg[port-1].tostart = 1;
        adect_acim_usrcfg[port-1].channum = port;
    }

    ast_mutex_lock(&adt_lock);
    ret = dahdi_pstn_acim_autodetect_start(a->fd, concise);
    ast_mutex_unlock(&adt_lock);

    if ( ret<0 )
    {
        if ( concise == 1 )
        {
            ast_cli(a->fd, "FXO=all!state=failed\n"); 
        }
        else
        {
            ast_cli(a->fd, "PSTN ACIM Auto Detect failed....\n"); 
        }
    }
	/* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    else
    {
        if ( concise == 1 )
        {
            ast_cli(a->fd, "FXO=all!state=successfully\n"); 
        }
        else
        {
            ast_cli(a->fd, "PSTN ACIM Auto Detect successfully....\n"); 
        }    
    }
	/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */

    return CLI_SUCCESS;

}

static char *dahdi_pstn_acim_autodetect_stop_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int port;
    int concise = 0;
    char *portlist;
    char *portstr;

    switch ( cmd )
    {
        case CLI_INIT:
            e->command = "pstn autodetect acim_stop";
            e->usage = "Usage: pstn autodetect acim_stop [port list] [concise]\n"
                       "   [port list]: port list: separated by ',' eg: 1,2,3\n";
            return NULL;
        case CLI_GENERATE:
            return NULL;
    }

    if ( a->argc < 3 )
    {
        return CLI_SHOWUSAGE;
    }
    
    if ( ( a->argv[4] != NULL && strcmp(a->argv[4], "concise") == 0 )
         || ( a->argv[3] != NULL && strcmp(a->argv[3], "concise") == 0 ) )
    {
        concise = 1;
    }
	/* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    if(concise == 1)
    {
        ast_cli(a->fd, "PSTN AutoDetect Stop\n"); 
    }
	/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */

    if ( ast_strlen_zero(a->argv[3]) || strcmp(a->argv[3], "concise") == 0 )
    {
        for ( port=1;port<=g_fxo_amount;port++ )
            adect_acim_usrcfg[port-1].release = 1;
        if ( concise == 1 )
        {
            ast_cli(a->fd, "FXO=all!state=Stoping\n");
        }
        else
        {
            ast_cli(a->fd, "PSTN ACIM Auto Detect Stop....all the ports \n");
        }
    }
    else
    {
        if ( concise != 1 )
        {
            ast_cli(a->fd, "PSTN ACIM Auto Detect Stop.... ports:%s \n", a->argv[3]);
        }
        portlist = ast_strdupa(a->argv[3]);
        while ( (portstr = strsep(&portlist, ",")) )
        {
            port = atoi(portstr);
            if ( port > g_fxo_amount || port < 1 )
            {
                if ( concise == 1 )
                {
                    ast_cli(a->fd, "FXO=%d!state=invalid\n", port); 
                }
                else
                {
                    ast_cli(a->fd, "PSTN ACIM Auto Detect invalid port %d\n", port); 
                }
                continue;
            }
            adect_acim_usrcfg[port-1].release = 1;
            adect_acim_usrcfg[port-1].channum = port;
        }
    }
    ast_mutex_lock(&adt_lock);
    dahdi_pstn_acim_autodetect_thread_release(a->fd, concise);
    ast_mutex_unlock(&adt_lock);

    ast_cli(a->fd, "FXO=all!state=Stop done\n");
    return CLI_SUCCESS;
}

static char *dahdi_pstn_acim_autodetect_getstate_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int port;
    int concise = 0;
    char *portlist;
    char *portstr;

    switch ( cmd )
    {
        case CLI_INIT:
            e->command = "pstn autodetect acim_state";
            e->usage = "Usage: pstn autodetect acim_state [port list] concise \n"
                       "   [port list]: port list: separated by ',' eg: 1,2,3\n";
            return NULL;
        case CLI_GENERATE:
            return NULL;
    }

    if ( a->argc < 2 )
    {
        return CLI_SHOWUSAGE;
    }

    if ( (a->argv[4] != NULL && strcmp( a->argv[4], "concise") == 0)
         || (a->argv[3] != NULL && strcmp( a->argv[3], "concise") == 0)  )
    {
        concise = 1;
    }

     /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */
    if(concise == 1)
    {
        ast_cli(a->fd, "PSTN AutoDetect State\n"); 
    }
	/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] add */

    if ( ast_strlen_zero(a->argv[3]) || strcmp( a->argv[3], "concise") == 0 )
    {
        for ( port=1;port<=g_fxo_amount;port++ )
        {
            if ( concise == 1 )
            {
                if ( adect_acim_usrcfg[port-1].best_result.get_result )
                {
                    ast_cli(a->fd, "FXO=%d!state=%s!ACIM=%d!coefs=%d,%d,%d,%d,%d,%d,%d,%d!ERL=%fdB\nstateID=%d!ErrCod=%d\n",
                            port, adt_acim_state_str[adect_acim_usrcfg[port-1].acim_state],
                            adect_acim_usrcfg[port-1].best_result.coefs.acim,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef1,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef2,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef3,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef4,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef5,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef6,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef7,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef8,
                            adect_acim_usrcfg[port-1].best_result.erl,
                           adect_acim_usrcfg[port-1].acim_state, adect_acim_usrcfg[port-1].acim_errcode);
                }
                else
                {
                    ast_cli(a->fd, "FXO=%d!state=%s\nstateID=%d!ErrCod=%d\n", port,         
                            adt_acim_state_str[adect_acim_usrcfg[port-1].acim_state],
                            adect_acim_usrcfg[port-1].acim_state, adect_acim_usrcfg[port-1].acim_errcode);
                }
            }
            else
            {
                ast_cli(a->fd, "Port [%d] PSTN ACIM Auto Detect state %s erro:%s....\n", port,
                        adt_acim_state_str[adect_acim_usrcfg[port-1].acim_state],
                        adt_acim_error_str[adect_acim_usrcfg[port-1].acim_errcode]);

                if ( adect_acim_usrcfg[port-1].best_result.get_result )
                {
                    ast_cli(a->fd, "Port [%d] PSTN ACIM Auto Detect result: acim %d,"
                            " coefs:%d %d %d %d %d %d %d %d, ERL:%f dB\n", 
                            port,
                            adect_acim_usrcfg[port-1].best_result.coefs.acim,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef1,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef2,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef3,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef4,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef5,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef6,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef7,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef8,
                            adect_acim_usrcfg[port-1].best_result.erl);
                }
            }
        }
    }
    else
    {
        portlist = ast_strdupa(a->argv[3]);
        while ( (portstr = strsep(&portlist, ",")) )
        {
            port = atoi(portstr);
            if ( port > g_fxo_amount || port < 1 )
            {
                if ( concise == 1 )
                {
                    ast_cli(a->fd, "FXO=%d!state=invalid\n", port); 
                }
                else
                {
                    ast_cli(a->fd, "PSTN ACIM Auto Detect invalid port %d\n", port); 
                }
                continue;
            }

            if ( concise == 1 )
            {
			    /* BEGIN: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
                ast_cli(a->fd, "FXO=%d!state=%s!ACIM=%d!coefs=%d,%d,%d,%d,%d,%d,%d,%d!ERL=%f dB\nstateID=%d!ErrCod=%d\n", 
                            port, adt_acim_state_str[adect_acim_usrcfg[port-1].acim_state],
                            adect_acim_usrcfg[port-1].best_result.coefs.acim,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef1,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef2,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef3,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef4,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef5,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef6,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef7,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef8,
                            adect_acim_usrcfg[port-1].best_result.erl,
                            adect_acim_usrcfg[port-1].acim_state, adect_acim_usrcfg[port-1].acim_errcode);
				/* END: jilzhang 20140104 [Event cpt_autodetect_state and acim_autodetect_state] modify */
            }
            else
            {
                ast_cli(a->fd, "Port [%d] PSTN CPT Auto Detect state %s erro:%s....\n", port,
                        adt_acim_state_str[adect_acim_usrcfg[port-1].acim_state],
                        adt_acim_error_str[adect_acim_usrcfg[port-1].acim_errcode]);
    
                if ( adect_acim_usrcfg[port-1].best_result.get_result )
                {
                    ast_cli(a->fd, "Port [%d] PSTN CPT Auto Detect result: acim %d,"
                            " coefs:%d %d %d %d %d %d %d %d, ERL:%f dB\n", 
                            port,
                            adect_acim_usrcfg[port-1].best_result.coefs.acim,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef1,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef2,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef3,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef4,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef5,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef6,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef7,
                            adect_acim_usrcfg[port-1].best_result.coefs.coef8,
                            adect_acim_usrcfg[port-1].best_result.erl);
                }
            }
        }
    }

    return CLI_SUCCESS;
}

static char *dahdi_pstn_acim_autodetect_doset_cli(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
    int dfd;
    char dahdidev[80]="";
    char *cofesstr;
    int ret;
    int cli_ret;
    int i;
    int port;
    char *portlist;
    char *portstr;
    int myacim,mycoef1,mycoef2,mycoef3,mycoef4,mycoef5,mycoef6,mycoef7,mycoef8;
    int doset[MAX_FXO_CHANNEL]={0,};

    switch ( cmd )
    {
        case CLI_INIT:
            e->command = "pstn autodetect acim_doset {auto|manual|mode|cfg}";
            e->usage = "Usage: pstn autodetect acim_doset auto|manual|mode [port list]\n"
                       "       pstn autodetect acim_doset cfg [cofes] [port list]\n";
            return NULL;
        case CLI_GENERATE:
            return NULL;
    }

    if ( a->argc < 4 )
    {
        return CLI_SHOWUSAGE;
    }

    if ( !strcasecmp(S_OR(a->argv[3], ""), "cfg") )
    {
        if ( a->argc < 5 )
            return CLI_SHOWUSAGE;

        if ( ast_strlen_zero(a->argv[5]) )
        {
            for ( i=0;i<g_fxo_amount;i++ )
                doset[i] = 1;
        }
        else
        {
            portlist = ast_strdupa(a->argv[5]);
            while ( (portstr = strsep(&portlist, ",")) )
            {
                port = atoi(portstr);
                if ( port > g_fxo_amount || port < 1 )
                {
                    ast_cli(a->fd, "PSTN ACIM Auto Detect invalid port %d\n", port); 
                    continue;
                }
                doset[port-1] = 1;
            }
        }
    }
    else//for auto,manual,mode 
    {
        if ( ast_strlen_zero(a->argv[4]) )
            for ( i=0;i<g_fxo_amount;i++ )
                doset[i] = 1;
        else
        {
            portlist = ast_strdupa(a->argv[4]);
            while ( (portstr = strsep(&portlist, ",")) )
            {
                port = atoi(portstr);
                if ( port > g_fxo_amount || port < 1 )
                {
                    ast_cli(a->fd, "PSTN ACIM Auto Detect invalid port %d\n", port); 
                    continue;
                }
                doset[port-1] = 1;
            }
        }
    }

    ast_mutex_lock(&adt_lock);

    if ( ADT_CPT == g_adt_item )
    {
        if ( !cpt_autodetect_is_finished() )
        {
            ast_log(LOG_ERROR, "PSTN Auto detect CPT is running!\n");
            ast_mutex_unlock(&adt_lock);
            return CLI_FAILURE;
        }
    }
    if ( ADT_ACIM == g_adt_item )
    {
        if ( !acim_autodetect_is_finished() )
        {
            ast_log(LOG_ERROR, "PSTN Auto detect ACIM is running!\n");
            ast_mutex_unlock(&adt_lock);
            return CLI_FAILURE;
        }
    }
    unload_chan_dahdi();
    cli_ret = 0;

    for ( i=0;i<g_fxo_amount;i++ )
    {
        if ( !doset[i] )
            continue;
        port = i+1;
        if ( !strcasecmp(S_OR(a->argv[3], ""), "auto") )
        {
            adect_acim_usrcfg[i].acim_doset_mode = ACIM_DOSET_AUTO;
        }
        else if ( !strcasecmp(S_OR(a->argv[3], ""), "manual") )
        {
            adect_acim_usrcfg[i].acim_doset_mode = ACIM_DOSET_MANUAL;
        }
        else if ( !strcasecmp(S_OR(a->argv[3], ""), "mode") )
        {
            ast_cli(a->fd, "Port[%d] PSTN CPT Auto Detect doset mode: %s\n", port, 
                    (adect_acim_usrcfg[i].acim_doset_mode == ACIM_DOSET_MANUAL)?
                    "Manual Setting":"Auto Setting");
        }
        else if ( !strcasecmp(S_OR(a->argv[3], ""), "cfg") )
        {
            snprintf(dahdidev,sizeof(dahdidev),"%s/%d",dahdipath, port);

            dfd = open(dahdidev,O_RDWR);
            if ( dfd < 0 )
            {
                ast_cli(a->fd, "%s absent: %s\n",dahdidev,strerror(errno));
                cli_ret = 1;
                goto doset_return;
            }

            if ( ast_strlen_zero(a->argv[4]) )
            {
                close(dfd);
                cli_ret = 2;
                goto doset_return;
            }

            cofesstr = ast_strdupa(a->argv[4]);

            if ( sscanf(cofesstr, "%d,%d,%d,%d,%d,%d,%d,%d,%d",
                        &myacim, &mycoef1, &mycoef2, &mycoef3, &mycoef4,
                        &mycoef5, &mycoef6, &mycoef7, &mycoef8)!= 9 )
            {
                ast_cli(a->fd, "invalid acim coefs number\n");
                close(dfd);
                cli_ret = 1;
                goto doset_return;
            }

            adect_acim_usrcfg[i].best_result.coefs.acim = myacim;
            adect_acim_usrcfg[i].best_result.coefs.coef1 = mycoef1;
            adect_acim_usrcfg[i].best_result.coefs.coef2 = mycoef2;
            adect_acim_usrcfg[i].best_result.coefs.coef3 = mycoef3;
            adect_acim_usrcfg[i].best_result.coefs.coef4 = mycoef4;
            adect_acim_usrcfg[i].best_result.coefs.coef5 = mycoef5;
            adect_acim_usrcfg[i].best_result.coefs.coef6 = mycoef6;
            adect_acim_usrcfg[i].best_result.coefs.coef7 = mycoef7;
            adect_acim_usrcfg[i].best_result.coefs.coef8 = mycoef8;

            ast_cli(a->fd, "Set acim cofes:%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
                    myacim, mycoef1, mycoef2, mycoef3, mycoef4,
                    mycoef5, mycoef6, mycoef7, mycoef8);

            ret = do_set(dfd, i);

            if ( ret<0 )
            {
                close(dfd);
                ast_cli(a->fd,"port %d set acim cofes error!\n", port);
                continue;
            }
            close(dfd);
        }
        else
        {
            cli_ret = 2;
            goto doset_return;
        }
    }
    doset_return:

    load_chan_dahdi();
    ast_mutex_unlock(&adt_lock);

    if ( 0 == cli_ret )
        return CLI_SUCCESS;
    else if ( 1 == cli_ret )
        return CLI_FAILURE;
    else
        return CLI_SHOWUSAGE;

}

static int load_module(void)
{
    ast_mutex_init(&app_lock);
    ast_mutex_init(&adt_lock);
    init_acim_autodetect();
    ast_cli_register_multiple(cli, ARRAY_LEN(cli));
    return 0;
}

static int unload_module(void)
{
    ast_mutex_destroy(&adt_lock);
    ast_mutex_destroy(&app_lock);
    ast_cli_unregister_multiple(cli, ARRAY_LEN(cli));
    return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PSTN Auto Detect Channel Driver",
                .load = load_module,
                .unload = unload_module,
                .load_pri = AST_MODPRI_CHANNEL_DRIVER,
               );
