/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \author Joshua Colp <jcolp@digium.com>
 *
 * \brief PSJIP SIP Channel Driver
 *
 * \ingroup channel_drivers
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_session</depend>
	<support_level>core</support_level>
 ***/

#ifdef GRANDSTREAM_NETWORKS
#define AST_MODULE_LOG "pjsip"
#endif

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>
#include <pjlib.h>

#ifdef GRANDSTREAM_NETWORKS
#include <pjmedia/sdp_neg.h>
#include <asterisk/astdb.h>
#endif

ASTERISK_FILE_VERSION(__FILE__, "$Revision$")

#include "asterisk/lock.h"
#include "asterisk/channel.h"
#include "asterisk/module.h"
#include "asterisk/pbx.h"
#include "asterisk/rtp_engine.h"
#include "asterisk/acl.h"
#include "asterisk/callerid.h"
#include "asterisk/file.h"
#include "asterisk/cli.h"
#include "asterisk/app.h"
#include "asterisk/musiconhold.h"
#include "asterisk/causes.h"
#include "asterisk/taskprocessor.h"
#include "asterisk/dsp.h"
#include "asterisk/stasis_endpoints.h"
#include "asterisk/stasis_channels.h"
#include "asterisk/indications.h"
#include "asterisk/format_cache.h"
#include "asterisk/translate.h"
#include "asterisk/threadstorage.h"
#include "asterisk/features_config.h"
#include "asterisk/pickup.h"
#include "asterisk/test.h"

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"

#include "pjsip/include/chan_pjsip.h"
#include "pjsip/include/dialplan_functions.h"

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/app.h"
#include "asterisk/pjsip_cei.h"
#include "asterisk/ccss.h"
#include "asterisk/stasis_message_router.h"


/*** DOCUMENTATION
    <application name="SIPSetCallType" language="en_US">
        <synopsis>
            Set Call Type Header.
        </synopsis>
        <syntax>
            <parameter name="type" required="true" />
        </syntax>
        <description>
            <para>Set Call Type header for SIP response</para>
        </description>
    </application>
	<application name="SIPSetMCBResp" language="en_US">
		<synopsis>
			Set MCB Header Response.
		</synopsis>
		<syntax>
			<parameter name="header" required="true" />
		</syntax>
		<description>
			<para>Set MCB header for SIP response</para>
		</description>
	</application>
    <application name="SIPSetConfFocus" language="en_US">
		<synopsis>
			Add isfocus in a contact parameter.
		</synopsis>
		<syntax>
			<parameter name="room_number" required="true" />
		</syntax>
		<description>
			<para>Add isfocus in a contact parameter
                  during responding of an incoming INVITE</para>
		</description>
	</application>
    <application name="SIPSetContactUser" language="en_US">
		<synopsis>
			Set Contact URI Host User Part.
		</synopsis>
		<syntax>
			<parameter name="user" required="true" />
		</syntax>
		<description>
			<para>Set Contact URI Host User Part
                  during responding of an incoming INVITE</para>
		</description>
	</application>
    <application name="SIPTerminatedNotify" language="en_US">
        <synopsis>
            Send SIP NOTIFY message with terminated state.
        </synopsis>
        <syntax>
            <parameter name="conference" required="true" />
        </syntax>
        <description>
            <para>Send SIP NOTIFY message with terminated state and clean up
                  SIP subscription.</para>
        </description>
    </application>
    <application name="SIPCEIJoined" language="en_US">
        <synopsis>
            Given channel joined to a conference with valid authentication.
        </synopsis>
        <syntax>
            <parameter name="conference" required="true" />
        </syntax>
        <description>
            <para>Given channel joined to a conference with valid authentication.</para>
        </description>
    </application>
    <application name="SIPSubscribeExtMCB" language="en_US">
		<synopsis>
            Subscribe an external MCB.
		</synopsis>
		<syntax>
			<parameter name="local_conference" required="true" />
            <parameter name="local_conference_id" required="true" />
            <parameter name="remote_conference" required="true" />
            <parameter name="remote_conference_id" required="true" />
            <parameter name="remote_channel" required="true" />
            <parameter name="endpoint_name" required="true" />
		</syntax>
		<description>
			<para>Subscribe to and handle subscription process of
                  an external MCB</para>
		</description>
	</application>
    <application name="SIPUnsubscribeExtMCB" language="en_US">
		<synopsis>
            Unsubscribe an external MCB.
		</synopsis>
		<syntax>
			<parameter name="local_conference_id" required="true" />
            <parameter name="remote_conference_id" required="true" />
            <parameter name="endpoint_name" required="true" />
		</syntax>
		<description>
			<para>Unsubscribe to and handle unsubscription process of
                  an external MCB</para>
		</description>
	</application>
  <application name="SIPSubscribeMCB" language="en_US">
  <synopsis>
          Subscribe an internal MCB.
  </synopsis>
  <syntax>
          <parameter name="local_conference" required="true" />
          <parameter name="local_conference_id" required="true" />
          <parameter name="remote_conference" required="true" />
          <parameter name="remote_conference_id" required="true" />
  </syntax>
  <description>
    <para>Subscribe to and handle subscription process of
                an internal MCB</para>
  </description>
</application>
  <application name="SIPUnsubscribeMCB" language="en_US">
  <synopsis>
          Unsubscribe an internal MCB.
  </synopsis>
  <syntax>
          <parameter name="local_conference_id" required="true" />
          <parameter name="remote_conference_id" required="true" />
  </syntax>
  <description>
    <para>Unsubscribe to and handle unsubscription process of
                an internal MCB</para>
  </description>
</application>
 ***/
#endif

AST_THREADSTORAGE(uniqueid_threadbuf);
#define UNIQUEID_BUFSIZE 256

static const char channel_type[] = "PJSIP";
#ifdef GRANDSTREAM_NETWORKS
static int globalEnabledCEI = 0;
static struct stasis_message_router *cei_toggle_enable_router = NULL;

static char *app_sipsetcalltype = "SIPSetCallType";
static char *app_sipsetmcbresp = "SIPSetMCBResp";
static char *app_sipsetconffocus = "SIPSetConfFocus";
static char *app_sipsetcontactuser = "SIPSetContactUser";
static char *app_sipterminatednotify = "SIPTerminatedNotify";
static char *app_sipceijoined = "SIPCEIJoined";
// Internal MCB
 static char *app_sipsubscribemcb = "SIPSubscribeMCB";
 static char *app_sipunsubscribemcb = "SIPUnsubscribeMCB";
// External MCB
static char *app_sipsubscribeextmcb = "SIPSubscribeExtMCB";
static char *app_sipunsubscribeextmcb = "SIPUnsubscribeExtMCB";
static int sip_setcalltype(struct ast_channel *chan, const char *data);
static int sip_setmcbresp(struct ast_channel *chan, const char *data);
static int sip_set_conf_focus(struct ast_channel *chan, const char *data);
static int sip_set_contact_user(struct ast_channel *chan, const char *data);
static int sip_terminated_notify(struct ast_channel *chan, const char *data);
static int sip_cei_joined(struct ast_channel *chan, const char *data);
static int sip_external_mcb_sub(struct ast_channel *chan, const char *data);
static int sip_external_mcb_unsub(struct ast_channel *chan, const char *data);
static int sip_internal_mcb_sub(struct ast_channel *chan, const char *data);
static int sip_internal_mcb_unsub(struct ast_channel *chan, const char *data);
static int sip_voip_trunk_enable_fax_detect(struct ast_channel *chan, unsigned int fax_detect);

struct ast_gs_feature_map gs_feature_map[GS_FEATURE_MAP_COUNT] = {
    { "blindxfer", "#", GS_FEATURE_DISABLE },
    { "atxfer",    "",  GS_FEATURE_DISABLE },
    { "parkcall",  "",  GS_FEATURE_DISABLE },
    { "automixmon", "",  GS_FEATURE_DISABLE }
};
#endif

static unsigned int chan_idx;

static void chan_pjsip_pvt_dtor(void *obj)
{
	struct chan_pjsip_pvt *pvt = obj;
	int i;

	for (i = 0; i < SIP_MEDIA_SIZE; ++i) {
		ao2_cleanup(pvt->media[i]);
		pvt->media[i] = NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
    if (pvt->call_type != NULL) {
        free(pvt->call_type);
        pvt->call_type = NULL;
    }

	if (pvt->mcb_request != NULL) {
		free(pvt->mcb_request);
		pvt->mcb_request = NULL;
	}

	if (pvt->mcb_response != NULL) {
		free(pvt->mcb_response);
		pvt->mcb_response = NULL;
	}

    if (pvt->contact_user != NULL) {
		free(pvt->contact_user);
		pvt->contact_user = NULL;
	}
#endif
}

/* \brief Asterisk core interaction functions */
static struct ast_channel *chan_pjsip_request(const char *type, struct ast_format_cap *cap, const struct ast_assigned_ids *assignedids, const struct ast_channel *requestor, const char *data, int *cause);
static int chan_pjsip_sendtext(struct ast_channel *ast, const char *text);
static int chan_pjsip_digit_begin(struct ast_channel *ast, char digit);
static int chan_pjsip_digit_end(struct ast_channel *ast, char digit, unsigned int duration);
static int chan_pjsip_call(struct ast_channel *ast, const char *dest, int timeout);
static int chan_pjsip_hangup(struct ast_channel *ast);
static int chan_pjsip_answer(struct ast_channel *ast);
#ifdef GRANDSTREAM_NETWORKS
static int chan_pjsip_special_answer(struct ast_channel *chan, struct ast_channel *peer);

void add_feature_map_header(pjsip_tx_data *tdata, int outgoing_call);
static int get_feature_code(void);

// CEI
//static void clear_peer_conferences(struct ast_sip_endpoint *endpoint);
static void clear_peer_conference(struct ast_sip_endpoint *endpoint, const char *channel);
static void destroy_conference(struct sip_conference *conference);
static int add_peer_conference(const char *channel,
                               struct ast_sip_endpoint *endpoint,
                               const char *conference);
static void publish_cei_unsub(const char *conference, const char *channel);
static void publish_cei_joined(const char *conference, const char *channel);

/* Add by bfdu for video reinvite */
static int send_video_reinvite_request(void *data);
static int send_video_reinvite_response(void *data);
static void ast_sip_session_send_reinvite_response(struct ast_sip_session *session);
static void send_video_reinvite_request_to_callee(struct ast_sip_session *session, const struct ast_sip_video_reinvite_data *reinvite_data);
static void send_video_reinvite_response_to_caller(struct ast_sip_session *session, const struct ast_sip_video_reinvite_data *reinvite_data);
static void handle_response_video_reinvite(struct ast_sip_session *session, struct ast_sip_video_reinvite_data *reinvite_data);
static unsigned char judge_if_support_video_media(struct ast_sip_session *session);
static unsigned char judge_if_both_type_is_sip(struct ast_channel* chan);
/* End by bfdu for video reinvite */
#endif
static struct ast_frame *chan_pjsip_read(struct ast_channel *ast);
static int chan_pjsip_write(struct ast_channel *ast, struct ast_frame *f);
static int chan_pjsip_indicate(struct ast_channel *ast, int condition, const void *data, size_t datalen);
static int chan_pjsip_transfer(struct ast_channel *ast, const char *target);
static int chan_pjsip_fixup(struct ast_channel *oldchan, struct ast_channel *newchan);
static int chan_pjsip_devicestate(const char *data);
static int chan_pjsip_queryoption(struct ast_channel *ast, int option, void *data, int *datalen);
static const char *chan_pjsip_get_uniqueid(struct ast_channel *ast);

/*! \brief PBX interface structure for channel registration */
struct ast_channel_tech chan_pjsip_tech = {
	.type = channel_type,
	.description = "PJSIP Channel Driver",
	.requester = chan_pjsip_request,
	.send_text = chan_pjsip_sendtext,
	.send_digit_begin = chan_pjsip_digit_begin,
	.send_digit_end = chan_pjsip_digit_end,
	.call = chan_pjsip_call,
	.hangup = chan_pjsip_hangup,
	.answer = chan_pjsip_answer,
#ifdef GRANDSTREAM_NETWORKS
	.special_answer = chan_pjsip_special_answer,
#endif
	.read = chan_pjsip_read,
	.write = chan_pjsip_write,
	.write_video = chan_pjsip_write,
	.exception = chan_pjsip_read,
	.indicate = chan_pjsip_indicate,
	.transfer = chan_pjsip_transfer,
	.fixup = chan_pjsip_fixup,
	.devicestate = chan_pjsip_devicestate,
	.queryoption = chan_pjsip_queryoption,
	.func_channel_read = pjsip_acf_channel_read,
	.get_pvt_uniqueid = chan_pjsip_get_uniqueid,
	.properties = AST_CHAN_TP_WANTSJITTER | AST_CHAN_TP_CREATESJITTER
};

/*! \brief SIP session interaction functions */
static void chan_pjsip_session_begin(struct ast_sip_session *session);
static void chan_pjsip_session_end(struct ast_sip_session *session);
static int chan_pjsip_incoming_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata);
static void chan_pjsip_incoming_response(struct ast_sip_session *session, struct pjsip_rx_data *rdata);

/*! \brief SIP session supplement structure */
static struct ast_sip_session_supplement chan_pjsip_supplement = {
	.method = "INVITE",
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_CHANNEL,
	.session_begin = chan_pjsip_session_begin,
	.session_end = chan_pjsip_session_end,
	.incoming_request = chan_pjsip_incoming_request,
	.incoming_response = chan_pjsip_incoming_response,
	/* It is important that this supplement runs after media has been negotiated */
	.response_priority = AST_SIP_SESSION_AFTER_MEDIA,
};

static int chan_pjsip_incoming_ack(struct ast_sip_session *session, struct pjsip_rx_data *rdata);

static struct ast_sip_session_supplement chan_pjsip_ack_supplement = {
	.method = "ACK",
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_CHANNEL,
	.incoming_request = chan_pjsip_incoming_ack,
};

/*! \brief Function called by RTP engine to get local audio RTP peer */
static enum ast_rtp_glue_result chan_pjsip_get_rtp_peer(struct ast_channel *chan, struct ast_rtp_instance **instance)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt = channel->pvt;
	struct ast_sip_endpoint *endpoint;

	if (!pvt || !channel->session || !pvt->media[SIP_MEDIA_AUDIO]->rtp) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	endpoint = channel->session->endpoint;

	*instance = pvt->media[SIP_MEDIA_AUDIO]->rtp;
	ao2_ref(*instance, +1);

	ast_assert(endpoint != NULL);
	if (endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	if (endpoint->media.direct_media.enabled) {
		return AST_RTP_GLUE_RESULT_REMOTE;
	}

	return AST_RTP_GLUE_RESULT_LOCAL;
}

/*! \brief Function called by RTP engine to get local video RTP peer */
static enum ast_rtp_glue_result chan_pjsip_get_vrtp_peer(struct ast_channel *chan, struct ast_rtp_instance **instance)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt = channel->pvt;
	struct ast_sip_endpoint *endpoint;

	if (!pvt || !channel->session || !pvt->media[SIP_MEDIA_VIDEO]->rtp) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	endpoint = channel->session->endpoint;

	*instance = pvt->media[SIP_MEDIA_VIDEO]->rtp;
	ao2_ref(*instance, +1);

	ast_assert(endpoint != NULL);
	if (endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	return AST_RTP_GLUE_RESULT_LOCAL;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for FEC/FECC/Slides video RTP P2P mode, 2016.03.15 */
static enum ast_rtp_glue_result chan_pjsip_get_slides_vrtp_peer(struct ast_channel *chan, struct ast_rtp_instance **instance)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct ast_sip_endpoint *endpoint;

	if (!channel->session || !channel->session->slides_video_media.rtp) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	endpoint = channel->session->endpoint;

	*instance = channel->session->slides_video_media.rtp;
	ao2_ref(*instance, +1);

	ast_assert(endpoint != NULL);
	if (endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	if (endpoint->media.direct_media.enabled) {
		return AST_RTP_GLUE_RESULT_REMOTE;
	}

	return AST_RTP_GLUE_RESULT_LOCAL;
}

static enum ast_rtp_glue_result chan_pjsip_get_fec_rtp_peer(struct ast_channel *chan, struct ast_rtp_instance **instance)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct ast_sip_endpoint *endpoint;

	if (!channel->session || !channel->session->fec_media.rtp) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	endpoint = channel->session->endpoint;

	*instance = channel->session->fec_media.rtp;
	ao2_ref(*instance, +1);

	ast_assert(endpoint != NULL);
	if (endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	if (endpoint->media.direct_media.enabled) {
		return AST_RTP_GLUE_RESULT_REMOTE;
	}

	return AST_RTP_GLUE_RESULT_LOCAL;
}

static enum ast_rtp_glue_result chan_pjsip_get_fecc_rtp_peer(struct ast_channel *chan, struct ast_rtp_instance **instance)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct ast_sip_endpoint *endpoint;

	if (!channel->session || !channel->session->fecc_media.rtp) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	endpoint = channel->session->endpoint;

	*instance = channel->session->fecc_media.rtp;
	ao2_ref(*instance, +1);

	ast_assert(endpoint != NULL);
	if (endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	if (endpoint->media.direct_media.enabled) {
		return AST_RTP_GLUE_RESULT_REMOTE;
	}

	return AST_RTP_GLUE_RESULT_LOCAL;
}
#endif

/*! \brief Function called by RTP engine to get peer capabilities */
static void chan_pjsip_get_codec(struct ast_channel *chan, struct ast_format_cap *result)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
#ifdef GRANDSTREAM_NETWORKS
	ast_format_cap_append_from_cap(result, channel->session->final_caps, AST_MEDIA_TYPE_UNKNOWN);
#else
	ast_format_cap_append_from_cap(result, channel->session->endpoint->media.codecs, AST_MEDIA_TYPE_UNKNOWN);
#endif
}

static int send_direct_media_request(void *data)
{
	struct ast_sip_session *session = data;
	int res;
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP and canreinvite, 2015.08.12 */
	session->is_send_direct_media_request = 1;
#endif

	res = ast_sip_session_refresh(session, NULL, NULL, NULL,
		session->endpoint->media.direct_media.method, 1);
	ao2_ref(session, -1);
	return res;
}

/*! \brief Destructor function for \ref transport_info_data */
static void transport_info_destroy(void *obj)
{
	struct transport_info_data *data = obj;
	ast_free(data);
}

/*! \brief Datastore used to store local/remote addresses for the
 * INVITE request that created the PJSIP channel */
static struct ast_datastore_info transport_info = {
	.type = "chan_pjsip_transport_info",
	.destroy = transport_info_destroy,
};

static struct ast_datastore_info direct_media_mitigation_info = { };

static int direct_media_mitigate_glare(struct ast_sip_session *session)
{
	RAII_VAR(struct ast_datastore *, datastore, NULL, ao2_cleanup);

	if (session->endpoint->media.direct_media.glare_mitigation ==
			AST_SIP_DIRECT_MEDIA_GLARE_MITIGATION_NONE) {
		return 0;
	}

	datastore = ast_sip_session_get_datastore(session, "direct_media_glare_mitigation");
	if (!datastore) {
		return 0;
	}

	/* Removing the datastore ensures we won't try to mitigate glare on subsequent reinvites */
	ast_sip_session_remove_datastore(session, "direct_media_glare_mitigation");

	if ((session->endpoint->media.direct_media.glare_mitigation ==
			AST_SIP_DIRECT_MEDIA_GLARE_MITIGATION_OUTGOING &&
			session->inv_session->role == PJSIP_ROLE_UAC) ||
			(session->endpoint->media.direct_media.glare_mitigation ==
			AST_SIP_DIRECT_MEDIA_GLARE_MITIGATION_INCOMING &&
			session->inv_session->role == PJSIP_ROLE_UAS)) {
		return 1;
	}

	return 0;
}

static int check_for_rtp_changes(struct ast_channel *chan, struct ast_rtp_instance *rtp,
		struct ast_sip_session_media *media, int rtcp_fd)
{
	int changed = 0;

	if (rtp) {
		changed = ast_rtp_instance_get_and_cmp_remote_address(rtp, &media->direct_media_addr);
		if (media->rtp) {
			ast_channel_set_fd(chan, rtcp_fd, -1);
			ast_rtp_instance_set_prop(media->rtp, AST_RTP_PROPERTY_RTCP, 0);
		}
	} else if (!ast_sockaddr_isnull(&media->direct_media_addr)){
		ast_sockaddr_setnull(&media->direct_media_addr);
		changed = 1;
		if (media->rtp) {
			ast_rtp_instance_set_prop(media->rtp, AST_RTP_PROPERTY_RTCP, 1);
			ast_channel_set_fd(chan, rtcp_fd, ast_rtp_instance_fd(media->rtp, 1));
		}
	}

	return changed;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for BFCP/FEC/FECC, when open canreinvite, need to chang direct media addr */
static int check_for_bfcp_changes(struct ast_sip_session_media *media, struct ast_sip_session_media *peer_media)
{
	int changed = 0;

	if (media && peer_media && peer_media->bfcp) {
		changed = ast_bfcp_get_and_cmp_requested_target_address(peer_media->bfcp, &media->direct_media_addr);
	}

	if (media && peer_media && peer_media->bfcp_tcp) {
		changed = ast_bfcp_tcp_get_and_cmp_requested_target_address(peer_media->bfcp_tcp, &media->direct_media_addr);
	}

	return changed;
}

static int check_for_slides_video_rtp_changes(struct ast_channel *chan, struct ast_rtp_instance *rtp,
		struct ast_sip_extra_media *slides_video_media, int rtcp_fd)
{
	int changed = 0;

	if (rtp) {
		changed = ast_rtp_instance_get_and_cmp_remote_address(rtp, &slides_video_media->direct_media_addr);
		if (slides_video_media->rtp) {
			ast_channel_set_fd(chan, rtcp_fd, -1);
			ast_rtp_instance_set_prop(slides_video_media->rtp, AST_RTP_PROPERTY_RTCP, 0);
		}
	} else if (!ast_sockaddr_isnull(&slides_video_media->direct_media_addr)){
		ast_sockaddr_setnull(&slides_video_media->direct_media_addr);
		changed = 1;
		if (slides_video_media->rtp) {
			ast_rtp_instance_set_prop(slides_video_media->rtp, AST_RTP_PROPERTY_RTCP, 1);
			ast_channel_set_fd(chan, rtcp_fd, ast_rtp_instance_fd(slides_video_media->rtp, 1));
		}
	}

	return changed;
}

static int check_for_fec_rtp_changes(struct ast_channel *chan, struct ast_rtp_instance *rtp,
		struct ast_sip_extra_media *fec_media, int rtcp_fd)
{
	int changed = 0;

	if (rtp) {
		changed = ast_rtp_instance_get_and_cmp_remote_address(rtp, &fec_media->direct_media_addr);
		if (fec_media->rtp) {
			ast_channel_set_fd(chan, rtcp_fd, -1);
			ast_rtp_instance_set_prop(fec_media->rtp, AST_RTP_PROPERTY_RTCP, 0);
		}
	} else if (!ast_sockaddr_isnull(&fec_media->direct_media_addr)){
		ast_sockaddr_setnull(&fec_media->direct_media_addr);
		changed = 1;
		if (fec_media->rtp) {
			ast_rtp_instance_set_prop(fec_media->rtp, AST_RTP_PROPERTY_RTCP, 1);
			ast_channel_set_fd(chan, rtcp_fd, ast_rtp_instance_fd(fec_media->rtp, 1));
		}
	}

	return changed;
}

static int check_for_fecc_rtp_changes(struct ast_channel *chan, struct ast_rtp_instance *rtp,
		struct ast_sip_extra_media *fecc_media, int rtcp_fd)
{
	int changed = 0;

	if (rtp) {
		changed = ast_rtp_instance_get_and_cmp_remote_address(rtp, &fecc_media->direct_media_addr);
		if (fecc_media->rtp) {
			ast_channel_set_fd(chan, rtcp_fd, -1);
			ast_rtp_instance_set_prop(fecc_media->rtp, AST_RTP_PROPERTY_RTCP, 0);
		}
	} else if (!ast_sockaddr_isnull(&fecc_media->direct_media_addr)){
		ast_sockaddr_setnull(&fecc_media->direct_media_addr);
		changed = 1;
		if (fecc_media->rtp) {
			ast_rtp_instance_set_prop(fecc_media->rtp, AST_RTP_PROPERTY_RTCP, 1);
			ast_channel_set_fd(chan, rtcp_fd, ast_rtp_instance_fd(fecc_media->rtp, 1));
		}
	}

	return changed;
}
#endif

/*! \brief Function called by RTP engine to change where the remote party should send media */
static int chan_pjsip_set_rtp_peer(struct ast_channel *chan,
		struct ast_rtp_instance *rtp,
		struct ast_rtp_instance *vrtp,
		struct ast_rtp_instance *tpeer,
		const struct ast_format_cap *cap,
		int nat_active)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt = channel->pvt;
	struct ast_sip_session *session = channel->session;
	int changed = 0;

	/* Don't try to do any direct media shenanigans on early bridges */
	if ((rtp || vrtp || tpeer) && !ast_channel_is_bridged(chan)) {
		ast_debug(4, "Disregarding setting RTP on %s: channel is not bridged\n", ast_channel_name(chan));
		return 0;
	}

	if (nat_active && session->endpoint->media.direct_media.disable_on_nat) {
		ast_debug(4, "Disregarding setting RTP on %s: NAT is active\n", ast_channel_name(chan));
		return 0;
	}

	if (pvt->media[SIP_MEDIA_AUDIO]) {
		changed |= check_for_rtp_changes(chan, rtp, pvt->media[SIP_MEDIA_AUDIO], 1);
	}
	if (pvt->media[SIP_MEDIA_VIDEO]) {
		changed |= check_for_rtp_changes(chan, vrtp, pvt->media[SIP_MEDIA_VIDEO], 3);
	}
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC direct media */
	/* Only do this when has bfcp/fec/fecc */
	if (session->bfcp_para.has_bfcp == 1 || session->bfcp_para.has_tcp_bfcp == 1 ||
		session->fec_para.has_fec == 1 || session->fecc_para.has_fecc == 1) {
		ast_channel_unlock(chan);
		RAII_VAR(struct ast_channel *, bridged, ast_channel_bridge_peer(chan), ast_channel_cleanup);
		ast_channel_lock(chan);
		if (NULL == bridged) {
			ast_log(LOG_DEBUG, "set peer bridge is not exsit");
			return 0;
		}
		struct ast_sip_channel_pvt *peer_channel = ast_channel_tech_pvt(bridged);
		struct chan_pjsip_pvt *peer_pvt = peer_channel->pvt;
		/* if audio/video not change, don't mark change flag, if need support later.
		 * just change direct media address.
		 */
		if (peer_pvt->media[SIP_MEDIA_APPLICATION]) {
			check_for_bfcp_changes(pvt->media[SIP_MEDIA_APPLICATION], peer_pvt->media[SIP_MEDIA_APPLICATION]);
		}
		if (peer_channel->session && channel->session) {
			check_for_slides_video_rtp_changes(chan, peer_channel->session->slides_video_media.rtp, &channel->session->slides_video_media, 8);
		}
		if (peer_channel->session && channel->session) {
			check_for_fec_rtp_changes(chan, peer_channel->session->fec_media.rtp, &channel->session->fec_media, 10);
		}
		if (peer_channel->session && channel->session) {
			check_for_fecc_rtp_changes(chan, peer_channel->session->fecc_media.rtp, &channel->session->fecc_media, 12);
		}
	}
#endif

	if (direct_media_mitigate_glare(session)) {
		ast_debug(4, "Disregarding setting RTP on %s: mitigating re-INVITE glare\n", ast_channel_name(chan));
		return 0;
	}

	if (cap && ast_format_cap_count(cap) && !ast_format_cap_identical(session->direct_media_cap, cap)) {
		ast_format_cap_remove_by_type(session->direct_media_cap, AST_MEDIA_TYPE_UNKNOWN);
		ast_format_cap_append_from_cap(session->direct_media_cap, cap, AST_MEDIA_TYPE_UNKNOWN);
		changed = 1;
	}

	if (changed) {
		ao2_ref(session, +1);

		ast_debug(4, "RTP changed on %s; initiating direct media update\n", ast_channel_name(chan));
		if (ast_sip_push_task(session->serializer, send_direct_media_request, session)) {
			ao2_cleanup(session);
		}
	}

	return 0;
}

/*! \brief Local glue for interacting with the RTP engine core */
static struct ast_rtp_glue chan_pjsip_rtp_glue = {
	.type = "PJSIP",
	.get_rtp_info = chan_pjsip_get_rtp_peer,
	.get_vrtp_info = chan_pjsip_get_vrtp_peer,
	.get_codec = chan_pjsip_get_codec,
	.update_peer = chan_pjsip_set_rtp_peer,
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for FEC/FECC/Slides video RTP P2P mode, 2016.03.15 */
	.get_slides_vrtp_info = chan_pjsip_get_slides_vrtp_peer,
	.get_fec_rtp_info = chan_pjsip_get_fec_rtp_peer,
	.get_fecc_rtp_info = chan_pjsip_get_fecc_rtp_peer,
#endif
};

static void set_channel_on_rtp_instance(struct chan_pjsip_pvt *pvt, const char *channel_id)
{
	if (pvt->media[SIP_MEDIA_AUDIO] && pvt->media[SIP_MEDIA_AUDIO]->rtp) {
		ast_rtp_instance_set_channel_id(pvt->media[SIP_MEDIA_AUDIO]->rtp, channel_id);
	}
	if (pvt->media[SIP_MEDIA_VIDEO] && pvt->media[SIP_MEDIA_VIDEO]->rtp) {
		ast_rtp_instance_set_channel_id(pvt->media[SIP_MEDIA_VIDEO]->rtp, channel_id);
	}
}

#ifdef GRANDSTREAM_NETWORKS
static struct ast_channel *chan_pjsip_new(struct ast_sip_session *session, int state, const char *exten, const char *title, const struct ast_assigned_ids *assignedids, const struct ast_channel *requestor, const char *cid_name, struct ast_format_cap *req_caps_requestor)
#else
/*! \brief Function called to create a new PJSIP Asterisk channel */
static struct ast_channel *chan_pjsip_new(struct ast_sip_session *session, int state, const char *exten, const char *title, const struct ast_assigned_ids *assignedids, const struct ast_channel *requestor, const char *cid_name)
#endif
{
	struct ast_channel *chan;
	struct ast_format_cap *caps;
	RAII_VAR(struct chan_pjsip_pvt *, pvt, NULL, ao2_cleanup);
	struct ast_sip_channel_pvt *channel;
	struct ast_variable *var;

	if (!(pvt = ao2_alloc(sizeof(*pvt), chan_pjsip_pvt_dtor))) {
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	pvt->call_type = NULL;
	pvt->mcb_request = NULL;
	pvt->mcb_response = NULL;
    pvt->contact_user = NULL;
    pvt->isConference = FALSE;
#endif

	caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!caps) {
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	chan = ast_channel_alloc_with_endpoint(1, AST_STATE_DOWN,
#else
	chan = ast_channel_alloc_with_endpoint(1, state,
#endif
		S_COR(session->id.number.valid, session->id.number.str, ""),
		S_COR(session->id.name.valid, session->id.name.str, ""),
		session->endpoint->accountcode, "", "", assignedids, requestor, 0,
		session->endpoint->persistent, "PJSIP/%s-%08x",
		ast_sorcery_object_get_id(session->endpoint),
		(unsigned) ast_atomic_fetchadd_int((int *) &chan_idx, +1));
	if (!chan) {
		ao2_ref(caps, -1);
		return NULL;
	}

	ast_channel_tech_set(chan, &chan_pjsip_tech);
#ifdef GRANDSTREAM_NETWORKS
	if (!(channel = ast_sip_channel_pvt_alloc(pvt, session, req_caps_requestor))) {
#else
	if (!(channel = ast_sip_channel_pvt_alloc(pvt, session))) {
#endif
		ao2_ref(caps, -1);
		ast_channel_unlock(chan);
		ast_hangup(chan);
		return NULL;
	}

	ast_channel_stage_snapshot(chan);

#ifdef GRANDSTREAM_NETWORKS
	if (strcasecmp(session->endpoint->anonymous, "anonymous") == 0) {
		pbx_builtin_setvar_helper(chan, "__ANONYMOUS", "anonymous");
		ast_string_field_set(session->endpoint, anonymous, "");
	}
	ast_channel_cc_params_init(chan, session->endpoint->cc_params);
#endif

	ast_channel_tech_pvt_set(chan, channel);

	if (!ast_format_cap_count(session->req_caps) ||
		!ast_format_cap_iscompatible(session->req_caps, session->endpoint->media.codecs)) {
#ifdef GRANDSTREAM_NETWORKS
		/* There is no the same codec in session->req_caps and endpoint codecs. only need to add audio codecs because asterisk can translate it. */
		ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_AUDIO);
#else
		ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_UNKNOWN);
#endif
	} else {
#ifdef GRANDSTREAM_NETWORKS
		ast_format_cap_get_compatible(session->req_caps, session->endpoint->media.codecs, caps);
		ast_format_cap_append_from_cap(caps, session->req_caps, AST_MEDIA_TYPE_AUDIO);
#else
		ast_format_cap_append_from_cap(caps, session->req_caps, AST_MEDIA_TYPE_UNKNOWN);
#endif
	}

	ast_channel_nativeformats_set(chan, caps);
#ifdef GRANDSTREAM_NETWORKS
	{
		struct ast_str *format_buf = ast_str_alloca(64);
		ast_log(LOG_CODEC_NEG, " set channel nativeformats: %s\n", ast_format_cap_get_names(caps, &format_buf));
	}
#endif

	if (!ast_format_cap_empty(caps)) {
		struct ast_format *fmt;

		fmt = ast_format_cap_get_best_by_type(caps, AST_MEDIA_TYPE_AUDIO);
		if (!fmt) {
			/* Since our capabilities aren't empty, this will succeed */
			fmt = ast_format_cap_get_format(caps, 0);
		}
		ast_channel_set_writeformat(chan, fmt);
		ast_channel_set_rawwriteformat(chan, fmt);
		ast_channel_set_readformat(chan, fmt);
		ast_channel_set_rawreadformat(chan, fmt);
		ao2_ref(fmt, -1);
	}

	ao2_ref(caps, -1);

	if (state == AST_STATE_RING) {
		ast_channel_rings_set(chan, 1);
#ifdef GRANDSTREAM_NETWORKS
		ast_setstate(chan, AST_STATE_RING);
#endif
	}

	ast_channel_adsicpe_set(chan, AST_ADSI_UNAVAILABLE);

	ast_party_id_copy(&ast_channel_caller(chan)->id, &session->id);
	ast_party_id_copy(&ast_channel_caller(chan)->ani, &session->id);

	ast_channel_context_set(chan, session->endpoint->context);
	ast_channel_exten_set(chan, S_OR(exten, "s"));
	ast_channel_priority_set(chan, 1);

	ast_channel_callgroup_set(chan, session->endpoint->pickup.callgroup);
	ast_channel_pickupgroup_set(chan, session->endpoint->pickup.pickupgroup);

	ast_channel_named_callgroups_set(chan, session->endpoint->pickup.named_callgroups);
	ast_channel_named_pickupgroups_set(chan, session->endpoint->pickup.named_pickupgroups);

	if (!ast_strlen_zero(session->endpoint->language)) {
		ast_channel_language_set(chan, session->endpoint->language);
	}

	if (!ast_strlen_zero(session->endpoint->zone)) {
		struct ast_tone_zone *zone = ast_get_indication_zone(session->endpoint->zone);
		if (!zone) {
			ast_log(LOG_ERROR, "Unknown country code '%s' for tonezone. Check indications.conf for available country codes.\n", session->endpoint->zone);
		}
		ast_channel_zone_set(chan, zone);
	}

	for (var = session->endpoint->channel_vars; var; var = var->next) {
		char buf[512];
		pbx_builtin_setvar_helper(chan, var->name, ast_get_encoded_str(
						  var->value, buf, sizeof(buf)));
	}

	ast_channel_stage_snapshot_done(chan);
	ast_channel_unlock(chan);

	/* If res_pjsip_session is ever updated to create/destroy ast_sip_session_media
	 * during a call such as if multiple same-type stream support is introduced,
	 * these will need to be recaptured as well */
	pvt->media[SIP_MEDIA_AUDIO] = ao2_find(session->media, "audio", OBJ_KEY);
	pvt->media[SIP_MEDIA_VIDEO] = ao2_find(session->media, "video", OBJ_KEY);
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.13 */
	pvt->media[SIP_MEDIA_APPLICATION] = ao2_find(session->media, "application", OBJ_KEY);
#endif
	set_channel_on_rtp_instance(pvt, ast_channel_uniqueid(chan));

#ifdef GRANDSTREAM_NETWORKS
	sip_voip_trunk_enable_fax_detect(chan, session->wFaxDetect);
#endif
	return chan;
}

#ifdef GRANDSTREAM_NETWORKS
struct pjmedia_sdp_neg
{
	pjmedia_sdp_neg_state state;	    /**< Negotiator state.	     */
	pj_bool_t		  prefer_remote_codec_order;
	pj_bool_t             answer_with_multiple_codecs;
	pj_bool_t		  has_remote_answer;
	pj_bool_t		  answer_was_remote;

	pjmedia_sdp_session	*initial_sdp,	    /**< Initial local SDP	     */
			*active_local_sdp,  /**< Currently active local SDP. */
			*active_remote_sdp, /**< Currently active remote's.  */
			*neg_local_sdp,	    /**< Temporary local SDP.	     */
			*neg_remote_sdp;    /**< Temporary remote SDP.	     */
};

typedef struct pjmedia_sdp_neg pjmedia_sdp_neg;

static int append_isfocus_contact(pjsip_tx_data *packet)
{
    pjsip_contact_hdr *contact = NULL;
    static const pj_str_t isfocus_str = { "isfocus", 7 };
    static const pj_str_t blank_str = { "", 0 };
    pjsip_param *isfocus = NULL;

    // Sanity check
    if (packet == NULL) {
        ast_log(LOG_WARNING, "TX packet is NULL\n");
        return -1;
    }

    contact = pjsip_msg_find_hdr(packet->msg, PJSIP_H_CONTACT, NULL);
    if (contact == NULL) {
        ast_log(LOG_WARNING, "contact header is NULL\n");
        return -1;
    }

    isfocus = PJ_POOL_ALLOC_T(packet->pool, pjsip_param);
    if (isfocus == NULL) {
        ast_log(LOG_WARNING, "isfocus is NULL\n");
        return -1;
    }

    isfocus->name = isfocus_str;
    isfocus->value = blank_str;
    pj_list_push_back(&contact->other_param, isfocus);

    return 0;
}

static int set_contact_user(pjsip_tx_data *packet, char *user)
{
    pjsip_contact_hdr *contact = NULL;
    pjsip_sip_uri *contact_uri = NULL;

    // Sanity check
    if (packet == NULL) {
        ast_log(LOG_WARNING, "TX packet is NULL\n");
        return -1;
    }

    if (user == NULL) {
        ast_log(LOG_WARNING, "user is NULL\n");
        return -1;
    }

    contact = pjsip_msg_find_hdr(packet->msg, PJSIP_H_CONTACT, NULL);
    if (contact == NULL || contact->uri == NULL) {
        ast_log(LOG_WARNING, "contact URI header is NULL\n");
        return -1;
    }

    contact_uri = pjsip_uri_get_uri(contact->uri);
    if (contact_uri == NULL) {
        ast_log(LOG_WARNING, "Failed to get contact URI\n");
        return -1;
    }

    pj_strdup2(packet->pool, &contact_uri->user, user);

    return 0;
}

static int set_from_name(pjsip_tx_data *packet, const char *name)
{
    pjsip_name_addr *name_addr = NULL;

    // Sanity check
    if (packet == NULL) {
        ast_log(LOG_WARNING, "TX packet is NULL\n");
        return -1;
    }

    if (name == NULL) {
        ast_log(LOG_WARNING, "name is NULL\n");
        return -1;
    }

    name_addr = (pjsip_name_addr *)PJSIP_MSG_FROM_HDR(packet->msg)->uri;
    pj_strdup2(packet->pool, &name_addr->display, name);

    return 0;
}

/* Added by yfyang, Grandstream networks, Bug 46889 */
static void add_call_info_to_response_tmp(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	char uuid_str[AST_UUID_STR_LEN] = {0};
	char *user = NULL;
	pjsip_fromto_hdr *hdr;
	pjsip_sip_uri    *uri;
	pjsip_name_addr  *name_addr;
	pjsip_param *param_p;
	pjsip_param *param_m;
	pjsip_contact_hdr  *contact_hdr = (pjsip_contact_hdr*) pjsip_msg_find_hdr(tdata->msg,PJSIP_H_CONTACT, NULL);
	pjsip_uri *base;
	char cc_uri[PJSIP_MAX_URL_SIZE] = {0};
	pj_str_t callinfo_name = { "Call-Info", 9 };

	if (!session || !tdata || !contact_hdr) {
		return;
	}

	base = contact_hdr->uri;
	hdr = pjsip_from_hdr_create(tdata->pool); 
	hdr->type = PJSIP_H_OTHER;
	pj_strdup(tdata->pool, &hdr->name, &callinfo_name);
	hdr->sname = hdr->name;

	name_addr = pjsip_uri_clone(tdata->pool, base);
	uri = pjsip_uri_get_uri(name_addr->uri);
	user = ast_uuid_generate_str(uuid_str, sizeof(uuid_str));
	pj_strdup2(tdata->pool, &name_addr->display, "");
	pj_strdup2(tdata->pool, &uri->user, user);

	pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, pjsip_uri_clone(tdata->pool, (pjsip_uri *) name_addr), cc_uri, sizeof(cc_uri)); 
	ast_log(LOG_NOTICE, "Get subscribe_uri[%s]\n", cc_uri);

	hdr->uri = (pjsip_uri *) name_addr;

	pj_strdup2(session->inv_session->pool, &session->cc_subs_user, user);
	session->cc_subs_uri = pjsip_uri_clone(session->inv_session->pool, (pjsip_uri *) name_addr); 

	param_p = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
	param_p->name = pj_str("purpose");
	param_p->value = pj_str("call-completion");
	pj_list_insert_before(&hdr->other_param, param_p);

	param_m = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
	param_m->name = pj_str("m");
	param_m->value = pj_str("BS");
	pj_list_insert_before(&hdr->other_param, param_m);

	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr);
}
#endif

static int answer(void *data)
{
	pj_status_t status = PJ_SUCCESS;
	pjsip_tx_data *packet = NULL;
	struct ast_sip_session *session = data;

#ifdef GRANDSTREAM_NETWORKS
	RAII_VAR(struct ast_channel *, peer, ast_channel_bridge_peer(session->channel), ast_channel_cleanup);
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(session->channel);
	struct chan_pjsip_pvt *pvt = NULL;
	int pjsip_channel = 0;
#endif

	if (session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		return 0;
	}

#ifdef GRANDSTREAM_NETWORKS
    if (peer) {
		if (0 != strncmp(ast_channel_name(peer), "PJSIP/trunk_", 12)) {
			session->cc_offer = 0;
		}
    }

	pvt = channel->pvt;
	if (strcmp(ast_channel_tech(session->channel)->type, "PJSIP") == 0 && pvt) {
		pjsip_channel = 1;

		ast_log(LOG_DEBUG, "[TK] channel: %s, mcb_response: %s, isConference: %d\n",
							ast_channel_name(session->channel),
							(pvt->mcb_response != NULL) ? pvt->mcb_response : "NULL",
                            pvt->isConference);
	}
#endif

	pjsip_dlg_inc_lock(session->inv_session->dlg);
	if (session->inv_session->invite_tsx) {
#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for 183 neg, 2016.01.20 */
		if (session->has_early_media && session->inv_session->use_final_rsp_sdp) {
			session->has_early_media = 0;
			ast_log(LOG_DEBUG, "Has send 183 before, when send 200OK according callee answer\n");
			status = pjsip_inv_answer(session->inv_session, 200, NULL, session->inv_session->neg->neg_local_sdp, &packet);
		} else {
			status = pjsip_inv_answer(session->inv_session, 200, NULL, NULL, &packet);
		}
#else
		status = pjsip_inv_answer(session->inv_session, 200, NULL, NULL, &packet);
#endif

#ifdef GRANDSTREAM_NETWORKS
		if (status == PJ_SUCCESS) {
			if (session->inv_session->neg && session->inv_session->neg->active_local_sdp) {
				char buf[1024] = {0};
				pjmedia_sdp_session* caller_answer_sdp = session->inv_session->neg->active_local_sdp;
				pjmedia_sdp_print(caller_answer_sdp, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
				ast_log(LOG_CODEC_NEG, "[caller] A <<<=== UCM. send 200 ok sdp to the caller!\n%s", buf);
			}
		}
#endif
	} else {
		ast_log(LOG_ERROR,"Cannot answer '%s' because there is no associated SIP transaction\n",
			ast_channel_name(session->channel));
	}

#ifndef GRANDSTREAM_NETWORKS /*Moved it to below, rzhang*/
	pjsip_dlg_dec_lock(session->inv_session->dlg);
#endif

	if (status == PJ_SUCCESS && packet) {
#ifdef GRANDSTREAM_NETWORKS
		if (pjsip_channel) {
            if (pvt->mcb_response != NULL) {
                ast_sip_add_header(packet, "X-Grandstream-Mcb-Resp", pvt->mcb_response);
            }

            //not add isfocus if CEI notify disabled, modified by yfyang, for bug58890
            if (pvt->isConference == TRUE && globalEnabledCEI) {
                append_isfocus_contact(packet);
            }

            if (pvt->contact_user != NULL) {
                set_contact_user(packet, pvt->contact_user);
            } else if (session->contact_user != NULL) {
                set_contact_user(packet, session->contact_user);
            }

            if (pvt->call_type != NULL) {
                ast_sip_add_header(packet, "X-Grandstream-Call-Type", pvt->call_type);
            }
		}

		add_feature_map_header(packet, 0);

		/* Added by yfyang, Grandstream, Bug 46889 */
		if (!peer && (!strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) &&
			(PJSIP_INV_STATE_INCOMING == session->inv_session->state) && 
			(AST_CC_AGENT_NATIVE == ast_get_cc_agent_policy(session->endpoint->cc_params)) &&
			(strcmp(S_OR(pbx_builtin_getvar_helper(session->channel, "DIALSTATUS"), " "), "CHANUNAVAIL"))) {

			ast_log(LOG_NOTICE, "answer, add callinfo header!");
			if (!ast_cc_check_max_agents(session->channel)) {
				add_call_info_to_response_tmp(session, packet); 
			}
		}
#endif
                ast_log(LOG_NOTICE, "answer, sleep start!!!!!  tran %p tsx %p", session->inv_session, session->inv_session->invite_tsx);
                //sleep( 60 );
                //ast_log(LOG_NOTICE, "answer, sleep end  !!!!!  tran %p tsx %p", session->inv_session, session->inv_session->invite_tsx);

		ast_sip_session_send_response(session, packet);
	}
#ifdef GRANDSTREAM_NETWORKS
       /*Added by rzhang, for avoiding crash issue bug 47538*/
        pjsip_dlg_dec_lock(session->inv_session->dlg);

#endif
	return (status == PJ_SUCCESS) ? 0 : -1;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add the following interfaces to make further codec negotiation between the caller and the callee leg. */
static void str_swap(pj_str_t *str1, pj_str_t *str2)
{
	pj_str_t tmp = *str1;
	*str1 = *str2;
	*str2 = tmp;
}

static void print_sdp_media(const pjmedia_sdp_media* media)
{
	unsigned j;
	for (j=0; j<media->desc.fmt_count; ++j) {
		ast_log(LOG_CODEC_NEG, "media->desc.fmt[%d]: %ld\n", j, pj_strtoul(&media->desc.fmt[j]));
	}
}

static int get_pjmedia_sdp_attr_index (unsigned count,
						pjmedia_sdp_attr *const attr_array[],
						const pj_str_t *name,
						const pj_str_t *c_fmt)
{
	int i;
	unsigned c_pt = 0xFFFF;

	if (c_fmt)
	c_pt = pj_strtoul(c_fmt);

	for (i=0; i<count; ++i) {
		if (pj_strcmp(&attr_array[i]->name, name) == 0) {
			const pjmedia_sdp_attr *a = attr_array[i];
			if (c_fmt) {
				unsigned pt = (unsigned) pj_strtoul2(&a->value, NULL, 10);
				if (pt == c_pt) {
					return i;
				}
			} else
				return i;
		}
	}
	return -1;
}

static int index_of_pjmedia_sdp_attr(unsigned count,
						pjmedia_sdp_attr *const attr_array[],
						const char *c_name,
						const pj_str_t *c_fmt)
{
	pj_str_t name;

	name.ptr = (char*)c_name;
	name.slen = strlen(c_name);

	return get_pjmedia_sdp_attr_index(count, attr_array, &name, c_fmt);
}

static void arrange_pjmedia_sdp_attr_by_m_line(pjmedia_sdp_media *media)
{
	pjmedia_sdp_attr   *attr;
	unsigned j;
	int attr_index = 0;
	int index;

	for (j=0; j<media->desc.fmt_count; j++) {
		index = index_of_pjmedia_sdp_attr(media->attr_count, media->attr, "rtpmap", &media->desc.fmt[j]);
		if (index > -1) {
			if (index != attr_index)
			{
				attr = media->attr[attr_index];
				media->attr[attr_index] = media->attr[index];
				media->attr[index] = attr;
			}
			attr_index++;
		}

		index = index_of_pjmedia_sdp_attr(media->attr_count, media->attr, "fmtp", &media->desc.fmt[j]);
		if (index > -1) {
			if (index != attr_index)
			{
				attr = media->attr[attr_index];
				media->attr[attr_index] = media->attr[index];
				media->attr[index] = attr;
			}
			attr_index++;
		}
	}
}

static int match_callee_answer_app_media_bfcp(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	char tmp[512];
	pj_str_t stmp;
	pjmedia_sdp_attr *attr_old;
	pjmedia_sdp_attr *attr_new;

	ast_log(LOG_DEBUG, "match callee bfcp answer begin\n");

	if (0 == peer_session->bfcp_para.has_bfcp && 0 == peer_session->bfcp_para.has_tcp_bfcp) {
		ast_log(LOG_DEBUG, "match callee bfcp answer, peer bfcp is decline, local should decline too\n");
		*p_result = pjmedia_sdp_media_clone_deactivate(pool, local);
		return 0;
	}

	if (local_session->bfcp_para.confid != peer_session->bfcp_para.confid) {
		ast_log(LOG_DEBUG, "match callee bfcp answer, confid is not the same, need change\n");
		attr_old = pjmedia_sdp_media_find_attr2(local, "confid", NULL);
		if (attr_old) {
			pjmedia_sdp_media_remove_attr(local, attr_old);
		}
		if (peer_session->bfcp_para.confid) {
			snprintf(tmp, sizeof(tmp), "%d", peer_session->bfcp_para.confid);
			attr_new = pjmedia_sdp_attr_create(pool, "confid", pj_cstr(&stmp, tmp));
			pjmedia_sdp_media_add_attr(local, attr_new);
		} else {
			ast_log(LOG_DEBUG, "bfcp confid is zero, don't add confid attr\n");
		}
	}

	if (local_session->bfcp_para.userid != peer_session->bfcp_para.userid) {
		ast_log(LOG_DEBUG, "match callee bfcp answer, userid is not the same, need change\n");
		attr_old = pjmedia_sdp_media_find_attr2(local, "userid", NULL);
		if (attr_old) {
			pjmedia_sdp_media_remove_attr(local, attr_old);
		}
		if (peer_session->bfcp_para.userid) {
			snprintf(tmp, sizeof(tmp), "%d", peer_session->bfcp_para.userid);
			attr_new = pjmedia_sdp_attr_create(pool, "userid", pj_cstr(&stmp, tmp));
			pjmedia_sdp_media_add_attr(local, attr_new);
		} else {
			ast_log(LOG_DEBUG, "bfcp userid is zero, don't add userid attr\n");
		}
	}

	if (strcmp(local_session->bfcp_para.floorid , peer_session->bfcp_para.floorid)) {
		ast_log(LOG_DEBUG, "match callee bfcp answer, floorid is not the same, need change\n");
		attr_old = pjmedia_sdp_media_find_attr2(local, "floorid", NULL);
		if (attr_old) {
			pjmedia_sdp_media_remove_attr(local, attr_old);
		}
		if (strlen(peer_session->bfcp_para.floorid)) {
			snprintf(tmp, sizeof(tmp), "%s", peer_session->bfcp_para.floorid);
			attr_new = pjmedia_sdp_attr_create(pool, "floorid", pj_cstr(&stmp, tmp));
			pjmedia_sdp_media_add_attr(local, attr_new);
		} else {
			ast_log(LOG_DEBUG, "bfcp floorid is zero, don't add floorid attr\n");
		}
	}

	if (strcmp(local_session->bfcp_para.floorctrl , peer_session->bfcp_para.floorctrl)) {
		ast_log(LOG_DEBUG, "match callee bfcp answer, floorctrl is not the same, need change\n");
		attr_old = pjmedia_sdp_media_find_attr2(local, "floorctrl", NULL);
		if (attr_old) {
			pjmedia_sdp_media_remove_attr(local, attr_old);
		}
		if (strlen(peer_session->bfcp_para.floorctrl)) {
			snprintf(tmp, sizeof(tmp), "%s", peer_session->bfcp_para.floorctrl);
			attr_new = pjmedia_sdp_attr_create(pool, "floorctrl", pj_cstr(&stmp, tmp));
			pjmedia_sdp_media_add_attr(local, attr_new);
		} else {
			ast_log(LOG_DEBUG, "bfcp floorctrl is zero, don't add floorctrl attr\n");
		}
	}

	if (strcmp(local_session->bfcp_para.setup , peer_session->bfcp_para.setup)) {
		ast_log(LOG_DEBUG, "match callee bfcp answer, setup is not the same, need change\n");
		attr_old = pjmedia_sdp_media_find_attr2(local, "setup", NULL);
		if (attr_old) {
			pjmedia_sdp_media_remove_attr(local, attr_old);
		}
		if (strlen(peer_session->bfcp_para.setup)) {
			snprintf(tmp, sizeof(tmp), "%s", peer_session->bfcp_para.setup);
			attr_new = pjmedia_sdp_attr_create(pool, "setup", pj_cstr(&stmp, tmp));
			pjmedia_sdp_media_add_attr(local, attr_new);
		} else {
			ast_log(LOG_DEBUG, "bfcp setup is zero, don't add setup attr\n");
		}
	}

	if (strcmp(local_session->bfcp_para.connection , peer_session->bfcp_para.connection)) {
		ast_log(LOG_DEBUG, "match callee bfcp answer, connection is not the same, need change\n");
		attr_old = pjmedia_sdp_media_find_attr2(local, "connection", NULL);
		if (attr_old) {
			pjmedia_sdp_media_remove_attr(local, attr_old);
		}
		if (strlen(peer_session->bfcp_para.connection)) {
			snprintf(tmp, sizeof(tmp), "%s", peer_session->bfcp_para.connection);
			attr_new = pjmedia_sdp_attr_create(pool, "connection", pj_cstr(&stmp, tmp));
			pjmedia_sdp_media_add_attr(local, attr_new);
		} else {
			ast_log(LOG_DEBUG, "bfcp connection is zero, don't add connection attr\n");
		}
	}

	if (strcmp(local_session->bfcp_para.fingerprint , peer_session->bfcp_para.fingerprint)) {
		ast_log(LOG_DEBUG, "match callee bfcp answer, fingerprint is not the same, need change\n");
		attr_old = pjmedia_sdp_media_find_attr2(local, "fingerprint", NULL);
		if (attr_old) {
			pjmedia_sdp_media_remove_attr(local, attr_old);
		}
		if (strlen(peer_session->bfcp_para.fingerprint)) {
			snprintf(tmp, sizeof(tmp), "%s", peer_session->bfcp_para.fingerprint);
			attr_new = pjmedia_sdp_attr_create(pool, "fingerprint", pj_cstr(&stmp, tmp));
			pjmedia_sdp_media_add_attr(local, attr_new);
		} else {
			ast_log(LOG_DEBUG, "bfcp fingerprint is zero, don't add fingerprint attr\n");
		}
	}

	*p_result = pjmedia_sdp_media_clone(pool, local);
	return 0;
}

static int match_callee_answer_app_media_fec(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	/* now we only need to process peer disable fec, if change fmtp param, support later */
	ast_log(LOG_DEBUG, "match_callee_answer_app_media_fec begin\n");
	if (0 == callee_answer->desc.port) {
		*p_result = pjmedia_sdp_media_clone_deactivate(pool, local);
		return 0;
	}

	if (0 == local->desc.port) {
		*p_result = pjmedia_sdp_media_clone(pool, local);
		return 0;
	}

	*p_result = pjmedia_sdp_media_clone(pool, local);
	return 0;
}

static int match_callee_answer_app_media_fecc(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	/* now we only need to process peer disable fecc, if change fmtp param, support later */
	ast_log(LOG_DEBUG, "match_callee_answer_app_media_fecc begin\n");
	if (0 == callee_answer->desc.port) {
		*p_result = pjmedia_sdp_media_clone_deactivate(pool, local);
		return 0;
	}

	if (0 == local->desc.port) {
		*p_result = pjmedia_sdp_media_clone(pool, local);
		return 0;
	}

	*p_result = pjmedia_sdp_media_clone(pool, local);
	return 0;
}

static int match_callee_answer_app_media_rtp(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	int res = 0;
	ast_log(LOG_DEBUG, "match callee answer rtp application media begin\n");

	if (ast_sip_session_judge_if_fec_stream(peer_session, callee_answer) &&
		ast_sip_session_judge_if_fec_stream(local_session, local)) {
		ast_log(LOG_DEBUG, "It's private FEC for grandstream, match as FEC\n");
		res = match_callee_answer_app_media_fec(pool, local, callee_answer, local_session, peer_session, p_result);
		return res;
	}

	if (ast_sip_session_judge_if_fecc_stream(peer_session, callee_answer) &&
		ast_sip_session_judge_if_fecc_stream(local_session, local)) {
		ast_log(LOG_DEBUG, "It's FECC, match as FECC\n");
		res = match_callee_answer_app_media_fecc(pool, local, callee_answer, local_session, peer_session, p_result);
		return res;
	}

	/* If callee answer decline media without attrs, modify_local_sdp_with_callee_answer will deactivate media when return -1,*/
	ast_log(LOG_DEBUG, "Can't match FEC/FECC media\n");
	return -1;
}

static int match_callee_answer_app_media(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	int res = -1;
	enum ast_sip_application_type callee_app_type = ast_sip_session_get_application_type(callee_answer->desc.transport);
	enum ast_sip_application_type caller_app_type = ast_sip_session_get_application_type(local->desc.transport);
	if (callee_app_type != caller_app_type) {
		ast_log(LOG_DEBUG, "match_callee_answer_app_media transport type is not the same\n");
		return res;
	}
	switch(callee_app_type) {
		case AST_SIP_APP_UDP_BFCP:
		case AST_SIP_APP_TCP_BFCP:
			res = match_callee_answer_app_media_bfcp(pool, local, callee_answer, local_session, peer_session, p_result);
			return res;
		case AST_SIP_APP_RTP:
			res = match_callee_answer_app_media_rtp(pool, local, callee_answer, local_session, peer_session, p_result);
			return res;
		default:
			ast_log(LOG_DEBUG, "Can't match other application media, except bfcp/fec/fecc");
	}
	return res;
}

static int match_callee_answer_extra_video_attr(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	char tmp[512];
	pj_str_t stmp;
	pjmedia_sdp_attr *attr_old;
	pjmedia_sdp_attr *attr_new;

	ast_log(LOG_DEBUG, "match callee extra video attr answer begin\n");

	if (FALSE == ast_sip_session_judge_if_slides_video(local) &&
		FALSE == ast_sip_session_judge_if_slides_video(callee_answer)) {
		/* if caller's framerate or label is different between callee's, need to change, but content can't change */
		if (local_session->main_video_para.framerate != peer_session->main_video_para.framerate) {
			ast_log(LOG_DEBUG, "match callee extra main video attr answer, framerate is not the same, need change\n");
			attr_old = pjmedia_sdp_media_find_attr2(*p_result, "framerate", NULL);
			if (attr_old) {
				pjmedia_sdp_media_remove_attr(*p_result, attr_old);
			}
			if (peer_session->main_video_para.framerate) {
				snprintf(tmp, sizeof(tmp), "%d", peer_session->main_video_para.framerate);
				attr_new = pjmedia_sdp_attr_create(pool, "framerate", pj_cstr(&stmp, tmp));
				pjmedia_sdp_media_add_attr(*p_result, attr_new);
			} else {
				ast_log(LOG_DEBUG, "main video framerate is zero, don't add framerate attr\n");
			}
		}
		if (local_session->main_video_para.label != peer_session->main_video_para.label) {
			ast_log(LOG_DEBUG, "match callee extra main video attr answer, label is not the same, need change\n");
			attr_old = pjmedia_sdp_media_find_attr2(*p_result, "label", NULL);
			if (attr_old) {
				pjmedia_sdp_media_remove_attr(*p_result, attr_old);
			}
			if (peer_session->main_video_para.label) {
				snprintf(tmp, sizeof(tmp), "%d", peer_session->main_video_para.label);
				attr_new = pjmedia_sdp_attr_create(pool, "label", pj_cstr(&stmp, tmp));
				pjmedia_sdp_media_add_attr(*p_result, attr_new);
			} else {
				ast_log(LOG_DEBUG, "main video label is zero, don't add label attr\n");
			}
		}
	}

	if (TRUE == ast_sip_session_judge_if_slides_video(local) &&
		TRUE == ast_sip_session_judge_if_slides_video(callee_answer)) {
		if (local_session->slides_video_para.framerate != peer_session->slides_video_para.framerate) {
			ast_log(LOG_DEBUG, "match callee extra slides video attr answer, framerate is not the same, need change\n");
			attr_old = pjmedia_sdp_media_find_attr2(*p_result, "framerate", NULL);
			if (attr_old) {
				pjmedia_sdp_media_remove_attr(*p_result, attr_old);
			}
			if (peer_session->slides_video_para.framerate) {
				snprintf(tmp, sizeof(tmp), "%d", peer_session->slides_video_para.framerate);
				attr_new = pjmedia_sdp_attr_create(pool, "framerate", pj_cstr(&stmp, tmp));
				pjmedia_sdp_media_add_attr(*p_result, attr_new);
			} else {
				ast_log(LOG_DEBUG, "slides video framerate is zero, don't add framerate attr\n");
			}
		}
		if (local_session->slides_video_para.label != peer_session->slides_video_para.label) {
			ast_log(LOG_DEBUG, "match callee extra slides video attr answer, label is not the same, need change\n");
			attr_old = pjmedia_sdp_media_find_attr2(*p_result, "label", NULL);
			if (attr_old) {
				pjmedia_sdp_media_remove_attr(*p_result, attr_old);
			}
			if (peer_session->slides_video_para.label) {
				snprintf(tmp, sizeof(tmp), "%d", peer_session->slides_video_para.label);
				attr_new = pjmedia_sdp_attr_create(pool, "label", pj_cstr(&stmp, tmp));
				pjmedia_sdp_media_add_attr(*p_result, attr_new);
			} else {
				ast_log(LOG_DEBUG, "slides video label is zero, don't add label attr\n");
			}
		}
	}

	return 0;
}

/* video payload and attrs need to modify result according to callee answer */
static void match_video_payload_callee_answer(pjmedia_sdp_media *result, const pjmedia_sdp_media *callee_answer,
	int pt_answer_count, pj_str_t pt_local[PJMEDIA_MAX_SDP_FMT])
{
	int res_fmt_count = 0;
	int local_fmt_count = 0;
	int fmt_found_in_callee = FALSE;

	if (!strncmp(result->desc.media.ptr, "video", 5) &&
		!strncmp(callee_answer->desc.media.ptr, "video", 5)) {
		ast_log(LOG_DEBUG, "It's video, just use callee answer result\n");
		/* to remove rtpmap/fmtp for payload can't find in callee answer */
		for (local_fmt_count = 0; local_fmt_count < result->desc.fmt_count; ++local_fmt_count) {
			for(res_fmt_count = 0; res_fmt_count < pt_answer_count; ++res_fmt_count) {
				if (pj_strtoul(&pt_local[res_fmt_count]) == pj_strtoul(&result->desc.fmt[local_fmt_count])) {
					fmt_found_in_callee = TRUE;
					break;
				}
			}
			if (fmt_found_in_callee) {
				fmt_found_in_callee = FALSE;
			} else {
				ast_log(LOG_DEBUG, "Video media can't find local fmt:%ld in callee answer, just remove this attr\n", pj_strtoul(&result->desc.fmt[local_fmt_count]));
				pjmedia_sdp_attr *a_rtpmap;
				a_rtpmap = pjmedia_sdp_media_find_attr2(result, "rtpmap", &result->desc.fmt[local_fmt_count]);
				if (NULL != a_rtpmap) {
					pjmedia_sdp_media_remove_attr(result, a_rtpmap);
				}

				pjmedia_sdp_attr *a_fmtp;
				a_fmtp = pjmedia_sdp_media_find_attr2(result, "fmtp", &result->desc.fmt[local_fmt_count]);
				if (NULL != a_fmtp) {
					pjmedia_sdp_media_remove_attr(result, a_fmtp);
				}
			}
		}

		/* to remove payload in m line */
		result->desc.fmt_count = pt_answer_count;
		for (res_fmt_count = 0; res_fmt_count < pt_answer_count; ++res_fmt_count) {
			result->desc.fmt[res_fmt_count] = pt_local[res_fmt_count];
		}
	}

	return;
}
/* End add by bfdu for BFCP/FEC/FECC, 2015.07.09 */

static int match_callee_answer(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					pjmedia_sdp_media **p_result)
{
	unsigned i;
	unsigned pt_answer_count = 0;
	pj_str_t pt_local[PJMEDIA_MAX_SDP_FMT];
	pjmedia_sdp_media *result;
#ifdef GRANDSTREAM_NETWORKS
	pj_str_t DTMF_STR_LOCAL = pj_str(NULL);
#endif

	if (0 == callee_answer->desc.port) {
		*p_result = pjmedia_sdp_media_clone_deactivate(pool, local);
		return 0;
	}

	if (0 == local->desc.port) {
		*p_result = pjmedia_sdp_media_clone(pool, local);
		return 0;
	}

	for (i=0; i<callee_answer->desc.fmt_count; ++i) {
		unsigned j;

		if (pj_isdigit(*callee_answer->desc.fmt[i].ptr)) {
			/* This is normal/standard payload type, where it's identified
			 * by payload number.
			 */
			unsigned pt;

			pt = pj_strtoul(&callee_answer->desc.fmt[i]);

			if (pt < 96) {
				/* For static payload type, it's enough to compare just
				 * the payload number.
				 */

				/* Find matching codec in local descriptor. */
				for (j=0; j<local->desc.fmt_count; ++j) {
					unsigned p;
					p = pj_strtoul(&local->desc.fmt[j]);
					if (p == pt && pj_isdigit(*local->desc.fmt[j].ptr)) {
						pt_local[pt_answer_count++] = local->desc.fmt[j];
						break;
					}
				}

			} else {
				/* This is dynamic payload type.
				 * For dynamic payload type, we must look the rtpmap and
				 * compare the encoding name.
				 */
				const pjmedia_sdp_attr *a;
				pjmedia_sdp_rtpmap or_;
				//pj_bool_t is_codec;

				/* Get the rtpmap for the payload type. */
				a = pjmedia_sdp_media_find_attr2(callee_answer, "rtpmap",
								 &callee_answer->desc.fmt[i]);
				if (!a) {
					pj_assert(!"Bug! Offer should have been validated");
					return -1;
				}
				pjmedia_sdp_attr_get_rtpmap(a, &or_);

				/* Delete by bfdu for Bug 43126, DTMF is dynamic payload too */
				/*if (!pj_stricmp2(&or_.enc_name, "telephone-event")) {
					//is_codec = 0;
					continue;
				} else {
					//is_codec = 1;
				}*/

				/* Find paylaod in our initial SDP with matching
				 * encoding name and clock rate.
				 */
				for (j=0; j<local->desc.fmt_count; ++j) {
				    a = pjmedia_sdp_media_find_attr2(local, "rtpmap",
								     &local->desc.fmt[j]);
				    if (a) {
						pjmedia_sdp_rtpmap lr;
						pjmedia_sdp_attr_get_rtpmap(a, &lr);

#ifdef GRANDSTREAM_NETWORKS
						if (!pj_stricmp2(&lr.enc_name, "telephone-event")) {
						    DTMF_STR_LOCAL = local->desc.fmt[j];
						}
#endif

						/* See if encoding name, clock rate, and
						 * channel count  match
						 */
						if (!pj_stricmp(&or_.enc_name, &lr.enc_name) &&
						    or_.clock_rate == lr.clock_rate &&
						    (pj_stricmp(&or_.param, &lr.param)==0 ||
						     (lr.param.slen==0 && or_.param.slen==1 &&
									 *or_.param.ptr=='1') ||
						     (or_.param.slen==0 && lr.param.slen==1 &&
									  *lr.param.ptr=='1')))
						{
							/* need to update fmtp attributes to pass through some video attributes. */
							if (!strncmp(local->desc.media.ptr, "video", 5)) {
								pjmedia_sdp_attr *a_o_fmtp, *a_a_fmtp, *new_a_o_fmtp;

								a_o_fmtp = pjmedia_sdp_media_find_attr2(local, "fmtp", &local->desc.fmt[j]);
								a_a_fmtp = pjmedia_sdp_media_find_attr2(callee_answer, "fmtp", &callee_answer->desc.fmt[i]);

								if (a_o_fmtp && a_a_fmtp) {
									pjmedia_sdp_media_remove_attr(local, a_o_fmtp);

									new_a_o_fmtp = pjmedia_sdp_attr_change_fmtp_payload(pool, a_a_fmtp, &callee_answer->desc.fmt[i], &local->desc.fmt[j]);

									if (new_a_o_fmtp) {
										pjmedia_sdp_media_add_attr(local, new_a_o_fmtp);
									}
								}
							}
						    /* Match! */
							pt_local[pt_answer_count++] = local->desc.fmt[j];
							break;
						}
				    }
				}
			}

		} else {
		/* This is a non-standard, brain damaged SDP where the payload
		 * type is non-numeric. It exists e.g. in Microsoft RTC based
		 * UA, to indicate instant messaging capability.
		 * Example:
		 *	- m=x-ms-message 5060 sip null
		 */

			for (j=0; j<local->desc.fmt_count; ++j) {
				if (!pj_strcmp(&callee_answer->desc.fmt[i], &local->desc.fmt[j])) {
					/* Match */
					pt_local[pt_answer_count++] = local->desc.fmt[j];
					break;
				}
			}
		}
	}

	//const pj_str_t DTMF_STR = {"101", 3};
	/* rearrange the payload to suit the offer. */
	result = pjmedia_sdp_media_clone(pool, local);
	ast_log(LOG_CODEC_NEG, "before swap, result is:\n");
	print_sdp_media(result);
	for (i=0; i<pt_answer_count; ++i) {
		unsigned j;
		for (j=i; j<result->desc.fmt_count; ++j) {
			if (!pj_strcmp(&result->desc.fmt[j], &pt_local[i]))
				break;
		}
		pj_assert(j != result->desc.fmt_count);
		/* add by chdruan for Bug 45198*/
		//if (!pj_strcmp(&pt_local[i], &DTMF_STR))
			//continue;
		/* end by chdruan for Bug 45198*/
		ast_log(LOG_CODEC_NEG, "pt_local[%d]: %ld, swap format position: (result->desc.fmt[%d])%ld <==> (result->desc.fmt[%d])%ld\n",
			i, pj_strtoul(&pt_local[i]), i, pj_strtoul(&result->desc.fmt[i]), j, pj_strtoul(&result->desc.fmt[j]));
		str_swap(&result->desc.fmt[i], &result->desc.fmt[j]);
	}
#ifdef GRANDSTREAM_NETWORKS
	for (i = 0; i < result->desc.fmt_count; ++i) {
		if ( !pj_strcmp(&result->desc.fmt[i], &DTMF_STR_LOCAL)) {
			str_swap(&result->desc.fmt[i], &result->desc.fmt[result->desc.fmt_count - 1]);
		}
	}
#else
	for (i = 0; i < result->desc.fmt_count; ++i) {
		if ( !pj_strcmp(&result->desc.fmt[i], &DTMF_STR)) {
			str_swap(&result->desc.fmt[i], &result->desc.fmt[result->desc.fmt_count - 1]);
		}
	}
#endif
	/* Begin add by bfdu for FEC, 2015.11.24 */
	match_video_payload_callee_answer(result, callee_answer, pt_answer_count, pt_local);
	/* End add by bfdu for FEC, 2015.11.24 */
	arrange_pjmedia_sdp_attr_by_m_line(result);
	ast_log(LOG_CODEC_NEG, "after swap, result is:\n");
	print_sdp_media(result);

	*p_result = result;
	return 0;
}

/* Begin add by bfdu for Bug 43278, 2015.05.20 */
/*! \brief figure out if media stream has crypto lines for sdes */
static int media_stream_has_crypto(const struct pjmedia_sdp_media *stream)
{
	int i;

	for (i = 0; i < stream->attr_count; i++) {
		pjmedia_sdp_attr *attr;

		/* check the stream for the required crypto attribute */
		attr = stream->attr[i];
		if (pj_strcmp2(&attr->name, "crypto")) {
			continue;
		}

		return 1;
	}

	return 0;
}

/*! \brief figure out media transport encryption type from the media transport string */
static enum ast_sip_session_media_encryption get_media_encryption_type(pj_str_t transport,
	const struct pjmedia_sdp_media *stream, unsigned int *optimistic)
{
	RAII_VAR(char *, transport_str, ast_strndup(transport.ptr, transport.slen), ast_free);

	*optimistic = 0;

	if (strstr(transport_str, "UDP/TLS")) {
		return AST_SIP_MEDIA_ENCRYPT_DTLS;
	} else if (strstr(transport_str, "SAVP")) {
		return AST_SIP_MEDIA_ENCRYPT_SDES;
	} else if (media_stream_has_crypto(stream)) {
		*optimistic = 1;
		return AST_SIP_MEDIA_ENCRYPT_SDES;
	} else {
		return AST_SIP_MEDIA_ENCRYPT_NONE;
	}
}

static enum ast_sip_session_media_encryption check_endpoint_media_transport(
	struct ast_sip_endpoint *endpoint,
	const struct pjmedia_sdp_media *stream)
{
	enum ast_sip_session_media_encryption incoming_encryption;
	char transport_end = stream->desc.transport.ptr[stream->desc.transport.slen - 1];
	unsigned int optimistic;

	if ((transport_end == 'F' && !endpoint->media.rtp.use_avpf)
		|| (transport_end != 'F' && endpoint->media.rtp.use_avpf)) {
		return AST_SIP_MEDIA_TRANSPORT_INVALID;
	}

	incoming_encryption = get_media_encryption_type(stream->desc.transport, stream, &optimistic);

	if (incoming_encryption == endpoint->media.rtp.encryption) {
		return incoming_encryption;
	}

#ifdef GRANDSTREAM_NETWORKS
	if ((endpoint->media.rtp.encryption == AST_SIP_MEDIA_AUTO_NO_ENCRYPT)
		|| (endpoint->media.rtp.encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP)
		|| (endpoint->media.rtp.encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS)) {
		return incoming_encryption;
	}
#endif

	if (endpoint->media.rtp.force_avp ||
		endpoint->media.rtp.encryption_optimistic) {
		return incoming_encryption;
	}

	/* If an optimistic offer has been made but encryption is not enabled consider it as having
	 * no offer of crypto at all instead of invalid so the session proceeds.
	 */
	if (optimistic) {
		return AST_SIP_MEDIA_ENCRYPT_NONE;
	}

	return AST_SIP_MEDIA_TRANSPORT_INVALID;
}

static int judge_if_transport_compatible(const struct ast_sip_session *session, const pjmedia_sdp_media *stream)
{
	if (NULL == session || NULL == stream) {
		ast_log(LOG_WARNING, "judge_if_transport_compatible para is null\n");
		return 0;
	}
	/* FEC stream, should not use our config, just use received encryption type */
	if (ast_sip_session_judge_if_fec_stream(session, stream)) {
		return 1;
	}
	enum ast_sip_application_type app_type = ast_sip_session_get_application_type(stream->desc.transport);
	enum ast_sip_session_media_encryption encryption = AST_SIP_MEDIA_ENCRYPT_NONE;
	if (!session->endpoint->media.rtp.use_received_transport && 
		AST_SIP_APP_UDP_BFCP != app_type && AST_SIP_APP_TCP_BFCP != app_type && 
		AST_SIP_APP_RTP != app_type) {
		encryption = check_endpoint_media_transport(session->endpoint, stream);

		if (encryption == AST_SIP_MEDIA_TRANSPORT_INVALID) {
			ast_log(LOG_WARNING, "transport is not compatible\n");
			return 0;
		}
	}
	return 1;
}
/* End add by bfdu for Bug 43278, 2015.05.20 */

/* Add by bfdu for reinvite when open direct media, 2015.09.16 */
static void reset_caller_media_addr(const struct ast_sip_session *callee_session, const struct ast_sip_session *caller_session, 
	const pjmedia_sdp_media *callee_media, pjmedia_sdp_media *caller_media)
{
	ast_log(LOG_DEBUG, "reset_caller_media_addr begin\n");

	if (!callee_session || !callee_session->endpoint || !callee_session->inv_session || !callee_session->inv_session->pool_prov) {
		ast_log(LOG_ERROR, "reset_caller_media_addr callee session or endpoint is not exist\n");
		return;
	}
	if (!caller_session || !caller_session->endpoint || !caller_session->inv_session || !caller_session->inv_session->pool_prov) {
		ast_log(LOG_ERROR, "reset_caller_media_addr caller session or endpoint is not exist\n");
		return;
	}

	if(!callee_media || !caller_media) {
		ast_log(LOG_ERROR, "reset_caller_media_addr callee_media or caller_media is not exist\n");
		return;
	}

	if (callee_session->endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		ast_log(LOG_DEBUG, "reset_caller_media_addr callee endpoint has encryption\n");
		return;
	}
	if (caller_session->endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		ast_log(LOG_DEBUG, "reset_caller_media_addr caller endpoint has encryption\n");
		return;
	}
	if (!callee_session->endpoint->media.direct_media.enabled) {
		ast_log(LOG_DEBUG, "reset_caller_media_addr callee direct media is not enabled\n");
		return;
	}
	if (!caller_session->endpoint->media.direct_media.enabled) {
		ast_log(LOG_DEBUG, "reset_caller_media_addr caller direct media is not enabled\n");
		return;
	}

	pj_pool_t *pool = caller_session->inv_session->pool_prov;
	if (callee_media->conn) {
		caller_media->conn = pjmedia_sdp_conn_clone(pool, callee_media->conn);
	} else {
		ast_log(LOG_DEBUG, "callee media conn is null, set caller null\n");
		caller_media->conn = NULL;
	}
	caller_media->desc.port = callee_media->desc.port;

	return;
}

static void reset_caller_media_sdp_session_addr(const struct ast_sip_session *callee_session, const struct ast_sip_session *caller_session, 
	pjmedia_sdp_session *caller_local, const pjmedia_sdp_session *callee_answer)
{
	ast_log(LOG_DEBUG, "reset_caller_media_sdp_session_addr begin\n");

	if (!callee_session || !callee_session->endpoint || !callee_session->inv_session || !callee_session->inv_session->pool_prov) {
		ast_log(LOG_ERROR, "reset_caller_media_sdp_session_addr callee session or endpoint is not exist\n");
		return;
	}
	if (!caller_session || !caller_session->endpoint || !caller_session->inv_session || !caller_session->inv_session->pool_prov) {
		ast_log(LOG_ERROR, "reset_caller_media_sdp_session_addr caller session or endpoint is not exist\n");
		return;
	}

	if(!caller_local || !callee_answer) {
		ast_log(LOG_ERROR, "reset_caller_media_sdp_session_addr callee_media or caller_media is not exist\n");
		return;
	}

	if (callee_session->endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		ast_log(LOG_DEBUG, "reset_caller_media_sdp_session_addr callee endpoint has encryption\n");
		return;
	}
	if (caller_session->endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		ast_log(LOG_DEBUG, "reset_caller_media_sdp_session_addr caller endpoint has encryption\n");
		return;
	}
	if (!callee_session->endpoint->media.direct_media.enabled) {
		ast_log(LOG_DEBUG, "reset_caller_media_sdp_session_addr callee direct media is not enabled\n");
		return;
	}
	if (!caller_session->endpoint->media.direct_media.enabled) {
		ast_log(LOG_DEBUG, "reset_caller_media_sdp_session_addr caller direct media is not enabled\n");
		return;
	}

	pj_pool_t *pool = caller_session->inv_session->pool_prov;
	if (callee_answer->conn) {
		caller_local->conn = pjmedia_sdp_conn_clone(pool, callee_answer->conn);
	} else {
		ast_log(LOG_ERROR, "callee media conn is null\n");
	}

	return;
}


static pjmedia_sdp_session* modify_local_sdp_with_callee_answer( pj_pool_t *pool,
				  pjmedia_sdp_session *caller_local,
				  const pjmedia_sdp_session *callee_answer,
				  const struct ast_sip_session *local_session,
				  const struct ast_sip_session *peer_session
				  )
{
	pjmedia_sdp_session *result;
	char media_used[PJMEDIA_MAX_SDP_MEDIA];
	unsigned found = 0;
	unsigned i, j;
	pjmedia_sdp_media *im;
	pjmedia_sdp_session *callee_active_remote_sdp;
	callee_active_remote_sdp = peer_session->inv_session->neg->active_remote_sdp;

	result = pjmedia_sdp_session_clone(pool, caller_local);

	result->media_count = 0;

	pj_bzero(media_used, sizeof(media_used));
	for (im=NULL, j=0; j<caller_local->media_count; ++j) {
		const pjmedia_sdp_media *om;
		pjmedia_sdp_media *am = NULL;
		const pjmedia_sdp_media *active_remote_media = NULL;

		im = caller_local->media[j];
		found = 0;

		for (i=0; i<callee_answer->media_count; ++i) {
			om = callee_answer->media[i];

			if (pj_strcmp(&om->desc.media, &im->desc.media)==0 &&
			judge_if_transport_compatible(local_session, im) &&
			judge_if_transport_compatible(peer_session, om) &&
			media_used[i] == 0)
			{
				pj_status_t status;
				/* Begin modify by bfdu for BFCP, 2015.07.13 */
				if (!strncmp(om->desc.media.ptr, "application", 11)) {
					ast_log(LOG_CODEC_NEG, "match callee answer, it's application media\n");
					status = match_callee_answer_app_media(pool, im, om, local_session, peer_session, &am);
				} else {
					/* See if it has matching codec. */
					status = match_callee_answer(pool, im, om, &am);
					if (!strncmp(om->desc.media.ptr, "video", 5)) {
						status = match_callee_answer_extra_video_attr(pool, im, om, local_session, peer_session, &am);
						if (local_session->is_yealink_msg || peer_session->is_yealink_msg) {
							ast_sip_session_modify_outgoing_yealink_sdp(local_session, pool, am);
						}
					}
				}
				/* End modify by bfdu for BFCP, 2015.07.13 */
				if (status == 0) {
					media_used[i] = 1;
					found = 1;
					/* Add by bfdu for reinvite when open direct media, 2015.09.16 */
					if (local_session->is_video_media_change) {
						if (callee_active_remote_sdp && (i < callee_active_remote_sdp->media_count)) {
							active_remote_media = callee_active_remote_sdp->media[i];
						}
						if (active_remote_media && pj_strcmp(&om->desc.media, &active_remote_media->desc.media)==0) {
							reset_caller_media_addr(peer_session, local_session, active_remote_media, am);
						}
					}
					break;
				}
			}
		}

		if (found) {
			result->media[result->media_count++] = am;
		} else {
			result->media[result->media_count++] = pjmedia_sdp_media_clone_deactivate(pool, im);
		}
	}
	return result;
}

/*added by jklou for fec/fecc attribure passthrough*/
static int set_app_attr_local_remote_app_fec_sdp(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
					const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	ast_log(LOG_NOTICE, "set_app_attr_local_remote_app_fec_sdp\n");
	pjmedia_sdp_media *tmp_local;
	tmp_local = pjmedia_sdp_media_clone(pool, local);
	int i = 0, j =0 ;
	tmp_local->attr_count = 0;
	for (i = 0; i < local->attr_count; i++) {
		if (strncasecmp(local->attr[i]->name.ptr, "sendrecv", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "sendonly", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "recvonly", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "inactive", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "rtpmap", 6) &&
				strncasecmp(local->attr[i]->name.ptr, "crypto", 6)) {
			int found = 0;
			for (j = 0; j < callee_answer->attr_count; j++) {
				if (pj_strcmp(&local->attr[i]->name, &callee_answer->attr[j]->name) == 0) {
					found = 1;
					break;
				}
			}
			if (found == 0) {
				ast_log(LOG_NOTICE, "Not find the attribute , do not add\n");
			} else {
				tmp_local->attr[tmp_local->attr_count++] = local->attr[i];
			}

		}else {
			tmp_local->attr[tmp_local->attr_count++] = local->attr[i];
		}
	}
	tmp_local->bandw_count = 0;
	for (i = 0; i < callee_answer->bandw_count; i++) {
		tmp_local->bandw[tmp_local->bandw_count++] = callee_answer->bandw[i];
	}
	*p_result = pjmedia_sdp_media_clone(pool, tmp_local);
	return 0;
}

static int set_app_attr_local_remote_app_fecc_sdp(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
					const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	ast_log(LOG_NOTICE, "set_app_attr_local_remote_app_fecc_sdp\n");
	pjmedia_sdp_media *tmp_local;
	tmp_local = pjmedia_sdp_media_clone(pool, local);
	int i = 0, j =0 ;
	tmp_local->attr_count = 0;
	for (i = 0; i < local->attr_count; i++) {
		if (strncasecmp(local->attr[i]->name.ptr, "sendrecv", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "sendonly", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "recvonly", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "inactive", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "rtpmap", 6) &&
				strncasecmp(local->attr[i]->name.ptr, "crypto", 6)) {
			int found = 0;
			for (j = 0; j < callee_answer->attr_count; j++) {
				if (pj_strcmp(&local->attr[i]->name, &callee_answer->attr[j]->name) == 0) {
					found = 1;
					break;
				}
			}
			if (found == 0) {
				ast_log(LOG_NOTICE, "Not find the attribute , do not add\n");
			} else {
				tmp_local->attr[tmp_local->attr_count++] = local->attr[i];
			}

		}else {
			tmp_local->attr[tmp_local->attr_count++] = local->attr[i];
		}
	}
	tmp_local->bandw_count = 0;
	for (i = 0; i < callee_answer->bandw_count; i++) {
		tmp_local->bandw[tmp_local->bandw_count++] = callee_answer->bandw[i];
	}
	*p_result = pjmedia_sdp_media_clone(pool, tmp_local);
	return 0;
}


static int set_app_attr_local_remote_app_sdp(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
					const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	int res = 0;

	if (ast_sip_session_judge_if_fec_stream(local_session, local) &&
		ast_sip_session_judge_if_fec_stream(peer_session, callee_answer)) {
		ast_log(LOG_DEBUG, "It's private FEC for grandstream, match as FEC\n");
		res = set_app_attr_local_remote_app_fec_sdp(pool, local, callee_answer, local_session, peer_session, p_result);
		return res;
    } else if (ast_sip_session_judge_if_fecc_stream(local_session, local) &&
		ast_sip_session_judge_if_fecc_stream(peer_session, callee_answer)) {
		ast_log(LOG_DEBUG, "It's FECC, match as FECC\n");
		res = set_app_attr_local_remote_app_fecc_sdp(pool, local, callee_answer, local_session, peer_session, p_result);
		return res;
    } else {
		ast_log(LOG_DEBUG, "Don't have rtpmap, just deactivate media\n");
		return -1;
    }

	return -1;
}

/*added by jklou for bfcp attribure passthrough*/
static int set_app_attr_local_remote_bfcp_sdp(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
					const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	ast_log(LOG_NOTICE, "set_app_attr_local_remote_bfcp_sdp\n");
	pjmedia_sdp_media *tmp_local;
	tmp_local = pjmedia_sdp_media_clone(pool, local);
	int i = 0, j =0 ;
	tmp_local->attr_count = 0;
	for (i = 0; i < local->attr_count; i++) {
		if (strncasecmp(local->attr[i]->name.ptr, "sendrecv", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "sendonly", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "recvonly", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "inactive", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "rtpmap", 6) &&
				strncasecmp(local->attr[i]->name.ptr, "confid", 6) &&
				strncasecmp(local->attr[i]->name.ptr, "userid", 6) &&
				strncasecmp(local->attr[i]->name.ptr, "floorctrl", 9) &&
				strncasecmp(local->attr[i]->name.ptr, "setup", 5) &&
				strncasecmp(local->attr[i]->name.ptr, "connection", 10) &&
				strncasecmp(local->attr[i]->name.ptr, "floorid", 7) &&
				strncasecmp(local->attr[i]->name.ptr, "fingerprint", 11) &&
				strncasecmp(local->attr[i]->name.ptr, "crypto", 6)) {
			int found = 0;
			for (j = 0; j < callee_answer->attr_count; j++) {
				if (pj_strcmp(&local->attr[i]->name, &callee_answer->attr[j]->name) == 0) {
					found = 1;
					break;
				}
			}
			if (found == 0) {
				ast_log(LOG_NOTICE, "Not find the attribute , do not add\n");
			} else {
				tmp_local->attr[tmp_local->attr_count++] = local->attr[i];
			}

		}else {
			tmp_local->attr[tmp_local->attr_count++] = local->attr[i];
		}
	}
	tmp_local->bandw_count = 0;
	for (i = 0; i < callee_answer->bandw_count; i++) {
		tmp_local->bandw[tmp_local->bandw_count++] = callee_answer->bandw[i];
	}
	*p_result = pjmedia_sdp_media_clone(pool, tmp_local);
	return 0;
}
static int set_main_video_attr_local_remote_sdp(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
					const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	ast_log(LOG_DEBUG, "set_main_video_attr_local_remote_sdp\n");

	pjmedia_sdp_media *tmp_local;
	tmp_local = pjmedia_sdp_media_clone(pool, local);
	int i = 0, j =0 ;
	tmp_local->attr_count = 0;
	for (i = 0; i < local->attr_count; i++) {
		if (strncasecmp(local->attr[i]->name.ptr, "rtpmap", 6) &&
				strncasecmp(local->attr[i]->name.ptr, "fmtp", 4) &&
				strncasecmp(local->attr[i]->name.ptr, "sendrecv", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "sendonly", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "recvonly", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "inactive", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "framerate", 9) &&
				strncasecmp(local->attr[i]->name.ptr, "label", 5) &&
				strncasecmp(local->attr[i]->name.ptr, "content", 7) &&
				strncasecmp(local->attr[i]->name.ptr, "crypto", 6)) {
			int found = 0;
			for (j = 0; j < callee_answer->attr_count; j++) {
				if (pj_strcmp(&local->attr[i]->name, &callee_answer->attr[j]->name) == 0) {
					found = 1;
					break;
				}
			}
			if (found == 0) {
				ast_log(LOG_NOTICE, "Not find the attribute , do not add\n");
			} else {
				tmp_local->attr[tmp_local->attr_count++] = local->attr[i];
			}

		}else {
			tmp_local->attr[tmp_local->attr_count++] = local->attr[i];
		}
	}
	tmp_local->bandw_count = 0;
	for (i = 0; i < callee_answer->bandw_count; i++) {
		tmp_local->bandw[tmp_local->bandw_count++] = callee_answer->bandw[i];
	}
	*p_result = pjmedia_sdp_media_clone(pool, tmp_local);
	return 0;
}

static int set_slides_video_attr_local_remote_sdp(pj_pool_t *pool,
						pjmedia_sdp_media *local,
						const pjmedia_sdp_media *callee_answer,
						const struct ast_sip_session *local_session,
						const struct ast_sip_session *peer_session,
						pjmedia_sdp_media **p_result)
{
	ast_log(LOG_DEBUG, "set_slides_video_attr_local_remote_sdp\n");

	pjmedia_sdp_media *tmp_local;
	tmp_local = pjmedia_sdp_media_clone(pool, local);
	int i = 0, j =0 ;
	tmp_local->attr_count = 0;
	for (i = 0; i < local->attr_count; i++) {
		if (strncasecmp(local->attr[i]->name.ptr, "rtpmap", 6) &&
				strncasecmp(local->attr[i]->name.ptr, "fmtp", 4) &&
				strncasecmp(local->attr[i]->name.ptr, "sendrecv", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "sendonly", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "recvonly", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "inactive", 8) &&
				strncasecmp(local->attr[i]->name.ptr, "framerate", 9) &&
				strncasecmp(local->attr[i]->name.ptr, "label", 5) &&
				strncasecmp(local->attr[i]->name.ptr, "content", 7) &&
				strncasecmp(local->attr[i]->name.ptr, "crypto", 6)) {
			int found = 0;
			for (j = 0; j < callee_answer->attr_count; j++) {
				if (pj_strcmp(&local->attr[i]->name, &callee_answer->attr[j]->name) == 0) {
					found = 1;
					break;
				}
			}
			if (found == 0) {
				ast_log(LOG_NOTICE, "Not find the attribute , do not add\n");
			} else {
				tmp_local->attr[tmp_local->attr_count++] = local->attr[i];
			}

		}else {
			tmp_local->attr[tmp_local->attr_count++] = local->attr[i];
		}
	}
	tmp_local->bandw_count = 0;
	for (i = 0; i < callee_answer->bandw_count; i++) {
		tmp_local->bandw[tmp_local->bandw_count++] = callee_answer->bandw[i];
	}
	*p_result = pjmedia_sdp_media_clone(pool, tmp_local);
	return 0;
}

static int set_video_bandw_sdp(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
					const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	ast_log(LOG_DEBUG, "set_video_bandw_sdp\n");

	pjmedia_sdp_media *tmp_local;
	tmp_local = pjmedia_sdp_media_clone(pool, local);
	int i = 0;
	tmp_local->bandw_count = 0;
	for (i = 0; i < callee_answer->bandw_count; i++) {
		tmp_local->bandw[tmp_local->bandw_count++] = callee_answer->bandw[i];
	}
	*p_result = pjmedia_sdp_media_clone(pool, tmp_local);
	return 0;
}

static int set_video_attr_local_remote_sdp(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
					const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	pjmedia_sdp_attr *attr1;
	pjmedia_sdp_attr *attr2;
	char content1[8];
	char content2[8];
	int res = 0;
	memset(content1, 0, sizeof(content1));
	memset(content2, 0, sizeof(content2));
	ast_log(LOG_DEBUG, "set_video_attr_local_remote_sdp begin\n");

	if ((attr1 = pjmedia_sdp_media_find_attr2(local, "content", NULL)) && 
			(attr2 = pjmedia_sdp_media_find_attr2(callee_answer, "content", NULL))) {
		ast_copy_pj_str(content1, pj_strltrim(&attr1->value), sizeof(content1));
		ast_copy_pj_str(content2, pj_strltrim(&attr2->value), sizeof(content2));
		if ((0 != strlen(content1)) && (0 != strlen(content2))) {
			if ((!strncmp(content1, "main", 4)) && (!strncmp(content2, "main", 4))){
				res = set_main_video_attr_local_remote_sdp(pool, local, callee_answer, local_session, peer_session, p_result);
				return res;
			} else if ((!strncmp(content1, "slides", 6)) && (!strncmp(content2, "slides", 6))) {
				res = set_slides_video_attr_local_remote_sdp(pool, local, callee_answer, local_session, peer_session, p_result);
				return res;
			} else  if (strncmp(content1, "main", 4) &&
							strncmp(content2, "main", 4)&&
							strncmp(content1, "slides", 6) &&
							strncmp(content2, "slides", 6)){
				ast_log(LOG_DEBUG, "can't support third video stream or not the same main or slides, normal video\n");
				res = set_video_bandw_sdp(pool, local, callee_answer, local_session, peer_session, p_result);
				return res;
			} else {
				ast_log(LOG_WARNING, "can't support third video stream or not the same main or slides\n");
				return -1;
			}
		}
	}else {
		ast_log(LOG_DEBUG, "can't support main/slides video stream , normal video\n");
		res = set_video_bandw_sdp(pool, local, callee_answer, local_session, peer_session, p_result);
		return res;
	}
	return -1;
}

/*added by jklou for fec/fecc/bfcp attribure passthrough*/
static pjmedia_sdp_session* set_app_attr_local_remote_sdp( pj_pool_t *pool,
				  pjmedia_sdp_session *caller_local,
				  const pjmedia_sdp_session *callee_answer,
				  const struct ast_sip_session *local_session,
				  const struct ast_sip_session *peer_session
				  )
{
	pjmedia_sdp_session *result;
	char media_used[PJMEDIA_MAX_SDP_MEDIA];
	unsigned found = 0;
	unsigned i, j;
	pjmedia_sdp_media *im;

	result = pjmedia_sdp_session_clone(pool, caller_local);
	result->media_count = 0;

	pj_bzero(media_used, sizeof(media_used));
	for (im = NULL, j = 0; j < caller_local->media_count; ++j) {
		const pjmedia_sdp_media *om;
		pjmedia_sdp_media *am = NULL;

		im = caller_local->media[j];
		found = 0;

		for (i = 0; i < callee_answer->media_count; ++i) {
			om = callee_answer->media[i];
			if (pj_strcmp(&om->desc.media, &im->desc.media) == 0 &&
				judge_if_transport_compatible(local_session, im) &&
				judge_if_transport_compatible(peer_session, om) &&
				media_used[i] == 0)
			{
				int status = -1;
				if (!strncmp(om->desc.media.ptr, "application", 11)) {
					enum ast_sip_application_type app_type = ast_sip_session_get_application_type(om->desc.transport);
					enum ast_sip_application_type i_app_type = ast_sip_session_get_application_type(im->desc.transport);
					if (app_type == i_app_type) {
						switch(app_type) {
							case AST_SIP_APP_UDP_BFCP:
							case AST_SIP_APP_TCP_BFCP:
								status = set_app_attr_local_remote_bfcp_sdp(pool, im, om, local_session, peer_session, &am);
								break;
							case AST_SIP_APP_RTP:
								status = set_app_attr_local_remote_app_sdp(pool, im, om, local_session, peer_session, &am);
								break;
							default:
								ast_log(LOG_DEBUG, "Can't match other application media, except bfcp/fec/fecc\n");
						}
					}
				}
				if (!strncmp(om->desc.media.ptr, "video", 5)) {
					status = set_video_attr_local_remote_sdp(pool, im, om, local_session, peer_session, &am);
				}
				if (status == 0){
					found = 1;
					break;
				}
			}
		}
		if (found) {
			result->media[result->media_count++] = am;
		} else {
			result->media[result->media_count++] = im;
		}
	}
	return result;
}

/*ended by jklou*/

struct ast_format {
	/*! Name of the format */
	const char *name;
	/*! \brief Pointer to the codec in use for this format */
	struct ast_codec *codec;
	/*! \brief Attribute specific data, implementation specific */
	void *attribute_data;
	/*! \brief Pointer to the optional format interface */
	const struct ast_format_interface *interface;
};

struct format_cap_framed {
	/*! \brief A pointer to the format */
	struct ast_format *format;
	/*! \brief The format framing size */
	unsigned int framing;
	/*! \brief Linked list information */
	AST_LIST_ENTRY(format_cap_framed) entry;
};

struct ast_format_cap {
	/*! \brief Vector of formats, indexed using the codec identifier */
	AST_VECTOR(, struct format_cap_framed_list) formats;
	/*! \brief Vector of formats, added in preference order */
	AST_VECTOR(, struct format_cap_framed *) preference_order;
	/*! \brief Global framing size, applies to all formats if no framing present on format */
	unsigned int framing;
};

/* Begin add by bfdu for bfcp tcp, 2015.09.11 */
static void set_bfcp_tcp_peer_remote_addr(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	const pjmedia_sdp_session *remote, *peer_remote;
	char host[NI_MAXHOST];
	char peer_host[NI_MAXHOST];
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	RAII_VAR(struct ast_sockaddr *, peer_addrs, NULL, ast_free);
	pjmedia_sdp_media *m_remote;
	pjmedia_sdp_media *m_peer_remote;
	unsigned i, j;

	ast_log(LOG_DEBUG, "set peer bfcp tcp remote addr begin\n");

	if (!session || !peer_session || !session->inv_session || !peer_session->inv_session) {
		ast_log(LOG_DEBUG, "set peer bfcp tcp remote addr failed\n");
		return;
	}

	if (pjmedia_sdp_neg_get_neg_remote(session->inv_session->neg, &remote) != PJ_SUCCESS) {
		ast_log(LOG_WARNING, "Can't get neg remote sdp\n");
		return;
	}

	for (i = 0; i < remote->media_count; ++i) {
		m_remote = remote->media[i];
		enum ast_sip_application_type app_type = ast_sip_session_get_application_type(m_remote->desc.transport);
		if (AST_SIP_APP_TCP_BFCP == app_type) {
			if (0 == m_remote->desc.port) {
				ast_log(LOG_DEBUG, "Remote BFCP is at port 0, don't set remote addr\n");
				break;
			}
			ast_copy_pj_str(host, m_remote->conn ? &m_remote->conn->addr : &remote->conn->addr, sizeof(host));
			if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
				ast_log(LOG_ERROR, "Set bfcp tcp remote addr, provided remote addr was actually invalid\n");
				return;
			}
			ast_sockaddr_set_port(addrs, m_remote->desc.port);
			ast_verbose("Remote BFCP is at port %s\n", ast_sockaddr_stringify(addrs));
			ast_bfcp_tcp_set_peer_remote(peer_session->bfcp_tcp, addrs);
			break;
		}
	}

	if (pjmedia_sdp_neg_get_active_remote(peer_session->inv_session->neg, &peer_remote) != PJ_SUCCESS) {
		ast_log(LOG_WARNING, "Can't get peer remote active sdp\n");
		return;
	}

	for (j = 0; j < peer_remote->media_count; ++j) {
		m_peer_remote = peer_remote->media[j];
		enum ast_sip_application_type peer_app_type = ast_sip_session_get_application_type(m_peer_remote->desc.transport);
		if (AST_SIP_APP_TCP_BFCP == peer_app_type) {
			if (0 == m_peer_remote->desc.port) {
				ast_log(LOG_DEBUG, "Peer remote BFCP is at port 0, don't set peer remote addr\n");
				break;
			}
			ast_copy_pj_str(peer_host, m_peer_remote->conn ? &m_peer_remote->conn->addr : &peer_remote->conn->addr, sizeof(peer_host));
			if (ast_sockaddr_resolve(&peer_addrs, peer_host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
				ast_log(LOG_ERROR, "Set bfcp tcp peer remote addr, provided remote addr was actually invalid\n");
				return;
			}
			ast_sockaddr_set_port(peer_addrs, m_peer_remote->desc.port);
			ast_verbose("Peer remote BFCP is at port %s\n", ast_sockaddr_stringify(peer_addrs));
			ast_bfcp_tcp_set_peer_remote(session->bfcp_tcp, peer_addrs);
			break;
		}
	}

	return;
}
/* End add by bfdu for bfcp tcp, 2015.09.11 */
static int send_183_invite_request(void *data)
{
	struct ast_sip_session *session = data;
	ast_sip_session_refresh(session, NULL, NULL, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1);
	ao2_ref(session, -1);

	return 0;
}

static void chan_pjsip_use_final_response_sdp(struct ast_sip_channel_pvt *chan_pvt, struct ast_sip_channel_pvt *peer_pvt)
{
	if (!chan_pvt || !peer_pvt) {
		ast_log(LOG_WARNING, "chan_pjsip_use_final_response_sdp param is invalid\n");
		return;
	}
	if (!chan_pvt->session || !chan_pvt->session->inv_session || !chan_pvt->session->inv_session->neg ||
		!chan_pvt->session->inv_session->neg->active_local_sdp) {
		ast_log(LOG_WARNING, "chan_pjsip_use_final_response_sdp chan_pvt session is invalid\n");
		return;
	}
	if (!peer_pvt->session || !peer_pvt->session->inv_session || !peer_pvt->session->inv_session->neg ||
		!peer_pvt->session->inv_session->neg->active_local_sdp ||
		!peer_pvt->session->inv_session->neg->active_remote_sdp) {
		ast_log(LOG_WARNING, "chan_pjsip_use_final_response_sdp peer_pvt session is invalid\n");
		return;
	}
	pjmedia_sdp_session *callee_answer_sdp, *caller_local_sdp, *callee_sdp_local;
	pj_pool_t *pool = chan_pvt->session->inv_session->pool_prov;

	callee_sdp_local = peer_pvt->session->inv_session->neg->active_local_sdp;
	caller_local_sdp = chan_pvt->session->inv_session->neg->active_local_sdp;
	callee_answer_sdp = peer_pvt->session->inv_session->neg->active_remote_sdp;

	char buf[1024] = {0};
	{
		memset(buf, 0, sizeof(buf));
		pjmedia_sdp_print(caller_local_sdp, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
		ast_log(LOG_CODEC_NEG, "[caller] caller local sdp is:\n%s", buf);

		memset(buf, 0, sizeof(buf));
		pjmedia_sdp_print(callee_answer_sdp, buf, sizeof(buf) - 1);// Must use sizeof(buf) - 1, otherwise there is a potential overflow
		ast_log(LOG_CODEC_NEG, "[callee] UCM <<<=== B . received 200 ok from the callee\n%s", buf);

		memset(buf, 0, sizeof(buf));
		pjmedia_sdp_print(callee_sdp_local, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
		ast_log(LOG_CODEC_NEG, "[callee] local sdp of the callee is:\n%s", buf);
	}

	ast_log(LOG_CODEC_NEG, "[caller] modify caller local sdp according to callee answer sdp!\n");

    chan_pvt->session->inv_session->neg->neg_local_sdp = modify_local_sdp_with_callee_answer(pool, caller_local_sdp, callee_sdp_local, chan_pvt->session, peer_pvt->session);

	chan_pvt->session->inv_session->neg->neg_local_sdp = set_app_attr_local_remote_sdp(pool, chan_pvt->session->inv_session->neg->neg_local_sdp,
		callee_answer_sdp, chan_pvt->session, peer_pvt->session);

	{
		memset(buf, 0, sizeof(buf));
		pjmedia_sdp_print(chan_pvt->session->inv_session->neg->neg_local_sdp, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
		ast_log(LOG_CODEC_NEG, "[caller] caller local sdp is:\n%s", buf);
	}

	chan_pvt->session->has_early_media = 1;

	pjmedia_sdp_neg_set_remote_offer(pool, chan_pvt->session->inv_session->neg, chan_pvt->session->inv_session->neg->active_remote_sdp);

	set_bfcp_tcp_peer_remote_addr(chan_pvt->session, peer_pvt->session);

	return;
}

static void chan_pjsip_send_reinvite_after_early_media(struct ast_sip_channel_pvt *chan_pvt, struct ast_sip_channel_pvt *peer_pvt)
{
	if (!chan_pvt || !peer_pvt) {
		ast_log(LOG_WARNING, "chan_pjsip_send_reinvite_after_early_media param is invalid\n");
		return;
	}

	struct ast_str *format_buf = ast_str_alloca(64);
	ast_log(LOG_CODEC_NEG, "[caller] final_caps: %s\n", ast_format_cap_get_names(chan_pvt->session->final_caps, &format_buf));
	ast_log(LOG_CODEC_NEG, "[callee] final_caps: %s\n", ast_format_cap_get_names(peer_pvt->session->final_caps, &format_buf));
	ast_log(LOG_CODEC_NEG, "[callee] req_caps: %s\n", ast_format_cap_get_names(peer_pvt->session->req_caps, &format_buf));

	/* Add by bfdu, we only need to compare first format, if not the same, will need translate, don't send reinvite */
	RAII_VAR(struct ast_format_cap *, peer_caps_tmp, NULL, ao2_cleanup);
	peer_caps_tmp = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!peer_caps_tmp) {
		ast_log(LOG_ERROR, "Alloc peer caps failed\n");
		return;
	}
	ast_format_cap_append_from_cap(peer_caps_tmp, peer_pvt->session->final_caps, AST_MEDIA_TYPE_UNKNOWN);

	RAII_VAR(struct ast_format *, local_format, NULL, ao2_cleanup);
	local_format = ast_format_cap_get_format(peer_caps_tmp, 0);
	if (local_format) {
		ast_format_cap_remove_by_type(peer_caps_tmp, AST_MEDIA_TYPE_AUDIO);
		ast_format_cap_append(peer_caps_tmp, local_format, 0);
	} else {
		ast_log(LOG_ERROR, "Can't find format\n");
		return;
	}

	RAII_VAR(struct ast_format_cap *, joint_caps, NULL, ao2_cleanup);
	joint_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	ast_format_cap_get_compatible(peer_caps_tmp, chan_pvt->session->final_caps, joint_caps);
	ast_log(LOG_CODEC_NEG, "joint_caps: %s\n", ast_format_cap_get_names(joint_caps, &format_buf));

	if (AST_VECTOR_SIZE(&joint_caps->preference_order))
	{
		struct ast_format *audio_early_format, *audio_neg_format, *video_early_format, *video_neg_format;
		audio_early_format = ast_format_cap_get_best_by_type(chan_pvt->session->final_caps, AST_MEDIA_TYPE_AUDIO);
		audio_neg_format = ast_format_cap_get_best_by_type(joint_caps, AST_MEDIA_TYPE_AUDIO);

		video_early_format = ast_format_cap_get_best_by_type(chan_pvt->session->final_caps, AST_MEDIA_TYPE_VIDEO);
		video_neg_format = ast_format_cap_get_best_by_type(joint_caps, AST_MEDIA_TYPE_VIDEO);

		if ( (ast_format_cmp(audio_early_format, audio_neg_format) == AST_FORMAT_CMP_NOT_EQUAL
				&& ast_format_cap_has_type(joint_caps, AST_MEDIA_TYPE_AUDIO))
				|| (!(video_early_format == NULL && video_neg_format == NULL)
						&& ast_format_cmp(video_early_format, video_neg_format) == AST_FORMAT_CMP_NOT_EQUAL
						&& (ast_format_cap_has_type(joint_caps, AST_MEDIA_TYPE_VIDEO))))
		{
			ast_format_cap_remove_by_type(chan_pvt->session->req_caps, AST_MEDIA_TYPE_UNKNOWN);
			ast_format_cap_append_from_cap(chan_pvt->session->req_caps, joint_caps, AST_MEDIA_TYPE_UNKNOWN);

			ast_format_cap_append_from_cap(chan_pvt->session->req_caps, chan_pvt->session->endpoint->media.codecs, AST_MEDIA_TYPE_AUDIO);

			ast_log(LOG_CODEC_NEG, "[caller] req_caps: %s\n", ast_format_cap_get_names(chan_pvt->session->req_caps, &format_buf));

			ast_channel_nativeformats_set(chan_pvt->session->channel, joint_caps);
			ast_channel_nativeformats_set(peer_pvt->session->channel, joint_caps);

			ao2_ref(chan_pvt->session, +1);
			if (ast_sip_push_task(chan_pvt->session->serializer, send_183_invite_request, chan_pvt->session)) {
				ao2_cleanup(chan_pvt->session);
			}
		}
		if (audio_early_format) {
			ao2_ref(audio_early_format, -1);
		}
		if (audio_neg_format) {
			ao2_ref(audio_neg_format, -1);
		}
		if (video_early_format) {
			ao2_ref(video_early_format, -1);
		}
		if (video_neg_format) {
			ao2_ref(video_neg_format, -1);
		}
	}

	return;
}

static int chan_pjsip_special_answer(struct ast_channel *chan, struct ast_channel *peer)
{
	if (peer) {
		if (0 == strcmp(ast_channel_tech(peer)->type, "PJSIP")) {
			struct ast_sip_channel_pvt *chan_pvt, *peer_pvt;
			pjmedia_sdp_neg_state chan_neg_state, peer_neg_state;

			chan_pvt = ast_channel_tech_pvt(chan);
			peer_pvt = ast_channel_tech_pvt(peer);

			if (!chan_pvt || !chan_pvt->session || !chan_pvt->session->inv_session || !chan_pvt->session->inv_session->neg) {
				ast_log(LOG_WARNING, "special answer chan_pvt session is invalid\n");
				return -1;
			}
			if (!peer_pvt || !peer_pvt->session || !peer_pvt->session->inv_session || !peer_pvt->session->inv_session->neg) {
				ast_log(LOG_WARNING, "special answer peer_pvt session is invalid\n");
				return -1;
			}

			chan_neg_state = chan_pvt->session->inv_session->neg->state;
			peer_neg_state = peer_pvt->session->inv_session->neg->state;

			if (chan_neg_state == PJMEDIA_SDP_NEG_STATE_WAIT_NEGO && peer_neg_state == PJMEDIA_SDP_NEG_STATE_DONE)
			{
				if (!chan_pvt->session->inv_session->neg->neg_local_sdp) {
					ast_log(LOG_WARNING, "special answer chan_pvt neg_local_sdp is invalid\n");
					return -1;
				}
				if (!peer_pvt->session->inv_session->neg->active_local_sdp ||
					!peer_pvt->session->inv_session->neg->active_remote_sdp) {
					ast_log(LOG_WARNING, "special answer peer_pvt active local/remote sdp is invalid\n");
					return -1;
				}
				pjmedia_sdp_session *callee_answer_sdp, *caller_local_sdp, *callee_sdp_local;
				pj_pool_t *pool = chan_pvt->session->inv_session->pool_prov;

				callee_sdp_local = peer_pvt->session->inv_session->neg->active_local_sdp;
				caller_local_sdp = chan_pvt->session->inv_session->neg->neg_local_sdp;
				callee_answer_sdp = peer_pvt->session->inv_session->neg->active_remote_sdp;

				char buf[1024] = {0};
				{
					memset(buf, 0, sizeof(buf));
					pjmedia_sdp_print(caller_local_sdp, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
					ast_log(LOG_CODEC_NEG, "[caller] caller local sdp is:\n%s", buf);

					memset(buf, 0, sizeof(buf));
					pjmedia_sdp_print(callee_answer_sdp, buf, sizeof(buf) - 1);// Must use sizeof(buf) - 1, otherwise there is a potential overflow
					ast_log(LOG_CODEC_NEG, "[callee] UCM <<<=== B . received 200 ok from the callee\n%s", buf);

					memset(buf, 0, sizeof(buf));
					pjmedia_sdp_print(callee_sdp_local, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
					ast_log(LOG_CODEC_NEG, "[callee] local sdp of the callee is:\n%s", buf);
				}

				ast_log(LOG_CODEC_NEG, "[caller] modify caller local sdp according to callee answer sdp!\n");

				/* Add by bfdu for modify yealink profile-level-id value, 2015.12.02 */
				chan_pvt->session->is_yealink_msg = peer_pvt->session->is_yealink_msg;

			    chan_pvt->session->inv_session->neg->neg_local_sdp = modify_local_sdp_with_callee_answer(pool, caller_local_sdp, callee_sdp_local, chan_pvt->session, peer_pvt->session);
				/* added by jklou for fec/fecc/bfcp attribute passthrough */

				chan_pvt->session->inv_session->neg->neg_local_sdp = set_app_attr_local_remote_sdp(pool, chan_pvt->session->inv_session->neg->neg_local_sdp,
					callee_answer_sdp, chan_pvt->session, peer_pvt->session);
				/* ended */
				{
					memset(buf, 0, sizeof(buf));
					pjmedia_sdp_print(chan_pvt->session->inv_session->neg->neg_local_sdp, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
					ast_log(LOG_CODEC_NEG, "[caller] caller local sdp is:\n%s", buf);
				}
				/* Add by bfdu for bfcp tcp, 2015.09.11 */
				set_bfcp_tcp_peer_remote_addr(chan_pvt->session, peer_pvt->session);
			}

			if (chan_neg_state == PJMEDIA_SDP_NEG_STATE_DONE && peer_neg_state == PJMEDIA_SDP_NEG_STATE_DONE)
			{
				if (chan_pvt->session->inv_session->use_final_rsp_sdp) {
					chan_pjsip_use_final_response_sdp(chan_pvt, peer_pvt);
				} else {
					chan_pjsip_send_reinvite_after_early_media(chan_pvt, peer_pvt);
				}
			}
		}
	}
	return chan_pjsip_answer(chan);
}
#endif

/*! \brief Function called by core when we should answer a PJSIP session */
static int chan_pjsip_answer(struct ast_channel *ast)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct ast_sip_session *session;

	if (ast_channel_state(ast) == AST_STATE_UP) {
		return 0;
	}

	ast_setstate(ast, AST_STATE_UP);
	session = ao2_bump(channel->session);

	/* the answer task needs to be pushed synchronously otherwise a race condition
	   can occur between this thread and bridging (specifically when native bridging
	   attempts to do direct media) */
	ast_channel_unlock(ast);
	if (ast_sip_push_task_synchronous(session->serializer, answer, session)) {
		ast_log(LOG_WARNING, "Unable to push answer task to the threadpool. Cannot answer call\n");
		ao2_ref(session, -1);
		ast_channel_lock(ast);
		return -1;
	}
	ao2_ref(session, -1);
	ast_channel_lock(ast);

	return 0;
}

/*! \brief Internal helper function called when CNG tone is detected */
static struct ast_frame *chan_pjsip_cng_tone_detected(struct ast_sip_session *session, struct ast_frame *f)
{
	const char *target_context;
	int exists;

	/* If we only needed this DSP for fax detection purposes we can just drop it now */
	if (session->endpoint->dtmf == AST_SIP_DTMF_INBAND || session->endpoint->dtmf == AST_SIP_DTMF_AUTO) {
		ast_dsp_set_features(session->dsp, DSP_FEATURE_DIGIT_DETECT);
	} else {
		ast_dsp_free(session->dsp);
		session->dsp = NULL;
	}

	/* If already executing in the fax extension don't do anything */
	if (!strcmp(ast_channel_exten(session->channel), "fax")) {
		return f;
	}

	target_context = S_OR(ast_channel_macrocontext(session->channel), ast_channel_context(session->channel));

	/* We need to unlock the channel here because ast_exists_extension has the
	 * potential to start and stop an autoservice on the channel. Such action
	 * is prone to deadlock if the channel is locked.
	 */
	ast_channel_unlock(session->channel);
	exists = ast_exists_extension(session->channel, target_context, "fax", 1,
		S_COR(ast_channel_caller(session->channel)->id.number.valid,
			ast_channel_caller(session->channel)->id.number.str, NULL));
	ast_channel_lock(session->channel);

	if (exists) {
		ast_verb(2, "Redirecting '%s' to fax extension due to CNG detection\n",
			ast_channel_name(session->channel));
		pbx_builtin_setvar_helper(session->channel, "FAXEXTEN", ast_channel_exten(session->channel));
		if (ast_async_goto(session->channel, target_context, "fax", 1)) {
			ast_log(LOG_ERROR, "Failed to async goto '%s' into fax extension in '%s'\n",
				ast_channel_name(session->channel), target_context);
		}
		ast_frfree(f);
		f = &ast_null_frame;
	} else {
#ifdef GRANDSTREAM_NETWORKS
		short intelligent_route_flag = 0;
		const char *callee_num = pbx_builtin_getvar_helper(session->channel, "EXTTOCALL");
		const char *from_trunk = pbx_builtin_getvar_helper(session->channel, "FROM_TRUNK");

		if (callee_num && session->endpoint->fax_intelligent_route_destination
			&& !strcasecmp(callee_num, session->endpoint->fax_intelligent_route_destination)) {
			intelligent_route_flag = 1;
			ast_log(LOG_VERBOSE, "The same purpose(destination[%s]) and called extension(callee[%s])",
				session->endpoint->fax_intelligent_route_destination, callee_num);
		}
		/*! Added by pxie@grandstream.cn for bug[42104], 2015-04-14*/
		if (!intelligent_route_flag && !strcasecmp(session->endpoint->faxcontext, "faxcontext")) {
			ast_verb(2, "Redirecting '%s' to fax extension due to CNG detection\n",
				ast_channel_name(session->channel));
			pbx_builtin_setvar_helper(session->channel, "FAXEXTEN", ast_channel_exten(session->channel));
			if (!strcasecmp(session->endpoint->fax_intelligent_route, "yes")
				&& strcasecmp(session->endpoint->fax_intelligent_route_destination, "no")
				&& from_trunk && !strcasecmp(from_trunk, "yes")) {
				ast_channel_set_is_enable_fax_detect(session->channel, 0);
				pbx_builtin_setvar_helper(session->channel, "_FAX_INTELLIGENT_ROUTE", "yes");
				pbx_builtin_setvar_helper(session->channel, "_FAX_INTELLIGENT_ROUTE_DETINATION",
					session->endpoint->fax_intelligent_route_destination);
				session->is_intelligent_route = 1;
			}
			if (ast_async_goto(session->channel, session->endpoint->faxcontext, "fax", 1)) {
				ast_log(LOG_NOTICE, "Failed to async goto '%s' into fax of '%s'\n",
					ast_channel_name(session->channel), session->endpoint->faxcontext);
			}
			f = &ast_null_frame;
		} else
		/*! Added by pxie@grandstream.cn for bug[42104], 2015-04-14*/
#endif
		ast_log(LOG_NOTICE, "FAX CNG detected on '%s' but no fax extension in '%s'\n",
			ast_channel_name(session->channel), target_context);
	}

	return f;
}

/*! \brief Function called by core to read any waiting frames */
static struct ast_frame *chan_pjsip_read(struct ast_channel *ast)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct chan_pjsip_pvt *pvt = channel->pvt;
	struct ast_frame *f;
	struct ast_sip_session_media *media = NULL;
	int rtcp = 0;
	int fdno = ast_channel_fdno(ast);

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC */
	if (NULL == channel->session) {
		ast_log(LOG_ERROR, "session is not exsit, don't read frame");
		return &ast_null_frame;
	}
#endif

	switch (fdno) {
	case 0:
		media = pvt->media[SIP_MEDIA_AUDIO];
		break;
	case 1:
		media = pvt->media[SIP_MEDIA_AUDIO];
		rtcp = 1;
		break;
	case 2:
		media = pvt->media[SIP_MEDIA_VIDEO];
		break;
	case 3:
		media = pvt->media[SIP_MEDIA_VIDEO];
		rtcp = 1;
		break;
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC */
	case 6:
		media = pvt->media[SIP_MEDIA_APPLICATION];
		f = ast_bfcp_read(media->bfcp);		/* UDP for bfcp */
		return f;
	case 7:
		if (channel->session->slides_video_media.rtp == NULL) {
			ast_log(LOG_ERROR, "Slides video media rtp is not exsit, don't read frame");
			return &ast_null_frame;
		}
		if (!(f = ast_rtp_instance_read(channel->session->slides_video_media.rtp, rtcp))) {
			ast_log(LOG_ERROR, "can't read slides video rtp");
			return f;
		}
		f->video_sub_type = AST_FRAME_SLIDES_VIDEO;
		return f;
	case 8:
		if (channel->session->slides_video_media.rtp == NULL) {
			ast_log(LOG_ERROR, "Slides video media rtp is not exsit, don't read rtcp frame");
			return &ast_null_frame;
		}
		rtcp = 1;
		if (!(f = ast_rtp_instance_read(channel->session->slides_video_media.rtp, rtcp))) {
			ast_log(LOG_ERROR, "can't read slides video rtcp");
			return f;
		}
		f->rtcp_media_type = AST_RTCP_SLIDES_VIDEO;
		return f;
	case 9:
		if (channel->session->fec_media.rtp == NULL) {
			ast_log(LOG_ERROR, "FEC media rtp is not exsit, don't read frame");
			return &ast_null_frame;
		}
		if (!(f = ast_rtp_instance_read(channel->session->fec_media.rtp, rtcp))) {
			ast_log(LOG_ERROR, "can't read fec rtp");
			return f;
		}
		f->app_rtp_sub_type = AST_FRAME_APP_FEC;
		return f;
	case 10:
		if (channel->session->fec_media.rtp == NULL) {
			ast_log(LOG_ERROR, "FEC media rtp is not exsit, don't read rtcp frame");
			return &ast_null_frame;
		}
		rtcp = 1;
		if (!(f = ast_rtp_instance_read(channel->session->fec_media.rtp, rtcp))) {
			ast_log(LOG_ERROR, "can't read fec rtcp");
			return f;
		}
		f->rtcp_media_type = AST_RTCP_FEC;
		return f;
	case 11:
		if (channel->session->fecc_media.rtp == NULL) {
			ast_log(LOG_ERROR, "FECC media rtp is not exsit, don't read frame");
			return &ast_null_frame;
		}
		if (!(f = ast_rtp_instance_read(channel->session->fecc_media.rtp, rtcp))) {
			ast_log(LOG_ERROR, "can't read fecc rtp");
			return f;
		}
		f->app_rtp_sub_type = AST_FRAME_APP_FECC;
		return f;
	case 12:
		if (channel->session->fecc_media.rtp == NULL) {
			ast_log(LOG_ERROR, "FECC media rtp is not exsit, don't read rtcp frame");
			return &ast_null_frame;
		}
		rtcp = 1;
		if (!(f = ast_rtp_instance_read(channel->session->fecc_media.rtp, rtcp))) {
			ast_log(LOG_ERROR, "can't read fecc rtcp");
			return f;
		}
		f->rtcp_media_type = AST_RTCP_FECC;
		return f;
#endif
	}

	if (!media || !media->rtp) {
		return &ast_null_frame;
	}

#ifdef GRANDSTREAM_NETWORKS
	/*! Begin: Added by pxie@grandstream.cn for bug[42104], 2015-04-14*/
	if (channel->session->wFaxDetect) {
		ast_rtp_fax_detect_set_flag(media->rtp, channel->session->wFaxDetect);
	} else {
		ast_rtp_fax_detect_set_flag(media->rtp, 0);
	}
	/*! End: Added by pxie@grandstream.cn for bug[42104], 2015-04-14*/
#endif
	if (!(f = ast_rtp_instance_read(media->rtp, rtcp))) {
		return f;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Begin add by bfdu for BFCP/FEC/FECC, 2015.08.08 */
	switch (fdno) {
	case 1:
		f->rtcp_media_type = AST_RTCP_AUDIO;
		break;
	case 2:
		/* Add by bfdu for FEC multiplex, 2015.11.24 */
		f->app_rtp_sub_type = AST_FRAME_APP_FEC_MULTIPLEXD;
		break;
	case 3:
		f->rtcp_media_type = AST_RTCP_MAIN_VIDEO;
		break;
	}
	/* End add by bfdu for BFCP/FEC/FECC, 2015.08.08 */

	if (!rtcp) {
		ast_rtp_instance_set_last_rx(media->rtp, time(NULL));
	}

	if (f->frametype != AST_FRAME_VOICE) {
		return f;
	}

	/*! Begin: Added by pxie@grandstream.cn for bug[47337, 50564], 2016-03-17.*/
	if (ast_channel_get_is_send_fax(ast) && !ast_channel_get_caller_extension_flag(ast)
		&& channel->session->dsp && !strncasecmp(ast_channel_name(ast), "PJSIP/trunk_", 12)) {
		return f;
	}
	/*! End: Added by pxie@grandstream.cn for bug[47337, 50564], 2016-03-17.*/

	if (channel->session->dsp && !channel->session->is_intelligent_route)
#else
	if (f->frametype != AST_FRAME_VOICE) {
		return f;
	}

	if (channel->session->dsp)
#endif
	{
		f = ast_dsp_process(ast, channel->session->dsp, f);

		if (f && (f->frametype == AST_FRAME_DTMF)) {
			if (f->subclass.integer == 'f') {
				ast_debug(3, "Fax CNG detected on %s\n", ast_channel_name(ast));
				f = chan_pjsip_cng_tone_detected(channel->session, f);
			} else {
				ast_debug(3, "* Detected inband DTMF '%c' on '%s'\n", f->subclass.integer,
					ast_channel_name(ast));
			}
		}
	}

	return f;
}

/*! \brief Function called by core to write frames */
static int chan_pjsip_write(struct ast_channel *ast, struct ast_frame *frame)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct chan_pjsip_pvt *pvt = channel->pvt;
	struct ast_sip_session_media *media;
	int res = 0;

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC */
	if (NULL == channel->session) {
		ast_log(LOG_ERROR, "session is not exsit, don't read frame");
		return 0;
	}
#endif

	switch (frame->frametype) {
	case AST_FRAME_VOICE:
		media = pvt->media[SIP_MEDIA_AUDIO];

		if (!media) {
			return 0;
		}
		if (ast_format_cap_iscompatible_format(ast_channel_nativeformats(ast), frame->subclass.format) == AST_FORMAT_CMP_NOT_EQUAL) {
			struct ast_str *cap_buf = ast_str_alloca(128);
			struct ast_str *write_transpath = ast_str_alloca(256);
			struct ast_str *read_transpath = ast_str_alloca(256);

			ast_log(LOG_WARNING,
				"Channel %s asked to send %s frame when native formats are %s (rd:%s->%s;%s wr:%s->%s;%s)\n",
				ast_channel_name(ast),
				ast_format_get_name(frame->subclass.format),
				ast_format_cap_get_names(ast_channel_nativeformats(ast), &cap_buf),
				ast_format_get_name(ast_channel_rawreadformat(ast)),
				ast_format_get_name(ast_channel_readformat(ast)),
				ast_translate_path_to_str(ast_channel_readtrans(ast), &read_transpath),
				ast_format_get_name(ast_channel_writeformat(ast)),
				ast_format_get_name(ast_channel_rawwriteformat(ast)),
				ast_translate_path_to_str(ast_channel_writetrans(ast), &write_transpath));
			return 0;
		}
		if (media->rtp) {
			res = ast_rtp_instance_write(media->rtp, frame);
		}
		break;
	case AST_FRAME_VIDEO:
#ifdef GRANDSTREAM_NETWORKS
		/* Modify by bfdu for BFCP/FEC/FECC, 2015.07.13 */
		if (frame->video_sub_type == AST_FRAME_SLIDES_VIDEO) {
			if (channel && channel->session && channel->session->slides_video_media.rtp) {
				res = ast_rtp_instance_write(channel->session->slides_video_media.rtp, frame);
			}
		} else {
			if ((media = pvt->media[SIP_MEDIA_VIDEO]) && media->rtp) {
				res = ast_rtp_instance_write(media->rtp, frame);
			}
		}
#else
		if ((media = pvt->media[SIP_MEDIA_VIDEO]) && media->rtp) {
			res = ast_rtp_instance_write(media->rtp, frame);
		}
#endif
		break;
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.13 */
	case AST_FRAME_BFCP:
		if ((media = pvt->media[SIP_MEDIA_APPLICATION]) && media->bfcp) {
			res = ast_bfcp_write(media->bfcp, frame);
		}
		break;
	case AST_FRAME_APP_RTP:
		switch(frame->app_rtp_sub_type) {
			case AST_FRAME_APP_FEC:
				if (channel && channel->session && channel->session->fec_media.rtp) {
					res = ast_rtp_instance_write(channel->session->fec_media.rtp, frame);
				}
				break;
			case AST_FRAME_APP_FECC:
				if (channel && channel->session && channel->session->fecc_media.rtp) {
					res = ast_rtp_instance_write(channel->session->fecc_media.rtp, frame);
				}
				break;
			case AST_FRAME_APP_FEC_MULTIPLEXD:
				if ((media = pvt->media[SIP_MEDIA_VIDEO]) && media->rtp) {
					res = ast_rtp_instance_write(media->rtp, frame);
				}
				break;
			default:
				ast_log(LOG_WARNING, "Can't process other application RTP(%d), execpt fec/fecc", frame->app_rtp_sub_type);
				break;
		}
		break;
	case AST_FRAME_RTCP:
		switch(frame->rtcp_media_type) {
			case AST_RTCP_AUDIO:
				if ((media = pvt->media[SIP_MEDIA_AUDIO]) && media->rtp) {
					res = ast_rtp_instance_write(media->rtp, frame);
				}
				break;
			case AST_RTCP_MAIN_VIDEO:
				if ((media = pvt->media[SIP_MEDIA_VIDEO]) && media->rtp) {
					res = ast_rtp_instance_write(media->rtp, frame);
				}
				break;
			case AST_RTCP_SLIDES_VIDEO:
				if (channel && channel->session && channel->session->slides_video_media.rtp) {
					res = ast_rtp_instance_write(channel->session->slides_video_media.rtp, frame);
				}
				break;
			case AST_RTCP_FEC:
				if (channel && channel->session && channel->session->fec_media.rtp) {
					res = ast_rtp_instance_write(channel->session->fec_media.rtp, frame);
				}
				break;
			case AST_RTCP_FECC:
				if (channel && channel->session && channel->session->fecc_media.rtp) {
					res = ast_rtp_instance_write(channel->session->fecc_media.rtp, frame);
				}
				break;
			default:
				ast_log(LOG_WARNING, "can't process this rtcp_media_type: %d", frame->rtcp_media_type);
				break;
		}
		break;
#endif
	case AST_FRAME_MODEM:
		break;
	default:
		ast_log(LOG_WARNING, "Can't send %u type frames with PJSIP\n", frame->frametype);
		break;
	}

	return res;
}

/*! \brief Function called by core to change the underlying owner channel */
static int chan_pjsip_fixup(struct ast_channel *oldchan, struct ast_channel *newchan)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(newchan);
	struct chan_pjsip_pvt *pvt = channel->pvt;

	if (channel->session->channel != oldchan) {
		return -1;
	}

	/*
	 * The masquerade has suspended the channel's session
	 * serializer so we can safely change it outside of
	 * the serializer thread.
	 */
	channel->session->channel = newchan;

	set_channel_on_rtp_instance(pvt, ast_channel_uniqueid(newchan));

	return 0;
}

/*! AO2 hash function for on hold UIDs */
static int uid_hold_hash_fn(const void *obj, const int flags)
{
	const char *key = obj;

	switch (flags & OBJ_SEARCH_MASK) {
	case OBJ_SEARCH_KEY:
		break;
	case OBJ_SEARCH_OBJECT:
		break;
	default:
		/* Hash can only work on something with a full key. */
		ast_assert(0);
		return 0;
	}
	return ast_str_hash(key);
}

/*! AO2 sort function for on hold UIDs */
static int uid_hold_sort_fn(const void *obj_left, const void *obj_right, const int flags)
{
	const char *left = obj_left;
	const char *right = obj_right;
	int cmp;

	switch (flags & OBJ_SEARCH_MASK) {
	case OBJ_SEARCH_OBJECT:
	case OBJ_SEARCH_KEY:
		cmp = strcmp(left, right);
		break;
	case OBJ_SEARCH_PARTIAL_KEY:
		cmp = strncmp(left, right, strlen(right));
		break;
	default:
		/* Sort can only work on something with a full or partial key. */
		ast_assert(0);
		cmp = 0;
		break;
	}
	return cmp;
}

static struct ao2_container *pjsip_uids_onhold;

/*!
 * \brief Add a channel ID to the list of PJSIP channels on hold
 *
 * \param chan_uid - Unique ID of the channel being put into the hold list
 *
 * \retval 0 Channel has been added to or was already in the hold list
 * \retval -1 Failed to add channel to the hold list
 */
static int chan_pjsip_add_hold(const char *chan_uid)
{
	RAII_VAR(char *, hold_uid, NULL, ao2_cleanup);

	hold_uid = ao2_find(pjsip_uids_onhold, chan_uid, OBJ_SEARCH_KEY);
	if (hold_uid) {
		/* Device is already on hold. Nothing to do. */
		return 0;
	}

	/* Device wasn't in hold list already. Create a new one. */
	hold_uid = ao2_alloc_options(strlen(chan_uid) + 1, NULL,
		AO2_ALLOC_OPT_LOCK_NOLOCK);
	if (!hold_uid) {
		return -1;
	}

	ast_copy_string(hold_uid, chan_uid, strlen(chan_uid) + 1);

	if (ao2_link(pjsip_uids_onhold, hold_uid) == 0) {
		return -1;
	}

	return 0;
}

/*!
 * \brief Remove a channel ID from the list of PJSIP channels on hold
 *
 * \param chan_uid - Unique ID of the channel being taken out of the hold list
 */
static void chan_pjsip_remove_hold(const char *chan_uid)
{
	ao2_find(pjsip_uids_onhold, chan_uid, OBJ_SEARCH_KEY | OBJ_UNLINK | OBJ_NODATA);
}

/*!
 * \brief Determine whether a channel ID is in the list of PJSIP channels on hold
 *
 * \param chan_uid - Channel being checked
 *
 * \retval 0 The channel is not in the hold list
 * \retval 1 The channel is in the hold list
 */
static int chan_pjsip_get_hold(const char *chan_uid)
{
	RAII_VAR(char *, hold_uid, NULL, ao2_cleanup);

	hold_uid = ao2_find(pjsip_uids_onhold, chan_uid, OBJ_SEARCH_KEY);
	if (!hold_uid) {
		return 0;
	}

	return 1;
}

/*! \brief Function called to get the device state of an endpoint */
static int chan_pjsip_devicestate(const char *data)
{
	RAII_VAR(struct ast_sip_endpoint *, endpoint, ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", data), ao2_cleanup);
	enum ast_device_state state = AST_DEVICE_UNKNOWN;
	RAII_VAR(struct ast_endpoint_snapshot *, endpoint_snapshot, NULL, ao2_cleanup);
	RAII_VAR(struct stasis_cache *, cache, NULL, ao2_cleanup);
	struct ast_devstate_aggregate aggregate;
	int num, inuse = 0;

	if (!endpoint) {
		return AST_DEVICE_INVALID;
	}

	endpoint_snapshot = ast_endpoint_latest_snapshot(ast_endpoint_get_tech(endpoint->persistent),
		ast_endpoint_get_resource(endpoint->persistent));

	if (!endpoint_snapshot) {
		return AST_DEVICE_INVALID;
	}

	if (endpoint_snapshot->state == AST_ENDPOINT_OFFLINE) {
		state = AST_DEVICE_UNAVAILABLE;
	} else if (endpoint_snapshot->state == AST_ENDPOINT_ONLINE) {
		state = AST_DEVICE_NOT_INUSE;
	}

	if (!endpoint_snapshot->num_channels || !(cache = ast_channel_cache())) {
		return state;
	}

	ast_devstate_aggregate_init(&aggregate);

	ao2_ref(cache, +1);

	for (num = 0; num < endpoint_snapshot->num_channels; num++) {
		RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
		struct ast_channel_snapshot *snapshot;

		msg = stasis_cache_get(cache, ast_channel_snapshot_type(),
			endpoint_snapshot->channel_ids[num]);

		if (!msg) {
			continue;
		}

		snapshot = stasis_message_data(msg);

		if (chan_pjsip_get_hold(snapshot->uniqueid)) {
			ast_devstate_aggregate_add(&aggregate, AST_DEVICE_ONHOLD);
		} else {
			ast_devstate_aggregate_add(&aggregate, ast_state_chan2dev(snapshot->state));
		}

		if ((snapshot->state == AST_STATE_UP) || (snapshot->state == AST_STATE_RING) ||
			(snapshot->state == AST_STATE_BUSY)) {
			inuse++;
		}
	}

	if (endpoint->devicestate_busy_at && (inuse == endpoint->devicestate_busy_at)) {
		state = AST_DEVICE_BUSY;
	} else if (ast_devstate_aggregate_result(&aggregate) != AST_DEVICE_INVALID) {
		state = ast_devstate_aggregate_result(&aggregate);
	}

	return state;
}

/*! \brief Function called to query options on a channel */
static int chan_pjsip_queryoption(struct ast_channel *ast, int option, void *data, int *datalen)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for crash, 2015.12.14 */
	if (!channel) {
		ast_log(LOG_ERROR, "Query option error, channel is null");
		return -1;
	}
#endif
	struct ast_sip_session *session = channel->session;
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for crash, 2015.12.14 */
	if (!session) {
		ast_log(LOG_ERROR, "Query option error, session is null");
		return -1;
	}
#endif
	int res = -1;
	enum ast_t38_state state = T38_STATE_UNAVAILABLE;

	switch (option) {
	case AST_OPTION_T38_STATE:
		if (session->endpoint->media.t38.enabled) {
			switch (session->t38state) {
			case T38_LOCAL_REINVITE:
			case T38_PEER_REINVITE:
				state = T38_STATE_NEGOTIATING;
				break;
			case T38_ENABLED:
				state = T38_STATE_NEGOTIATED;
				break;
			case T38_REJECTED:
				state = T38_STATE_REJECTED;
				break;
			default:
				state = T38_STATE_UNKNOWN;
				break;
			}
		}

		*((enum ast_t38_state *) data) = state;
		res = 0;

		break;
	default:
		break;
	}

	return res;
}

static const char *chan_pjsip_get_uniqueid(struct ast_channel *ast)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	char *uniqueid = ast_threadstorage_get(&uniqueid_threadbuf, UNIQUEID_BUFSIZE);

	if (!uniqueid) {
		return "";
	}

	ast_copy_pj_str(uniqueid, &channel->session->inv_session->dlg->call_id->id, UNIQUEID_BUFSIZE);

	return uniqueid;
}

struct indicate_data {
	struct ast_sip_session *session;
	int condition;
	int response_code;
	void *frame_data;
	size_t datalen;
};

static void indicate_data_destroy(void *obj)
{
	struct indicate_data *ind_data = obj;

	ast_free(ind_data->frame_data);
	ao2_ref(ind_data->session, -1);
}

static struct indicate_data *indicate_data_alloc(struct ast_sip_session *session,
		int condition, int response_code, const void *frame_data, size_t datalen)
{
	struct indicate_data *ind_data = ao2_alloc(sizeof(*ind_data), indicate_data_destroy);

	if (!ind_data) {
		return NULL;
	}

	ind_data->frame_data = ast_malloc(datalen);
	if (!ind_data->frame_data) {
		ao2_ref(ind_data, -1);
		return NULL;
	}

	memcpy(ind_data->frame_data, frame_data, datalen);
	ind_data->datalen = datalen;
	ind_data->condition = condition;
	ind_data->response_code = response_code;
	ao2_ref(session, +1);
	ind_data->session = session;

	return ind_data;
}

static int indicate(void *data)
{
	pjsip_tx_data *packet = NULL;
	struct indicate_data *ind_data = data;
	struct ast_sip_session *session = ind_data->session;
	int response_code = ind_data->response_code;

#ifdef GRANDSTREAM_NETWORKS
	/* Begin add by chdruan for Bug 42574 , 2015.7.9 */
	if (response_code == 183 && session->inv_session->options == PJSIP_INV_REQUIRE_100REL) {
		/* pjsip_inv_answer will clear sdp when 100rel is active, then "200 OK" without sdp*/
		session->inv_session->options &= ~PJSIP_INV_REQUIRE_100REL;
		if ((session->inv_session->state != PJSIP_INV_STATE_DISCONNECTED) &&
			(pjsip_inv_answer(session->inv_session, response_code, NULL, NULL, &packet) == PJ_SUCCESS)) {
			ast_sip_session_send_response(session, packet);
		}
		session->inv_session->options &= PJSIP_INV_REQUIRE_100REL;
	} else if ((session->inv_session->state != PJSIP_INV_STATE_DISCONNECTED) &&
		(pjsip_inv_answer(session->inv_session, response_code, NULL, NULL, &packet) == PJ_SUCCESS)) {
		ast_sip_session_send_response(session, packet);
	}
#else
	if ((session->inv_session->state != PJSIP_INV_STATE_DISCONNECTED) &&
		(pjsip_inv_answer(session->inv_session, response_code, NULL, NULL, &packet) == PJ_SUCCESS)) {
		ast_sip_session_send_response(session, packet);
	}
	/* End add by chdruan for Bug 42574 , 2015.7.9 */
#endif

	ao2_ref(ind_data, -1);

	return 0;
}

/*! \brief Send SIP INFO with video update request */
static int transmit_info_with_vidupdate(void *data)
{
	const char * xml =
		"<?xml version=\"1.0\" encoding=\"utf-8\" ?>\r\n"
		" <media_control>\r\n"
		"  <vc_primitive>\r\n"
		"   <to_encoder>\r\n"
		"    <picture_fast_update/>\r\n"
		"   </to_encoder>\r\n"
		"  </vc_primitive>\r\n"
		" </media_control>\r\n";

	const struct ast_sip_body body = {
		.type = "application",
		.subtype = "media_control+xml",
		.body_text = xml
	};

	RAII_VAR(struct ast_sip_session *, session, data, ao2_cleanup);
	struct pjsip_tx_data *tdata;

	if (ast_sip_create_request("INFO", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create text video update INFO request\n");
		return -1;
	}
	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to text video update INFO request\n");
		return -1;
	}
	ast_sip_session_send_request(session, tdata);

	return 0;
}

/*!
 * \internal
 * \brief TRUE if a COLP update can be sent to the peer.
 * \since 13.3.0
 *
 * \param session The session to see if the COLP update is allowed.
 *
 * \retval 0 Update is not allowed.
 * \retval 1 Update is allowed.
 */
static int is_colp_update_allowed(struct ast_sip_session *session)
{
	struct ast_party_id connected_id;
	int update_allowed = 0;

	if (!session->endpoint->id.send_pai && !session->endpoint->id.send_rpid) {
		return 0;
	}

#ifdef GRANDSTREAM_NETWORKS  /* added by yfyang, for Bug60850 */
	if (!strncmp(ast_sorcery_object_get_id(session->endpoint), "trunk_", 6)) {
		return 0;
	}
#endif
	/*
	 * Check if privacy allows the update.  Check while the channel
	 * is locked so we can work with the shallow connected_id copy.
	 */
	ast_channel_lock(session->channel);
	connected_id = ast_channel_connected_effective_id(session->channel);
	if (connected_id.number.valid
		&& (session->endpoint->id.trust_outbound
			|| (ast_party_id_presentation(&connected_id) & AST_PRES_RESTRICTION) == AST_PRES_ALLOWED)) {
		update_allowed = 1;
	}
	ast_channel_unlock(session->channel);

	return update_allowed;
}

/*! \brief Update connected line information */
static int update_connected_line_information(void *data)
{
	struct ast_sip_session *session = data;

	if (ast_channel_state(session->channel) == AST_STATE_UP
		|| session->inv_session->role == PJSIP_ROLE_UAC) {
		if (is_colp_update_allowed(session)) {
			enum ast_sip_session_refresh_method method;
			int generate_new_sdp;

			method = session->endpoint->id.refresh_method;
			if (session->inv_session->invite_tsx
				&& (session->inv_session->options & PJSIP_INV_SUPPORT_UPDATE)) {
				method = AST_SIP_SESSION_REFRESH_METHOD_UPDATE;
			}

			/* Only the INVITE method actually needs SDP, UPDATE can do without */
			generate_new_sdp = (method == AST_SIP_SESSION_REFRESH_METHOD_INVITE);

			ast_sip_session_refresh(session, NULL, NULL, NULL, method, generate_new_sdp);
		}
	} else if (session->endpoint->rpid_immediate
		&& session->inv_session->state != PJSIP_INV_STATE_DISCONNECTED
		&& is_colp_update_allowed(session)) {
		int response_code = 0;

		if (ast_channel_state(session->channel) == AST_STATE_RING) {
			response_code = !session->endpoint->inband_progress ? 180 : 183;
		} else if (ast_channel_state(session->channel) == AST_STATE_RINGING) {
			response_code = 183;
		}

		if (response_code) {
			struct pjsip_tx_data *packet = NULL;

			if (pjsip_inv_answer(session->inv_session, response_code, NULL, NULL, &packet) == PJ_SUCCESS) {
				ast_sip_session_send_response(session, packet);
			}
		}
	}

	ao2_ref(session, -1);
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for video reinivite */
static void copy_reinvite_caps(struct ast_sip_session *session, struct ast_format_cap *caps, struct ast_format_cap *reinvite_caps)
{
	if (!session || !caps || !reinvite_caps) {
		ast_log(LOG_ERROR, "Copy reinvite caps error, param is invalid\n");
		return;
	}
	ast_format_cap_remove_by_type(caps, AST_MEDIA_TYPE_UNKNOWN);
	if (ast_format_cap_count(reinvite_caps)) {
		ast_format_cap_append_from_cap(caps, reinvite_caps, AST_MEDIA_TYPE_UNKNOWN);
	}

	return;
}

static void copy_direct_media_addrs(struct ast_sip_session *session, const struct ast_sip_video_reinvite_data *reinvite_data)
{
	ast_log(LOG_DEBUG, "copy_direct_media_addrs begin\n");

	if (!session || !session->endpoint || !session->inv_session || !session->inv_session->pool_prov) {
		ast_log(LOG_ERROR, "copy_direct_media_addrs session or endpoint is not exist\n");
		return;
	}

	if(!reinvite_data || !reinvite_data->reinvite_session) {
		ast_log(LOG_ERROR, "copy_direct_media_addrs reinvite_data is not exist\n");
		return;
	}

	if (session->endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		ast_log(LOG_DEBUG, "copy_direct_media_addrs endpoint has encryption\n");
		return;
	}
	if (!session->endpoint->media.direct_media.enabled) {
		ast_log(LOG_DEBUG, "copy_direct_media_addrs direct media is not enabled\n");
		return;
	}

	memcpy(&session->direct_media_para, &reinvite_data->reinvite_session->peer_direct_media_para, sizeof(session->direct_media_para));

	return;
}

static void reset_media_encrytion(struct ast_sip_session *session)
{
	if (session->endpoint->media.rtp.use_received_transport) {
		ast_log(LOG_DEBUG, "use received transport, don't reset media encryption\n");
		return;
	}

	RAII_VAR(struct ast_sip_session_media *, audio_session_media, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_session_media *, video_session_media, NULL, ao2_cleanup);

	audio_session_media = ao2_find(session->media, "audio", OBJ_KEY);
	if (audio_session_media) {
		audio_session_media->encryption = session->endpoint->media.rtp.encryption;
	}

	video_session_media = ao2_find(session->media, "video", OBJ_KEY);
	if (video_session_media) {
		video_session_media->encryption = session->endpoint->media.rtp.encryption;
	}

	session->slides_video_media.encryption = session->endpoint->media.rtp.encryption;
	session->fecc_media.encryption = session->endpoint->media.rtp.encryption;

	return;
}

static void copy_yealink_param(struct ast_sip_session *session, const struct ast_sip_video_reinvite_data *reinvite_data)
{
	if (!session || !reinvite_data || !reinvite_data->reinvite_session) {
		ast_log(LOG_ERROR, "copy_yealink_param session or peer session is not exist\n");
		return;
	}

	session->is_yealink_msg = reinvite_data->reinvite_session->is_yealink_msg;

	return;
}

static void send_video_reinvite_request_to_callee(struct ast_sip_session *session, const struct ast_sip_video_reinvite_data *reinvite_data)
{
	ast_log(LOG_DEBUG, "send_video_reinvite_request_to_callee begin\n");
	if (session && session->channel && ast_channel_tech(session->channel) && reinvite_data) {
		if (0 == strcmp(ast_channel_tech(session->channel)->type, "PJSIP")) {
			memcpy(&session->bfcp_para, &reinvite_data->reinvite_session->bfcp_para, sizeof(session->bfcp_para));
			memcpy(&session->main_video_para, &reinvite_data->reinvite_session->main_video_para, sizeof(session->main_video_para));
			memcpy(&session->slides_video_para, &reinvite_data->reinvite_session->slides_video_para, sizeof(session->slides_video_para));
			memcpy(&session->fec_para, &reinvite_data->reinvite_session->fec_para, sizeof(session->fec_para));
			memcpy(&session->fecc_para, &reinvite_data->reinvite_session->fecc_para, sizeof(session->fecc_para));
			session->bfcp_other_attr_count = 0;
			session->main_video_para.other_attr_count = 0;
			session->slides_video_para.other_attr_count = 0;
			session->fec_para.other_attr_count = 0;
			session->fecc_para.other_attr_count = 0;
			int i = 0;
			pj_pool_t *pool = session->inv_session->pool_prov;
			for (i = 0; i < reinvite_data->reinvite_session->bfcp_other_attr_count; i++) {
				session->bfcp_other_attr[session->bfcp_other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
							 reinvite_data->reinvite_session->bfcp_other_attr[i]);
			}
			for (i = 0; i < reinvite_data->reinvite_session->main_video_para.other_attr_count; i++) {
				session->main_video_para.other_attr[session->main_video_para.other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
							 reinvite_data->reinvite_session->main_video_para.other_attr[i]);
			}
			for (i = 0; i < reinvite_data->reinvite_session->slides_video_para.other_attr_count; i++) {
				session->slides_video_para.other_attr[session->slides_video_para.other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
							 reinvite_data->reinvite_session->slides_video_para.other_attr[i]);
			}
			for (i = 0; i < reinvite_data->reinvite_session->fec_para.other_attr_count; i++) {
				session->fec_para.other_attr[session->fec_para.other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
							 reinvite_data->reinvite_session->fec_para.other_attr[i]);
			}
			for (i = 0; i < reinvite_data->reinvite_session->fecc_para.other_attr_count; i++) {
				session->fecc_para.other_attr[session->fecc_para.other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
							 reinvite_data->reinvite_session->fecc_para.other_attr[i]);
			}
			session->bfcp_bandw_count = 0;
			session->main_video_para.bandw_count = 0;
			session->slides_video_para.bandw_count = 0;
			session->fec_para.bandw_count = 0;
			session->fecc_para.bandw_count = 0;
			for (i = 0; i < reinvite_data->reinvite_session->bfcp_bandw_count; i++) {
				session->bfcp_bandw[session->bfcp_bandw_count++] = pjmedia_sdp_bandw_clone(pool, 
					reinvite_data->reinvite_session->bfcp_bandw[i]);
			}
			for (i = 0; i < reinvite_data->reinvite_session->main_video_para.bandw_count; i++) {
				session->main_video_para.bandw[session->main_video_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool, 
					reinvite_data->reinvite_session->main_video_para.bandw[i]);
			}
			for (i = 0; i < reinvite_data->reinvite_session->slides_video_para.bandw_count; i++) {
				session->slides_video_para.bandw[session->slides_video_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool, 
					reinvite_data->reinvite_session->slides_video_para.bandw[i]);
			}
			for (i = 0; i < reinvite_data->reinvite_session->fec_para.bandw_count; i++) {
				session->fec_para.bandw[session->fec_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool, 
					reinvite_data->reinvite_session->fec_para.bandw[i]);
			}
			for (i = 0; i < reinvite_data->reinvite_session->fecc_para.bandw_count; i++) {
				session->fecc_para.bandw[session->fecc_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool, 
					reinvite_data->reinvite_session->fecc_para.bandw[i]);
			}

			memcpy(&session->fec_multiplexed_para, &reinvite_data->reinvite_session->fec_multiplexed_para, sizeof(session->fec_multiplexed_para));

			copy_direct_media_addrs(session, reinvite_data);

			copy_yealink_param(session, reinvite_data);

			copy_reinvite_caps(session, session->req_caps, reinvite_data->reinvite_session->req_caps);
			copy_reinvite_caps(session, session->slides_video_req_caps, reinvite_data->reinvite_session->slides_video_req_caps);
			copy_reinvite_caps(session, session->fec_req_caps, reinvite_data->reinvite_session->fec_req_caps);
			copy_reinvite_caps(session, session->fecc_req_caps, reinvite_data->reinvite_session->fecc_req_caps);
			copy_reinvite_caps(session, session->fec_multiplexed_req_caps, reinvite_data->reinvite_session->fec_multiplexed_req_caps);

			reset_media_encrytion(session);

			memcpy(&session->reinvite_attr, &reinvite_data->reinvite_session->reinvite_attr, sizeof(session->reinvite_attr));
			reinvite_data->reinvite_session->reinvite_attr.reinvite_attr_audio= REINVITE_SENDRECV;
			reinvite_data->reinvite_session->reinvite_attr.reinvite_attr_m_video = REINVITE_SENDRECV;
			reinvite_data->reinvite_session->reinvite_attr.reinvite_attr_s_video = REINVITE_SENDRECV;
			reinvite_data->reinvite_session->reinvite_attr.reinvite_attr_fec = REINVITE_SENDRECV;
			reinvite_data->reinvite_session->reinvite_attr.reinvite_attr_fecc = REINVITE_SENDRECV;
			reinvite_data->reinvite_session->reinvite_attr.reinvite_attr_bfcp= REINVITE_SENDRECV;

			ao2_ref(session, +1);
			session->is_outgoing = 1;
			if (ast_sip_push_task(session->serializer, send_video_reinvite_request, session)) {
				ao2_cleanup(session);
			}
			ast_log(LOG_DEBUG, "send video re-invite to peer\n");
		} else {
			ast_log(LOG_WARNING, "channel type is not SIP, can not send video reinvte to callee\n");
		}
	} else {
		ast_log(LOG_WARNING, "can not send video re-invite to peer\n");
	}

	return;
}

static void modify_local_sdp_attr_with_callee_answer( pj_pool_t *pool,
				  pjmedia_sdp_session *caller_local,
				  const pjmedia_sdp_session *callee_answer,
				  const struct ast_sip_session *local_session,
				  const struct ast_sip_session *peer_session
				  )
{
	pjmedia_sdp_session *result;
	char media_used[PJMEDIA_MAX_SDP_MEDIA];
	unsigned i, j;
	pjmedia_sdp_media *im;

	result = pjmedia_sdp_session_clone(pool, caller_local);

	result->media_count = 0;

	pj_bzero(media_used, sizeof(media_used));
	for (im=NULL, j=0; j<caller_local->media_count; ++j) {
		const pjmedia_sdp_media *om;

		im = caller_local->media[j];

		for (i=0; i<callee_answer->media_count; ++i) {
			om = callee_answer->media[i];
			char media_type[16];
			memset(media_type, 0, sizeof(media_type));
			ast_copy_pj_str(media_type, &om->desc.media, sizeof(media_type));

			if (pj_strcmp(&om->desc.media, &im->desc.media)==0 &&
					judge_if_transport_compatible(local_session, im) &&
					judge_if_transport_compatible(peer_session, om) &&
					media_used[i] == 0) {
				if (strncmp(media_type, "audio", 5) == 0) {
					enum ast_reinvite_attr flag = REINVITE_SENDRECV;
					if (pjmedia_sdp_media_find_attr2(om, "sendonly", NULL)) {
						flag = REINVITE_SENDONLY;
					} else if (pjmedia_sdp_media_find_attr2(om, "inactive", NULL)) {
						flag = REINVITE_INACTIVE;
					} else if (pjmedia_sdp_media_find_attr2(om, "recvonly", NULL)) {
						flag = REINVITE_RECVONLY;
					}

					if (flag != REINVITE_SENDRECV) {
						pjmedia_sdp_attr *attr;
						pjmedia_sdp_media_remove_all_attr(im, "inactive");
						pjmedia_sdp_media_remove_all_attr(im, "sendrecv");
						pjmedia_sdp_media_remove_all_attr(im, "sendonly");
						pjmedia_sdp_media_remove_all_attr(im, "recvonly");

						switch (flag) {
							case REINVITE_SENDONLY:
								attr = pjmedia_sdp_attr_create(pool, "sendonly", NULL);
								break;
							case REINVITE_INACTIVE:
								attr = pjmedia_sdp_attr_create(pool, "inactive", NULL);
								break;
							case REINVITE_RECVONLY:
								attr = pjmedia_sdp_attr_create(pool, "recvonly", NULL);
								break;
							default:
								attr = pjmedia_sdp_attr_create(pool, "sendrecv", NULL);
								break;
						}
						pjmedia_sdp_media_add_attr(im, attr);
					}
				}
			}
		}
	}
}

static void send_video_reinvite_response_to_caller(struct ast_sip_session *session, const struct ast_sip_video_reinvite_data *reinvite_data)
{
	ast_log(LOG_DEBUG, "send_video_reinvite_response_to_caller begin\n");
	if (session && session->channel && ast_channel_tech(session->channel) && reinvite_data) {
		if (0 == strcmp(ast_channel_tech(session->channel)->type, "PJSIP")) {

			struct ast_sip_channel_pvt *chan_pvt;
			pjmedia_sdp_neg_state chan_neg_state;

			chan_pvt = ast_channel_tech_pvt(session->channel);
			if (!chan_pvt || !chan_pvt->session || !chan_pvt->session->inv_session || !chan_pvt->session->inv_session->neg) {
				ast_log(LOG_WARNING, "send reinvite response chan_pvt session is invalid\n");
				return;
			}
			chan_neg_state = chan_pvt->session->inv_session->neg->state;

			if (chan_neg_state == PJMEDIA_SDP_NEG_STATE_WAIT_NEGO)
			{
				if (!chan_pvt->session->inv_session->neg->neg_local_sdp) {
					ast_log(LOG_WARNING, "send reinvite response chan_pvt neg_local_sdp is invalid\n");
					return;
				}
				if (!reinvite_data || !reinvite_data->reinvite_session ||
					!reinvite_data->reinvite_session->inv_session ||
					!reinvite_data->reinvite_session->inv_session->neg ||
					!reinvite_data->reinvite_session->inv_session->neg->active_local_sdp ||
					!reinvite_data->reinvite_session->inv_session->neg->active_remote_sdp) {
					ast_log(LOG_WARNING, "send reinvite response peer_pvt active local/remote sdp is invalid\n");
					return;
				}
				ast_log(LOG_DEBUG, "send_video_reinvite_response_to_caller need future neg\n");
				pjmedia_sdp_session *callee_answer_sdp, *caller_local_sdp, *callee_sdp_local;
				pj_pool_t *pool = chan_pvt->session->inv_session->pool_prov;

				copy_direct_media_addrs(session, reinvite_data);

				callee_sdp_local = reinvite_data->reinvite_session->inv_session->neg->active_local_sdp;
				caller_local_sdp = chan_pvt->session->inv_session->neg->neg_local_sdp;
				callee_answer_sdp = reinvite_data->reinvite_session->inv_session->neg->active_remote_sdp;

				char buf[1024];
				{
					memset(buf, 0, sizeof(buf));
					pjmedia_sdp_print(caller_local_sdp, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
					ast_log(LOG_CODEC_NEG, "[caller] caller reinvite local sdp is:\n%s", buf);

					memset(buf, 0, sizeof(buf));
					pjmedia_sdp_print(callee_answer_sdp, buf, sizeof(buf) - 1);// Must use sizeof(buf) - 1, otherwise there is a potential overflow
					ast_log(LOG_CODEC_NEG, "[callee] UCM <<<=== B . received reinvite 200 ok from the callee\n%s", buf);

					memset(buf, 0, sizeof(buf));
					pjmedia_sdp_print(callee_sdp_local, buf, sizeof(buf) - 1);// Must use sizeof(buf) - 1, otherwise there is a potential overflow
					ast_log(LOG_CODEC_NEG, "[callee] reinvite local sdp of the callee is:\n%s", buf);
				}

				ast_log(LOG_CODEC_NEG, "[caller] modify caller reinvite local sdp according to callee answer sdp!");

				/* Add by bfdu for modify yealink profile-level-id value, 2015.12.02 */
				chan_pvt->session->is_yealink_msg = reinvite_data->reinvite_session->is_yealink_msg;

			    chan_pvt->session->inv_session->neg->neg_local_sdp = modify_local_sdp_with_callee_answer(pool, caller_local_sdp, callee_sdp_local, chan_pvt->session, reinvite_data->reinvite_session);
				/* Add by bfdu for canreinvite, 2016.01.19 */
				if (chan_pvt->session->is_video_media_change) {
					reset_caller_media_sdp_session_addr(reinvite_data->reinvite_session, chan_pvt->session, chan_pvt->session->inv_session->neg->neg_local_sdp, callee_answer_sdp);
				}
				/* added by jklou for fec/fecc/bfcp attribute passthrough */

				chan_pvt->session->inv_session->neg->neg_local_sdp = set_app_attr_local_remote_sdp(pool, chan_pvt->session->inv_session->neg->neg_local_sdp,
					callee_answer_sdp, chan_pvt->session, reinvite_data->reinvite_session);
				/* ended */
				modify_local_sdp_attr_with_callee_answer(pool, chan_pvt->session->inv_session->neg->neg_local_sdp, callee_answer_sdp, chan_pvt->session, reinvite_data->reinvite_session);

				{
					memset(buf, 0, sizeof(buf));
					pjmedia_sdp_print(chan_pvt->session->inv_session->neg->neg_local_sdp, buf, sizeof(buf) - 1);// Must use sizeof(buf) - 1, otherwise there is a potential overflow
					ast_log(LOG_CODEC_NEG, "[caller] caller reinvite local sdp is:\n%s", buf);
				}
            }
			ao2_ref(session, +1);
			ast_channel_unlock(session->channel);
			if (ast_sip_push_task_synchronous(session->serializer, send_video_reinvite_response, session)) {
				ast_log(LOG_WARNING, "Unable to push send_video_reinvite_response task to the threadpool.\n");
				ao2_ref(session, -1);
				ast_channel_lock(session->channel);
				return;
			}
			ast_channel_lock(session->channel);
			ast_log(LOG_CODEC_NEG, "send video re-invite rsp to peer\n");
		} else {
			ast_log(LOG_WARNING, "channel type is not PJSIP, can not send video reinvte rsp to callee\n");
		}
	} else {
		ast_log(LOG_WARNING, "can not send video re-invite rsp to peer\n");
	}

	return;
}

static int send_video_reinvite_request(void *data)
{
	ast_log(LOG_DEBUG, "send_video_reinvite_request begin\n");
	struct ast_sip_session *session = data;
	session->is_video_media_change = 1;
	ast_log(LOG_DEBUG, "send_video_reinvite_request set is_video_media_change true\n");
	ast_sip_session_refresh(session, NULL, NULL, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1);
	ao2_ref(session, -1);

	return 0;
}

static int send_video_reinvite_response(void *data)
{
	ast_log(LOG_DEBUG, "send_video_reinvite_response begin\n");
	struct ast_sip_session *session = data;
	session->is_video_media_change = 0;
	ast_log(LOG_DEBUG, "send_video_reinvite_response set is_video_media_change False\n");
	ast_sip_session_send_reinvite_response(session);
	ao2_ref(session, -1);

	return 0;
}

static void ast_sip_session_send_reinvite_response(struct ast_sip_session *session)
{
	pj_status_t status = PJ_SUCCESS;
	pjsip_tx_data *packet = NULL;

	if (session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_CODEC_NEG, "Session is terminated\n");
		return;
	}

	if (strcmp(ast_channel_tech(session->channel)->type, "PJSIP")) {
		ast_log(LOG_CODEC_NEG, "channel's type is not pjsip\n");
		return;
    }

	pjsip_dlg_inc_lock(session->inv_session->dlg);
	if (session->inv_session->invite_tsx) {
		status = pjsip_inv_answer(session->inv_session, 200, NULL, NULL, &packet);
		if (status == PJ_SUCCESS) {
			if (session->inv_session->neg && session->inv_session->neg->active_local_sdp) {
				char buf[1024] = {0};
				pjmedia_sdp_session* caller_answer_sdp = session->inv_session->neg->active_local_sdp;
				pjmedia_sdp_print(caller_answer_sdp, buf, sizeof(buf) - 1);// Must use sizeof(buf) - 1, otherwise there is a potential overflow
				ast_log(LOG_CODEC_NEG, "[caller] A <<<=== UCM. send reinvite 200 ok sdp to the caller!\n%s", buf);
			}
		}
	} else {
		ast_log(LOG_ERROR,"Cannot answer reinvite '%s' because there is no associated SIP transaction\n",
			ast_channel_name(session->channel));
	}
	pjsip_dlg_dec_lock(session->inv_session->dlg);

	if (status == PJ_SUCCESS && packet) {
		ast_sip_session_send_response(session, packet);
	}

	return;
}

static void handle_response_video_reinvite(struct ast_sip_session *session, struct ast_sip_video_reinvite_data *reinvite_data)
{
	ast_log(LOG_DEBUG, "handle_response_video_reinvite start\n");
	if (session && session->channel && ast_channel_tech(session->channel) && reinvite_data) {
		if (0 == strcmp(ast_channel_tech(session->channel)->type, "PJSIP")) {
			memcpy(&reinvite_data->reinvite_session, &session, sizeof(reinvite_data->reinvite_session));
			ast_log(LOG_DEBUG, "handle video re-invite rsp\n");
		}
		else {
			ast_log(LOG_WARNING, "channel type is not SIP, can not handle video reinvite req\n");
		}
	} else {
		ast_log(LOG_WARNING, "param illegal, can not handle video re-invite\n");
	}
	return;
}

static unsigned char judge_if_support_video_media(struct ast_sip_session *session)
{
	ast_log(LOG_DEBUG, "judge_if_support_video_media begin\n");
	unsigned char res = FALSE;

	RAII_VAR(struct ast_sip_session_media *, session_media, NULL, ao2_cleanup);

	session_media = ao2_find(session->media, "video", OBJ_KEY);
	if (!session_media) {
		/* if the session_media doesn't exist, there weren't
		 * any handlers at the time of its creation */
		return FALSE;
	}

	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate %s capabilities\n", session_media->stream_type);
		res = FALSE;
	}

	ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_VIDEO);

	if (!ast_format_cap_count(caps)) {
		ast_log(LOG_WARNING, "endpoint has no video cap\n");
		res = FALSE;
	} else {
		res = TRUE;
	}

	return res;
}

static unsigned char judge_if_both_type_is_sip(struct ast_channel* chan)
{
	if (NULL == chan){
		ast_log(LOG_WARNING, "chan is NULL");
		return FALSE;
	}
	RAII_VAR(struct ast_channel *, bridged, ast_channel_bridge_peer(chan), ast_channel_cleanup);
	if (NULL == bridged) {
		ast_log(LOG_DEBUG, "bridge is not exsit");
		return FALSE;
	}
	if (ast_channel_tech(bridged) && (0 == strcmp(ast_channel_tech(bridged)->type, "PJSIP")) &&
		ast_channel_tech(chan) && (0 == strcmp(ast_channel_tech(chan)->type, "PJSIP"))) {
		return TRUE;
	} else {
		ast_log(LOG_DEBUG, "type is not PJSIP");
		return FALSE;
	}
}

static void modify_early_answer_sdp(struct ast_channel *chan, const struct ast_channel *peer)
{
	ast_log(LOG_DEBUG, "modify_early_answer_sdp begin\n");
	if (peer) {
		if (0 == strcmp(ast_channel_tech(peer)->type, "PJSIP")) {
			struct ast_sip_channel_pvt *chan_pvt, *peer_pvt;
			pjmedia_sdp_neg_state chan_neg_state, peer_neg_state;

			chan_pvt = ast_channel_tech_pvt(chan);
			peer_pvt = ast_channel_tech_pvt(peer);

			if (!chan_pvt || !chan_pvt->session || !chan_pvt->session->inv_session || !chan_pvt->session->inv_session->neg) {
				ast_log(LOG_WARNING, "modify early answer chan_pvt session is invalid\n");
				return;
			}
			if (!peer_pvt || !peer_pvt->session || !peer_pvt->session->inv_session || !peer_pvt->session->inv_session->neg) {
				ast_log(LOG_WARNING, "modify early answer peer_pvt session is invalid\n");
				return;
			}

			chan_neg_state = chan_pvt->session->inv_session->neg->state;
			peer_neg_state = peer_pvt->session->inv_session->neg->state;

			ast_log(LOG_DEBUG, "modify_early_answer_sdp chan_neg_state:%d, peer_neg_state:%d\n", chan_neg_state, peer_neg_state);

			if (chan_neg_state == PJMEDIA_SDP_NEG_STATE_WAIT_NEGO && peer_neg_state == PJMEDIA_SDP_NEG_STATE_DONE)
			{
				if (!chan_pvt->session->inv_session->neg->neg_local_sdp) {
					ast_log(LOG_WARNING, "modify early answer chan_pvt neg_local_sdp is invalid\n");
					return;
				}
				if (!peer_pvt->session->inv_session->neg->active_local_sdp ||
					!peer_pvt->session->inv_session->neg->active_remote_sdp) {
					ast_log(LOG_WARNING, "modify early answer peer_pvt active local/remote sdp is invalid\n");
					return;
				}
				ast_log(LOG_DEBUG, "modify_early_answer_sdp neg begin\n");
				pjmedia_sdp_session *callee_answer_sdp, *caller_local_sdp, *callee_sdp_local;
				pj_pool_t *pool = chan_pvt->session->inv_session->pool_prov;

				callee_sdp_local = peer_pvt->session->inv_session->neg->active_local_sdp;
				caller_local_sdp = chan_pvt->session->inv_session->neg->neg_local_sdp;
				callee_answer_sdp = peer_pvt->session->inv_session->neg->active_remote_sdp;

				char buf[1024] = {0};
				{
					memset(buf, 0, sizeof(buf));
					pjmedia_sdp_print(caller_local_sdp, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
					ast_log(LOG_CODEC_NEG, "[caller] caller early media(183) local sdp is:\n%s", buf);

					memset(buf, 0, sizeof(buf));
					pjmedia_sdp_print(callee_answer_sdp, buf, sizeof(buf) - 1);// Must use sizeof(buf) - 1, otherwise there is a potential overflow
					ast_log(LOG_CODEC_NEG, "[callee] UCM <<<=== B . received 183 from the callee\n%s", buf);

					memset(buf, 0, sizeof(buf));
					pjmedia_sdp_print(callee_sdp_local, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
					ast_log(LOG_CODEC_NEG, "[callee] early media(183) local sdp of the callee is:\n%s", buf);
				}

				ast_log(LOG_CODEC_NEG, "[caller] modify caller early media(183) local sdp according to callee answer sdp!");

				/* Add by bfdu for modify yealink profile-level-id value, 2015.12.02 */
				chan_pvt->session->is_yealink_msg = peer_pvt->session->is_yealink_msg;

			    chan_pvt->session->inv_session->neg->neg_local_sdp = modify_local_sdp_with_callee_answer(pool, caller_local_sdp, callee_sdp_local, chan_pvt->session, peer_pvt->session);
				/* added by jklou for fec/fecc/bfcp attribute passthrough */

				chan_pvt->session->inv_session->neg->neg_local_sdp = set_app_attr_local_remote_sdp(pool, chan_pvt->session->inv_session->neg->neg_local_sdp,
					callee_answer_sdp, chan_pvt->session, peer_pvt->session);
				/* ended */
				{
					memset(buf, 0, sizeof(buf));
					pjmedia_sdp_print(chan_pvt->session->inv_session->neg->neg_local_sdp, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
					ast_log(LOG_CODEC_NEG, "[caller] caller early media(183) local sdp is:\n%s", buf);
				}

				set_bfcp_tcp_peer_remote_addr(chan_pvt->session, peer_pvt->session);
			}
		}
	}
	return;
}
#endif

/*! \brief Function called by core to ask the channel to indicate some sort of condition */
static int chan_pjsip_indicate(struct ast_channel *ast, int condition, const void *data, size_t datalen)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct chan_pjsip_pvt *pvt = channel->pvt;
	struct ast_sip_session_media *media;
	int response_code = 0;
	int res = 0;
	char *device_buf;
	size_t device_buf_size;

	switch (condition) {
	case AST_CONTROL_RINGING:
#ifdef GRANDSTREAM_NETWORKS
		if (ast_channel_state(ast) == AST_STATE_RING && !ast_channel_cHasSend183(ast)) {
#else
		if (ast_channel_state(ast) == AST_STATE_RING) {
#endif
			if (channel->session->endpoint->inband_progress) {
				response_code = 183;
				res = -1;
			} else {
				response_code = 180;
			}
		} else {
			res = -1;
		}
		ast_devstate_changed(AST_DEVICE_UNKNOWN, AST_DEVSTATE_CACHABLE, "PJSIP/%s", ast_sorcery_object_get_id(channel->session->endpoint));
		break;
	case AST_CONTROL_BUSY:
		if (ast_channel_state(ast) != AST_STATE_UP) {
			response_code = 486;
		} else {
			res = -1;
		}
		break;
	case AST_CONTROL_CONGESTION:
#ifdef GRANDSTREAM_NETWORKS
		/*Begin added by zhywang . See asterisk 1.8.23 sip_indicate */
		if(ast_channel_state(ast) == AST_STATE_RING){
			ast_log(LOG_NOTICE, "When CONGESTION, channel is RING, response 480");
			response_code = 480;
		}
		/*End added by zhywang */
		else if (ast_channel_state(ast) != AST_STATE_UP) {
			response_code = 503;
		} else {
			res = -1;
		}
#else
		if (ast_channel_state(ast) != AST_STATE_UP) {
			response_code = 503;
		} else {
			res = -1;
		}
#endif
		break;
	case AST_CONTROL_INCOMPLETE:
		if (ast_channel_state(ast) != AST_STATE_UP) {
			response_code = 484;
		} else {
			res = -1;
		}
		break;
	case AST_CONTROL_PROCEEDING:
		if (ast_channel_state(ast) != AST_STATE_UP) {
			response_code = 100;
		} else {
			res = -1;
		}
		break;
	case AST_CONTROL_PROGRESS:
		if (ast_channel_state(ast) != AST_STATE_UP) {
			response_code = 183;
#ifdef GRANDSTREAM_NETWORKS
			/* Begin add by bfdu for 183 neg, 2015.09.29 */
			const struct ast_channel *peer_chan = data;
			modify_early_answer_sdp(ast, peer_chan);
			/* End add by bfdu for 183 neg, 2015.09.29 */
			ast_channel_cHasSend183_set(ast,1);
#endif
		} else {
			res = -1;
		}
		break;
	case AST_CONTROL_VIDUPDATE:
		media = pvt->media[SIP_MEDIA_VIDEO];
		if (media && media->rtp) {
			/* FIXME: Only use this for VP8. Additional work would have to be done to
			 * fully support other video codecs */

			if (ast_format_cap_iscompatible_format(ast_channel_nativeformats(ast), ast_format_vp8) != AST_FORMAT_CMP_NOT_EQUAL) {
				/* FIXME Fake RTP write, this will be sent as an RTCP packet. Ideally the
				 * RTP engine would provide a way to externally write/schedule RTCP
				 * packets */
				struct ast_frame fr;
				fr.frametype = AST_FRAME_CONTROL;
				fr.subclass.integer = AST_CONTROL_VIDUPDATE;
				res = ast_rtp_instance_write(media->rtp, &fr);
			} else {
				ao2_ref(channel->session, +1);

				if (ast_sip_push_task(channel->session->serializer, transmit_info_with_vidupdate, channel->session)) {
					ao2_cleanup(channel->session);
				}
			}
			ast_test_suite_event_notify("AST_CONTROL_VIDUPDATE", "Result: Success");
		} else {
			ast_test_suite_event_notify("AST_CONTROL_VIDUPDATE", "Result: Failure");
			res = -1;
		}
		break;
	case AST_CONTROL_CONNECTED_LINE:
		ao2_ref(channel->session, +1);
		if (ast_sip_push_task(channel->session->serializer, update_connected_line_information, channel->session)) {
			ao2_cleanup(channel->session);
		}
		break;
	case AST_CONTROL_UPDATE_RTP_PEER:
		break;
	case AST_CONTROL_PVT_CAUSE_CODE:
		res = -1;
		break;
	case AST_CONTROL_MASQUERADE_NOTIFY:
		ast_assert(datalen == sizeof(int));
		if (*(int *) data) {
			/*
			 * Masquerade is beginning:
			 * Wait for session serializer to get suspended.
			 */
			ast_channel_unlock(ast);
			ast_sip_session_suspend(channel->session);
			ast_channel_lock(ast);
		} else {
			/*
			 * Masquerade is complete:
			 * Unsuspend the session serializer.
			 */
			ast_sip_session_unsuspend(channel->session);
		}
		break;
	case AST_CONTROL_HOLD:
		chan_pjsip_add_hold(ast_channel_uniqueid(ast));
		device_buf_size = strlen(ast_channel_name(ast)) + 1;
		device_buf = alloca(device_buf_size);
		ast_channel_get_device_name(ast, device_buf, device_buf_size);
		ast_devstate_changed_literal(AST_DEVICE_ONHOLD, 1, device_buf);
#ifdef GRANDSTREAM_NETWORKS
        ast_log(LOG_DEBUG, "Channel: %s, Call Type: %s\n",
                           ast_channel_name(ast), S_OR(pvt->call_type, "NULL"));
        // Skip Music on hold in a conference
        if (pvt->call_type == NULL ||
            strcmp(pvt->call_type, "Conference") != 0) {
    		ast_moh_start(ast, data, NULL);
        }
#else
		ast_moh_start(ast, data, NULL);
#endif
		break;
	case AST_CONTROL_UNHOLD:
		chan_pjsip_remove_hold(ast_channel_uniqueid(ast));
		device_buf_size = strlen(ast_channel_name(ast)) + 1;
		device_buf = alloca(device_buf_size);
		ast_channel_get_device_name(ast, device_buf, device_buf_size);
		ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, 1, device_buf);
		ast_moh_stop(ast);
		break;
	case AST_CONTROL_SRCUPDATE:
		break;
	case AST_CONTROL_SRCCHANGE:
#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for bfcp/fec/fecc, Grandstream Networks, 2015.08.08 */
		ao2_ref(channel->session, +1);
		struct ast_sip_session_media *audio_media;
		audio_media = pvt->media[SIP_MEDIA_AUDIO];
		if (audio_media && audio_media->rtp) {
			ast_log(LOG_DEBUG, "audio change source\n");
			ast_rtp_instance_change_source(audio_media->rtp);
		}
		struct ast_sip_session_media *video_media;
		video_media = pvt->media[SIP_MEDIA_VIDEO];
		if (video_media && video_media->rtp) {
			ast_log(LOG_DEBUG, "main video change source\n");
			ast_rtp_instance_change_source(video_media->rtp);
		}

		if (channel->session->slides_video_media.rtp) {
			ast_log(LOG_DEBUG, "slides video change source\n");
			ast_rtp_instance_change_source(channel->session->slides_video_media.rtp);
		}
		if (channel->session->fecc_media.rtp) {
			ast_log(LOG_DEBUG, "fecc change source\n");
			ast_rtp_instance_change_source(channel->session->fecc_media.rtp);
		}
		if (channel->session->fec_media.rtp) {
			ast_log(LOG_DEBUG, "fec change source\n");
			ast_rtp_instance_change_source(channel->session->fec_media.rtp);
			media = pvt->media[SIP_MEDIA_VIDEO];
			if (media && media->rtp) {
				ast_rtp_instance_set_ssrc(channel->session->fec_media.rtp, media->rtp);
			}
		}
		ao2_ref(channel->session, -1);
#endif
		break;
	case AST_CONTROL_REDIRECTING:
		if (ast_channel_state(ast) != AST_STATE_UP) {
			response_code = 181;
		} else {
			res = -1;
		}
		break;
	case AST_CONTROL_T38_PARAMETERS:
		res = 0;

		if (channel->session->t38state == T38_PEER_REINVITE) {
			const struct ast_control_t38_parameters *parameters = data;

			if (parameters->request_response == AST_T38_REQUEST_PARMS) {
				res = AST_T38_REQUEST_PARMS;
			}
		}

		break;
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for video reinivite */
	case AST_CONTROL_VIDEO_REINVITE_REQ:
		if (datalen != sizeof(struct ast_sip_video_reinvite_data)) {
			res = -1;
			ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_VIDEO_REINVITE_REQ. Expected %d, got %d\n", (int) sizeof(struct ast_sip_video_reinvite_data), (int) datalen);
		} else {
			if ((FALSE == judge_if_support_video_media(channel->session)) ||
			    (FALSE == judge_if_both_type_is_sip(ast))) {
				struct ast_sip_video_reinvite_data video_not_support_param;
				video_not_support_param.reinvite_session = channel->session;
				ast_format_cap_remove_by_type(video_not_support_param.reinvite_session->req_caps, AST_MEDIA_TYPE_VIDEO);
				ast_queue_control_data(ast, AST_CONTROL_VIDEO_REINVITE_RSP, &video_not_support_param, sizeof(video_not_support_param));
				ast_log(LOG_DEBUG, "handle video re-invite frame, peer do not support video media\n");
			} else if (1 == channel->session->is_video_media_change) {
				ast_log(LOG_DEBUG, "It's dealing with video reinvite now, reject this request");
				ast_format_cap_remove_by_type(channel->session->slides_video_req_caps, AST_MEDIA_TYPE_VIDEO);
				struct ast_sip_video_reinvite_data reinvite_conflict_data;
				reinvite_conflict_data.reinvite_session = channel->session;
				/* Now we just deal case that GVC open slides video reinvite conflict.
				   Todo: need to deal with other case.
				*/
				ast_queue_control_data(ast, AST_CONTROL_VIDEO_REINVITE_RSP, &reinvite_conflict_data, sizeof(reinvite_conflict_data));
				send_video_reinvite_response_to_caller(channel->session, &reinvite_conflict_data);
			} else {
				const struct ast_sip_video_reinvite_data *req_data = data;
				send_video_reinvite_request_to_callee(channel->session, req_data);
			}
		}
		break;
	case AST_CONTROL_VIDEO_REINVITE_RSP:
		if (datalen != sizeof(struct ast_sip_video_reinvite_data)) {
			res = -1;
			ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_VIDEO_REINVITE_RSP. Expected %d, got %d\n", (int) sizeof(struct ast_sip_video_reinvite_data), (int) datalen);
		} else {
			if (TRUE == channel->session->is_video_media_change) {
				const struct ast_sip_video_reinvite_data *rsp_data = data;
				send_video_reinvite_response_to_caller(channel->session, rsp_data);
			} else {
				ast_log(LOG_DEBUG, "Don't need to send video re-invite rsp\n");
			}
		}
		break;
	case AST_CONTROL_BFCP_SET_REMOTE_ADDRS:
		if (datalen != sizeof(struct ast_sockaddr)) {
			res = -1;
			ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_BFCP_SET_REMOTE_ADDRS. Expected %d, got %d\n", (int) sizeof(struct ast_frame), (int) datalen);
		} else {
			const struct ast_sockaddr *addrs = data;
			ast_log(LOG_DEBUG, "chan_pjsip_indicate set bfcp old remote at port %s\n", ast_sockaddr_stringify(addrs));
			ast_bfcp_tcp_set_peer_remote(channel->session->bfcp_tcp, addrs);
		}
		break;
	case AST_CONTROL_REINVITE_BYE_CONFLICT:
		ao2_ref(channel->session, +1);
		if (ast_sip_push_task(channel->session->serializer, send_video_reinvite_response, channel->session)) {
			ao2_cleanup(channel->session);
		}
		break;
#endif
	case -1:
		res = -1;
		break;
	default:
		ast_log(LOG_WARNING, "Don't know how to indicate condition %d\n", condition);
		res = -1;
		break;
	}

	if (response_code) {
		struct indicate_data *ind_data = indicate_data_alloc(channel->session, condition, response_code, data, datalen);
		if (!ind_data || ast_sip_push_task(channel->session->serializer, indicate, ind_data)) {
			ast_log(LOG_NOTICE, "Cannot send response code %d to endpoint %s. Could not queue task properly\n",
					response_code, ast_sorcery_object_get_id(channel->session->endpoint));
			ao2_cleanup(ind_data);
			res = -1;
		}
	}

	return res;
}

struct transfer_data {
	struct ast_sip_session *session;
	char *target;
};

static void transfer_data_destroy(void *obj)
{
	struct transfer_data *trnf_data = obj;

	ast_free(trnf_data->target);
	ao2_cleanup(trnf_data->session);
}

static struct transfer_data *transfer_data_alloc(struct ast_sip_session *session, const char *target)
{
	struct transfer_data *trnf_data = ao2_alloc(sizeof(*trnf_data), transfer_data_destroy);

	if (!trnf_data) {
		return NULL;
	}

	if (!(trnf_data->target = ast_strdup(target))) {
		ao2_ref(trnf_data, -1);
		return NULL;
	}

	ao2_ref(session, +1);
	trnf_data->session = session;

	return trnf_data;
}

static void transfer_redirect(struct ast_sip_session *session, const char *target)
{
	pjsip_tx_data *packet;
	enum ast_control_transfer message = AST_TRANSFER_SUCCESS;
	pjsip_contact_hdr *contact;
	pj_str_t tmp;

	if (pjsip_inv_end_session(session->inv_session, 302, NULL, &packet) != PJ_SUCCESS) {
		ast_log(LOG_WARNING, "Failed to redirect PJSIP session for channel %s\n",
			ast_channel_name(session->channel));
		message = AST_TRANSFER_FAILED;
		ast_queue_control_data(session->channel, AST_CONTROL_TRANSFER, &message, sizeof(message));

		return;
	}

	if (!(contact = pjsip_msg_find_hdr(packet->msg, PJSIP_H_CONTACT, NULL))) {
		contact = pjsip_contact_hdr_create(packet->pool);
	}

	pj_strdup2_with_null(packet->pool, &tmp, target);
	if (!(contact->uri = pjsip_parse_uri(packet->pool, tmp.ptr, tmp.slen, PJSIP_PARSE_URI_AS_NAMEADDR))) {
		ast_log(LOG_WARNING, "Failed to parse destination URI '%s' for channel %s\n",
			target, ast_channel_name(session->channel));
		message = AST_TRANSFER_FAILED;
		ast_queue_control_data(session->channel, AST_CONTROL_TRANSFER, &message, sizeof(message));
		pjsip_tx_data_dec_ref(packet);

		return;
	}
	pjsip_msg_add_hdr(packet->msg, (pjsip_hdr *) contact);

	ast_sip_session_send_response(session, packet);
	ast_queue_control_data(session->channel, AST_CONTROL_TRANSFER, &message, sizeof(message));
}

static void transfer_refer(struct ast_sip_session *session, const char *target)
{
	pjsip_evsub *sub;
	enum ast_control_transfer message = AST_TRANSFER_SUCCESS;
	pj_str_t tmp;
	pjsip_tx_data *packet;

	if (pjsip_xfer_create_uac(session->inv_session->dlg, NULL, &sub) != PJ_SUCCESS) {
		message = AST_TRANSFER_FAILED;
		ast_queue_control_data(session->channel, AST_CONTROL_TRANSFER, &message, sizeof(message));

		return;
	}

	if (pjsip_xfer_initiate(sub, pj_cstr(&tmp, target), &packet) != PJ_SUCCESS) {
		message = AST_TRANSFER_FAILED;
		ast_queue_control_data(session->channel, AST_CONTROL_TRANSFER, &message, sizeof(message));
		pjsip_evsub_terminate(sub, PJ_FALSE);

		return;
	}

	pjsip_xfer_send_request(sub, packet);
	ast_queue_control_data(session->channel, AST_CONTROL_TRANSFER, &message, sizeof(message));
}

static int transfer(void *data)
{
	struct transfer_data *trnf_data = data;
	struct ast_sip_endpoint *endpoint = NULL;
	struct ast_sip_contact *contact = NULL;
	const char *target = trnf_data->target;

	/* See if we have an endpoint; if so, use its contact */
	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", target);
	if (endpoint) {
		contact = ast_sip_location_retrieve_contact_from_aor_list(endpoint->aors);
		if (contact && !ast_strlen_zero(contact->uri)) {
			target = contact->uri;
		}
	}

	if (ast_channel_state(trnf_data->session->channel) == AST_STATE_RING) {
		transfer_redirect(trnf_data->session, target);
	} else {
		transfer_refer(trnf_data->session, target);
	}

	ao2_ref(trnf_data, -1);
	ao2_cleanup(endpoint);
	ao2_cleanup(contact);
	return 0;
}

/*! \brief Function called by core for Asterisk initiated transfer */
static int chan_pjsip_transfer(struct ast_channel *chan, const char *target)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct transfer_data *trnf_data = transfer_data_alloc(channel->session, target);

	if (!trnf_data) {
		return -1;
	}

	if (ast_sip_push_task(channel->session->serializer, transfer, trnf_data)) {
		ast_log(LOG_WARNING, "Error requesting transfer\n");
		ao2_cleanup(trnf_data);
		return -1;
	}

	return 0;
}

/*! \brief Function called by core to start a DTMF digit */
static int chan_pjsip_digit_begin(struct ast_channel *chan, char digit)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt = channel->pvt;
	struct ast_sip_session_media *media = pvt->media[SIP_MEDIA_AUDIO];
	int res = 0;

	switch (channel->session->endpoint->dtmf) {
	case AST_SIP_DTMF_RFC_4733:
		if (!media || !media->rtp) {
			return -1;
		}

		ast_rtp_instance_dtmf_begin(media->rtp, digit);
                break;
	case AST_SIP_DTMF_AUTO:
                       if (!media || !media->rtp || (ast_rtp_instance_dtmf_mode_get(media->rtp) == AST_RTP_DTMF_MODE_INBAND)) {
                        return -1;
                }

                ast_rtp_instance_dtmf_begin(media->rtp, digit);
                break;
	case AST_SIP_DTMF_NONE:
		break;
	case AST_SIP_DTMF_INBAND:
		res = -1;
		break;
	default:
		break;
	}

	return res;
}

struct info_dtmf_data {
	struct ast_sip_session *session;
	char digit;
	unsigned int duration;
};

static void info_dtmf_data_destroy(void *obj)
{
	struct info_dtmf_data *dtmf_data = obj;
	ao2_ref(dtmf_data->session, -1);
}

static struct info_dtmf_data *info_dtmf_data_alloc(struct ast_sip_session *session, char digit, unsigned int duration)
{
	struct info_dtmf_data *dtmf_data = ao2_alloc(sizeof(*dtmf_data), info_dtmf_data_destroy);
	if (!dtmf_data) {
		return NULL;
	}
	ao2_ref(session, +1);
	dtmf_data->session = session;
	dtmf_data->digit = digit;
	dtmf_data->duration = duration;
	return dtmf_data;
}

static int transmit_info_dtmf(void *data)
{
	RAII_VAR(struct info_dtmf_data *, dtmf_data, data, ao2_cleanup);

	struct ast_sip_session *session = dtmf_data->session;
	struct pjsip_tx_data *tdata;

	RAII_VAR(struct ast_str *, body_text, NULL, ast_free_ptr);

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "dtmf-relay",
	};

	if (!(body_text = ast_str_create(32))) {
		ast_log(LOG_ERROR, "Could not allocate buffer for INFO DTMF.\n");
		return -1;
	}
	ast_str_set(&body_text, 0, "Signal=%c\r\nDuration=%u\r\n", dtmf_data->digit, dtmf_data->duration);

	body.body_text = ast_str_buffer(body_text);

	if (ast_sip_create_request("INFO", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create DTMF INFO request\n");
		return -1;
	}
	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to DTMF INFO request\n");
		pjsip_tx_data_dec_ref(tdata);
		return -1;
	}
	ast_sip_session_send_request(session, tdata);

	return 0;
}

/*! \brief Function called by core to stop a DTMF digit */
static int chan_pjsip_digit_end(struct ast_channel *ast, char digit, unsigned int duration)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct chan_pjsip_pvt *pvt = channel->pvt;
	struct ast_sip_session_media *media = pvt->media[SIP_MEDIA_AUDIO];
	int res = 0;

	switch (channel->session->endpoint->dtmf) {
	case AST_SIP_DTMF_INFO:
	{
		struct info_dtmf_data *dtmf_data = info_dtmf_data_alloc(channel->session, digit, duration);

		if (!dtmf_data) {
			return -1;
		}

		if (ast_sip_push_task(channel->session->serializer, transmit_info_dtmf, dtmf_data)) {
			ast_log(LOG_WARNING, "Error sending DTMF via INFO.\n");
			ao2_cleanup(dtmf_data);
			return -1;
		}
		break;
	}
	case AST_SIP_DTMF_RFC_4733:
		if (!media || !media->rtp) {
			return -1;
		}

		ast_rtp_instance_dtmf_end_with_duration(media->rtp, digit, duration);
                break;
        case AST_SIP_DTMF_AUTO:
                if (!media || !media->rtp || (ast_rtp_instance_dtmf_mode_get(media->rtp) == AST_RTP_DTMF_MODE_INBAND)) {
                        return -1;
                }

                ast_rtp_instance_dtmf_end_with_duration(media->rtp, digit, duration);
                break;

	case AST_SIP_DTMF_NONE:
		break;
	case AST_SIP_DTMF_INBAND:
		res = -1;
		break;
	}

	return res;
}

static void update_initial_connected_line(struct ast_sip_session *session)
{
	struct ast_party_connected_line connected;

	/*
	 * Use the channel CALLERID() as the initial connected line data.
	 * The core or a predial handler may have supplied missing values
	 * from the session->endpoint->id.self about who we are calling.
	 */
	ast_channel_lock(session->channel);
	ast_party_id_copy(&session->id, &ast_channel_caller(session->channel)->id);
	ast_channel_unlock(session->channel);

	/* Supply initial connected line information if available. */
	if (!session->id.number.valid && !session->id.name.valid) {
		return;
	}

	ast_party_connected_line_init(&connected);
	connected.id = session->id;
	connected.source = AST_CONNECTED_LINE_UPDATE_SOURCE_ANSWER;

	ast_channel_queue_connected_line_update(session->channel, &connected, NULL);
}

static int call(void *data)
{
	struct ast_sip_channel_pvt *channel = data;
	struct ast_sip_session *session = channel->session;
	struct chan_pjsip_pvt *pvt = channel->pvt;
	pjsip_tx_data *tdata;

	int res = ast_sip_session_create_invite(session, &tdata);

	if (res) {
		ast_set_hangupsource(session->channel, ast_channel_name(session->channel), 0);
		ast_queue_hangup(session->channel);
	} else {
		set_channel_on_rtp_instance(pvt, ast_channel_uniqueid(session->channel));
		update_initial_connected_line(session);

#ifdef GRANDSTREAM_NETWORKS
		add_feature_map_header(tdata, 1);

		/*add by chdruan, Bug 42500 - Diversion:the ucm carry 4 diversion header when going through register trunk*/
		if (session->channel != NULL && pbx_builtin_getvar_helper(session->channel, "CF_REASON") != NULL
				&& !ast_channel_redirecting(session->channel)->count) {
			ast_channel_redirecting(session->channel)->count++;
		}

		//==========================================
		// Support CEI for user invitation
		//==========================================
		const char *cei_support = pbx_builtin_getvar_helper(session->channel, "CEI_SUPPORT");
		const char *conf_room = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");
		const char *mcb_contact = pbx_builtin_getvar_helper(session->channel, "MCB_CONTACT");
		const char *cid_name = pbx_builtin_getvar_helper(session->channel, "CID_NAME");
        const char *transport_type = pbx_builtin_getvar_helper(session->channel, "TRANSPORT_TYPE");
		struct ast_sip_endpoint *endpoint = session->endpoint;
		const char *endpoint_id = ast_sorcery_object_get_id(endpoint);

        if (cei_support != NULL && strcmp(cei_support, "yes") == 0 &&
			!ast_strlen_zero(conf_room) && endpoint != NULL &&
			endpoint_id != NULL && strstr(endpoint_id, "trunk") == NULL) {
			// Set isfocus
			pvt->isConference = TRUE;

            if (globalEnabledCEI) {
                add_peer_conference(ast_channel_name(session->channel), endpoint, conf_room);
            }

			// Set contact user
			if (pvt->contact_user != NULL) {
				ast_free(pvt->contact_user);
			}
			pvt->contact_user = ast_strdup(conf_room);

			// Update tdata
			if (globalEnabledCEI) {  //not add isfocus if CEI notify disabled, modified by yfyang, for bug58890
				append_isfocus_contact(tdata);
			}
			set_contact_user(tdata, pvt->contact_user);
        }

        //==========================================
		// Update SIP headers
		//==========================================
        // Assign From header information
		if (cid_name != NULL) {
			set_from_name(tdata, cid_name ? cid_name : "");
		}

		// Update contact information
		if (mcb_contact != NULL && strlen(mcb_contact) != 0) {
			set_contact_user(tdata, (char *)mcb_contact);
		}

        // Add SIP custom header to indicate transport type
        if (transport_type != NULL) {
            ast_sip_add_header(tdata, "X-Grandstream-Transport",
                               S_OR(transport_type, ""));
        }
#endif
		ast_sip_session_send_request(session, tdata);
	}
	ao2_ref(channel, -1);
	return res;
}

/*! \brief Function called by core to actually start calling a remote party */
static int chan_pjsip_call(struct ast_channel *ast, const char *dest, int timeout)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);

	ao2_ref(channel, +1);
	if (ast_sip_push_task(channel->session->serializer, call, channel)) {
		ast_log(LOG_WARNING, "Error attempting to place outbound call to '%s'\n", dest);
		ao2_cleanup(channel);
		return -1;
	}

	return 0;
}

/*! \brief Internal function which translates from Asterisk cause codes to SIP response codes */
static int hangup_cause2sip(int cause)
{
	switch (cause) {
	case AST_CAUSE_UNALLOCATED:             /* 1 */
	case AST_CAUSE_NO_ROUTE_DESTINATION:    /* 3 IAX2: Can't find extension in context */
	case AST_CAUSE_NO_ROUTE_TRANSIT_NET:    /* 2 */
		return 404;
	case AST_CAUSE_CONGESTION:              /* 34 */
	case AST_CAUSE_SWITCH_CONGESTION:       /* 42 */
		return 503;
	case AST_CAUSE_NO_USER_RESPONSE:        /* 18 */
		return 408;
	case AST_CAUSE_NO_ANSWER:               /* 19 */
	case AST_CAUSE_UNREGISTERED:        /* 20 */
		return 480;
	case AST_CAUSE_CALL_REJECTED:           /* 21 */
		return 403;
	case AST_CAUSE_NUMBER_CHANGED:          /* 22 */
		return 410;
	case AST_CAUSE_NORMAL_UNSPECIFIED:      /* 31 */
		return 480;
	case AST_CAUSE_INVALID_NUMBER_FORMAT:
		return 484;
	case AST_CAUSE_USER_BUSY:
		return 486;
	case AST_CAUSE_FAILURE:
		return 500;
	case AST_CAUSE_FACILITY_REJECTED:       /* 29 */
		return 501;
	case AST_CAUSE_CHAN_NOT_IMPLEMENTED:
		return 503;
	case AST_CAUSE_DESTINATION_OUT_OF_ORDER:
		return 502;
	case AST_CAUSE_BEARERCAPABILITY_NOTAVAIL:       /* Can't find codec to connect to host */
		return 488;
	case AST_CAUSE_INTERWORKING:    /* Unspecified Interworking issues */
		return 500;
	case AST_CAUSE_NOTDEFINED:
	default:
		ast_debug(1, "AST hangup cause %d (no match found in PJSIP)\n", cause);
		return 0;
	}

	/* Never reached */
	return 0;
}

struct hangup_data {
	int cause;
	struct ast_channel *chan;
};

static void hangup_data_destroy(void *obj)
{
	struct hangup_data *h_data = obj;

	h_data->chan = ast_channel_unref(h_data->chan);
}

static struct hangup_data *hangup_data_alloc(int cause, struct ast_channel *chan)
{
	struct hangup_data *h_data = ao2_alloc(sizeof(*h_data), hangup_data_destroy);

	if (!h_data) {
		return NULL;
	}

	h_data->cause = cause;
	h_data->chan = ast_channel_ref(chan);

	return h_data;
}

/*! \brief Clear a channel from a session along with its PVT */
static void clear_session_and_channel(struct ast_sip_session *session, struct ast_channel *ast, struct chan_pjsip_pvt *pvt)
{
#ifdef GRANDSTREAM_NETWORKS
	if (session == NULL) {
		ast_log(LOG_ERROR, "Invalid SIP session!");
		return;
	}
	ao2_lock(session);
#endif
	session->channel = NULL;
	set_channel_on_rtp_instance(pvt, "");
	ast_channel_tech_pvt_set(ast, NULL);
#ifdef GRANDSTREAM_NETWORKS
	ao2_unlock(session);
#endif
}

static int hangup(void *data)
{
	struct hangup_data *h_data = data;
	struct ast_channel *ast = h_data->chan;
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct chan_pjsip_pvt *pvt = channel->pvt;
	struct ast_sip_session *session = channel->session;
#ifdef GRANDSTREAM_NETWORKS
    struct ast_sip_endpoint *endpoint = session ? session->endpoint: NULL;
#endif
	int cause = h_data->cause;

#ifdef GRANDSTREAM_NETWORKS
    // Clear current channel's subscription
    if (endpoint) {
        AST_LIST_LOCK(&endpoint->conferences);
        clear_peer_conference(endpoint, ast_channel_name(ast));
        AST_LIST_UNLOCK(&endpoint->conferences);
    }
#endif

	ast_sip_session_terminate(session, cause);
	clear_session_and_channel(session, ast, pvt);
	ao2_cleanup(channel);
	ao2_cleanup(h_data);

	return 0;
}

/*! \brief Function called by core to hang up a PJSIP session */
static int chan_pjsip_hangup(struct ast_channel *ast)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct chan_pjsip_pvt *pvt = channel->pvt;
	int cause = hangup_cause2sip(ast_channel_hangupcause(channel->session->channel));
	struct hangup_data *h_data = hangup_data_alloc(cause, ast);

	if (!h_data) {
		goto failure;
	}

	if (ast_sip_push_task(channel->session->serializer, hangup, h_data)) {
		ast_log(LOG_WARNING, "Unable to push hangup task to the threadpool. Expect bad things\n");
		goto failure;
	}

	return 0;

failure:
	/* Go ahead and do our cleanup of the session and channel even if we're not going
	 * to be able to send our SIP request/response
	 */
	clear_session_and_channel(channel->session, ast, pvt);
	ao2_cleanup(channel);
	ao2_cleanup(h_data);

	return -1;
}

struct request_data {
	struct ast_sip_session *session;
	struct ast_format_cap *caps;
	const char *dest;
	int cause;
};

static int request(void *obj)
{
	struct request_data *req_data = obj;
	struct ast_sip_session *session = NULL;
	char *tmp = ast_strdupa(req_data->dest), *endpoint_name = NULL, *request_user = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
#ifdef GRANDSTREAM_NETWORKS
	enum ast_endpoint_state state;
#endif

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(endpoint);
		AST_APP_ARG(aor);
	);

	if (ast_strlen_zero(tmp)) {
		ast_log(LOG_ERROR, "Unable to create PJSIP channel with empty destination\n");
		req_data->cause = AST_CAUSE_CHANNEL_UNACCEPTABLE;
		return -1;
	}

	AST_NONSTANDARD_APP_ARGS(args, tmp, '/');

	/* If a request user has been specified extract it from the endpoint name portion */
	if ((endpoint_name = strchr(args.endpoint, '@'))) {
		request_user = args.endpoint;
		*endpoint_name++ = '\0';
	} else {
		endpoint_name = args.endpoint;
	}

	if (ast_strlen_zero(endpoint_name)) {
		ast_log(LOG_ERROR, "Unable to create PJSIP channel with empty endpoint name\n");
		req_data->cause = AST_CAUSE_CHANNEL_UNACCEPTABLE;
		return -1;
	} else if (!(endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", endpoint_name))) {
		ast_log(LOG_ERROR, "Unable to create PJSIP channel - endpoint '%s' was not found\n", endpoint_name);
		req_data->cause = AST_CAUSE_NO_ROUTE_DESTINATION;
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	state = ast_endpoint_get_state(endpoint->persistent);
	if (state == AST_ENDPOINT_OFFLINE) {
		ast_log(LOG_ERROR, "endpoint[%s] is OFFLINE!\n", endpoint_name);
		req_data->cause = AST_CAUSE_NO_ROUTE_DESTINATION;
		return -1;
	}
#endif

	if (!(session = ast_sip_session_create_outgoing(endpoint, NULL, args.aor, request_user, req_data->caps))) {
		ast_log(LOG_ERROR, "Failed to create outgoing session to endpoint '%s'\n", endpoint_name);
		req_data->cause = AST_CAUSE_NO_ROUTE_DESTINATION;
		return -1;
	}

	req_data->session = session;

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
static unsigned char judge_chan_type_is_sip(const struct ast_channel *channel)
{
	if (NULL == channel){
		ast_log(LOG_WARNING, "chan is NULL\n");
		return FALSE;
	}

	if (ast_channel_tech(channel) && 
		(0 == strcmp(ast_channel_tech(channel)->type, "PJSIP"))) {
		return TRUE;
	} else {
		return FALSE;
	}
}

static void get_peer_bfcp_para(struct ast_sip_session *session, const struct ast_channel *requestor)
{
	if (NULL == session || NULL == requestor || NULL == session->channel) {
		ast_log(LOG_ERROR, "get peer bfcp para error\n");
		return;
	}
	ast_log(LOG_DEBUG, "get peer bfcp para begin\n");

	if (!judge_chan_type_is_sip(session->channel) || !judge_chan_type_is_sip(requestor)) {
		ast_log(LOG_DEBUG, "get peer bfcp para, channel type is not PJSIP\n");
		return;
	}

	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(requestor);
	if (!channel || !channel->session) {
		ast_log(LOG_WARNING, "Requestor has no session");
		return;
	}
	memcpy(&session->bfcp_para, &channel->session->bfcp_para, sizeof(session->bfcp_para));
	session->bfcp_other_attr_count = 0;
	int i = 0;
	for (i = 0; i < channel->session->bfcp_other_attr_count; i++) {
		pj_pool_t *pool = session->inv_session->pool;
		session->bfcp_other_attr[session->bfcp_other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
					 channel->session->bfcp_other_attr[i]);
	}
	session->bfcp_bandw_count = 0;
	for (i = 0; i < channel->session->bfcp_bandw_count; i++) {
		pj_pool_t *pool = session->inv_session->pool;
		session->bfcp_bandw[session->bfcp_bandw_count++] = pjmedia_sdp_bandw_clone(pool, 
					 channel->session->bfcp_bandw[i]);
	}

	return;
}

static void get_peer_video_extra_attr_para(struct ast_sip_session *session, const struct ast_channel *requestor)
{
	if (NULL == session || NULL == requestor) {
		ast_log(LOG_ERROR, "get peer video extra para error\n");
		return;
	}
	ast_log(LOG_DEBUG, "get peer video extra para begin\n");

	if (!judge_chan_type_is_sip(session->channel) || !judge_chan_type_is_sip(requestor)) {
		ast_log(LOG_DEBUG, "get peer video extra para, channel type is not PJSIP\n");
		return;
	}

	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(requestor);
	if (!channel || !channel->session) {
		ast_log(LOG_WARNING, "Requestor has no session\n");
		return;
	}
	struct ast_str *format_buf = ast_str_alloca(64);

	memcpy(&session->main_video_para, &channel->session->main_video_para, sizeof(session->main_video_para));
	int i = 0;
	session->main_video_para.other_attr_count = 0;
	for (i = 0; i < channel->session->main_video_para.other_attr_count; i++) {
			pj_pool_t *pool = session->inv_session->pool;
			session->main_video_para.other_attr[session->main_video_para.other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
									 channel->session->main_video_para.other_attr[i]);
	}
	session->main_video_para.bandw_count = 0;
	for (i = 0; i < channel->session->main_video_para.bandw_count; i++) {
		pj_pool_t *pool = session->inv_session->pool;
		session->main_video_para.bandw[session->main_video_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool, 
					 channel->session->main_video_para.bandw[i]);
	}
	memcpy(&session->slides_video_para, &channel->session->slides_video_para, sizeof(session->slides_video_para));
	session->slides_video_para.other_attr_count = 0;
	for (i = 0; i < channel->session->slides_video_para.other_attr_count; i++) {
			pj_pool_t *pool = session->inv_session->pool;
			session->slides_video_para.other_attr[session->slides_video_para.other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
									 channel->session->slides_video_para.other_attr[i]);
	}
	session->slides_video_para.bandw_count = 0;
	for (i = 0; i < channel->session->slides_video_para.bandw_count; i++) {
		pj_pool_t *pool = session->inv_session->pool;
		session->slides_video_para.bandw[session->slides_video_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool, 
					 channel->session->slides_video_para.bandw[i]);
	}

	if (channel->session->slides_video_req_caps && ast_format_cap_count(channel->session->slides_video_req_caps)) {
		/* get joint caps between req_caps and endpoint caps */
		struct ast_format_cap *joint_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
		ast_format_cap_get_compatible(channel->session->slides_video_req_caps, session->endpoint->media.codecs, joint_caps);

		ast_log(LOG_CODEC_NEG, "[callee] slides_video_req_caps inherited from the caller: %s\n",	 
			ast_format_cap_get_names(channel->session->slides_video_req_caps, &format_buf));
		ast_log(LOG_CODEC_NEG, "[callee] endpoint codecs: %s\n",
			ast_format_cap_get_names(session->endpoint->media.codecs, &format_buf));

		/* if joint caps */
		if (ast_format_cap_count(joint_caps)) {
			ast_format_cap_append_from_cap(session->slides_video_req_caps, joint_caps, AST_MEDIA_TYPE_UNKNOWN);

			ast_log(LOG_CODEC_NEG, "[callee] slides video joint codecs(endpoint & req_caps): %s\n",
				ast_format_cap_get_names(joint_caps, &format_buf));
			ast_log(LOG_CODEC_NEG, "[callee] final slides_video_req_caps of the callee: %s\n",
				ast_format_cap_get_names(session->slides_video_req_caps, &format_buf));

			/* replace instances of joint caps equivalents in session->req_caps */
			ast_format_cap_replace_from_cap(session->slides_video_req_caps, joint_caps, AST_MEDIA_TYPE_UNKNOWN);
		}
		ao2_cleanup(joint_caps);
	}
	return;
}

static void get_peer_fec_para(struct ast_sip_session *session, const struct ast_channel *requestor)
{
	if (NULL == session || NULL == requestor) {
		ast_log(LOG_ERROR, "get peer fec para error\n");
		return;
	}
	ast_log(LOG_DEBUG, "get peer fec para begin\n");

	if (!judge_chan_type_is_sip(session->channel) || !judge_chan_type_is_sip(requestor)) {
		ast_log(LOG_DEBUG, "get peer fec para, channel type is not PJSIP\n");
		return;
	}

	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(requestor);
	if (!channel || !channel->session) {
		ast_log(LOG_WARNING, "Requestor has no session\n");
		return;
	}
	struct ast_str *format_buf = ast_str_alloca(64);

	memcpy(&session->fec_para, &channel->session->fec_para, sizeof(session->fec_para));
	session->fec_para.other_attr_count = 0;
	int i = 0;
	for (i = 0; i < channel->session->fec_para.other_attr_count; i++) {
		pj_pool_t *pool = session->inv_session->pool;
		session->fec_para.other_attr[session->fec_para.other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
					 channel->session->fec_para.other_attr[i]);
	}
	session->fec_para.bandw_count = 0;
	for (i = 0; i < channel->session->fec_para.bandw_count; i++) {
		pj_pool_t *pool = session->inv_session->pool;
		session->fec_para.bandw[session->fec_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool, 
					 channel->session->fec_para.bandw[i]);
	}
	/* FEC will not use srtp normally, if peer use we use */
	session->fec_media.encryption = channel->session->fec_media.encryption;
	pj_strdup(session->inv_session->pool, &session->fec_media.transport, &channel->session->fec_media.transport);

	if (channel->session->fec_req_caps && ast_format_cap_count(channel->session->fec_req_caps)) {
		ast_log(LOG_CODEC_NEG, "[callee] fec_req_caps inherited from the caller: %s\n",
			ast_format_cap_get_names(channel->session->fec_req_caps, &format_buf));
		ast_format_cap_append_from_cap(session->fec_req_caps, channel->session->fec_req_caps, AST_MEDIA_TYPE_UNKNOWN);
	}
	return;
}

static void get_peer_fecc_para(struct ast_sip_session *session, const struct ast_channel *requestor)
{
	if (NULL == session || NULL == requestor) {
		ast_log(LOG_ERROR, "get peer fecc para error\n");
		return;
	}
	ast_log(LOG_DEBUG, "get peer fecc para begin\n");

	if (!judge_chan_type_is_sip(session->channel) || !judge_chan_type_is_sip(requestor)) {
		ast_log(LOG_DEBUG, "get peer fecc para, channel type is not PJSIP\n");
		return;
	}

	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(requestor);
	if (!channel || !channel->session) {
		ast_log(LOG_WARNING, "Requestor has no session\n");
		return;
	}
	struct ast_str *format_buf = ast_str_alloca(64);

	memcpy(&session->fecc_para, &channel->session->fecc_para, sizeof(session->fecc_para));
	session->fecc_para.other_attr_count = 0;
	int i = 0;
	for (i = 0; i < channel->session->fecc_para.other_attr_count; i++) {
		pj_pool_t *pool = session->inv_session->pool;
		session->fecc_para.other_attr[session->fecc_para.other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
					 channel->session->fecc_para.other_attr[i]);
	}
	session->fecc_para.bandw_count = 0;
	for (i = 0; i < channel->session->fecc_para.bandw_count; i++) {
		pj_pool_t *pool = session->inv_session->pool;
		session->fecc_para.bandw[session->fecc_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool, 
					 channel->session->fecc_para.bandw[i]);
	}

	if (channel->session->fecc_req_caps && ast_format_cap_count(channel->session->fecc_req_caps)) {
		ast_log(LOG_CODEC_NEG, "[callee] fecc_req_caps inherited from the caller: %s\n",
			ast_format_cap_get_names(channel->session->fecc_req_caps, &format_buf));
		ast_format_cap_append_from_cap(session->fecc_req_caps, channel->session->fecc_req_caps, AST_MEDIA_TYPE_UNKNOWN);
	}
	return;
}

static void get_peer_fec_multiplexed_para(struct ast_sip_session *session, const struct ast_channel *requestor)
{
	if (NULL == session || NULL == requestor) {
		ast_log(LOG_ERROR, "get peer fec multiplexed para error\n");
		return;
	}
	ast_log(LOG_DEBUG, "get peer fec multiplexed para begin\n");

	if (!judge_chan_type_is_sip(session->channel) || !judge_chan_type_is_sip(requestor)) {
		ast_log(LOG_DEBUG, "get peer fec multiplexed para, channel type is not PJSIP\n");
		return;
	}

	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(requestor);
	if (!channel || !channel->session) {
		ast_log(LOG_WARNING, "Requestor has no session\n");
		return;
	}
	struct ast_str *format_buf = ast_str_alloca(64);

	memcpy(&session->fec_multiplexed_para, &channel->session->fec_multiplexed_para, sizeof(session->fec_multiplexed_para));

	if (channel->session->fec_multiplexed_req_caps && ast_format_cap_count(channel->session->fec_multiplexed_req_caps)) {
		ast_log(LOG_CODEC_NEG, "[callee] fec_multiplexed_req_caps inherited from the caller: %s\n",
			ast_format_cap_get_names(channel->session->fec_multiplexed_req_caps, &format_buf));
		ast_format_cap_append_from_cap(session->fec_multiplexed_req_caps, channel->session->fec_multiplexed_req_caps, AST_MEDIA_TYPE_UNKNOWN);
	}
	return;
}

static void get_peer_yealink_para(struct ast_sip_session *session, const struct ast_channel *requestor)
{
	if (NULL == session || NULL == requestor) {
		ast_log(LOG_ERROR, "get peer yealink para error");
		return;
	}
	ast_log(LOG_DEBUG, "get peer yealink para begin");

	if (!judge_chan_type_is_sip(session->channel) || !judge_chan_type_is_sip(requestor)) {
		ast_log(LOG_DEBUG, "get peer yealink para, channel type is not PJSIP");
		return;
	}

	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(requestor);
	if (!channel || !channel->session) {
		ast_log(LOG_WARNING, "Requestor has no session");
		return;
	}

	ast_log(LOG_DEBUG, "get_peer_yealink_para, is_yealink_msg:%d", channel->session->is_yealink_msg);

	session->is_yealink_msg = channel->session->is_yealink_msg;

	return;
}
#endif

/*! \brief Function called by core to create a new outgoing PJSIP session */
static struct ast_channel *chan_pjsip_request(const char *type, struct ast_format_cap *cap, const struct ast_assigned_ids *assignedids, const struct ast_channel *requestor, const char *data, int *cause)
{
	struct request_data req_data;
	RAII_VAR(struct ast_sip_session *, session, NULL, ao2_cleanup);

	req_data.caps = cap;
	req_data.dest = data;

	if (ast_sip_push_task_synchronous(NULL, request, &req_data)) {
		*cause = req_data.cause;
		return NULL;
	}

	session = req_data.session;

#ifdef GRANDSTREAM_NETWORKS
	if (!(session->channel = chan_pjsip_new(session, AST_STATE_DOWN, NULL, NULL, assignedids, requestor, NULL, cap))) {
#else
	if (!(session->channel = chan_pjsip_new(session, AST_STATE_DOWN, NULL, NULL, assignedids, requestor, NULL))) {
#endif
		/* Session needs to be terminated prematurely */
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	/*! Begin: Added by pxie@grandstream.cn for bug[45683 and 42261], 2015-07-16.*/
	if (requestor) {
		const char *send_fax_file_path = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "filepath");
		if (send_fax_file_path) {
			session->is_fax_sending_request = 1;
		} else {
			session->is_fax_sending_request = 0;
		}
	}
	/*! End: Added by pxie@grandstream.cn for bug[45683 and 42261], 2015-07-16.*/
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	get_peer_bfcp_para(session, requestor);
	get_peer_video_extra_attr_para(session, requestor);
	get_peer_fec_para(session, requestor);
	get_peer_fecc_para(session, requestor);
	get_peer_fec_multiplexed_para(session, requestor);
	ast_sip_session_set_use_final_sdp_val(session);
	get_peer_yealink_para(session, requestor);
#endif

	return session->channel;
}

struct sendtext_data {
	struct ast_sip_session *session;
	char text[0];
};

static void sendtext_data_destroy(void *obj)
{
	struct sendtext_data *data = obj;
	ao2_ref(data->session, -1);
}

static struct sendtext_data* sendtext_data_create(struct ast_sip_session *session, const char *text)
{
	int size = strlen(text) + 1;
	struct sendtext_data *data = ao2_alloc(sizeof(*data)+size, sendtext_data_destroy);

	if (!data) {
		return NULL;
	}

	data->session = session;
	ao2_ref(data->session, +1);
	ast_copy_string(data->text, text, size);
	return data;
}

static int sendtext(void *obj)
{
	RAII_VAR(struct sendtext_data *, data, obj, ao2_cleanup);
	pjsip_tx_data *tdata;

	const struct ast_sip_body body = {
		.type = "text",
		.subtype = "plain",
		.body_text = data->text
	};

	ast_debug(3, "Sending in dialog SIP message\n");

	ast_sip_create_request("MESSAGE", data->session->inv_session->dlg, data->session->endpoint, NULL, NULL, &tdata);
	ast_sip_add_body(tdata, &body);
	ast_sip_send_request(tdata, data->session->inv_session->dlg, data->session->endpoint, NULL, NULL);

	return 0;
}

/*! \brief Function called by core to send text on PJSIP session */
static int chan_pjsip_sendtext(struct ast_channel *ast, const char *text)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct sendtext_data *data = sendtext_data_create(channel->session, text);

	if (!data || ast_sip_push_task(channel->session->serializer, sendtext, data)) {
		ao2_ref(data, -1);
		return -1;
	}
	return 0;
}

/*! \brief Convert SIP hangup causes to Asterisk hangup causes */
static int hangup_sip2cause(int cause)
{
	/* Possible values taken from causes.h */

	switch(cause) {
	case 401:       /* Unauthorized */
		return AST_CAUSE_CALL_REJECTED;
	case 403:       /* Not found */
		return AST_CAUSE_CALL_REJECTED;
	case 404:       /* Not found */
		return AST_CAUSE_UNALLOCATED;
	case 405:       /* Method not allowed */
		return AST_CAUSE_INTERWORKING;
	case 407:       /* Proxy authentication required */
		return AST_CAUSE_CALL_REJECTED;
	case 408:       /* No reaction */
		return AST_CAUSE_NO_USER_RESPONSE;
	case 409:       /* Conflict */
		return AST_CAUSE_NORMAL_TEMPORARY_FAILURE;
	case 410:       /* Gone */
		return AST_CAUSE_NUMBER_CHANGED;
	case 411:       /* Length required */
		return AST_CAUSE_INTERWORKING;
	case 413:       /* Request entity too large */
		return AST_CAUSE_INTERWORKING;
	case 414:       /* Request URI too large */
		return AST_CAUSE_INTERWORKING;
	case 415:       /* Unsupported media type */
		return AST_CAUSE_INTERWORKING;
	case 420:       /* Bad extension */
		return AST_CAUSE_NO_ROUTE_DESTINATION;
	case 480:       /* No answer */
		return AST_CAUSE_NO_ANSWER;
	case 481:       /* No answer */
		return AST_CAUSE_INTERWORKING;
	case 482:       /* Loop detected */
		return AST_CAUSE_INTERWORKING;
	case 483:       /* Too many hops */
		return AST_CAUSE_NO_ANSWER;
	case 484:       /* Address incomplete */
		return AST_CAUSE_INVALID_NUMBER_FORMAT;
	case 485:       /* Ambiguous */
		return AST_CAUSE_UNALLOCATED;
	case 486:       /* Busy everywhere */
		return AST_CAUSE_BUSY;
	case 487:       /* Request terminated */
		return AST_CAUSE_INTERWORKING;
	case 488:       /* No codecs approved */
		return AST_CAUSE_BEARERCAPABILITY_NOTAVAIL;
	case 491:       /* Request pending */
		return AST_CAUSE_INTERWORKING;
	case 493:       /* Undecipherable */
		return AST_CAUSE_INTERWORKING;
	case 500:       /* Server internal failure */
		return AST_CAUSE_FAILURE;
	case 501:       /* Call rejected */
		return AST_CAUSE_FACILITY_REJECTED;
	case 502:
		return AST_CAUSE_DESTINATION_OUT_OF_ORDER;
	case 503:       /* Service unavailable */
		return AST_CAUSE_CONGESTION;
	case 504:       /* Gateway timeout */
		return AST_CAUSE_RECOVERY_ON_TIMER_EXPIRE;
	case 505:       /* SIP version not supported */
		return AST_CAUSE_INTERWORKING;
	case 600:       /* Busy everywhere */
		return AST_CAUSE_USER_BUSY;
	case 603:       /* Decline */
		return AST_CAUSE_CALL_REJECTED;
	case 604:       /* Does not exist anywhere */
		return AST_CAUSE_UNALLOCATED;
	case 606:       /* Not acceptable */
		return AST_CAUSE_BEARERCAPABILITY_NOTAVAIL;
	default:
		if (cause < 500 && cause >= 400) {
			/* 4xx class error that is unknown - someting wrong with our request */
			return AST_CAUSE_INTERWORKING;
		} else if (cause < 600 && cause >= 500) {
			/* 5xx class error - problem in the remote end */
			return AST_CAUSE_CONGESTION;
		} else if (cause < 700 && cause >= 600) {
			/* 6xx - global errors in the 4xx class */
			return AST_CAUSE_INTERWORKING;
		}
		return AST_CAUSE_NORMAL;
	}
	/* Never reached */
	return 0;
}

static void chan_pjsip_session_begin(struct ast_sip_session *session)
{
	RAII_VAR(struct ast_datastore *, datastore, NULL, ao2_cleanup);
#ifdef GRANDSTREAM_NETWORKS
	if (session == NULL) {
		ast_log(LOG_ERROR, "Invalid SIP session!");
		return;
	}
	ao2_lock(session);
#endif
	if (session->endpoint->media.direct_media.glare_mitigation ==
			AST_SIP_DIRECT_MEDIA_GLARE_MITIGATION_NONE) {
#ifdef GRANDSTREAM_NETWORKS
		ao2_unlock(session);
#endif
		return;
	}

	datastore = ast_sip_session_alloc_datastore(&direct_media_mitigation_info,
			"direct_media_glare_mitigation");

	if (!datastore) {
#ifdef GRANDSTREAM_NETWORKS
		ao2_unlock(session);
#endif
		return;
	}

	ast_sip_session_add_datastore(session, datastore);
#ifdef GRANDSTREAM_NETWORKS
	ao2_unlock(session);
#endif
}

/*! \brief Function called when the session ends */
static void chan_pjsip_session_end(struct ast_sip_session *session)
{
#ifdef GRANDSTREAM_NETWORKS
	if (session == NULL) {
		ast_log(LOG_ERROR, "Invalid SIP session");
		return;
	}
	ao2_lock(session);
#endif
	if (!session->channel) {
#ifdef GRANDSTREAM_NETWORKS
		ao2_unlock(session);
#endif
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for Bug 52368, 2015.10.29 */
	if (session->is_video_media_change && session->inv_session && session->inv_session->invite_tsx) {
		session->is_video_media_change = 0;
		ast_queue_control(session->channel, AST_CONTROL_REINVITE_BYE_CONFLICT);
		ast_log(LOG_DEBUG, "Recv bye, but has a reinvite pending, just decline this reinvite request\n");
	}
#endif

	chan_pjsip_remove_hold(ast_channel_uniqueid(session->channel));

	ast_set_hangupsource(session->channel, ast_channel_name(session->channel), 0);
	if (!ast_channel_hangupcause(session->channel) && session->inv_session) {
		int cause = hangup_sip2cause(session->inv_session->cause);

		ast_queue_hangup_with_cause(session->channel, cause);
	} else {
		ast_queue_hangup(session->channel);
	}
#ifdef GRANDSTREAM_NETWORKS
	ao2_unlock(session);
#endif
}

/*! \brief Function called when a request is received on the session */
static int chan_pjsip_incoming_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	RAII_VAR(struct ast_datastore *, datastore, NULL, ao2_cleanup);
	struct transport_info_data *transport_data;
	pjsip_tx_data *packet = NULL;

#ifdef GRANDSTREAM_NETWORKS
	if (session == NULL) {
		ast_log(LOG_ERROR, "Invalid SIP session!");
		return -1;
	}
	ao2_lock(session);
#endif
	if (session->channel) {
#ifdef GRANDSTREAM_NETWORKS
		ao2_unlock(session);
#endif
		return 0;
	}

	if (session->inv_session->state >= PJSIP_INV_STATE_CONFIRMED) {
		/* Weird case. We've received a reinvite but we don't have a channel. The most
		 * typical case for this happening is that a blind transfer fails, and so the
		 * transferer attempts to reinvite himself back into the call. We already got
		 * rid of that channel, and the other side of the call is unrecoverable.
		 *
		 * We treat this as a failure, so our best bet is to just hang this call
		 * up and not create a new channel. Clearing defer_terminate here ensures that
		 * calling ast_sip_session_terminate() can result in a BYE being sent ASAP.
		 */
		session->defer_terminate = 0;
		ast_sip_session_terminate(session, 400);
#ifdef GRANDSTREAM_NETWORKS
		ao2_unlock(session);
#endif
		return -1;
	}

	datastore = ast_sip_session_alloc_datastore(&transport_info, "transport_info");
	if (!datastore) {
#ifdef GRANDSTREAM_NETWORKS
		ao2_unlock(session);
#endif
		return -1;
	}

	transport_data = ast_calloc(1, sizeof(*transport_data));
	if (!transport_data) {
#ifdef GRANDSTREAM_NETWORKS
		ao2_unlock(session);
#endif
		return -1;
	}
	pj_sockaddr_cp(&transport_data->local_addr, &rdata->tp_info.transport->local_addr);
	pj_sockaddr_cp(&transport_data->remote_addr, &rdata->pkt_info.src_addr);
	datastore->data = transport_data;
	ast_sip_session_add_datastore(session, datastore);
#ifdef GRANDSTREAM_NETWORKS
	if (!(session->channel = chan_pjsip_new(session, AST_STATE_RING, session->exten, NULL, NULL, NULL, NULL, NULL))) {
#else
	if (!(session->channel = chan_pjsip_new(session, AST_STATE_RING, session->exten, NULL, NULL, NULL, NULL))) {
#endif
		if (pjsip_inv_end_session(session->inv_session, 503, NULL, &packet) == PJ_SUCCESS) {
			ast_sip_session_send_response(session, packet);
		}

		ast_log(LOG_ERROR, "Failed to allocate new PJSIP channel on incoming SIP INVITE\n");
#ifdef GRANDSTREAM_NETWORKS
		ao2_unlock(session);
#endif
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (session->core_id >= 0) {
		ast_setup_cc_recall_datastore(session->channel, session->core_id);
		ast_cc_agent_set_interfaces_chanvar(session->channel);
	}
#endif

	/* channel gets created on incoming request, but we wait to call start
           so other supplements have a chance to run */
#ifdef GRANDSTREAM_NETWORKS
	ao2_unlock(session);
#endif
	return 0;
}

static int call_pickup_incoming_request(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	struct ast_features_pickup_config *pickup_cfg;
	struct ast_channel *chan;

#ifdef GRANDSTREAM_NETWORKS
	if (session == NULL) {
		ast_log(LOG_ERROR, "Invalid SIP session!");
		return -1;
	}
	ao2_lock(session);
#endif
	/* We don't care about reinvites */
	if (session->inv_session->state >= PJSIP_INV_STATE_CONFIRMED) {
#ifdef GRANDSTREAM_NETWORKS
		ao2_unlock(session);
#endif
		return 0;
	}

	pickup_cfg = ast_get_chan_features_pickup_config(session->channel);
	if (!pickup_cfg) {
		ast_log(LOG_ERROR, "Unable to retrieve pickup configuration options. Unable to detect call pickup extension.\n");
#ifdef GRANDSTREAM_NETWORKS
		ao2_unlock(session);
#endif
		return 0;
	}

	if (strcmp(session->exten, pickup_cfg->pickupexten)) {
		ao2_ref(pickup_cfg, -1);
#ifdef GRANDSTREAM_NETWORKS
		ao2_unlock(session);
#endif
		return 0;
	}
	ao2_ref(pickup_cfg, -1);

	/* We can't directly use session->channel because the pickup operation will cause a masquerade to occur,
	 * changing the channel pointer in session to a different channel. To ensure we work on the right channel
	 * we store a pointer locally before we begin and keep a reference so it remains valid no matter what.
	 */
	chan = ast_channel_ref(session->channel);
	if (ast_pickup_call(chan)) {
		ast_channel_hangupcause_set(chan, AST_CAUSE_CALL_REJECTED);
	} else {
#ifdef GRANDSTREAM_NETWORKS
		/*jklou Bug 41546 - Pickup Group:It will have a missed call record when pickup a call*/
		if (ast_channel_hangupcause(chan) != AST_CAUSE_ANSWERED_ELSEWHERE){
			ast_channel_hangupcause_set(chan, AST_CAUSE_NORMAL_CLEARING);
		}
#else
		ast_channel_hangupcause_set(chan, AST_CAUSE_NORMAL_CLEARING);
#endif
	}
	/* A hangup always occurs because the pickup operation will have either failed resulting in the call
	 * needing to be hung up OR the pickup operation was a success and the channel we now have is actually
	 * the channel that was replaced, which should be hung up since it is literally in limbo not connected
	 * to anything at all.
	 */
	ast_hangup(chan);
	ast_channel_unref(chan);

#ifdef GRANDSTREAM_NETWORKS
	ao2_unlock(session);
#endif
	return 1;
}

static struct ast_sip_session_supplement call_pickup_supplement = {
	.method = "INVITE",
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_LAST - 1,
	.incoming_request = call_pickup_incoming_request,
};

static int pbx_start_incoming_request(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	int res;

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for Bug 55708, 2015.12.23
	 * pjsip uas tsx state will not turn to Terminated until recv ack,
	 * inv state will still in CONNECTING. Don't start pbx send invite.
	 */
	if (session->inv_session->state >= PJSIP_INV_STATE_CONNECTING) {
		ast_log(LOG_DEBUG, "pbx needn't process incoming request, invite state:%d", session->inv_session->state);
		return 0;
	}
#else
	/* We don't care about reinvites */
	if (session->inv_session->state >= PJSIP_INV_STATE_CONFIRMED) {
		return 0;
	}
#endif

	res = ast_pbx_start(session->channel);

	switch (res) {
	case AST_PBX_FAILED:
		ast_log(LOG_WARNING, "Failed to start PBX ;(\n");
		ast_channel_hangupcause_set(session->channel, AST_CAUSE_SWITCH_CONGESTION);
		ast_hangup(session->channel);
		break;
	case AST_PBX_CALL_LIMIT:
		ast_log(LOG_WARNING, "Failed to start PBX (call limit reached) \n");
		ast_channel_hangupcause_set(session->channel, AST_CAUSE_SWITCH_CONGESTION);
		ast_hangup(session->channel);
		break;
	case AST_PBX_SUCCESS:
	default:
		break;
	}

	ast_debug(3, "Started PBX on new PJSIP channel %s\n", ast_channel_name(session->channel));

	return (res == AST_PBX_SUCCESS) ? 0 : -1;
}

static struct ast_sip_session_supplement pbx_start_supplement = {
	.method = "INVITE",
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_LAST,
	.incoming_request = pbx_start_incoming_request,
};

/*! \brief Function called when a response is received on the session */
static void chan_pjsip_incoming_response(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	struct pjsip_status_line status = rdata->msg_info.msg->line.status;
	struct ast_control_pvt_cause_code *cause_code;
	int data_size = sizeof(*cause_code);
#ifdef GRANDSTREAM_NETWORKS
	const pj_str_t mcb_resp_name = { "X-Grandstream-Mcb-Resp", 22 };
    const pj_str_t call_type_name = { "X-Grandstream-Call-Type", 23 };
	pjsip_generic_string_hdr *hdr = NULL;
	pj_str_t value;
    const char *cei_support = NULL;
    const char *conf_room = NULL;
#endif

	if (!session->channel) {
		return;
	}

	switch (status.code) {
	case 180:
		ast_queue_control(session->channel, AST_CONTROL_RINGING);
		ast_channel_lock(session->channel);
		if (ast_channel_state(session->channel) != AST_STATE_UP) {
			ast_setstate(session->channel, AST_STATE_RINGING);
		}
		ast_channel_unlock(session->channel);
		break;
	case 183:
		ast_queue_control(session->channel, AST_CONTROL_PROGRESS);
		break;
	case 200:
#ifdef GRANDSTREAM_NETWORKS
		hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &mcb_resp_name, NULL);

		if (hdr) {
			pj_strdup_with_null(rdata->tp_info.pool, &value, &hdr->hvalue);
			ast_log(LOG_DEBUG, "[TK] X-Grandstream-Mcb-Resp: %s, len: %d",
								 pj_strbuf(&value), pj_strlen(&value));
			if (pj_strlen(&value) > 0) {
				sip_setmcbresp(session->channel, pj_strbuf(&value));
			}
		}

        hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg,
                                         &call_type_name, NULL);

		if (hdr) {
			pj_strdup_with_null(rdata->tp_info.pool, &value, &hdr->hvalue);

            if (pj_strlen(&value) > 0) {
                sip_setcalltype(session->channel, pj_strbuf(&value));
            }
		}
#endif

#ifdef GRANDSTREAM_NETWORKS
		if (0 == session->is_video_media_change) {
			ast_queue_control(session->channel, AST_CONTROL_ANSWER);
			session->is_send_direct_media_request = 0;
		} else {
			if (session->is_send_direct_media_request) {
				ast_queue_control(session->channel, AST_CONTROL_ANSWER);
				session->is_send_direct_media_request = 0;
			} else {
				struct ast_sip_video_reinvite_data reinvite_data;
				session->is_video_media_change = 0;
				ast_log(LOG_DEBUG, "chan_pjsip_incoming_response set is_video_media_change false\n");
				handle_response_video_reinvite(session, &reinvite_data);
				ast_queue_control_data(session->channel, AST_CONTROL_VIDEO_REINVITE_RSP, &reinvite_data, sizeof(reinvite_data));
			}
		}
#else
		ast_queue_control(session->channel, AST_CONTROL_ANSWER);
#endif

		break;
#ifdef GRANDSTREAM_NETWORKS
        case 486:
            cei_support = pbx_builtin_getvar_helper(session->channel, "CEI_SUPPORT");
    		conf_room = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");

            // Invited user rejected a call, so we need to terminate its subscription
            if (!ast_strlen_zero(cei_support) && !ast_strlen_zero(conf_room)) {
                publish_cei_unsub(conf_room, ast_channel_name(session->channel));
            }
        break;
#endif
	default:
		break;
	}

	/* Build and send the tech-specific cause information */
	/* size of the string making up the cause code is "SIP " number + " " + reason length */
	data_size += 4 + 4 + pj_strlen(&status.reason);
	cause_code = ast_alloca(data_size);
	memset(cause_code, 0, data_size);

	ast_copy_string(cause_code->chan_name, ast_channel_name(session->channel), AST_CHANNEL_NAME);

	snprintf(cause_code->code, data_size - sizeof(*cause_code) + 1, "SIP %d %.*s", status.code,
		(int) pj_strlen(&status.reason), pj_strbuf(&status.reason));

	cause_code->ast_cause = hangup_sip2cause(status.code);
	ast_queue_control_data(session->channel, AST_CONTROL_PVT_CAUSE_CODE, cause_code, data_size);
	ast_channel_hangupcause_hash_set(session->channel, cause_code, data_size);
}

static int chan_pjsip_incoming_ack(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	if (rdata->msg_info.msg->line.req.method.id == PJSIP_ACK_METHOD) {
		if (session->endpoint->media.direct_media.enabled && session->channel) {
			ast_queue_control(session->channel, AST_CONTROL_SRCCHANGE);
		}
	}
	return 0;
}

static int update_devstate(void *obj, void *arg, int flags)
{
	ast_devstate_changed(AST_DEVICE_UNKNOWN, AST_DEVSTATE_CACHABLE,
			     "PJSIP/%s", ast_sorcery_object_get_id(obj));
	return 0;
}

static struct ast_custom_function chan_pjsip_dial_contacts_function = {
	.name = "PJSIP_DIAL_CONTACTS",
	.read = pjsip_acf_dial_contacts_read,
};

static struct ast_custom_function media_offer_function = {
	.name = "PJSIP_MEDIA_OFFER",
	.read = pjsip_acf_media_offer_read,
	.write = pjsip_acf_media_offer_write
};

#ifdef GRANDSTREAM_NETWORKS
static int is_pjsip_channel(struct ast_channel *chan)
{
    struct ast_sip_channel_pvt *channel = NULL;

    if (chan == NULL) {
    	return -1;
    }

    if (strcmp(ast_channel_tech(chan)->type, "PJSIP") != 0) {
    	ast_log(LOG_WARNING, "This application supports only PJSIP channel\n");
    	return -1;
    }

    channel = ast_channel_tech_pvt(chan);
    if (channel == NULL || channel->pvt == NULL) {
    	ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n",
                             ast_channel_name(chan));
    	return -1;
    }

    return 0;
}

static int sip_voip_trunk_enable_fax_detect(struct ast_channel *chan, unsigned int fax_detect)
{
	if (fax_detect) {
		ast_channel_set_is_enable_fax_detect(chan, 1);
	} else {
		ast_channel_set_is_enable_fax_detect(chan, 0);
	}
	return 0;
}

static int add_peer_conference(const char *channel,
                    		   struct ast_sip_endpoint *endpoint,
							   const char *conference)
{
    struct sip_conference* conf = NULL;

    // Sanity check
    if (channel == NULL || endpoint == NULL || conference == NULL) {
        ast_log(LOG_WARNING, "Invalid arguments\n");
        return -1;
    }

    conf = ast_sip_cei_find_peer_conference(endpoint, conference);
    if (conf != NULL) {
        ast_log(LOG_WARNING, "[TK] Channel %s already subscribed to Conference %s\n",
                             channel, conference);
        return -1;
    }

	AST_LIST_LOCK(&endpoint->conferences);

    conf = ast_calloc(1, sizeof(*conf));
    if (conf == NULL) {
        ast_log(LOG_WARNING, "Failed to allocate new peer conference\n");
        AST_LIST_UNLOCK(&endpoint->conferences);
        return -1;
    }

    conf->conference = ast_strdup(conference);
    conf->channelName = ast_strdup(channel);
	conf->state = SIP_CONFERENCE_STATE_WAITSUB;
    conf->update_number = 1;
    conf->ceiJoined = 0;
    conf->ceiRole = CEI_ROLE_UNKNOWN; /* Added by achan, Grandstream for CCM */
    AST_LIST_INSERT_TAIL(&endpoint->conferences, conf, entry);

	AST_LIST_UNLOCK(&endpoint->conferences);

    ast_debug(LOG_DEBUG, "[TK] Channel %s subscribed to Conference %s",
                         conf->channelName, conf->conference);

    return 0;
}

/* BEGIN: Added by achan, Grandstream */
/*! Destroy conference subscriptions */
static void destroy_conference(struct sip_conference *conference)
{
    // Sanity check
    if (conference == NULL) {
        return;
    }

    ast_debug(LOG_DEBUG, "[TK] Destroying Channel %s, Conference %s\n",
                          conference->channelName, conference->conference);

    if (conference->channelName != NULL) {
        ast_free(conference->channelName);
        conference->channelName = NULL;
    }

    if (conference->conference != NULL) {
        ast_free(conference->conference);
        conference->conference = NULL;
    }

	ast_free(conference);
}

///*! Destroy all peer-related conference subscriptions, REQUIRES LOCK */
//static void clear_peer_conferences(struct ast_sip_endpoint *endpoint)
//{
//	struct sip_conference *conference;
//
//    // Sanity check
//    if (endpoint == NULL) {
//        ast_log(LOG_WARNING, "endpoint is NULL\n");
//        return;
//    }
//
//	while ((conference = AST_LIST_REMOVE_HEAD(&endpoint->conferences, entry))) {
//        destroy_conference(conference);
//    }
//}

/*! Destroy given peer-related conference subscription, REQUIRES LOCK */
static void clear_peer_conference(struct ast_sip_endpoint *endpoint,
                                  const char *channel)
{
	struct sip_conference *conference;

    // Sanity check
    if (endpoint == NULL || channel == NULL) {
        ast_log(LOG_WARNING, "Invalid arguments\n");
        return;
    }

    AST_LIST_TRAVERSE_SAFE_BEGIN(&endpoint->conferences, conference, entry) {
        if (strcmp(conference->channelName, channel) == 0) {
            AST_LIST_REMOVE_CURRENT(entry);
            destroy_conference(conference);
        }        
    }
    AST_LIST_TRAVERSE_SAFE_END;
}
/* END: Added by achan, Grandstream */

// /*! \brief Retrive remote address in ip v4 format from session datastore.
//     This returned remote address MUST be freed.
// */
// static char* get_session_remote_addr_ipv4(struct ast_sip_session *session)
// {
//     struct ast_datastore * datastore = NULL;
//     struct transport_info_data *transport_data = NULL;
//     char *ipv4 = NULL;
//
//     // Sanity check
//     if (session == NULL) {
//         ast_log(LOG_WARNING, "session is NULL\n");
//         return NULL;
//     }
//
//     datastore = ast_sip_session_get_datastore(session, "transport_info");
//     if (datastore == NULL || datastore->data == NULL) {
//         ast_log(LOG_WARNING, "No transport_info in session datastore\n");
//     }
//
//     transport_data = datastore->data;
//     ipv4 = pj_inet_ntoa(transport_data->remote_addr.ipv4.sin_addr);
//
//     if (ast_strlen_zero(ipv4)) {
//         ast_log(LOG_WARNING, "Failed to get session remote address ipv4\n");
//         ao2_ref(datastore, -1);
//         return NULL;
//     }
//
//     // Clean up
//     ao2_ref(datastore, -1);
//
//     return ast_strdup(ipv4);
// }

/*! \brief Retrive contact URI from AORs moved sip:.
    This returned URI MUST be freed.
*/
static char* get_contact_uri_from_aors(const char *aors)
{
    struct ast_sip_contact *contact = NULL;
    const char *origin_uri = NULL;
    char *uri = NULL;
    char *pUri = NULL;

    // Sanity check
    if (aors == NULL) {
        return NULL;
    }

    contact = ast_sip_location_retrieve_contact_from_aor_list(aors);
    if (contact == NULL || ast_strlen_zero(contact->uri)) {
        return NULL;
    }

    // Valid URI starts with sip: or sips:
    if (strncmp(contact->uri, "sip:", 4) == 0) {
        origin_uri = contact->uri + 4;
    }
    else if (strncmp(contact->uri, "sips:", 5) == 0) {
        origin_uri = contact->uri + 5;
    }
    else {
        return NULL;
    }

    // IP address is next to username@
    pUri = strchr(origin_uri, '@');
    if (pUri == NULL) {
        uri = ast_strdup(origin_uri);
    }
    else {
        uri = ast_strdup(pUri + 1);
    }

    // Remove port
    pUri = strchr(uri, ':');
    if (pUri != NULL) {
        *pUri = '\0';
    }

    // Remove transport
    pUri = strchr(uri, ';');
    if (pUri != NULL) {
        *pUri = '\0';
    }
    
    return uri;
}

/*!
 * \brief Prepare and publish an \ref ast_cei_ext_mcb_sub_type message to the
 * \ref ast_cei_mcb_sub_topic_all.
 */
static void publish_cei_ext_mcb_sub(const char *peer_name,
                                    struct cei_external_mcb_peer *mcb_peer)
{
    struct stasis_topic *cei_mcb_sub_topic = NULL;
    RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
    RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

    // Sanity check
    if (peer_name == NULL || mcb_peer == NULL) {
        return;
    }

    cei_mcb_sub_topic = ast_cei_mcb_sub_topic_all();
	if (cei_mcb_sub_topic == NULL) {
        ast_log(LOG_WARNING, "Failed to get ast_cei_mcb_sub_topic_all");
        return;
	}

    json_object = ast_json_pack("{s: s, s: s}",
                                "peer_name", peer_name,
                                "mcb_peer_id", mcb_peer->ref_id);

    if (json_object == NULL) {
        ast_log(LOG_WARNING, "Failed to create json_object");
        return;
    }

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
		return;
	}

    msg = stasis_message_create(ast_cei_ext_mcb_sub_type(), payload);

    if (msg == NULL) {
        ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_ext_mcb_sub_type");
        return;
    }

    stasis_publish(cei_mcb_sub_topic, msg);
}

/*!
 * \brief Prepare and publish an \ref ast_cei_ext_mcb_unsub_type message to the
 * \ref ast_cei_mcb_sub_topic_all.
 */
static int publish_cei_ext_mcb_unsub(const char *peer_name,
                                      struct cei_external_mcb_peer *mcb_peer)
{
    struct stasis_topic *cei_mcb_sub_topic = NULL;
    RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
    RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

    // Sanity check
    if (peer_name == NULL || mcb_peer == NULL) {
        return -1;
    }

    cei_mcb_sub_topic = ast_cei_mcb_sub_topic_all();
	if (cei_mcb_sub_topic == NULL) {
        ast_log(LOG_WARNING, "Failed to get ast_cei_mcb_sub_topic_all");
        return -1;
	}

    json_object = ast_json_pack("{s: s, s: s, s: s, s: s}",
                                "peer_name", peer_name,
                                "mcb_peer_id", mcb_peer->ref_id,
                                "remote_conf_id", mcb_peer->remote_conf_id,
                                "local_conf_id", mcb_peer->local_conf_id);

    if (json_object == NULL) {
        ast_log(LOG_WARNING, "Failed to create json_object");
        return -1;
    }

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
		return -1;
	}

    msg = stasis_message_create(ast_cei_ext_mcb_unsub_type(), payload);

    if (msg == NULL) {
        ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_ext_mcb_unsub_type");
        return -1;
    }

    stasis_publish(cei_mcb_sub_topic, msg);

    return 0;
}

/*!
 * \brief Prepare and publish an \ref ast_cei_int_mcb_sub_type message to the
 * \ref ast_cei_mcb_sub_topic_all.
 */
static void publish_cei_int_mcb_sub(struct cei_internal_mcb_peer *mcb_peer)
{
  ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
  struct stasis_topic *cei_mcb_sub_topic = NULL;
  RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
  RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
  RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

  // Sanity check
  if (mcb_peer == NULL) {
    return;
  }

  cei_mcb_sub_topic = ast_cei_mcb_sub_topic_all();
	if (cei_mcb_sub_topic == NULL) {
    ast_log(LOG_WARNING, "Failed to get ast_cei_mcb_sub_topic_all");
    return;
	}

  json_object = ast_json_pack("{s: s}",
                              "mcb_peer_id", mcb_peer->ref_id);

  ast_log(LOG_DEBUG, "[WL] publish_cei_int_mcb_sub prepare to send data: %s,%s,%s,%s,%s\n",
                          mcb_peer->local_conf,
                          mcb_peer->local_conf_id,
                          mcb_peer->remote_conf,
                          mcb_peer->remote_conf_id,
                          mcb_peer->channelName);

  if (json_object == NULL) {
    ast_log(LOG_WARNING, "Failed to create json_object");
    return;
  }

  payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
		return;
	}

  msg = stasis_message_create(ast_cei_int_mcb_sub_type(), payload);

  if (msg == NULL) {
    ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_ext_mcb_sub_type");
    return;
  }

  stasis_publish(cei_mcb_sub_topic, msg);
}

/*!
 * \brief Prepare and publish an \ref ast_cei_int_mcb_unsub_type message to the
 * \ref ast_cei_mcb_sub_topic_all.
 */
static void publish_cei_int_mcb_unsub(struct cei_internal_mcb_peer *mcb_peer)
{
  struct stasis_topic *cei_mcb_sub_topic = NULL;
  RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
  RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
  RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

  // Sanity check
  if (mcb_peer == NULL) {
      return;
  }

  cei_mcb_sub_topic = ast_cei_mcb_sub_topic_all();
	if (cei_mcb_sub_topic == NULL) {
    ast_log(LOG_WARNING, "Failed to get ast_cei_mcb_sub_topic_all");
    return;
	}

  json_object = ast_json_pack("{s: s}",
                              "mcb_peer_id", mcb_peer->ref_id);

  ast_log(LOG_DEBUG, "[WL] publish_cei_int_mcb_unsub prepare to send data: %s,%s,%s,%s,%s\n",
                          mcb_peer->local_conf,
                          mcb_peer->local_conf_id,
                          mcb_peer->remote_conf,
                          mcb_peer->remote_conf_id,
                          mcb_peer->channelName);

  if (json_object == NULL) {
    ast_log(LOG_WARNING, "Failed to create json_object");
    return;
  }

  payload = ast_json_payload_create(json_object);
	if(NULL == payload) {
		return;
	}

  msg = stasis_message_create(ast_cei_int_mcb_unsub_type(), payload);

  if (msg == NULL) {
    ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_int_mcb_unsub_type");
    return;
  }

    stasis_publish(cei_mcb_sub_topic, msg);
}

/*! \brief Set call type in SIP response */
static int sip_setcalltype(struct ast_channel *chan, const char *data) {
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt;
    const char *inbuf = data;
    char *subbuf = NULL;
    size_t len = 0;

	if (ast_strlen_zero(inbuf)) {
		ast_log(LOG_WARNING, "This application requires the argument: Type\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "This application supports only PJSIP channel\n");
		return 0;
	}

    ast_channel_lock(chan);

	pvt = channel->pvt;
	if (!pvt) {
		ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n",
                             ast_channel_name(chan));
        ast_channel_unlock(chan);
		return -1;
	}

    len = strlen(inbuf);
	subbuf = (char*)malloc(len + 1);
	ast_get_encoded_str(inbuf, subbuf, len + 1);

	if (pvt->call_type != NULL) {
		ast_log(LOG_DEBUG, "Channel: %s, call_type: %s\n",
							ast_channel_name(chan), pvt->call_type);

		ast_free(pvt->call_type);
		pvt->call_type = NULL;
	}
	pvt->call_type = subbuf;

    ast_log(LOG_DEBUG, "[TK] Channel: %s sets call type response hash to: %s\n",
						ast_channel_name(chan), pvt->call_type);

	ast_channel_unlock(chan);

	return 0;
}

static struct ast_custom_function pjsip_mcb_resp_function = {
	.name = "MCB_RESPONSE",
	.read = func_mcb_resp_get,
};

/*! \brief Add a SIP header to an outbound INVITE */
static int sip_setmcbresp(struct ast_channel *chan, const char *data) {
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt;
    const char *inbuf = data;
    char *subbuf = NULL;
    size_t len = 0;

	if (ast_strlen_zero(inbuf)) {
		ast_log(LOG_WARNING, "This application requires the argument: Header\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "This application supports only PJSIP channel\n");
		return 0;
	}

    ast_channel_lock(chan);

	pvt = channel->pvt;
	if (!pvt) {
		ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n",
                             ast_channel_name(chan));
        ast_channel_unlock(chan);
		return -1;
	}

    len = strlen(inbuf);
	subbuf = (char*)malloc(len + 1);
	ast_get_encoded_str(inbuf, subbuf, len + 1);

	if (pvt->mcb_response != NULL) {
		ast_log(LOG_DEBUG, "Channel: %s, mcb_response: %s\n",
							ast_channel_name(chan), pvt->mcb_response);

		ast_free(pvt->mcb_response);
		pvt->mcb_response = NULL;
	}
	pvt->mcb_response = subbuf;

    ast_log(LOG_DEBUG, "[TK] Channel: %s sets MCB response hash to: %s\n",
						ast_channel_name(chan), pvt->mcb_response);

	ast_channel_unlock(chan);

	return 0;
}

static int sip_set_conf_focus(struct ast_channel *chan, const char *data)
{
    struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt = NULL;
    struct ast_sip_session *session = NULL;
    struct ast_sip_endpoint *endpoint = NULL;
    const char *inbuf = data;
    const char *id = NULL;

	if (ast_strlen_zero(inbuf)) {
		ast_log(LOG_WARNING, "This application requires the argument: Room number\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "This application supports only PJSIP channel\n");
		return 0;
	}

    ast_channel_lock(chan);

	pvt = channel->pvt;
	if (pvt == NULL) {
		ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n",
                             ast_channel_name(chan));
        ast_channel_unlock(chan);
		return 0;
	}

    session = channel->session;
    if (session == NULL || session->endpoint == NULL) {
        ast_log(LOG_WARNING, "Channel %s has no end point!\n",
                             ast_channel_name(chan));
        ast_channel_unlock(chan);
        return 0;
    }

    endpoint = session->endpoint;
    id = ast_sorcery_object_get_id(endpoint);
    if (id == NULL || strstr(id, "trunk") != NULL) {
        ast_log(LOG_WARNING, "Channel %s has invalid end point (%s)!\n",
                             ast_channel_name(chan), S_COR(id, id, "NULL"));
        ast_channel_unlock(chan);
        return 0;
	}

    pvt->isConference = TRUE;
    if (globalEnabledCEI) {
      add_peer_conference(ast_channel_name(chan), endpoint, inbuf);
    }

	ast_channel_unlock(chan);

	return 0;
}

/*! \brief Set Contact URI Host User Part during responding of an incoming INVITE */
static int sip_set_contact_user(struct ast_channel *chan, const char *data) {
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt;
    const char *inbuf = data;
    char *subbuf = NULL;
    size_t len = 0;

	if (ast_strlen_zero(inbuf)) {
		ast_log(LOG_WARNING, "This application requires the argument: User\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "This application supports only PJSIP channel\n");
		return 0;
	}

    ast_channel_lock(chan);

	pvt = channel->pvt;
	if (!pvt) {
		ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n",
                             ast_channel_name(chan));
        ast_channel_unlock(chan);
		return -1;
	}

    len = strlen(inbuf);
	subbuf = (char*)malloc(len + 1);
	ast_get_encoded_str(inbuf, subbuf, len + 1);

	if (pvt->contact_user != NULL) {
		ast_log(LOG_DEBUG, "Channel: %s, contact_user: %s\n",
							ast_channel_name(chan), pvt->contact_user);

		ast_free(pvt->contact_user);
		pvt->contact_user = NULL;
	}
	pvt->contact_user = subbuf;

    ast_log(LOG_DEBUG, "[TK] Channel: %s sets Contact URI Host User Part to: %s\n",
						ast_channel_name(chan), pvt->contact_user);

	ast_channel_unlock(chan);

	return 0;
}

/*!
 * \brief Prepare and publish an \ref ast_cei_unsub_type message to the
 * \ref ast_cei_sub_topic_all.
 */
static void publish_cei_unsub(const char *conference, const char *channel)
{
    struct stasis_topic *cei_sub_topic = NULL;
    RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
    RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

    // Sanity check
    if (conference == NULL || channel == NULL) {
        return;
    }

    cei_sub_topic = ast_cei_sub_topic_all();
	if (cei_sub_topic == NULL) {
        ast_log(LOG_WARNING, "Failed to get ast_cei_sub_topic_all");
        return;
	}

    json_object = ast_json_pack("{s: s, s: s}",
                                "conference", conference,
                                "channel", channel);

    if (json_object == NULL) {
        ast_log(LOG_WARNING, "Failed to create json_object");
        return;
    }

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
		return;
	}

    msg = stasis_message_create(ast_cei_unsub_type(), payload);
    if (msg == NULL) {
        ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_unsub_type");
        return;
    }

    stasis_publish(cei_sub_topic, msg);
}

/*!
 * \brief Prepare and publish an \ref ast_cei_joined_type message to the
 * \ref ast_cei_sub_topic_all.
 */
static void publish_cei_joined(const char *conference, const char *channel)
{
    struct stasis_topic *cei_sub_topic = NULL;
    RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
    RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

    // Sanity check
    if (conference == NULL || channel == NULL) {
        return;
    }

    cei_sub_topic = ast_cei_sub_topic_all();
	if (cei_sub_topic == NULL) {
        ast_log(LOG_WARNING, "Failed to get ast_cei_sub_topic_all");
        return;
	}

    json_object = ast_json_pack("{s: s, s: s}",
                                "conference", conference,
                                "channel", channel);

    if (json_object == NULL) {
        ast_log(LOG_WARNING, "Failed to create json_object");
        return;
    }

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
		return;
	}

    msg = stasis_message_create(ast_cei_joined_type(), payload);

    if (msg == NULL) {
        ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_joined_type");
        return;
    }

    stasis_publish(cei_sub_topic, msg);
}

/*! \brief Send stasis message to generate SIP NOTIFY message with terminated state */
static int sip_terminated_notify(struct ast_channel *chan, const char *data)
{
    // Sanity check
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "This application requires the argument: Conference\n");
		return 0;
	}

    ast_channel_lock(chan);

    publish_cei_unsub(data, ast_channel_name(chan));

	ast_channel_unlock(chan);

	return 0;
}

/*! \brief Given channel joined to a conference with valid authentication */
static int sip_cei_joined(struct ast_channel *chan, const char *data)
{
    struct ast_sip_channel_pvt *channel = NULL;
    struct ast_sip_endpoint *endpoint = NULL;
    struct sip_conference* conf = NULL;

    // Sanity check
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "This application requires the argument: Conference\n");
		return 0;
	}

    if (is_pjsip_channel(chan) == -1) {
		return 0;
	}

    ast_channel_lock(chan);

    // Update ceiJoined status
    channel = ast_channel_tech_pvt(chan);
    if (channel == NULL || channel->session == NULL ||
        channel->session->endpoint == NULL) {
        ast_log(LOG_WARNING, "Failed to get endpoint of channel %s\n",
                             ast_channel_name(chan));
        ast_channel_unlock(chan);
        return 0;
    }

    endpoint = channel->session->endpoint;
    conf = ast_sip_cei_find_peer_conference(endpoint, data);
    if (conf == NULL) {
        ast_log(LOG_WARNING, "Failed to get peer conference: chan: %s, conf: %s\n",
                             ast_channel_name(chan), data);
        ast_channel_unlock(chan);
        return 0;
    }
    conf->ceiJoined = 1;

    ast_log(LOG_WARNING, "[TK] Channel %s in conference %s has valid authentication\n",
                         ast_channel_name(chan), data);

    // Notify ceiJoined status to CEI module
    publish_cei_joined(data, ast_channel_name(chan));

	ast_channel_unlock(chan);

	return 0;
}

/*! \brief find the caller id of incoming call bridged by register sip trunk */
/*! \return need to be free*/
static char* get_register_trunk_caller_id(const char * extension_num)
{
  char family[80] = {0};
  char caller_id[64] = {0};
  char outbound_cid[128] = {0};
  char *ret;

  // Sanity check
  if (extension_num == NULL || ast_strlen_zero(extension_num)) {
    return NULL;
  }

  snprintf(family, sizeof(family), "AAUSER/%s", extension_num);
  ast_ucm_db_get(family, "OUTBOUNDCID", outbound_cid, sizeof(outbound_cid));
  if (ast_strlen_zero(outbound_cid)) {
      ast_log(LOG_WARNING, "[WL] No customized caller id of %s\n", family);
      return NULL;
  }

  ret = strrchr(outbound_cid, '<');
  if (ret == NULL || ast_strlen_zero(ret + 2)) {
    return NULL;
  }

  strncpy(caller_id, ret + 1, strlen(ret) -2);
  return ast_strdup(caller_id);
}

static int sip_external_mcb_sub(struct ast_channel *chan, const char *data)
{
  ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
  if (!globalEnabledCEI) {
    return 0;
  }

    struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt = NULL;
    struct ast_sip_session *session = NULL;
    struct ast_sip_endpoint *endpoint = NULL;
    char *inbuf = NULL;
	char *local_conf = NULL;
    char *local_conf_id = NULL;
    char *remote_conf = NULL;
    char *remote_conf_id = NULL;
    char *remote_channel = NULL;
    char *endpoint_name = NULL;
    char *call_type = NULL;
	char *saveptr = NULL;
    char local_conf_private_num[128];
    char remote_conf_public_num[128];
  	char *remote_conf_private_num = NULL;
    char *ip_addr = NULL;
    char outbound_prepend[MAX_PREPEND_LEN] = {0};
    char inbound_prepend[MAX_PREPEND_LEN]  = {0};
    int pjsip_channel = 0;
    int outbound_strip = 0;
    int inbound_strip = 0;
    struct cei_external_mcb_peer *mcb_peer = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint0, NULL, ao2_cleanup);

    if (ast_strlen_zero(data)) {
        ast_log(LOG_WARNING, "This application requires the arguments: "
    		                 "local conference, local conference ID, "
                             "remote conference, remote conference ID, "
                             "remote channel, endpoint name, call type\n");
		return 0;
	}

    if (strcmp(ast_channel_tech(chan)->type, "PJSIP") &&
        strcmp(ast_channel_tech(chan)->type, "Local")) {
		ast_log(LOG_WARNING, "This application supports only PJSIP/Local channel\n");
		return 0;
	}

    if (strcmp(ast_channel_tech(chan)->type, "PJSIP") == 0) {
        pjsip_channel = 1;
    }

    ast_channel_lock(chan);

    //=================================================
    // Parse arguments
    //=================================================
    inbuf = ast_strdupa(data);

    // Get local conference
    local_conf = strtok_r(inbuf, ",", &saveptr);
    if (ast_strlen_zero(local_conf)) {
        ast_log(LOG_WARNING, "Failed to parse local conference\n");
        ast_channel_unlock(chan);
		return 0;
	}

    local_conf_id = strtok_r(NULL, ",", &saveptr);
	if (!local_conf_id || strlen(local_conf_id) != 40) {
		ast_log(LOG_WARNING, "Failed to parse local conference ID in SHA1 hash\n");
        ast_channel_unlock(chan);
		return 0;
	}

    remote_conf = strtok_r(NULL, ",", &saveptr);
    if (ast_strlen_zero(remote_conf)) {
        ast_log(LOG_WARNING, "Failed to parse remote conference\n");
        ast_channel_unlock(chan);
		return 0;
	}

    remote_conf_id = strtok_r(NULL, ",", &saveptr);
	if (!remote_conf_id || strlen(remote_conf_id) != 40) {
		ast_log(LOG_WARNING, "Failed to parse remote conference ID in SHA1 hash\n");
        ast_channel_unlock(chan);
		return 0;
	}

    remote_channel = strtok_r(NULL, ",", &saveptr);
    if (ast_strlen_zero(remote_channel)) {
        ast_log(LOG_WARNING, "Failed to parse remote channel\n");
        ast_channel_unlock(chan);
		return 0;
	}

    endpoint_name = strtok_r(NULL, ",", &saveptr);
    if (ast_strlen_zero(endpoint_name)) {
        ast_log(LOG_WARNING, "Failed to parse endpoint name\n");
        ast_channel_unlock(chan);
		return 0;
	}

    call_type  = strtok_r(NULL, "", &saveptr);
    if (ast_strlen_zero(call_type)) {
        ast_log(LOG_WARNING, "Failed to parse call type\n");
        ast_channel_unlock(chan);
		return 0;
	}

    //=================================================
    // Get session and endpoint
    //=================================================
    // PJSIP channel
    if (pjsip_channel) {
        pvt = channel->pvt;
    	if (pvt == NULL) {
    		ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n",
                                 ast_channel_name(chan));
            ast_channel_unlock(chan);
    		return 0;
    	}

        // Get session
        session = channel->session;

        // Get endpoint
        if (session != NULL) {
            endpoint = session->endpoint;
        }
    }
    // Local channel
    else {
        // Get session
        session = channel->session;

        // Get endpoint
        endpoint = endpoint0 = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
                                              endpoint_name);
    }

    if (session == NULL || endpoint == NULL) {
        ast_log(LOG_WARNING, "Channel %s has no end point!\n",
                             ast_channel_name(chan));
        ast_channel_unlock(chan);
        return 0;
    }

    //=================================================
    // Save new subscription in endpoint
    //=================================================
    if (ast_strlen_zero(endpoint->aors)) {
        ast_log(LOG_WARNING, "Failed to get endpoint aors\n");
        ast_channel_unlock(chan);
		return 0;
	}

    memset(remote_conf_public_num, 0, sizeof(remote_conf_public_num));
    if (strcmp(call_type, CEI_MCB_DIR_OUTGOING) == 0) {
		if (ast_cei_get_trunk_info_by_ext(endpoint->aors, CEI_MCB_DIR_OUTGOING, remote_conf, outbound_prepend, &outbound_strip)) {
			ast_log(LOG_WARNING, "Failed to get outgoing informations for trunk: %s\n", endpoint->aors);
			ast_channel_unlock(chan);
			return 0;
		}
        if (!ast_strlen_zero(outbound_prepend)) {
          strncpy(remote_conf_public_num, outbound_prepend, strlen(outbound_prepend));
        }
    }

    strcat(remote_conf_public_num, remote_conf + outbound_strip);

    // Get remote address
    ip_addr = get_contact_uri_from_aors(endpoint->aors);
    if (ast_strlen_zero(ip_addr)) {
        ast_channel_unlock(chan);
        return 0;
    }

    mcb_peer = ast_cei_add_external_mcb_peer(endpoint, local_conf, local_conf_id,
                                             remote_conf_public_num, remote_conf_id, ip_addr,
                                             (pjsip_channel) ? ast_channel_name(chan) : remote_channel);

    AST_LIST_LOCK(&endpoint->external_mcb_peers);

    if(mcb_peer == NULL) {
        ast_channel_unlock(chan);
        ast_free(ip_addr);
        AST_LIST_UNLOCK(&endpoint->external_mcb_peers);

        return 0;
    }

    //Retrieve correct provate num of local conf and remote conf
    if (strcmp(call_type, CEI_MCB_DIR_INCOMING) == 0) {
      if (strstr(endpoint_name, "trunk_") == NULL) {
        ast_log(LOG_DEBUG, "[WL] Incoming mcb via Register SIP Trunk \n");
        remote_conf_private_num = get_register_trunk_caller_id(endpoint_name);

        if (remote_conf_private_num != NULL && !ast_strlen_zero(remote_conf_private_num)) {
          ast_string_field_set(mcb_peer, conf_private_num, remote_conf_private_num);
          ast_free(remote_conf_private_num);
        }
      } else {
        ast_log(LOG_DEBUG, "[WL] Incoming mcb via Peer SIP Trunk \n");

        memset(local_conf_private_num, 0, sizeof(local_conf_private_num));
		if (ast_cei_get_trunk_info_by_ext(endpoint->aors, CEI_MCB_DIR_INCOMING, mcb_peer->local_conf, inbound_prepend, &inbound_strip)) {
			ast_log(LOG_WARNING, "Failed to get incoming informations for trunk: %s\n", mcb_peer->local_conf);
			ast_channel_unlock(chan);
			return 0;
		}

        if (!ast_strlen_zero(inbound_prepend)) {
          strncpy(local_conf_private_num, inbound_prepend, strlen(inbound_prepend));
        }

         strcat(local_conf_private_num, mcb_peer->local_conf + inbound_strip);

      }
    } else if (strcmp(call_type, CEI_MCB_DIR_OUTGOING) == 0) {
      ast_log(LOG_DEBUG, "[WL] Outgoing mcb \n");
      ast_string_field_set(mcb_peer, conf_private_num, remote_conf);
    }

    // Notify res_pjsip_cei to generate SIP SUBSCRIBE
    publish_cei_ext_mcb_sub(ast_sorcery_object_get_id(endpoint), mcb_peer);

	// Clean up
	ast_channel_unlock(chan);
    ast_free(ip_addr);
    AST_LIST_UNLOCK(&endpoint->external_mcb_peers);

	return 0;
}

static int sip_external_mcb_unsub(struct ast_channel *chan, const char *data)
{
  ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
  if (!globalEnabledCEI) {
    return 0;
  }
    struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt = NULL;
    struct ast_sip_session *session = NULL;
    struct ast_sip_endpoint *endpoint = NULL;
    char *inbuf = NULL;
    char *local_conf_id = NULL;
    char *remote_conf_id = NULL;
    char *endpoint_name = NULL;
	char *saveptr = NULL;
    int pjsip_channel = 0;
    struct cei_external_mcb_peer *mcb_peer = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint0, NULL, ao2_cleanup);

    if (ast_strlen_zero(data)) {
        ast_log(LOG_WARNING, "This application requires the arguments: "
    		                 "local conference ID, remote conference ID, "
                             "endpoint name\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP") &&
        strcmp(ast_channel_tech(chan)->type, "Local")) {
		ast_log(LOG_WARNING, "This application supports only PJSIP/Local channel\n");
		return 0;
	}

    if (strcmp(ast_channel_tech(chan)->type, "PJSIP") == 0) {
        pjsip_channel = 1;
    }

    ast_channel_lock(chan);

    //=================================================
    // Parse arguments
    //=================================================
    inbuf = ast_strdupa(data);

    // Get local conference
    local_conf_id = strtok_r(inbuf, ",", &saveptr);
	if (!local_conf_id || strlen(local_conf_id) != 40) {
		ast_log(LOG_WARNING, "Failed to parse local conference ID in SHA1 hash\n");
        ast_channel_unlock(chan);
		return 0;
	}

    remote_conf_id = strtok_r(NULL, ",", &saveptr);
	if (!remote_conf_id || strlen(remote_conf_id) != 40) {
		ast_log(LOG_WARNING, "Failed to parse remote conference ID in SHA1 hash\n");
        ast_channel_unlock(chan);
		return 0;
	}

    endpoint_name = strtok_r(NULL, "", &saveptr);
    if (ast_strlen_zero(endpoint_name)) {
        ast_log(LOG_WARNING, "Failed to parse endpoint name\n");
        ast_channel_unlock(chan);
		return 0;
	}

    //=================================================
    // Get session and endpoint
    //=================================================
    // PJSIP channel
    if (pjsip_channel) {
        pvt = channel->pvt;
    	if (pvt == NULL) {
    		ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n",
                                 ast_channel_name(chan));
            ast_channel_unlock(chan);
    		return 0;
    	}

        // Get session
        session = channel->session;

        // Get endpoint
        if (session != NULL) {
            endpoint = session->endpoint;
        }
    }
    // Local channel
    else {
        // Get session
        session = channel->session;

        // Get endpoint
        endpoint = endpoint0 = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
                                              endpoint_name);
    }

    if (session == NULL || endpoint == NULL) {
        ast_log(LOG_WARNING, "Channel %s has no end point!\n",
                             ast_channel_name(chan));
        ast_channel_unlock(chan);
        return 0;
    }

    //=================================================
    // Unsubscription
    //=================================================
    // Get MCB peer
    mcb_peer = ast_cei_find_external_mcb_peer_by_id(endpoint, local_conf_id, remote_conf_id);
    if (mcb_peer == NULL) {
        ast_log(LOG_WARNING, "Failed to get mcb peer of local id: %s, remote id: %s\n",
                             local_conf_id, remote_conf_id);
        ast_channel_unlock(chan);
        return 0;
    }

    // Notify res_pjsip_pubsub to generate SIP Un-SUBSCRIBE
    if (publish_cei_ext_mcb_unsub(ast_sorcery_object_get_id(endpoint), mcb_peer) == -1) {
        // Remove given subscription in endpoint if the subscription is failed.
        ast_cei_remove_external_mcb_peer(endpoint, local_conf_id, remote_conf_id);
    }

	// Clean up
	ast_channel_unlock(chan);

	return 0;
}

static int sip_internal_mcb_sub_rev(const char *local_conf, const char *local_conf_id,
                                    const char *remote_conf, const char *remote_conf_id,
                                    const char *channelName)
{
  // Sanity check
  if (local_conf == NULL || local_conf_id == NULL ||
      remote_conf == NULL || remote_conf_id == NULL || channelName == NULL) {
    ast_log(LOG_WARNING, "Invalid arguments\n");
    return 0;
  }

  struct cei_internal_mcb_peer *mcb_peer = NULL;

  mcb_peer = ast_cei_add_int_mcb_peer(local_conf, local_conf_id,
                                      remote_conf, remote_conf_id,
                                      channelName);
  if(mcb_peer == NULL) {
    return 0;
  }

  // Notify res_pjsip_cei to generate SIP SUBSCRIBE
  publish_cei_int_mcb_sub(mcb_peer);

  return 0;
}

static int sip_internal_mcb_sub(struct ast_channel *chan, const char *data)
{
  ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
  if (!globalEnabledCEI) {
    return 0;
  }
  char *inbuf = NULL;
  char *local_conf = NULL;
  char *local_conf_id = NULL;
  char *remote_conf = NULL;
  char *remote_conf_id = NULL;
  char *saveptr = NULL;
  struct cei_internal_mcb_peer *mcb_peer = NULL;

  if (ast_strlen_zero(data)) {
    ast_log(LOG_WARNING, "[WL] Invalid data\n");
    return 0;
  }

  if (strcmp(ast_channel_tech(chan)->type, "Local")) {
    ast_log(LOG_WARNING, "This application supports only Local channel\n");
    return 0;
  }

  ast_channel_lock(chan);

  //=================================================
  // Parse arguments
  //=================================================
  inbuf = ast_strdupa(data);

  // Get local conference
  local_conf = strtok_r(inbuf, ",", &saveptr);
  if (ast_strlen_zero(local_conf)) {
      ast_log(LOG_WARNING, "Failed to parse local conference\n");
      ast_channel_unlock(chan);
  return 0;
  }

  local_conf_id = strtok_r(NULL, ",", &saveptr);
  if (!local_conf_id || strlen(local_conf_id) != 40) {
    ast_log(LOG_WARNING, "Failed to parse local conference ID in SHA1 hash\n");
    ast_channel_unlock(chan);
    return 0;
  }

  remote_conf = strtok_r(NULL, ",", &saveptr);
  if (ast_strlen_zero(remote_conf)) {
    ast_log(LOG_WARNING, "Failed to parse remote conference\n");
    ast_channel_unlock(chan);
    return 0;
  }

  remote_conf_id = strtok_r(NULL, "", &saveptr);
  if (!remote_conf_id || strlen(remote_conf_id) != 40) {
    ast_log(LOG_WARNING, "Failed to parse remote conference ID in SHA1 hash\n");
    ast_channel_unlock(chan);
    return 0;
  }

  mcb_peer = ast_cei_add_int_mcb_peer(local_conf, local_conf_id,
                                   remote_conf, remote_conf_id,
                                   ast_channel_name(chan));
  if(mcb_peer == NULL) {
    ast_channel_unlock(chan);
    return 0;
  }

  // Notify res_pjsip_cei to generate SIP SUBSCRIBE
  publish_cei_int_mcb_sub(mcb_peer);

  sip_internal_mcb_sub_rev(remote_conf, remote_conf_id,
                          local_conf, local_conf_id,
                          ast_channel_name(chan));
  // Clean up
  ast_channel_unlock(chan);

  return 0;
}

static int sip_internal_mcb_unsub_rev(const char *local_conf_id,
                                      const char *remote_conf_id)
{
  // Sanity check
  if (local_conf_id == NULL || remote_conf_id == NULL) {
      ast_log(LOG_WARNING, "Invalid arguments\n");
      return 0;
  }

  struct cei_internal_mcb_peer *mcb_peer = NULL;

  mcb_peer = ast_cei_find_int_mcb_peer(local_conf_id, remote_conf_id);
  if(mcb_peer == NULL) {
    ast_log(LOG_WARNING, "Failed to get mcb peer of local id: %s, remote id: %s\n",
                         local_conf_id, remote_conf_id);
    return 0;
  }

  // Notify res_pjsip_cei to generate SIP SUBSCRIBE
  publish_cei_int_mcb_unsub(mcb_peer);

  return 0;
}

static int sip_internal_mcb_unsub(struct ast_channel *chan, const char *data)
{
  ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
  if (!globalEnabledCEI) {
    return 0;
  }
  char *inbuf = NULL;
  char *local_conf_id = NULL;
  char *remote_conf_id = NULL;
  char *saveptr = NULL;
  struct cei_internal_mcb_peer *mcb_peer = NULL;

  if (ast_strlen_zero(data)) {
    ast_log(LOG_WARNING, "[WL] Invalid Data\n");
    return 0;
  }
  if (strcmp(ast_channel_tech(chan)->type, "Local")) {
    ast_log(LOG_WARNING, "This application supports only Local channel\n");
    return 0;
  }

  ast_channel_lock(chan);

  //=================================================
  // Parse arguments
  //=================================================
  inbuf = ast_strdupa(data);

  // Get local conference
  local_conf_id = strtok_r(inbuf, ",", &saveptr);
  if (!local_conf_id || strlen(local_conf_id) != 40) {
    ast_log(LOG_WARNING, "Failed to parse local conference ID in SHA1 hash\n");
    ast_channel_unlock(chan);
    return 0;
  }

  remote_conf_id = strtok_r(NULL, "", &saveptr);
  if (!remote_conf_id || strlen(remote_conf_id) != 40) {
    ast_log(LOG_WARNING, "Failed to parse remote conference ID in SHA1 hash\n");
    ast_channel_unlock(chan);
    return 0;
  }
  //=================================================
  // Save new subscription in endpoint
  //=================================================


  mcb_peer = ast_cei_find_int_mcb_peer(local_conf_id, remote_conf_id);
  if(mcb_peer == NULL) {
    ast_log(LOG_WARNING, "Failed to get mcb peer of local id: %s, remote id: %s\n",
                         local_conf_id, remote_conf_id);
    ast_channel_unlock(chan);
    return 0;
  }

  // [TK] TODO: Notify CEI manager about new subscription

  // Notify res_pjsip_cei to generate SIP SUBSCRIBE
  publish_cei_int_mcb_unsub(mcb_peer);

  sip_internal_mcb_unsub_rev(remote_conf_id, local_conf_id);
  // Clean up
  ast_channel_unlock(chan);

  return 0;
}

static void cei_toggle_enable_cb(void *data, struct stasis_subscription *sub,
								  struct stasis_message *message)
{
	struct ast_json			*json_object = NULL;
	struct ast_json_payload *payload	 = NULL;
	const char *enableCEI = NULL;

	// Sanity check
	if (message == NULL) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	enableCEI = ast_json_string_get(ast_json_object_get(json_object, "enableCEI"));
	if (enableCEI == NULL) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return;
	}
  
	if (strcmp(enableCEI, "true") == 0) {
		globalEnabledCEI = 1;
	} else if (strcmp(enableCEI, "false") == 0) {
		globalEnabledCEI = 0;
	}
}

static int ast_cei_get_enable_configuration(void)
{
	char buf[64] = { 0 };
	char buffer[256] = { 0 };
	int enableCEI = 0;

	FILE *fp = NULL;
	fp = popen("nvram get enableCEI", "r");

	if (fp != NULL) {
		while (fp != NULL && (!feof(fp))
		      && (fgets(buf, sizeof(buf) - 1, fp) != NULL)) {
			strcat(buffer, buf);
		}
		pclose(fp);
		fp = NULL;

		if (buffer[0] != '\0') {
		   /*parsing result*/
		   enableCEI = atoi( buffer );
		}
	}

	return enableCEI;
}

/*! dial_options convert to feature right */
static void set_feature_map(const char* dial_options)
{
	if (NULL == dial_options) {
		return;
	}

	if (strchr(dial_options, 't')) {
		gs_feature_map[GS_FEATURE_BLINDXFER].f_right |= GS_FEATURE_ALLOW_CALLEE;
		gs_feature_map[GS_FEATURE_ATXFER].f_right    |= GS_FEATURE_ALLOW_CALLEE;
	}

	if (strchr(dial_options, 'T')) {
		gs_feature_map[GS_FEATURE_BLINDXFER].f_right |= GS_FEATURE_ALLOW_CALLER;
		gs_feature_map[GS_FEATURE_ATXFER].f_right    |= GS_FEATURE_ALLOW_CALLER;
	}

	if (strchr(dial_options, 'x')) {
		gs_feature_map[GS_FEATURE_AUTOMIXMON].f_right |= GS_FEATURE_ALLOW_CALLEE;
	}

	if (strchr(dial_options, 'X')) {
		gs_feature_map[GS_FEATURE_AUTOMIXMON].f_right |= GS_FEATURE_ALLOW_CALLER;
	}

	if (strchr(dial_options, 'k')) {
		gs_feature_map[GS_FEATURE_PARKCALL].f_right |= GS_FEATURE_ALLOW_CALLEE;
	}

	if (strchr(dial_options, 'K')) {
		gs_feature_map[GS_FEATURE_PARKCALL].f_right |= GS_FEATURE_ALLOW_CALLER;
	}

	return;
}

/*! \brief get feature code map from features.conf, and get feature right from extensions.conf (DIAL_OPTIONS) */
int get_feature_code(void)
{
	struct ast_flags config_flags = {0};
	struct ast_config    *pstCfg  = NULL;
	struct ast_config    *pstECfg = NULL;
	struct ast_variable  *pstVarGen = NULL;
	int    i;

	/* clear feature right */
	for (i = 0; i < GS_FEATURE_MAP_COUNT; i++) {
		gs_feature_map[i].f_right = GS_FEATURE_DISABLE;
	}

	/* load features.conf file*/
	pstCfg = ast_config_load("features.conf", config_flags);

	if (pstCfg == CONFIG_STATUS_FILEUNCHANGED || NULL == pstCfg) {
		return -1;
	}

	if (pstCfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Contents of features.conf are invalid and cannot be parsed\n");
		return -1;
	}

	pstVarGen = ast_variable_browse(pstCfg, "featuremap");

	for (; (NULL != pstVarGen); pstVarGen = pstVarGen->next) {
		for (i = 0; i < GS_FEATURE_MAP_COUNT; i++) {
			if (0 == strcasecmp(gs_feature_map[i].f_name, pstVarGen->name)) {
				ast_copy_string(gs_feature_map[i].f_code, pstVarGen->value,  sizeof(gs_feature_map[i].f_code));
				break;
			}
		}
	}

	ast_config_destroy(pstCfg);

	/* load extensions.conf file*/
	pstECfg = ast_config_load("extensions.conf", config_flags);

	if (pstECfg == CONFIG_STATUS_FILEUNCHANGED || NULL == pstECfg) {
		return -1;
	}

	if (pstECfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Contents of extensions.conf are invalid and cannot be parsed\n");
		return -1;
	}

	pstVarGen = ast_variable_browse(pstECfg, "globals");

	for (; (NULL != pstVarGen); pstVarGen = pstVarGen->next) {
		if (0 == strcasecmp(pstVarGen->name, "DIAL_OPTIONS")) {
			if (NULL == pstVarGen->value) {
				break;
			}

			set_feature_map(pstVarGen->value);
			break;
		}
	}

	ast_config_destroy(pstECfg);
	return 0;
}

/*! \brief Add "X-UCM-AudioRecord" header and "X-UCM-CallPark" header to outgoing message
 *
 * if the right is allow callee or allow both, add the header in Invite message;
   if the right is allow caller or allow both, add the header in 200 for Invite message
 *
 * \param req The request/response to which we will add the header
 */
void add_feature_map_header(pjsip_tx_data *tdata, int outgoing_call)
{
	char  audio_record_code[12] = {0};
	char  call_park_code[12]    = {0};

	if (NULL == tdata) {
		return;
	}

	/* check right */
	if (outgoing_call == 0) {
		/* copy code */
		if (0 != (gs_feature_map[GS_FEATURE_AUTOMIXMON].f_right & GS_FEATURE_ALLOW_CALLER)) {
			strncpy(audio_record_code, gs_feature_map[GS_FEATURE_AUTOMIXMON].f_code, sizeof(audio_record_code));
		}

		if (0 != (gs_feature_map[GS_FEATURE_PARKCALL].f_right & GS_FEATURE_ALLOW_CALLER)) {
			strncpy(call_park_code, gs_feature_map[GS_FEATURE_PARKCALL].f_code, sizeof(call_park_code));
		}
	}
	else {
		if (0 != (gs_feature_map[GS_FEATURE_AUTOMIXMON].f_right & GS_FEATURE_ALLOW_CALLEE)) {
			strncpy(audio_record_code, gs_feature_map[GS_FEATURE_AUTOMIXMON].f_code, sizeof(audio_record_code));
		}

		if (0 != (gs_feature_map[GS_FEATURE_PARKCALL].f_right & GS_FEATURE_ALLOW_CALLEE)) {
			strncpy(call_park_code, gs_feature_map[GS_FEATURE_PARKCALL].f_code, sizeof(call_park_code));
		}
	}

	/* add header */
	if (strlen(audio_record_code) > 0) {
		ast_sip_add_header(tdata, "X-UCM-AudioRecord", audio_record_code);
	}

	if (strlen(call_park_code) > 0) {
		ast_sip_add_header(tdata, "X-UCM-CallPark", call_park_code);
	}

	return;
}

static int reload_module(void)
{
#ifdef  GRANDSTREAM_NETWORKS
	globalEnabledCEI = ast_cei_get_enable_configuration();
#endif
	get_feature_code();

	return 0;
}

#endif

/*!
 * \brief Load the module
 *
 * Module loading including tests for configuration or dependencies.
 * This function can return AST_MODULE_LOAD_FAILURE, AST_MODULE_LOAD_DECLINE,
 * or AST_MODULE_LOAD_SUCCESS. If a dependency or environment variable fails
 * tests return AST_MODULE_LOAD_FAILURE. If the module can not load the
 * configuration file or other non-critical problem return
 * AST_MODULE_LOAD_DECLINE. On success return AST_MODULE_LOAD_SUCCESS.
 */
static int load_module(void)
{
	struct ao2_container *endpoints;

	CHECK_PJSIP_SESSION_MODULE_LOADED();

	if (!(chan_pjsip_tech.capabilities = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		return AST_MODULE_LOAD_DECLINE;
	}

	ast_format_cap_append_by_type(chan_pjsip_tech.capabilities, AST_MEDIA_TYPE_AUDIO);

	ast_rtp_glue_register(&chan_pjsip_rtp_glue);

	if (ast_channel_register(&chan_pjsip_tech)) {
		ast_log(LOG_ERROR, "Unable to register channel class %s\n", channel_type);
		goto end;
	}

	if (ast_custom_function_register(&chan_pjsip_dial_contacts_function)) {
		ast_log(LOG_ERROR, "Unable to register PJSIP_DIAL_CONTACTS dialplan function\n");
		goto end;
	}

	if (ast_custom_function_register(&media_offer_function)) {
		ast_log(LOG_WARNING, "Unable to register PJSIP_MEDIA_OFFER dialplan function\n");
		goto end;
	}

	if (ast_sip_session_register_supplement(&chan_pjsip_supplement)) {
		ast_log(LOG_ERROR, "Unable to register PJSIP supplement\n");
		goto end;
	}

	if (!(pjsip_uids_onhold = ao2_container_alloc_hash(AO2_ALLOC_OPT_LOCK_RWLOCK,
			AO2_CONTAINER_ALLOC_OPT_DUPS_REJECT, 37, uid_hold_hash_fn,
			uid_hold_sort_fn, NULL))) {
		ast_log(LOG_ERROR, "Unable to create held channels container\n");
		goto end;
	}

	if (ast_sip_session_register_supplement(&call_pickup_supplement)) {
		ast_log(LOG_ERROR, "Unable to register PJSIP call pickup supplement\n");
		ast_sip_session_unregister_supplement(&chan_pjsip_supplement);
		goto end;
	}

	if (ast_sip_session_register_supplement(&pbx_start_supplement)) {
		ast_log(LOG_ERROR, "Unable to register PJSIP pbx start supplement\n");
		ast_sip_session_unregister_supplement(&chan_pjsip_supplement);
		ast_sip_session_unregister_supplement(&call_pickup_supplement);
		goto end;
	}

	if (ast_sip_session_register_supplement(&chan_pjsip_ack_supplement)) {
		ast_log(LOG_ERROR, "Unable to register PJSIP ACK supplement\n");
		ast_sip_session_unregister_supplement(&pbx_start_supplement);
		ast_sip_session_unregister_supplement(&chan_pjsip_supplement);
		ast_sip_session_unregister_supplement(&call_pickup_supplement);
		goto end;
	}

#ifdef GRANDSTREAM_NETWORKS
    globalEnabledCEI = ast_cei_get_enable_configuration();
    struct stasis_topic *cei_toggle_enable_topic = NULL;

  cei_toggle_enable_topic = ast_cei_toggle_enable_topic_all();
  if (!cei_toggle_enable_topic) {
    ast_log(LOG_WARNING, "[WL] Failed to get cei_toggle_enable_topic\n");
    goto end;
  }

  cei_toggle_enable_router = stasis_message_router_create(cei_toggle_enable_topic);
  if (cei_toggle_enable_router == NULL) {
    ast_log(LOG_WARNING, "[WL] Failed to create cei_toggle_enable_router\n");
    goto end;
  }

  if (stasis_message_router_add(cei_toggle_enable_router,
                  ast_cei_toggle_enable_type(),
                  cei_toggle_enable_cb, NULL))
  {
    ast_log(LOG_WARNING, "Failed to subscribe to ast_cei_toggle_enable_type\n");
    goto end;
  }

    if (ast_register_application_xml(app_sipsetcalltype, sip_setcalltype)) {
        ast_log(LOG_ERROR, "Unable to register SIPSetCallType dialplan function\n");
        goto end;
    }

	if (ast_custom_function_register(&pjsip_mcb_resp_function)) {
		ast_log(LOG_ERROR, "Unable to register MCB_RESPONSE dialplan function\n");
		goto end;
	}

	if (ast_register_application_xml(app_sipsetmcbresp, sip_setmcbresp)) {
		ast_log(LOG_ERROR, "Unable to register SIPSetMCBResp dialplan function\n");
		goto end;
	}

    if (ast_register_application_xml(app_sipsetconffocus, sip_set_conf_focus)) {
		ast_log(LOG_ERROR, "Unable to register SIPSetConfFocus dialplan function\n");
		goto end;
	}

    if (ast_register_application_xml(app_sipsetcontactuser, sip_set_contact_user)) {
		ast_log(LOG_ERROR, "Unable to register SIPSetContactUser dialplan function\n");
		goto end;
	}

    if (ast_register_application_xml(app_sipterminatednotify, sip_terminated_notify)) {
		ast_log(LOG_ERROR, "Unable to register SIPTerminatedNotify dialplan function\n");
		goto end;
	}

    if (ast_register_application_xml(app_sipceijoined, sip_cei_joined)) {
		ast_log(LOG_ERROR, "Unable to register SIPCEIJoined dialplan function\n");
		goto end;
	}

    if (ast_register_application_xml(app_sipsubscribeextmcb, sip_external_mcb_sub)) {
		ast_log(LOG_ERROR, "Unable to register SIPSubscribeExtMCB dialplan function\n");
		goto end;
	}

    if (ast_register_application_xml(app_sipunsubscribeextmcb, sip_external_mcb_unsub)) {
		ast_log(LOG_ERROR, "Unable to register SIPUnsubscribeExtMCB dialplan function\n");
		goto end;
	}

    if (ast_register_application_xml(app_sipsubscribemcb, sip_internal_mcb_sub)) {
    ast_log(LOG_ERROR, "Unable to register SIPSubscribeMCB dialplan function\n");
    goto end;
  }

    if (ast_register_application_xml(app_sipunsubscribemcb, sip_internal_mcb_unsub)) {
    ast_log(LOG_ERROR, "Unable to register SIPUnsubscribeMCB dialplan function\n");
    goto end;
  }

    get_feature_code();
#endif

	/* since endpoints are loaded before the channel driver their device
	   states get set to 'invalid', so they need to be updated */
	if ((endpoints = ast_sip_get_endpoints())) {
		ao2_callback(endpoints, OBJ_NODATA, update_devstate, NULL);
		ao2_ref(endpoints, -1);
	}

	return 0;

end:
	ao2_cleanup(pjsip_uids_onhold);
	pjsip_uids_onhold = NULL;
	ast_custom_function_unregister(&media_offer_function);
	ast_custom_function_unregister(&chan_pjsip_dial_contacts_function);
	ast_channel_unregister(&chan_pjsip_tech);
	ast_rtp_glue_unregister(&chan_pjsip_rtp_glue);
#ifdef GRANDSTREAM_NETWORKS
	ast_unregister_application(app_sipsetcalltype);
	ast_custom_function_unregister(&pjsip_mcb_resp_function);
	ast_unregister_application(app_sipsetmcbresp);
    ast_unregister_application(app_sipsetconffocus);
    ast_unregister_application(app_sipsetcontactuser);
    ast_unregister_application(app_sipterminatednotify);
    ast_unregister_application(app_sipceijoined);
    ast_unregister_application(app_sipsubscribeextmcb);
    ast_unregister_application(app_sipunsubscribeextmcb);
    ast_unregister_application(app_sipsubscribemcb);
    ast_unregister_application(app_sipunsubscribemcb);

    if (cei_toggle_enable_router) {
  		stasis_message_router_unsubscribe(cei_toggle_enable_router);
  		cei_toggle_enable_router = NULL;
  	}
#endif

	return AST_MODULE_LOAD_FAILURE;
}

/*! \brief Unload the PJSIP channel from Asterisk */
static int unload_module(void)
{
	ao2_cleanup(pjsip_uids_onhold);
	pjsip_uids_onhold = NULL;

	ast_sip_session_unregister_supplement(&chan_pjsip_supplement);
	ast_sip_session_unregister_supplement(&pbx_start_supplement);
	ast_sip_session_unregister_supplement(&chan_pjsip_ack_supplement);
	ast_sip_session_unregister_supplement(&call_pickup_supplement);

	ast_custom_function_unregister(&media_offer_function);
	ast_custom_function_unregister(&chan_pjsip_dial_contacts_function);

	ast_channel_unregister(&chan_pjsip_tech);
	ao2_ref(chan_pjsip_tech.capabilities, -1);
	ast_rtp_glue_unregister(&chan_pjsip_rtp_glue);

#ifdef GRANDSTREAM_NETWORKS
    ast_unregister_application(app_sipsetcalltype);
	ast_custom_function_unregister(&pjsip_mcb_resp_function);
	ast_unregister_application(app_sipsetmcbresp);
    ast_unregister_application(app_sipsetconffocus);
    ast_unregister_application(app_sipsetcontactuser);
    ast_unregister_application(app_sipterminatednotify);
    ast_unregister_application(app_sipceijoined);
    ast_unregister_application(app_sipsubscribeextmcb);
    ast_unregister_application(app_sipunsubscribeextmcb);
    ast_unregister_application(app_sipsubscribemcb);
    ast_unregister_application(app_sipunsubscribemcb);

    if (cei_toggle_enable_router) {
  		stasis_message_router_unsubscribe(cei_toggle_enable_router);
  		cei_toggle_enable_router = NULL;
  	}
#endif

	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP Channel Driver",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
#ifdef GRANDSTREAM_NETWORKS
		.reload = reload_module,
#endif
		.load_pri = AST_MODPRI_CHANNEL_DRIVER,
	       );
