/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2009, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Analog signaling module
 *
 * \author Matthew Fredrickson <creslin@digium.com>
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#ifdef GRANDSTREAM_NETWORKS
#define AST_MODULE_LOG "chan_dahdi"
#endif
#include "asterisk.h"

#ifdef GRANDSTREAM_NETWORKS
ASTERISK_FILE_VERSION(__FILE__, "$Revision: 13.98 $")
#endif
#include <errno.h>
#include <ctype.h>

#include "asterisk/utils.h"
#include "asterisk/options.h"
#include "asterisk/pickup.h"
#include "asterisk/pbx.h"
#include "asterisk/file.h"
#include "asterisk/callerid.h"
#include "asterisk/say.h"
#include "asterisk/manager.h"
#include "asterisk/astdb.h"
#include "asterisk/features.h"
#include "asterisk/causes.h"
#include "asterisk/features_config.h"
#include "asterisk/bridge.h"
#include "asterisk/parking.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/cdr.h"

#include <dahdi/user.h>
#endif
#include "sig_analog.h"
#include "chan_dahdi.h"

/*** DOCUMENTATION
 ***/

/*! \note
 * Define if you want to check the hook state for an FXO (FXS signalled) interface
 * before dialing on it.  Certain FXO interfaces always think they're out of
 * service with this method however.
 */
/* #define DAHDI_CHECK_HOOKSTATE */

#define POLARITY_IDLE 0
#define POLARITY_REV    1
#define MIN_MS_SINCE_FLASH			( (2000) )	/*!< 2000 ms */
static int analog_matchdigittimeout = 3000;
static int analog_gendigittimeout = 8000;
static int analog_firstdigittimeout = 16000;
static char analog_defaultcic[64] = "";
static char analog_defaultozz[64] = "";

static const struct {
	enum analog_sigtype sigtype;
	const char const *name;
} sigtypes[] = {
	{ ANALOG_SIG_FXOLS, "fxo_ls" },
	{ ANALOG_SIG_FXOKS, "fxo_ks" },
	{ ANALOG_SIG_FXOGS, "fxo_gs" },
	{ ANALOG_SIG_FXSLS, "fxs_ls" },
	{ ANALOG_SIG_FXSKS, "fxs_ks" },
	{ ANALOG_SIG_FXSGS, "fxs_gs" },
	{ ANALOG_SIG_EMWINK, "em_w" },
	{ ANALOG_SIG_EM, "em" },
	{ ANALOG_SIG_EM_E1, "em_e1" },
	{ ANALOG_SIG_FEATD, "featd" },
	{ ANALOG_SIG_FEATDMF, "featdmf" },
	{ ANALOG_SIG_FEATDMF_TA, "featdmf_ta" },
	{ ANALOG_SIG_FEATB, "featb" },
	{ ANALOG_SIG_FGC_CAMA, "fgccama" },
	{ ANALOG_SIG_FGC_CAMAMF, "fgccamamf" },
	{ ANALOG_SIG_SF, "sf" },
	{ ANALOG_SIG_SFWINK, "sf_w" },
	{ ANALOG_SIG_SF_FEATD, "sf_featd" },
	{ ANALOG_SIG_SF_FEATDMF, "sf_featdmf" },
	{ ANALOG_SIG_SF_FEATB, "sf_featb" },
	{ ANALOG_SIG_E911, "e911" },
};

static const struct {
	unsigned int cid_type;
	const char const *name;
} cidtypes[] = {
	{ CID_SIG_BELL,   "bell" },
	{ CID_SIG_V23,    "v23" },
	{ CID_SIG_V23_JP, "v23_jp" },
	{ CID_SIG_DTMF,   "dtmf" },
	/* "smdi" is intentionally not supported here, as there is a much better
	 * way to do this in the dialplan now. */
};

#define ISTRUNK(p) ((p->sig == ANALOG_SIG_FXSLS) || (p->sig == ANALOG_SIG_FXSKS) || \
					(p->sig == ANALOG_SIG_FXSGS))

#ifdef GRANDSTREAM_NETWORKS
#if defined(SVIP_DETECT_DTMF)
extern char g_svip_dtmf_digit[DAHDI_MAX_EVENTSIZE][DAHDI_MAX_DTMF_BUF];
#endif
#ifdef CURRENT_DISCONNECT
static int analog_proc_battlost_timeout(void *data);
#endif
static int analog_proc_dial_delay_timeout(void *data);
static int analog_proc_em_dial_delay_timeout(void *data);
static int analog_proc_FXO_Hangup_Delay_timeout(void *data);
#endif

enum analog_sigtype analog_str_to_sigtype(const char *name)
{
	int i;

	for (i = 0; i < ARRAY_LEN(sigtypes); i++) {
		if (!strcasecmp(sigtypes[i].name, name)) {
			return sigtypes[i].sigtype;
		}
	}

	return 0;
}

const char *analog_sigtype_to_str(enum analog_sigtype sigtype)
{
	int i;

	for (i = 0; i < ARRAY_LEN(sigtypes); i++) {
		if (sigtype == sigtypes[i].sigtype) {
			return sigtypes[i].name;
		}
	}

	return "Unknown";
}

unsigned int analog_str_to_cidtype(const char *name)
{
	int i;

	for (i = 0; i < ARRAY_LEN(cidtypes); i++) {
		if (!strcasecmp(cidtypes[i].name, name)) {
			return cidtypes[i].cid_type;
		}
	}

	return 0;
}

const char *analog_cidtype_to_str(unsigned int cid_type)
{
	int i;

	for (i = 0; i < ARRAY_LEN(cidtypes); i++) {
		if (cid_type == cidtypes[i].cid_type) {
			return cidtypes[i].name;
		}
	}

	return "Unknown";
}

static int analog_start_cid_detect(struct analog_pvt *p, int cid_signalling)
{
	if (analog_callbacks.start_cid_detect) {
		return analog_callbacks.start_cid_detect(p->chan_pvt, cid_signalling);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_start_cid_detect(%p) start_cid_detect is NULL.", p->channel, p);
#endif
	return -1;
}

static int analog_stop_cid_detect(struct analog_pvt *p)
{
	if (analog_callbacks.stop_cid_detect) {
		return analog_callbacks.stop_cid_detect(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_NOTICE, "[ch:%d]analog_stop_cid_detect(%p) start_cid_detect is NULL.", p->channel, p);
#endif
	return -1;
}

#ifdef GRANDSTREAM_NETWORKS
static int analog_start_cid_detect_auto(struct analog_pvt *p, int cid_signalling)
{
	if (analog_callbacks.start_cid_detect_auto) {
		return analog_callbacks.start_cid_detect_auto(p->chan_pvt, cid_signalling);
	}

    ast_log(LOG_NOTICE, "[ch:%d]analog_start_cid_detect_auto(%p) start_cid_detect_auto is NULL.", p->channel, p);
	return -1;
}

static int analog_stop_cid_detect_auto(struct analog_pvt *p)
{
	if (analog_callbacks.stop_cid_detect_auto) {
		return analog_callbacks.stop_cid_detect_auto(p->chan_pvt);
	}

    ast_log(LOG_NOTICE, "[ch:%d]analog_stop_cid_detect_auto(%p) stop_cid_detect_auto is NULL.", p->channel, p);
	return -1;
}

static int analog_get_ring_event(struct analog_pvt *p, enum analog_event *ev, size_t timeout)
{
	if (analog_callbacks.get_ring_event) {
		return analog_callbacks.get_ring_event(p->chan_pvt, ev, timeout);
	} else {
		ast_log(LOG_WARNING, "analog_callbacks.get_ring_event is NULL!");
	}

	return -1;
}
#endif

static int analog_get_callerid(struct analog_pvt *p, char *name, char *number, enum analog_event *ev, size_t timeout)
{
	if (analog_callbacks.get_callerid) {
		return analog_callbacks.get_callerid(p->chan_pvt, name, number, ev, timeout);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_get_callerid(%p) get_callerid is NULL.", p->channel, p);
#endif
	return -1;
}
#ifdef GRANDSTREAM_NETWORKS
static int analog_get_callerid_auto(struct analog_pvt *p, char *name, char *number, void *fsk_buf, int datalen, int *dwwhodetectcallerID)
{
	if (analog_callbacks.get_callerid_auto) {
		return analog_callbacks.get_callerid_auto(p->chan_pvt, name, number, fsk_buf, datalen, dwwhodetectcallerID);
	}
	else
	{
        ast_log(LOG_WARNING, "get_callerid_auto failed, analog_get_callerid_auto is NULL!");
	}
	return -1;
}
#endif
static const char *analog_get_orig_dialstring(struct analog_pvt *p)
{
	if (analog_callbacks.get_orig_dialstring) {
		return analog_callbacks.get_orig_dialstring(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_get_orig_dialstring(%p) get_orig_dialstring is NULL.", p->channel, p);
#endif
	return "";
}

static int analog_get_event(struct analog_pvt *p)
{
	if (analog_callbacks.get_event) {
		return analog_callbacks.get_event(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_get_event(%p) get_event is NULL.", p->channel, p);
#endif
	return -1;
}

static int analog_wait_event(struct analog_pvt *p)
{
	if (analog_callbacks.wait_event) {
		return analog_callbacks.wait_event(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_wait_event(%p) wait_event is NULL.", p->channel, p);
#endif
	return -1;
}

static int analog_have_progressdetect(struct analog_pvt *p)
{
	if (analog_callbacks.have_progressdetect) {
		return analog_callbacks.have_progressdetect(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_NOTICE, "[ch:%d]analog_have_progressdetect(%p) have_progressdetect is NULL.", p->channel, p);
#endif
	/* Don't have progress detection. */
	return 0;
}
#ifdef GRANDSTREAM_NETWORKS
#if defined(SVIP_T38_STACK)
static void * analog_get_t38_capabilities(struct analog_pvt *p)
{
	if (analog_callbacks.get_t38_capabilities) {
		return analog_callbacks.get_t38_capabilities(p->chan_pvt);
	}
    ast_log(LOG_NOTICE, "[ch:%d]analog_get_t38_capabilities(%p) get_t38_capabilities is NULL.", p->channel, p);
	return NULL;
}

static int analog_stop_t38_session(struct analog_pvt *p)
{
	if (analog_callbacks.stop_t38_session) {
		return analog_callbacks.stop_t38_session(p->chan_pvt);
	}
    ast_log(LOG_NOTICE, "[ch:%d]analog_stop_t38_session(%p) stop_t38_session is NULL.", p->channel, p);
	return -1;
}
static int analog_get_t38_state(struct analog_pvt *p)
{
	if (analog_callbacks.get_t38_state) {
		return analog_callbacks.get_t38_state(p->chan_pvt);
	}
    ast_log(LOG_NOTICE, "[ch:%d]analog_get_t38_state(%p) get_t38_state is NULL.", p->channel, p);
	return -1;
}
static int analog_set_t38_state(struct analog_pvt *p, int t38_state)
{
	if (analog_callbacks.set_t38_state) {
		return analog_callbacks.set_t38_state(p->chan_pvt, t38_state);
	}
    ast_log(LOG_NOTICE, "[ch:%d]analog_set_t38_state(%p, %d) get_t38_state is NULL.", p->channel, p, t38_state);
	return -1;
}

static void analog_unlock_t38_session(struct analog_pvt *p)
{
	if (analog_callbacks.unlock_t38_session) {
		analog_callbacks.unlock_t38_session(p->chan_pvt);
		return;
	}
	ast_log(LOG_NOTICE, "[ch:%d]analog_unlock_t38_session(%p) calls->unlock_t38_session is NULL.", p->channel, p);
}

static void analog_lock_t38_session(struct analog_pvt *p)
{
	if (analog_callbacks.lock_t38_session) {
		analog_callbacks.lock_t38_session(p->chan_pvt);
		return;
	}
	ast_log(LOG_NOTICE, "[ch:%d]analog_unlock_t38_session(%p) calls->lock_t38_session is NULL.", p->channel, p);
}
#endif
#endif
enum analog_cid_start analog_str_to_cidstart(const char *value)
{
	if (!strcasecmp(value, "ring")) {
		return ANALOG_CID_START_RING;
	} else if (!strcasecmp(value, "polarity")) {
		return ANALOG_CID_START_POLARITY;
	} else if (!strcasecmp(value, "polarity_in")) {
		return ANALOG_CID_START_POLARITY_IN;
	} else if (!strcasecmp(value, "dtmf")) {
		return ANALOG_CID_START_DTMF_NOALERT;
	}

	return 0;
}

const char *analog_cidstart_to_str(enum analog_cid_start cid_start)
{
	switch (cid_start) {
	case ANALOG_CID_START_RING:
		return "Ring";
	case ANALOG_CID_START_POLARITY:
		return "Polarity";
	case ANALOG_CID_START_POLARITY_IN:
		return "Polarity_In";
	case ANALOG_CID_START_DTMF_NOALERT:
		return "DTMF";
#ifdef GRANDSTREAM_NETWORKS
	case ANALOG_CID_START_AUTO_DETECT:
		return "AUTO_DETECT";
#endif
	}

	return "Unknown";
}

static char *analog_event2str(enum analog_event event)
{
	char *res;
	switch (event) {
	case ANALOG_EVENT_ONHOOK:
		res = "ANALOG_EVENT_ONHOOK";
		break;
	case ANALOG_EVENT_RINGOFFHOOK:
		res = "ANALOG_EVENT_RINGOFFHOOK";
		break;
	case ANALOG_EVENT_WINKFLASH:
		res = "ANALOG_EVENT_WINKFLASH";
		break;
	case ANALOG_EVENT_ALARM:
		res = "ANALOG_EVENT_ALARM";
		break;
	case ANALOG_EVENT_NOALARM:
		res = "ANALOG_EVENT_NOALARM";
		break;
	case ANALOG_EVENT_DIALCOMPLETE:
		res = "ANALOG_EVENT_DIALCOMPLETE";
		break;
	case ANALOG_EVENT_HOOKCOMPLETE:
		res = "ANALOG_EVENT_HOOKCOMPLETE";
		break;
	case ANALOG_EVENT_PULSE_START:
		res = "ANALOG_EVENT_PULSE_START";
		break;
	case ANALOG_EVENT_POLARITY:
		res = "ANALOG_EVENT_POLARITY";
		break;
	case ANALOG_EVENT_RINGBEGIN:
		res = "ANALOG_EVENT_RINGBEGIN";
		break;
	case ANALOG_EVENT_EC_DISABLED:
		res = "ANALOG_EVENT_EC_DISABLED";
		break;
	case ANALOG_EVENT_RINGERON:
		res = "ANALOG_EVENT_RINGERON";
		break;
	case ANALOG_EVENT_RINGEROFF:
		res = "ANALOG_EVENT_RINGEROFF";
		break;
	case ANALOG_EVENT_REMOVED:
		res = "ANALOG_EVENT_REMOVED";
		break;
	case ANALOG_EVENT_NEONMWI_ACTIVE:
		res = "ANALOG_EVENT_NEONMWI_ACTIVE";
		break;
	case ANALOG_EVENT_NEONMWI_INACTIVE:
		res = "ANALOG_EVENT_NEONMWI_INACTIVE";
		break;
#ifdef HAVE_DAHDI_ECHOCANCEL_FAX_MODE
	case ANALOG_EVENT_TX_CED_DETECTED:
		res = "ANALOG_EVENT_TX_CED_DETECTED";
		break;
	case ANALOG_EVENT_RX_CED_DETECTED:
		res = "ANALOG_EVENT_RX_CED_DETECTED";
		break;
	case ANALOG_EVENT_EC_NLP_DISABLED:
		res = "ANALOG_EVENT_EC_NLP_DISABLED";
		break;
	case ANALOG_EVENT_EC_NLP_ENABLED:
		res = "ANALOG_EVENT_EC_NLP_ENABLED";
		break;
#endif
#ifdef GRANDSTREAM_NETWORKS
    case ANALOG_EVENT_BATTERY_PRESENT:
		res = "ANALOG_EVENT_BATTERY_PRESENT";
		break;
    case ANALOG_EVENT_BATTERY_LOST:
		res = "ANALOG_EVENT_BATTERY_LOST";
		break;
    case ANALOG_EVENT_ERROR: 
        res = "ANALOG_EVENT_ERROR";
        break;
    case ANALOG_EVENT_DTMFCID:
        res = "ANALOG_EVENT_DTMFCID";
        break;
#endif
	case ANALOG_EVENT_PULSEDIGIT:
		res = "ANALOG_EVENT_PULSEDIGIT";
		break;
	case ANALOG_EVENT_DTMFDOWN:
		res = "ANALOG_EVENT_DTMFDOWN";
		break;
	case ANALOG_EVENT_DTMFUP:
		res = "ANALOG_EVENT_DTMFUP";
		break;
	default:
		res = "UNKNOWN/OTHER";
		break;
	}

	return res;
}

static void analog_swap_subs(struct analog_pvt *p, enum analog_sub a, enum analog_sub b)
{
	int tinthreeway;
	struct ast_channel *towner;

#ifdef GRANDSTREAM_NETWORKS
	ast_debug(1, "[ch:%d]Swapping %d and %d.", p->channel, a, b);
#else
	ast_debug(1, "Swapping %u and %u\n", a, b);
#endif

	towner = p->subs[a].owner;
	p->subs[a].owner = p->subs[b].owner;
	p->subs[b].owner = towner;

	tinthreeway = p->subs[a].inthreeway;
	p->subs[a].inthreeway = p->subs[b].inthreeway;
	p->subs[b].inthreeway = tinthreeway;

	if (analog_callbacks.swap_subs) {
		analog_callbacks.swap_subs(p->chan_pvt, a, p->subs[a].owner, b, p->subs[b].owner);
	}
}

static int analog_alloc_sub(struct analog_pvt *p, enum analog_sub x)
{
	if (analog_callbacks.allocate_sub) {
		int res;
		res = analog_callbacks.allocate_sub(p->chan_pvt, x);
		if (!res) {
			p->subs[x].allocd = 1;
		}
		return res;
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_alloc_sub(%p) allocate_sub is NULL.", p->channel, p);
#endif
	return 0;
}

static int analog_unalloc_sub(struct analog_pvt *p, enum analog_sub x)
{
	p->subs[x].allocd = 0;
	p->subs[x].owner = NULL;
	if (analog_callbacks.unallocate_sub) {
		return analog_callbacks.unallocate_sub(p->chan_pvt, x);
	}
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_NOTICE, "[ch:%d]analog_unalloc_sub(%p) unallocate_sub is NULL.", p->channel, p);
#endif
	return 0;
}

static int analog_send_callerid(struct analog_pvt *p, int cwcid, struct ast_party_caller *caller)
{
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "[ch:%d]analog_send_callerid(%p, %d, %p) CID_NAME: '%s' CID_NUM: '%s' ...",
		p->channel, p, cwcid, caller,
		caller->id.name.str,
		caller->id.number.str);
#else
	ast_debug(1, "Sending callerid.  CID_NAME: '%s' CID_NUM: '%s'\n",
		caller->id.name.str,
		caller->id.number.str);
#endif

	if (cwcid) {
		p->callwaitcas = 0;
	}

	if (analog_callbacks.send_callerid) {
		return analog_callbacks.send_callerid(p->chan_pvt, cwcid, caller);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_send_callerid(%p) send_callerid is NULL.", p->channel, p);
#endif
	return 0;
}

#define analog_get_index(ast, p, nullok)	_analog_get_index(ast, p, nullok, __PRETTY_FUNCTION__, __LINE__)
static int _analog_get_index(struct ast_channel *ast, struct analog_pvt *p, int nullok, const char *fname, unsigned long line)
{
	int res;
	if (p->subs[ANALOG_SUB_REAL].owner == ast) {
		res = ANALOG_SUB_REAL;
	} else if (p->subs[ANALOG_SUB_CALLWAIT].owner == ast) {
		res = ANALOG_SUB_CALLWAIT;
	} else if (p->subs[ANALOG_SUB_THREEWAY].owner == ast) {
		res = ANALOG_SUB_THREEWAY;
	} else {
		res = -1;
		if (!nullok) {
			ast_log(LOG_WARNING,
				"Unable to get index for '%s' on channel %d (%s(), line %lu)\n",
				ast ? ast_channel_name(ast) : "", p->channel, fname, line);
		}
	}
	return res;
}

static int analog_dsp_reset_and_flush_digits(struct analog_pvt *p)
{
	if (analog_callbacks.dsp_reset_and_flush_digits) {
		return analog_callbacks.dsp_reset_and_flush_digits(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_dsp_reset_and_flush_digits(%p) dsp_reset_and_flush_digits is NULL.", p->channel, p);
#endif
	/* Return 0 since I think this is unnecessary to do in most cases it is used.  Mostly only for ast_dsp */
	return 0;
}

static int analog_play_tone(struct analog_pvt *p, enum analog_sub sub, enum analog_tone tone)
{
	if (analog_callbacks.play_tone) {
		return analog_callbacks.play_tone(p->chan_pvt, sub, tone);
	}
#ifdef GRANDSTREAM_NETWORKS
    ast_log(LOG_WARNING, "[ch:%d]analog_play_tone(%p, %u, %u) play_tone is NULL.",
        p->channel, p, sub, tone);
#endif
	return -1;
}

static void analog_set_new_owner(struct analog_pvt *p, struct ast_channel *new_owner)
{
	p->owner = new_owner;
	if (analog_callbacks.set_new_owner) {
		analog_callbacks.set_new_owner(p->chan_pvt, new_owner);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_set_new_owner(%p) set_new_owner is NULL.", p->channel, p);
	}
#endif
}

static struct ast_channel * analog_new_ast_channel(struct analog_pvt *p, int state, int startpbx, enum analog_sub sub, const struct ast_channel *requestor)
{
	struct ast_channel *c;
#ifdef GRANDSTREAM_NETWORKS
	if (NULL == p)
	{
		ast_log(LOG_ERROR, "%s(0x%08tx, %d, %d, %u, 0x%08tx) parameter is NULL!!", 
			__FUNCTION__, (ptrdiff_t)p, state, startpbx, sub, (ptrdiff_t)requestor);
		return NULL;
	}
	ast_log(LOG_NOTICE, "[ch:%d]analog_new_ast_channel(0x%08tx, %d, %d, %u, 0x%08tx) ...", 
		p->channel, (ptrdiff_t)p, state, startpbx, sub, (ptrdiff_t)requestor);
#endif

	if (!analog_callbacks.new_ast_channel) 
	{
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_ERROR, "%s(0x%08tx, %d, %d, %u, 0x%08tx) new_ast_channel is NULL!!", 
			__FUNCTION__, (ptrdiff_t)p, state, startpbx, sub, (ptrdiff_t)requestor);
#endif
		return NULL;
	}

	c = analog_callbacks.new_ast_channel(p->chan_pvt, state, startpbx, sub, requestor);
	if (c) {
		ast_channel_call_forward_set(c, p->call_forward);
#ifdef GRANDSTREAM_NETWORKS
		/*! Begin: Fax processing flow optimization*/
		struct dahdi_pvt *pDahdiPvt = (struct dahdi_pvt *)p->chan_pvt;
		if (pDahdiPvt->wFaxGateway && p->wIsOutgoing) {
			ast_channel_set_soft_fax_gateway(c, 1);
			ast_log(LOG_VERBOSE, "Create channel[%s] is success and 61XX FXS configure soft gateway!", ast_channel_name(c));
		} else {
			ast_channel_set_soft_fax_gateway(c, 0);
		}
		/*! End: Fax processing flow optimization*/
#endif
	}
	p->subs[sub].owner = c;
	if (!p->owner) {
		analog_set_new_owner(p, c);
	}
	return c;
}

static int analog_set_echocanceller(struct analog_pvt *p, int enable)
{
	if (analog_callbacks.set_echocanceller) {
		return analog_callbacks.set_echocanceller(p->chan_pvt, enable);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_set_echocanceller(%p) set_echocanceller is NULL.", p->channel, p);
#endif
	return -1;
}

static int analog_train_echocanceller(struct analog_pvt *p)
{
	if (analog_callbacks.train_echocanceller) {
		return analog_callbacks.train_echocanceller(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_train_echocanceller(%p) train_echocanceller is NULL.", p->channel, p);
#endif
	return -1;
}

static int analog_is_off_hook(struct analog_pvt *p)
{
	if (analog_callbacks.is_off_hook) {
		return analog_callbacks.is_off_hook(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_is_off_hook(%p) is_off_hook is NULL.", p->channel, p);
#endif
	return -1;
}

static int analog_ring(struct analog_pvt *p)
{
	if (analog_callbacks.ring) {
		return analog_callbacks.ring(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_ring(%p) ring is NULL.", p->channel, p);
#endif
	return -1;
}

static int analog_flash(struct analog_pvt *p)
{
	if (analog_callbacks.flash) {
		return analog_callbacks.flash(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_flash(%p) flash is NULL.", p->channel, p);
#endif
	return -1;
}

static int analog_start(struct analog_pvt *p)
{
	if (analog_callbacks.start) {
		return analog_callbacks.start(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_start(%p) start is NULL.", p->channel, p);
#endif
	return -1;
}

static int analog_dial_digits(struct analog_pvt *p, enum analog_sub sub, struct analog_dialoperation *dop)
{
	if (analog_callbacks.dial_digits) {
		return analog_callbacks.dial_digits(p->chan_pvt, sub, dop);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_dial_digits(%p) dial_digits is NULL.", p->channel, p);
#endif
	return -1;
}

static int analog_on_hook(struct analog_pvt *p)
{
	if (analog_callbacks.on_hook) {
		return analog_callbacks.on_hook(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_on_hook(%p) on_hook is NULL.", p->channel, p);
#endif
	return -1;
}

static void analog_set_outgoing(struct analog_pvt *p, int is_outgoing)
{
	p->outgoing = is_outgoing;
	if (analog_callbacks.set_outgoing) {
		analog_callbacks.set_outgoing(p->chan_pvt, is_outgoing);
	}
}

static int analog_check_for_conference(struct analog_pvt *p)
{
	if (analog_callbacks.check_for_conference) {
		return analog_callbacks.check_for_conference(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_check_for_conference(%p) check_for_conference is NULL.", p->channel, p);
#endif
	return -1;
}

static void analog_all_subchannels_hungup(struct analog_pvt *p)
{
	if (analog_callbacks.all_subchannels_hungup) {
		analog_callbacks.all_subchannels_hungup(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_check_for_conference(%p) check_for_conference is NULL.", p->channel, p);
	}
#endif
}

static void analog_unlock_private(struct analog_pvt *p)
{
	if (analog_callbacks.unlock_private) {
		analog_callbacks.unlock_private(p->chan_pvt);
	}
}

static void analog_lock_private(struct analog_pvt *p)
{
	if (analog_callbacks.lock_private) {
		analog_callbacks.lock_private(p->chan_pvt);
	}
}

static void analog_deadlock_avoidance_private(struct analog_pvt *p)
{
	if (analog_callbacks.deadlock_avoidance_private) {
		analog_callbacks.deadlock_avoidance_private(p->chan_pvt);
	} else {
		/* Fallback to manual avoidance if callback not present. */
		analog_unlock_private(p);
		usleep(1);
		analog_lock_private(p);
	}
}

/*!
 * \internal
 * \brief Obtain the specified subchannel owner lock if the owner exists.
 *
 * \param pvt Analog private struct.
 * \param sub_idx Subchannel owner to lock.
 *
 * \note Assumes the analog_lock_private(pvt->chan_pvt) is already obtained.
 *
 * \note
 * Because deadlock avoidance may have been necessary, you need to confirm
 * the state of things before continuing.
 *
 * \return Nothing
 */
static void analog_lock_sub_owner(struct analog_pvt *pvt, enum analog_sub sub_idx)
{
	for (;;) {
		if (!pvt->subs[sub_idx].owner) {
			/* No subchannel owner pointer */
			break;
		}
		if (!ast_channel_trylock(pvt->subs[sub_idx].owner)) {
			/* Got subchannel owner lock */
			break;
		}
		/* We must unlock the private to avoid the possibility of a deadlock */
		analog_deadlock_avoidance_private(pvt);
	}
}

static int analog_off_hook(struct analog_pvt *p)
{
	if (analog_callbacks.off_hook) {
		return analog_callbacks.off_hook(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_off_hook(%p) off_hook is NULL.", p->channel, p);
#endif
	return -1;
}

static void analog_set_needringing(struct analog_pvt *p, int value)
{
	if (analog_callbacks.set_needringing) {
		analog_callbacks.set_needringing(p->chan_pvt, value);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_set_needringing(%p,%d) set_needringing is NULL.", p->channel, p, value);
	}
#endif
}

#if 0
static void analog_set_polarity(struct analog_pvt *p, int value)
{
	if (analog_callbacks.set_polarity) {
		analog_callbacks.set_polarity(p->chan_pvt, value);
	}
}
#endif

static void analog_start_polarityswitch(struct analog_pvt *p)
{
	if (analog_callbacks.start_polarityswitch) {
		analog_callbacks.start_polarityswitch(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_start_polarityswitch(%p) start_polarityswitch is NULL.", p->channel, p);
	}
#endif
}
static void analog_answer_polarityswitch(struct analog_pvt *p)
{
	if (analog_callbacks.answer_polarityswitch) {
		analog_callbacks.answer_polarityswitch(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_answer_polarityswitch(%p) answer_polarityswitch is NULL.", p->channel, p);
	}
#endif
}

static void analog_hangup_polarityswitch(struct analog_pvt *p)
{
	if (analog_callbacks.hangup_polarityswitch) {
		analog_callbacks.hangup_polarityswitch(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_hangup_polarityswitch(%p) hangup_polarityswitch is NULL.", p->channel, p);
	}
#endif
}

static int analog_dsp_set_digitmode(struct analog_pvt *p, enum analog_dsp_digitmode mode)
{
	if (analog_callbacks.dsp_set_digitmode) {
		return analog_callbacks.dsp_set_digitmode(p->chan_pvt, mode);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_dsp_set_digitmode(%p,%d) dsp_set_digitmode is NULL.", p->channel, p, mode);
#endif
	return -1;
}

static void analog_cb_handle_dtmf(struct analog_pvt *p, struct ast_channel *ast, enum analog_sub analog_index, struct ast_frame **dest)
{
	if (analog_callbacks.handle_dtmf) {
		analog_callbacks.handle_dtmf(p->chan_pvt, ast, analog_index, dest);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
        ast_log(LOG_NOTICE, "[ch:%d]analog_cb_handle_dtmf(%p, %p, %u, %p) handle_dtmf is NULL.", p->channel, p, ast, analog_index, dest);
	}
#endif
}

static int analog_wink(struct analog_pvt *p, enum analog_sub index)
{
	if (analog_callbacks.wink) {
		return analog_callbacks.wink(p->chan_pvt, index);
	}
#ifdef GRANDSTREAM_NETWORKS
    ast_log(LOG_NOTICE, "[ch:%d]analog_wink(%p, %u) wink is NULL.", p->channel, p, index);
#endif
	return -1;
}

static int analog_has_voicemail(struct analog_pvt *p)
{
	if (analog_callbacks.has_voicemail) {
		return analog_callbacks.has_voicemail(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_has_voicemail(%p) has_voicemail is NULL.", p->channel, p);
#endif
	return -1;
}

static int analog_is_dialing(struct analog_pvt *p, enum analog_sub index)
{
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "[ch:%d]analog_is_dialing(%p, %u)  ...", p->channel, p, index);
#endif
	if (analog_callbacks.is_dialing) {
		return analog_callbacks.is_dialing(p->chan_pvt, index);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_is_dialing(%p, %d) is_dialing is NULL.", p->channel, p, index);
#endif
	return -1;
}

/*!
 * \internal
 * \brief Attempt to transfer 3-way call.
 *
 * \param p Analog private structure.
 *
 * \note On entry these locks are held: real-call, private, 3-way call.
 * \note On exit these locks are held: real-call, private.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int analog_attempt_transfer(struct analog_pvt *p)
{
	struct ast_channel *owner_real;
	struct ast_channel *owner_3way;
	enum ast_transfer_result xfer_res;	
	int res = 0;

#ifdef GRANDSTREAM_NETWORKS
	if (NULL == p)
	{
		ast_log(LOG_VERBOSE, "%s(%p) parameter is NULL!!", __FUNCTION__, p);
		return -1;
	}

	ast_log(LOG_VERBOSE, "[ch:%d]analog_attempt_transfer(%p) ...", p->channel, p);
#endif
	owner_real = ast_channel_ref(p->subs[ANALOG_SUB_REAL].owner);
	owner_3way = ast_channel_ref(p->subs[ANALOG_SUB_THREEWAY].owner);

	ast_verb(3, "TRANSFERRING %s to %s\n",
		ast_channel_name(owner_3way), ast_channel_name(owner_real));

	ast_channel_unlock(owner_real);
	ast_channel_unlock(owner_3way);
	analog_unlock_private(p);

	xfer_res = ast_bridge_transfer_attended(owner_3way, owner_real);
	if (xfer_res != AST_BRIDGE_TRANSFER_SUCCESS) {
		ast_softhangup(owner_3way, AST_SOFTHANGUP_DEV);
		res = -1;
	}

	/* Must leave with these locked. */
	ast_channel_lock(owner_real);
	analog_lock_private(p);

	ast_channel_unref(owner_real);
	ast_channel_unref(owner_3way);

	return res;
}

static int analog_update_conf(struct analog_pvt *p)
{
	int x;
	int needconf = 0;

	/* Start with the obvious, general stuff */
	for (x = 0; x < 3; x++) {
		/* Look for three way calls */
		if ((p->subs[x].allocd) && p->subs[x].inthreeway) {
			if (analog_callbacks.conf_add) {
				analog_callbacks.conf_add(p->chan_pvt, x);
			}
			needconf++;
		} else {
			if (analog_callbacks.conf_del) {
				analog_callbacks.conf_del(p->chan_pvt, x);
			}
		}
	}
	ast_debug(1, "Updated conferencing on %d, with %d conference users\n", p->channel, needconf);

	if (analog_callbacks.complete_conference_update) {
		analog_callbacks.complete_conference_update(p->chan_pvt, needconf);
	}
	return 0;
}

struct ast_channel * analog_request(struct analog_pvt *p, int *callwait, const struct ast_channel *requestor)
{
	struct ast_channel *ast;

#ifdef GRANDSTREAM_NETWORKS
	if (NULL == p)
	{
		ast_log(LOG_ERROR, "%s(0x%08tx, 0x%08tx, 0x%08tx) parameter is NULL !!", 
			__FUNCTION__, (ptrdiff_t)p, (ptrdiff_t)callwait, (ptrdiff_t)requestor);
		return NULL;
	}

	ast_debug(1, "[ch:%d]analog_request(%p, %p, %p) owner<%p:%s> ...",
		p->channel, p, callwait, requestor, p->owner, (p->owner ? ast_channel_name(p->owner) : "Null"));
#else
	ast_debug(1, "%s %d\n", __FUNCTION__, p->channel);
#endif
	*callwait = (p->owner != NULL);

	if (p->owner) {
		if (analog_alloc_sub(p, ANALOG_SUB_CALLWAIT)) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_ERROR, "[ch:%d]Unable to alloc subchannel !!", p->channel);
#else
			ast_log(LOG_ERROR, "Unable to alloc subchannel\n");
#endif
			return NULL;
		}
	}

	analog_set_outgoing(p, 1);
	ast = analog_new_ast_channel(p, AST_STATE_RESERVED, 0,
		p->owner ? ANALOG_SUB_CALLWAIT : ANALOG_SUB_REAL, requestor);
	if (!ast) {
		analog_set_outgoing(p, 0);
	}
	return ast;
}

int analog_available(struct analog_pvt *p)
{
	int offhook;
	
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "[ch:%d]analog_available(0x%08tx)...", p->channel, (ptrdiff_t)p);
#else
	ast_debug(1, "%s %d\n", __FUNCTION__, p->channel);
#endif

	/* If do not disturb, definitely not */
	if (p->dnd) {
		return 0;
	}
	/* If guard time, definitely not */
	if (p->guardtime && (time(NULL) < p->guardtime)) {
		return 0;
	}

	/* If no owner definitely available */
	if (!p->owner) {
		offhook = analog_is_off_hook(p);

		/* TDM FXO card, "onhook" means out of service (no battery on the line) */
		if ((p->sig == ANALOG_SIG_FXSLS) || (p->sig == ANALOG_SIG_FXSKS) || (p->sig == ANALOG_SIG_FXSGS)) {
#ifdef DAHDI_CHECK_HOOKSTATE
			if (offhook) {
				return 1;
			}
			return 0;
#endif
		/* TDM FXS card, "offhook" means someone took the hook off so it's unavailable! */
		} else if (offhook) {
			ast_debug(1, "Channel %d off hook, can't use\n", p->channel);
			/* Not available when the other end is off hook */
			return 0;
		}
		return 1;
	}

	/* If it's not an FXO, forget about call wait */
	if ((p->sig != ANALOG_SIG_FXOKS) && (p->sig != ANALOG_SIG_FXOLS) && (p->sig != ANALOG_SIG_FXOGS)) {
		return 0;
	}

	if (!p->callwaiting) {
		/* If they don't have call waiting enabled, then for sure they're unavailable at this point */
		return 0;
	}

	if (p->subs[ANALOG_SUB_CALLWAIT].allocd) {
		/* If there is already a call waiting call, then we can't take a second one */
		return 0;
	}

	if ((ast_channel_state(p->owner) != AST_STATE_UP) &&
	    ((ast_channel_state(p->owner) != AST_STATE_RINGING) || p->outgoing)) {
		/* If the current call is not up, then don't allow the call */
		return 0;
	}
	if ((p->subs[ANALOG_SUB_THREEWAY].owner) && (!p->subs[ANALOG_SUB_THREEWAY].inthreeway)) {
		/* Can't take a call wait when the three way calling hasn't been merged yet. */
		return 0;
	}
	/* We're cool */
	return 1;
}

static int analog_stop_callwait(struct analog_pvt *p)
{
	p->callwaitcas = 0;
	if (analog_callbacks.stop_callwait) {
		return analog_callbacks.stop_callwait(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_stop_callwait(%p) stop_callwait is NULL.", p->channel, p);
#endif
	return 0;
}

static int analog_callwait(struct analog_pvt *p)
{
	p->callwaitcas = p->callwaitingcallerid;
	if (analog_callbacks.callwait) {
		return analog_callbacks.callwait(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_callwait(%p) callwait is NULL.", p->channel, p);
#endif
	return 0;
}

static void analog_set_callwaiting(struct analog_pvt *p, int callwaiting_enable)
{
	p->callwaiting = callwaiting_enable;
	if (analog_callbacks.set_callwaiting) {
		analog_callbacks.set_callwaiting(p->chan_pvt, callwaiting_enable);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_set_callwaiting(%p, %d) set_callwaiting is NULL.", p->channel, p, callwaiting_enable);
	}
#endif
}

static void analog_set_cadence(struct analog_pvt *p, struct ast_channel *chan)
{
	if (analog_callbacks.set_cadence) {
		analog_callbacks.set_cadence(p->chan_pvt, &p->cidrings, chan);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_set_cadence(%p, %p) set_cadence is NULL.", p->channel, p, chan);
	}
#endif
}

static void analog_set_dialing(struct analog_pvt *p, int is_dialing)
{
	p->dialing = is_dialing;
	if (analog_callbacks.set_dialing) {
		analog_callbacks.set_dialing(p->chan_pvt, is_dialing);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_set_dialing(%p, %d) set_dialing is NULL.", p->channel, p, is_dialing);
	}
#endif
}

static void analog_set_alarm(struct analog_pvt *p, int in_alarm)
{
#ifdef GRANDSTREAM_NETWORKS
	if (NULL == p)
	{
		ast_log(LOG_ERROR, "%s(%p, %d) parameter is NULL!!", __FUNCTION__, p, in_alarm);
		return;
	}

	ast_log(LOG_NOTICE, "analog_set_alarm(%p, %d) ...",	p, in_alarm);
#endif
	p->inalarm = in_alarm;
	if (analog_callbacks.set_alarm) {
		analog_callbacks.set_alarm(p->chan_pvt, in_alarm);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_set_alarm(%p, %d) set_alarm is NULL.", p->channel, p, in_alarm);
	}
#endif
}

static void analog_set_ringtimeout(struct analog_pvt *p, int ringt)
{
#ifdef GRANDSTREAM_NETWORKS
	if (NULL == p)
	{
		ast_log(LOG_ERROR, "%s(%p, %d) parameter is NULL!!", __FUNCTION__, p, ringt);
		return;
	}

	ast_log(LOG_VERBOSE, "[ch:%d]analog_set_ringtimeout(%p, %d) ...", p->channel, p, ringt);
#endif
	p->ringt = ringt;
	if (analog_callbacks.set_ringtimeout) {
		analog_callbacks.set_ringtimeout(p->chan_pvt, ringt);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_set_ringtimeout(%p, %d) set_ringtimeout is NULL.", p->channel, p, ringt);
	}
#endif
}

static void analog_set_waitingfordt(struct analog_pvt *p, struct ast_channel *ast)
{
	if (analog_callbacks.set_waitingfordt) {
		analog_callbacks.set_waitingfordt(p->chan_pvt, ast);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_set_waitingfordt(%p, %p) set_waitingfordt is NULL.", p->channel, p, ast);
	}
#endif
}

static int analog_check_waitingfordt(struct analog_pvt *p)
{
	if (analog_callbacks.check_waitingfordt) {
		return analog_callbacks.check_waitingfordt(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_check_waitingfordt(%p) check_waitingfordt is NULL.", p->channel, p);
#endif
	return 0;
}

static void analog_set_confirmanswer(struct analog_pvt *p, int flag)
{
	if (analog_callbacks.set_confirmanswer) {
		analog_callbacks.set_confirmanswer(p->chan_pvt, flag);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_set_confirmanswer(%p, %d) set_confirmanswer is NULL.", p->channel, p, flag);
	}
#endif
}

static int analog_check_confirmanswer(struct analog_pvt *p)
{
	if (analog_callbacks.check_confirmanswer) {
		return analog_callbacks.check_confirmanswer(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_check_confirmanswer(%p) check_confirmanswer is NULL.", p->channel, p);
#endif
	return 0;
}

static void analog_cancel_cidspill(struct analog_pvt *p)
{
	if (analog_callbacks.cancel_cidspill) {
		analog_callbacks.cancel_cidspill(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_cancel_cidspill(%p) cancel_cidspill is NULL.", p->channel, p);
	}
#endif
}

static int analog_confmute(struct analog_pvt *p, int mute)
{
	if (analog_callbacks.confmute) {
		return analog_callbacks.confmute(p->chan_pvt, mute);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_confmute(%p,%d) confmute is NULL.", p->channel, p, mute);
#endif
	return 0;
}

static void analog_set_pulsedial(struct analog_pvt *p, int flag)
{
	if (analog_callbacks.set_pulsedial) {
		analog_callbacks.set_pulsedial(p->chan_pvt, flag);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_set_pulsedial(%p,%d) set_pulsedial is NULL.", p->channel, p,flag);
	}
#endif
}

static int analog_set_linear_mode(struct analog_pvt *p, enum analog_sub sub, int linear_mode)
{
	if (analog_callbacks.set_linear_mode) {
		/* Return provides old linear_mode setting or error indication */
		return analog_callbacks.set_linear_mode(p->chan_pvt, sub, linear_mode);
	}
#ifdef GRANDSTREAM_NETWORKS
    ast_log(LOG_NOTICE, "[ch:%d]analog_set_linear_mode(%p, %u, %d) set_linear_mode is NULL.", p->channel, p, sub, linear_mode);
#endif
	
	return -1;
}

static void analog_set_inthreeway(struct analog_pvt *p, enum analog_sub sub, int inthreeway)
{
	p->subs[sub].inthreeway = inthreeway;
	if (analog_callbacks.set_inthreeway) {
		analog_callbacks.set_inthreeway(p->chan_pvt, sub, inthreeway);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
	    ast_log(LOG_NOTICE, "[ch:%d]analog_set_inthreeway(%p, %u, %d) set_inthreeway is NULL.", p->channel, p, sub, inthreeway);
	}
#endif
}
#ifdef GRANDSTREAM_NETWORKS
static void analog_set_cw_send_voice(struct analog_pvt *p,int enable)
{
	if (analog_callbacks.set_cw_send_voice) {
		analog_callbacks.set_cw_send_voice(p->chan_pvt, enable);
	}
	else
	{
	    ast_log(LOG_NOTICE, "[ch:%d]analog_set_cw_send_voice(%p, %d) set_cw_send_voice is NULL.", p->channel, p, enable);
	}
}

static void analog_set_ring_after_onhook(struct analog_pvt *p,int enable)
{
    p->reringwhenhanguped = enable;
	if (analog_callbacks.set_ring_after_onhook) {
		analog_callbacks.set_ring_after_onhook(p->chan_pvt, enable);
	}
	else
	{
	    ast_log(LOG_NOTICE, "[ch:%d]analog_set_ring_after_onhook(%p, %d) set_ring_after_onhook is NULL.", p->channel, p, enable);
	}
}

static int analog_get_holdon(struct analog_pvt *p, enum analog_sub sub)
{
	if (analog_callbacks.get_holdon) {
		return analog_callbacks.get_holdon(p->chan_pvt, sub);
	}
	else
	{
	    ast_log(LOG_NOTICE, "[ch:%d]analog_get_holdon(%p, %u) get_holdon is NULL.", p->channel, p, sub);
        return -1;
    }
}

static void analog_play_moh(struct analog_pvt *p, int enable)
{
	if (analog_callbacks.play_moh) {
		analog_callbacks.play_moh(p->chan_pvt, enable);
	}
	else
	{
	    ast_log(LOG_NOTICE, "[ch:%d]analog_play_moh(%p, %d) play_moh is NULL.", p->channel, p, enable);
    }
}

static void analog_set_vmwi(struct analog_pvt *p, int msgnum)
{
	if (analog_callbacks.set_vmwi) {
		analog_callbacks.set_vmwi(p->chan_pvt, msgnum);
	}else{
	    ast_log(LOG_NOTICE, "[ch:%d]analog_set_vmwi(%p, %d) set_vmwi is NULL.", p->channel, p, msgnum);
    }
}
static void analog_set_exten(struct analog_pvt *p, char* exten)
{
	if (analog_callbacks.set_exten) {
		analog_callbacks.set_exten(p->chan_pvt, exten);
	}else{
	    ast_log(LOG_NOTICE, "[ch:%d]analog_set_exten(%p, %s) set_vmwi is NULL.", p->channel, p, exten);
    }
}

#endif
int analog_call(struct analog_pvt *p, struct ast_channel *ast, const char *rdest, int timeout)
{
	int res, idx, mysig;
	char *c, *n, *l;
	char dest[256]; /* must be same length as p->dialdest */
	
#ifdef GRANDSTREAM_NETWORKS
	if (NULL == p || NULL == ast)
	{
		ast_log(LOG_ERROR, "%s(%p, %p, %p, %d) parameter is NULL !!", __FUNCTION__, p, ast, rdest, timeout);
		return -1;
	}

	ast_log(LOG_VERBOSE, "[ch:%d]analog_call(%p, %p, %p, %d) channel:%s owner<%p:%s> CALLING CID_NAME:%s, CID_NUM::%s",
		p->channel, p, ast, rdest, timeout, ast_channel_name(ast), p->owner, (p->owner ?ast_channel_name(p->owner) : "NULL"),
		S_COR(ast_channel_connected(ast)->id.name.valid, ast_channel_connected(ast)->id.name.str, ""),
		S_COR(ast_channel_connected(ast)->id.number.valid, ast_channel_connected(ast)->id.number.str, ""));
#else
	ast_debug(1, "CALLING CID_NAME: %s CID_NUM:: %s\n",
		S_COR(ast_channel_connected(ast)->id.name.valid, ast_channel_connected(ast)->id.name.str, ""),
		S_COR(ast_channel_connected(ast)->id.number.valid, ast_channel_connected(ast)->id.number.str, ""));
#endif

	ast_copy_string(dest, rdest, sizeof(dest));
	ast_copy_string(p->dialdest, rdest, sizeof(p->dialdest));

	if ((ast_channel_state(ast) == AST_STATE_BUSY)) {
		ast_queue_control(p->subs[ANALOG_SUB_REAL].owner, AST_CONTROL_BUSY);
		return 0;
	}

	if ((ast_channel_state(ast) != AST_STATE_DOWN) && (ast_channel_state(ast) != AST_STATE_RESERVED)) {
		ast_log(LOG_WARNING, "analog_call called on %s, neither down nor reserved\n", ast_channel_name(ast));
		return -1;
	}

	p->dialednone = 0;
	analog_set_outgoing(p, 1);

	mysig = p->sig;
	if (p->outsigmod > -1) {
		mysig = p->outsigmod;
	}

	switch (mysig) {
	case ANALOG_SIG_FXOLS:
	case ANALOG_SIG_FXOGS:
	case ANALOG_SIG_FXOKS:
		if (p->owner == ast) {
			/* Normal ring, on hook */

			/* Don't send audio while on hook, until the call is answered */
			analog_set_dialing(p, 1);
			analog_set_cadence(p, ast); /* and set p->cidrings */

			/* nick@dccinc.com 4/3/03 mods to allow for deferred dialing */
			c = strchr(dest, '/');
			if (c) {
				c++;
			}
			if (c && (strlen(c) < p->stripmsd)) {
				ast_log(LOG_WARNING, "Number '%s' is shorter than stripmsd (%d)\n", c, p->stripmsd);
				c = NULL;
			}
			if (c) {
				p->dop.op = ANALOG_DIAL_OP_REPLACE;
				snprintf(p->dop.dialstr, sizeof(p->dop.dialstr), "Tw%s", c);
				ast_debug(1, "FXO: setup deferred dialstring: %s\n", c);
			} else {
				p->dop.dialstr[0] = '\0';
			}

#ifdef GRANDSTREAM_NETWORKS
			analog_set_vmwi(p, 0);
#endif

			if (analog_ring(p)) {
				ast_log(LOG_WARNING, "Unable to ring phone: %s\n", strerror(errno));
				return -1;
			}
			analog_set_dialing(p, 1);
		} else {
			/* Call waiting call */
			if (ast_channel_connected(ast)->id.number.valid && ast_channel_connected(ast)->id.number.str) {
				ast_copy_string(p->callwait_num, ast_channel_connected(ast)->id.number.str, sizeof(p->callwait_num));
			} else {
				p->callwait_num[0] = '\0';
			}
			if (ast_channel_connected(ast)->id.name.valid && ast_channel_connected(ast)->id.name.str) {
				ast_copy_string(p->callwait_name, ast_channel_connected(ast)->id.name.str, sizeof(p->callwait_name));
			} else {
				p->callwait_name[0] = '\0';
			}

			/* Call waiting tone instead */
			if (analog_callwait(p)) {
				return -1;
			}
			/* Make ring-back */
			if (analog_play_tone(p, ANALOG_SUB_CALLWAIT, ANALOG_TONE_RINGTONE)) {
				ast_log(LOG_WARNING, "Unable to generate call-wait ring-back on channel %s\n", ast_channel_name(ast));
			}

		}
		n = ast_channel_connected(ast)->id.name.valid ? ast_channel_connected(ast)->id.name.str : NULL;
		l = ast_channel_connected(ast)->id.number.valid ? ast_channel_connected(ast)->id.number.str : NULL;
		if (l) {
			ast_copy_string(p->lastcid_num, l, sizeof(p->lastcid_num));
		} else {
			p->lastcid_num[0] = '\0';
		}
		if (n) {
			ast_copy_string(p->lastcid_name, n, sizeof(p->lastcid_name));
		} else {
			p->lastcid_name[0] = '\0';
		}

		if (p->use_callerid) {
			p->caller.id.name.str = p->lastcid_name;
			p->caller.id.number.str = p->lastcid_num;
		}

		ast_setstate(ast, AST_STATE_RINGING);
		idx = analog_get_index(ast, p, 0);
		if (idx > -1) {
			struct ast_cc_config_params *cc_params;

			/* This is where the initial ringing frame is queued for an analog call.
			 * As such, this is a great time to offer CCNR to the caller if it's available.
			 */
			cc_params = ast_channel_get_cc_config_params(p->subs[idx].owner);
			if (cc_params) {
				switch (ast_get_cc_monitor_policy(cc_params)) {
				case AST_CC_MONITOR_NEVER:
					break;
				case AST_CC_MONITOR_NATIVE:
				case AST_CC_MONITOR_ALWAYS:
				case AST_CC_MONITOR_GENERIC:
					ast_queue_cc_frame(p->subs[idx].owner, AST_CC_GENERIC_MONITOR_TYPE,
						analog_get_orig_dialstring(p), AST_CC_CCNR, NULL);
					break;
				}
			}
			ast_queue_control(p->subs[idx].owner, AST_CONTROL_RINGING);
		}
		break;
	case ANALOG_SIG_FXSLS:
	case ANALOG_SIG_FXSGS:
	case ANALOG_SIG_FXSKS:
		if (p->answeronpolarityswitch || p->hanguponpolarityswitch) {
			ast_debug(1, "Ignore possible polarity reversal on line seizure\n");
			p->polaritydelaytv = ast_tvnow();
		}
		/* fall through */
	case ANALOG_SIG_EMWINK:
	case ANALOG_SIG_EM:
	case ANALOG_SIG_EM_E1:
	case ANALOG_SIG_FEATD:
	case ANALOG_SIG_FEATDMF:
	case ANALOG_SIG_E911:
	case ANALOG_SIG_FGC_CAMA:
	case ANALOG_SIG_FGC_CAMAMF:
	case ANALOG_SIG_FEATB:
	case ANALOG_SIG_SFWINK:
	case ANALOG_SIG_SF:
	case ANALOG_SIG_SF_FEATD:
	case ANALOG_SIG_SF_FEATDMF:
	case ANALOG_SIG_FEATDMF_TA:
	case ANALOG_SIG_SF_FEATB:
		c = strchr(dest, '/');
		if (c) {
			c++;
		} else {
			c = "";
		}
		if (strlen(c) < p->stripmsd) {
			ast_log(LOG_WARNING, "Number '%s' is shorter than stripmsd (%d)\n", c, p->stripmsd);
			return -1;
		}
		res = analog_start(p);
		if (res < 0) {
			if (errno != EINPROGRESS) {
				return -1;
			}
		}
		ast_debug(1, "Dialing '%s'\n", c);
		p->dop.op = ANALOG_DIAL_OP_REPLACE;

		c += p->stripmsd;

		switch (mysig) {
		case ANALOG_SIG_FEATD:
			l = ast_channel_connected(ast)->id.number.valid ? ast_channel_connected(ast)->id.number.str : NULL;
			if (l) {
				snprintf(p->dop.dialstr, sizeof(p->dop.dialstr), "T*%s*%s*", l, c);
			} else {
				snprintf(p->dop.dialstr, sizeof(p->dop.dialstr), "T**%s*", c);
			}
			break;
		case ANALOG_SIG_FEATDMF:
			l = ast_channel_connected(ast)->id.number.valid ? ast_channel_connected(ast)->id.number.str : NULL;
			if (l) {
				snprintf(p->dop.dialstr, sizeof(p->dop.dialstr), "M*00%s#*%s#", l, c);
			} else {
				snprintf(p->dop.dialstr, sizeof(p->dop.dialstr), "M*02#*%s#", c);
			}
			break;
		case ANALOG_SIG_FEATDMF_TA:
		{
			const char *cic = "", *ozz = "";

			/* If you have to go through a Tandem Access point you need to use this */
#ifndef STANDALONE
			ozz = pbx_builtin_getvar_helper(p->owner, "FEATDMF_OZZ");
			if (!ozz) {
				ozz = analog_defaultozz;
			}
			cic = pbx_builtin_getvar_helper(p->owner, "FEATDMF_CIC");
			if (!cic) {
				cic = analog_defaultcic;
			}
#endif
			if (!ozz || !cic) {
				ast_log(LOG_WARNING, "Unable to dial channel of type feature group D MF tandem access without CIC or OZZ set\n");
				return -1;
			}
			snprintf(p->dop.dialstr, sizeof(p->dop.dialstr), "M*%s%s#", ozz, cic);
			snprintf(p->finaldial, sizeof(p->finaldial), "M*%s#", c);
			p->whichwink = 0;
		}
			break;
		case ANALOG_SIG_E911:
			ast_copy_string(p->dop.dialstr, "M*911#", sizeof(p->dop.dialstr));
			break;
		case ANALOG_SIG_FGC_CAMA:
			snprintf(p->dop.dialstr, sizeof(p->dop.dialstr), "P%s", c);
			break;
		case ANALOG_SIG_FGC_CAMAMF:
		case ANALOG_SIG_FEATB:
			snprintf(p->dop.dialstr, sizeof(p->dop.dialstr), "M*%s#", c);
			break;
		default:
			if (p->pulse) {
				snprintf(p->dop.dialstr, sizeof(p->dop.dialstr), "P%sw", c);
			} else {
				snprintf(p->dop.dialstr, sizeof(p->dop.dialstr), "T%sw", c);
			}
			break;
		}

		if (p->echotraining && (strlen(p->dop.dialstr) > 4)) {
			memset(p->echorest, 'w', sizeof(p->echorest) - 1);
			strcpy(p->echorest + (p->echotraining / 400) + 1, p->dop.dialstr + strlen(p->dop.dialstr) - 2);
			p->echorest[sizeof(p->echorest) - 1] = '\0';
			p->echobreak = 1;
			p->dop.dialstr[strlen(p->dop.dialstr)-2] = '\0';
		} else {
			p->echobreak = 0;
		}
		analog_set_waitingfordt(p, ast);
		if (!res) {
			if (analog_dial_digits(p, ANALOG_SUB_REAL, &p->dop)) {
				analog_on_hook(p);
				return -1;
			}
		} 
		else 
		{
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_NOTICE, "[ch:%d]Res:%d errno:%d(%s), Deferring dialing...", p->channel, res, errno, strerror(errno));
#else
			ast_debug(1, "Deferring dialing...\n");
#endif
		}
		analog_set_dialing(p, 1);
		if (ast_strlen_zero(c)) {
			p->dialednone = 1;
		}
		ast_setstate(ast, AST_STATE_DIALING);
		break;
	default:
		ast_debug(1, "not yet implemented\n");
		return -1;
	}
	return 0;
}

int analog_hangup(struct analog_pvt *p, struct ast_channel *ast)
{
	int res;
	int idx, x;

#ifdef GRANDSTREAM_NETWORKS
	if (NULL == p || NULL == ast)
	{
		ast_log(LOG_ERROR, "%s(0x%08tx, 0x%08tx) parameter is NULL !!", __FUNCTION__, (ptrdiff_t)p, (ptrdiff_t)ast);
		return -1;
	}
	p->ucHasReceiveRing = 0;
	ast_log(LOG_VERBOSE, "[ch:%d]analog_hangup(0x%08tx, 0x%08tx) Real:0x%08tx, CallWait:0x%08tx, ThreeWay:0x%08tx ...",
		p->channel, (ptrdiff_t)p, (ptrdiff_t)ast, 
		(ptrdiff_t)p->subs[ANALOG_SUB_REAL].owner, (ptrdiff_t)p->subs[ANALOG_SUB_CALLWAIT].owner, (ptrdiff_t)p->subs[ANALOG_SUB_THREEWAY].owner);
#else
	ast_debug(1, "%s %d\n", __FUNCTION__, p->channel);
#endif

	if (!ast_channel_tech_pvt(ast)) {
		ast_log(LOG_WARNING, "Asked to hangup channel not connected\n");
		return 0;
	}

	idx = analog_get_index(ast, p, 1);

	x = 0;
	if (p->origcid_num) {
		ast_copy_string(p->cid_num, p->origcid_num, sizeof(p->cid_num));
		ast_free(p->origcid_num);
		p->origcid_num = NULL;
	}
	if (p->origcid_name) {
		ast_copy_string(p->cid_name, p->origcid_name, sizeof(p->cid_name));
		ast_free(p->origcid_name);
		p->origcid_name = NULL;
	}

	analog_dsp_set_digitmode(p, ANALOG_DIGITMODE_DTMF);
	
#ifdef GRANDSTREAM_NETWORKS
	ast_debug(1, "[ch:%d]Hangup index:%d, real:<%s,allocd:%d>, callwait:<%s,allocd:%d>, thirdcall:<%s,allocd:%d>",
		p->channel, idx,
		(p->subs[ANALOG_SUB_REAL].owner ? ast_channel_name(p->subs[ANALOG_SUB_REAL].owner): "Null"), p->subs[ANALOG_SUB_REAL].allocd,
		(p->subs[ANALOG_SUB_CALLWAIT].owner ? ast_channel_name(p->subs[ANALOG_SUB_CALLWAIT].owner): "Null"), p->subs[ANALOG_SUB_CALLWAIT].allocd,
		(p->subs[ANALOG_SUB_THREEWAY].owner ? ast_channel_name(p->subs[ANALOG_SUB_THREEWAY].owner): "Null"), p->subs[ANALOG_SUB_THREEWAY].allocd);
#else
	ast_debug(1, "Hangup: channel: %d index = %d, normal = %d, callwait = %d, thirdcall = %d\n",
		p->channel, idx, p->subs[ANALOG_SUB_REAL].allocd, p->subs[ANALOG_SUB_CALLWAIT].allocd, p->subs[ANALOG_SUB_THREEWAY].allocd);
#endif
	if (idx > -1) {
		/* Real channel, do some fixup */
		p->subs[idx].owner = NULL;
		p->polarity = POLARITY_IDLE;
		analog_set_linear_mode(p, idx, 0);
		switch (idx) {
		case ANALOG_SUB_REAL:
			if (p->subs[ANALOG_SUB_CALLWAIT].allocd && p->subs[ANALOG_SUB_THREEWAY].allocd) 
			{
#ifdef GRANDSTREAM_NETWORKS
				ast_debug(1, "[ch:%d]Normal call hung up with both three way call and a call waiting call in place? callwaitsub_inthreeway:%u.",
					p->channel, p->subs[ANALOG_SUB_CALLWAIT].inthreeway);
#else
				ast_debug(1, "Normal call hung up with both three way call and a call waiting call in place?\n");
#endif
				if (p->subs[ANALOG_SUB_CALLWAIT].inthreeway) {
					/* We had flipped over to answer a callwait and now it's gone */
					ast_debug(1, "We were flipped over to the callwait, moving back and unowning.\n");
					/* Move to the call-wait, but un-own us until they flip back. */
					analog_swap_subs(p, ANALOG_SUB_CALLWAIT, ANALOG_SUB_REAL);
					analog_unalloc_sub(p, ANALOG_SUB_CALLWAIT);
					analog_set_new_owner(p, NULL);
				} else {
					/* The three way hung up, but we still have a call wait */
#ifdef GRANDSTREAM_NETWORKS
					ast_debug(1, "[ch:%d]We were in the threeway and have a callwait still. Ditching the threeway. realsub_inthreeway:%u.",
						p->channel, p->subs[ANALOG_SUB_REAL].inthreeway);
#else
					ast_debug(1, "We were in the threeway and have a callwait still.  Ditching the threeway.\n");
#endif
					analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
					analog_unalloc_sub(p, ANALOG_SUB_THREEWAY);
#ifdef GRANDSTREAM_NETWORKS
					analog_set_inthreeway(p, ANALOG_SUB_THREEWAY, 0);
#endif
					if (p->subs[ANALOG_SUB_REAL].inthreeway) {
						/* This was part of a three way call.  Immediately make way for
						   another call */
#ifdef GRANDSTREAM_NETWORKS
						ast_debug(1, "[ch:%d]Call was complete, setting owner to former third call", p->channel);
#else
						ast_debug(1, "Call was complete, setting owner to former third call\n");
#endif
						analog_set_inthreeway(p, ANALOG_SUB_REAL, 0);
						analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
					} else {
						/* This call hasn't been completed yet...  Set owner to NULL */
#ifdef GRANDSTREAM_NETWORKS
						ast_debug(1, "[ch:%d]Call was incomplete, setting owner to NULL", p->channel);
#else
						ast_debug(1, "Call was incomplete, setting owner to NULL\n");
#endif
						analog_set_new_owner(p, NULL);
					}
				}
			} else if (p->subs[ANALOG_SUB_CALLWAIT].allocd) {
				/* Need to hold the lock for real-call, private, and call-waiting call */
				analog_lock_sub_owner(p, ANALOG_SUB_CALLWAIT);
				if (!p->subs[ANALOG_SUB_CALLWAIT].owner) {
					/* The call waiting call dissappeared. */
					analog_set_new_owner(p, NULL);
					break;
				}
				
#ifdef GRANDSTREAM_NETWORKS
				{					
					ast_debug(1, "[ch:%d]callwait Call was incomplete, setting owner to NULL and play busy tone.", p->channel); 				 
					analog_swap_subs(p, ANALOG_SUB_CALLWAIT, ANALOG_SUB_REAL);
					analog_unalloc_sub(p, ANALOG_SUB_CALLWAIT); 
					analog_play_tone(p, idx, ANALOG_TONE_BUSY);
					analog_set_cw_send_voice(p,0);
					analog_set_new_owner(p, NULL);					
					ast_channel_unlock(p->subs[ANALOG_SUB_REAL].owner);
					break;
				}
#endif
				/* Move to the call-wait and switch back to them. */
				analog_swap_subs(p, ANALOG_SUB_CALLWAIT, ANALOG_SUB_REAL);
				analog_unalloc_sub(p, ANALOG_SUB_CALLWAIT);
				analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
				if (ast_channel_state(p->owner) != AST_STATE_UP) {
					ast_queue_control(p->subs[ANALOG_SUB_REAL].owner, AST_CONTROL_ANSWER);
#ifdef GRANDSTREAM_NETWORKS
					ast_log(LOG_NOTICE, "[ch:%d]When channel answer, callwait subchannel is %d(%s), callwait...",
						p->channel, ast_channel_state(p->owner), ast_state2str(ast_channel_state(p->owner)));
#endif
				}
				ast_queue_unhold(p->subs[ANALOG_SUB_REAL].owner);
				/* Unlock the call-waiting call that we swapped to real-call. */
				ast_channel_unlock(p->subs[ANALOG_SUB_REAL].owner);
			} else if (p->subs[ANALOG_SUB_THREEWAY].allocd) {
				analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
				analog_unalloc_sub(p, ANALOG_SUB_THREEWAY);
#ifdef GRANDSTREAM_NETWORKS
				analog_set_inthreeway(p, ANALOG_SUB_THREEWAY, 0);
				analog_set_inthreeway(p, ANALOG_SUB_THREEWAY, 0);
#endif
				if (p->subs[ANALOG_SUB_REAL].inthreeway) {
					/* This was part of a three way call.  Immediately make way for
					   another call */
#ifdef GRANDSTREAM_NETWORKS
					ast_debug(1, "[ch:%d]Call was complete, setting owner to former third call", p->channel);
#else
					ast_debug(1, "Call was complete, setting owner to former third call\n");
#endif
					analog_set_inthreeway(p, ANALOG_SUB_REAL, 0);
					analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
				} else {
					/* This call hasn't been completed yet...  Set owner to NULL */
#ifdef GRANDSTREAM_NETWORKS
					ast_debug(1, "[ch:%d]Call was incomplete, setting owner to NULL and play busy tone.", p->channel);
#else
					ast_debug(1, "Call was incomplete, setting owner to NULL\n");
#endif
					analog_set_new_owner(p, NULL);
#ifdef GRANDSTREAM_NETWORKS
					analog_play_tone(p, idx, ANALOG_TONE_BUSY);
#endif
				}
			}
			break;
		case ANALOG_SUB_CALLWAIT:
			/* Ditch the holding callwait call, and immediately make it available */
			if (p->subs[ANALOG_SUB_CALLWAIT].inthreeway) {
				/* Need to hold the lock for call-waiting call, private, and 3-way call */
				analog_lock_sub_owner(p, ANALOG_SUB_THREEWAY);

				/* This is actually part of a three way, placed on hold.  Place the third part
				   on music on hold now */
				if (p->subs[ANALOG_SUB_THREEWAY].owner) {
					ast_queue_hold(p->subs[ANALOG_SUB_THREEWAY].owner, p->mohsuggest);
				}
				analog_set_inthreeway(p, ANALOG_SUB_THREEWAY, 0);
				/* Make it the call wait now */
				analog_swap_subs(p, ANALOG_SUB_CALLWAIT, ANALOG_SUB_THREEWAY);
				analog_unalloc_sub(p, ANALOG_SUB_THREEWAY);
				if (p->subs[ANALOG_SUB_CALLWAIT].owner) {
					/* Unlock the 3-way call that we swapped to call-waiting call. */
					ast_channel_unlock(p->subs[ANALOG_SUB_CALLWAIT].owner);
				}
			} else {
				analog_unalloc_sub(p, ANALOG_SUB_CALLWAIT);
#ifdef GRANDSTREAM_NETWORKS
				if(p->owner && (ast_channel_state(p->owner) == AST_STATE_RINGING))
				{
				   analog_play_tone(p, ANALOG_SUB_REAL, ANALOG_TONE_RINGTONE);
				}
#endif
			}
			break;
		case ANALOG_SUB_THREEWAY:
			/* Need to hold the lock for 3-way call, private, and call-waiting call */
			analog_lock_sub_owner(p, ANALOG_SUB_CALLWAIT);
			if (p->subs[ANALOG_SUB_CALLWAIT].inthreeway) {
				/* The other party of the three way call is currently in a call-wait state.
				   Start music on hold for them, and take the main guy out of the third call */
				analog_set_inthreeway(p, ANALOG_SUB_CALLWAIT, 0);
				if (p->subs[ANALOG_SUB_CALLWAIT].owner) {
					ast_queue_hold(p->subs[ANALOG_SUB_CALLWAIT].owner, p->mohsuggest);
				}
			}
			if (p->subs[ANALOG_SUB_CALLWAIT].owner) {
				ast_channel_unlock(p->subs[ANALOG_SUB_CALLWAIT].owner);
			}
			analog_set_inthreeway(p, ANALOG_SUB_REAL, 0);
			/* If this was part of a three way call index, let us make
			   another three way call */
			analog_unalloc_sub(p, ANALOG_SUB_THREEWAY);
#ifdef GRANDSTREAM_NETWORKS
			analog_set_inthreeway(p, ANALOG_SUB_THREEWAY, 0);
			if(1 == p->hangupallwheninthreeway)
			{
				if (p->subs[ANALOG_SUB_CALLWAIT].allocd)
				{
					/* Need to hold the lock for real-call, private, and call-waiting call */
					analog_lock_sub_owner(p, ANALOG_SUB_CALLWAIT);

					/* There's a call waiting call, so ring the phone, but make it unowned in the mean time */
					analog_swap_subs(p, ANALOG_SUB_CALLWAIT, ANALOG_SUB_REAL);
					ast_verb(3, "Channel %d still has (callwait) call, ringing phone\n", p->channel);
					analog_unalloc_sub(p, ANALOG_SUB_CALLWAIT);
					analog_stop_callwait(p);
					analog_set_new_owner(p, NULL);
					/* Don't start streaming audio yet if the incoming call isn't up yet */
					if (ast_channel_state(p->subs[ANALOG_SUB_REAL].owner) != AST_STATE_UP)
					{
						analog_set_dialing(p, 1);
					}
					/* Unlock the call-waiting call that we swapped to real-call. */
					ast_channel_unlock(p->subs[ANALOG_SUB_REAL].owner);
					analog_set_cadence(p, p->subs[ANALOG_SUB_REAL].owner); /* and set p->cidrings */
					analog_ring(p); 				
				}
				else
				{
					analog_lock_sub_owner(p, ANALOG_SUB_REAL);
					ast_queue_hangup_with_cause(p->subs[ANALOG_SUB_REAL].owner, AST_CAUSE_NO_ANSWER);
					ast_softhangup_nolock(p->subs[ANALOG_SUB_REAL].owner, AST_SOFTHANGUP_DEV);
					ast_channel_unlock(p->subs[ANALOG_SUB_REAL].owner);
				}
				p->hangupallwheninthreeway = 0;
			}
#endif
			break;
		default:
			/*
			 * Should never happen.
			 * This wasn't any sort of call, so how are we an index?
			 */
			ast_log(LOG_ERROR, "Index found but not any type of call?\n");
			break;
		}
	}

	if (!p->subs[ANALOG_SUB_REAL].owner && !p->subs[ANALOG_SUB_CALLWAIT].owner && !p->subs[ANALOG_SUB_THREEWAY].owner) {
		analog_set_new_owner(p, NULL);
		analog_set_ringtimeout(p, 0);
		analog_set_confirmanswer(p, 0);
		analog_set_pulsedial(p, 0);
		analog_set_outgoing(p, 0);
		p->onhooktime = time(NULL);
		p->cidrings = 1;
#ifdef GRANDSTREAM_NETWORKS
		p->msgstate = -1;
#if defined(SVIP_T38_STACK)
		if(3 == analog_get_t38_state(p))
		{
			sleep(1);
			analog_lock_t38_session(p);
			analog_stop_t38_session(p);
			analog_unlock_t38_session(p);
		}
		else
		{
			analog_set_t38_state(p, 0);
		}
#endif
#endif

		/* Perform low level hangup if no owner left */
		res = analog_on_hook(p);
		if (res < 0) {
			ast_log(LOG_WARNING, "Unable to hangup line %s\n", ast_channel_name(ast));
		}
		switch (p->sig) {
		case ANALOG_SIG_FXOGS:
		case ANALOG_SIG_FXOLS:
		case ANALOG_SIG_FXOKS:
			/* If they're off hook, try playing congestion */
			if (analog_is_off_hook(p)) {
				analog_hangup_polarityswitch(p);
				analog_play_tone(p, ANALOG_SUB_REAL, ANALOG_TONE_CONGESTION);
#ifdef GRANDSTREAM_NETWORKS
				p->fxsoffhookstate = 1;
#endif
			} else {
				analog_play_tone(p, ANALOG_SUB_REAL, -1);
#ifdef GRANDSTREAM_NETWORKS
				p->fxsoffhookstate = 0;
#endif
			}
#ifdef GRANDSTREAM_NETWORKS
			analog_set_cw_send_voice(p,1);
			analog_set_ring_after_onhook(p,0);
			p->reringswhenhanguped = 0;
#endif
			break;
		case ANALOG_SIG_FXSGS:
		case ANALOG_SIG_FXSLS:
		case ANALOG_SIG_FXSKS:
			/* Make sure we're not made available for at least two seconds assuming
			   we were actually used for an inbound or outbound call. */
			if (ast_channel_state(ast) != AST_STATE_RESERVED) {
				time(&p->guardtime);
				p->guardtime += 2;
			}
			break;
		default:
			analog_play_tone(p, ANALOG_SUB_REAL, -1);
			break;
		}

		analog_set_echocanceller(p, 0);

		x = 0;
		ast_channel_setoption(ast,AST_OPTION_TONE_VERIFY,&x,sizeof(char),0);
		ast_channel_setoption(ast,AST_OPTION_TDD,&x,sizeof(char),0);
		p->callwaitcas = 0;
		analog_set_callwaiting(p, p->permcallwaiting);
		p->hidecallerid = p->permhidecallerid;
		analog_set_dialing(p, 0);
		analog_update_conf(p);
		analog_all_subchannels_hungup(p);
	}

	analog_stop_callwait(p);

#ifdef GRANDSTREAM_NETWORKS
#ifdef CURRENT_DISCONNECT

    if (p->cd_timer_index >= 0)
    {
        ast_log(AST_LOG_WARNING, "[ch:%d]When channel hangup, current disconnect timier(%d) is exist!, delete it!!",
            p->channel, p->cd_timer_index);
            
        dahdi_timer_del(p->cd_timer_index);
        p->cd_timer_index = -1;
    }
#endif
	gettimeofday(&p->tv_hangup, NULL);
	p->ucHangupFlag = 1;
#endif

	ast_verb(3, "Hanging up on '%s'\n", ast_channel_name(ast));

	return 0;
}

int analog_answer(struct analog_pvt *p, struct ast_channel *ast)
{
	int res = 0;
	int idx;
	int oldstate = ast_channel_state(ast);
	
#ifdef GRANDSTREAM_NETWORKS
	if (NULL == p || NULL == ast)
	{
		ast_log(LOG_ERROR, "%s(0x%08tx, 0x%08tx) parameter is NULL !!", __FUNCTION__, (ptrdiff_t)p, (ptrdiff_t)ast);
		return -1;
	}

	ast_log(LOG_NOTICE, "[ch:%d]analog_answer(0x%08tx, 0x%08tx) channel signal:%u ...",
		p->channel, (ptrdiff_t)p, (ptrdiff_t)ast, p->sig);
#else
	ast_debug(1, "%s %d\n", __FUNCTION__, p->channel);
#endif
	ast_setstate(ast, AST_STATE_UP);
	idx = analog_get_index(ast, p, 1);
	if (idx < 0) {
		idx = ANALOG_SUB_REAL;
	}
	switch (p->sig) {
	case ANALOG_SIG_FXSLS:
	case ANALOG_SIG_FXSGS:
	case ANALOG_SIG_FXSKS:
		analog_set_ringtimeout(p, 0);
		/* Fall through */
	case ANALOG_SIG_EM:
	case ANALOG_SIG_EM_E1:
	case ANALOG_SIG_EMWINK:
	case ANALOG_SIG_FEATD:
	case ANALOG_SIG_FEATDMF:
	case ANALOG_SIG_FEATDMF_TA:
	case ANALOG_SIG_E911:
	case ANALOG_SIG_FGC_CAMA:
	case ANALOG_SIG_FGC_CAMAMF:
	case ANALOG_SIG_FEATB:
	case ANALOG_SIG_SF:
	case ANALOG_SIG_SFWINK:
	case ANALOG_SIG_SF_FEATD:
	case ANALOG_SIG_SF_FEATDMF:
	case ANALOG_SIG_SF_FEATB:
	case ANALOG_SIG_FXOLS:
	case ANALOG_SIG_FXOGS:
	case ANALOG_SIG_FXOKS:
		/* Pick up the line */
#ifdef GRANDSTREAM_NETWORKS
#else
		ast_debug(1, "Took %s off hook\n", ast_channel_name(ast));
#endif
		if (p->hanguponpolarityswitch) {
			gettimeofday(&p->polaritydelaytv, NULL);
		}
		res = analog_off_hook(p);
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_NOTICE, "[ch:%d]Took %s off hook, res:%d.", p->channel, ast_channel_name(ast), res);
#endif
		analog_play_tone(p, idx, -1);
		analog_set_dialing(p, 0);
		if ((idx == ANALOG_SUB_REAL) && p->subs[ANALOG_SUB_THREEWAY].inthreeway) {
			if (oldstate == AST_STATE_RINGING) {
				ast_debug(1, "Finally swapping real and threeway\n");
				analog_play_tone(p, ANALOG_SUB_THREEWAY, -1);
				analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
				analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
			}
		}

		switch (p->sig) {
		case ANALOG_SIG_FXSLS:
		case ANALOG_SIG_FXSKS:
		case ANALOG_SIG_FXSGS:
			analog_set_echocanceller(p, 1);
			analog_train_echocanceller(p);
			break;
		case ANALOG_SIG_FXOLS:
		case ANALOG_SIG_FXOKS:
		case ANALOG_SIG_FXOGS:
			analog_answer_polarityswitch(p);
			break;
		default:
			break;
		}
		break;
	default:
		ast_log(LOG_WARNING, "Don't know how to answer signalling %d (channel %d)\n", p->sig, p->channel);
		res = -1;
		break;
	}
	ast_setstate(ast, AST_STATE_UP);
	return res;
}

static int analog_handles_digit(struct ast_frame *f)
{
	char subclass = toupper(f->subclass.integer);

	switch (subclass) {
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '9':
	case 'A':
	case 'B':
	case 'C':
	case 'D':
	case 'E':
	case 'F':
		return 1;
	default:
		return 0;
	}
}

void analog_handle_dtmf(struct analog_pvt *p, struct ast_channel *ast, enum analog_sub idx, struct ast_frame **dest)
{
	struct ast_frame *f = *dest;
	
#ifdef GRANDSTREAM_NETWORKS
	ast_debug(1, "[ch:%d]%s(%d) DTMF digit: 0x%02X '%c' on %s, _state:%d(%s).",
		p->channel,
		f->frametype == AST_FRAME_DTMF_BEGIN ? "Begin" : "End",
		f->frametype,
		f->subclass.integer, f->subclass.integer, ast_channel_name(ast),
		ast_channel_state(ast), ast_state2str(ast_channel_state(ast)));
	/*Start:modified [Bug 36847] by chzhdeng*/
	/*analog_play_tone(p, idx, -1);*/
	/*End  :modified [Bug 36847] by chzhdeng*/
#else
	ast_debug(1, "%s DTMF digit: 0x%02X '%c' on %s\n",
		f->frametype == AST_FRAME_DTMF_BEGIN ? "Begin" : "End",
		(unsigned)f->subclass.integer, f->subclass.integer, ast_channel_name(ast));
#endif

	if (analog_check_confirmanswer(p)) {
		if (f->frametype == AST_FRAME_DTMF_END) {
			ast_debug(1, "Confirm answer on %s!\n", ast_channel_name(ast));
			/* Upon receiving a DTMF digit, consider this an answer confirmation instead
			of a DTMF digit */
			p->subs[idx].f.frametype = AST_FRAME_CONTROL;
			p->subs[idx].f.subclass.integer = AST_CONTROL_ANSWER;
			/* Reset confirmanswer so DTMF's will behave properly for the duration of the call */
			analog_set_confirmanswer(p, 0);
		} else {
			p->subs[idx].f.frametype = AST_FRAME_NULL;
			p->subs[idx].f.subclass.integer = 0;
		}
		*dest = &p->subs[idx].f;
	} else if (p->callwaitcas) {
		if (f->frametype == AST_FRAME_DTMF_END) {
			if ((f->subclass.integer == 'A') || (f->subclass.integer == 'D')) {
				ast_debug(1, "Got some DTMF, but it's for the CAS\n");
				p->caller.id.name.str = p->callwait_name;
				p->caller.id.number.str = p->callwait_num;
				analog_send_callerid(p, 1, &p->caller);
			}
			if (analog_handles_digit(f)) {
				p->callwaitcas = 0;
			}
		}
		p->subs[idx].f.frametype = AST_FRAME_NULL;
		p->subs[idx].f.subclass.integer = 0;
		*dest = &p->subs[idx].f;
	} else {
		analog_cb_handle_dtmf(p, ast, idx, dest);
	}
}

static int analog_my_getsigstr(struct ast_channel *chan, char *str, const char *term, int ms)
{
	char c;

	*str = 0; /* start with empty output buffer */
	for (;;) {
		/* Wait for the first digit (up to specified ms). */
		c = ast_waitfordigit(chan, ms);
		/* if timeout, hangup or error, return as such */
		if (c < 1) {
			return c;
		}
		*str++ = c;
		*str = 0;
		if (strchr(term, c)) {
			return 1;
		}
	}
}

static int analog_handle_notify_message(struct ast_channel *chan, struct analog_pvt *p, int cid_flags, int neon_mwievent)
{
	if (analog_callbacks.handle_notify_message) {
		analog_callbacks.handle_notify_message(chan, p->chan_pvt, cid_flags, neon_mwievent);
		return 0;
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_handle_notify_message(%p, %d, %d) handle_notify_message is NULL.", p->channel, p, cid_flags, neon_mwievent);
#endif
	return -1;
}

static void analog_increase_ss_count(void)
{
	if (analog_callbacks.increase_ss_count) {
		analog_callbacks.increase_ss_count();
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "analog_increase_ss_count() increase_ss_count is NULL.");
	}
#endif
}

static void analog_decrease_ss_count(void)
{
	if (analog_callbacks.decrease_ss_count) {
		analog_callbacks.decrease_ss_count();
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "analog_decrease_ss_count() decrease_ss_count is NULL.");
	}
#endif
}

static int analog_distinctive_ring(struct ast_channel *chan, struct analog_pvt *p, int idx, int *ringdata)
{
	if (!p->usedistinctiveringdetection) {
		return 0;
	}
	if (analog_callbacks.distinctive_ring) {
		return analog_callbacks.distinctive_ring(chan, p->chan_pvt, idx, ringdata);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "[ch:%d]analog_distinctive_ring(%p, %p, %d, %p) distinctive_ring is NULL.", p->channel, chan, p, idx, ringdata);
#endif
	return -1;

}

static void analog_get_and_handle_alarms(struct analog_pvt *p)
{
	if (analog_callbacks.get_and_handle_alarms) {
		analog_callbacks.get_and_handle_alarms(p->chan_pvt);
	}
#ifdef GRANDSTREAM_NETWORKS
	else
	{
		ast_log(LOG_NOTICE, "[ch:%d]analog_get_and_handle_alarms(%p) get_and_handle_alarms is NULL.", p->channel, p);
	}
#endif
}

static void *analog_get_bridged_channel(struct ast_channel *chan)
{
	if (analog_callbacks.get_sigpvt_bridged_channel) {
		return analog_callbacks.get_sigpvt_bridged_channel(chan);
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "analog_get_bridged_channel(%p) get_sigpvt_bridged_channel is NULL.", chan);
#endif
	return NULL;
}

static int analog_get_sub_fd(struct analog_pvt *p, enum analog_sub sub)
{
	if (analog_callbacks.get_sub_fd) {
		return analog_callbacks.get_sub_fd(p->chan_pvt, sub);
	}
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_NOTICE, "[ch:%d]analog_get_sub_fd(%p,%d) get_sub_fd is NULL.", p->channel, p, sub);
#endif
	return -1;
}

#define ANALOG_NEED_MFDETECT(p) (((p)->sig == ANALOG_SIG_FEATDMF) || ((p)->sig == ANALOG_SIG_FEATDMF_TA) || ((p)->sig == ANALOG_SIG_E911) || ((p)->sig == ANALOG_SIG_FGC_CAMA) || ((p)->sig == ANALOG_SIG_FGC_CAMAMF) || ((p)->sig == ANALOG_SIG_FEATB))

static int analog_canmatch_featurecode(const char *pickupexten, const char *exten)
{
	int extlen = strlen(exten);
	if (!extlen) {
		return 1;
	}
	if (extlen < strlen(pickupexten) && !strncmp(pickupexten, exten, extlen)) {
		return 1;
	}
	/* hardcoded features are *60, *67, *69, *70, *72, *73, *78, *79, *82, *0 */
	if (exten[0] == '*' && extlen < 3) {
		if (extlen == 1) {
			return 1;
		}
		/* "*0" should be processed before it gets here */
		switch (exten[1]) {
		case '6':
		case '7':
		case '8':
			return 1;
		}
	}
	return 0;
}

static void *__analog_ss_thread(void *data)
{
	struct analog_pvt *p = data;
	struct ast_channel *chan = p->ss_astchan;
	char exten[AST_MAX_EXTENSION] = "";
	char exten2[AST_MAX_EXTENSION] = "";
	char dtmfcid[300];
	char dtmfbuf[300];
	char namebuf[ANALOG_MAX_CID];
#ifdef GRANDSTREAM_NETWORKS
	char numbuf[ANALOG_MAX_CID]="Unavailable";
#else
	char numbuf[ANALOG_MAX_CID];
#endif
	struct callerid_state *cs = NULL;
	char *name = NULL, *number = NULL;
	int flags = 0;
	struct ast_smdi_md_message *smdi_msg = NULL;
	int timeout;
	int getforward = 0;
	char *s1, *s2;
	int len = 0;
	int res = 0;
	int idx;
	struct ast_callid *callid;
	RAII_VAR(struct ast_features_pickup_config *, pickup_cfg, NULL, ao2_cleanup);
	const char *pickupexten;
#ifdef GRANDSTREAM_NETWORKS
	char number_default[5]="0000",name_default[11]="Undetected";
    int ringTimeoutBase = 0;
	char *dashptr;
	char device_name[AST_CHANNEL_NAME];
#if defined(SVIP_DETECT_CID) && defined(SVIP_DETECT_DTMF)
    char *dtmf_end_index = NULL;
    int svip_dtmf_flags = -1;
#endif	
	int dw_who_detect_callerid = 0;
	int dwwhodetectcallerID=0;
	int dwhas_start_fsk_detect=0;
	int dwFirstDigitC=0;
	int dwHasRecvAorD=0;
#endif

	analog_increase_ss_count();

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "[ch:%d]__analog_ss_thread(0x%08tx) sigtype:%u(%s) use_cid:%u ...",
		p->channel, (ptrdiff_t)data, p->sig, analog_sigtype_to_str(p->sig), p->use_callerid);
#else
	ast_debug(1, "%s %d\n", __FUNCTION__, p->channel);
#endif

	if (!chan) {
		/* What happened to the channel? */
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_WARNING, "[ch:%d] analog_pvt->ss_astchan is NULL !", p->channel);
#endif
		goto quit;
	}

	if ((callid = ast_channel_callid(chan))) {
		ast_callid_threadassoc_add(callid);
		ast_callid_unref(callid);
	}

	/* in the bizarre case where the channel has become a zombie before we
	   even get started here, abort safely
	*/
	if (!ast_channel_tech_pvt(chan)) {
#ifdef GRANDSTREAM_NETWORKS 
		ast_log(LOG_WARNING, "Channel(%d) became a zombie before simple switch could be started (%s) state:%u(%s)",
			p->channel, ast_channel_name(chan), ast_channel_state(chan), ast_state2str(ast_channel_state(chan)));
		//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
#else
		ast_log(LOG_WARNING, "Channel became a zombie before simple switch could be started (%s)\n", ast_channel_name(chan));
#endif
		ast_hangup(chan);
		goto quit;
	}
	
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "[ch:%d]Channel starting simple switch on '%s'\n", p->channel, ast_channel_name(chan));	
#else
	ast_verb(3, "Starting simple switch on '%s'\n", ast_channel_name(chan));
#endif
	idx = analog_get_index(chan, p, 0);
	if (idx < 0) {
#ifdef GRANDSTREAM_NETWORKS
		//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
		ast_hangup(chan);
		ast_log(LOG_WARNING, "Channel(%d) analog_get_index() %d failed!", p->channel, idx);
#else
		ast_hangup(chan);
#endif
		goto quit;
	}

	ast_channel_lock(chan);
	pickup_cfg = ast_get_chan_features_pickup_config(chan);
	if (!pickup_cfg) {
		ast_log(LOG_ERROR, "Unable to retrieve pickup configuration options. Unable to detect call pickup extension\n");
		pickupexten = "";
	} else {
		pickupexten = ast_strdupa(pickup_cfg->pickupexten);
	}
	ast_channel_unlock(chan);

	analog_dsp_reset_and_flush_digits(p);
#ifdef GRANDSTREAM_NETWORKS
	ast_copy_string(device_name, ast_channel_name(chan), sizeof(device_name));
	dashptr = strrchr(device_name, '-');
	if (dashptr) {
		*dashptr = '\0';
	}

	ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, AST_DEVSTATE_NOT_CACHABLE, device_name);
#endif
	switch (p->sig) {
	case ANALOG_SIG_FEATD:
	case ANALOG_SIG_FEATDMF:
	case ANALOG_SIG_FEATDMF_TA:
	case ANALOG_SIG_E911:
	case ANALOG_SIG_FGC_CAMAMF:
	case ANALOG_SIG_FEATB:
	case ANALOG_SIG_EMWINK:
	case ANALOG_SIG_SF_FEATD:
	case ANALOG_SIG_SF_FEATDMF:
	case ANALOG_SIG_SF_FEATB:
	case ANALOG_SIG_SFWINK:
		if (analog_wink(p, idx))
			goto quit;
		/* Fall through */
	case ANALOG_SIG_EM:
	case ANALOG_SIG_EM_E1:
	case ANALOG_SIG_SF:
	case ANALOG_SIG_FGC_CAMA:
#ifdef GRANDSTREAM_NETWORKS
		if(p->sig != ANALOG_SIG_EM)
		{
			res = analog_play_tone(p, idx, -1);
		}
#else
		res = analog_play_tone(p, idx, -1);
#endif

		analog_dsp_reset_and_flush_digits(p);

		/* set digit mode appropriately */
		if (ANALOG_NEED_MFDETECT(p)) {
			analog_dsp_set_digitmode(p, ANALOG_DIGITMODE_MF);
		} else {
			analog_dsp_set_digitmode(p, ANALOG_DIGITMODE_DTMF);
		}

		memset(dtmfbuf, 0, sizeof(dtmfbuf));
		/* Wait for the first digit only if immediate=no */
		if (!p->immediate) {
			/* Wait for the first digit (up to 5 seconds). */
			res = ast_waitfordigit(chan, 5000);
		} else {
			res = 0;
		}
		if (res > 0) {
			/* save first char */
			dtmfbuf[0] = res;
			switch (p->sig) {
			case ANALOG_SIG_FEATD:
			case ANALOG_SIG_SF_FEATD:
				res = analog_my_getsigstr(chan, dtmfbuf + 1, "*", 3000);
				if (res > 0) {
					res = analog_my_getsigstr(chan, dtmfbuf + strlen(dtmfbuf), "*", 3000);
				}
				if (res < 1) {
					analog_dsp_reset_and_flush_digits(p);
				}
				break;
			case ANALOG_SIG_FEATDMF_TA:
				res = analog_my_getsigstr(chan, dtmfbuf + 1, "#", 3000);
				if (res < 1) {
					analog_dsp_reset_and_flush_digits(p);
				}
				if (analog_wink(p, idx)) {
					goto quit;
				}
				dtmfbuf[0] = 0;
				/* Wait for the first digit (up to 5 seconds). */
				res = ast_waitfordigit(chan, 5000);
				if (res <= 0) {
					break;
				}
				dtmfbuf[0] = res;
				/* fall through intentionally */
			case ANALOG_SIG_FEATDMF:
			case ANALOG_SIG_E911:
			case ANALOG_SIG_FGC_CAMAMF:
			case ANALOG_SIG_SF_FEATDMF:
				res = analog_my_getsigstr(chan, dtmfbuf + 1, "#", 3000);
				/* if international caca, do it again to get real ANO */
				if ((p->sig == ANALOG_SIG_FEATDMF) && (dtmfbuf[1] != '0')
					&& (strlen(dtmfbuf) != 14)) {
					if (analog_wink(p, idx)) {
						goto quit;
					}
					dtmfbuf[0] = 0;
					/* Wait for the first digit (up to 5 seconds). */
					res = ast_waitfordigit(chan, 5000);
					if (res <= 0) {
						break;
					}
					dtmfbuf[0] = res;
					res = analog_my_getsigstr(chan, dtmfbuf + 1, "#", 3000);
				}
				if (res > 0) {
					/* if E911, take off hook */
					if (p->sig == ANALOG_SIG_E911) {
						analog_off_hook(p);
					}
					res = analog_my_getsigstr(chan, dtmfbuf + strlen(dtmfbuf), "#", 3000);
				}
				if (res < 1) {
					analog_dsp_reset_and_flush_digits(p);
				}
				break;
			case ANALOG_SIG_FEATB:
			case ANALOG_SIG_SF_FEATB:
				res = analog_my_getsigstr(chan, dtmfbuf + 1, "#", 3000);
				if (res < 1) {
					analog_dsp_reset_and_flush_digits(p);
				}
				break;
			case ANALOG_SIG_EMWINK:
				/* if we received a '*', we are actually receiving Feature Group D
				   dial syntax, so use that mode; otherwise, fall through to normal
				   mode
				*/
				if (res == '*') {
					res = analog_my_getsigstr(chan, dtmfbuf + 1, "*", 3000);
					if (res > 0) {
						res = analog_my_getsigstr(chan, dtmfbuf + strlen(dtmfbuf), "*", 3000);
					}
					if (res < 1) {
						analog_dsp_reset_and_flush_digits(p);
					}
					break;
				}
			default:
				/* If we got the first digit, get the rest */
				len = 1;
				dtmfbuf[len] = '\0';
#ifdef GRANDSTREAM_NETWORKS
				//added by zhywang for bug 50853
				char* callerid = p->cid_num;
				if( (p->sig == ANALOG_SIG_EMWINK) || (p->sig == ANALOG_SIG_EM)){
					callerid = ast_channel_caller(chan)->id.number.valid ? ast_channel_caller(chan)->id.number.str : NULL;//anonymous
				}
				while ((len < AST_MAX_EXTENSION-1) && ast_matchmore_extension(chan, ast_channel_context(chan), dtmfbuf, 1, callerid)) {
					if (ast_exists_extension(chan, ast_channel_context(chan), dtmfbuf, 1,callerid)) {
#else
				while ((len < AST_MAX_EXTENSION-1) && ast_matchmore_extension(chan, ast_channel_context(chan), dtmfbuf, 1, p->cid_num)) {
					if (ast_exists_extension(chan, ast_channel_context(chan), dtmfbuf, 1, p->cid_num)) {
#endif
						timeout = analog_matchdigittimeout;
					} else {
						timeout = analog_gendigittimeout;
					}
					res = ast_waitfordigit(chan, timeout);
					if (res < 0) {
						ast_debug(1, "waitfordigit returned < 0...\n");
						ast_hangup(chan);
						goto quit;
					} else if (res) {
						dtmfbuf[len++] = res;
						dtmfbuf[len] = '\0';
					} else {
						break;
					}
				}
				break;
			}
		}
		if (res == -1) {
			ast_log(LOG_WARNING, "getdtmf on channel %d: %s\n", p->channel, strerror(errno));
			ast_hangup(chan);
			goto quit;
		} else if (res < 0) {
			ast_debug(1, "Got hung up before digits finished\n");
			ast_hangup(chan);
			goto quit;
		}

		if (p->sig == ANALOG_SIG_FGC_CAMA) {
			char anibuf[100];

			if (ast_safe_sleep(chan,1000) == -1) {
				ast_hangup(chan);
				goto quit;
			}
			analog_off_hook(p);
			analog_dsp_set_digitmode(p, ANALOG_DIGITMODE_MF);
			res = analog_my_getsigstr(chan, anibuf, "#", 10000);
			if ((res > 0) && (strlen(anibuf) > 2)) {
				if (anibuf[strlen(anibuf) - 1] == '#') {
					anibuf[strlen(anibuf) - 1] = 0;
				}
				ast_set_callerid(chan, anibuf + 2, NULL, anibuf + 2);
			}
			analog_dsp_set_digitmode(p, ANALOG_DIGITMODE_DTMF);
		}

		ast_copy_string(exten, dtmfbuf, sizeof(exten));
		if (ast_strlen_zero(exten)) {
			ast_copy_string(exten, "s", sizeof(exten));
		}
		if (p->sig == ANALOG_SIG_FEATD || p->sig == ANALOG_SIG_EMWINK) {
			/* Look for Feature Group D on all E&M Wink and Feature Group D trunks */
			if (exten[0] == '*') {
				char *stringp=NULL;
				ast_copy_string(exten2, exten, sizeof(exten2));
				/* Parse out extension and callerid */
				stringp=exten2 +1;
				s1 = strsep(&stringp, "*");
				s2 = strsep(&stringp, "*");
				if (s2) {
					if (!ast_strlen_zero(p->cid_num)) {
						ast_set_callerid(chan, p->cid_num, NULL, p->cid_num);
					} else {
						ast_set_callerid(chan, s1, NULL, s1);
					}
					ast_copy_string(exten, s2, sizeof(exten));
				} else {
					ast_copy_string(exten, s1, sizeof(exten));
				}
			} else if (p->sig == ANALOG_SIG_FEATD) {
				ast_log(LOG_WARNING, "Got a non-Feature Group D input on channel %d.  Assuming E&M Wink instead\n", p->channel);
			}
		}
		if ((p->sig == ANALOG_SIG_FEATDMF) || (p->sig == ANALOG_SIG_FEATDMF_TA)) {
			if (exten[0] == '*') {
				char *stringp=NULL;
				struct ast_party_caller *caller;

				ast_copy_string(exten2, exten, sizeof(exten2));
				/* Parse out extension and callerid */
				stringp=exten2 +1;
				s1 = strsep(&stringp, "#");
				s2 = strsep(&stringp, "#");
				if (s2) {
					if (!ast_strlen_zero(p->cid_num)) {
						ast_set_callerid(chan, p->cid_num, NULL, p->cid_num);
					} else {
						if (*(s1 + 2)) {
							ast_set_callerid(chan, s1 + 2, NULL, s1 + 2);
						}
					}
					ast_copy_string(exten, s2 + 1, sizeof(exten));
				} else {
					ast_copy_string(exten, s1 + 2, sizeof(exten));
				}

				/* The first two digits are ani2 information. */
				caller = ast_channel_caller(chan);
				s1[2] = '\0';
				caller->ani2 = atoi(s1);
			} else {
				ast_log(LOG_WARNING, "Got a non-Feature Group D input on channel %d.  Assuming E&M Wink instead\n", p->channel);
			}
		}
		if ((p->sig == ANALOG_SIG_E911) || (p->sig == ANALOG_SIG_FGC_CAMAMF)) {
			if (exten[0] == '*') {
				char *stringp=NULL;
				ast_copy_string(exten2, exten, sizeof(exten2));
				/* Parse out extension and callerid */
				stringp=exten2 +1;
				s1 = strsep(&stringp, "#");
				s2 = strsep(&stringp, "#");
				if (s2 && (*(s2 + 1) == '0')) {
					if (*(s2 + 2)) {
						ast_set_callerid(chan, s2 + 2, NULL, s2 + 2);
					}
				}
				if (s1) {
					ast_copy_string(exten, s1, sizeof(exten));
				} else {
					ast_copy_string(exten, "911", sizeof(exten));
				}
			} else {
				ast_log(LOG_WARNING, "Got a non-E911/FGC CAMA input on channel %d.  Assuming E&M Wink instead\n", p->channel);
			}
		}
		if (p->sig == ANALOG_SIG_FEATB) {
			if (exten[0] == '*') {
				char *stringp=NULL;
				ast_copy_string(exten2, exten, sizeof(exten2));
				/* Parse out extension and callerid */
				stringp=exten2 +1;
				s1 = strsep(&stringp, "#");
				ast_copy_string(exten, exten2 + 1, sizeof(exten));
			} else {
				ast_log(LOG_WARNING, "Got a non-Feature Group B input on channel %d.  Assuming E&M Wink instead\n", p->channel);
			}
		}
		if ((p->sig == ANALOG_SIG_FEATDMF) || (p->sig == ANALOG_SIG_FEATDMF_TA)) {
			analog_wink(p, idx);
			/*
			 * Some switches require a minimum guard time between the last
			 * FGD wink and something that answers immediately.  This
			 * ensures it.
			 */
			if (ast_safe_sleep(chan, 100)) {
				ast_hangup(chan);
				goto quit;
			}
		}
		analog_set_echocanceller(p, 1);

		analog_dsp_set_digitmode(p, ANALOG_DIGITMODE_DTMF);

		if (ast_exists_extension(chan, ast_channel_context(chan), exten, 1,
			ast_channel_caller(chan)->id.number.valid ? ast_channel_caller(chan)->id.number.str : NULL)) {
			ast_channel_exten_set(chan, exten);
			analog_dsp_reset_and_flush_digits(p);
			res = ast_pbx_run(chan);
			if (res) {
				ast_log(LOG_WARNING, "PBX exited non-zero\n");
				res = analog_play_tone(p, idx, ANALOG_TONE_CONGESTION);
			}
			goto quit;
		} else {
			ast_verb(3, "Unknown extension '%s' in context '%s' requested\n", exten, ast_channel_context(chan));
			sleep(2);
			res = analog_play_tone(p, idx, ANALOG_TONE_INFO);
			if (res < 0) {
				ast_log(LOG_WARNING, "Unable to start special tone on %d\n", p->channel);
			} else {
				sleep(1);
			}
			res = ast_streamfile(chan, "ss-noservice", ast_channel_language(chan));
			if (res >= 0) {
				ast_waitstream(chan, "");
			}
			res = analog_play_tone(p, idx, ANALOG_TONE_CONGESTION);
			ast_hangup(chan);
			goto quit;
		}
		break;
	case ANALOG_SIG_FXOLS:
	case ANALOG_SIG_FXOGS:
	case ANALOG_SIG_FXOKS:
		/* Read the first digit */
		timeout = analog_firstdigittimeout;
#ifdef GRANDSTREAM_NETWORKS
		/*Start: chzhdeng  [Bug 37676 ]  add */
		pbx_builtin_setvar_helper(chan, "CALLED_REALHOOK", "no");
		/*End: chzhdeng  [Bug 37676 ]  add */
#endif
		/* If starting a threeway call, never timeout on the first digit so someone
		   can use flash-hook as a "hold" feature */
		if (p->subs[ANALOG_SUB_THREEWAY].owner) {
			timeout = 999999;
		}
		while (len < AST_MAX_EXTENSION-1) {
			int is_exten_parking = 0;

			/* Read digit unless it's supposed to be immediate, in which case the
			   only answer is 's' */
			if (p->immediate) {
				res = 's';
#ifdef GRANDSTREAM_NETWORKS
			}else if(p->hotline_delay && p->hotline_exten){
				if(len == 0){
					ast_log(LOG_VERBOSE, " Delay Hotline ! exten = %s \n",p->hotline_exten);
					res = ast_waitfordigit(chan, 5000);
					if(res == 0){
						snprintf(exten,sizeof(exten),"%s",p->hotline_exten);
						len = strlen(exten);
					}
				}else{
					res = ast_waitfordigit(chan, timeout);
				}
#endif
			} else {
				res = ast_waitfordigit(chan, timeout);
			}
#ifdef GRANDSTREAM_NETWORKS
			featurecode:
				ast_debug(1,"dial digital key :res = '%c' \n",res);
#endif
			timeout = 0;
			if (res < 0) {
				ast_debug(1, "waitfordigit returned < 0...\n");
				res = analog_play_tone(p, idx, -1);
#ifdef GRANDSTREAM_NETWORKS
				//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
#endif
				ast_hangup(chan);
				goto quit;
			}
#ifdef GRANDSTREAM_NETWORKS
            else if (p->sharp_is_sendkey && ('#' == res) && len >= 1)
            {
                res = 0;
            }
#endif
			else if (res) 
			{
				ast_debug(1,"waitfordigit returned '%c' (%d), timeout = %d\n", res, res, timeout);
				exten[len++]=res;
				exten[len] = '\0';
			}
			if (!ast_ignore_pattern(ast_channel_context(chan), exten)) {
				analog_play_tone(p, idx, -1);
			} else {
				analog_play_tone(p, idx, ANALOG_TONE_DIALTONE);
			}
			if (ast_parking_provider_registered()) {
				is_exten_parking = ast_parking_is_exten_park(ast_channel_context(chan), exten);
			}
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_NOTICE, "exten = %s, pickupexten = %s\n", exten, pickupexten);
			if (!strncmp(exten, pickupexten, strlen(exten)) && strlen(exten) < strlen(pickupexten) &&
					pickupexten[0] >= '0' && pickupexten[0] <= '9') {
				;/* Bug 56884 - pickup group:when change feature cade to digital,FXS extension pickup is failure. */
			}
			else if (!strcmp(exten, pickupexten)) /*modified by jklou for Bug 41569 - FXS failed to use *8 to pick up*/
			{//when dialed number not exist in dialplan,recv '#' or timeout
				/* Scan all channels and see if there are any
				* ringing channels that have call groups
				* that equal this channels pickup group*/
				if (idx == ANALOG_SUB_REAL) 
				{
					/* Switch us from Third call to Call Wait */
					if (p->subs[ANALOG_SUB_THREEWAY].owner) 
					{
						/* If you make a threeway call and the *8# a call, it should actually
						   look like a callwait */
						analog_alloc_sub(p, ANALOG_SUB_CALLWAIT);
						analog_swap_subs(p, ANALOG_SUB_CALLWAIT, ANALOG_SUB_THREEWAY);
						analog_unalloc_sub(p, ANALOG_SUB_THREEWAY);
					}
					analog_set_echocanceller(p, 1);
					if (ast_pickup_call(chan)) 
					{
						ast_debug(1, "No call pickup possible...\n");
						res = analog_play_tone(p, idx, ANALOG_TONE_CONGESTION);
						/* Set refresh device state */
						ast_copy_string(device_name, ast_channel_name(chan), sizeof(device_name));
						dashptr = strrchr(device_name, '-');
						if (dashptr) 
						{
							*dashptr = '\0';
						}

						ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, AST_DEVSTATE_NOT_CACHABLE, device_name);
						analog_wait_event(p);
						//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
					}
					ast_hangup(chan);
					goto quit;
				}
				else
				{
					ast_log(LOG_WARNING, "Huh?	Got *8# on call not on real\n");
					ast_hangup(chan);
					goto quit;
				}
			}else if (ast_exists_extension(chan, ast_channel_context(chan), exten, 1, p->cid_num)) 
#else
			if (ast_exists_extension(chan, ast_channel_context(chan), exten, 1, p->cid_num) && !is_exten_parking) 
#endif
			{
				if (!res || !ast_matchmore_extension(chan, ast_channel_context(chan), exten, 1, p->cid_num)) {
					if (getforward) {
						/* Record this as the forwarding extension */
						ast_copy_string(p->call_forward, exten, sizeof(p->call_forward));
						ast_verb(3, "Setting call forward to '%s' on channel %d\n", p->call_forward, p->channel);
						res = analog_play_tone(p, idx, ANALOG_TONE_DIALRECALL);
						if (res) {
							break;
						}
						usleep(500000);
						res = analog_play_tone(p, idx, -1);
						sleep(1);
						memset(exten, 0, sizeof(exten));
						res = analog_play_tone(p, idx, ANALOG_TONE_DIALTONE);
						len = 0;
						getforward = 0;
					} 
#ifdef GRANDSTREAM_NETWORKS
					else if (0)
					{
					   //The dialed number exist in dialplan and not can match, but it equals pick up exten 
                        /* Scan all channels and see if there are any
		                        * ringing channels that have call groups
		                        * that equal this channels pickup group */
						if (idx == ANALOG_SUB_REAL) 
						{
							/* Switch us from Third call to Call Wait */
							if (p->subs[ANALOG_SUB_THREEWAY].owner) 
							{
								/* If you make a threeway call and the *8# a call, it should actually
								   look like a callwait */
								analog_alloc_sub(p, ANALOG_SUB_CALLWAIT);
								analog_swap_subs(p, ANALOG_SUB_CALLWAIT, ANALOG_SUB_THREEWAY);
								analog_unalloc_sub(p, ANALOG_SUB_THREEWAY);
							}
							analog_set_echocanceller(p, 1);
							if (ast_pickup_call(chan)) 
							{
								ast_debug(1, "No call pickup possible...\n");
								res = analog_play_tone(p, idx, ANALOG_TONE_CONGESTION);
						        /* Set refresh device state */
						        ast_copy_string(device_name, ast_channel_name(chan), sizeof(device_name));
						        dashptr = strrchr(device_name, '-');
						        if (dashptr) 
								{
						            *dashptr = '\0';
						        }

						        ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, AST_DEVSTATE_NOT_CACHABLE, device_name);
								analog_wait_event(p);
								//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
							}
							ast_hangup(chan);
							goto quit;
						}
						else 
						{
							ast_log(LOG_WARNING, "Huh?	Got *8# on call not on real\n");
							ast_hangup(chan);
							goto quit;
						}
					}
#endif
					else 
					{
						res = analog_play_tone(p, idx, -1);
						ast_channel_lock(chan);
						ast_channel_exten_set(chan, exten);
#ifdef GRANDSTREAM_NETWORKS
						ast_channel_dialed(chan)->number.str = ast_strdup(exten);
#endif
						if (!ast_strlen_zero(p->cid_num)) {
							if (!p->hidecallerid) {
								ast_set_callerid(chan, p->cid_num, NULL, p->cid_num);
							} else {
								ast_set_callerid(chan, NULL, NULL, p->cid_num);
							}
						}
						if (!ast_strlen_zero(p->cid_name)) {
							if (!p->hidecallerid) {
								ast_set_callerid(chan, NULL, p->cid_name, NULL);
							}
						}
						ast_setstate(chan, AST_STATE_RING);
						ast_channel_unlock(chan);
						analog_set_echocanceller(p, 1);
						res = ast_pbx_run(chan);
						if (res) {
							ast_log(LOG_WARNING, "PBX exited non-zero\n");
							res = analog_play_tone(p, idx, ANALOG_TONE_CONGESTION);
						}
						goto quit;
					}
				} else {
					/* It's a match, but they just typed a digit, and there is an ambiguous match,
					   so just set the timeout to analog_matchdigittimeout and wait some more */
					timeout = analog_matchdigittimeout;
				}
			} else if (res == 0) {
#ifdef GRANDSTREAM_NETWORKS
				//if (0)
				#if 0
				if (!strcmp(exten, pickupexten)) /*modified by jklou for Bug 41569 - FXS failed to use *8 to pick up*/
				{//when dialed number not exist in dialplan,recv '#' or timeout
					/* Scan all channels and see if there are any
					* ringing channels that have call groups
					* that equal this channels pickup group*/
					if (idx == ANALOG_SUB_REAL) 
					{
						/* Switch us from Third call to Call Wait */
						if (p->subs[ANALOG_SUB_THREEWAY].owner) 
						{
							/* If you make a threeway call and the *8# a call, it should actually
							   look like a callwait */
							analog_alloc_sub(p, ANALOG_SUB_CALLWAIT);
							analog_swap_subs(p, ANALOG_SUB_CALLWAIT, ANALOG_SUB_THREEWAY);
							analog_unalloc_sub(p, ANALOG_SUB_THREEWAY);
						}
						analog_set_echocanceller(p, 1);
						if (ast_pickup_call(chan)) 
						{
							ast_debug(1, "No call pickup possible...\n");
							res = analog_play_tone(p, idx, ANALOG_TONE_CONGESTION);
							/* Set refresh device state */
							ast_copy_string(device_name, ast_channel_name(chan), sizeof(device_name));
							dashptr = strrchr(device_name, '-');
							if (dashptr) 
							{
								*dashptr = '\0';
							}

							ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, AST_DEVSTATE_NOT_CACHABLE, device_name);
							analog_wait_event(p);
							//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
						}
						ast_hangup(chan);
						goto quit;
					}
					else
					{
						ast_log(LOG_WARNING, "Huh?	Got *8# on call not on real\n");
						ast_hangup(chan);
						goto quit;
					}
				}
				#endif
				//else 
				//{			
					ast_debug(1, "not enough digits (and no ambiguous match)...\n");
					res = analog_play_tone(p, idx, ANALOG_TONE_CONGESTION);
					/* Set refresh device state */
					ast_copy_string(device_name, ast_channel_name(chan), sizeof(device_name));
					dashptr = strrchr(device_name, '-');
					if (dashptr) 
					{
						*dashptr = '\0';
					}

					ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, AST_DEVSTATE_NOT_CACHABLE, device_name);
					analog_wait_event(p);
					//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
					ast_hangup(chan);
					goto quit;
				//}
#else
				ast_debug(1, "not enough digits (and no ambiguous match)...\n");
				res = analog_play_tone(p, idx, ANALOG_TONE_CONGESTION);
				analog_wait_event(p);
				ast_hangup(chan);
				goto quit;
#endif
			} else if (p->callwaiting && !strcmp(exten, "*70")) {
				ast_verb(3, "Disabling call waiting on %s\n", ast_channel_name(chan));
				/* Disable call waiting if enabled */
				analog_set_callwaiting(p, 0);
				res = analog_play_tone(p, idx, ANALOG_TONE_DIALRECALL);
				if (res) {
					ast_log(LOG_WARNING, "Unable to do dial recall on channel %s: %s\n",
						ast_channel_name(chan), strerror(errno));
				}
				len = 0;
				memset(exten, 0, sizeof(exten));
				timeout = analog_firstdigittimeout;

			}
#ifdef GRANDSTREAM_NETWORKS
			else if (!strcmp(exten, pickupexten) && !ast_matchmore_extension(chan, ast_channel_context(chan), exten, 1, p->cid_num))
#else
			else if (!strcmp(exten, pickupexten)) 
#endif
			{
				/* Scan all channels and see if there are any
				 * ringing channels that have call groups
				 * that equal this channels pickup group
				 */
				if (idx == ANALOG_SUB_REAL) {
					/* Switch us from Third call to Call Wait */
					if (p->subs[ANALOG_SUB_THREEWAY].owner) {
						/* If you make a threeway call and the *8# a call, it should actually
						   look like a callwait */
						analog_alloc_sub(p, ANALOG_SUB_CALLWAIT);
						analog_swap_subs(p, ANALOG_SUB_CALLWAIT, ANALOG_SUB_THREEWAY);
						analog_unalloc_sub(p, ANALOG_SUB_THREEWAY);
					}
					analog_set_echocanceller(p, 1);
					if (ast_pickup_call(chan)) {
						ast_debug(1, "No call pickup possible...\n");
						res = analog_play_tone(p, idx, ANALOG_TONE_CONGESTION);
#ifdef GRANDSTREAM_NETWORKS
						ast_copy_string(device_name, ast_channel_name(chan), sizeof(device_name));
						dashptr = strrchr(device_name, '-');
						if (dashptr) 
						{
							*dashptr = '\0';
						}
                        ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, AST_DEVSTATE_NOT_CACHABLE, device_name);
#endif
						analog_wait_event(p);
#ifdef GRANDSTREAM_NETWORKS
						//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
#endif
					}
					ast_hangup(chan);
					goto quit;
				} else {
					ast_log(LOG_WARNING, "Huh?  Got *8# on call not on real\n");
					ast_hangup(chan);
					goto quit;
				}

			} else if (!p->hidecallerid && !strcmp(exten, "*67")) {
				ast_verb(3, "Disabling Caller*ID on %s\n", ast_channel_name(chan));
				/* Disable Caller*ID if enabled */
				p->hidecallerid = 1;
				ast_party_number_free(&ast_channel_caller(chan)->id.number);
				ast_party_number_init(&ast_channel_caller(chan)->id.number);
				ast_party_name_free(&ast_channel_caller(chan)->id.name);
				ast_party_name_init(&ast_channel_caller(chan)->id.name);
				res = analog_play_tone(p, idx, ANALOG_TONE_DIALRECALL);
				if (res) {
					ast_log(LOG_WARNING, "Unable to do dial recall on channel %s: %s\n",
						ast_channel_name(chan), strerror(errno));
				}
				len = 0;
				memset(exten, 0, sizeof(exten));
				timeout = analog_firstdigittimeout;
			} else if (p->callreturn && !strcmp(exten, "*69")) {
				res = 0;
				if (!ast_strlen_zero(p->lastcid_num)) {
					res = ast_say_digit_str(chan, p->lastcid_num, "", ast_channel_language(chan));
				}
				if (!res) {
					res = analog_play_tone(p, idx, ANALOG_TONE_DIALRECALL);
				}
				break;
			} else if (!strcmp(exten, "*78")) {
				/* Do not disturb enabled */
				analog_dnd(p, 1);
				res = analog_play_tone(p, idx, ANALOG_TONE_DIALRECALL);
				getforward = 0;
				memset(exten, 0, sizeof(exten));
				len = 0;
			} else if (!strcmp(exten, "*79")) {
				/* Do not disturb disabled */
				analog_dnd(p, 0);
				res = analog_play_tone(p, idx, ANALOG_TONE_DIALRECALL);
				getforward = 0;
				memset(exten, 0, sizeof(exten));
				len = 0;
			} else if (p->cancallforward && !strcmp(exten, "*72")) {
				res = analog_play_tone(p, idx, ANALOG_TONE_DIALRECALL);
				getforward = 1;
				memset(exten, 0, sizeof(exten));
				len = 0;
			} else if (p->cancallforward && !strcmp(exten, "*73")) {
				ast_verb(3, "Cancelling call forwarding on channel %d\n", p->channel);
				res = analog_play_tone(p, idx, ANALOG_TONE_DIALRECALL);
				memset(p->call_forward, 0, sizeof(p->call_forward));
				getforward = 0;
				memset(exten, 0, sizeof(exten));
				len = 0;
			} else if ((p->transfer || p->canpark) && is_exten_parking
				&& p->subs[ANALOG_SUB_THREEWAY].owner) {
				struct ast_bridge_channel *bridge_channel;

				/*
				 * This is a three way call, the main call being a real channel,
				 * and we're parking the first call.
				 */
				ast_channel_lock(p->subs[ANALOG_SUB_THREEWAY].owner);
				bridge_channel = ast_channel_get_bridge_channel(p->subs[ANALOG_SUB_THREEWAY].owner);
				ast_channel_unlock(p->subs[ANALOG_SUB_THREEWAY].owner);
				if (bridge_channel) {
					if (!ast_parking_blind_transfer_park(bridge_channel, ast_channel_context(chan), exten, NULL, NULL)) {
						/*
						 * Swap things around between the three-way and real call so we
						 * can hear where the channel got parked.
						 */
						analog_lock_private(p);
						analog_set_new_owner(p, p->subs[ANALOG_SUB_THREEWAY].owner);
						analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
						analog_unlock_private(p);

						ast_verb(3, "%s: Parked call\n", ast_channel_name(chan));
						ast_hangup(chan);
						ao2_ref(bridge_channel, -1);
						goto quit;
					}
					ao2_ref(bridge_channel, -1);
				}
				break;
			} else if (!ast_strlen_zero(p->lastcid_num) && !strcmp(exten, "*60")) {
				ast_verb(3, "Blacklisting number %s\n", p->lastcid_num);
				res = ast_db_put("blacklist", p->lastcid_num, "1");
				if (!res) {
					res = analog_play_tone(p, idx, ANALOG_TONE_DIALRECALL);
					memset(exten, 0, sizeof(exten));
					len = 0;
				}
			} else if (p->hidecallerid && !strcmp(exten, "*82")) {
				ast_verb(3, "Enabling Caller*ID on %s\n", ast_channel_name(chan));
				/* Enable Caller*ID if enabled */
				p->hidecallerid = 0;
				ast_set_callerid(chan, p->cid_num, p->cid_name, NULL);
				res = analog_play_tone(p, idx, ANALOG_TONE_DIALRECALL);
				if (res) {
					ast_log(LOG_WARNING, "Unable to do dial recall on channel %s: %s\n",
						ast_channel_name(chan), strerror(errno));
				}
				len = 0;
				memset(exten, 0, sizeof(exten));
				timeout = analog_firstdigittimeout;
			} else if (!strcmp(exten, "*0")) {
				struct ast_channel *nbridge = p->subs[ANALOG_SUB_THREEWAY].owner;
				struct analog_pvt *pbridge = NULL;
#ifdef GRANDSTREAM_NETWORKS
				res = ast_waitfordigit(chan, 8000);
				//exten=="*0" or exten=="*0#"
				if(res == 0 || res == 35)
				{
					ast_debug(1,"execute '*0' feature code ! \n");
#endif
				/* set up the private struct of the bridged one, if any */
				if (nbridge) {
					pbridge = analog_get_bridged_channel(nbridge);
				}
				if (pbridge && ISTRUNK(pbridge)) {
					/* Clear out the dial buffer */
					p->dop.dialstr[0] = '\0';
					/* flash hookswitch */
					if ((analog_flash(pbridge) == -1) && (errno != EINPROGRESS)) {
						ast_log(LOG_WARNING,
							"Unable to flash-hook bridged trunk from channel %s: %s\n",
							ast_channel_name(nbridge), strerror(errno));
					}
					analog_swap_subs(p, ANALOG_SUB_REAL, ANALOG_SUB_THREEWAY);
					analog_unalloc_sub(p, ANALOG_SUB_THREEWAY);
					analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
					ast_queue_unhold(p->subs[ANALOG_SUB_REAL].owner);
					ast_hangup(chan);
					goto quit;
				} else {
					analog_play_tone(p, idx, ANALOG_TONE_CONGESTION);
					analog_wait_event(p);
					analog_play_tone(p, idx, -1);
					analog_swap_subs(p, ANALOG_SUB_REAL, ANALOG_SUB_THREEWAY);
					analog_unalloc_sub(p, ANALOG_SUB_THREEWAY);
					analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
					ast_hangup(chan);
					goto quit;
				}
#ifdef GRANDSTREAM_NETWORKS
				}
				else 
				{
					ast_debug(1,"execute:goto featurecode !\n");
					goto featurecode ;
				}
#endif
			} else if (!ast_canmatch_extension(chan, ast_channel_context(chan), exten, 1,
				ast_channel_caller(chan)->id.number.valid ? ast_channel_caller(chan)->id.number.str : NULL)
				&& !analog_canmatch_featurecode(pickupexten, exten)) {
				ast_debug(1, "Can't match %s from '%s' in context %s\n", exten,
					ast_channel_caller(chan)->id.number.valid && ast_channel_caller(chan)->id.number.str
						? ast_channel_caller(chan)->id.number.str : "<Unknown Caller>",
					ast_channel_context(chan));
				break;
			}
			if (!timeout) {
				timeout = analog_gendigittimeout;
			}
			if (len && !ast_ignore_pattern(ast_channel_context(chan), exten)) {
				analog_play_tone(p, idx, -1);
			}
		}
		break;
	case ANALOG_SIG_FXSLS:
	case ANALOG_SIG_FXSGS:
	case ANALOG_SIG_FXSKS:
		/* check for SMDI messages */
		if (p->use_smdi && p->smdi_iface) {
			smdi_msg = ast_smdi_md_message_wait(p->smdi_iface, ANALOG_SMDI_MD_WAIT_TIMEOUT);
			if (smdi_msg != NULL) {
				ast_channel_exten_set(chan, smdi_msg->fwd_st);

				if (smdi_msg->type == 'B')
					pbx_builtin_setvar_helper(chan, "_SMDI_VM_TYPE", "b");
				else if (smdi_msg->type == 'N')
					pbx_builtin_setvar_helper(chan, "_SMDI_VM_TYPE", "u");

				ast_debug(1, "Received SMDI message on %s\n", ast_channel_name(chan));
			} else {
				ast_log(LOG_WARNING, "SMDI enabled but no SMDI message present\n");
			}
		}
		
#ifdef GRANDSTREAM_NETWORKS
		ringTimeoutBase = (p->ringt_base / 2) * READ_SIZE / 8;

		if (ringTimeoutBase < 4000)
		{
			ringTimeoutBase = 4000;
		}
		if (ringTimeoutBase > 10000)
		{
			ringTimeoutBase = 10000;
		}
#endif
		if (p->use_callerid && (p->cid_signalling == CID_SIG_SMDI && smdi_msg)) {
			number = smdi_msg->calling_st;
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_DEBUG, "[ch:%d]Receiving SMDI cid(%s) on channel %s\n", p->channel, number, ast_channel_name(chan));
#endif
		/* If we want caller id, we're in a prering state due to a polarity reversal
		 * and we're set to use a polarity reversal to trigger the start of caller id,
		 * grab the caller id and wait for ringing to start... */
		} else if (p->use_callerid && (ast_channel_state(chan) == AST_STATE_PRERING
			&& (p->cid_start == ANALOG_CID_START_POLARITY
				|| p->cid_start == ANALOG_CID_START_POLARITY_IN
				|| p->cid_start == ANALOG_CID_START_DTMF_NOALERT))) {
			/* If set to use DTMF CID signalling, listen for DTMF */
			if (p->cid_signalling == CID_SIG_DTMF) {
				int k = 0;
				int oldlinearity;
#ifdef GRANDSTREAM_NETWORKS
				int haveEvent = 0;
				int ucCount_timeout=0;
#endif
				int timeout_ms;
				int ms;
				struct timeval start = ast_tvnow();
				cs = NULL;
#ifdef GRANDSTREAM_NETWORKS
				ast_log(LOG_DEBUG, "[ch:%d]Receiving DTMF cid on channel %s", p->channel, ast_channel_name(chan));
#else
				ast_debug(1, "Receiving DTMF cid on channel %s\n", ast_channel_name(chan));
#endif

				oldlinearity = analog_set_linear_mode(p, idx, 0);

				/*
				 * We are the only party interested in the Rx stream since
				 * we have not answered yet.  We don't need or even want DTMF
				 * emulation.  The DTMF digits can come so fast that emulation
				 * can drop some of them.
				 */
				ast_set_flag(ast_channel_flags(chan), AST_FLAG_END_DTMF_ONLY);
				timeout_ms = 4000;/* This is a typical OFF time between rings. */
				for (;;) {
					struct ast_frame *f;

					ms = ast_remaining_ms(start, timeout_ms);
					res = ast_waitfor(chan, ms);
#ifdef GRANDSTREAM_NETWORKS
					if (res < 0) 
#else
					if (res <= 0) 
#endif
					{
						/*
						 * We do not need to restore the analog_set_linear_mode()
						 * or AST_FLAG_END_DTMF_ONLY flag settings since we
						 * are hanging up the channel.
						 */
#ifdef GRANDSTREAM_NETWORKS
						ast_log(LOG_ERROR, "[ch:%d]DTMFCID timed out waiting for ring:%d, errno:%d(%s), ringBase:%d. "
							"Exiting simple switch!", p->channel, res, errno, strerror(errno), ringTimeoutBase);
						//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
#else
						ast_log(LOG_WARNING, "DTMFCID timed out waiting for ring. "
							"Exiting simple switch\n");
#endif
						ast_hangup(chan);
						goto quit;
					}
#ifdef GRANDSTREAM_NETWORKS
					else if (0 == res)
					{
						if(0 == ucCount_timeout)
						{
							res = 500;/*append 0.5 second, waiting for callerid or first alerting tone*/
							ucCount_timeout++;
							ast_log(LOG_WARNING, "ucCount_timeout=%d !",ucCount_timeout);
						}
						else
						{
							if(p->ucHasReceiveRing)
							{
								p->ucHasReceiveRing = 0;
								if (haveEvent)
								{	/* Still working */
									ast_log(LOG_WARNING, "[ch:%d]DTMFCID detect timeout. haveEvent:%d, res:%d, errno:%d(%s), ringBase:%d. continue call; Maybe unmatched CID scheme!",
										p->channel, haveEvent, res, errno, strerror(errno), ringTimeoutBase);
									break;
								}
								else
								{
									ast_log(LOG_WARNING, "[ch:%d]DTMFCID detect timeout. haveEvent:%d, res:%d, errno:%d(%s), ringBase:%d. Hangup; Maybe caller was canceled!",
										p->channel, haveEvent, res, errno, strerror(errno), ringTimeoutBase);
									//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
									ast_hangup(chan);
									goto quit;
								}
							}
							else
							{
								ast_log(LOG_WARNING, "DTMF detect timeout,and has not ringbegin event arrive,not a real call,release");
								//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
								ast_hangup(chan);
								goto quit;
							}
						}
					}
#endif
					f = ast_read(chan);
					if (!f) {
						break;
					}
					if (f->frametype == AST_FRAME_DTMF) {
						if (k < ARRAY_LEN(dtmfbuf) - 1) {
							dtmfbuf[k++] = f->subclass.integer;
						}
#ifdef GRANDSTREAM_NETWORKS
						ast_log(LOG_NOTICE, "[ch:%d]CID got digit %d '%c'", p->channel, f->subclass.integer, f->subclass.integer);
#else
						ast_debug(1, "CID got digit '%c'\n", f->subclass.integer);
#endif
						start = ast_tvnow();
#ifdef GRANDSTREAM_NETWORKS
						res = ringTimeoutBase;/* This is a typical OFF time between rings. */
						haveEvent = 0;
#endif
					}
#ifdef GRANDSTREAM_NETWORKS
					else if (AST_FRAME_VOICE != f->frametype
						&& AST_FRAME_VIDEO != f->frametype)
					{
						haveEvent = 1;
						ast_log(LOG_NOTICE, "[ch:%d]Got frame type:%d, subclass:0x%x.",
							p->channel, f->frametype, f->subclass.integer);
					}
#endif
					ast_frfree(f);
					if (ast_channel_state(chan) == AST_STATE_RING ||
						ast_channel_state(chan) == AST_STATE_RINGING) {
						break; /* Got ring */
					}
#ifdef GRANDSTREAM_NETWORKS
#if defined(SVIP_DETECT_CID) && defined(SVIP_DETECT_DTMF)
					if (p->channel > 0 && p->channel <= ARRAY_LEN(g_svip_dtmf_digit))
					{
						if(dwFirstDigitC==0 && (strchr(g_svip_dtmf_digit[p->channel - 1], 'A')
							|| strchr(g_svip_dtmf_digit[p->channel - 1], 'D')))
						{
							dwHasRecvAorD=1;
						}

						dtmf_end_index = strchr(g_svip_dtmf_digit[p->channel - 1], 'C');
						
						if (dtmf_end_index != NULL)
						{
							if(dwHasRecvAorD)
							{
								ast_log(LOG_DEBUG,"got end flag C, break");
								break;
							}

							if(1 == dwFirstDigitC)/*C123C*/
							{
								ast_log(LOG_DEBUG,"got end flag C, break");
								break;
							}
							else
							{
								ast_log(LOG_DEBUG,"got flag C");
								if(!dwHasRecvAorD)
								{
									if('C' == g_svip_dtmf_digit[p->channel - 1][0])/*C123*/
									{
										*dtmf_end_index = 'A';
										dwFirstDigitC = 1;
										ast_log(LOG_DEBUG,"got first flag C");
									}
									else/*123C*/
									{
										ast_log(LOG_DEBUG,"got first flag C,break");
										break;
									}
								}
							}
						}

						dtmf_end_index = strchr(g_svip_dtmf_digit[p->channel - 1], '#');

						if (dtmf_end_index != NULL)
						{
							break;
						}
					}
#endif
#endif
				}
				ast_clear_flag(ast_channel_flags(chan), AST_FLAG_END_DTMF_ONLY);
				dtmfbuf[k] = '\0';

				analog_set_linear_mode(p, idx, oldlinearity);
#ifdef GRANDSTREAM_NETWORKS
#if defined(SVIP_DETECT_CID) && defined(SVIP_DETECT_DTMF)
				if (p->channel > 0 && p->channel <= ARRAY_LEN(g_svip_dtmf_digit))
				{
					ast_log(AST_LOG_DEBUG, "[ch:%d]SVIP DTMF CID got string '%s'", 
						p->channel, g_svip_dtmf_digit[p->channel - 1]);
					svip_dtmf_flags = 0;
					dw_who_detect_callerid =3;
					dtmf_end_index = strchr(g_svip_dtmf_digit[p->channel - 1], 'A');
					if(dtmf_end_index == NULL)
						dtmf_end_index = strchr(g_svip_dtmf_digit[p->channel - 1], 'D');
					if (dtmf_end_index != NULL){
						if(g_svip_dtmf_digit[p->channel - 1][0] != 'A' && g_svip_dtmf_digit[p->channel - 1][0] != 'D'){
							ast_log(LOG_VERBOSE, "[ch:%d]Total Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
							strncpy(dtmfcid, dtmf_end_index, sizeof(dtmfcid) - 1);
							ast_log(LOG_VERBOSE, "[ch:%d]Really Receiving DTMF cid(%s) on channel %s\n", p->channel, dtmfcid, ast_channel_name(chan));
							strncpy(g_svip_dtmf_digit[p->channel - 1], dtmfcid, DAHDI_MAX_DTMF_BUF - 1);
							ast_log(LOG_VERBOSE, "[ch:%d]complete handle, Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
							dtmfcid[0] = 0;
						}
						if(!strchr(g_svip_dtmf_digit[p->channel - 1], 'C')
							&&!strchr(g_svip_dtmf_digit[p->channel - 1], '#'))
						{
							int len=0;
							len=strlen(g_svip_dtmf_digit[p->channel - 1]);
							if(len < DAHDI_MAX_DTMF_BUF-1)
							g_svip_dtmf_digit[p->channel - 1][len] = 'C';
							g_svip_dtmf_digit[p->channel - 1][len+1] = '\0';
							ast_log(LOG_VERBOSE, "[ch:%d]complete handle, Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
						}
					}
					else
					{
						dtmf_end_index = strchr(g_svip_dtmf_digit[p->channel - 1], 'C');
						if(dtmf_end_index != NULL)
						{
							if(strchr(dtmf_end_index+1, 'C')||strchr(dtmf_end_index+1, '#')) /*C1234C and C1234# MODE*/
							{
								*dtmf_end_index='A';
								ast_log(LOG_VERBOSE, "[ch:%d]complete handle, Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
							}
							else
							{
								if(1 == strlen(dtmf_end_index))/*1234C	MODE*/
								{
									dtmfcid[0]='A';
									strncpy(dtmfcid+1, g_svip_dtmf_digit[p->channel - 1], sizeof(dtmfcid) - 1);				
									strncpy(g_svip_dtmf_digit[p->channel - 1], dtmfcid, DAHDI_MAX_DTMF_BUF - 1);					
									ast_log(LOG_VERBOSE, "[ch:%d]complete handle, Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
								}
								else /*C1234  MODE*/
								{
									int len=0;
									strncpy(dtmfcid, dtmf_end_index, sizeof(dtmfcid) - 2);
									len=strlen(dtmfcid);
									dtmfcid[len]='C';
									dtmfcid[len+1]='\0';
									strncpy(g_svip_dtmf_digit[p->channel - 1], dtmfcid, DAHDI_MAX_DTMF_BUF - 1);						
									ast_log(LOG_VERBOSE, "[ch:%d]complete handle, Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
								}
							}
						}
						else/*1234	MODE*/
						{
							if(0!=strlen(g_svip_dtmf_digit[p->channel - 1]))
							{
								int len=0;
								dtmfcid[0]='A';
								strncpy(dtmfcid+1, g_svip_dtmf_digit[p->channel - 1], sizeof(dtmfcid) - 2);
								len=strlen(dtmfcid);
								dtmfcid[len]='C';
								dtmfcid[len+1]='\0';
								strncpy(g_svip_dtmf_digit[p->channel - 1], dtmfcid, DAHDI_MAX_DTMF_BUF - 1);
							}
							else
							{
								dw_who_detect_callerid = 0;
							}
							ast_log(LOG_VERBOSE, "[ch:%d]complete handle, Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
						}
						ast_log(LOG_VERBOSE, "[ch:%d]Really Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
					}
					callerid_get_dtmf(g_svip_dtmf_digit[p->channel - 1], dtmfcid, &svip_dtmf_flags);
					if(strlen(dtmfcid)>= ANALOG_MAX_CID)
					{
						dtmfcid[ANALOG_MAX_CID-1] = '\0';
					}
				}
				else
				{
					svip_dtmf_flags = -1;
				}

				if (0 == svip_dtmf_flags)
				{
					flags = 0;
					ast_log(LOG_NOTICE, "[ch:%d]SVIP DTMF CID is '%s', flags %d.", p->channel, dtmfcid, flags);
				}
				else
#endif
#endif
					
#ifdef GRANDSTREAM_NETWORKS
				{
					/* Got cid and ring. */
					ast_debug(1, "[ch:%d]CID got string '%s'", p->channel, dtmfbuf);
					callerid_get_dtmf(dtmfbuf, dtmfcid, &flags);
					ast_log(LOG_NOTICE, "[ch:%d]CID is '%s', flags %d.", p->channel, dtmfcid, flags);
				}
#else
				/* Got cid and ring. */
				ast_debug(1, "CID got string '%s'\n", dtmfbuf);
				callerid_get_dtmf(dtmfbuf, dtmfcid, &flags);
				ast_debug(1, "CID is '%s', flags %d\n", dtmfcid, flags);
#endif
				/* If first byte is NULL, we have no cid */
				if (!ast_strlen_zero(dtmfcid)) {
					number = dtmfcid;
				} else {
					number = NULL;
				}

			/* If set to use V23 Signalling, launch our FSK gubbins and listen for it */
			} else if ((p->cid_signalling == CID_SIG_V23) || (p->cid_signalling == CID_SIG_V23_JP)) {
				namebuf[0] = 0;
				numbuf[0] = 0;

				if (!analog_start_cid_detect(p, p->cid_signalling)) {
					int timeout = 10000;  /* Ten seconds */
					struct timeval start = ast_tvnow();
					enum analog_event ev;
					int off_ms;
					int ms;
					struct timeval off_start;

					if (!p->usedistinctiveringdetection) {
						/* Disable distinctive ring timeout count */
						analog_set_ringtimeout(p, 0);
					}
					while ((ms = ast_remaining_ms(start, timeout))) {
						res = analog_get_callerid(p, namebuf, numbuf, &ev, ms);
						if (res < 0) {
							ast_log(LOG_WARNING,
								"CallerID returned with error on channel '%s'\n",
								ast_channel_name(chan));
							break;
						}
						if (res == 0) {
							break;
						}

						if (res != 1) {
							continue;
						}
						if (ev == ANALOG_EVENT_NOALARM) {
							analog_set_alarm(p, 0);
						}
						if (p->cid_signalling == CID_SIG_V23_JP) {
							if (ev == ANALOG_EVENT_RINGBEGIN) {
								analog_off_hook(p);
								usleep(1);
							}
						} else {
							break;
						}
					}

					name = namebuf;
					number = numbuf;
					analog_stop_cid_detect(p);

					if (p->cid_signalling == CID_SIG_V23_JP) {
						res = analog_on_hook(p);
						usleep(1);
					}

					/* Finished with Caller*ID, now wait for a ring to make sure there really is a call coming */
#ifdef GRANDSTREAM_NETWORKS
					res = ringTimeoutBase;/* This is a typical OFF time between rings. */
#endif
					off_start = ast_tvnow();
					off_ms = 4000;/* This is a typical OFF time between rings. */
					while ((ms = ast_remaining_ms(off_start, off_ms))) {
						struct ast_frame *f;

						res = ast_waitfor(chan, ms);
#ifdef GRANDSTREAM_NETWORKS
						if (res < 0) 
#else
						if (res <= 0) 
#endif
						{
#ifdef GRANDSTREAM_NETWORKS
							ast_log(LOG_ERROR, "[ch:%d]CID timed out waiting for ring:%d, errno:%d(%s), ringBase:%d. "
								"Exiting simple switch.", p->channel, res, errno, strerror(errno), ringTimeoutBase);
							//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
#else
							ast_log(LOG_WARNING, "CID timed out waiting for ring. "
								"Exiting simple switch\n");
#endif
							ast_hangup(chan);
							goto quit;
						}
#ifdef GRANDSTREAM_NETWORKS
						else if (0 == res)
						{
							ast_log(LOG_WARNING, "[ch:%d]Detect ring timeout, CID type %s. res:%d, errno:%d(%s), ringBase:%d. hangup; Maybe caller was canceled!",
									p->channel, CID_SIG_V23 == p->cid_signalling ? "V23" : "V23_JP", res, errno, strerror(errno), ringTimeoutBase);
							//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
							ast_hangup(chan);
							goto quit;
						}
#endif
						if (!(f = ast_read(chan))) 
						{
#ifdef GRANDSTREAM_NETWORKS
							ast_log(LOG_WARNING, "[ch:%d]Hangup received waiting for ring. Exiting simple switch.", p->channel);
							//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
#else
							ast_log(LOG_WARNING, "Hangup received waiting for ring. Exiting simple switch\n");
#endif
							ast_hangup(chan);
							goto quit;
						}
#ifdef GRANDSTREAM_NETWORKS
						else if (AST_FRAME_VOICE != f->frametype
							&& AST_FRAME_VIDEO != f->frametype
							&& AST_FRAME_NULL != f->frametype)
						{
							ast_log(LOG_NOTICE, "[ch:%d]Got %s frame type:%d, subclass:0x%x.",
								p->channel, 
								CID_SIG_V23 == p->cid_signalling ? "V23" : "V23_JP",
								f->frametype, f->subclass.integer);
							ast_frfree(f);
							break;
						}
#endif
						ast_frfree(f);
						if (ast_channel_state(chan) == AST_STATE_RING ||
							ast_channel_state(chan) == AST_STATE_RINGING)
							break; /* Got ring */
					}

					if (analog_distinctive_ring(chan, p, idx, NULL)) {
						goto quit;
					}

					if (res < 0) 
					{
#ifdef GRANDSTREAM_NETWORKS
						ast_log(LOG_WARNING, "[ch:%d]CallerID returned with error on channel '%s', res:%d!", p->channel, ast_channel_name(chan), res);
#else
						ast_log(LOG_WARNING, "CallerID returned with error on channel '%s'\n", ast_channel_name(chan));
#endif
					}
				}
				else
				{
#ifdef GRANDSTREAM_NETWORKS
					ast_log(LOG_WARNING, "[ch:%d]Unable to get caller ID space. res:%d!", p->channel, res);
#else
					ast_log(LOG_WARNING, "Unable to get caller ID space\n");
#endif
				}
			} 
			else 
			{
#ifdef GRANDSTREAM_NETWORKS
				ast_log(LOG_WARNING, "[ch:%d]Channel %s in prering "
					"state, cid_signalling is %d(%s), but I have nothing to do. "
					"Terminating simple switch, should be "
					"restarted by the actual ring.",
					p->channel, ast_channel_name(chan), p->cid_signalling, ast_cid_signal2string(p->cid_signalling));
				//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
#else
				ast_log(LOG_WARNING, "Channel %s in prering "
					"state, but I have nothing to do. "
					"Terminating simple switch, should be "
					"restarted by the actual ring.\n",
					ast_channel_name(chan));
#endif
				ast_hangup(chan);
				goto quit;
			}
		} else if (p->use_callerid && p->cid_start == ANALOG_CID_START_RING) {
			int timeout = 10000;  /* Ten seconds */
			struct timeval start = ast_tvnow();
			enum analog_event ev;
			int curRingData[RING_PATTERNS] = { 0 };
			int receivedRingT = 0;
#ifdef GRANDSTREAM_NETWORKS
			long long int det_time = 0; /* ms */
			int start_cid_res;
#endif

			namebuf[0] = 0;
			numbuf[0] = 0;
			
#ifdef GRANDSTREAM_NETWORKS
			if (!(start_cid_res = analog_start_cid_detect(p, p->cid_signalling)))
#else
			if (!analog_start_cid_detect(p, p->cid_signalling)) 
#endif
			{
				while (1) 
				{
#ifdef GRANDSTREAM_NETWORKS
					det_time = ast_tvdiff_ms(ast_tvnow(), start);
					res = analog_get_callerid(p, namebuf, numbuf, &ev, timeout - det_time);
#else
					res = analog_get_callerid(p, namebuf, numbuf, &ev, timeout - ast_tvdiff_ms(ast_tvnow(), start));
#endif

					if (res == 0) 
					{
						break;
					}

					if (res == 1 || res == 2)
					{
						if (ev == ANALOG_EVENT_NOALARM) 
						{
							analog_set_alarm(p, 0);
						} 
						else if (ev == ANALOG_EVENT_POLARITY && p->hanguponpolarityswitch && p->polarity == POLARITY_REV)
						{
#ifdef GRANDSTREAM_NETWORKS
							ast_debug(1, "[ch:%d]Hanging up due to polarity on channel %s reversal while detecting callerid.",
								p->channel, ast_channel_name(chan));
							//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
#else
							ast_debug(1, "Hanging up due to polarity reversal on channel %d while detecting callerid\n", p->channel);
#endif
							p->polarity = POLARITY_IDLE;
							ast_hangup(chan);
							goto quit;
						} 
						else if (ev != ANALOG_EVENT_NONE && ev != ANALOG_EVENT_RINGBEGIN && ev != ANALOG_EVENT_RINGOFFHOOK) 
						{
#ifdef GRANDSTREAM_NETWORKS
							ast_log(AST_LOG_WARNING, "[ch:%d]Unfriendliness event :%d(%s)!",
								p->channel, ev, analog_event2str(ev));
#endif						
							break;
						}
						if (res != 2) 
						{
							/* Let us detect callerid when the telco uses distinctive ring */
							curRingData[receivedRingT] = p->ringt;

							if (p->ringt < p->ringt_base/2)
							{
								break;
							}
							/* Increment the ringT counter so we can match it against
							   values in chan_dahdi.conf for distinctive ring */
							if (++receivedRingT == RING_PATTERNS) 
							{
								break;
							}
						}
					}

					if (ast_tvdiff_ms(ast_tvnow(), start) > timeout) 
					{
						break;
					}

				}
				name = namebuf;
				number = numbuf;

				analog_stop_cid_detect(p);

#ifdef GRANDSTREAM_NETWORKS
				if ('\0' == namebuf[0] && '\0' == numbuf[0]
					&& 0 == receivedRingT && det_time > ringTimeoutBase) /* No ring */
				{
					/* When end detected callerid, if no ring,maybe is caller onhook */
					ast_log(LOG_NOTICE, "[ch:%d]Hanging up channel %s, not ring detected(%lld > %d), maybe caller is cancel call.",
						p->channel, ast_channel_name(chan), det_time, ringTimeoutBase);
					//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
					ast_hangup(chan);
					goto quit;
				}
#endif

				if (analog_distinctive_ring(chan, p, idx, curRingData)) {
					goto quit;
				}

				if (res < 0) 
				{
#ifdef GRANDSTREAM_NETWORKS
					ast_log(LOG_WARNING, "[ch:%d]CallerID returned with error on channel '%s', res:%d.",
						p->channel, ast_channel_name(chan), res);
#else
					ast_log(LOG_WARNING, "CallerID returned with error on channel '%s'\n", ast_channel_name(chan));
#endif
				}
			}
			else 
			{
#ifdef GRANDSTREAM_NETWORKS
				ast_log(LOG_WARNING, "[ch:%d]Unable to get caller ID space, cid_sig:%d(%s), res:%d",
					p->channel, p->cid_signalling, ast_cid_signal2string(p->cid_signalling), start_cid_res);
#else
				ast_log(LOG_WARNING, "Unable to get caller ID space\n");
#endif
			}
		}
#ifdef GRANDSTREAM_NETWORKS
		else if(p->use_callerid && ANALOG_CID_START_AUTO_DETECT == p->cid_start)
		{
			int oldlinearity; 
			struct timeval start = ast_tvnow();
			/*enum analog_event ev;*/
			int timeout = 5000;
			
			namebuf[0] = 0;
			numbuf[0] = 0;
			oldlinearity = analog_set_linear_mode(p, idx, 0);
			ast_log(LOG_NOTICE, "Caller ID auto detect");
			/*
			 * We are the only party interested in the Rx stream since
			 * we have not answered yet.  We don't need or even want DTMF
			 * emulation.  The DTMF digits can come so fast that emulation
			 * can drop some of them.
			 */
			ast_clear_flag(ast_channel_flags(chan), AST_FLAG_END_DTMF_ONLY);
			res = ringTimeoutBase;/* This is a typical OFF time between rings. */
			
			for (;;)
			{
				struct ast_frame *f = NULL; 
				res = ast_waitfor(chan, res);
				if (res < 0) 
				{
					/** We do not need to restore the analog_set_linear_mode()
					 * or AST_FLAG_END_DTMF_ONLY flag settings since we
					 * are hanging up the channel. */
					ast_log(LOG_ERROR, "[ch:%d]DTMFCID timed out waiting for ring:%d, errno:%d(%s), ringBase:%d. "
						"Exiting simple switch!", p->channel, res, errno, strerror(errno), ringTimeoutBase);
					if(1 == dwhas_start_fsk_detect)
					{
						analog_stop_cid_detect_auto(p);
						dwhas_start_fsk_detect = 0;
					}
        		    //ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
					ast_hangup(chan);
					if(f)
					{
						ast_frfree(f);
						f = NULL;
					}
					goto quit;
				}
				else if (0 == res)
				{
					if(p->ucHasReceiveRing)
					{
						ast_log(LOG_NOTICE, "**************, Has receive ring before!");
						p->ucHasReceiveRing = 0;
						if(f)
						{
							ast_frfree(f);
							f = NULL;
						}
						break;
					}
					else
					{
						ast_log(LOG_WARNING, "DTMF detect timeout,and has not ringbegin event arrive,not a real call,release");
						if(1 == dwhas_start_fsk_detect)
						{
							analog_stop_cid_detect_auto(p);
							dwhas_start_fsk_detect = 0;
						}
        				//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
						ast_hangup(chan);
						if(f)
						{
							ast_frfree(f);
							f = NULL;
						}
						goto quit;
					} 
				}
				f = ast_read(chan);
				
				if (!f)
				{
					break;
				}
				
				if (f->frametype == AST_FRAME_DTMF)
				{
					res = 500;/* This is a typical OFF time between rings. */
				}
				else if(AST_FRAME_VOICE == f->frametype)
				{
					if(0 == dwhas_start_fsk_detect)
					{
						if(0==analog_start_cid_detect_auto(p, 0))/*start seccess*/
						{
							dwhas_start_fsk_detect = 1;
						}
						else
						{
							ast_log(LOG_ERROR, "[ch:%d]start cid auto detect fail,hungup",p->channel);
        				    //ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
							ast_hangup(chan);
							if(f)
							{
								ast_frfree(f);
								f = NULL;
							}
							goto quit;
						}
					}
					
					analog_lock_private(p);
					res = analog_get_callerid_auto(p, namebuf, numbuf, f->data.ptr, f->datalen, &dwwhodetectcallerID);
					analog_unlock_private(p);

					if (res == 0) 
					{
						ast_log(LOG_NOTICE, "[ch:%d]Got caller number =%s, dwwhodetectcallerID=%d",p->channel, numbuf, dwwhodetectcallerID);
						dw_who_detect_callerid = dwwhodetectcallerID;
						if(f)
						{
							ast_frfree(f);
							f = NULL;
						}
						break;
					}
					if (ast_tvdiff_ms(ast_tvnow(), start) > timeout)
					{
						ast_log(LOG_ERROR, "[ch:%d]FSK timeout-------------",p->channel);
						if(p->ucHasReceiveRing)
						{
							ast_log(LOG_ERROR, "[ch:%d]FSK timeout,and has recv ring,continue calling ...",p->channel);
							p->ucHasReceiveRing = 0;
							if(f)
							{
								ast_frfree(f);
								f = NULL;
							}
							break;
						}
						else
						{
							ast_log(LOG_ERROR, "[ch:%d]FSK timeout,and has not recv ring,hungup",p->channel);
							if(1 == dwhas_start_fsk_detect)
							{
								analog_stop_cid_detect_auto(p);
								dwhas_start_fsk_detect = 0;
							}
            				//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED); 
							ast_hangup(chan);
							if(f)
							{
								ast_frfree(f);
								f = NULL;
							}
							goto quit;
						}
					}
					res = 500;
				}
				else if (AST_FRAME_VOICE != f->frametype
					&& AST_FRAME_VIDEO != f->frametype)
				{
					ast_log(LOG_NOTICE, "[ch:%d]Got frame type:%d, subclass:0x%x.",
						p->channel, f->frametype, f->subclass.integer);
					res =500;
				}

				if(f)
				{
					ast_frfree(f);
					f = NULL;
				}

#if defined(SVIP_DETECT_CID) && defined(SVIP_DETECT_DTMF)
				if (p->channel > 0 && p->channel <= ARRAY_LEN(g_svip_dtmf_digit))
				{
					if(dwFirstDigitC==0 && (strchr(g_svip_dtmf_digit[p->channel - 1], 'A')
						|| strchr(g_svip_dtmf_digit[p->channel - 1], 'D')))
					{
						dwHasRecvAorD=1;
					}

					dtmf_end_index = strchr(g_svip_dtmf_digit[p->channel - 1], 'C');
					
					if (dtmf_end_index != NULL)
					{
						if(dwHasRecvAorD)
						{
							ast_log(LOG_DEBUG,"got end flag C, break");
							dw_who_detect_callerid = 3;
							break;
						}

						if(1 == dwFirstDigitC)/*C123C*/
						{
							ast_log(LOG_DEBUG,"got end flag C, break");
							dw_who_detect_callerid = 3;
							break;
						}
						else
						{
							ast_log(LOG_DEBUG,"got flag C");
							if(!dwHasRecvAorD)
							{
								if('C' == g_svip_dtmf_digit[p->channel - 1][0])/*C123*/
								{
									*dtmf_end_index = 'A';
									dwFirstDigitC = 1;
									ast_log(LOG_DEBUG,"got first flag C");
								}
								else/*123C*/
								{
									dw_who_detect_callerid = 3;
									ast_log(LOG_DEBUG,"got first flag C,break");
									break;
								}
							}
						}
					}
					dtmf_end_index = strchr(g_svip_dtmf_digit[p->channel - 1], '#');
					if (dtmf_end_index != NULL)
					{
						dw_who_detect_callerid = 3;
						break;
					}
				}
#endif
			}
			analog_stop_cid_detect_auto(p);
			ast_clear_flag(ast_channel_flags(chan), AST_FLAG_END_DTMF_ONLY);

			analog_set_linear_mode(p, idx, oldlinearity);
#if defined(SVIP_DETECT_CID) && defined(SVIP_DETECT_DTMF)
			if(1 != dw_who_detect_callerid && 2 != dw_who_detect_callerid)
			{
				if (p->channel > 0 && p->channel <= ARRAY_LEN(g_svip_dtmf_digit))
				{
					ast_log(AST_LOG_DEBUG, "[ch:%d]SVIP DTMF CID got string '%s'",
						p->channel, g_svip_dtmf_digit[p->channel - 1]);
					svip_dtmf_flags = 0;
					dw_who_detect_callerid =3;
					dtmf_end_index = strchr(g_svip_dtmf_digit[p->channel - 1], 'A');
					if(dtmf_end_index == NULL)
						dtmf_end_index = strchr(g_svip_dtmf_digit[p->channel - 1], 'D');
					if (dtmf_end_index != NULL){
						if(g_svip_dtmf_digit[p->channel - 1][0] != 'A' && g_svip_dtmf_digit[p->channel - 1][0] != 'D'){
							ast_log(LOG_VERBOSE, "[ch:%d]Total Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
							strncpy(dtmfcid, dtmf_end_index, sizeof(dtmfcid) - 1);
							ast_log(LOG_VERBOSE, "[ch:%d]Really Receiving DTMF cid(%s) on channel %s\n", p->channel, dtmfcid, ast_channel_name(chan));
							strncpy(g_svip_dtmf_digit[p->channel - 1], dtmfcid, DAHDI_MAX_DTMF_BUF - 1);
							ast_log(LOG_VERBOSE, "[ch:%d]complete handle, Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
							dtmfcid[0] = 0;
						}
						if(!strchr(g_svip_dtmf_digit[p->channel - 1], 'C')
							&&!strchr(g_svip_dtmf_digit[p->channel - 1], '#'))
						{
							int len=0;
							len=strlen(g_svip_dtmf_digit[p->channel - 1]);
							if(len < DAHDI_MAX_DTMF_BUF-1)
							g_svip_dtmf_digit[p->channel - 1][len] = 'C';
							g_svip_dtmf_digit[p->channel - 1][len+1] = '\0';
							ast_log(LOG_VERBOSE, "[ch:%d]complete handle, Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
						}
					}
					else
					{
						dtmf_end_index = strchr(g_svip_dtmf_digit[p->channel - 1], 'C');
						if(dtmf_end_index != NULL)
						{
							if(strchr(dtmf_end_index+1, 'C')||strchr(dtmf_end_index+1, '#')) /*C1234C and C1234# MODE*/
							{
								*dtmf_end_index='A';
								ast_log(LOG_VERBOSE, "[ch:%d]complete handle, Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
							}
							else
							{
								if(1 == strlen(dtmf_end_index))/*1234C	MODE*/
								{
									dtmfcid[0]='A';
									strncpy(dtmfcid+1, g_svip_dtmf_digit[p->channel - 1], sizeof(dtmfcid) - 1);
									strncpy(g_svip_dtmf_digit[p->channel - 1], dtmfcid, DAHDI_MAX_DTMF_BUF - 1);
									ast_log(LOG_VERBOSE, "[ch:%d]complete handle, Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
								}
								else /*C1234  MODE*/
								{
									int len=0;
									strncpy(dtmfcid, dtmf_end_index, sizeof(dtmfcid) - 2);
									len=strlen(dtmfcid);
									dtmfcid[len]='C';
									dtmfcid[len+1]='\0';
									strncpy(g_svip_dtmf_digit[p->channel - 1], dtmfcid, DAHDI_MAX_DTMF_BUF - 1);
									ast_log(LOG_VERBOSE, "[ch:%d]complete handle, Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
								}
							}
						}
						else/*1234	MODE*/
						{
							if(0!=strlen(g_svip_dtmf_digit[p->channel - 1]))
							{
								int len=0;
								dtmfcid[0]='A';
								strncpy(dtmfcid+1, g_svip_dtmf_digit[p->channel - 1], sizeof(dtmfcid) - 2);
								len=strlen(dtmfcid);
								dtmfcid[len]='C';
								dtmfcid[len+1]='\0';
								strncpy(g_svip_dtmf_digit[p->channel - 1], dtmfcid, DAHDI_MAX_DTMF_BUF - 1);
							}
							else
							{
								dw_who_detect_callerid = 0;
							}
							ast_log(LOG_VERBOSE, "[ch:%d]complete handle, Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
						}
						ast_log(LOG_VERBOSE, "[ch:%d]Really Receiving DTMF cid(%s) on channel %s\n", p->channel, g_svip_dtmf_digit[p->channel - 1], ast_channel_name(chan));
					}
					callerid_get_dtmf(g_svip_dtmf_digit[p->channel - 1], dtmfcid, &svip_dtmf_flags);
					if(strlen(dtmfcid)>= ANALOG_MAX_CID)
					{
						dtmfcid[ANALOG_MAX_CID-1] = '\0';
					}
				}
				else
				{
					svip_dtmf_flags = -1;
				}
			}
			if (0 == svip_dtmf_flags)
			{
				flags = 0;
				ast_log(LOG_NOTICE, "[ch:%d]SVIP DTMF CID is '%s', flags %d.", p->channel, dtmfcid, flags);
			}
			else
#endif
			{
			
			}
			/* If first byte is NULL, we have no cid */

			if(1 == dw_who_detect_callerid || 2 == dw_who_detect_callerid)
			{
				if (!ast_strlen_zero(numbuf))
				{
					number = numbuf;
					name = namebuf;
				} else {
					number = NULL;
					name = NULL;
				}
			}
			else if((3 == dw_who_detect_callerid)
#if defined(SVIP_DETECT_CID) && defined(SVIP_DETECT_DTMF)
				&&(0 == svip_dtmf_flags)
#endif
					)
			{
				if (!ast_strlen_zero(dtmfcid))
				{
					number = dtmfcid;
				} else {
					number = NULL;
				}
				name = NULL;
			}
			else
			{
				number = NULL;
				name = NULL;
				ast_log(LOG_WARNING, "can not detect caller ID");
			}
		}
#endif
		else
		{
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_WARNING, 
				"[ch:%d]Cann't do CID, use_callerid:%u, cid_start:%u(%s), "
				"cid_signalling:%d(%s), _state:%u(%s), smdi_msg:0x%08tx",
				p->channel, p->use_callerid, p->cid_start, analog_cid_start2string(p->cid_start),
				p->cid_signalling, ast_cid_signal2string(p->cid_signalling), 
				ast_channel_state(chan), ast_state2str(ast_channel_state(chan)),
				(ptrdiff_t)smdi_msg
				);
			number = numbuf;
#endif
			cs = NULL;
#ifdef GRANDSTREAM_NETWORKS
			if( ANALOG_CID_START_DTMF_NOALERT == p->cid_start) {
				int timeout = 5000;  /* Ten seconds */
				struct timeval start = ast_tvnow();
				enum analog_event ev;
				ast_log(LOG_DEBUG,"Not detect callerid,CID start is ANALOG_CID_START_DTMF_NOALERT, wait for Ring");
				while(1) {
					res = analog_get_ring_event(p, &ev, timeout - ast_tvdiff_ms(ast_tvnow(), start));
					if(res==0) {
						if(ANALOG_EVENT_RINGOFFHOOK == ev) {
							ast_log(LOG_DEBUG,"Has received ring,start to call");
							break;
						}
					}

					if (ast_tvdiff_ms(ast_tvnow(), start) >= timeout) {
						ast_log(LOG_DEBUG,"Has not received ring,hangup call");
						ast_hangup(chan);
						goto quit;
					}
				}
			}
#endif
		}

		if (number) {
			ast_shrink_phone_number(number);
		}
#ifdef GRANDSTREAM_NETWORKS
		else
		{
			number = number_default;
			name = name_default;
		}
#endif
		
		ast_set_callerid(chan, number, name, number);

		if (cs) {
			callerid_free(cs);
		}

		analog_handle_notify_message(chan, p, flags, -1);

		ast_channel_lock(chan);
		ast_setstate(chan, AST_STATE_RING);
		ast_channel_rings_set(chan, 1);
		ast_channel_unlock(chan);
		analog_set_ringtimeout(p, p->ringt_base);
#ifdef GRANDSTREAM_NETWORKS
		p->ucHasReceiveRing = 0;
		if(ast_context_find(ast_channel_context(chan)))
		{  
			if (ast_exists_extension(chan, ast_channel_context(chan), "s", 1, S_OR(p->cid_num, number)))
			{
				res = ast_pbx_run(chan);
				if (res) 
				{
					//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED);
					ast_hangup(chan);
					ast_log(LOG_WARNING, "[ch:%d]PBX exited non-zero!", p->channel);
				}
			}
			else
			{
				//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED);
				ast_hangup(chan);
				ast_log(LOG_WARNING, "[ch:%d]not match caller %s!", p->channel, S_OR(p->cid_num, number));
			}
		}
		else
		{
			ast_log(LOG_WARNING, "[ch:%d]not find %s!", p->channel, ast_channel_context(chan));
			//ast_set_flag(chan->cdr, AST_CDR_FLAG_POST_DISABLED);
			ast_hangup(chan);
		}
#else
		res = ast_pbx_run(chan);
		if (res) {
			ast_hangup(chan);
			ast_log(LOG_WARNING, "PBX exited non-zero\n");
		}
#endif
		goto quit;
	default:
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_WARNING, "[ch:%d]Don't know how to handle simple switch with signalling:%d(%s).",
			p->channel, p->sig, analog_sigtype_to_str(p->sig));
#else
		ast_log(LOG_WARNING, "Don't know how to handle simple switch with signalling %s on channel %d\n", analog_sigtype_to_str(p->sig), p->channel);
#endif
		break;
	}
	res = analog_play_tone(p, idx, ANALOG_TONE_CONGESTION);
	if (res < 0) {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_WARNING, "[ch:%d]Unable to play congestion tone, return %d!", p->channel, res);
#else
		ast_log(LOG_WARNING, "Unable to play congestion tone on channel %d\n", p->channel);
#endif
	}
	ast_hangup(chan);
quit:
	ao2_cleanup(smdi_msg);
	analog_decrease_ss_count();
	return NULL;
}

int analog_ss_thread_start(struct analog_pvt *p, struct ast_channel *chan)
{
	pthread_t threadid;

	return ast_pthread_create_detached(&threadid, NULL, __analog_ss_thread, p);
}

static void analog_publish_channel_alarm_clear(int channel)
{
	RAII_VAR(struct ast_json *, body, NULL, ast_json_unref);

	ast_log(LOG_NOTICE, "Alarm cleared on channel %d\n", channel);
	body = ast_json_pack("{s: i}", "Channel", channel);
	if (!body) {
		return;
	}

	ast_manager_publish_event("AlarmClear", EVENT_FLAG_SYSTEM, body);
}

static struct ast_frame *__analog_handle_event(struct analog_pvt *p, struct ast_channel *ast)
{
	int res, x;
	int mysig;
	enum analog_sub idx;
	char *c;
	pthread_t threadid;
	struct ast_channel *chan;
	struct ast_frame *f;
	struct ast_control_pvt_cause_code *cause_code = NULL;
	int data_size = sizeof(*cause_code);
	char *subclass = NULL;

#ifdef GRANDSTREAM_NETWORKS
	char device_name[AST_CHANNEL_NAME];
	char *dashptr;
#if defined(SVIP_T38_STACK)  
	struct ast_control_t38_parameters t38_parameters;
#endif
	ast_log(LOG_VERBOSE, "[ch:%d]%s(%p, %p) ...", p->channel, __FUNCTION__, p, ast);
#else
	ast_debug(1, "%s %d\n", __FUNCTION__, p->channel);
#endif

	idx = analog_get_index(ast, p, 0);
	if (idx < 0) {
		return &ast_null_frame;
	}
	if (idx != ANALOG_SUB_REAL) {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_ERROR, "[ch:%d]We got an event on a non real sub(idx:%d).  Fix it!!", p->channel, idx);
#else
		ast_log(LOG_ERROR, "We got an event on a non real sub.	Fix it!\n");
#endif
	}

	mysig = p->sig;
	if (p->outsigmod > -1) {
		mysig = p->outsigmod;
	}

	p->subs[idx].f.frametype = AST_FRAME_NULL;
	p->subs[idx].f.subclass.integer = 0;
	p->subs[idx].f.datalen = 0;
	p->subs[idx].f.samples = 0;
	p->subs[idx].f.mallocd = 0;
	p->subs[idx].f.offset = 0;
	p->subs[idx].f.src = "dahdi_handle_event";
	p->subs[idx].f.data.ptr = NULL;
	f = &p->subs[idx].f;

	res = analog_get_event(p);

	ast_debug(1, "Got event %s(%d) on channel %d (index %u)\n", analog_event2str(res), res, p->channel, idx);

#ifdef GRANDSTREAM_NETWORKS
	if ((res & (ANALOG_EVENT_PULSEDIGIT | ANALOG_EVENT_DTMFUP)) && ast_channel_state(ast) != AST_STATE_RESERVED) 
#else
	if (res & (ANALOG_EVENT_PULSEDIGIT | ANALOG_EVENT_DTMFUP)) 
#endif
	{
		analog_set_pulsedial(p, (res & ANALOG_EVENT_PULSEDIGIT) ? 1 : 0);
#ifdef GRANDSTREAM_NETWORKS
		ast_debug(1, "[ch:%d]Detected %s digit '%c'", p->channel, (res & ANALOG_EVENT_PULSEDIGIT) ? "pulse ": "", res & 0xff);
#else
		ast_debug(1, "Detected %sdigit '%c'\n", (res & ANALOG_EVENT_PULSEDIGIT) ? "pulse ": "", res & 0xff);
#endif
		analog_confmute(p, 0);
		p->subs[idx].f.frametype = AST_FRAME_DTMF_END;
		p->subs[idx].f.subclass.integer = res & 0xff;
		analog_handle_dtmf(p, ast, idx, &f);
		return f;
	}

	if (res & ANALOG_EVENT_DTMFDOWN) {
#ifdef GRANDSTREAM_NETWORKS
		ast_debug(1, "[ch:%d]DTMF Down '%c'", p->channel, res & 0xff);
#else
		ast_debug(1, "DTMF Down '%c'\n", res & 0xff);
#endif
		/* Mute conference */
		analog_confmute(p, 1);
		p->subs[idx].f.frametype = AST_FRAME_DTMF_BEGIN;
		p->subs[idx].f.subclass.integer = res & 0xff;
		analog_handle_dtmf(p, ast, idx, &f);
		return f;
	}

	switch (res) {
	case ANALOG_EVENT_ALARM:
	case ANALOG_EVENT_POLARITY:
	case ANALOG_EVENT_ONHOOK:
		/* add length of "ANALOG " */
		data_size += 7;
		subclass = analog_event2str(res);
		data_size += strlen(subclass);
		cause_code = ast_alloca(data_size);
		memset(cause_code, 0, data_size);
		cause_code->ast_cause = AST_CAUSE_NORMAL_CLEARING;
		ast_copy_string(cause_code->chan_name, ast_channel_name(ast), AST_CHANNEL_NAME);
		snprintf(cause_code->code, data_size - sizeof(*cause_code) + 1, "ANALOG %s", subclass);
		break;
	default:
		break;
	}

	switch (res) {
	case ANALOG_EVENT_EC_DISABLED:
		ast_verb(3, "Channel %d echo canceler disabled due to CED detection\n", p->channel);
		analog_set_echocanceller(p, 0);
		break;
#ifdef HAVE_DAHDI_ECHOCANCEL_FAX_MODE
	case ANALOG_EVENT_TX_CED_DETECTED:
		ast_verb(3, "Channel %d detected a CED tone towards the network.\n", p->channel);
#ifdef GRANDSTREAM_NETWORKS
		analog_set_echocanceller(p, 0);  
#endif
		break;
	case ANALOG_EVENT_RX_CED_DETECTED:
		ast_log(LOG_DEBUG, "[Channel %d]ANALOG_EVENT_RX_CED_DETECTED p->owner=%p\n", p->channel,p->owner);
#ifdef GRANDSTREAM_NETWORKS
#if defined(SVIP_T38_STACK)
		ast_channel_lock(p->owner);
		struct ast_bridge_channel *peer_bridge_channel = ast_channel_get_bridge_channel(p->owner);
		ast_channel_unlock(p->owner);
		if(p->owner && ast_bridge_channel_get_channel_from_bridge_channel(peer_bridge_channel) && strcasecmp(ast_channel_tech(ast_bridge_channel_get_channel_from_bridge_channel(peer_bridge_channel))->type, "DAHDI") && !analog_get_t38_state(p))//The other end is not T30
		{
			ast_log(LOG_DEBUG, "[Channel %d]read the capabilities of the T.38 stack implementation for the used firmware version.\n", p->channel);
			/* Read the capabilities of the T.38 stack implementation for the used firmware version */
			struct ast_control_t38_parameters * parameters = analog_get_t38_capabilities(p);
			if(parameters)
			{
				memcpy(&t38_parameters, parameters, sizeof(struct ast_control_t38_parameters));
				t38_parameters.request_response = AST_T38_REQUEST_NEGOTIATE;
				ast_queue_control_data(p->owner, AST_CONTROL_T38_PARAMETERS, &t38_parameters, sizeof(t38_parameters));
				analog_set_t38_state(p, 1);
			}
		}
		ao2_ref(peer_bridge_channel, -1);
#endif
		analog_set_echocanceller(p, 0);  
#endif
		ast_verb(3, "Channel %d detected a CED tone from the network.\n", p->channel);
		break;
	case ANALOG_EVENT_EC_NLP_DISABLED:
		ast_verb(3, "Channel %d echo canceler disabled its NLP.\n", p->channel);
#ifdef GRANDSTREAM_NETWORKS
		analog_set_echocanceller(p, 0);  
#endif
		break;
	case ANALOG_EVENT_EC_NLP_ENABLED:
		ast_verb(3, "Channel %d echo canceler enabled its NLP.\n", p->channel);
#ifdef GRANDSTREAM_NETWORKS
		analog_set_echocanceller(p, 1);  
#endif
		break;
#endif
	case ANALOG_EVENT_PULSE_START:
		/* Stop tone if there's a pulse start and the PBX isn't started */
#ifdef GRANDSTREAM_NETWORKS
		if (!ast_channel_pbx(ast) && ast_channel_state(ast) != AST_STATE_RESERVED)
#else
		if (!ast_channel_pbx(ast))
#endif
		{
			analog_play_tone(p, ANALOG_SUB_REAL, -1);
		}
		break;
	case ANALOG_EVENT_DIALCOMPLETE:
		if (p->inalarm) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_WARNING, "[ch:%d]Channel inalarm, break;", p->channel);
#endif
			break;
		}
		x = analog_is_dialing(p, idx);
		if (!x) { /* if not still dialing in driver */
			analog_set_echocanceller(p, 1);
			if (p->echobreak) {
				analog_train_echocanceller(p);
				ast_copy_string(p->dop.dialstr, p->echorest, sizeof(p->dop.dialstr));
				p->dop.op = ANALOG_DIAL_OP_REPLACE;
				analog_dial_digits(p, ANALOG_SUB_REAL, &p->dop);
				p->echobreak = 0;
			} else {
				analog_set_dialing(p, 0);
				if ((mysig == ANALOG_SIG_E911) || (mysig == ANALOG_SIG_FGC_CAMA) || (mysig == ANALOG_SIG_FGC_CAMAMF)) {
					/* if thru with dialing after offhook */
					if (ast_channel_state(ast) == AST_STATE_DIALING_OFFHOOK) {
						ast_setstate(ast, AST_STATE_UP);
						p->subs[idx].f.frametype = AST_FRAME_CONTROL;
						p->subs[idx].f.subclass.integer = AST_CONTROL_ANSWER;
						break;
					} else { /* if to state wait for offhook to dial rest */
						/* we now wait for off hook */
						ast_setstate(ast,AST_STATE_DIALING_OFFHOOK);
					}
				}
				if (ast_channel_state(ast) == AST_STATE_DIALING) {
					if (analog_have_progressdetect(p)) {
#ifdef GRANDSTREAM_NETWORKS
						ast_debug(1, "[ch:%d]Done dialing, but waiting for progress detection before doing more...", p->channel);
#else
						ast_debug(1, "Done dialing, but waiting for progress detection before doing more...\n");
#endif
					} else if (analog_check_confirmanswer(p) || (!p->dialednone
						&& ((mysig == ANALOG_SIG_EM) || (mysig == ANALOG_SIG_EM_E1)
							|| (mysig == ANALOG_SIG_EMWINK) || (mysig == ANALOG_SIG_FEATD)
							|| (mysig == ANALOG_SIG_FEATDMF_TA) || (mysig == ANALOG_SIG_FEATDMF)
							|| (mysig == ANALOG_SIG_E911) || (mysig == ANALOG_SIG_FGC_CAMA)
							|| (mysig == ANALOG_SIG_FGC_CAMAMF) || (mysig == ANALOG_SIG_FEATB)
							|| (mysig == ANALOG_SIG_SF) || (mysig == ANALOG_SIG_SFWINK)
							|| (mysig == ANALOG_SIG_SF_FEATD) || (mysig == ANALOG_SIG_SF_FEATDMF)
							|| (mysig == ANALOG_SIG_SF_FEATB)))) {
						ast_setstate(ast, AST_STATE_RINGING);
					} else if (!p->answeronpolarityswitch) {
						ast_setstate(ast, AST_STATE_UP);
						p->subs[idx].f.frametype = AST_FRAME_CONTROL;
						p->subs[idx].f.subclass.integer = AST_CONTROL_ANSWER;
						/* If aops=0 and hops=1, this is necessary */
						p->polarity = POLARITY_REV;
					} else {
						/* Start clean, so we can catch the change to REV polarity when party answers */
#ifdef GRANDSTREAM_NETWORKS
						ast_setstate(ast, AST_STATE_RINGING);
						p->subs[idx].f.frametype = AST_FRAME_CONTROL;
						p->subs[idx].f.subclass.integer = AST_CONTROL_RINGING;
#endif
						p->polarity = POLARITY_IDLE;
					}
				}
			}
		}
		break;
	case ANALOG_EVENT_ALARM:
		analog_set_alarm(p, 1);
		analog_get_and_handle_alarms(p);
		cause_code->ast_cause = AST_CAUSE_NETWORK_OUT_OF_ORDER;
	case ANALOG_EVENT_ONHOOK:
		ast_queue_control_data(ast, AST_CONTROL_PVT_CAUSE_CODE, cause_code, data_size);
		ast_channel_hangupcause_hash_set(ast, cause_code, data_size);
		switch (p->sig) {
		case ANALOG_SIG_FXOLS:
		case ANALOG_SIG_FXOGS:
		case ANALOG_SIG_FXOKS:
			analog_start_polarityswitch(p);
			p->fxsoffhookstate = 0;
			p->onhooktime = time(NULL);
#ifdef GRANDSTREAM_NETWORKS
#if defined(SVIP_T38_STACK)
			/* When T38 session is active,Stop T.38 session. */
			if(analog_get_t38_state(p))
			{
				t38_parameters.request_response = AST_T38_REQUEST_TERMINATE;
				ast_queue_control_data(p->owner, AST_CONTROL_T38_PARAMETERS, &t38_parameters, sizeof(t38_parameters));
			}
#endif
#else
			p->msgstate = -1;
#endif

			/* Check for some special conditions regarding call waiting */
			if (idx == ANALOG_SUB_REAL) {
#ifdef GRANDSTREAM_NETWORKS
				if(p->subs[ANALOG_SUB_REAL].inthreeway)
				{
					ast_log(LOG_NOTICE, "[ch:%d]ready to hangup 3-way call with %s and %s",
							p->channel,
							ast_channel_name(p->subs[ANALOG_SUB_THREEWAY].owner),
							ast_channel_name(p->subs[ANALOG_SUB_REAL].owner));
					/* Need to hold the lock for real-call, private, and call-waiting call */
					analog_lock_sub_owner(p, ANALOG_SUB_THREEWAY);
					p->hangupallwheninthreeway = 1;
					
					ast_queue_hangup_with_cause(p->subs[ANALOG_SUB_THREEWAY].owner, AST_CAUSE_NO_ANSWER);
					ast_softhangup_nolock(p->subs[ANALOG_SUB_THREEWAY].owner, AST_SOFTHANGUP_DEV);
					
					ast_channel_unlock(p->subs[ANALOG_SUB_THREEWAY].owner);  
					break;
				}
#endif
				/* The normal line was hung up */
				if (p->subs[ANALOG_SUB_CALLWAIT].owner) {
					/* Need to hold the lock for real-call, private, and call-waiting call */
					analog_lock_sub_owner(p, ANALOG_SUB_CALLWAIT);
					if (!p->subs[ANALOG_SUB_CALLWAIT].owner) {
						/*
						 * The call waiting call dissappeared.
						 * This is now a normal hangup.
						 */
						analog_set_echocanceller(p, 0);
						return NULL;
					}

					/* There's a call waiting call, so ring the phone, but make it unowned in the mean time */
					analog_swap_subs(p, ANALOG_SUB_CALLWAIT, ANALOG_SUB_REAL);
					ast_verb(3, "Channel %d still has (callwait) call, ringing phone\n", p->channel);
					analog_unalloc_sub(p, ANALOG_SUB_CALLWAIT);
					analog_stop_callwait(p);
					analog_set_new_owner(p, NULL);
					/* Don't start streaming audio yet if the incoming call isn't up yet */
					if (ast_channel_state(p->subs[ANALOG_SUB_REAL].owner) != AST_STATE_UP) {
						analog_set_dialing(p, 1);
					}
					/* Unlock the call-waiting call that we swapped to real-call. */
					ast_channel_unlock(p->subs[ANALOG_SUB_REAL].owner);
#ifdef GRANDSTREAM_NETWORKS
					analog_set_cadence(p, p->subs[ANALOG_SUB_REAL].owner); /* and set p->cidrings */
					if (p->use_callerid)
					{
						if (ast_channel_connected(p->subs[ANALOG_SUB_REAL].owner)->id.name.valid)
						{
							ast_copy_string(p->lastcid_name, ast_channel_connected(p->subs[ANALOG_SUB_REAL].owner)->id.name.str, sizeof(p->lastcid_name));
							p->caller.id.name.str = p->lastcid_name;
						}

						if (ast_channel_connected(p->subs[ANALOG_SUB_REAL].owner)->id.number.valid)
						{
							ast_copy_string(p->lastcid_num, ast_channel_connected(p->subs[ANALOG_SUB_REAL].owner)->id.number.str, sizeof(p->lastcid_num));
							p->caller.id.number.str = p->lastcid_num;
						}
						else if(ast_channel_dialed(p->subs[ANALOG_SUB_REAL].owner)->number.str)
						{
							ast_copy_string(p->lastcid_num, ast_channel_dialed(p->subs[ANALOG_SUB_REAL].owner)->number.str, sizeof(p->lastcid_num));
							p->caller.id.number.str = p->lastcid_num;
						}
					}
#endif
					analog_ring(p);
				} else if (p->subs[ANALOG_SUB_THREEWAY].owner) {
					unsigned int mssinceflash;
#ifdef GRANDSTREAM_NETWORKS
					unsigned int real_inthreeway = p->subs[ANALOG_SUB_REAL].inthreeway;
#endif
					/* Need to hold the lock for real-call, private, and 3-way call */
					analog_lock_sub_owner(p, ANALOG_SUB_THREEWAY);
					if (!p->subs[ANALOG_SUB_THREEWAY].owner) {
#ifdef GRANDSTREAM_NETWORKS
						ast_log(LOG_NOTICE, "[ch:%d]Whoa, threeway disappeared kinda randomly.", p->channel);
#else
						ast_log(LOG_NOTICE, "Whoa, threeway disappeared kinda randomly.\n");
#endif
						/* Just hangup */
						return NULL;
					}
					if (p->owner != ast) {
						ast_channel_unlock(p->subs[ANALOG_SUB_THREEWAY].owner);
#ifdef GRANDSTREAM_NETWORKS
						ast_log(LOG_WARNING, "[ch:%d]This isn't good... ", p->channel);
#else
						ast_log(LOG_WARNING, "This isn't good...\n");
#endif
						/* Just hangup */
						return NULL;
					}

					mssinceflash = ast_tvdiff_ms(ast_tvnow(), p->flashtime);
#ifdef GRANDSTREAM_NETWORKS
					ast_debug(1, "[ch:%d]Last flash was %d ms ago; pbx:%p, _state:%u(%s), transfer:%u, transfertobusy:%u, inthreeway:%u, 3waysub_state:%d(%s)",
						p->channel, mssinceflash, ast_channel_pbx(ast), ast_channel_state(ast), ast_state2str(ast_channel_state(ast)), p->transfer, p->transfertobusy,
						real_inthreeway, ast_channel_state(p->subs[ANALOG_SUB_THREEWAY].owner), ast_state2str(ast_channel_state(p->subs[ANALOG_SUB_THREEWAY].owner)));
#else
					ast_debug(1, "Last flash was %u ms ago\n", mssinceflash);
#endif
					if (mssinceflash < MIN_MS_SINCE_FLASH) {
						/* It hasn't been long enough since the last flashook.  This is probably a bounce on
						   hanging up.  Hangup both channels now */
#ifdef GRANDSTREAM_NETWORKS
						ast_debug(1, "[ch:%d]Looks like a bounced flash, hanging up both calls.", p->channel);
#else
						ast_debug(1, "Looks like a bounced flash, hanging up both calls on %d\n", p->channel);
#endif
						ast_queue_hangup_with_cause(p->subs[ANALOG_SUB_THREEWAY].owner, AST_CAUSE_NO_ANSWER);
						ast_softhangup_nolock(p->subs[ANALOG_SUB_THREEWAY].owner, AST_SOFTHANGUP_DEV);
						ast_channel_unlock(p->subs[ANALOG_SUB_THREEWAY].owner);
					} else if ((ast_channel_pbx(ast)) || (ast_channel_state(ast) == AST_STATE_UP)) {
						if (p->transfer) {
#ifdef GRANDSTREAM_NETWORKS
							/* In any case this isn't a threeway call anymore */
							analog_set_inthreeway(p, ANALOG_SUB_REAL, 0);
							analog_set_inthreeway(p, ANALOG_SUB_THREEWAY, 0);

							/* Only attempt transfer if the phone is ringing; why transfer to busy tone eh? */
							if (real_inthreeway
							    || (p->transfertobusy && (AST_STATE_BUSY == ast_channel_state(ast) || AST_STATE_RINGING == ast_channel_state(ast))))
							{
								res = analog_attempt_transfer(p);
								if (res < 0) {
									/* Transfer attempt failed. */
									ast_softhangup_nolock(p->subs[ANALOG_SUB_THREEWAY].owner, AST_SOFTHANGUP_DEV);
									ast_channel_unlock(p->subs[ANALOG_SUB_THREEWAY].owner);
								} else if (res) {
									/* Don't actually hang up at this point */
									break;
								}
							}
							else
							{
								/* Swap subs and dis-own channel */
								analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
								/* Unlock the 3-way call that we swapped to real-call. */
								ast_channel_unlock(p->subs[ANALOG_SUB_REAL].owner);
								analog_set_new_owner(p, NULL);
            					analog_set_cadence(p, p->subs[ANALOG_SUB_REAL].owner); /* and set p->cidrings */
                        		if (p->use_callerid)
                                {
                                    if (ast_channel_connected(p->subs[ANALOG_SUB_REAL].owner)->id.name.valid)
                                    {
                                        ast_copy_string(p->lastcid_name, ast_channel_connected(p->subs[ANALOG_SUB_REAL].owner)->id.name.str, sizeof(p->lastcid_name));
                                        p->caller.id.name.str = p->lastcid_name;
                                    }

                                    if (ast_channel_connected(p->subs[ANALOG_SUB_REAL].owner)->id.number.valid)
                                    {
                                        ast_copy_string(p->lastcid_num, ast_channel_connected(p->subs[ANALOG_SUB_REAL].owner)->id.number.str, sizeof(p->lastcid_num));
                                        p->caller.id.number.str = p->lastcid_num;
                                    }
                                    else if(ast_channel_dialed(p->subs[ANALOG_SUB_REAL].owner)->number.str)
                                    {
                                        ast_copy_string(p->lastcid_num, ast_channel_dialed(p->subs[ANALOG_SUB_REAL].owner)->number.str, sizeof(p->lastcid_num));
                                        p->caller.id.number.str = p->lastcid_num;
                                    }
                                }
								/* Ring the phone */
								analog_ring(p);
							}
#else
							/* In any case this isn't a threeway call anymore */
							analog_set_inthreeway(p, ANALOG_SUB_REAL, 0);
							analog_set_inthreeway(p, ANALOG_SUB_THREEWAY, 0);

							/* Only attempt transfer if the phone is ringing; why transfer to busy tone eh? */
							if (!p->transfertobusy && ast_channel_state(ast) == AST_STATE_BUSY) {
								/* Swap subs and dis-own channel */
								analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
								/* Unlock the 3-way call that we swapped to real-call. */
								ast_channel_unlock(p->subs[ANALOG_SUB_REAL].owner);
								analog_set_new_owner(p, NULL);
								/* Ring the phone */
								analog_ring(p);
							} else if (!analog_attempt_transfer(p)) {
								/*
								 * Transfer successful.  Don't actually hang up at this point.
								 * Let our channel legs of the calls die off as the transfer
								 * percolates through the core.
								 */
								break;
							}
#endif
						} else {
							ast_softhangup_nolock(p->subs[ANALOG_SUB_THREEWAY].owner, AST_SOFTHANGUP_DEV);
							ast_channel_unlock(p->subs[ANALOG_SUB_THREEWAY].owner);
						}
					} else {
						/* Swap subs and dis-own channel */
						analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
						/* Unlock the 3-way call that we swapped to real-call. */
						ast_channel_unlock(p->subs[ANALOG_SUB_REAL].owner);
						analog_set_new_owner(p, NULL);
#ifdef GRANDSTREAM_NETWORKS
						analog_set_cadence(p, p->subs[ANALOG_SUB_REAL].owner); /* and set p->cidrings */
						if (p->use_callerid)
						{
							if (ast_channel_connected(p->subs[ANALOG_SUB_REAL].owner)->id.name.valid)
							{	 
								ast_copy_string(p->lastcid_name, ast_channel_connected(p->subs[ANALOG_SUB_REAL].owner)->id.name.str, sizeof(p->lastcid_name));
								p->caller.id.name.str = p->lastcid_name;
							}

							if (ast_channel_connected(p->subs[ANALOG_SUB_REAL].owner)->id.number.valid)
							{
								ast_copy_string(p->lastcid_num, ast_channel_connected(p->subs[ANALOG_SUB_REAL].owner)->id.number.str, sizeof(p->lastcid_num));
								p->caller.id.number.str = p->lastcid_num;
							}
							else if(ast_channel_dialed(p->subs[ANALOG_SUB_REAL].owner)->number.str)
							{
								ast_copy_string(p->lastcid_num, ast_channel_dialed(p->subs[ANALOG_SUB_REAL].owner)->number.str, sizeof(p->lastcid_num));
								p->caller.id.number.str = p->lastcid_num;
							}
						}
#endif
						/* Ring the phone */
						analog_ring(p);
					}
				}
			} else {
#ifdef GRANDSTREAM_NETWORKS
				ast_log(LOG_WARNING, "[ch:%d]Got a hangup and my index is %d?",p->channel, idx);
#else
				ast_log(LOG_WARNING, "Got a hangup and my index is %u?\n", idx);
#endif
			}
			/* Fall through */
		default:
			analog_set_echocanceller(p, 0);
			return NULL;
		}
		break;
	case ANALOG_EVENT_RINGOFFHOOK:
		if (p->inalarm) {
			break;
		}
		/* for E911, its supposed to wait for offhook then dial
		   the second half of the dial string */
		if (((mysig == ANALOG_SIG_E911) || (mysig == ANALOG_SIG_FGC_CAMA) || (mysig == ANALOG_SIG_FGC_CAMAMF)) && (ast_channel_state(ast) == AST_STATE_DIALING_OFFHOOK)) {
			c = strchr(p->dialdest, '/');
			if (c) {
				c++;
			} else {
				c = p->dialdest;
			}
			if (*c) {
				snprintf(p->dop.dialstr, sizeof(p->dop.dialstr), "M*0%s#", c);
			} else {
				ast_copy_string(p->dop.dialstr,"M*2#", sizeof(p->dop.dialstr));
			}
			if (strlen(p->dop.dialstr) > 4) {
				memset(p->echorest, 'w', sizeof(p->echorest) - 1);
				strcpy(p->echorest + (p->echotraining / 401) + 1, p->dop.dialstr + strlen(p->dop.dialstr) - 2);
				p->echorest[sizeof(p->echorest) - 1] = '\0';
				p->echobreak = 1;
				p->dop.dialstr[strlen(p->dop.dialstr)-2] = '\0';
			} else {
				p->echobreak = 0;
			}
			if (analog_dial_digits(p, ANALOG_SUB_REAL, &p->dop)) {
				analog_on_hook(p);
				return NULL;
			}
			analog_set_dialing(p, 1);
			return &p->subs[idx].f;
		}
		switch (p->sig) {
		case ANALOG_SIG_FXOLS:
		case ANALOG_SIG_FXOGS:
		case ANALOG_SIG_FXOKS:
			p->fxsoffhookstate = 1;
			switch (ast_channel_state(ast)) {
			case AST_STATE_RINGING:
				analog_set_echocanceller(p, 1);
				analog_train_echocanceller(p);
				p->subs[idx].f.frametype = AST_FRAME_CONTROL;
				p->subs[idx].f.subclass.integer = AST_CONTROL_ANSWER;
				/* Make sure it stops ringing */
				analog_set_needringing(p, 0);
				analog_off_hook(p);
#ifdef GRANDSTREAM_NETWORKS
				analog_set_cw_send_voice(p,1);
#endif
				ast_debug(1, "channel %d answered\n", p->channel);

				/* Cancel any running CallerID spill */
				analog_cancel_cidspill(p);

				analog_set_dialing(p, 0);
				p->callwaitcas = 0;
				if (analog_check_confirmanswer(p)) {
					/* Ignore answer if "confirm answer" is enabled */
					p->subs[idx].f.frametype = AST_FRAME_NULL;
					p->subs[idx].f.subclass.integer = 0;
				} else if (!ast_strlen_zero(p->dop.dialstr)) {
					/* nick@dccinc.com 4/3/03 - fxo should be able to do deferred dialing */
					res = analog_dial_digits(p, ANALOG_SUB_REAL, &p->dop);
					if (res) {
						p->dop.dialstr[0] = '\0';
						return NULL;
					} else {
						ast_debug(1, "Sent FXO deferred digit string: %s\n", p->dop.dialstr);
						p->subs[idx].f.frametype = AST_FRAME_NULL;
						p->subs[idx].f.subclass.integer = 0;
						analog_set_dialing(p, 1);
					}
					p->dop.dialstr[0] = '\0';
					ast_setstate(ast, AST_STATE_DIALING);
				} else {
					ast_setstate(ast, AST_STATE_UP);
					analog_answer_polarityswitch(p);
				}
				return &p->subs[idx].f;
			case AST_STATE_DOWN:
				ast_setstate(ast, AST_STATE_RING);
				ast_channel_rings_set(ast, 1);
				p->subs[idx].f.frametype = AST_FRAME_CONTROL;
				p->subs[idx].f.subclass.integer = AST_CONTROL_OFFHOOK;
				ast_debug(1, "channel %d picked up\n", p->channel);
				return &p->subs[idx].f;
			case AST_STATE_UP:
#ifdef GRANDSTREAM_NETWORKS
				analog_set_echocanceller(p, 1);
				analog_set_cw_send_voice(p,1);
				analog_set_ring_after_onhook(p,0);
				if(1 == analog_get_holdon(p,ANALOG_SUB_REAL))
				{
					ast_debug(1, "[ch:%d]%s is hold on\n", p->channel, ast_channel_name(p->owner));
					analog_play_moh(p, 1);
				}
				else
				{
					ast_debug(1, "[ch:%d]%s is not hold on\n", p->channel, ast_channel_name(p->owner));
					analog_play_moh(p, 0);
				}
				ast_log(LOG_NOTICE, "update state on channel %d\n", p->channel);
				ast_copy_string(device_name, ast_channel_name(p->owner), sizeof(device_name));
				dashptr = strrchr(device_name, '-');
				if (dashptr) {
					*dashptr = '\0';
				}
				ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, AST_DEVSTATE_NOT_CACHABLE, device_name);
#endif
				/* Make sure it stops ringing */
				analog_off_hook(p);
				/* Okay -- probably call waiting */
				ast_queue_unhold(p->owner);
				break;
			case AST_STATE_RESERVED:
				/* Start up dialtone */
				if (analog_has_voicemail(p)) {
					res = analog_play_tone(p, ANALOG_SUB_REAL, ANALOG_TONE_STUTTER);
				} else {
					res = analog_play_tone(p, ANALOG_SUB_REAL, ANALOG_TONE_DIALTONE);
				}
				break;
			default:
#ifdef GRANDSTREAM_NETWORKS
				ast_log(LOG_WARNING, "[ch:%d]FXO phone off hook in weird state %d(%s)??",
					p->channel, ast_channel_state(ast), ast_state2str(ast_channel_state(ast)));
#else
				ast_log(LOG_WARNING, "FXO phone off hook in weird state %u??\n", ast_channel_state(ast));
#endif
			}
			break;
		case ANALOG_SIG_FXSLS:
		case ANALOG_SIG_FXSGS:
		case ANALOG_SIG_FXSKS:
#ifdef GRANDSTREAM_NETWORKS
			if(1 == p->ucBatLost)
			{
				ast_log(LOG_DEBUG, "[ch:%d]has receive BatLOST before,and the bat present event is not arrived,ignore this ringoffhook event!\n", p->channel);
				break;
			}
#endif
			if (ast_channel_state(ast) == AST_STATE_RING) {
				analog_set_ringtimeout(p, p->ringt_base);
			}

			/* Fall through */
		case ANALOG_SIG_EM:
		case ANALOG_SIG_EM_E1:
		case ANALOG_SIG_EMWINK:
		case ANALOG_SIG_FEATD:
		case ANALOG_SIG_FEATDMF:
		case ANALOG_SIG_FEATDMF_TA:
		case ANALOG_SIG_E911:
		case ANALOG_SIG_FGC_CAMA:
		case ANALOG_SIG_FGC_CAMAMF:
		case ANALOG_SIG_FEATB:
		case ANALOG_SIG_SF:
		case ANALOG_SIG_SFWINK:
		case ANALOG_SIG_SF_FEATD:
		case ANALOG_SIG_SF_FEATDMF:
		case ANALOG_SIG_SF_FEATB:
			switch (ast_channel_state(ast)) {
			case AST_STATE_PRERING:
				ast_setstate(ast, AST_STATE_RING);
				/* Fall through */
			case AST_STATE_DOWN:
			case AST_STATE_RING:
#ifdef GRANDSTREAM_NETWORKS
				ast_debug(1, "[ch:%d]Ring detected\n", p->channel);
#else
				ast_debug(1, "Ring detected\n");
#endif
				p->subs[idx].f.frametype = AST_FRAME_CONTROL;
				p->subs[idx].f.subclass.integer = AST_CONTROL_RING;
				break;
			case AST_STATE_RINGING:
			case AST_STATE_DIALING:
				if (p->outgoing) {
#ifdef GRANDSTREAM_NETWORKS
					ast_debug(1, "[ch:%d]Line answered\n", p->channel);
#else
					ast_debug(1, "Line answered\n");
#endif
					if (analog_check_confirmanswer(p)) {
						p->subs[idx].f.frametype = AST_FRAME_NULL;
						p->subs[idx].f.subclass.integer = 0;
					} else {
						p->subs[idx].f.frametype = AST_FRAME_CONTROL;
						p->subs[idx].f.subclass.integer = AST_CONTROL_ANSWER;
						ast_setstate(ast, AST_STATE_UP);
					}
					break;
				}
				/* Fall through */
			default:
				ast_log(LOG_WARNING, "Ring/Off-hook in strange state %u on channel %d\n", ast_channel_state(ast), p->channel);
				break;
			}
			break;
		default:
			ast_log(LOG_WARNING, "Don't know how to handle ring/off hook for signalling %d\n", p->sig);
			break;
		}
		break;
	case ANALOG_EVENT_RINGBEGIN:
		switch (p->sig) {
		case ANALOG_SIG_FXSLS:
		case ANALOG_SIG_FXSGS:
		case ANALOG_SIG_FXSKS:
			if (ast_channel_state(ast) == AST_STATE_RING) {
				analog_set_ringtimeout(p, p->ringt_base);
			}
#ifdef GRANDSTREAM_NETWORKS
			p->ucHasReceiveRing=1;
			ast_log(LOG_NOTICE, "Receive ringbegin on channel %d\n",p->channel);
#endif
			break;
		default:
			break;
		}
		break;
	case ANALOG_EVENT_RINGEROFF:
		if (p->inalarm) break;
		ast_channel_rings_set(ast, ast_channel_rings(ast) + 1);
		if (ast_channel_rings(ast) == p->cidrings) {
			analog_send_callerid(p, 0, &p->caller);
		}

		if (ast_channel_rings(ast) > p->cidrings) {
			analog_cancel_cidspill(p);
			p->callwaitcas = 0;
		}
#ifdef GRANDSTREAM_NETWORKS
		/*Start: chzhdeng  [ Bug 36039 ]  add */
		p->subs[idx].f.frametype = AST_FRAME_NULL;
		/*End  : chzhdeng  [ Bug 36039 ]  add */
#else
		p->subs[idx].f.frametype = AST_FRAME_CONTROL;
#endif
		p->subs[idx].f.subclass.integer = AST_CONTROL_RINGING;
		break;
	case ANALOG_EVENT_RINGERON:
		break;
	case ANALOG_EVENT_NOALARM:
		analog_set_alarm(p, 0);
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_NOTICE, "Alarm cleared on channel %d\n", p->channel);
#endif
		analog_publish_channel_alarm_clear(p->channel);
		break;
	case ANALOG_EVENT_WINKFLASH:
		if (p->inalarm) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_WARNING, "[ch:%d]When event:%d(%s), channel is inalarm:%u!",
				p->channel, res, analog_event2str(res), p->inalarm);
#endif
			break;
		}
		/* Remember last time we got a flash-hook */
		gettimeofday(&p->flashtime, NULL);
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_DEBUG, "[ch:%d]When ANALOG_EVENT_WINKFLASH, mysig is %d.", p->channel, mysig);
#endif
		switch (mysig) {
		case ANALOG_SIG_FXOLS:
		case ANALOG_SIG_FXOGS:
		case ANALOG_SIG_FXOKS:
#ifdef GRANDSTREAM_NETWORKS
			ast_debug(1, "[ch:%d]Winkflash, index:%d, real<owner:%p,fd:%d>, callwait:<owner:%p,fd:%d>, thirdcall:<owner:%p,fd:%d>",
				p->channel, idx, 
				p->subs[ANALOG_SUB_REAL].owner, analog_get_sub_fd(p, ANALOG_SUB_REAL), 
				p->subs[ANALOG_SUB_CALLWAIT].owner, analog_get_sub_fd(p, ANALOG_SUB_CALLWAIT), 
				p->subs[ANALOG_SUB_THREEWAY].owner, analog_get_sub_fd(p, ANALOG_SUB_THREEWAY));
#else
			ast_debug(1, "Winkflash, index: %u, normal: %d, callwait: %d, thirdcall: %d\n",
				idx, analog_get_sub_fd(p, ANALOG_SUB_REAL), analog_get_sub_fd(p, ANALOG_SUB_CALLWAIT), analog_get_sub_fd(p, ANALOG_SUB_THREEWAY));
#endif

			/* Cancel any running CallerID spill */
			analog_cancel_cidspill(p);
			p->callwaitcas = 0;

			if (idx != ANALOG_SUB_REAL) {
				ast_log(LOG_WARNING, "Got flash hook with index %u on channel %d?!?\n", idx, p->channel);
				goto winkflashdone;
			}
#ifdef GRANDSTREAM_NETWORKS
			if(p->subs[ANALOG_SUB_REAL].inthreeway || ast_channel_state(p->subs[ANALOG_SUB_REAL].owner)!=AST_STATE_UP || analog_check_for_conference(p)){
				if(p->subs[ANALOG_SUB_THREEWAY].owner && ast_channel_state(p->subs[ANALOG_SUB_THREEWAY].owner)==AST_STATE_UP){
					if(ast_channel_state(p->subs[ANALOG_SUB_REAL].owner) == AST_STATE_RESERVED){
						if(p->tone!=ANALOG_TONE_CONGESTION && p->tone!=ANALOG_TONE_BUSY){
							/* Need to hold the lock for real-call, private, and call-waiting call */
							analog_lock_sub_owner(p, ANALOG_SUB_THREEWAY);

							/* Swap to ANALOG_SUB_THREEWAY */
							analog_swap_subs(p, ANALOG_SUB_REAL, ANALOG_SUB_THREEWAY);
							analog_play_tone(p, ANALOG_SUB_REAL, -1);
							analog_set_echocanceller(p, 1);//added by zhywang for bug 50075
							analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
							ast_debug(1, "[ch:%d]Making %s the new owner\n", p->channel, ast_channel_name(p->owner) );
							ast_softhangup_nolock(p->subs[ANALOG_SUB_THREEWAY].owner, AST_SOFTHANGUP_DEV);

							ast_queue_hold(p->subs[ANALOG_SUB_THREEWAY].owner, p->mohsuggest);
							ast_queue_hold(p->subs[ANALOG_SUB_REAL].owner, p->mohsuggest);

							ast_queue_control(p->subs[ANALOG_SUB_REAL].owner, AST_CONTROL_UNHOLD);
							/* Unlock the call-waiting call that we swapped to real-call. */
							ast_channel_unlock(p->subs[ANALOG_SUB_REAL].owner);
						}
						else
							ast_log(LOG_NOTICE, "[ch:%d]Whoa,flash takes no effect except line is up or hold..", p->channel);
					}
					else
						ast_log(LOG_NOTICE, "[ch:%d]Whoa,flash takes no effect except line is up or hold..", p->channel);
				}
				else
				ast_log(LOG_NOTICE, "[ch:%d]Whoa,flash takes no effect except line is up or hold..", p->channel);

				goto winkflashdone;
			}

			if (p->subs[ANALOG_SUB_CALLWAIT].owner){
				if(ast_channel_state(p->subs[ANALOG_SUB_CALLWAIT].owner)!=AST_STATE_UP||!p->threewaycalling){
					/* Need to hold the lock for real-call, private, and call-waiting call */
					analog_lock_sub_owner(p, ANALOG_SUB_CALLWAIT);
					if (!p->subs[ANALOG_SUB_CALLWAIT].owner) {
						/*
						 * The call waiting call dissappeared.
						 * Let's just ignore this flash-hook.
						 */
						ast_log(LOG_NOTICE, "Whoa, the call-waiting call disappeared.\n");
						goto winkflashdone;
					}

				/* Swap to call-wait */
				analog_swap_subs(p, ANALOG_SUB_REAL, ANALOG_SUB_CALLWAIT);
				analog_play_tone(p, ANALOG_SUB_REAL, -1);
				analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
				ast_debug(1, "Making %s the new owner\n", ast_channel_name(p->owner));
				if (ast_channel_state(p->subs[ANALOG_SUB_REAL].owner) == AST_STATE_RINGING) {
					ast_setstate(p->subs[ANALOG_SUB_REAL].owner, AST_STATE_UP);
					ast_queue_control(p->subs[ANALOG_SUB_REAL].owner, AST_CONTROL_ANSWER);
				}
				analog_stop_callwait(p);

				/* Start music on hold if appropriate */
				if (!p->subs[ANALOG_SUB_CALLWAIT].inthreeway) {
					ast_queue_hold(p->subs[ANALOG_SUB_CALLWAIT].owner, p->mohsuggest);
				}
				ast_queue_hold(p->subs[ANALOG_SUB_REAL].owner, p->mohsuggest);
				ast_queue_unhold(p->subs[ANALOG_SUB_REAL].owner);

					/* Unlock the call-waiting call that we swapped to real-call. */
					ast_channel_unlock(p->subs[ANALOG_SUB_REAL].owner);
				}else{
					ast_debug(1, "[ch:%d]Building 3-way call with %s and %s",
							p->channel,
							ast_channel_name(p->subs[ANALOG_SUB_CALLWAIT].owner),
							ast_channel_name(p->subs[ANALOG_SUB_REAL].owner) );

					if (analog_alloc_sub(p, ANALOG_SUB_THREEWAY)){
						ast_log(LOG_WARNING, "[ch:%d]Unable to allocate three-way subchannel", p->channel);
						goto winkflashdone;
					}

					/* Need to hold the lock for real-call, private, and call-waiting call */
					analog_lock_sub_owner(p, ANALOG_SUB_CALLWAIT);
					analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_CALLWAIT);

					analog_unalloc_sub(p, ANALOG_SUB_CALLWAIT);

					/* Put them in the threeway, and flip */
					analog_set_inthreeway(p, ANALOG_SUB_THREEWAY, 1);
					analog_set_inthreeway(p, ANALOG_SUB_REAL, 1);

					analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);

					ast_queue_control(p->subs[ANALOG_SUB_REAL].owner, AST_CONTROL_UNHOLD);

					analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);

					/* Unlock the call-waiting call that we swapped to real-call. */
					ast_channel_unlock(p->subs[ANALOG_SUB_REAL].owner);
				}

				goto winkflashdone;
			}else if (p->subs[ANALOG_SUB_THREEWAY].owner){
				if (!p->threewaycalling){
					/* Need to hold the lock for real-call, private, and call-waiting call */
					analog_lock_sub_owner(p, ANALOG_SUB_THREEWAY);

					/* Swap to ANALOG_SUB_THREEWAY */
					analog_swap_subs(p, ANALOG_SUB_REAL, ANALOG_SUB_THREEWAY);
					analog_play_tone(p, ANALOG_SUB_REAL, -1);
					analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
					ast_debug(1, "[ch:%d]Making %s the new owner\n", p->channel, ast_channel_name(p->owner));

					ast_queue_hold(p->subs[ANALOG_SUB_THREEWAY].owner, p->mohsuggest);
					ast_queue_hold(p->subs[ANALOG_SUB_REAL].owner, p->mohsuggest);
					ast_queue_control(p->subs[ANALOG_SUB_REAL].owner, AST_CONTROL_UNHOLD);

					if(1 == analog_get_holdon(p,ANALOG_SUB_REAL)){
						ast_debug(1, "[ch:%d]%s is hold on\n", p->channel, ast_channel_name(p->owner));
						analog_play_moh(p, 1);
					}else{
						ast_debug(1, "[ch:%d]%s is not hold on\n", p->channel, ast_channel_name(p->owner));
						analog_play_moh(p, 0);
					}

					/* Unlock the call-waiting call that we swapped to real-call. */
					ast_channel_unlock(p->subs[ANALOG_SUB_REAL].owner);
				}else{
					/* Already have a 3 way call */
					enum analog_sub orig_3way_sub;

				/* Need to hold the lock for real-call, private, and 3-way call */
				analog_lock_sub_owner(p, ANALOG_SUB_THREEWAY);
				if (!p->subs[ANALOG_SUB_THREEWAY].owner) {
					/*
					 * The 3-way call dissappeared.
					 * Let's just ignore this flash-hook.
					 */
					ast_log(LOG_NOTICE, "Whoa, the 3-way call disappeared.\n");
					goto winkflashdone;
				}
				orig_3way_sub = ANALOG_SUB_THREEWAY;

					if (p->subs[ANALOG_SUB_THREEWAY].inthreeway) {
						/* Call is already up, drop the last person */
						ast_debug(1, "Got flash with three way call up, dropping last call on %d\n", p->channel);
						/* If the primary call isn't answered yet, use it */
						if ((ast_channel_state(p->subs[ANALOG_SUB_REAL].owner) != AST_STATE_UP) &&
							(ast_channel_state(p->subs[ANALOG_SUB_THREEWAY].owner) == AST_STATE_UP)) {
							/* Swap back -- we're dropping the real 3-way that isn't finished yet*/
							analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
							orig_3way_sub = ANALOG_SUB_REAL;
							analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
						}
						/* Drop the last call and stop the conference */
						ast_verb(3, "Dropping three-way call on %s\n", ast_channel_name(p->subs[ANALOG_SUB_THREEWAY].owner));
						ast_softhangup_nolock(p->subs[ANALOG_SUB_THREEWAY].owner, AST_SOFTHANGUP_DEV);
						analog_set_inthreeway(p, ANALOG_SUB_REAL, 0);
						analog_set_inthreeway(p, ANALOG_SUB_THREEWAY, 0);
					} else {
						/* Lets see what we're up to */
						if (((ast_channel_pbx(ast)) || (ast_channel_state(ast) == AST_STATE_UP)) &&
							(p->transfertobusy || (ast_channel_state(ast) != AST_STATE_BUSY))) {
							ast_verb(3, "Building conference call with %s and %s\n",
								ast_channel_name(p->subs[ANALOG_SUB_THREEWAY].owner),
								ast_channel_name(p->subs[ANALOG_SUB_REAL].owner));
							/* Put them in the threeway, and flip */
							analog_set_inthreeway(p, ANALOG_SUB_THREEWAY, 1);
							analog_set_inthreeway(p, ANALOG_SUB_REAL, 1);
							if (ast_channel_state(ast) == AST_STATE_UP) {
								analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
								orig_3way_sub = ANALOG_SUB_REAL;
							}
							ast_queue_unhold(p->subs[orig_3way_sub].owner);
							analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
						} else {
							ast_verb(3, "Dumping incomplete call on %s\n", ast_channel_name(p->subs[ANALOG_SUB_THREEWAY].owner));
							analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
							orig_3way_sub = ANALOG_SUB_REAL;
							ast_softhangup_nolock(p->subs[ANALOG_SUB_THREEWAY].owner, AST_SOFTHANGUP_DEV);
							analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
							ast_queue_unhold(p->subs[ANALOG_SUB_REAL].owner);
							analog_set_echocanceller(p, 1);
						}
					}
					ast_channel_unlock(p->subs[orig_3way_sub].owner);
				}
				goto winkflashdone;
			}else{
				struct ast_callid *callid = NULL;
				int callid_created;
				char cid_num[256];
				char cid_name[256];
				const char* real_hook_flag = NULL;

				cid_num[0] = '\0';
				cid_name[0] = '\0';

				/*Start: chzhdeng  [Bug 37676 ]  add */
				real_hook_flag = pbx_builtin_getvar_helper(p->subs[ANALOG_SUB_REAL].owner, "CALLED_REALHOOK");
				if (real_hook_flag){
					ast_debug(1," called real hook flag:%s \n",real_hook_flag);
					if ((strcmp(real_hook_flag, "no")==0) ||p->subs[ANALOG_SUB_THREEWAY].owner){
						ast_debug(1,"called still not off_hook!\n");
						goto winkflashdone;
					}
					/*End: chzhdeng  [Bug 37676 ]  add */
				}

				if (p->dahditrcallerid && p->owner) {
					if (ast_channel_caller(p->owner)->id.number.valid
						&& ast_channel_caller(p->owner)->id.number.str) {
						ast_copy_string(cid_num, ast_channel_caller(p->owner)->id.number.str,
							sizeof(cid_num));
					}
					if (ast_channel_caller(p->owner)->id.name.valid
						&& ast_channel_caller(p->owner)->id.name.str) {
						ast_copy_string(cid_name, ast_channel_caller(p->owner)->id.name.str,
							sizeof(cid_name));
					}
				}

				/* XXX This section needs much more error checking!!! XXX */
				/* Start a 3-way call if feasible */
				if (!((ast_channel_pbx(ast)) ||
					(ast_channel_state(ast) == AST_STATE_UP) ||
					(ast_channel_state(ast) == AST_STATE_RING))) {
					ast_debug(1, "Flash when call not up or ringing\n");
					goto winkflashdone;
				}
				if (analog_alloc_sub(p, ANALOG_SUB_THREEWAY)) {
					ast_log(LOG_WARNING, "Unable to allocate three-way subchannel\n");
					goto winkflashdone;
				}

				callid_created = ast_callid_threadstorage_auto(&callid);

				/*
				 * Make new channel
				 *
				 * We cannot hold the p or ast locks while creating a new
				 * channel.
				 */
				chan = analog_new_ast_channel(p, AST_STATE_RESERVED, 0, ANALOG_SUB_THREEWAY, NULL);
				if (!chan) {
					ast_log(LOG_WARNING,
						"Cannot allocate new call structure on channel %d\n",
						p->channel);
					analog_unalloc_sub(p, ANALOG_SUB_THREEWAY);
					ast_callid_threadstorage_auto_clean(callid, callid_created);
					goto winkflashdone;
				}
				if (p->dahditrcallerid) {
					if (!p->origcid_num) {
						p->origcid_num = ast_strdup(p->cid_num);
					}
					if (!p->origcid_name) {
						p->origcid_name = ast_strdup(p->cid_name);
					}
					ast_copy_string(p->cid_num, cid_num, sizeof(p->cid_num));
					ast_copy_string(p->cid_name, cid_name, sizeof(p->cid_name));
				}
				/* Swap things around between the three-way and real call */
				analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
				/* Disable echo canceller for better dialing */
				analog_set_echocanceller(p, 0);
				res = analog_play_tone(p, ANALOG_SUB_REAL, ANALOG_TONE_DIALRECALL);
				if (res) {
					ast_log(LOG_WARNING, "Unable to start dial recall tone on channel %d\n", p->channel);
				}
				analog_set_new_owner(p, chan);
				p->ss_astchan = chan;
				if (ast_pthread_create_detached(&threadid, NULL, __analog_ss_thread, p)){
					ast_log(LOG_WARNING, "Unable to start simple switch on channel %d\n", p->channel);
					res = analog_play_tone(p, ANALOG_SUB_REAL, ANALOG_TONE_CONGESTION);
					analog_set_echocanceller(p, 1);
					ast_hangup(chan);
				} else{
						ast_log(LOG_DEBUG, "[ch:%d]Started three way call.", p->channel);

						/* Start music on hold */
						ast_queue_hold(p->subs[ANALOG_SUB_THREEWAY].owner, p->mohsuggest);
					}

				goto winkflashdone;
			}
#else
			if (p->subs[ANALOG_SUB_CALLWAIT].owner) {
				/* Need to hold the lock for real-call, private, and call-waiting call */
				analog_lock_sub_owner(p, ANALOG_SUB_CALLWAIT);
				if (!p->subs[ANALOG_SUB_CALLWAIT].owner) {
					/*
					 * The call waiting call dissappeared.
					 * Let's just ignore this flash-hook.
					 */
					ast_log(LOG_NOTICE, "Whoa, the call-waiting call disappeared.\n");
					goto winkflashdone;
				}

				/* Swap to call-wait */
				analog_swap_subs(p, ANALOG_SUB_REAL, ANALOG_SUB_CALLWAIT);
				analog_play_tone(p, ANALOG_SUB_REAL, -1);
				analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
				ast_debug(1, "Making %s the new owner\n", ast_channel_name(p->owner));
				if (ast_channel_state(p->subs[ANALOG_SUB_REAL].owner) == AST_STATE_RINGING) {
					ast_setstate(p->subs[ANALOG_SUB_REAL].owner, AST_STATE_UP);
					ast_queue_control(p->subs[ANALOG_SUB_REAL].owner, AST_CONTROL_ANSWER);
				}
				analog_stop_callwait(p);

				/* Start music on hold if appropriate */
				if (!p->subs[ANALOG_SUB_CALLWAIT].inthreeway) {
					ast_queue_hold(p->subs[ANALOG_SUB_CALLWAIT].owner, p->mohsuggest);
				}
				ast_queue_hold(p->subs[ANALOG_SUB_REAL].owner, p->mohsuggest);
				ast_queue_unhold(p->subs[ANALOG_SUB_REAL].owner);

				/* Unlock the call-waiting call that we swapped to real-call. */
				ast_channel_unlock(p->subs[ANALOG_SUB_REAL].owner);
			} else if (!p->subs[ANALOG_SUB_THREEWAY].owner) {
				if (!p->threewaycalling) {
					/* Just send a flash if no 3-way calling */
					ast_queue_control(p->subs[ANALOG_SUB_REAL].owner, AST_CONTROL_FLASH);
					goto winkflashdone;
				} else if (!analog_check_for_conference(p)) {
					struct ast_callid *callid = NULL;
					int callid_created;
					char cid_num[256];
					char cid_name[256];

					cid_num[0] = '\0';
					cid_name[0] = '\0';
					if (p->dahditrcallerid && p->owner) {
						if (ast_channel_caller(p->owner)->id.number.valid
							&& ast_channel_caller(p->owner)->id.number.str) {
							ast_copy_string(cid_num, ast_channel_caller(p->owner)->id.number.str,
								sizeof(cid_num));
						}
						if (ast_channel_caller(p->owner)->id.name.valid
							&& ast_channel_caller(p->owner)->id.name.str) {
							ast_copy_string(cid_name, ast_channel_caller(p->owner)->id.name.str,
								sizeof(cid_name));
						}
					}
					/* XXX This section needs much more error checking!!! XXX */
					/* Start a 3-way call if feasible */
					if (!((ast_channel_pbx(ast)) ||
						(ast_channel_state(ast) == AST_STATE_UP) ||
						(ast_channel_state(ast) == AST_STATE_RING))) {
						ast_debug(1, "Flash when call not up or ringing\n");
						goto winkflashdone;
					}
					if (analog_alloc_sub(p, ANALOG_SUB_THREEWAY)) {
						ast_log(LOG_WARNING, "Unable to allocate three-way subchannel\n");
						goto winkflashdone;
					}

					callid_created = ast_callid_threadstorage_auto(&callid);

					/*
					 * Make new channel
					 *
					 * We cannot hold the p or ast locks while creating a new
					 * channel.
					 */
					analog_unlock_private(p);
					ast_channel_unlock(ast);
					chan = analog_new_ast_channel(p, AST_STATE_RESERVED, 0, ANALOG_SUB_THREEWAY, NULL);
					ast_channel_lock(ast);
					analog_lock_private(p);
					if (!chan) {
						ast_log(LOG_WARNING,
							"Cannot allocate new call structure on channel %d\n",
							p->channel);
						analog_unalloc_sub(p, ANALOG_SUB_THREEWAY);
						ast_callid_threadstorage_auto_clean(callid, callid_created);
						goto winkflashdone;
					}
					if (p->dahditrcallerid) {
						if (!p->origcid_num) {
							p->origcid_num = ast_strdup(p->cid_num);
						}
						if (!p->origcid_name) {
							p->origcid_name = ast_strdup(p->cid_name);
						}
						ast_copy_string(p->cid_num, cid_num, sizeof(p->cid_num));
						ast_copy_string(p->cid_name, cid_name, sizeof(p->cid_name));
					}
					/* Swap things around between the three-way and real call */
					analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
					/* Disable echo canceller for better dialing */
					analog_set_echocanceller(p, 0);
					res = analog_play_tone(p, ANALOG_SUB_REAL, ANALOG_TONE_DIALRECALL);
					if (res) {
						ast_log(LOG_WARNING, "Unable to start dial recall tone on channel %d\n", p->channel);
					}
					analog_set_new_owner(p, chan);
					p->ss_astchan = chan;
					if (ast_pthread_create_detached(&threadid, NULL, __analog_ss_thread, p)) {
						ast_log(LOG_WARNING, "Unable to start simple switch on channel %d\n", p->channel);
						res = analog_play_tone(p, ANALOG_SUB_REAL, ANALOG_TONE_CONGESTION);
						analog_set_echocanceller(p, 1);
						ast_hangup(chan);
					} else {
						ast_verb(3, "Started three way call on channel %d\n", p->channel);

						/* Start music on hold */
						ast_queue_hold(p->subs[ANALOG_SUB_THREEWAY].owner, p->mohsuggest);
					}
					ast_callid_threadstorage_auto_clean(callid, callid_created);
				}
			} else {
				/* Already have a 3 way call */
				enum analog_sub orig_3way_sub;

				/* Need to hold the lock for real-call, private, and 3-way call */
				analog_lock_sub_owner(p, ANALOG_SUB_THREEWAY);
				if (!p->subs[ANALOG_SUB_THREEWAY].owner) {
					/*
					 * The 3-way call dissappeared.
					 * Let's just ignore this flash-hook.
					 */
					ast_log(LOG_NOTICE, "Whoa, the 3-way call disappeared.\n");
					goto winkflashdone;
				}
				orig_3way_sub = ANALOG_SUB_THREEWAY;

				if (p->subs[ANALOG_SUB_THREEWAY].inthreeway) {
					/* Call is already up, drop the last person */
					ast_debug(1, "Got flash with three way call up, dropping last call on %d\n", p->channel);
					/* If the primary call isn't answered yet, use it */
					if ((ast_channel_state(p->subs[ANALOG_SUB_REAL].owner) != AST_STATE_UP) &&
						(ast_channel_state(p->subs[ANALOG_SUB_THREEWAY].owner) == AST_STATE_UP)) {
						/* Swap back -- we're dropping the real 3-way that isn't finished yet*/
						analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
						orig_3way_sub = ANALOG_SUB_REAL;
						analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
					}
					/* Drop the last call and stop the conference */
					ast_verb(3, "Dropping three-way call on %s\n", ast_channel_name(p->subs[ANALOG_SUB_THREEWAY].owner));
					ast_softhangup_nolock(p->subs[ANALOG_SUB_THREEWAY].owner, AST_SOFTHANGUP_DEV);
					analog_set_inthreeway(p, ANALOG_SUB_REAL, 0);
					analog_set_inthreeway(p, ANALOG_SUB_THREEWAY, 0);
				} else {
					/* Lets see what we're up to */
					if (((ast_channel_pbx(ast)) || (ast_channel_state(ast) == AST_STATE_UP)) &&
						(p->transfertobusy || (ast_channel_state(ast) != AST_STATE_BUSY))) {
						ast_verb(3, "Building conference call with %s and %s\n",
							ast_channel_name(p->subs[ANALOG_SUB_THREEWAY].owner),
							ast_channel_name(p->subs[ANALOG_SUB_REAL].owner));
						/* Put them in the threeway, and flip */
						analog_set_inthreeway(p, ANALOG_SUB_THREEWAY, 1);
						analog_set_inthreeway(p, ANALOG_SUB_REAL, 1);
						if (ast_channel_state(ast) == AST_STATE_UP) {
							analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
							orig_3way_sub = ANALOG_SUB_REAL;
						}
						ast_queue_unhold(p->subs[orig_3way_sub].owner);
						analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
					} else {
						ast_verb(3, "Dumping incomplete call on %s\n", ast_channel_name(p->subs[ANALOG_SUB_THREEWAY].owner));
						analog_swap_subs(p, ANALOG_SUB_THREEWAY, ANALOG_SUB_REAL);
						orig_3way_sub = ANALOG_SUB_REAL;
						ast_softhangup_nolock(p->subs[ANALOG_SUB_THREEWAY].owner, AST_SOFTHANGUP_DEV);
						analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
						ast_queue_unhold(p->subs[ANALOG_SUB_REAL].owner);
						analog_set_echocanceller(p, 1);
					}
				}
				ast_channel_unlock(p->subs[orig_3way_sub].owner);
			}
#endif
winkflashdone:
			analog_update_conf(p);
			break;
		case ANALOG_SIG_EM:
		case ANALOG_SIG_EM_E1:
		case ANALOG_SIG_FEATD:
		case ANALOG_SIG_SF:
		case ANALOG_SIG_SFWINK:
		case ANALOG_SIG_SF_FEATD:
		case ANALOG_SIG_FXSLS:
		case ANALOG_SIG_FXSGS:
			if (p->dialing) {
				ast_debug(1, "Ignoring wink on channel %d\n", p->channel);
			} else {
#ifdef GRANDSTREAM_NETWORKS
				ast_debug(1, "Got wink in weird state %d(%s) on channel %d\n",
					ast_channel_state(ast), ast_state2str(ast_channel_state(ast)), p->channel);
#else
				ast_debug(1, "Got wink in weird state %u on channel %d\n", ast_channel_state(ast), p->channel);				
#endif
			}
			break;
		case ANALOG_SIG_FEATDMF_TA:
			switch (p->whichwink) {
			case 0:
				ast_debug(1, "ANI2 set to '%d' and ANI is '%s'\n", ast_channel_caller(p->owner)->ani2,
					S_COR(ast_channel_caller(p->owner)->ani.number.valid,
						ast_channel_caller(p->owner)->ani.number.str, ""));
				snprintf(p->dop.dialstr, sizeof(p->dop.dialstr), "M*%d%s#",
					ast_channel_caller(p->owner)->ani2,
					S_COR(ast_channel_caller(p->owner)->ani.number.valid,
						ast_channel_caller(p->owner)->ani.number.str, ""));
				break;
			case 1:
				ast_copy_string(p->dop.dialstr, p->finaldial, sizeof(p->dop.dialstr));
				break;
			case 2:
				ast_log(LOG_WARNING, "Received unexpected wink on channel of type ANALOG_SIG_FEATDMF_TA\n");
				return NULL;
			}
			p->whichwink++;
			/* Fall through */
		case ANALOG_SIG_FEATDMF:
		case ANALOG_SIG_E911:
		case ANALOG_SIG_FGC_CAMAMF:
		case ANALOG_SIG_FGC_CAMA:
		case ANALOG_SIG_FEATB:
		case ANALOG_SIG_SF_FEATDMF:
		case ANALOG_SIG_SF_FEATB:
		case ANALOG_SIG_EMWINK:
			/* FGD MF and EMWINK *Must* wait for wink */
#ifdef GRANDSTREAM_NETWORKS
			if(ANALOG_SIG_EMWINK == p->sig)
			{
				if(p->dwem_outband_call_dial_delay)
				{
					struct timeval when;
					if (p->dwEM_Dialdelay_timer_index >= 0)
					{
							ast_log(LOG_WARNING, "[ch:%d]EM Dial delay time is not null, delete it first!",
									p->channel);
							dahdi_timer_del(p->dwEM_Dialdelay_timer_index);
							p->dwEM_Dialdelay_timer_index = -1;
					}

					when.tv_sec = p->dwem_outband_call_dial_delay / 1000;
					when.tv_usec = p->dwem_outband_call_dial_delay % 1000 * 1000;
					when = ast_tvadd(ast_tvnow(), when);
					p->dwEM_Dialdelay_timer_index = dahdi_timer_add(&when, analog_proc_em_dial_delay_timeout, p);

					if (p->dwEM_Dialdelay_timer_index < 0)
					{
							ast_log(LOG_WARNING, "[ch:%d]When receive WinkFlash, start dial delay timer failed, dial directly!",
									p->channel);
					}
					else
					{
							ast_log(LOG_DEBUG, "[ch:%d]When receive WinkFlash, EM start dial delay timer successfully, wait %d ms to dial!",
									p->channel,p->dwem_outband_call_dial_delay);
							break;
					}
				}
			}
#endif
			if (!ast_strlen_zero(p->dop.dialstr)) {
				res = analog_dial_digits(p, ANALOG_SUB_REAL, &p->dop);
				if (res) {
					p->dop.dialstr[0] = '\0';
					return NULL;
				} else {
					ast_debug(1, "Sent deferred digit string on channel %d: %s\n", p->channel, p->dop.dialstr);
				}
			}
			p->dop.dialstr[0] = '\0';
			break;
		default:
			ast_log(LOG_WARNING, "Don't know how to handle ring/off hook for signalling %d\n", p->sig);
		}
		break;
	case ANALOG_EVENT_HOOKCOMPLETE:
		if (p->inalarm) break;
		if (analog_check_waitingfordt(p)) {
			break;
		}
		switch (mysig) {
		case ANALOG_SIG_FXSLS:  /* only interesting for FXS */
		case ANALOG_SIG_FXSGS:
		case ANALOG_SIG_FXSKS:
#ifdef GRANDSTREAM_NETWORKS
			if(p->dwfxo_outband_call_dial_delay)
			{
					struct timeval when;
					if (p->dwDialdelay_timer_index >= 0)
					{
							ast_log(LOG_WARNING, "[ch:%d] Dial delay time is not null, delete it first!",
									p->channel);
							dahdi_timer_del(p->dwDialdelay_timer_index);
							p->dwDialdelay_timer_index = -1;
					}

					when.tv_sec = p->dwfxo_outband_call_dial_delay / 1000;
					when.tv_usec = p->dwfxo_outband_call_dial_delay % 1000 * 1000;
					when = ast_tvadd(ast_tvnow(), when);
					p->dwDialdelay_timer_index = dahdi_timer_add(&when, analog_proc_dial_delay_timeout, p);

					if (p->dwDialdelay_timer_index < 0)
					{
							ast_log(LOG_WARNING, "[ch:%d]When Hookcomplete, start dial delay timer failed, dial directly!",
									p->channel);
					}
					else
					{
							ast_log(LOG_DEBUG, "[ch:%d]When Hookcomplete, start dial delay timer successfully, wait %d ms to dial!",
									p->channel,p->dwfxo_outband_call_dial_delay);
							break;
					}
			}
#endif
		case ANALOG_SIG_EM:
		case ANALOG_SIG_EM_E1:
		case ANALOG_SIG_EMWINK:
		case ANALOG_SIG_FEATD:
		case ANALOG_SIG_SF:
		case ANALOG_SIG_SFWINK:
		case ANALOG_SIG_SF_FEATD:
#ifdef GRANDSTREAM_NETWORKS
			if(ANALOG_SIG_EMWINK == p->sig && (-1 != p->dwEM_Dialdelay_timer_index))
			{
				ast_debug(1, "[ch:%d]EM daildely timer is still on, wait for timeout\n", p->channel);
				break;
			}
#endif
			if (!ast_strlen_zero(p->dop.dialstr)) {
				res = analog_dial_digits(p, ANALOG_SUB_REAL, &p->dop);
				if (res) {
					p->dop.dialstr[0] = '\0';
					return NULL;
				} else {
					ast_debug(1, "Sent deferred digit string on channel %d: %s\n", p->channel, p->dop.dialstr);
				}
			}
			p->dop.dialstr[0] = '\0';
			p->dop.op = ANALOG_DIAL_OP_REPLACE;
			break;
		case ANALOG_SIG_FEATDMF:
		case ANALOG_SIG_FEATDMF_TA:
		case ANALOG_SIG_E911:
		case ANALOG_SIG_FGC_CAMA:
		case ANALOG_SIG_FGC_CAMAMF:
		case ANALOG_SIG_FEATB:
		case ANALOG_SIG_SF_FEATDMF:
		case ANALOG_SIG_SF_FEATB:
			ast_debug(1, "Got hook complete in MF FGD, waiting for wink now on channel %d\n",p->channel);
			break;
		default:
			break;
		}
		break;
	case ANALOG_EVENT_POLARITY:
		/*
		 * If we get a Polarity Switch event, this could be
		 * due to line seizure, remote end connect or remote end disconnect.
		 *
		 * Check to see if we should change the polarity state and
		 * mark the channel as UP or if this is an indication
		 * of remote end disconnect.
		 */

		if (p->polarityonanswerdelay > 0) {
			/* check if event is not too soon after OffHook or Answer */
			if (ast_tvdiff_ms(ast_tvnow(), p->polaritydelaytv) > p->polarityonanswerdelay) {
				switch (ast_channel_state(ast)) {
				case AST_STATE_DIALING:			/*!< Digits (or equivalent) have been dialed */
				case AST_STATE_RINGING:			/*!< Remote end is ringing */
					if (p->answeronpolarityswitch) {
#ifdef GRANDSTREAM_NETWORKS
						ast_log(LOG_NOTICE, "[ch:%d]Answering on polarity switch!", p->channel);
#else
						ast_debug(1, "Answering on polarity switch! channel %d\n", p->channel);
#endif
						ast_setstate(p->owner, AST_STATE_UP);
						p->polarity = POLARITY_REV;
						if (p->hanguponpolarityswitch) {
							p->polaritydelaytv = ast_tvnow();
						}
					} else {
						ast_debug(1, "Ignore Answer on polarity switch, channel %d\n", p->channel);
					}
					break;

				case AST_STATE_RING:			/*!< Line is ringing */
#ifdef GRANDSTREAM_NETWORKS
					ast_log(LOG_NOTICE, "[ch:%d]Ignore polarity switch when line is ringing!", p->channel);
					break;
#endif
				case AST_STATE_UP:				/*!< Line is up */
					if (p->hanguponpolarityswitch) {
#ifdef GRANDSTREAM_NETWORKS
						if(!p->dwFXOhangupFlag)
						{
							struct timeval when;
							when.tv_sec  = 0;
							when.tv_usec = 100*1000;
							when = ast_tvadd(ast_tvnow(), when);

							p->dwFXOHangupDelayonPolarityTimeIndex = dahdi_timer_add(&when, analog_proc_FXO_Hangup_Delay_timeout, p);

							if (p->dwFXOHangupDelayonPolarityTimeIndex < 0)
							{
								ast_log(LOG_ERROR, "[ch:%d]start FXO Hangup delay timer failed!",p->channel);
							}
						}
#else
						ast_debug(1, "HangingUp on polarity switch! channel %d\n", p->channel);
						ast_queue_control_data(ast, AST_CONTROL_PVT_CAUSE_CODE, cause_code, data_size);
						ast_channel_hangupcause_hash_set(ast, cause_code, data_size);
						ast_softhangup(p->owner, AST_SOFTHANGUP_EXPLICIT);
						p->polarity = POLARITY_IDLE;
#endif
					} else {
						ast_debug(1, "Ignore Hangup on polarity switch, channel %d\n", p->channel);
					}
					break;

				case AST_STATE_DOWN:				/*!< Channel is down and available */
				case AST_STATE_RESERVED:			/*!< Channel is down, but reserved */
				case AST_STATE_OFFHOOK:				/*!< Channel is off hook */
				case AST_STATE_BUSY:				/*!< Line is busy */
				case AST_STATE_DIALING_OFFHOOK:		/*!< Digits (or equivalent) have been dialed while offhook */
				case AST_STATE_PRERING:				/*!< Channel has detected an incoming call and is waiting for ring */
				default:
					if (p->answeronpolarityswitch || p->hanguponpolarityswitch) {
#ifdef GRANDSTREAM_NETWORKS
						ast_debug(1, "Ignoring Polarity switch on channel %d, state %d(%s)\n",
							p->channel, ast_channel_state(ast), ast_state2str(ast_channel_state(ast)));
#else
						ast_debug(1, "Ignoring Polarity switch on channel %d, state %u\n", p->channel, ast_channel_state(ast));
#endif
					}
					break;
				}

			} else {
				/* event is too soon after OffHook or Answer */
				switch (ast_channel_state(ast)) {
				case AST_STATE_DIALING:		/*!< Digits (or equivalent) have been dialed */
				case AST_STATE_RINGING:		/*!< Remote end is ringing */
					if (p->answeronpolarityswitch) {
						ast_debug(1, "Polarity switch detected but NOT answering (too close to OffHook event) on channel %d, state %u\n", p->channel, ast_channel_state(ast));
					}
					break;

				case AST_STATE_UP:			/*!< Line is up */
				case AST_STATE_RING:		/*!< Line is ringing */
					if (p->hanguponpolarityswitch) {
						ast_debug(1, "Polarity switch detected but NOT hanging up (too close to Answer event) on channel %d, state %u\n", p->channel, ast_channel_state(ast));
					}
					break;

				default:
					if (p->answeronpolarityswitch || p->hanguponpolarityswitch) {
						ast_debug(1, "Polarity switch detected (too close to previous event) on channel %d, state %u\n", p->channel, ast_channel_state(ast));
					}
					break;
				}
			}
		}

		/* Added more log_debug information below to provide a better indication of what is going on */
		ast_debug(1, "Polarity Reversal event occured - DEBUG 2: channel %d, state %u, pol= %d, aonp= %d, honp= %d, pdelay= %d, tv= %" PRIi64 "\n", p->channel, ast_channel_state(ast), p->polarity, p->answeronpolarityswitch, p->hanguponpolarityswitch, p->polarityonanswerdelay, ast_tvdiff_ms(ast_tvnow(), p->polaritydelaytv) );
		break;
#ifdef GRANDSTREAM_NETWORKS
#ifdef CURRENT_DISCONNECT
    case ANALOG_EVENT_BATTERY_PRESENT:
    {
		p->ucBatLost = 0;
		if (p->cd_timer_index < 0)
        {
            ast_log(LOG_NOTICE, "[ch:%d]When BATTERY_PRESENT occur, timer(%d) not exist, maybe is recovery event!",
                p->channel, p->cd_timer_index);
        }
        else
        {
            ast_log(LOG_NOTICE, "[ch:%d]When event BATTERY_PRESENT, cancel current disconnect detect, del timer(%d).",
                p->channel, p->cd_timer_index);
                
            dahdi_timer_del(p->cd_timer_index);
            p->cd_timer_index = -1;
        }
        
        break;
    }
    
    case ANALOG_EVENT_BATTERY_LOST:
    {
        struct timeval when;
        int threshold;

		p->ucBatLost = 1;
        if (p->currentdisconnectthreshold <= 0)
        {
            ast_log(LOG_NOTICE, "[ch:%d]When ANALOG_EVENT_BATTERY_LOST, Current Disconnect detect(%d) is off, ignore it!",
                p->channel, p->currentdisconnectthreshold);
            break;
        }

        if (p->cd_timer_index >= 0)
        {
            ast_log(LOG_WARNING, "[ch:%d]Duplicate ANALOG_EVENT_BATTERY_LOST, ingore it!",
                p->channel);
        }
        else if (p->owner)
        {
            threshold = p->currentdisconnectthreshold - 3;
            if (threshold < 1)
            {
                threshold = 1;
            }
        
            when.tv_sec  = threshold / 1000;
            when.tv_usec = threshold % 1000 * 1000;
            when = ast_tvadd(ast_tvnow(), when);

            p->cd_timer_index = dahdi_timer_add(&when, analog_proc_battlost_timeout, p);

            if (p->cd_timer_index < 0)
            {
                ast_log(LOG_ERROR, "[ch:%d]When battery lost, start timer failed!",
                    p->channel);
            }
        }

        break;
    }
#endif
#endif
	default:
		ast_debug(1, "Dunno what to do with event %d on channel %d\n", res, p->channel);
	}
	return &p->subs[idx].f;
}

struct ast_frame *analog_exception(struct analog_pvt *p, struct ast_channel *ast)
{
	int res;
	int idx;
	struct ast_frame *f;
#ifdef GRANDSTREAM_NETWORKS
	char device_name[AST_CHANNEL_NAME];
	char *dashptr;
#endif

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "[ch:%d]analog_exception(%p, %p) ...", p->channel, p, ast);
#else
	ast_debug(1, "%s %d\n", __FUNCTION__, p->channel);
#endif

	idx = analog_get_index(ast, p, 1);
	if (idx < 0) {
		idx = ANALOG_SUB_REAL;
	}

	p->subs[idx].f.frametype = AST_FRAME_NULL;
	p->subs[idx].f.datalen = 0;
	p->subs[idx].f.samples = 0;
	p->subs[idx].f.mallocd = 0;
	p->subs[idx].f.offset = 0;
	p->subs[idx].f.subclass.integer = 0;
	p->subs[idx].f.delivery = ast_tv(0,0);
	p->subs[idx].f.src = "dahdi_exception";
	p->subs[idx].f.data.ptr = NULL;

	if (!p->owner) {
		/* If nobody owns us, absorb the event appropriately, otherwise
		   we loop indefinitely.  This occurs when, during call waiting, the
		   other end hangs up our channel so that it no longer exists, but we
		   have neither FLASH'd nor ONHOOK'd to signify our desire to
		   change to the other channel. */
		res = analog_get_event(p);
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_DEBUG, "[ch:%d]Get event:%d(%s) ", p->channel, res, analog_event2str(res));
#endif

		/* Switch to real if there is one and this isn't something really silly... */
		if ((res != ANALOG_EVENT_RINGEROFF)
			&& (res != ANALOG_EVENT_RINGERON) 
#ifdef GRANDSTREAM_NETWORKS
			&& (res != ANALOG_EVENT_WINKFLASH)
			&& (res != ANALOG_EVENT_PULSE_START)
			&& ((res & 0xFFFF0000) != (ANALOG_EVENT_PULSEDIGIT))
			&& ((res & 0xFFFF0000) != (ANALOG_EVENT_DTMFDOWN))
			&& ((res & 0xFFFF0000) != (ANALOG_EVENT_DTMFUP))
#endif
			&& (res != ANALOG_EVENT_HOOKCOMPLETE)) 
		{
			ast_debug(1, "Restoring owner of channel %d on event %d\n", p->channel, res);
			analog_set_new_owner(p, p->subs[ANALOG_SUB_REAL].owner);
			if (p->owner && ast != p->owner) {
				/*
				 * Could this even happen?
				 * Possible deadlock because we do not have the real-call lock.
				 */
				ast_log(LOG_WARNING, "Event %s on %s is not restored owner %s\n",
					analog_event2str(res), ast_channel_name(ast), ast_channel_name(p->owner));
			}
#ifdef GRANDSTREAM_NETWORKS
			if (ANALOG_EVENT_ONHOOK != res) /* add by wangjiafeng 20130412 for [Bug14815] */
			{
				if (p->owner) 
				{
					ast_queue_unhold(p->owner);
				}
			}
#else
			if (p->owner) 
			{
				ast_queue_unhold(p->owner);
			}
#endif			
		}
		switch (res) {
		case ANALOG_EVENT_ONHOOK:
			analog_set_echocanceller(p, 0);
#ifdef GRANDSTREAM_NETWORKS
			p->fxsoffhookstate = 0;
#endif
			if (p->owner) 
			{
#ifdef GRANDSTREAM_NETWORKS
				ast_verb(3, "Channel %s still has call, ringing phone. cname:%s, cnum:%s.",
					ast_channel_name(p->owner),
					ast_channel_connected(ast)->id.name.valid ? ast_channel_connected(ast)->id.name.str : "NULL",
					ast_channel_connected(ast)->id.number.valid ? ast_channel_connected(ast)->id.number.str : "NULL");
#else
				ast_verb(3, "Channel %s still has call, ringing phone\n", ast_channel_name(p->owner));
#endif

#ifdef GRANDSTREAM_NETWORKS
                if (p->use_callerid)
                {
					if (ast_channel_connected(ast)->id.name.valid)
                    {
                        ast_copy_string(p->lastcid_name, ast_channel_connected(ast)->id.name.str, sizeof(p->lastcid_name));
                        p->caller.id.name.str = p->lastcid_name;
                    }

                    if (ast_channel_connected(ast)->id.number.valid)
                    {
                        ast_copy_string(p->lastcid_num, ast_channel_connected(ast)->id.number.str, sizeof(p->lastcid_num));
                        p->caller.id.number.str = p->lastcid_num;
                    }
                    else if(ast_channel_dialed(p->subs[ANALOG_SUB_REAL].owner)->number.str)
                    {
                        ast_copy_string(p->lastcid_num, ast_channel_dialed(p->subs[ANALOG_SUB_REAL].owner)->number.str, sizeof(p->lastcid_num));
                        p->caller.id.number.str = p->lastcid_num;
                    }
                }
				analog_set_cadence(p, p->owner); /* and set p->cidrings */
				if(p->owner && (ast_channel_state(p->owner) == AST_STATE_UP))
				{
				   analog_set_ring_after_onhook(p,1);	
				}
				ast_log(LOG_NOTICE, "update state on channel %d\n", p->channel);
            	ast_copy_string(device_name, ast_channel_name(p->owner), sizeof(device_name));
            	dashptr = strrchr(device_name, '-');
            	if (dashptr) {
            		*dashptr = '\0';
            	}
            	ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, AST_DEVSTATE_NOT_CACHABLE, device_name);
#endif
				analog_ring(p);
				analog_stop_callwait(p);
			} else {
				ast_log(LOG_WARNING, "Absorbed %s, but nobody is left!?!?\n",
					analog_event2str(res));
			}
			analog_update_conf(p);
			break;
		case ANALOG_EVENT_RINGOFFHOOK:
			analog_set_echocanceller(p, 1);
			analog_off_hook(p);
#ifdef GRANDSTREAM_NETWORKS
			p->fxsoffhookstate = 1;
			/* Set refresh device state */
			analog_set_ring_after_onhook(p,0);
			analog_set_cw_send_voice(p,1);
			p->reringswhenhanguped = 0;
			if(1 == analog_get_holdon(p,ANALOG_SUB_REAL))
			{
				ast_debug(1, "[ch:%d]%s is hold on\n", p->channel, ast_channel_name(p->owner));
				analog_play_moh(p, 1);
			}
			else
			{
				ast_debug(1, "[ch:%d]%s is not hold on\n", p->channel, ast_channel_name(p->owner));
				analog_play_moh(p, 0);
			}
			if(p->owner && (ast_channel_state(p->owner) == AST_STATE_UP))
			{
				ast_log(LOG_NOTICE, "update state on channel %d\n", p->channel);
				ast_copy_string(device_name, ast_channel_name(p->owner), sizeof(device_name));
				dashptr = strrchr(device_name, '-');
				if (dashptr) {
					*dashptr = '\0';
				}
				ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, AST_DEVSTATE_NOT_CACHABLE, device_name);
			}
#endif
			if (p->owner && (ast_channel_state(p->owner) == AST_STATE_RINGING)) {
				ast_queue_control(p->owner, AST_CONTROL_ANSWER);
				analog_set_dialing(p, 0);
			}
			break;
		case ANALOG_EVENT_HOOKCOMPLETE:
		case ANALOG_EVENT_RINGERON:
#ifdef GRANDSTREAM_NETWORKS
			break;
#endif
		case ANALOG_EVENT_RINGEROFF:
#ifdef GRANDSTREAM_NETWORKS
			if(p->reringwhenhanguped!=1)
			{		
				analog_set_ring_after_onhook(p,1);
				ast_log(LOG_NOTICE, "update state on channel %d\n", p->channel);
				ast_copy_string(device_name, ast_channel_name(p->subs[ANALOG_SUB_REAL].owner), sizeof(device_name));
				dashptr = strrchr(device_name, '-');
				if (dashptr) {
					*dashptr = '\0';
				}
				ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, AST_DEVSTATE_NOT_CACHABLE, device_name);
			}

			p->reringswhenhanguped++;
			ast_debug(1, "[ch:%d]cidrings=%d,reringswhenhanguped=%d", p->channel,p->cidrings,p->reringswhenhanguped);
			if (p->reringswhenhanguped == p->cidrings) 
			{
				analog_send_callerid(p, 0, &p->caller);
			}
			if (p->reringswhenhanguped > p->cidrings)
			{
				analog_cancel_cidspill(p);
				p->callwaitcas = 0;
			}

#endif
			break;
		case ANALOG_EVENT_WINKFLASH:
			gettimeofday(&p->flashtime, NULL);
			if (p->owner) {
				ast_verb(3, "Channel %d flashed to other channel %s\n", p->channel, ast_channel_name(p->owner));
				if (ast_channel_state(p->owner) != AST_STATE_UP) {
					/* Answer if necessary */
					ast_queue_control(p->owner, AST_CONTROL_ANSWER);
					ast_setstate(p->owner, AST_STATE_UP);
				}
				analog_stop_callwait(p);
				ast_queue_unhold(p->owner);
			} else {
				ast_log(LOG_WARNING, "Absorbed %s, but nobody is left!?!?\n",
					analog_event2str(res));
			}
			analog_update_conf(p);
			break;
		default:
			ast_log(LOG_WARNING, "Don't know how to absorb event %s\n", analog_event2str(res));
			break;
		}
		f = &p->subs[idx].f;
		return f;
	}
	ast_debug(1, "Exception on %d, channel %d\n", ast_channel_fd(ast, 0), p->channel);
	/* If it's not us, return NULL immediately */
	if (ast != p->owner) {
		ast_log(LOG_WARNING, "We're %s, not %s\n", ast_channel_name(ast), ast_channel_name(p->owner));
		f = &p->subs[idx].f;
		return f;
	}

	f = __analog_handle_event(p, ast);
	if (!f) {
		const char *name = ast_strdupa(ast_channel_name(ast));

		/* Tell the CDR this DAHDI device hung up */
		analog_unlock_private(p);
		ast_channel_unlock(ast);
		ast_set_hangupsource(ast, name, 0);
		ast_channel_lock(ast);
		analog_lock_private(p);
	}
	return f;
}

#ifdef GRANDSTREAM_NETWORKS
extern int g_svip_fd ;
extern int dahdi_svip_get_cid(int svip_fd, int chan, struct callerid_state *cid);
#endif

void *analog_handle_init_event(struct analog_pvt *i, int event)
{
	int res;
	pthread_t threadid;
	struct ast_channel *chan;
	struct ast_callid *callid = NULL;
	int callid_created;
#ifdef GRANDSTREAM_NETWORKS
	char device_name[AST_CHANNEL_NAME]={0};
#endif

	ast_debug(1, "channel (%d) - signaling (%d) - event (%s)\n",
				i->channel, i->sig, analog_event2str(event));

	/* Handle an event on a given channel for the monitor thread. */
	switch (event) {
	case ANALOG_EVENT_WINKFLASH:
	case ANALOG_EVENT_RINGOFFHOOK:
		if (i->inalarm) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_NOTICE, "[ch:%d]channel in alarm:%d, signaling:%d(%s) event:%d(%s); break ",
				i->channel, i->inalarm, i->sig, analog_sigtype_to_str(i->sig), event, analog_event2str(event));
#endif
			break;
		}
		/* Got a ring/answer.  What kind of channel are we? */
		switch (i->sig) {
		case ANALOG_SIG_FXOLS:
		case ANALOG_SIG_FXOGS:
		case ANALOG_SIG_FXOKS:
			res = analog_off_hook(i);
			i->fxsoffhookstate = 1;
			if (res && (errno == EBUSY)) {
				break;
			}
			callid_created = ast_callid_threadstorage_auto(&callid);

			/* Cancel VMWI spill */
#ifdef GRANDSTREAM_NETWORKS
			analog_set_vmwi(i, 0);
#endif
			analog_cancel_cidspill(i);
#ifdef GRANDSTREAM_NETWORKS
			if (i->immediate || i->hotline_immediate) {
#else
			if (i->immediate) {
#endif
				analog_set_echocanceller(i, 1);
				/* The channel is immediately up.  Start right away */
				res = analog_play_tone(i, ANALOG_SUB_REAL, ANALOG_TONE_RINGTONE);
#ifdef GRANDSTREAM_NETWORKS
				analog_set_exten(i,i->hotline_exten);
#endif
				chan = analog_new_ast_channel(i, AST_STATE_RING, 1, ANALOG_SUB_REAL, NULL);
				if (!chan) {
					ast_log(LOG_WARNING, "Unable to start PBX on channel %d\n", i->channel);
					res = analog_play_tone(i, ANALOG_SUB_REAL, ANALOG_TONE_CONGESTION);
					if (res < 0) {
						ast_log(LOG_WARNING, "Unable to play congestion tone on channel %d\n", i->channel);
					}
				}
			} else {
				/* Check for callerid, digits, etc */
				chan = analog_new_ast_channel(i, AST_STATE_RESERVED, 0, ANALOG_SUB_REAL, NULL);
				i->ss_astchan = chan;
				if (chan) {
					if (analog_has_voicemail(i)) {
						res = analog_play_tone(i, ANALOG_SUB_REAL, ANALOG_TONE_STUTTER);
					} else {
						res = analog_play_tone(i, ANALOG_SUB_REAL, ANALOG_TONE_DIALTONE);
					}
					if (res < 0)
						ast_log(LOG_WARNING, "Unable to play dialtone on channel %d, do you have defaultzone and loadzone defined?\n", i->channel);

					if (ast_pthread_create_detached(&threadid, NULL, __analog_ss_thread, i)) {
						ast_log(LOG_WARNING, "Unable to start simple switch thread on channel %d\n", i->channel);
						res = analog_play_tone(i, ANALOG_SUB_REAL, ANALOG_TONE_CONGESTION);
						if (res < 0) {
							ast_log(LOG_WARNING, "Unable to play congestion tone on channel %d\n", i->channel);
						}
						ast_hangup(chan);
					}
				} else
					ast_log(LOG_WARNING, "Unable to create channel\n");
			}
			ast_callid_threadstorage_auto_clean(callid, callid_created);
			break;
		case ANALOG_SIG_FXSLS:
		case ANALOG_SIG_FXSGS:
		case ANALOG_SIG_FXSKS:
#ifdef GRANDSTREAM_NETWORKS
			if(1 == i->ucBatLost)
			{
				ast_log(LOG_DEBUG, "[ch:%d]has receive BatLOST before,and the bat present event is not arrived,ignore this ringoffhook event!\n", i->channel);
				break;
			}
#endif
			analog_set_ringtimeout(i, i->ringt_base);
			/* Fall through */
		case ANALOG_SIG_EMWINK:
		case ANALOG_SIG_FEATD:
		case ANALOG_SIG_FEATDMF:
		case ANALOG_SIG_FEATDMF_TA:
		case ANALOG_SIG_E911:
		case ANALOG_SIG_FGC_CAMA:
		case ANALOG_SIG_FGC_CAMAMF:
		case ANALOG_SIG_FEATB:
		case ANALOG_SIG_EM:
		case ANALOG_SIG_EM_E1:
		case ANALOG_SIG_SFWINK:
		case ANALOG_SIG_SF_FEATD:
		case ANALOG_SIG_SF_FEATDMF:
		case ANALOG_SIG_SF_FEATB:
		case ANALOG_SIG_SF:
			callid_created = ast_callid_threadstorage_auto(&callid);
			/* Check for callerid, digits, etc */
			if (i->cid_start == ANALOG_CID_START_POLARITY_IN 
				|| i->cid_start == ANALOG_CID_START_DTMF_NOALERT
#ifdef GRANDSTREAM_NETWORKS
				|| i->cid_start == ANALOG_CID_START_POLARITY
#endif
				) 
			{
				chan = analog_new_ast_channel(i, AST_STATE_PRERING, 0, ANALOG_SUB_REAL, NULL);
			} else {
				chan = analog_new_ast_channel(i, AST_STATE_RING, 0, ANALOG_SUB_REAL, NULL);
			}
			i->ss_astchan = chan;
			if (!chan) {
				ast_log(LOG_WARNING, "Cannot allocate new structure on channel %d\n", i->channel);
			} else if (ast_pthread_create_detached(&threadid, NULL, __analog_ss_thread, i)) {
				ast_log(LOG_WARNING, "Unable to start simple switch thread on channel %d\n", i->channel);
				res = analog_play_tone(i, ANALOG_SUB_REAL, ANALOG_TONE_CONGESTION);
				if (res < 0) {
					ast_log(LOG_WARNING, "Unable to play congestion tone on channel %d\n", i->channel);
				}
				ast_hangup(chan);
			}
#ifdef GRANDSTREAM_NETWORKS
			if(i->sig == ANALOG_SIG_EM)
			{
				res = analog_off_hook(i);
				ast_log(LOG_NOTICE, "[ch:%d]Took %s off hook, res:%d.", i->channel, ast_channel_name(chan), res);
				analog_play_tone(i, 0, ANALOG_TONE_DIALTONE);
			}
#endif
			ast_callid_threadstorage_auto_clean(callid, callid_created);
			break;
		default:
			ast_log(LOG_WARNING, "Don't know how to handle ring/answer with signalling %s on channel %d\n", analog_sigtype_to_str(i->sig), i->channel);
			res = analog_play_tone(i, ANALOG_SUB_REAL, ANALOG_TONE_CONGESTION);
			if (res < 0) {
				ast_log(LOG_WARNING, "Unable to play congestion tone on channel %d\n", i->channel);
			}
			return NULL;
		}
		break;
	case ANALOG_EVENT_NOALARM:
		analog_set_alarm(i, 0);
		analog_publish_channel_alarm_clear(i->channel);
		break;
	case ANALOG_EVENT_ALARM:
		analog_set_alarm(i, 1);
		analog_get_and_handle_alarms(i);
		/* fall thru intentionally */
	case ANALOG_EVENT_ONHOOK:
		/* Back on hook.  Hang up. */
		switch (i->sig) {
		case ANALOG_SIG_FXOLS:
		case ANALOG_SIG_FXOGS:
			i->fxsoffhookstate = 0;
			analog_start_polarityswitch(i);
			/* Fall through */
		case ANALOG_SIG_FEATD:
		case ANALOG_SIG_FEATDMF:
		case ANALOG_SIG_FEATDMF_TA:
		case ANALOG_SIG_E911:
		case ANALOG_SIG_FGC_CAMA:
		case ANALOG_SIG_FGC_CAMAMF:
		case ANALOG_SIG_FEATB:
		case ANALOG_SIG_EM:
		case ANALOG_SIG_EM_E1:
		case ANALOG_SIG_EMWINK:
		case ANALOG_SIG_SF_FEATD:
		case ANALOG_SIG_SF_FEATDMF:
		case ANALOG_SIG_SF_FEATB:
		case ANALOG_SIG_SF:
		case ANALOG_SIG_SFWINK:
		case ANALOG_SIG_FXSLS:
		case ANALOG_SIG_FXSGS:
		case ANALOG_SIG_FXSKS:
			analog_set_echocanceller(i, 0);
			res = analog_play_tone(i, ANALOG_SUB_REAL, -1);
			analog_on_hook(i);
#ifdef GRANDSTREAM_NETWORKS
			snprintf(device_name, sizeof(device_name), "DAHDI/%d", i->channel);
			ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, AST_DEVSTATE_NOT_CACHABLE, device_name);
#endif
			break;
		case ANALOG_SIG_FXOKS:
			i->fxsoffhookstate = 0;
			analog_start_polarityswitch(i);
			analog_set_echocanceller(i, 0);
#ifdef GRANDSTREAM_NETWORKS
			snprintf(device_name, sizeof(device_name), "DAHDI/%d", i->channel);
			ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, AST_DEVSTATE_NOT_CACHABLE, device_name);
#endif
			/* Diddle the battery for the zhone */
#ifdef ZHONE_HACK
			analog_off_hook(i);
			usleep(1);
#endif
			res = analog_play_tone(i, ANALOG_SUB_REAL, -1);
			analog_on_hook(i);
			break;
		default:
			ast_log(LOG_WARNING, "Don't know how to handle on hook with signalling %s on channel %d\n", analog_sigtype_to_str(i->sig), i->channel);
			res = analog_play_tone(i, ANALOG_SUB_REAL, -1);
			return NULL;
		}
		break;
#ifdef GRANDSTREAM_NETWORKS
	case ANALOG_EVENT_RINGBEGIN:
		switch (i->sig) 
		{
			case ANALOG_SIG_FXSLS:
			case ANALOG_SIG_FXSKS:
			case ANALOG_SIG_FXSGS:
				i->ucHasReceiveRing = 1;
				ast_log(LOG_NOTICE, "Receive ringbegin on channel %d\n",i->channel);
				break;
			default:
				break;
		}
		break;
#endif
	case ANALOG_EVENT_POLARITY:
		switch (i->sig) {
		case ANALOG_SIG_FXSLS:
		case ANALOG_SIG_FXSKS:
		case ANALOG_SIG_FXSGS:
			callid_created = ast_callid_threadstorage_auto(&callid);
			/* We have already got a PR before the channel was
			   created, but it wasn't handled. We need polarity
			   to be REV for remote hangup detection to work.
			   At least in Spain */
#ifdef GRANDSTREAM_NETWORKS
			if(1 == i->dweve_bat_pre_flag)
			{
				i->dweve_bat_pre_flag = 0;
				ast_log(LOG_WARNING, "The last eve_battery_present till now is : %" PRIi64 " ms\n", ast_tvdiff_ms(ast_tvnow(), i->tv_eve_bat_pre));
				if(20 > ast_tvdiff_ms(ast_tvnow(), i->tv_eve_bat_pre))
				{
					ast_log(LOG_WARNING, "The last eve_battery_present till now is : %" PRIi64 " ms\n", ast_tvdiff_ms(ast_tvnow(), i->tv_eve_bat_pre));
					return NULL;
				}
				else
				{
					ast_log(LOG_WARNING, "The last eve_battery_present till now is : %" PRIi64 " ms\n", ast_tvdiff_ms(ast_tvnow(), i->tv_eve_bat_pre));
				}
			}
			if(1 == i->dweve_bat_lost_flag)
			{
				i->dweve_bat_lost_flag = 0;
				ast_log(LOG_WARNING, "The last eve_battery_lost till now is : %" PRIi64 " ms\n", ast_tvdiff_ms(ast_tvnow(), i->tv_eve_bat_lost));
				if(200 > ast_tvdiff_ms(ast_tvnow(), i->tv_eve_bat_lost))
				{
					ast_log(LOG_WARNING, "The last eve_battery_lost till now is : %" PRIi64 " ms\n", ast_tvdiff_ms(ast_tvnow(), i->tv_eve_bat_lost));
					return NULL;
				}
				else
				{
					ast_log(LOG_WARNING, "The last eve_battery_lost till now is : %" PRIi64 " ms\n", ast_tvdiff_ms(ast_tvnow(), i->tv_eve_bat_lost));
				}
			}
			if(1 == i->ucHangupFlag)
			{
				i->ucHangupFlag = 0;
				ast_log(LOG_DEBUG, "The last hangup till now is : %" PRIi64 " ms\n", ast_tvdiff_ms(ast_tvnow(), i->tv_hangup));
				if(1300 > ast_tvdiff_ms(ast_tvnow(), i->tv_hangup))
				{
					ast_log(LOG_DEBUG, "Too close to hangup(smaller than 1300ms), Disgard this ANALOG_EVENT_POLARITY event\n");
					return NULL;
				}
			}
#endif
			if (i->hanguponpolarityswitch) {
				i->polarity = POLARITY_REV;
			}
			if (i->cid_start == ANALOG_CID_START_POLARITY 
				|| i->cid_start == ANALOG_CID_START_POLARITY_IN
#ifdef GRANDSTREAM_NETWORKS
				|| i->cid_start == ANALOG_CID_START_AUTO_DETECT
#endif
				) 
			{
				i->polarity = POLARITY_REV;
				ast_verb(2, "Starting post polarity "
					"CID detection on channel %d\n",
					i->channel);
				chan = analog_new_ast_channel(i, AST_STATE_PRERING, 0, ANALOG_SUB_REAL, NULL);
				i->ss_astchan = chan;
				if (!chan) {
					ast_log(LOG_WARNING, "Cannot allocate new structure on channel %d\n", i->channel);
				} else if (ast_pthread_create_detached(&threadid, NULL, __analog_ss_thread, i)) {
					ast_log(LOG_WARNING, "Unable to start simple switch thread on channel %d\n", i->channel);
					ast_hangup(chan);
				}
			}
			ast_callid_threadstorage_auto_clean(callid, callid_created);
			break;
		default:
			ast_log(LOG_WARNING,
				"handle_init_event detected polarity reversal on non-FXO (ANALOG_SIG_FXS) interface %d\n",
				i->channel);
			break;
		}
		break;
	case ANALOG_EVENT_DTMFCID:
		switch (i->sig) {
		case ANALOG_SIG_FXSLS:
		case ANALOG_SIG_FXSKS:
		case ANALOG_SIG_FXSGS:
			callid_created = ast_callid_threadstorage_auto(&callid);
			if (i->cid_start == ANALOG_CID_START_DTMF_NOALERT) {
				ast_verb(2, "Starting DTMF CID detection on channel %d\n",
					i->channel);
				chan = analog_new_ast_channel(i, AST_STATE_PRERING, 0, ANALOG_SUB_REAL, NULL);
				i->ss_astchan = chan;
				if (!chan) {
					ast_log(LOG_WARNING, "Cannot allocate new structure on channel %d\n", i->channel);
				} else if (ast_pthread_create_detached(&threadid, NULL, __analog_ss_thread, i)) {
					ast_log(LOG_WARNING, "Unable to start simple switch thread on channel %d\n", i->channel);
					ast_hangup(chan);
				}
			}
			ast_callid_threadstorage_auto_clean(callid, callid_created);
			break;
		default:
			ast_log(LOG_WARNING,
				"handle_init_event detected dtmfcid generation event on non-FXO (ANALOG_SIG_FXS) interface %d\n",
				i->channel);
			break;
		}
		break;
	case ANALOG_EVENT_REMOVED: /* destroy channel, will actually do so in do_monitor */
		ast_log(LOG_NOTICE, "Got ANALOG_EVENT_REMOVED. Destroying channel %d\n",
			i->channel);
		return i->chan_pvt;
	case ANALOG_EVENT_NEONMWI_ACTIVE:
		analog_handle_notify_message(NULL, i, -1, ANALOG_EVENT_NEONMWI_ACTIVE);
		break;
	case ANALOG_EVENT_NEONMWI_INACTIVE:
		analog_handle_notify_message(NULL, i, -1, ANALOG_EVENT_NEONMWI_INACTIVE);
		break;
#ifdef GRANDSTREAM_NETWORKS
#ifdef CURRENT_DISCONNECT
    case ANALOG_EVENT_BATTERY_PRESENT:
    {
		i->ucBatLost = 0;
		i->dweve_bat_lost_flag = 0;
		if (i->cd_timer_index < 0)
        {
            ast_log(LOG_NOTICE, "[ch:%d]When BATTERY_PRESENT occur, timer(%d) not exist, maybe is recovery event!",
                i->channel, i->cd_timer_index);
        }
        else
        {

            ast_log(LOG_NOTICE, "[ch:%d]When event BATTERY_PRESENT, cancel current disconnect detect, del timer(%d).",
                i->channel, i->cd_timer_index);
                
            dahdi_timer_del(i->cd_timer_index);
            i->cd_timer_index = -1;
        }
		gettimeofday(&i->tv_eve_bat_pre, NULL);
		i->dweve_bat_pre_flag = 1;
        break;
    }
    
    case ANALOG_EVENT_BATTERY_LOST:
    {
        struct timeval when;
        int threshold;
		
		gettimeofday(&i->tv_eve_bat_lost, NULL);
		i->dweve_bat_lost_flag = 1;
		i->ucBatLost = 1;
        if (i->currentdisconnectthreshold <= 0)
        {
            ast_log(LOG_NOTICE, "[ch:%d]When ANALOG_EVENT_BATTERY_LOST, Current Disconnect detect(%d) is off, ignore it!",
                i->channel, i->currentdisconnectthreshold);
            break;
        }

        if (i->cd_timer_index >= 0)
        {
            ast_log(LOG_WARNING, "[ch:%d]Duplicate ANALOG_EVENT_BATTERY_LOST, ingore it!",
                i->channel);
        }
        else if (i->owner)
        {
            threshold = i->currentdisconnectthreshold - 3;
            if (threshold < 1)
            {
                threshold = 1;
            }
        
            when.tv_sec  = threshold / 1000;
            when.tv_usec = threshold % 1000 * 1000;
            when = ast_tvadd(ast_tvnow(), when);

            i->cd_timer_index = dahdi_timer_add(&when, analog_proc_battlost_timeout, i);

            if (i->cd_timer_index < 0)
            {
                ast_log(LOG_ERROR, "[ch:%d]When battery lost, start timer failed!",
                    i->channel);
            }
        }
		else
			{
#if defined(SVIP_DETECT_DTMF)
				/* init dtmf detect */
				if (i->channel > 0 && i->channel <= ARRAY_LEN(g_svip_dtmf_digit))
				{
					ast_log(LOG_DEBUG, "[ch:%d]Clear DTMF detect buffer[%s].",
						i->channel, g_svip_dtmf_digit[i->channel - 1]);
					g_svip_dtmf_digit[i->channel - 1][0] = '\0';
				}
				else
				{
					ast_log(LOG_ERROR, "[ch:%d]Invalid DTMF channel clear failed, buffer index:%d, max index:%d!!",
						i->channel, i->channel	- 1, ARRAY_LEN(g_svip_dtmf_digit));
				}
#endif
		
#if defined(SVIP_DETECT_CID)
				if (ANALOG_SIG_FXSLS == i->sig || ANALOG_SIG_FXSGS == i->sig || ANALOG_SIG_FXSKS == i->sig)
				{
					int res;
					struct callerid_state cid;
					char buf[2048];
					int  buf_size = 0;
		
					memset(&cid, 0, sizeof(struct callerid_state));
		
					res = dahdi_svip_get_cid(g_svip_fd, i->channel, &cid);
		
					if (0 == res)
					{
						ast_log(LOG_DEBUG, "[ch:%d]Clear SVIP CID detect buffer.", i->channel);
					}
					else if (1 == res)
					{
						ast_log(LOG_NOTICE, "[ch:%d]When clear SVIP CID buffer, Got cid res[%d] number[%s] name[%s] len[%d] data(drop it):",
							i->channel, res, cid.number, cid.name, cid.len);
		
						buf_size += snprintf(buf + buf_size, sizeof(buf) - buf_size, 
							"\n=============================================\n");
							
						buf_size += ast_snprintf_binary(buf + buf_size, sizeof(buf) - buf_size, cid.rawdata, cid.len);
		
						buf_size += snprintf(buf + buf_size, sizeof(buf) - buf_size, 
							"\n=============================================\n");
							
						ast_log(LOG_NOTICE, "%s", buf);
					}
					else
					{
						ast_log(LOG_WARNING, "[ch:%d]Clear SVIP CID detect buffer failed, fd:%d!!",
							i->channel, g_svip_fd);
					}
				}
#endif
		}
        break;
    }
#endif
    
    default:
        break;
#endif
	}
	return NULL;
}


struct analog_pvt *analog_new(enum analog_sigtype signallingtype, void *private_data)
{
	struct analog_pvt *p;
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "analog_new(%u, %p) ...", signallingtype,  private_data);
#endif
	p = ast_calloc(1, sizeof(*p));
	if (!p) {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_ERROR, "When analog_new(%u, 0x%08zx), ast_calloc(1, %zu) failed!!\n", 
			signallingtype, (ptrdiff_t)private_data, sizeof(*p));
#endif
		return p;
	}

	p->outsigmod = ANALOG_SIG_NONE;
	p->sig = signallingtype;
	p->chan_pvt = private_data;

	/* Some defaults for values */
	p->cid_start = ANALOG_CID_START_RING;
	p->cid_signalling = CID_SIG_BELL;
	/* Sub real is assumed to always be alloc'd */
	p->subs[ANALOG_SUB_REAL].allocd = 1;
#ifdef GRANDSTREAM_NETWORKS
	p->tone = ANALOG_TONE_STOP;
	p->cd_timer_index = -1;
	p->dwDialdelay_timer_index = -1;
	p->dwEM_Dialdelay_timer_index = -1;
	p->dwFXOHangupDelayonPolarityTimeIndex = -1;
	p->dwFXOhangupFlag= 0;
#endif

	return p;
}

/*!
 * \brief Delete the analog private structure.
 * \since 1.8
 *
 * \param doomed Analog private structure to delete.
 *
 * \return Nothing
 */
void analog_delete(struct analog_pvt *doomed)
{
	ast_free(doomed);
}

int analog_config_complete(struct analog_pvt *p)
{
	/* No call waiting on non FXS channels */
	if ((p->sig != ANALOG_SIG_FXOKS) && (p->sig != ANALOG_SIG_FXOLS) && (p->sig != ANALOG_SIG_FXOGS)) {
		p->permcallwaiting = 0;
	}

	analog_set_callwaiting(p, p->permcallwaiting);

	return 0;
}

void analog_free(struct analog_pvt *p)
{
	ast_free(p);
}

/* called while dahdi_pvt is locked in dahdi_fixup */
int analog_fixup(struct ast_channel *oldchan, struct ast_channel *newchan, void *newp)
{
	struct analog_pvt *new_pvt = newp;
	int x;
	ast_debug(1, "New owner for channel %d is %s\n", new_pvt->channel, ast_channel_name(newchan));
	if (new_pvt->owner == oldchan) {
		analog_set_new_owner(new_pvt, newchan);
	}
	for (x = 0; x < 3; x++) {
		if (new_pvt->subs[x].owner == oldchan) {
			new_pvt->subs[x].owner = newchan;
		}
	}

	analog_update_conf(new_pvt);
	return 0;
}

static void analog_publish_dnd_state(int channel, const char *status)
{
	RAII_VAR(struct ast_json *, body, NULL, ast_json_unref);
	RAII_VAR(struct ast_str *, dahdichan, ast_str_create(32), ast_free);
	if (!dahdichan) {
		return;
	}

	ast_str_set(&dahdichan, 0, "DAHDI/%d", channel);

	body = ast_json_pack("{s: s, s: s}",
		"Channel", ast_str_buffer(dahdichan),
		"Status", status);
	if (!body) {
		return;
	}

	ast_manager_publish_event("DNDState", EVENT_FLAG_SYSTEM, body);
}

int analog_dnd(struct analog_pvt *p, int flag)
{
	if (flag == -1) {
		return p->dnd;
	}

	p->dnd = flag;

	ast_verb(3, "%s DND on channel %d\n",
			flag ? "Enabled" : "Disabled",
			p->channel);
	analog_publish_dnd_state(p->channel, flag ? "enabled" : "disabled");

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
const char* analog_cid_start2string(enum analog_cid_start cid_start)
{
    switch (cid_start)
    {
    case ANALOG_CID_START_RING:
        return "ring";
    case ANALOG_CID_START_POLARITY:
        return "polarity";
    case ANALOG_CID_START_POLARITY_IN:
        return "polarity_in";
    case ANALOG_CID_START_DTMF_NOALERT:
        return "dtmf";
    default:
        ast_log(LOG_ERROR, "Unknow analog_cid_start:%u.", cid_start);
        return "unknown";
    }
    
}

DahdiState analog_get_channel_state(struct analog_pvt *p)
{
	DahdiState state_result = DAHDI_STATE_UNKNOWN;
    if (NULL == p)
    {
        ast_log(LOG_ERROR, "%s(%p) parameter is NULL!!", __FUNCTION__, p);
        return DAHDI_STATE_UNKNOWN;
    }

    if (NULL == p->owner)
    {
        return DAHDI_STATE_IDLE;
    }

    if (AST_STATE_RESERVED == ast_channel_state(p->owner))
    {
    	switch (p->tone)
    	{
        case ANALOG_TONE_STUTTER:
            return DAHDI_STATE_DIALTONE;
        case ANALOG_TONE_CONGESTION:
            return DAHDI_STATE_HOWLER;
        case ANALOG_TONE_DIALTONE:
            return DAHDI_STATE_DIALTONE;
        case ANALOG_TONE_DIALRECALL:
            return DAHDI_STATE_DIALRECALL;
        case ANALOG_TONE_BUSY:
            return DAHDI_STATE_BUSY;
        case ANALOG_TONE_STOP:
            return DAHDI_STATE_DIALING;
        default:
            ast_log(LOG_WARNING, "[ch:%d]%s(%u) unknow analog tone, return Unknown!", p->channel, __FUNCTION__, p->tone);
            return DAHDI_STATE_UNKNOWN;
    	}
    }

	state_result = ast_state2dahdi_state(ast_channel_state(p->owner));

    return state_result;
}

DahdiState ast_state2dahdi_state(enum ast_channel_state state)
{

    switch (state)
    {
	    case AST_STATE_DIALING:
	        return DAHDI_STATE_DIALING;
	    case AST_STATE_RING:
	        return DAHDI_STATE_RING;
	    case AST_STATE_RINGING:
	        return DAHDI_STATE_RINGING;
	    case AST_STATE_UP:
	        return DAHDI_STATE_UP;
	    case AST_STATE_BUSY:
	        return DAHDI_STATE_BUSY;
	    default:
	        ast_log(LOG_WARNING, "%s(%u) unknow 'enum ast_channel_state'!", __FUNCTION__, state);
	        return DAHDI_STATE_UNKNOWN;
    }
}

const char* dahdi_state2string(DahdiState state)
{

    switch (state)
    {
	    case DAHDI_STATE_UNKNOWN:
	        return "Unknown";
	    case DAHDI_STATE_ALARM:
	        return "Alarm";
	    case DAHDI_STATE_BUSY:
	        return "Busy";
	    case DAHDI_STATE_DIALING:
	        return "Dialing";
	    case DAHDI_STATE_DIALRECALL:
	        return "Dialrecall";
	    case DAHDI_STATE_DIALTONE:
	        return "Dialtone";
	    case DAHDI_STATE_HOWLER:
	        return "Howler";
	    case DAHDI_STATE_IDLE:
	        return "Idle";
	    case DAHDI_STATE_RING:
	        return "Ring";
	    case DAHDI_STATE_RINGING:
	        return "Ringing";
	    case DAHDI_STATE_UP:
	        return "Up";
	    case DAHDI_STATE_CONFIG_ERR:
	        return "ErrConfigured";
	    case DAHDI_STATE_LOCAL_BLOCK:
	        return "LocalBlock";
	    case DAHDI_STATE_REMOTE_BLOCK:
	        return "RemoteBlock";
	    case DAHDI_STATE_BLOCK:
	        return "BidirectionalBlock";
	    default:
		    ast_log(LOG_WARNING, "%s(%u) Unknow DahdiState!", __FUNCTION__, state);
			return "Unknown";

    }
}

enum ast_device_state dahdi_state2device_state(DahdiState state)
{
    switch (state)
    {
    case DAHDI_STATE_UNKNOWN:
        return AST_DEVICE_UNKNOWN;
    case DAHDI_STATE_ALARM:
        return AST_DEVICE_INVALID;
    case DAHDI_STATE_BUSY:
        return AST_DEVICE_BUSY;
    case DAHDI_STATE_DIALING:
        return AST_DEVICE_INUSE;
    case DAHDI_STATE_DIALRECALL:
        return AST_DEVICE_ONHOLD;
    case DAHDI_STATE_DIALTONE:
        return AST_DEVICE_INUSE;
    case DAHDI_STATE_HOWLER:
        return AST_DEVICE_BUSY;
    case DAHDI_STATE_IDLE:
        return AST_DEVICE_NOT_INUSE;
    case DAHDI_STATE_RING:
        return AST_DEVICE_INUSE; 
    case DAHDI_STATE_RINGING:
        return AST_DEVICE_RINGING;
    case DAHDI_STATE_UP:
        return AST_DEVICE_INUSE;
    default:
	    ast_log(LOG_WARNING, "%s(%u) Unknow DahdiState!", __FUNCTION__, state);
		return AST_DEVICE_UNKNOWN;
    }
}

PortState dahdi_state2port_state(DahdiState state)
{
   switch (state)
    {
    case DAHDI_STATE_UNKNOWN:
        return PORT_STATE_IDLE;
    case DAHDI_STATE_ALARM:
        return PORT_STATE_DISCONNECT;
    case DAHDI_STATE_BUSY:
    case DAHDI_STATE_DIALING:
    case DAHDI_STATE_DIALRECALL:
    case DAHDI_STATE_DIALTONE:
    case DAHDI_STATE_HOWLER:
    case DAHDI_STATE_RING:
    case DAHDI_STATE_RINGING:
    case DAHDI_STATE_UP:
         return PORT_STATE_INUSE;
    case DAHDI_STATE_IDLE:
        return PORT_STATE_IDLE;
    default:
	    ast_log(LOG_WARNING, "%s(%u) Unknow DahdiState!", __FUNCTION__, state);
		return PORT_STATE_IDLE;
    }
}

const char* port_state2string(PortState state)
{
   switch (state)
    {
    case PORT_STATE_DISCONNECT:
        return "Disconnect";
    case PORT_STATE_IDLE:
        return "Idle";
    case PORT_STATE_INUSE:
        return "InUse";
    case PORT_STATE_UNCONFIGURED:
        return "UnConfigured";
    case (PORT_STATE_UNCONFIGURED|PORT_STATE_DISCONNECT):
        return "Disconnect and UnConfigured";
    default:
	    ast_log(LOG_WARNING, "%s(%u) Unknow PortState!", __FUNCTION__, state);
		return "Unknown";
    }
}

const char* analog_tone2string(enum analog_tone tone)
{
	switch (tone) {
	case ANALOG_TONE_RINGTONE:
		return "TONE_RING";
	case ANALOG_TONE_STUTTER:
		return "TONE_STUTTER";
	case ANALOG_TONE_CONGESTION:
		return "TONE_CONGESTION";
	case ANALOG_TONE_DIALTONE:
		return "TONE_DIALTONE";
	case ANALOG_TONE_DIALRECALL:
		return "TONE_DIALRECALL";
	case ANALOG_TONE_INFO:
		return "TONE_INFO";
	case ANALOG_TONE_BUSY:
	    return "TONE_BUSY";
	case ANALOG_TONE_STOP:
	    return "TONE_STOP";
	default:
	    ast_log(LOG_WARNING, "%s(%u) unknow analog tone, return Unknown!", __FUNCTION__, tone);
		return "Unknown";
	}
}

#ifdef CURRENT_DISCONNECT
static int analog_proc_battlost_timeout(void *data)
{
    struct analog_pvt *p = data;

    if (NULL == p)
    {
	    ast_log(AST_LOG_ERROR, "analog_proc_battlost_timeout(%p), parameter is NULL!!", p);
	    return -1;
    }

    ast_log(AST_LOG_NOTICE, "[ch:%d]Current disconnect timer(%d) timeout, hangup!",
         p->channel, p->cd_timer_index);
         
    p->cd_timer_index = -1;

    if (p->owner) 
    {
        ast_softhangup(p->owner, AST_SOFTHANGUP_DEV);
    }

    return 0;
}
#endif

#define dis_ascii(d) ((((d) & 0xFF) < 32 || ((d) & 0xFF) >=127) ? '.' : (d))

int ast_snprintf_binary(char *pBuf, int bufLen, const char *pData, int dataLen)
{
    int buf_size = 0;
    int u, v;

    for (v = 0; v <= dataLen - 8; v += 8)
    {
        buf_size += snprintf(pBuf + buf_size, bufLen - buf_size, 
            "  0x%04x | %02x%02x %02x%02x %02x%02x %02x%02x"
            "  %c%c%c%c%c%c%c%c\n",
            v,
            pData[v + 0] & 0xFF, pData[v + 1] & 0xFF,
            pData[v + 2] & 0xFF, pData[v + 3] & 0xFF,
            pData[v + 4] & 0xFF, pData[v + 5] & 0xFF,
            pData[v + 6] & 0xFF, pData[v + 7] & 0xFF,
            
            dis_ascii(pData[v + 0]), dis_ascii(pData[v + 1]),
            dis_ascii(pData[v + 2]), dis_ascii(pData[v + 3]),
            dis_ascii(pData[v + 4]), dis_ascii(pData[v + 5]),
            dis_ascii(pData[v + 6]), dis_ascii(pData[v + 7]));
    }

    if (dataLen <= v)
    {
        return buf_size;
    }

    buf_size += snprintf(pBuf + buf_size, bufLen - buf_size,
        "  0x%04x |", v);

    for (u = 0; u < 8; u++)
    {
        if (u % 2 == 0)
        {
            buf_size += snprintf(pBuf + buf_size, bufLen - buf_size, " ");
        }

        if (v + u < dataLen)
        {
            buf_size += snprintf(pBuf + buf_size, bufLen - buf_size,
                "%02x", pData[v + u] & 0xFF);
        }
        else
        {
            buf_size += snprintf(pBuf + buf_size, bufLen - buf_size, "  ");
        }
    }
    
    buf_size += snprintf(pBuf + buf_size, bufLen - buf_size, "  ");

    for (u = 0; u < 8; u++)
    {
        if (v + u < dataLen)
        {
            buf_size += snprintf(pBuf + buf_size, bufLen - buf_size,
                "%c", dis_ascii(pData[v + u]));
        }
        else
        {
            buf_size += snprintf(pBuf + buf_size, bufLen - buf_size, " ");
        }
    }

    return buf_size;
}

static int analog_proc_dial_delay_timeout(void *data)
{
	struct analog_pvt *p = data;
	int res;

	if (NULL == p)
	{
		ast_log(LOG_ERROR, "analog_proc_dial_delay_timeout(%p), parameter is NULL!!", p);
		return -1;
	}

	ast_log(LOG_NOTICE, "[ch:%d]fxo dial delay timer(%d) timeout, start dialing!",
		p->channel, p->dwDialdelay_timer_index);

	p->dwDialdelay_timer_index = -1;

	if (!ast_strlen_zero(p->dop.dialstr))
	{
		res = analog_dial_digits(p, ANALOG_SUB_REAL, &p->dop);
		if (res < 0)
		{
			ast_log(LOG_WARNING, "Unable to initiate dialing on trunk channel %d: %s\n", p->channel, strerror(errno));
			p->dop.dialstr[0] = '\0';
			return -1;
		}
		else
		{
			ast_debug(1, "Sent deferred digit string on channel %d: %s\n", p->channel, p->dop.dialstr);
		}
	}
	p->dop.dialstr[0] = '\0';
	p->dop.op = ANALOG_DIAL_OP_REPLACE;
	return 0;
}

static int analog_proc_em_dial_delay_timeout(void *data)
{
	struct analog_pvt *p = data;
	int res;

	if (NULL == p)
	{
		ast_log(LOG_ERROR, "analog_proc_em_dial_delay_timeout(%p), parameter is NULL!!", p);
		return -1;
	}

	ast_log(LOG_NOTICE, "[ch:%d]EM dial delay timer(%d) timeout, start dialing!",
		p->channel, p->dwEM_Dialdelay_timer_index);

	p->dwEM_Dialdelay_timer_index = -1;

	if (!ast_strlen_zero(p->dop.dialstr))
	{
		res = analog_dial_digits(p, ANALOG_SUB_REAL, &p->dop);
		if (res < 0)
		{
			ast_log(LOG_WARNING, "Unable to initiate dialing on trunk channel %d: %s\n", p->channel, strerror(errno));
			p->dop.dialstr[0] = '\0';
			return -1;
		}
		else
		{
			ast_debug(1, "Sent deferred digit string on channel %d: %s\n", p->channel, p->dop.dialstr);
		}
	}
	p->dop.dialstr[0] = '\0';
	p->dop.op = ANALOG_DIAL_OP_REPLACE;
	return 0;
}

static int analog_proc_FXO_Hangup_Delay_timeout(void *data)
{
    struct analog_pvt *p = data;

    if (NULL == p)
    {
	    ast_log(LOG_ERROR, "analog_proc_FXO_Hangup_Delay_timeout(%p), parameter is NULL!!", p);
	    return -1;
    }

    ast_log(LOG_NOTICE, "[ch:%d]FXO Hangup Delay timer(%d) timeout, hangup!",
         p->channel, p->dwFXOHangupDelayonPolarityTimeIndex);

    p->dwFXOHangupDelayonPolarityTimeIndex = -1;
	p->dwFXOhangupFlag = 0;

    if (p->owner)
    {
		ast_softhangup(p->owner, AST_SOFTHANGUP_EXPLICIT);
		p->polarity = POLARITY_IDLE;
    }

    return 0;
}
#endif
