/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 * Kevin Harwell <kharwell@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \author Joshua Colp <jcolp@digium.com>
 *
 * \brief SIP SDP media stream handling
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_session</depend>
	<support_level>core</support_level>
 ***/

#ifdef GRANDSTREAM_NETWORKS
#define AST_MODULE_LOG "pjsip"
#endif

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>
#include <pjmedia.h>
#include <pjlib.h>

ASTERISK_FILE_VERSION(__FILE__, "$Revision$")

#include "asterisk/module.h"
#include "asterisk/format.h"
#include "asterisk/format_cap.h"
#include "asterisk/rtp_engine.h"
#include "asterisk/netsock2.h"
#include "asterisk/channel.h"
#include "asterisk/causes.h"
#include "asterisk/sched.h"
#include "asterisk/acl.h"
#include "asterisk/sdp_srtp.h"
#include "asterisk/dsp.h"

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/pbx.h"
#include "asterisk/bridge.h"
#endif

/*! \brief Scheduler for RTCP purposes */
static struct ast_sched_context *sched;

/*! \brief Address for IPv4 RTP */
static struct ast_sockaddr address_ipv4;

/*! \brief Address for IPv6 RTP */
static struct ast_sockaddr address_ipv6;

static const char STR_AUDIO[] = "audio";
static const int FD_AUDIO = 0;

static const char STR_VIDEO[] = "video";
static const int FD_VIDEO = 2;

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
static const char STR_APPLICATION[] = "application";
static const int FD_APPLICATION_BFCP = 6;	/* BFCP FD */
static const int FD_SLIDES_VIDEO     = 7;	/* slides video fd, rtp:7 rtcp:8 */
static const int FD_APPLICATION_FEC  = 9;	/* fec fd, rtp:9 rtcp:10 */
static const int FD_APPLICATION_FECC = 11;	/* fec fd, rtp:11 rtcp:12 */

static int negotiate_incoming_application_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream);
static int negotiate_incoming_slides_video_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream);
static int create_outgoing_application_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp);
static int create_outgoing_slides_video_sdp_stream(struct ast_sip_session *session, struct pjmedia_sdp_session *sdp);
static int apply_negotiated_application_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
	const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream);
static int apply_negotiated_slides_video_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
	const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream);
static void create_outgoing_main_video_extra_attr_sdp(struct ast_sip_session *session, pj_pool_t *pool, pjmedia_sdp_media *media);
static void create_outgoing_slides_video_extra_attr_sdp(struct ast_sip_session *session, pj_pool_t *pool, pjmedia_sdp_media *media);
static void create_outgoing_disable_main_video_sdp(struct ast_sip_session *session, struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp);
static void set_video_extra_attr_para(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream);
static void set_multiplexed_fec_flag(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream);
static void set_app_other_attr_para(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream);
static int setup_extra_media_encryption(struct ast_sip_session *session, struct ast_sip_extra_media *extra_media,
	const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream);
static int add_crypto_to_extra_media_stream(struct ast_sip_session *session, struct ast_sip_extra_media *extra_media,
	pj_pool_t *pool, pjmedia_sdp_media *media);
#endif

/*! \brief Retrieves an ast_format_type based on the given stream_type */
static enum ast_media_type stream_to_media_type(const char *stream_type)
{
	if (!strcasecmp(stream_type, STR_AUDIO)) {
		return AST_MEDIA_TYPE_AUDIO;
	} else if (!strcasecmp(stream_type, STR_VIDEO)) {
		return AST_MEDIA_TYPE_VIDEO;
	}
#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	else if (!strcasecmp(stream_type, STR_APPLICATION)) {
		return AST_MEDIA_TYPE_APPLICATION;
	}
#endif
	return 0;
}

/*! \brief Get the starting descriptor for a media type */
static int media_type_to_fdno(enum ast_media_type media_type)
{
	switch (media_type) {
	case AST_MEDIA_TYPE_AUDIO: return FD_AUDIO;
	case AST_MEDIA_TYPE_VIDEO: return FD_VIDEO;
	case AST_MEDIA_TYPE_TEXT:
	case AST_MEDIA_TYPE_UNKNOWN:
	case AST_MEDIA_TYPE_IMAGE: break;
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	case AST_MEDIA_TYPE_APPLICATION: return FD_APPLICATION_BFCP;
#endif
	}
	return -1;
}

/*! \brief Remove all other cap types but the one given */
static void format_cap_only_type(struct ast_format_cap *caps, enum ast_media_type media_type)
{
	int i = 0;
	while (i <= AST_MEDIA_TYPE_TEXT) {
		if (i != media_type && i != AST_MEDIA_TYPE_UNKNOWN) {
			ast_format_cap_remove_by_type(caps, i);
		}
		i += 1;
	}
}

#ifdef GRANDSTREAM_NETWORKS
static int send_keepalive(const void *data)
{
	struct ast_sip_session_media *session_media = (struct ast_sip_session_media *) data;
	struct ast_rtp_instance *rtp = session_media->rtp;
	int keepalive;
	time_t interval;
	int send_keepalive;

	keepalive = ast_rtp_instance_get_keepalive(rtp);

	if (!ast_sockaddr_isnull(&session_media->direct_media_addr)) {
		ast_debug(3, "Not sending RTP keepalive on RTP instance %p since direct media is in use\n", rtp);
		return keepalive * 1000;
	}

	interval = time(NULL) - ast_rtp_instance_get_last_tx(rtp);
	send_keepalive = interval >= keepalive;

	ast_debug(3, "It has been %d seconds since RTP was last sent on instance %p. %sending keepalive\n",
			(int) interval, rtp, send_keepalive ? "S" : "Not s");

	if (send_keepalive) {
		ast_rtp_instance_sendcng(rtp, 0);
		return keepalive * 1000;
	}

	return (keepalive - interval) * 1000;
}

/*! \brief judge if video and audio all timeout */
static int  ast_video_audio_all_timeout(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	if (!channel || !channel->session) {
		ast_log(LOG_WARNING, "Chan has no session\n");
		return 0;
	}

	RAII_VAR(struct ast_sip_session_media *, video_session_media, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_session_media *, audio_session_media, NULL, ao2_cleanup);
	video_session_media = ao2_find(channel->session->media, "video", OBJ_KEY);
	audio_session_media = ao2_find(channel->session->media, "audio", OBJ_KEY);
	if (video_session_media) {
		struct ast_rtp_instance *rtp_v = video_session_media->rtp;
		if (rtp_v) {
			int interval_v = 0;
			interval_v = time(NULL) - ast_rtp_instance_get_last_rx(rtp_v);
			if (interval_v < ast_rtp_instance_get_timeout(rtp_v)) {
				return 1;
			}
		}
	}
	if (audio_session_media) {
		struct ast_rtp_instance *rtp_a = audio_session_media->rtp;
		if (rtp_a){
			int interval_a = 0;
			interval_a = time(NULL) - ast_rtp_instance_get_last_rx(rtp_a);
			if (interval_a < ast_rtp_instance_get_timeout(rtp_a)) {
				return 1;
			}
		}
	}
	return 0;
}

/*! \brief Check whether RTP is being received or not */
static int rtp_check_timeout(const void *data)
{
	struct ast_sip_session_media *session_media = (struct ast_sip_session_media *)data;
	struct ast_rtp_instance *rtp = session_media->rtp;
	int elapsed;
	struct ast_channel *chan;

	if (!rtp) {
		return 0;
	}

	elapsed = time(NULL) - ast_rtp_instance_get_last_rx(rtp);
	if (elapsed < ast_rtp_instance_get_timeout(rtp)) {
		return (ast_rtp_instance_get_timeout(rtp) - elapsed) * 1000;
	}

	chan = ast_channel_get_by_name(ast_rtp_instance_get_channel_id(rtp));
	if (!chan) {
		return 0;
	}
	if (ast_video_audio_all_timeout (chan)) { /*bug 58445 */
		return 0;
	}
	/*jklou bug 55469*/
	struct ast_bridge *bridge;

	ast_channel_lock(chan);
	bridge = ast_channel_get_bridge(chan);
	ast_channel_unlock(chan);
	if (bridge && bridge->is_directmedia == 1) {
		ao2_ref(bridge, -1);
		ast_channel_unref(chan);
		return 0;
	} else if (bridge) {
		ao2_ref(bridge, -1);
	}
	/*ended jklou*/

	if (ast_channel_get_is_t38_fax(chan)) {
		ast_log(LOG_NOTICE, "This channel[%s] is T.38 FAX channel!\n", ast_channel_name(chan));
		ast_channel_unref(chan);
		return 0;
	}

	ast_log(LOG_NOTICE, "Disconnecting channel '%s' for lack of RTP activity in %d seconds\n",
		ast_channel_name(chan), elapsed);

	ast_softhangup(chan, AST_SOFTHANGUP_DEV);
	ast_channel_unref(chan);

	return 0;
}
#endif

/*! \brief Internal function which creates an RTP instance */
static int create_rtp(struct ast_sip_session *session, struct ast_sip_session_media *session_media, unsigned int ipv6)
{
	struct ast_rtp_engine_ice *ice;

#ifdef GRANDSTREAM_NETWORKS
	int endpt_ice_support = session->endpoint->media.rtp.ice_support ?
		session->endpoint->media.rtp.ice_support : session->is_incoming_webrtc_call;

	if (session && session->endpoint && session->endpoint->intranet_ip_filter) {
		endpt_ice_support |= AST_RTP_ICE_IP_FILTER;
	}

	if (!(session_media->rtp = ast_rtp_instance_new(session->endpoint->media.rtp.engine, sched, ipv6 ? &address_ipv6 : &address_ipv4, &endpt_ice_support))) {
#else
	if (!(session_media->rtp = ast_rtp_instance_new(session->endpoint->media.rtp.engine, sched, ipv6 ? &address_ipv6 : &address_ipv4, NULL))) {
#endif
		ast_log(LOG_ERROR, "Unable to create RTP instance using RTP engine '%s'\n", session->endpoint->media.rtp.engine);
		return -1;
	}

	ast_rtp_instance_set_prop(session_media->rtp, AST_RTP_PROPERTY_RTCP, 1);
	ast_rtp_instance_set_prop(session_media->rtp, AST_RTP_PROPERTY_NAT, session->endpoint->media.rtp.symmetric);

#ifdef GRANDSTREAM_NETWORKS
	if ((!session->endpoint->media.rtp.ice_support && !session->is_incoming_webrtc_call) && (ice = ast_rtp_instance_get_ice(session_media->rtp))) {
		ice->stop(session_media->rtp);
	}
#else
	if (!session->endpoint->media.rtp.ice_support && (ice = ast_rtp_instance_get_ice(session_media->rtp))) {
		ice->stop(session_media->rtp);
	}
#endif

	if (session->endpoint->dtmf == AST_SIP_DTMF_RFC_4733 || session->endpoint->dtmf == AST_SIP_DTMF_AUTO) {
		ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_RFC2833);
		ast_rtp_instance_set_prop(session_media->rtp, AST_RTP_PROPERTY_DTMF, 1);
	} else if (session->endpoint->dtmf == AST_SIP_DTMF_INBAND) {
		ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_INBAND);
	}

	if (!strcmp(session_media->stream_type, STR_AUDIO) &&
			(session->endpoint->media.tos_audio || session->endpoint->media.cos_video)) {
		ast_rtp_instance_set_qos(session_media->rtp, session->endpoint->media.tos_audio,
				session->endpoint->media.cos_audio, "SIP RTP Audio");
	} else if (!strcmp(session_media->stream_type, STR_VIDEO) &&
			(session->endpoint->media.tos_video || session->endpoint->media.cos_video)) {
		ast_rtp_instance_set_qos(session_media->rtp, session->endpoint->media.tos_video,
				session->endpoint->media.cos_video, "SIP RTP Video");
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_rtp_instance_set_last_rx(session_media->rtp, time(NULL));
	/* Add by bfdu for srtp crash, 2015.07.27 */
	ast_rtp_instance_set_srtp_flag(session_media->rtp, 0);
#endif

	/*BEGIN: Add by shfguan for [Bug 47439] 2015-09-08 */
#ifdef GRANDSTREAM_NETWORKS
	if (session->endpoint->media.rtp.ice_support)
	{
		pbx_builtin_setvar_helper(session->channel,"WEBRTC","yes");
	}
#endif
	/*BEGIN: Add by shfguan for [Bug 47439] 2015-09-08 */


	return 0;
}

static void get_codecs(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream, struct ast_rtp_codecs *codecs,
       struct ast_sip_session_media *session_media)
{
	pjmedia_sdp_attr *attr;
	pjmedia_sdp_rtpmap *rtpmap;
	pjmedia_sdp_fmtp fmtp;
	struct ast_format *format;
	int i, num = 0, tel_event = 0;
	char name[256];
	char media[20];
	char fmt_param[256];

	ast_rtp_codecs_payloads_initialize(codecs);

	/* Iterate through provided formats */
	for (i = 0; i < stream->desc.fmt_count; ++i) {
		/* The payload is kept as a string for things like t38 but for video it is always numerical */
		ast_rtp_codecs_payloads_set_m_type(codecs, NULL, pj_strtoul(&stream->desc.fmt[i]));
		/* Look for the optional rtpmap attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "rtpmap", &stream->desc.fmt[i]))) {
			continue;
		}

		/* Interpret the attribute as an rtpmap */
		if ((pjmedia_sdp_attr_to_rtpmap(session->inv_session->pool_prov, attr, &rtpmap)) != PJ_SUCCESS) {
			continue;
		}

		ast_copy_pj_str(name, &rtpmap->enc_name, sizeof(name));
                if (strcmp(name,"telephone-event") == 0) {
                        tel_event++;
                }
#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for BFCP, 2015.11.17 */
		if (session_media && !strncmp(session_media->stream_type, "video", 5)) {
			if (strcmp(name,"GS-FEC") == 0) {
				session->fec_multiplexed_para.has_multiplexed_fec = TRUE;
				ast_rtp_codecs_payloads_set_rtpmap_type_rate(codecs, NULL, pj_strtoul(&stream->desc.fmt[i]),
							     "application", name, 0, rtpmap->clock_rate);
			}
		}
#endif
		ast_copy_pj_str(media, (pj_str_t*)&stream->desc.media, sizeof(media));
		ast_rtp_codecs_payloads_set_rtpmap_type_rate(codecs, NULL, pj_strtoul(&stream->desc.fmt[i]),
							     media, name, 0, rtpmap->clock_rate);
		/* Look for an optional associated fmtp attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "fmtp", &rtpmap->pt))) {
			continue;
		}

		if ((pjmedia_sdp_attr_get_fmtp(attr, &fmtp)) == PJ_SUCCESS) {
			ast_copy_pj_str(fmt_param, &fmtp.fmt, sizeof(fmt_param));
			if (sscanf(fmt_param, "%30d", &num) != 1) {
				continue;
			}

			if ((format = ast_rtp_codecs_get_payload_format(codecs, num))) {
				struct ast_format *format_parsed;

				ast_copy_pj_str(fmt_param, &fmtp.fmt_param, sizeof(fmt_param));

				format_parsed = ast_format_parse_sdp_fmtp(format, fmt_param);
				if (format_parsed) {
					ast_rtp_codecs_payload_replace_format(codecs, num, format_parsed);
					ao2_ref(format_parsed, -1);
				}

				ao2_ref(format, -1);
			}
		}
	}
	if ((tel_event==0) && (session->endpoint->dtmf == AST_SIP_DTMF_AUTO)) {
                ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_INBAND);
	}
	/* Get the packetization, if it exists */
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "ptime", NULL))) {
		unsigned long framing = pj_strtoul(pj_strltrim(&attr->value));
		if (framing && session->endpoint->media.rtp.use_ptime) {
			ast_rtp_codecs_set_framing(codecs, framing);
		}
	}
#ifdef GRANDSTREAM_NETWORKS
	/* Add by jklou*/
	if (session_media && !strncmp(session_media->stream_type, "application", 11)) {
		char *passthrough = ast_sip_get_attr_passthrough();
		ast_log(LOG_NOTICE, "attr passthroug = %s\n", passthrough);
		if (strcasecmp(passthrough, "yes") == 0) {
			set_app_other_attr_para(session, stream);
		}
		ast_free(passthrough);
	}
#endif
}

#ifdef GRANDSTREAM_NETWORKS
static void get_media_formats_by_order(struct ast_rtp_codecs *codecs, const struct pjmedia_sdp_media *stream,
	struct ast_format_cap *astformats, int *nonastformats)
{
	unsigned j;	
	struct ast_rtp_payload_type *type;
	
	ast_rwlock_rdlock(&codecs->codecs_lock);
	for (j=0; j<stream->desc.fmt_count; ++j) {
		int payload = pj_strtoul(&stream->desc.fmt[j]);

		if (payload < 0 || payload >= AST_RTP_MAX_PT) {			
			break;		
		}

		if (payload < AST_VECTOR_SIZE(&codecs->payloads)) {
			type = AST_VECTOR_GET(&codecs->payloads, payload);
			if (type) {
				if ( type->asterisk_format) {					
					ast_format_cap_append(astformats, type->format, 0);
				} else {				
					*nonastformats |= type->rtp_code;
				}
			}
		}
	}
	
	if (codecs->framing) {
		ast_format_cap_set_framing(astformats, codecs->framing);
	}
	ast_rwlock_unlock(&codecs->codecs_lock);
}

/* Begin Add by bfdu for Bug 43313/Bug 43783, 2015.05.26 */
struct pjmedia_sdp_neg
{
    pjmedia_sdp_neg_state state;	    /**< Negotiator state.	     */
    pj_bool_t		  prefer_remote_codec_order;
    pj_bool_t             answer_with_multiple_codecs;
    pj_bool_t		  has_remote_answer;
    pj_bool_t		  answer_was_remote;

    pjmedia_sdp_session	*initial_sdp,	    /**< Initial local SDP	     */
			*active_local_sdp,  /**< Currently active local SDP. */
			*active_remote_sdp, /**< Currently active remote's.  */
			*neg_local_sdp,	    /**< Temporary local SDP.	     */
			*neg_remote_sdp;    /**< Temporary remote SDP.	     */
};

static void set_caller_nativeformat_after_neg(struct ast_sip_session *session)
{
	RAII_VAR(struct ast_sip_session_media *, local_session_media, NULL, ao2_cleanup);
	RAII_VAR(struct ast_format_cap *, local_caps, NULL, ao2_cleanup);
	int fmts = 0;
	char local_media[20];
	int i = 0;
	pjmedia_sdp_session *local_sdp;
	pjmedia_sdp_media *local_stream = NULL;
	if (!session || !session->inv_session || !session->inv_session->neg) {
		ast_log(LOG_WARNING, "set caller nativeformat after neg param is invalid\n");
		return;
	}
	local_sdp = session->inv_session->neg->active_local_sdp;
	if (!local_sdp) {
		ast_log(LOG_WARNING, "set_caller_nativeformat_after_neg can't get active local sdp\n");
		return;
	}

	local_session_media = ao2_find(session->media, "audio", OBJ_KEY);

	ast_log(LOG_CODEC_NEG, "set_caller_nativeformat begin\n");

	for (i = 0; i < local_sdp->media_count; ++i) {
		memset(local_media, 0, sizeof(local_media));
		ast_copy_pj_str(local_media, &local_sdp->media[i]->desc.media, sizeof(local_media));
		if (!strcasecmp(local_media, "audio")) {
			local_stream = local_sdp->media[i];
		}
	}
	if (NULL == local_stream) {
		ast_log(LOG_ERROR, "can't get local audio media stream.\n");
		return;
	}

	if (!(local_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate capabilities\n");
		return;
	}

	if (local_session_media) {
		get_media_formats_by_order(ast_rtp_instance_get_codecs(local_session_media->rtp), local_stream, local_caps, &fmts);
	}

	struct ast_format *local_format;
	local_format = ast_format_cap_get_format(local_caps, 0);
	if (local_format) {
		ast_format_cap_remove_by_type(local_caps, AST_MEDIA_TYPE_AUDIO);
		ast_format_cap_append(local_caps, local_format, 0);
		ast_channel_nativeformats_set(session->channel, local_caps);

		struct ast_str *format_buf = ast_str_alloca(64);
		ast_log(LOG_CODEC_NEG, "set_caller_nativeformat nativeformat: %s\n",	ast_format_cap_get_names(local_caps, &format_buf));
	}
	if (local_format) {
		ao2_ref(local_format, -1);
	}
	return;
}
/* End Add by bfdu for Bug 43313/Bug 43783, 2015.05.26 */
#endif

static int set_caps(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
		    const struct pjmedia_sdp_media *stream)
{
	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);
	RAII_VAR(struct ast_format_cap *, peer, NULL, ao2_cleanup);
	RAII_VAR(struct ast_format_cap *, joint, NULL, ao2_cleanup);
	enum ast_media_type media_type = stream_to_media_type(session_media->stream_type);
	struct ast_rtp_codecs codecs = AST_RTP_CODECS_NULL_INIT;
	int fmts = 0;
	int direct_media_enabled = !ast_sockaddr_isnull(&session_media->direct_media_addr) &&
		ast_format_cap_count(session->direct_media_cap);
	int dsp_features = 0;

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT)) ||
	    !(peer = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT)) ||
	    !(joint = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate %s capabilities\n", session_media->stream_type);
		return -1;
	}

	/* get the endpoint capabilities */
	if (direct_media_enabled) {
		ast_format_cap_get_compatible(session->endpoint->media.codecs, session->direct_media_cap, caps);
		format_cap_only_type(caps, media_type);
	} else {
#ifdef GRANDSTREAM_NETWORKS
		/*! Begin: Added by pxie@grandstream.cn for bug[45683 and 42261], 2015-07-16.*/
		if (session && session->wImageReInvite) {
			ast_format_cap_sequence_append_from_cap(caps, session->endpoint->media.codecs, media_type);
		} else {
			ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, media_type);
		}
		/*! End: Added by pxie@grandstream.cn for bug[45683 and 42261], 2015-07-16.*/
#else
		ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, media_type);
#endif
	}

	/* get the capabilities on the peer */
	get_codecs(session, stream, &codecs,  session_media);

#ifdef GRANDSTREAM_NETWORKS
	if (session_media && !strncmp(session_media->stream_type, "video", 5)) {
		set_multiplexed_fec_flag(session, stream);
	}
	get_media_formats_by_order(&codecs, stream, peer, &fmts);
	if (TRUE == session->fec_multiplexed_para.has_multiplexed_fec) {
		ast_format_cap_append_from_cap(session->fec_multiplexed_req_caps, peer, AST_MEDIA_TYPE_APPLICATION);
	}
#else
	ast_rtp_codecs_payload_formats(&codecs, peer, &fmts);
#endif

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_CODEC_NEG, "aros:%s, fromdomain:%s\n", session->endpoint->aors, session->endpoint->fromdomain);
	if (1 == session->is_outgoing && session->endpoint->aors && 0 == strncmp(session->endpoint->aors, "trunk_", 6)) {
		ast_log(LOG_CODEC_NEG, "an inbound call from trunk!\n");
		ast_format_cap_get_compatible(caps, peer, joint);
	} else {
		ast_format_cap_get_compatible(peer, caps, joint);
	}
#else
	/* get the joint capabilities between peer and endpoint */
	ast_format_cap_get_compatible(caps, peer, joint);
#endif
#ifdef GRANDSTREAM_NETWORKS
	int is_direct_media = 0;
	if (session->channel) {
		struct ast_bridge *bridge;
		ast_channel_lock(session->channel);
		bridge = ast_channel_get_bridge(session->channel);
		ast_channel_unlock(session->channel);
		if (!bridge) {
			;
		} else {
			if(1 == bridge->is_directmedia) {
				is_direct_media = 1;
			} else {
				is_direct_media = 0;
			}
			ao2_ref(bridge, -1);
		}
	}

	if (!ast_format_cap_count(joint) && !(is_direct_media == 1 && !strncmp(session_media->stream_type, "video", 5))) {
#else
	if (!ast_format_cap_count(joint)) {
#endif
		struct ast_str *usbuf = ast_str_alloca(256);
		struct ast_str *thembuf = ast_str_alloca(256);

		ast_rtp_codecs_payloads_destroy(&codecs);
		ast_log(LOG_NOTICE, "No joint capabilities for '%s' media stream between our configuration(%s) and incoming SDP(%s)\n",
			session_media->stream_type,
			ast_format_cap_get_names(caps, &usbuf),
			ast_format_cap_get_names(peer, &thembuf));
		return -1;
	}

	ast_rtp_codecs_payloads_copy(&codecs, ast_rtp_instance_get_codecs(session_media->rtp),
				     session_media->rtp);	
#ifdef GRANDSTREAM_NETWORKS
	{		
		struct ast_str *format_buf = ast_str_alloca(64);		
		ast_log(LOG_CODEC_NEG, " peer: %s\n",	ast_format_cap_get_names(peer, &format_buf));		
		ast_log(LOG_CODEC_NEG, " caps: %s\n",	ast_format_cap_get_names(caps, &format_buf));
		ast_log(LOG_CODEC_NEG, " set session->req_caps of caller: %s\n",			 
			ast_format_cap_get_names(joint, &format_buf));
	}
	/* Add by bfdu for canreinvite, 2015.07.09 */
	ast_format_cap_remove_by_type(session->req_caps, media_type);
#endif

	ast_format_cap_append_from_cap(session->req_caps, joint, AST_MEDIA_TYPE_UNKNOWN);

	if (session->channel) {
#ifdef GRANDSTREAM_NETWORKS
		RAII_VAR(struct ast_channel *, bridged, ast_channel_bridge_peer(session->channel), ast_channel_cleanup);
#endif
		ast_channel_lock(session->channel);
		ast_format_cap_remove_by_type(caps, AST_MEDIA_TYPE_UNKNOWN);
		ast_format_cap_append_from_cap(caps, ast_channel_nativeformats(session->channel),
			AST_MEDIA_TYPE_UNKNOWN);
		ast_format_cap_remove_by_type(caps, media_type);
#ifdef GRANDSTREAM_NETWORKS
		/*! Begin: Added by pxie@grandstream.cn for bug[45683], 2015-07-28.*/
		if (session->is_fax_sending_request) {
			ast_format_cap_sequence_append_from_cap(caps, joint, media_type);
		} else {
			ast_format_cap_append_from_cap(caps, joint, media_type);
		}
		/*! End: Added by pxie@grandstream.cn for bug[45683], 2015-07-28.*/
#else
		ast_format_cap_append_from_cap(caps, joint, media_type);
#endif

		/*
		 * Apply the new formats to the channel, potentially changing
		 * raw read/write formats and translation path while doing so.
		 */
#ifdef GRANDSTREAM_NETWORKS
		/* Modify by bfdu for Bug 43313/Bug 43783, 2015.05.26 */
		if (session->is_outgoing || media_type == AST_MEDIA_TYPE_VIDEO) {
			ast_channel_nativeformats_set(session->channel, caps);
		} else {
			/* nativeformat should also set here when some condition don't go speical_answer like voicemail */
			set_caller_nativeformat_after_neg(session);
		}
#else
		ast_channel_nativeformats_set(session->channel, caps);
#endif
		if (media_type == AST_MEDIA_TYPE_AUDIO) {
#ifdef GRANDSTREAM_NETWORKS
			if (!ast_channel_get_vfax_flag(session->channel)) {
				struct ast_format *session_format;
				ast_format_cap_remove_by_type(caps, AST_MEDIA_TYPE_VIDEO);
				session_format = ast_format_cap_get_format(caps, 0);
				if (session_format) {
					ast_channel_set_readformat(session->channel, session_format);
					ast_channel_set_writeformat(session->channel, session_format);
					ao2_ref(session_format, -1);
				}
			}
#endif

			ast_set_read_format(session->channel, ast_channel_readformat(session->channel));
			ast_set_write_format(session->channel, ast_channel_writeformat(session->channel));
		}
		if ((session->endpoint->dtmf == AST_SIP_DTMF_AUTO)
		    && (ast_rtp_instance_dtmf_mode_get(session_media->rtp) == AST_RTP_DTMF_MODE_RFC2833)
		    && (session->dsp)) {
			dsp_features = ast_dsp_get_features(session->dsp);
			dsp_features &= ~DSP_FEATURE_DIGIT_DETECT;
			if (dsp_features) {
				ast_dsp_set_features(session->dsp, dsp_features);
			} else {
				ast_dsp_free(session->dsp);
				session->dsp = NULL;
			}
		}
		ast_channel_unlock(session->channel);
	}

	ast_rtp_codecs_payloads_destroy(&codecs);
	return 0;
}

static pjmedia_sdp_attr* generate_rtpmap_attr(pjmedia_sdp_media *media, pj_pool_t *pool, int rtp_code,
					      int asterisk_format, struct ast_format *format, int code)
{
	pjmedia_sdp_rtpmap rtpmap;
	pjmedia_sdp_attr *attr = NULL;
	char tmp[64];

	snprintf(tmp, sizeof(tmp), "%d", rtp_code);
	pj_strdup2(pool, &media->desc.fmt[media->desc.fmt_count++], tmp);
	rtpmap.pt = media->desc.fmt[media->desc.fmt_count - 1];
	rtpmap.clock_rate = ast_rtp_lookup_sample_rate2(asterisk_format, format, code);
	pj_strdup2(pool, &rtpmap.enc_name, ast_rtp_lookup_mime_subtype2(asterisk_format, format, code, 0));
	rtpmap.param.slen = 0;
	rtpmap.param.ptr = NULL;

	pjmedia_sdp_rtpmap_to_attr(pool, &rtpmap, &attr);

	return attr;
}

static pjmedia_sdp_attr* generate_fmtp_attr(pj_pool_t *pool, struct ast_format *format, int rtp_code)
{
	struct ast_str *fmtp0 = ast_str_alloca(256);
	pj_str_t fmtp1;
	pjmedia_sdp_attr *attr = NULL;
	char *tmp;

	ast_format_generate_sdp_fmtp(format, rtp_code, &fmtp0);
	if (ast_str_strlen(fmtp0)) {
		tmp = ast_str_buffer(fmtp0) + ast_str_strlen(fmtp0) - 1;
		/* remove any carriage return line feeds */
		while (*tmp == '\r' || *tmp == '\n') --tmp;
		*++tmp = '\0';
		/* ast...generate gives us everything, just need value */
		tmp = strchr(ast_str_buffer(fmtp0), ':');
		if (tmp && tmp + 1) {
			fmtp1 = pj_str(tmp + 1);
		} else {
			fmtp1 = pj_str(ast_str_buffer(fmtp0));
		}
		attr = pjmedia_sdp_attr_create(pool, "fmtp", &fmtp1);
	}
	return attr;
}

/*! \brief Function which adds ICE attributes to a media stream */
static void add_ice_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media, pj_pool_t *pool, pjmedia_sdp_media *media)
{
	struct ast_rtp_engine_ice *ice;
	struct ao2_container *candidates;
	const char *username, *password;
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	struct ao2_iterator it_candidates;
	struct ast_rtp_engine_ice_candidate *candidate;

#ifdef GRANDSTREAM_NETWORKS
	if ((!session->endpoint->media.rtp.ice_support && !session->is_incoming_webrtc_call) || !(ice = ast_rtp_instance_get_ice(session_media->rtp)) ||
		!(candidates = ice->get_local_candidates(session_media->rtp))) {
		return;
	}
#else
	if ((!session->endpoint->media.rtp.ice_support || !session->is_incoming_webrtc_call) || !(ice = ast_rtp_instance_get_ice(session_media->rtp)) ||
		!(candidates = ice->get_local_candidates(session_media->rtp))) {
		return;
	}
#endif

	if ((username = ice->get_ufrag(session_media->rtp))) {
		attr = pjmedia_sdp_attr_create(pool, "ice-ufrag", pj_cstr(&stmp, username));
		media->attr[media->attr_count++] = attr;
	}

	if ((password = ice->get_password(session_media->rtp))) {
		attr = pjmedia_sdp_attr_create(pool, "ice-pwd", pj_cstr(&stmp, password));
		media->attr[media->attr_count++] = attr;
	}

	it_candidates = ao2_iterator_init(candidates, 0);
	for (; (candidate = ao2_iterator_next(&it_candidates)); ao2_ref(candidate, -1)) {
		struct ast_str *attr_candidate = ast_str_create(128);

		ast_str_set(&attr_candidate, -1, "%s %u %s %d %s ", candidate->foundation, candidate->id, candidate->transport,
					candidate->priority, ast_sockaddr_stringify_addr_remote(&candidate->address));
		ast_str_append(&attr_candidate, -1, "%s typ ", ast_sockaddr_stringify_port(&candidate->address));

		switch (candidate->type) {
			case AST_RTP_ICE_CANDIDATE_TYPE_HOST:
				ast_str_append(&attr_candidate, -1, "host");
				break;
			case AST_RTP_ICE_CANDIDATE_TYPE_SRFLX:
				ast_str_append(&attr_candidate, -1, "srflx");
				break;
			case AST_RTP_ICE_CANDIDATE_TYPE_RELAYED:
				ast_str_append(&attr_candidate, -1, "relay");
				break;
		}

		if (!ast_sockaddr_isnull(&candidate->relay_address)) {
			ast_str_append(&attr_candidate, -1, " raddr %s rport", ast_sockaddr_stringify_addr_remote(&candidate->relay_address));
			ast_str_append(&attr_candidate, -1, " %s", ast_sockaddr_stringify_port(&candidate->relay_address));
		}

		attr = pjmedia_sdp_attr_create(pool, "candidate", pj_cstr(&stmp, ast_str_buffer(attr_candidate)));

#ifdef GRANDSTREAM_NETWORKS
		RAII_VAR(char *, candidate_ip, NULL, ast_free);
		if (ast_rtp_get_ice_stun_server_ip(session_media->rtp) && ast_rtp_get_ice_turn_server_ip(session_media->rtp)
			&& session->endpoint->intranet_ip_filter && strstr(attr->name.ptr, "candidate")
			&& !ast_rtp_regex_match("((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)",
			attr->value.ptr, &candidate_ip)) {
			if (candidate_ip && !ast_rtp_judge_ipaddr_private(candidate_ip)) {
				ast_log(LOG_NOTICE, "This ip address '%s' is private in candidate!\n", candidate_ip);
				continue;
			}
		} else if (!ast_rtp_get_ice_stun_server_ip(session_media->rtp) && ast_rtp_get_ice_turn_server_ip(session_media->rtp)
			&& session->endpoint->intranet_ip_filter && strstr(attr->name.ptr, "candidate")
			&& !ast_rtp_regex_match("((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)",
			attr->value.ptr, &candidate_ip)) {
			if (candidate_ip && !ast_rtp_judge_ipaddr_private(candidate_ip)) {
				ast_log(LOG_NOTICE, "This ip address '%s' is private in candidate!\n", candidate_ip);
				continue;
			}
		}
#endif

		media->attr[media->attr_count++] = attr;

		ast_free(attr_candidate);
	}

	ao2_iterator_destroy(&it_candidates);
	ao2_ref(candidates, -1);
}

/*! \brief Function which processes ICE attributes in an audio stream */
static void process_ice_attributes(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				   const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream)
{
	struct ast_rtp_engine_ice *ice;
	const pjmedia_sdp_attr *attr;
	char attr_value[256];
	unsigned int attr_i;

	/* If ICE support is not enabled or available exit early */
#ifdef GRANDSTREAM_NETWORKS
	if ((!session->endpoint->media.rtp.ice_support && !session->is_incoming_webrtc_call) || !(ice = ast_rtp_instance_get_ice(session_media->rtp))) {
		return;
	}
#else
	if (!session->endpoint->media.rtp.ice_support || !(ice = ast_rtp_instance_get_ice(session_media->rtp))) {
		return;
	}
#endif

	attr = pjmedia_sdp_media_find_attr2(remote_stream, "ice-ufrag", NULL);
	if (!attr) {
		attr = pjmedia_sdp_attr_find2(remote->attr_count, remote->attr, "ice-ufrag", NULL);
	}
	if (attr) {
		ast_copy_pj_str(attr_value, (pj_str_t*)&attr->value, sizeof(attr_value));
		ice->set_authentication(session_media->rtp, attr_value, NULL);
	} else {
		return;
	}

	attr = pjmedia_sdp_media_find_attr2(remote_stream, "ice-pwd", NULL);
	if (!attr) {
		attr = pjmedia_sdp_attr_find2(remote->attr_count, remote->attr, "ice-pwd", NULL);
	}
	if (attr) {
		ast_copy_pj_str(attr_value, (pj_str_t*)&attr->value, sizeof(attr_value));
		ice->set_authentication(session_media->rtp, NULL, attr_value);
	} else {
		return;
	}

	if (pjmedia_sdp_media_find_attr2(remote_stream, "ice-lite", NULL)) {
		ice->ice_lite(session_media->rtp);
	}

	/* Find all of the candidates */
	for (attr_i = 0; attr_i < remote_stream->attr_count; ++attr_i) {
		char foundation[32], transport[32], address[PJ_INET6_ADDRSTRLEN + 1], cand_type[6], relay_address[PJ_INET6_ADDRSTRLEN + 1] = "";
		unsigned int port, relay_port = 0;
		struct ast_rtp_engine_ice_candidate candidate = { 0, };

		attr = remote_stream->attr[attr_i];

		/* If this is not a candidate line skip it */
		if (pj_strcmp2(&attr->name, "candidate")) {
			continue;
		}

		ast_copy_pj_str(attr_value, (pj_str_t*)&attr->value, sizeof(attr_value));

		if (sscanf(attr_value, "%31s %30u %31s %30u %46s %30u typ %5s %*s %23s %*s %30u", foundation, &candidate.id, transport,
			(unsigned *)&candidate.priority, address, &port, cand_type, relay_address, &relay_port) < 7) {
			/* Candidate did not parse properly */
			continue;
		}

		candidate.foundation = foundation;
		candidate.transport = transport;

		ast_sockaddr_parse(&candidate.address, address, PARSE_PORT_FORBID);
		ast_sockaddr_set_port(&candidate.address, port);

		if (!strcasecmp(cand_type, "host")) {
			candidate.type = AST_RTP_ICE_CANDIDATE_TYPE_HOST;
		} else if (!strcasecmp(cand_type, "srflx")) {
			candidate.type = AST_RTP_ICE_CANDIDATE_TYPE_SRFLX;
		} else if (!strcasecmp(cand_type, "relay")) {
			candidate.type = AST_RTP_ICE_CANDIDATE_TYPE_RELAYED;
		} else {
			continue;
		}

		if (!ast_strlen_zero(relay_address)) {
			ast_sockaddr_parse(&candidate.relay_address, relay_address, PARSE_PORT_FORBID);
		}

		if (relay_port) {
			ast_sockaddr_set_port(&candidate.relay_address, relay_port);
		}

		ice->add_remote_candidate(session_media->rtp, &candidate);
	}

	ice->set_role(session_media->rtp, pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg) == PJ_TRUE ?
		AST_RTP_ICE_ROLE_CONTROLLING : AST_RTP_ICE_ROLE_CONTROLLED);
	ice->start(session_media->rtp);
}

/*! \brief figure out if media stream has crypto lines for sdes */
static int media_stream_has_crypto(const struct pjmedia_sdp_media *stream)
{
	int i;

	for (i = 0; i < stream->attr_count; i++) {
		pjmedia_sdp_attr *attr;

		/* check the stream for the required crypto attribute */
		attr = stream->attr[i];
		if (pj_strcmp2(&attr->name, "crypto")) {
			continue;
		}

		return 1;
	}

	return 0;
}

/*! \brief figure out media transport encryption type from the media transport string */
static enum ast_sip_session_media_encryption get_media_encryption_type(pj_str_t transport,
	const struct pjmedia_sdp_media *stream, unsigned int *optimistic)
{
	RAII_VAR(char *, transport_str, ast_strndup(transport.ptr, transport.slen), ast_free);

	*optimistic = 0;

	if (strstr(transport_str, "UDP/TLS")) {
		return AST_SIP_MEDIA_ENCRYPT_DTLS;
	} else if (strstr(transport_str, "SAVP")) {
		return AST_SIP_MEDIA_ENCRYPT_SDES;
	} else if (media_stream_has_crypto(stream)) {
		*optimistic = 1;
		return AST_SIP_MEDIA_ENCRYPT_SDES;
	} else {
		return AST_SIP_MEDIA_ENCRYPT_NONE;
	}
}

/*!
 * \brief Checks whether the encryption offered in SDP is compatible with the endpoint's configuration
 * \internal
 *
 * \param endpoint_encryption Media encryption configured for the endpoint
 * \param stream pjmedia_sdp_media stream description
 *
 * \retval AST_SIP_MEDIA_TRANSPORT_INVALID on encryption mismatch
 * \retval The encryption requested in the SDP
 */
static enum ast_sip_session_media_encryption check_endpoint_media_transport(
	struct ast_sip_endpoint *endpoint,
	const struct pjmedia_sdp_media *stream)
{
	enum ast_sip_session_media_encryption incoming_encryption;
	char transport_end = stream->desc.transport.ptr[stream->desc.transport.slen - 1];
	unsigned int optimistic;

	if ((transport_end == 'F' && !endpoint->media.rtp.use_avpf)
		|| (transport_end != 'F' && endpoint->media.rtp.use_avpf)) {
		return AST_SIP_MEDIA_TRANSPORT_INVALID;
	}

	incoming_encryption = get_media_encryption_type(stream->desc.transport, stream, &optimistic);

	if (incoming_encryption == endpoint->media.rtp.encryption) {
		return incoming_encryption;
	}
#ifdef GRANDSTREAM_NETWORKS
	if ((endpoint->media.rtp.encryption == AST_SIP_MEDIA_AUTO_NO_ENCRYPT)
		|| (endpoint->media.rtp.encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP)
		|| (endpoint->media.rtp.encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS)) {
		return incoming_encryption;
	}
#endif

	if (endpoint->media.rtp.force_avp ||
		endpoint->media.rtp.encryption_optimistic) {
		return incoming_encryption;
	}

	/* If an optimistic offer has been made but encryption is not enabled consider it as having
	 * no offer of crypto at all instead of invalid so the session proceeds.
	 */
	if (optimistic) {
		return AST_SIP_MEDIA_ENCRYPT_NONE;
	}

	return AST_SIP_MEDIA_TRANSPORT_INVALID;
}

static int setup_srtp(struct ast_sip_session_media *session_media)
{
	if (!session_media->srtp) {
		session_media->srtp = ast_sdp_srtp_alloc();
		if (!session_media->srtp) {
			return -1;
		}
	}

	if (!session_media->srtp->crypto) {
		session_media->srtp->crypto = ast_sdp_crypto_alloc();
		if (!session_media->srtp->crypto) {
			return -1;
		}
	}

	return 0;
}

static int setup_dtls_srtp(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media)
{
	struct ast_rtp_engine_dtls *dtls;

	if (!session->endpoint->media.rtp.dtls_cfg.enabled || !session_media->rtp) {
		return -1;
	}

	dtls = ast_rtp_instance_get_dtls(session_media->rtp);
	if (!dtls) {
		return -1;
	}

	session->endpoint->media.rtp.dtls_cfg.suite = ((session->endpoint->media.rtp.srtp_tag_32) ? AST_AES_CM_128_HMAC_SHA1_32 : AST_AES_CM_128_HMAC_SHA1_80);
	if (dtls->set_configuration(session_media->rtp, &session->endpoint->media.rtp.dtls_cfg)) {
		ast_log(LOG_ERROR, "Attempted to set an invalid DTLS-SRTP configuration on RTP instance '%p'\n",
			session_media->rtp);
		return -1;
	}

	if (setup_srtp(session_media)) {
		return -1;
	}
	return 0;
}

static void apply_dtls_attrib(struct ast_sip_session_media *session_media,
	pjmedia_sdp_attr *attr)
{
	struct ast_rtp_engine_dtls *dtls = ast_rtp_instance_get_dtls(session_media->rtp);
	pj_str_t *value;

	if (!attr->value.ptr) {
		return;
	}

	value = pj_strtrim(&attr->value);

	if (!pj_strcmp2(&attr->name, "setup")) {
		if (!pj_stricmp2(value, "active")) {
			dtls->set_setup(session_media->rtp, AST_RTP_DTLS_SETUP_ACTIVE);
		} else if (!pj_stricmp2(value, "passive")) {
			dtls->set_setup(session_media->rtp, AST_RTP_DTLS_SETUP_PASSIVE);
		} else if (!pj_stricmp2(value, "actpass")) {
			dtls->set_setup(session_media->rtp, AST_RTP_DTLS_SETUP_ACTPASS);
		} else if (!pj_stricmp2(value, "holdconn")) {
			dtls->set_setup(session_media->rtp, AST_RTP_DTLS_SETUP_HOLDCONN);
		} else {
			ast_log(LOG_WARNING, "Unsupported setup attribute value '%*s'\n", (int)value->slen, value->ptr);
		}
	} else if (!pj_strcmp2(&attr->name, "connection")) {
		if (!pj_stricmp2(value, "new")) {
			dtls->reset(session_media->rtp);
		} else if (!pj_stricmp2(value, "existing")) {
			/* Do nothing */
		} else {
			ast_log(LOG_WARNING, "Unsupported connection attribute value '%*s'\n", (int)value->slen, value->ptr);
		}
	} else if (!pj_strcmp2(&attr->name, "fingerprint")) {
		char hash_value[256], hash[32];
		char fingerprint_text[value->slen + 1];
		ast_copy_pj_str(fingerprint_text, value, sizeof(fingerprint_text));
			if (sscanf(fingerprint_text, "%31s %255s", hash, hash_value) == 2) {
			if (!strcasecmp(hash, "sha-1")) {
				dtls->set_fingerprint(session_media->rtp, AST_RTP_DTLS_HASH_SHA1, hash_value);
			} else if (!strcasecmp(hash, "sha-256")) {
				dtls->set_fingerprint(session_media->rtp, AST_RTP_DTLS_HASH_SHA256, hash_value);
			} else {
				ast_log(LOG_WARNING, "Unsupported fingerprint hash type '%s'\n",
				hash);
			}
		}
	}
}

static int parse_dtls_attrib(struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_session *sdp,
	const struct pjmedia_sdp_media *stream)
{
	int i;

	for (i = 0; i < sdp->attr_count; i++) {
		apply_dtls_attrib(session_media, sdp->attr[i]);
	}

	for (i = 0; i < stream->attr_count; i++) {
		apply_dtls_attrib(session_media, stream->attr[i]);
	}

	ast_set_flag(session_media->srtp, AST_SRTP_CRYPTO_OFFER_OK);

	return 0;
}

static int setup_sdes_srtp(struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *stream)
{
	int i;

	for (i = 0; i < stream->attr_count; i++) {
		pjmedia_sdp_attr *attr;
		RAII_VAR(char *, crypto_str, NULL, ast_free);

		/* check the stream for the required crypto attribute */
		attr = stream->attr[i];
		if (pj_strcmp2(&attr->name, "crypto")) {
			continue;
		}

		crypto_str = ast_strndup(attr->value.ptr, attr->value.slen);
		if (!crypto_str) {
			return -1;
		}

		if (setup_srtp(session_media)) {
			return -1;
		}

		if (!ast_sdp_crypto_process(session_media->rtp, session_media->srtp, crypto_str)) {
			/* found a valid crypto attribute */
			return 0;
		}

		ast_debug(1, "Ignoring crypto offer with unsupported parameters: %s\n", crypto_str);
	}

	/* no usable crypto attributes found */
	return -1;
}

static int setup_media_encryption(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_session *sdp,
	const struct pjmedia_sdp_media *stream)
{
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for Bug 41366 */
	if (session->endpoint->media.rtp.use_received_transport) {
		ast_log(LOG_DEBUG, "setup_media_encryption use received transport");
		unsigned int optimistic;
		session_media->encryption = get_media_encryption_type(stream->desc.transport, stream, &optimistic);
	}
#endif
	switch (session_media->encryption) {
	case AST_SIP_MEDIA_ENCRYPT_SDES:
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP:
#endif
		if (setup_sdes_srtp(session_media, stream)) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_ERROR, "Setup SRTP error!");
#endif
			return -1;
		}
		break;
	case AST_SIP_MEDIA_ENCRYPT_DTLS:
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS:
#endif
		if (setup_dtls_srtp(session, session_media)) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_ERROR, "Setup DTLS SRTP error!");
#endif
			return -1;
		}
		if (parse_dtls_attrib(session_media, sdp, stream)) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_ERROR, "Setup DTLS attributes error!");
#endif
			return -1;
		}
		break;
	case AST_SIP_MEDIA_TRANSPORT_INVALID:
	case AST_SIP_MEDIA_ENCRYPT_NONE:
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_MEDIA_AUTO_NO_ENCRYPT:
		/* Add by bfdu for disable SRTP, 2015.07.21 */
		if (session_media->srtp) {
			ast_log(LOG_DEBUG, "Have srtp, but transport is not srtp, need destory");
			ast_rtp_instance_destroy_srtp(session_media->rtp);

			ast_sdp_srtp_destroy(session_media->srtp);
			session_media->srtp = NULL;
		}
#endif
		break;
	}

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for reinvite when open direct media, 2015.09.16 */
static void copy_peer_direct_media_addr(struct ast_sip_session *session, struct ast_sockaddr *direct_media_addr, struct ast_sockaddr *addr)
{
	ast_log(LOG_DEBUG, "copy_peer_direct_media_addr begin\n");
	if (!session || !session->endpoint || !session->inv_session || !session->inv_session->pool_prov) {
		ast_log(LOG_ERROR, "copy_peer_direct_media_addr session or endpoint is not exist\n");
		return;
	}
	if (session->endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		ast_log(LOG_DEBUG, "copy_peer_direct_media_addr endpoint has encryption\n");
		return;
	}
	if (!session->endpoint->media.direct_media.enabled) {
		ast_log(LOG_DEBUG, "copy_peer_direct_media_addr direct media is not enabled\n");
		return;
	}
	if (ast_sockaddr_isnull(addr)) {
		ast_log(LOG_DEBUG, "copy_peer_direct_media_addr addr is null\n");
		return;
	}

	ast_verbose("Direct media is at port %s\n", ast_sockaddr_stringify(addr));

	ast_sockaddr_copy(direct_media_addr, addr);

	return;
}
#endif


/*! \brief Function which negotiates an incoming media stream */
static int negotiate_incoming_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
					 const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream)
{
	char host[NI_MAXHOST];
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	enum ast_media_type media_type = stream_to_media_type(session_media->stream_type);
	enum ast_sip_session_media_encryption encryption = AST_SIP_MEDIA_ENCRYPT_NONE;
	int res;

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	if (AST_MEDIA_TYPE_APPLICATION == media_type) {
		ast_log(LOG_DEBUG, "negotiated sdp, get application media type\n");
		res = negotiate_incoming_application_sdp_stream(session, session_media, sdp, stream);
		return res;
	}

	if (AST_MEDIA_TYPE_VIDEO == media_type) {
		if (ast_sip_session_judge_if_slides_video(stream)) {
			res = negotiate_incoming_slides_video_sdp_stream(session, session_media, sdp, stream);
			return res;
		}
	}
#endif

	/* If port is 0, ignore this media stream */
	if (!stream->desc.port) {
		ast_debug(3, "Media stream '%s' is already declined\n", session_media->stream_type);
#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for BFCP/FEC/FECC, 2015.08.13 */
		ast_format_cap_remove_by_type(session->req_caps, media_type);
#endif
		return 0;
	}

	/* If no type formats have been configured reject this stream */
	if (!ast_format_cap_has_type(session->endpoint->media.codecs, media_type)) {
		ast_debug(3, "Endpoint has no codecs for media type '%s', declining stream\n", session_media->stream_type);
		return 0;
	}

	/* Ensure incoming transport is compatible with the endpoint's configuration */
	if (!session->endpoint->media.rtp.use_received_transport) {
		encryption = check_endpoint_media_transport(session->endpoint, stream);

		if (encryption == AST_SIP_MEDIA_TRANSPORT_INVALID) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_DEBUG, "check transport is invalid");
#endif
			return -1;
		}
	}

	ast_copy_pj_str(host, stream->conn ? &stream->conn->addr : &sdp->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		return -1;
	}

	/* Using the connection information create an appropriate RTP instance */
	if (!session_media->rtp && create_rtp(session, session_media, ast_sockaddr_is_ipv6(addrs))) {
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for reinvite when open direct media, 2015.09.16 */
	ast_sockaddr_set_port(addrs, stream->desc.port);
	if (media_type == AST_MEDIA_TYPE_AUDIO) {
		copy_peer_direct_media_addr(session, &session->peer_direct_media_para.audio_direct_media_addr, addrs);
	} else if (media_type == AST_MEDIA_TYPE_VIDEO) {
		copy_peer_direct_media_addr(session, &session->peer_direct_media_para.main_video_direct_media_addr, addrs);
	}
#endif

	res = setup_media_encryption(session, session_media, sdp, stream);
	if (res) {
		if (!session->endpoint->media.rtp.encryption_optimistic) {
			/* If optimistic encryption is disabled and crypto should have been enabled
			 * but was not this session must fail.
			 */
			return -1;
		}
		/* There is no encryption, sad. */
		session_media->encryption = AST_SIP_MEDIA_ENCRYPT_NONE;
 	}

	/* If we've been explicitly configured to use the received transport OR if
	 * encryption is on and crypto is present use the received transport.
	 * This is done in case of optimistic because it may come in as RTP/AVP or RTP/SAVP depending
	 * on the configuration of the remote endpoint (optimistic themselves or mandatory).
	 */
	if ((session->endpoint->media.rtp.use_received_transport) ||
		((encryption == AST_SIP_MEDIA_ENCRYPT_SDES) && !res)) {
		pj_strdup(session->inv_session->pool, &session_media->transport, &stream->desc.transport);
 	}

	if (set_caps(session, session_media, stream)) {
		return 0;
	}
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for modify video extra attr, 2015.12.16 */
	if (session_media && !strncmp(session_media->stream_type, "video", 5)) {
		set_video_extra_attr_para(session, stream);
	}
#endif
	return 1;
}

static int add_crypto_to_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	enum ast_rtp_dtls_hash hash;
	const char *crypto_attribute;
	struct ast_rtp_engine_dtls *dtls;
	static const pj_str_t STR_NEW = { "new", 3 };
	static const pj_str_t STR_EXISTING = { "existing", 8 };
	static const pj_str_t STR_ACTIVE = { "active", 6 };
	static const pj_str_t STR_PASSIVE = { "passive", 7 };
	static const pj_str_t STR_ACTPASS = { "actpass", 7 };
	static const pj_str_t STR_HOLDCONN = { "holdconn", 8 };

	switch (session_media->encryption) {
	case AST_SIP_MEDIA_ENCRYPT_NONE:
	case AST_SIP_MEDIA_TRANSPORT_INVALID:
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_MEDIA_AUTO_NO_ENCRYPT:
#endif
		break;
	case AST_SIP_MEDIA_ENCRYPT_SDES:
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP:
#endif
		if (!session_media->srtp) {
			session_media->srtp = ast_sdp_srtp_alloc();
			if (!session_media->srtp) {
#ifdef GRANDSTREAM_NETWORKS
				ast_log(LOG_ERROR, "SRTP alloc error!");
#endif
				return -1;
			}
		}

		crypto_attribute = ast_sdp_srtp_get_attrib(session_media->srtp,
			0 /* DTLS running? No */,
			session->endpoint->media.rtp.srtp_tag_32 /* 32 byte tag length? */);
		if (!crypto_attribute) {
			/* No crypto attribute to add, bad news */
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_ERROR, "SRTP crypto attribute error!");
#endif
			return -1;
		}

		attr = pjmedia_sdp_attr_create(pool, "crypto", pj_cstr(&stmp, crypto_attribute));
		media->attr[media->attr_count++] = attr;
		break;
	case AST_SIP_MEDIA_ENCRYPT_DTLS:
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS:
#endif
		if (setup_dtls_srtp(session, session_media)) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_ERROR, "Setup DTLS SRTP error!");
#endif
			return -1;
		}

		dtls = ast_rtp_instance_get_dtls(session_media->rtp);
		if (!dtls) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_ERROR, "RTP instance get DTLS error!");
#endif
			return -1;
		}

		switch (dtls->get_connection(session_media->rtp)) {
		case AST_RTP_DTLS_CONNECTION_NEW:
			attr = pjmedia_sdp_attr_create(pool, "connection", &STR_NEW);
			media->attr[media->attr_count++] = attr;
			break;
		case AST_RTP_DTLS_CONNECTION_EXISTING:
			attr = pjmedia_sdp_attr_create(pool, "connection", &STR_EXISTING);
			media->attr[media->attr_count++] = attr;
			break;
		default:
			break;
		}

		switch (dtls->get_setup(session_media->rtp)) {
		case AST_RTP_DTLS_SETUP_ACTIVE:
/*Begin: Add for [Bug 56159] */
#ifdef GRANDSTREAM_NETWORKS
			if (session->inv_session != NULL && (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED)){
				attr = pjmedia_sdp_attr_create(pool, "setup", &STR_ACTPASS);
			}
			else{
				attr = pjmedia_sdp_attr_create(pool, "setup", &STR_ACTIVE);
			}
#else
			attr = pjmedia_sdp_attr_create(pool, "setup", &STR_ACTIVE);
#endif
			media->attr[media->attr_count++] = attr;
			break;
		case AST_RTP_DTLS_SETUP_PASSIVE:
#ifdef GRANDSTREAM_NETWORKS
			if (session->inv_session != NULL && (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED)){
				attr = pjmedia_sdp_attr_create(pool, "setup", &STR_ACTPASS);
			}
			else{
				attr = pjmedia_sdp_attr_create(pool, "setup", &STR_PASSIVE);
			}
#else
			attr = pjmedia_sdp_attr_create(pool, "setup", &STR_PASSIVE);
#endif
/*End: Add for [Bug 56159] */
			media->attr[media->attr_count++] = attr;
			break;
		case AST_RTP_DTLS_SETUP_ACTPASS:
			attr = pjmedia_sdp_attr_create(pool, "setup", &STR_ACTPASS);
			media->attr[media->attr_count++] = attr;
			break;
		case AST_RTP_DTLS_SETUP_HOLDCONN:
			attr = pjmedia_sdp_attr_create(pool, "setup", &STR_HOLDCONN);
			media->attr[media->attr_count++] = attr;
			break;
		default:
			break;
		}

		hash = dtls->get_fingerprint_hash(session_media->rtp);
		crypto_attribute = dtls->get_fingerprint(session_media->rtp);
		if (crypto_attribute && (hash == AST_RTP_DTLS_HASH_SHA1 || hash == AST_RTP_DTLS_HASH_SHA256)) {
			RAII_VAR(struct ast_str *, fingerprint, ast_str_create(64), ast_free);
			if (!fingerprint) {
#ifdef GRANDSTREAM_NETWORKS
				ast_log(LOG_ERROR, "Fingerprint check error!");
#endif
				return -1;
			}

			if (hash == AST_RTP_DTLS_HASH_SHA1) {
				ast_str_set(&fingerprint, 0, "SHA-1 %s", crypto_attribute);
			} else {
				ast_str_set(&fingerprint, 0, "SHA-256 %s", crypto_attribute);
			}

			attr = pjmedia_sdp_attr_create(pool, "fingerprint", pj_cstr(&stmp, ast_str_buffer(fingerprint)));
			media->attr[media->attr_count++] = attr;
		}
		break;
	}

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for reinvite when open direct media, 2015.09.16 */
static void reset_direct_media_addr(struct ast_sip_session *session, struct ast_sockaddr direct_media_addr, pjmedia_sdp_media *media)
{
	char hostip[PJ_INET6_ADDRSTRLEN+2];
	memset(hostip, 0, sizeof(hostip));
	ast_log(LOG_DEBUG, "reset_direct_media_addr begin\n");
	if (!session || !session->endpoint || !session->inv_session || !session->inv_session->pool_prov) {
		ast_log(LOG_ERROR, "reset_direct_media_addr session or endpoint is not exist\n");
		return;
	}
	if (session->endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		ast_log(LOG_DEBUG, "reset_direct_media_addr endpoint has encryption\n");
		return;
	}
	if (!session->endpoint->media.direct_media.enabled) {
		ast_log(LOG_DEBUG, "reset_direct_media_addr direct media is not enabled\n");
		return;
	}
	if (ast_sockaddr_isnull(&direct_media_addr)) {
		ast_log(LOG_DEBUG, "reset_direct_media_addr direct media addr is null\n");
		return;
	}
	ast_verbose("Direct media is at port %s\n", ast_sockaddr_stringify(&direct_media_addr));
	pj_pool_t *pool = session->inv_session->pool_prov;
	ast_copy_string(hostip, ast_sockaddr_stringify_fmt(&direct_media_addr, AST_SOCKADDR_STR_ADDR), sizeof(hostip));
	pj_strdup2(pool, &media->conn->addr, hostip);
	media->desc.port = ast_sockaddr_port(&direct_media_addr);

	return;
}
#endif

/*! \brief Function which creates an outgoing stream */
static int create_outgoing_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				      struct pjmedia_sdp_session *sdp)
{
	pj_pool_t *pool = session->inv_session->pool_prov;
	static const pj_str_t STR_IN = { "IN", 2 };
	static const pj_str_t STR_IP4 = { "IP4", 3};
	static const pj_str_t STR_IP6 = { "IP6", 3};
	static const pj_str_t STR_SENDRECV = { "sendrecv", 8 };
	pjmedia_sdp_media *media;
	char hostip[PJ_INET6_ADDRSTRLEN+2];
	struct ast_sockaddr addr;
	char tmp[512];
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	int index = 0;
	int noncodec = (session->endpoint->dtmf == AST_SIP_DTMF_RFC_4733 || session->endpoint->dtmf == AST_SIP_DTMF_AUTO) ? AST_RTP_DTMF : 0;
	int min_packet_size = 0, max_packet_size = 0;
	int rtp_code;
	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);
	enum ast_media_type media_type = stream_to_media_type(session_media->stream_type);
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	if (AST_MEDIA_TYPE_APPLICATION == media_type) {
		ast_log(LOG_DEBUG, "create outgoing sdp, get application media type\n");
		int app_res = 0;
		app_res = create_outgoing_application_sdp_stream(session, session_media, sdp);
		return app_res;
	}
#endif

	int use_override_prefs = ast_format_cap_count(session->req_caps);
#ifdef GRANDSTREAM_NETWORKS
	struct ast_str *format_buf = ast_str_alloca(64);
#endif

	int direct_media_enabled = !ast_sockaddr_isnull(&session_media->direct_media_addr) &&
		ast_format_cap_count(session->direct_media_cap);

	if ((use_override_prefs && !ast_format_cap_has_type(session->req_caps, media_type)) ||
	    (!use_override_prefs && !ast_format_cap_has_type(session->endpoint->media.codecs, media_type))) {
		/* If no type formats are configured don't add a stream */
#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for BFCP, 2015.07.21 */
		create_outgoing_disable_main_video_sdp(session, session_media, sdp);
#endif
		return 0;
	} else if (!session_media->rtp && create_rtp(session, session_media, session->endpoint->media.rtp.ipv6)) {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_ERROR, "Create RTP error");
#endif
		return -1;
	}

	if (!(media = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_media))) ||
		!(media->conn = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_conn)))) {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_ERROR, "Alloc pj pool error!");
#endif
		return -1;
	}

	if (add_crypto_to_stream(session, session_media, pool, media)) {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_ERROR, "Add crypto to stream error!");
#endif
		return -1;
	}

	media->desc.media = pj_str(session_media->stream_type);
	if (pj_strlen(&session_media->transport)) {
		/* If a transport has already been specified use it */
		media->desc.transport = session_media->transport;
	} else {
#ifdef GRANDSTREAM_NETWORKS
		unsigned int use_avpf = 0;
		if (!session->session_uas) {
			if ((session_media->encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS) ||
				(session_media->encryption == AST_SIP_MEDIA_ENCRYPT_DTLS)) {
				ast_log(LOG_NOTICE, "use_avpf = session->endpoint->media.rtp.use_avpf[%d]\n", use_avpf = session->endpoint->media.rtp.use_avpf);
				use_avpf = session->endpoint->media.rtp.use_avpf;
			}
		}

		media->desc.transport = pj_str(ast_sdp_get_rtp_profile(
			/* Optimistic encryption places crypto in the normal RTP/AVP profile */
			!session->endpoint->media.rtp.encryption_optimistic &&
				((session_media->encryption == AST_SIP_MEDIA_ENCRYPT_SDES)
				|| (session_media->encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP)),
			session_media->rtp, use_avpf,
			session->endpoint->media.rtp.force_avp));
#else
		media->desc.transport = pj_str(ast_sdp_get_rtp_profile(
			!session->endpoint->media.rtp.encryption_optimistic &&
				(session_media->encryption == AST_SIP_MEDIA_ENCRYPT_SDES),
			session_media->rtp, session->endpoint->media.rtp.use_avpf,
			session->endpoint->media.rtp.force_avp));
#endif
	}

	/* Add connection level details */
	if (direct_media_enabled) {
		ast_copy_string(hostip, ast_sockaddr_stringify_fmt(&session_media->direct_media_addr, AST_SOCKADDR_STR_ADDR), sizeof(hostip));
	} else if (ast_strlen_zero(session->endpoint->media.address)) {
		pj_sockaddr localaddr;

		if (pj_gethostip(session->endpoint->media.rtp.ipv6 ? pj_AF_INET6() : pj_AF_INET(), &localaddr)) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_ERROR, "Pj get host ip error!");
#endif
			return -1;
		}
		pj_sockaddr_print(&localaddr, hostip, sizeof(hostip), 2);
	} else {
		ast_copy_string(hostip, session->endpoint->media.address, sizeof(hostip));
	}

	media->conn->net_type = STR_IN;
	media->conn->addr_type = session->endpoint->media.rtp.ipv6 ? STR_IP6 : STR_IP4;
	pj_strdup2(pool, &media->conn->addr, hostip);
	ast_rtp_instance_get_local_address(session_media->rtp, &addr);
	media->desc.port = direct_media_enabled ? ast_sockaddr_port(&session_media->direct_media_addr) : (pj_uint16_t) ast_sockaddr_port(&addr);
	media->desc.port_count = 1;

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for reinvite when use direct media, 2015.09.16 */
	if (media_type == AST_MEDIA_TYPE_AUDIO) {
		if (!ast_sockaddr_isnull(&session->direct_media_para.audio_direct_media_addr)) {
			reset_direct_media_addr(session, session->direct_media_para.audio_direct_media_addr, media);
			ast_sockaddr_setnull(&session->direct_media_para.audio_direct_media_addr);
		}
	} else if (media_type == AST_MEDIA_TYPE_VIDEO) {
		if (!ast_sockaddr_isnull(&session->direct_media_para.main_video_direct_media_addr)) {
			reset_direct_media_addr(session, session->direct_media_para.main_video_direct_media_addr, media);
			ast_sockaddr_setnull(&session->direct_media_para.main_video_direct_media_addr);
		}
	}
#endif

	/* Add ICE attributes and candidates */
	add_ice_to_stream(session, session_media, pool, media);

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate %s capabilities\n", session_media->stream_type);
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_CODEC_NEG, "%s session->req_caps: %s\n", 
		session->is_outgoing ? "[callee]" : "[caller]",
		ast_format_cap_get_names(session->req_caps, &format_buf));		
	ast_log(LOG_CODEC_NEG, "%s endpoint codecs: %s\n",			
		session->is_outgoing ? "[callee]" : "[caller]",
		ast_format_cap_get_names(session->endpoint->media.codecs, &format_buf));
#endif

	if (direct_media_enabled) {
#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for Bug 43713 */
		if (1 == session->is_video_media_change) {
			ast_log(LOG_DEBUG, "It's video reinvite, need change direct media caps\n");
			ast_format_cap_remove_by_type(session->direct_media_cap, AST_MEDIA_TYPE_UNKNOWN);
			ast_format_cap_append_from_cap(session->direct_media_cap, session->req_caps, AST_MEDIA_TYPE_UNKNOWN);
		}
		if ((1 == session->is_outgoing && session->endpoint->aors && 0 == strncmp(session->endpoint->aors, "trunk_", 6))) {
			ast_format_cap_get_compatible(session->endpoint->media.codecs, session->direct_media_cap, caps);
		} else {
			ast_log(LOG_CODEC_NEG, "direct media enabled, use direct_media_cap\n");
			ast_format_cap_get_compatible(session->direct_media_cap, session->endpoint->media.codecs, caps);
		}
#else
		ast_format_cap_get_compatible(session->endpoint->media.codecs, session->direct_media_cap, caps);
#endif
	} else if (!ast_format_cap_count(session->req_caps) ||
		!ast_format_cap_iscompatible(session->req_caps, session->endpoint->media.codecs)) {
#ifdef GRANDSTREAM_NETWORKS
		/* There is no the same codec in session->req_caps and endpoint codecs. only need to add audio codecs because asterisk can translate it. */		
		ast_log(LOG_CODEC_NEG, " There is no same (type:%s) codec in req_caps and endpoint codecs.\n",
			media_type == AST_MEDIA_TYPE_AUDIO ? "audio" : "video");
		
		if (media_type == AST_MEDIA_TYPE_AUDIO) {
			/*! Begin: Added by pxie@grandstream.cn for bug[45683 and 42261], 2015-07-16.*/
			if (session->is_fax_sending_request) {
				ast_format_cap_sequence_append_from_cap(caps, session->endpoint->media.codecs, media_type);
			} else {
				ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, media_type);
			}
			/*! End: Added by pxie@grandstream.cn for bug[45683 and 42261], 2015-07-16.*/
		}
#else
		ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, media_type);
#endif
	} else {	
#ifdef GRANDSTREAM_NETWORKS
		if ((1 == session->is_outgoing && session->endpoint->aors && 0 == strncmp(session->endpoint->aors, "trunk_", 6))) {
			ast_format_cap_get_compatible(session->endpoint->media.codecs, session->req_caps, caps);
		} else {
			if (session->wSwitchImageInvite488ToAudioInvite) {
				ast_format_cap_get_compatible_sequence(session->req_caps, session->endpoint->media.codecs, caps);
				ast_channel_nativeformats_set(session->channel, caps);
			} else {
				ast_format_cap_get_compatible(session->req_caps, session->endpoint->media.codecs, caps);
			}
		}
		format_cap_only_type(caps, media_type);
		
		ast_log(LOG_CODEC_NEG, "%s The same (type:%s) codecs in req_caps and endpoint codes: %s\n",			
			session->is_outgoing ? "[callee]" : "[caller]",
			media_type == AST_MEDIA_TYPE_AUDIO ? "audio" : "video",
			ast_format_cap_get_names(caps, &format_buf));
		
		if (session->is_outgoing && media_type == AST_MEDIA_TYPE_AUDIO) {
			ast_log(LOG_CODEC_NEG, "[callee] For the callee, add the rest codecs in endpoint codecs.\n");
			if (0 == strncmp(session->endpoint->aors, "trunk_", 6)){
				ast_format_cap_remove_by_type(caps, AST_MEDIA_TYPE_AUDIO);
			}

			/*! Begin: Added by pxie@grandstream.cn for bug[45683 and 42261], 2015-07-16.*/
			if (session->wSwitchImageInvite488ToAudioInvite) {
				ast_format_cap_sequence_append_from_cap(caps, session->endpoint->media.codecs, media_type);
			} else {
				ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, media_type);
			}
			/*! End: Added by pxie@grandstream.cn for bug[45683 and 42261], 2015-07-16.*/
		}
#else
		ast_format_cap_append_from_cap(caps, session->req_caps, media_type);
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_CODEC_NEG, "%s final request codecs: (type:%s) %s\n",	
		session->is_outgoing ? "[callee]" : "[caller]",
		media_type == AST_MEDIA_TYPE_AUDIO ? "audio" : "video", 
		ast_format_cap_get_names(caps, &format_buf));
#endif

	for (index = 0; index < ast_format_cap_count(caps); ++index) {
		struct ast_format *format = ast_format_cap_get_format(caps, index);

		if (ast_format_get_type(format) != media_type) {
			ao2_ref(format, -1);
			continue;
		}

		if ((rtp_code = ast_rtp_codecs_payload_code(ast_rtp_instance_get_codecs(session_media->rtp), 1, format, 0)) == -1) {
			ast_log(LOG_WARNING,"Unable to get rtp codec payload code for %s\n", ast_format_get_name(format));
			ao2_ref(format, -1);
			continue;
		}

		if (!(attr = generate_rtpmap_attr(media, pool, rtp_code, 1, format, 0))) {
			ao2_ref(format, -1);
			continue;
		}
		media->attr[media->attr_count++] = attr;

		if ((attr = generate_fmtp_attr(pool, format, rtp_code))) {
			media->attr[media->attr_count++] = attr;
		}

		if (ast_format_get_maximum_ms(format) &&
			((ast_format_get_maximum_ms(format) < max_packet_size) || !max_packet_size)) {
			max_packet_size = ast_format_get_maximum_ms(format);
		}
		ao2_ref(format, -1);
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for FEC, 2015.11.16 */
	if (media_type == AST_MEDIA_TYPE_VIDEO) {
		if (session->fec_multiplexed_para.has_multiplexed_fec) {
			int fec_index = 0;
			struct ast_str *fec_format_buf = ast_str_alloca(64);
			ast_log(LOG_DEBUG, "fec_multiplexed_req_caps: %s\n", ast_format_cap_get_names(session->fec_multiplexed_req_caps, &fec_format_buf));
			for (fec_index = 0; fec_index < ast_format_cap_count(session->fec_multiplexed_req_caps); ++fec_index) {
				struct ast_format *fec_format = ast_format_cap_get_format(session->fec_multiplexed_req_caps, fec_index);

				if ((rtp_code = ast_rtp_codecs_payload_code(ast_rtp_instance_get_codecs(session_media->rtp), 1, fec_format, 0)) == -1) {
					ast_log(LOG_WARNING,"Unable to get rtp codec payload code for %s\n", ast_format_get_name(fec_format));
					ao2_ref(fec_format, -1);
					continue;
				}

				if (!(attr = generate_rtpmap_attr(media, pool, rtp_code, 1, fec_format, 0))) {
					ao2_ref(fec_format, -1);
					continue;
				}
				media->attr[media->attr_count++] = attr;

				if ((attr = generate_fmtp_attr(pool, fec_format, rtp_code))) {
					media->attr[media->attr_count++] = attr;
				}
				ao2_ref(fec_format, -1);
			}
		}
	}
#endif

	/* Add non-codec formats */
	if (media_type != AST_MEDIA_TYPE_VIDEO) {
		for (index = 1LL; index <= AST_RTP_MAX; index <<= 1) {
			if (!(noncodec & index) || (rtp_code = ast_rtp_codecs_payload_code(ast_rtp_instance_get_codecs(session_media->rtp),
											   0, NULL, index)) == -1) {
				continue;
			}

			if (!(attr = generate_rtpmap_attr(media, pool, rtp_code, 0, NULL, index))) {
				continue;
			}

			media->attr[media->attr_count++] = attr;

			if (index == AST_RTP_DTMF) {
				snprintf(tmp, sizeof(tmp), "%d 0-16", rtp_code);
				attr = pjmedia_sdp_attr_create(pool, "fmtp", pj_cstr(&stmp, tmp));
				media->attr[media->attr_count++] = attr;
			}
		}
	}

	/* If no formats were actually added to the media stream don't add it to the SDP */
	if (!media->desc.fmt_count) {
		return 1;
	}

	/* If ptime is set add it as an attribute */
	min_packet_size = ast_rtp_codecs_get_framing(ast_rtp_instance_get_codecs(session_media->rtp));
	if (!min_packet_size) {
		min_packet_size = ast_format_cap_get_framing(caps);
	}
	if (min_packet_size) {
		snprintf(tmp, sizeof(tmp), "%d", min_packet_size);
		attr = pjmedia_sdp_attr_create(pool, "ptime", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	if (max_packet_size) {
		snprintf(tmp, sizeof(tmp), "%d", max_packet_size);
		attr = pjmedia_sdp_attr_create(pool, "maxptime", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.13 */
	if (AST_MEDIA_TYPE_VIDEO == media_type) {
		create_outgoing_main_video_extra_attr_sdp(session, pool, media);
		ast_sip_session_modify_outgoing_yealink_sdp(session, pool, media);
	}
#endif
	/* Add the sendrecv attribute - we purposely don't keep track because pjmedia-sdp will automatically change our offer for us */
	attr = PJ_POOL_ZALLOC_T(pool, pjmedia_sdp_attr);
	attr->name = STR_SENDRECV;
	media->attr[media->attr_count++] = attr;

	/* Add the media stream to the SDP */
	sdp->media[sdp->media_count++] = media;

	return 1;
}

static int apply_negotiated_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				       const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
				       const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream)
{
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	enum ast_media_type media_type = stream_to_media_type(session_media->stream_type);
	char host[NI_MAXHOST];
	int fdno, res;

	if (!session->channel) {
		return 1;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	if (AST_MEDIA_TYPE_APPLICATION == media_type) {
		ast_log(LOG_DEBUG, "apply negotiated sdp, get application media type\n");
		int app_res = 1;
		app_res = apply_negotiated_application_sdp_stream(session, session_media, local, local_stream, remote, remote_stream);
		return app_res;
	}

	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	if (AST_MEDIA_TYPE_VIDEO == media_type) {
		if (ast_sip_session_judge_if_slides_video(remote_stream)) {
			ast_log(LOG_DEBUG, "content is slides, use slides video apply\n");
			res = apply_negotiated_slides_video_sdp_stream(session, session_media, local, local_stream, remote, remote_stream);
			return res;
		}
	}
#endif

	if (!local_stream->desc.port || !remote_stream->desc.port) {
#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
		ast_format_cap_remove_by_type(session->req_caps, media_type);
#endif
		return 1;
	}

	/* Ensure incoming transport is compatible with the endpoint's configuration */
	if (!session->endpoint->media.rtp.use_received_transport &&
		check_endpoint_media_transport(session->endpoint, remote_stream) == AST_SIP_MEDIA_TRANSPORT_INVALID) {
		return -1;
	}

	/* Create an RTP instance if need be */
	if (!session_media->rtp && create_rtp(session, session_media, session->endpoint->media.rtp.ipv6)) {
		return -1;
	}

	res = setup_media_encryption(session, session_media, remote, remote_stream);
	if (!session->endpoint->media.rtp.encryption_optimistic && res) {
		/* If optimistic encryption is disabled and crypto should have been enabled but was not
		 * this session must fail.
		 */
		return -1;
	}

	if (!remote_stream->conn && !remote->conn) {
		return 1;
	}

	ast_copy_pj_str(host, remote_stream->conn ? &remote_stream->conn->addr : &remote->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		return -1;
	}

	/* Apply connection information to the RTP instance */
	ast_sockaddr_set_port(addrs, remote_stream->desc.port);
	ast_rtp_instance_set_remote_address(session_media->rtp, addrs);

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for reinvite when use direct media, 2015.09.16 */
	if (media_type == AST_MEDIA_TYPE_AUDIO) {
		copy_peer_direct_media_addr(session, &session->peer_direct_media_para.audio_direct_media_addr, addrs);
	} else if (media_type == AST_MEDIA_TYPE_VIDEO) {
		copy_peer_direct_media_addr(session, &session->peer_direct_media_para.main_video_direct_media_addr, addrs);
	}
#endif

	if (set_caps(session, session_media, local_stream)) {
#ifdef GRANDSTREAM_NETWORKS
		return -1;
#else
		return 1;
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for modify video extra attr, 2015.12.16
	   pjsip can't change video extra attr, need use remote stream to modify */
	if (session_media && !strncmp(session_media->stream_type, "video", 5)) {
		set_video_extra_attr_para(session, remote_stream);
	}
#endif

	if ((fdno = media_type_to_fdno(media_type)) < 0) {
		return -1;
	}
	ast_channel_set_fd(session->channel, fdno, ast_rtp_instance_fd(session_media->rtp, 0));
	ast_channel_set_fd(session->channel, fdno + 1, ast_rtp_instance_fd(session_media->rtp, 1));

	/* If ICE support is enabled find all the needed attributes */
	process_ice_attributes(session, session_media, remote, remote_stream);

	/* Ensure the RTP instance is active */
	ast_rtp_instance_activate(session_media->rtp);
#ifdef GRANDSTREAM_NETWORKS
	/* Remove by bfdu for video hold also stop rtp instance, set held flag, 2015.11.13 */
#else
	/* audio stream handles music on hold */
	if (media_type != AST_MEDIA_TYPE_AUDIO) {
		if ((pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg) == PJ_FALSE)
			&& (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED)) {
			ast_queue_control(session->channel, AST_CONTROL_UPDATE_RTP_PEER);
		}
		return 1;
	}
#endif

	if (ast_sockaddr_isnull(addrs) ||
		ast_sockaddr_is_any(addrs) ||
		pjmedia_sdp_media_find_attr2(remote_stream, "sendonly", NULL) ||
		pjmedia_sdp_media_find_attr2(remote_stream, "inactive", NULL)) {
		if (!session_media->held) {
#ifdef GRANDSTREAM_NETWORKS
			/* Modify by bfdu, when 183 or initial 200OK has sendonly attr, don't play hold music, 2016.05.26 */
			if (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED) {
				/* The remote side has put us on hold */
				ast_queue_hold(session->channel, session->endpoint->mohsuggest);
			}
#else
			/* The remote side has put us on hold */
			ast_queue_hold(session->channel, session->endpoint->mohsuggest);
#endif
			ast_rtp_instance_stop(session_media->rtp);
			ast_queue_frame(session->channel, &ast_null_frame);
			session_media->held = 1;
		}
	} else if (session_media->held) {
#ifdef GRANDSTREAM_NETWORKS
		/* Modify by bfdu, when 183 or initial 200OK has sendonly attr, don't play hold music, 2016.05.26 */
		if (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED) {
			/* The remote side has taken us off hold */
			ast_queue_unhold(session->channel);
		}
#else
		/* The remote side has taken us off hold */
		ast_queue_unhold(session->channel);
#endif
		ast_queue_frame(session->channel, &ast_null_frame);
		session_media->held = 0;
	} else if ((pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg) == PJ_FALSE)
		&& (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED)) {
		ast_queue_control(session->channel, AST_CONTROL_UPDATE_RTP_PEER);
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Modify by bfdu for encryption, 2015.01.05 */
	if (session->endpoint->media.rtp.use_received_transport) {
		ast_log(LOG_DEBUG, "use received transport, don't reset media encryption\n");
	} else {
		session_media->encryption = session->endpoint->media.rtp.encryption;
	}
#else
	/* This purposely resets the encryption to the configured in case it gets added later */
	session_media->encryption = session->endpoint->media.rtp.encryption;
#endif

#ifdef GRANDSTREAM_NETWORKS
	/* Begin add by bfdu for Bug 59239, 2016.03.31
	 * send_keepalive is only support audio, which use comfort noise payload.
	 * and just consider audio rtp timeout.
	 */
	if (media_type != AST_MEDIA_TYPE_AUDIO) {
		return 1;
	}
	/* End add by bfdu for Bug 59239, 2016.03.31 */
	if (session->endpoint->media.rtp.keepalive > 0 &&
			stream_to_media_type(session_media->stream_type) == AST_MEDIA_TYPE_AUDIO) {
		ast_rtp_instance_set_keepalive(session_media->rtp, session->endpoint->media.rtp.keepalive);
		/* Schedule the initial keepalive early in case this is being used to punch holes through
		 * a NAT. This way there won't be an awkward delay before media starts flowing in some
		 * scenarios.
		 */
		session_media->keepalive_sched_id = ast_sched_add_variable(sched, 500, send_keepalive,
			session_media, 1);
	}

	/* As the channel lock is not held during this process the scheduled item won't block if
	 * it is hanging up the channel at the same point we are applying this negotiated SDP.
	 */
	AST_SCHED_DEL(sched, session_media->timeout_sched_id);

	/* Due to the fact that we only ever have one scheduled timeout item for when we are both
	 * off hold and on hold we don't need to store the two timeouts differently on the RTP
	 * instance itself.
	 */
	ast_rtp_instance_set_timeout(session_media->rtp, 0);
	if (session->endpoint->media.rtp.timeout && !session_media->held) {
		ast_rtp_instance_set_timeout(session_media->rtp, session->endpoint->media.rtp.timeout);
	} else if (session->endpoint->media.rtp.timeout_hold && session_media->held) {
		ast_rtp_instance_set_timeout(session_media->rtp, session->endpoint->media.rtp.timeout_hold);
	}

	if (ast_rtp_instance_get_timeout(session_media->rtp)) {
		session_media->timeout_sched_id = ast_sched_add_variable(sched,
			ast_rtp_instance_get_timeout(session_media->rtp) * 1000, rtp_check_timeout,
			session_media, 1);
	}
#endif

	return 1;
}

/*! \brief Function which updates the media stream with external media address, if applicable */
static void change_outgoing_sdp_stream_media_address(pjsip_tx_data *tdata, struct pjmedia_sdp_media *stream, struct ast_sip_transport *transport)
{
	char host[NI_MAXHOST];
	struct ast_sockaddr addr = { { 0, } };

	/* If the stream has been rejected there will be no connection line */
	if (!stream->conn) {
		return;
	}

	ast_copy_pj_str(host, &stream->conn->addr, sizeof(host));
	ast_sockaddr_parse(&addr, host, PARSE_PORT_FORBID);

	/* Is the address within the SDP inside the same network? */
	if (ast_apply_ha(transport->localnet, &addr) == AST_SENSE_ALLOW) {
		return;
	}

	pj_strdup2(tdata->pool, &stream->conn->addr, transport->external_media_address);
}

/*! \brief Function which destroys the RTP instance when session ends */
static void stream_destroy(struct ast_sip_session_media *session_media)
{
	if (session_media->rtp) {
#ifdef GRANDSTREAM_NETWORKS
		AST_SCHED_DEL(sched, session_media->keepalive_sched_id);
		AST_SCHED_DEL(sched, session_media->timeout_sched_id);
#endif
		ast_rtp_instance_stop(session_media->rtp);
		ast_rtp_instance_destroy(session_media->rtp);
	}
	session_media->rtp = NULL;

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	if (session_media->bfcp) {
		ast_bfcp_destroy(session_media->bfcp);
	}
	session_media->bfcp = NULL;

	if (session_media->bfcp_tcp) {
		ast_bfcp_tcp_destroy(session_media->bfcp_tcp);
	}
	session_media->bfcp_tcp = NULL;
#endif
}

/*! \brief SDP handler for 'audio' media stream */
static struct ast_sip_session_sdp_handler audio_sdp_handler = {
	.id = STR_AUDIO,
	.negotiate_incoming_sdp_stream = negotiate_incoming_sdp_stream,
	.create_outgoing_sdp_stream = create_outgoing_sdp_stream,
	.apply_negotiated_sdp_stream = apply_negotiated_sdp_stream,
	.change_outgoing_sdp_stream_media_address = change_outgoing_sdp_stream_media_address,
	.stream_destroy = stream_destroy,
};

/*! \brief SDP handler for 'video' media stream */
static struct ast_sip_session_sdp_handler video_sdp_handler = {
	.id = STR_VIDEO,
	.negotiate_incoming_sdp_stream = negotiate_incoming_sdp_stream,
	.create_outgoing_sdp_stream = create_outgoing_sdp_stream,
	.apply_negotiated_sdp_stream = apply_negotiated_sdp_stream,
	.change_outgoing_sdp_stream_media_address = change_outgoing_sdp_stream_media_address,
	.stream_destroy = stream_destroy,
};

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
/*! \brief SDP handler for 'application' media stream */
static struct ast_sip_session_sdp_handler application_sdp_handler = {
	.id = STR_APPLICATION,
	.negotiate_incoming_sdp_stream = negotiate_incoming_sdp_stream,
	.create_outgoing_sdp_stream = create_outgoing_sdp_stream,
	.apply_negotiated_sdp_stream = apply_negotiated_sdp_stream,
	.change_outgoing_sdp_stream_media_address = change_outgoing_sdp_stream_media_address,
	.stream_destroy = stream_destroy,
};
#endif

static int video_info_incoming_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	struct pjsip_transaction *tsx;
	pjsip_tx_data *tdata;

	if (!session->channel
		|| !ast_sip_is_content_type(&rdata->msg_info.msg->body->content_type,
			"application",
			"media_control+xml")) {
		return 0;
	}

	tsx = pjsip_rdata_get_tsx(rdata);

	ast_queue_control(session->channel, AST_CONTROL_VIDUPDATE);

	if (pjsip_dlg_create_response(session->inv_session->dlg, rdata, 200, NULL, &tdata) == PJ_SUCCESS) {
		pjsip_dlg_send_response(session->inv_session->dlg, tsx, tdata);
	}

	return 0;
}

static struct ast_sip_session_supplement video_info_supplement = {
	.method = "INFO",
	.incoming_request = video_info_incoming_request,
};

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
static void set_main_video_framerate(struct ast_sip_session *session, unsigned int framerate)
{
	if (!framerate) {
		return;
	}
	session->main_video_para.framerate = framerate;
}

static unsigned int get_main_video_framerate(struct ast_sip_session *session)
{
	unsigned int framerate;
	framerate = session->main_video_para.framerate;

	return framerate;
}

static void set_main_video_content(struct ast_sip_session *session, char *content)
{
	if (!content) {
		return;
	}

	strncpy(session->main_video_para.content, content, sizeof(session->main_video_para.content) - 1);
}

static void set_main_video_label(struct ast_sip_session *session, unsigned int label)
{
	if (!label) {
		return;
	}

	session->main_video_para.label = label;
}

static unsigned int get_main_video_label(struct ast_sip_session *session)
{
	unsigned int label;
	label = session->main_video_para.label;

	return label;
}

static void set_slides_video_framerate(struct ast_sip_session *session, unsigned int framerate)
{
	if (!framerate) {
		return;
	}
	session->slides_video_para.framerate = framerate;
}

static unsigned int get_slides_video_framerate(struct ast_sip_session *session)
{
	unsigned int framerate;
	framerate = session->slides_video_para.framerate;

	return framerate;
}

static void set_slides_video_content(struct ast_sip_session *session, char *content)
{
	if (!content) {
		return;
	}

	strncpy(session->slides_video_para.content, content, sizeof(session->slides_video_para.content) - 1);
}

static void set_slides_video_label(struct ast_sip_session *session, unsigned int label)
{
	if (!label) {
		return;
	}

	session->slides_video_para.label = label;
}

static unsigned int get_slides_video_label(struct ast_sip_session *session)
{
	unsigned int label;
	label = session->slides_video_para.label;

	return label;
}

static void set_video_extra_attr_para(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream)
{
	pjmedia_sdp_attr *attr;
	char content[8];
	memset(content, 0, sizeof(content));
	ast_log(LOG_DEBUG, "set video extra attr para begin\n");

	char *passthrough = ast_sip_get_attr_passthrough();
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "content", NULL))) {
		ast_copy_pj_str(content, pj_strltrim(&attr->value), sizeof(content));
		if (0 != strlen(content)) {
			if (!strncmp(content, "main", 4)) {
				ast_log(LOG_DEBUG, "set main video extra attr para, content:%s\n", content);
				set_main_video_content(session, content);
			} else if (!strncmp(content, "slides", 6)) {
				ast_log(LOG_DEBUG, "set slides video extra attr para, content:%s\n", content);
				set_slides_video_content(session, content);
			} else {
				int i = 0;
				if (strcasecmp(passthrough, "yes") == 0) {
					session->main_video_para.bandw_count = 0;
					for(i = 0; i < stream->bandw_count; i++) {
						pj_pool_t *pool = session->inv_session->pool;
						session->main_video_para.bandw[session->main_video_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool,
								stream->bandw[i]);
					}
					ast_log(LOG_DEBUG, "can't support third video stream now passthroug line band\n");
				}
				ast_log(LOG_WARNING, "can't support third video stream now\n");
				ast_free(passthrough);
				return;
			}
		}
	}else {
		int i = 0;
		if (strcasecmp(passthrough, "yes") == 0) {
			session->main_video_para.bandw_count = 0;
			for(i = 0; i < stream->bandw_count; i++) {
				pj_pool_t *pool = session->inv_session->pool;
				session->main_video_para.bandw[session->main_video_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool,
						stream->bandw[i]);
			}
			ast_log(LOG_DEBUG, "can't support third video stream now passthroug line band\n");
		}
		ast_free(passthrough);
		return;
	}

	if (!strncmp(content, "main", 4)) {
		if ((attr = pjmedia_sdp_media_find_attr2(stream, "framerate", NULL))) {
			unsigned int framerate = pj_strtoul(pj_strltrim(&attr->value));
			if (framerate) {
				ast_log(LOG_DEBUG, "set main video extra attr para, framerate:%d\n", framerate);
				set_main_video_framerate(session, framerate);
			}
		}

		if ((attr = pjmedia_sdp_media_find_attr2(stream, "label", NULL))) {
			unsigned int label = pj_strtoul(pj_strltrim(&attr->value));
			if (label) {
				ast_log(LOG_DEBUG, "set main video extra attr para, label:%d\n", label);
				set_main_video_label(session, label);
			}
		}
		if (strcasecmp(passthrough, "yes") == 0) {
			int i = 0;
			session->main_video_para.other_attr_count = 0;
			for(i = 0; i < stream->attr_count; i++) {
				if ( strncasecmp(stream->attr[i]->name.ptr, "rtpmap", 6) &&
						strncasecmp(stream->attr[i]->name.ptr, "fmtp", 4) &&
						strncasecmp(stream->attr[i]->name.ptr, "sendrecv", 8) &&
						strncasecmp(stream->attr[i]->name.ptr, "sendonly", 8) &&
						strncasecmp(stream->attr[i]->name.ptr, "recvonly", 8) &&
						strncasecmp(stream->attr[i]->name.ptr, "inactive", 8) &&
						strncasecmp(stream->attr[i]->name.ptr, "framerate", 9) &&
						strncasecmp(stream->attr[i]->name.ptr, "label", 5) &&
						strncasecmp(stream->attr[i]->name.ptr, "content", 7) &&
						strncasecmp(stream->attr[i]->name.ptr, "crypto", 6)) {
					pj_pool_t *pool = session->inv_session->pool;
					session->main_video_para.other_attr[session->main_video_para.other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
											 stream->attr[i]);
				}
			}
			session->main_video_para.bandw_count = 0;
			for(i = 0; i < stream->bandw_count; i++) {
				pj_pool_t *pool = session->inv_session->pool;
				session->main_video_para.bandw[session->main_video_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool,
								stream->bandw[i]);
			}
		}
	} else if (!strncmp(content, "slides", 6)) {
		if ((attr = pjmedia_sdp_media_find_attr2(stream, "framerate", NULL))) {
			unsigned int framerate = pj_strtoul(pj_strltrim(&attr->value));
			if (framerate) {
				ast_log(LOG_DEBUG, "set slides video extra attr para, framerate:%d\n", framerate);
				set_slides_video_framerate(session, framerate);
			}
		}

		if ((attr = pjmedia_sdp_media_find_attr2(stream, "label", NULL))) {
			unsigned int label = pj_strtoul(pj_strltrim(&attr->value));
			if (label) {
				ast_log(LOG_DEBUG, "set slides video extra attr para, label:%d\n", label);
				set_slides_video_label(session, label);
			}
		}
		if (strcasecmp(passthrough, "yes") == 0) {
			int i = 0;
			session->slides_video_para.other_attr_count = 0;
			for(i = 0; i < stream->attr_count; i++) {
				if ( strncasecmp(stream->attr[i]->name.ptr, "rtpmap", 6) &&
						strncasecmp(stream->attr[i]->name.ptr, "fmtp", 4) &&
						strncasecmp(stream->attr[i]->name.ptr, "sendrecv", 8) &&
						strncasecmp(stream->attr[i]->name.ptr, "sendonly", 8) &&
						strncasecmp(stream->attr[i]->name.ptr, "recvonly", 8) &&
						strncasecmp(stream->attr[i]->name.ptr, "inactive", 8) &&
						strncasecmp(stream->attr[i]->name.ptr, "framerate", 9) &&
						strncasecmp(stream->attr[i]->name.ptr, "label", 5) &&
						strncasecmp(stream->attr[i]->name.ptr, "content", 7) &&
						strncasecmp(stream->attr[i]->name.ptr, "crypto", 6)) {
					pj_pool_t *pool = session->inv_session->pool;
					session->slides_video_para.other_attr[session->slides_video_para.other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
											 stream->attr[i]);
				}
			}
			session->slides_video_para.bandw_count = 0;
			for(i = 0; i < stream->bandw_count; i++) {
				pj_pool_t *pool = session->inv_session->pool;
				session->slides_video_para.bandw[session->slides_video_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool,
								stream->bandw[i]);
			}
		}
	} else {
		ast_log(LOG_DEBUG, "has no content value\n");
	}
	ast_free(passthrough);
	return;
}

static void set_multiplexed_fec_flag(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream)
{
	ast_log(LOG_DEBUG, "Set multiplexed fec flag begin\n");

	if (ast_sip_session_judge_if_fec_stream(session, stream)) {
		ast_log(LOG_DEBUG, "Find FEC format, set multiplexed fec ture\n");
		session->fec_multiplexed_para.has_multiplexed_fec = TRUE;
	} else {
		ast_log(LOG_DEBUG, "Can't find FEC format, set multiplexed fec false\n");
		session->fec_multiplexed_para.has_multiplexed_fec = FALSE;
	}

	return;
}

/* FEC undefine attribute */
static void set_app_fec_other_attr_para(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream)
{
	int i = 0;
	session->fec_para.other_attr_count = 0;
	for (i = 0; i < stream->attr_count; i++) {
		if (strncasecmp(stream->attr[i]->name.ptr, "sendrecv", 8) &&
				strncasecmp(stream->attr[i]->name.ptr, "sendonly", 8) &&
				strncasecmp(stream->attr[i]->name.ptr, "recvonly", 8) &&
				strncasecmp(stream->attr[i]->name.ptr, "inactive", 8) &&
				strncasecmp(stream->attr[i]->name.ptr, "rtpmap", 6) &&
				strncasecmp(stream->attr[i]->name.ptr, "crypto", 6)) {
			pj_pool_t *pool = session->inv_session->pool;
			session->fec_para.other_attr[session->fec_para.other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
						 stream->attr[i]);

		}
	}
	session->fec_para.bandw_count= 0;
	for (i = 0; i < stream->bandw_count; i++) {
		pj_pool_t *pool = session->inv_session->pool;
		session->fec_para.bandw[session->fec_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool,
			stream->bandw[i]);
	}
}

/* FECC undefine attribute */
static void set_app_fecc_other_attr_para(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream)
{
	int i = 0;
	session->fecc_para.other_attr_count = 0;
	for (i = 0; i < stream->attr_count; i++) {
		if (strncasecmp(stream->attr[i]->name.ptr, "sendrecv", 8) &&
				strncasecmp(stream->attr[i]->name.ptr, "sendonly", 8) &&
				strncasecmp(stream->attr[i]->name.ptr, "recvonly", 8) &&
				strncasecmp(stream->attr[i]->name.ptr, "inactive", 8) &&
				strncasecmp(stream->attr[i]->name.ptr, "rtpmap", 6) &&
				strncasecmp(stream->attr[i]->name.ptr, "crypto", 6)) {
			pj_pool_t *pool = session->inv_session->pool;
			session->fecc_para.other_attr[session->fecc_para.other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
						 stream->attr[i]);
		}
	}
	session->fecc_para.bandw_count= 0;
	for (i = 0; i < stream->bandw_count; i++) {
		pj_pool_t *pool = session->inv_session->pool;
		session->fecc_para.bandw[session->fecc_para.bandw_count++] = pjmedia_sdp_bandw_clone(pool,
			stream->bandw[i]);
	}
}

/* FEC/FECC undefine attribute*/
static void set_app_other_rtp_attr_para(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream)
{
	ast_log(LOG_DEBUG, "set_app_other_rtp_attr_para begin\n");

	if (ast_sip_session_judge_if_fec_stream(session, stream)) {
		ast_log(LOG_DEBUG, "It's private FEC for grandstream, use FEC extra attr\n");
		set_app_fec_other_attr_para(session, stream);
    } else if (ast_sip_session_judge_if_fecc_stream(session, stream)) {
		ast_log(LOG_DEBUG, "It's FECC, use FECC extra attr\n");
		set_app_fecc_other_attr_para(session, stream);
	} else {
		ast_log(LOG_DEBUG, "Can't support other rtp application accept fec/fecc now\n");
	}
}

/* BFCP undefine attribute */
static void set_app_other_bfcp_attr_para(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream)
{
	int i = 0;
	session->bfcp_other_attr_count = 0;
	for (i = 0; i < stream->attr_count; i++) {
		if (strncasecmp(stream->attr[i]->name.ptr, "sendrecv", 8) &&
				strncasecmp(stream->attr[i]->name.ptr, "sendonly", 8) &&
				strncasecmp(stream->attr[i]->name.ptr, "recvonly", 8) &&
				strncasecmp(stream->attr[i]->name.ptr, "inactive", 8) &&
				strncasecmp(stream->attr[i]->name.ptr, "rtpmap", 6) &&
				strncasecmp(stream->attr[i]->name.ptr, "confid", 6) &&
				strncasecmp(stream->attr[i]->name.ptr, "userid", 6) &&
				strncasecmp(stream->attr[i]->name.ptr, "floorctrl", 9) &&
				strncasecmp(stream->attr[i]->name.ptr, "setup", 5) &&
				strncasecmp(stream->attr[i]->name.ptr, "connection", 10) &&
				strncasecmp(stream->attr[i]->name.ptr, "floorid", 7) &&
				strncasecmp(stream->attr[i]->name.ptr, "fingerprint", 11) &&
				strncasecmp(stream->attr[i]->name.ptr, "crypto", 6)) {
			pj_pool_t *pool = session->inv_session->pool;
			session->bfcp_other_attr[session->bfcp_other_attr_count++] = pjmedia_sdp_attr_clone(pool, 
						 stream->attr[i]);
		}
	}
	session->bfcp_bandw_count= 0;
	for (i = 0; i < stream->bandw_count; i++) {
		pj_pool_t *pool = session->inv_session->pool;
		session->bfcp_bandw[session->bfcp_bandw_count++] = pjmedia_sdp_bandw_clone(pool,
		      stream->bandw[i]);
	}
}

/* BFCP/FEC/FECC undefine attribute*/
static void set_app_other_attr_para(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream)
{
	enum ast_sip_application_type app_type = ast_sip_session_get_application_type(stream->desc.transport);
	switch (app_type) {
		case AST_SIP_APP_UDP_BFCP:
			set_app_other_bfcp_attr_para(session, stream);
			break;
		case AST_SIP_APP_TCP_BFCP:
			set_app_other_bfcp_attr_para(session, stream);
			break;
		case AST_SIP_APP_RTP:
			set_app_other_rtp_attr_para(session, stream);
			break;
		default:
			ast_log(LOG_ERROR, "Can't process other application media, ignore this");
			break;
	}
}


static int create_udp_bfcp(struct ast_sip_session *session, struct ast_sip_session_media *session_media, unsigned int ipv6)
{
	ast_log(LOG_DEBUG, "create udp bfcp begin\n");

	/* If we've already initialized bfcp, don't take any further action */
	if (session_media->bfcp) {
		ast_log(LOG_WARNING, "bfcp has already initialized\n");
		return 0;
	}

	if ((session_media->bfcp = ast_bfcp_new_with_bindaddr("UDP", NULL, NULL, ipv6 ? &address_ipv6 : &address_ipv4))) {
		/* todo: set bfcp qos use video now, need add a bfcp qos config */
		ast_bfcp_setqos(session_media->bfcp, session->endpoint->media.tos_video,
				session->endpoint->media.cos_video);
		ast_log(LOG_DEBUG, "Setting NAT on BFCP to %s\n", session->endpoint->media.rtp.symmetric ? "On" : "Off");
		ast_bfcp_setnat(session_media->bfcp, session->endpoint->media.rtp.symmetric);
	} else {
		ast_log(LOG_WARNING, "bfcp creation failed\n");
		return 1;
	}

	return 0;
}

static int create_tcp_bfcp(struct ast_sip_session *session, struct ast_sip_session_media *session_media, unsigned int ipv6)
{
	ast_log(LOG_DEBUG, "create tcp bfcp begin\n");

	/* If we've already initialized bfcp, don't take any further action */
	if (session_media->bfcp_tcp) {
		ast_log(LOG_WARNING, "bfcp tcp has already initialized\n");
		return 0;
	}

	if ((session_media->bfcp_tcp = ast_tcp_bfcp_new_with_bindaddr(ipv6 ? &address_ipv6 : &address_ipv4))) {
		session->bfcp_tcp = session_media->bfcp_tcp;
	} else {
		ast_log(LOG_WARNING, "bfcp tcp creation failed\n");
		return 1;
	}

	return 0;
}

/* begin add by bfdu for bfcp, Grandstream Networks, 2015.01.19 */
static void process_sdp_a_app_bfcp(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream)
{
	pjmedia_sdp_attr *attr;
	int confid  = 0;
	int userid  = 0;

	ast_log(LOG_DEBUG, "process_sdp_a_app_bfcp begin\n");

	if ((attr = pjmedia_sdp_media_find_attr2(stream, "confid", NULL))) {
		confid = pj_strtoul(pj_strltrim(&attr->value));
		ast_log(LOG_DEBUG, "bfcp confid:%d\n", confid);
		if (confid) {
			session->bfcp_para.confid = confid;
		}
	}
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "userid", NULL))) {
		userid = pj_strtoul(pj_strltrim(&attr->value));
		ast_log(LOG_DEBUG, "bfcp userid:%d\n", userid);
		if (userid) {
			session->bfcp_para.userid = userid;
		}
	}
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "floorctrl", NULL))) {
		if (pj_strltrim(&attr->value)) {
			ast_copy_pj_str(session->bfcp_para.floorctrl, pj_strltrim(&attr->value), sizeof(session->bfcp_para.floorctrl));
			ast_log(LOG_DEBUG, "bfcp floorctrl:%s\n", session->bfcp_para.floorctrl);
		}
	}
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "setup", NULL))) {
		if (pj_strltrim(&attr->value)) {
			ast_copy_pj_str(session->bfcp_para.setup, pj_strltrim(&attr->value), sizeof(session->bfcp_para.setup));
			ast_log(LOG_DEBUG, "bfcp setup %s\n", session->bfcp_para.setup);
		}
	}
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "connection", NULL))) {
		if (pj_strltrim(&attr->value)) {
			ast_copy_pj_str(session->bfcp_para.connection, pj_strltrim(&attr->value), sizeof(session->bfcp_para.connection));
			ast_log(LOG_DEBUG, "bfcp connection %s\n", session->bfcp_para.connection);
		}
	}
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "floorid", NULL))) {
		if (pj_strltrim(&attr->value)) {
			ast_copy_pj_str(session->bfcp_para.floorid, pj_strltrim(&attr->value), sizeof(session->bfcp_para.floorid));
			ast_log(LOG_DEBUG, "bfcp floorid %s\n", session->bfcp_para.floorid);
		}
	}
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "fingerprint", NULL))) {
		if (pj_strltrim(&attr->value)) {
			ast_copy_pj_str(session->bfcp_para.fingerprint, pj_strltrim(&attr->value), sizeof(session->bfcp_para.fingerprint));
			ast_log(LOG_DEBUG, "bfcp fingerprint %s\n", session->bfcp_para.fingerprint);
		}
	}
	char *passthrough = ast_sip_get_attr_passthrough();
	ast_log(LOG_NOTICE, "attr passthroug = %s\n", passthrough);
	if (strcasecmp(passthrough, "yes") == 0) {
		set_app_other_attr_para(session, stream);
	}
	ast_free(passthrough);
	return;
}

static int negotiate_incoming_app_bfcp_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
					 const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream)
{
	char host[NI_MAXHOST];
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);

	ast_log(LOG_DEBUG, "negotiate incoming bfcp sdp stream begin\n");

	/* If port is 0, ignore this media stream */
	if (!stream->desc.port) {
		ast_debug(3, "Media stream bfcp is already declined\n");
		return 0;
	}

	ast_copy_pj_str(host, stream->conn ? &stream->conn->addr : &sdp->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		return -1;
	}

	ast_sockaddr_set_port(addrs, stream->desc.port);
	copy_peer_direct_media_addr(session, &session->peer_direct_media_para.bfcp_direct_media_addr, addrs);

	/* Using the connection information create an appropriate RTP instance */
	if (!session_media->bfcp && create_udp_bfcp(session, session_media, ast_sockaddr_is_ipv6(addrs))) {
		return -1;
	}

	/* for bfcp, just use recieved transport */
	pj_strdup(session->inv_session->pool, &session_media->transport, &stream->desc.transport);

	process_sdp_a_app_bfcp(session, stream);

	session->bfcp_para.has_bfcp = 1;

	return 1;
}

static int negotiate_incoming_app_bfcp_tcp_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
					 const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream)
{
	char host[NI_MAXHOST];
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);

	ast_log(LOG_DEBUG, "negotiate incoming bfcp tcp sdp stream begin\n");

	/* If port is 0, ignore this media stream */
	if (!stream->desc.port) {
		ast_debug(3, "Media stream bfcp is already declined\n");
		return 0;
	}

	ast_copy_pj_str(host, stream->conn ? &stream->conn->addr : &sdp->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		return -1;
	}

	ast_sockaddr_set_port(addrs, stream->desc.port);
	copy_peer_direct_media_addr(session, &session->peer_direct_media_para.bfcp_direct_media_addr, addrs);

	/* Using the connection information create an appropriate RTP instance */
	if (!session_media->bfcp_tcp && create_tcp_bfcp(session, session_media, ast_sockaddr_is_ipv6(addrs))) {
		return -1;
	}

	/* for bfcp, just use recieved transport */
	pj_strdup(session->inv_session->pool, &session_media->transport, &stream->desc.transport);

	process_sdp_a_app_bfcp(session, stream);

	session->bfcp_para.has_tcp_bfcp = 1;

	return 1;
}


/* for fec */
static int create_fec_rtp(struct ast_sip_session *session, unsigned int ipv6)
{
	ast_log(LOG_DEBUG, "Create fec rtp begin\n");
	if (!(session->fec_media.rtp = ast_rtp_instance_new(session->endpoint->media.rtp.engine, sched, ipv6 ? &address_ipv6 : &address_ipv4, NULL))) {
		ast_log(LOG_ERROR, "Unable to create RTP instance using RTP engine '%s'\n", session->endpoint->media.rtp.engine);
		return -1;
	}

	ast_rtp_instance_set_prop(session->fec_media.rtp, AST_RTP_PROPERTY_RTCP, 1);
	ast_rtp_instance_set_prop(session->fec_media.rtp, AST_RTP_PROPERTY_NAT, session->endpoint->media.rtp.symmetric);

	/* Use default video TOS */
	if (session->endpoint->media.tos_video || session->endpoint->media.cos_video) {
		ast_rtp_instance_set_qos(session->fec_media.rtp, session->endpoint->media.tos_video,
				session->endpoint->media.cos_video, "SIP RTP FEC");
	}
	ast_rtp_instance_set_srtp_flag(session->fec_media.rtp, 0);
	/* Set FEC ssrc as main video ssrc */
	RAII_VAR(struct ast_sip_session_media *, main_video_media, NULL, ao2_cleanup);
	main_video_media = ao2_find(session->media, "video", OBJ_KEY);
	if (main_video_media) {
		ast_rtp_instance_set_ssrc(session->fec_media.rtp, main_video_media->rtp);
	}

	return 0;
}

/* for fecc */
static int create_fecc_rtp(struct ast_sip_session *session, unsigned int ipv6)
{
	ast_log(LOG_DEBUG, "Create fecc rtp begin\n");
	if (!(session->fecc_media.rtp = ast_rtp_instance_new(session->endpoint->media.rtp.engine, sched, ipv6 ? &address_ipv6 : &address_ipv4, NULL))) {
		ast_log(LOG_ERROR, "Unable to create RTP instance using RTP engine '%s'\n", session->endpoint->media.rtp.engine);
		return -1;
	}

	ast_rtp_instance_set_prop(session->fecc_media.rtp, AST_RTP_PROPERTY_RTCP, 1);
	ast_rtp_instance_set_prop(session->fecc_media.rtp, AST_RTP_PROPERTY_NAT, session->endpoint->media.rtp.symmetric);

	/* Use default video TOS */
	if (session->endpoint->media.tos_video || session->endpoint->media.cos_video) {
		ast_rtp_instance_set_qos(session->fecc_media.rtp, session->endpoint->media.tos_video,
				session->endpoint->media.cos_video, "SIP RTP FECC");
	}
	ast_rtp_instance_set_srtp_flag(session->fecc_media.rtp, 0);

	return 0;
}

static int set_fec_caps(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
		    const struct pjmedia_sdp_media *stream)
{
	RAII_VAR(struct ast_format_cap *, peer, NULL, ao2_cleanup);
	enum ast_media_type media_type = stream_to_media_type(session_media->stream_type);
	struct ast_rtp_codecs codecs = AST_RTP_CODECS_NULL_INIT;
	int fmts = 0;
	ast_log(LOG_DEBUG, "set fec caps begin\n");

	if (!(peer = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate %s capabilities\n", session_media->stream_type);
		return -1;
	}

	/* get the capabilities on the peer */
	get_codecs(session, stream, &codecs, session_media);

	get_media_formats_by_order(&codecs, stream, peer, &fmts);

	if (!ast_format_cap_count(peer)) {
		ast_rtp_codecs_payloads_destroy(&codecs);
		ast_log(LOG_NOTICE, "Can't get peer fec capabilities for application media stream\n");
		return 1;
	}

	ast_rtp_codecs_payloads_copy(&codecs, ast_rtp_instance_get_codecs(session->fec_media.rtp),
				     session->fec_media.rtp);

	{
		struct ast_str *format_buf = ast_str_alloca(64);
		ast_log(LOG_DEBUG, " peer: %s\n",	ast_format_cap_get_names(peer, &format_buf));
	}

	ast_format_cap_remove_by_type(session->fec_req_caps, media_type);
	ast_format_cap_append_from_cap(session->fec_req_caps, peer, AST_MEDIA_TYPE_APPLICATION);
	struct ast_str *fec_format_buf = ast_str_alloca(64);
	ast_log(LOG_DEBUG, "fec_req_caps: %s\n", ast_format_cap_get_names(session->fec_req_caps, &fec_format_buf));

	ast_rtp_codecs_payloads_destroy(&codecs);
	return 0;
}

static int set_fecc_caps(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
		    const struct pjmedia_sdp_media *stream)
{
	RAII_VAR(struct ast_format_cap *, peer, NULL, ao2_cleanup);
	enum ast_media_type media_type = stream_to_media_type(session_media->stream_type);
	struct ast_rtp_codecs codecs = AST_RTP_CODECS_NULL_INIT;
	int fmts = 0;
	ast_log(LOG_DEBUG, "set fecc caps begin\n");

	if (!(peer = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate %s capabilities\n", session_media->stream_type);
		return -1;
	}

	/* get the capabilities on the peer */
	get_codecs(session, stream, &codecs, session_media);

	get_media_formats_by_order(&codecs, stream, peer, &fmts);

	if (!ast_format_cap_count(peer)) {
		ast_rtp_codecs_payloads_destroy(&codecs);
		ast_log(LOG_NOTICE, "Can't get peer fecc capabilities for application media stream\n");
		return 1;
	}

	ast_rtp_codecs_payloads_copy(&codecs, ast_rtp_instance_get_codecs(session->fecc_media.rtp),
				     session->fecc_media.rtp);

	{
		struct ast_str *format_buf = ast_str_alloca(64);
		ast_log(LOG_DEBUG, " peer: %s\n",	ast_format_cap_get_names(peer, &format_buf));
	}

	ast_format_cap_remove_by_type(session->fecc_req_caps, media_type);
	ast_format_cap_append_from_cap(session->fecc_req_caps, peer, AST_MEDIA_TYPE_APPLICATION);
	struct ast_str *fecc_format_buf = ast_str_alloca(64);
	ast_log(LOG_DEBUG, "fecc_req_caps: %s\n", ast_format_cap_get_names(session->fecc_req_caps, &fecc_format_buf));

	ast_rtp_codecs_payloads_destroy(&codecs);
	return 0;
}

static int negotiate_incoming_app_fec_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
					 const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream)
{
	ast_log(LOG_DEBUG, "negotiate incoming fec sdp stream begin\n");

	char host[NI_MAXHOST];
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	enum ast_sip_session_media_encryption encryption = AST_SIP_MEDIA_ENCRYPT_NONE;
	int res = 0;

	/* If port is 0, ignore this media stream */
	if (!stream->desc.port) {
		ast_debug(3, "Media stream fec is already declined\n");
		return 0;
	}

	ast_copy_pj_str(host, stream->conn ? &stream->conn->addr : &sdp->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		ast_log(LOG_DEBUG, "address provided is invalid\n");
		return -1;
	}

	ast_sockaddr_set_port(addrs, stream->desc.port);
	copy_peer_direct_media_addr(session, &session->peer_direct_media_para.fec_direct_media_addr, addrs);

	/* Using the connection information create an appropriate RTP instance */
	if (!session->fec_media.rtp && create_fec_rtp(session, ast_sockaddr_is_ipv6(addrs))) {
		ast_log(LOG_DEBUG, "create_rtp invalid\n");
		return -1;
	}

	res = setup_extra_media_encryption(session, &session->fec_media, sdp, stream);
	if (res) {
		if (!session->endpoint->media.rtp.encryption_optimistic) {
			/* If optimistic encryption is disabled and crypto should have been enabled
			 * but was not this session must fail.
			 */
			ast_log(LOG_DEBUG, "setup_media_encryption invalid\n");
			return -1;
		}
		/* There is no encryption, sad. */
		session_media->encryption = AST_SIP_MEDIA_ENCRYPT_NONE;
 	}

	/* If we've been explicitly configured to use the received transport OR if
	 * encryption is on and crypto is present use the received transport.
	 * This is done in case of optimistic because it may come in as RTP/AVP or RTP/SAVP depending
	 * on the configuration of the remote endpoint (optimistic themselves or mandatory).
	 */
	if ((session->endpoint->media.rtp.use_received_transport) ||
		((encryption == AST_SIP_MEDIA_ENCRYPT_SDES) && !res)) {
		pj_strdup(session->inv_session->pool, &session->fec_media.transport, &stream->desc.transport);
 	}

	if (set_fec_caps(session, session_media, stream)) {
		return 0;
	}

	session->fec_para.has_fec = 1;

	return 1;
}

static int negotiate_incoming_app_fecc_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
					 const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream)
{
	ast_log(LOG_DEBUG, "negotiate incoming fecc sdp stream begin\n");

	char host[NI_MAXHOST];
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	enum ast_sip_session_media_encryption encryption = AST_SIP_MEDIA_ENCRYPT_NONE;
	int res = 0;

	/* If port is 0, ignore this media stream */
	if (!stream->desc.port) {
		ast_debug(3, "Media stream fecc is already declined\n");
		return 0;
	}

	/* Ensure incoming transport is compatible with the endpoint's configuration */
	if (!session->endpoint->media.rtp.use_received_transport) {
		encryption = check_endpoint_media_transport(session->endpoint, stream);

		if (encryption == AST_SIP_MEDIA_TRANSPORT_INVALID) {
			ast_log(LOG_DEBUG, "AST_SIP_MEDIA_TRANSPORT_INVALID\n");
			return -1;
		}
	}

	ast_copy_pj_str(host, stream->conn ? &stream->conn->addr : &sdp->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		ast_log(LOG_DEBUG, "address provided is invalid\n");
		return -1;
	}

	ast_sockaddr_set_port(addrs, stream->desc.port);
	copy_peer_direct_media_addr(session, &session->peer_direct_media_para.fecc_direct_media_addr, addrs);

	/* Using the connection information create an appropriate RTP instance */
	if (!session->fecc_media.rtp && create_fecc_rtp(session, ast_sockaddr_is_ipv6(addrs))) {
		ast_log(LOG_DEBUG, "create fecc rtp invalid\n");
		return -1;
	}

	res = setup_extra_media_encryption(session, &session->fecc_media, sdp, stream);
	if (res) {
		if (!session->endpoint->media.rtp.encryption_optimistic) {
			/* If optimistic encryption is disabled and crypto should have been enabled
			 * but was not this session must fail.
			 */
			ast_log(LOG_DEBUG, "setup_media_encryption invalid\n");
			return -1;
		}
		/* There is no encryption, sad. */
		session_media->encryption = AST_SIP_MEDIA_ENCRYPT_NONE;
 	}

	/* If we've been explicitly configured to use the received transport OR if
	 * encryption is on and crypto is present use the received transport.
	 * This is done in case of optimistic because it may come in as RTP/AVP or RTP/SAVP depending
	 * on the configuration of the remote endpoint (optimistic themselves or mandatory).
	 */
	if ((session->endpoint->media.rtp.use_received_transport) ||
		((encryption == AST_SIP_MEDIA_ENCRYPT_SDES) && !res)) {
		pj_strdup(session->inv_session->pool, &session->fecc_media.transport, &stream->desc.transport);
 	}

	if (set_fecc_caps(session, session_media, stream)) {
		return 0;
	}

	session->fecc_para.has_fecc = 1;

	return 1;
}

static int negotiate_incoming_app_rtp_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
					 const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream)
{
	int res = 0;
	ast_log(LOG_DEBUG, "negotiate incoming app rtp sdp stream begin\n");

	if (ast_sip_session_judge_if_fec_stream(session, stream)) {
		ast_log(LOG_DEBUG, "It's private FEC for grandstream, use FEC neg process\n");
		res = negotiate_incoming_app_fec_sdp_stream(session, session_media, sdp, stream);
		return res;
	}

	if (ast_sip_session_judge_if_fecc_stream(session, stream)) {
		ast_log(LOG_DEBUG, "It's FECC, use FECC neg process\n");
		res = negotiate_incoming_app_fecc_sdp_stream(session, session_media, sdp, stream);
		return res;
	}

	ast_log(LOG_DEBUG, "Can't support other rtp application accept fec/fecc now\n");

	return res;
}

static int negotiate_incoming_application_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
					 const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream)
{
	int res = 0;
	if (NULL == session || NULL == session_media || NULL == sdp || NULL == stream) {
		ast_log(LOG_ERROR, "negotiate application sdp, param is invalid\n");
		return -1;
	}

	enum ast_sip_application_type app_type = ast_sip_session_get_application_type(stream->desc.transport);
	switch (app_type) {
		case AST_SIP_APP_UDP_BFCP:
			res = negotiate_incoming_app_bfcp_sdp_stream(session, session_media, sdp, stream);
			break;
		case AST_SIP_APP_TCP_BFCP:
			res = negotiate_incoming_app_bfcp_tcp_sdp_stream(session, session_media, sdp, stream);
			break;
		case AST_SIP_APP_RTP:
			res = negotiate_incoming_app_rtp_sdp_stream(session, session_media, sdp, stream);
			break;
		default:
			ast_log(LOG_ERROR, "Can't process other application media, ignore this\n");
			break;
	}

	return res;
}

static int create_outgoing_app_bfcp_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				      struct pjmedia_sdp_session *sdp)
{
	ast_log(LOG_DEBUG, "create outgoing application bfcp sdp stream begin\n");
	pj_pool_t *pool = session->inv_session->pool_prov;
	pjmedia_sdp_media *media;
	static const pj_str_t STR_IN = { "IN", 2 };
	static const pj_str_t STR_IP4 = { "IP4", 3};
	static const pj_str_t STR_IP6 = { "IP6", 3};
	struct ast_sockaddr bfcpaddr = { {0,} };
	pjmedia_sdp_attr *attr;
	char tmp[512];
	pj_str_t stmp;
	char hostip[PJ_INET6_ADDRSTRLEN+2];
	pj_str_t str_transport_bfcp = { "UDP/BFCP", 8 };

	if (!session || !session_media || !sdp) {
		ast_log(LOG_DEBUG, "create outgoing application bfcp sdp stream para is null\n");
		return -1;
	}

	if (1 == session->bfcp_para.has_bfcp) {
		pj_cstr(&str_transport_bfcp, "UDP/BFCP");
		if (!session_media->bfcp && create_udp_bfcp(session, session_media, session->endpoint->media.rtp.ipv6)) {
			ast_log(LOG_ERROR, "create outgoing bfcp sdp stream, create udp error\n");
			return -1;
		}
	} else if (1 == session->bfcp_para.has_tcp_bfcp) {
		pj_cstr(&str_transport_bfcp, "TCP/BFCP");
		if (!session_media->bfcp_tcp && create_tcp_bfcp(session, session_media, session->endpoint->media.rtp.ipv6)) {
			ast_log(LOG_ERROR, "create outgoing bfcp sdp stream, create udp error\n");
			return -1;
		}
	} else {
		ast_log(LOG_ERROR, "Can't create outgoing bfcp sdp\n");
		return -1;
	}

	if (!(media = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_media))) ||
		!(media->conn = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_conn)))) {
		return -1;
	}
	int direct_media_enabled = !ast_sockaddr_isnull(&session_media->direct_media_addr) &&
		ast_format_cap_count(session->direct_media_cap);
	media->desc.media = pj_str(session_media->stream_type);
	media->desc.transport = str_transport_bfcp;
	char tmp_fmt[64];
	snprintf(tmp_fmt, sizeof(tmp_fmt), "%s", "*");
	pj_strdup2(pool, &media->desc.fmt[media->desc.fmt_count++], tmp_fmt);
	media->conn->net_type = STR_IN;
	media->conn->addr_type = session->endpoint->media.rtp.ipv6 ? STR_IP6 : STR_IP4;
	if (direct_media_enabled) {
		ast_log(LOG_DEBUG, "hostip:%s\n", ast_sockaddr_stringify_fmt(&session_media->direct_media_addr, AST_SOCKADDR_STR_ADDR));
		ast_copy_string(hostip, ast_sockaddr_stringify_fmt(&session_media->direct_media_addr, AST_SOCKADDR_STR_ADDR), sizeof(hostip));
	} else if (ast_strlen_zero(session->endpoint->media.address)) {
		pj_sockaddr localaddr;
		if (pj_gethostip(session->endpoint->media.rtp.ipv6 ? pj_AF_INET6() : pj_AF_INET(), &localaddr)) {
			ast_log(LOG_ERROR, "create bfcp sdp, Pj get host ip error!\n");
			return -1;
		}
		pj_sockaddr_print(&localaddr, hostip, sizeof(hostip), 2);
	} else {
		ast_copy_string(hostip, session->endpoint->media.address, sizeof(hostip));
	}
	pj_strdup2(pool, &media->conn->addr, hostip);
	if (1 == session->bfcp_para.has_bfcp) {
		ast_bfcp_get_us(session_media->bfcp, &bfcpaddr);
	} else if (1 == session->bfcp_para.has_tcp_bfcp) {
		ast_bfcp_tcp_get_us(session_media->bfcp_tcp, &bfcpaddr);
	}
	media->desc.port = direct_media_enabled ? ast_sockaddr_port(&session_media->direct_media_addr) : (pj_uint16_t) ast_sockaddr_port(&bfcpaddr);
	media->desc.port_count = 1;

	if (!ast_sockaddr_isnull(&session->direct_media_para.bfcp_direct_media_addr)) {
		reset_direct_media_addr(session, session->direct_media_para.bfcp_direct_media_addr, media);
		ast_sockaddr_setnull(&session->direct_media_para.bfcp_direct_media_addr);
	}

	/* add bfcp attr */
	if (session->bfcp_para.confid) {
		ast_log(LOG_DEBUG, "creat bfcp confid %d\n", session->bfcp_para.confid);
		snprintf(tmp, sizeof(tmp), "%d", session->bfcp_para.confid);
		attr = pjmedia_sdp_attr_create(pool, "confid", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
		memset(tmp, 0, sizeof(tmp));
	}
	if (session->bfcp_para.userid) {
		ast_log(LOG_DEBUG, "creat bfcp userid %d\n", session->bfcp_para.userid);
		snprintf(tmp, sizeof(tmp), "%d", session->bfcp_para.userid);
		attr = pjmedia_sdp_attr_create(pool, "userid", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
		memset(tmp, 0, sizeof(tmp));
	}
	if (strlen(session->bfcp_para.floorid)) {
		ast_log(LOG_DEBUG, "creat bfcp floorid %s\n", session->bfcp_para.floorid);
		snprintf(tmp, sizeof(tmp), "%s", session->bfcp_para.floorid);
		attr = pjmedia_sdp_attr_create(pool, "floorid", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
		memset(tmp, 0, sizeof(tmp));
	}
	if (strlen(session->bfcp_para.floorctrl)) {
		ast_log(LOG_DEBUG, "creat bfcp floorctrl %s\n", session->bfcp_para.floorctrl);
		snprintf(tmp, sizeof(tmp), "%s", session->bfcp_para.floorctrl);
		attr = pjmedia_sdp_attr_create(pool, "floorctrl", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
		memset(tmp, 0, sizeof(tmp));
	}
	if (strlen(session->bfcp_para.setup)) {
		ast_log(LOG_DEBUG, "creat bfcp setup %s\n", session->bfcp_para.setup);
		snprintf(tmp, sizeof(tmp), "%s", session->bfcp_para.setup);
		attr = pjmedia_sdp_attr_create(pool, "setup", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
		memset(tmp, 0, sizeof(tmp));
	}
	if (strlen(session->bfcp_para.connection)) {
		ast_log(LOG_DEBUG, "creat bfcp connection %s\n", session->bfcp_para.connection);
		snprintf(tmp, sizeof(tmp), "%s", session->bfcp_para.connection);
		attr = pjmedia_sdp_attr_create(pool, "connection", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
		memset(tmp, 0, sizeof(tmp));
	}
	if (strlen(session->bfcp_para.fingerprint)) {
		ast_log(LOG_DEBUG, "creat bfcp fingerprint %s\n", session->bfcp_para.fingerprint);
		snprintf(tmp, sizeof(tmp), "%s", session->bfcp_para.fingerprint);
		attr = pjmedia_sdp_attr_create(pool, "fingerprint", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
		memset(tmp, 0, sizeof(tmp));
	}

	/* added by jklou */
	int i = 0;
	for (i = 0; i < session->bfcp_other_attr_count; i++) {
		media->attr[media->attr_count++] = session->bfcp_other_attr[i];
	}
	for (i = 0; i < session->bfcp_bandw_count; i++) {
		media->bandw[media->bandw_count++] = session->bfcp_bandw[i];
	}
	/* end */

	/* Add the media stream to the SDP */
	sdp->media[sdp->media_count++] = media;

	ast_log(LOG_DEBUG, "creat bfcp sdp end, media count:%d\n", sdp->media_count);

	if (1 == session->slides_video_para.has_slides_video) {
		ast_log(LOG_DEBUG, "create slides video outgoing sdp\n");
		create_outgoing_slides_video_sdp_stream(session, sdp);
	}

	return 1;
}

static int create_outgoing_app_fec_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				      struct pjmedia_sdp_session *sdp)
{
	pj_pool_t *pool = session->inv_session->pool_prov;
	static const pj_str_t STR_IN = { "IN", 2 };
	static const pj_str_t STR_IP4 = { "IP4", 3};
	static const pj_str_t STR_IP6 = { "IP6", 3};
	static const pj_str_t STR_SENDRECV = { "sendrecv", 8 };
	pjmedia_sdp_media *media;
	char hostip[PJ_INET6_ADDRSTRLEN+2];
	struct ast_sockaddr addr;
	char tmp[512];
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	int index = 0;
	int min_packet_size = 0, max_packet_size = 0;
	int rtp_code;
	enum ast_media_type media_type = AST_MEDIA_TYPE_APPLICATION;
	struct ast_str *format_buf = ast_str_alloca(64);
	ast_log(LOG_DEBUG, "create outgoing fec sdp begin\n");

	int direct_media_enabled = !ast_sockaddr_isnull(&session->fec_media.direct_media_addr) &&
		ast_format_cap_count(session->direct_media_cap);

	if (!session->fec_media.rtp && create_fec_rtp(session, session->endpoint->media.rtp.ipv6)) {
		ast_log(LOG_ERROR, "Create FEC RTP error\n");
		return -1;
	}

	if (!(media = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_media))) ||
		!(media->conn = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_conn)))) {
		ast_log(LOG_ERROR, "Alloc FEC pj pool error!\n");
		return -1;
	}

	if (add_crypto_to_extra_media_stream(session, &session->fec_media, pool, media)) {
		ast_log(LOG_ERROR, "Add FEC crypto to stream error!\n");
		return -1;
	}
	/* media type must be application */
	media->desc.media = pj_str("application");
	if (pj_strlen(&session->fec_media.transport)) {
		/* If a transport has already been specified use it */
		media->desc.transport = session->fec_media.transport;
	} else {
		media->desc.transport = pj_str(ast_sdp_get_rtp_profile(
			/* Optimistic encryption places crypto in the normal RTP/AVP profile */
			!session->endpoint->media.rtp.encryption_optimistic &&
				(session->fec_media.encryption == AST_SIP_MEDIA_ENCRYPT_SDES),
			session->fec_media.rtp, session->endpoint->media.rtp.use_avpf,
			session->endpoint->media.rtp.force_avp));
	}

	/* Add connection level details */
	if (direct_media_enabled) {
		ast_log(LOG_DEBUG, "hostip:%s\n", ast_sockaddr_stringify_fmt(&session_media->direct_media_addr, AST_SOCKADDR_STR_ADDR));
		ast_copy_string(hostip, ast_sockaddr_stringify_fmt(&session->fec_media.direct_media_addr, AST_SOCKADDR_STR_ADDR), sizeof(hostip));
	} else if (ast_strlen_zero(session->endpoint->media.address)) {
		pj_sockaddr localaddr;

		if (pj_gethostip(session->endpoint->media.rtp.ipv6 ? pj_AF_INET6() : pj_AF_INET(), &localaddr)) {
			ast_log(LOG_ERROR, "Pj get slides video host ip error!\n");
			return -1;
		}
		pj_sockaddr_print(&localaddr, hostip, sizeof(hostip), 2);
	} else {
		ast_copy_string(hostip, session->endpoint->media.address, sizeof(hostip));
	}

	media->conn->net_type = STR_IN;
	media->conn->addr_type = session->endpoint->media.rtp.ipv6 ? STR_IP6 : STR_IP4;
	pj_strdup2(pool, &media->conn->addr, hostip);
	ast_rtp_instance_get_local_address(session->fec_media.rtp, &addr);
	media->desc.port = direct_media_enabled ? ast_sockaddr_port(&session->fec_media.direct_media_addr) : (pj_uint16_t) ast_sockaddr_port(&addr);
	media->desc.port_count = 1;

	if (!ast_sockaddr_isnull(&session->direct_media_para.fec_direct_media_addr)) {
		reset_direct_media_addr(session, session->direct_media_para.fec_direct_media_addr, media);
		ast_sockaddr_setnull(&session->direct_media_para.fec_direct_media_addr);
	}

	ast_log(LOG_CODEC_NEG, "%s final fec request codecs: (type:%s) %s\n",
		session->is_outgoing ? "[callee]" : "[caller]",
		"application",
		ast_format_cap_get_names(session->fec_req_caps, &format_buf));

	for (index = 0; index < ast_format_cap_count(session->fec_req_caps); ++index) {
		struct ast_format *format = ast_format_cap_get_format(session->fec_req_caps, index);

		if (ast_format_get_type(format) != media_type) {
			ao2_ref(format, -1);
			continue;
		}

		if ((rtp_code = ast_rtp_codecs_payload_code(ast_rtp_instance_get_codecs(session->fec_media.rtp), 1, format, 0)) == -1) {
			ast_log(LOG_WARNING,"Unable to get rtp codec payload code for %s\n", ast_format_get_name(format));
			ao2_ref(format, -1);
			continue;
		}

		if (!(attr = generate_rtpmap_attr(media, pool, rtp_code, 1, format, 0))) {
			ao2_ref(format, -1);
			continue;
		}
		media->attr[media->attr_count++] = attr;

		if ((attr = generate_fmtp_attr(pool, format, rtp_code))) {
			media->attr[media->attr_count++] = attr;
		}

		if (ast_format_get_maximum_ms(format) &&
			((ast_format_get_maximum_ms(format) < max_packet_size) || !max_packet_size)) {
			max_packet_size = ast_format_get_maximum_ms(format);
		}
		ao2_ref(format, -1);
	}

	/* If no formats were actually added to the media stream don't add it to the SDP */
	if (!media->desc.fmt_count) {
		return 1;
	}

	/* If ptime is set add it as an attribute */
	min_packet_size = ast_rtp_codecs_get_framing(ast_rtp_instance_get_codecs(session->fec_media.rtp));
	if (!min_packet_size) {
		min_packet_size = ast_format_cap_get_framing(session->fec_req_caps);
	}
	if (min_packet_size) {
		snprintf(tmp, sizeof(tmp), "%d", min_packet_size);
		attr = pjmedia_sdp_attr_create(pool, "ptime", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	if (max_packet_size) {
		snprintf(tmp, sizeof(tmp), "%d", max_packet_size);
		attr = pjmedia_sdp_attr_create(pool, "maxptime", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	attr = PJ_POOL_ZALLOC_T(pool, pjmedia_sdp_attr);
	attr->name = STR_SENDRECV;
	media->attr[media->attr_count++] = attr;

	/* added by jklou */
	int i = 0;
	for (i = 0; i < session->fec_para.other_attr_count; i++) {
		media->attr[media->attr_count++] = session->fec_para.other_attr[i];
	}
	for (i = 0; i < session->fec_para.bandw_count; i++) {
		media->bandw[media->bandw_count++] = session->fec_para.bandw[i];
	}
	/* end */

	/* Add the media stream to the SDP */
	sdp->media[sdp->media_count++] = media;
	ast_log(LOG_DEBUG, "create outgoing fec sdp end\n");

	return 1;
}

static int create_outgoing_app_fecc_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				      struct pjmedia_sdp_session *sdp)
{
	pj_pool_t *pool = session->inv_session->pool_prov;
	static const pj_str_t STR_IN = { "IN", 2 };
	static const pj_str_t STR_IP4 = { "IP4", 3};
	static const pj_str_t STR_IP6 = { "IP6", 3};
	static const pj_str_t STR_SENDRECV = { "sendrecv", 8 };
	pjmedia_sdp_media *media;
	char hostip[PJ_INET6_ADDRSTRLEN+2];
	struct ast_sockaddr addr;
	char tmp[512];
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	int index = 0;
	int min_packet_size = 0, max_packet_size = 0;
	int rtp_code;
	enum ast_media_type media_type = AST_MEDIA_TYPE_APPLICATION;
	struct ast_str *format_buf = ast_str_alloca(64);
	ast_log(LOG_DEBUG, "create outgoing fecc sdp begin\n");

	int direct_media_enabled = !ast_sockaddr_isnull(&session->fecc_media.direct_media_addr) &&
		ast_format_cap_count(session->direct_media_cap);

	if (!session->fecc_media.rtp && create_fecc_rtp(session, session->endpoint->media.rtp.ipv6)) {
		ast_log(LOG_ERROR, "Create FECC RTP error\n");
		return -1;
	}

	if (!(media = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_media))) ||
		!(media->conn = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_conn)))) {
		ast_log(LOG_ERROR, "Alloc FECC pj pool error!\n");
		return -1;
	}

	if (add_crypto_to_extra_media_stream(session, &session->fecc_media, pool, media)) {
		ast_log(LOG_ERROR, "Add FECC crypto to stream error!\n");
		return -1;
	}
	/* media type must be application */
	media->desc.media = pj_str("application");
	if (pj_strlen(&session->fecc_media.transport)) {
		/* If a transport has already been specified use it */
		media->desc.transport = session->fecc_media.transport;
	} else {
		media->desc.transport = pj_str(ast_sdp_get_rtp_profile(
			/* Optimistic encryption places crypto in the normal RTP/AVP profile */
			!session->endpoint->media.rtp.encryption_optimistic &&
				(session->fecc_media.encryption == AST_SIP_MEDIA_ENCRYPT_SDES),
			session->fecc_media.rtp, session->endpoint->media.rtp.use_avpf,
			session->endpoint->media.rtp.force_avp));
	}

	/* Add connection level details */
	if (direct_media_enabled) {
		ast_log(LOG_DEBUG, "hostip:%s\n", ast_sockaddr_stringify_fmt(&session_media->direct_media_addr, AST_SOCKADDR_STR_ADDR));
		ast_copy_string(hostip, ast_sockaddr_stringify_fmt(&session->fecc_media.direct_media_addr, AST_SOCKADDR_STR_ADDR), sizeof(hostip));
	} else if (ast_strlen_zero(session->endpoint->media.address)) {
		pj_sockaddr localaddr;

		if (pj_gethostip(session->endpoint->media.rtp.ipv6 ? pj_AF_INET6() : pj_AF_INET(), &localaddr)) {
			ast_log(LOG_ERROR, "Pj get slides video host ip error!\n");
			return -1;
		}
		pj_sockaddr_print(&localaddr, hostip, sizeof(hostip), 2);
	} else {
		ast_copy_string(hostip, session->endpoint->media.address, sizeof(hostip));
	}

	media->conn->net_type = STR_IN;
	media->conn->addr_type = session->endpoint->media.rtp.ipv6 ? STR_IP6 : STR_IP4;
	pj_strdup2(pool, &media->conn->addr, hostip);
	ast_rtp_instance_get_local_address(session->fecc_media.rtp, &addr);
	media->desc.port = direct_media_enabled ? ast_sockaddr_port(&session->fecc_media.direct_media_addr) : (pj_uint16_t) ast_sockaddr_port(&addr);
	media->desc.port_count = 1;

	if (!ast_sockaddr_isnull(&session->direct_media_para.fecc_direct_media_addr)) {
		reset_direct_media_addr(session, session->direct_media_para.fecc_direct_media_addr, media);
		ast_sockaddr_setnull(&session->direct_media_para.fecc_direct_media_addr);
	}

	ast_log(LOG_CODEC_NEG, "%s final fecc request codecs: (type:%s) %s\n",
		session->is_outgoing ? "[callee]" : "[caller]",
		"application",
		ast_format_cap_get_names(session->fecc_req_caps, &format_buf));

	for (index = 0; index < ast_format_cap_count(session->fecc_req_caps); ++index) {
		struct ast_format *format = ast_format_cap_get_format(session->fecc_req_caps, index);

		if (ast_format_get_type(format) != media_type) {
			ao2_ref(format, -1);
			continue;
		}

		if ((rtp_code = ast_rtp_codecs_payload_code(ast_rtp_instance_get_codecs(session->fecc_media.rtp), 1, format, 0)) == -1) {
			ast_log(LOG_WARNING,"Unable to get rtp codec payload code for %s\n", ast_format_get_name(format));
			ao2_ref(format, -1);
			continue;
		}

		if (!(attr = generate_rtpmap_attr(media, pool, rtp_code, 1, format, 0))) {
			ao2_ref(format, -1);
			continue;
		}
		media->attr[media->attr_count++] = attr;

		if ((attr = generate_fmtp_attr(pool, format, rtp_code))) {
			media->attr[media->attr_count++] = attr;
		}

		if (ast_format_get_maximum_ms(format) &&
			((ast_format_get_maximum_ms(format) < max_packet_size) || !max_packet_size)) {
			max_packet_size = ast_format_get_maximum_ms(format);
		}
		ao2_ref(format, -1);
	}

	/* If no formats were actually added to the media stream don't add it to the SDP */
	if (!media->desc.fmt_count) {
		return 1;
	}

	/* If ptime is set add it as an attribute */
	min_packet_size = ast_rtp_codecs_get_framing(ast_rtp_instance_get_codecs(session->fecc_media.rtp));
	if (!min_packet_size) {
		min_packet_size = ast_format_cap_get_framing(session->fecc_req_caps);
	}
	if (min_packet_size) {
		snprintf(tmp, sizeof(tmp), "%d", min_packet_size);
		attr = pjmedia_sdp_attr_create(pool, "ptime", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	if (max_packet_size) {
		snprintf(tmp, sizeof(tmp), "%d", max_packet_size);
		attr = pjmedia_sdp_attr_create(pool, "maxptime", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	attr = PJ_POOL_ZALLOC_T(pool, pjmedia_sdp_attr);
	attr->name = STR_SENDRECV;
	media->attr[media->attr_count++] = attr;

	/* added by jklou */
	int i = 0;
	for (i = 0; i < session->fecc_para.other_attr_count; i++) {
		media->attr[media->attr_count++] = session->fecc_para.other_attr[i];
	}
	for (i = 0; i < session->fecc_para.bandw_count; i++) {
		media->bandw[media->bandw_count++] = session->fecc_para.bandw[i];
	}
	/* end */

	/* Add the media stream to the SDP */
	sdp->media[sdp->media_count++] = media;
	ast_log(LOG_DEBUG, "create outgoing fecc sdp end\n");

	return 1;
}

static int create_outgoing_disable_app_fec_sdp_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp)
{
	if (!session || !session_media || !sdp) {
		ast_log(LOG_DEBUG, "create outgoing disable fec sdp stream, para is null\n");
		return 0;
	}

	enum ast_media_type media_type = stream_to_media_type(session_media->stream_type);
	if (AST_MEDIA_TYPE_APPLICATION != media_type) {
		ast_log(LOG_DEBUG, "create disable fec sdp, media type is not application\n");
		return 0;
	}

	if (!session->inv_session || !session->inv_session->neg) {
		ast_log(LOG_DEBUG, "Have no inv session or neg, don't add disable fec\n");
		return 0;
	}

	ast_log(LOG_DEBUG, "Create outgoing disable fec sdp stream begin\n");
	pjmedia_sdp_session *active_local_sdp;
	pj_pool_t *pool = session->inv_session->pool_prov;
	int i = 0;

	active_local_sdp = session->inv_session->neg->active_local_sdp;
	if (!active_local_sdp) {
		ast_log(LOG_WARNING, "Have no active_local_sdp, don't add disable fec\n");
		return 0;
	}

	for (i = 0; i < active_local_sdp->media_count; ++i) {
		const pjmedia_sdp_media *im;	/* initial active media */
		im = active_local_sdp->media[i];
		if (!im) {
			continue;
		}
		/* FEC will use with video media line, need to judge media type */
		char media_type[16];
		memset(media_type, 0, sizeof(media_type));
		ast_copy_pj_str(media_type, &im->desc.media, sizeof(media_type));
		if (strncmp(media_type, "application", 11) == 0) {
			pjmedia_sdp_media *res = NULL;
			if (ast_sip_session_judge_if_fec_stream(session, im)) {
				ast_log(LOG_DEBUG, "It's private FEC for grandstream, disable FEC\n");
				res = pjmedia_sdp_media_clone_deactivate(pool, im);
				sdp->media[sdp->media_count++] = res;
				return 1;
			}
		}
	}
	return 1;
}

static int create_outgoing_disable_app_fecc_sdp_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp)
{
	if (!session || !session_media || !sdp) {
		ast_log(LOG_DEBUG, "create outgoing disable fecc sdp stream, para is null\n");
		return 0;
	}

	enum ast_media_type media_type = stream_to_media_type(session_media->stream_type);
	if (AST_MEDIA_TYPE_APPLICATION != media_type) {
		ast_log(LOG_DEBUG, "create disable fecc sdp, media type is not application\n");
		return 0;
	}

	if (!session->inv_session || !session->inv_session->neg) {
		ast_log(LOG_WARNING, "Have no inv session or neg, don't add disable fecc\n");
		return 0;
	}

	ast_log(LOG_DEBUG, "Create outgoing disable fecc sdp stream begin\n");
	pjmedia_sdp_session *active_local_sdp;
	pj_pool_t *pool = session->inv_session->pool_prov;
	int i = 0;

	active_local_sdp = session->inv_session->neg->active_local_sdp;
	if (!active_local_sdp) {
		ast_log(LOG_DEBUG, "Have no active_local_sdp, don't add disable fecc\n");
		return 0;
	}

	for (i = 0; i < active_local_sdp->media_count; ++i) {
		const pjmedia_sdp_media *im;	/* initial active media */
		im = active_local_sdp->media[i];
		pjmedia_sdp_media *res = NULL;
		if (ast_sip_session_judge_if_fecc_stream(session, im)) {
			ast_log(LOG_DEBUG, "It's FECC, disable FECC\n");
			res = pjmedia_sdp_media_clone_deactivate(pool, im);
			sdp->media[sdp->media_count++] = res;
			return 1;
		}
	}
	return 1;
}

static void create_outgoing_disable_slides_video_sdp(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp)
{
	if (!session || !session_media || !sdp) {
		ast_log(LOG_DEBUG, "create outgoing disable video sdp stream, para is null\n");
		return;
	}

	if (!session->inv_session || !session->inv_session->neg) {
		ast_log(LOG_WARNING, "Have no inv session or neg, don't add disable slides video\n");
		return;
	}

	ast_log(LOG_DEBUG, "Create outgoing disable slides video sdp stream begin\n");
	pjmedia_sdp_session *active_local_sdp;
	pj_pool_t *pool = session->inv_session->pool_prov;
	int i = 0;

	active_local_sdp = session->inv_session->neg->active_local_sdp;
	if (!active_local_sdp) {
		ast_log(LOG_DEBUG, "Have no active_local_sdp, don't add disable slides video\n");
		return;
	}

	for (i = 0; i < active_local_sdp->media_count; ++i) {
		const pjmedia_sdp_media *im;	/* initial active media */
		im = active_local_sdp->media[i];
		pjmedia_sdp_media *res = NULL;
		if (ast_sip_session_judge_if_slides_video(im)) {
			ast_log(LOG_DEBUG, "content is main, add decline video media\n");
			res = pjmedia_sdp_media_clone_deactivate(pool, im);
			sdp->media[sdp->media_count++] = res;
			return;
		}
	}
	return;
}


static int create_outgoing_disable_app_bfcp_sdp_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp)
{
	if (!session || !session_media || !sdp) {
		ast_log(LOG_DEBUG, "create outgoing disable bfcp sdp stream, para is null\n");
		return 0;
	}

	enum ast_media_type media_type = stream_to_media_type(session_media->stream_type);
	if (AST_MEDIA_TYPE_APPLICATION != media_type) {
		ast_log(LOG_DEBUG, "create disable bfcp sdp, media type is not application\n");
		return 0;
	}

	if (!session->inv_session || !session->inv_session->neg) {
		ast_log(LOG_WARNING, "Have no inv session or neg, don't add disable bfcp\n");
		return 0;
	}
	pjmedia_sdp_session *active_local_sdp;
	pj_pool_t *pool = session->inv_session->pool_prov;
	int i = 0;

	active_local_sdp = session->inv_session->neg->active_local_sdp;
	if (!active_local_sdp) {
		ast_log(LOG_DEBUG, "Have no active_local_sdp, don't add disable bfcp\n");
		return 0;
	}

	for (i = 0; i < active_local_sdp->media_count; ++i) {
		const pjmedia_sdp_media *im;	/* initial active media */
		im = active_local_sdp->media[i];
		pjmedia_sdp_media *res = NULL;
		enum ast_sip_application_type app_type = ast_sip_session_get_application_type(im->desc.transport);
		if (AST_SIP_APP_UDP_BFCP == app_type || AST_SIP_APP_TCP_BFCP == app_type) {
			ast_log(LOG_DEBUG, "It's bfcp, disable bfcp\n");
			res = pjmedia_sdp_media_clone_deactivate(pool, im);
			sdp->media[sdp->media_count++] = res;
		}
	}
	return 1;
}

static int create_outgoing_application_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				      struct pjmedia_sdp_session *sdp)
{
	/* todo: We use default sdp order as FEC-BFCP-sildes video-FECC
	 * other terminal(not GVC) may have other sdp order, use caller's order support later.
	 */
	int res = 0;
	if (!session || !session_media || !sdp) {
		ast_log(LOG_DEBUG, "create outgoing application sdp stream para is null\n");
		return -1;
	}
	if (1 == session->fec_para.has_fec) {
		res = create_outgoing_app_fec_sdp_stream(session, session_media, sdp);
	} else {
		res = create_outgoing_disable_app_fec_sdp_stream(session, session_media, sdp);
	}
	if (1 == session->bfcp_para.has_bfcp) {
		res = create_outgoing_app_bfcp_sdp_stream(session, session_media, sdp);
		if (0 == session->slides_video_para.has_slides_video) {
			create_outgoing_disable_slides_video_sdp(session, session_media, sdp);
		}
	} else if (1 == session->bfcp_para.has_tcp_bfcp) {
		res = create_outgoing_app_bfcp_sdp_stream(session, session_media, sdp);
		if (0 == session->slides_video_para.has_slides_video) {
			create_outgoing_disable_slides_video_sdp(session, session_media, sdp);
		}
	} else {
		res = create_outgoing_disable_app_bfcp_sdp_stream(session, session_media, sdp);
		if (0 == session->slides_video_para.has_slides_video) {
			create_outgoing_disable_slides_video_sdp(session, session_media, sdp);
		}
	}
	if (1 == session->fecc_para.has_fecc) {
		res = create_outgoing_app_fecc_sdp_stream(session, session_media, sdp);
	} else {
		res = create_outgoing_disable_app_fecc_sdp_stream(session, session_media, sdp);
	}

	return res;
}

static int apply_negotiated_app_bfcp_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				       const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
				       const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream)
{
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	char host[NI_MAXHOST];
	int fdno;

	if (!local_stream->desc.port || !remote_stream->desc.port) {
		ast_log(LOG_DEBUG, "apply bfcp, peer bfcp is declined\n");
		session->bfcp_para.has_bfcp = 0;
		return 1;
	}

	if (!session_media->bfcp && create_udp_bfcp(session, session_media, session->endpoint->media.rtp.ipv6)) {
		ast_log(LOG_ERROR, "apply bfcp, create udp error\n");
		return -1;
	}

	if (!remote_stream->conn && !remote->conn) {
		return 1;
	}

	ast_copy_pj_str(host, remote_stream->conn ? &remote_stream->conn->addr : &remote->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		ast_log(LOG_ERROR, "apply bfcp, provided host was actually invalid\n");
		return -1;
	}

	/* Apply connection information to the bfcp */
	ast_sockaddr_set_port(addrs, remote_stream->desc.port);
	ast_bfcp_set_peer(session_media->bfcp, addrs);
	ast_verbose("Peer BFCP is at port %s\n", ast_sockaddr_stringify(addrs));

	copy_peer_direct_media_addr(session, &session->peer_direct_media_para.bfcp_direct_media_addr, addrs);

	if ((fdno = media_type_to_fdno(stream_to_media_type(session_media->stream_type))) < 0) {
		ast_log(LOG_ERROR, "apply bfcp, get fdno error\n");
		return -1;
	}
	if (session->channel) {
		ast_channel_set_fd(session->channel, fdno, ast_bfcp_fd(session_media->bfcp));
	}

	process_sdp_a_app_bfcp(session, remote_stream);

	session->bfcp_para.has_bfcp = 1;

	return 1;
}

static int apply_negotiated_app_bfcp_tcp_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				       const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
				       const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream)
{
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	char host[NI_MAXHOST];

	if (!local_stream->desc.port || !remote_stream->desc.port) {
		ast_log(LOG_DEBUG, "apply bfcp, peer bfcp tcp is declined\n");
		session->bfcp_para.has_tcp_bfcp = 0;
		return 1;
	}

	if (!session_media->bfcp_tcp && create_tcp_bfcp(session, session_media, session->endpoint->media.rtp.ipv6)) {
		ast_log(LOG_ERROR, "apply bfcp, create tcp error\n");
		return -1;
	}

	if (!remote_stream->conn && !remote->conn) {
		return 1;
	}

	ast_copy_pj_str(host, remote_stream->conn ? &remote_stream->conn->addr : &remote->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		ast_log(LOG_ERROR, "apply bfcp, provided host was actually invalid\n");
		return -1;
	}

	/* Apply connection information to the bfcp */
	ast_sockaddr_set_port(addrs, remote_stream->desc.port);
	ast_bfcp_tcp_set_peer(session_media->bfcp_tcp, addrs);
	ast_verbose("Peer BFCP is at port %s\n", ast_sockaddr_stringify(addrs));
	struct ast_sockaddr peer_bfcp_addrs;
	ast_sockaddr_copy(&peer_bfcp_addrs, addrs);
	if (session && session->channel) {
		ast_queue_control_data(session->channel, AST_CONTROL_BFCP_SET_REMOTE_ADDRS, &peer_bfcp_addrs, sizeof(peer_bfcp_addrs));
	}

	copy_peer_direct_media_addr(session, &session->peer_direct_media_para.bfcp_direct_media_addr, addrs);

	process_sdp_a_app_bfcp(session, remote_stream);

	session->bfcp_para.has_tcp_bfcp = 1;

	return 1;
}

static int apply_negotiated_app_fec_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				       const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
				       const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream)
{
	int res = 1;
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	char host[NI_MAXHOST];
	ast_log(LOG_DEBUG, "apply negotiated fec sdp stream begin\n");

	if (!session || !session_media || !local || !local_stream || !remote || !remote_stream) {
		ast_log(LOG_DEBUG, "apply negotiated application fec sdp stream para is null\n");
		return -1;
	}

	if (!local_stream->desc.port || !remote_stream->desc.port) {
		session->fec_para.has_fec = 0;
		ast_log(LOG_DEBUG, "apply fec, peer is declined, local port:%d, remote port:%d\n", local_stream->desc.port, remote_stream->desc.port);
		return 1;
	}

	/* Create an RTP instance if need be */
	if (!session->fec_media.rtp && create_fec_rtp(session, session->endpoint->media.rtp.ipv6)) {
		ast_log(LOG_ERROR, "apply fec, create fec rtp failed\n");
		return -1;
	}

	res = setup_extra_media_encryption(session, &session->fec_media, remote, remote_stream);
	if (!session->endpoint->media.rtp.encryption_optimistic && res) {
		/* If optimistic encryption is disabled and crypto should have been enabled but was not
		 * this session must fail.
		 */
		ast_log(LOG_ERROR, "apply fec, setup encryption failed\n");
		return -1;
	}

	if (!remote_stream->conn && !remote->conn) {
		ast_log(LOG_ERROR, "apply fec, has no connection\n");
		return 1;
	}

	ast_copy_pj_str(host, remote_stream->conn ? &remote_stream->conn->addr : &remote->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		ast_log(LOG_ERROR, "apply fec, host invalid\n");
		return -1;
	}

	/* Apply connection information to the RTP instance */
	ast_sockaddr_set_port(addrs, remote_stream->desc.port);
	ast_rtp_instance_set_remote_address(session->fec_media.rtp, addrs);
	ast_verbose("Peer fec RTP is at port %s\n", ast_sockaddr_stringify(addrs));

	copy_peer_direct_media_addr(session, &session->peer_direct_media_para.fec_direct_media_addr, addrs);

	if (set_fec_caps(session, session_media, local_stream)) {
		ast_log(LOG_ERROR, "apply fec, set fec caps failed\n");
		return 1;
	}

	ast_channel_set_fd(session->channel, FD_APPLICATION_FEC, ast_rtp_instance_fd(session->fec_media.rtp, 0));
	ast_channel_set_fd(session->channel, FD_APPLICATION_FEC + 1, ast_rtp_instance_fd(session->fec_media.rtp, 1));

	/* If ICE support is enabled find all the needed attributes */
	//process_ice_attributes(session, session_media, remote, remote_stream);

	/* Ensure the RTP instance is active */
	ast_rtp_instance_activate(session->fec_media.rtp);

	if (ast_sockaddr_isnull(addrs) ||
		ast_sockaddr_is_any(addrs) ||
		pjmedia_sdp_media_find_attr2(remote_stream, "sendonly", NULL) ||
		pjmedia_sdp_media_find_attr2(remote_stream, "inactive", NULL)) {
		if (!session->fec_media.held) {
			ast_rtp_instance_stop(session->fec_media.rtp);
			session->fec_media.held = 1;
		}
	} else if (session->fec_media.held) {
		session->fec_media.held = 0;
	} else if ((pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg) == PJ_FALSE)
		&& (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED)) {
		ast_queue_control(session->channel, AST_CONTROL_UPDATE_RTP_PEER);
	}

	return 1;
}

static int apply_negotiated_app_fecc_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				       const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
				       const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream)
{
	int res = 1;
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	char host[NI_MAXHOST];
	ast_log(LOG_DEBUG, "apply negotiated fecc sdp stream begin\n");

	if (!session || !session_media || !local || !local_stream || !remote || !remote_stream) {
		ast_log(LOG_DEBUG, "apply negotiated application fecc sdp stream para is null\n");
		return -1;
	}

	if (!local_stream->desc.port || !remote_stream->desc.port) {
		session->fecc_para.has_fecc = 0;
		ast_log(LOG_DEBUG, "apply fecc, peer is declined, l_p:%d, r_p:%d\n", local_stream->desc.port, remote_stream->desc.port);
		return 1;
	}

	/* Ensure incoming transport is compatible with the endpoint's configuration */
	if (!session->endpoint->media.rtp.use_received_transport &&
		check_endpoint_media_transport(session->endpoint, remote_stream) == AST_SIP_MEDIA_TRANSPORT_INVALID) {
		ast_log(LOG_ERROR, "apply fecc, transport is not compatible with the endpoint\n");
		return -1;
	}

	/* Create an RTP instance if need be */
	if (!session->fecc_media.rtp && create_fec_rtp(session, session->endpoint->media.rtp.ipv6)) {
		ast_log(LOG_ERROR, "apply fecc, create fecc rtp failed\n");
		return -1;
	}

	res = setup_extra_media_encryption(session, &session->fecc_media, remote, remote_stream);
	if (!session->endpoint->media.rtp.encryption_optimistic && res) {
		/* If optimistic encryption is disabled and crypto should have been enabled but was not
		 * this session must fail.
		 */
		ast_log(LOG_ERROR, "apply fecc, setup encryption failed\n");
		return -1;
	}

	if (!remote_stream->conn && !remote->conn) {
		ast_log(LOG_ERROR, "apply fecc, has no connection\n");
		return 1;
	}

	ast_copy_pj_str(host, remote_stream->conn ? &remote_stream->conn->addr : &remote->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		ast_log(LOG_ERROR, "apply fecc, host invalid\n");
		return -1;
	}

	/* Apply connection information to the RTP instance */
	ast_sockaddr_set_port(addrs, remote_stream->desc.port);
	ast_rtp_instance_set_remote_address(session->fecc_media.rtp, addrs);
	ast_verbose("Peer fecc RTP is at port %s\n", ast_sockaddr_stringify(addrs));

	copy_peer_direct_media_addr(session, &session->peer_direct_media_para.fecc_direct_media_addr, addrs);

	if (set_fecc_caps(session, session_media, local_stream)) {
		ast_log(LOG_ERROR, "apply fecc, set fecc caps failed\n");
		return 1;
	}

	ast_channel_set_fd(session->channel, FD_APPLICATION_FECC, ast_rtp_instance_fd(session->fecc_media.rtp, 0));
	ast_channel_set_fd(session->channel, FD_APPLICATION_FECC + 1, ast_rtp_instance_fd(session->fecc_media.rtp, 1));

	/* If ICE support is enabled find all the needed attributes */
	//process_ice_attributes(session, session_media, remote, remote_stream);

	/* Ensure the RTP instance is active */
	ast_rtp_instance_activate(session->fecc_media.rtp);

	if (ast_sockaddr_isnull(addrs) ||
		ast_sockaddr_is_any(addrs) ||
		pjmedia_sdp_media_find_attr2(remote_stream, "sendonly", NULL) ||
		pjmedia_sdp_media_find_attr2(remote_stream, "inactive", NULL)) {
		if (!session->fecc_media.held) {
			ast_rtp_instance_stop(session->fecc_media.rtp);
			session->fecc_media.held = 1;
		}
	} else if (session->fecc_media.held) {
		session->fecc_media.held = 0;
	} else if ((pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg) == PJ_FALSE)
		&& (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED)) {
		ast_queue_control(session->channel, AST_CONTROL_UPDATE_RTP_PEER);
	}

	/* This purposely resets the encryption to the configured in case it gets added later */
	if (session->endpoint->media.rtp.use_received_transport) {
		ast_log(LOG_DEBUG, "FECC use received transport, don't reset media encryption\n");
	} else {
		session->fecc_media.encryption = session->endpoint->media.rtp.encryption;
	}

	return 1;
}

static int apply_negotiated_app_rtp_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				       const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
				       const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream)
{
	int res = 0;
	ast_log(LOG_DEBUG, "Apply incoming app rtp sdp stream begin\n");

	if (ast_sip_session_judge_if_fec_stream(session, remote_stream)) {
		ast_log(LOG_DEBUG, "It's private FEC for grandstream, use FEC apply process\n");
		res = apply_negotiated_app_fec_sdp_stream(session, session_media, local, local_stream, remote, remote_stream);
		return res;
	}

	if (ast_sip_session_judge_if_fecc_stream(session, remote_stream)) {
		ast_log(LOG_DEBUG, "It's FECC, use FECC apply process\n");
		res = apply_negotiated_app_fecc_sdp_stream(session, session_media, local, local_stream, remote, remote_stream);
		return res;
	}

	ast_log(LOG_DEBUG, "Can't support other rtp application accept fec/fecc now\n");

	return res;
}

static int apply_negotiated_application_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				       const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
				       const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream)
{
	int res = 0;
	if (!session || !session_media || !local || !local_stream || !remote || !remote_stream) {
		ast_log(LOG_DEBUG, "apply negotiated application sdp stream para is null\n");
		return -1;
	}

	enum ast_sip_application_type app_type = ast_sip_session_get_application_type(remote_stream->desc.transport);
	switch (app_type) {
		case AST_SIP_APP_UDP_BFCP:
			res = apply_negotiated_app_bfcp_sdp_stream(session, session_media, local, local_stream, remote, remote_stream);
			break;
		case AST_SIP_APP_TCP_BFCP:
			res = apply_negotiated_app_bfcp_tcp_sdp_stream(session, session_media, local, local_stream, remote, remote_stream);
			break;
		case AST_SIP_APP_RTP:
			res = apply_negotiated_app_rtp_sdp_stream(session, session_media, local, local_stream, remote, remote_stream);
			break;
		default:
			ast_log(LOG_DEBUG, "Can't apply other application sdp stream, except bfcp/fec/fecc\n");
			break;
	}

	return res;
}

/* for extra media srtp like slides video/FEC/FECC */
static int setup_extra_media_srtp(struct ast_sip_extra_media *extra_media)
{
	ast_log(LOG_DEBUG, "setup extra media srtp\n");
	if (!extra_media->srtp) {
		extra_media->srtp = ast_sdp_srtp_alloc();
		if (!extra_media->srtp) {
			ast_log(LOG_ERROR, "setup extra media srtp alloc failed\n");
			return -1;
		}
	}

	if (!extra_media->srtp->crypto) {
		extra_media->srtp->crypto = ast_sdp_crypto_alloc();
		if (!extra_media->srtp->crypto) {
			ast_log(LOG_ERROR, "setup extra media srtp crypto alloc failed\n");
			return -1;
		}
	}

	return 0;
}

static int setup_sdes_extra_media_srtp(struct ast_sip_extra_media *extra_media,
	const struct pjmedia_sdp_media *stream)
{
	int i;

	for (i = 0; i < stream->attr_count; i++) {
		pjmedia_sdp_attr *attr;
		RAII_VAR(char *, crypto_str, NULL, ast_free);

		/* check the stream for the required crypto attribute */
		attr = stream->attr[i];
		if (pj_strcmp2(&attr->name, "crypto")) {
			continue;
		}

		crypto_str = ast_strndup(attr->value.ptr, attr->value.slen);
		if (!crypto_str) {
			return -1;
		}

		if (setup_extra_media_srtp(extra_media)) {
			return -1;
		}

		if (!ast_sdp_crypto_process(extra_media->rtp, extra_media->srtp, crypto_str)) {
			/* found a valid crypto attribute */
			return 0;
		}

		ast_debug(1, "Ignoring extra media crypto offer with unsupported parameters: %s\n", crypto_str);
	}

	/* no usable crypto attributes found */
	return -1;
}

static int setup_dtls_extra_media_srtp(struct ast_sip_session *session,
	struct ast_sip_extra_media *extra_media)
{
	struct ast_rtp_engine_dtls *dtls;

	if (!session->endpoint->media.rtp.dtls_cfg.enabled || !extra_media->rtp) {
		return -1;
	}

	dtls = ast_rtp_instance_get_dtls(extra_media->rtp);

	if (!dtls) {
		return -1;
	}

	session->endpoint->media.rtp.dtls_cfg.suite = ((session->endpoint->media.rtp.srtp_tag_32) ? AST_AES_CM_128_HMAC_SHA1_32 : AST_AES_CM_128_HMAC_SHA1_80);
	if (dtls->set_configuration(extra_media->rtp, &session->endpoint->media.rtp.dtls_cfg)) {
		ast_log(LOG_ERROR, "Attempted to set an invalid DTLS-SRTP configuration on RTP instance '%p'\n",
			extra_media->rtp);
		return -1;
	}

	if (setup_extra_media_srtp(extra_media)) {
		return -1;
	}
	return 0;
}

static void apply_dtls_slides_video_attrib(struct ast_sip_extra_media *extra_media,
	pjmedia_sdp_attr *attr)
{
	struct ast_rtp_engine_dtls *dtls = ast_rtp_instance_get_dtls(extra_media->rtp);

	pj_str_t *value;

	if (!attr->value.ptr) {
		return;
	}

	value = pj_strtrim(&attr->value);

	if (!pj_strcmp2(&attr->name, "setup")) {
		if (!pj_stricmp2(value, "active")) {
			dtls->set_setup(extra_media->rtp, AST_RTP_DTLS_SETUP_ACTIVE);
		} else if (!pj_stricmp2(value, "passive")) {
			dtls->set_setup(extra_media->rtp, AST_RTP_DTLS_SETUP_PASSIVE);
		} else if (!pj_stricmp2(value, "actpass")) {
			dtls->set_setup(extra_media->rtp, AST_RTP_DTLS_SETUP_ACTPASS);
		} else if (!pj_stricmp2(value, "holdconn")) {
			dtls->set_setup(extra_media->rtp, AST_RTP_DTLS_SETUP_HOLDCONN);
		} else {
			ast_log(LOG_WARNING, "Unsupported setup attribute value '%*s'\n", (int)value->slen, value->ptr);
		}
	} else if (!pj_strcmp2(&attr->name, "connection")) {
		if (!pj_stricmp2(value, "new")) {
			dtls->reset(extra_media->rtp);
		} else if (!pj_stricmp2(value, "existing")) {
			/* Do nothing */
		} else {
			ast_log(LOG_WARNING, "Unsupported connection attribute value '%*s'\n", (int)value->slen, value->ptr);
		}
	} else if (!pj_strcmp2(&attr->name, "fingerprint")) {
		char hash_value[256], hash[32];
		char fingerprint_text[value->slen + 1];
		ast_copy_pj_str(fingerprint_text, value, sizeof(fingerprint_text));
			if (sscanf(fingerprint_text, "%31s %255s", hash, hash_value) == 2) {
			if (!strcasecmp(hash, "sha-1")) {
				dtls->set_fingerprint(extra_media->rtp, AST_RTP_DTLS_HASH_SHA1, hash_value);
			} else if (!strcasecmp(hash, "sha-256")) {
				dtls->set_fingerprint(extra_media->rtp, AST_RTP_DTLS_HASH_SHA256, hash_value);
			} else {
				ast_log(LOG_WARNING, "Unsupported fingerprint hash type '%s'\n",
				hash);
			}
		}
	}
}

static int parse_dtls_extra_media_attrib(struct ast_sip_extra_media *extra_media,
	const struct pjmedia_sdp_session *sdp,
	const struct pjmedia_sdp_media *stream)
{
	int i;

	for (i = 0; i < sdp->attr_count; i++) {
		apply_dtls_slides_video_attrib(extra_media, sdp->attr[i]);
	}

	for (i = 0; i < stream->attr_count; i++) {
		apply_dtls_slides_video_attrib(extra_media, stream->attr[i]);
	}

	ast_set_flag(extra_media->srtp, AST_SRTP_CRYPTO_OFFER_OK);

	return 0;
}

static int setup_extra_media_encryption(struct ast_sip_session *session,
	struct ast_sip_extra_media *extra_media,
	const struct pjmedia_sdp_session *sdp,
	const struct pjmedia_sdp_media *stream)
{
	unsigned int optimistic;

	if (session->endpoint->media.rtp.use_received_transport) {
		ast_log(LOG_DEBUG, "setup extra media encryption use received transport\n");
		extra_media->encryption = get_media_encryption_type(stream->desc.transport, stream, &optimistic);
	}

	/* FEC stream, should not use our config, just use received encryption type */
	if (ast_sip_session_judge_if_fec_stream(session, stream)) {
		ast_log(LOG_DEBUG, "It's FEC stream, use received encryption type\n");
		extra_media->encryption = get_media_encryption_type(stream->desc.transport, stream, &optimistic);
	}

	switch (extra_media->encryption) {
	case AST_SIP_MEDIA_ENCRYPT_SDES:
	case AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP:
		if (setup_sdes_extra_media_srtp(extra_media, stream)) {
			ast_log(LOG_ERROR, "Setup SRTP error!");
			return -1;
		}
		break;
	case AST_SIP_MEDIA_ENCRYPT_DTLS:
	case AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS:
		if (setup_dtls_extra_media_srtp(session, extra_media)) {
			ast_log(LOG_ERROR, "Setup DTLS SRTP error!");
			return -1;
		}
		if (parse_dtls_extra_media_attrib(extra_media, sdp, stream)) {
			ast_log(LOG_ERROR, "Setup DTLS attributes error!");
			return -1;
		}
		break;
	case AST_SIP_MEDIA_TRANSPORT_INVALID:
	case AST_SIP_MEDIA_ENCRYPT_NONE:
	case AST_SIP_MEDIA_AUTO_NO_ENCRYPT:
		if (extra_media->srtp) {
			ast_log(LOG_DEBUG, "Extra media have srtp, but transport is not srtp, need destory\n");
			ast_rtp_instance_destroy_srtp(extra_media->rtp);

			ast_sdp_srtp_destroy(extra_media->srtp);
			extra_media->srtp = NULL;
		}
		break;
	}

	return 0;
}

static int add_crypto_to_extra_media_stream(struct ast_sip_session *session,
	struct ast_sip_extra_media *extra_media,
	pj_pool_t *pool, pjmedia_sdp_media *media)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	enum ast_rtp_dtls_hash hash;
	const char *crypto_attribute;
	struct ast_rtp_engine_dtls *dtls;
	static const pj_str_t STR_NEW = { "new", 3 };
	static const pj_str_t STR_EXISTING = { "existing", 8 };
	static const pj_str_t STR_ACTIVE = { "active", 6 };
	static const pj_str_t STR_PASSIVE = { "passive", 7 };
	static const pj_str_t STR_ACTPASS = { "actpass", 7 };
	static const pj_str_t STR_HOLDCONN = { "holdconn", 8 };

	switch (extra_media->encryption) {
	case AST_SIP_MEDIA_ENCRYPT_NONE:
	case AST_SIP_MEDIA_TRANSPORT_INVALID:
	case AST_SIP_MEDIA_AUTO_NO_ENCRYPT:
		break;
	case AST_SIP_MEDIA_ENCRYPT_SDES:
	case AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP:
		if (!extra_media->srtp) {
			extra_media->srtp = ast_sdp_srtp_alloc();
			if (!extra_media->srtp) {
				ast_log(LOG_ERROR, "Extra media SRTP alloc error!\n");
				return -1;
			}
		}

		crypto_attribute = ast_sdp_srtp_get_attrib(extra_media->srtp,
			0 /* DTLS running? No */,
			session->endpoint->media.rtp.srtp_tag_32 /* 32 byte tag length? */);
		if (!crypto_attribute) {
			/* No crypto attribute to add, bad news */
			ast_log(LOG_ERROR, "Extra media SRTP crypto attribute error!\n");
			return -1;
		}

		attr = pjmedia_sdp_attr_create(pool, "crypto", pj_cstr(&stmp, crypto_attribute));
		media->attr[media->attr_count++] = attr;
		break;
	case AST_SIP_MEDIA_ENCRYPT_DTLS:
	case AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS:
		if (setup_dtls_extra_media_srtp(session, extra_media)) {
			ast_log(LOG_ERROR, "Extra media Setup DTLS SRTP error!\n");
			return -1;
		}

		dtls = ast_rtp_instance_get_dtls(extra_media->rtp);
		if (!dtls) {
			ast_log(LOG_ERROR, "Extra media RTP instance get DTLS error!\n");
			return -1;
		}

		switch (dtls->get_connection(extra_media->rtp)) {
		case AST_RTP_DTLS_CONNECTION_NEW:
			attr = pjmedia_sdp_attr_create(pool, "connection", &STR_NEW);
			media->attr[media->attr_count++] = attr;
			break;
		case AST_RTP_DTLS_CONNECTION_EXISTING:
			attr = pjmedia_sdp_attr_create(pool, "connection", &STR_EXISTING);
			media->attr[media->attr_count++] = attr;
			break;
		default:
			break;
		}

		switch (dtls->get_setup(extra_media->rtp)) {
		case AST_RTP_DTLS_SETUP_ACTIVE:
			attr = pjmedia_sdp_attr_create(pool, "setup", &STR_ACTIVE);
			media->attr[media->attr_count++] = attr;
			break;
		case AST_RTP_DTLS_SETUP_PASSIVE:
			attr = pjmedia_sdp_attr_create(pool, "setup", &STR_PASSIVE);
			media->attr[media->attr_count++] = attr;
			break;
		case AST_RTP_DTLS_SETUP_ACTPASS:
			attr = pjmedia_sdp_attr_create(pool, "setup", &STR_ACTPASS);
			media->attr[media->attr_count++] = attr;
			break;
		case AST_RTP_DTLS_SETUP_HOLDCONN:
			attr = pjmedia_sdp_attr_create(pool, "setup", &STR_HOLDCONN);
			media->attr[media->attr_count++] = attr;
			break;
		default:
			break;
		}

		hash = dtls->get_fingerprint_hash(extra_media->rtp);
		crypto_attribute = dtls->get_fingerprint(extra_media->rtp);
		if (crypto_attribute && (hash == AST_RTP_DTLS_HASH_SHA1 || hash == AST_RTP_DTLS_HASH_SHA256)) {
			RAII_VAR(struct ast_str *, fingerprint, ast_str_create(64), ast_free);
			if (!fingerprint) {
				ast_log(LOG_ERROR, "Extra media Fingerprint check error!\n");
				return -1;
			}

			if (hash == AST_RTP_DTLS_HASH_SHA1) {
				ast_str_set(&fingerprint, 0, "SHA-1 %s", crypto_attribute);
			} else {
				ast_str_set(&fingerprint, 0, "SHA-256 %s", crypto_attribute);
			}

			attr = pjmedia_sdp_attr_create(pool, "fingerprint", pj_cstr(&stmp, ast_str_buffer(fingerprint)));
			media->attr[media->attr_count++] = attr;
		}
		break;
	}

	return 0;
}

/* for slides video */
static int create_slides_video_rtp(struct ast_sip_session *session, unsigned int ipv6)
{
	ast_log(LOG_DEBUG, "create slides video rtp begin\n");
	if (!(session->slides_video_media.rtp = ast_rtp_instance_new(session->endpoint->media.rtp.engine, sched, ipv6 ? &address_ipv6 : &address_ipv4, NULL))) {
		ast_log(LOG_ERROR, "Unable to create RTP instance using RTP engine '%s'\n", session->endpoint->media.rtp.engine);
		return -1;
	}

	ast_rtp_instance_set_prop(session->slides_video_media.rtp, AST_RTP_PROPERTY_RTCP, 1);
	ast_rtp_instance_set_prop(session->slides_video_media.rtp, AST_RTP_PROPERTY_NAT, session->endpoint->media.rtp.symmetric);

	if (session->endpoint->media.tos_video || session->endpoint->media.cos_video) {
		ast_rtp_instance_set_qos(session->slides_video_media.rtp, session->endpoint->media.tos_video,
				session->endpoint->media.cos_video, "SIP RTP slides Video");
	}
	ast_rtp_instance_set_srtp_flag(session->slides_video_media.rtp, 0);

	return 0;
}

static int set_slides_video_caps(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
		    const struct pjmedia_sdp_media *stream)
{
	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);
	RAII_VAR(struct ast_format_cap *, peer, NULL, ao2_cleanup);
	RAII_VAR(struct ast_format_cap *, joint, NULL, ao2_cleanup);
	enum ast_media_type media_type = stream_to_media_type(session_media->stream_type);
	struct ast_rtp_codecs codecs = AST_RTP_CODECS_NULL_INIT;
	int fmts = 0;
	ast_log(LOG_DEBUG, "set_slides_video_caps begin\n");

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT)) ||
	    !(peer = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT)) ||
	    !(joint = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate %s capabilities\n", session_media->stream_type);
		return -1;
	}

	/* get the endpoint capabilities */
	ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, media_type);

	/* get the capabilities on the peer */
	get_codecs(session, stream, &codecs,  session_media);

	get_media_formats_by_order(&codecs, stream, peer, &fmts);

	ast_log(LOG_DEBUG, "aros:%s, fromdomain:%s\n", session->endpoint->aors, session->endpoint->fromdomain);
	if (1 == session->is_outgoing && session->endpoint->aors && 0 == strncmp(session->endpoint->aors, "trunk_", 6)) {
		ast_log(LOG_CODEC_NEG, "an inbound call from trunk!\n");
		ast_format_cap_get_compatible(caps, peer, joint);
	} else {
		ast_format_cap_get_compatible(peer, caps, joint);
	}

	if (!ast_format_cap_count(joint)) {
		struct ast_str *usbuf = ast_str_alloca(64);
		struct ast_str *thembuf = ast_str_alloca(64);

		ast_rtp_codecs_payloads_destroy(&codecs);
		ast_log(LOG_NOTICE, "No joint capabilities for '%s' media stream between our configuration(%s) and incoming SDP(%s)\n",
			session_media->stream_type,
			ast_format_cap_get_names(caps, &usbuf),
			ast_format_cap_get_names(peer, &thembuf));
		return 1;
	}

	ast_rtp_codecs_payloads_copy(&codecs, ast_rtp_instance_get_codecs(session->slides_video_media.rtp),
				     session->slides_video_media.rtp);

	{
		struct ast_str *format_buf = ast_str_alloca(64);
		ast_log(LOG_DEBUG, " peer: %s\n",	ast_format_cap_get_names(peer, &format_buf));
		ast_log(LOG_DEBUG, " caps: %s\n",	ast_format_cap_get_names(caps, &format_buf));
		ast_log(LOG_DEBUG, " set session->slides_video_req_caps of caller: %s\n",
			ast_format_cap_get_names(joint, &format_buf));
	}
	ast_format_cap_remove_by_type(session->slides_video_req_caps, media_type);
	ast_format_cap_append_from_cap(session->slides_video_req_caps, joint, AST_MEDIA_TYPE_VIDEO);
	struct ast_str *slides_format_buf = ast_str_alloca(64);
	ast_log(LOG_DEBUG, "slides_video_req_caps: %s %d\n", ast_format_cap_get_names(session->slides_video_req_caps, &slides_format_buf), ast_format_cap_count(session->slides_video_req_caps));

	ast_rtp_codecs_payloads_destroy(&codecs);
	return 0;
}

static int negotiate_incoming_slides_video_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
					 const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream)
{
	ast_log(LOG_DEBUG, "negotiate_incoming_slides_video_sdp_stream begin\n");
	if (NULL == session || NULL == session_media || NULL == sdp || NULL == stream) {
		ast_log(LOG_ERROR, "negotiate slides video sdp, param is invalid\n");
		return -1;
	}

	/* If port is 0, ignore this media stream */
	if (!stream->desc.port) {
		ast_debug(3, "Media stream slides video is already declined\n");
		return 0;
	}

	char host[NI_MAXHOST];
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	enum ast_media_type media_type = AST_MEDIA_TYPE_VIDEO;
	enum ast_sip_session_media_encryption encryption = AST_SIP_MEDIA_ENCRYPT_NONE;
	int res = 0;

	/* If no type formats have been configured reject this stream */
	if (!ast_format_cap_has_type(session->endpoint->media.codecs, media_type)) {
		ast_debug(3, "Endpoint has no codecs for media type '%s', declining stream\n", session_media->stream_type);
		return 0;
	}

	/* Ensure incoming transport is compatible with the endpoint's configuration */
	if (!session->endpoint->media.rtp.use_received_transport) {
		encryption = check_endpoint_media_transport(session->endpoint, stream);

		if (encryption == AST_SIP_MEDIA_TRANSPORT_INVALID) {
			ast_log(LOG_DEBUG, "AST_SIP_MEDIA_TRANSPORT_INVALID\n");
			return -1;
		}
	}

	ast_copy_pj_str(host, stream->conn ? &stream->conn->addr : &sdp->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		ast_log(LOG_DEBUG, "address provided is invalid\n");
		return -1;
	}

	ast_sockaddr_set_port(addrs, stream->desc.port);
	copy_peer_direct_media_addr(session, &session->peer_direct_media_para.slides_video_direct_media_addr, addrs);

	/* Using the connection information create an appropriate RTP instance */
	if (!session->slides_video_media.rtp && create_slides_video_rtp(session, ast_sockaddr_is_ipv6(addrs))) {
		ast_log(LOG_DEBUG, "create_rtp invalid\n");
		return -1;
	}

	res = setup_extra_media_encryption(session, &session->slides_video_media, sdp, stream);
	if (res) {
		if (!session->endpoint->media.rtp.encryption_optimistic) {
			/* If optimistic encryption is disabled and crypto should have been enabled
			 * but was not this session must fail.
			 */
			ast_log(LOG_DEBUG, "setup_media_encryption invalid\n");
			return -1;
		}
		/* There is no encryption, sad. */
		session_media->encryption = AST_SIP_MEDIA_ENCRYPT_NONE;
 	}

	/* If we've been explicitly configured to use the received transport OR if
	 * encryption is on and crypto is present use the received transport.
	 * This is done in case of optimistic because it may come in as RTP/AVP or RTP/SAVP depending
	 * on the configuration of the remote endpoint (optimistic themselves or mandatory).
	 */
	if ((session->endpoint->media.rtp.use_received_transport) ||
		((encryption == AST_SIP_MEDIA_ENCRYPT_SDES) && !res)) {
		pj_strdup(session->inv_session->pool, &session->slides_video_media.transport, &stream->desc.transport);
 	}

	if (set_slides_video_caps(session, session_media, stream)) {
		return 0;
	}

	/* Add by bfdu for modify video extra attr, 2015.12.16 */
	set_video_extra_attr_para(session, stream);

	session->slides_video_para.has_slides_video = 1;

	return 1;
}

/*! \brief Function which creates an outgoing stream */
static int create_outgoing_slides_video_sdp_stream(struct ast_sip_session *session, struct pjmedia_sdp_session *sdp)
{
	pj_pool_t *pool = session->inv_session->pool_prov;
	static const pj_str_t STR_IN = { "IN", 2 };
	static const pj_str_t STR_IP4 = { "IP4", 3};
	static const pj_str_t STR_IP6 = { "IP6", 3};
	static const pj_str_t STR_SENDRECV = { "sendrecv", 8 };
	pjmedia_sdp_media *media;
	char hostip[PJ_INET6_ADDRSTRLEN+2];
	struct ast_sockaddr addr;
	char tmp[512];
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	int index = 0;
	int min_packet_size = 0, max_packet_size = 0;
	int rtp_code;
	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);
	enum ast_media_type media_type = AST_MEDIA_TYPE_VIDEO;
	int use_override_prefs = ast_format_cap_count(session->slides_video_req_caps);
	struct ast_str *format_buf = ast_str_alloca(64);
	ast_log(LOG_DEBUG, "create outgoing slides video sdp begin\n");

	int direct_media_enabled = !ast_sockaddr_isnull(&session->slides_video_media.direct_media_addr) &&
		ast_format_cap_count(session->direct_media_cap);

	if ((use_override_prefs && !ast_format_cap_has_type(session->slides_video_req_caps, media_type)) ||
	    (!use_override_prefs && !ast_format_cap_has_type(session->endpoint->media.codecs, media_type))) {
		/* If no type formats are configured don't add a stream */
		ast_log(LOG_DEBUG, "No type formats are configured don't add a stream\n");
		return 0;
	} else if (!session->slides_video_media.rtp && create_slides_video_rtp(session, session->endpoint->media.rtp.ipv6)) {
		ast_log(LOG_ERROR, "Create slides video RTP error\n");
		return -1;
	}

	if (!(media = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_media))) ||
		!(media->conn = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_conn)))) {
		ast_log(LOG_ERROR, "Alloc slides video pj pool error!\n");
		return -1;
	}

	if (add_crypto_to_extra_media_stream(session, &session->slides_video_media, pool, media)) {
		ast_log(LOG_ERROR, "Add slides video crypto to stream error!\n");
		return -1;
	}
	/* media type must be video */
	media->desc.media = pj_str("video");
	if (pj_strlen(&session->slides_video_media.transport)) {
		/* If a transport has already been specified use it */
		media->desc.transport = session->slides_video_media.transport;
	} else {
		media->desc.transport = pj_str(ast_sdp_get_rtp_profile(
			/* Optimistic encryption places crypto in the normal RTP/AVP profile */
			!session->endpoint->media.rtp.encryption_optimistic &&
				(session->slides_video_media.encryption == AST_SIP_MEDIA_ENCRYPT_SDES),
			session->slides_video_media.rtp, session->endpoint->media.rtp.use_avpf,
			session->endpoint->media.rtp.force_avp));
	}

	/* Add connection level details */
	if (direct_media_enabled) {
		ast_log(LOG_DEBUG, "hostip:%s\n", ast_sockaddr_stringify_fmt(&session->slides_video_media.direct_media_addr, AST_SOCKADDR_STR_ADDR));
		ast_copy_string(hostip, ast_sockaddr_stringify_fmt(&session->slides_video_media.direct_media_addr, AST_SOCKADDR_STR_ADDR), sizeof(hostip));
	} else if (ast_strlen_zero(session->endpoint->media.address)) {
		pj_sockaddr localaddr;

		if (pj_gethostip(session->endpoint->media.rtp.ipv6 ? pj_AF_INET6() : pj_AF_INET(), &localaddr)) {
			ast_log(LOG_ERROR, "Pj get slides video host ip error!\n");
			return -1;
		}
		pj_sockaddr_print(&localaddr, hostip, sizeof(hostip), 2);
	} else {
		ast_copy_string(hostip, session->endpoint->media.address, sizeof(hostip));
	}

	media->conn->net_type = STR_IN;
	media->conn->addr_type = session->endpoint->media.rtp.ipv6 ? STR_IP6 : STR_IP4;
	pj_strdup2(pool, &media->conn->addr, hostip);
	ast_rtp_instance_get_local_address(session->slides_video_media.rtp, &addr);
	media->desc.port = direct_media_enabled ? ast_sockaddr_port(&session->slides_video_media.direct_media_addr) : (pj_uint16_t) ast_sockaddr_port(&addr);
	media->desc.port_count = 1;

	if (!ast_sockaddr_isnull(&session->direct_media_para.slides_video_direct_media_addr)) {
		reset_direct_media_addr(session, session->direct_media_para.slides_video_direct_media_addr, media);
		ast_sockaddr_setnull(&session->direct_media_para.slides_video_direct_media_addr);
	}

	/* Add ICE attributes and candidates */
	//add_ice_to_stream(session, session_media, pool, media);

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate slides video capabilities\n");
		return -1;
	}

	ast_log(LOG_CODEC_NEG, "%s session->slides_video_req_caps: %s\n",
		session->is_outgoing ? "[callee]" : "[caller]",
		ast_format_cap_get_names(session->slides_video_req_caps, &format_buf));
	ast_log(LOG_CODEC_NEG, "%s endpoint codecs: %s\n",
		session->is_outgoing ? "[callee]" : "[caller]",
		ast_format_cap_get_names(session->endpoint->media.codecs, &format_buf));

	if (direct_media_enabled) {
		if ((1 == session->is_outgoing && session->endpoint->aors && 0 == strncmp(session->endpoint->aors, "trunk_", 6))) {
			ast_format_cap_get_compatible(session->endpoint->media.codecs, session->slides_video_req_caps, caps);
		} else {
			ast_log(LOG_CODEC_NEG, "direct media enabled, use direct_media_cap\n");
			ast_format_cap_get_compatible(session->slides_video_req_caps, session->endpoint->media.codecs, caps);
		}
	} else if (!ast_format_cap_count(session->slides_video_req_caps) ||
		!ast_format_cap_iscompatible(session->slides_video_req_caps, session->endpoint->media.codecs)) {
		/* There is no the same codec in session->req_caps and endpoint codecs. only need to add audio codecs because asterisk can translate it. */		
		ast_log(LOG_CODEC_NEG, " There is no same (type:%s) codec in slides video req_caps and endpoint codecs.\n",
			media_type == AST_MEDIA_TYPE_AUDIO ? "audio" : "video");
	} else {
		if ((1 == session->is_outgoing && session->endpoint->aors && 0 == strncmp(session->endpoint->aors, "trunk_", 6))) {
			ast_format_cap_get_compatible(session->endpoint->media.codecs, session->slides_video_req_caps, caps);
		} else {
			ast_format_cap_get_compatible(session->slides_video_req_caps, session->endpoint->media.codecs, caps);
		}
		format_cap_only_type(caps, media_type);

		ast_log(LOG_CODEC_NEG, "%s The same (type:%s) codecs in slides video req_caps and endpoint codes: %s\n",			
			session->is_outgoing ? "[callee]" : "[caller]",
			media_type == AST_MEDIA_TYPE_AUDIO ? "audio" : "video",
			ast_format_cap_get_names(caps, &format_buf));
	}

	ast_log(LOG_CODEC_NEG, "%s final slides video request codecs: (type:%s) %s\n",
		session->is_outgoing ? "[callee]" : "[caller]",
		media_type == AST_MEDIA_TYPE_AUDIO ? "audio" : "video",
		ast_format_cap_get_names(caps, &format_buf));

	for (index = 0; index < ast_format_cap_count(caps); ++index) {
		struct ast_format *format = ast_format_cap_get_format(caps, index);

		if (ast_format_get_type(format) != media_type) {
			ao2_ref(format, -1);
			continue;
		}

		if ((rtp_code = ast_rtp_codecs_payload_code(ast_rtp_instance_get_codecs(session->slides_video_media.rtp), 1, format, 0)) == -1) {
			ast_log(LOG_WARNING,"Unable to get rtp codec payload code for %s\n", ast_format_get_name(format));
			ao2_ref(format, -1);
			continue;
		}

		if (!(attr = generate_rtpmap_attr(media, pool, rtp_code, 1, format, 0))) {
			ao2_ref(format, -1);
			continue;
		}
		media->attr[media->attr_count++] = attr;

		if ((attr = generate_fmtp_attr(pool, format, rtp_code))) {
			media->attr[media->attr_count++] = attr;
		}

		if (ast_format_get_maximum_ms(format) &&
			((ast_format_get_maximum_ms(format) < max_packet_size) || !max_packet_size)) {
			max_packet_size = ast_format_get_maximum_ms(format);
		}
		ao2_ref(format, -1);
	}

	/* If no formats were actually added to the media stream don't add it to the SDP */
	if (!media->desc.fmt_count) {
		return 1;
	}

	/* If ptime is set add it as an attribute */
	min_packet_size = ast_rtp_codecs_get_framing(ast_rtp_instance_get_codecs(session->slides_video_media.rtp));
	if (!min_packet_size) {
		min_packet_size = ast_format_cap_get_framing(caps);
	}
	if (min_packet_size) {
		snprintf(tmp, sizeof(tmp), "%d", min_packet_size);
		attr = pjmedia_sdp_attr_create(pool, "ptime", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	if (max_packet_size) {
		snprintf(tmp, sizeof(tmp), "%d", max_packet_size);
		attr = pjmedia_sdp_attr_create(pool, "maxptime", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.13 */
	if (AST_MEDIA_TYPE_VIDEO == media_type) {
		create_outgoing_slides_video_extra_attr_sdp(session, pool, media);
		ast_sip_session_modify_outgoing_yealink_sdp(session, pool, media);
	}

	attr = PJ_POOL_ZALLOC_T(pool, pjmedia_sdp_attr);
	attr->name = STR_SENDRECV;
	media->attr[media->attr_count++] = attr;

	/* Add the media stream to the SDP */
	sdp->media[sdp->media_count++] = media;
	ast_log(LOG_DEBUG, "create outgoing slides video sdp end\n");

	return 1;
}

static int apply_negotiated_slides_video_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				       const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
				       const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream)
{
	int res = 1;
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	char host[NI_MAXHOST];
	ast_log(LOG_DEBUG, "apply negotiated slides video sdp stream begin\n");

	if (!session || !session_media || !local || !local_stream || !remote || !remote_stream) {
		ast_log(LOG_DEBUG, "apply negotiated application sdp stream para is null\n");
		return -1;
	}

	if (!local_stream->desc.port || !remote_stream->desc.port) {
		session->slides_video_para.has_slides_video = 0;
		ast_log(LOG_DEBUG, "apply slides video, peer is declined\n");
		return 1;
	}

	/* Ensure incoming transport is compatible with the endpoint's configuration */
	if (!session->endpoint->media.rtp.use_received_transport &&
		check_endpoint_media_transport(session->endpoint, remote_stream) == AST_SIP_MEDIA_TRANSPORT_INVALID) {
		ast_log(LOG_ERROR, "apply slides video, transport is not compatible with the endpoint\n");
		return -1;
	}

	/* Create an RTP instance if need be */
	if (!session->slides_video_media.rtp && create_slides_video_rtp(session, session->endpoint->media.rtp.ipv6)) {
		ast_log(LOG_ERROR, "apply slides video, create slides video rtp failed\n");
		return -1;
	}

	res = setup_extra_media_encryption(session, &session->slides_video_media, remote, remote_stream);
	if (!session->endpoint->media.rtp.encryption_optimistic && res) {
		/* If optimistic encryption is disabled and crypto should have been enabled but was not
		 * this session must fail.
		 */
		ast_log(LOG_ERROR, "apply slides video, setup encryption failed\n");
		return -1;
	}

	if (!remote_stream->conn && !remote->conn) {
		ast_log(LOG_ERROR, "apply slides video, has no connection\n");
		return 1;
	}

	ast_copy_pj_str(host, remote_stream->conn ? &remote_stream->conn->addr : &remote->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		ast_log(LOG_ERROR, "apply slides video, host invalid\n");
		return -1;
	}

	/* Apply connection information to the RTP instance */
	ast_sockaddr_set_port(addrs, remote_stream->desc.port);
	ast_rtp_instance_set_remote_address(session->slides_video_media.rtp, addrs);
	ast_verbose("Peer slides video RTP is at port %s\n", ast_sockaddr_stringify(addrs));

	copy_peer_direct_media_addr(session, &session->peer_direct_media_para.slides_video_direct_media_addr, addrs);

	if (set_slides_video_caps(session, session_media, local_stream)) {
		ast_log(LOG_ERROR, "apply slides video, set slides video caps failed\n");
		return 1;
	}

	/* pjsip can't change video extra attr, need use remote stream to modify */
	set_video_extra_attr_para(session, remote_stream);

	ast_channel_set_fd(session->channel, FD_SLIDES_VIDEO, ast_rtp_instance_fd(session->slides_video_media.rtp, 0));
	ast_channel_set_fd(session->channel, FD_SLIDES_VIDEO + 1, ast_rtp_instance_fd(session->slides_video_media.rtp, 1));

	/* If ICE support is enabled find all the needed attributes */
	//process_ice_attributes(session, session_media, remote, remote_stream);

	/* Ensure the RTP instance is active */
	ast_rtp_instance_activate(session->slides_video_media.rtp);

	if (ast_sockaddr_isnull(addrs) ||
		ast_sockaddr_is_any(addrs) ||
		pjmedia_sdp_media_find_attr2(remote_stream, "sendonly", NULL) ||
		pjmedia_sdp_media_find_attr2(remote_stream, "inactive", NULL)) {
		if (!session->slides_video_media.held) {
			ast_rtp_instance_stop(session->slides_video_media.rtp);
			session->slides_video_media.held = 1;
		}
	} else if (session->slides_video_media.held) {
		session->slides_video_media.held = 0;
	} else if ((pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg) == PJ_FALSE)
		&& (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED)) {
		ast_queue_control(session->channel, AST_CONTROL_UPDATE_RTP_PEER);
	}

	/* This purposely resets the encryption to the configured in case it gets added later */
	if (session->endpoint->media.rtp.use_received_transport) {
		ast_log(LOG_DEBUG, "Slides video use received transport, don't reset media encryption\n");
	} else {
		session->slides_video_media.encryption = session->endpoint->media.rtp.encryption;
	}

	return 1;
}

static void create_outgoing_main_video_extra_attr_sdp(struct ast_sip_session *session,
	pj_pool_t *pool, pjmedia_sdp_media *media)
{
	char tmp[512];
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	ast_log(LOG_DEBUG, "create outgoing main video extra attr sdp begin\n");

	unsigned int framerate = get_main_video_framerate(session);
	if (framerate) {
		ast_log(LOG_DEBUG, "create outgoing main video sdp framerate %d\n", framerate);
		snprintf(tmp, sizeof(tmp), "%d", framerate);
		attr = pjmedia_sdp_attr_create(pool, "framerate", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	char content[8];
	memset(content, 0, sizeof(content));
	memcpy(content, session->main_video_para.content, sizeof(content));
	if (strlen(content)) {
		ast_log(LOG_DEBUG, "create outgoing main video sdp content %s\n", content);
		snprintf(tmp, sizeof(tmp), "%s", content);
		attr = pjmedia_sdp_attr_create(pool, "content", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	unsigned int label = get_main_video_label(session);
	if (label) {
		ast_log(LOG_DEBUG, "create outgoing main video sdp label %d\n", label);
		snprintf(tmp, sizeof(tmp), "%d", label);
		attr = pjmedia_sdp_attr_create(pool, "label", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}
	/*added by jklou*/
	int i = 0;
	for (i = 0; i < session->main_video_para.other_attr_count; i++) {
		media->attr[media->attr_count++] = session->main_video_para.other_attr[i];
	}
	for (i = 0; i < session->main_video_para.bandw_count; i++) {
		media->bandw[media->bandw_count++] = session->main_video_para.bandw[i];
	}
}

static void create_outgoing_slides_video_extra_attr_sdp(struct ast_sip_session *session,
	pj_pool_t *pool, pjmedia_sdp_media *media)
{
	char tmp[512];
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	ast_log(LOG_DEBUG, "create outgoing slides video extra attr sdp begin\n");

	unsigned int framerate = get_slides_video_framerate(session);
	if (framerate) {
		ast_log(LOG_DEBUG, "create outgoing slides video sdp framerate %d\n", framerate);
		snprintf(tmp, sizeof(tmp), "%d", framerate);
		attr = pjmedia_sdp_attr_create(pool, "framerate", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	char content[8];
	memset(content, 0, sizeof(content));
	memcpy(content, session->slides_video_para.content, sizeof(content));
	if (strlen(content)) {
		ast_log(LOG_DEBUG, "create outgoing slides video sdp content %s\n", content);
		snprintf(tmp, sizeof(tmp), "%s", content);
		attr = pjmedia_sdp_attr_create(pool, "content", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	unsigned int label = get_slides_video_label(session);
	if (label) {
		ast_log(LOG_DEBUG, "create outgoing slides video sdp label %d\n", label);
		snprintf(tmp, sizeof(tmp), "%d", label);
		attr = pjmedia_sdp_attr_create(pool, "label", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}
	/*added by jklou*/
	int i = 0;
	for (i = 0; i < session->slides_video_para.other_attr_count; i++) {
		media->attr[media->attr_count++] = session->slides_video_para.other_attr[i];
	}
	for (i = 0; i < session->slides_video_para.bandw_count; i++) {
		media->bandw[media->bandw_count++] = session->slides_video_para.bandw[i];
	}
}

static void create_outgoing_disable_main_video_sdp(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp)
{
	if (!session || !session_media || !sdp) {
		ast_log(LOG_DEBUG, "create outgoing disable video sdp stream, para is null\n");
		return;
	}

	enum ast_media_type media_type = stream_to_media_type(session_media->stream_type);
	if (AST_MEDIA_TYPE_VIDEO != media_type) {
		ast_log(LOG_DEBUG, "create disable sdp, media type is not video\n");
		return;
	}

	if (!session->inv_session || !session->inv_session->neg) {
		ast_log(LOG_WARNING, "Have no inv session or neg, don't add disable main video\n");
		return;
	}
	pjmedia_sdp_session *active_local_sdp;
	pj_pool_t *pool = session->inv_session->pool_prov;
	int i = 0;

	active_local_sdp = session->inv_session->neg->active_local_sdp;
	if (!active_local_sdp) {
		ast_log(LOG_WARNING, "Have no active_local_sdp, don't add disable main video\n");
		return;
	}

	for (i = 0; i < active_local_sdp->media_count; ++i) {
		const pjmedia_sdp_media *im;	/* initial active media */
		im = active_local_sdp->media[i];
		char media_type[16];
		memset(media_type, 0, sizeof(media_type));
		ast_copy_pj_str(media_type, &im->desc.media, sizeof(media_type));
		int direct_media_enabled = !ast_sockaddr_isnull(&session_media->direct_media_addr) &&
		ast_format_cap_count(session->direct_media_cap);
		if (strncmp(media_type, "video", 5) == 0) {
			pjmedia_sdp_attr *attr;
			char content[8];
			memset(content, 0, sizeof(content));
			pjmedia_sdp_media *res = NULL;
			if ((attr = pjmedia_sdp_media_find_attr2(im, "content", NULL))) {
				ast_copy_pj_str(content, pj_strltrim(&attr->value), sizeof(content));
				if (0 != strlen(content)) {
					if (!strncmp(content, "main", 4)) {
						ast_log(LOG_DEBUG, "content is main, add decline video media\n");
						res = pjmedia_sdp_media_clone_deactivate(pool, im);
						sdp->media[sdp->media_count++] = res;
						return;
					} else {
						ast_log(LOG_DEBUG, "content is not main, don't add decline video media\n");
					}
				}
			} else if (1 == direct_media_enabled) {
				ast_log(LOG_DEBUG, "direct_media_enabled, add decline video media\n");
				res = pjmedia_sdp_media_clone_deactivate(pool, im);
				sdp->media[sdp->media_count++] = res;
				return;
			} else {
				ast_log(LOG_DEBUG, "is not main content or video reinvite, don't add disable video media\n");
			}
		}
	}
	return;
}
#endif

/*! \brief Unloads the sdp RTP/AVP module from Asterisk */
static int unload_module(void)
{
	ast_sip_session_unregister_supplement(&video_info_supplement);
	ast_sip_session_unregister_sdp_handler(&video_sdp_handler, STR_VIDEO);
	ast_sip_session_unregister_sdp_handler(&audio_sdp_handler, STR_AUDIO);
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	ast_sip_session_unregister_sdp_handler(&application_sdp_handler, STR_APPLICATION);
#endif

	if (sched) {
		ast_sched_context_destroy(sched);
	}

	return 0;
}

/*!
 * \brief Load the module
 *
 * Module loading including tests for configuration or dependencies.
 * This function can return AST_MODULE_LOAD_FAILURE, AST_MODULE_LOAD_DECLINE,
 * or AST_MODULE_LOAD_SUCCESS. If a dependency or environment variable fails
 * tests return AST_MODULE_LOAD_FAILURE. If the module can not load the
 * configuration file or other non-critical problem return
 * AST_MODULE_LOAD_DECLINE. On success return AST_MODULE_LOAD_SUCCESS.
 */
static int load_module(void)
{
	CHECK_PJSIP_SESSION_MODULE_LOADED();

	ast_sockaddr_parse(&address_ipv4, "0.0.0.0", 0);
	ast_sockaddr_parse(&address_ipv6, "::", 0);

	if (!(sched = ast_sched_context_create())) {
		ast_log(LOG_ERROR, "Unable to create scheduler context.\n");
		goto end;
	}

	if (ast_sched_start_thread(sched)) {
		ast_log(LOG_ERROR, "Unable to create scheduler context thread.\n");
		goto end;
	}

	if (ast_sip_session_register_sdp_handler(&audio_sdp_handler, STR_AUDIO)) {
		ast_log(LOG_ERROR, "Unable to register SDP handler for %s stream type\n", STR_AUDIO);
		goto end;
	}

	if (ast_sip_session_register_sdp_handler(&video_sdp_handler, STR_VIDEO)) {
		ast_log(LOG_ERROR, "Unable to register SDP handler for %s stream type\n", STR_VIDEO);
		goto end;
	}
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	if (ast_sip_session_register_sdp_handler(&application_sdp_handler, STR_APPLICATION)) {
		ast_log(LOG_ERROR, "Unable to register SDP handler for %s stream type\n", STR_APPLICATION);
		goto end;
	}
#endif

	ast_sip_session_register_supplement(&video_info_supplement);

	return AST_MODULE_LOAD_SUCCESS;
end:
	unload_module();

	return AST_MODULE_LOAD_FAILURE;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP SDP RTP/AVP stream handler",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.load_pri = AST_MODPRI_CHANNEL_DRIVER,
	);
